Package: lgspline
Type: Package
Title: Lagrangian Multiplier Smoothing Splines for Smooth Function
        Estimation
Version: 1.1.0
Authors@R: 
    person("Matthew", "Davis", email = "matthewlouisdavis@gmail.com", 
           role = c("aut", "cre"),
           comment = c(ORCID = "0000-0001-9714-1018"))
Description: Implements Lagrangian multiplier smoothing splines for flexible
    nonparametric regression and function estimation. Provides tools for fitting,
    prediction, and inference using a constrained optimization approach to
    enforce smoothness. Supports generalized linear models, Weibull accelerated
    failure time (AFT) models, Cox proportional hazards models, 
    quadratic programming constraints, and customizable working-correlation 
    structures, with options for parallel fitting. The core spline construction 
    builds on Ezhov et al. (2018) <doi:10.1515/jag-2017-0029>. 
    Quadratic-programming and SQP details follow Goldfarb & Idnani (1983) 
    <doi:10.1007/BF02591962> and Nocedal & Wright (2006) 
    <doi:10.1007/978-0-387-40065-5>. For smoothing spline and penalized spline 
    background, see Wahba (1990) <doi:10.1137/1.9781611970128> and Wood (2017)
    <doi:10.1201/9781315370279>. For variance-component and correlation-parameter
    estimation, see Searle et al. (2006) <ISBN:978-0470009598>. The default
    multivariate partitioning step uses k-means clustering as in MacQueen
    (1967).
License: MIT + file LICENSE
Language: en-US
Depends: R (>= 3.5.0)
Imports: Rcpp (>= 1.0.7), RcppArmadillo, FNN, RColorBrewer, plotly,
        quadprog, methods, stats
LinkingTo: Rcpp, RcppArmadillo
Suggests: testthat (>= 3.0.0), spelling, knitr, rmarkdown, parallel,
        survival, MASS, graphics
URL: https://github.com/matthewlouisdavisBioStat/lgspline
BugReports: https://github.com/matthewlouisdavisBioStat/lgspline/issues
Encoding: UTF-8
RoxygenNote: 7.3.2
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-05-08 12:58:44 UTC; matth
Author: Matthew Davis [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-9714-1018>)
Maintainer: Matthew Davis <matthewlouisdavis@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-08 13:20:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-08 15:46:55 UTC; unix
Archs: lgspline.so.dSYM

Package: lgspline
Type: Package
Title: Lagrangian Multiplier Smoothing Splines for Smooth Function
        Estimation
Version: 1.2.1
Authors@R: 
    person("Matthew", "Davis", email = "matthewlouisdavis@gmail.com", 
           role = c("aut", "cre"),
           comment = c(ORCID = "0000-0001-9714-1018"))
Description: Implements Lagrangian multiplier smoothing splines for flexible
    nonparametric regression and function estimation. Provides tools for fitting,
    prediction, and inference using a constrained optimization approach to
    enforce smoothness. Supports generalized linear models, Weibull accelerated
    failure time (AFT) models, Cox proportional hazards models, 
    quadratic programming constraints, and customizable working-correlation 
    structures, with options for parallel fitting. The core spline construction 
    builds on Ezhov et al. (2018) <doi:10.1515/jag-2017-0029>. 
    Quadratic-programming and SQP details follow Goldfarb & Idnani (1983) 
    <doi:10.1007/BF02591962> and Nocedal & Wright (2006) 
    <doi:10.1007/978-0-387-40065-5>. For smoothing spline and penalized spline 
    background, see Wahba (1990) <doi:10.1137/1.9781611970128> and Wood (2017)
    <doi:10.1201/9781315370279>. For variance-component and correlation-parameter
    estimation, see Searle et al. (2006) <ISBN:978-0470009598>. The default
    multivariate partitioning step uses k-means clustering as in MacQueen
    (1967).
License: MIT + file LICENSE
Language: en-US
Depends: R (>= 3.5.0)
Imports: Rcpp (>= 1.0.7), FNN, RColorBrewer, plotly, quadprog, methods,
        stats
LinkingTo: Rcpp, RcppArmadillo
Suggests: testthat (>= 3.0.0), spelling, knitr, rmarkdown, parallel,
        survival, MASS, graphics
URL: https://github.com/matthewlouisdavisBioStat/lgspline
BugReports: https://github.com/matthewlouisdavisBioStat/lgspline/issues
Encoding: UTF-8
RoxygenNote: 7.3.2
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-08-04 18:51:40 UTC; matth
Author: Matthew Davis [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-9714-1018>)
Maintainer: Matthew Davis <matthewlouisdavis@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-04 20:30:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-08-04 23:20:57 UTC; unix
Archs: lgspline.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(coef,lgspline)
S3method(coef,wald_lgspline)
S3method(confint,lgspline)
S3method(confint,wald_lgspline)
S3method(equation,additive_lgspline)
S3method(equation,lgspline)
S3method(integrate,default)
S3method(integrate,lgspline)
S3method(logLik,lgspline)
S3method(plot,lgspline)
S3method(plot,wald_lgspline)
S3method(predict,lgspline)
S3method(print,additive_lgspline)
S3method(print,equation)
S3method(print,lgspline)
S3method(print,summary.lgspline)
S3method(print,wald_lgspline)
S3method(summary,additive_lgspline)
S3method(summary,lgspline)
S3method(summary,wald_lgspline)
export("%**%")
export(AGAmult_wrapper)
export(GAmult_wrapper)
export(approx_grad)
export(armaInv)
export(blockfit_solve)
export(collapse_block_diagonal)
export(compute_G_eigen)
export(compute_GhalfXy_temp_wrapper)
export(compute_Lambda)
export(compute_dG_dlambda)
export(compute_dG_u_dlambda_xy)
export(compute_dGhalf)
export(compute_gram_block_diagonal)
export(compute_trace_H)
export(compute_trace_UGXX_wrapper)
export(cox_dispersion_function)
export(cox_family)
export(cox_glm_weight_function)
export(cox_qp_score_function)
export(cox_schur_correction)
export(create_block_diagonal)
export(create_onehot)
export(damped_newton_r)
export(efficient_bfgs)
export(efficient_matrix_mult)
export(equation)
export(expgrid)
export(find_extremum)
export(find_neighbors)
export(generate_posterior)
export(generate_posterior_correlation)
export(get_2ndDerivPenalty)
export(get_2ndDerivPenalty_wrapper)
export(get_B)
export(get_U)
export(get_centers)
export(get_interaction_patterns)
export(get_polynomial_expansions)
export(gramMatrix)
export(info_cox)
export(info_negbin)
export(integrate)
export(invert)
export(is_binary)
export(knot_expand_list)
export(leave_one_out)
export(lgspline)
export(lgspline.fit)
export(lgspline_cox)
export(lgspline_negbin)
export(lgspline_weibull)
export(loglik_cox)
export(loglik_negbin)
export(loglik_weibull)
export(make_constraint_matrix)
export(make_derivative_matrix)
export(make_partitions)
export(matinvsqrt)
export(matmult_U)
export(matmult_block_diagonal)
export(matsqrt)
export(negbin_dispersion_function)
export(negbin_family)
export(negbin_glm_weight_function)
export(negbin_qp_score_function)
export(negbin_schur_correction)
export(negbin_theta)
export(nr_iterate)
export(prior_loglik)
export(process_input)
export(process_qp)
export(reml_grad_from_dV)
export(score_cox)
export(score_negbin)
export(softplus)
export(std)
export(take_derivative)
export(take_interaction_2ndderivative)
export(tune_Lambda)
export(unconstrained_fit_cox)
export(unconstrained_fit_default)
export(unconstrained_fit_negbin)
export(unconstrained_fit_weibull)
export(vectorproduct_block_diagonal)
export(wald_univariate)
export(weibull_dispersion_function)
export(weibull_family)
export(weibull_glm_weight_function)
export(weibull_qp_score_function)
export(weibull_scale)
export(weibull_schur_correction)
export(weibull_shur_correction)
import(Rcpp)
import(methods)
import(stats)
importFrom(FNN,get.knnx)
importFrom(RColorBrewer,brewer.pal)
importFrom(Rcpp,sourceCpp)
importFrom(graphics,legend)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(plotly,plot_ly)
importFrom(quadprog,solve.QP)
useDynLib(lgspline)

# Introduction

This R package implements Lagrangian multiplier smoothing splines, which reformulate smoothing splines through constrained optimization. This approach provides direct access to predictor-response relationships through interpretable coefficients, unlike other formulations that require post-fitting algebraic manipulation.

# Installation 

devtools::install_github("matthewlouisdavisBioStat/lgspline")

# Citation 

If you use this package in your research, please cite:

Davis, M. (2025). Lagrangian Multiplier Smoothing Splines. https://github.com/matthewlouisdavisBioStat/lgspline/

# Contact

For questions or feedback, please open an issue on GitHub.

set.seed(1234)
library(lgspline)

################################################################################

## ## 95% Coverage Operating Characteristics ## ##
#  This script runs simulation checks for coefficient and correlation-parameter
#  coverage across the main lgspline likelihood paths. It is intentionally not a
#  CRAN test: at reviewer-scale settings it can take a long time.
#
#  The common regression target is
#
#      y ~ spl(t1, t2) + t1*t3
#
#  so the spline surface is a joined two-predictor smooth and the coefficient
#  target is the explicit t1-by-t3 interaction, repeated over all partitions.
#
#  Environment overrides:
#    LGSPLINE_COVERAGE_N          sample size; default 2000
#    LGSPLINE_COVERAGE_REPS       Monte Carlo replicates; default 100
#    LGSPLINE_COVERAGE_FAST       TRUE uses K = 1 and opt = FALSE for smoke runs
#    LGSPLINE_COVERAGE_FAST_K     K used in fast smoke mode; default 1
#    LGSPLINE_COVERAGE_SCENARIOS  comma-separated scenario names
#    LGSPLINE_COVERAGE_CSV        output CSV path

## Simulation controls
N <- as.integer(Sys.getenv("LGSPLINE_COVERAGE_N", "2000"))
n_sims <- as.integer(Sys.getenv("LGSPLINE_COVERAGE_REPS", "100"))
fast_smoke <- tolower(Sys.getenv("LGSPLINE_COVERAGE_FAST", "FALSE")) %in%
  c("true", "t", "1", "yes", "y")
fast_K <- as.integer(Sys.getenv("LGSPLINE_COVERAGE_FAST_K", "1"))
scenario_filter <- Sys.getenv("LGSPLINE_COVERAGE_SCENARIOS", "")
output_csv <- Sys.getenv(
  "LGSPLINE_COVERAGE_CSV",
  file.path(getwd(), "coverage_operating_characteristics.csv")
)

## True parameter values
beta_int <- 0.35
beta_t3 <- -0.25
sigma_gaussian <- 0.75
gamma_shape <- 6
negbin_size <- 4
weibull_scale_true <- 0.75

## Correlation parameters are stored by lgspline on working scales.
rho_ar1 <- 0.45
rho_exchangeable <- 0.25
omega_spatial <- 1.10
matern_ell <- 0.70
matern_nu <- 1.25
true_working <- list(
  ar1 = log(-log(rho_ar1)),
  exchangeable = log(-log(rho_exchangeable)),
  spatial_exponential = log(omega_spatial),
  matern = log(c(matern_ell, matern_nu))
)

## Fit controls: leave defaults alone unless doing a smoke run.
base_fit_args <- list(
  return_varcovmat = TRUE,
  include_warnings = FALSE
)
if(fast_smoke) {
  base_fit_args$K <- fast_K
  base_fit_args$opt <- FALSE
}

################################################################################

## ## Data-Generating Helpers ## ##

make_covariates <- function(N, block_size = NULL) {
  dat <- data.frame(
    t1 = runif(N, -1, 1),
    t2 = runif(N, -1, 1),
    t3 = runif(N, -1, 1)
  )

  if(!is.null(block_size)) {
    if(N %% block_size != 0) {
      stop("N must be divisible by block_size for correlated scenarios.")
    }
    dat$id <- rep(seq_len(N / block_size), each = block_size)
    dat$spacetime <- rep(seq(0, 1, length.out = block_size), N / block_size)
  }
  dat
}

smooth_signal <- function(t1, t2) {
  0.35 * sin(2.1 * t1 + 0.3 * cos(3 * t2)) +
    0.25 * cos(1.7 * t2 - 0.4 * t1^2) -
    0.30 * exp(-((t1 - 0.35)^2 + (t2 + 0.25)^2) / 0.18) +
    0.12 * atan(3 * t1 * t2)
}

eta_common <- function(dat, intercept = 0, scale = 1) {
  intercept +
    scale * smooth_signal(dat$t1, dat$t2) +
    beta_t3 * dat$t3 +
    beta_int * dat$t1 * dat$t3
}

block_cor_ar1 <- function(d, rho = rho_ar1) {
  ranked <- abs(outer(seq_along(d), seq_along(d), "-"))
  rho^ranked
}

block_cor_exchangeable <- function(d, rho = rho_exchangeable) {
  m <- length(d)
  (1 - rho) * diag(m) + rho * matrix(1, m, m)
}

block_cor_spatial_exponential <- function(d, omega = omega_spatial) {
  exp(-omega * abs(outer(d, d, "-")))
}

block_cor_matern <- function(d, ell = matern_ell, nu = matern_nu) {
  diffs <- abs(outer(d, d, "-"))
  scaled <- sqrt(2 * nu) * diffs / ell
  V <- matrix(1, length(d), length(d))
  nz <- scaled > 0
  V[nz] <- (2^(1 - nu) / gamma(nu)) *
    scaled[nz]^nu * besselK(scaled[nz], nu)
  V
}

make_exchangeable_vhalf_functions <- function(id) {
  list(
    VhalfInv_fxn = function(par) {
      rho <- exp(-exp(par))
      out <- matrix(0, length(id), length(id))
      for(clust in unique(id)) {
        inds <- which(id == clust)
        m <- length(inds)
        V <- (1 - rho) * diag(m) + rho * matrix(1, m, m)
        out[inds, inds] <- matinvsqrt(V)
      }
      out
    },
    Vhalf_fxn = function(par) {
      rho <- exp(-exp(par))
      out <- matrix(0, length(id), length(id))
      for(clust in unique(id)) {
        inds <- which(id == clust)
        m <- length(inds)
        V <- (1 - rho) * diag(m) + rho * matrix(1, m, m)
        out[inds, inds] <- matsqrt(V)
      }
      out
    },
    VhalfInv_logdet = function(par) {
      rho <- exp(-exp(par))
      log_det <- 0
      for(clust in unique(id)) {
        inds <- which(id == clust)
        m <- length(inds)
        V <- (1 - rho) * diag(m) + rho * matrix(1, m, m)
        log_det <- log_det -
          0.5 * determinant(V, logarithm = TRUE)$modulus[1]
      }
      log_det
    }
  )
}

make_spatial_exponential_vhalf_functions <- function(id, spacetime) {
  list(
    VhalfInv_fxn = function(par) {
      omega <- exp(par)
      out <- matrix(0, length(id), length(id))
      for(clust in unique(id)) {
        inds <- which(id == clust)
        d <- abs(outer(spacetime[inds], spacetime[inds], "-"))
        out[inds, inds] <- matinvsqrt(exp(-omega * d))
      }
      out
    },
    Vhalf_fxn = function(par) {
      omega <- exp(par)
      out <- matrix(0, length(id), length(id))
      for(clust in unique(id)) {
        inds <- which(id == clust)
        d <- abs(outer(spacetime[inds], spacetime[inds], "-"))
        out[inds, inds] <- matsqrt(exp(-omega * d))
      }
      out
    },
    VhalfInv_logdet = function(par) {
      omega <- exp(par)
      log_det <- 0
      for(clust in unique(id)) {
        inds <- which(id == clust)
        d <- abs(outer(spacetime[inds], spacetime[inds], "-"))
        V <- exp(-omega * d)
        log_det <- log_det -
          0.5 * determinant(V, logarithm = TRUE)$modulus[1]
      }
      log_det
    }
  )
}

draw_block_normal <- function(id, spacetime, block_cor_fxn) {
  z <- numeric(length(id))
  for(clust in unique(id)) {
    inds <- which(id == clust)
    V <- block_cor_fxn(spacetime[inds])
    R <- chol(V)
    z[inds] <- c(t(R) %*% rnorm(length(inds)))
  }
  z
}

draw_correlated_count <- function(mu, id, spacetime, block_cor_fxn) {
  ## Count-valued quasi-Poisson DGP with approximately correct working
  #  Pearson-residual correlation:
  #      y_i = round(mu_i + sqrt(mu_i) * z_i), z ~ N(0, V).
  #  The pmax() truncation is negligible for the moderately large means used
  #  below, but keeps the response in the count support.
  z <- draw_block_normal(id, spacetime, block_cor_fxn)
  as.integer(pmax(0, round(mu + sqrt(pmax(mu, .Machine$double.eps)) * z)))
}

simulate_weibull_aft <- function(mu, scale) {
  ## log(T) = log(mu) + scale * log(E), E ~ Exp(1)
  mu * rexp(length(mu))^scale
}

apply_censoring <- function(time, rate) {
  cens <- rexp(length(time), rate = rate)
  list(
    y = pmin(time, cens),
    status = as.integer(time <= cens)
  )
}

################################################################################

## ## Fit and Extraction Helpers ## ##

call_fit <- function(fit_fun, formula, data, extra_args = list()) {
  args <- c(list(formula = formula, data = data), base_fit_args, extra_args)
  do.call(fit_fun, args)
}

extract_interaction_coverage <- function(fit, scenario, sim_id) {
  ci <- confint(fit)
  wald <- wald_univariate(fit)$coefficients
  rows <- grep("(^|\\.)t1xt3$|(^|\\.)t3xt1$", rownames(ci), value = TRUE)
  if(length(rows) == 0) {
    rows <- grep("t1xt3|t3xt1", rownames(ci), value = TRUE)
  }

  if(length(rows) == 0) {
    return(data.frame(
      scenario = scenario,
      sim = sim_id,
      estimand = "interaction",
      parameter = "t1xt3",
      true = beta_int,
      estimate = NA_real_,
      lower = NA_real_,
      upper = NA_real_,
      covered = NA,
      status = "missing_interaction_ci"
    ))
  }

  lower <- ci[rows, 1]
  upper <- ci[rows, 2]
  covered <- lower <= beta_int & beta_int <= upper
  status <- ifelse(is.finite(lower) & is.finite(upper),
                   "ok", "nonfinite_interaction_ci")

  data.frame(
    scenario = scenario,
    sim = sim_id,
    estimand = "interaction",
    parameter = rows,
    true = beta_int,
    estimate = wald[rows, "Estimate"],
    lower = lower,
    upper = upper,
    covered = covered,
    status = status,
    row.names = NULL
  )
}

extract_correlation_coverage <- function(fit, scenario, sim_id, true_par) {
  ci <- confint(fit)
  rows <- grep("^Correlation parameter", rownames(ci), value = TRUE)
  if(length(rows) == 0) {
    return(data.frame(
      scenario = scenario,
      sim = sim_id,
      estimand = "correlation",
      parameter = paste0("Correlation parameter ", seq_along(true_par)),
      true = true_par,
      estimate = NA_real_,
      lower = NA_real_,
      upper = NA_real_,
      covered = NA,
      status = "missing_correlation_ci"
    ))
  }

  true_par <- rep(true_par, length.out = length(rows))
  lower <- ci[rows, 1]
  upper <- ci[rows, 2]
  covered <- lower <= true_par & true_par <= upper
  status <- ifelse(is.finite(lower) & is.finite(upper),
                   "ok", "nonfinite_correlation_ci")

  data.frame(
    scenario = scenario,
    sim = sim_id,
    estimand = "correlation",
    parameter = rows,
    true = true_par,
    estimate = fit$VhalfInv_params_estimates[seq_along(rows)],
    lower = lower,
    upper = upper,
    covered = covered,
    status = status,
    row.names = NULL
  )
}

summarize_results <- function(results) {
  ok <- results[results$status == "ok" & !is.na(results$covered), , drop = FALSE]
  if(nrow(ok) == 0) return(data.frame())

  split_res <- split(ok, list(ok$scenario, ok$estimand, ok$parameter),
                     drop = TRUE)
  out <- do.call(rbind, lapply(split_res, function(x) {
    p <- mean(x$covered)
    data.frame(
      scenario = x$scenario[1],
      estimand = x$estimand[1],
      parameter = x$parameter[1],
      n = nrow(x),
      coverage = p,
      mc_se = sqrt(p * (1 - p) / nrow(x)),
      mean_estimate = mean(x$estimate),
      mean_width = mean(x$upper - x$lower),
      row.names = NULL
    )
  }))
  out[order(out$scenario, out$estimand, out$parameter), ]
}

write_results <- function(results, output_csv) {
  if(is.null(results) || nrow(results) == 0) return(invisible(NULL))
  dir.create(dirname(output_csv), recursive = TRUE, showWarnings = FALSE)
  write.csv(results, output_csv, row.names = FALSE)
  summary_csv <- sub("\\.csv$", "_summary.csv", output_csv)
  write.csv(summarize_results(results), summary_csv, row.names = FALSE)
  invisible(summary_csv)
}

record_failure <- function(scenario, sim_id, err) {
  data.frame(
    scenario = scenario,
    sim = sim_id,
    estimand = "fit",
    parameter = "fit",
    true = NA_real_,
    estimate = NA_real_,
    lower = NA_real_,
    upper = NA_real_,
    covered = NA,
    status = paste("error:", conditionMessage(err)),
    row.names = NULL
  )
}

################################################################################

## ## Scenario Definitions ## ##

fit_formula <- y ~ spl(t1, t2) + t1*t3
time_formula <- time ~ spl(t1, t2) + t1*t3

scenario_gaussian <- function(sim_id) {
  dat <- make_covariates(N)
  eta <- eta_common(dat, intercept = 0.2, scale = 1)
  dat$y <- eta + rnorm(N, 0, sigma_gaussian)
  fit <- call_fit(lgspline, fit_formula, dat)
  extract_interaction_coverage(fit, "gaussian", sim_id)
}

scenario_logistic <- function(sim_id) {
  dat <- make_covariates(N)
  eta <- eta_common(dat, intercept = -0.10, scale = 0.75)
  dat$y <- rbinom(N, 1, plogis(eta))
  fit <- call_fit(lgspline, fit_formula, dat,
                  list(family = binomial()))
  extract_interaction_coverage(fit, "logistic", sim_id)
}

scenario_poisson <- function(sim_id) {
  dat <- make_covariates(N)
  eta <- eta_common(dat, intercept = 0.55, scale = 0.65)
  dat$y <- rpois(N, exp(eta))
  fit <- call_fit(lgspline, fit_formula, dat,
                  list(family = poisson()))
  extract_interaction_coverage(fit, "poisson", sim_id)
}

scenario_gamma_log <- function(sim_id) {
  dat <- make_covariates(N)
  eta <- eta_common(dat, intercept = 0.45, scale = 0.55)
  mu <- exp(eta)
  dat$y <- rgamma(N, shape = gamma_shape, scale = mu / gamma_shape)
  fit <- call_fit(lgspline, fit_formula, dat,
                  list(family = Gamma(link = "log"),
                       keep_weighted_Lambda = TRUE))
  extract_interaction_coverage(fit, "gamma_log", sim_id)
}

scenario_gaussian_ar1 <- function(sim_id) {
  dat <- make_covariates(N, block_size = 5)
  eta <- eta_common(dat, intercept = 0.2, scale = 1)
  err <- draw_block_normal(dat$id, dat$spacetime, block_cor_ar1)
  dat$y <- eta + sigma_gaussian * err
  fit <- call_fit(
    lgspline, fit_formula, dat,
    list(correlation_id = dat$id,
         spacetime = dat$spacetime,
         correlation_structure = "ar(1)",
         VhalfInv_par_init = true_working$ar1)
  )
  rbind(
    extract_interaction_coverage(fit, "gaussian_ar1", sim_id),
    extract_correlation_coverage(fit, "gaussian_ar1", sim_id,
                                 true_working$ar1)
  )
}

scenario_gaussian_matern <- function(sim_id) {
  dat <- make_covariates(N, block_size = 5)
  eta <- eta_common(dat, intercept = 0.2, scale = 1)
  err <- draw_block_normal(dat$id, dat$spacetime, block_cor_matern)
  dat$y <- eta + sigma_gaussian * err
  fit <- call_fit(
    lgspline, fit_formula, dat,
    list(correlation_id = dat$id,
         spacetime = dat$spacetime,
         correlation_structure = "matern",
         VhalfInv_par_init = true_working$matern)
  )
  rbind(
    extract_interaction_coverage(fit, "gaussian_matern", sim_id),
    extract_correlation_coverage(fit, "gaussian_matern", sim_id,
                                 true_working$matern)
  )
}

scenario_poisson_gee_spatial <- function(sim_id) {
  dat <- make_covariates(N, block_size = 5)
  eta <- eta_common(dat, intercept = 1.60, scale = 0.50)
  mu <- exp(eta)
  dat$y <- draw_correlated_count(
    mu, dat$id, dat$spacetime, block_cor_spatial_exponential)
  vf <- make_spatial_exponential_vhalf_functions(dat$id, dat$spacetime)
  fit <- call_fit(
    lgspline, fit_formula, dat,
    list(family = quasipoisson(),
         VhalfInv_fxn = vf$VhalfInv_fxn,
         Vhalf_fxn = vf$Vhalf_fxn,
         VhalfInv_logdet = vf$VhalfInv_logdet,
         VhalfInv_par_init = true_working$spatial_exponential)
  )
  rbind(
    extract_interaction_coverage(fit, "poisson_gee_spatial", sim_id),
    extract_correlation_coverage(fit, "poisson_gee_spatial", sim_id,
                                 true_working$spatial_exponential)
  )
}

scenario_poisson_gee_exchangeable <- function(sim_id) {
  dat <- make_covariates(N, block_size = 5)
  eta <- eta_common(dat, intercept = 1.60, scale = 0.50)
  mu <- exp(eta)
  dat$y <- draw_correlated_count(
    mu, dat$id, dat$spacetime, block_cor_exchangeable)
  vf <- make_exchangeable_vhalf_functions(dat$id)
  fit <- call_fit(
    lgspline, fit_formula, dat,
    list(family = quasipoisson(),
         VhalfInv_fxn = vf$VhalfInv_fxn,
         Vhalf_fxn = vf$Vhalf_fxn,
         VhalfInv_logdet = vf$VhalfInv_logdet,
         VhalfInv_par_init = true_working$exchangeable)
  )
  rbind(
    extract_interaction_coverage(fit, "poisson_gee_exchangeable", sim_id),
    extract_correlation_coverage(fit, "poisson_gee_exchangeable", sim_id,
                                 true_working$exchangeable)
  )
}

scenario_negbin <- function(sim_id) {
  dat <- make_covariates(N)
  eta <- eta_common(dat, intercept = 0.55, scale = 0.65)
  dat$y <- rnbinom(N, size = negbin_size, mu = exp(eta))
  fit <- call_fit(lgspline_negbin, fit_formula, dat)
  extract_interaction_coverage(fit, "negative_binomial", sim_id)
}

scenario_cox <- function(sim_id) {
  dat <- make_covariates(N)
  eta <- eta_common(dat, intercept = 0, scale = 0.70)
  event_time <- rexp(N, rate = 0.12 * exp(eta))
  obs <- apply_censoring(event_time, rate = 0.08)
  dat$time <- obs$y
  fit <- call_fit(lgspline_cox, time_formula, dat,
                  list(status = obs$status))
  extract_interaction_coverage(fit, "cox_ph", sim_id)
}

scenario_weibull_aft <- function(sim_id) {
  dat <- make_covariates(N)
  eta <- eta_common(dat, intercept = 1.5, scale = 0.65)
  event_time <- simulate_weibull_aft(exp(eta), weibull_scale_true)
  obs <- apply_censoring(event_time, rate = 0.06)
  dat$time <- obs$y
  fit <- call_fit(lgspline_weibull, time_formula, dat,
                  list(status = obs$status))
  extract_interaction_coverage(fit, "weibull_aft", sim_id)
}

scenarios <- list(
  gaussian = scenario_gaussian,
  logistic = scenario_logistic,
  poisson = scenario_poisson,
  gamma_log = scenario_gamma_log,
  gaussian_ar1 = scenario_gaussian_ar1,
  gaussian_matern = scenario_gaussian_matern,
  poisson_gee_spatial = scenario_poisson_gee_spatial,
  poisson_gee_exchangeable = scenario_poisson_gee_exchangeable,
  negative_binomial = scenario_negbin,
  cox_ph = scenario_cox,
  weibull_aft = scenario_weibull_aft
)

if(nzchar(scenario_filter)) {
  keep <- trimws(strsplit(scenario_filter, ",")[[1]])
  missing <- setdiff(keep, names(scenarios))
  if(length(missing) > 0) {
    stop("Unknown scenario(s): ", paste(missing, collapse = ", "))
  }
  scenarios <- scenarios[keep]
}

################################################################################

## ## Run Simulation ## ##

cat("lgspline coverage operating-characteristic simulation\n")
cat("N:", N, "\n")
cat("Replicates:", n_sims, "\n")
cat("Fast smoke mode:", fast_smoke, "\n")
if(fast_smoke) cat("Fast smoke K:", fast_K, "\n")
cat("Scenarios:", paste(names(scenarios), collapse = ", "), "\n")
cat("Output CSV:", output_csv, "\n\n")
flush.console()

all_results <- list()
counter <- 0L

start_time <- Sys.time()
for(scenario_name in names(scenarios)) {
  for(sim_id in seq_len(n_sims)) {
    counter <- counter + 1L
    cat(sprintf("[%s] replicate %d/%d\n", scenario_name, sim_id, n_sims))
    flush.console()
    one <- tryCatch(
      scenarios[[scenario_name]](sim_id),
      error = function(e) record_failure(scenario_name, sim_id, e)
    )
    all_results[[counter]] <- one
    write_results(do.call(rbind, all_results), output_csv)
  }
}

results <- do.call(rbind, all_results)
summary_results <- summarize_results(results)

cat("\nRaw result status counts:\n")
print(table(results$scenario, results$status), quote = FALSE)

cat("\nCoverage summary:\n")
print(summary_results, row.names = FALSE)

summary_csv <- write_results(results, output_csv)

cat("\nSaved raw results to:\n", output_csv, "\n", sep = "")
cat("Saved summary results to:\n", summary_csv, "\n", sep = "")
cat("Elapsed time:", format(Sys.time() - start_time), "\n")

invisible(list(raw = results, summary = summary_results))

test_that("basic correlation structure runs without error", {
  set.seed(1234)
  x <- seq(-9, 9, length.out = 1000)
  y <- sin(x) + rnorm(1000, 0, 0.1)
  model_fit <- lgspline(x, y, K = 1, standardize_response = FALSE,
                  VhalfInv = diag(length(y)))
  expect_error(model_fit$model_fit$VhalfInv_params_estimates, NA)
})


test_that("built-in Gaussian correlation structures provide Vhalf_fxn", {
  set.seed(123)

  id <- rep(1:12, each = 4)
  x <- rep(seq(0, 1, length.out = 4), 12)
  y <- sin(2 * pi * x) + rnorm(length(x), 0, 0.05)

  fit <- lgspline(cbind(x), y,
                  K = 1,
                  correlation_id = id,
                  correlation_structure = "exchangeable",
                  include_warnings = FALSE)

  fit_core <- if(!is.null(fit$model_fit)) fit$model_fit else fit

  expect_true(is.function(fit_core$Vhalf_fxn))
  expect_false(is.null(fit_core$VhalfInv_params_estimates))

  Vhalf <- fit_core$Vhalf_fxn(fit_core$VhalfInv_params_estimates)
  expect_true(is.matrix(Vhalf))
  expect_equal(dim(Vhalf), c(length(y), length(y)))
  expect_true(all(is.finite(Vhalf)))
})

test_that("exchangeable correlation supports unconstrained K = 0 GLM fits", {
  set.seed(456)

  id <- rep(1:18, each = 4)
  x <- rep(seq(0, 1, length.out = 4), 18)
  eta <- 0.25 + 0.9 * x
  y <- rpois(length(x), lambda = exp(eta))

  expect_error({
    fit <- lgspline(cbind(x), y,
                    family = quasipoisson(),
                    K = 0,
                    opt = FALSE,
                    correlation_id = id,
                    correlation_structure = "exchangeable",
                    include_warnings = FALSE)
  }, NA)

  fit_core <- if(!is.null(fit$model_fit)) fit$model_fit else fit

  expect_true(is.finite(fit_core$sigmasq_tilde))
  expect_true(all(is.finite(fit_core$varcovmat)))
})

test_that("weighted correlated Gaussian varcov and logLik use the whitened-system weights once", {
  set.seed(20260314)

  n <- 10
  x <- seq(-1, 1, length.out = n)
  y <- 1 + 0.4 * x + rnorm(n, 0, 0.03)
  w <- seq(0.8, 1.7, length.out = n)
  rho <- 0.35
  V <- rho ^ abs(outer(seq_len(n), seq_len(n), `-`))
  VhalfInv <- t(chol(solve(V)))

  fit <- lgspline(
    cbind(x),
    y,
    K = 0,
    opt = FALSE,
    VhalfInv = VhalfInv,
    observation_weights = w,
    just_linear_without_interactions = 1,
    standardize_response = FALSE,
    include_warnings = FALSE
  )

  fit_core <- if(!is.null(fit$model_fit)) fit$model_fit else fit

  X_full <- collapse_block_diagonal(fit_core$X)[unlist(fit_core$og_order), , drop = FALSE]
  VinvhalfX <- fit_core$VhalfInv %*% X_full
  VinvhalfX <- t(t(VinvhalfX) * sqrt(fit_core$weights))

  Lambda_full <- collapse_block_diagonal(
    lapply(seq_len(fit_core$K + 1), function(k) {
      if(length(fit_core$penalties$L_partition_list) == (fit_core$K + 1)) {
        fit_core$penalties$Lambda + fit_core$penalties$L_partition_list[[k]]
      } else {
        fit_core$penalties$Lambda
      }
    })
  )

  gram_expected <- crossprod(VinvhalfX) + Lambda_full
  varcov_expected <- solve(gram_expected) * fit_core$sigmasq_tilde

  resid_w <- c(fit_core$VhalfInv %*% cbind(fit_core$y - fit_core$ytilde))
  logdet_VhalfInv <- determinant(fit_core$VhalfInv, logarithm = TRUE)$modulus[[1L]]
  ll_expected <- -0.5 * fit_core$N * log(2 * pi * fit_core$sigmasq_tilde) +
    logdet_VhalfInv -
    0.5 * sum(fit_core$weights * resid_w^2) / fit_core$sigmasq_tilde

  expect_equal(fit_core$varcovmat, varcov_expected, tolerance = 1e-4)
  expect_equal(as.numeric(logLik(fit_core, include_prior = FALSE)), ll_expected,
               tolerance = 1e-4)
})

#' Log-Prior Distribution Evaluation for lgspline Models
#'
#' @description
#' Evaluates the log-prior on the spline coefficients conditional on the
#' dispersion and penalty matrices.
#'
#' @details
#' Returns the quadratic form of \eqn{\beta^{T}\Lambda\beta} evaluated at the
#' tuned or fixed penalties, scaled by negative one-half inverse dispersion.
#'
#' Assuming fixed penalties, the prior on \eqn{\beta} is taken to be
#' \deqn{\beta | \sigma^2 \sim \mathcal{N}(\textbf{0}, \sigma^2\Lambda^{-1})}
#' so that, up to a normalizing constant \eqn{C} with respect to \eqn{\beta},
#' \deqn{\implies \log P(\beta|\sigma^2) = C-\frac{1}{2\sigma^2}\beta^{T}\Lambda\beta}
#'
#' The value of \eqn{C} is included when \code{include_constant = TRUE}, and
#' omitted when \code{FALSE}.
#'
#' This is useful for computing joint penalized log-likelihoods and related
#' MAP-style diagnostics for a fitted \code{lgspline} object.
#'
#' @param model_fit An lgspline model object.
#' @param sigmasq Numeric scalar dispersion parameter. If NULL,
#'   \code{model_fit$sigmasq_tilde} is used.
#' @param include_constant Logical; if TRUE (default), include the
#'   multivariate normal normalizing constant.
#'
#' @return A numeric scalar representing the prior log-likelihood.
#'
#' @examples
#'
#' \donttest{
#' ## Data
#' t <- sort(runif(100, -5, 5))
#' y <- sin(t) - 0.1*t^2 + rnorm(100)
#'
#' ## Model keeping penalties fixed
#' model_fit <- lgspline(t, y, opt = FALSE)
#'
#' ## Full joint log-likelihood, conditional upon known sigma^2 = 1
#' jntloglik <- sum(dnorm(model_fit$y,
#'                     model_fit$ytilde,
#'                     1,
#'                     log = TRUE)) +
#'           prior_loglik(model_fit, sigmasq = 1)
#' print(jntloglik)
#' }
#'
#' @seealso \code{\link{lgspline}}
#'
#' @export
prior_loglik <- function(model_fit,

                         sigmasq = NULL,

                         include_constant = TRUE){



  if(is.null(sigmasq)){

    sigmasq <- model_fit$sigmasq_tilde

  }



  ## If no partition-specific penalties were stored, treat them as zero.
  if(length(model_fit$penalties$L_partition_list) == 0){

    model_fit$penalties$L_partition_list <- lapply(

      1:(model_fit$K + 1), function(k) 0

    )

  }



  running_sum <- 0

  total_constant <- 0



  ## Sum the quadratic penalty contribution partition by partition.
  for(k in 1:(model_fit$K + 1)){



    beta_k <- model_fit$B_raw[[k]]

    Lambda_total <- model_fit$penalties$Lambda +

      model_fit$penalties$L_partition_list[[k]]



    p_k <- length(beta_k)



    quad_form <- as.numeric(

      t(beta_k) %**% Lambda_total %**% beta_k

    )



    running_sum <- running_sum -

      0.5 * quad_form / sigmasq



    ## Add the multivariate-normal normalizing constant when requested.
    #  For beta | sigma^2 ~ N(0, sigma^2 Lambda^{-1}),
    #  the constant contribution is
    #    -0.5 * (p log(2 pi) + p log(sigma^2) - log |Lambda|).
    if(include_constant){



      logdet <- as.numeric(

        determinant(Lambda_total, logarithm = TRUE)$modulus

      )



      total_constant <- total_constant -

        0.5 * (

          p_k * log(2 * pi) +

            p_k * log(sigmasq) -

            logdet

        )

    }

  }



  return(as.numeric(running_sum + total_constant))

}

test_that("process_input resolves names, offsets, and encoded factor groups", {
  df <- data.frame(
    y = c(1, 2, 3, 4, 5, 6),
    t = seq(-1, 1, length.out = 6),
    z = c(10, 11, 12, 13, 14, 15),
    grp = factor(c("a", "b", "c", "a", "b", "c"))
  )

  processed <- process_input(
    predictors = y ~ spl(t) + grp + offset(z),
    data = df,
    just_linear_without_interactions = "z",
    do_not_cluster_on_these = "z",
    auto_encode_factors = TRUE,
    include_warnings = FALSE
  )

  expect_true(is.matrix(processed$predictors))
  expect_equal(length(processed$offset), 1)
  expect_equal(length(processed$do_not_cluster_on_these), 1)
  expect_identical(processed$offset, processed$do_not_cluster_on_these)
  expect_true("grp" %in% names(processed$factor_groups))
  expect_true(length(processed$factor_groups$grp) > 1)
  expect_true(all(processed$factor_groups$grp %in%
                    processed$just_linear_without_interactions))
})

test_that("integration and prior helper methods return coherent values", {
  t <- seq(-1, 2, length.out = 100)
  y <- 2 * t + 1

  fit <- lgspline(
    t,
    y,
    K = 1,
    opt = FALSE,
    just_linear_without_interactions = 1,
    standardize_response = FALSE
  )
  fit_linear_only <- lgspline(
    t,
    y,
    K = 1,
    opt = FALSE,
    just_linear_with_interactions = 1,
    standardize_response = FALSE
  )

  integ_resp <- integrate(fit, lower = -1, upper = 2, n_quad = 20)
  integ_link <- integrate(fit, lower = -1, upper = 2, n_quad = 20,
                          link_scale = TRUE)
  integ_linear_only <- integrate(fit_linear_only, lower = -1, upper = 2,
                                 n_quad = 20)

  expect_equal(integ_resp, 6, tolerance = 1e-4)
  expect_equal(integ_link, integ_resp, tolerance = 1e-4)
  expect_equal(integ_linear_only, integ_resp, tolerance = 1e-4)

  lp_with_const <- prior_loglik(fit, include_constant = TRUE)
  lp_no_const <- prior_loglik(fit, include_constant = FALSE)

  part_penalties <- fit$penalties$L_partition_list
  if(length(part_penalties) == 0){
    part_penalties <- lapply(seq_len(fit$K + 1), function(k) 0)
  }
  expected_const <- sum(vapply(seq_len(fit$K + 1), function(k) {
    Lambda_total <- fit$penalties$Lambda + part_penalties[[k]]
    -0.5 * (
      length(fit$B_raw[[k]]) * log(2 * pi) +
        length(fit$B_raw[[k]]) * log(fit$sigmasq_tilde) -
        as.numeric(determinant(Lambda_total, logarithm = TRUE)$modulus)
    )
  }, numeric(1)))

  expect_true(is.numeric(lp_with_const) && length(lp_with_const) == 1)
  expect_true(is.numeric(lp_no_const) && length(lp_no_const) == 1)
  expect_true(is.finite(lp_with_const))
  expect_true(is.finite(lp_no_const))
  expect_equal(lp_with_const - lp_no_const, expected_const, tolerance = 1e-4)
})

test_that("leave_one_out matches explicit Gaussian refits in a linear case", {
  t <- seq(-2, 2, length.out = 8)
  y <- 1 + 0.5 * t + c(0.1, -0.05, 0.02, -0.03, 0.04, -0.02, 0.01, -0.04)

  fit <- lgspline(
    cbind(t),
    y,
    K = 0,
    opt = FALSE,
    just_linear_without_interactions = 1,
    standardize_response = FALSE
  )

  loo_fast <- leave_one_out(fit)

  loo_refit <- sapply(seq_along(y), function(i) {
    refit <- lgspline(
      cbind(t[-i]),
      y[-i],
      K = 0,
      opt = FALSE,
      just_linear_without_interactions = 1,
      standardize_response = FALSE
    )
    as.numeric(predict(refit, cbind(t[i])))
  })

  expect_equal(loo_fast, loo_refit, tolerance = 1e-4)
})

#' Lagrangian Multiplier Smoothing Splines: Mathematical Details
#'
#' @description
#' This document provides the mathematical and implementation details for
#' Lagrangian Multiplier Smoothing Splines as implemented in \pkg{lgspline}.
#'
#' This package provides exhaustive resources for fitting multivariate smoothing
#' smoothing splines with a monomial basis and analytical form cubic smoothing
#' spline penalty.
#'
#' The material is presented such that a programmer or statistician of
#' reasonable experience and background can understand and implement the
#' procedure from scratch, and also potentially critique some of the modelling
#' choices that went into designing this package.
#'
#' Informally, \pkg{lgspline} answers the following question:
#' How can we best adapt a useful functionality of basis splines, under the
#' alternative interpretation of smoothing as explicit external constraints instead?
#'
#' The obvious benefit is a much more flexible and interpretable final model that for
#' non-experienced users is simply easier to understand without post-hoc processing, and
#' for experienced users can be used to customize models more easily.
#'
#' The drawback is that the interpretation of constraints as external adds a new layer of complexity
#' to each step of the model fitting process, whereas for implicit design matrix
#' construction these complications are bypassed.
#'
#' While it is true a B-spline can always be converted back into monomial form,
#' tensor-product splines that generalize this to multiple dimensions often
#' explodes the number and degree of interaction terms, the conversion may not
#' be computationally stable, and it is not available in standard software.
#'
#' @section Statistical Problem Formulation:
#'
#' Consider an \eqn{N \times q} matrix of predictors
#' \eqn{\mathbf{T} = (\mathbf{t}_1, \dots, \mathbf{t}_N)^{\top}} and an
#' \eqn{N \times 1} response vector \eqn{\mathbf{y} = (y_1, \dots, y_N)^{\top}}.
#' We assume the relationship follows a generalized linear model with unknown
#' smooth function \eqn{f}:
#' \deqn{y_i \sim \mathcal{D}(g^{-1}(f(\mathbf{t}_i)),\, \sigma^2)}
#' where \eqn{\mathcal{D}} is a distribution (e.g. exponential family or related) with mean
#' \eqn{\mu_i = g^{-1}(f(\mathbf{t}_i))}, link function \eqn{g(\cdot)}, and
#' dispersion parameter \eqn{\sigma^2}. For Gaussian response with identity
#' link, observations are independently distributed as
#' \eqn{y_i \mid \mathbf{t}_i, \sigma^2 \sim \mathcal{N}(f(\mathbf{t}_i), \sigma^2)}.
#'
#' The objective is to estimate \eqn{f} by:
#' \enumerate{
#'   \item Partitioning the predictor space into \eqn{K+1} mutually exclusive regions.
#'   \item Fitting local polynomial models within each partition.
#'   \item Enforcing smoothness at partition boundaries via Lagrangian multipliers.
#'   \item Penalizing the integrated squared second derivative to discourage roughness.
#' }
#'
#' Unlike other smoothing spline formulations, no post-fitting algebraic
#' rearrangement or disentanglement of a spline basis is needed to obtain
#' interpretable models. The polynomial expansions are homogeneous across
#' partitions, and the relationship between predictor and response is explicit
#' at the coefficient level.
#'
#' To anchor the notation, in the single-predictor cubic case one would write
#' \deqn{\hat{f}(t_i) = \hat{\beta}_{(0)} + \hat{\beta}_{(1)}t_i +
#'   \hat{\beta}_{(2)}t_i^2 + \hat{\beta}_{(3)}t_i^3 =
#'   \mathbf{x}_i^{\top}\hat{\boldsymbol{\beta}},}
#' where \eqn{\mathbf{x}_i = (1, t_i, t_i^2, t_i^3)^{\top}}. The LMSS
#' formulation preserves exactly this kind of polynomial representation, but now
#' does so within each partition and then forces neighboring pieces to agree in
#' the smoothness conditions described below.
#'
#' Core notation used throughout:
#' \itemize{
#'   \item \eqn{\mathbf{y}_{(N \times 1)}}: Response vector.
#'   \item \eqn{\mathbf{T}_{(N \times q)}}: Matrix of predictors.
#'   \item \eqn{\mathbf{X}_{(N \times P)}}: Block-diagonal matrix of polynomial
#'     expansions, with diagonal blocks \eqn{\mathbf{X}_k} of dimension
#'     \eqn{n_k \times p}.
#'   \item \eqn{\boldsymbol{\Lambda}_{(P \times P)}}: Block-diagonal penalty matrix,
#'     with blocks \eqn{\boldsymbol{\Lambda}_k} of dimension \eqn{p \times p}.
#'   \item \eqn{\hat{\boldsymbol{\beta}}_{(P \times 1)}}: Unconstrained penalized
#'     estimate.
#'   \item \eqn{\tilde{\boldsymbol{\beta}}_{(P \times 1)}}: Constrained coefficient
#'     estimates.
#'   \item \eqn{\mathbf{G}_{(P \times P)}}: Block-diagonal matrix with blocks
#'     \eqn{\mathbf{G}_k = (\mathbf{X}_k^{\top}\mathbf{W}_k\mathbf{D}_k\mathbf{X}_k + \boldsymbol{\Lambda}_k)^{-1}},
#'     where \eqn{\mathbf{W}_k} and \eqn{\mathbf{D}_k} are defined below.
#'   \item \eqn{\mathbf{A}_{(P \times r)}}: Constraint matrix encoding smoothness
#'     conditions. Reduced to linearly independent columns via pivoted QR
#'     decomposition.
#'   \item \eqn{\mathbf{U}_{(P \times P)}}:
#'     \eqn{\mathbf{I} - \mathbf{G}\mathbf{A}(\mathbf{A}^{\top}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{\top}}.
#'   \item \eqn{\mathbf{D}_{(N \times N)}}: Diagonal matrix of user-supplied
#'     observation weights (\code{observation_weights} or \code{weights}).
#'     Defaults to the identity. These play the role of prior precision on
#'     individual observations: a weight of 2 is equivalent to seeing that
#'     observation twice.
#'   \item \eqn{\mathbf{W}_{(N \times N)}}: Diagonal matrix of GLM working
#'     weights. In the implementation these diagonal entries are whatever is
#'     returned by \code{glm_weight_function}; by default this is
#'     \code{family$variance(mu)}, optionally multiplied by user-supplied
#'     observation weights. For Gaussian response with identity link,
#'     \eqn{\mathbf{W} = \mathbf{I}}. For other families,
#'     \eqn{\mathbf{W}} depends on the current fitted values and is updated at
#'     each Newton-Raphson iteration. For the common canonical families used
#'     by default, this matches the familiar Fisher-scoring weighting role.
#'   \item \eqn{\mathbf{V}_{(N \times N)}}: Correlation matrix of errors.
#'     When no correlation structure is specified, \eqn{\mathbf{V} = \mathbf{I}}.
#'     Otherwise supplied via \code{VhalfInv} or estimated through
#'     \code{VhalfInv_fxn}.
#' }
#'
#' In the Gaussian identity case with unit weights and no correlation,
#' \eqn{\mathbf{G}_k = (\mathbf{X}_k^{\top}\mathbf{X}_k + \boldsymbol{\Lambda}_k)^{-1}}
#' and most formulas simplify accordingly. When \eqn{\mathbf{D}} or
#' \eqn{\mathbf{W}} appear in a formula, the product \eqn{\mathbf{W}\mathbf{D}}
#' means ``GLM working weights times observation weights''; whenever one of
#' them is the identity it drops out.
#'
#' Before these quantities reach the main fitting stage, the user-facing inputs
#' are parsed, standardized, and organized by \code{\link{process_input}}. When
#' the formula interface is used and \code{auto_encode_factors = TRUE}, that
#' preprocessing step also relies on helpers such as \code{\link{create_onehot}}
#' to encode factor levels before the design reaches \code{\link{lgspline.fit}()}.
#' The notation in the remainder of this document therefore refers to the internal
#' objects that actually enter \code{\link{lgspline.fit}()}, not necessarily the raw
#' objects originally supplied by the user.
#'
#' From the user side, many of the arguments that control these internal objects
#' can be supplied either individually or through the grouped lists
#' \code{penalty_args}, \code{tuning_args}, \code{expansion_args},
#' \code{constraint_args}, \code{qp_args}, \code{parallel_args},
#' \code{covariance_args}, \code{return_args}, and \code{glm_args}, as
#' documented in \code{\link{lgspline}}. These grouped lists are unpacked before
#' dispatch into the same fitting pipeline, so they are a convenience layer
#' rather than a separate modeling abstraction. A closely related exploratory
#' mode is \code{dummy_fit = TRUE} in \code{\link{lgspline}} or
#' \code{\link{lgspline.fit}}, which runs the preprocessing, partition
#' construction, expansion building, and penalty setup without solving for
#' nonzero coefficients, making it a practical way to inspect objects such as
#' \code{X}, \code{A}, the returned \code{make_partition_list} from
#' \code{\link{make_partitions}}, and the assembled \code{penalties} from
#' \code{\link{compute_Lambda}} before a full fit.
#'
#'
#' @section Model Formulation and Estimation:
#'
#' \subsection{Piecewise Polynomial Structure}{
#' For \eqn{K} knots (one predictor) or \eqn{K+1} partitions (multiple
#' predictors) there are \eqn{K+1} mutually exclusive partitions
#' \eqn{\mathcal{P}_0, \dots, \mathcal{P}_{K}}. Each observation \eqn{i}
#' belongs to exactly one partition. Within partition \eqn{k}, the function is
#' represented as a polynomial of degree \eqn{p-1} in each predictor:
#' \deqn{\hat{f}_k(\mathbf{t}) = \mathbf{x}^{\top}\tilde{\boldsymbol{\beta}}_k}
#' where \eqn{\mathbf{x}} collects the polynomial basis terms (intercept,
#' linear, quadratic, cubic, and optionally quartic and interaction terms) and
#' \eqn{\tilde{\boldsymbol{\beta}}_k} are the corresponding coefficients. In
#' one predictor, the same idea can be written more explicitly as
#' \deqn{\hat{f}(t_i) = \sum_{k=0}^{K}
#'   \mathbf{x}_{ik}^{\top}\hat{\boldsymbol{\beta}}_k
#'   \mathbf{1}(t_i \in \mathcal{P}_k),}
#' which highlights that the unconstrained problem is just a collection of
#' local polynomial regressions. The expansions are homogeneous across
#' partitions, so coefficients are directly comparable. This is implemented via
#' \code{\link{get_polynomial_expansions}}.
#'
#' The exact contents of \eqn{\mathbf{x}} are controlled by the basis-expansion
#' arguments documented in \code{\link{lgspline}}: \code{include_quadratic_terms},
#' \code{include_cubic_terms}, \code{include_quartic_terms},
#' \code{include_2way_interactions}, \code{include_3way_interactions},
#' \code{include_quadratic_interactions}, \code{exclude_interactions_for},
#' \code{exclude_these_expansions}, and \code{custom_basis_fxn}. Likewise,
#' \code{just_linear_with_interactions} and
#' \code{just_linear_without_interactions} determine which predictors remain
#' structurally linear even though they still participate in the same
#' partition-wise polynomial bookkeeping described here.
#'
#' Letting \eqn{p} denote the number of basis terms per partition,
#' \eqn{P = p(K+1)} is the total number of coefficients. The full design matrix
#' \eqn{\mathbf{X}} and penalty matrix \eqn{\boldsymbol{\Lambda}} are
#' block-diagonal with \eqn{K+1} blocks, so unconstrained estimation reduces to
#' \eqn{K+1} independent penalized regressions, which appears as follows for the identity link case:
#' \deqn{\hat{\boldsymbol{\beta}}_k = \mathbf{G}_k \mathbf{X}_k^{\top} \mathbf{W}_k\mathbf{D}_k\mathbf{y}_k, \quad
#'   \mathbf{G}_k = (\mathbf{X}_k^{\top}\mathbf{W}_k\mathbf{D}_k\mathbf{X}_k + \boldsymbol{\Lambda}_k)^{-1}.}
#' For Gaussian identity with unit weights this reduces to the familiar
#' \eqn{\mathbf{G}_k = (\mathbf{X}_k^{\top}\mathbf{X}_k + \boldsymbol{\Lambda}_k)^{-1}}.
#' The block-diagonal structure means these can be computed in parallel across
#' partitions. In the user-facing interface this is realized by supplying a
#' cluster through \code{cl}, optionally controlling work splitting with
#' \code{chunk_size}, and enabling stages such as \code{parallel_eigen}
#' for the eigendecompositions and, in non-Gaussian Path 3, \code{parallel_unconstrained}
#' for the partition-wise unconstrained fits; nearby stages can likewise use
#' \code{parallel_penalty} and \code{parallel_make_constraint}. The eigenvalue
#' decomposition and matrix square roots of each \eqn{\mathbf{G}_k} are
#' computed by \code{\link{compute_G_eigen}}, and can be returned in the fitted
#' object as \code{G} and \code{Ghalf} when \code{return_G = TRUE} and
#' \code{return_Ghalf = TRUE}.
#'
#' Fitted values for the canonical Gaussian case appear as
#' \eqn{\tilde{\mathbf{y}} = \mathbf{X}\tilde{\boldsymbol{\beta}} = \mathbf{H}\mathbf{y}}
#' for \eqn{\mathbf{H} = \mathbf{X}\mathbf{U}\mathbf{G}\mathbf{X}^{\top}}.
#' }
#'
#' \subsection{Smoothing Constraints and the Constraint Matrix}{
#' Without further intervention the piecewise polynomial will be discontinuous.
#' The central idea of LMSS is that smoothness is not hidden inside a special
#' basis, but instead imposed directly where neighboring partitions meet.
#' At each knot \eqn{t_{k,k+1}} between neighboring partitions \eqn{k} and
#' \eqn{k+1}, up to three smoothing constraints are imposed:
#' \enumerate{
#'   \item Continuity: \eqn{\mathbf{x}_{k,k+1}^{\top}\boldsymbol{\beta}_k = \mathbf{x}_{k,k+1}^{\top}\boldsymbol{\beta}_{k+1}}.
#'   \item First-derivative continuity: \eqn{\mathbf{x}_{k,k+1}^{\prime\top}\boldsymbol{\beta}_k = \mathbf{x}_{k,k+1}^{\prime\top}\boldsymbol{\beta}_{k+1}}.
#'   \item Second-derivative continuity: \eqn{\mathbf{x}_{k,k+1}^{\prime\prime\top}\boldsymbol{\beta}_k = \mathbf{x}_{k,k+1}^{\prime\prime\top}\boldsymbol{\beta}_{k+1}}.
#' }
#' where \eqn{\mathbf{x}^{\prime}} and \eqn{\mathbf{x}^{\prime\prime}} are
#' elementwise first and second derivatives of the basis with respect to
#' \eqn{\mathbf{t}}. For the familiar cubic single-predictor basis
#' \eqn{\mathbf{x} = (1, t, t^2, t^3)^{\top}}, these derivative vectors are
#' \deqn{\mathbf{x}' = (0, 1, 2t, 3t^2)^{\top}, \qquad
#'   \mathbf{x}'' = (0, 0, 2, 6t)^{\top}.}
#' With \eqn{K} knots this yields up to \eqn{3K} scalar
#' constraints (for a single predictor; more for multiple predictors with
#' interactions), collected as linear equations
#' \eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{0}}
#' in a \eqn{P \times r} matrix \eqn{\mathbf{A}}. The constraint matrix is
#' built by \code{\link{make_constraint_matrix}} and returned in the fitted
#' object as \code{A}.
#'
#' In higher dimensions or with many partitions, the constraints can become
#' over-specified and force the model toward a single global polynomial. In these
#' cases it is recommended to drop second-derivative constraints or include quartic
#' terms, allowing the model to fit a richer surface while maintaining
#' perceived smoothness at knots. The appropriate constraint level can be
#' controlled via \code{include_constrain_fitted},
#' \code{include_constrain_first_deriv}, and
#' \code{include_constrain_second_deriv}.
#' The companion flag \code{include_constrain_interactions} determines whether
#' the analogous mixed-partial constraints are imposed for interaction terms,
#' and \code{no_intercept} adds the special homogeneous equality constraint that
#' fixes the intercept at zero (the same behavior triggered by using
#' \code{0 +} in the formula interface).
#'
#' Before computing the projection \eqn{\mathbf{U}}, the constraint matrix is
#' reduced to a linearly independent subset of columns via pivoted QR
#' decomposition. This avoids numerical instability from redundant constraints
#' and ensures \eqn{\mathbf{A}^{\top}\mathbf{G}\mathbf{A}} is invertible.
#' }
#'
#' \subsection{Lagrangian Projection}{
#' The constrained estimate is derived via Lagrangian multipliers. Define the
#' \eqn{P \times P} projection matrix:
#' \deqn{\mathbf{U} = \mathbf{I} - \mathbf{G}\mathbf{A}(\mathbf{A}^{\top}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{\top}.}
#' Then the constrained estimate is:
#' \deqn{\tilde{\boldsymbol{\beta}} = \mathbf{U}\hat{\boldsymbol{\beta}}.}
#' The matrix \eqn{\mathbf{U}} has the property that
#' \eqn{\mathbf{U}\mathbf{G}\mathbf{U}^{\top} = \mathbf{U}\mathbf{G}}, which is
#' used extensively in variance estimation and posterior draws. In words, the
#' unconstrained penalized estimate is projected back into the coefficient space
#' that satisfies the smoothness restrictions, and all subsequent uncertainty
#' calculations inherit that same projected geometry.
#' The projection is computed via \code{\link{get_U}} and, when requested,
#' returned in the fitted object as \code{U} through \code{return_U = TRUE}.
#'
#' When the constraints are inhomogeneous
#' (\eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{c}} with
#' \eqn{\mathbf{c} \neq \mathbf{0}}), a particular solution
#' \eqn{\boldsymbol{\beta}_0} satisfying
#' \eqn{\mathbf{A}^{\top}\boldsymbol{\beta}_0 = \mathbf{c}} is added back
#' after projection, yielding the full Lagrangian solution
#' \eqn{\mathbf{U}\hat{\boldsymbol{\beta}} + (\mathbf{I} - \mathbf{U})\boldsymbol{\beta}_0}.
#' In \code{\link{lgspline}} and \code{\link{lgspline.fit}}, users realize
#' this by supplying extra equality columns in \code{constraint_vectors}
#' together with matching right-hand sides in \code{constraint_values};
#' \code{null_constraint} provides the alternate shorthand documented in
#' \code{\link{lgspline}} when \code{constraint_vectors} is supplied and
#' \code{constraint_values} is left empty.
#'
#' In practice \eqn{\mathbf{U}} is never explicitly formed during fitting.
#' The constrained estimate is obtained from a transformed OLS residual problem (the
#' \eqn{\mathbf{G}^{1/2}\mathbf{r}^{*}} trick) in four steps:
#' \enumerate{
#'   \item Obtain the unconstrained partition-wise
#'     unconstrained estimate \eqn{\hat{\boldsymbol{\beta}}}.
#'   \item Set \eqn{\mathbf{y}^{*} = \mathbf{G}^{-1/2}\hat{\boldsymbol{\beta}}} and
#'     \eqn{\mathbf{X}^{*} = \mathbf{G}^{1/2}\mathbf{A}}.
#'   \item Fit the linear model \eqn{\mathbb{E}[\mathbf{y}^{*}] = \mathbf{X}^{*}\boldsymbol{\gamma}}
#'     by OLS using QR decomposition.
#'   \item Compute the residuals \eqn{\mathbf{r}^{*} = \mathbf{y}^{*} - \mathbf{X}^{*}(\mathbf{X}^{*\top}\mathbf{X}^{*})^{-1}\mathbf{X}^{*\top}\mathbf{y}^{*}} from that transformed OLS
#'     fit and recover the constrained estimate by
#'     \eqn{\tilde{\boldsymbol{\beta}} = \mathbf{G}^{1/2}\mathbf{r}^{*}}.
#' }
#'
#' A scaling factor \eqn{1/\sqrt{K+1}} is applied to both
#' \eqn{\mathbf{X}^{*}} and \eqn{\mathbf{y}^{*}} prior to the OLS call
#' and divided out afterward, improving numerical conditioning when the
#' constraint matrix has many rows.
#'
#' The most expensive operation in this approach is the QR decomposition of the
#' \eqn{P \times r} matrix \eqn{\mathbf{X}^{*} = \mathbf{G}^{1/2}\mathbf{A}},
#' which is far cheaper than working with the full \eqn{P \times P} system directly.
#' Without correlation or SQP constraints, \eqn{\mathbf{G}} is stored and operated upon as a
#' list of \eqn{K+1} small \eqn{p \times p} matrices rather than the full
#' \eqn{P \times P} block-diagonal, saving substantial memory when \eqn{K}
#' is large and allowing for parallelism.
#'
#' When correlation is present, \eqn{\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X}}
#' is no longer block-diagonal, so the full-dimensional system must be handled
#' directly unless the Woodbury acceleration
#' (see \code{\link{.woodbury_decompose_V}}) applies.
#' When additional inequality constraints are present, the code either augments the equality system with a
#' partition-wise active-set refinement (block-separable case) or falls back to
#' dense SQP via the Goldfarb-Idnani dual active-set method implemented in
#' \code{\link[quadprog]{solve.QP}}.
#' }
#'
#' @section GLM Extension and Iterative Updates:
#'
#' \subsection{Working Quantities}{
#' For GLM responses with mean \eqn{\mu_i = g^{-1}(\eta_i)} and working
#' weights \eqn{w_i = [V(\mu_i)\{g'(\mu_i)\}^2]^{-1}}, the penalized
#' information becomes
#' \eqn{\mathbf{G}_k^{-1} = \mathbf{X}_k^{\top}\mathbf{W}_k\mathbf{X}_k
#' + \boldsymbol{\Lambda}_k} with
#' \eqn{\mathbf{W}_k = \mathrm{diag}(w_i : i \in \mathcal{P}_k)}.
#' Because \eqn{\mathbf{W}} is diagonal,
#' \eqn{\mathbf{X}^{\top}\mathbf{W}\mathbf{X}} remains block-diagonal and
#' the four-step procedure carries through with
#' \eqn{\mathbf{G}_k = (\mathbf{X}_k^{\top}\mathbf{W}_k\mathbf{X}_k +
#' \boldsymbol{\Lambda}_k)^{-1}} in place of
#' \eqn{(\mathbf{X}_k^{\top}\mathbf{X}_k + \boldsymbol{\Lambda}_k)^{-1}}.
#' Under the default \code{glm_weight_function}, the diagonal entries reduce
#' to \code{family$variance(mu)} (optionally scaled by observation weights),
#' which for canonical families is the usual Fisher scoring weight.
#'
#' The partition-wise unconstrained estimates are obtained by
#' \code{unconstrained_fit_fxn}, by default
#' \code{\link{unconstrained_fit_default}}, which initializes via an augmented
#' ridge trick (appending \eqn{\boldsymbol{\Lambda}^{1/2}} as
#' pseudo-observations to \code{glm.fit}) then refines using
#' \code{\link{damped_newton_r}} with \code{\link{nr_iterate}}. For
#' non-canonical log-link gamma regression, the \code{keep_weighted_Lambda = TRUE}
#' option correctly returns the augmented ridge estimate directly.
#' }
#'
#' \subsection{Three Fitting Paths}{
#'
#' \strong{Path 1: Correlation structure present.}
#' When a working correlation \eqn{\mathbf{V}} is present, such as for
#' marginal means models or generalized estimating equation (GEE)-like models,
#' \eqn{\mathbf{V}^{-1}} couples observations across partitions, so
#' partition-wise fitting is not directly available. Two sub-paths handle
#' this.
#'
#' \emph{Path 1a} (Gaussian identity + GEE) solves the whitened system
#' directly via
#' \eqn{\tilde{\mathbf{G}} =
#' (\tilde{\mathbf{X}}^{\top}\tilde{\mathbf{X}} +
#' \boldsymbol{\Lambda}_{\mathrm{block}})^{-1}} where
#' \eqn{\tilde{\mathbf{X}} = \mathbf{V}^{-1/2}\mathbf{X}}, then applies
#' the Lagrangian projection in the full \eqn{P}-space.
#' See \code{\link{.get_B_gee_gaussian}}.
#'
#' \emph{Path 1b} (non-Gaussian GEE) uses a damped SQP iteration on the
#' whitened system. The first iterate is a constrained Newton step from the
#' projection matrix \eqn{\mathbf{U}}; subsequent iterates solve the
#' quadratic subproblem via \code{\link[quadprog]{solve.QP}}.
#' See \code{\link{.get_B_gee_glm}}.
#'
#' Both sub-paths have Woodbury-accelerated variants described below.
#'
#' \strong{Path 2: Gaussian identity, no correlation.}
#' The constrained estimate is obtained by a single Lagrangian projection
#' from the per-partition penalized least-squares cross-products (the four
#' steps in closed form). No outer iteration is needed. When \eqn{K = 0} and
#' there are no additional constraints, this reduces to the ordinary
#' penalized closed form
#' \eqn{\hat{\boldsymbol{\beta}} = \mathbf{G}\mathbf{X}^{\top}\mathbf{y}}.
#' Implemented in \code{\link{.get_B_gaussian_nocorr}}.
#'
#' \strong{Path 3: Non-Gaussian GLM, no correlation.}
#' Unconstrained estimates are obtained separately within each partition,
#' then projected onto the constraint space. When the link is non-identity,
#' the information \eqn{\mathbf{G}} depends on the current fitted values
#' through the working weights, so the projection must be iterated. The
#' unconstrained anchor \eqn{\hat{\boldsymbol{\beta}}} is held fixed while
#' \eqn{\mathbf{G}} is recomputed at each iterate's constrained estimate:
#' \deqn{\tilde{\boldsymbol{\beta}}^{(s+1)} =
#'   \mathbf{U}^{(s)}\hat{\boldsymbol{\beta}}, \quad
#'   \mathbf{U}^{(s)} = \mathbf{I} - \mathbf{G}^{(s)}\mathbf{A}
#'   (\mathbf{A}^{\top}\mathbf{G}^{(s)}\mathbf{A})^{-1}\mathbf{A}^{\top}.}
#' Iteration stops when the mean absolute coefficient change falls below
#' \code{tol}, or when the update begins increasing (in which case the
#' previous iterate is restored). The recomputation of weighted Gram
#' matrices, Schur corrections, and square-root information factors at each
#' step is handled by \code{.solver_recompute_G_at_estimate}. Implemented
#' in \code{\link{.get_B_glm_nocorr}}.
#' }
#'
#' \subsection{Woodbury Acceleration for Structured Correlation}{
#' For structured correlation, write the penalized information as
#' \deqn{\mathbf{G}^{-1}
#'   = \mathbf{G}_{\mathrm{on}}^{-1} + \mathbf{G}_{\mathrm{off}}^{-1},}
#' where \eqn{\mathbf{G}_{\mathrm{on}}^{-1}} contains the \eqn{K+1}
#' block-diagonal \eqn{p \times p} pieces of
#' \eqn{\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X} + \boldsymbol{\Lambda}},
#' and \eqn{\mathbf{G}_{\mathrm{off}}^{-1}} contains the cross-partition part.
#' Without correlation, \eqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{0}}.
#'
#' When the cross-partition part is low rank, the code factors it as
#' \deqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{E}\mathbf{J}\mathbf{E}^{\top},}
#' with \eqn{\mathbf{E}} of size \eqn{P \times r} and \eqn{\mathbf{J}}
#' diagonal with entries \eqn{\pm 1}. Then
#' \deqn{\mathbf{G}
#'   = \mathbf{G}_{\mathrm{on}}
#'   - \mathbf{G}_{\mathrm{on}}\mathbf{E}
#'   (\mathbf{J}^{-1} + \mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}\mathbf{E})^{-1}
#'   \mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}.}
#' The only dense inverse is the small \eqn{r \times r} matrix
#' \eqn{(\mathbf{J}^{-1} + \mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}\mathbf{E})^{-1}}.
#'
#' Define \eqn{\mathbf{N} = \mathbf{G}_{\mathrm{on}}^{1/2}\mathbf{E}} and a
#' thin QR decomposition \eqn{\mathbf{N} = \mathbf{Q}\mathbf{R}}. With
#' \eqn{\mathbf{P} = \mathbf{Q}\mathbf{Q}^{\top}} and
#' \eqn{\mathbf{S} = \mathbf{I}_r + \mathbf{R}\mathbf{J}\mathbf{R}^{\top}},
#' the supplement writes
#' \deqn{\mathbf{F}
#'   = (\mathbf{I}_P + \mathbf{N}\mathbf{J}\mathbf{N}^{\top})^{-1}
#'   = \mathbf{I}_P - \mathbf{P} + \mathbf{Q}\mathbf{S}^{-1}\mathbf{Q}^{\top},}
#' so that
#' \deqn{\mathbf{G}^{1/2}
#'   = \mathbf{G}_{\mathrm{on}}^{1/2}\mathbf{F}^{1/2}, \qquad
#'   \mathbf{G}^{-1/2}
#'   = \mathbf{F}^{-1/2}\mathbf{G}_{\mathrm{on}}^{-1/2},}
#' with
#' \deqn{\mathbf{F}^{1/2}
#'   = (\mathbf{I}_P - \mathbf{P}) +
#'   \mathbf{Q}\mathbf{S}^{-1/2}\mathbf{Q}^{\top}, \qquad
#'   \mathbf{F}^{-1/2}
#'   = (\mathbf{I}_P - \mathbf{P}) +
#'   \mathbf{Q}\mathbf{S}^{1/2}\mathbf{Q}^{\top}.}
#' This preserves the same transformed OLS projection used in the
#' independent case:
#' \deqn{\mathbf{y}^{*} = \mathbf{G}^{1/2}\mathbf{w}, \qquad
#'   \mathbf{X}^{*} = \mathbf{G}^{1/2}\mathbf{A}, \qquad
#'   \tilde{\boldsymbol{\beta}} = \mathbf{G}^{1/2}\mathbf{r}^{*},}
#' where \eqn{\mathbf{w} = \mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{y}}.
#' The full-space operations remain the QR on \eqn{\mathbf{X}^{*}} and a
#' rank-\eqn{r} correction through \eqn{\mathbf{Q}}.
#'
#' \strong{Implementation map.} The code uses the same algebra but stores a
#' numerically convenient representation. The correspondence is:
#' \itemize{
#'   \item manuscript \eqn{\mathbf{E}} \eqn{\leftrightarrow} helper output
#'     \code{E} (legacy alias \code{L});
#'   \item manuscript diagonal sign matrix \eqn{\mathbf{J}}
#'     \eqn{\leftrightarrow} helper output \code{J_signs}
#'     (legacy alias \code{S_signs});
#'   \item manuscript inverse
#'     \eqn{(\mathbf{J}^{-1} + \mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}\mathbf{E})^{-1}}
#'     \eqn{\leftrightarrow} helper output \code{inner_inv}
#'     (legacy alias \code{M});
#'   \item manuscript projector basis \eqn{\mathbf{Q}}
#'     \eqn{\leftrightarrow} internal basis \code{U_Q}
#'     (alias \code{Q_basis});
#'   \item manuscript projector \eqn{\mathbf{P} = \mathbf{Q}\mathbf{Q}^{\top}}
#'     is not stored explicitly, but is represented through rank-\eqn{r}
#'     multiplies involving \code{U_Q};
#'   \item manuscript factors \eqn{\mathbf{F}^{1/2}} and
#'     \eqn{\mathbf{F}^{-1/2}} are stored internally as
#'     \eqn{\mathbf{I} - \mathbf{U}_Q\mathbf{C}\mathbf{U}_Q^{\top}} and
#'     \eqn{\mathbf{I} + \mathbf{U}_Q\mathbf{C}_{\mathrm{inv}}\mathbf{U}_Q^{\top}},
#'     where \code{C} and \code{C_inv} are diagonal.
#' }
#' These are just storage choices; the notation matches the supplement.
#'
#' For the non-Gaussian Woodbury path
#' (\code{\link{.get_B_gee_glm_woodbury}}), the fixed correlation
#' perturbation \eqn{\mathbf{V}^{-1} - \mathbf{I}} and its product with
#' \eqn{\mathbf{X}} are precomputed once and held fixed
#' across Newton iterations. At each step, the weighted perturbation
#' \eqn{\mathbf{X}^{\top}\mathbf{W}(\mathbf{V}^{-1} -
#' \mathbf{I})\mathbf{X}} is formed using the
#' precomputed product, then split and re-factored via
#' \code{\link{.woodbury_redecompose_weighted}}.
#'
#' The Woodbury path is used when \eqn{r < P/3}; otherwise the code falls
#' back to the dense whitened solve. If \eqn{\mathbf{F}} is not positive
#' definite, the dense path is also used.
#' See \code{\link{.get_B_gee_woodbury}} (Gaussian) and
#' \code{\link{.get_B_gee_glm_woodbury}} (non-Gaussian).
#' }
#'
#' \subsection{Step Control}{
#' Different paths use different step-control strategies.
#'
#' In the GEE paths (Paths 1a and 1b), the coefficient update is damped:
#' \deqn{\boldsymbol{\beta}^{(s+1)} = (1 - \alpha_s)\boldsymbol{\beta}^{(s)}
#'   + \alpha_s\boldsymbol{\beta}_{\mathrm{cand}}^{(s)},}
#' with \eqn{\alpha_s = 2^{-d_s}} and \eqn{d_s} increased whenever the
#' candidate gives non-finite or larger deviance. The loop terminates after
#' 10 consecutive rejections or 100 total iterations, and exits early when
#' both coefficient change and deviance improvement fall below \code{tol}
#' after a burn-in period.
#'
#' In Path 3 (non-Gaussian, no correlation), the outer projection loop has
#' no line search. The algorithm recomputes \eqn{\mathbf{G}} at the current
#' constrained estimate and applies a fresh projection. If the mean absolute
#' coefficient change begins increasing, the previous iterate is restored
#' and the loop stops. The partition-wise unconstrained estimates themselves
#' are obtained by damped Newton-Raphson inside
#' \code{\link{unconstrained_fit_default}}, where
#' \code{\link{damped_newton_r}} computes a Newton direction once per
#' iteration and halves the step size until the penalized log-likelihood
#' improves.
#' }
#'
#' @section Accommodating Correlation Structures:
#'
#' \subsection{Parametric Correlation Structures}{
#' Suppose \eqn{\mathrm{Cov}(\mathbf{y}) = \sigma^2\,\mathbf{V}(\boldsymbol{\theta})}
#' for a known parametric family indexed by \eqn{\boldsymbol{\theta}} (e.g.,
#' AR(1) with \eqn{\theta = \rho}, Matern with
#' \eqn{\boldsymbol{\theta} = (\ell, \nu)}, exchangeable with
#' \eqn{\theta = \rho}). The penalized generalized least-squares problem
#' becomes
#' \deqn{\min_{\boldsymbol{\beta}}\;
#'   (\mathbf{y} - \mathbf{X}\boldsymbol{\beta})^{\top}\mathbf{V}^{-1}
#'   (\mathbf{y} - \mathbf{X}\boldsymbol{\beta})
#'   + \boldsymbol{\beta}^{\top}\boldsymbol{\Lambda}\boldsymbol{\beta}
#'   \quad \text{s.t.}\; \mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{0}.}
#' The correlation matrix \eqn{\mathbf{V}} is supplied through the
#' fitted-object components \code{Vhalf} and \code{VhalfInv}, either
#' directly or via user functions \code{Vhalf_fxn} and \code{VhalfInv_fxn}.
#' When both are non-\code{NULL}, \code{\link{get_B}} dispatches to the GEE
#' paths (Path 1a or 1b). For built-in correlation structures, the required
#' square-root matrices are assembled numerically using
#' \code{\link{matsqrt}}, \code{\link{matinvsqrt}}, and \code{\link{invert}}.
#' }
#'
#' \subsection{Whitening and Permutation}{
#' Because the data are stored in partition ordering (all observations from
#' partition 0, then partition 1, etc.) while \eqn{\mathbf{V}} is in the
#' original observation ordering, a permutation is applied internally:
#' \eqn{\mathbf{V}_{\mathrm{perm}}^{-1/2} =
#' \mathbf{V}^{-1/2}[\boldsymbol{\pi}, \boldsymbol{\pi}]}, where
#' \eqn{\boldsymbol{\pi} = \texttt{unlist(order\_list)}} maps original
#' indices to partition-ordered indices. The whitened design and response are
#' \eqn{\tilde{\mathbf{X}} =
#' \mathbf{V}_{\mathrm{perm}}^{-1/2}\mathbf{X}_{\mathrm{block}}} and
#' \eqn{\tilde{\mathbf{y}} =
#' \mathbf{V}_{\mathrm{perm}}^{-1/2}\mathbf{y}}.
#'
#' The \eqn{\mathbf{X}} and \eqn{\mathbf{y}} inputs to
#' \code{\link{lgspline.fit}} are preserved in their unwhitened form.
#' Whitening is applied inside \code{\link{get_B}} and
#' \code{\link{blockfit_solve}} where the full \eqn{N \times P}
#' block-diagonal design is available, since applying
#' \eqn{\mathbf{V}^{-1/2}} to only the diagonal blocks of the partitioned
#' design would silently discard cross-partition contributions and corrupt
#' the Gram matrix.
#' }
#'
#' \subsection{Loss of Block-Diagonal Structure}{
#' Unlike the independent-errors case,
#' \eqn{\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X}} is not block-diagonal
#' because \eqn{\mathbf{V}^{-1}} introduces cross-partition coupling. The
#' unconstrained estimator
#' \deqn{\hat{\boldsymbol{\beta}} =
#'   (\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X} +
#'   \boldsymbol{\Lambda})^{-1}
#'   \mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{y}}
#' requires a full \eqn{P \times P} solve, and the constraint projection
#' proceeds with
#' \eqn{\mathbf{G} = (\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X} +
#' \boldsymbol{\Lambda})^{-1}}.
#'
#' For structured correlation matrices (AR(1), exchangeable, banded), the
#' perturbation \eqn{\mathbf{V}^{-1} - \mathbf{I}} is sparse and the
#' cross-partition coupling has low effective rank. In these cases the
#' Woodbury-accelerated paths
#' (\code{\link{.get_B_gee_woodbury}},
#' \code{\link{.get_B_gee_glm_woodbury}}) recover the partition-wise
#' computational structure by decomposing the coupling into a
#' block-diagonal correction plus a low-rank remainder, as described in the
#' GLM Extension section. For dense or high-rank
#' \eqn{\mathbf{V}^{-1}}, the code falls back to the full whitened system
#' (\code{\link{.get_B_gee_gaussian}}, \code{\link{.get_B_gee_glm}}).
#' }
#'
#' \subsection{GEE Deviance Monitoring}{
#' For non-Gaussian models with correlation, the deviance used for
#' convergence monitoring is computed in the whitened space by
#' \code{.bf_gee_deviance}. When the family supplies
#' \code{custom_dev.resids}, the raw deviance residuals
#' \eqn{r_i = \mathrm{sign}(d_i)\sqrt{|d_i|}} are divided by
#' \eqn{\sqrt{w_i}} and pre=dmultiplied by
#' \eqn{\mathbf{V}_{\mathrm{perm}}^{-1/2}} before squaring and averaging:
#' \deqn{D_{\mathrm{GEE}} = \frac{1}{N}\left\|
#'   \mathbf{V}_{\mathrm{perm}}^{-1/2}\,
#'   \mathbf{w}^{-1/2}\mathbf{r}\right\|^{2},}
#' where \eqn{\mathbf{w}} is the vector of working weights at the current
#' iterate, clamped below at \eqn{\sqrt{\varepsilon_{\mathrm{mach}}}}. When
#' only \code{dev.resids} is available, the function falls back to the
#' standard mean deviance; otherwise it uses mean squared error.
#' }
#'
#' \subsection{REML Estimation of Correlation Parameters}{
#' Correlation parameters \eqn{\boldsymbol{\theta}} are estimated by minimizing
#' a negative restricted log-likelihood (REML) objective. The criterion
#' implemented in \pkg{lgspline} is a central-limit-theorem-based working
#' approximation to a Laplace-style marginal likelihood criterion,
#' applied here solely to correlation structure estimation rather than
#' penalty parameter selection.
#' Let \eqn{\mathbf{D} = \mathrm{diag}(d_i)} be the observation weight matrix,
#' \eqn{\tilde{\mathbf{W}} = \mathrm{diag}(\tilde{w}_i)} the GLM working weight
#' matrix at the current fitted values, \eqn{\mathbf{V}} the correlation matrix
#' parameterized by \eqn{\boldsymbol{\rho}} (a vector on the unconstrained
#' real line), and \eqn{\tilde{\sigma}^{2}} the dispersion profiled at its
#' restricted maximum likelihood estimate. The negative REML objective
#' implemented in \pkg{lgspline}, scaled by \eqn{1/N}, is
#' \deqn{-\ell_{R}(\boldsymbol{\rho}) = \frac{1}{N}\!\left[
#'   -\log|\mathbf{V}^{-1/2}|
#'   + \frac{N}{2}\log\tilde{\sigma}^{2}
#'   + \frac{1}{2\tilde{\sigma}^{2}}
#'     (\mathbf{y} - \boldsymbol{\mu})^{\top}
#'     \mathbf{D}\tilde{\mathbf{W}}^{-1}\mathbf{V}^{-1}
#'     (\mathbf{y} - \boldsymbol{\mu})
#'   + \frac{1}{2}\log|(\tilde{\sigma}^{2}\mathbf{U}\mathbf{G})^{-1}|^{+}
#' \right],}
#' where \eqn{|\cdot|^{+}} denotes the generalized determinant (product of
#' nonzero eigenvalues), and
#' \eqn{\boldsymbol{\mu} = g^{-1}(\mathbf{X}\tilde{\boldsymbol{\beta}})} are
#' the fitted values on the response scale.
#'
#' Gradients with respect to correlation parameters are available in
#' closed form for all built-in structures except Matern, which uses
#' finite-difference approximation due to the complexity of differentiating
#' the modified Bessel function \eqn{K_{\nu}} with respect to \eqn{\nu}. See
#' \code{\link{reml_grad_from_dV}} for the full gradient derivation and
#' notation. Custom analytic gradients can be supplied through \code{REML_grad},
#' and a fully custom criterion can replace REML through
#' \code{custom_VhalfInv_loss}. The Toeplitz example in \code{\link{lgspline}}
#' demonstrates how to supply custom correlation structures with user-defined
#' gradient functions. Optimization over these working correlation parameters is
#' then carried out by the same quasi-Newton engine used elsewhere in the
#' package, with a finite-difference fallback when needed. In the user-facing interface, this
#' machinery is activated through \code{correlation_structure},
#' \code{correlation_id}, \code{spacetime}, \code{VhalfInv}, \code{Vhalf},
#' \code{VhalfInv_fxn}, \code{Vhalf_fxn}, \code{VhalfInv_par_init},
#' \code{REML_grad}, \code{custom_VhalfInv_loss}, and
#' \code{VhalfInv_logdet}.
#'
#' The gradient of the negative REML has three terms per parameter:
#' \enumerate{
#'   \item \eqn{\frac{1}{2}\mathrm{tr}(\mathbf{V}^{-1}\partial\mathbf{V}/\partial\theta_j)}:
#'     the log-determinant contribution.
#'   \item \eqn{-\frac{1}{2\tilde{\sigma}^{2}}\mathbf{r}^{\top}
#'     (\partial\mathbf{V}/\partial\theta_j)\mathbf{r}}:
#'     the residual quadratic form contribution, where
#'     \eqn{\mathbf{r} = \mathrm{diag}(\sqrt{d_i}/\sqrt{\tilde{w}_i})
#'     \mathbf{V}^{-1/2}(\mathbf{y} - \boldsymbol{\mu})}.
#'   \item \eqn{-\frac{1}{2}\mathrm{tr}\!\left(\mathbf{M}^{+}\mathbf{X}_{*}^{\top}
#'     \mathbf{V}^{-1}(\partial\mathbf{V}/\partial\theta_j)\mathbf{V}^{-1}
#'     \tilde{\mathbf{W}}\mathbf{D}\mathbf{X}_{*}\right)}:
#'     the REML correction, where
#'     \eqn{\mathbf{X}_{*} = \mathbf{X}\mathbf{U}} is the constrained design and
#'     \eqn{\mathbf{M} = \mathbf{X}_{*}^{\top}\mathbf{V}^{-1}
#'     \tilde{\mathbf{W}}\mathbf{D}\mathbf{X}_{*}
#'     + \mathbf{U}^{\top}\boldsymbol{\Lambda}\mathbf{U}} is the projected
#'     penalized information.
#' }
#'
#' For each supported correlation family, the derivatives
#' \eqn{\partial\mathbf{V}/\partial\theta_j} are available in closed form,
#' enabling analytic gradient computation for use with the quasi-Newton
#' optimizer.
#' }
#'
#' \subsection{Connection to Standard Mixed Model REML}{
#' The quadratic penalty \eqn{\boldsymbol{\Lambda}} acts as the inverse prior
#' covariance of a Gaussian random effect on the spline coefficients, with
#' the smoothing parameter satisfying \eqn{\lambda = \tau^{2}/\sigma^{2}};
#' this is the same mixed model representation used by \pkg{mgcv}, and
#' Monte Carlo draws under the resulting Laplace-style approximation are
#' available via \code{\link{generate_posterior}}. For Gaussian responses
#' with identity link and no penalization, the implemented criterion
#' coincides exactly with classical REML. For non-Gaussian responses, the
#' criterion substitutes the Fisher information for the full penalized
#' log-likelihood Hessian, exploiting the CLT approximation that
#' \eqn{\tilde{\mathbf{W}}^{-1/2}(\mathbf{y} - \boldsymbol{\mu})} is
#' approximately Gaussian; this yields a method-of-moments style estimator
#' for the correlation and variance parameters that depends only on
#' mean-variance relationships through \eqn{\mathbf{W}}, and therefore
#' generalizes naturally to quasi-likelihood and other settings where a
#' fully specified log-likelihood is unavailable. The REML correction term
#' \eqn{\log|(\tilde{\sigma}^{2}\mathbf{U}\mathbf{G})^{-1}|^{+}} uses a
#' generalized log-determinant so that only nonzero eigenvalues contribute
#' when rank deficiency arises from smoothness constraints or
#' identifiability conditions in \eqn{\mathbf{A}}. When
#' \eqn{\boldsymbol{\Lambda}} is full rank, the criterion coincides with
#' the exact marginal likelihood from integrating out
#' \eqn{\boldsymbol{\beta}} under its Gaussian prior; when
#' \eqn{\boldsymbol{\Lambda}} is rank-deficient, unpenalized coefficients
#' are projected out in the REML sense while penalized coefficients are
#' integrated through their prior. During penalty tuning, the block-diagonal
#' approximation is retained for GCV criteria and gradients; since GCV is
#' rotation-invariant the practical effect on automatic selection is expected
#' to be negligible, though this is not confirmed and the tuned penalties can
#' always be overridden.
#' }
#'
#' \subsection{Built-In Correlation Structures}{
#' The package provides several built-in correlation structures for modeling
#' spatial and temporal dependence. These are specified via
#' \code{correlation_structure} with group membership in \code{correlation_id}
#' and spatial or temporal coordinates in \code{spacetime} (an \eqn{N}-row
#' matrix). Exchangeable correlation does not require \code{spacetime}.
#'
#' All positive scale parameters are estimated on the log scale,
#' with back-transform \eqn{\exp(\cdot)}. Parameters constrained to
#' \eqn{(0, 1)} use a double-exponential back-transform of the form
#' \eqn{\exp(-\exp(\eta))}, so optimization still occurs on the
#' unconstrained real line while the correlation remains bounded.
#'
#' \describe{
#'   \item{Exchangeable}{
#'     Aliases: \code{'exchangeable'}, \code{'cs'}, \code{'CS'},
#'     \code{'compoundsymmetric'}, \code{'compound-symmetric'}.
#'
#'     A constant correlation \eqn{\nu} between any two observations within
#'     the same cluster. Parameterization: \eqn{\nu = \exp(-\exp(\rho))},
#'     so \eqn{\nu \in (0, 1)}. Only positive within-cluster correlation is
#'     supported under this parameterization.
#'   }
#'   \item{Spatial Exponential}{
#'     Aliases: \code{'spatial-exponential'}, \code{'spatialexponential'},
#'     \code{'exp'}, \code{'exponential'}.
#'
#'     Correlation decays exponentially with distance:
#'     \eqn{\exp(-\omega d)} where \eqn{d} is Euclidean distance and
#'     \eqn{\omega > 0}. Parameterization: \eqn{\omega = \exp(\rho)}.
#'     Mathematically equivalent to the power correlation \eqn{\theta^{d}}
#'     with \eqn{\theta = e^{-\omega}}, but with better numerical properties
#'     during optimization.
#'   }
#'   \item{AR(1)}{
#'     Aliases: \code{'ar1'}, \code{'ar(1)'}, \code{'AR(1)'}, \code{'AR1'}.
#'
#'     Correlation depends on rank difference between observations:
#'     \eqn{\nu^{r}} where \eqn{r} is the rank difference within cluster.
#'     Parameterization: \eqn{\nu = \exp(-\exp(\rho))},
#'     so \eqn{\nu \in (0, 1)}. Only positive autocorrelation is supported.
#'   }
#'   \item{Gaussian / Squared Exponential}{
#'     Aliases: \code{'gaussian'}, \code{'rbf'}, \code{'squared-exponential'}.
#'
#'     Smooth decay with squared distance:
#'     \eqn{\exp(-d^{2}/(2\ell^{2}))} where \eqn{\ell} is the length scale.
#'     Parameterization: \eqn{\ell = \exp(\rho)}.
#'   }
#'   \item{Spherical}{
#'     Aliases: \code{'spherical'}, \code{'Spherical'}, \code{'cubic'},
#'     \code{'sphere'}.
#'
#'     Polynomial decay with a hard cutoff at range \eqn{r}:
#'     \eqn{1 - 1.5(d/r) + 0.5(d/r)^{3}} for \eqn{d \le r}, and \eqn{0}
#'     otherwise. Parameterization: \eqn{r = \exp(\rho)}.
#'   }
#'   \item{Matern}{
#'     Aliases: \code{'matern'}, \code{'Matern'}.
#'
#'     Flexible correlation with adjustable smoothness:
#'     \eqn{(2^{1-\nu}/\Gamma(\nu))(\sqrt{2\nu}\,d/\ell)^{\nu}K_{\nu}(\sqrt{2\nu}\,d/\ell)}.
#'     Two parameters: length scale \eqn{\ell = \exp(\rho_1)} and smoothness
#'     \eqn{\nu = \exp(\rho_2)}. No analytical gradient is available for
#'     \eqn{\nu} due to the difficulty of differentiating the modified Bessel
#'     function \eqn{K_{\nu}} with respect to its order, so finite differences
#'     are used; this makes Matern slower and potentially less stable than
#'     other structures.
#'   }
#'   \item{Gamma-Cosine}{
#'     Aliases: \code{'gamma-cosine'}, \code{'gammacosine'},
#'     \code{'GammaCosine'}.
#'
#'     Oscillatory dependence:
#'     \eqn{(d^{\alpha-1}e^{-\gamma d})/(\Gamma(\alpha)/\gamma^{\alpha})\cdot\cos(\omega d)}.
#'     Three parameters: shape \eqn{\alpha = \exp(\rho_1)}, rate
#'     \eqn{\gamma = \exp(\rho_2)}, frequency \eqn{\omega = \exp(\rho_3)}.
#'     Reduces to exponential when \eqn{\alpha = 1} and \eqn{\omega \approx 0}.
#'   }
#'   \item{Gaussian-Cosine}{
#'     Aliases: \code{'gaussian-cosine'}, \code{'gaussiancosine'},
#'     \code{'GaussianCosine'}.
#'
#'     Smooth oscillatory correlation:
#'     \eqn{\exp(-d^{2}/(2\ell^{2}))\cdot\cos(\omega d)}.
#'     Two parameters: length scale \eqn{\ell = \exp(\rho_1)} and frequency
#'     \eqn{\omega = \exp(\rho_2)}. Reduces to Gaussian when
#'     \eqn{\omega \approx 0}.
#'   }
#' }
#' }
#'
#' \subsection{Interpreting Estimated Correlation Parameters}{
#' Correlation parameters are estimated on transformed scales; they must be
#' back-transformed for interpretation. When \code{\link{confint.lgspline}} is called and the
#' inverse Hessian from BFGS is available, confidence intervals are returned
#' on the untransformed (working) scale and should be back-transformed as
#' described in the examples for \code{\link{lgspline}}.
#' }
#'
#' \subsection{Custom Correlation Structures}{
#' Custom correlation structures can be specified through:
#' \itemize{
#'   \item \code{VhalfInv_fxn}: Creates \eqn{\mathbf{V}^{-1/2}}.
#'   \item \code{Vhalf_fxn}: Creates \eqn{\mathbf{V}^{1/2}}. When omitted,
#'     the code computes it by explicit inversion of \code{VhalfInv}.
#'   \item \code{REML_grad}: Provides the analytical gradient of the REML
#'     objective.
#'   \item \code{VhalfInv_logdet}: Efficient log-determinant computation.
#'   \item \code{custom_VhalfInv_loss}: Replaces the REML objective entirely.
#' }
#' These functions enter \code{\link{lgspline}} through
#' \code{correlation_structure}, \code{VhalfInv_fxn}, \code{Vhalf_fxn},
#' \code{REML_grad}, and \code{custom_VhalfInv_loss}, and the fitted object
#' retains the resulting correlation machinery in components such as
#' \code{VhalfInv_fxn}, \code{Vhalf_fxn}, and
#' \code{VhalfInv_params_estimates}. When \code{VhalfInv} is supplied but
#' \code{Vhalf} is not, \code{Vhalf} is computed unconditionally as the inverse
#' of \code{VhalfInv} for all family/link combinations, since both
#' \code{\link{get_B}} and \code{\link{blockfit_solve}} require it for GEE
#' estimation.
#' }
#'
#' @section Variance and Dispersion Estimation:
#'
#' Once the constrained estimate has been obtained, the next questions are how
#' much flexibility the fitted model effectively used and how uncertainty should
#' be propagated through the same constrained geometry. The quantities in this
#' section are therefore all built on the projected information matrices from
#' the previous sections.
#'
#' \subsection{Effective Degrees of Freedom and Dispersion}{
#' The effective degrees of freedom is the trace of the hat matrix. In the
#' Gaussian identity case with observation weights and correlation, the fitted
#' linear operator is built from the dense GLS analogue
#' \eqn{\mathbf{G}_{\mathrm{correct}}} and can be written schematically as
#' \deqn{\mathbf{H} =
#'   \mathbf{V}^{-1/2}(\mathbf{W}\mathbf{D})^{1/2}
#'   \mathbf{X}\mathbf{U}\mathbf{G}_{\mathrm{correct}}\mathbf{X}^{\top}
#'   (\mathbf{W}\mathbf{D})^{1/2}\mathbf{V}^{-1/2},}
#' where for Gaussian identity \eqn{\mathbf{W} = \mathbf{I}}. In the
#' no-correlation Gaussian case this reduces to the familiar
#' \eqn{\mathbf{H} = \mathbf{X}\mathbf{U}\mathbf{G}\mathbf{X}^{\top}\mathbf{D}}.
#'
#' For Gaussian identity fits, the dispersion estimate is computed as a
#' weighted mean squared residual, optionally scaled by
#' \eqn{N/(N - \mathrm{tr}(\mathbf{H}))} when
#' \code{unbias_dispersion = TRUE}:
#' \deqn{\tilde{\sigma}^{2} =
#'   \frac{1}{N - \mathrm{tr}(\mathbf{H})}\|\mathbf{y} - \mathbf{\tilde{y}}_i \|^{2}.}
#'
#' More generally with weights, a correlation structure and non-linear link function:
#' \deqn{\tilde{\sigma}^{2} =
#'   \frac{1}{N - \mathrm{tr}(\mathbf{H})}\| \mathbf{V}^{-1/2}\mathbf{W}^{-1/2}\mathbf{D}^{1/2}(\mathbf{y} - \tilde{\mathbf{y}})\|^{2}.}
#'
#' This estimated dispersion is returned as \code{sigmasq_tilde}, and the
#' corresponding effective degrees of freedom trace is returned as
#' \code{trace_XUGX}. For non-Gaussian families, the fitting code delegates
#' dispersion estimation to \code{dispersion_function}; thus the package does
#' not assume a single closed-form Pearson-style formula outside the Gaussian
#' identity setting. The hat-matrix trace itself is assembled by
#' \code{\link{compute_trace_H}} in the dense correlation-aware case and by the
#' same blockwise products summarized by \code{trace_XUGX} in the simpler
#' no-correlation paths.
#' A concrete built-in non-Gaussian example is the Weibull AFT path, which pairs
#' \code{\link{weibull_family}} with \code{\link{weibull_dispersion_function}},
#' \code{\link{weibull_glm_weight_function}}, and
#' \code{\link{weibull_schur_correction}}.
#' Users who want these quantities available for downstream inference should
#' keep \code{estimate_dispersion = TRUE} and \code{return_varcovmat = TRUE}
#' (the defaults), since \code{\link{wald_univariate}},
#' \code{\link{confint.lgspline}}, and the prediction-standard-error path in
#' \code{\link{predict.lgspline}} all rely on the post-fit dispersion and
#' covariance components documented here.
#' }
#'
#' \subsection{Variance-Covariance Matrix}{
#' The variance-covariance matrix of \eqn{\tilde{\boldsymbol{\beta}}} is
#' estimated as:
#' \deqn{\mathrm{Var}(\tilde{\boldsymbol{\beta}}) = \tilde{\sigma}^{2}(\mathbf{U}\mathbf{G}^{1/2})(\mathbf{U}\mathbf{G}^{1/2})^{\top}}
#' using the outer-product form for numerical stability. The result is returned
#' as \code{varcovmat} when \code{return_varcovmat = TRUE}. The algebraically
#' equivalent expression \eqn{\tilde{\sigma}^{2}\mathbf{U}\mathbf{G}} is not
#' used because \eqn{\mathbf{G}} is only positive semi-definite when the
#' penalty matrix \eqn{\boldsymbol{\Lambda}} has zero eigenvalues (e.g., the
#' intercept and linear terms under the smoothing spline penalty when
#' \code{flat_ridge_penalty} = 0), which can introduce negative diagonal
#' entries in finite precision arithmetic. The outer-product form also
#' guarantees symmetry.
#'
#' This is the Bayesian posterior covariance, treating the penalty as a
#' Gaussian prior on the coefficients. When \code{exact_varcovmat = TRUE},
#' a frequentist correction is additionally computed:
#' \deqn{\mathrm{Var}_{\mathrm{exact}}(\tilde{\boldsymbol{\beta}}) =
#'   \tilde{\sigma^2}\mathbf{U}\mathbf{G}^{1/2}(\mathbf{X}^{\top}\mathbf{W}\mathbf{D}\mathbf{V}^{-1}\mathbf{X})\mathbf{G}^{1/2}\mathbf{U}^{\top} =
#'   \tilde{\sigma}^{2}\mathbf{U}\mathbf{G}\mathbf{U}^{\top}
#'   - \tilde{\sigma}^{2}\mathbf{U}\mathbf{G}\boldsymbol{\Lambda}\mathbf{G}\mathbf{U}^{\top}.}
#' The first term is the Bayesian posterior covariance; the second is a frequentist
#' correction such that for Gaussian identity link (with or without correlation),
#' this is the exact variance-covariance matrix of the constrained estimator.
#'
#' When a correlation structure is present (\code{VhalfInv} non-\code{NULL}),
#' the block-diagonal \eqn{\mathbf{G}} is replaced by the full weighted GLS
#' analogue
#' \deqn{\mathbf{G}_{\mathrm{correct}} =
#'   \left(\mathbf{X}^{\top}\mathbf{W}\mathbf{D}\mathbf{V}^{-1}\mathbf{X}
#'   + \boldsymbol{\Lambda}\right)^{-1},}
#' where \eqn{\mathbf{W} = \mathbf{I}} in the Gaussian identity case.
#' This dense matrix is what enters the correlation-aware \eqn{\mathbf{U}},
#' \eqn{\mathrm{Var}(\tilde{\boldsymbol{\beta}})}, and
#' \eqn{\mathrm{Var}_{\mathrm{exact}}(\tilde{\boldsymbol{\beta}})}
#' computations.
#'
#' In user-facing terms, \code{return_varcovmat} controls whether this matrix is
#' stored at all, while \code{exact_varcovmat} switches between the default
#' posterior/Laplace approximation and the exact frequentist correction in the
#' Gaussian-identity setting. The stored covariance is what powers
#' \code{\link{wald_univariate}}, \code{\link{confint.lgspline}}, and
#' \code{se.fit = TRUE} in \code{\link{predict.lgspline}}; the
#' \code{critical_value} argument supplied at fit time is carried forward as the
#' default cutoff for those interval-producing helpers.
#' }
#'
#' \subsection{Recomputation of G at Convergence}{
#' At the final iterate, \eqn{\mathbf{G}} is recomputed to reflect the
#' converged working weights and Schur corrections. The implementation
#' computes the weighted design
#' \eqn{\mathbf{X}_{w}^{(k)} = \mathbf{X}_k \cdot \mathrm{diag}(\sqrt{\mathbf{w}_k})},
#' forms the weighted Gram matrix
#' \eqn{\mathbf{X}_{w}^{(k)\top}\mathbf{X}_{w}^{(k)}}, adds the Schur
#' correction, and performs eigendecomposition via
#' \code{\link{compute_G_eigen}} to obtain \eqn{\mathbf{G}_k},
#' \eqn{\mathbf{G}_k^{1/2}}, and \eqn{\mathbf{G}_k^{-1/2}}. The relationship
#' \eqn{\mathbf{G}_k = \mathbf{G}_k^{1/2}(\mathbf{G}_k^{1/2})^{\top}} is
#' enforced exactly by construction, and the fitted object can retain these as
#' \code{G} and \code{Ghalf} when \code{return_G = TRUE} and
#' \code{return_Ghalf = TRUE}. The square-root factors are numerically
#' stabilized through the helper routines \code{\link{matsqrt}} and
#' \code{\link{matinvsqrt}}, which are also used elsewhere in the package when
#' dense analogues of \eqn{\mathbf{G}^{1/2}} or \eqn{\mathbf{G}^{-1/2}} are
#' required.
#' }
#'
#' @section Bayesian Interpretation:
#'
#' The penalty has a natural Gaussian-prior interpretation, so once the
#' constrained estimator and its covariance are available, Bayesian-style
#' posterior simulation follows almost immediately. This section records the
#' interpretation that is already implicit in the fitted object and in the
#' package's posterior simulation helpers.
#'
#' A Bayesian interpretation follows from viewing the penalty as a Gaussian
#' prior on the coefficients. Conditional on the fitted smoothing parameters,
#' the code samples on the coefficient scale from
#' \deqn{\boldsymbol{\beta}^{(m)} =
#'   \tilde{\boldsymbol{\beta}} +
#'   \sqrt{\tilde{\sigma}^{2}},\mathbf{U}\mathbf{G}^{1/2}\mathbf{z}^{(m)},
#'   \qquad \mathbf{z}^{(m)} \sim \mathcal{N}(\mathbf{0}, \mathbf{I}).}
#' The coefficients are then back-transformed to the original response and
#' predictor scales.
#'
#' When inequality constraints are absent, these draws are i.i.d. Gaussian
#' posterior draws around the fitted mode.
#' The underlying coefficient-draw closure also contains an elliptical slice
#' sampling route for active inequality constraints, using the same covariance
#' factor to keep retained draws in the feasible region.
#'
#' At the implementation level, standard Gaussian posterior draws may place
#' positive mass on the infeasible region, so the constrained-draw closure
#' instead targets the truncated posterior
#' \deqn{\pi(\boldsymbol{\beta} \mid \mathbf{y}) \propto
#'   \exp\!\left(-\frac{1}{2}(\boldsymbol{\beta} - \tilde{\boldsymbol{\beta}})^{\top}
#'   \mathbf{G}^{-1}(\boldsymbol{\beta} - \tilde{\boldsymbol{\beta}})\right)
#'   \mathbf{1}(\mathbf{C}^{\top}\boldsymbol{\beta} \succeq \mathbf{c}),}
#' yielding credible intervals that respect the constraint boundaries.
#' The public \code{\link{generate_posterior}} wrapper forwards
#' \code{enforce_qp_constraints} to the stored constrained-draw closure, so
#' constrained draws can be requested directly from the user-facing interface.
#' When a working correlation structure is present, the companion helper
#' \code{\link{generate_posterior_correlation}} extends this idea by propagating
#' uncertainty in the fitted correlation parameters through the same
#' \code{VhalfInv_fxn}/\code{Vhalf_fxn} machinery described in the correlation
#' section, rather than conditioning only on fixed covariance parameters. Correlation
#' parameters are drawn from a multivariate normal distribution centered about their estimates
#' with the inverse approxiamte BFGS Hessian of the REML optimization problem
#' \code{VhalfInv_params_vcov} used by default (or a custom alternative as
#' supplied to the argument \code{correlation_param_vcov_sc}).
#'
#'
#' @section Inequality Constraints via Sequential Quadratic Programming:
#'
#' \subsection{Overview}{
#' Inequality constraints of the form
#' \eqn{\mathbf{C}^{\top}\boldsymbol{\beta} \succeq \mathbf{c}} handle
#' shape restrictions such as monotonicity, convexity, or boundedness. In
#' the monomial basis, these are linear in \eqn{\boldsymbol{\beta}}.
#' Monotonicity at a grid of points requires the first derivative polynomial
#' to be non-negative there; convexity requires the second derivative to be
#' non-negative; range constraints bound the fitted values directly. All of
#' these translate to linear inequality constraints on the coefficient
#' vector.
#'
#' The inequality pieces are assembled by \code{\link{process_qp}}, which
#' returns the \code{qp_Amat}, \code{qp_bvec}, and \code{qp_meq} objects
#' passed to \code{\link[quadprog]{solve.QP}}, together with a
#' \code{quadprog} flag indicating whether any inequality constraints are
#' active. For derivative-sign constraints, \code{\link{process_qp}} calls
#' \code{.build_deriv_qp}, which uses
#' \code{\link{make_derivative_matrix}} on the expansion-standardized
#' design and maps the derivative rows into the full \eqn{P}-dimensional
#' coefficient space partition by partition.
#' }
#'
#' \subsection{Partition-Wise Active-Set Method}{
#' The sparsity pattern of \eqn{\mathbf{C}} is inspected automatically by
#' \code{\link{.detect_qp_global}} (equivalently
#' \code{.solver_detect_qp_global}). When every column of
#' \eqn{\mathbf{C}} has nonzero entries in only a single partition block,
#' the constraint system is block-separable and an active-set method can
#' replace the dense QP.
#'
#' At each iteration, the active set \eqn{\mathcal{A}} (constraints
#' satisfied at equality) is appended to \eqn{\mathbf{A}} as additional
#' equality constraints:
#' \deqn{\mathbf{A}_{\mathrm{aug}} = [\mathbf{A} \mid
#'   \mathbf{C}_{\mathcal{A}}].}
#' The constrained estimate is obtained by the same OLS projection with
#' \eqn{\mathbf{A}_{\mathrm{aug}}} in place of \eqn{\mathbf{A}}, and
#' since \eqn{\mathbf{A}_{\mathrm{aug}}} retains block-diagonal
#' compatibility, all operations remain partition-wise.
#'
#' The active set is updated by checking primal feasibility (adding the
#' most-violated inactive constraint) and dual feasibility (dropping the
#' active constraint with the most negative Lagrange multiplier) until the
#' KKT conditions are satisfied. Lagrange multipliers for active
#' inequalities are recovered from the OLS fit used in the Lagrangian
#' projection: the fitted coefficients on
#' \eqn{\mathbf{X}^* = \mathbf{G}^{1/2}\mathbf{A}_{\mathrm{aug}}} give
#' the multipliers up to sign. Implemented in
#' \code{\link{.active_set_refine}} and
#' \code{\link{.check_kkt_partitionwise}}.
#'
#' When any column of \eqn{\mathbf{C}} spans multiple partition blocks
#' (for example, cross-knot monotonicity constraints), the block-diagonal
#' structure is broken and the dense SQP approach is required. The
#' selection is made automatically. If the active-set method does not
#' converge within its iteration limit (default 50), the code falls back
#' to dense SQP as well.
#' }
#'
#' \subsection{Dense SQP Iteration}{
#' The dense SQP approach, implemented in \code{\link{.qp_refine}},
#' solves a sequence of quadratic subproblems approximating the penalized
#' log-likelihood. At each iteration \eqn{s}:
#' \enumerate{
#'   \item Compute the information matrix
#'     \eqn{\mathbf{M}^{(s)} =
#'     \mathbf{X}^{\top}\mathbf{W}^{(s)}\mathbf{X} +
#'     \boldsymbol{\Lambda}_{\mathrm{block}} + \mathbf{S}^{(s)}},
#'     where \eqn{\mathbf{S}^{(s)}} is the Schur complement correction.
#'   \item Compute the score vector via \code{qp_score_function}.
#'   \item Solve the QP with \code{\link[quadprog]{solve.QP}}, where the
#'     combined constraint matrix is
#'     \eqn{[\mathbf{A} \mid \mathbf{C}]} with the first \eqn{R} columns
#'     treated as equalities.
#'   \item Apply a damped update
#'     \eqn{\boldsymbol{\beta}^{(s+1)} =
#'     (1 - \alpha)\boldsymbol{\beta}^{(s)} +
#'     \alpha\boldsymbol{\beta}_{\mathrm{QP}}}, with
#'     \eqn{\alpha = 2^{-d}} and \eqn{d} incremented upon deviance
#'     increase.
#' }
#' A rescaling factor
#' \eqn{\mathrm{sc} = \sqrt{\mathrm{mean}(|\mathbf{M}^{(s)}|)}} is applied
#' to the Hessian and linear term before calling the QP solver. The
#' equality-constrained estimate from the Lagrangian projection serves as
#' a warm start. The \code{qp_score_function} defaults to the canonical GLM
#' score
#' \eqn{\mathbf{X}^{\top}(\mathbf{y} - \boldsymbol{\mu})}; for custom
#' models a different score can be supplied.
#' }
#'
#' \subsection{Active Set and Lagrange Multipliers}{
#' The active set at the solution identifies binding inequality constraints.
#' The Lagrange multipliers quantify the cost of each: a multiplier of zero
#' means the constraint is not binding. The implementation stores active
#' constraint indices, the corresponding submatrix of the constraint matrix,
#' and the multiplier vector in the \code{qp_info} list returned alongside
#' coefficient estimates, with components \code{lagrangian}, \code{iact},
#' and \code{Amat_active}. When
#' \code{return_lagrange_multipliers = TRUE}, the fitted object also stores
#' the final multiplier vector directly as \code{lagrange_multipliers}.
#'
#' The original assembled inequality data are retained in the fitted
#' object's \code{quadprog_list} component (containing \code{qp_Amat},
#' \code{qp_bvec}, and \code{qp_meq} from \code{\link{process_qp}}) so
#' the final active set can be interpreted relative to the full
#' specification. The final \eqn{\mathbf{U}} used in constructing the
#' posterior variance-covariance matrix is built from both the equality
#' constraints and the active inequality constraints at the solution.
#' }
#'
#' \subsection{Built-In Constraints}{
#' Built-in inequality constraints include:
#' \itemize{
#'   \item Monotonicity: \code{qp_monotonic_increase},
#'     \code{qp_monotonic_decrease}. Enforced by requiring consecutive
#'     fitted values to be non-decreasing (or non-increasing):
#'     \eqn{(\mathbf{x}_i - \mathbf{x}_{i-1})^{\top}\boldsymbol{\beta}
#'     \geq 0}. These are constructed by \code{\link{process_qp}} from
#'     the partition-stacked block design reordered to observation order.
#'   \item Derivative sign: \code{qp_positive_derivative},
#'     \code{qp_negative_derivative}. Enforced through the
#'     first-derivative design matrix from
#'     \code{\link{make_derivative_matrix}}. May be \code{TRUE}/\code{FALSE}
#'     or a character/integer vector selecting specific predictors.
#'   \item Second-derivative sign: \code{qp_positive_2ndderivative},
#'     \code{qp_negative_2ndderivative}. Same construction using the
#'     second-derivative design matrix.
#'   \item Response range: \code{qp_range_lower}, \code{qp_range_upper}.
#'     Bounds on the linear predictor; for non-identity links, the bounds
#'     are transformed to the link scale inside \code{\link{process_qp}}.
#'   \item Custom constraints via \code{qp_Amat_fxn}, \code{qp_bvec_fxn},
#'     and \code{qp_meq_fxn}, which receive the design matrix structure
#'     and return the constraint matrix, bound vector, and number of
#'     equalities. These are commonly paired with a custom
#'     \code{qp_score_function} when the quadratic approximation uses a
#'     non-default likelihood. The low-level objects \code{qp_Amat},
#'     \code{qp_bvec}, and \code{qp_meq} remain documented in
#'     \code{\link{lgspline}} but in the current implementation serve as
#'     activation markers rather than being merged into the constraint set
#'     assembled by \code{\link{process_qp}}.
#' }
#' All constraints can be thinned to a user-specified subset of rows via
#' \code{qp_observations}, which \code{\link{process_qp}} applies before
#' assembly.
#' }
#'
#' @section Blockfit Backfitting for Linear Non-Interactive Effects:
#'
#' \subsection{Motivation}{
#' When a model contains both spline terms (receiving \eqn{K+1}
#' partition-specific coefficient vectors constrained to smoothness) and
#' non-interactive linear terms (``flat'' terms, specified via
#' \code{just_linear_without_interactions}, receiving a single shared
#' coefficient vector \eqn{\mathbf{v}} across all partitions), the standard
#' solver carries \eqn{K+1} copies of \eqn{\mathbf{v}} linked by equality
#' constraints. Backfitting avoids this inflation by solving a
#' lower-dimensional problem at each step. Write the partition-\eqn{k}
#' design as
#' \eqn{\mathbf{X}_k = [\mathbf{Z}_k \mid
#' \mathbf{X}_{\mathrm{flat}}^{(k)}]}, where \eqn{\mathbf{Z}_k} contains
#' the spline columns and \eqn{\mathbf{X}_{\mathrm{flat}}^{(k)}} the flat
#' columns. This is invoked when \code{blockfit = TRUE}, flat columns are
#' non-empty, and \eqn{K > 0}.
#'
#' The design, penalty, and constraint matrices are split into spline and
#' flat components by \code{\link{.bf_split_components}}, which extracts
#' spline rows from \eqn{\mathbf{A}}, drops null columns, rank-reduces via
#' QR, and detects mixed constraints (columns of \eqn{\mathbf{A}} with
#' nonzero entries on both spline and flat rows).
#' }
#'
#' \subsection{Block-Coordinate Descent}{
#' \strong{Spline step.} Holding \eqn{\mathbf{v}} fixed, the code forms
#' the adjusted response
#' \eqn{\mathbf{y}_k - \mathbf{X}_{\mathrm{flat}}^{(k)}\mathbf{v}} and
#' applies the spline-only Lagrangian projection via
#' \code{\link{.bf_lagrangian_project}}:
#' \deqn{\tilde{\boldsymbol{\beta}}_{\mathrm{spline}}^{(k)} =
#'   \mathbf{U}_{\mathrm{spline}}\mathbf{G}_{\mathrm{spline}}
#'   \mathbf{Z}_k^{\top}(\mathbf{y}_k -
#'   \mathbf{X}_{\mathrm{flat}}^{(k)}\mathbf{v}).}
#'
#' \strong{Flat step.} Holding spline coefficients fixed, the shared flat
#' vector is updated by pooled penalized regression on residuals:
#' \deqn{\mathbf{v} = \left(\sum_{k=0}^{K}
#'   \mathbf{X}_{\mathrm{flat}}^{(k)\top}
#'   \mathbf{X}_{\mathrm{flat}}^{(k)} +
#'   \boldsymbol{\Lambda}_{\mathrm{flat}}\right)^{-1}
#'   \sum_{k=0}^{K}
#'   \mathbf{X}_{\mathrm{flat}}^{(k)\top}
#'   (\mathbf{y}_k - \mathbf{Z}_k
#'   \tilde{\boldsymbol{\beta}}_{\mathrm{spline}}^{(k)}).}
#' When the constraint matrix \eqn{\mathbf{A}} has mixed columns (nonzero
#' entries on both spline and flat rows), the flat update instead solves a
#' KKT system enforcing the residual equality constraint
#' \eqn{\mathbf{A}_{\mathrm{flat}}^{\top}\mathbf{v} = \mathbf{c} -
#' \mathbf{A}_{\mathrm{spline}}^{\top}
#' \tilde{\boldsymbol{\beta}}_{\mathrm{spline}}}
#' via \code{\link{.bf_constrained_flat_update}}.
#'
#' Convergence is checked using the maximum absolute change across both
#' spline and flat coefficients. In the weighted inner loop used by the GLM
#' solvers, the flat-block change alone determines stopping.
#' }
#'
#' \subsection{Four Estimation Cases}{
#' \code{\link{blockfit_solve}} dispatches to one of four paths.
#'
#' \strong{Case (a): Gaussian identity + GEE}
#' (\code{\link{.bf_case_gauss_gee}}).
#' Whitening destroys block-diagonal structure, so the code skips
#' backfitting and performs the same full-system Gaussian GEE projection
#' used by \code{\link{get_B}} Path 1a. The result is split back into
#' spline and flat components for downstream assembly.
#'
#' \strong{Case (b): Gaussian identity, no correlation}
#' (\code{\link{.bf_case_gauss_no_corr}}).
#' Standard block-coordinate descent as described above. The spline-only
#' \eqn{\mathbf{G}_{\mathrm{spline}}} factors are precomputed once, and the
#' pooled flat penalized inverse is reused across iterations.
#'
#' \strong{Case (c): GLM + GEE}
#' (\code{\link{.bf_case_glm_gee}}).
#' Two stages. Stage 1 forms a warm start by running damped Newton-Raphson
#' on the unwhitened working response: each outer iteration computes
#' working responses and weights at the current linear predictor, then
#' calls \code{\link{.bf_inner_weighted}} for the inner backfitting loop.
#' Stage 2 refines the warm start on the full whitened system using the
#' damped SQP loop (\code{\link{.bf_sqp_loop}}), replicating the approach
#' used by \code{\link{.get_B_gee_glm}}.
#'
#' \strong{Case (d): GLM without GEE}
#' (\code{\link{.bf_case_glm_no_corr}}).
#' A damped Newton-Raphson outer loop updates working responses and weights,
#' while each inner iteration alternates between weighted spline and
#' weighted flat updates via \code{\link{.bf_inner_weighted}}. Deviance is
#' monitored across outer iterations for convergence and damping.
#' }
#'
#' \subsection{Inequality Constraints and Reassembly}{
#' Because flat coefficients are shared by construction, the corresponding
#' equality constraints are satisfied exactly. Smoothness constraints on the
#' spline block are enforced by the spline-only Lagrangian projection.
#' After backfitting convergence, inequality constraints are enforced via
#' the same partition-wise active-set or dense SQP refinement used by
#' \code{\link{get_B}}. The method is selected automatically by
#' \code{.solver_detect_qp_global}: block-separable constraints use the
#' active-set method through \code{\link{.active_set_refine}}, while
#' cross-partition constraints trigger the dense SQP loop through
#' \code{\link{.bf_sqp_loop}}. For GEE (Case c), inequality handling
#' occurs inside Stage 2 on the whitened system.
#'
#' After convergence, the shared flat vector \eqn{\mathbf{v}} is copied
#' into each partition's coefficient vector, yielding
#' \eqn{\boldsymbol{\beta}_k =
#' [\tilde{\boldsymbol{\beta}}_{\mathrm{spline}}^{(k)\top},
#' \mathbf{v}^{\top}]^{\top}} for compatibility with downstream inference.
#' If \code{\link{blockfit_solve}} throws an error, a warning is issued
#' and the code falls back to \code{\link{get_B}}.
#' }
#'
#' @section Knot Selection and Partitioning:
#'
#' The topic of knot selection is not the main focus of the package, but the
#' partition structure is central because every later design matrix, penalty,
#' and smoothness constraint depends on it. The defaults in \pkg{lgspline} are
#' therefore meant to be practical and transparent rather than theoretically
#' final.
#'
#' \subsection{Univariate Case}{
#' For a single predictor, the default partitioning is now handled by
#' \code{\link{make_partitions}} in the same \eqn{k}-means framework used more
#' generally: \eqn{K+1} centers are fit on an internally standardized copy
#' of the predictor, controlled by \code{standardize_predictors_for_knots}, and
#' then returned on the raw scale. Custom knots can still be supplied via
#' \code{custom_knots}, in which case partition assignment is built directly
#' from those raw-scale breakpoints. The default number of knots \eqn{K} is
#' chosen adaptively based on \eqn{N}, \eqn{p}, \eqn{q}, and the GLM family.
#' For multivariate fits, the resulting partition metadata are returned as
#' \code{make_partition_list} and can be re-used in later calls to
#' \code{\link{lgspline}}. This is particularly useful when one wants to hold
#' the partition geometry fixed across repeated fits, for example while varying
#' penalties, families, or correlation structures.
#' }
#' \subsection{Multivariate Case}{
#' For multiple predictors, \eqn{K+1} cluster centers are identified by
#' \eqn{k}-means on an internally standardized predictor matrix via
#' \code{\link{make_partitions}}. This is the partitioning mechanism used to
#' determine the multivariate spline regions; see MacQueen (1967) for the
#' classical clustering formulation and Kisi et al. (2025) for a recent applied
#' example of \eqn{k}-means-driven partitioning in a nonlinear prediction
#' setting. Midpoints between neighboring centers
#' (those whose midpoint  does not fall into a third cluster) serve as knot
#' locations. Observations are assigned to the nearest cluster center using
#' \code{\link[FNN]{get.knnx}}, and the returned centers and knots are on
#' the original predictor scale. The resulting partition structure is a type
#' of Voronoi diagram and is stored in the fitted object as
#' \code{make_partition_list}. The \code{do_not_cluster_on_these} argument can
#' exclude certain predictors from clustering (e.g., a treatment indicator that
#' should not drive partitioning). The lower-level clustering behavior can be
#' further controlled by \code{cluster_args} and \code{neighbor_tolerance},
#' while \code{cluster_on_indicators} determines whether binary predictors are
#' allowed to influence the partition geometry at all.
#' }
#'
#' \subsection{Standardizing Predictors}{
#' Higher-order polynomial terms can dramatically inflate or deflate the
#' magnitude of basis expansions, introducing numerical instability. All
#' polynomial basis expansions are scaled by
#' \eqn{q_{0.69} - q_{0.31}}, where \eqn{q_{\zeta}} is the \eqn{\zeta}-th
#' quantile of the expansion. For a standard normal distribution this quantity
#' is approximately 1, so the scaling is close to one standard deviation for
#' symmetric distributions. This fitting-stage rescaling is controlled by
#' \code{standardize_expansions_for_fitting}, while knot construction is
#' controlled separately by \code{standardize_predictors_for_knots}. The same
#' scaling is applied to the constraint matrix to maintain smoothness, and
#' coefficients are back-transformed to the original scale after fitting.
#' }
#'
#' @section Smoothing Spline Penalty:
#'
#' \subsection{Penalty Construction}{
#' The penalty matrix \eqn{\boldsymbol{\Lambda}_s} penalizes the integrated
#' squared total curvature of the fitted function over the observed predictor
#' ranges. This is the step that makes the piecewise polynomial fit genuinely
#' behave like a smoothing spline rather than merely a constrained regression
#' spline. The package computes this penalty directly from the monomial
#' structure of the basis rather than by appealing to a pre-tabulated spline
#' basis. For a single partition \eqn{k} with basis expansion
#' \eqn{\mathbf{x}_k = (\phi_1(\mathbf{t}), \ldots, \phi_p(\mathbf{t}))^{\top}}
#' where each \eqn{\phi_i(\mathbf{t}) = \prod_{j=1}^{q} t_j^{\alpha_{ij}}}
#' is a multivariate monomial:
#' \deqn{\boldsymbol{\beta}_k^{\top}\boldsymbol{\Lambda}_s\boldsymbol{\beta}_k
#'   = \int_{\mathbf{a}}^{\mathbf{b}}
#'     \|\tilde{f}_k''(\mathbf{t})\|^{2}\,d\mathbf{t},}
#' where \eqn{\mathbf{a}} and \eqn{\mathbf{b}} are the observed predictor
#' minimums and maximums (computed globally from the data, not
#' partition-specific), and
#' \eqn{\tilde{f}_k(\mathbf{t}) = \mathbf{x}_k^{\top}\boldsymbol{\beta}_k}
#' is the fitted function for partition \eqn{\mathcal{P}_k}.
#'
#' \strong{Total curvature operator.}
#' The integrated squared second derivative decomposes into \eqn{q}
#' curvature operators, one per predictor. For predictor \eqn{v}, the
#' curvature operator \eqn{D_v} is defined as
#' \deqn{D_v = \frac{\partial^{2}}{\partial t_v^{2}}
#'   + \sum_{s \neq v}\frac{\partial^{2}}{\partial t_v\,\partial t_s}.}
#' That is, \eqn{D_v} captures both the pure second derivative with respect
#' to \eqn{t_v} and all mixed second partial derivatives involving \eqn{t_v}.
#' The penalty matrix entries are then
#' \deqn{[\boldsymbol{\Lambda}_s]_{ij}
#'   = \sum_{v=1}^{q}\int_{\mathbf{a}}^{\mathbf{b}}
#'     D_v(\phi_i)\,D_v(\phi_j)\,d\mathbf{t}.}
#'
#' \strong{Monomial derivative rule.}
#' For a monomial \eqn{\phi(\mathbf{t}) = \prod_j t_j^{\alpha_j}}, the
#' derivatives entering \eqn{D_v} have closed forms. The pure second
#' derivative is
#' \deqn{\frac{\partial^{2}}{\partial t_v^{2}}\prod_j t_j^{\alpha_j}
#'   = \alpha_v(\alpha_v - 1)\,t_v^{\alpha_v - 2}\prod_{j \neq v} t_j^{\alpha_j},}
#' which is zero when \eqn{\alpha_v < 2}. The mixed second derivative is
#' \deqn{\frac{\partial^{2}}{\partial t_v\,\partial t_s}\prod_j t_j^{\alpha_j}
#'   = \alpha_v\alpha_s\,t_v^{\alpha_v - 1}t_s^{\alpha_s - 1}
#'   \prod_{j \neq v,s} t_j^{\alpha_j},}
#' which is zero when \eqn{\alpha_v < 1} or \eqn{\alpha_s < 1}. Applying
#' \eqn{D_v} to a monomial \eqn{\phi_i} produces a sum of monomials with
#' known coefficients and exponent vectors.
#'
#' \strong{Factorized integration.}
#' Because every \eqn{D_v(\phi_i)} is polynomial, the product
#' \eqn{D_v(\phi_i)\,D_v(\phi_j)} is also polynomial and the multivariate
#' integral factorizes over predictors:
#' \deqn{\int_{\mathbf{a}}^{\mathbf{b}}\prod_{j=1}^{q} t_j^{e_j}\,d\mathbf{t}
#'   = \prod_{j=1}^{q}\frac{b_j^{e_j+1} - a_j^{e_j+1}}{e_j + 1}.}
#' Crucially, this integral runs over \emph{all} \eqn{q} predictor ranges,
#' including predictors that do not appear in the integrand (for which
#' \eqn{e_j = 0} and the factor reduces to \eqn{b_j - a_j}). This ensures
#' that the penalty is properly scaled relative to the volume of the
#' predictor space.
#'
#' \strong{Single-predictor verification.}
#' For \eqn{q = 1} with expansion
#' \eqn{\mathbf{x} = (1, t, t^{2}, t^{3})^{\top}} on \eqn{[a, b]}, the
#' curvature operator reduces to \eqn{D_1 = \partial^{2}/\partial t^{2}} (no
#' mixed partials exist), and the penalty matrix reduces to
#' \deqn{\boldsymbol{\Lambda}_s
#'   = \int_a^b \mathbf{x}''\mathbf{x}''^{\top}\,dt
#'   = \begin{pmatrix}
#'       0 & 0 & 0 & 0 \\
#'       0 & 0 & 0 & 0 \\
#'       0 & 0 & 4(b - a) & 6(b^{2} - a^{2}) \\
#'       0 & 0 & 6(b^{2} - a^{2}) & 12(b^{3} - a^{3})
#'     \end{pmatrix},}
#' equivalent to the classical cubic smoothing spline penalty originally proposed by Reinsch.
#'
#' \strong{Handling of non-spline predictors.}
#' Predictors specified via \code{just_linear_without_interactions} or
#' \code{just_linear_with_interactions} do not receive higher-order
#' polynomial expansions in the design matrix. To ensure their curvature
#' contributions are still correctly computed (particularly through
#' interaction terms), the implementation temporarily appends phantom
#' higher-order columns (with zero data) for these predictors, computes
#' the full curvature penalty on the augmented basis, and then subsets the
#' result back to the original \eqn{p \times p} dimensions. This ensures
#' that interaction terms involving non-spline predictors receive appropriate
#' penalty contributions without affecting the rest of the estimation
#' pipeline.
#'
#' \strong{Parallel computation.}
#' Because the total penalty is an additive sum over predictors
#' (\eqn{\boldsymbol{\Lambda}_s = \sum_{v=1}^{q}\boldsymbol{\Lambda}_{s,v}}),
#' the computation can be parallelized by distributing the per-predictor
#' curvature matrices across workers via \code{parallel::parLapply} and
#' summing the results. This is controlled by the \code{parallel_penalty}
#' argument and is beneficial when \eqn{q} is large.
#'
#' The penalty is computed by \code{\link{get_2ndDerivPenalty}} (single
#' predictor or subset) and \code{\link{get_2ndDerivPenalty_wrapper}}
#' (full assembly with optional parallelism and non-spline handling).
#'
#' Because the smoothing penalty has zero eigenvalues for the intercept and
#' linear terms (whose second derivatives vanish), an optional ridge penalty on
#' lower-order terms is added for computational stability. The full penalty
#' block for partition \eqn{k} is:
#' \deqn{\boldsymbol{\Lambda}_k = \lambda_w\bigl(\boldsymbol{\Lambda}_s + \lambda_r\boldsymbol{\Lambda}_r + \sum_{m=1}^{M}\xi_{mk}\mathbf{L}_{mk}\bigr)}
#' where \eqn{\lambda_w} is the global wiggle penalty (\code{wiggle_penalty}),
#' \eqn{\lambda_r} is ridge penalty on linear and intercept terms
#' (\code{flat_ridge_penalty}) multiplied by the wiggle penalty, and
#' \eqn{\xi_{mk}} and \eqn{\mathbf{L}_{mk}} denote optional additional
#' penalty multipliers and matrices, including the predictor- and
#' partition-specific components activated through
#' \code{unique_penalty_per_predictor}, \code{unique_penalty_per_partition},
#' \code{predictor_penalties}, and \code{partition_penalties}. This assembly is handled by
#' \code{\link{compute_Lambda}}.
#'
#' The penalty matrix \eqn{\boldsymbol{\Lambda}} is stored as a list
#' of \eqn{K+1} \eqn{p \times p} square, symmetric, positive semi-definite
#' matrices.
#' }
#'
#' \subsection{Penalty Optimization via Leave-One-Out or Generalized Cross-Validation}{
#' After the structural pieces of the model are fixed, the main remaining
#' question is how much smoothing to apply. In \pkg{lgspline}, that tuning is
#' performed with exact leave-one-out by default, or with generalized
#' cross-validation when \code{tuning_criterion = "gcv"}. In either case, the
#' criterion is evaluated using the same constrained estimator that will be
#' used in the final model fit.
#' Penalty parameters are estimated on the log scale via exponential
#' parameterization (\eqn{\lambda = \exp(\theta)},
#' \eqn{\theta \in \mathbb{R}}), ensuring positivity. The chain rule factor
#' \eqn{\partial\exp(\theta)/\partial\theta = \exp(\theta) = \lambda} is
#' applied throughout. User-facing arguments (\code{initial_wiggle},
#' \code{initial_flat}, \code{predictor_penalties},
#' \code{partition_penalties}) accept values on the raw, natural scale;
#' conversion to log scale is handled internally. The final tuned values and
#' assembled penalty pieces are returned in the fitted object's
#' \code{penalties} component.
#'
#' The total penalty matrix \eqn{\boldsymbol{\Lambda}} is constructed as:
#' \deqn{\boldsymbol{\Lambda} = \lambda_w\mathbf{L}_{w} + \lambda_r\mathbf{L}_{r}
#'   + \sum_{j}\nu_j\mathbf{L}_j^{(\mathrm{pred})}
#'   + \sum_{k}\tau_k\mathbf{L}_k^{(\mathrm{part})},}
#' where \eqn{\mathbf{L}_{w}} is the integrated squared second-derivative
#' penalty (i.e., \eqn{\boldsymbol{\Lambda}_s} above), \eqn{\mathbf{L}_{r}} is
#' a ridge penalty on intercept and linear coefficients,
#' \eqn{\mathbf{L}_j^{(\mathrm{pred})}} are predictor-specific penalties, and
#' \eqn{\mathbf{L}_k^{(\mathrm{part})}} are partition-specific penalties. The
#' scalars \eqn{\lambda_w} (\code{wiggle_penalty}), \eqn{\lambda_r}
#' (\code{flat_ridge_penalty}), \eqn{\{\nu_j\}}
#' (\code{predictor_penalties}), and \eqn{\{\tau_k\}}
#' (\code{partition_penalties}) are tuned.
#' With \code{tuning_criterion = "loo"}, the criterion is
#' \deqn{\mathrm{LOO}
#'   = \frac{1}{N}\sum_{i=1}^{N}\left(\frac{r_i}{1-h_{ii}}\right)^2,}
#' where \eqn{h_{ii}} is the diagonal of the tuning hat matrix. For Gaussian
#' identity-link tuning with fixed transformed design, this is exact. With
#' correlation structure and/or non-identity links, \pkg{lgspline} applies the
#' same transformed or linearized tuning problem already used by GCV, so the
#' leave-one-out calculation is carried out on that working scale. In empirical
#' diagnostics, the criterion itself behaves stably, while the fully analytic
#' derivative of the observation-wise leverage term can be numerically delicate
#' in some problems; when that sensitivity matters in practice, the finite-
#' difference optimizer path (\code{use_custom_bfgs = FALSE}) remains
#' available. This exact observation-wise calculation is also the main reason
#' LOO tuning becomes more expensive than GCV as \eqn{N} grows; in routine use,
#' \code{tuning_criterion = "gcv"} is often the more practical choice once the
#' sample size is above about 250,000.
#'
#' With \code{tuning_criterion = "gcv"}, the package instead uses
#' \deqn{\mathrm{GCV}_{u,\gamma}
#'   = \frac{\sum_{i=1}^{N}D_{ii}\,r_i^{2}}
#'           {N(1 - \gamma\bar{W})^{2}}}
#' where \eqn{\bar{W} = \mathrm{tr}(\mathbf{H})/N} is the mean of the
#' hat-matrix diagonal. The symbol \eqn{\gamma} is set by
#' \code{gcv_gamma} and controls optional inflation of the
#' effective-degrees-of-freedom term.
#'
#' For identity link, \eqn{r_i = y_i - \hat{\eta}_i}. For non-identity links,
#' the residuals are
#' \eqn{r_i = g((y_i + \delta)/(1+2\delta)) - (\hat{\eta}_i + \delta)/(1+2\delta)},
#' where \eqn{\delta \geq 0} is a pseudocount that stabilizes the link
#' transformation, automatically tuned within \code{\link{tune_Lambda}} if not supplied
#' to \code{delta}.
#' Non-Gaussian families with observation weights
#' \eqn{\omega_i} have their residuals scaled by \eqn{\omega_i}. When the
#' family provides a custom deviance residual function, that function is used
#' in place of the link-scale residuals.
#'
#' \strong{Pseudocount selection.} The pseudocount \eqn{\delta} is chosen to
#' make the transformed response distribution most closely approximate a
#' \eqn{t}-distribution with \eqn{N-1} degrees of freedom, in the sense of
#' minimizing the (optionally weighted) mean absolute deviation between the
#' sorted standardized transformed responses and the corresponding
#' \eqn{t}-quantiles. This is solved via Brent's method over
#' \eqn{[10^{-64}, 1]}. When the link is identity, or when the response
#' naturally lies in the domain of the link function, \eqn{\delta = 0}.
#' This behavior is exposed through the \code{delta} argument in
#' \code{\link{lgspline}}: supplying a fixed numeric value bypasses the internal
#' search, while leaving it \code{NULL} allows the tuning code to choose the
#' stabilizing pseudocount automatically when needed.
#'
#' \strong{Meta-penalty regularization.} A regularization term pulls the
#' predictor- and partition-specific penalty parameters toward 1 on the raw
#' scale:
#' \deqn{P_{\mathrm{meta}}(\lambda_w, \nu_j, \tau_k)
#'   = \frac{1}{2}c_{\mathrm{meta}}\sum_j(\nu_j - 1)^{2}
#'   + \frac{1}{2}\cdot 10^{-32}(\lambda_w - 1)^{2}}
#' where \eqn{c_{\mathrm{meta}}} is a user-specified coefficient
#' (\code{meta_penalty}). The gradient of \eqn{P_{\mathrm{meta}}} on the log
#' scale, incorporating the exp chain rule, is
#' \eqn{\partial P_{\mathrm{meta}}/\partial\theta_j = c_{\mathrm{meta}}(\nu_j - 1)\nu_j}
#' and
#' \eqn{\partial P_{\mathrm{meta}}/\partial\theta_1 = 10^{-32}(\lambda_w - 1)\lambda_w}.
#' The total objective is the selected tuning criterion plus
#' \eqn{P_{\mathrm{meta}}}.
#' }
#'
#' \subsection{Closed-Form Gradients for Tuning Criteria}{
#' The gradient of the GCV objective with respect to
#' \eqn{\theta_1 = \log\lambda_w} is computed analytically via the quotient
#' rule:
#' \deqn{\frac{\partial\mathrm{GCV}_{u,\gamma}}{\partial\theta_1}
#'   = \frac{1}{D^{2}}\left(\frac{\partial\mathcal{N}}{\partial\theta_1}D
#'   - \mathcal{N}\frac{\partial D}{\partial\theta_1}\right),}
#' where \eqn{\mathcal{N} = \sum r_i^{2}} (numerator) and
#' \eqn{D = N(1 - \gamma\bar{W})^{2}} (denominator). The key intermediates are:
#' \itemize{
#'   \item \eqn{\partial\mathbf{G}/\partial\lambda_w}, computed from the
#'     matrix identity
#'     \eqn{\partial(\mathbf{X}^{\top}\mathbf{X} + \boldsymbol{\Lambda})^{-1}/\partial\lambda = -\mathbf{G}(\partial\boldsymbol{\Lambda}/\partial\lambda)\mathbf{G}}.
#'   \item \eqn{\partial\mathbf{G}^{1/2}/\partial\lambda_w}, derived from
#'     \eqn{\partial\mathbf{G}/\partial\lambda_w} via the eigendecomposition
#'     chain rule.
#'   \item \eqn{\partial\bar{W}/\partial\lambda_w}, the derivative of the
#'     trace of the hat matrix
#'     \eqn{\mathbf{H} = \mathbf{X}\mathbf{U}\mathbf{G}\mathbf{X}^{\top}},
#'     which depends on both
#'     \eqn{\partial\mathbf{G}/\partial\lambda_w} and
#'     \eqn{\partial\mathbf{G}^{1/2}/\partial\lambda_w}.
#'   \item \eqn{\partial\mathcal{N}/\partial\theta_1 = -2\mathbf{r}^{\top}\mathbf{X}(\partial(\mathbf{U}\mathbf{G})/\partial\lambda_w)\mathbf{X}^{\top}\mathbf{y}\cdot\lambda_w},
#'     via the chain rule applied to the residual vector.
#'   \item \eqn{\partial D/\partial\theta_1 =
#'     2(1 - \gamma\bar{W})(-\gamma\,\partial\bar{W}/\partial\lambda_w)\cdot\lambda_w}.
#' }
#' In the implementation, these quantities are assembled by a small set of
#' helper routines: \code{\link{compute_dG_dlambda}} for
#' \eqn{\partial\mathbf{G}/\partial\lambda}, \code{\link{compute_dGhalf}}
#' for \eqn{\partial\mathbf{G}^{1/2}/\partial\lambda},
#' \code{\link{compute_trace_H}} together with the tuning hat-diagonal
#' derivative helpers for the effective-degrees-of-freedom term, and
#' \code{\link{compute_dG_u_dlambda_xy}} for the derivative of the fitted-value
#' quadratic form.
#'
#' For exact leave-one-out, the implementation instead differentiates the
#' per-observation PRESS form directly, requiring both
#' \eqn{\partial r_i / \partial\lambda_w} and
#' \eqn{\partial h_{ii} / \partial\lambda_w}. The latter is assembled
#' partition-wise from the same constrained-\eqn{\mathbf{G}} ingredients used
#' elsewhere in tuning. The exact LOO path does not explicitly form the full
#' projection matrix \eqn{\mathbf{U}} or the full hat matrix \eqn{\mathbf{H}};
#' instead it works with blockwise matrices
#' \eqn{\mathbf{C}_k = \mathbf{G}_k - \mathbf{G}_k\mathbf{A}_k(\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}\mathbf{A}_k^\top\mathbf{G}_k}
#' and their derivatives. Empirically, this observation-wise leverage
#' derivative is the most numerically delicate part of the analytic LOO
#' gradient: most observations are typically well behaved, but a diffuse set of
#' small local errors can remain even when the overall LOO objective and the
#' corresponding GCV gradient are stable.
#'
#' The shared wiggle and flat directions are differentiated directly. For
#' predictor- and partition-specific penalties, a lower-cost ratio heuristic is
#' used:
#' \deqn{\frac{\partial\mathrm{GCV}_{u,\gamma}}{\partial\theta_l} \approx
#'   \frac{\mathrm{mean}(\mathrm{diag}(\mathbf{L}_l))}{\mathrm{mean}(\mathrm{diag}(\boldsymbol{\Lambda}))}
#'   \frac{\partial\mathrm{GCV}_{u,\gamma}}{\partial\theta_w},}
#' where \eqn{\theta_l = \log \lambda_l} and
#' \eqn{\mathbf{L}_l} denotes the current contribution of the corresponding
#' predictor- or partition-specific penalty to \eqn{\boldsymbol{\Lambda}}. This
#' keeps the extra penalty directions inexpensive while preserving the direct
#' derivatives for the shared wiggle and flat penalties.
#' }
#'
#' \subsection{Optimization Procedure}{
#' \strong{Grid search initialization.} The selected tuning criterion is
#' evaluated over a grid of candidate values for
#' \eqn{(\lambda_w, \lambda_r)} on the log scale. All combinations of
#' user-supplied candidate vectors (\code{initial_wiggle} and
#' \code{initial_flat}) are formed, and the combination yielding
#' the smallest finite criterion value is selected as the starting
#' point for BFGS optimization. Grid points producing non-finite
#' objective values are discarded. If all grid points fail, an error
#' is raised advising the user to check the data or adjust the grid.
#'
#' \strong{Damped BFGS optimizer.} A custom damped BFGS quasi-Newton optimizer,
#' implemented in \code{\link{tune_Lambda}} through
#' \code{.damped_bfgs()}, minimizes the selected tuning criterion plus
#' \eqn{P_{\mathrm{meta}}}. When analytic gradients are not preferred, or when
#' the exact LOO leverage derivative appears too noisy for a given problem, a
#' base-R \code{\link[stats]{optim}} call with method \code{"BFGS"} provides
#' the finite-difference fallback.
#'
#' \emph{Iterations 1-2: steepest descent.} The first two iterations use
#' steepest descent with a damping factor \eqn{\alpha}:
#' \eqn{\boldsymbol{\phi}^{(t+1)} = \boldsymbol{\phi}^{(t)} - \alpha\nabla_{\boldsymbol{\phi}}}.
#'
#' \emph{Iterations 3+: BFGS.} From iteration 3, an inverse Hessian
#' approximation \eqn{\mathbf{J}^{(t)}} is maintained via the standard secant
#' update. Let
#' \eqn{\mathbf{s}^{(t)} = \boldsymbol{\phi}^{(t)} - \boldsymbol{\phi}^{(t-1)}}
#' and
#' \eqn{\mathbf{v}^{(t)} = \nabla^{(t)} - \nabla^{(t-1)}}. The BFGS update
#' is:
#' \deqn{\mathbf{J}^{(t+1)}
#'   = (\mathbf{I} - \mathbf{u}\mathbf{s}\mathbf{v}^{\top})
#'   \mathbf{J}^{(t)}
#'   (\mathbf{I} - \mathbf{u}\mathbf{v}\mathbf{s}^{\top})
#'   + \mathbf{u}\mathbf{s}\mathbf{s}^{\top},
#'   \qquad \mathbf{u} = (\mathbf{v}^{\top}\mathbf{s})^{-1}.}
#' When \eqn{|\mathbf{v}^{\top}\mathbf{s}| < 10^{-64}}, the approximation is
#' reset to \eqn{\mathbf{I}} and the iteration is flagged for restart. The
#' search direction is
#' \eqn{\mathbf{d}^{(t)} = -\mathbf{J}^{(t)}\nabla^{(t)}}.
#'
#' \emph{Step acceptance.} A step is accepted if the new tuning criterion value
#' is no larger than the old one.
#' On rejection, \eqn{\alpha} is halved. If \eqn{\alpha < 2^{-10}} (early
#' iterations) or \eqn{\alpha < 2^{-12}} (later iterations), the optimizer
#' terminates with the best solution found.
#'
#' \emph{Convergence.} The optimizer terminates when
#' the absolute change in the tuning criterion is less than \eqn{\epsilon}, or
#' \eqn{\|\boldsymbol{\phi}^{(t)} - \boldsymbol{\phi}^{(t-1)}\|_{\infty} < \epsilon},
#' provided at least 10 iterations have elapsed, for penalties
#' \eqn{\boldsymbol{\phi} = \lambda_w, \lambda_r, \nu_j, \tau_j, ...}.
#'
#' \emph{Alternative.} A base-R
#' \code{\link[stats]{optim}} call with method \code{"BFGS"} and
#' finite-difference gradients can be used instead via
#' \code{use_custom_bfgs = FALSE}.
#'
#' \strong{Post-optimization sample-size adjustment.} After optimization, the
#' tuned penalty parameters are divided by \eqn{((N+2)/(N-2))^{2}}
#' (equivalently multiplied by \eqn{((N-2)/(N+2))^{2}}) for both GCV- and
#' LOO-based tuning. This decreases the final penalties at small sample sizes.
#'
#' The tuning loop is implemented in \code{\link{tune_Lambda}}.
#' }
#'
#' @section Incorporating Non-Spline Effects:
#'
#' Multiple fixed effects are accommodated naturally in the LMSS framework
#' because spline effects, linear effects, and many interaction terms all live
#' in the same partition-wise polynomial expansion. The distinction is therefore
#' not whether a term is "allowed" by the solver, but whether it receives full
#' spline treatment or remains structurally linear across partitions.
#'
#' The constrained framework naturally accommodates non-spline terms. If only
#' linear terms are included for a predictor (via
#' \code{just_linear_without_interactions} or
#' \code{just_linear_with_interactions}), the first-derivative smoothing
#' constraint forces the linear coefficient to be identical across all
#' partitions, since the derivative of a linear function is its slope. This
#' is not an algorithmic modification but a natural consequence of the
#' constraint structure.
#'
#' For example, a model with one spline effect and a linear treatment indicator
#' interaction will naturally keep the treatment-time interaction coefficient
#' constant across partitions while allowing the time effect to vary
#' nonlinearly. This conveniently extends to arbitrary combinations of spline
#' and linear terms without requiring special handling.
#'
#' When \code{blockfit = TRUE} is specified alongside
#' \code{just_linear_without_interactions}, the flat-block path provides an
#' alternative enforcement mechanism. Rather than relying on constraint
#' projection, flat coefficients are pooled structurally across partitions
#' during backfitting. The two approaches agree at the point estimate but
#' differ in their uncertainty quantification; see the Blockfit section above.
#'
#' @section Integration:
#'
#' Because the fitted object retains an explicit polynomial representation in
#' each partition, numerical integration can be carried out in a fairly direct
#' way. The package wraps that calculation in a user-facing S3 method so the
#' user does not need to manage knot boundaries or partition membership by hand.
#'
#' In the user-facing interface, numerical integration is applied through
#' \code{\link{integrate.lgspline}}, which applies Gauss-Legendre quadrature to
#' predictions from the fitted model produced by \code{\link{predict.lgspline}}.
#'
#' \subsection{Implementation}{
#' For a user-supplied rectangular domain, \code{\link{integrate.lgspline}} constructs a
#' tensor-product grid of Gauss-Legendre nodes, evaluates the fitted model at
#' those points, and forms the weighted sum. This works for both univariate and
#' multivariate models, respects the fitted partition structure automatically,
#' and avoids requiring the user to keep track of knot boundaries by hand.
#'
#' The \code{vars} argument selects which predictors are integrated over.
#' Predictors not listed in \code{vars} are held fixed at
#' \code{initial_values} when supplied, or otherwise at the midpoint of their
#' observed training range. The optional \code{B_predict} argument makes it
#' possible to integrate posterior draws or other alternate coefficient sets,
#' and \code{n_quad} controls the number of Gauss-Legendre nodes used per
#' integrated dimension.
#'
#' Integration is performed on the response scale by default. Setting
#' \code{link_scale = TRUE} instead integrates the linear predictor
#' \eqn{\eta = f(\mathbf{t})}. For identity-link Gaussian models the two scales
#' coincide.
#' }
#'
#'
#' @section Lagrange Multipliers:
#'
#' When \code{return_lagrange_multipliers = TRUE}, the multiplier vector
#' \deqn{\boldsymbol{\lambda} = (\mathbf{A}^{\top}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{\top}\hat{\boldsymbol{\beta}}}
#' is returned. These quantify the sensitivity of the penalized objective to
#' relaxing each smoothness or user-supplied equality constraint. When
#' constraint target values are nonzero
#' (\eqn{\mathbf{A}^{\top}\boldsymbol{\beta}_0 \neq \mathbf{0}}), the modified
#' formulation is used:
#' \deqn{\boldsymbol{\lambda} = (\mathbf{A}^{\top}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{\top}(\hat{\boldsymbol{\beta}} - \boldsymbol{\beta}_0)}
#' where \eqn{\mathbf{A}^{\top}\boldsymbol{\beta}_0} is the vector of
#' constraint target values. Multipliers are \code{NULL} when no constraints
#' are active (\eqn{\mathbf{A}} is \code{NULL} or \eqn{K = 0}).
#'
#' For inequality constraints, multipliers are returned as computed by
#' \code{\link[quadprog]{solve.QP}}. The Lagrange multipliers for active
#' inequality constraints can be used diagnostically to identify which shape
#' constraints are most costly in terms of goodness of fit.
#'
#' @section S3 Methods:
#'
#' Standard S3 methods are provided for objects of class \code{lgspline}:
#' \itemize{
#'   \item \code{\link{print.lgspline}} and \code{\link{summary.lgspline}}:
#'     Provide concise model summaries, with
#'     \code{\link{print.summary.lgspline}} formatting coefficient tables in a
#'     familiar regression-style layout.
#'
#'   \item \code{\link{logLik.lgspline}}: Returns a standard \code{logLik}
#'     object. For Gaussian responses with identity link, the exact
#'     log-likelihood is computed. When a correlation structure is present via
#'     \code{VhalfInv}, the log-likelihood includes the
#'     \eqn{\log|\mathbf{V}^{-1/2}|} adjustment and the corresponding
#'     whitened quadratic form. For other families, the method falls back to
#'     \code{family$aic} or a deviance-based approximation. An
#'     \code{include_prior} argument (default \code{TRUE}) optionally adds
#'     the Gaussian prior penalty interpretation of the smoothing spline
#'     penalty
#'     \eqn{-\frac{1}{2\sigma^{2}}\tilde{\boldsymbol{\beta}}^{\top}\boldsymbol{\Lambda}\tilde{\boldsymbol{\beta}}}
#'     to obtain a penalized MAP log-likelihood.
#'
#'   \item \code{\link{predict.lgspline}}: Produces fitted values and related
#'     quantities (e.g., derivatives and standard errors through
#'     \code{se.fit = TRUE}), lets \code{new_predictors} override
#'     \code{newdata}, accepts alternate coefficient lists through
#'     \code{B_predict}, and supports prediction on new predictor matrices
#'     consistent with the original spline expansions.
#'
#'   \item \code{\link{coef.lgspline}}: Extracts partition-specific
#'     coefficient vectors.
#'
#'   \item \code{\link{confint.lgspline}}: Extracts confidence intervals.
#'     When the inverse Hessian from BFGS optimization is available for
#'     correlation parameters, intervals for those correlation parameters are
#'     returned on the working (transformed) scale and should be
#'     back-transformed as described in the correlation section.
#'   \item \code{\link{plot.lgspline}}: For one-dimensional fits, produces
#'     base R graphics showing the fitted function (with optional partition-wise
#'     formulas) and supports overlay via \code{add = TRUE}. For two or more
#'     predictors, an interactive \pkg{plotly}-based visualization is
#'     returned. Specific predictors may be selected via \code{vars}.
#'
#'   \item \code{\link{integrate.lgspline}}: Computes definite integrals of
#'     the fitted surface over rectangular domains by Gauss-Legendre
#'     quadrature.
#' }
#'
#' Additional user-facing helpers include \code{\link{wald_univariate}} for
#' coefficient-wise Wald inference, \code{\link{generate_posterior}} for
#' posterior and posterior-predictive sampling,
#' \code{\link{generate_posterior_correlation}} for correlation-aware posterior
#' simulation, \code{\link{equation}} for closed-form display of the fitted
#' partition formulas, and
#' \code{\link{find_extremum}} for optimizing the fitted surface or a custom
#' acquisition function built from it.
#'
#' @references
#'
#' Buse, A. and Lim, L. (1977). Cubic Splines as a Special Case of
#' Restricted Least Squares. \emph{Journal of the American Statistical
#' Association}, 72, 64-68.
#'
#' Eilers, P. H. and Marx, B. D. (1996). Flexible Smoothing with B-splines
#' and Penalties. \emph{Statistical Science}, 11(2), 89-121.
#'
#' Ezhov, N., Neitzel, F. and Petrovic, S. (2018). Spline Approximation,
#' Part 1: Basic Methodology. \emph{Journal of Applied Geodesy}, 12(2),
#' 139-155.
#'
#' Goldfarb, D. and Idnani, A. (1983). A Numerically Stable Dual Method for
#' Solving Strictly Convex Quadratic Programs. \emph{Mathematical
#' Programming}, 27(1), 1-33.
#'
#' Harville, D. A. (1977). Maximum Likelihood Approaches to Variance
#' Component Estimation and to Related Problems. \emph{Journal of the
#' American Statistical Association}, 72(358), 320-338.
#'
#' Hastie, T. J. and Tibshirani, R. J. (1990). \emph{Generalized Additive
#' Models}. Chapman & Hall/CRC.
#'
#' Kisi, O., Heddam, S., Parmar, K. S., Petroselli, A., Kulls, C. and
#' Zounemat-Kermani, M. (2025). Integration of Gaussian Process Regression
#' and K Means Clustering for Enhanced Short Term Rainfall Runoff Modeling.
#' \emph{Scientific Reports}, 15, 7444.
#'
#' MacQueen, J. B. (1967). Some Methods for Classification and Analysis of
#' Multivariate Observations. In \emph{Proceedings of the Fifth Berkeley
#' Symposium on Mathematical Statistics and Probability}, Volume 1,
#' 281-297. University of California Press.
#'
#' McCullagh, P. and Nelder, J. A. (1989). \emph{Generalized Linear Models}.
#' Chapman & Hall, 2nd edition.
#'
#' Murray, I., Adams, R. P. and MacKay, D. J. C. (2010). Elliptical Slice
#' Sampling. \emph{Proceedings of the 13th International Conference on
#' Artificial Intelligence and Statistics (AISTATS)}, 9, 541-548.
#'
#' Nocedal, J. and Wright, S. J. (2006). \emph{Numerical Optimization}
#' (2nd ed.). Springer.
#'
#' Kim, Y.-J. and Gu, C. (2004). Smoothing Spline Gaussian Regression:
#' More Scalable Computation via Efficient Approximation. \emph{Journal of
#' the Royal Statistical Society: Series B}, 66(2), 337-356.
#'
#' Patterson, H. D. and Thompson, R. (1971). Recovery of Inter-Block
#' Information When Block Sizes Are Unequal. \emph{Biometrika}, 58,
#' 545-554.
#'
#' Pya, N. and Wood, S. N. (2015). Shape Constrained Additive Models.
#' \emph{Statistics and Computing}, 25(3), 543-559.
#'
#' Reinsch, C. H. (1967). Smoothing by Spline Functions. \emph{Numerische
#' Mathematik}, 10, 177-183.
#'
#' Ruppert, D., Wand, M. P. and Carroll, R. J. (2003).
#' \emph{Semiparametric Regression}. Cambridge University Press.
#'
#' Searle, S. R., Casella, G. and McCulloch, C. E. (2006). \emph{Variance
#' Components}. Wiley.
#'
#' Wahba, G. (1990). \emph{Spline Models for Observational Data}. SIAM.
#'
#' Wood, S. N. (2006). On Confidence Intervals for Generalized Additive
#' Models Based on Penalized Regression Splines. \emph{Australian & New
#' Zealand Journal of Statistics}, 48(4), 445-464.
#'
#' Wood, S. N. (2011). Fast Stable Restricted Maximum Likelihood and Marginal
#' Likelihood Estimation of Semiparametric Generalized Linear Models.
#' \emph{Journal of the Royal Statistical Society: Series B}, 73(1), 3-36.
#'
#' Wood, S. N. (2017). \emph{Generalized Additive Models: An Introduction
#' with R}. CRC Press, 2nd edition.
#'
#' @name Details
NULL

#' Process Input Arguments for lgspline
#'
#' @description
#' Parses formula and data arguments, performs factor encoding,
#' resolves variable roles (spline, linear-with-interactions,
#' linear-without-interactions), constructs exclusion patterns,
#' and validates inputs. Called internally by \code{\link{lgspline}}.
#'
#' Users may call this function directly to inspect how their formula
#' and data are interpreted before fitting.
#'
#' @param predictors Default: NULL. Formula or numeric matrix/data frame
#'   of predictor variables.
#' @param y Default: NULL. Numeric response vector.
#' @param formula Default: NULL. Optional formula; alias for predictors
#'   when a formula object.
#' @param response Default: NULL. Alternative name for \code{y}.
#' @param data Default: NULL. Data frame for formula interface.
#' @param weights Default: NULL. Alias for \code{observation_weights}.
#' @param observation_weights Default: NULL. Numeric observation weight
#'   vector.
#' @param family Default: \code{gaussian()}. GLM family object.
#' @param K Default: NULL. Number of interior knots.
#' @param custom_knots Default: NULL. Custom knot matrix.
#' @param auto_encode_factors Default: TRUE. Logical; auto one-hot encode
#'   factor and character columns when using the formula interface.
#' @param include_2way_interactions Default: TRUE. Logical.
#' @param include_3way_interactions Default: TRUE. Logical.
#' @param just_linear_with_interactions Default: NULL. Integer vector or
#'   character vector of column names.
#' @param just_linear_without_interactions Default: NULL. Integer vector or
#'   character vector of column names.
#' @param exclude_interactions_for Default: NULL. Integer vector or character
#'   vector of column names.
#' @param exclude_these_expansions Default: NULL. Character vector of
#'   expansion names to exclude.
#' @param offset Default: \code{c()}. Vector of column indices or names
#'   to include as offsets.
#' @param no_intercept Default: FALSE. Logical; remove intercept.
#' @param do_not_cluster_on_these Default: \code{c()}. Vector of column
#'   indices or names to exclude from clustering.
#' @param include_quartic_terms Default: NULL. Logical or NULL.
#' @param cluster_args Default: \code{c(custom_centers = NA, nstart = 10)}.
#'   Named vector of clustering arguments.
#' @param include_warnings Default: TRUE. Logical.
#' @param dummy_fit Default: FALSE. Logical; run the full preprocessing path but stop short of fitting nonzero coefficients.
#' @param include_constrain_second_deriv Default: TRUE. Logical.
#' @param standardize_response Default: TRUE. Logical.
#' @param ... Additional arguments (checked for depreciated names).
#'
#' @return A named list containing:
#' \describe{
#'   \item{predictors}{Numeric matrix of predictor variables with column
#'     names stripped for positional indexing.}
#'   \item{y}{Numeric response vector.}
#'   \item{og_cols}{Character vector of original predictor column names,
#'     or NULL if none were available.}
#'   \item{replace_colnames}{Logical; TRUE if og_cols is available and
#'     column renaming should be applied post-fit.}
#'   \item{just_linear_with_interactions}{Integer vector of column indices,
#'     or NULL.}
#'   \item{just_linear_without_interactions}{Integer vector of column indices,
#'     or NULL.}
#'   \item{exclude_interactions_for}{Integer vector of column indices,
#'     or NULL.}
#'   \item{exclude_these_expansions}{Character vector of positional-notation
#'     expansion names, or NULL.}
#'   \item{offset}{Integer vector of column indices, or \code{c()}.}
#'   \item{no_intercept}{Logical.}
#'   \item{do_not_cluster_on_these}{Numeric vector of column indices,
#'     or \code{c()}.}
#'   \item{observation_weights}{Numeric vector or NULL.}
#'   \item{K}{Integer or NULL, possibly updated by cluster_args or
#'     all-linear detection.}
#'   \item{include_3way_interactions}{Logical, possibly updated by
#'     formula parsing.}
#'   \item{include_quartic_terms}{Logical or NULL, possibly updated
#'     by number of predictors.}
#'   \item{data}{Data frame, possibly with factor columns one-hot encoded.}
#'   \item{include_constrain_second_deriv}{Logical, possibly set FALSE
#'     when no numeric predictors remain.}
#'   \item{factor_groups}{Named list mapping original factor column names to
#'     integer vectors of their one-hot indicator column positions within
#'     the predictor matrix. Used by lgspline.fit to impose sum-to-zero
#'     constraints on encoded factor levels. NULL when no factors were
#'     encoded, or when the formula interface was not used.}
#' }
#'
#' @seealso \code{\link{lgspline}} for the main fitting interface.
#'
#' @examples
#' \dontrun{
#' data("Theoph")
#' df <- Theoph[, c("Time", "Dose", "conc", "Subject")]
#' processed <- process_input(
#'   predictors = conc ~ spl(Time) + Time*Dose,
#'   data = df,
#'   auto_encode_factors = TRUE,
#'   include_warnings = TRUE
#' )
#' str(processed$predictors)
#' processed$og_cols
#' processed$just_linear_without_interactions
#' processed$factor_groups
#' }
#'
#' @keywords internal
#' @export
process_input <- function(
    predictors = NULL,
    y = NULL,
    formula = NULL,
    response = NULL,
    data = NULL,
    weights = NULL,
    observation_weights = NULL,
    family = gaussian(),
    K = NULL,
    custom_knots = NULL,
    auto_encode_factors = TRUE,
    include_2way_interactions = TRUE,
    include_3way_interactions = TRUE,
    just_linear_with_interactions = NULL,
    just_linear_without_interactions = NULL,
    exclude_interactions_for = NULL,
    exclude_these_expansions = NULL,
    offset = c(),
    no_intercept = FALSE,
    do_not_cluster_on_these = c(),
    include_quartic_terms = NULL,
    cluster_args = c(custom_centers = NA, nstart = 10),
    include_warnings = TRUE,
    dummy_fit = FALSE,
    include_constrain_second_deriv = TRUE,
    standardize_response = TRUE,
    ...
){

  ## Backward compat: expansions_only in ... maps
  #  to dummy_fit with deprecation warning
  #  shur -> schur for correct spelling
  dots <- list(...)
  if(!is.null(dots$expansions_only) && dots$expansions_only){
    if(include_warnings){
      warning(
        "'expansions_only' is depreciated; use 'dummy_fit = TRUE' instead.",
        " The full pipeline now executes with coefficients set to 0.")
    }
    dummy_fit <- TRUE
  }
  if(!is.null(dots$shur_correction_function)){
    if(include_warnings){
      warning("'shur_correction_function' option is depreciated, and won't be",
              "applied; repeat with 'schur_correction_function' instead.")
    }
  }

  ## Track one-hot factor groups for later sum-to-zero constraints
  #  Stores original factor name -> indicator column names until the
  #  final predictor matrix has been resolved.
  factor_groups <- list()

  ## Update naming conventions, if first argument is a formula and second is a
  # data frame, assumed by R-like interfaces for user convenience.
  if(any(!is.null(predictors)) & any(!is.null(y))){
    if(any(inherits(y,'data.frame') & inherits(predictors, "formula"))){
      data <- y
    }
  }

  ## Update naming conventions, if response supplied in place of "y" for user
  # convenience.
  if(any(is.null(predictors)) & any(!is.null(formula))){
    predictors <- formula
  } else if(any(is.null(predictors))){
    stop(
      '\n \t Predictors argument is NULL without formula supplied.',
      ' Either supply a formula to predictors OR formula argument, or a',
      ' data frame to predictors argument, or a matrix of numeric predictors',
      ' to the predictor argument. \n')
  }

  ## Update naming conventions, if response supplied in place of "y" for user
  # convenience.
  if(any(is.null(y)) & any(!is.null(response))){
    y <- response
    response <- NULL
  }

  ## Weights is just an R-friendly argument to be passed to observation_weights
  # actually used by the function.
  if(any(is.null(observation_weights)) &
     any(!is.null(weights))){
    observation_weights <- weights
    weights <- NULL
  }

  ## Check cluster args for compatibility
  if(any(!is.na(cluster_args[[1]]))){
    ncluster <- try({nrow(cluster_args[[1]])},
                    silent = TRUE)
    if(any(inherits(ncluster, 'try-error'))){
      stop('\n \t custom_centers should be a matrix; do not include any other',
           ' arguments within cluster_args if you include custom_centers. \n')
    }
    if(!is.null(K)){
      if(ncluster != (K+1) & include_warnings){
        warning('\n \t K must be equal to number of custom_centers minus 1. ',
                'Updating K for compatibility. \n')
        K <- ncluster - 1
      }
    } else {
      K <- ncluster - 1
    }
  }

  ## Check data and formula argument
  if(!is.null(data) & !inherits(predictors, "formula")) {
    stop("\n \t If submitting data argument, formula must be supplied and',
         ' variables must match. Otherwise, use predictors and y (or response)",
         " arguments directly.\n",
         "\t Example: lgspline(y ~ spl(x1, x2) + x3 + x4*x5, data = my_data)\n",
         "\t Example: lgspline(y ~ ., data = my_data)\n")
  }

  ## Handle formula interface
  if(inherits(predictors, "formula")) {

    ## Allow s() as an alias for spl() for mgcv-style formulas.
    #  Replace s(...) with spl(...) in formula before any parsing.
    #  Uses pattern matching to avoid replacing 's' inside other function
    #  names like abs(), cos(), is(), etc. The regex anchors on a
    #  non-alphanumeric/non-underscore/non-dot character (or start of string)
    #  immediately before the 's(', so only bare 's(' is matched.
    form_str_alias <- paste(deparse(predictors, width.cutoff = 500L),
                            collapse = "")
    form_str_alias <- gsub("(^|[^a-zA-Z0-9_.])s\\(", "\\1spl(", form_str_alias)
    predictors <- as.formula(form_str_alias)

    ## Check data argument
    if(is.null(data)) {
      stop(
        "\n \t When using formula interface, data argument must be ",
        "provided.\n",
        "\t Example: lgspline(y ~ spl(x1, x2) + x3 + x4*x5, data = my_data)\n",
        "\t Example: lgspline(y ~ ., data = my_data)\n")
    }

    ## Try to coerce data to data.frame
    tryCatch({
      data <- as.data.frame(data)
    }, error = function(e) {
      stop("\n \t Could not coerce data argument to data.frame. ",
           "Please provide data in a format coercible to data.frame. ",
           "Examples: data.frame, tibble, or matrix.")
    })

    ## Auto one-hot encode factor/character columns
    # For reviewers: the decision to include one-hot rather than
    # dummy-intercept is for interpretation as conditional random effects
    # models. We would not like this to drop a random effect level,
    # The flat_ridge_penalty should address lack of identifiability to an extent
    # Users can manually supply the variable names after one-hot encoding
    # if dummy-intercept is truly desired.
    if(auto_encode_factors && !is.null(data)){
      non_numeric_cols <- which(!sapply(data, is.numeric))
      ## Exclude response column from encoding
      resp_name <- NULL
      if(inherits(predictors, "formula")){
        resp_name <- as.character(predictors[[2]])
      }
      non_numeric_cols <- non_numeric_cols[
        !names(non_numeric_cols) %in% resp_name
      ]
      if(length(non_numeric_cols) > 0){
        for(col_idx in rev(non_numeric_cols)){
          col_name <- colnames(data)[col_idx]
          ## One-hot encode this column
          encoded <- create_onehot(data[[col_name]])
          colnames(encoded) <- paste0(col_name, "_", colnames(encoded))
          ## Remove original, append indicators
          data <- cbind(data[, -col_idx, drop = FALSE], encoded)
          ## Track these as linear-without-interactions only if the
          #  original column was not involved in any interaction term
          new_names <- colnames(encoded)
          ## Track these indicator columns for the later sum-to-zero constraints.
          #  Stored by original factor name, then resolved to indices below.
          factor_groups[[col_name]] <- new_names
          form_check <- paste(deparse(predictors, width.cutoff = 500L),
                              collapse = "")
          col_in_interaction <- grepl(
            paste0("(\\*|:)\\s*", col_name, "(\\s|\\)|\\+|$)"), form_check
          ) || grepl(
            paste0("(\\+|\\(|~|\\s)", col_name, "\\s*(\\*|:)"), form_check
          )
          if(!col_in_interaction){
            if(is.null(just_linear_without_interactions)){
              just_linear_without_interactions <- new_names
            } else {
              just_linear_without_interactions <- c(
                just_linear_without_interactions, new_names
              )
            }
          }
          ## Update formula if needed: replace col_name with encoded names
          if(inherits(predictors, "formula")){
            form_str <- deparse(predictors)
            replacement <- paste0("(", paste(new_names, collapse = " + "), ")")
            form_str <- gsub(col_name, replacement, form_str, fixed = TRUE)
            predictors <- as.formula(form_str)
          }
        }
      }
    }

    ## Check column names are present
    if(any(is.null(colnames(data)))){
      stop('\n \t Column names of data must be supplied if formula is',
           ' supplied, and column names must match what is provided in formula',
           '.\n')
    }

    ## Stringed formula
    form_paste0 <- paste0(predictors)

    ## If formula is y ~ ., replace with y ~ spl(x1, x2, ...)
    if(form_paste0[1] == '~' &
       (gsub(' ', '', form_paste0[3]) %in% c('.', '0+.','.+0')) &
       length(form_paste0) == 3){
      if(form_paste0[2] == '.'){
        stop('\n \t . ~ . is not valid for this function, specify the',
             ' response directly, like "y ~ . \n"')
      } else if(gsub(' ', '', form_paste0[3]) == '0+.' |
                gsub(' ', '', form_paste0[3]) == '.+0'){
        ## No intercept
        predictors <- as.formula(
          paste0(form_paste0[2],
                 ' ~ 0 + spl(',
                 paste(colnames(data)[-which(colnames(data) ==
                                               form_paste0[2])],
                       collapse = ', '),
                 ')')
        )
      } else {
        ## With intercept
        predictors <- as.formula(
          paste0(form_paste0[2],
                 ' ~ spl(',
                 paste(colnames(data)[-which(colnames(data) ==
                                               form_paste0[2])],
                       collapse = ', '),
                 ')')
        )
      }
    }

    ## [Change PATCH] Handle spl(...) + . formula pattern.
    #  When the formula contains spl() terms AND a trailing dot,
    #  replace the dot with explicit column names of non-spl,
    #  non-response columns and treat them as
    #  just_linear_without_interactions.
    #  For example:
    #    conc ~ spl(Time) + Time*Dose + .
    #  becomes (with data having columns Time, Dose, Subject):
    #    conc ~ spl(Time) + Time*Dose + Subject
    #  and Subject is added to just_linear_without_interactions.
    #  Variables already named elsewhere in the formula (inside spl(),
    #  in interactions, etc.) are excluded from the dot expansion so
    #  they do not appear twice. The downstream formula parser then
    #  correctly classifies variables that appear in interactions as
    #  linear_with_int rather than linear_without_int.
    form_str_full <- paste(deparse(predictors, width.cutoff = 500L),
                           collapse = "")
    has_spl <- grepl("spl\\(", form_str_full)
    has_dot <- grepl("\\+\\s*\\.", form_str_full) |
      grepl("\\.\\s*\\+", form_str_full)
    if(has_spl & has_dot){
      ## Extract spl variable names
      spl_matches <- gregexpr("spl\\(([^)]+)\\)", form_str_full)
      spl_contents <- regmatches(form_str_full, spl_matches)[[1]]
      spl_var_names <- unique(trimws(unlist(strsplit(
        gsub("spl\\((.*)\\)", "\\1", spl_contents), ","
      ))))

      ## Response name
      resp_nm <- as.character(predictors[[2]])

      ## All variables already explicitly named in the formula
      #  (outside spl() and the dot itself). Parse the terms to find them.
      form_terms_tmp <- try(terms(predictors), silent = TRUE)
      if(!inherits(form_terms_tmp, 'try-error')){
        explicit_vars <- unique(c(
          spl_var_names,
          rownames(attr(form_terms_tmp, "factors"))[-1]
        ))
      } else {
        explicit_vars <- spl_var_names
      }

      ## Remaining columns = all data columns minus response minus explicit
      remaining_cols <- setdiff(colnames(data),
                                c(resp_nm, explicit_vars))

      if(length(remaining_cols) > 0){
        ## Replace the dot with explicit column names
        dot_replacement <- paste(remaining_cols, collapse = " + ")
        ## Handle both "+ ." and ". +" patterns
        form_str_new <- sub("\\+\\s*\\.", paste0("+ ", dot_replacement),
                            form_str_full)
        form_str_new <- sub("\\.\\s*\\+", paste0(dot_replacement, " + "),
                            form_str_new)
        predictors <- as.formula(form_str_new)

        ## Append to just_linear_without_interactions
        if(is.null(just_linear_without_interactions)){
          just_linear_without_interactions <- remaining_cols
        } else {
          just_linear_without_interactions <- unique(c(
            just_linear_without_interactions, remaining_cols
          ))
        }
      }
    }

    ## If offset() appears in the formula, remove offset from the formula
    # and set offset = names of terms inside the offset() operator
    if(any(grepl('offset', paste0(predictors)))){
      if(any(grepl("offset\\(", deparse(predictors)))){
        ## Extract the original formula as string
        form_str <- deparse(predictors)

        ## Find all offset terms using regex
        offset_matches <- gregexpr("offset\\(([^)]+)\\)", form_str)
        offset_vars <- regmatches(form_str, offset_matches)[[1]]

        ## Extract variable names from offset terms
        offset_var_names <- gsub("offset\\((.*)\\)", "\\1", offset_vars)

        ## Remove whitespace
        offset_var_names <- trimws(offset_var_names)

        ## Add to offset vector
        offset <- c(offset, offset_var_names)

        ## Replace offset(var) with var in formula
        for(i in seq_along(offset_vars)) {
          form_str <- gsub(offset_vars[i], offset_var_names[i],
                           form_str,
                           fixed=TRUE)
        }

        ## Convert back to formula
        predictors <- as.formula(form_str)
      }
    }

    ## Check that all formula predictors are numeric in data
    terms <- terms(predictors)
    term_labels <- attr(terms, "term.labels")
    non_numeric <- c(1:ncol(data))[!sapply(data, is.numeric)]
    if(length(non_numeric) > 0) {
      for(v in non_numeric){
        if(any(term_labels == colnames(data)[v])){
          if(auto_encode_factors){
            stop("Auto-encoding failed for column '",
                 colnames(data)[v], "'. Check that it is a factor ",
                 "or character vector.")
          } else {
            stop("Non-numeric columns detected in predictors. ",
                 "Set auto_encode_factors = TRUE to automatically ",
                 "one-hot encode factor/character columns, or ",
                 "convert these to numeric before proceeding. ",
                 "You can use create_onehot() on categorical ",
                 "columns of your dataset to obtain binary ",
                 "indicator variables if you need to. Remember ",
                 "to remove the original categorical variable ",
                 "and append the indicators to your data ",
                 "before adjusting your formula and call ",
                 "to lgspline(...).\n\nSee ?create_onehot for an example.")
          }
        }
      }
    }

    ## Parse rest of formula
    has_3way <- any(attr(terms, "order") == 3)
    if(include_3way_interactions) include_3way_interactions <- has_3way

    ## Check for no intercept specification in formula
    if(inherits(predictors, "formula")) {
      formula_text <- gsub(" ", "", Reduce(paste, deparse(predictors)))
      if(grepl("\\+0|0\\+", formula_text)) {
        no_intercept <- TRUE
      }
    }

    ## Get the factors matrix which shows interaction structure
    factors <- attr(terms, "factors")

    ## Initialize term containers
    spline_terms <- character()
    linear_no_int <- character()
    linear_with_int <- character()

    ## First pass to identify spline terms and extract their variables
    for(term in term_labels) {
      if(grepl("^spl\\(.*\\)$", term)) {
        vars <- gsub("^spl\\((.*)\\)$", "\\1", term)
        spline_terms <- c(spline_terms, trimws(strsplit(vars, ",")[[1]]))
      }
    }

    ## Extract explicit interactions from formula
    formula_interactions <- term_labels[attr(terms, "order") > 1]
    allowed_interactions <- sapply(formula_interactions, function(term) {
      if(grepl(":", term)) {
        vars <- strsplit(term, ":")[[1]]
        # Only keep interactions where NO term is a spline term
        if(!any(vars %in% spline_terms)) {
          return(term)
        }
      }
      return(NULL)
    })

    ## Identify interaction structure using factors matrix
    var_names <- rownames(factors)[-1] # Remove responses

    ## Extract all types of variables in formula
    all_formula_vars <- unique(c(spline_terms, var_names))
    formula_cols <- which(colnames(data) %in% all_formula_vars)

    ## Match custom exclusions nominal to numeric formula columns, if available
    if(inherits(exclude_interactions_for, 'character')){
      exclude_interactions_for <-
        unlist(lapply(exclude_interactions_for,
                      function(var){
                        grep(var,
                             colnames(data)[formula_cols])
                      }))

    }
    if(inherits(just_linear_with_interactions,  'character')){
      just_linear_with_interactions <-
        unlist(lapply(just_linear_with_interactions,
                      function(var){
                        grep(var,
                             colnames(data)[formula_cols])
                      }))

    }
    if(inherits(just_linear_without_interactions, 'character')){
      just_linear_without_interactions <-
        unlist(lapply(just_linear_without_interactions,
                      function(var){
                        grep(var,
                             colnames(data)[formula_cols])
                      }))

    }
    if(length(exclude_these_expansions) > 0){
      for(ii in 1:length(exclude_these_expansions)){
        exps <- exclude_these_expansions[ii]
        if(substr(exps, 1, 1) != "_" |
           substr(exps, length(exps), length(exps)) != "_"){
          for(jj in 1:length(formula_cols)){
            if(grepl(colnames(data)[formula_cols[jj]],
                     exclude_these_expansions[ii])){
              exclude_these_expansions[ii] <- gsub(
                colnames(data)[formula_cols[jj]],
                paste0('_', jj, '_'),
                exclude_these_expansions[ii]
              )
            }
          }
        }
      }
    }
    if(inherits(offset,  'character')){
      offset <-
        unlist(lapply(offset,
                      function(var){
                        grep(var,
                             colnames(data)[formula_cols])
                      }))
    }

    ## Resolve character do_not_cluster_on_these to numeric
    #  column indices within the formula path. Without this,
    #  make_partitions' assign_partition closure fails when indexing
    #  matrices after transf() strips colnames.
    if(length(do_not_cluster_on_these) > 0 &&
       is.character(do_not_cluster_on_these)){
      do_not_cluster_on_these <-
        unlist(lapply(do_not_cluster_on_these,
                      function(var){
                        idx <- which(colnames(data)[formula_cols] == var)
                        if(length(idx) == 0){
                          idx <- grep(var, colnames(data)[formula_cols])
                        }
                        idx
                      }))
      do_not_cluster_on_these <- unique(do_not_cluster_on_these)
    }

    ## Convert factor_groups from column names to predictor-matrix indices.
    #  lgspline.fit uses these positions to build the sum-to-zero constraints.
    #  Groups that cannot be resolved cleanly are dropped silently.
    if(length(factor_groups) > 0){
      factor_groups <- lapply(factor_groups, function(col_names){
        unlist(lapply(col_names, function(nm){
          idx <- which(colnames(data)[formula_cols] == nm)
          if(length(idx) == 0) idx <- grep(paste0("^", nm, "$"),
                                           colnames(data)[formula_cols])
          idx
        }))
      })
      ## Remove any groups that could not be resolved or have only 1 level
      factor_groups <- factor_groups[sapply(factor_groups, length) > 1]
    }

    ## Non-spline variables that appear in interactions (from factors matrix)
    interaction_terms <- term_labels[attr(terms, "order") > 1]
    nonspline_interact_vars <- unique(sapply(interaction_terms,
                                             function(term) {
                                               if(grepl(":", term)) {
                                                 vars <-
                                                   strsplit(term,":")[[1]]
                                                 vars[!vars %in%
                                                        spline_terms]
                                               }
                                             }))

    ## Safe initialization of allowed_interaction_pairs
    if(!exists("allowed_interaction_pairs")){
      allowed_interaction_pairs <- list()
    }

    ## Add explicit interactions for spline terms
    if(length(spline_terms) > 1) {
      ## Generate all possible interactions between spline terms
      spline_interactions <- utils::combn(spline_terms, 2, simplify=FALSE)
      spline_triplets <- if(length(spline_terms) >= 3 &
                            include_3way_interactions) {
        utils::combn(spline_terms, 3, simplify=FALSE)
      } else {
        list()
      }

      ## Add 2-way interactions
      for(pair in spline_interactions) {
        interaction_terms <- c(interaction_terms, paste(pair, collapse=":"))
      }

      ## Add 3-way interactions
      for(triplet in spline_triplets) {
        interaction_terms <- c(interaction_terms, paste(triplet, collapse=":"))
      }

      ## Also register spline triplets as allowed interaction pairs
      ## so they are not excluded by the grouping logic below
      for(triplet in spline_triplets) {
        allowed_interaction_pairs[[length(allowed_interaction_pairs) + 1]] <-
          list(pair = triplet,
               transforms = c(
                 paste(triplet, collapse = "x"),
                 paste(rev(triplet), collapse = "x")
               )
          )
      }
    }

    ## Get indices of variables in raw expansions (after response removed)
    resp_ind <- which(colnames(data) == paste0(terms[[2]]))
    var_positions <- match(var_names, colnames(data[,formula_cols]))

    ## Get allowed interaction pairs
    allowed_pairs <- lapply(interaction_terms[grepl(":", interaction_terms)],
                            function(term) {
                              vars <- strsplit(term, ":")[[1]]
                              match(vars, colnames(data[,formula_cols]))
                            })

    ## Generate exclusion patterns for non-spline vars
    exclude_patterns <- c()

    ## Get all possible 2-way interactions between ANY variables
    vars <- colnames(data[,formula_cols])
    if(include_2way_interactions){
      for(ii in seq_along(vars)) {
        for(jj in seq_along(vars)) {
          if(ii != jj) {
            pattern <- get_interaction_patterns(c(vars[ii], vars[jj]))

            ## Skip if not in formula
            if(any(!(c(vars[ii], vars[jj]) %in% all_formula_vars))){
              next
            }

            ## Skip if any variable in exclude_interactions_for
            if(!is.null(exclude_interactions_for)){
              if(any(c(ii, jj) %in% exclude_interactions_for)){
                next
              }
            }

            ## Only keep spline-spline interactions and explicit interactions
            if(!all(c(vars[ii], vars[jj]) %in% spline_terms) &
               length(spline_terms) > 0 &
               all(c(ii, jj) %in% formula_cols)){
              exclude_patterns <- c(exclude_patterns, pattern)
            }
          }
        }
      }
    }

    ## Add all possible 3-way interactions to exclusions
    if(include_3way_interactions) {
      for(ii in seq_along(vars)) {
        for(jj in seq_along(vars)) {
          for(kk in seq_along(vars)) {
            if(ii != jj && jj != kk && ii != kk) {
              triplet_vars <- c(vars[ii], vars[jj], vars[kk])
              pattern <- get_interaction_patterns(triplet_vars)

              ## Skip if not in formula
              if(any(!(triplet_vars %in% all_formula_vars))){
                next
              }

              ## Skip if any variable in exclude_interactions_for
              if(!is.null(exclude_interactions_for)){
                if(any(c(ii, jj, kk) %in% exclude_interactions_for)){
                  next
                }
              }

              ## Skip if only spline terms
              if(all(triplet_vars %in% spline_terms)){
                next
              }

              ## If ANY var is a spline term but not ALL are spline terms,
              # we should exclude this interaction
              if(any(triplet_vars %in% spline_terms) &&
                 !all(triplet_vars %in% spline_terms)) {
                exclude_patterns <- c(exclude_patterns, pattern)
                next
              }

              ## For non-spline terms, allow explicitly specified interactions
              if(!any(triplet_vars %in% spline_terms)) {
                ## Get interaction terms that could involve these variables
                relevant_terms <- interaction_terms[grepl(paste(triplet_vars,
                                                                collapse="|"),
                                                          interaction_terms)]

                ## Check if this exact triplet exists in any order
                matches_interaction <- any(sapply(strsplit(relevant_terms, ":"),
                                                  function(term) {
                                                    length(term) == 3 &&
                                                      all(sort(triplet_vars) ==
                                                            sort(term))
                                                  }))

                if(!matches_interaction) {
                  exclude_patterns <- c(exclude_patterns, pattern)
                }
              }
            }
          }
        }
      }
    }

    ## Helper-vector for the next step, in isolating which interactions are
    #  not part of the same grouping/block.
    #  The above only works if we have only 1 additive interaction effect.
    #  the code below corrects for when we have multiple
    #  (for example, a*b + c*d)
    #  and want to prevent unspecified interactions
    #  (for example, remove a*d and b*c)
    different_grouping_exclusions <- c()

    ## Explicitly track allowed interactions from * and : terms
    if(!exists("allowed_interaction_pairs")){
      allowed_interaction_pairs <- list()
    }

    ## Track which interactions are explicitly specified
    for(term in term_labels) {
      # Handle * interactions
      if(grepl("\\*", term)) {
        vars <- trimws(strsplit(term, "\\*")[[1]])

        # Ensure exactly 2 variables in the interaction
        if(length(vars) == 2) {
          allowed_interaction_pairs[[length(allowed_interaction_pairs) + 1]] <-
            list(pair = vars,
                 transforms = c(
                   paste(vars, collapse = "x"),
                   paste(rev(vars), collapse = "x")
                 )
            )
        }
      }

      # Handle : interactions
      if(grepl(":", term)) {
        vars <- trimws(strsplit(term, ":")[[1]])

        # Ensure exactly 2 variables in the interaction
        if(length(vars) == 2) {
          allowed_interaction_pairs[[length(allowed_interaction_pairs) + 1]] <-
            list(pair = vars,
                 transforms = c(
                   paste(vars, collapse = "x"),
                   paste(rev(vars), collapse = "x")
                 )
            )
        }
      }
    }

    ## Identify spline block variables
    spline_block_vars <- c()
    for(term in term_labels) {
      if(grepl("^spl\\(.*\\)$", term)) {
        vars <- gsub("^spl\\((.*)\\)$", "\\1", term)
        term_vars <- trimws(strsplit(vars, ",")[[1]])
        spline_block_vars <- c(spline_block_vars, term_vars)
      }
    }

    ## Generate exclusions for any interaction not in the allowed pairs
    vars <- colnames(data[,formula_cols])
    if(include_2way_interactions | include_3way_interactions){
      for(ii in seq_along(vars)) {
        for(jj in seq_along(vars)) {
          if(ii != jj) {

            pair <- c(vars[ii], vars[jj])
            current_transform <- paste(pair, collapse = "x")

            ## Skip if not in formula
            if(any(!(pair %in% all_formula_vars))){
              next
            }

            ## Skip if any variable in exclude_interactions_for
            if(!is.null(exclude_interactions_for)){
              if(any(c(ii, jj) %in% exclude_interactions_for)){
                next
              }
            }

            ## Check if this is an allowed interaction
            is_allowed_interaction <- FALSE
            for(allowed in allowed_interaction_pairs) {
              if(setequal(pair, allowed$pair) ||
                 any(current_transform == allowed$transforms)) {
                is_allowed_interaction <- TRUE
                break
              }
            }

            ## Check if this is a spline block interaction (2-way or 3-way)
            is_spline_block_interaction <- all(pair %in% spline_block_vars)

            ## If not an allowed or spline block interaction,
            # generate exclusion patterns
            if(!is_allowed_interaction && !is_spline_block_interaction) {
              patterns <- get_interaction_patterns(pair)
              different_grouping_exclusions <- c(different_grouping_exclusions,
                                                 patterns)
            }
          }
        }
      }
    }

    ## Remove duplicates
    different_grouping_exclusions <- unique(different_grouping_exclusions)

    ## Remove explicitly allowed interactions from exclusions
    for(term in interaction_terms) {
      vars <- strsplit(term, ":")[[1]]
      allowed <- get_interaction_patterns(vars)
      exclude_patterns <- setdiff(exclude_patterns, allowed)
    }
    exclude_patterns <- unique(c(exclude_patterns,
                                 different_grouping_exclusions))

    ## Convert to positional notation
    vars <- colnames(data)[formula_cols]
    for(ii in seq_along(formula_cols)) {
      exclude_patterns <- gsub(vars[ii], paste0("_", ii, "_"), exclude_patterns)
    }

    ## Append to custom exclusions
    if(!is.null(exclude_these_expansions)){
      exclude_these_expansions <- c(exclude_these_expansions,
                                    exclude_patterns)
    } else if (length(exclude_patterns) > 0){
      exclude_these_expansions <- exclude_patterns
    }

    ## For each variable, determine if linear with or without interactions
    for(var in var_names) {
      if(length(spline_terms) > 0){
        if(var %in% spline_terms) next # Skip spline terms
      }

      ## Check if this variable appears in any interactions
      var_terms <- which(factors[var,] > 0)
      if(length(var_terms) > 0) {
        ## If any term containing this variable has order > 1,
        # its in an interaction
        if(any(attr(terms, "order")[var_terms] > 1)) {
          linear_with_int <- c(linear_with_int, var)
        } else {
          linear_no_int <- c(linear_no_int, var)
        }
      } else {
        linear_no_int <- c(linear_no_int, var)
      }
    }

    ## Remove duplicates and ensure proper separation
    linear_with_int <- unique(linear_with_int)
    linear_no_int <- setdiff(unique(linear_no_int),
                             c(spline_terms, linear_with_int))
    linear_no_int <- linear_no_int[!(substr(linear_no_int,
                                            1,
                                            4) == 'spl(')]

    ## Create predictors matrix and response for compatibility
    predictors <- data[, formula_cols, drop = FALSE]
    y <- data[, resp_ind]

    ## Convert variable names to column indices
    new_just_linear_without_interactions <- match(linear_no_int,
                                                  colnames(predictors))
    new_just_linear_with_interactions <- match(linear_with_int,
                                               colnames(predictors))
    new_just_linear_without_interactions <- new_just_linear_without_interactions[
      !(new_just_linear_without_interactions %in% spline_terms)
    ]
    new_just_linear_with_interactions <- new_just_linear_with_interactions[
      !(new_just_linear_with_interactions %in% spline_terms)
    ]
    if(is.null(just_linear_with_interactions)){
      just_linear_with_interactions <- new_just_linear_with_interactions
    } else{
      just_linear_with_interactions <- unique(c(
        just_linear_with_interactions,
        new_just_linear_with_interactions
      ))
    }
    if(is.null(just_linear_without_interactions)){
      just_linear_without_interactions <- new_just_linear_without_interactions
    } else {
      just_linear_without_interactions <- unique(c(
        just_linear_without_interactions,
        new_just_linear_without_interactions
      ))
    }
  }

  ## Not a formula - try to coerce to matrix
  tryCatch({
    predictors <- as.matrix(predictors)
  }, error = function(e) {
    stop("\n \t Could not coerce predictors to matrix. ",
         "predictors must be either a formula or an object coercible to matrix.",
         "Examples:\n",
         "  Formula: lgspline(y ~ spl(x1, x2) + x3, data = my_data)\n",
         "  Matrix:  lgspline(predictors = Tmat, y = y) \n")
  })

  ## Check numeric type
  if(any(!is.numeric(predictors))){
    stop("\n \t predictors matrix must be numeric. ",
         "Please convert categorical variables to numeric indicators. \n")
  }

  ## Check response for missings
  if(any(is.na(y) | is.nan(y) | !is.finite(y))){
    stop("\n \t NA, NaN, or infinite value detected in response. \n")
  }

  ## Original predictor names
  og_cols <- colnames(predictors)
  if(!any(is.null(og_cols))){
    replace_colnames <- TRUE
  } else {
    replace_colnames <- FALSE
  }

  ## Check nrow of input predictors and matrix coersion
  t <- try({if(nrow(methods::as(predictors,'matrix')) < 3){
    stop('\n \t Need at least 3 observations to fit model \n')
  }}, silent = TRUE)
  if(inherits(t, 'try-error')){
    stop('\n \t Cannot coerce predictors to a matrix \n')
  }

  ## Check if no spline terms - if so, set K = 0
  if(length(unique(c(just_linear_with_interactions,
                     just_linear_without_interactions))) == ncol(predictors)){
    K <- 0
  }

  ## Check if custom knots is not missing, that it can be
  # coerced to a matrix
  if(any(!(is.null(custom_knots)))){
    custom_knots <- try(cbind(custom_knots),
                        silent = TRUE)
    if(any(inherits(custom_knots, 'try-error')) & include_warnings){
      warning('\n \t custom_knots must be a matrix, or should be coercible ',
              'to it. Custom_knots will be ignored. \n')
      custom_knots <- NULL
    }
  }

  ## If ncol(predictors) > 1 and include_quartic_terms is NULL,
  # set include_quartic_terms = TRUE
  # otherwise, if ncol(predictors) == 1 then FALSE
  if(is.null(include_quartic_terms)){
    if(ncol(cbind(predictors)) == 1){
      include_quartic_terms <- FALSE
    } else {
      include_quartic_terms <- TRUE
    }
  }

  ## factor_groups carries one-hot indicator positions back to lgspline.fit.
  #  An empty list means no encoded factor groups were detected.
  list(
    predictors                       = predictors,
    y                                = y,
    og_cols                          = og_cols,
    replace_colnames                 = replace_colnames,
    just_linear_with_interactions    = just_linear_with_interactions,
    just_linear_without_interactions = just_linear_without_interactions,
    exclude_interactions_for         = exclude_interactions_for,
    exclude_these_expansions         = exclude_these_expansions,
    offset                           = offset,
    no_intercept                     = no_intercept,
    do_not_cluster_on_these          = do_not_cluster_on_these,
    observation_weights              = observation_weights,
    K                                = K,
    include_3way_interactions        = include_3way_interactions,
    include_quartic_terms            = include_quartic_terms,
    data                             = data,
    include_constrain_second_deriv   = include_constrain_second_deriv,
    custom_knots                     = custom_knots,
    dummy_fit                        = dummy_fit,
    factor_groups                    = factor_groups
  )
}

#' Negative Binomial Regression Helpers for lgspline
#'
#' Functions for fitting negative binomial (NB2) regression models within the
#' lgspline framework. Analogous to the Weibull AFT and Cox PH helpers, these
#' provide the log-likelihood, score, information, and all interface functions
#' needed by lgspline's unconstrained fitting, penalty tuning, and inference
#' machinery.
#'
#' The parameterization follows NB2: \eqn{Y \sim \mathrm{NB}(\mu, \theta)}
#' with \eqn{\mathrm{Var}(Y) = \mu + \mu^2/\theta}, where \eqn{\theta > 0}
#' is the shape (size) parameter. The canonical link is log:
#' \eqn{\eta = \log\mu}. The dispersion stored in
#' \code{lgspline$sigmasq_tilde} is \eqn{\theta} itself, not
#' \eqn{1/\theta}.
#'
#' @name negbin_helpers
NULL


#' Compute Negative Binomial Log-Likelihood
#'
#' @description
#' Evaluates the NB2 log-likelihood for given mean vector and shape parameter.
#'
#' @param y Non-negative integer response vector.
#' @param mu Positive mean vector, same length as \code{y}.
#' @param theta Positive scalar shape parameter.
#' @param weights Optional observation weights (default 1).
#'
#' @return Scalar log-likelihood value.
#'
#' @details
#' The log-likelihood is
#' \deqn{\ell(\mu, \theta) = \sum_i w_i \bigl[
#'   \log\Gamma(y_i + \theta) - \log\Gamma(\theta) - \log\Gamma(y_i + 1)
#'   + \theta\log\theta - \theta\log(\mu_i + \theta)
#'   + y_i\log\mu_i - y_i\log(\mu_i + \theta)\bigr]}
#'
#' @examples
#' set.seed(1234)
#' mu <- exp(rnorm(50))
#' y <- rpois(50, mu)
#' loglik_negbin(y, mu, theta = 5)
#'
#' @export
loglik_negbin <- function(y, mu, theta, weights = 1) {

  N <- length(y)
  if(length(weights) == 1) weights <- rep(weights, N)

  logL <- lgamma(y + theta) - lgamma(theta) - lgamma(y + 1) +
    theta * log(theta) - theta * log(mu + theta) +
    y * log(mu) - y * log(mu + theta)

  sum(logL * weights)
}


#' Compute Negative Binomial Score Vector
#'
#' @description
#' Gradient of the NB2 log-likelihood with respect to
#' \eqn{\boldsymbol{\beta}} under the log link.
#'
#' @param X Design matrix (N x p).
#' @param y Response vector.
#' @param mu Mean vector \eqn{\exp(\mathbf{X}\boldsymbol{\beta})}.
#' @param theta Shape parameter.
#' @param weights Observation weights (default 1).
#'
#' @return Numeric column vector of length p (gradient).
#'
#' @details
#' Under log link, \eqn{d\mu_i/d\eta_i = \mu_i}, so the score is
#' \deqn{\mathbf{u} = \mathbf{X}^{\top}\mathbf{w}_{\mathrm{obs}}
#'   \odot \frac{(y_i - \mu_i)\theta}{\theta + \mu_i}}
#' where the per-observation contribution
#' \eqn{(y_i - \mu_i)\theta/(\theta + \mu_i)} is the derivative of the
#' log-likelihood with respect to \eqn{\eta_i}.
#'
#' @keywords internal
#' @export
score_negbin <- function(X, y, mu, theta, weights = 1) {

  N <- length(y)
  if(length(weights) == 1) weights <- rep(weights, N)

  ## d ell_i / d eta_i = (y_i - mu_i) * theta / (theta + mu_i)
  r <- weights * (y - mu) * theta / (theta + mu)

  crossprod(X, cbind(r))
}


#' Compute Negative Binomial Observed Information Matrix
#'
#' @description
#' Negative Hessian of the NB2 log-likelihood with respect to
#' \eqn{\boldsymbol{\beta}} under the log link.
#'
#' @param X Design matrix (N x p).
#' @param mu Mean vector.
#' @param theta Shape parameter.
#' @param weights Observation weights (default 1).
#'
#' @return Symmetric p x p observed information matrix.
#'
#' @details
#' The expected information under log link is
#' \deqn{\mathbf{I} = \mathbf{X}^{\top}\mathbf{W}\mathbf{X}}
#' where \eqn{W_{ii} = w_i \mu_i \theta / (\theta + \mu_i)}.
#' This is the IRLS weight for NB2 with log link.
#'
#' @keywords internal
#' @export
info_negbin <- function(X, y, mu, theta, weights = 1) {

  N <- length(mu)
  if(length(weights) == 1) weights <- rep(weights, N)

  ## Observed information under log link:
  #  J_bb = X^T diag(w_i * mu_i * theta * (theta + y_i)/(theta + mu_i)^2) X
  W <- weights * mu * theta * (theta + y) / (theta + mu)^2
  crossprod(X, W * X)
}


#' Estimate Negative Binomial Shape Parameter
#'
#' @description
#' Computes the profile MLE of the shape parameter \eqn{\theta} given
#' current mean estimates \eqn{\mu}.
#'
#' @param y Response vector.
#' @param mu Mean vector.
#' @param weights Observation weights (default 1).
#' @param init Optional initial value for \eqn{\theta}. If NULL, uses a
#'   moment-based estimate.
#'
#' @return Scalar MLE of \eqn{\theta}.
#'
#' @details
#' Maximizes the profile log-likelihood over \eqn{\theta} via Brent's
#' method on \eqn{[10^{-4}, 10^4]}.
#'
#' @examples
#' set.seed(1234)
#' mu <- rep(5, 200)
#' y <- rnbinom(200, size = 3, mu = 5)
#' negbin_theta(y, mu)
#'
#' @export
negbin_theta <- function(y, mu, weights = 1, init = NULL) {

  N <- length(y)
  if(length(weights) == 1) weights <- rep(weights, N)

  ## Moment-based initialization: theta = mu^2 / (var - mu)
  if(is.null(init)) {
    v <- max(weighted.mean((y - mu)^2, weights), mean(mu) + 0.1)
    m <- weighted.mean(mu, weights)
    init <- max(m^2 / max(v - m, 0.1), 0.1)
  }

  optim(
    init,
    fn = function(par) -loglik_negbin(y, mu, par, weights),
    method = 'Brent',
    lower = 1e-4,
    upper = 1e4
  )$par
}


#' Negative Binomial Family for lgspline
#'
#' @description
#' Creates a family-like object for NB2 regression. The link is log, the
#' variance function is \eqn{V(\mu) = \mu + \mu^2/\theta}, and the
#' dispersion stored by lgspline (\code{sigmasq_tilde}) is the shape
#' parameter \eqn{\theta}.
#'
#' @return A list with family components used by \code{lgspline}.
#'
#' @details
#' The NB2 model has a nuisance shape parameter \eqn{\theta} analogous
#' to the Weibull scale parameter. It is estimated jointly with
#' \eqn{\boldsymbol{\beta}} and its uncertainty is propagated via the
#' Schur complement correction.
#'
#' @examples
#' fam <- negbin_family()
#' fam$family
#' fam$link
#'
#' @export
negbin_family <- function() list(
  family = "negbin",
  link = "log",
  linkfun = log,
  linkinv = exp,

  ## Variance function: V(mu) = mu + mu^2/theta
  #  theta is stored in dispersion; when called from standard glm
  #  machinery without theta, return mu (Poisson-like) as fallback.
  variance = function(mu) mu,

  ## dev.resids for base R compatibility.
  #  Simplified version without theta; full version uses
  #  custom_dev.resids below.
  dev.resids = function(y, mu, wt) {
    ## Poisson-like deviance as fallback
    2 * wt * (y * log(pmax(y, 1) / mu) - (y - mu))
  },

  ## Custom deviance for GCV tuning.
  #  Receives full vectors and estimates theta internally.
  custom_dev.resids = function(y, mu, order_indices, family,
                               observation_weights) {

    wt <- c(observation_weights)

    ## Estimate theta given current mu
    theta <- negbin_theta(y, mu, wt)

    ## NB deviance: -2 * log-likelihood per observation
    dev <- -2 * (
      lgamma(y + theta) - lgamma(theta) - lgamma(y + 1) +
        theta * log(theta) - theta * log(mu + theta) +
        y * log(pmax(mu, .Machine$double.eps)) - y * log(mu + theta)
    )
    dev * wt
  },

  ## Log-likelihood extraction from fitted model
  loglik = function(model_fit) {
    y_fit  <- model_fit$y
    mu_fit <- model_fit$ytilde
    ## sigmasq_tilde stores theta directly
    theta <- model_fit$sigmasq_tilde

    wt <- if(!is.null(model_fit$weights)) model_fit$weights
    else rep(1, length(y_fit))

    if(is.null(theta) || !is.finite(theta) || theta <= 0) {
      warning("Shape parameter theta not found or invalid; ",
              "estimating from fitted values.")
      theta <- negbin_theta(y_fit, mu_fit, wt)
    }

    loglik_negbin(y_fit, mu_fit, theta, wt)
  },

  ## AIC = -2*loglik + 2*(edf + 1) where +1 is for theta
  aic = function(model_fit) {
    ll <- negbin_family()$loglik(model_fit)
    edf <- model_fit$edf
    if(is.null(edf)) edf <- model_fit$trace_XUGX
    if(is.null(edf)) {
      warning("Cannot compute AIC: effective degrees of freedom not found.")
      return(NA)
    }
    ## +1 for the shape parameter
    -2 * ll + 2 * (edf + 1)
  }
)


#' NB GLM Weight Function
#'
#' @description
#' Computes working weights for the NB2 information matrix used by
#' lgspline when updating \eqn{\mathbf{G}} after obtaining constrained
#' estimates.
#'
#' @param mu Predicted values \eqn{\exp(\eta)}.
#' @param y Observed counts.
#' @param order_indices Observation indices in partition order.
#' @param family NB family object (unused, for interface compatibility).
#' @param dispersion Shape parameter \eqn{\theta}.
#' @param observation_weights Observation weights.
#'
#' @return Numeric vector of working weights, length N.
#'
#' @details
#' The IRLS weight for NB2 with log link is
#' \deqn{W_i = w_i \mu_i \theta / (\theta + \mu_i)}
#'
#' Falls back to observation weights when natural weights are degenerate.
#'
#' @export
negbin_glm_weight_function <- function(mu,
                                       y,
                                       order_indices,
                                       family,
                                       dispersion,
                                       observation_weights) {

  theta <- dispersion
  N <- length(mu)
  if(length(observation_weights) == 1) {
    observation_weights <- rep(observation_weights, N)
  }

  ## Observed-information weight for J_bb
  val <- mu * theta * (theta + y) / (theta + mu)^2

  if(any(!is.finite(val)) || all(val == 0)) {
    return(rep(1, N))
  }

  val <- pmax(val, .Machine$double.eps)
  val * c(observation_weights)
}


#' NB Dispersion Function
#'
#' @description
#' Estimates the shape parameter \eqn{\theta} from current fitted values.
#' When a correlation structure is present (\code{VhalfInv} is non-NULL),
#' the Pearson residuals are whitened before computing the moment-based
#' initial value, giving a better starting point for the profile MLE
#' under correlated data. The final estimate is always the profile MLE
#' over \eqn{\theta}.
#'
#' @param mu Predicted values.
#' @param y Observed counts.
#' @param order_indices Observation indices.
#' @param family NB family object.
#' @param observation_weights Observation weights.
#' @param VhalfInv Inverse square root of the correlation matrix, or NULL
#'   for independent observations. When non-NULL, used to whiten residuals
#'   for the moment-based initialization of \eqn{\theta}.
#'
#' @return Scalar \eqn{\theta} estimate (stored as \code{sigmasq_tilde}).
#'
#' @details
#' The profile MLE maximizes \eqn{\ell(\theta \mid \mu)} via Brent's
#' method. When \code{VhalfInv} is provided, the Pearson residuals
#' \eqn{r_i = (y_i - \mu_i) / \sqrt{V(\mu_i)}} are pre-whitened as
#' \eqn{\tilde{r} = V^{-1/2} r} before computing the moment estimator
#' used for initialization. This accounts for the correlation structure
#' in the variance decomposition and produces a more stable starting
#' point for the optimizer, particularly when the correlation inflates
#' the marginal variance beyond what the NB model alone would predict.
#'
#' The profile MLE itself does not use \code{VhalfInv} because the NB
#' log-likelihood is a marginal quantity; the correlation structure
#' affects estimation only through the mean model (handled by the GEE
#' paths in \code{get_B}).
#'
#' @export
negbin_dispersion_function <- function(mu,
                                       y,
                                       order_indices,
                                       family,
                                       observation_weights,
                                       VhalfInv) {

  mu <- c(mu)
  y <- c(y)
  observation_weights <- c(observation_weights)
  N <- length(y)
  if(length(observation_weights) == 1) {
    observation_weights <- rep(observation_weights, N)
  }

  ## Moment-based initialization, optionally whitened for GEE
  init <- NULL
  if(!is.null(VhalfInv)) {
    ## Whiten Pearson residuals: r_i = (y_i - mu_i) / sqrt(mu_i)
    #  then tilde_r = V^{-1/2} r
    #  The whitened variance should be ~ 1 + mu/theta under the model,
    #  so theta_init = mean(mu) / max(var(tilde_r) - 1, 0.1)
    pear_r <- (y - mu) / sqrt(pmax(mu, .Machine$double.eps))
    tilde_r <- c(VhalfInv %**% cbind(pear_r))
    v_tilde <- weighted.mean(tilde_r^2, observation_weights)
    m_mu <- weighted.mean(mu, observation_weights)
    init <- max(m_mu / max(v_tilde - 1, 0.1), 0.1)
  }

  ## Profile MLE over theta
  negbin_theta(y, mu, observation_weights, init)
}


#' NB Score Function for Quadratic Programming and Blockfit
#'
#' @description
#' Computes the score (gradient of NB log-likelihood) in the format
#' expected by lgspline's \code{qp_score_function} interface.
#'
#' @param X Block-diagonal design matrix (N x P).
#' @param y Response vector (N x 1).
#' @param mu Predicted values (N x 1), same order as X and y.
#' @param order_list List of observation indices per partition.
#' @param dispersion Shape parameter \eqn{\theta}.
#' @param VhalfInv Inverse square root of correlation matrix; when non-NULL
#'   the score is computed on the whitened scale as
#'   \eqn{\tilde{X}^{\top}\tilde{r}} where
#'   \eqn{\tilde{X} = V^{-1/2}X} and the residual vector accounts for
#'   the correlation.
#' @param observation_weights Observation weights.
#'
#' @return Numeric column vector of length P.
#'
#' @details
#' Without correlation (\code{VhalfInv = NULL}), the score is
#' \eqn{\mathbf{X}^{\top}\mathbf{w} \odot (y - \mu)\theta/(\theta + \mu)}.
#'
#' With correlation, the GEE score is
#' \eqn{\tilde{\mathbf{X}}^{\top}\mathrm{diag}(\mathbf{W})
#' \mathbf{V}^{-1}(\mathbf{y} - \boldsymbol{\mu})}
#' where \eqn{\mathbf{W}} contains the NB working weights. The
#' whitening is absorbed by pre-multiplying both \eqn{\mathbf{X}} and
#' the residual by \eqn{\mathbf{V}^{-1/2}}.
#'
#' @export
negbin_qp_score_function <- function(X, y, mu, order_list, dispersion,
                                     VhalfInv, observation_weights) {

  theta <- dispersion
  N <- length(y)
  if(length(observation_weights) == 1) {
    observation_weights <- rep(observation_weights, N)
  }
  mu <- c(mu)
  y <- c(y)

  if(!is.null(VhalfInv)) {
    ## GEE score: X^T W V^{-1} (y - mu) where W is NB working weight
    #  On whitened scale: (V^{-1/2} X)^T * W * V^{-1/2}(y - mu)
    #  But the caller already passes X_tilde = V^{-1/2} X in the GEE
    #  paths, so X here may already be whitened. The mu passed is
    #  V^{-1/2} mu_original in GEE. We follow the Weibull convention:
    #  compute residual on original scale, then let the caller handle
    #  whitening.
    resid <- observation_weights * (y - mu) * theta / (theta + mu)
    return(crossprod(X, cbind(resid)))
  }

  ## Independent case
  resid <- observation_weights * (y - mu) * theta / (theta + mu)
  crossprod(X, cbind(resid))
}


#' NB Schur Correction
#'
#' @description
#' Computes the Schur complement correction to account for uncertainty in
#' estimating \eqn{\theta}. Structure is identical to
#' \code{\link{weibull_schur_correction}}: the joint information is
#' partitioned into \eqn{(\boldsymbol{\beta}, \theta)} blocks and the
#' correction is \eqn{-\mathbf{I}_{\beta\theta}
#' I_{\theta\theta}^{-1}\mathbf{I}_{\beta\theta}^{\top}}.
#'
#' @param X List of partition design matrices.
#' @param y List of partition response vectors.
#' @param B List of partition coefficient vectors.
#' @param dispersion Scalar shape parameter \eqn{\theta}.
#' @param order_list List of observation indices per partition.
#' @param K Number of knots.
#' @param family Family object.
#' @param observation_weights Observation weights.
#'
#' @return List of K+1 correction matrices, with \code{0} for empty
#'   partitions.
#'
#' @details
#' The cross-derivative (score of \eqn{\eta_i} w.r.t. \eqn{\theta}) is
#' \deqn{\frac{\partial^2 \ell}{\partial \eta_i \partial \theta}
#'   = \frac{(y_i - \mu_i)\mu_i}{(\theta + \mu_i)^2}}
#' The second derivative of the log-likelihood w.r.t. \eqn{\theta} is
#' \deqn{I_{\theta\theta} = -\sum_i w_i \bigl[
#'   \psi'(y_i + \theta) - \psi'(\theta) +
#'   \frac{1}{\theta} - \frac{2}{\mu_i + \theta} +
#'   \frac{y_i + \theta}{(\mu_i + \theta)^2}\bigr]}
#' where \eqn{\psi'} is the trigamma function.
#'
#' @export
negbin_schur_correction <- function(X, y, B, dispersion, order_list, K,
                                    family, observation_weights) {

  theta <- dispersion
  lapply(1:(K + 1), function(k) {
    if(nrow(X[[k]]) < 1) {
      return(0)
    } else {
      mu <- family$linkinv(c(X[[k]] %**% B[[k]]))
      obs <- c(y[[k]])
      weights <- c(observation_weights[[k]])

      ## Cross-derivative: d^2 ell / (d eta_i d theta)
      #  = (y_i - mu_i) * mu_i / (theta + mu_i)^2
      I_bt <- crossprod(X[[k]], cbind(
        weights * (obs - mu) * mu / (theta + mu)^2
      ))

      ## Information for theta (negative of second derivative)
      I_tt <- -sum(weights * (
        trigamma(obs + theta) - trigamma(theta) +
          1 / theta -
          2 / (mu + theta) +
          (obs + theta) / (mu + theta)^2
      ))

      ## Schur complement: -I_bt (1/I_tt) I_bt^T
      #  Gets elementwise added to G[[k]]^{-1}, so should be negative
      #  semi-definite for well-behaved data (I_tt > 0 implies
      #  compl is negative semi-definite).
      if(abs(I_tt) < .Machine$double.eps) {
        return(0)
      }
      compl <- I_bt %**% matrix(-1 / I_tt) %**% t(I_bt)
      return(compl)
    }
  })
}


#' Unconstrained NB Estimation for lgspline
#'
#' @description
#' Estimates penalized NB2 coefficients for a single partition (or the
#' full model when K = 0) using a two-stage optimization: outer loop
#' profiles over \eqn{\theta} via Brent's method, inner loop estimates
#' \eqn{\boldsymbol{\beta}} via damped Newton-Raphson on the penalized
#' log-likelihood.
#'
#' @param X Design matrix (N_k x p) for partition k.
#' @param y Response counts for partition k.
#' @param LambdaHalf Square root of penalty matrix.
#' @param Lambda Penalty matrix.
#' @param keep_weighted_Lambda Logical; if TRUE, return hot-start estimates
#'   from augmented Poisson regression without refinement.
#' @param family NB family object.
#' @param tol Convergence tolerance.
#' @param K Number of knots.
#' @param parallel Logical for parallel processing.
#' @param cl Cluster object.
#' @param chunk_size, num_chunks, rem_chunks Parallelism parameters.
#' @param order_indices Observation indices mapping partition to full data.
#' @param weights Observation weights.
#'
#' @return Numeric column vector of penalized coefficient estimates.
#'
#' @details
#' The penalized log-likelihood is
#' \deqn{\ell_p(\boldsymbol{\beta}, \theta) = \ell(\boldsymbol{\beta},
#'   \theta) - \tfrac{1}{2}\boldsymbol{\beta}^{\top}
#'   \boldsymbol{\Lambda}\boldsymbol{\beta}}
#'
#' The outer loop optimizes \eqn{\theta} given the inner-loop-optimal
#' \eqn{\boldsymbol{\beta}(\theta)}. This mirrors the Weibull AFT
#' approach where scale (\eqn{\sigma}) is profiled out.
#'
#' Initialization uses Poisson regression coefficients as a hot start
#' (equivalent to \eqn{\theta \to \infty}).
#'
#' @keywords internal
#' @export
unconstrained_fit_negbin <- function(X, y, LambdaHalf, Lambda,
                                     keep_weighted_Lambda, family,
                                     tol = 1e-8, K, parallel, cl,
                                     chunk_size, num_chunks, rem_chunks,
                                     order_indices, weights) {

  ## Handle empty partitions
  if(nrow(X) == 0) {
    return(cbind(rep(0, ncol(X))))
  }

  N <- nrow(X)
  p <- ncol(X)

  ## Observation weights
  if(any(!is.null(weights))) {
    weights <- c(weights)
  } else {
    weights <- rep(1, N)
  }

  y <- c(y)

  ## Hot-start via Poisson (theta -> Inf limit of NB)
  pois_init <- try({
    glm.fit(
      x = rbind(X, LambdaHalf),
      y = c(y, rep(1, nrow(LambdaHalf))),
      family = poisson(link = 'log'),
      weights = c(weights, rep(mean(weights), nrow(LambdaHalf)))
    )$coefficients
  }, silent = TRUE)

  if(inherits(pois_init, 'try-error') || any(!is.finite(pois_init))) {
    pois_init <- c(log(max(mean(y), 0.1)), rep(0, p - 1))
  }

  if(keep_weighted_Lambda) {
    return(cbind(pois_init))
  }

  ## Initial theta from Poisson residuals
  mu_init <- c(exp(X %**% cbind(pois_init)))
  theta_init <- negbin_theta(y, mu_init, weights)

  ## Outer loop: profile over theta via Brent
  theta <- optim(
    theta_init,
    fn = function(par) {
      th <- par
      ## Inner loop: optimize beta given theta
      beta <- damped_newton_r(
        pois_init,
        function(b) {
          mu <- c(exp(X %**% cbind(b)))
          loglik_negbin(y, mu, th, weights) -
            0.5 * c(crossprod(b, Lambda) %**% cbind(b))
        },
        function(b) {
          mu <- c(exp(X %**% cbind(b)))
          c(score_negbin(X, y, mu, th, weights) -
              Lambda %**% cbind(b))
        },
        function(b) {
          mu <- c(exp(X %**% cbind(b)))
          info_negbin(X, y, mu, th, weights) + Lambda
        },
        tol
      )
      mu <- c(exp(X %**% cbind(beta)))
      -loglik_negbin(y, mu, par, weights) +
        0.5 * c(crossprod(beta, Lambda) %**% cbind(beta))
    },
    lower = theta_init / 10,
    upper = theta_init * 10,
    method = 'Brent'
  )$par

  ## Final beta given optimal theta
  beta <- cbind(damped_newton_r(
    pois_init,
    function(b) {
      mu <- c(exp(X %**% cbind(b)))
      loglik_negbin(y, mu, theta, weights) -
        0.5 * c(crossprod(b, Lambda) %**% cbind(b))
    },
    function(b) {
      mu <- c(exp(X %**% cbind(b)))
      c(score_negbin(X, y, mu, theta, weights) -
          Lambda %**% cbind(b))
    },
    function(b) {
      mu <- c(exp(X %**% cbind(b)))
      info_negbin(X, y, mu, theta, weights) + Lambda
    },
    tol
  ))

  return(beta)
}


#' Fit Negative Binomial Model via lgspline
#'
#' @description
#' Convenience wrapper that calls \code{lgspline} with the correct
#' family, weight, dispersion, score, and unconstrained-fit functions for
#' NB2 regression. All standard lgspline arguments (knots, penalties,
#' constraints, parallel, correlation structures, etc.) are passed through.
#'
#' @param formula Formula specifying the model. The response should be
#'   non-negative integer counts.
#' @param data Data frame.
#' @param ... Additional arguments passed to \code{lgspline}, including
#'   \code{Vhalf} and \code{VhalfInv} for correlation structures.
#'
#' @return An object of class \code{"lgspline"}.
#'
#' @details
#' Internally sets:
#' \itemize{
#'   \item \code{family = negbin_family()}
#'   \item \code{unconstrained_fit_fxn = unconstrained_fit_negbin}
#'   \item \code{glm_weight_function = negbin_glm_weight_function}
#'   \item \code{qp_score_function = negbin_qp_score_function}
#'   \item \code{dispersion_function = negbin_dispersion_function}
#'   \item \code{schur_correction_function = negbin_schur_correction}
#'   \item \code{need_dispersion_for_estimation = TRUE}
#'   \item \code{estimate_dispersion = TRUE}
#'   \item \code{standardize_response = FALSE}
#' }
#'
#' A formula interface is needed, e.g. \code{lgspline_negbin(t, y)} won't work,
#' unlike for ordinary \code{\link{lgspline}}.
#'
#' When a correlation structure is supplied via \code{Vhalf}/\code{VhalfInv},
#' the model is fitted through the GEE Path 1b machinery in \code{get_B}.
#' The dispersion function uses \code{VhalfInv} to whiten Pearson
#' residuals for a better moment-based initialization of \eqn{\theta},
#' which stabilizes the profile MLE under moderate to strong correlation.
#' The score function handles the whitened design consistently with the
#' Weibull AFT GEE convention.
#'
#'
#' @examples
#'
#' set.seed(1234)
#' N <- 300
#' t <- rnorm(N)
#' mu <- exp(1 + 0.5 * sin(2 * t))
#' y <- rnbinom(N, size = 3, mu = mu)
#' df <- data.frame(response = y, predictor = t)
#'
#' fit <- lgspline_negbin(
#'   response ~ spl(predictor),
#'   data = df,
#'   K = 2,
#'   opt = FALSE,
#'   wiggle_penalty = 1e-2
#' )
#' print(summary(fit))
#' plot(fit, show_formulas = TRUE,
#'      custom_response_lab = 'Count')
#' points(t, mu, col = 'grey', cex=0.67)
#'
#' @seealso \code{\link{lgspline_cox}} for Cox PH,
#'   \code{\link{lgspline_weibull}} for Weibull AFT,
#'   \code{\link{negbin_family}}, \code{\link{unconstrained_fit_negbin}}
#'
#' @export
lgspline_negbin <- function(formula, data, ...) {
  lgspline(
    formula = formula,
    data = data,
    family = negbin_family(),
    unconstrained_fit_fxn = unconstrained_fit_negbin,
    glm_weight_function = negbin_glm_weight_function,
    qp_score_function = negbin_qp_score_function,
    dispersion_function = negbin_dispersion_function,
    schur_correction_function = negbin_schur_correction,
    need_dispersion_for_estimation = TRUE,
    estimate_dispersion = TRUE,
    standardize_response = FALSE,
    ...
  )
}

test_that("Negative Binomial helpers", {
  ## Test 1: Log-likelihood at known values matches manual computation
  set.seed(1234)
  N <- 50
  y <- rpois(N, 5)
  mu <- rep(5, N)
  theta <- 3
  ll_manual <- sum(
    lgamma(y + theta) - lgamma(theta) - lgamma(y + 1) +
      theta * log(theta) - theta * log(mu + theta) +
      y * log(mu) - y * log(mu + theta)
  )
  ll_fn <- loglik_negbin(y, mu, theta)
  expect_equal(ll_fn, ll_manual, tolerance = 1e-4)

  ## Test 2: Score at MLE should be approximately zero
  set.seed(42)
  N <- 300
  t1 <- rnorm(N)
  t2 <- rbinom(N, 1, 0.5)
  true_theta <- 4
  mu_true <- exp(1 + 0.3 * t1 + 0.2 * t2)
  y <- rnbinom(N, size = true_theta, mu = mu_true)
  X_s <- cbind(1, t1, t2)
  b <- cbind(c(log(mean(y)), 0, 0))
  ## Profile: alternate theta and beta
  for(outer in 1:10) {
    mu_cur <- c(exp(X_s %**% b))
    th <- negbin_theta(y, mu_cur)
    for(i in 1:20) {
      mu_cur <- c(exp(X_s %**% b))
      sc <- score_negbin(X_s, y, mu_cur, th)
      I <- info_negbin(X_s, y, mu_cur, th)
      step <- solve(I, sc)
      b <- b + c(step)
    }
  }
  mu_final <- c(exp(X_s %**% b))
  th_final <- negbin_theta(y, mu_final)
  sc_final <- score_negbin(X_s, y, mu_final, th_final)
  expect_true(all(abs(sc_final) < 1e-4))

  ## Test 3: Compare with MASS::glm.nb
  skip_if_not_installed("MASS")
  df_test <- data.frame(y = y, t1 = t1, t2 = t2)
  nbfit <- MASS::glm.nb(y ~ t1 + t2, data = df_test)
  b_mass <- coef(nbfit)
  names(b_mass) <- NULL
  expect_equal(c(b), b_mass, tolerance = 5e-2)
  ## Theta comparison
  expect_equal(th_final, nbfit$theta, tolerance = 1)

  ## Test 4: Information matrix is positive definite at the MLE
  I_mle <- info_negbin(X_s, y, mu_final, th_final)
  eigvals <- eigen(I_mle, symmetric = TRUE, only.values = TRUE)$values
  expect_true(all(eigvals > 0))

  ## Test 5: Log-likelihood increases along Newton direction
  b0 <- cbind(c(log(mean(y)), 0, 0))
  mu0 <- c(exp(X_s %**% b0))
  th0 <- negbin_theta(y, mu0)
  ll0 <- loglik_negbin(y, mu0, th0)
  sc0 <- score_negbin(X_s, y, mu0, th0)
  I0 <- info_negbin(X_s, y, mu0, th0)
  step0 <- solve(I0, sc0)
  b1 <- b0 + c(step0)
  mu1 <- c(exp(X_s %**% b1))
  th1 <- negbin_theta(y, mu1)
  ll1 <- loglik_negbin(y, mu1, th1)
  expect_gt(ll1, ll0)

  ## Test 6: unconstrained_fit_negbin recovers MLE (no penalty)
  p <- 3
  Lambda_zero <- matrix(0, p, p)
  LambdaHalf_zero <- matrix(0, p, p)
  b_fit <- unconstrained_fit_negbin(
    X = X_s, y = y,
    LambdaHalf = LambdaHalf_zero, Lambda = Lambda_zero,
    keep_weighted_Lambda = FALSE, family = negbin_family(),
    tol = 1e-10, K = 0,
    parallel = FALSE, cl = NULL,
    chunk_size = NULL, num_chunks = NULL, rem_chunks = NULL,
    order_indices = 1:N, weights = rep(1, N)
  )
  expect_equal(c(b_fit), c(b), tolerance = 5e-2)

  ## Test 7: negbin_qp_score_function near zero at MLE
  mu_test <- exp(c(X_s %**% b))
  order_list_test <- list(1:N)
  sc_qp <- negbin_qp_score_function(
    X = X_s, y = y, mu = mu_test,
    order_list = order_list_test, dispersion = th_final,
    VhalfInv = NULL, observation_weights = rep(1, N)
  )
  expect_true(all(abs(sc_qp) < 1e-4))

  ## Test 8: Weighted log-likelihood
  set.seed(99)
  N_wt <- 30
  y_wt <- rpois(N_wt, 3)
  mu_wt <- rep(3, N_wt)
  th_wt <- 5
  wt_wt <- runif(N_wt, 0.5, 2.0)
  ll_manual <- sum(wt_wt * (
    lgamma(y_wt + th_wt) - lgamma(th_wt) - lgamma(y_wt + 1) +
      th_wt * log(th_wt) - th_wt * log(mu_wt + th_wt) +
      y_wt * log(mu_wt) - y_wt * log(mu_wt + th_wt)
  ))
  ll_fn <- loglik_negbin(y_wt, mu_wt, th_wt, wt_wt)
  expect_equal(ll_fn, ll_manual, tolerance = 1e-4)
  ll_unit <- loglik_negbin(y_wt, mu_wt, th_wt, rep(1, N_wt))
  ll_default <- loglik_negbin(y_wt, mu_wt, th_wt)
  expect_equal(ll_unit, ll_default, tolerance = 1e-4)

  ## Test 9: Dispersion function returns positive theta
  th_est <- negbin_dispersion_function(
    mu = mu_test, y = y, order_indices = 1:N,
    family = negbin_family(),
    observation_weights = rep(1, N),
    VhalfInv = NULL
  )
  expect_true(th_est > 0)
  expect_true(is.finite(th_est))

  ## Test 10: GLM weight function returns positive values
  wt_vals <- negbin_glm_weight_function(
    mu = mu_test, y = y, order_indices = 1:N,
    family = negbin_family(), dispersion = th_final,
    observation_weights = rep(1, N)
  )
  expect_true(all(wt_vals > 0))
  expect_length(wt_vals, N)

  ## Test 11: negbin_theta recovers known theta from large sample
  set.seed(2025)
  N_big <- 5000
  mu_big <- rep(10, N_big)
  theta_true <- 3.0
  y_big <- rnbinom(N_big, size = theta_true, mu = 10)
  th_hat <- negbin_theta(y_big, mu_big)
  expect_equal(th_hat, theta_true, tolerance = 0.5)

  ## Test 12: Schur correction is negative semi-definite
  X_test <- cbind(1, t1, t2)
  B_test <- list(b)
  order_list_sc <- list(1:N)
  obs_wt_sc <- list(rep(1, N))
  schur <- negbin_schur_correction(
    X = list(X_test), y = list(y),
    B = B_test, dispersion = th_final,
    order_list = order_list_sc, K = 0,
    family = negbin_family(),
    observation_weights = obs_wt_sc
  )
  ## Should be a matrix (not 0) for nondegenerate data
  expect_true(is.matrix(schur[[1]]))
  ## Eigenvalues of Schur correction should be <= 0 (negative semi-definite)
  eig_schur <- eigen(schur[[1]], symmetric = TRUE, only.values = TRUE)$values
  expect_true(all(eig_schur <= sqrt(.Machine$double.eps)))

  ## Test 13: Score with weights is consistent
  set.seed(77)
  wt_sc <- runif(N, 0.5, 2.0)
  mu_sc <- c(exp(X_s %**% b))
  sc_weighted <- score_negbin(X_s, y, mu_sc, th_final, wt_sc)
  ## Manual: X^T (wt * (y - mu) * theta / (theta + mu))
  r_manual <- wt_sc * (y - mu_sc) * th_final / (th_final + mu_sc)
  sc_manual <- crossprod(X_s, cbind(r_manual))
  expect_equal(c(sc_weighted), c(sc_manual), tolerance = 1e-4)

  ## Test 14: Info with weights matches X^T W X manually
  W_manual <- wt_sc * mu_sc * th_final * (th_final + y) / (th_final + mu_sc)^2
  I_manual <- crossprod(X_s, W_manual * X_s)
  I_fn <- info_negbin(X_s, y, mu_sc, th_final, wt_sc)
  expect_equal(I_fn, I_manual, tolerance = 1e-4)

  ## Test 15: Dispersion function with VhalfInv gives valid result
  N_gee <- 50
  y_gee <- rnbinom(N_gee, size = 3, mu = 5)
  mu_gee <- rep(5, N_gee)
  rho <- 0.2
  ## Simple diagonal perturbation for testing
  VhalfInv_test <- diag(1 / sqrt(1 - rho), N_gee, N_gee)
  th_gee <- negbin_dispersion_function(
    mu = mu_gee, y = y_gee, order_indices = 1:N_gee,
    family = negbin_family(),
    observation_weights = rep(1, N_gee),
    VhalfInv = VhalfInv_test
  )
  expect_true(th_gee > 0)
  expect_true(is.finite(th_gee))

  ## Test 16: Poisson limit (theta -> Inf gives Poisson log-likelihood)
  set.seed(88)
  N_pois <- 100
  mu_pois <- rep(3, N_pois)
  y_pois <- rpois(N_pois, 3)
  ll_nb_big <- loglik_negbin(y_pois, mu_pois, theta = 1e8)
  ll_pois <- sum(dpois(y_pois, mu_pois, log = TRUE))
  expect_equal(ll_nb_big, ll_pois, tolerance = 1e-2)

  ## Test 17: keep_weighted_Lambda returns Poisson hot-start
  b_hot <- unconstrained_fit_negbin(
    X = X_s, y = y,
    LambdaHalf = diag(0.01, p), Lambda = diag(0.01^2, p),
    keep_weighted_Lambda = TRUE, family = negbin_family(),
    tol = 1e-8, K = 0,
    parallel = FALSE, cl = NULL,
    chunk_size = NULL, num_chunks = NULL, rem_chunks = NULL,
    order_indices = 1:N, weights = rep(1, N)
  )
  expect_length(b_hot, p)
  expect_true(all(is.finite(b_hot)))

  ## Test 18: Family object components exist and are correct
  fam <- negbin_family()
  expect_equal(fam$family, "negbin")
  expect_equal(fam$link, "log")
  expect_equal(fam$linkfun(exp(1)), 1)
  expect_equal(fam$linkinv(0), 1)
  expect_true(is.function(fam$loglik))
  expect_true(is.function(fam$aic))
  expect_true(is.function(fam$dev.resids))
  expect_true(is.function(fam$custom_dev.resids))

  ## Test 19: Schur correction returns 0 for empty partition
  schur_empty <- negbin_schur_correction(
    X = list(matrix(0, nrow = 0, ncol = 3), X_test),
    y = list(numeric(0), y),
    B = list(cbind(rep(0, 3)), b),
    dispersion = th_final,
    order_list = list(integer(0), 1:N),
    K = 1,
    family = negbin_family(),
    observation_weights = list(numeric(0), rep(1, N))
  )
  expect_identical(schur_empty[[1]], 0)
  expect_true(is.matrix(schur_empty[[2]]))

  ## Test 20: Numerical gradient check for score_negbin
  set.seed(123)
  X_ng <- cbind(1, rnorm(80))
  y_ng <- rnbinom(80, size = 2, mu = 3)
  b_ng <- c(log(3), 0.1)
  th_ng <- 2
  eps_ng <- 1e-6
  sc_analytic <- score_negbin(X_ng, y_ng, exp(X_ng %**% cbind(b_ng)), th_ng)
  sc_numerical <- sapply(1:2, function(j) {
    bp <- bm <- b_ng
    bp[j] <- bp[j] + eps_ng
    bm[j] <- bm[j] - eps_ng
    (loglik_negbin(y_ng, exp(X_ng %**% cbind(bp)), th_ng) -
        loglik_negbin(y_ng, exp(X_ng %**% cbind(bm)), th_ng)) / (2 * eps_ng)
  })
  expect_equal(c(sc_analytic), sc_numerical, tolerance = 1e-4)
})

#' lgspline: S3 Methods
#'
#' @description
#' S3 methods for lgspline objects: print, summary, coef, plot, predict,
#' confint, logLik, and inference helpers.
#'
#' @docType methods
#' @keywords internal
#' @name lgspline-methods
#' @rdname lgspline-methods
#' @aliases lgspline-methods
NULL

#' Print Method for lgspline Objects
#'
#' Prints a concise summary of the fitted model to the console.
#'
#' @param x An lgspline model object.
#' @param ... Not used.
#'
#' @return Invisibly returns \code{x}.
#'
#' @export
#' @method print lgspline
print.lgspline <- function(x, ...) {
  cat("Lagrangian Multiplier Smoothing Spline Model\n")
  cat("============================================\n")
  cat("Model Family (Link Function):",
      paste0(paste0(x$family)[1],
             " (",
             paste0(x$family)[2],
             ")", collapse = ""),
      "\n")
  cat("Number of Observations:", x$N, "\n")
  cat("Number of Predictors:", x$q, "\n")
  cat("Number of Partitions:", x$K + 1, "\n")
  cat("Basis Functions per Partition:", x$p, "\n")
  invisible(x)
}


#' Summary Method for lgspline Objects
#'
#' @param object An lgspline model object.
#' @param ... Not used.
#'
#' @return An object of class \code{summary.lgspline}, a list containing:
#' \describe{
#'   \item{model_family}{The \code{\link[stats]{family}} object.}
#'   \item{observations}{Number of observations N.}
#'   \item{predictors}{Number of predictor variables q.}
#'   \item{knots}{Number of knots K.}
#'   \item{basis_functions}{Basis functions per partition p.}
#'   \item{estimate_dispersion}{"Yes" or "No".}
#'   \item{cv}{Critical value used for confidence intervals.}
#'   \item{coefficients}{Coefficient matrix from \code{\link{wald_univariate}},
#'         or a single-column estimate matrix if \code{return_varcovmat = FALSE}.}
#'   \item{sigmasq_tilde}{Estimated dispersion \eqn{\tilde{\sigma}^2}.}
#'   \item{trace_XUGX}{Trace of the hat matrix \eqn{\mathrm{trace}(\mathbf{XUGX}^\top)}.}
#'   \item{N}{Number of observations.}
#' }
#'
#' @export
summary.lgspline <- function(object, ...) {
  summary_list <- list(
    model_family = object$family,
    observations = object$N,
    predictors = object$q,
    knots = object$K,
    basis_functions = object$p,
    estimate_dispersion = ifelse(!is.null(object$estimate_dispersion) && object$estimate_dispersion &&
                                   !is.null(object$sigmasq_tilde) && object$sigmasq_tilde != 1,
                                 'Yes',
                                 'No'),
    cv = object$critical_value,
    coefficients = NULL,
    sigmasq_tilde = object$sigmasq_tilde,
    trace_XUGX = object$trace_XUGX,
    N = object$N
  )

  ## Wald inference table (requires return_varcovmat = TRUE)
  #  Column names are compatible with stats::printCoefmat()
  if(!is.null(object$return_varcovmat) &&
     object$return_varcovmat &&
     !is.null(object$wald_univariate)){
    tr <- try({
      wald_obj <- wald_univariate(object)
      wald_obj$coefficients
    }, silent = TRUE)

    if(!inherits(tr, 'try-error') && !is.null(tr)){
      summary_list$coefficients <- tr
    } else {
      summary_list$coefficients <- cbind(Estimate = unlist(object$B))
    }
  } else {
    summary_list$coefficients <- cbind(Estimate = unlist(object$B))
  }

  class(summary_list) <- "summary.lgspline"
  return(summary_list)
}


#' Print Method for lgspline Summaries
#'
#' Displays a formatted model summary using \code{\link[stats]{printCoefmat}}
#' for the coefficient table.
#'
#' @param x A \code{summary.lgspline} object.
#' @param ... Not used.
#'
#' @return Invisibly returns \code{x}.
#'
#' @seealso \code{\link[stats]{printCoefmat}}
#' @export
print.summary.lgspline <- function(x, ...) {
  cat("Lagrangian Multiplier Smoothing Spline Model Summary\n")
  cat("====================================================\n")
  cat("Model Family:", x$model_family[[1]], "\n")
  cat("Model Family (Link Function):",
      paste0(paste0(x$model_family)[1],
             " (",
             paste0(x$model_family)[2],
             ")", collapse = ""),
      "\n")
  cat("Observations:", x$observations, "\n")
  cat("Predictors:", x$predictors, "\n")
  cat("Partitions:", x$knots + 1, "\n")
  cat("Basis Functions per Partition:", x$basis_functions, "\n")
  if(length(unlist(x$coefficients)) > 1 && ncol(x$coefficients) > 1){
    cat("----------------------------------------------------\n")
    cat("Univariate Inference: \n")
    ## Reorder so p-value column is last, as required by printCoefmat
    pval_col <- grep("^Pr\\(", colnames(x$coefficients))
    if(length(pval_col) > 0){
      other_cols <- setdiff(seq_len(ncol(x$coefficients)), pval_col)
      print_mat <- x$coefficients[, c(other_cols, pval_col), drop = FALSE]
      pval_col_new <- ncol(print_mat)
      tst_col <- grep("value$", colnames(print_mat))
      tst_col <- tst_col[!tst_col %in% pval_col_new]
      stats::printCoefmat(print_mat,
                          cs.ind = 1:2,
                          tst.ind = if(length(tst_col) > 0) tst_col[1] else NULL,
                          P.values = TRUE,
                          has.Pvalue = TRUE,
                          signif.stars = TRUE)
    } else {
      print(x$coefficients)
    }
    cat('\n')
    cat("Dispersion:", x$sigmasq_tilde, "\n")
    if(!is.null(x$N) && !is.null(x$trace_XUGX)){
      cat("Effective degrees of freedom:", x$N - x$trace_XUGX, "\n")
    } else {
      cat("Effective degrees of freedom: Not Available\n")
    }
    cat("Critical value for confidence intervals: ", x$cv, "\n")
    cat("----------------------------------------------------\n")
  }
  invisible(x)
}


#' Find the Extremum of a Fitted lgspline
#'
#' Finds the global maximum or minimum of a fitted lgspline using L-BFGS-B,
#' with options for partition-based heuristics, stochastic exploration, and
#' custom objective functions (e.g., acquisition functions for Bayesian
#' optimization).
#'
#' @param object A fitted lgspline model object.
#' @param vars Integer or character vector; indices or names of predictors to
#'        optimize over. Default NULL optimizes all predictors.
#' @param quick_heuristic Logical; if TRUE (default) searches only the
#'        best-performing partition. If FALSE, initiates searches from all
#'        partition local maxima.
#' @param initial Numeric vector; optional starting values. Useful for
#'        fixing binary predictors. Default NULL.
#' @param B_predict List; optional coefficient list for prediction, e.g.
#'        from \code{\link{generate_posterior}}. Default NULL uses
#'        \code{object$B}.
#' @param minimize Logical; find minimum instead of maximum. Default FALSE.
#' @param stochastic Logical; add noise during optimization for exploration.
#'        Default FALSE.
#' @param stochastic_draw Function; generates noise for stochastic
#'        optimization. Takes \code{mu}, \code{sigma}, and \code{...}.
#'        Default \code{rnorm(length(mu), mu, sigma)}.
#' @param sigmasq_predict Numeric; variance for stochastic draws.
#'        Default \code{object$sigmasq_tilde}.
#' @param custom_objective_function Function; optional custom objective.
#'        Takes \code{mu}, \code{sigma}, \code{y_best}, \code{...}.
#'        Default NULL.
#' @param custom_objective_derivative Function; optional gradient of
#'        \code{custom_objective_function}. Takes \code{mu}, \code{sigma},
#'        \code{y_best}, \code{d_mu}, \code{...}. Default NULL.
#' @param ... Additional arguments passed to internal optimization routines.
#'
#' @return A list with elements:
#' \describe{
#'   \item{t}{Numeric vector; predictor values at the extremum.}
#'   \item{y}{Numeric; objective value at the extremum.}
#' }
#'
#' @examples
#'
#' set.seed(1234)
#' t <- runif(1000, -10, 10)
#' y <- 2*sin(t) + -0.06*t^2 + rnorm(length(t))
#' model_fit <- lgspline(t, y)
#' plot(model_fit)
#'
#' max_point <- find_extremum(model_fit)
#' min_point <- find_extremum(model_fit, minimize = TRUE)
#' abline(v = max_point$t, col = 'blue')
#' abline(v = min_point$t, col = 'red')
#'
#' ## Expected improvement acquisition function
#' ei_obj <- function(mu, sigma, y_best, ...) {
#'   d <- y_best - mu
#'   d * pnorm(d/sigma) + sigma * dnorm(d/sigma)
#' }
#' ei_deriv <- function(mu, sigma, y_best, d_mu, ...) {
#'   d <- y_best - mu
#'   z <- d/sigma
#'   d_z <- -d_mu/sigma
#'   pnorm(z)*d_mu - d*dnorm(z)*d_z + sigma*z*dnorm(z)*d_z
#' }
#'
#' post_draw <- generate_posterior(model_fit)
#' acq <- find_extremum(model_fit,
#'                      stochastic = TRUE,
#'                      B_predict = post_draw$post_draw_coefficients,
#'                      sigmasq_predict = post_draw$post_draw_sigmasq,
#'                      custom_objective_function = ei_obj,
#'                      custom_objective_derivative = ei_deriv)
#' abline(v = acq$t, col = 'green')
#'
#' @seealso \code{\link{lgspline}}, \code{\link{generate_posterior}}
#' @export
find_extremum <- function(object,
                          vars = NULL,
                          quick_heuristic = TRUE,
                          initial = NULL,
                          B_predict = NULL,
                          minimize = FALSE,
                          stochastic = FALSE,
                          stochastic_draw = function(mu, sigma, ...){
                            N <- length(mu)
                            rnorm(N, mu, sigma)
                          },
                          sigmasq_predict = object$sigmasq_tilde,
                          custom_objective_function = NULL,
                          custom_objective_derivative = NULL,
                          ...) {
  internal_find_extremum_func <- object$find_extremum
  if (!is.null(internal_find_extremum_func) && is.function(internal_find_extremum_func)) {
    return(internal_find_extremum_func(
      vars = vars,
      quick_heuristic = quick_heuristic,
      initial = initial,
      B_predict = B_predict,
      minimize = minimize,
      stochastic = stochastic,
      stochastic_draw = stochastic_draw,
      sigmasq_predict = sigmasq_predict,
      custom_objective_function = custom_objective_function,
      custom_objective_derivative = custom_objective_derivative,
      ...
    ))
  } else {
    stop("Internal find_extremum method not found or not a function.")
  }
}


#' Generate Posterior Samples from a Fitted lgspline
#'
#' Draws from the posterior distribution of model coefficients, with optional
#' dispersion sampling, posterior predictive draws, and propagation of
#' uncertainty in estimated correlation parameters.
#'
#' Uses a Laplace approximation centred at the MAP estimate for non-Gaussian
#' responses.
#'
#' @details
#' \strong{Dispersion posterior.}
#' When \code{draw_dispersion = TRUE}, \eqn{\sigma^2} is drawn from
#' \deqn{
#'   \sigma^2 \mid \mathbf{y} \sim
#'   \mathrm{InvGamma}(\alpha_1, \alpha_2),
#' }
#' where
#' \deqn{
#'   \alpha_1 = \theta_1 + \tfrac{1}{2}(N - s \cdot \mathrm{tr}(\mathbf{H})),
#'   \quad
#'   \alpha_2 = \theta_2 + \tfrac{1}{2}(N - s \cdot \mathrm{tr}(\mathbf{H}))
#'              \tilde{\sigma}^2,
#' }
#' \eqn{\mathbf{H} = \mathbf{XUGX}^\top} is the hat matrix, \eqn{s = 1}
#' when \code{unbias_dispersion = TRUE} (else \eqn{s = 0}), and
#' \eqn{\theta_1 = \theta_2 = 0} recovers an improper uniform prior.
#'
#' \strong{Correlation parameter posterior.}
#' When \code{draw_correlation = TRUE} and the fitted model contains an
#' estimated correlation structure, each draw first samples
#' \eqn{\boldsymbol{\rho}} from
#' \deqn{
#'   \boldsymbol{\rho}^{(m)} \sim
#'   \mathcal{N}(\hat{\boldsymbol{\rho}}_{\mathrm{REML}},
#'               \mathbf{H}^{-1}_{\mathrm{BFGS}}),
#' }
#' rebuilds the posterior covariance under the drawn correlation
#' structure (reusing all pre-computed design matrices, constraints, and
#' penalty matrices) and then draws coefficients from the updated
#' posterior. Knot placement, partitioning, coefficient re-estimation,
#' and GCV tuning are skipped entirely.
#' Draws producing non-positive-definite correlation matrices are
#' rejected and redrawn (up to 50 attempts).
#'
#' When \code{draw_correlation = FALSE} (default), correlation parameters
#' are fixed at their estimated values.
#'
#' \strong{Inequality constraints.}
#' Active QP inequalities can be enforced during posterior sampling via
#' elliptical slice sampling, producing draws from the corresponding
#' truncated multivariate normal posterior on the coefficient scale.
#' The public \code{enforce_qp_constraints} argument is forwarded to the
#' stored sampler for both the standard and correlation-aware posterior
#' paths.
#'
#' @param object A fitted \code{lgspline} model object.
#' @param new_sigmasq_tilde Numeric; dispersion \eqn{\tilde{\sigma}^2} used
#'        as the point estimate when \code{draw_dispersion = FALSE}.
#'        Default \code{object$sigmasq_tilde}.
#' @param new_predictors Matrix; predictor matrix for posterior predictive
#'        sampling. Default uses in-sample predictors.
#' @param theta_1 Numeric; shape increment for the inverse-gamma prior on
#'        \eqn{\sigma^2}. Default 0.
#' @param theta_2 Numeric; rate increment for the inverse-gamma prior.
#'        Default 0.
#' @param posterior_predictive_draw Function; sampler for posterior predictive
#'        realisations. Must accept \code{N}, \code{mean},
#'        \code{sqrt_dispersion}, \code{...}. Default \code{rnorm}.
#' @param draw_dispersion Logical; sample \eqn{\sigma^2} from its posterior.
#'        Default TRUE.
#' @param include_posterior_predictive Logical; generate posterior predictive
#'        draws at \code{new_predictors}. Default FALSE.
#' @param num_draws Positive integer; number of draws. Default 1.
#' @param enforce_qp_constraints Logical; if TRUE, enforce active QP
#'        inequality constraints during posterior sampling via the stored
#'        elliptical-slice constrained sampler. Default TRUE.
#' @param draw_correlation Logical; propagate correlation parameter
#'        uncertainty. Requires \code{VhalfInv_fxn} and
#'        \code{VhalfInv_params_estimates} in the fitted object.
#'        Default FALSE.
#' @param correlation_param_mean Numeric vector; mean of the approximate
#'        normal posterior for correlation parameters on the unbounded
#'        (working) scale. Default: \code{object$VhalfInv_params_estimates}.
#' @param correlation_param_vcov Matrix; variance-covariance for correlation
#'        parameter draws. Default: inverse Hessian from BFGS
#'        (\code{object$VhalfInv_params_vcov}).
#' @param correlation_VhalfInv_fxn Function; maps correlation parameter
#'        vector to \eqn{\mathbf{V}^{-1/2}}. Default \code{object$VhalfInv_fxn}.
#' @param correlation_Vhalf_fxn Function or NULL; maps to
#'        \eqn{\mathbf{V}^{1/2}}. Passed through to
#'        \code{\link{generate_posterior_correlation}}; the current correlation-aware posterior path only requires \code{correlation_VhalfInv_fxn}.
#' @param correlation_param_vcov_scale NULL or numeric; if supplied,
#'        divides a user-supplied \code{correlation_param_vcov} by this
#'        value before passing it to
#'        \code{\link{generate_posterior_correlation}}. When NULL,
#'        no additional scaling is applied.
#' @param include_warnings Logical; emit warnings for degenerate draws,
#'        constraint violations, etc. Default TRUE.
#' @param ... Additional arguments forwarded to the GLM weight function,
#'        dispersion function, and \code{posterior_predictive_draw}.
#'
#' @return When \code{num_draws = 1}, a named list:
#' \describe{
#'   \item{post_draw_coefficients}{List of length K+1; per-partition
#'         coefficient vectors on the original scale.}
#'   \item{post_draw_sigmasq}{Drawn (or fixed) dispersion.}
#'   \item{post_pred_draw}{Posterior predictive vector (only when
#'         \code{include_posterior_predictive = TRUE}).}
#'   \item{post_draw_correlation_params}{Drawn correlation parameters
#'         on the working scale (only when \code{draw_correlation = TRUE}).}
#' }
#' When \code{num_draws > 1}, each element becomes a list of length
#' \code{num_draws}, and \code{post_pred_draw} (if requested) is an
#' \eqn{N_{\mathrm{new}} \times M} matrix, where \eqn{M = \mathrm{num\_draws}}.
#'
#' @examples
#'
#' \donttest{
#' set.seed(1234)
#' n_blocks <- 100; block_size <- 5; N <- n_blocks * block_size
#' rho_true <- 0.3
#' t <- seq(-5, 5, length.out = N)
#' true_mean <- sin(t)
#' errors <- Reduce("rbind",
#'   lapply(1:n_blocks, function(i) {
#'     sigma <- diag(block_size) + rho_true *
#'       (matrix(1, block_size, block_size) - diag(block_size))
#'     matsqrt(sigma) %*% rnorm(block_size)
#'   })
#' )
#' y <- true_mean + errors * 0.5
#'
#' model_fit <- lgspline(t, y,
#'   K = 3,
#'   correlation_id = rep(1:n_blocks, each = block_size),
#'   correlation_structure = "exchangeable",
#'   include_warnings = FALSE
#' )
#'
#' ## Propagate correlation uncertainty across 50 draws
#' post <- generate_posterior(model_fit,
#'   draw_correlation = TRUE, num_draws = 50,
#'   include_warnings = FALSE
#' )
#'
#' ## Fixed correlation parameters for comparison
#' post_fixed <- generate_posterior(model_fit, num_draws = 50)
#'
#' corr_draws <- unlist(post$post_draw_correlation_params)
#' rho_draws <- exp(-exp(corr_draws))
#' print(summary(rho_draws))
#' }
#'
#' @seealso
#' \code{\link{lgspline}},
#' \code{\link{generate_posterior_correlation}},
#' \code{\link{wald_univariate}}
#' @export
generate_posterior <- function(object,
                               new_sigmasq_tilde = object$sigmasq_tilde,
                               new_predictors = NULL,
                               theta_1 = 0,
                               theta_2 = 0,
                               posterior_predictive_draw =
                                 function(N, mean, sqrt_dispersion, ...) {
                                   rnorm(N, mean, sqrt_dispersion)
                                 },
                               draw_dispersion = TRUE,
                               include_posterior_predictive = FALSE,
                               num_draws = 1,
                               enforce_qp_constraints = TRUE,
                               draw_correlation = FALSE,
                               correlation_param_mean = NULL,
                               correlation_param_vcov = NULL,
                               correlation_VhalfInv_fxn = NULL,
                               correlation_Vhalf_fxn = NULL,
                               correlation_param_vcov_scale = NULL,
                               include_warnings = TRUE,
                               ...) {

  has_qp <- !is.null(object$quadprog_list) &&
    !identical(object$quadprog_list, list(NA)) &&
    !is.null(object$quadprog_list$qp_Amat)

  ## Route to correlation-aware sampler when requested
  if(draw_correlation){

    ## Resolve the scaling divisor for the correlation parameter vcov.
    #  When the user supplies correlation_param_vcov_scale, use it directly.
    #  Otherwise default to 1.
    if(is.null(correlation_param_vcov_scale)){
      scale_by <- 1
    } else {
      scale_by <- correlation_param_vcov_scale
    }

    ## Apply scaling to the user-supplied vcov before handoff.
    #  If the user supplied correlation_param_vcov, scale it here.
    #  If NULL, generate_posterior_correlation builds it internally
    #  from object$VhalfInv_params_vcov with the same default scaling.
    if(!is.null(correlation_param_vcov)){
      scaled_vcov <- correlation_param_vcov / scale_by
    } else {
      scaled_vcov <- NULL
    }

    return(
      generate_posterior_correlation(
        object                     = object,
        new_sigmasq_tilde          = new_sigmasq_tilde,
        new_predictors             = new_predictors,
        theta_1                    = theta_1,
        theta_2                    = theta_2,
        posterior_predictive_draw  = posterior_predictive_draw,
        draw_dispersion            = draw_dispersion,
        include_posterior_predictive = include_posterior_predictive,
        num_draws                  = num_draws,
        enforce_qp_constraints     = enforce_qp_constraints,
        correlation_param_mean     = correlation_param_mean,
        correlation_param_vcov_sc  = scaled_vcov,
        correlation_VhalfInv_fxn   = correlation_VhalfInv_fxn,
        correlation_Vhalf_fxn      = correlation_Vhalf_fxn,
        include_warnings           = include_warnings,
        ...
      )
    )
  }

  ## Standard path: delegate to closure stored in the object
  internal_genpost_func <- object$generate_posterior
  if(!is.null(internal_genpost_func) && is.function(internal_genpost_func)){
    if(is.null(new_predictors)){
      return(internal_genpost_func(
        new_sigmasq_tilde            = new_sigmasq_tilde,
        theta_1                      = theta_1,
        theta_2                      = theta_2,
        posterior_predictive_draw    = posterior_predictive_draw,
        draw_dispersion              = draw_dispersion,
        include_posterior_predictive = include_posterior_predictive,
        num_draws                    = num_draws,
        enforce_qp_constraints       = enforce_qp_constraints,
        ...
      ))
    } else {
      return(internal_genpost_func(
        new_sigmasq_tilde            = new_sigmasq_tilde,
        new_predictors               = new_predictors,
        theta_1                      = theta_1,
        theta_2                      = theta_2,
        posterior_predictive_draw    = posterior_predictive_draw,
        draw_dispersion              = draw_dispersion,
        include_posterior_predictive = include_posterior_predictive,
        num_draws                    = num_draws,
        enforce_qp_constraints       = enforce_qp_constraints,
        ...
      ))
    }
  } else {
    stop("Internal generate_posterior method not found in lgspline object.")
  }
}


#' Generate Posterior Samples Propagating Correlation Parameter Uncertainty
#'
#' Called internally by \code{\link{generate_posterior}} when
#' \code{draw_correlation = TRUE}, but can be used directly for finer control.
#' For each draw, samples the correlation parameter vector from its approximate
#' normal posterior, rebuilds the posterior covariance under that drawn
#' correlation structure without re-solving for a new coefficient mode, then
#' draws coefficients from the updated posterior.
#'
#' @details
#' Each draw proceeds in three steps:
#' \enumerate{
#'   \item \strong{Draw correlation parameters.}
#'     \eqn{\boldsymbol{\rho}^{(m)} \sim
#'     \mathcal{N}(\hat{\boldsymbol{\rho}}_{\mathrm{REML}},
#'               \mathbf{H}^{-1}_{\mathrm{BFGS}})}
#'     on the unbounded working scale. Draws producing a non-PD correlation
#'     matrix are rejected and redrawn (up to 50 attempts); if all fail, the
#'     point estimate is used with a warning.
#'
#'   \item \strong{Rebuild posterior covariance.}
#'     Using the already-expanded \eqn{\mathbf{X}_k}, \eqn{\mathbf{A}}, and
#'     \eqn{\boldsymbol{\Lambda}} from the original fit, recompute only the
#'     covariance-side quantities implied by the drawn correlation structure:
#'     \deqn{
#'       \mathbf{G}_{\mathrm{correct}}^{(m)} =
#'       \left(\mathbf{X}^{\top} \mathbf{W}\mathbf{D}
#'       \mathbf{V}^{-1}(\boldsymbol{\rho}^{(m)}) \mathbf{X}
#'       + \boldsymbol{\Lambda}\right)^{-1},
#'     }
#'     and from this the updated constraint projection
#'     \eqn{\mathbf{U}^{(m)}} and effective degrees of freedom
#'     \eqn{\mathrm{trace}(\mathbf{H}^{(m)})}.
#'
#'     The coefficient mode \eqn{\hat{\boldsymbol{\beta}}_{\mathrm{raw}}}
#'     and fitted mean \eqn{\tilde{\mathbf{y}}} are held fixed at the
#'     original fit values. Knot placement, partitioning, polynomial
#'     expansion, penalty tuning, and coefficient re-estimation are all
#'     skipped entirely.
#'
#'   \item \strong{Draw coefficients.}
#'     Updated quantities (\code{U}, \code{Ghalf_correct}, \code{VhalfInv},
#'     \code{sigmasq_tilde}, \code{trace_XUGX}) are passed to the stored
#'     closure via \code{override_*} arguments so that the draw is centred
#'     at the original mode but uses the covariance implied by the drawn
#'     correlation structure. The stored mode \code{object$B_raw} is passed
#'     as \code{override_B_raw} so it is not recomputed.
#' }
#'
#' \strong{Why the mode is held fixed.}
#' Re-solving for a new MAP estimate under each drawn correlation structure
#' is expensive, requires iterative solvers, and risks convergence failures
#' on draws far from the REML estimate. The posterior draw is centred at the
#' original mode, which remains a reasonable approximation when the REML
#' surface is not sharply peaked. The covariance update captures the primary
#' effect of correlation uncertainty on posterior width and shape.
#'
#' \strong{BFGS inverse Hessian caveat.}
#' The BFGS inverse Hessian approximation for the correlation parameter
#' covariance is asymptotically valid but may be poor for small samples,
#' near-boundary estimates, or multimodal REML surfaces. It is not guaranteed
#' to converge to the observed information matrix. Users should inspect
#' \code{object$VhalfInv_params_vcov} before relying on these draws.
#'
#' @param object A fitted \code{lgspline} object with a correlation structure
#'        (i.e., \code{VhalfInv_fxn} and \code{VhalfInv_params_estimates}
#'        present, or supplied via override arguments).
#' @param new_sigmasq_tilde Numeric; dispersion starting value when
#'        \code{draw_dispersion = FALSE}. Default \code{object$sigmasq_tilde}.
#' @param new_predictors Matrix or NULL; predictor matrix for posterior
#'        predictive sampling. Default uses in-sample predictors.
#' @param theta_1 Numeric; shape increment for the inverse-gamma prior.
#'        Default 0.
#' @param theta_2 Numeric; rate increment for the inverse-gamma prior.
#'        Default 0.
#' @param posterior_predictive_draw Function; sampler for posterior predictive
#'        realisations. Default \code{rnorm}.
#' @param draw_dispersion Logical; sample \eqn{\sigma^2} within each
#'        draw. Default TRUE.
#' @param include_posterior_predictive Logical; generate posterior predictive
#'        draws. Default FALSE.
#' @param num_draws Positive integer; number of draws (each requires one
#'        correlation parameter sample and one covariance rebuild).
#'        Default 1.
#' @param enforce_qp_constraints Logical; if TRUE, enforce active QP
#'        inequality constraints during each coefficient draw via the
#'        stored elliptical-slice constrained sampler. Default TRUE.
#' @param correlation_param_mean Numeric vector or NULL; mean of the
#'        approximate normal posterior on the working scale. Default:
#'        \code{object$VhalfInv_params_estimates}. Supplying this allows
#'        correlation draws for models fit with a fixed (non-optimised)
#'        correlation structure.
#' @param correlation_param_vcov_sc Matrix or NULL; variance-covariance
#'        on the working scale. Default:
#'        \code{object$VhalfInv_params_vcov}. No further scaling is
#'        applied within this function.
#' @param correlation_VhalfInv_fxn Function or NULL; maps parameter vector
#'        to \eqn{\mathbf{V}^{-1/2}}. Default \code{object$VhalfInv_fxn}.
#' @param correlation_Vhalf_fxn Function or NULL; maps to
#'        \eqn{\mathbf{V}^{1/2}}. Not consumed in the current method body
#'        but resolved for interface consistency and potential future use.
#' @param include_warnings Logical; emit warnings. Default TRUE.
#' @param ... Additional arguments forwarded to the GLM weight function,
#'        dispersion function, and \code{posterior_predictive_draw}.
#'
#' @return When \code{num_draws = 1}, a named list:
#' \describe{
#'   \item{post_draw_coefficients}{List of length K+1; per-partition
#'         coefficient vectors on the original scale.}
#'   \item{post_draw_sigmasq}{Drawn dispersion.}
#'   \item{post_pred_draw}{Posterior predictive vector (only when
#'         \code{include_posterior_predictive = TRUE}).}
#'   \item{post_draw_correlation_params}{Drawn correlation parameters
#'         on the working scale.}
#' }
#' When \code{num_draws > 1}:
#' \describe{
#'   \item{post_draw_coefficients}{List of \code{num_draws} lists of K+1
#'         coefficient vectors.}
#'   \item{post_draw_sigmasq}{List of \code{num_draws} scalars.}
#'   \item{post_pred_draw}{\eqn{N_{\mathrm{new}} \times M} matrix, where
#'         \eqn{M = \mathrm{num\_draws}} (only when
#'         \code{include_posterior_predictive = TRUE}).}
#'   \item{post_draw_correlation_params}{List of \code{num_draws} vectors.}
#' }
#'
#' @examples
#' ## See ?generate_posterior for a complete worked example.
#'
#' @seealso
#' \code{\link{generate_posterior}},
#' \code{\link{lgspline}},
#' \code{\link{lgspline.fit}}
#' @export
generate_posterior_correlation <- function(
    object,
    new_sigmasq_tilde = object$sigmasq_tilde,
    new_predictors = NULL,
    theta_1 = 0,
    theta_2 = 0,
    posterior_predictive_draw =
      function(N, mean, sqrt_dispersion, ...) {
        rnorm(N, mean, sqrt_dispersion)
      },
    draw_dispersion = TRUE,
    include_posterior_predictive = FALSE,
    num_draws = 1,
    enforce_qp_constraints = TRUE,
    correlation_param_mean = NULL,
    correlation_param_vcov_sc = NULL,
    correlation_VhalfInv_fxn = NULL,
    correlation_Vhalf_fxn = NULL,
    include_warnings = TRUE,
    ...
) {


  ## 1. Resolve correlation parameter mean, vcov, and functions.
  #     These are required for sampling phi. If the user did not supply
  #     overrides, pull from the fitted object.

  if(is.null(correlation_param_mean)){
    correlation_param_mean <- object$VhalfInv_params_estimates
    if(is.null(correlation_param_mean)){
      stop(
        "\n\t Cannot draw correlation parameters: no point estimates ",
        "found (VhalfInv_params_estimates is NULL) and no ",
        "'correlation_param_mean' was supplied.\n"
      )
    }
  }
  correlation_param_mean <- c(correlation_param_mean)
  n_corr_par <- length(correlation_param_mean)

  ## Default vcov: no scaling
  if(is.null(correlation_param_vcov_sc)){
    correlation_param_vcov_sc <- object$VhalfInv_params_vcov
    if(is.null(correlation_param_vcov_sc)){
      stop(
        "\n\t Cannot draw correlation parameters: no inverse Hessian ",
        "found (VhalfInv_params_vcov is NULL) and no ",
        "'correlation_param_vcov_sc' was supplied.\n"
      )
    }
  }
  correlation_param_vcov_sc <- as.matrix(correlation_param_vcov_sc)
  if(nrow(correlation_param_vcov_sc) != n_corr_par ||
     ncol(correlation_param_vcov_sc) != n_corr_par){
    stop(
      "\n\t 'correlation_param_vcov_sc' must be a ",
      n_corr_par, " x ", n_corr_par, " matrix.\n"
    )
  }

  ## VhalfInv function: maps unbounded parameters to V^{-1/2}
  if(is.null(correlation_VhalfInv_fxn)){
    correlation_VhalfInv_fxn <- object$VhalfInv_fxn
    if(is.null(correlation_VhalfInv_fxn)){
      stop(
        "\n\t No VhalfInv_fxn found and no ",
        "'correlation_VhalfInv_fxn' was supplied.\n"
      )
    }
  }

  ## Vhalf function: resolved for interface consistency, not consumed.
  #  The covariance-only path only needs VhalfInv.
  if(is.null(correlation_Vhalf_fxn)){
    correlation_Vhalf_fxn <- object$Vhalf_fxn
  }


  ## 2. Cholesky factor of correlation parameter vcov for sampling.
  #     If the matrix is not positive definite (can happen with near-
  #     degenerate BFGS approximations), add a small ridge.

  vcov_chol <- tryCatch({
    chol(correlation_param_vcov_sc)
  }, error = function(e){
    ## The BFGS inverse Hessian can be indefinite when the optimizer
    #  converges poorly or the REML surface is nearly flat. A token
    #  ridge is not sufficient when eigenvalues are actually negative.
    #  Project to the nearest PD matrix by clamping eigenvalues, then
    #  add a small ridge for numerical headroom.
    eig <- eigen(correlation_param_vcov_sc, symmetric = TRUE)
    min_eval <- max(abs(eig$values)) * sqrt(.Machine$double.eps)
    eig$values <- pmax(eig$values, min_eval)
    vcov_pd <- crossprod(t(eig$vectors) * sqrt(eig$values))
    if(include_warnings){
      warning(
        "\n\t correlation_param_vcov_sc is not positive definite; ",
        "projecting to nearest PD matrix.\n"
      )
    }
    chol(vcov_pd)
  })

  ## 3. Extract stored components needed for the covariance rebuild.
  K  <- object$K
  nc <- object$p          # p_expansions: basis terms per partition
  N  <- object$N
  P_total <- nc * (K + 1) # total coefficient dimension

  A  <- object$A
  nca <- if(!is.null(A)) ncol(A) else 0

  ## Standardized design matrices (the scale on which the posterior
  #  sampler operates). These are the stored X, rescaled.
  X_std <- lapply(object$X, object$std_X)

  ## Full block-diagonal standardized design: N x P in partition order
  X_full_std <- collapse_block_diagonal(X_std)

  ## Permutation from original observation order to partition order.
  #  Under correlation, algebra must respect this mapping because
  #  VhalfInv is in original observation order while X is in
  #  partition order.
  order_list <- object$order_list
  og_order   <- object$og_order
  po <- unlist(order_list)  # partition order permutation

  y_og <- object$y         # response in original observation order
  family <- object$family
  penalties <- object$penalties
  expansion_scales <- object$expansion_scales
  mean_y <- object$mean_y
  sd_y   <- object$sd_y

  ## Observation weights D in original observation order.
  #  For the penalized information matrix we need sqrt(W * D) row-wise
  #  on the design, where W is the GLM working weight and D is the
  #  observation weight.
  wts_og <- object$weights
  if(is.list(wts_og)){
    obs_wts_vec <- unlist(wts_og)[og_order]
  } else {
    obs_wts_vec <- wts_og
  }

  ## GLM weight function.
  #  For the penalized information matrix we need the mean-variance
  #  relationship evaluated at the fitted values:
  #    gram_gls = X' (W D) V^{-1} X + Lambda
  #  where W = glm_weight_function(mu, y, ...).
  #  For Gaussian identity, W = 1 everywhere, so the weight is just D.
  #  For GLMs (logistic, Poisson, etc.), W carries the inverse
  #  variance function and is essential for correctness.
  #
  #  glm_weight_function is stored in .fit_call_args. If it is not
  #  available (e.g. objects created before .fit_call_args was introduced),
  #  fall back to the family$variance, which is correct for the default
  #  weight function.
  if(!is.null(object$.fit_call_args$glm_weight_function)){
    glm_weight_function <- object$.fit_call_args$glm_weight_function
  } else {
    ## Fallback: replicate the default glm_weight_function from lgspline()
    glm_weight_function <- function(mu, y, order_indices, family,
                                    dispersion, observation_weights, ...){
      if(any(!is.null(observation_weights))){
        family$variance(mu) * observation_weights
      } else {
        family$variance(mu)
      }
    }
  }

  ## Whether partition-specific penalties exist
  unique_pp <- (length(penalties$L_partition_list) == (K + 1))


  ## 4. Precompute the full block-diagonal penalty matrix Lambda_full.
  #     This is fixed across draws and can be computed once.
  #     Lambda_full = blockdiag(Lambda + L_partition_list[[k]])

  Lambda_full <- collapse_block_diagonal(
    lapply(1:(K + 1), function(k){
      if(unique_pp){
        penalties$Lambda + penalties$L_partition_list[[k]]
      } else {
        penalties$Lambda
      }
    })
  )


  ## 5. Determine whether we are Gaussian identity.
  #     This controls whether sigmasq_tilde is recomputed from
  #     whitened residuals under each drawn VhalfInv, or held fixed.

  is_gaussian_identity <- (paste0(family)[1] == "gaussian" &&
                             paste0(family)[2] == "identity")

  ## The fixed fitted mean and mode from the original fit.
  #  These are the quantities we explicitly do NOT recompute.
  ytilde_fixed <- object$ytilde
  B_raw_fixed  <- object$B_raw


  ## 5b. Precompute GLM-weighted design matrix.
  #      The penalized information matrix under correlation is:
  #        gram_gls = X' (W~ D) V^{-1} X + Lambda
  #      where W~ = glm_weight_function(mu, y, ...) is the GLM working
  #      weight (inverse mean-variance relationship) and D is the
  #      observation weight vector.
  #
  #      Since we hold ytilde fixed, the GLM weights W~ do not change
  #      across draws. We can precompute (W~ D)^{1/2} X once, so the
  #      per-draw helper only needs to apply V^{-1/2}.
  #
  #      For Gaussian identity, W~ = 1 everywhere and the combined
  #      weight reduces to sqrt(D). For GLMs (logistic, Poisson, etc.),
  #      W~ carries the inverse variance function and is essential for
  #      correctness of the information matrix.
  #
  #      We pass rep(1, N) for observation_weights inside
  #      glm_weight_function to avoid double-counting D, then fold D
  #      in separately. This mirrors the varcovmat block in lgspline.fit.
  W_glm <- c(glm_weight_function(
    ytilde_fixed[po],       # mu in partition order
    y_og[po],               # y in partition order
    1:N,                    # order_indices
    family,
    object$sigmasq_tilde,   # dispersion at the fixed estimate
    rep(1, N),              # unit obs weights to avoid double-counting D
    ...
  ))
  W_glm <- pmax(W_glm, .Machine$double.eps)

  ## Observation weights D in partition order
  D_po <- obs_wts_vec[po]

  ## Combined weight applied after whitening, matching the main GEE fit.
  combined_wt_po <- sqrt(W_glm * D_po)


  ## 6. Helper: rebuild posterior covariance given a drawn VhalfInv.
  #
  #     This replaces the previous .reestimate_with_VhalfInv() which
  #     re-solved for B_raw, B, and ytilde. Here we only rebuild:
  #       - Ghalf_correct  (dense correlated posterior square root)
  #       - U              (equality-constraint projection, G-dependent)
  #       - trace_XUGX     (effective df for dispersion draw)
  #       - sigmasq_tilde  (Gaussian identity: whitened residual-based;
  #                          non-Gaussian: held at original estimate)
  #
  #     The mode is held fixed at object$B_raw. The fitted mean is
  #     held fixed at object$ytilde.

  .rebuild_posterior_cov_with_VhalfInv <- function(VhalfInv_draw){

    ## Permute VhalfInv into partition order.
    #  VhalfInv_draw is N x N in original observation order.
    #  X_full_std is N x P in partition order.
    #  We need V^{-1/2}[po, po] to align with the design matrix.
    VhalfInv_po <- VhalfInv_draw[po, po]

    ## Whiten then apply the combined row weights, matching the solver path:
    ##   X' V^{-1/2} W~ D V^{-1/2} X.
    VhalfInvX <- VhalfInv_po %**% X_full_std
    VhalfInvX <- t(t(VhalfInvX) * combined_wt_po)

    ## Full penalized GLS Gram in the whitened system + Lambda  (P x P)
    #  This is the dense correlated information matrix. Under correlation,
    #  the exact posterior factor is based on this full-system Gram,
    #  NOT the stored diagonal blocks of Ghalf.
    gram_gls <- crossprod(VhalfInvX) + Lambda_full

    ## G_correct = gram_gls^{-1}     (needed for U)
    ## Ghalf_correct = gram_gls^{-1/2}  (posterior square root factor)
    G_correct     <- invert(gram_gls)
    Ghalf_correct <- matinvsqrt(gram_gls)

    ## Constraint projection U = I - G A (A' G A)^{-1} A'
    #  The feasible equality subspace itself is fixed by A, but the
    #  projection matrix U depends on the current G. If G changes
    #  with the drawn correlation structure, U changes too.
    #  If A is empty or NULL, U reduces to identity.
    if(!is.null(A) && nca > 0){
      GA      <- G_correct %**% A
      AGA_inv <- invert(crossprod(A, GA))
      U_draw  <- diag(P_total) - GA %**% tcrossprod(AGA_inv, A)
    } else {
      U_draw <- diag(P_total)
    }

    ## Effective df: trace(X' W~ D V^{-1} X U G_correct)
    #  = || V^{-1/2} (W~ D)^{1/2} X U G_correct^{1/2} ||_F^2
    #  This matters because the stored posterior closure uses it in
    #  the inverse-gamma dispersion draw. If the effective df changes
    #  under the drawn correlation structure, leaving it fixed is
    #  internally inconsistent.
    UGhalf_draw     <- U_draw %**% Ghalf_correct
    trace_XUGX_draw <- sum((VhalfInvX %**% UGhalf_draw)^2)

    ## Dispersion
    sigmasq_draw <- object$sigmasq_tilde

    ## Return only the covariance-side pieces. No B_raw, no B, no ytilde.
    list(
      Ghalf_correct = Ghalf_correct,
      U             = U_draw,
      trace_XUGX    = trace_XUGX_draw,
      sigmasq_tilde = sigmasq_draw,
      VhalfInv      = VhalfInv_draw
    )
  }


  ## 7. Single draw: sample phi, rebuild covariance, draw coefficients.

  .one_corr_draw <- function(){

    ## 7a. Draw correlation parameters from approximate normal posterior.
    #     Reject draws that produce non-finite or wrong-dimension V^{-1/2}.
    max_corr_reject <- 50L
    n_corr_reject   <- 0L
    phi_draw         <- NULL
    VhalfInv_draw    <- NULL

    repeat {
      z <- rnorm(n_corr_par)
      phi_candidate <- correlation_param_mean + c(crossprod(vcov_chol, z))

      tr <- try({
        VhalfInv_candidate <- correlation_VhalfInv_fxn(phi_candidate)
        stopifnot(
          is.matrix(VhalfInv_candidate),
          all(is.finite(VhalfInv_candidate)),
          nrow(VhalfInv_candidate) == N,
          ncol(VhalfInv_candidate) == N
        )
      }, silent = TRUE)

      if(!inherits(tr, "try-error")){
        phi_draw      <- phi_candidate
        VhalfInv_draw <- VhalfInv_candidate
        break
      }

      n_corr_reject <- n_corr_reject + 1L
      if(n_corr_reject >= max_corr_reject){
        if(include_warnings){
          warning(
            "\n\t Failed to draw a valid correlation parameter after ",
            max_corr_reject, " attempts. Using point estimate.\n"
          )
        }
        phi_draw      <- correlation_param_mean
        VhalfInv_draw <- correlation_VhalfInv_fxn(phi_draw)
        break
      }
    }

    ## 7b. Rebuild posterior covariance under drawn VhalfInv.
    #     This is the covariance-only replacement for the old
    #     .reestimate_with_VhalfInv(). The coefficient mode stays
    #     fixed at object$B_raw.
    cov_rebuild <- try(
      .rebuild_posterior_cov_with_VhalfInv(VhalfInv_draw),
      silent = TRUE
    )

    if(inherits(cov_rebuild, "try-error")){
      if(include_warnings){
        warning(
          "\n\t Covariance rebuild with drawn correlation parameters ",
          "failed. Falling back to original model for this draw.\n"
        )
      }
      ## Fall back: draw from original posterior with no overrides
      one_draw <- object$generate_posterior(
        new_sigmasq_tilde            = new_sigmasq_tilde,
        theta_1                      = theta_1,
        theta_2                      = theta_2,
        posterior_predictive_draw    = posterior_predictive_draw,
        draw_dispersion              = draw_dispersion,
        include_posterior_predictive = include_posterior_predictive,
        num_draws                    = 1,
        enforce_qp_constraints       = enforce_qp_constraints,
        ...
      )
      one_draw$post_draw_correlation_params <- phi_draw
      return(one_draw)
    }

    ## 7c. Draw coefficients from updated posterior via overrides.
    #
    #     The closure captures model_fit by reference, so mutating a
    #     shallow copy of 'object' would have no effect. Instead we
    #     pass the rebuilt covariance-side quantities through
    #     override_* args:
    #
    #       override_B_raw          = object$B_raw  (FIXED mode)
    #       override_U              = U_draw
    #       override_Ghalf_correct  = Ghalf_correct_draw
    #       override_VhalfInv       = VhalfInv_draw
    #       override_sigmasq_tilde  = sigmasq_draw
    #       override_trace_XUGX     = trace_draw
    #
    #     The closure will then compute:
    #       L_post = (1/sd_y) * U_draw %*% Ghalf_correct_draw
    #     and draw:
    #       beta ~ N(B_raw_fixed, sigmasq * L_post L_post^T)
    #     (or use ESS when inequalities are enforced).

    ## Use the rebuilt dispersion for the InvGamma draw
    refit_sigmasq <- cov_rebuild$sigmasq_tilde
    if(is.na(refit_sigmasq) || !is.finite(refit_sigmasq)){
      refit_sigmasq <- new_sigmasq_tilde
    }

    override_args <- list(
      new_sigmasq_tilde            = refit_sigmasq,
      theta_1                      = theta_1,
      theta_2                      = theta_2,
      posterior_predictive_draw    = posterior_predictive_draw,
      draw_dispersion              = draw_dispersion,
      include_posterior_predictive = include_posterior_predictive,
      num_draws                    = 1,
      enforce_qp_constraints       = enforce_qp_constraints,
      override_B_raw               = B_raw_fixed,
      override_U                   = cov_rebuild$U,
      override_Ghalf_correct       = cov_rebuild$Ghalf_correct,
      override_VhalfInv            = cov_rebuild$VhalfInv,
      override_sigmasq_tilde       = refit_sigmasq,
      override_trace_XUGX          = cov_rebuild$trace_XUGX
    )
    if(!is.null(new_predictors)){
      override_args$new_predictors <- new_predictors
    }

    one_draw <- do.call(object$generate_posterior,
                        c(override_args, list(...)))
    one_draw$post_draw_correlation_params <- phi_draw
    return(one_draw)
  }


  ## 8. Execute draws and repack into the standard return shape.

  results <- lapply(seq_len(num_draws), function(m) .one_corr_draw())

  ## Single-draw return: flat list
  if(num_draws == 1){
    return(results[[1]])
  }

  ## Multi-draw return: repack per-draw results into the usual shape
  #  that matches the generate_posterior() multi-draw contract.
  post_draw_coefficients <- lapply(results, `[[`, "post_draw_coefficients")
  post_draw_sigmasq <- lapply(results, `[[`, "post_draw_sigmasq")
  post_draw_correlation_params <- lapply(
    results, `[[`, "post_draw_correlation_params"
  )

  if(include_posterior_predictive){
    post_pred_draw <- Reduce(
      "cbind",
      lapply(results, `[[`, "post_pred_draw")
    )
    return(list(
      post_draw_coefficients       = post_draw_coefficients,
      post_draw_sigmasq            = post_draw_sigmasq,
      post_pred_draw               = post_pred_draw,
      post_draw_correlation_params = post_draw_correlation_params
    ))
  }

  return(list(
    post_draw_coefficients       = post_draw_coefficients,
    post_draw_sigmasq            = post_draw_sigmasq,
    post_draw_correlation_params = post_draw_correlation_params
  ))
}


#' Plot Method for lgspline Objects
#'
#' Wrapper for the internal lgspline plot function. Produces a 1D line plot
#' (base R) or interactive 3D surface plot (plotly) depending on the number
#' of predictors, with optional formula overlays per partition. When plotting
#' a subset of variables via \code{vars}, non-plotted predictors are
#' automatically set to zero (or a user-specified value via
#' \code{fixed_values}).
#'
#' @details
#' Partition boundaries are indicated by color changes. For 1D models,
#' observation points can be overlaid. For 2D models, plotly is used.
#'
#' When using \code{vars} to plot a subset of predictors, the non-plotted
#' predictors are automatically set to zero. This can be overridden by
#' passing a named list to \code{fixed_values} (e.g.,
#' \code{fixed_values = list(Height = 75)}). The automatic zeroing replaces
#' the previous behavior where the user had to manually construct
#' \code{new_predictors} with non-plotted variables set to fixed values.
#'
#' When \code{se.fit = TRUE}, pointwise confidence bands are drawn around
#' the fitted function. These are Wald-type intervals constructed on the
#' link scale and back-transformed to the response scale, using
#' \code{cv} as the critical value to multiply se.fit by (default 1 for actual se).
#'
#' The function extracts predictor positions from linear expansion terms.
#' If linear terms are excluded (e.g., via \code{exclude_these_expansions}),
#' plotting will fail. As a workaround, constrain those terms to zero via
#' \code{constraint_vectors} / \code{constraint_values} so they remain in
#' the expansion but are zeroed out.
#'
#' @param x A fitted lgspline model object.
#' @param show_formulas Logical; display partition-level polynomial formulas.
#'        Default FALSE.
#' @param include_all_terms_in_formulas Logical; when \code{show_formulas = TRUE}
#'        and plotting only a subset of predictors via \code{vars}, include all
#'        fitted terms in the displayed formulas rather than only the terms
#'        involving the plotted predictor(s). Default FALSE retains the current
#'        marginal-only formula display.
#' @param digits Integer; decimal places for formula coefficients. Default 4.
#' @param legend_pos Character; legend position for 1D plots. Default
#'        \code{"topright"}.
#' @param custom_response_lab Character; response axis label. Default
#'        \code{"y"}.
#' @param custom_predictor_lab Character; predictor axis label (1D). Default
#'        NULL uses the column name.
#' @param custom_predictor_lab1 Character; first predictor axis label (2D).
#'        Default NULL.
#' @param custom_predictor_lab2 Character; second predictor axis label (2D).
#'        Default NULL.
#' @param custom_formula_lab Character; fitted response label on the link
#'        scale. Default NULL.
#' @param custom_title Character; plot title. Default \code{"Fitted Function"}.
#' @param text_size_formula Numeric; formula text size. Passed to \code{cex}
#'        (1D) or hover font size (2D). Default NULL (0.8 for 1D, 8 for 2D).
#' @param legend_args List; additional arguments passed to \code{legend()}
#'        (1D only).
#' @param new_predictors Matrix; optional predictor values for prediction.
#'        Default NULL. When \code{vars} is specified and
#'        \code{new_predictors} is NULL, a grid is automatically generated
#'        with non-plotted variables set to zero (or values from
#'        \code{fixed_values}).
#' @param xlim Numeric vector; x-axis limits (1D only). Default NULL.
#' @param ylim Numeric vector; y-axis limits (1D only). Default NULL.
#' @param color_function Function; returns K+1 colors, one per partition.
#'        Default NULL uses \code{grDevices::rainbow(K+1)} for 1D and a
#'        Spectral palette for 2D.
#' @param add Logical; add to an existing plot (1D only). Default FALSE.
#' @param vars Numeric or character vector; predictor indices or names to
#'        plot. Default \code{c()} plots all.
#' @param legend_order Numeric; re-ordered partition indices for the legend.
#' @param se.fit Logical; if TRUE, plot pointwise confidence bands. Default
#'        FALSE.
#' @param cv Numeric; critical value for confidence bands. Default 1.
#' @param band_col Character; color for confidence band fill. Default
#'        \code{"grey80"}.
#' @param band_border Character or NA; border color for confidence band
#'        polygon. Default NA (no border).
#' @param fixed_values Named list; fixed values for non-plotted predictors
#'        when \code{vars} is specified. Names should match predictor names.
#'        Default NULL sets non-plotted predictors to zero.
#' @param n_grid Integer; number of grid points for automatic grid
#'        generation when \code{vars} is specified and
#'        \code{new_predictors} is NULL. Default 200.
#' @param ... Additional arguments passed to \code{\link[graphics]{plot}}
#'        (1D) or \code{\link[plotly]{plot_ly}} (2D).
#'
#' @return For 1D models: invisibly returns NULL (base R plot drawn to
#'   device). For 2D models: returns a plotly object.
#'
#' @examples
#'
#' set.seed(1234)
#' t_data <- runif(1000, -10, 10)
#' y_data <- 2*sin(t_data) + -0.06*t_data^2 + rnorm(length(t_data))
#' model_fit <- lgspline(t_data, y_data, K = 9)
#'
#' ## Basic plot
#' plot(model_fit)
#'
#' ## Plot with confidence bands
#' plot(model_fit,
#'      se.fit = TRUE,
#'      cv = 1.96,
#'      custom_title = 'Fitted Function with 95% CI')
#'
#' ## Multi-predictor: automatically zeros non-plotted variables
#' # plot(model_fit_2d, vars = 'x1', se.fit = TRUE)
#'
#' @seealso
#' \code{\link{lgspline}},
#' \code{\link[graphics]{plot}},
#' \code{\link[plotly]{plot_ly}}
#' @export
plot.lgspline <- function(x,
                          show_formulas = FALSE,
                          include_all_terms_in_formulas = FALSE,
                          digits = 4,
                          legend_pos = "topright",
                          custom_response_lab = "y",
                          custom_predictor_lab = NULL,
                          custom_predictor_lab1 = NULL,
                          custom_predictor_lab2 = NULL,
                          custom_formula_lab = NULL,
                          custom_title = "Fitted Function",
                          text_size_formula = NULL,
                          legend_args = list(),
                          new_predictors = NULL,
                          xlim = NULL,
                          ylim = NULL,
                          color_function = NULL,
                          add = FALSE,
                          vars = c(),
                          legend_order = NULL,
                          se.fit = FALSE,
                          cv = 1,
                          band_col = "grey80",
                          band_border = NA,
                          fixed_values = NULL,
                          n_grid = 200,
                          ...) {

  ## When vars is supplied without new_predictors, build a plotting grid
  #  from the training predictors and hold the other variables fixed.

  og_cols <- x$.fit_call_args$og_cols
  q_pred <- x$q

  ## Only auto-generate new_predictors when vars is specified and
  ## new_predictors is NULL and we have >1 predictor.
  if(length(vars) > 0 && is.null(new_predictors) && q_pred > 1){

    ## Resolve vars to numeric indices
    if(is.character(vars)){
      if(is.null(og_cols)){
        stop('\n\t Character vars requires named predictor columns. ',
             'Use numeric column indices instead.\n')
      }
      vars_idx <- match(vars, og_cols)
      if(any(is.na(vars_idx))){
        stop('\n\t vars "', paste(vars[is.na(vars_idx)], collapse='", "'),
             '" not found in predictor names.\n')
      }
    } else {
      vars_idx <- as.integer(vars)
    }

    ## Recover training predictors from the design matrices.
    #  x$X is a list of unstandardized design matrices per partition.
    #  The linear columns (power1_cols and nonspline_cols) contain the
    #  original predictor values. We extract these in partition order
    #  and reorder to observation order.
    all_linear_cols <- sort(c(x$power1_cols, x$nonspline_cols))
    training_preds <- matrix(NA, x$N, q_pred)
    for(k in 1:(x$K + 1)){
      if(nrow(x$X[[k]]) == 0) next
      obs_idx <- x$order_list[[k]]
      ## Extract the q linear columns from the unstandardized expansion
      training_preds[obs_idx, ] <- x$X[[k]][, all_linear_cols, drop = FALSE]
    }

    ## Build the grid
    if(length(vars_idx) == 1){
      ## 1D plot: vary the single plotted variable
      var_range <- range(training_preds[, vars_idx], na.rm = TRUE)
      grid_vals <- seq(var_range[1], var_range[2], length.out = n_grid)

      ## Initialize all columns to their fixed values (default 0)
      new_pred_mat <- matrix(0, nrow = n_grid, ncol = q_pred)
      if(!is.null(og_cols)){
        colnames(new_pred_mat) <- og_cols
      }

      ## Set plotted variable
      new_pred_mat[, vars_idx] <- grid_vals

      ## Override non-plotted variables with fixed_values if provided
      if(!is.null(fixed_values)){
        for(nm in names(fixed_values)){
          if(!is.null(og_cols)){
            fix_idx <- match(nm, og_cols)
          } else {
            fix_idx <- as.integer(nm)
          }
          if(!is.na(fix_idx) && !(fix_idx %in% vars_idx)){
            new_pred_mat[, fix_idx] <- fixed_values[[nm]]
          }
        }
      }

    } else if(length(vars_idx) == 2){
      ## 2D plot: vary both plotted variables
      var1_range <- range(training_preds[, vars_idx[1]], na.rm = TRUE)
      var2_range <- range(training_preds[, vars_idx[2]], na.rm = TRUE)
      n_side <- ceiling(sqrt(n_grid))
      grid_expand <- expand.grid(
        seq(var1_range[1], var1_range[2], length.out = n_side),
        seq(var2_range[1], var2_range[2], length.out = n_side)
      )

      new_pred_mat <- matrix(0, nrow = nrow(grid_expand), ncol = q_pred)
      if(!is.null(og_cols)){
        colnames(new_pred_mat) <- og_cols
      }
      new_pred_mat[, vars_idx[1]] <- grid_expand[, 1]
      new_pred_mat[, vars_idx[2]] <- grid_expand[, 2]

      ## Override non-plotted variables with fixed_values if provided
      if(!is.null(fixed_values)){
        for(nm in names(fixed_values)){
          if(!is.null(og_cols)){
            fix_idx <- match(nm, og_cols)
          } else {
            fix_idx <- as.integer(nm)
          }
          if(!is.na(fix_idx) && !(fix_idx %in% vars_idx)){
            new_pred_mat[, fix_idx] <- fixed_values[[nm]]
          }
        }
      }
    }

    new_predictors <- new_pred_mat
  }

  ## For 1D plots, se.fit draws the confidence band first and then
  #  overlays the fitted curve.

  internal_plot_func <- x$plot
  if (!is.null(internal_plot_func) && is.function(internal_plot_func)) {

    ## Determine if this is a 1D plot scenario
    is_1d <- (x$q == 1) || (length(vars) == 1)

    if(se.fit && is_1d){
      ## Get predictions with standard errors
      #  First, determine which predictors to use for the SE computation
      if(!is.null(new_predictors)){
        se_predictors <- new_predictors
      } else {
        ## Use the training data; extract from X
        all_linear_cols <- sort(c(x$power1_cols, x$nonspline_cols))
        se_predictors <- matrix(NA, x$N, x$q)
        for(k in 1:(x$K + 1)){
          if(nrow(x$X[[k]]) == 0) next
          obs_idx <- x$order_list[[k]]
          se_predictors[obs_idx, ] <- x$X[[k]][, all_linear_cols, drop = FALSE]
        }
        if(!is.null(og_cols)){
          colnames(se_predictors) <- og_cols
        }
      }

      ## Get predictions with SE
      se_result <- x$predict(new_predictors = se_predictors,
                             se.fit = TRUE,
                             cv = cv)

      ## Order the fitted values by the plotted x-variable before drawing
      #  the confidence band polygon.
      if(length(vars) == 1){
        if(is.character(vars)){
          var_col <- match(vars, og_cols)
        } else {
          var_col <- as.integer(vars)
        }
      } else {
        var_col <- 1
      }
      x_vals <- se_predictors[, var_col]
      sort_order <- order(x_vals)
      x_sorted <- x_vals[sort_order]
      lower_sorted <- se_result$lower[sort_order]
      upper_sorted <- se_result$upper[sort_order]

      ## Set up the plot region first (if not adding to existing)
      if(!add){
        ## Set up a fresh plotting region, then draw the band underneath
        #  the fitted curves.
        ## Determine y-limits including the bands
        all_y <- c(se_result$lower, se_result$upper, se_result$fit, x$y)
        if(is.null(ylim)){
          ylim_use <- range(all_y, na.rm = TRUE)
        } else {
          ylim_use <- ylim
        }

        ## Determine axis label
        if(is.null(custom_predictor_lab)){
          if(!is.null(og_cols) && length(vars) == 1){
            if(is.character(vars)){
              custom_predictor_lab_use <- vars
            } else {
              custom_predictor_lab_use <- og_cols[var_col]
            }
          } else if(!is.null(og_cols)){
            custom_predictor_lab_use <- og_cols[1]
          } else {
            custom_predictor_lab_use <- "x"
          }
        } else {
          custom_predictor_lab_use <- custom_predictor_lab
        }

        ## Draw empty plot frame
        graphics::plot(range(x_sorted), ylim_use,
                       type = 'n',
                       xlab = custom_predictor_lab_use,
                       ylab = custom_response_lab,
                       main = custom_title,
                       xlim = xlim,
                       ...)

        ## Draw confidence band polygon
        graphics::polygon(
          x = c(x_sorted, rev(x_sorted)),
          y = c(lower_sorted, rev(upper_sorted)),
          col = band_col,
          border = band_border
        )

        ## Now overlay the fitted curves using add = TRUE
        plot_result <- internal_plot_func(
          model_fit_in = x,
          show_formulas = show_formulas,
          include_all_terms_in_formulas = include_all_terms_in_formulas,
          digits = digits,
          legend_pos = legend_pos,
          custom_response_lab = custom_response_lab,
          custom_predictor_lab = custom_predictor_lab,
          custom_predictor_lab1 = custom_predictor_lab1,
          custom_predictor_lab2 = custom_predictor_lab2,
          custom_formula_lab = custom_formula_lab,
          custom_title = custom_title,
          text_size_formula = text_size_formula,
          legend_args = legend_args,
          new_predictors = new_predictors,
          xlim = xlim,
          ylim = ylim_use,
          color_function = color_function,
          add = TRUE,  # overlay on the band
          vars = vars,
          legend_order = legend_order,
          ...)
      } else {
        ## When adding to an existing plot, just layer the band and curves.
        graphics::polygon(
          x = c(x_sorted, rev(x_sorted)),
          y = c(lower_sorted, rev(upper_sorted)),
          col = band_col,
          border = band_border
        )

        plot_result <- internal_plot_func(
          model_fit_in = x,
          show_formulas = show_formulas,
          include_all_terms_in_formulas = include_all_terms_in_formulas,
          digits = digits,
          legend_pos = legend_pos,
          custom_response_lab = custom_response_lab,
          custom_predictor_lab = custom_predictor_lab,
          custom_predictor_lab1 = custom_predictor_lab1,
          custom_predictor_lab2 = custom_predictor_lab2,
          custom_formula_lab = custom_formula_lab,
          custom_title = custom_title,
          text_size_formula = text_size_formula,
          legend_args = legend_args,
          new_predictors = new_predictors,
          xlim = xlim,
          ylim = ylim,
          color_function = color_function,
          add = TRUE,
          vars = vars,
          legend_order = legend_order,
          ...)
      }

      invisible(NULL)

    } else {
      ## Non-SE path: delegate directly to the internal plot function
      plot_result <- internal_plot_func(
        model_fit_in = x,
        show_formulas = show_formulas,
        include_all_terms_in_formulas = include_all_terms_in_formulas,
        digits = digits,
        legend_pos = legend_pos,
        custom_response_lab = custom_response_lab,
        custom_predictor_lab = custom_predictor_lab,
        custom_predictor_lab1 = custom_predictor_lab1,
        custom_predictor_lab2 = custom_predictor_lab2,
        custom_formula_lab = custom_formula_lab,
        custom_title = custom_title,
        text_size_formula = text_size_formula,
        legend_args = legend_args,
        new_predictors = new_predictors,
        xlim = xlim,
        ylim = ylim,
        color_function = color_function,
        add = add,
        vars = vars,
        legend_order = legend_order,
        ...)

      if(inherits(plot_result, "plotly")) {
        return(plot_result)
      } else {
        invisible(NULL)
      }
    }
  } else {
    stop("Internal plot method not found or not a function.")
  }
}


#' Predict Method for lgspline Objects
#'
#' Generates predictions, derivatives, and basis expansions from a fitted
#' lgspline model. Wrapper for the internal predict closure stored in the
#' object.
#'
#' @details
#' \code{new_predictors} takes priority over \code{newdata} when both are
#' supplied. When both are NULL, the training data is used.
#'
#' Fitted values are also accessible directly as \code{model_fit$ytilde} or
#' via \code{model_fit$predict()}.
#'
#' The parallel processing feature is experimental.
#'
#' Additional arguments passed through \code{...} include \code{se.fit}
#' and \code{cv} for pointwise interval summaries.
#'
#' Predictor input should use the original predictor columns. Named extra
#' columns are dropped when they can be identified as irrelevant to the
#' fitted expansions.
#'
#' @param object A fitted lgspline model object.
#' @param newdata Matrix or data.frame; new predictor values. Default NULL.
#' @param parallel Logical; use parallel processing (experimental).
#'        Default FALSE.
#' @param cl Cluster object for parallel processing. Default NULL.
#' @param chunk_size Integer; chunk size for parallel processing. Default NULL.
#' @param num_chunks Integer; number of chunks. Default NULL.
#' @param rem_chunks Integer; remainder chunks. Default NULL.
#' @param B_predict List; per-partition coefficient list for prediction, e.g.
#'        from \code{\link{generate_posterior}}. Default NULL uses
#'        \code{object$B}.
#' @param take_first_derivatives Logical; compute first derivatives.
#'        Default FALSE.
#' @param take_second_derivatives Logical; compute second derivatives.
#'        Default FALSE.
#' @param expansions_only Logical; return basis expansion matrix only.
#'        Default FALSE.
#' @param new_predictors Matrix or data.frame; overrides \code{newdata}.
#' @param ... Additional arguments passed to the internal predict method.
#'
#' @return A numeric vector of predictions, or a list when derivatives or
#'   interval summaries are requested:
#' \describe{
#'   \item{preds}{Numeric vector of predictions when derivatives are requested.}
#'   \item{fit}{Numeric vector of predictions when \code{se.fit = TRUE} and no
#'         derivatives are requested.}
#'   \item{first_deriv}{Numeric vector or named list of first derivatives (if
#'         requested).}
#'   \item{second_deriv}{Numeric vector or named list of second derivatives (if
#'         requested).}
#'   \item{se.fit}{Pointwise standard errors on the link scale (if requested).}
#'   \item{lower}{Pointwise lower interval bound (if requested).}
#'   \item{upper}{Pointwise upper interval bound (if requested).}
#'   \item{cv}{Critical value returned when \code{se.fit = TRUE} without
#'         derivative requests.}
#' }
#' If \code{expansions_only = TRUE}, returns a list of basis expansions.
#'
#' @examples
#'
#' set.seed(1234)
#' t <- runif(1000, -10, 10)
#' y <- 2*sin(t) + -0.06*t^2 + rnorm(length(t))
#' model_fit <- lgspline(t, y)
#'
#' newdata <- matrix(sort(rnorm(10000)), ncol = 1)
#' preds <- predict(model_fit, newdata)
#'
#' deriv1_res <- predict(model_fit, newdata, take_first_derivatives = TRUE)
#' deriv2_res <- predict(model_fit, newdata, take_second_derivatives = TRUE)
#'
#' oldpar <- par(no.readonly = TRUE)
#' layout(matrix(c(1,1,2,2,3,3), byrow = TRUE, ncol = 2))
#'
#' plot(newdata[,1], preds, main = 'Fitted Function',
#'      xlab = 't', ylab = "f(t)", type = 'l')
#' plot(newdata[,1], deriv1_res$first_deriv, main = 'First Derivative',
#'      xlab = 't', ylab = "f'(t)", type = 'l')
#' plot(newdata[,1], deriv2_res$second_deriv, main = 'Second Derivative',
#'      xlab = 't', ylab = "f''(t)", type = 'l')
#'
#' par(oldpar)
#'
#' @seealso \code{\link{lgspline}}, \code{\link{plot.lgspline}}
#' @export
predict.lgspline <- function(object,
                             newdata = NULL,
                             parallel = FALSE,
                             cl = NULL,
                             chunk_size = NULL,
                             num_chunks = NULL,
                             rem_chunks = NULL,
                             B_predict = NULL,
                             take_first_derivatives = FALSE,
                             take_second_derivatives = FALSE,
                             expansions_only = FALSE,
                             new_predictors = NULL,
                             ...) {

  internal_predict_func <- object$predict
  if(is.null(internal_predict_func) || !is.function(internal_predict_func)){
    stop("Internal predict method not found or not a function.")
  }

  ## Default to the fitted coefficients unless B_predict is supplied.
  B_predict_val <- if(!is.null(B_predict)) B_predict else object$B

  ## new_predictors takes priority over newdata
  if(!is.null(new_predictors)){
    predictors_val <- new_predictors
  } else if(!is.null(newdata)){
    predictors_val <- newdata
  } else {
    predictors_val <- NULL
  }

  ## Unwrap nested data.frame: data.frame(new_predictors = data.frame(...))
  #  produces a single-column df whose sole column is itself a df.
  if(inherits(predictors_val, "data.frame") && ncol(predictors_val) == 1 &&
     inherits(predictors_val[[1]], "data.frame")){
    predictors_val <- predictors_val[[1]]
  }

  ## Leave data.frames with non-numeric columns as-is so the internal
  #  predict function can handle factor encoding before matrix coercion.
  if(!is.null(predictors_val)){
    if(inherits(predictors_val, "data.frame") &&
       any(!sapply(predictors_val, function(x) is.numeric(x) ||
                   is.integer(x)))){
      ## has non-numeric columns, leave as data.frame
    } else {
      predictors_val <- try(
        methods::as(cbind(predictors_val), 'matrix'),
        silent = TRUE
      )
      if(inherits(predictors_val, 'try-error')){
        stop('\n \t newdata / new_predictors cannot be coerced to a matrix. \n')
      }
    }
  }

  ## Omit new_predictors for in-sample predictions; otherwise pass the
  #  requested predictor matrix through explicitly.
  if(is.null(predictors_val)){
    internal_predict_func(
      parallel                = parallel,
      cl                      = cl,
      chunk_size              = chunk_size,
      num_chunks              = num_chunks,
      rem_chunks              = rem_chunks,
      B_predict               = B_predict_val,
      take_first_derivatives  = take_first_derivatives,
      take_second_derivatives = take_second_derivatives,
      expansions_only         = expansions_only,
      ...
    )
  } else {
    internal_predict_func(
      new_predictors          = predictors_val,
      parallel                = parallel,
      cl                      = cl,
      chunk_size              = chunk_size,
      num_chunks              = num_chunks,
      rem_chunks              = rem_chunks,
      B_predict               = B_predict_val,
      take_first_derivatives  = take_first_derivatives,
      take_second_derivatives = take_second_derivatives,
      expansions_only         = expansions_only,
      ...
    )
  }
}


#' Extract Coefficients from a Fitted lgspline
#'
#' Returns the per-partition polynomial coefficient lists from a fitted
#' lgspline model.
#'
#' @param object A fitted lgspline model object.
#' @param ... Not used.
#'
#' @details
#' Coefficient names reflect the polynomial expansion terms, e.g.:
#' \itemize{
#'   \item intercept
#'   \item v: linear term for predictor v
#'   \item v_^2: quadratic term
#'   \item v^3: cubic term
#'   \item _v_x_w_: two-way interaction
#' }
#' Column/variable names replace numeric indices when available.
#'
#' To get all coefficients as a single matrix:
#' \code{Reduce('cbind', coef(model_fit))}.
#'
#' @return A list of per-partition coefficient vectors. Returns NULL with a
#'   warning if \code{object$B} is not found.
#'
#' @examples
#'
#' set.seed(1234)
#' t <- runif(1000, -10, 10)
#' y <- 2*sin(t) + -0.06*t^2 + rnorm(length(t))
#' model_fit <- lgspline(t, y)
#'
#' coefficients <- coef(model_fit)
#' print(coefficients[[1]])
#' print(Reduce('cbind', coefficients))
#'
#' @seealso \code{\link{lgspline}}
#' @export
coef.lgspline <- function(object, ...) {
  if(is.null(object$B)) {
    warning("Coefficient component 'B' not found in object.", call. = FALSE)
    return(NULL)
  }
  object$B
}


#' Univariate Wald Tests and Confidence Intervals for lgspline Coefficients
#'
#' Computes per-coefficient Wald tests and confidence intervals from a fitted
#' lgspline. For Gaussian identity-link models, t-statistics and t-intervals
#' are used; otherwise z-statistics.
#'
#' @param object A fitted lgspline object. Must have been fit with
#'        \code{return_varcovmat = TRUE}.
#' @param scale_vcovmat_by Numeric; scaling factor for the variance-covariance
#'        matrix. Default 1.
#' @param cv Numeric; critical value for confidence intervals. If missing,
#'        defaults to \code{object$critical_value} or \code{qnorm(0.975)}.
#' @param ... Additional arguments passed to the internal \code{wald_univariate}
#'        method.
#'
#' @return An object of class \code{"wald_lgspline"}, a list with:
#' \describe{
#'   \item{coefficients}{Matrix with columns: Estimate, Std. Error,
#'         t value or z value, Pr(>|t|) or Pr(>|z|), CI LB, CI UB.}
#'   \item{critical_value}{Critical value used.}
#'   \item{family}{GLM family from the fitted model.}
#'   \item{N}{Number of observations.}
#'   \item{trace_XUGX}{Effective df trace term.}
#'   \item{statistic_name}{"t value" or "z value".}
#'   \item{p_value_name}{"Pr(>|t|)" or "Pr(>|z|)".}
#'   \item{df.residual}{Residual degrees of freedom when supplied by the
#'         internal Wald method.}
#' }
#' Print, summary, and plot methods are available; see
#' \code{\link{print.wald_lgspline}}, \code{\link{summary.wald_lgspline}},
#' \code{\link{plot.wald_lgspline}}.
#'
#' @examples
#'
#' set.seed(1234)
#' t <- runif(1000, -10, 10)
#' y <- 2*sin(t) + -0.06*t^2 + rnorm(length(t))
#' model_fit <- lgspline(t, y, return_varcovmat = TRUE)
#'
#' wald_default <- wald_univariate(model_fit)
#' print(wald_default)
#'
#' ## t-distribution critical value
#' eff_df <- model_fit$N - model_fit$trace_XUGX
#' wald_t <- wald_univariate(model_fit, cv = qt(0.975, eff_df))
#' print(wald_t)
#'
#' coef_table <- wald_default$coefficients
#' plot(wald_default)
#'
#' @seealso
#' \code{\link{lgspline}}, \code{\link{confint.lgspline}},
#' \code{\link{print.wald_lgspline}}, \code{\link{summary.wald_lgspline}},
#' \code{\link{plot.wald_lgspline}}
#' @export
wald_univariate <- function(object, scale_vcovmat_by = 1, cv, ...) {
  if (is.null(object$varcovmat)) {
    stop("Wald tests require return_varcovmat = TRUE during model fitting")
  }

  if (missing(cv)) {
    if (!is.null(object$critical_value)) {
      cv <- object$critical_value
    } else {
      cv <- stats::qnorm(0.975)
      warning("Critical value 'cv' not provided, defaulting to qnorm(0.975).",
              call. = FALSE)
    }
  }

  ## t-tests for Gaussian identity; z-tests otherwise
  if(object$family$family == 'gaussian' && object$family$link == 'identity'){
    stat_name <- 't value'
    p_name <- 'Pr(>|t|)'
  } else {
    stat_name <- 'z value'
    p_name <- 'Pr(>|z|)'
  }

  internal_wald_func <- object$wald_univariate
  if (is.null(internal_wald_func) || !is.function(internal_wald_func)) {
    stop("Internal wald_univariate method not found or not a function.")
  }

  res <- internal_wald_func(scale_vcovmat_by = scale_vcovmat_by, cv = cv, ...)

  ## If internal method already returns a wald_lgspline object, rebuild the
  #  wrapper return for a consistent structure.
  if (inherits(res, "wald_lgspline")) {
    out <- list(
      coefficients   = res$coefficients,
      critical_value = cv,
      family         = object$family,
      N              = object$N,
      trace_XUGX     = object$trace_XUGX,
      statistic_name = stat_name,
      p_value_name   = p_name
    )
    class(out) <- "wald_lgspline"
    return(out)
  }

  ## Normalize to a labelled matrix.
  #  The internal method may return a list of vectors (old-style interface).
  coef_mat <- NULL
  if (is.list(res) && !is.data.frame(res) && !is.matrix(res)) {
    coef_mat <- tryCatch({
      mat <- cbind(
        Estimate      = res$est,
        `Std. Error`  = res$se,
        Statistic     = res$stat,
        `CI LB`       = res$interval_lb,
        `CI UB`       = res$interval_ub,
        p.value       = res$pval
      )
      colnames(mat)[3] <- stat_name
      colnames(mat)[6] <- p_name
      mat <- mat[, c("Estimate", "Std. Error", stat_name,
                     p_name, "CI LB", "CI UB"), drop = FALSE]
      mat
    }, error = function(e) NULL)
    if (is.null(coef_mat)) {
      coef_mat <- tryCatch(Reduce('cbind', res), error = function(e) NULL)
    }
  } else if (is.matrix(res) || is.data.frame(res)) {
    coef_mat <- as.matrix(res)
  }

  if (is.null(coef_mat)) {
    warning("Could not normalize wald_univariate output to matrix.",
            call. = FALSE)
    coef_mat <- cbind(Estimate = unlist(object$B))
  }

  ## Assign row names from coefficient names when missing
  if (is.null(rownames(coef_mat)) && !is.null(object$B) && is.list(object$B) &&
      length(unlist(object$B)) == nrow(coef_mat)) {
    rn <- tryCatch({
      unlist(lapply(seq_along(object$B), function(k) {
        part_names <- names(object$B[[k]])
        if (is.null(part_names))
          part_names <- paste0("Term", seq_len(length(object$B[[k]])))
        paste0("partition", k, "_", part_names)
      }))
    }, error = function(e) NULL)
    if (!is.null(rn) && length(rn) == nrow(coef_mat)) {
      rownames(coef_mat) <- rn
    }
  }

  out <- list(
    coefficients   = coef_mat,
    critical_value = cv,
    family         = object$family,
    N              = object$N,
    trace_XUGX     = object$trace_XUGX,
    statistic_name = stat_name,
    p_value_name   = p_name,
    df.residual    = res$df.residual
  )
  class(out) <- "wald_lgspline"
  return(out)
}


#' Print Method for wald_lgspline Objects
#'
#' Prints the coefficient table using \code{\link[stats]{printCoefmat}} with
#' significance stars.
#'
#' @param x A \code{"wald_lgspline"} object from \code{\link{wald_univariate}}.
#' @param digits Number of significant digits.
#' @param signif.stars Logical; show significance stars.
#' @param ... Additional arguments passed to \code{\link[stats]{printCoefmat}}.
#'
#' @return Invisibly returns \code{x}.
#'
#' @seealso \code{\link{wald_univariate}}, \code{\link[stats]{printCoefmat}}
#' @method print wald_lgspline
#' @export
print.wald_lgspline <- function(x, digits = max(3, getOption("digits") - 3),
                                signif.stars = getOption("show.signif.stars"),
                                ...) {
  cat("\nWald Inference for lgspline Coefficients\n")
  cat("Family:", paste0(x$family)[1],
      " Link:", paste0(x$family)[2], "\n")
  if(!is.null(x$N)) cat("N =", x$N, "\n\n")

  mat <- x$coefficients
  pval_col <- grep("^Pr\\(", colnames(mat))
  if(length(pval_col) > 0){
    ## printCoefmat expects p-value column last
    other_cols <- setdiff(seq_len(ncol(mat)), pval_col)
    print_mat <- mat[, c(other_cols, pval_col), drop = FALSE]
    pval_col_new <- ncol(print_mat)
    tst_col <- grep("value$", colnames(print_mat))
    tst_col <- tst_col[!tst_col %in% pval_col_new]
    stats::printCoefmat(print_mat,
                        digits = digits,
                        signif.stars = signif.stars,
                        cs.ind = 1:2,
                        tst.ind = if(length(tst_col) > 0) tst_col[1] else NULL,
                        P.values = TRUE,
                        has.Pvalue = TRUE,
                        df = if(!is.null(x$df.residual) &&
                                length(x$df.residual) > 0 &&
                                is.finite(x$df.residual)) x$df.residual else NULL,
                        ...)
  } else {
    print(mat, ...)
  }
  cat("\nCritical value:", x$critical_value, "\n")
  if(!is.null(x$df.residual) && is.finite(x$df.residual)){
    cat("Residual degrees of freedom:", round(x$df.residual, 2), "\n")
  }
  invisible(x)
}


#' Summary Method for wald_lgspline Objects
#'
#' Prints a header with model info then delegates to
#' \code{\link{print.wald_lgspline}}.
#'
#' @param object A \code{"wald_lgspline"} object.
#' @param ... Passed to \code{\link{print.wald_lgspline}}.
#'
#' @return Invisibly returns \code{object}.
#'
#' @seealso \code{\link{wald_univariate}}, \code{\link{print.wald_lgspline}}
#' @method summary wald_lgspline
#' @export
summary.wald_lgspline <- function(object, ...) {
  cat("Univariate Wald Tests Summary\n")
  cat("=============================\n")
  cat("Family:", object$family$family, "\n")
  cat("Link:", object$family$link, "\n")
  cat("N:", object$N, "\n")
  if(!is.null(object$N) && !is.null(object$trace_XUGX)){
    cat("Effective df:", object$N - object$trace_XUGX, "\n")
  }
  cat("Test statistic:", object$statistic_name, "\n")
  cat("Critical value:", object$critical_value, "\n")
  cat("-----------------------------\n")
  print(object)
  invisible(object)
}


#' Plot Method for wald_lgspline Objects
#'
#' Forest-style plot of coefficient estimates with confidence intervals.
#'
#' @param x A \code{"wald_lgspline"} object.
#' @param parm Integer vector of coefficient indices or character vector of
#'        names to plot. Default NULL plots all.
#' @param which Integer vector of coefficient indices to plot (alternative to
#'        \code{parm}). Default NULL.
#' @param main Plot title. Default \code{"Coefficient Estimates and CIs"}.
#' @param xlab x-axis label. Default \code{"Estimate"}.
#' @param ... Additional arguments passed to \code{\link[graphics]{plot}}.
#'
#' @return Invisibly returns NULL.
#'
#' @seealso \code{\link{wald_univariate}}, \code{\link{confint.lgspline}}
#' @method plot wald_lgspline
#' @export
plot.wald_lgspline <- function(x,
                               parm = NULL,
                               which = NULL,
                               main = "Coefficient Estimates and CIs",
                               xlab = "Estimate",
                               ...) {
  mat <- x$coefficients

  ## Support both parm (named/indexed) and which (index-only)
  if(!is.null(parm)){
    if(is.character(parm)){
      mat <- mat[rownames(mat) %in% parm, , drop = FALSE]
    } else {
      mat <- mat[parm, , drop = FALSE]
    }
  } else if(!is.null(which)){
    mat <- mat[which, , drop = FALSE]
  }

  est_col <- which(colnames(mat) == "Estimate")
  lb_col  <- grep("CI LB|^Lower", colnames(mat))
  ub_col  <- grep("CI UB|^Upper", colnames(mat))

  if(length(est_col) == 0 || length(lb_col) == 0 || length(ub_col) == 0){
    ## Try x$est path for old-style objects
    if(!is.null(x$est)){
      idx_sel <- if(!is.null(which)) which else seq_along(x$est)
      est <- x$est[idx_sel]
      lb  <- x$interval_lb[idx_sel]
      ub  <- x$interval_ub[idx_sel]
      nms <- names(est)
      if(is.null(nms)) nms <- paste0("beta[", idx_sel, "]")
      n <- length(est)
      xlim <- range(c(lb, ub), na.rm = TRUE)
      graphics::plot(est, n:1, xlim = xlim, yaxt = "n",
                     ylab = "", xlab = xlab, main = main,
                     pch = 16, ...)
      graphics::segments(lb, n:1, ub, n:1)
      graphics::axis(2, at = n:1, labels = nms, las = 1, cex.axis = 0.7)
      graphics::abline(v = 0, lty = 2, col = "grey50")
      return(invisible(x))
    }
    warning("Cannot produce forest plot: missing Estimate, CI LB, or CI UB columns.",
            call. = FALSE)
    print(mat)
    return(invisible(NULL))
  }

  ests <- mat[, est_col]
  lbs  <- mat[, lb_col]
  ubs  <- mat[, ub_col]
  n_coef <- length(ests)
  idx <- seq_len(n_coef)

  labs <- rownames(mat)
  if(is.null(labs)) labs <- paste0("Coef ", idx)

  xlims <- range(c(lbs, ubs), na.rm = TRUE)
  xlims <- xlims + diff(xlims) * c(-0.05, 0.05)

  oldpar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(oldpar))
  graphics::par(mar = c(4.5, max(nchar(labs)) * 0.5 + 2, 3, 1))

  graphics::plot(ests, idx, xlim = xlims, yaxt = "n",
                 ylab = "", xlab = xlab, main = main,
                 pch = 19, ...)
  graphics::axis(2, at = idx, labels = labs, las = 1, cex.axis = 0.7)
  graphics::segments(lbs, idx, ubs, idx, lwd = 2)
  graphics::abline(v = 0, lty = 2, col = "grey50")
  invisible(NULL)
}


#' Extract Coefficients from a wald_lgspline Object
#'
#' @param object A \code{"wald_lgspline"} object.
#' @param ... Not used.
#'
#' @return Named numeric vector of coefficient estimates, or NULL if no
#'   estimate column is available.
#' @export
coef.wald_lgspline <- function(object, ...){
  if(!is.null(object$est)) return(object$est)
  if(!is.null(object$coefficients)){
    est_col <- which(colnames(object$coefficients) == "Estimate")
    if(length(est_col) > 0) return(object$coefficients[, est_col])
  }
  NULL
}


#' Extract Confidence Intervals from a wald_lgspline Object
#'
#' @param object A \code{"wald_lgspline"} object.
#' @param parm Parameter specification (ignored; all returned).
#' @param level Confidence level (ignored; uses the object's critical value).
#' @param ... Not used.
#'
#' @return Matrix with columns lower and upper, or NULL if confidence
#'   limits are not available.
#' @export
confint.wald_lgspline <- function(object, parm = NULL, level = NULL, ...){
  if(!is.null(object$interval_lb) && !is.null(object$interval_ub)){
    return(cbind(lower = object$interval_lb, upper = object$interval_ub))
  }
  if(!is.null(object$coefficients)){
    lb_col <- grep("CI LB|^Lower", colnames(object$coefficients))
    ub_col <- grep("CI UB|^Upper", colnames(object$coefficients))
    if(length(lb_col) > 0 && length(ub_col) > 0){
      return(cbind(lower = object$coefficients[, lb_col],
                   upper = object$coefficients[, ub_col]))
    }
  }
  NULL
}


#' Confidence Intervals for lgspline Coefficients
#'
#' Wald-based confidence intervals for regression coefficients and, when
#' available, correlation parameters (on the working scale).
#'
#' @details
#' For Gaussian identity-link models, t-distribution quantiles are used with
#' effective degrees of freedom
#' \eqn{N - \mathrm{trace}(\mathbf{XUGX}^\top)}.
#' All other families use normal quantiles.
#'
#' Correlation parameter intervals (if \code{VhalfInv_params_estimates} and
#' \code{VhalfInv_params_vcov} are present) are computed on the unbounded
#' working scale via a Wald interval.
#' @param object A fitted lgspline object with \code{return_varcovmat = TRUE}.
#' @param parm Optional vector of parameter indices or names. Default returns
#'        all regression parameters; working-scale correlation parameters are
#'        appended when available.
#' @param level Confidence level. Default 0.95.
#' @param ... Additional arguments passed to \code{\link{wald_univariate}}.
#'
#' @return A matrix with columns giving lower and upper confidence limits,
#'   named e.g. \code{2.5 \%} and \code{97.5 \%} for 95\% intervals.
#'   When available, rows for working-scale correlation parameters are
#'   appended after the regression coefficients.
#'
#' @method confint lgspline
#' @export
confint.lgspline <- function(object, parm, level = 0.95, ...) {
  if(is.null(object$varcovmat)){
    stop("confint requires return_varcovmat = TRUE during model fitting")
  }

  alpha <- 1 - level

  if(object$family$family == "gaussian" && object$family$link == "identity"){
    eff_df <- object$N - object$trace_XUGX
    if(!is.null(eff_df) && !is.na(eff_df) && is.finite(eff_df) && eff_df > 0){
      cv <- stats::qt(1 - alpha / 2, df = eff_df)
    } else {
      cv <- stats::qnorm(1 - alpha / 2)
      warning("Effective df non-positive; using normal quantiles.", call. = FALSE)
    }
  } else {
    cv <- stats::qnorm(1 - alpha / 2)
  }

  wald_res <- wald_univariate(object, cv = cv, ...)
  coef_mat <- wald_res$coefficients

  ## CI column detection
  lb_col <- grep("CI LB|Lower", colnames(coef_mat), ignore.case = TRUE)
  ub_col <- grep("CI UB|Upper", colnames(coef_mat), ignore.case = TRUE)
  if(length(lb_col) == 0 || length(ub_col) == 0){
    stop("Could not extract CI columns from wald_univariate output.")
  }
  lb_col <- lb_col[1]
  ub_col <- ub_col[1]

  ci <- coef_mat[, c(lb_col, ub_col), drop = FALSE]
  pct <- format(100 * c(alpha / 2, 1 - alpha / 2), digits = 3, trim = TRUE)
  colnames(ci) <- paste0(pct, " %")

  if(!missing(parm)){
    if(is.character(parm)){
      ci <- ci[rownames(ci) %in% parm, , drop = FALSE]
    } else {
      ci <- ci[parm, , drop = FALSE]
    }
  }

  ## Append correlation parameter intervals on the working scale
  if(!is.null(object$VhalfInv_params_estimates) &&
     !is.null(object$VhalfInv_params_vcov) &&
     all(!is.na(object$VhalfInv_params_estimates)) &&
     all(!is.na(object$VhalfInv_params_vcov))){

    par_est <- object$VhalfInv_params_estimates
    vcov_mat <- object$VhalfInv_params_vcov
    se <- sqrt(diag(vcov_mat))

    for(i in seq_along(par_est)){
      work_ci <- par_est[i] + c(-cv, cv) * se[i]
      ci_corr <- matrix(work_ci, nrow = 1)
      colnames(ci_corr) <- colnames(ci)
      rownames(ci_corr) <- paste0("Correlation parameter ", i)
      ci <- rbind(ci, ci_corr)
    }
  }

  return(ci)
}


#' Extract Log-Likelihood from a Fitted lgspline
#'
#' Returns the log-likelihood as a \code{"logLik"} object for use with
#' \code{\link[stats]{AIC}}, \code{\link[stats]{BIC}}, and other model
#' comparison tools.
#'
#' @details
#' \strong{Gaussian identity, no correlation.}
#' \deqn{
#'   \ell = -\frac{N}{2}\log(2\pi\tilde{\sigma}^2) -
#'   \frac{1}{2\tilde{\sigma}^2}\sum_{i}(y_i - \hat{y}_i)^2
#' }
#'
#' \strong{Gaussian identity, with correlation.}
#' GLS log-likelihood:
#' \deqn{
#'   \ell = -\frac{N}{2}\log(2\pi\tilde{\sigma}^2)
#'   + \log|\mathbf{V}^{-1/2}|
#'   - \frac{1}{2\tilde{\sigma}^2}
#'     \|\mathbf{V}^{-1/2}(\mathbf{y} - \hat{\mathbf{y}})\|^2
#' }
#' \eqn{\log|\mathbf{V}^{-1/2}|} is obtained from \code{VhalfInv_logdet}
#' when available, or computed directly from \code{VhalfInv}.
#'
#' \strong{Prior contribution.}
#' When \code{include_prior = TRUE} (default), the log-prior
#' \deqn{
#'   -\frac{1}{2\tilde{\sigma}^2}
#'   \sum_{k}\boldsymbol{\beta}_k^\top\boldsymbol{\Lambda}_k
#'   \boldsymbol{\beta}_k
#' }
#' is added, giving the penalised MAP log-likelihood coherent with the
#' smoothing spline objective. Set \code{include_prior = FALSE} for the
#' unpenalised marginal likelihood, which is more appropriate when comparing
#' models with different penalty structures or numbers of knots.
#'
#' \strong{Other GLM families.}
#' Uses \code{family$aic()} when available. For correlated models the
#' whitened residuals and fitted values are passed. When \code{family$aic()}
#' is unavailable, a deviance-based approximation is used (valid for
#' relative comparisons; a warning is emitted).
#'
#' This function returns the marginal (full) GLS log-likelihood, not the
#' REML log-likelihood. This is consistent with \code{REML = FALSE} in
#' \code{lme} and \code{gls}, and is the conventional choice for AIC/BIC
#' comparisons of fixed-effects structure.
#'
#' The \code{df} attribute is set to \eqn{N - \mathrm{trace}(\mathbf{XUGX}^\top)}.
#'
#' @param object A fitted lgspline model object.
#' @param include_prior Logical; add the log-prior penalty term. Default TRUE.
#' @param new_weights Numeric scalar or N-vector; optional observation weights
#'        overriding \code{object$weights}.
#' @param ... Not used.
#'
#' @return A \code{"logLik"} object with attributes \code{df} (effective
#'   degrees of freedom) and \code{nobs} (number of observations).
#'
#' @examples
#'
#' set.seed(1234)
#' t <- runif(1000, -10, 10)
#' y <- 2*sin(t) + -0.06*t^2 + rnorm(length(t))
#' model_fit <- lgspline(t, y)
#'
#' logLik(model_fit)
#' logLik(model_fit, include_prior = FALSE)
#'
#' AIC(model_fit)
#' BIC(model_fit)
#'
#' ## Compare models with different K using unpenalized likelihood
#' fit_k3 <- lgspline(t, y, K = 3)
#' fit_k7 <- lgspline(t, y, K = 7)
#' AIC(fit_k3, fit_k7)
#'
#' @seealso
#' \code{\link{lgspline}}, \code{\link{prior_loglik}},
#' \code{\link[stats]{logLik}}, \code{\link[stats]{AIC}},
#' \code{\link[stats]{BIC}}
#' @method logLik lgspline
#' @export
logLik.lgspline <- function(object,
                            include_prior = TRUE,
                            new_weights = NULL, ...) {
  N  <- object$N
  mu <- object$ytilde
  y  <- object$y
  sigma2 <- object$sigmasq_tilde
  fam <- object$family
  if(!is.null(new_weights)){
    if(length(new_weights) %in% c(1, N)){
      wt <- new_weights
    } else {
      stop('\n\t weights argument must be a scalar or N-length vector\n')
    }
  } else {
    wt  <- object$weights
  }

  edf <- if(!is.null(object$trace_XUGX)) N - object$trace_XUGX else NA_real_
  has_corr <- !is.null(object$VhalfInv)

  if(has_corr){
    logdet_Vhalfinv <- tryCatch({
      if(!is.null(object$VhalfInv_logdet) &&
         !is.null(object$VhalfInv_params_estimates)){
        object$VhalfInv_logdet(object$VhalfInv_params_estimates)
      } else {
        determinant(object$VhalfInv, logarithm = TRUE)$modulus[[1L]]
      }
    }, error = function(e) NA_real_)

    resid_w <- tryCatch(
      c(object$VhalfInv %**% cbind(y - mu)),
      error = function(e) NULL
    )
  }

  ll <- NA_real_

  ## First try the exact Gaussian formulas when they apply.
  if(fam$family == "gaussian" && fam$link == "identity"){

    if(has_corr && !is.null(resid_w) && is.finite(logdet_Vhalfinv)){
      ss_w <- sum(wt*(resid_w)^2)
      ll   <- -0.5 * N * log(2 * pi * sigma2) +
        logdet_Vhalfinv -
        0.5 * ss_w / sigma2
    } else {
      resid <- y - mu
      ss    <- sum(wt*(resid)^2)
      ll    <- -0.5 * N * log(2 * pi * sigma2) - 0.5 * ss / sigma2
    }

  } else if(!is.null(fam$aic)){

    ## Otherwise ask the family for its AIC/log-likelihood contribution.
    ## For correlated models, pass whitened quantities to family$aic().
    #  For families where aic() depends on the raw scale (e.g. binomial)
    #  this is an approximation.
    if(has_corr && !is.null(resid_w)){
      y_eval  <- tryCatch(c(object$VhalfInv %*% cbind(y)), error = function(e) y)
      mu_eval <- tryCatch(c(object$VhalfInv %*% cbind(mu)), error = function(e) mu)
    } else {
      y_eval  <- y
      mu_eval <- mu
    }

    dev_resids <- tryCatch(fam$dev.resids(y_eval, mu_eval, wt), error = function(e) NULL)

    if(!is.null(dev_resids)){
      dev <- sum(dev_resids)
      aic_val <- tryCatch(
        fam$aic(y_eval, rep(1, N), mu_eval, wt, dev),
        error = function(e) NULL
      )
      if(!is.null(aic_val) && is.finite(aic_val)){
        ll_base <- -0.5 * aic_val
        ll <- if(has_corr && is.finite(logdet_Vhalfinv)){
          ll_base + logdet_Vhalfinv
        } else {
          ll_base
        }
      }
    }
  }

  ## Fallback: deviance-based approximation
  if(!is.finite(ll)){

    if(has_corr && !is.null(resid_w)){
      y_fb  <- tryCatch(c(object$VhalfInv %*% cbind(y)), error = function(e) y)
      mu_fb <- tryCatch(c(object$VhalfInv %*% cbind(mu)), error = function(e) mu)
    } else {
      y_fb  <- y
      mu_fb <- mu
    }

    dev_resids <- tryCatch(fam$dev.resids(y_fb, mu_fb, wt), error = function(e) NULL)

    if(!is.null(dev_resids)){
      dev <- sum(dev_resids)
      ll_base <- -0.5 * dev / sigma2
      ll <- if(has_corr && is.finite(logdet_Vhalfinv)){
        ll_base + logdet_Vhalfinv
      } else {
        ll_base
      }
      warning(
        paste0(
          "Log-likelihood computed via deviance approximation",
          if(has_corr) " with correlation structure log-determinant correction" else "",
          "; a family-specific constant may be omitted. ",
          "Valid for relative model comparison only."
        ),
        call. = FALSE
      )
    } else {
      warning("Could not compute log-likelihood for this family.", call. = FALSE)
    }
  }

  ## Add the penalty-induced prior term when requested.
  if(include_prior && is.finite(ll)){
    lp <- tryCatch(
      as.numeric(prior_loglik(object, sigmasq = sigma2)),
      error = function(e) {
        warning("Could not evaluate prior_loglik; prior term omitted.", call. = FALSE)
        0
      }
    )
    ll <- ll + lp
  }

  attr(ll, "df")   <- edf
  attr(ll, "nobs") <- N
  class(ll)        <- "logLik"
  return(ll)
}


#' Print Closed-Form Fitted Equation from lgspline Model
#'
#' @description
#' Displays the closed-form polynomial equation for each partition of a fitted
#' lgspline model, along with partition boundary or cluster center information.
#' Optionally prints the first derivative, second derivative, or antiderivative
#' of the fitted equation with respect to a single specified variable.
#'
#' @param object A fitted lgspline model object.
#' @param x An object returned by \code{equation()} for printing.
#' @param digits Integer; decimal places for coefficient display. Default 4.
#' @param scientific Logical; use scientific notation for coefficients with
#'        absolute value < 1e-3 or > 1e4. Default FALSE.
#' @param show_bounds Logical; display partition bounds (1D) or knot midpoint
#'        boundaries (multi-D). Default TRUE.
#' @param predictor_names Character vector; custom names for predictor variables.
#'        If NULL (default), uses original column names or "_j_" labels.
#' @param response_name Character; label for response. If NULL (default), uses
#'        "y" for identity link Gaussian, or "link(E[y])" otherwise.
#' @param collapse_zero Logical; omit terms with coefficient exactly 0.
#'        Default TRUE.
#' @param first_derivative Default: NULL. Character name or integer index of
#'        the predictor variable with respect to which the first derivative
#'        is printed. Only one variable at a time is supported. When non-NULL,
#'        the printed equations show \eqn{df/dx_j} for each partition.
#' @param second_derivative Default: NULL. Character name or integer index of
#'        the predictor variable with respect to which the second derivative
#'        is printed. Only one variable at a time is supported. When non-NULL,
#'        the printed equations show \eqn{d^2f/dx_j^2} for each partition.
#'        Ignored if \code{first_derivative} is also non-NULL.
#' @param antiderivative Default: NULL. Character name or integer index of
#'        the predictor variable with respect to which the antiderivative
#'        (indefinite integral) is printed. Only one variable at a time is
#'        supported. When non-NULL, the printed equations show
#'        \eqn{\int f\, dx_j} for each partition, with an unspecified
#'        constant of integration \eqn{C}. Ignored if \code{first_derivative}
#'        or \code{second_derivative} is also non-NULL.
#' @param ... Not used.
#'
#' @details
#' For 1D models with K knots, partition boundaries are displayed as intervals
#' on the predictor scale. For multi-predictor models, partition boundaries are
#' computed as the midpoints between adjacent cluster centers along each
#' predictor dimension. When the model's \code{make_partition_list} contains
#' \code{knots} (midpoint boundaries between clusters), those are used directly.
#' Otherwise, cluster centers are displayed.
#'
#' Coefficients are displayed on the original (unstandardized) predictor scale.
#' For GLMs with non-identity link, the left-hand side shows the link function
#' applied to the expected response.
#'
#' \strong{Derivative and antiderivative modes.}
#' Only one of \code{first_derivative}, \code{second_derivative}, or
#' \code{antiderivative} may be non-NULL. If more than one is supplied, the
#' priority order is: first derivative, second derivative, antiderivative.
#'
#' Derivatives and antiderivatives are computed symbolically from the
#' polynomial coefficients. For a term \eqn{a x^n}, the first derivative is
#' \eqn{n a x^{n-1}}, the second derivative is \eqn{n(n-1) a x^{n-2}}, and
#' the antiderivative is \eqn{a x^{n+1}/(n+1)}. Cross-terms (interactions)
#' involving the target variable are differentiated or integrated with respect
#' to that variable only, treating all other variables as constants.
#'
#' A warning is emitted if the user attempts to differentiate or integrate with
#' respect to more than one variable simultaneously. Multi-variable calculus
#' operations should be performed one variable at a time by calling
#' \code{equation()} repeatedly.
#'
#' @return Invisibly returns a list with components:
#' \describe{
#'   \item{formulas}{Character vector of equation strings per partition.}
#'   \item{bounds}{Matrix or list of partition boundary information.}
#'   \item{link}{Character; link function name.}
#'   \item{mode}{Character; one of "equation", "first_derivative",
#'         "second_derivative", or "antiderivative".}
#'   \item{variable}{Character; the variable name for the calculus operation,
#'         or NULL if mode is "equation".}
#' }
#'
#' @examples
#'
#' ## 1D example
#' set.seed(1234)
#' t <- runif(500, -5, 5)
#' y <- 2*sin(t) + 0.1*t^2 + rnorm(length(t), 0, 0.5)
#' fit <- lgspline(t, y, K = 2)
#' equation(fit)
#' equation(fit, digits = 2, predictor_names = "time")
#'
#' ## First derivative with respect to predictor
#' equation(fit, first_derivative = 1)
#'
#' ## Second derivative
#' equation(fit, second_derivative = 1)
#'
#' ## Antiderivative
#' equation(fit, antiderivative = 1)
#'
#' ## 2D example with named predictors
#' x1 <- runif(300, 0, 10)
#' x2 <- runif(300, 0, 10)
#' y <- x1 + 0.5*x2 + 0.1*x1*x2 + rnorm(300)
#' fit2d <- lgspline(cbind(x1, x2), y, K = 3)
#' equation(fit2d, predictor_names = c("Length", "Width"))
#'
#' ## Derivative w.r.t. first variable only
#' equation(fit2d, first_derivative = "Length",
#'          predictor_names = c("Length", "Width"))
#'
#' ## GLM example
#' y_bin <- rbinom(500, 1, plogis(0.5*t))
#' fit_glm <- lgspline(t, y_bin, family = binomial(), K = 1)
#' equation(fit_glm)
#'
#' @seealso \code{\link{lgspline}}, \code{\link{plot.lgspline}},
#'   \code{\link{coef.lgspline}}
#'
#' @export
equation <- function(object, ...) {

  UseMethod("equation")
}

#' @rdname equation
#' @method equation lgspline
#' @export
equation.lgspline <- function(object,
                              digits = 4,
                              scientific = FALSE,
                              show_bounds = TRUE,
                              predictor_names = NULL,
                              response_name = NULL,
                              collapse_zero = TRUE,
                              first_derivative = NULL,
                              second_derivative = NULL,
                              antiderivative = NULL,
                              ...) {

  K  <- object$K
  nc <- object$p
  q  <- object$q
  b_names <- names(object$B[[1]])
  raw_nms <- object$raw_expansion_names

  ## Resolve predictor display names from B[[1]] at linear positions
  all_linear_cols <- sort(c(object$power1_cols, object$nonspline_cols))
  if (is.null(predictor_names)) {
    if (!is.null(b_names) && length(all_linear_cols) > 0) {
      predictor_names <- b_names[all_linear_cols]
    } else if (!is.null(raw_nms) && length(all_linear_cols) > 0) {
      predictor_names <- raw_nms[all_linear_cols]
    } else {
      predictor_names <- paste0("x", 1:q)
    }
  }
  if (length(predictor_names) < q) {
    predictor_names <- c(
      predictor_names,
      paste0("x", (length(predictor_names) + 1):q)
    )
  }

  ## Map from internal column index (the number inside "_j_") to
  #  1-based predictor display index into predictor_names.
  internal_idx_to_pred <- integer(0)
  if (!is.null(raw_nms) && length(all_linear_cols) > 0) {
    for (jj in seq_along(all_linear_cols)) {
      col_idx <- all_linear_cols[jj]
      raw_nm <- raw_nms[col_idx]
      idx_match <- regmatches(raw_nm, regexec("^_([0-9]+)_$", raw_nm))[[1]]
      if (length(idx_match) == 2) {
        internal_idx_to_pred[as.integer(idx_match[2])] <- jj
      }
    }
  }

  ## Calculus mode
  calc_mode <- "equation"
  calc_var_idx <- NULL
  calc_var_name <- NULL

  .resolve_calc_var <- function(spec, predictor_names, q, arg_name) {
    if (length(spec) > 1) {
      warning(
        "\n\t ", arg_name, " accepts only one variable at a time. ",
        "Only the first element will be used. Call equation() separately ",
        "for each variable.\n"
      )
      spec <- spec[1]
    }
    if (is.character(spec)) {
      idx <- match(spec, predictor_names)
      if (is.na(idx)) {
        stop(
          "\n\t ", arg_name, " variable '", spec,
          "' not found in predictor names: ",
          paste(predictor_names, collapse = ", "), ".\n"
        )
      }
      list(idx = idx, name = predictor_names[idx])
    } else {
      idx <- as.integer(spec)
      if (idx < 1 || idx > q) {
        stop(
          "\n\t ", arg_name, " index ", idx,
          " is out of range [1, ", q, "].\n"
        )
      }
      list(idx = idx, name = predictor_names[idx])
    }
  }

  if (!is.null(first_derivative)) {
    calc_mode <- "first_derivative"
    res <- .resolve_calc_var(first_derivative, predictor_names, q,
                             "first_derivative")
    calc_var_idx  <- res$idx
    calc_var_name <- res$name
  } else if (!is.null(second_derivative)) {
    calc_mode <- "second_derivative"
    res <- .resolve_calc_var(second_derivative, predictor_names, q,
                             "second_derivative")
    calc_var_idx  <- res$idx
    calc_var_name <- res$name
  } else if (!is.null(antiderivative)) {
    calc_mode <- "antiderivative"
    res <- .resolve_calc_var(antiderivative, predictor_names, q,
                             "antiderivative")
    calc_var_idx  <- res$idx
    calc_var_name <- res$name
  }

  ## Response name from link function
  fam <- object$family
  if (is.null(response_name)) {
    if (fam$family == "gaussian" && fam$link == "identity") {
      response_name <- "y"
    } else {
      response_name <- paste0(fam$link, "(E[y])")
    }
  }

  ## Parse a raw_expansion_name into structured factors.
  #  Raw names use "_j_" for variables and "x" (no underscores) as
  #  the interaction separator between complete "_j_" tokens.
  #  Examples: "_1_", "_1_^2", "_1_x_2_", "_2_x_1_^2", "_1_x_2_x_3_"
  .parse_term <- function(nm) {
    if (nm == "intercept" ||
        grepl("^intercept$", nm, ignore.case = TRUE)) {
      return(list(is_intercept = TRUE, factors = list()))
    }

    ## Extract all "_j_" and "_j_^n" tokens from the raw name
    #  Pattern: underscore, digits, underscore, optionally ^digits
    matches <- gregexpr("_([0-9]+)_(\\^[0-9]+)?", nm)
    tokens <- regmatches(nm, matches)[[1]]

    if (length(tokens) == 0) {
      return(list(is_intercept = FALSE,
                  factors = list(list(pred_idx = NA_integer_, power = 1L))))
    }

    factors <- lapply(tokens, function(tok) {
      ## Extract power if present
      pow_match <- regmatches(tok, regexec("\\^([0-9]+)$", tok))[[1]]
      if (length(pow_match) == 2) {
        power <- as.integer(pow_match[2])
        base_tok <- sub("\\^[0-9]+$", "", tok)
      } else {
        power <- 1L
        base_tok <- tok
      }
      ## Extract the column number
      idx_match <- regmatches(
        base_tok, regexec("^_([0-9]+)_$", base_tok)
      )[[1]]
      if (length(idx_match) == 2) {
        raw_col <- as.integer(idx_match[2])
        pred_idx <- internal_idx_to_pred[raw_col]
        if (is.na(pred_idx)) pred_idx <- NA_integer_
      } else {
        pred_idx <- NA_integer_
      }
      list(pred_idx = pred_idx, power = power)
    })
    list(is_intercept = FALSE, factors = factors)
  }

  ## Differentiate a single term w.r.t. variable at index var_idx
  .differentiate_term <- function(coef_val, parsed, var_idx) {
    if (parsed$is_intercept) return(NULL)
    hit <- NULL
    hit_pos <- 0L
    for (i in seq_along(parsed$factors)) {
      if (!is.na(parsed$factors[[i]]$pred_idx) &&
          parsed$factors[[i]]$pred_idx == var_idx) {
        hit <- parsed$factors[[i]]
        hit_pos <- i
        break
      }
    }
    if (is.null(hit)) return(NULL)
    new_coef <- coef_val * hit$power
    new_power <- hit$power - 1L
    new_factors <- parsed$factors
    if (new_power == 0L) {
      new_factors[[hit_pos]] <- NULL
    } else {
      new_factors[[hit_pos]]$power <- new_power
    }
    list(coef = new_coef,
         parsed = list(is_intercept = (length(new_factors) == 0),
                       factors = new_factors))
  }

  ## Second derivative: differentiate twice
  .differentiate2_term <- function(coef_val, parsed, var_idx) {
    first <- .differentiate_term(coef_val, parsed, var_idx)
    if (is.null(first)) return(NULL)
    if (first$parsed$is_intercept) return(NULL)
    .differentiate_term(first$coef, first$parsed, var_idx)
  }

  ## Antidifferentiate a single term w.r.t. variable at index var_idx
  .antidifferentiate_term <- function(coef_val, parsed, var_idx) {
    if (parsed$is_intercept) {
      new_factors <- list(list(pred_idx = var_idx, power = 1L))
      return(list(coef = coef_val,
                  parsed = list(is_intercept = FALSE,
                                factors = new_factors)))
    }
    hit_pos <- 0L
    for (i in seq_along(parsed$factors)) {
      if (!is.na(parsed$factors[[i]]$pred_idx) &&
          parsed$factors[[i]]$pred_idx == var_idx) {
        hit_pos <- i
        break
      }
    }
    new_factors <- parsed$factors
    if (hit_pos > 0) {
      old_power <- new_factors[[hit_pos]]$power
      new_power <- old_power + 1L
      new_coef <- coef_val / new_power
      new_factors[[hit_pos]]$power <- new_power
    } else {
      new_coef <- coef_val
      new_factors <- c(new_factors,
                       list(list(pred_idx = var_idx, power = 1L)))
    }
    list(coef = new_coef,
         parsed = list(is_intercept = FALSE, factors = new_factors))
  }

  ## Rebuild human-readable term from parsed structure
  .rebuild_term_name <- function(parsed, predictor_names) {
    if (parsed$is_intercept) return("intercept")
    parts <- vapply(parsed$factors, function(f) {
      if (is.na(f$pred_idx)) {
        nm <- "UNKNOWN"
      } else if (f$pred_idx <= length(predictor_names)) {
        nm <- predictor_names[f$pred_idx]
      } else {
        nm <- paste0("x", f$pred_idx)
      }
      if (f$power == 1) return(nm)
      paste0(nm, "^", f$power)
    }, character(1))
    paste(parts, collapse = "*")
  }

  ## Format one coefficient * term pair for display
  .fmt_coef <- function(val, term_nm, first_term = FALSE) {
    if (collapse_zero && abs(val) < .Machine$double.eps * 100) {
      return(NULL)
    }
    if (scientific && (abs(val) < 1e-3 || abs(val) > 1e4) && val != 0) {
      val_str <- format(val, digits = digits, scientific = TRUE)
    } else {
      val_str <- format(round(val, digits), nsmall = digits, trim = TRUE)
    }
    ## Intercept / constant
    if (term_nm == "intercept" || term_nm == "") {
      if (first_term) return(val_str)
      if (val >= 0) return(paste0(" + ", val_str))
      return(paste0(" - ", gsub("^-", "", val_str)))
    }
    ## Coefficient * term
    if (first_term) {
      if (abs(val - 1) < .Machine$double.eps * 100) return(term_nm)
      if (abs(val + 1) < .Machine$double.eps * 100) return(paste0("-", term_nm))
      return(paste0(val_str, "*", term_nm))
    }
    if (val >= 0) {
      if (abs(val - 1) < .Machine$double.eps * 100) return(paste0(" + ", term_nm))
      return(paste0(" + ", val_str, "*", term_nm))
    }
    if (abs(val + 1) < .Machine$double.eps * 100) return(paste0(" - ", term_nm))
    paste0(" - ", gsub("^-", "", val_str), "*", term_nm)
  }

  ## LHS label
  if (calc_mode == "first_derivative") {
    lhs_label <- paste0("d(", response_name, ")/d(", calc_var_name, ")")
  } else if (calc_mode == "second_derivative") {
    lhs_label <- paste0("d2(", response_name, ")/d(", calc_var_name, ")2")
  } else if (calc_mode == "antiderivative") {
    lhs_label <- paste0("integral ", response_name, " d(", calc_var_name, ")")
  } else {
    lhs_label <- response_name
  }

  ## Build formula string for each partition.
  #  Always parse from raw_expansion_names (reliable _j_ structure),
  #  display with predictor_names (human-readable).
  formulas <- character(K + 1)
  for (k in 1:(K + 1)) {
    coefs <- object$B[[k]]
    terms_out <- character(0)
    first <- TRUE

    for (i in seq_along(coefs)) {
      orig_coef <- coefs[i]
      parsed <- .parse_term(raw_nms[i])

      if (calc_mode == "equation") {
        display_nm <- .rebuild_term_name(parsed, predictor_names)
        term_str <- .fmt_coef(orig_coef, display_nm, first_term = first)
      } else if (calc_mode == "first_derivative") {
        result <- .differentiate_term(orig_coef, parsed, calc_var_idx)
        if (is.null(result)) next
        display_nm <- .rebuild_term_name(result$parsed, predictor_names)
        term_str <- .fmt_coef(result$coef, display_nm, first_term = first)
      } else if (calc_mode == "second_derivative") {
        result <- .differentiate2_term(orig_coef, parsed, calc_var_idx)
        if (is.null(result)) next
        display_nm <- .rebuild_term_name(result$parsed, predictor_names)
        term_str <- .fmt_coef(result$coef, display_nm, first_term = first)
      } else if (calc_mode == "antiderivative") {
        result <- .antidifferentiate_term(orig_coef, parsed, calc_var_idx)
        if (is.null(result)) next
        display_nm <- .rebuild_term_name(result$parsed, predictor_names)
        term_str <- .fmt_coef(result$coef, display_nm, first_term = first)
      }

      if (!is.null(term_str)) {
        terms_out <- c(terms_out, term_str)
        first <- FALSE
      }
    }

    if (length(terms_out) == 0) {
      formulas[k] <- paste0(lhs_label, " = 0")
    } else {
      rhs <- paste0(terms_out, collapse = "")
      if (calc_mode == "antiderivative") rhs <- paste0(rhs, " + C")
      formulas[k] <- paste0(lhs_label, " = ", rhs)
    }
  }

  ## Partition boundary info
  bounds_info <- NULL
  if (show_bounds && K > 0) {
    if (q == 1 && !is.null(object$knots)) {
      ## 1D interval bounds
      knot_vals <- sort(c(object$knots))
      bounds_info <- matrix(NA, nrow = K + 1, ncol = 2)
      colnames(bounds_info) <- c("lower", "upper")
      rownames(bounds_info) <- paste0("Partition ", 1:(K + 1))

      all_x <- unlist(lapply(object$X, function(Xk) {
        if (nrow(Xk) > 0 && length(object$power1_cols) > 0) {
          Xk[, object$power1_cols[1]]
        } else if (nrow(Xk) > 0 && length(object$nonspline_cols) > 0) {
          Xk[, object$nonspline_cols[1]]
        } else {
          NULL
        }
      }))
      if (length(all_x) > 0) {
        x_min <- min(all_x, na.rm = TRUE)
        x_max <- max(all_x, na.rm = TRUE)
      } else {
        x_min <- -Inf
        x_max <- Inf
      }

      bounds_info[1, 1] <- x_min
      bounds_info[1, 2] <- knot_vals[1]
      if (K >= 2) {
        for (kk in 2:K) {
          bounds_info[kk, 1] <- knot_vals[kk - 1]
          bounds_info[kk, 2] <- knot_vals[kk]
        }
      }
      bounds_info[K + 1, 1] <- knot_vals[K]
      bounds_info[K + 1, 2] <- x_max

    } else if (!is.null(object$make_partition_list)) {
      ## Multi-D bounding boxes from cluster centers
      centers <- object$make_partition_list$centers

      if (!is.null(centers)) {
        n_centers <- nrow(centers)
        n_dims <- ncol(centers)

        ## Only show spline predictor dimensions, not indicator columns
        n_spline <- length(object$power1_cols)
        if (n_spline > 0 && n_spline < n_dims) {
          show_dims <- seq_len(n_spline)
        } else {
          show_dims <- seq_len(n_dims)
        }

        center_colnames <- colnames(centers)
        if (is.null(center_colnames)) {
          if (!is.null(predictor_names) && n_dims <= length(predictor_names)) {
            center_colnames <- predictor_names[1:n_dims]
          } else {
            center_colnames <- paste0("x", 1:n_dims)
          }
        }

        ## Training predictor ranges per dimension
        training_ranges <- matrix(NA, 2, n_dims)
        rownames(training_ranges) <- c("min", "max")
        for (jj in seq_len(n_dims)) {
          col_vals <- unlist(lapply(object$X, function(Xk) {
            if (nrow(Xk) > 0 && jj <= length(all_linear_cols)) {
              Xk[, all_linear_cols[jj]]
            } else {
              NULL
            }
          }))
          if (length(col_vals) > 0) {
            training_ranges[1, jj] <- min(col_vals, na.rm = TRUE)
            training_ranges[2, jj] <- max(col_vals, na.rm = TRUE)
          }
        }

        ## Per-partition bounding boxes via midpoints between sorted centers
        bounds_lower <- matrix(NA, n_centers, n_dims)
        bounds_upper <- matrix(NA, n_centers, n_dims)
        colnames(bounds_lower) <- center_colnames
        colnames(bounds_upper) <- center_colnames

        for (jj in seq_len(n_dims)) {
          center_vals_j <- centers[, jj]
          ord_j <- order(center_vals_j)
          sorted_j <- center_vals_j[ord_j]
          if (length(sorted_j) > 1) {
            midpts_j <- (sorted_j[-length(sorted_j)] + sorted_j[-1]) / 2
          } else {
            midpts_j <- numeric(0)
          }
          for (rank in seq_along(ord_j)) {
            part_idx <- ord_j[rank]
            lo <- if (rank == 1 && !is.na(training_ranges[1, jj])) {
              training_ranges[1, jj]
            } else if (rank == 1) {
              -Inf
            } else {
              midpts_j[rank - 1]
            }
            hi <- if (rank == length(ord_j) && !is.na(training_ranges[2, jj])) {
              training_ranges[2, jj]
            } else if (rank == length(ord_j)) {
              Inf
            } else {
              midpts_j[rank]
            }
            bounds_lower[part_idx, jj] <- lo
            bounds_upper[part_idx, jj] <- hi
          }
        }

        bounds_info <- list(
          lower = bounds_lower[, show_dims, drop = FALSE],
          upper = bounds_upper[, show_dims, drop = FALSE],
          centers = centers,
          colnames = center_colnames[show_dims]
        )
      }
    }
  }

  ## Print
  cat("\n")
  div_major <- paste(rep("=", 72), collapse = "")
  div_minor <- paste(rep("-", 72), collapse = "")

  cat(div_major, "\n", sep = "")
  if (calc_mode == "equation") {
    cat("Fitted Equations: lgspline Model\n")
  } else if (calc_mode == "first_derivative") {
    cat("First Derivative w.r.t. ", calc_var_name,
        ": lgspline Model\n", sep = "")
  } else if (calc_mode == "second_derivative") {
    cat("Second Derivative w.r.t. ", calc_var_name,
        ": lgspline Model\n", sep = "")
  } else if (calc_mode == "antiderivative") {
    cat("Antiderivative w.r.t. ", calc_var_name,
        ": lgspline Model\n", sep = "")
  }
  cat(div_major, "\n", sep = "")
  cat("Family:", fam$family, "  Link:", fam$link, "\n")
  cat("Partitions:", K + 1, "  Basis functions per partition:", nc, "\n")
  cat(div_minor, "\n", sep = "")

  for (k in 1:(K + 1)) {
    cat("\n")
    cat("Partition", k, "\n")

    if (show_bounds && !is.null(bounds_info)) {
      if (q == 1 && is.matrix(bounds_info) && ncol(bounds_info) == 2) {
        lb <- bounds_info[k, 1]
        ub <- bounds_info[k, 2]
        lb_str <- if (is.finite(lb)) {
          format(round(lb, digits), nsmall = digits)
        } else { "-Inf" }
        ub_str <- if (is.finite(ub)) {
          format(round(ub, digits), nsmall = digits)
        } else { "+Inf" }
        lb_bracket <- if (k == 1) "[" else "("
        ub_bracket <- if (k == K + 1) "]" else ")"
        cat("  Bounds: ", predictor_names[1], " in ",
            lb_bracket, lb_str, ", ", ub_str, ub_bracket,
            "\n", sep = "")

      } else if (is.list(bounds_info) && !is.null(bounds_info$lower)) {
        n_dims <- ncol(bounds_info$lower)
        cnames <- bounds_info$colnames
        cat("  Bounds:\n")
        for (jj in seq_len(n_dims)) {
          lo <- bounds_info$lower[k, jj]
          hi <- bounds_info$upper[k, jj]
          lo_str <- if (is.finite(lo)) {
            format(round(lo, min(digits, 2)), nsmall = min(digits, 2))
          } else { "-Inf" }
          hi_str <- if (is.finite(hi)) {
            format(round(hi, min(digits, 2)), nsmall = min(digits, 2))
          } else { "+Inf" }
          lb_bracket <- if (lo == bounds_info$lower[
            which.min(bounds_info$lower[, jj]), jj]) "[" else "("
          ub_bracket <- if (hi == bounds_info$upper[
            which.max(bounds_info$upper[, jj]), jj]) "]" else ")"
          cat("    ", cnames[jj], " in ",
              lb_bracket, lo_str, ", ", hi_str, ub_bracket,
              "\n", sep = "")
        }
      }
    }

    cat("  ", formulas[k], "\n", sep = "")
  }

  cat("\n")
  cat(div_minor, "\n", sep = "")
  if (!is.null(object$sigmasq_tilde) && is.finite(object$sigmasq_tilde)) {
    cat("Dispersion (sigma^2):",
        format(object$sigmasq_tilde, digits = digits), "\n")
  }
  if (!is.null(object$trace_XUGX) && is.finite(object$trace_XUGX)) {
    cat("Effective df:",
        format(object$N - object$trace_XUGX, digits = 2), "\n")
  }
  cat(div_major, "\n", sep = "")
  cat("\n")

  invisible(list(
    formulas = formulas,
    bounds   = bounds_info,
    link     = fam$link,
    mode     = calc_mode,
    variable = calc_var_name
  ))
}


#' @rdname equation
#' @method print equation
#' @export
print.equation <- function(x, ...) {
  cat("\nFitted Equations")
  if (!is.null(x$mode) && x$mode != "equation") {
    mode_label <- switch(
      x$mode,
      "first_derivative"  = "First Derivative",
      "second_derivative" = "Second Derivative",
      "antiderivative"    = "Antiderivative"
    )
    cat(" (", mode_label, " w.r.t. ", x$variable, ")", sep = "")
  }
  cat(":\n")
  for (k in seq_along(x$formulas)) {
    cat("  Partition", k, ":", x$formulas[k], "\n")
  }
  if (!is.null(x$bounds)) {
    if (is.matrix(x$bounds)) {
      cat("\nPartition Bounds:\n")
      print(x$bounds)
    } else if (is.list(x$bounds) && !is.null(x$bounds$lower)) {
      cat("\nPartition Bounds (lower):\n")
      print(x$bounds$lower)
      cat("\nPartition Bounds (upper):\n")
      print(x$bounds$upper)
    }
  }
  invisible(x)
}


#' @rdname equation
#' @method equation lgspline
#' @export
equation.lgspline <- function(object,
                              digits = 4,
                              scientific = FALSE,
                              show_bounds = TRUE,
                              predictor_names = NULL,
                              response_name = NULL,
                              collapse_zero = TRUE,
                              first_derivative = NULL,
                              second_derivative = NULL,
                              antiderivative = NULL,
                              ...) {

  K  <- object$K
  nc <- object$p
  q  <- object$q
  b_names <- names(object$B[[1]])
  raw_nms <- object$raw_expansion_names

  ## Resolve predictor display names from B[[1]] at linear positions
  all_linear_cols <- sort(c(object$power1_cols, object$nonspline_cols))
  if (is.null(predictor_names)) {
    if (!is.null(b_names) && length(all_linear_cols) > 0) {
      predictor_names <- b_names[all_linear_cols]
    } else if (!is.null(raw_nms) && length(all_linear_cols) > 0) {
      predictor_names <- raw_nms[all_linear_cols]
    } else {
      predictor_names <- paste0("x", 1:q)
    }
  }
  if (length(predictor_names) < q) {
    predictor_names <- c(
      predictor_names,
      paste0("x", (length(predictor_names) + 1):q)
    )
  }

  ## Map from internal column index (the number inside "_j_") to
  #  1-based predictor display index into predictor_names.
  internal_idx_to_pred <- integer(0)
  if (!is.null(raw_nms) && length(all_linear_cols) > 0) {
    for (jj in seq_along(all_linear_cols)) {
      col_idx <- all_linear_cols[jj]
      raw_nm <- raw_nms[col_idx]
      idx_match <- regmatches(raw_nm, regexec("^_([0-9]+)_$", raw_nm))[[1]]
      if (length(idx_match) == 2) {
        internal_idx_to_pred[as.integer(idx_match[2])] <- jj
      }
    }
  }

  ## Calculus mode
  calc_mode <- "equation"
  calc_var_idx <- NULL
  calc_var_name <- NULL

  .resolve_calc_var <- function(spec, predictor_names, q, arg_name) {
    if (length(spec) > 1) {
      warning(
        "\n\t ", arg_name, " accepts only one variable at a time. ",
        "Only the first element will be used. Call equation() separately ",
        "for each variable.\n"
      )
      spec <- spec[1]
    }
    if (is.character(spec)) {
      idx <- match(spec, predictor_names)
      if (is.na(idx)) {
        stop(
          "\n\t ", arg_name, " variable '", spec,
          "' not found in predictor names: ",
          paste(predictor_names, collapse = ", "), ".\n"
        )
      }
      list(idx = idx, name = predictor_names[idx])
    } else {
      idx <- as.integer(spec)
      if (idx < 1 || idx > q) {
        stop(
          "\n\t ", arg_name, " index ", idx,
          " is out of range [1, ", q, "].\n"
        )
      }
      list(idx = idx, name = predictor_names[idx])
    }
  }

  if (!is.null(first_derivative)) {
    calc_mode <- "first_derivative"
    res <- .resolve_calc_var(first_derivative, predictor_names, q,
                             "first_derivative")
    calc_var_idx  <- res$idx
    calc_var_name <- res$name
  } else if (!is.null(second_derivative)) {
    calc_mode <- "second_derivative"
    res <- .resolve_calc_var(second_derivative, predictor_names, q,
                             "second_derivative")
    calc_var_idx  <- res$idx
    calc_var_name <- res$name
  } else if (!is.null(antiderivative)) {
    calc_mode <- "antiderivative"
    res <- .resolve_calc_var(antiderivative, predictor_names, q,
                             "antiderivative")
    calc_var_idx  <- res$idx
    calc_var_name <- res$name
  }

  ## Response name from link function
  fam <- object$family
  if (is.null(response_name)) {
    if (fam$family == "gaussian" && fam$link == "identity") {
      response_name <- "y"
    } else {
      response_name <- paste0(fam$link, "(E[y])")
    }
  }

  ## Parse a raw_expansion_name into structured factors.
  #  Raw names use "_j_" for variables and "x" (no underscores) as
  #  the interaction separator between complete "_j_" tokens.
  #  Examples: "_1_", "_1_^2", "_1_x_2_", "_2_x_1_^2", "_1_x_2_x_3_"
  #
  #  Key insight: strsplit on "_x_" eats the trailing/leading underscores

  #  of the _j_ tokens. Instead, use a regex that finds all _j_ or _j_^n
  #  tokens directly.
  .parse_term <- function(nm) {
    if (nm == "intercept" ||
        grepl("^intercept$", nm, ignore.case = TRUE)) {
      return(list(is_intercept = TRUE, factors = list()))
    }

    ## Extract all "_j_" and "_j_^n" tokens from the raw name
    #  Pattern: underscore, digits, underscore, optionally ^digits
    matches <- gregexpr("_([0-9]+)_(\\^[0-9]+)?", nm)
    tokens <- regmatches(nm, matches)[[1]]

    if (length(tokens) == 0) {
      return(list(is_intercept = FALSE,
                  factors = list(list(pred_idx = NA_integer_, power = 1L))))
    }

    factors <- lapply(tokens, function(tok) {
      ## Extract power if present
      pow_match <- regmatches(tok, regexec("\\^([0-9]+)$", tok))[[1]]
      if (length(pow_match) == 2) {
        power <- as.integer(pow_match[2])
        base_tok <- sub("\\^[0-9]+$", "", tok)
      } else {
        power <- 1L
        base_tok <- tok
      }
      ## Extract the column number
      idx_match <- regmatches(
        base_tok, regexec("^_([0-9]+)_$", base_tok)
      )[[1]]
      if (length(idx_match) == 2) {
        raw_col <- as.integer(idx_match[2])
        pred_idx <- internal_idx_to_pred[raw_col]
        if (is.na(pred_idx)) pred_idx <- NA_integer_
      } else {
        pred_idx <- NA_integer_
      }
      list(pred_idx = pred_idx, power = power)
    })
    list(is_intercept = FALSE, factors = factors)
  }

  ## Differentiate a single term w.r.t. variable at index var_idx
  .differentiate_term <- function(coef_val, parsed, var_idx) {
    if (parsed$is_intercept) return(NULL)
    hit <- NULL
    hit_pos <- 0L
    for (i in seq_along(parsed$factors)) {
      if (!is.na(parsed$factors[[i]]$pred_idx) &&
          parsed$factors[[i]]$pred_idx == var_idx) {
        hit <- parsed$factors[[i]]
        hit_pos <- i
        break
      }
    }
    if (is.null(hit)) return(NULL)
    new_coef <- coef_val * hit$power
    new_power <- hit$power - 1L
    new_factors <- parsed$factors
    if (new_power == 0L) {
      new_factors[[hit_pos]] <- NULL
    } else {
      new_factors[[hit_pos]]$power <- new_power
    }
    list(coef = new_coef,
         parsed = list(is_intercept = (length(new_factors) == 0),
                       factors = new_factors))
  }

  ## Second derivative: differentiate twice
  .differentiate2_term <- function(coef_val, parsed, var_idx) {
    first <- .differentiate_term(coef_val, parsed, var_idx)
    if (is.null(first)) return(NULL)
    if (first$parsed$is_intercept) return(NULL)
    .differentiate_term(first$coef, first$parsed, var_idx)
  }

  ## Antidifferentiate a single term w.r.t. variable at index var_idx
  .antidifferentiate_term <- function(coef_val, parsed, var_idx) {
    if (parsed$is_intercept) {
      new_factors <- list(list(pred_idx = var_idx, power = 1L))
      return(list(coef = coef_val,
                  parsed = list(is_intercept = FALSE,
                                factors = new_factors)))
    }
    hit_pos <- 0L
    for (i in seq_along(parsed$factors)) {
      if (!is.na(parsed$factors[[i]]$pred_idx) &&
          parsed$factors[[i]]$pred_idx == var_idx) {
        hit_pos <- i
        break
      }
    }
    new_factors <- parsed$factors
    if (hit_pos > 0) {
      old_power <- new_factors[[hit_pos]]$power
      new_power <- old_power + 1L
      new_coef <- coef_val / new_power
      new_factors[[hit_pos]]$power <- new_power
    } else {
      new_coef <- coef_val
      new_factors <- c(new_factors,
                       list(list(pred_idx = var_idx, power = 1L)))
    }
    list(coef = new_coef,
         parsed = list(is_intercept = FALSE, factors = new_factors))
  }

  ## Rebuild human-readable term from parsed structure
  .rebuild_term_name <- function(parsed, predictor_names) {
    if (parsed$is_intercept) return("intercept")
    parts <- vapply(parsed$factors, function(f) {
      if (is.na(f$pred_idx)) {
        nm <- "UNKNOWN"
      } else if (f$pred_idx <= length(predictor_names)) {
        nm <- predictor_names[f$pred_idx]
      } else {
        nm <- paste0("x", f$pred_idx)
      }
      if (f$power == 1) return(nm)
      paste0(nm, "^", f$power)
    }, character(1))
    paste(parts, collapse = "*")
  }

  ## Format one coefficient * term pair for display
  .fmt_coef <- function(val, term_nm, first_term = FALSE) {
    if (collapse_zero && abs(val) < .Machine$double.eps * 100) {
      return(NULL)
    }
    if (scientific && (abs(val) < 1e-3 || abs(val) > 1e4) && val != 0) {
      val_str <- format(val, digits = digits, scientific = TRUE)
    } else {
      val_str <- format(round(val, digits), nsmall = digits, trim = TRUE)
    }
    ## Intercept / constant
    if (term_nm == "intercept" || term_nm == "") {
      if (first_term) return(val_str)
      if (val >= 0) return(paste0(" + ", val_str))
      return(paste0(" - ", gsub("^-", "", val_str)))
    }
    ## Coefficient * term
    if (first_term) {
      if (abs(val - 1) < .Machine$double.eps * 100) return(term_nm)
      if (abs(val + 1) < .Machine$double.eps * 100) return(paste0("-", term_nm))
      return(paste0(val_str, "*", term_nm))
    }
    if (val >= 0) {
      if (abs(val - 1) < .Machine$double.eps * 100) return(paste0(" + ", term_nm))
      return(paste0(" + ", val_str, "*", term_nm))
    }
    if (abs(val + 1) < .Machine$double.eps * 100) return(paste0(" - ", term_nm))
    paste0(" - ", gsub("^-", "", val_str), "*", term_nm)
  }

  ## LHS label
  if (calc_mode == "first_derivative") {
    lhs_label <- paste0("d(", response_name, ")/d(", calc_var_name, ")")
  } else if (calc_mode == "second_derivative") {
    lhs_label <- paste0("d2(", response_name, ")/d(", calc_var_name, ")2")
  } else if (calc_mode == "antiderivative") {
    lhs_label <- paste0("integral ", response_name, " d(", calc_var_name, ")")
  } else {
    lhs_label <- response_name
  }

  ## Build formula string for each partition.
  #  Always parse from raw_expansion_names (reliable _j_ structure),
  #  display with predictor_names (human-readable).
  formulas <- character(K + 1)
  for (k in 1:(K + 1)) {
    coefs <- object$B[[k]]
    terms_out <- character(0)
    first <- TRUE

    for (i in seq_along(coefs)) {
      orig_coef <- coefs[i]
      parsed <- .parse_term(raw_nms[i])

      if (calc_mode == "equation") {
        display_nm <- .rebuild_term_name(parsed, predictor_names)
        term_str <- .fmt_coef(orig_coef, display_nm, first_term = first)
      } else if (calc_mode == "first_derivative") {
        result <- .differentiate_term(orig_coef, parsed, calc_var_idx)
        if (is.null(result)) next
        display_nm <- .rebuild_term_name(result$parsed, predictor_names)
        term_str <- .fmt_coef(result$coef, display_nm, first_term = first)
      } else if (calc_mode == "second_derivative") {
        result <- .differentiate2_term(orig_coef, parsed, calc_var_idx)
        if (is.null(result)) next
        display_nm <- .rebuild_term_name(result$parsed, predictor_names)
        term_str <- .fmt_coef(result$coef, display_nm, first_term = first)
      } else if (calc_mode == "antiderivative") {
        result <- .antidifferentiate_term(orig_coef, parsed, calc_var_idx)
        if (is.null(result)) next
        display_nm <- .rebuild_term_name(result$parsed, predictor_names)
        term_str <- .fmt_coef(result$coef, display_nm, first_term = first)
      }

      if (!is.null(term_str)) {
        terms_out <- c(terms_out, term_str)
        first <- FALSE
      }
    }

    if (length(terms_out) == 0) {
      formulas[k] <- paste0(lhs_label, " = 0")
    } else {
      rhs <- paste0(terms_out, collapse = "")
      if (calc_mode == "antiderivative") rhs <- paste0(rhs, " + C")
      formulas[k] <- paste0(lhs_label, " = ", rhs)
    }
  }

  ## Partition boundary info
  bounds_info <- NULL
  if (show_bounds && K > 0) {
    if (q == 1 && !is.null(object$knots)) {
      ## 1D interval bounds
      knot_vals <- sort(c(object$knots))
      bounds_info <- matrix(NA, nrow = K + 1, ncol = 2)
      colnames(bounds_info) <- c("lower", "upper")
      rownames(bounds_info) <- paste0("Partition ", 1:(K + 1))

      all_x <- unlist(lapply(object$X, function(Xk) {
        if (nrow(Xk) > 0 && length(object$power1_cols) > 0) {
          Xk[, object$power1_cols[1]]
        } else if (nrow(Xk) > 0 && length(object$nonspline_cols) > 0) {
          Xk[, object$nonspline_cols[1]]
        } else {
          NULL
        }
      }))
      if (length(all_x) > 0) {
        x_min <- min(all_x, na.rm = TRUE)
        x_max <- max(all_x, na.rm = TRUE)
      } else {
        x_min <- -Inf
        x_max <- Inf
      }

      bounds_info[1, 1] <- x_min
      bounds_info[1, 2] <- knot_vals[1]
      if (K >= 2) {
        for (kk in 2:K) {
          bounds_info[kk, 1] <- knot_vals[kk - 1]
          bounds_info[kk, 2] <- knot_vals[kk]
        }
      }
      bounds_info[K + 1, 1] <- knot_vals[K]
      bounds_info[K + 1, 2] <- x_max

    } else if (!is.null(object$make_partition_list)) {
      ## Multi-D bounding boxes from cluster centers
      centers <- object$make_partition_list$centers

      if (!is.null(centers)) {
        n_centers <- nrow(centers)
        n_dims <- ncol(centers)

        ## Only show spline predictor dimensions, not indicator columns
        n_spline <- length(object$power1_cols)
        if (n_spline > 0 && n_spline < n_dims) {
          show_dims <- seq_len(n_spline)
        } else {
          show_dims <- seq_len(n_dims)
        }

        center_colnames <- colnames(centers)
        if (is.null(center_colnames)) {
          if (!is.null(predictor_names) && n_dims <= length(predictor_names)) {
            center_colnames <- predictor_names[1:n_dims]
          } else {
            center_colnames <- paste0("x", 1:n_dims)
          }
        }

        ## Training predictor ranges per dimension
        training_ranges <- matrix(NA, 2, n_dims)
        rownames(training_ranges) <- c("min", "max")
        for (jj in seq_len(n_dims)) {
          col_vals <- unlist(lapply(object$X, function(Xk) {
            if (nrow(Xk) > 0 && jj <= length(all_linear_cols)) {
              Xk[, all_linear_cols[jj]]
            } else {
              NULL
            }
          }))
          if (length(col_vals) > 0) {
            training_ranges[1, jj] <- min(col_vals, na.rm = TRUE)
            training_ranges[2, jj] <- max(col_vals, na.rm = TRUE)
          }
        }

        ## Per-partition bounding boxes via midpoints between sorted centers
        bounds_lower <- matrix(NA, n_centers, n_dims)
        bounds_upper <- matrix(NA, n_centers, n_dims)
        colnames(bounds_lower) <- center_colnames
        colnames(bounds_upper) <- center_colnames

        for (jj in seq_len(n_dims)) {
          center_vals_j <- centers[, jj]
          ord_j <- order(center_vals_j)
          sorted_j <- center_vals_j[ord_j]
          if (length(sorted_j) > 1) {
            midpts_j <- (sorted_j[-length(sorted_j)] + sorted_j[-1]) / 2
          } else {
            midpts_j <- numeric(0)
          }
          for (rank in seq_along(ord_j)) {
            part_idx <- ord_j[rank]
            lo <- if (rank == 1 && !is.na(training_ranges[1, jj])) {
              training_ranges[1, jj]
            } else if (rank == 1) {
              -Inf
            } else {
              midpts_j[rank - 1]
            }
            hi <- if (rank == length(ord_j) && !is.na(training_ranges[2, jj])) {
              training_ranges[2, jj]
            } else if (rank == length(ord_j)) {
              Inf
            } else {
              midpts_j[rank]
            }
            bounds_lower[part_idx, jj] <- lo
            bounds_upper[part_idx, jj] <- hi
          }
        }

        bounds_info <- list(
          lower = bounds_lower[, show_dims, drop = FALSE],
          upper = bounds_upper[, show_dims, drop = FALSE],
          centers = centers,
          colnames = center_colnames[show_dims]
        )
      }
    }
  }

  ## Print
  cat("\n")
  div_major <- paste(rep("=", 72), collapse = "")
  div_minor <- paste(rep("-", 72), collapse = "")

  cat(div_major, "\n", sep = "")
  if (calc_mode == "equation") {
    cat("Fitted Equations: lgspline Model\n")
  } else if (calc_mode == "first_derivative") {
    cat("First Derivative w.r.t. ", calc_var_name,
        ": lgspline Model\n", sep = "")
  } else if (calc_mode == "second_derivative") {
    cat("Second Derivative w.r.t. ", calc_var_name,
        ": lgspline Model\n", sep = "")
  } else if (calc_mode == "antiderivative") {
    cat("Antiderivative w.r.t. ", calc_var_name,
        ": lgspline Model\n", sep = "")
  }
  cat(div_major, "\n", sep = "")
  cat("Family:", fam$family, "  Link:", fam$link, "\n")
  cat("Partitions:", K + 1, "  Basis functions per partition:", nc, "\n")
  cat(div_minor, "\n", sep = "")

  for (k in 1:(K + 1)) {
    cat("\n")
    cat("Partition", k, "\n")

    if (show_bounds && !is.null(bounds_info)) {
      if (q == 1 && is.matrix(bounds_info) && ncol(bounds_info) == 2) {
        lb <- bounds_info[k, 1]
        ub <- bounds_info[k, 2]
        lb_str <- if (is.finite(lb)) {
          format(round(lb, digits), nsmall = digits)
        } else { "-Inf" }
        ub_str <- if (is.finite(ub)) {
          format(round(ub, digits), nsmall = digits)
        } else { "+Inf" }
        lb_bracket <- if (k == 1) "[" else "("
        ub_bracket <- if (k == K + 1) "]" else ")"
        cat("  Bounds: ", predictor_names[1], " in ",
            lb_bracket, lb_str, ", ", ub_str, ub_bracket,
            "\n", sep = "")

      } else if (is.list(bounds_info) && !is.null(bounds_info$lower)) {
        n_dims <- ncol(bounds_info$lower)
        cnames <- bounds_info$colnames
        cat("  Bounds:\n")
        for (jj in seq_len(n_dims)) {
          lo <- bounds_info$lower[k, jj]
          hi <- bounds_info$upper[k, jj]
          lo_str <- if (is.finite(lo)) {
            format(round(lo, min(digits, 2)), nsmall = min(digits, 2))
          } else { "-Inf" }
          hi_str <- if (is.finite(hi)) {
            format(round(hi, min(digits, 2)), nsmall = min(digits, 2))
          } else { "+Inf" }
          lb_bracket <- if (lo == bounds_info$lower[
            which.min(bounds_info$lower[, jj]), jj]) "[" else "("
          ub_bracket <- if (hi == bounds_info$upper[
            which.max(bounds_info$upper[, jj]), jj]) "]" else ")"
          cat("    ", cnames[jj], " in ",
              lb_bracket, lo_str, ", ", hi_str, ub_bracket,
              "\n", sep = "")
        }
      }
    }

    cat("  ", formulas[k], "\n", sep = "")
  }

  cat("\n")
  cat(div_minor, "\n", sep = "")
  if (!is.null(object$sigmasq_tilde) && is.finite(object$sigmasq_tilde)) {
    cat("Dispersion (sigma^2):",
        format(object$sigmasq_tilde, digits = digits), "\n")
  }
  if (!is.null(object$trace_XUGX) && is.finite(object$trace_XUGX)) {
    cat("Effective df:",
        format(object$N - object$trace_XUGX, digits = 2), "\n")
  }
  cat(div_major, "\n", sep = "")
  cat("\n")

  invisible(list(
    formulas = formulas,
    bounds   = bounds_info,
    link     = fam$link,
    mode     = calc_mode,
    variable = calc_var_name
  ))
}


#' @rdname equation
#' @method print equation
#' @export
print.equation <- function(x, ...) {
  cat("\nFitted Equations")
  if (!is.null(x$mode) && x$mode != "equation") {
    mode_label <- switch(
      x$mode,
      "first_derivative"  = "First Derivative",
      "second_derivative" = "Second Derivative",
      "antiderivative"    = "Antiderivative"
    )
    cat(" (", mode_label, " w.r.t. ", x$variable, ")", sep = "")
  }
  cat(":\n")
  for (k in seq_along(x$formulas)) {
    cat("  Partition", k, ":", x$formulas[k], "\n")
  }
  if (!is.null(x$bounds)) {
    if (is.matrix(x$bounds)) {
      cat("\nPartition Bounds:\n")
      print(x$bounds)
    } else if (is.list(x$bounds) && !is.null(x$bounds$lower)) {
      cat("\nPartition Bounds (lower):\n")
      print(x$bounds$lower)
      cat("\nPartition Bounds (upper):\n")
      print(x$bounds$upper)
    }
  }
  invisible(x)
}

#' Generic for Numerical Integration
#'
#' @description
#' S3 generic that dispatches to \code{integrate.lgspline} for fitted
#' \code{lgspline} objects and falls back to \code{\link[stats]{integrate}}
#' for ordinary functions.
#'
#' @name integrate
#' @param f A fitted model object or a function.
#' @param ... Arguments passed to methods.
#'
#' @export
integrate <- function(f, ...) UseMethod("integrate")


#' Default method for the integrate generic
#'
#' @rdname integrate
#' @method integrate default
#' @export
integrate.default <- function(f, ...){
  stats::integrate(f, ...)
}

test_that("basic S3 methods work correctly", {
  x <- seq(-9, 9, length.out = 100)
  y <- sin(x) + rnorm(100, 0, 0.1) + 0.1*x^2
  fit <- lgspline(cbind(x), y, K = 5)

  ## Test print method
  expect_output(print(fit), "Lagrangian Multiplier Smoothing Spline Model")

  ## Test predict method
  newx <- matrix(seq(-9, 9, length.out = 10))
  pred <- predict(fit, newx)
  expect_length(pred, 10)

  ## Test coef method
  coefs <- coef(fit)
  expect_type(coefs, "list")
  expect_length(coefs, fit$K + 1)

  ## Test find extremum
  extr <- find_extremum(fit)
  expect_type(extr, "list")
  expect_length(extr, 2)

  ## Test generate posterior
  extr <- generate_posterior(fit)
  expect_type(extr, "list")
  expect_length(extr, 2)

})

test_that("generate_posterior forwards enforce_qp_constraints to the stored sampler", {
  set.seed(2026)

  x <- seq(0, 1, length.out = 40)
  y <- exp(x) + rnorm(length(x), 0, 0.02)

  fit <- lgspline(cbind(x), y,
                  K = 1,
                  opt = FALSE,
                  qp_monotonic_increase = TRUE,
                  include_warnings = FALSE)

  fit_core <- if(!is.null(fit$model_fit)) fit$model_fit else fit
  sampler <- fit_core$generate_posterior

  expect_true("enforce_qp_constraints" %in% names(formals(sampler)))

  set.seed(99)
  draw_direct <- sampler(draw_dispersion = FALSE,
                         num_draws = 1,
                         enforce_constraints = TRUE,
                         max_slice_iterations = 200)

  set.seed(99)
  draw_public <- generate_posterior(fit,
                                    draw_dispersion = FALSE,
                                    num_draws = 1,
                                    enforce_qp_constraints = TRUE,
                                    max_slice_iterations = 200)

  expect_equal(draw_public$post_draw_coefficients,
               draw_direct$post_draw_coefficients)
})

test_that("lgspline handles basic GLM and quadratic programming constraints", {
  set.seed(1234)

  ## Generate test data
  x <- seq(-9, 9, length.out = 250) # Reduced size for testing

  ## Helper functions
  slinky <- function(x) {
    (50 * cos(x * 2) + -2 * x^2 + (0.25 * x)^4 + 80)
  }

  coil <- function(x) {
    (100 * cos(x * 2) + -1.5 * x^2 + (0.1 * x)^4 + (0.05 * x^3) +
       (-0.01 * x^5) + (0.00002 * x^6) - (0.000001 * x^7) + 100)
  }

  exponential_log <- function(x) {
    unlist(sapply(x, function(xx) {
      if (xx <= 1) {
        100 * (exp(xx) - exp(1))
      } else {
        100 * (log(xx))
      }
    }))
  }

  ## Combined function
  fxn <- function(x) {
    slinky(x) + coil(x) + 0.5 * exponential_log(x) + 25 * x
  }

  ## Create dataset
  dat <- cbind(x, NA)
  colnames(dat) <- c('x', 'y')

  ## Transform mean for quasi-poisson
  m <- mean(fxn(x))
  s <- sd(fxn(x))
  mu <- (fxn(x) - m) / s
  mu <- mu + abs(min(mu)) + 1

  ## Generate poisson responses
  set.seed(1234)  # For reproducibility
  dat[,'y'] <- sapply(mu, function(m) rpois(1, m))

  ## Test fitting with monotonicity constraint and some non-default settings
  fit <- lgspline(dat[,'x', drop=FALSE],
                  dat[,'y'],
                  K = 1,
                  opt = FALSE,
                  wiggle_penalty = 1e-2,
                  flat_ridge_penalty = 1e-2,
                  unique_penalty_per_partition = FALSE,
                  tol = 1e-2,
                  qp_range_lower = 1,
                  qp_monotonic_increase = TRUE,
                  family = quasipoisson())

  ## Basic checks
  expect_s3_class(fit, "lgspline")
  expect_length(fit$B, fit$K + 1)
  expect_true(all(!is.na(fit$ytilde)))

  ## Check monotonicity constraint
  newx <- matrix(sort(x))
  preds <- fit$predict(newx)
  diffs <- diff(preds)
  expect_true(all(diffs >= -1e-10))  # Allow for numerical imprecision

  ## Check range constraint
  expect_true(all(preds > 1-1e-10)) # Allow for numeric imprecision

  ## Test GLM-specific components
  expect_equal(fit$family$family, quasipoisson()$family)
  expect_equal(fit$family$link, quasipoisson()$link)

  ## Test plotting
  expect_error(plot(fit, show_formulas = TRUE,
                    text_size_formula = 2), NA)

  ## Test predictions
  expect_length(predict(fit, matrix(x[1:10])), 10)
})

test_that("Basic lgspline handles logistic regression without constraints", {
  x <- seq(-3, 3, length.out = 250)

  ## Binary response (logistic regression)
  y_bin <- rbinom(250, 1, plogis(sin(x)))
  fit_bin <- lgspline(cbind(x),
                      unique_penalty_per_partition = FALSE,
                      log_initial_flat = 1,
                      log_initial_wiggle = 1e-1,
                      y_bin,
                      K = 10,
                      opt = FALSE,
                      iterate_tune = FALSE,
                      iterate_final_fit = FALSE,
                      include_constrain_first_deriv = FALSE,
                      include_constrain_second_deriv = FALSE,
                      include_constrain_fitted = FALSE,
                      family = quasibinomial())

  ## Class is right
  expect_s3_class(fit_bin, "lgspline")

  ## Family is right
  expect_equal(fit_bin$family$family, quasibinomial()$family)


  ## Functionality works
  generate_posterior(fit_bin, draw_dispersion = FALSE)
  print(summary(fit_bin))
  fit_bin$find_extremum(minimize = TRUE)

  ## Range is right
  preds <- predict(fit_bin, new_predictors = cbind(sample(x)+rnorm(length(x),
                                                             0,
                                                             0.00001)))
  expect_true(all(abs(preds) <= 1))
})

test_that("lgspline handles various quadratic programming constraints", {
  x <- seq(-3, 3, length.out = 100)
  y <- exp(x) + rnorm(100, 0, 0.1)

  ## Test monotone increasing
  fit_inc <- lgspline(cbind(x), y,
                      K = 2,
                      opt = FALSE,
                      qp_monotonic_increase = TRUE)
  preds_inc <- predict(fit_inc, matrix(sort(x)))
  expect_true(all(diff(preds_inc) >= -1e-10))

  ## Test monotone decreasing
  y_dec <- -y
  fit_dec <- lgspline(cbind(x), y_dec,
                      K = 0,
                      opt = FALSE,
                      qp_monotonic_decrease = TRUE)
  preds_dec <- predict(fit_dec, matrix(sort(x)))
  expect_true(all(diff(preds_dec) <= 1e-10))

  ## Test bounded range
  y_bound <- y - mean(y)
  fit_bound <- lgspline(cbind(x), y_bound,
                       K = 1,
                       qp_range_lower = -1,
                       qp_range_upper = 1,
                       opt = FALSE)
  preds_bound <- predict(fit_bound, matrix(x))
  expect_true(all(preds_bound >= -1.01)) # Allow small numerical error
  expect_true(all(preds_bound <= 1.01))
})

#' Definite Integral of a Fitted lgspline
#'
#' @description
#' Given a fitted \code{lgspline} object, computes the definite integral of
#' the fitted surface over a rectangular domain using Gauss--Legendre
#' quadrature on \code{predict()}.
#'
#' @section Method:
#' The integration domain is discretised into a tensor-product grid of
#' Gauss--Legendre quadrature nodes.  Predicted values at each node come
#' from the model's \code{predict()} method, which correctly handles
#' partition assignment and piecewise polynomial evaluation.  The integral
#' is the weighted sum
#' \deqn{\int_{a_1}^{b_1} \cdots \int_{a_d}^{b_d}
#'       \hat{f}(\mathbf{t})\,\mathrm{d}t_1 \cdots \mathrm{d}t_d
#'       \;\approx\;
#'       \sum_{i=1}^{M} w_i\,\hat{f}(\mathbf{t}_i)}
#' where \eqn{M = n_{\mathrm{quad}}^d} and each weight incorporates the
#' Jacobian \eqn{(b_j - a_j)/2} for the affine map from \eqn{[-1, 1]} to
#' \eqn{[a_j, b_j]}.  Nodes and weights on \eqn{[-1, 1]} are computed via
#' the Golub--Welsch algorithm (eigenvalues of the symmetric tridiagonal
#' Jacobi matrix).
#'
#' For smooth polynomials, 30--50 nodes per dimension is typically
#' sufficient; highly partitioned models (large \eqn{K}) may benefit from
#' more.  Total evaluation points scale as \eqn{n_{\mathrm{quad}}^d}, so
#' problems with \eqn{d \ge 4} may require reducing \code{n_quad}.
#'
#' @section Integration scale:
#' By default (\code{link_scale = FALSE}), integration is on the response
#' scale \eqn{\mu = g^{-1}(\eta)}.  Setting \code{link_scale = TRUE}
#' integrates the linear predictor \eqn{\eta = \mathbf{x}^{\top}\boldsymbol{\beta}}
#' directly, which is useful when the quantity of interest is the area
#' under the link-transformed surface rather than the response.  For the
#' identity link the two coincide.
#'
#' @param f A fitted \code{lgspline} object.
#' @param lower Numeric vector of lower bounds, one per integration variable.
#'   Scalar values are recycled.
#' @param upper Numeric vector of upper bounds, one per integration variable.
#'   Scalar values are recycled.
#' @param vars Default: NULL.  Character or integer vector identifying which
#'   predictor(s) to integrate over.  When NULL all numeric predictors are
#'   integrated simultaneously.
#' @param initial_values Default: NULL.  Numeric vector of length \eqn{q}
#'   supplying fixed values for predictors not among the integration
#'   variables.  When NULL, non-integration predictors are held at the
#'   midpoint of their training range.
#' @param B_predict Default: NULL.  Optional list of coefficient vectors,
#'   one per partition.  When NULL the fitted coefficients are used.
#' @param link_scale Default: FALSE.  Logical; when TRUE the integral is
#'   computed on the link (linear predictor) scale \eqn{\eta} rather than
#'   the response scale \eqn{\mu}.
#' @param n_quad Default: 50.  Number of Gauss--Legendre nodes per
#'   integration dimension.
#' @param \ldots Additional arguments (currently unused; present for S3
#'   method compatibility).
#'
#' @return A numeric scalar: the estimated definite integral.
#'
#' @examples
#'
#' ## 1-D: integral of fitted sin(t) over [-pi, pi] should be near 0
#' set.seed(1234)
#' t <- seq(-pi, pi, length.out = 1000)
#' y <- sin(t) + rnorm(length(t), 0, 0.01)
#' fit <- lgspline(t, y, K = 4, opt = FALSE)
#' integrate(fit, lower = -pi, upper = pi)
#'
#' ## Base R integrate still works as expected
#' integrate(sin, lower = -pi, upper = pi)
#'
#' ## 2-D: volume under fitted volcano surface
#' data(volcano)
#' vlong <- cbind(
#'   rep(seq_len(nrow(volcano)), ncol(volcano)),
#'   rep(seq_len(ncol(volcano)), each = nrow(volcano)),
#'   as.vector(volcano)
#' )
#' colnames(vlong) <- c("Length", "Width", "Height")
#' fit_v <- lgspline(vlong[, 1:2], vlong[, 3], K = 18,
#'                   include_quadratic_interactions = TRUE, opt = FALSE)
#' integrate(fit_v, lower = c(1, 1), upper = c(87, 61))
#'
#' @export
integrate.lgspline <- function(
    f,
    lower,
    upper,
    vars           = NULL,
    initial_values = NULL,
    B_predict      = NULL,
    link_scale     = FALSE,
    n_quad         = 50L,
    ...
){

  model_fit <- f

  ## Pull structural info and link function from the fitted model
  if(is.null(B_predict)) B_predict <- model_fit$B

  q_total <- model_fit$q
  link_name <- tryCatch(model_fit$family$link, error = function(e) "identity")
  if(is.null(link_name)) link_name <- "identity"

  ## Original predictor names (if the formula interface was used)
  og_cols <- tryCatch(
    get("og_cols", envir = environment(model_fit$predict)),
    error = function(e) NULL
  )

  ## Resolve integration variables from character names or integer indices
  #  When vars is NULL, all numeric predictors are integrated simultaneously.
  if(is.null(vars)){
    vars_idx <- model_fit$numerics
    if(length(vars_idx) == 0L) vars_idx <- seq_len(q_total)
  } else if(is.character(vars)){
    if(!is.null(og_cols) && all(vars %in% og_cols)){
      vars_idx <- match(vars, og_cols)
    } else {
      stop(
        "\n\tCharacter vars could not be resolved. Use integer predictor ",
        "indices unless original predictor names are available.\n"
      )
    }
  } else {
    vars_idx <- as.integer(vars)
  }

  n_integ <- length(vars_idx)
  if(n_integ == 0L) stop("\n\tNo integration variables identified.\n")
  if(any(is.na(vars_idx)) || any(vars_idx < 1L) || any(vars_idx > q_total)){
    stop("\n\tvars must be valid predictor indices between 1 and ", q_total, ".\n")
  }

  ## Validate and recycle bounds to match the number of integration variables
  lower <- as.numeric(lower)
  upper <- as.numeric(upper)
  if(length(lower) == 1L) lower <- rep(lower, n_integ)
  if(length(upper) == 1L) upper <- rep(upper, n_integ)
  if(length(lower) != n_integ || length(upper) != n_integ){
    stop(
      "\n\tlower and upper must each have length equal to the number of ",
      "integration variables (", n_integ, ").\n"
    )
  }

  ## Golub--Welsch algorithm for Gauss--Legendre nodes and weights on [-1, 1]
  #  Eigenvalues of the symmetric tridiagonal Jacobi matrix give the nodes;
  #  squared first-row eigenvector entries (times 2) give the weights.
  .gauss_legendre <- function(n){
    i <- seq_len(n - 1L)
    b <- i / sqrt(4 * i^2 - 1)
    J <- matrix(0, n, n)
    for(k in seq_along(b)){
      J[k, k + 1L] <- b[k]
      J[k + 1L, k] <- b[k]
    }
    eig <- eigen(J, symmetric = TRUE)
    ord <- order(eig$values)
    list(nodes = eig$values[ord], weights = (2 * eig$vectors[1, ]^2)[ord])
  }

  ## Reconstruct raw training predictors from the stored design matrices
  #  Extracts the linear column for each predictor from each partition's
  #  X matrix and reassembles into an N x q matrix of raw predictor values.
  .reconstruct_training_predictors <- function(){
    X_parts <- model_fit$X
    ords <- model_fit$order_list
    nm <- model_fit$raw_expansion_names
    pred_mat <- matrix(0, nrow = model_fit$N, ncol = q_total)
    for(j in seq_len(q_total)){
      ci <- which(nm == paste0("_", j, "_"))
      if(length(ci) == 0L) next
      for(k in seq_along(X_parts)){
        if(is.null(X_parts[[k]]) || nrow(X_parts[[k]]) == 0L) next
        pred_mat[as.integer(unlist(ords[[k]])), j] <- X_parts[[k]][, ci[1L]]
      }
    }
    pred_mat
  }

  gl <- .gauss_legendre(as.integer(n_quad))

  ## Affine map from [-1, 1] to [lower[vi], upper[vi]] for each dimension
  mapped_nodes <- vector("list", n_integ)
  mapped_weights <- vector("list", n_integ)
  for(vi in seq_len(n_integ)){
    hw <- (upper[vi] - lower[vi]) / 2
    mp <- (upper[vi] + lower[vi]) / 2
    mapped_nodes[[vi]] <- hw * gl$nodes + mp
    mapped_weights[[vi]] <- gl$weights * hw
  }

  ## Tensor-product grid and corresponding product weights
  grid <- expand.grid(mapped_nodes)
  product_weights <- apply(expand.grid(mapped_weights), 1L, prod)
  n_grid <- nrow(grid)

  ## Fixed values for non-integration predictors
  #  When initial_values is NULL, each predictor is held at the midpoint
  #  of its training range. Otherwise the user-supplied values are used.
  if(is.null(initial_values)){
    training_preds <- .reconstruct_training_predictors()
    base_vec <- rep(0, q_total)
    for(j in seq_len(q_total)){
      xj <- training_preds[, j]
      if(length(unique(xj)) > 1L){
        base_vec[j] <- (min(xj, na.rm = TRUE) + max(xj, na.rm = TRUE)) / 2
      } else {
        base_vec[j] <- xj[1L]
      }
    }
  } else {
    base_vec <- as.numeric(initial_values)
    if(length(base_vec) != q_total){
      stop("\n\tinitial_values must have length ", q_total, ".\n")
    }
  }

  ## Build the evaluation matrix: base values everywhere, overwritten at
  #  integration-variable columns with the quadrature node coordinates
  pred_mat <- matrix(rep(base_vec, each = n_grid), nrow = n_grid)
  for(vi in seq_len(n_integ)) pred_mat[, vars_idx[vi]] <- grid[, vi]

  ## Evaluate the fitted surface and optionally transform to link scale
  predicted <- model_fit$predict(new_predictors = pred_mat,
                                 B_predict = B_predict)
  if(link_scale && link_name != "identity"){
    predicted <- model_fit$family$linkfun(predicted)
  }

  ## Weighted sum gives the quadrature estimate of the definite integral
  sum(as.numeric(predicted) * product_weights)
}

#' Cox Proportional Hazards Helpers for lgspline
#'
#' Functions for fitting Cox proportional hazards regression models within the
#' lgspline framework. Analogous to the Weibull AFT helpers, these provide
#' the partial log-likelihood, score, information, and all interface functions
#' needed by lgspline's unconstrained fitting, penalty tuning, and inference
#' machinery.
#'
#'
#' @name cox_helpers
NULL


#' Compute Cox Partial Log-Likelihood
#'
#' @description
#' Evaluates the Cox partial log-likelihood for a given coefficient vector,
#' using the Breslow approximation for tied event times.
#'
#' Observations must be sorted in ascending order of survival time before
#' calling this function. The internal helpers handle sorting automatically;
#' this function is exposed for diagnostics and testing.
#'
#' @param eta Numeric vector of linear predictors \eqn{\mathbf{X}\boldsymbol{\beta}},
#'   length N, sorted by ascending event time.
#' @param status Integer/logical vector of event indicators (1 = event,
#'   0 = censored), same length and order as \code{eta}.
#' @param y Optional numeric vector of observed event/censor times, same length
#'   and order as \code{eta}. When supplied, tied event times are handled by the
#'   Breslow approximation using a common risk-set denominator within each tied
#'   event-time block. When omitted, the function assumes there are no ties (or
#'   that ties have already been expanded appropriately).
#' @param weights Optional numeric vector of observation weights (default 1).
#'
#' @return Scalar partial log-likelihood value.
#'
#' @details
#' The partial log-likelihood (Breslow) is
#' \deqn{\ell(\boldsymbol{\beta}) =
#'   \sum_{g} \Bigl[\sum_{i \in D_g} w_i \eta_i -
#'   d_g^{(w)} \log\Bigl(\sum_{j \in R_g} w_j \exp(\eta_j)\Bigr)\Bigr]}
#' where \eqn{D_g} is the event set at tied event time \eqn{t_g},
#' \eqn{R_g = \{j : t_j \ge t_g\}} is the corresponding risk set, and
#' \eqn{d_g^{(w)} = \sum_{i \in D_g} w_i}.
#'
#' @examples
#' set.seed(1234)
#' eta <- rnorm(50)
#' status <- rbinom(50, 1, 0.6)
#' y <- rexp(50)
#' loglik_cox(eta, status, y)
#'
#' @export
loglik_cox <- function(eta, status, y = NULL, weights = 1) {

  N <- length(eta)
  if(length(weights) == 1) weights <- rep(weights, N)

  ## Weighted hazard contributions
  haz <- weights * exp(eta)

  ## No event-time vector supplied: fall back to no-ties convention
  if(is.null(y)) {
    rsk <- rev(cumsum(rev(haz)))
    return(sum(status * weights * (eta - log(rsk))))
  }

  ## Breslow ties by unique event times
  event_times <- sort(unique(y[status == 1]))
  if(length(event_times) == 0) {
    return(0)
  }

  ll <- 0
  for(tt in event_times) {
    died_g <- which(y == tt & status == 1)
    surv_g <- which(y >= tt)
    sum_w <- sum(weights[died_g])
    ll <- ll + sum(weights[died_g] * eta[died_g]) -
      sum_w * log(sum(haz[surv_g]))
  }

  ll
}


#' Compute Cox Partial Log-Likelihood Score Vector
#'
#' @description
#' Gradient of the Cox partial log-likelihood with respect to
#' \eqn{\boldsymbol{\beta}}. Data must be sorted by ascending event time.
#'
#' @param X Design matrix (N x p), sorted by ascending event time.
#' @param eta Linear predictor vector \eqn{\mathbf{X}\boldsymbol{\beta}}.
#' @param status Event indicator (1 = event, 0 = censored).
#' @param y Optional numeric vector of observed event/censor times, same length
#'   and order as \code{eta}. When supplied, tied event times are handled with
#'   the Breslow approximation. When omitted, the function assumes there are no
#'   ties.
#' @param weights Observation weights (default 1).
#'
#' @return Numeric column vector of length p (gradient).
#'
#' @details
#' Under the Breslow approximation, the score is
#' \deqn{\mathbf{u} =
#'   \sum_g \Bigl[\sum_{i \in D_g} w_i \mathbf{x}_i -
#'   d_g^{(w)} \frac{\sum_{j \in R_g} w_j e^{\eta_j}\mathbf{x}_j}
#'   {\sum_{j \in R_g} w_j e^{\eta_j}}\Bigr]}
#' where \eqn{D_g} is the tied event set at time \eqn{t_g},
#' \eqn{R_g = \{j : t_j \ge t_g\}} is the risk set, and
#' \eqn{d_g^{(w)} = \sum_{i \in D_g} w_i}.
#'
#' @keywords internal
#' @export
score_cox <- function(X, eta, status, y = NULL, weights = 1) {

  N <- length(eta)
  p <- ncol(X)
  if(length(weights) == 1) weights <- rep(weights, N)

  haz <- weights * exp(eta)

  ## No event-time vector supplied: fall back to no-ties convention
  if(is.null(y)) {
    rsk <- rev(cumsum(rev(haz)))
    d_over_rsk <- (weights * status) / rsk
    cum_d_over_rsk <- cumsum(d_over_rsk)
    Pd <- haz * cum_d_over_rsk
    return(crossprod(X, cbind(weights * status - Pd)))
  }

  event_times <- sort(unique(y[status == 1]))
  if(length(event_times) == 0) {
    return(matrix(0, p, 1))
  }

  sc <- matrix(0, p, 1)
  for(tt in event_times) {
    died_g <- which(y == tt & status == 1)
    surv_g <- which(y >= tt)
    sum_w <- sum(weights[died_g])
    score0 <- sum(haz[surv_g])
    score1 <- crossprod(X[surv_g, , drop = FALSE], cbind(haz[surv_g]))
    sc <- sc + crossprod(X[died_g, , drop = FALSE], cbind(weights[died_g])) -
      sum_w * (score1 / score0)
  }

  sc
}


#' Compute Cox Observed Information Matrix
#'
#' @description
#' Negative Hessian of the Cox partial log-likelihood under the Breslow
#' approximation for tied event times. Data must be sorted by ascending event
#' time.
#'
#' @param X Design matrix (N x p), sorted by ascending event time.
#' @param eta Linear predictor vector.
#' @param status Event indicator (1 = event, 0 = censored).
#' @param y Optional numeric vector of observed event/censor times, same length
#'   and order as \code{eta}. When supplied, tied event times are handled with
#'   the Breslow approximation. When omitted, the function assumes there are no
#'   ties.
#' @param weights Observation weights (default 1).
#'
#' @return Symmetric p x p observed information matrix.
#'
#' @details
#' Under the Breslow approximation, the observed information is
#' \deqn{\sum_g d_g^{(w)}
#'   \Bigl[\frac{S_{2g}}{S_{0g}} -
#'   \Bigl(\frac{S_{1g}}{S_{0g}}\Bigr)
#'   \Bigl(\frac{S_{1g}}{S_{0g}}\Bigr)^{\top}\Bigr]}
#' where
#' \eqn{S_{0g} = \sum_{j \in R_g} w_j e^{\eta_j}},
#' \eqn{S_{1g} = \sum_{j \in R_g} w_j e^{\eta_j}\mathbf{x}_j},
#' and \eqn{S_{2g} = \sum_{j \in R_g} w_j e^{\eta_j}\mathbf{x}_j\mathbf{x}_j^{\top}}.
#'
#' @keywords internal
#' @export
info_cox <- function(X, eta, status, y = NULL, weights = 1) {

  N <- length(eta)
  p <- ncol(X)
  if(length(weights) == 1) weights <- rep(weights, N)

  haz <- weights * exp(eta)

  ## No event-time vector supplied: fall back to no-ties convention
  if(is.null(y)) {
    info <- matrix(0, p, p)
    s0 <- sum(haz)
    s1 <- crossprod(X, cbind(haz))
    s2 <- crossprod(X, haz * X)

    for(i in 1:N) {
      if(status[i] == 1 && s0 > 0) {
        xbar <- s1 / s0
        info <- info + weights[i] * (s2 / s0 - tcrossprod(xbar))
      }
      xi <- X[i, ]
      s0 <- s0 - haz[i]
      s1 <- s1 - haz[i] * xi
      s2 <- s2 - haz[i] * tcrossprod(xi)
    }
    return(info)
  }

  event_times <- sort(unique(y[status == 1]))
  if(length(event_times) == 0) {
    return(matrix(0, p, p))
  }

  info <- matrix(0, p, p)
  for(tt in event_times) {
    died_g <- which(y == tt & status == 1)
    surv_g <- which(y >= tt)
    sum_w <- sum(weights[died_g])

    s0 <- sum(haz[surv_g])
    s1 <- crossprod(X[surv_g, , drop = FALSE],
                    cbind(haz[surv_g]))
    s2 <- crossprod(X[surv_g, , drop = FALSE],
                    haz[surv_g] * X[surv_g, , drop = FALSE])

    xbar <- s1 / s0
    info <- info + sum_w * (s2 / s0 - tcrossprod(xbar))
  }

  info
}


#' Cox Proportional Hazards Family for lgspline
#'
#' @description
#' Creates a family-like object for Cox PH models. The link function is
#' \code{log} (the linear predictor is log-relative-hazard), but unlike
#' standard GLM families there is no dispersion parameter and no closed-form
#' mean-variance relationship.
#'
#' The family provides \code{$loglik} and \code{$aic} methods compatible
#' with \code{\link{logLik.lgspline}}.
#'
#' @return A list with family components used by \code{lgspline}.
#'
#' @details
#' Cox PH is semiparametric: the baseline hazard is unspecified. The
#' partial log-likelihood depends only on the order of event times and the
#' linear predictor \eqn{\eta = \mathbf{X}\boldsymbol{\beta}}. Consequently:
#' \itemize{
#'   \item No dispersion parameter is estimated (\code{sigmasq_tilde} is
#'         fixed at 1).
#'   \item \code{dev.resids} returns martingale-style residuals for GCV
#'         tuning compatibility.
#'   \item The response variable is survival time (positive), and the link
#'         is \code{log}.
#' }
#'
#' @examples
#' fam <- cox_family()
#' fam$family
#' fam$link
#'
#' @export
cox_family <- function() list(
  family = "cox",
  link = "log",
  linkfun = log,
  linkinv = exp,

  ## Variance function: Cox PH has no natural mean-variance relationship
  #  in the GLM sense. Return 1 so that default glm_weight_function
  #  produces unit weights (overridden by cox_glm_weight_function anyway).
  variance = function(mu) rep(1, length(mu)),

  ## dev.resids for GCV tuning compatibility.
  #  Uses -2 * partial-log-likelihood contribution per observation as a
  #  proxy. This is NOT the standard martingale residual, but serves the
  #  same role as family$dev.resids in the penalty tuning loop.
  dev.resids = function(y, mu, wt) {
    ## Fallback: squared residual on log scale
    wt * (log(y) - log(mu))^2
  },

  ## Custom deviance used internally by tune_Lambda for GCV.
  #  Receives the full vectors in partition order and the status vector
  #  via .... Returns per-observation deviance contributions.
  custom_dev.resids = function(y, mu, order_indices, family,
                               observation_weights, status) {

    ## Sort by event time
    ord <- order(y)
    y_s <- y[ord]
    mu_s <- mu[ord]
    st_s <- status[order_indices][ord]
    wt_s <- observation_weights[ord]

    eta_s <- log(mu_s)
    haz_s <- wt_s * exp(eta_s)

    pll_i_s <- rep(0, length(y_s))
    event_times <- sort(unique(y_s[st_s == 1]))
    if(length(event_times) > 0) {
      for(tt in event_times) {
        died_g <- which(y_s == tt & st_s == 1)
        surv_g <- which(y_s >= tt)
        pll_i_s[died_g] <- wt_s[died_g] * (eta_s[died_g] -
                           log(sum(haz_s[surv_g])))
      }
    }

    pll_i <- rep(0, length(y))
    pll_i[ord] <- pll_i_s

    ## Return -2 * pll_i as deviance contribution
    -2 * pll_i
  },

  ## Log-likelihood extraction from fitted model
  loglik = function(model_fit) {
    y_fit   <- model_fit$y
    mu_fit  <- model_fit$ytilde
    wt_fit  <- if(!is.null(model_fit$weights)) model_fit$weights
    else rep(1, length(y_fit))

    ## Retrieve status
    status <- NULL
    if(!is.null(model_fit$status)) {
      status <- model_fit$status
    } else if(!is.null(model_fit$.fit_call_args$status)) {
      status <- model_fit$.fit_call_args$status
    } else {
      warning("Censoring status not found; assuming all events observed.")
      status <- rep(1, length(y_fit))
    }

    ## Sort by event time
    ord <- order(y_fit)
    y_s   <- y_fit[ord]
    eta_s <- log(mu_fit[ord])
    st_s  <- status[ord]
    wt_s  <- wt_fit[ord]

    loglik_cox(eta_s, st_s, y_s, wt_s)
  },

  ## AIC = -2*loglik + 2*edf (no +1 for scale, unlike Weibull)
  aic = function(model_fit) {
    ll <- cox_family()$loglik(model_fit)
    edf <- model_fit$trace_XUGX
    if(is.null(edf)) {
      warning("Cannot compute AIC: effective degrees of freedom not found.")
      return(NA)
    }
    -2 * ll + 2 * edf
  }
)


#' Cox PH GLM Weight Function
#'
#' @description
#' Computes working weights for the Cox PH information matrix, used by
#' lgspline when updating \eqn{\mathbf{G}} after obtaining constrained
#' estimates. The weights are a diagonal approximation built from the Breslow
#' tied-event information contributions.
#'
#' @param mu Predicted values (exp(eta), i.e., relative hazard).
#' @param y Observed survival times.
#' @param order_indices Observation indices in partition order.
#' @param family Cox family object (unused, for interface compatibility).
#' @param dispersion Dispersion parameter (fixed at 1 for Cox PH).
#' @param observation_weights Observation weights.
#' @param status Event indicator (1 = event, 0 = censored).
#'
#' @return Numeric vector of working weights, length N.
#'
#' @details
#' For a tied event-time block \eqn{g}, the diagonal approximation uses
#' \deqn{W_{jj}^{(g)} = d_g^{(w)} \frac{h_j}{S_g}
#'   \Bigl(1 - \frac{h_j}{S_g}\Bigr), \qquad j \in R_g}
#' where \eqn{h_j = w_j \exp(\eta_j)},
#' \eqn{S_g = \sum_{k \in R_g} h_k}, and
#' \eqn{d_g^{(w)} = \sum_{i \in D_g} w_i}.
#'
#' When the natural weights are degenerate (all zero or non-finite), falls
#' back to a vector of ones.
#'
#' @examples
#' ## Used internally by lgspline; see cox_helpers examples below.
#'
#' @export
cox_glm_weight_function <- function(mu,
                                    y,
                                    order_indices,
                                    family,
                                    dispersion,
                                    observation_weights,
                                    status) {

  N <- length(mu)
  if(length(observation_weights) == 1) {
    observation_weights <- rep(observation_weights, N)
  }

  ## Sort by event time
  ord <- order(y)
  y_s   <- y[ord]
  mu_s  <- mu[ord]
  st_s  <- status[order_indices][ord]
  wt_s  <- observation_weights[ord]

  haz_s <- wt_s * mu_s
  W_diag_s <- rep(0, N)

  event_times <- sort(unique(y_s[st_s == 1]))
  if(length(event_times) > 0) {
    for(tt in event_times) {
      died_g <- which(y_s == tt & st_s == 1)
      surv_g <- which(y_s >= tt)
      sum_w <- sum(wt_s[died_g])
      s0 <- sum(haz_s[surv_g])
      frac <- haz_s[surv_g] / s0
      W_diag_s[surv_g] <- W_diag_s[surv_g] + sum_w * frac * (1 - frac)
    }
  }

  ## Unsort back to original order
  W_diag <- rep(0, N)
  W_diag[ord] <- W_diag_s

  ## Fallback for degenerate cases
  if(any(!is.finite(W_diag)) || all(W_diag == 0)) {
    return(rep(1, N))
  }

  ## Floor at machine epsilon to avoid zero weights
  W_diag <- pmax(W_diag, .Machine$double.eps)
  return(W_diag)
}


#' Cox PH Dispersion Function
#'
#' @description
#' Returns 1 unconditionally. Cox PH has no dispersion parameter; this
#' function exists solely for interface compatibility with lgspline's
#' \code{dispersion_function} argument.
#'
#' @param mu Predicted values.
#' @param y Observed survival times.
#' @param order_indices Observation indices.
#' @param family Family object.
#' @param observation_weights Observation weights.
#' @param VhalfInv Inverse square root of correlation matrix.
#' @param ... Additional arguments (including \code{status}).
#'
#' @return Scalar 1.
#'
#' @export
cox_dispersion_function <- function(mu, y, order_indices, family,
                                    observation_weights, VhalfInv, ...) {
  1
}


#' Cox PH Score Function for Quadratic Programming and Blockfit
#'
#' @description
#' Computes the score (gradient of partial log-likelihood) in the format
#' expected by lgspline's \code{qp_score_function} interface. The block-
#' diagonal design matrix \eqn{\mathbf{X}} and response \eqn{\mathbf{y}} are
#' in partition order; this function internally sorts by event time, computes
#' the Cox score using the Breslow approximation for tied event times, and
#' returns the result in the original partition order.
#'
#' @param X Block-diagonal design matrix (N x P).
#' @param y Response vector (survival times, N x 1).
#' @param mu Predicted values (N x 1), same order as X and y.
#' @param order_list List of observation indices per partition.
#' @param dispersion Dispersion (fixed at 1).
#' @param VhalfInv Inverse square root correlation matrix (NULL for
#'   independent observations).
#' @param observation_weights Observation weights.
#' @param status Event indicator (1 = event, 0 = censored).
#'
#' @return Numeric column vector of length P (gradient w.r.t. coefficients).
#'
#' @export
cox_qp_score_function <- function(X, y, mu, order_list, dispersion,
                                  VhalfInv, observation_weights, status) {

  N <- length(y)
  if(length(observation_weights) == 1) {
    observation_weights <- rep(observation_weights, N)
  }
  order_indices <- unlist(order_list)

  ## Sort by event time
  ord <- order(y)
  X_s <- X[ord, , drop = FALSE]
  y_s <- y[ord]
  mu_s <- mu[ord]
  st_s <- status[order_indices][ord]
  wt_s <- observation_weights[ord]

  eta_s <- log(mu_s)
  score_cox(X_s, eta_s, st_s, y_s, wt_s)
}


#' Cox PH Schur Correction
#'
#' @description
#' Returns zero corrections for all partitions. Cox PH has no nuisance
#' dispersion parameter, so no Schur complement correction to the
#' information matrix is needed. This function exists for interface
#' compatibility with lgspline's \code{schur_correction_function}.
#'
#' @param X List of partition design matrices.
#' @param y List of partition response vectors.
#' @param B List of partition coefficient vectors.
#' @param dispersion Dispersion (fixed at 1).
#' @param order_list List of observation indices per partition.
#' @param K Number of knots.
#' @param family Family object.
#' @param observation_weights Observation weights.
#' @param ... Additional arguments.
#'
#' @return List of K+1 zeros.
#'
#' @export
cox_schur_correction <- function(X, y, B, dispersion, order_list, K,
                                 family, observation_weights, ...) {
  lapply(1:(K + 1), function(k) 0)
}


#' Unconstrained Cox PH Estimation for lgspline
#'
#' @description
#' Estimates penalized Cox PH coefficients for a single partition (or the
#' full model when K = 0) using damped Newton-Raphson on the penalized
#' partial log-likelihood with the Breslow approximation for tied event times.
#'
#' This function is passed to lgspline's \code{unconstrained_fit_fxn}
#' argument. It receives the partition-level design matrix and response,
#' sorts internally by event time, and returns the coefficient vector.
#'
#' @param X Design matrix (N_k x p) for partition k.
#' @param y Survival times for partition k.
#' @param LambdaHalf Square root of penalty matrix (unused here, retained
#'   for interface compatibility).
#' @param Lambda Penalty matrix.
#' @param keep_weighted_Lambda Logical; if TRUE, return hot-start estimates
#'   without refinement (not recommended for Cox).
#' @param family Cox family object.
#' @param tol Convergence tolerance.
#' @param K Number of knots.
#' @param parallel Logical for parallel processing.
#' @param cl Cluster object.
#' @param chunk_size, num_chunks, rem_chunks Parallelism parameters.
#' @param order_indices Observation indices mapping partition to full data.
#' @param weights Observation weights.
#' @param status Event indicator (1 = event, 0 = censored), full-data length.
#'
#' @return Numeric column vector of penalized partial-likelihood coefficient
#'   estimates.
#'
#' @details
#' The penalized partial log-likelihood is
#' \deqn{\ell_p(\boldsymbol{\beta}) = \ell(\boldsymbol{\beta})
#'   - \tfrac{1}{2}\boldsymbol{\beta}^{\top}
#'   \boldsymbol{\Lambda}\boldsymbol{\beta}}
#' Newton-Raphson updates use the Cox score and observed information
#' (computed via \code{score_cox} and \code{info_cox}), plus the penalty
#' term. The step is damped: the step size is halved until the penalized
#' log-likelihood improves.
#'
#' @examples
#' ## Used internally by lgspline; see the full-model example below.
#'
#' @keywords internal
#' @export
unconstrained_fit_cox <- function(X, y, LambdaHalf, Lambda,
                                  keep_weighted_Lambda, family,
                                  tol = 1e-8, K, parallel, cl,
                                  chunk_size, num_chunks, rem_chunks,
                                  order_indices, weights, status) {

  ## Handle empty partitions
  if(nrow(X) == 0) {
    return(cbind(rep(0, ncol(X))))
  }

  N <- nrow(X)
  p <- ncol(X)

  ## Observation weights
  if(any(!is.null(weights))) {
    weights <- c(weights)
  } else {
    weights <- rep(1, N)
  }

  ## Extract status for this partition
  st <- status[order_indices]

  ## Sort by event time within this partition
  ord <- order(y)
  X_s <- X[ord, , drop = FALSE]
  y_s <- y[ord]
  st_s <- st[ord]
  wt_s <- weights[ord]

  ## Initialize at zero (log-relative-hazard = 0 is the null model)
  init <- rep(0, p)

  ## Damped Newton-Raphson on penalized partial log-likelihood
  beta <- damped_newton_r(
    init,
    ## Penalized partial log-likelihood
    function(par) {
      eta <- c(X_s %**% cbind(par))
      loglik_cox(eta, st_s, y_s, wt_s) -
        0.5 * c(crossprod(par, Lambda) %**% cbind(par))
    },
    ## Score (gradient)
    function(par) {
      eta <- c(X_s %**% cbind(par))
      sc <- score_cox(X_s, eta, st_s, y_s, wt_s)
      c(sc - Lambda %**% cbind(par))
    },
    ## Information (negative Hessian)
    function(par) {
      eta <- c(X_s %**% cbind(par))
      info_cox(X_s, eta, st_s, y_s, wt_s) + Lambda
    },
    tol
  )

  cbind(beta)
}


#' Fit Cox Proportional Hazards Model via lgspline
#'
#' @description
#' Convenience wrapper that calls \code{lgspline} with the correct
#' family, weight, dispersion, score, and unconstrained-fit functions for
#' Cox proportional hazards regression. All standard lgspline arguments
#' (knots, penalties, constraints, parallel, etc.) are passed through.
#'
#' @param formula Formula specifying the model. The response should be
#'   survival time; \code{status} is passed separately.
#' @param data Data frame.
#' @param status Integer vector of event indicators (1 = event,
#'   0 = censored), same length as the number of rows in \code{data}.
#' @param ... Additional arguments passed to \code{lgspline}.
#'
#' @return An object of class \code{"lgspline"}.
#'
#' @details
#' Internally sets:
#' \itemize{
#'   \item \code{family = cox_family()}
#'   \item \code{unconstrained_fit_fxn = unconstrained_fit_cox}
#'   \item \code{glm_weight_function = cox_glm_weight_function}
#'   \item \code{qp_score_function = cox_qp_score_function}
#'   \item \code{dispersion_function = cox_dispersion_function}
#'   \item \code{schur_correction_function = cox_schur_correction}
#'   \item \code{need_dispersion_for_estimation = FALSE}
#'   \item \code{estimate_dispersion = FALSE}
#'   \item \code{standardize_response = FALSE}
#' }
#'
#' A formula interface is needed, e.g. \code{lgspline_cox(t, y, ...)} won't work,
#' unlike for ordinary \code{\link{lgspline}}.
#'
#' @examples
#'
#' ## Cox PH with a nonlinear age effect on lung cancer survival
#' if(requireNamespace("survival", quietly = TRUE)) {
#'   library(survival)
#'   set.seed(1234)
#'   lung <- na.omit(lung[, c("time", "status", "age")])
#'   lung$age_std <- std(lung$age)
#'
#'   ## survival codes status as 1 = censored, 2 = dead
#'   event <- as.integer(lung$status == 2)
#'
#'   ## Spline on age
#'   fit <- lgspline_cox(
#'     time ~ spl(age_std),
#'     data = lung,
#'     status = event,
#'     K = 1
#'   )
#'   print(summary(fit))
#'   plot(fit,
#'        show_formulas = TRUE,
#'        custom_response_lab = 'HR',
#'        custom_predictor_lab = 'Standardized Age',
#'        ylim = c(0, 5))
#'
#' }
#'
#' @export
lgspline_cox <- function(formula, data, status, ...) {
  lgspline(
    formula = formula,
    data = data,
    family = cox_family(),
    unconstrained_fit_fxn = unconstrained_fit_cox,
    glm_weight_function = cox_glm_weight_function,
    qp_score_function = cox_qp_score_function,
    dispersion_function = cox_dispersion_function,
    schur_correction_function = cox_schur_correction,
    need_dispersion_for_estimation = FALSE,
    estimate_dispersion = FALSE,
    standardize_response = FALSE,
    status = status,
    ...
  )
}

test_that("Cox PH helpers", {

  ## Test 1: Partial log-likelihood at beta = 0 matches manual computation
  set.seed(1234)
  n <- 50
  time <- rexp(n, 1)
  status <- rbinom(n, 1, 0.7)
  ord <- order(time)
  time_s <- time[ord]
  status_s <- status[ord]

  event_times <- sort(unique(time_s[status_s == 1]))
  ll_manual <- 0
  for(tt in event_times) {
    Dg <- which(time_s == tt & status_s == 1)
    Rg <- which(time_s >= tt)
    ll_manual <- ll_manual + sum(0 - log(length(Rg)))
  }
  ll_fn <- loglik_cox(rep(0, n), status_s, time_s)
  expect_equal(ll_fn, ll_manual, tolerance = 1e-4)

  ## Test 2: Score at MLE should be approximately zero
  set.seed(42)
  n <- 200
  x1 <- rnorm(n)
  x2 <- rbinom(n, 1, 0.5)
  time <- rexp(n, exp(0.3 * x1 + 0.2 * x2))
  status <- rbinom(n, 1, 0.8)

  ord <- order(time)
  time_s <- time[ord]
  X_s <- cbind(x1, x2)[ord, ]
  st_s <- status[ord]

  b <- cbind(rep(0, 2))
  for(i in 1:50) {
    eta <- c(X_s %**% b)
    sc <- score_cox(X_s, eta, st_s, time_s)
    info <- info_cox(X_s, eta, st_s, time_s)
    step <- solve(info, sc)
    b <- b + c(step)
  }

  eta_final <- c(X_s %**% b)
  sc_final <- score_cox(X_s, eta_final, st_s, time_s)
  expect_true(all(abs(sc_final) < 1e-4))

  ## Test 3: Compare with survival::coxph
  skip_if_not_installed("survival")
  df_test <- data.frame(time = time, status = status, x1 = x1, x2 = x2)
  coxfit <- survival::coxph(
    survival::Surv(time, status) ~ x1 + x2,
    data = df_test, ties = "breslow"
  )
  b_coxph <- coef(coxfit)
  names(b_coxph) <- NULL
  expect_equal(c(b), b_coxph, tolerance = 1e-2)

  ## Test 4: Information matrix is positive definite at the MLE
  eta_mle <- c(X_s %**% b)
  I <- info_cox(X_s, eta_mle, st_s, time_s)
  eigvals <- eigen(I, symmetric = TRUE, only.values = TRUE)$values
  expect_true(all(eigvals > 0))

  ## Test 5: Partial log-likelihood increases along Newton direction
  b0 <- cbind(rep(0, 2))
  eta0 <- c(X_s %**% b0)
  ll0 <- loglik_cox(eta0, st_s, time_s)

  sc0 <- score_cox(X_s, eta0, st_s, time_s)
  I0  <- info_cox(X_s, eta0, st_s, time_s)
  step0 <- solve(I0, sc0)
  b1 <- b0 + c(step0)
  eta1 <- c(X_s %**% b1)
  ll1 <- loglik_cox(eta1, st_s, time_s)
  expect_gt(ll1, ll0)

  ## Test 6: unconstrained_fit_cox recovers MLE (no penalty)
  p <- 2
  Lambda_zero <- matrix(0, p, p)
  LambdaHalf_zero <- matrix(0, p, p)

  b_fit <- unconstrained_fit_cox(
    X = cbind(x1, x2), y = time,
    LambdaHalf = LambdaHalf_zero, Lambda = Lambda_zero,
    keep_weighted_Lambda = FALSE, family = cox_family(),
    tol = 1e-10, K = 0,
    parallel = FALSE, cl = NULL,
    chunk_size = NULL, num_chunks = NULL, rem_chunks = NULL,
    order_indices = 1:n, weights = rep(1, n), status = status
  )
  expect_equal(c(b_fit), c(b), tolerance = 1e-2)

  ## Test 7: cox_qp_score_function near zero at MLE
  X_full <- cbind(x1, x2)
  mu_test <- exp(c(X_full %**% b))
  order_list_test <- list(1:n)

  sc_qp <- cox_qp_score_function(
    X = X_full, y = time, mu = mu_test,
    order_list = order_list_test, dispersion = 1,
    VhalfInv = NULL, observation_weights = rep(1, n),
    status = status
  )
  expect_true(all(abs(sc_qp) < 1e-4))

  ## Test 8: Weighted partial log-likelihood
  set.seed(99)
  n_wt <- 20
  time_wt <- rexp(n_wt, 1)
  eta_wt <- rnorm(n_wt, 0, 0.3)
  st_wt <- rbinom(n_wt, 1, 0.6)
  wt_wt <- runif(n_wt, 0.5, 2.0)

  ord_wt <- order(time_wt)
  time_wt_s <- time_wt[ord_wt]
  eta_wt_s <- eta_wt[ord_wt]
  st_wt_s <- st_wt[ord_wt]
  wt_wt_s <- wt_wt[ord_wt]

  haz_wt <- wt_wt_s * exp(eta_wt_s)
  event_times_wt <- sort(unique(time_wt_s[st_wt_s == 1]))
  ll_manual <- 0
  for(tt in event_times_wt) {
    Dg <- which(time_wt_s == tt & st_wt_s == 1)
    Rg <- which(time_wt_s >= tt)
    dgw <- sum(wt_wt_s[Dg])
    ll_manual <- ll_manual +
      sum(wt_wt_s[Dg] * eta_wt_s[Dg]) -
      dgw * log(sum(haz_wt[Rg]))
  }
  ll_fn <- loglik_cox(eta_wt_s, st_wt_s, time_wt_s, wt_wt_s)
  expect_equal(ll_fn, ll_manual, tolerance = 1e-4)

  ll_unit <- loglik_cox(eta_wt_s, st_wt_s, time_wt_s, rep(1, n_wt))
  ll_default <- loglik_cox(eta_wt_s, st_wt_s, time_wt_s)
  expect_equal(ll_unit, ll_default, tolerance = 1e-4)

  ## Test 9: Dispersion function returns 1
  d_val <- cox_dispersion_function(mu_test, time, 1:n, cox_family(),
                                   rep(1, n), NULL)
  expect_identical(d_val, 1)

  ## Test 10: GLM weight function returns positive values
  wt_vals <- cox_glm_weight_function(
    mu = mu_test, y = time, order_indices = 1:n,
    family = cox_family(), dispersion = 1,
    observation_weights = rep(1, n), status = status
  )
  expect_true(all(wt_vals > 0))
  expect_length(wt_vals, n)
})

## tune_Lambda.R
#  Tuning for the smoothing and ridge penalties via exact leave-one-out
#  or modified generalized cross-validation.
#
## Main pieces:
#    tune_Lambda()
#    .compute_gcvu() and .compute_gcvu_gradient()
#    .damped_bfgs() and .tune_grid_search()
#    small helpers for predictions, residuals, and meta-penalties


## Residuals used in tuning criteria

#' Compute Residuals During Penalty Tuning
#'
#' @description
#' Computes the residuals used by the tuning criteria. Handles identity link,
#' general GLM link functions with pseudocount delta, custom deviance residual
#' functions, and observation weights.
#'
#' @param y List; response vectors by partition.
#' @param preds List; prediction vectors by partition.
#' @param delta Numeric; pseudocount for link function stabilization.
#' @param family GLM family object.
#' @param observation_weights List; observation weights by partition.
#' @param K Integer; number of interior knots (partitions - 1).
#' @param order_list List; observation indices per partition.
#' @param ... Additional arguments passed to \code{family$custom_dev.resids}.
#'
#' @return List of residual vectors, one per partition.
#'
#' @details
#' Three computation paths:
#' \enumerate{
#'   \item \strong{Identity link or no custom deviance}: Standard
#'     \eqn{g(y) - \hat{\eta}} residuals on the link scale, optionally
#'     weighted by observation weights for non-Gaussian families.
#'     For Gaussian identity link with heterogeneous weights, the weights
#'     have already been absorbed into X and y prior to this call.
#'   \item \strong{Custom deviance residuals}: Delegates to
#'     \code{family$custom_dev.resids(y, mu, order_indices, family,
#'     observation_weights, ...)}.
#' }
#'
#' @keywords internal
.compute_tuning_residuals <- function(y,
                                      preds,
                                      delta,
                                      family,
                                      observation_weights,
                                      K,
                                      order_list,
                                      ...) {
  if (paste0(family)[2] == "identity" ||
      is.null(family$custom_dev.resids)) {

    ## If not canonical Gaussian and weights are present, use them.
    ## Recall: for Gaussian identity, X and y were pre-weighted upstream.
    if (any(!is.null(observation_weights[[1]])) &&
        (paste0(family)[2] != "identity" ||
         paste0(family)[1] != "gaussian")) {
      residuals <- lapply(1:(K + 1), function(k) {
        (family$linkfun((y[[k]] + delta) / (1 + 2 * delta)) -
           (preds[[k]] + delta) / (1 + 2 * delta)) *
          c(observation_weights[[k]])
      })
    } else {
      residuals <- lapply(1:(K + 1), function(k) {
        family$linkfun((y[[k]] + delta) / (1 + 2 * delta)) -
          (preds[[k]] + delta) / (1 + 2 * delta)
      })
    }

  } else {
    ## Otherwise use the family's custom deviance residuals.
    residuals <- lapply(1:(K + 1), function(k) {
      family$custom_dev.resids(y[[k]],
                               family$linkinv(c(preds[[k]])),
                               order_list[[k]],
                               family,
                               observation_weights[[k]],
                               ...)
    })
  }

  return(residuals)
}


## Predictions used in tuning criteria

#' Compute Predictions During Penalty Tuning
#'
#' @description
#' Wrapper around \code{matmult_block_diagonal} for computing partition-wise
#' predictions \eqn{\mathbf{X}_{k} \boldsymbol{\beta}_{k}} during penalty
#' tuning.
#'
#' @param X List; design matrices by partition.
#' @param B List; coefficient vectors by partition.
#' @param K Integer; number of interior knots.
#' @param parallel Logical; use parallel computation.
#' @param cl Parallel cluster object.
#' @param chunk_size,num_chunks,rem_chunks Integer; parallel chunking parameters.
#'
#' @return List of prediction vectors, one per partition.
#'
#' @keywords internal
.compute_tuning_predictions <- function(X, B, K,
                                        parallel, cl,
                                        chunk_size, num_chunks,
                                        rem_chunks) {
  matmult_block_diagonal(X, B, K,
                         parallel, cl,
                         chunk_size, num_chunks, rem_chunks)
}


## Meta-penalty on the tuning parameters

#' Compute Regularization (Meta) Penalty on Penalty Parameters
#'
#' @description
#' Computes the regularization term that pulls predictor- and partition-specific
#' penalty parameters toward 1 on the raw (positive) scale. This acts as a
#' "meta-penalty" on the penalty magnitudes themselves.
#'
#' @param wiggle_penalty Numeric; current wiggle penalty on raw scale.
#' @param penalty_vec Numeric vector; current predictor/partition penalties
#'   on raw scale. May be empty (\code{c()}).
#' @param meta_penalty_coef Numeric; coefficient for the meta-penalty.
#' @param unique_penalty_per_predictor Logical; whether predictor-specific
#'   penalties are active.
#' @param unique_penalty_per_partition Logical; whether partition-specific
#'   penalties are active.
#'
#' @return Numeric scalar; the regularization penalty value.
#'
#' @details
#' The penalty takes the form:
#' \deqn{0.5 \times c_{\mathrm{meta}} \times \sum_{j} (\lambda_{j} - 1)^{2}
#'       + 0.5 \times 10^{-32} \times (\lambda_{w} - 1)^{2}}
#' where \eqn{\lambda_{j}} are predictor/partition penalties and
#' \eqn{\lambda_{w}} is the wiggle penalty.
#'
#' @keywords internal
.compute_meta_penalty <- function(wiggle_penalty,
                                  penalty_vec,
                                  meta_penalty_coef,
                                  unique_penalty_per_predictor,
                                  unique_penalty_per_partition) {
  if (unique_penalty_per_partition || unique_penalty_per_predictor) {
    0.5 * meta_penalty_coef * sum((penalty_vec - 1)^2) +
      0.5 * 1e-32 * ((wiggle_penalty - 1))^2
  } else {
    0.5 * 1e-32 * ((wiggle_penalty - 1))^2
  }
}


## Gradient of the meta-penalty

#' Compute Gradient of Regularization (Meta) Penalty
#'
#' @description
#' Computes the gradient of the meta-penalty with respect to the log-scale
#' penalty parameters, incorporating the exp parameterization chain rule.
#'
#' @param wiggle_penalty Numeric; current wiggle penalty on raw scale.
#' @param penalty_vec Numeric vector; current predictor/partition penalties
#'   on raw scale. May be empty (\code{c()}).
#' @param meta_penalty_coef Numeric; coefficient for the meta-penalty.
#' @param unique_penalty_per_predictor Logical; whether predictor-specific
#'   penalties are active.
#' @param unique_penalty_per_partition Logical; whether partition-specific
#'   penalties are active.
#'
#' @return Numeric vector; gradient of the meta-penalty on the log scale.
#'   Length equals 2 + length(penalty_vec).
#'
#' @details
#' Under exp parameterization \eqn{\lambda = \exp(\theta)}:
#' \deqn{\frac{\partial}{\partial \theta}
#'   \left[ 0.5 c (\exp(\theta) - 1)^{2} \right]
#'   = c (\lambda - 1) \lambda}
#'
#' @keywords internal
.compute_meta_penalty_gradient <- function(wiggle_penalty,
                                           penalty_vec,
                                           meta_penalty_coef,
                                           unique_penalty_per_predictor,
                                           unique_penalty_per_partition) {
  if (unique_penalty_per_partition || unique_penalty_per_predictor) {
    c(1e-32 * (wiggle_penalty - 1) * wiggle_penalty,
      0,
      meta_penalty_coef * (penalty_vec - 1) * penalty_vec)
  } else {
    c(1e-32 * (wiggle_penalty - 1) * wiggle_penalty,
      0)
  }
}


## Coefficient fit inside tuning

#' Fit Coefficients During Penalty Tuning: blockfit_solve or get_B
#'
#' @description
#' Dispatches to \code{blockfit_solve} when the blockfit conditions are met
#' (i.e. \code{env$use_blockfit} is TRUE), otherwise calls \code{get_B}.
#' On \code{blockfit_solve} failure, falls back to \code{get_B} automatically.
#'
#' @details
#' The blockfit condition mirrors \code{lgspline.fit}:
#' \code{blockfit && length(flat_cols) > 0 && K > 0}, pre-computed in
#' \code{tune_Lambda} and stored in \code{env$use_blockfit}.
#'
#' \code{return_G_getB} is set to TRUE by the callers so that
#' \code{B_list$G_list} contains the updated G matrices (after any GLM
#' weight iteration inside \code{get_B} or \code{blockfit_solve}).
#' These are needed immediately after this call for \code{AGAmult_wrapper},
#' \code{GXX}, and the trace computation.
#'
#' When correlation structure inputs are supplied, this wrapper does not
#' introduce a separate tuning-specific notation or solver path. Instead, the
#' same correlated coefficient estimator used in the final model fit is called
#' here inside each tuning-objective evaluation. In particular, any
#' structured-correlation Woodbury factorization
#' \eqn{\mathbf{G}^{-1} =
#' \mathbf{G}_{\mathrm{on}}^{-1} + \mathbf{G}_{\mathrm{off}}^{-1}}
#' is handled inside \code{get_B} and documented in \code{lgspline-details}.
#'
#' @param G_list List; eigendecomposition results from \code{compute_G_eigen},
#'   containing \code{Ghalf} and \code{GhalfInv}.
#' @param Lambda Matrix; current combined penalty matrix.
#' @param L_partition_list List; partition-specific penalty matrices.
#' @param env List; tuning environment from \code{.build_tuning_env}.
#' @param return_G_getB Logical; whether to return G inside the fit. Set to
#'   TRUE within \code{.compute_gcvu} and \code{.compute_gcvu_gradient} so
#'   that \code{B_list$G_list} carries the (possibly GLM-iterated) G matrices
#'   required for the subsequent \code{AGAmult_wrapper} and trace computations.
#' @param ... Additional arguments forwarded to the fitting routine.
#'
#' @return List; output of \code{blockfit_solve} or \code{get_B}, containing
#'   at minimum \code{$B} (coefficient list) and \code{$G_list}.
#'
#' @keywords internal
.fit_coefficients <- function(G_list,
                              Lambda,
                              L_partition_list,
                              env,
                              return_G_getB,
                              ...) {

  ## Attempt blockfit_solve when conditions are met
  if (env$use_blockfit) {
    B_list <- try({
      blockfit_solve(
        X                              = env$X,
        y                              = env$y,
        flat_cols                      = env$flat_cols,
        K                              = env$K,
        p_expansions                   = env$p_expansions,
        Lambda                         = Lambda,
        L_partition_list               = L_partition_list,
        unique_penalty_per_partition   = env$unique_penalty_per_partition,
        A                              = env$A,
        R_constraints                  = env$R_constraints,
        constraint_values              = env$constraint_value_vectors,
        X_gram                         = env$X_gram,
        Ghalf_full                     = G_list$Ghalf,
        GhalfInv_full                  = G_list$GhalfInv,
        family                         = env$family,
        order_list                     = env$order_list,
        glm_weight_function            = env$glm_weight_function,
        schur_correction_function      = env$schur_correction_function,
        need_dispersion_for_estimation = env$need_dispersion_for_estimation,
        dispersion_function            = env$dispersion_function,
        observation_weights            = env$observation_weights,
        homogenous_weights             = env$homogenous_weights,
        iterate                        = env$iterate,
        tol                            = env$tol,
        parallel_eigen                 = env$parallel & env$parallel_eigen,
        cl                             = env$cl,
        chunk_size                     = env$chunk_size,
        num_chunks                     = env$num_chunks,
        rem_chunks                     = env$rem_chunks,
        return_G_getB                  = return_G_getB,
        quadprog                       = env$quadprog,
        qp_Amat                        = env$qp_Amat,
        qp_bvec                        = env$qp_bvec,
        qp_meq                         = env$qp_meq,
        qp_score_function              = env$qp_score_function,
        keep_weighted_Lambda           = env$keep_weighted_Lambda,
        max_backfit_iter               = 100,
        Vhalf                          = env$Vhalf,
        VhalfInv                       = env$VhalfInv,
        include_warnings               = env$include_warnings,
        verbose                        = env$verbose,
        ...
      )
    }, silent = TRUE)

    ## Fall back to get_B on failure
    if (inherits(B_list, "try-error")) {
      if (env$include_warnings) {
        warning("\n \t blockfit_solve failed during GCV tuning, ",
                "falling back to get_B \n")
      }
      B_list <- .fit_get_B(G_list, Lambda, L_partition_list,
                           env, return_G_getB, ...)
    }

    return(B_list)
  }

  ## Otherwise use the usual get_B path.
  .fit_get_B(G_list, Lambda, L_partition_list, env, return_G_getB, ...)
}


## Small get_B wrapper for tuning

#' Call get_B During Penalty Tuning
#'
#' @description
#' Internal wrapper that calls \code{get_B} with all arguments drawn from
#' the tuning environment \code{env}. Separated from \code{.fit_coefficients}
#' so the fallback path in \code{.fit_coefficients} is clean and does not
#' repeat the full argument list.
#'
#' This helper is the point at which penalty tuning inherits the full
#' constrained solver described in \code{lgspline-details}, including GLM
#' reweighting, dense correlation whitening, and the structured-correlation
#' Woodbury acceleration when available. No additional tuning-specific
#' approximation is introduced here beyond whatever \code{get_B} itself uses.
#'
#' @inheritParams .fit_coefficients
#' @keywords internal
.fit_get_B <- function(G_list,
                       Lambda,
                       L_partition_list,
                       env,
                       return_G_getB,
                       ...) {
  get_B(
    env$X,
    env$X_gram,
    Lambda,
    env$keep_weighted_Lambda,
    env$unique_penalty_per_partition,
    L_partition_list,
    env$A,
    env$Xy,
    env$y,
    env$K,
    env$p_expansions,
    env$R_constraints,
    G_list$Ghalf,
    G_list$GhalfInv,
    env$parallel & env$parallel_eigen,
    env$parallel & env$parallel_aga,
    env$parallel & env$parallel_matmult,
    env$parallel & env$parallel_unconstrained,
    env$cl,
    env$chunk_size,
    env$num_chunks,
    env$rem_chunks,
    env$family,
    env$unconstrained_fit_fxn,
    env$iterate,
    env$qp_score_function,
    env$quadprog,
    env$qp_Amat,
    env$qp_bvec,
    env$qp_meq,
    prevB                          = NULL,
    prevUnconB                     = NULL,
    iter_count                     = 0,
    prev_diff                      = Inf,
    env$tol,
    env$constraint_value_vectors,
    env$order_list,
    env$glm_weight_function,
    env$schur_correction_function,
    env$need_dispersion_for_estimation,
    env$dispersion_function,
    env$observation_weights,
    env$homogenous_weights,
    return_G_getB,
    env$blockfit,
    env$just_linear_without_interactions,
    env$Vhalf,
    env$VhalfInv,
    ...
  )
}


## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
# Subfunction: .compute_gcvu
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

#' Evaluate Modified GCV_u Criterion at a Given Penalty Configuration
#'
#' @description
#' Computes the modified GCV_u criterion for a given set of penalty
#' parameters. This is the objective function minimized during penalty tuning;
#' \code{gcv_gamma = 1} recovers ordinary GCV.
#'
#' @param par Numeric vector; log-scale penalty parameters. First two elements
#'   are log(wiggle_penalty) and log(flat_ridge_penalty). Remaining elements
#'   (if any) are log-scale predictor/partition penalties.
#' @param log_penalty_vec Numeric vector; log-scale predictor/partition
#'   penalties (passed separately for compatibility with the grid search).
#' @param env List; pre-computed objects and tuning configuration. Contains:
#'   \describe{
#'     \item{X, y, X_gram, Xy}{Design matrices, response, Gram matrices,
#'       cross-products.}
#'     \item{A, R_constraints, K, p_expansions, N_obs}{Constraint matrix and
#'       dimensions.}
#'     \item{smoothing_spline_penalty}{Integrated second-derivative penalty
#'       matrix.}
#'     \item{custom_penalty_mat}{Optional custom penalty matrix.}
#'     \item{colnm_expansions}{Character vector of expansion column names.}
#'     \item{unique_penalty_per_predictor, unique_penalty_per_partition}{Logicals.}
#'     \item{family}{GLM family object.}
#'     \item{delta}{Pseudocount for link function stabilization.}
#'     \item{gcv_gamma}{Modified-GCV inflation factor applied to the
#'       effective-degrees-of-freedom term in the denominator.}
#'     \item{meta_penalty}{Meta-penalty coefficient.}
#'     \item{order_list}{List of observation indices per partition.}
#'     \item{observation_weights, homogenous_weights}{Observation weighting.}
#'     \item{parallel flags, cl, chunk_size, num_chunks, rem_chunks}{Parallel
#'       config.}
#'     \item{unconstrained_fit_fxn, keep_weighted_Lambda, iterate}{Fitting
#'       config.}
#'     \item{qp_score_function, quadprog, qp_Amat, qp_bvec, qp_meq}{QP config.}
#'     \item{tol, sd_y}{Convergence tolerance and response scale.}
#'     \item{constraint_value_vectors}{Constraint values.}
#'     \item{glm_weight_function, schur_correction_function}{GLM functions.}
#'     \item{need_dispersion_for_estimation, dispersion_function}{Dispersion.}
#'     \item{blockfit, just_linear_without_interactions}{Blockfit config.}
#'     \item{use_blockfit}{Logical; pre-computed dispatch flag for blockfit.}
#'     \item{flat_cols}{Integer vector; pre-computed flat column indices.}
#'     \item{Vhalf, VhalfInv}{Correlation structure matrices.}
#'     \item{verbose, include_warnings}{Output control.}
#'   }
#' @param ... Additional arguments passed to fitting functions.
#'
#' @return List containing:
#' \describe{
#'   \item{GCV_u}{Numeric; GCV_u criterion value including meta-penalty.}
#'   \item{B}{List; fitted coefficient vectors by partition.}
#'   \item{GXX}{List; \eqn{\mathbf{G}_{k} \mathbf{X}_{k}^{\top}\mathbf{X}_{k}}
#'     matrices.}
#'   \item{G_list}{List; eigendecomposition results from
#'     \code{compute_G_eigen}.}
#'   \item{mean_W}{Numeric; \eqn{\mathrm{tr}(\mathbf{H})/N}, the average
#'     leverage entering the denominator of \eqn{\mathrm{GCV}_u}.}
#'   \item{sum_W}{Numeric; \eqn{\mathrm{tr}(\mathbf{H})}, the effective
#'     degrees of freedom.}
#'   \item{Lambda}{Matrix; combined penalty matrix
#'     \eqn{\boldsymbol{\Lambda}}.}
#'   \item{L1}{Matrix; baseline smoothness penalty component.}
#'   \item{L2}{Matrix; baseline ridge penalty component.}
#'   \item{L_predictor_list}{List; predictor-specific penalty matrices.}
#'   \item{L_partition_list}{List; partition-specific penalty matrices.}
#'   \item{numerator}{Numeric; sum of squared residuals.}
#'   \item{denominator}{Numeric; modified-GCV denominator
#'     \eqn{N(1 - \gamma\bar{W})^{2}}.}
#'   \item{residuals}{List; residual vectors by partition.}
#'   \item{denom_sq}{Numeric; squared denominator.}
#'   \item{AGAInv}{Matrix; \eqn{(\mathbf{A}^{\top}\mathbf{G}\mathbf{A})^{-1}}.}
#' }
#'
#' @keywords internal
.compute_gcvu <- function(par,
                          log_penalty_vec,
                          env,
                          ...) {
  verbose <- env$verbose
  gamma <- env$gcv_gamma

  if (verbose) cat("        gcvu_fxn start\n")

  ## Unpack penalties from log scale
  wiggle_penalty     <- exp(par[1])
  flat_ridge_penalty <- exp(par[2])
  if (env$unique_penalty_per_predictor || env$unique_penalty_per_partition) {
    penalty_vec <- exp(c(par[-c(1:2)]))
  } else {
    penalty_vec <- c()
  }

  ## Compute penalty matrix Lambda and components
  if (verbose) cat("        compute_Lambda\n")
  Lambda_list <- compute_Lambda(env$custom_penalty_mat,
                                env$smoothing_spline_penalty,
                                wiggle_penalty,
                                flat_ridge_penalty,
                                env$K,
                                env$p_expansions,
                                env$unique_penalty_per_predictor,
                                env$unique_penalty_per_partition,
                                penalty_vec,
                                env$colnm_expansions,
                                just_Lambda = FALSE)
  Lambda <- Lambda_list[[1]]
  L1     <- Lambda_list[[2]]
  L2     <- Lambda_list[[3]]

  ## Compute G matrices via eigendecomposition.
  #  keep_G = TRUE always: G is needed for the trace computation below
  #  regardless of whether get_B or blockfit_solve is used for fitting.
  if (verbose) cat("        compute_G_eigen\n")
  schur_corrections <- lapply(1:(env$K + 1), function(k) 0)
  G_list <- compute_G_eigen(env$X_gram,
                            Lambda,
                            env$K,
                            env$parallel & env$parallel_eigen,
                            env$cl,
                            env$chunk_size,
                            env$num_chunks,
                            env$rem_chunks,
                            env$family,
                            env$unique_penalty_per_partition,
                            Lambda_list$L_partition_list,
                            keep_G = TRUE,
                            schur_corrections)

  ## Fit coefficients: dispatches to blockfit_solve or get_B via
  #  .fit_coefficients, with automatic fallback on failure.
  #  return_G_getB = TRUE: B_list$G_list carries the (possibly GLM-iterated)
  #  G matrices needed for AGAmult_wrapper and trace computations below.
  if (verbose) cat("        gcvu_fxn fit coefficients\n")
  return_G_getB <- TRUE
  B_list <- .fit_coefficients(G_list, Lambda,
                              Lambda_list$L_partition_list,
                              env, return_G_getB, ...)
  G_list <- B_list$G_list
  B      <- B_list$B

  ## Compute (A^T G A)^{-1}
  if (verbose) cat("        gcvu_fxn AGAmult_wrapper\n")
  AGAInv <- invert(AGAmult_wrapper(G_list$G,
                                   env$A,
                                   env$K,
                                   env$p_expansions,
                                   env$R_constraints,
                                   env$parallel & env$parallel_aga,
                                   env$cl,
                                   env$chunk_size,
                                   env$num_chunks,
                                   env$rem_chunks) +
                     1e-16 * diag(ncol(env$A)))

  ## Compute G * X^T X
  if (verbose) cat("        gcvu_fxn matmult_block_diagonal for GXX\n")
  GXX <- matmult_block_diagonal(G_list$G,
                                env$X_gram,
                                env$K,
                                env$parallel & env$parallel_matmult,
                                env$cl,
                                env$chunk_size,
                                env$num_chunks,
                                env$rem_chunks)

  ## Compute trace of hat matrix
  if (verbose) cat("        gcvu_fxn compute_trace_UGXX_wrapper\n")
  sum_W <- compute_trace_H(G_list$G,
                           Lambda,
                           env$A,
                           AGAInv,
                           env$p_expansions,
                           env$R_constraints,
                           env$K,
                           env$parallel,
                           env$cl,
                           env$chunk_size,
                           env$num_chunks,
                           env$rem_chunks,
                           env$unique_penalty_per_partition,
                           Lambda_list$L_partition_list)

  ## Predictions
  if (verbose) cat("        gcvu_fxn get predictions\n")
  preds <- .compute_tuning_predictions(env$X, B, env$K,
                                       env$parallel & env$parallel_matmult,
                                       env$cl,
                                       env$chunk_size,
                                       env$num_chunks,
                                       env$rem_chunks)

  ## Residuals
  if (verbose) cat("        gcvu_fxn custom or default residuals\n")
  residuals <- .compute_tuning_residuals(env$y, preds, env$delta,
                                         env$family,
                                         env$observation_weights,
                                         env$K, env$order_list,
                                         ...)

  ## Compute GCV_u components
  if (verbose) cat("        gcvu_fxn GCVu operations\n")
  numerator   <- sum(unlist(residuals)^2)
  mean_W      <- sum_W / env$N_obs
  denominator <- env$N_obs * (1 - gamma * mean_W)^2
  denom_sq    <- denominator^2
  GCV_u       <- numerator / denominator

  ## Meta-penalty regularization
  if (verbose) cat("        gcvu_fxn penalization operations\n")
  mp <- .compute_meta_penalty(wiggle_penalty, penalty_vec,
                              env$meta_penalty,
                              env$unique_penalty_per_predictor,
                              env$unique_penalty_per_partition)

  if (verbose) cat("        done GCVu,", GCV_u, "\n")

  return(list(criterion_value = GCV_u + mp,
              GCV_u       = GCV_u + mp,
              B           = B,
              GXX         = GXX,
              G_list      = G_list,
              mean_W      = mean_W,
              sum_W       = sum_W,
              Lambda      = Lambda,
              L1          = L1,
              L2          = L2,
              L_predictor_list = Lambda_list$L_predictor_list,
              L_partition_list = Lambda_list$L_partition_list,
              numerator   = numerator,
              denominator = denominator,
              residuals   = residuals,
              denom_sq    = denom_sq,
              AGAInv      = AGAInv))
}


## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
# Subfunction: .compute_gcvu_gradient
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

#' Compute Closed-Form Gradient of GCV_u Criterion
#'
#' @description
#' Computes the gradient of the modified GCV_u criterion with respect to the
#' log-scale penalty parameters using analytical derivatives of the hat matrix
#' trace and residual sum of squares.
#'
#' @param par Numeric vector; log-scale penalty parameters.
#' @param log_penalty_vec Numeric vector; log-scale predictor/partition
#'   penalties.
#' @param outlist List or NULL; pre-computed GCV_u components from
#'   \code{.compute_gcvu}. If NULL, they are computed internally.
#' @param env List; pre-computed objects and tuning configuration (same
#'   structure as in \code{.compute_gcvu}).
#' @param ... Additional arguments passed to fitting functions.
#'
#' @return List containing:
#' \describe{
#'   \item{GCV_u}{Numeric; GCV_u criterion value including meta-penalty.}
#'   \item{gradient}{Numeric vector; gradient on the log penalty scale.}
#'   \item{outlist}{List; GCV_u components (for reuse to avoid recomputation).}
#' }
#'
#' @details
#' The gradient is computed via:
#' \deqn{\frac{\partial \mathrm{GCV}_u}{\partial \theta}
#'   = \frac{1}{D^{2}} \left(
#'     \frac{\partial N}{\partial \theta} D
#'     - N \frac{\partial D}{\partial \theta}
#'   \right)}
#' where \eqn{N = \sum r_{i}^{2}} (numerator),
#' \eqn{D = n(1 - \gamma\bar{W})^{2}} (denominator), \eqn{\theta} is the
#' log-scale penalty parameter, and the chain rule
#' \eqn{d\lambda / d\theta = \lambda} (exp parameterization) is applied.
#'
#' The shared wiggle and flat directions are differentiated directly from the
#' fitted criterion. For predictor- and partition-specific penalties, a
#' lower-cost ratio heuristic is used on the log-penalty scale:
#' \deqn{\frac{\partial \mathrm{GCV}_u}{\partial \theta_{j}}
#'   \approx \frac{\mathrm{tr}(\mathbf{L}_{j})}{\mathrm{tr}(\boldsymbol{\Lambda})}
#'   \cdot \frac{\partial \mathrm{GCV}_u}{\partial \theta_{w}},}
#' where \eqn{\mathbf{L}_{j}} denotes the current contribution of that extra
#' penalty to \eqn{\boldsymbol{\Lambda}}.
#'
#' @keywords internal
.compute_gcvu_gradient <- function(par,
                                   log_penalty_vec,
                                   outlist = NULL,
                                   env,
                                   ...) {
  verbose <- env$verbose
  gamma <- env$gcv_gamma

  if (verbose) cat("        gr_fxn start\n")

  ## Unpack penalties from log scale
  wiggle_penalty     <- exp(par[1])
  flat_ridge_penalty <- exp(par[2])
  if (env$unique_penalty_per_predictor || env$unique_penalty_per_partition) {
    penalty_vec <- exp(c(par[-c(1:2)]))
  } else {
    penalty_vec <- c()
  }

  ## Reparameterize
  lambda_1 <- wiggle_penalty
  lambda_2 <- flat_ridge_penalty

  if (verbose) {
    cat("        lambda_1, lambda_2: ", lambda_1, ", ", lambda_2, "\n")
  }

  ## Recompute the fitted GCV_u pieces only when they were not already
  #  passed in from the same parameter value.
  if (any(is.null(outlist))) {

    if (verbose) cat("        Lambda list\n")
    Lambda_list <- compute_Lambda(env$custom_penalty_mat,
                                  env$smoothing_spline_penalty,
                                  wiggle_penalty,
                                  flat_ridge_penalty,
                                  env$K,
                                  env$p_expansions,
                                  env$unique_penalty_per_predictor,
                                  env$unique_penalty_per_partition,
                                  penalty_vec,
                                  env$colnm_expansions,
                                  just_Lambda = FALSE)
    Lambda <- Lambda_list[[1]]
    L1     <- Lambda_list[[2]]
    L2     <- Lambda_list[[3]]

    if (verbose) cat("        G list\n")
    schur_corrections <- lapply(1:(env$K + 1), function(k) 0)
    G_list <- compute_G_eigen(env$X_gram,
                              Lambda,
                              env$K,
                              env$parallel & env$parallel_eigen,
                              env$cl,
                              env$chunk_size,
                              env$num_chunks,
                              env$rem_chunks,
                              env$family,
                              env$unique_penalty_per_partition,
                              Lambda_list$L_partition_list,
                              keep_G = TRUE,
                              schur_corrections)

    ## Fit coefficients: dispatches to blockfit_solve or get_B
    #  return_G_getB = TRUE: B_list$G_list carries the (possibly GLM-iterated)
    #  G matrices needed for AGAmult_wrapper and trace computations below.
    if (verbose) cat("        gr fxn fit coefficients\n")
    return_G_getB <- TRUE
    B_list <- .fit_coefficients(G_list, Lambda,
                                Lambda_list$L_partition_list,
                                env, return_G_getB, ...)
    G_list <- B_list$G_list
    B      <- B_list$B

    if (verbose) cat("        AGAmult_wrapper\n")
    AGAInv <- invert(AGAmult_wrapper(G_list$G,
                                     env$A,
                                     env$K,
                                     env$p_expansions,
                                     env$R_constraints,
                                     env$parallel & env$parallel_aga,
                                     env$cl,
                                     env$chunk_size,
                                     env$num_chunks,
                                     env$rem_chunks) +
                       1e-16 * diag(ncol(env$A)))

    if (verbose) cat("        GXX matmult_block_diagonal\n")
    GXX <- matmult_block_diagonal(G_list$G,
                                  env$X_gram,
                                  env$K,
                                  env$parallel & env$parallel_matmult,
                                  env$cl,
                                  env$chunk_size,
                                  env$num_chunks,
                                  env$rem_chunks)

    if (verbose) cat("        sum_W compute_trace_UGXX_wrapper\n")
    sum_W <- compute_trace_H(G_list$G,
                             Lambda,
                             env$A,
                             AGAInv,
                             env$p_expansions,
                             env$R_constraints,
                             env$K,
                             env$parallel,
                             env$cl,
                             env$chunk_size,
                             env$num_chunks,
                             env$rem_chunks,
                             env$unique_penalty_per_partition,
                             Lambda_list$L_partition_list)

    if (verbose) cat("        gr fxn preds\n")
    preds <- .compute_tuning_predictions(env$X, B, env$K,
                                         env$parallel & env$parallel_matmult,
                                         env$cl,
                                         env$chunk_size,
                                         env$num_chunks,
                                         env$rem_chunks)

    if (verbose) cat("        gr fxn residuals\n")
    residuals <- .compute_tuning_residuals(env$y, preds, env$delta,
                                           env$family,
                                           env$observation_weights,
                                           env$K, env$order_list,
                                           ...)

    if (verbose) cat("        gr fxn compute GCV_u\n")
    numerator   <- sum(unlist(residuals)^2)
    mean_W      <- sum_W / env$N_obs
    denominator <- env$N_obs * (1 - gamma * mean_W)^2
    denom_sq    <- denominator^2
    GCV_u       <- numerator / denominator

    mp <- .compute_meta_penalty(wiggle_penalty, penalty_vec,
                                env$meta_penalty,
                                env$unique_penalty_per_predictor,
                                env$unique_penalty_per_partition)

    if (verbose) cat("        gr fxn outlist\n")
    outlist <- list(criterion_value = GCV_u + mp,
                    GCV_u       = GCV_u + mp,
                    B           = B,
                    GXX         = GXX,
                    G_list      = G_list,
                    mean_W      = mean_W,
                    sum_W       = sum_W,
                    Lambda      = Lambda,
                    L1          = L1,
                    L2          = L2,
                    L_predictor_list = Lambda_list$L_predictor_list,
                    L_partition_list = Lambda_list$L_partition_list,
                    numerator   = numerator,
                    denominator = denominator,
                    residuals   = residuals,
                    denom_sq    = denom_sq,
                    AGAInv      = AGAInv)
  }

  ## Differentiate the same fitted objects above so we do not have to
  #  refit again just to get the gradient.

  ## Key intermediate for the derivative calculations.
  if (verbose) cat("        GhalfXy_temp_list \n")
  GhalfXy_temp_list <- compute_GhalfXy_temp_wrapper(
    outlist$G_list$G,
    outlist$G_list$Ghalf,
    env$A,
    outlist$AGAInv,
    env$Xy,
    env$p_expansions,
    env$K,
    env$parallel & env$parallel_aga,
    env$cl,
    env$chunk_size,
    env$num_chunks,
    env$rem_chunks)
  GhalfXy_temp <- GhalfXy_temp_list[[1]]
  AGAInvAGXy   <- GhalfXy_temp_list[[2]]

  ## dG/dlambda
  if (verbose) cat("        compute_dG_dlambda \n")
  dPenalty_dlambda1 <- outlist$Lambda / lambda_1
  dPenalty_partition_dlambda1 <- if (env$unique_penalty_per_partition) {
    lapply(outlist$L_partition_list, function(L_k) L_k / lambda_1)
  } else {
    list()
  }
  dG_dlambda <- compute_dG_dlambda(outlist$G_list$G,
                                   dPenalty_dlambda1,
                                   env$K,
                                   env$unique_penalty_per_partition,
                                   dPenalty_partition_dlambda1,
                                   env$parallel & env$parallel_matmult,
                                   env$cl,
                                   env$chunk_size,
                                   env$num_chunks,
                                   env$rem_chunks)

  ## dG^{1/2}/dlambda
  if (verbose) cat("        Compute dGhalf \n")
  dGhalf <- compute_dGhalf(dG_dlambda,
                           env$p_expansions,
                           env$K,
                           env$parallel & env$parallel_eigen,
                           env$cl,
                           env$chunk_size,
                           env$num_chunks,
                           env$rem_chunks)

  ## d(U G)/dlambda * X^T y
  if (verbose) cat("        compute_dG_u_dlambda_xy \n")
  dG_u_dlambda1_Xyr <- compute_dG_u_dlambda_xy(
    AGAInvAGXy,
    outlist$AGAInv,
    outlist$G_list$G,
    env$A,
    dG_dlambda,
    env$p_expansions,
    env$R_constraints,
    env$K,
    env$Xy,
    outlist$G_list$Ghalf,
    dGhalf,
    GhalfXy_temp,
    env$parallel & env$parallel_matmult,
    env$cl,
    env$chunk_size,
    env$num_chunks,
    env$rem_chunks)

  ## dW/dlambda (trace derivative)
  #  trace(H) is the sum of the hat diagonals, so differentiate that
  #  quantity directly rather than relying on the older trace helper.
  if (verbose) cat("        compute_dhat_diag for dW \n")
  dW_dlambda <- sum(unlist(
    .compute_dhat_diag(
      env$X,
      outlist$G_list$G,
      dG_dlambda,
      env$A,
      outlist$AGAInv,
      env$K,
      env$p_expansions
    )$dh_diag
  ))
  dPenalty_dlambda2 <- (lambda_1 / lambda_2) * outlist$L2
  dPenalty_partition_dlambda2 <- if (env$unique_penalty_per_partition) {
    lapply(outlist$L_partition_list, function(L_k) 0 * L_k)
  } else {
    list()
  }
  dG_dlambda2 <- compute_dG_dlambda(outlist$G_list$G,
                                    dPenalty_dlambda2,
                                    env$K,
                                    env$unique_penalty_per_partition,
                                    dPenalty_partition_dlambda2,
                                    env$parallel & env$parallel_matmult,
                                    env$cl,
                                    env$chunk_size,
                                    env$num_chunks,
                                    env$rem_chunks)
  dGhalf2 <- compute_dGhalf(dG_dlambda2,
                            env$p_expansions,
                            env$K,
                            env$parallel & env$parallel_eigen,
                            env$cl,
                            env$chunk_size,
                            env$num_chunks,
                            env$rem_chunks)
  dG_u_dlambda2_Xyr <- compute_dG_u_dlambda_xy(
    AGAInvAGXy,
    outlist$AGAInv,
    outlist$G_list$G,
    env$A,
    dG_dlambda2,
    env$p_expansions,
    env$R_constraints,
    env$K,
    env$Xy,
    outlist$G_list$Ghalf,
    dGhalf2,
    GhalfXy_temp,
    env$parallel & env$parallel_matmult,
    env$cl,
    env$chunk_size,
    env$num_chunks,
    env$rem_chunks)
  dW_dlambda2 <- sum(unlist(
    .compute_dhat_diag(
      env$X,
      outlist$G_list$G,
      dG_dlambda2,
      env$A,
      outlist$AGAInv,
      env$K,
      env$p_expansions
    )$dh_diag
  ))

  ## -2 * residuals^T * X
  if (verbose) cat("        neg2tresidX\n")
  neg2tresidX <- Reduce("cbind",
                        matmult_block_diagonal(
                          lapply(outlist$residuals,
                                 function(r) -2 * t(r)),
                          env$X,
                          env$K,
                          env$parallel & env$parallel_matmult,
                          env$cl,
                          env$chunk_size,
                          env$num_chunks,
                          env$rem_chunks))

  ## Quotient rule for GCV_u derivative w.r.t. wiggle penalty
  dnumerator_dlambda1   <- c(neg2tresidX %**% dG_u_dlambda1_Xyr)
  ddenominator_dlambda1 <- 2 * (1 - gamma * outlist$mean_W) *
    (-gamma * dW_dlambda)
  dGCV_u_dlambda1 <- (dnumerator_dlambda1 * outlist$denominator -
                        outlist$numerator * ddenominator_dlambda1) /
    outlist$denom_sq

  dnumerator_dlambda2   <- c(neg2tresidX %**% dG_u_dlambda2_Xyr)
  ddenominator_dlambda2 <- 2 * (1 - gamma * outlist$mean_W) *
    (-gamma * dW_dlambda2)
  dGCV_u_dlambda2 <- (dnumerator_dlambda2 * outlist$denominator -
                        outlist$numerator * ddenominator_dlambda2) /
    outlist$denom_sq

  log_wiggle_gradient <- dGCV_u_dlambda1 * lambda_1
  log_flat_gradient <- dGCV_u_dlambda2 * lambda_2

  if (verbose) cat("        Gradient start \n")
  gradient <- cbind(c(log_wiggle_gradient,
                      log_flat_gradient))

  ## The remaining penalty gradients follow the wiggle derivative
  #  through the same trace-ratio approximation. These are already on the
  #  log-penalty scale because L_predictor_list/L_partition_list contain the
  #  current contribution of each penalty to Lambda, so no extra chain-rule
  #  multiplication by the raw penalty magnitude is needed here.
  if (env$unique_penalty_per_predictor) {
    predictor_penalties <- penalty_vec[grep("predictor", names(penalty_vec))]
    predictor_penalty_gradient <- sapply(
      seq_along(predictor_penalties), function(j) {
        mean(diag(outlist$L_predictor_list[[j]])) /
          mean(diag(outlist$Lambda)) *
          log_wiggle_gradient
      })
    gradient <- cbind(c(c(gradient), predictor_penalty_gradient))
  }

  ## Partition-specific penalty gradients
  if (env$unique_penalty_per_partition) {
    partition_penalties <- penalty_vec[grep("partition", names(penalty_vec))]
    partition_penalty_gradient <- sapply(
      seq_along(partition_penalties), function(j) {
        mean(diag(outlist$L_partition_list[[j]])) /
          mean(diag(outlist$Lambda +
                      outlist$L_partition_list[[j]])) *
          log_wiggle_gradient
      })
    gradient <- cbind(c(gradient, partition_penalty_gradient))
  }

  ## Add meta-penalty gradient
  regularizer <- .compute_meta_penalty_gradient(
    wiggle_penalty, penalty_vec, env$meta_penalty,
    env$unique_penalty_per_predictor,
    env$unique_penalty_per_partition)

  mp <- .compute_meta_penalty(wiggle_penalty, penalty_vec,
                              env$meta_penalty,
                              env$unique_penalty_per_predictor,
                              env$unique_penalty_per_partition)

  if (verbose) cat("        Gradient end \n")

  return(list(criterion_value = outlist$criterion_value,
              GCV_u    = outlist$GCV_u,
              gradient = gradient + regularizer,
              outlist  = outlist))
}


## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
# Subfunction: .damped_bfgs
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

#' Damped BFGS Optimizer for Penalty Tuning
#'
#' @description
#' Custom implementation of damped BFGS quasi-Newton optimization for
#' minimizing the selected tuning criterion. Uses step-size damping with
#' backtracking and Sherman-Morrison-Woodbury inverse Hessian updates.
#'
#' @param par Numeric vector; initial log-scale penalty parameters (first two
#'   elements are log(wiggle) and log(flat_ridge)).
#' @param log_penalty_vec Numeric vector; log-scale predictor/partition
#'   penalties appended to the optimization vector.
#' @param criterion_fxn Function; tuning-objective evaluation function with signature
#'   \code{function(par, log_penalty_vec, env, ...)}.
#' @param gr_fxn Function; gradient function with signature
#'   \code{function(par, log_penalty_vec, outlist, env, ...)}.
#' @param env List; tuning environment (passed through to
#'   \code{criterion_fxn} and \code{gr_fxn}).
#' @param tol Numeric; convergence tolerance for both criterion change and
#'   parameter change.
#' @param max_iter Integer; maximum number of BFGS iterations (default 100).
#' @param ... Additional arguments passed to fitting functions.
#'
#' @return List containing the best parameter vector found, the corresponding
#'   criterion value, and the number of iterations performed.
#'
#' @details
#' The optimizer uses the following strategy:
#' \enumerate{
#'   \item Iterations 1-2: steepest descent with damping.
#'   \item Iteration 3+: BFGS quasi-Newton with inverse Hessian approximation
#'     updated via the standard secant condition. Falls back to identity
#'     matrix when the update is numerically unstable.
#'   \item Step acceptance: Armijo-like criterion (accept if
#'     \eqn{\mathrm{GCV}_{u}^{(\mathrm{new})} \leq
#'     \mathrm{GCV}_{u}^{(\mathrm{old})}}).
#'   \item Backtracking: damping factor halved on rejection; terminates
#'     when damp < \eqn{2^{-10}} (early iterations) or \eqn{2^{-12}}
#'     (later iterations).
#' }
#'
#' @keywords internal
.damped_bfgs <- function(par,
                         log_penalty_vec,
                         criterion_fxn,
                         gr_fxn,
                         env,
                         tol,
                         max_iter = 100,
                         ...) {

  ## Initialize optimization parameters and storage
  lambda       <- cbind(par)
  old_lambda   <- lambda
  new_lambda   <- lambda
  n_params     <- length(lambda)
  Id           <- diag(n_params)
  initial_damp <- 0.5
  damp         <- initial_damp
  prev_gradient <- lambda * 0
  gradient      <- prev_gradient
  best_gradient <- prev_gradient
  old_gradient  <- prev_gradient
  outlist       <- NULL
  prev_outlist  <- NULL
  criterion_value <- Inf
  ridge         <- NULL
  dont_skip_gr  <- TRUE
  rho           <- NULL
  Inv           <- diag(n_params)
  best_lambda   <- lambda
  best_criterion_value <- criterion_value
  prev_lambda   <- old_lambda
  restart       <- TRUE

  ## Evaluate and retain the grid-search start point before taking any
  # optimizer step. Otherwise a bad first gradient step can displace the
  # best-known solution even when the initialization was already better.
  outlist <- criterion_fxn(c(lambda[1], lambda[2], log_penalty_vec),
                           log_penalty_vec, env, ...)
  criterion_value <- outlist$criterion_value
  best_lambda <- lambda
  best_criterion_value <- criterion_value
  prev_outlist <- outlist

  ## Main optimization loop
  for (iter in 1:max_iter) {

    ## Compute gradient if needed
    if (dont_skip_gr) {
      result   <- gr_fxn(c(lambda[1], lambda[2], log_penalty_vec),
                         log_penalty_vec, outlist, env, ...)
      gradient <- cbind(result$gradient)
      outlist  <- result$outlist
    }

    ## First two iterations: steepest descent
    if (iter <= 2) {
      new_lambda <- lambda - damp * gradient

      ## Reset to best if numerical issues
      if (any(!is.finite(exp(new_lambda))) ||
          any(is.nan(exp(new_lambda))) ||
          any(is.na(exp(new_lambda)))) {
        new_lambda <- best_lambda
      }

    } else {
      ## Add small ridge for stability if needed
      if (any(is.null(ridge))) ridge <- 1e-8 * diag(length(lambda))

      ## Update BFGS approximation
      if (dont_skip_gr) {

        ## Initial or restart BFGS approximation
        if (iter == 3 || restart) {
          diff_grad <- gradient - prev_gradient
          diff_lam  <- cbind(lambda - prev_lambda)
          denom     <- as.numeric(t(diff_grad) %**% diff_lam)
          Inv <- Inv +
            (denom + as.numeric(t(diff_grad) %**% Inv %**% diff_grad)) *
            (diff_lam %**% t(diff_lam)) / (denom^2) -
            (Inv %**% cbind(diff_grad) %**% t(diff_lam) +
               diff_lam %**% t(diff_grad) %**% Inv) / denom
          restart <- FALSE
        }

        ## Standard BFGS update
        if (iter > 3) {
          diff_grad <- gradient - prev_gradient
          diff_lam  <- cbind(lambda - prev_lambda)
          denom     <- as.numeric(t(diff_grad) %**% diff_lam)
          if (!is.na(denom) && abs(denom) > 1e-64) {
            rho   <- 1 / denom
            term1 <- Id - rho * (diff_lam %**% t(diff_grad))
            term2 <- Id - rho * (cbind(diff_grad) %**% t(diff_lam))
            Inv   <- term1 %**% Inv %**% term2 +
              rho * (diff_lam %**% t(diff_lam))
          } else {
            ## Reset if numerically unstable
            Inv     <- diag(length(gradient))
            restart <- TRUE
          }
        }
      }

      ## Compute BFGS step direction
      new_lambda <- lambda - damp * Inv %**% cbind(gradient)

      ## Reset to best if numerical issues
      if (any(!is.finite(exp(new_lambda))) ||
          any(is.nan(exp(new_lambda))) ||
          any(is.na(exp(new_lambda)))) {
        new_lambda <- best_lambda
      }
    }

    ## Evaluate GCV at new point
    if (any(is.na(new_lambda))) {
      ## Backtrack if invalid step
      lambda       <- old_lambda
      gradient     <- old_gradient
      dont_skip_gr <- FALSE
      damp         <- damp / 2
      if (damp < 2^-12 && iter > 9) {
        return(list(par        = c(best_lambda),
                    criterion_value = best_criterion_value,
                    gcv_u      = best_criterion_value,
                    iterations = iter))
      }
      next
    } else {
      prev_outlist <- outlist
      outlist      <- criterion_fxn(c(new_lambda[1], new_lambda[2],
                                      log_penalty_vec),
                                    log_penalty_vec, env, ...)
      new_criterion_value <- outlist$criterion_value
      if (any(is.na(new_criterion_value))) {
        new_criterion_value <- criterion_value
      }
    }

    ## Accept step if improvement or early iterations
    if (new_criterion_value <= criterion_value || iter <= 2) {
      ## Update solution history
      old_gradient  <- prev_gradient
      prev_gradient <- gradient
      dont_skip_gr  <- TRUE
      prev_outlist  <- outlist
      old_lambda    <- prev_lambda
      prev_lambda   <- lambda
      lambda        <- new_lambda
      damp          <- 1

      ## Track best solution globally, not just relative to the most recent
      # accepted iterate. Early damped steps can be exploratory and may be
      # accepted even when they are worse than the grid-search start.
      if (new_criterion_value <= best_criterion_value) {
        best_criterion_value <- new_criterion_value
        best_lambda <- lambda
      }

      ## Check convergence
      if (((abs(new_criterion_value - criterion_value) < tol) ||
           (max(abs(lambda - prev_lambda)) < tol)) &&
          (iter > 9)) {
        return(list(par        = c(best_lambda),
                    criterion_value = best_criterion_value,
                    gcv_u      = best_criterion_value,
                    iterations = iter))
      }
      criterion_value <- new_criterion_value
    } else {
      ## Reject step and backtrack
      dont_skip_gr <- FALSE
      outlist      <- prev_outlist
      damp         <- damp / 2
      if (damp < 2^(-10) && iter > 0) {
        return(list(par        = c(best_lambda),
                    criterion_value = best_criterion_value,
                    gcv_u      = best_criterion_value,
                    iterations = iter))
      }
    }
  }

  return(list(par        = c(best_lambda),
              criterion_value = best_criterion_value,
              gcv_u      = best_criterion_value,
              iterations = iter))
}


## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
# Subfunction: .tune_grid_search
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

#' Grid Search Initialization for Penalty Tuning
#'
#' @description
#' Evaluates the selected tuning criterion over a grid of initial wiggle and ridge
#' penalty values to find a good starting point for BFGS optimization.
#'
#' @param log_initial_wiggle Numeric vector; log-scale candidate values for
#'   the wiggle penalty.
#' @param log_initial_flat Numeric vector; log-scale candidate values for
#'   the flat ridge penalty.
#' @param log_penalty_vec Numeric vector; log-scale predictor/partition
#'   penalties.
#' @param criterion_fxn Function; tuning-objective evaluation function.
#' @param env List; tuning environment.
#' @param include_warnings Logical; whether to print warnings on failure.
#' @param ... Additional arguments passed to \code{criterion_fxn}.
#'
#' @return Numeric vector of length 2; the best (log_wiggle, log_flat) found.
#'
#' @keywords internal
.tune_grid_search <- function(log_initial_wiggle,
                              log_initial_flat,
                              log_penalty_vec,
                              criterion_fxn,
                              env,
                              include_warnings,
                              ...) {

  ## Create all combinations of the grid values
  initial_grid <- expand.grid(wiggle = log_initial_wiggle,
                              flat   = log_initial_flat)

  ## Function to safely evaluate the selected criterion
  safe_gcvu <- function(par) {
    tryCatch({
      result <- criterion_fxn(c(unlist(par), log_penalty_vec),
                              log_penalty_vec, env, ...)$criterion_value
      if (is.na(result) || is.nan(result)) {
        return(Inf)
      }
      return(result)
    }, error = function(e) {
      if (include_warnings) {
        return(Inf)
      }
      return(Inf)
    })
  }

  ## Evaluate the criterion for each grid point
  gcv_values <- apply(initial_grid, 1, safe_gcvu)
  bads <- which(is.na(gcv_values) |
                  is.nan(gcv_values) |
                  !is.finite(gcv_values))

  if (length(bads) == length(gcv_values)) {
    stop("All tuning criteria for the initial grid were computed as NA,",
         " NaN, or non-finite: check your data for corrupt or missing values,",
         " try changing initial tuning grid, or try manual tuning instead.",
         " If you are setting no_intercept = TRUE, try experimenting with",
         " standardize_response = FALSE vs. TRUE.")
  } else if (length(bads) > 0) {
    gcv_values <- gcv_values[-c(bads)]
    initial_grid <- initial_grid[-c(bads), , drop = FALSE]
  }

  ## Find the best starting point
  best_index <- which.min(gcv_values)[1]
  best_start <- as.numeric(initial_grid[best_index, ])

  return(best_start)
}


## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
# Subfunction: .compute_tuning_delta
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

#' Compute Pseudocount Delta for Link Function Stabilization
#'
#' @description
#' Determines the pseudocount \eqn{\delta} used to stabilize link function
#' transformations during penalty tuning. For identity link or when
#' the response is naturally in the domain of the link function, returns 0.
#' Otherwise, finds the \eqn{\delta} that makes the transformed response
#' distribution most closely approximate a t-distribution.
#'
#' @param family GLM family object.
#' @param unl_y Numeric vector; unlisted response values (concatenated across
#'   partitions).
#' @param N_obs Integer; total sample size.
#' @param observation_weights List or NULL; observation weights by partition.
#' @param opt Logical; whether optimization is being performed.
#'
#' @return Numeric scalar; the pseudocount \eqn{\delta \geq 0}.
#'
#' @keywords internal
.compute_tuning_delta <- function(family, unl_y, N_obs,
                                  observation_weights, opt) {

  ## Check if delta is needed at all
  if (!opt) return(0)

  ## Identity link: no pseudocount needed
  if (paste0(family)[2] == "identity") return(0)

  ## Log link with all positive y: no pseudocount needed
  if (paste0(family)[2] == "log" && (min(unl_y) > 0)) return(0)

  ## Inverse/1/mu^2 link with no zeros: no pseudocount needed
  if (any(paste0(family)[2] %in% c("inverse", "1/mu^2")) &&
      (!any(unl_y == 0))) return(0)

  ## Logit link with no boundary values: no pseudocount needed
  if (paste0(family)[2] == "logit" && (!any(unl_y %in% c(0, 1)))) return(0)

  ## Otherwise, find optimal pseudocount via Brent optimization
  t_quants <- qt((seq(0, 1, len = N_obs + 2))[-c(1, N_obs + 2)], df = N_obs - 1)
  delta <- stats::optim(
    1 / 16,
    fn = function(par) {
      y_delta <- std(sort(family$linkfun((unl_y + par) / (1 + 2 * par))))
      if (length(unlist(observation_weights)) == length(unl_y)) {
        mean(abs(y_delta - t_quants) * unlist(observation_weights))
      } else {
        mean(abs(y_delta - t_quants))
      }
    },
    method = "Brent",
    lower = 1e-64,
    upper = 1
  )$par

  return(delta)
}


## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
# Subfunction: .build_tuning_env
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

#' Build Tuning Environment
#'
#' @description
#' Assembles a named list containing all pre-computed objects and configuration
#' needed by the tuning-objective evaluation and gradient functions during
#' penalty tuning.
#' This avoids deep nesting of closures and makes the dependencies explicit.
#'
#' @details
#' In addition to the standard tuning arguments, the environment stores two
#' pre-computed blockfit dispatch items:
#' \describe{
#'   \item{use_blockfit}{Logical; TRUE when \code{blockfit} is enabled,
#'     \code{flat_cols} is non-empty, and \code{K > 0}. Mirrors the dispatch
#'     logic in \code{lgspline.fit} so that the same fitting path is used
#'     during tuning as during the final fit.}
#'   \item{flat_cols}{Integer vector; column indices of non-interactive linear
#'     terms derived from \code{just_linear_without_interactions} and
#'     \code{colnm_expansions}. Pre-computed once here rather than re-derived
#'     at every tuning-objective evaluation.}
#' }
#'
#' @param flat_cols Integer vector; pre-computed flat column indices (passed
#'   in from \code{tune_Lambda} to avoid recomputation).
#' @param use_blockfit Logical; pre-computed dispatch flag (passed in from
#'   \code{tune_Lambda}).
#' @param gcv_gamma Numeric scalar, at least 1; modified-GCV multiplier for the
#'   effective-degrees-of-freedom term.
#'
#' @return Named list (the "tuning environment").
#'
#' @keywords internal
.build_tuning_env <- function(y, X, X_gram, Xy,
                              smoothing_spline_penalty,
                              A, R_constraints, K, p_expansions, N_obs,
                              custom_penalty_mat,
                              colnm_expansions,
                              unique_penalty_per_predictor,
                              unique_penalty_per_partition,
                              meta_penalty,
                              family, delta,
                              order_list,
                              observation_weights,
                              homogenous_weights,
                              parallel, parallel_eigen,
                              parallel_trace, parallel_aga,
                              parallel_matmult,
                              parallel_unconstrained,
                              cl, chunk_size,
                              num_chunks, rem_chunks,
                              unconstrained_fit_fxn,
                              keep_weighted_Lambda,
                              iterate,
                              qp_score_function, quadprog,
                              qp_Amat, qp_bvec, qp_meq,
                              tol, sd_y, tuning_criterion, gcv_gamma,
                              constraint_value_vectors,
                              glm_weight_function,
                              schur_correction_function,
                              need_dispersion_for_estimation,
                              dispersion_function,
                              blockfit,
                              just_linear_without_interactions,
                              Vhalf, VhalfInv,
                              verbose, include_warnings,
                              flat_cols, use_blockfit) {

  list(
    y                             = y,
    X                             = X,
    X_gram                        = X_gram,
    Xy                            = Xy,
    smoothing_spline_penalty      = smoothing_spline_penalty,
    A                             = A,
    R_constraints                 = R_constraints,
    K                             = K,
    p_expansions                  = p_expansions,
    N_obs                         = N_obs,
    custom_penalty_mat            = custom_penalty_mat,
    colnm_expansions              = colnm_expansions,
    unique_penalty_per_predictor  = unique_penalty_per_predictor,
    unique_penalty_per_partition  = unique_penalty_per_partition,
    meta_penalty                  = meta_penalty,
    family                        = family,
    delta                         = delta,
    order_list                    = order_list,
    observation_weights           = observation_weights,
    homogenous_weights            = homogenous_weights,
    parallel                      = parallel,
    parallel_eigen                = parallel_eigen,
    parallel_trace                = parallel_trace,
    parallel_aga                  = parallel_aga,
    parallel_matmult              = parallel_matmult,
    parallel_unconstrained        = parallel_unconstrained,
    cl                            = cl,
    chunk_size                    = chunk_size,
    num_chunks                    = num_chunks,
    rem_chunks                    = rem_chunks,
    unconstrained_fit_fxn         = unconstrained_fit_fxn,
    keep_weighted_Lambda          = keep_weighted_Lambda,
    iterate                       = iterate,
    qp_score_function             = qp_score_function,
    quadprog                      = quadprog,
    qp_Amat                       = qp_Amat,
    qp_bvec                       = qp_bvec,
    qp_meq                        = qp_meq,
    tol                           = tol,
    sd_y                          = sd_y,
    tuning_criterion              = tuning_criterion,
    gcv_gamma                     = gcv_gamma,
    constraint_value_vectors      = constraint_value_vectors,
    glm_weight_function           = glm_weight_function,
    schur_correction_function     = schur_correction_function,
    need_dispersion_for_estimation = need_dispersion_for_estimation,
    dispersion_function           = dispersion_function,
    blockfit                      = blockfit,
    just_linear_without_interactions = just_linear_without_interactions,
    Vhalf                         = Vhalf,
    VhalfInv                      = VhalfInv,
    verbose                       = verbose,
    include_warnings              = include_warnings,
    flat_cols                     = flat_cols,
    use_blockfit                  = use_blockfit
  )
}


## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
# Main function: tune_Lambda (exported)
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

#' Tune Smoothing and Ridge Penalties via Leave-One-Out or Generalized Cross-Validation
#'
#' @description
#' Optimizes smoothing spline and ridge regression penalties by minimizing the
#' selected tuning criterion, either exact leave-one-out on the transformed
#' tuning problem or modified GCV. Uses BFGS optimization with analytical
#' gradients or finite differences. This is the top-level entry point that
#' orchestrates grid search initialization and quasi-Newton optimization via
#' internal subfunctions.
#'
#' @param y List; response vectors by partition.
#' @param X List; design matrices by partition.
#' @param X_gram List; Gram matrices by partition.
#' @param smoothing_spline_penalty Matrix; integrated squared second derivative
#'   penalty.
#' @param A Matrix; smoothness constraints at knots.
#' @param K Integer; number of interior knots in 1-D, number of partitions - 1
#'   in higher dimensions.
#' @param p_expansions Integer; columns per partition.
#' @param N_obs Integer; total sample size.
#' @param opt Logical; TRUE to optimize penalties, FALSE to use initial values.
#' @param use_custom_bfgs Logical; TRUE for analytic gradient BFGS as natively
#'   implemented, FALSE for finite differences as implemented by
#'   \code{stats::optim()}. The analytic path is typically faster; however, for
#'   exact LOO tuning the leverage derivative can be numerically delicate in
#'   some problems, in which case \code{use_custom_bfgs = FALSE} remains a
#'   conservative fallback.
#' @param C Matrix; polynomial expansion matrix (used for initialization).
#'   This is the monomial expansion design used to derive starting values and
#'   is not the inequality-constraint matrix sometimes denoted by
#'   \eqn{\mathbf{C}} elsewhere in the package documentation.
#' @param colnm_expansions Character vector; column names of the expansion
#'   matrix.
#' @param wiggle_penalty,flat_ridge_penalty Fixed penalty values if provided.
#' @param initial_wiggle,initial_flat Numeric vectors; candidate values for grid
#'   search initialization on the raw (non-negative) scale. Converted to log
#'   scale internally for optimization.
#' @param unique_penalty_per_predictor,unique_penalty_per_partition Logical;
#'   allow predictor/partition-specific penalties.
#' @param penalty_vec Numeric vector; initial values for predictor/partition
#'   penalties on the raw (non-negative) scale. Converted to log scale
#'   internally for optimization. Use \code{c()} when no per-predictor or
#'   per-partition penalties are needed.
#' @param meta_penalty The "meta" ridge penalty, a regularization for
#'   predictor/partition penalties to pull them towards 1 on the raw scale.
#' @param family GLM family with optional custom tuning loss.
#' @param unconstrained_fit_fxn Function for unconstrained fitting.
#' @param keep_weighted_Lambda,iterate Logical controlling GLM fitting.
#' @param qp_score_function,quadprog,qp_Amat,qp_bvec,qp_meq Quadratic
#'   programming parameters (see arguments of \code{\link[lgspline]{lgspline}}).
#' @param tol Numeric; convergence tolerance.
#' @param sd_y,delta Response standardization parameters.
#' @param tuning_criterion Character scalar selecting the tuning objective:
#'   \code{"loo"} for exact leave-one-out on the transformed tuning problem or
#'   \code{"gcv"} for generalized cross-validation. The LOO criterion itself is
#'   still evaluated exactly on that transformed problem; the caveat is that
#'   its fully analytic leverage derivative can be numerically sensitive in some
#'   datasets. For very large samples, \code{"gcv"} is often the more practical
#'   choice; as a rough guideline, it is recommended once the sample size is
#'   above about 250,000.
#' @param gcv_gamma Numeric scalar, at least 1, used only when
#'   \code{tuning_criterion = "gcv"}. Multiplies the effective degrees of
#'   freedom in the GCV denominator.
#' @param constraint_value_vectors List; constraint values.
#' @param parallel Logical; enable parallel computation.
#' @param parallel_eigen,parallel_trace,parallel_aga Logical; specific parallel
#'   flags.
#' @param parallel_matmult,parallel_unconstrained Logical; specific parallel
#'   flags.
#' @param cl Parallel cluster object.
#' @param chunk_size,num_chunks,rem_chunks Integer; parallel computation
#'   parameters.
#' @param shared_env Environment; shared variables exported to cluster workers.
#' @param custom_penalty_mat Optional custom penalty matrix.
#' @param order_list List; observation ordering by partition.
#' @param glm_weight_function,schur_correction_function Functions for GLM
#'   weights and corrections.
#' @param need_dispersion_for_estimation,dispersion_function Control dispersion
#'   estimation.
#' @param observation_weights Optional observation weights.
#' @param homogenous_weights Logical; TRUE if all weights equal.
#' @param blockfit Logical; when TRUE, the backfitting block decomposition
#'   (\code{blockfit_solve}) is used in place of \code{get_B} for the
#'   coefficient estimation step at each tuning-objective evaluation, provided
#'   \code{just_linear_without_interactions} is non-empty and \code{K > 0}.
#'   The dispatch condition is identical to that in \code{lgspline.fit} so
#'   that penalties are tuned under the same estimator that will be used for
#'   the final fit. Falls back to \code{get_B} automatically on failure.
#' @param just_linear_without_interactions Numeric; vector of columns for
#'   non-spline effects without interactions.
#' @param Vhalf,VhalfInv Square root and inverse square root correlation
#'   structure matrices. These are passed through to the coefficient-estimation
#'   step used inside each tuning-objective evaluation, so any dense or
#'   Woodbury-accelerated
#'   correlated solve is the same one used for the final fitted model.
#' @param verbose Logical; print progress.
#' @param include_warnings Logical; print warnings/try-errors.
#' @param ... Additional arguments passed to fitting functions.
#'
#' @return List containing:
#' \describe{
#'   \item{Lambda}{Final combined penalty matrix.}
#'   \item{flat_ridge_penalty}{Optimized ridge penalty.}
#'   \item{wiggle_penalty}{Optimized smoothing penalty.}
#'   \item{other_penalties}{Optimized predictor/partition penalties.}
#'   \item{L_predictor_list}{Predictor-specific penalty matrices.}
#'   \item{L_partition_list}{Partition-specific penalty matrices.}
#' }
#'
#' @details
#' The tuning procedure consists of the following steps:
#' \enumerate{
#'   \item \strong{Preprocessing}: Convert raw-scale penalties to log scale,
#'     compute cross-products, determine pseudocount delta, ensure constraint
#'     matrix compatibility.
#'   \item \strong{Blockfit dispatch}: Pre-compute \code{flat_cols} and the
#'     \code{use_blockfit} flag so that every tuning-objective evaluation uses the same
#'     coefficient estimator as the final fit. \code{flat_cols} are identified
#'     by matching column names against
#'     \code{paste0("_", just_linear_without_interactions, "_")} in
#'     \code{colnm_expansions}.
#'   \item \strong{Grid search}: Evaluate the selected tuning criterion over a grid of
#'     (wiggle, ridge) penalty candidates to find a good starting point
#'     (see \code{.tune_grid_search}).
#'   \item \strong{BFGS optimization}: Minimize the selected tuning criterion via either the custom
#'     damped BFGS with closed-form gradients (see \code{.damped_bfgs},
#'     \code{.compute_gcvu_gradient}, \code{.compute_loocv_gradient}) or base R's \code{stats::optim} with
#'     finite-difference gradients. The GCV gradient is differentiated directly
#'     in the shared wiggle and flat directions; for LOO, the same direct
#'     differentiation is used, but the leverage derivative may be numerically
#'     delicate in some problems and can motivate the finite-difference
#'     fallback.
#'   \item \strong{Criterion choice}: \code{"loo"} uses exact per-observation
#'     leverages on the transformed tuning problem, while \code{"gcv"} uses the
#'     existing trace-based denominator with optional \code{gcv_gamma}
#'     inflation.
#'   \item \strong{Sample-size adjustment}: After optimization, divide the tuned
#'     penalties by \eqn{((N+2)/(N-2))^{2}} (equivalently multiply by
#'     \eqn{((N-2)/(N+2))^{2}}) for both GCV- and LOO-based tuning so that the
#'     final penalties are decreased at small sample sizes.
#'   \item \strong{Final Lambda}: Compute the final penalty matrix from
#'     optimized parameters via \code{compute_Lambda}.
#' }
#'
#' Parameterization: initial penalty values are accepted on the raw
#' (non-negative) scale and converted to natural log-scale internally,
#' i.e. raw_penalty = exp(theta), so that raw penalties are always positive.
#' The chain rule factor d(exp(theta))/d(theta) = exp(theta) = raw_penalty.
#'
#' For predictor- and partition-specific penalties, the current implementation
#' still uses the existing low-cost ratio approximation rather than a full
#' direct derivative in each extra penalty direction. Thus the most accurate
#' closed-form tuning gradients are presently the shared wiggle and flat
#' directions.
#'
#' The resulting penalty matrix follows the same decomposition used throughout
#' the paper and package documentation:
#' \deqn{
#'   \boldsymbol{\Lambda}
#'   =
#'   \lambda_{w}\mathbf{L}_{1}
#'   +
#'   \lambda_{r}\mathbf{L}_{2}
#'   +
#'   \sum_{j}\nu_{j}\mathbf{L}^{(\mathrm{pred})}_{j}
#'   +
#'   \sum_{k}\tau_{k}\mathbf{L}^{(\mathrm{part})}_{k},
#' }
#' where the predictor- and partition-specific sums are included only when the
#' corresponding options are active. Internally these components are returned as
#' \code{L1}, \code{L2}, \code{L_predictor_list}, and \code{L_partition_list}.
#'
#' If correlation structure inputs are supplied, each tuning-objective evaluation calls the
#' same constrained correlated solver used by the final model fit. In the
#' structured-correlation case, the low-rank Woodbury correction described in
#' \code{lgspline-details} is therefore inherited automatically through
#' \code{get_B}; no separate tuning-specific notation is introduced here.
#'
#' @seealso
#' \itemize{
#'   \item \code{\link[stats]{optim}} for Hessian-free optimization
#'   \item \code{\link{compute_Lambda}} for penalty matrix construction
#'   \item \code{\link{compute_G_eigen}} for eigendecomposition of penalized
#'     Gram matrices
#'   \item \code{\link{get_B}} for constrained coefficient estimation
#'   \item \code{\link{blockfit_solve}} for the backfitting block-decomposition
#'     estimator used when \code{blockfit = TRUE}
#' }
#'
#' @examples
#' \dontrun{
#' ## ## Example 1: Basic usage within lgspline ## ## ## ## ## ## ## ## ## ## ##
#' ## tune_Lambda is called internally by lgspline; direct calls are
#' #  for advanced users. Here we verify that the refactored version
#' #  produces identical output to the original.
#'
#' set.seed(42)
#' t <- runif(200, -5, 5)
#' y <- sin(t) + rnorm(200, 0, 0.5)
#'
#' ## Fit with automatic penalty tuning (calls tune_Lambda internally)
#' fit1 <- lgspline(t, y, K = 3)
#' cat("Wiggle penalty:", fit1$penalties$wiggle_penalty, "\n")
#' cat("Ridge penalty:", fit1$penalties$flat_ridge_penalty, "\n")
#' cat("Trace (edf):", fit1$trace_XUGX, "\n")
#'
#' ## ## Example 2: Fixed penalties (no tuning) ## ## ## ## ## ## ## ## ## ## ##
#' fit2 <- lgspline(t, y, K = 3, opt = FALSE,
#'                  wiggle_penalty = 1e-4,
#'                  flat_ridge_penalty = 0.1)
#' cat("Fixed wiggle:", fit2$penalties$wiggle_penalty, "\n")
#'
#' ## ## Example 3: blockfit path #### ## ## ## ## ## ## ## ## ## ## ## ## ## ##
#' ## When blockfit = TRUE and just_linear_without_interactions is non-empty,
#' #  tune_Lambda dispatches to blockfit_solve at each GCV evaluation,
#' #  ensuring penalties are tuned under the same estimator used in the final
#' #  fit. Verify that tuned penalties are consistent across both paths.
#'
#' set.seed(7)
#' n  <- 300
#' x1 <- runif(n, 0, 5)
#' x2 <- rnorm(n)
#' y2 <- sin(x1) + 0.5 * x2 + rnorm(n, 0, 0.3)
#' df <- data.frame(x1 = x1, x2 = x2)
#'
#' ## blockfit = TRUE uses blockfit_solve during tuning
#' fit_bf  <- lgspline(df, y2, K = 2, blockfit = TRUE,
#'                     just_linear_without_interactions = 2)
#' ## blockfit = FALSE uses get_B during tuning (original path)
#' fit_std <- lgspline(df, y2, K = 2, blockfit = FALSE,
#'                     just_linear_without_interactions = 2)
#'
#' cat("blockfit wiggle  :", fit_bf$penalties$wiggle_penalty, "\n")
#' cat("standard wiggle  :", fit_std$penalties$wiggle_penalty, "\n")
#' ## Penalties may differ slightly; predictions should be close.
#' cat("Max pred diff:", max(abs(fit_bf$ytilde - fit_std$ytilde)), "\n")
#'
#' ## ## Example 4: Verify refactored subfunctions ## ## ## ## ## ## ## ## ## ##
#' ## The internal .compute_meta_penalty should match hand calculation
#' mp <- lgspline:::.compute_meta_penalty(
#'   wiggle_penalty = 0.5,
#'   penalty_vec = c(predictor1 = 1.2, partition1 = 0.8),
#'   meta_penalty_coef = 1e-8,
#'   unique_penalty_per_predictor = TRUE,
#'   unique_penalty_per_partition = TRUE
#' )
#' expected <- 0.5 * 1e-8 * ((1.2 - 1)^2 + (0.8 - 1)^2) +
#'             0.5 * 1e-32 * (0.5 - 1)^2
#' stopifnot(abs(mp - expected) < 1e-20)
#' cat("Meta-penalty check passed.\n")
#'
#' ## ## Example 5: Verify gradient of meta-penalty ## ## ## ## ## ## ## ## ## #
#' gr <- lgspline:::.compute_meta_penalty_gradient(
#'   wiggle_penalty = 2.0,
#'   penalty_vec = c(predictor1 = 1.5),
#'   meta_penalty_coef = 1e-8,
#'   unique_penalty_per_predictor = TRUE,
#'   unique_penalty_per_partition = FALSE
#' )
#' # gr[1] should be 1e-32 * (2 - 1) * 2 = 2e-32
#' # gr[2] should be 0
#' # gr[3] should be 1e-8 * (1.5 - 1) * 1.5 = 7.5e-9
#' stopifnot(abs(gr[1] - 2e-32) < 1e-40)
#' stopifnot(gr[2] == 0)
#' stopifnot(abs(gr[3] - 7.5e-9) < 1e-17)
#' cat("Meta-penalty gradient check passed.\n")
#'
#' ## ## Example 6: Residual computation paths ## ##
#' ## Identity link ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
#' r1 <- lgspline:::.compute_tuning_residuals(
#'   y = list(c(1, 2, 3)),
#'   preds = list(c(1.1, 1.9, 3.2)),
#'   delta = 0,
#'   family = gaussian(),
#'   observation_weights = list(NULL),
#'   K = 0,
#'   order_list = list(1:3)
#' )
#' stopifnot(max(abs(r1[[1]] - c(-0.1, 0.1, -0.2))) < 1e-10)
#' cat("Residual check passed.\n")
#' }
#'
#' @keywords internal
#' @export
tune_Lambda <- function(
    y,
    X,
    X_gram,
    smoothing_spline_penalty,
    A,
    K,
    p_expansions,
    N_obs,
    opt,
    use_custom_bfgs,
    C,
    colnm_expansions,
    wiggle_penalty,
    flat_ridge_penalty,
    initial_wiggle,
    initial_flat,
    unique_penalty_per_predictor,
    unique_penalty_per_partition,
    penalty_vec,
    meta_penalty,
    family,
    unconstrained_fit_fxn,
    keep_weighted_Lambda,
    iterate,
    qp_score_function, quadprog, qp_Amat, qp_bvec, qp_meq,
    tol,
    sd_y,
    delta,
    tuning_criterion,
    gcv_gamma,
    constraint_value_vectors,
    parallel,
    parallel_eigen,
    parallel_trace,
    parallel_aga,
    parallel_matmult,
    parallel_unconstrained,
    cl,
    chunk_size,
    num_chunks,
    rem_chunks,
    shared_env,
    custom_penalty_mat,
    order_list,
    glm_weight_function,
    schur_correction_function,
    need_dispersion_for_estimation,
    dispersion_function,
    observation_weights,
    homogenous_weights,
    blockfit,
    just_linear_without_interactions,
    Vhalf,
    VhalfInv,
    verbose,
    include_warnings,
    ...
) {

  if (verbose) cat("    Starting tuning\n")

  tuning_criterion <- match.arg(tuning_criterion, c("loo", "gcv"))

  if (tuning_criterion == "gcv" &&
      (!is.numeric(gcv_gamma) || length(gcv_gamma) != 1 ||
       !is.finite(gcv_gamma) || gcv_gamma < 1)) {
    stop("gcv_gamma must be a finite numeric scalar >= 1. ",
         "Set gcv_gamma = 1 to recover ordinary GCV.", call. = FALSE)
  }

  ## Step 1: Convert raw-scale inputs to log scale ## ##
  log_initial_wiggle <- log(initial_wiggle)
  log_initial_flat   <- log(initial_flat)
  if (length(penalty_vec) > 0) {
    log_penalty_vec <- log(penalty_vec)
  } else {
    log_penalty_vec <- c()
  }

  if (verbose) cat("    Xy\n")

  ## Step 2: Precompute cross-products and constants
  sN_obs <- sqrt(N_obs)
  Xy  <- vectorproduct_block_diagonal(X, y, K)
  Xyr <- Reduce("rbind", Xy)
  R_constraints <- ncol(A)
  Xt  <- lapply(X, t)
  unl_y <- unlist(y)

  ## Step 3: Compute pseudocount delta
  if (verbose) cat("    Getting pseudocount delta\n")

  if (is.null(delta) &&
      any(is.null(family$custom_dev.resids)) &&
      opt) {
    delta <- .compute_tuning_delta(family, unl_y, N_obs,
                                   observation_weights, opt)
  }
  if (is.null(delta)) delta <- 0

  if (verbose) cat("    GCV, gradient, and BFGS Function prep\n")

  ## Step 4: ensure the constraint matrix is always conformable
  if (any(is.null(A))) {
    A   <- cbind(rep(0, (K + 1) * p_expansions))
    A   <- cbind(A, A)
    R_constraints <- 2
  }

  ## Step 5: pre-compute the blockfit dispatch used during tuning
  #  by matching the flat linear columns the same way lgspline.fit does.
  flat_cols <- c()
  if (length(just_linear_without_interactions) > 0) {
    flat_cols <- which(colnm_expansions %in%
                         paste0("_", just_linear_without_interactions, "_"))
  }
  use_blockfit <- blockfit && length(flat_cols) > 0 && K > 0

  if (verbose && use_blockfit) {
    cat("    blockfit dispatch active during tuning (",
        length(flat_cols), "flat cols )\n")
  }

  ## Step 6: build the tuning environment
  env <- .build_tuning_env(
    y = y, X = X, X_gram = X_gram, Xy = Xy,
    smoothing_spline_penalty = smoothing_spline_penalty,
    A = A, R_constraints = R_constraints, K = K,
    p_expansions = p_expansions, N_obs = N_obs,
    custom_penalty_mat = custom_penalty_mat,
    colnm_expansions = colnm_expansions,
    unique_penalty_per_predictor = unique_penalty_per_predictor,
    unique_penalty_per_partition = unique_penalty_per_partition,
    meta_penalty = meta_penalty,
    family = family, delta = delta,
    order_list = order_list,
    observation_weights = observation_weights,
    homogenous_weights = homogenous_weights,
    parallel = parallel, parallel_eigen = parallel_eigen,
    parallel_trace = parallel_trace, parallel_aga = parallel_aga,
    parallel_matmult = parallel_matmult,
    parallel_unconstrained = parallel_unconstrained,
    cl = cl, chunk_size = chunk_size,
    num_chunks = num_chunks, rem_chunks = rem_chunks,
    unconstrained_fit_fxn = unconstrained_fit_fxn,
    keep_weighted_Lambda = keep_weighted_Lambda,
    iterate = iterate,
    qp_score_function = qp_score_function, quadprog = quadprog,
    qp_Amat = qp_Amat, qp_bvec = qp_bvec, qp_meq = qp_meq,
    tol = tol, sd_y = sd_y,
    tuning_criterion = tuning_criterion,
    gcv_gamma = gcv_gamma,
    constraint_value_vectors = constraint_value_vectors,
    glm_weight_function = glm_weight_function,
    schur_correction_function = schur_correction_function,
    need_dispersion_for_estimation = need_dispersion_for_estimation,
    dispersion_function = dispersion_function,
    blockfit = blockfit,
    just_linear_without_interactions = just_linear_without_interactions,
    Vhalf = Vhalf, VhalfInv = VhalfInv,
    verbose = verbose, include_warnings = include_warnings,
    flat_cols = flat_cols, use_blockfit = use_blockfit
  )

  ## Step 7: Optimization (if requested)
  if (opt) {
    criterion_fxn <- if (tuning_criterion == "loo") {
      .compute_loocv
    } else {
      .compute_gcvu
    }
    gr_fxn <- if (tuning_criterion == "loo") {
      .compute_loocv_gradient
    } else {
      .compute_gcvu_gradient
    }

    if (verbose) cat("    Starting grid search for initialization\n")

    ## Grid search for good starting values
    best_start <- .tune_grid_search(log_initial_wiggle,
                                    log_initial_flat,
                                    log_penalty_vec,
                                    criterion_fxn,
                                    env,
                                    include_warnings,
                                    ...)

    if (verbose) {
      cat("    Finished grid evaluations\n")
      cat("    Best from grid search: ", cbind(c(best_start)), "\n")
    }

    ## Run optimization
    if (use_custom_bfgs) {
      ## Custom damped BFGS with closed-form gradients
      res <- withCallingHandlers(
        try(.damped_bfgs(c(best_start, log_penalty_vec),
                         log_penalty_vec,
                         criterion_fxn,
                         gr_fxn,
                         env,
                         tol,
                         max_iter = 100,
                         ...),
            silent = TRUE),
        warning = function(w) {
          if (include_warnings) warning(w) else invokeRestart("muffleWarning")
        },
        message = function(m) {
          if (include_warnings) message(m) else invokeRestart("muffleMessage")
        }
      )
      if (any(inherits(res, "try-error"))) {
        if (include_warnings) print(res)
        if (include_warnings) {
          warning("Custom BFGS implementation failed. Try use_custom_bfgs =",
                  " FALSE, or manual tuning. Resorting to best as selected",
                  " from grid search.")
        }
        par <- c(best_start, log_penalty_vec)
      } else {
        par <- res$par
      }

    } else {
      ## Base R BFGS with finite-difference gradients
      res <- withCallingHandlers(
        try({
          optim(c(best_start, log_penalty_vec),
                fn = function(par) {
                  criterion_fxn(par, log_penalty_vec, env, ...)$criterion_value
                },
                method = "BFGS")
        }, silent = TRUE),
        warning = function(w) {
          if (include_warnings) warning(w) else invokeRestart("muffleWarning")
        },
        message = function(m) {
          if (include_warnings) message(m) else invokeRestart("muffleMessage")
        }
      )
      if (any(inherits(res, "try-error"))) {
        if (include_warnings) print(res)
        if (include_warnings) {
          warning("Base R BFGS failed. Try use_custom_bfgs = TRUE, or manual",
                  " tuning. Resorting to best as selected from grid search.")
        }
        par <- c(best_start, log_penalty_vec)
      } else {
        par <- res$par
      }
    }

    if (verbose) cat("    Finished tuning penalties\n")

    ## Decrease tuned penalties at small sample sizes by dividing the
    #  optimizer solution by ((N+2)/(N-2))^2 for both tuning criteria.
    sample_size_adjustment <- ((N_obs - 2) / (N_obs + 2))^2
    wiggle_penalty     <- exp(par[1]) * sample_size_adjustment
    flat_ridge_penalty <- exp(par[2]) * sample_size_adjustment

    ## Update penalty vec for predictor- and partition-specific penalties
    if (length(log_penalty_vec) > 0) {
      penalty_vec[1:length(penalty_vec)] <-
        exp(c(par[-c(1, 2)])) * sample_size_adjustment
    }

  } else if (length(penalty_vec) == 0) {
    ## No per-predictor or per-partition penalties; nothing to do
  }
  ## When !opt and length(penalty_vec) > 0, penalty_vec is already on raw
  #  scale from input and used as-is

  ## Step 8: Compute final penalty matrix
  if (verbose) cat("    Final update\n")

  Lambda_list <- compute_Lambda(custom_penalty_mat,
                                smoothing_spline_penalty,
                                wiggle_penalty,
                                flat_ridge_penalty,
                                K,
                                p_expansions,
                                unique_penalty_per_predictor,
                                unique_penalty_per_partition,
                                penalty_vec,
                                colnm_expansions,
                                just_Lambda = FALSE)

  return(list("Lambda"            = Lambda_list$Lambda,
              "flat_ridge_penalty" = flat_ridge_penalty,
              "wiggle_penalty"     = wiggle_penalty,
              "other_penalties"    = penalty_vec,
              "L_predictor_list"   = Lambda_list$L_predictor_list,
              "L_partition_list"   = Lambda_list$L_partition_list))
}

local_null_coalesce <- function(x, y) {
  if (is.null(x)) y else x
}

build_tuning_env_from_fit <- function(fit,
                                      tuning_criterion = "loo",
                                      gcv_gamma = 1.4) {
  X_std <- lapply(fit$X, fit$std_X)
  smoothing_spline_penalty <- lgspline:::get_2ndDerivPenalty(
    colnm_expansions = fit$raw_expansion_names,
    C = fit$X[[1]],
    power1_cols = fit$power1_cols,
    power2_cols = fit$power2_cols,
    power3_cols = fit$power3_cols,
    power4_cols = fit$power4_cols,
    interaction_single_cols = fit$interaction_single_cols,
    interaction_quad_cols = fit$interaction_quad_cols,
    triplet_cols = fit$triplet_cols,
    p_expansions = fit$p
  )
  y_vec <- fit$y
  y_list <- lapply(fit$order_list, function(idx) y_vec[idx])
  X_gram <- lapply(X_std, crossprod)
  Xy <- lapply(seq_along(X_std), function(k) crossprod(X_std[[k]], cbind(y_list[[k]])))
  delta <- fit$.fit_call_args$delta
  if (is.null(delta) && paste0(fit$family)[2] != "identity") {
    delta <- lgspline:::.compute_tuning_delta(
      family = fit$family,
      unl_y = y_vec,
      N_obs = fit$N,
      observation_weights = NULL,
      opt = TRUE
    )
  }
  if (is.null(delta)) {
    delta <- 0
  }

  A <- fit$A
  if (is.null(A)) {
    A <- cbind(rep(0, (fit$K + 1) * fit$p))
    A <- cbind(A, A)
  }
  R_constraints <- ncol(A)

  observation_weights <- fit$weights
  if (is.null(observation_weights)) {
    observation_weights <- rep(1, fit$N)
  }
  observation_weights_list <- lapply(fit$order_list, function(idx) {
    observation_weights[idx]
  })
  homogenous_weights <- length(unique(observation_weights)) == 1

  flat_cols <- c()
  flat_terms <- local_null_coalesce(
    fit$.fit_call_args$just_linear_without_interactions,
    c()
  )
  if (length(flat_terms) > 0) {
    flat_cols <- which(fit$raw_expansion_names %in% paste0("_", flat_terms, "_"))
  }
  use_blockfit <- isTRUE(fit$.fit_call_args$blockfit) &&
    length(flat_cols) > 0 && fit$K > 0

  lgspline:::.build_tuning_env(
    y = y_list,
    X = X_std,
    X_gram = X_gram,
    Xy = Xy,
    smoothing_spline_penalty = smoothing_spline_penalty,
    A = A,
    R_constraints = R_constraints,
    K = fit$K,
    p_expansions = fit$p,
    N_obs = fit$N,
    custom_penalty_mat = fit$.fit_call_args$custom_penalty_mat,
    colnm_expansions = fit$raw_expansion_names,
    unique_penalty_per_predictor = fit$.fit_call_args$unique_penalty_per_predictor,
    unique_penalty_per_partition = fit$.fit_call_args$unique_penalty_per_partition,
    meta_penalty = fit$.fit_call_args$meta_penalty,
    family = fit$family,
    delta = delta,
    order_list = fit$order_list,
    observation_weights = observation_weights_list,
    homogenous_weights = homogenous_weights,
    parallel = FALSE,
    parallel_eigen = FALSE,
    parallel_trace = FALSE,
    parallel_aga = FALSE,
    parallel_matmult = FALSE,
    parallel_unconstrained = FALSE,
    cl = NULL,
    chunk_size = 1,
    num_chunks = 0,
    rem_chunks = 0,
    unconstrained_fit_fxn = fit$.fit_call_args$unconstrained_fit_fxn,
    keep_weighted_Lambda = fit$.fit_call_args$keep_weighted_Lambda,
    iterate = fit$.fit_call_args$iterate_tune,
    qp_score_function = fit$.fit_call_args$qp_score_function,
    quadprog = FALSE,
    qp_Amat = NULL,
    qp_bvec = NULL,
    qp_meq = 0,
    tol = fit$.fit_call_args$tol,
    sd_y = fit$sd_y,
    tuning_criterion = tuning_criterion,
    gcv_gamma = gcv_gamma,
    constraint_value_vectors = local_null_coalesce(fit$constraint_values, cbind()),
    glm_weight_function = fit$.fit_call_args$glm_weight_function,
    schur_correction_function = fit$.fit_call_args$schur_correction_function,
    need_dispersion_for_estimation = fit$.fit_call_args$need_dispersion_for_estimation,
    dispersion_function = fit$.fit_call_args$dispersion_function,
    blockfit = fit$.fit_call_args$blockfit,
    just_linear_without_interactions = flat_terms,
    Vhalf = fit$Vhalf,
    VhalfInv = fit$VhalfInv,
    verbose = FALSE,
    include_warnings = FALSE,
    flat_cols = flat_cols,
    use_blockfit = use_blockfit
  )
}

make_gaussian_linear_fit <- function(wiggle_penalty = 1e-3,
                                     flat_ridge_penalty = 0.2) {
  t <- seq(-2, 2, length.out = 8)
  y <- 1 + 0.5 * t + c(0.1, -0.05, 0.02, -0.03, 0.04, -0.02, 0.01, -0.04)

  fit <- lgspline(
    cbind(t),
    y,
    K = 0,
    opt = FALSE,
    wiggle_penalty = wiggle_penalty,
    flat_ridge_penalty = flat_ridge_penalty,
    unique_penalty_per_predictor = FALSE,
    unique_penalty_per_partition = FALSE,
    just_linear_without_interactions = 1,
    standardize_response = FALSE,
    include_warnings = FALSE
  )

  list(
    t = t,
    y = y,
    fit = fit,
    wiggle_penalty = wiggle_penalty,
    flat_ridge_penalty = flat_ridge_penalty
  )
}

test_that(".compute_loocv matches explicit Gaussian refits at fixed penalties", {
  case <- make_gaussian_linear_fit()
  env <- build_tuning_env_from_fit(case$fit, tuning_criterion = "loo")
  par <- log(c(case$wiggle_penalty, case$flat_ridge_penalty))

  loo_obj <- lgspline:::.compute_loocv(par, c(), env)

  loo_refit <- sapply(seq_along(case$y), function(i) {
    refit <- lgspline(
      cbind(case$t[-i]),
      case$y[-i],
      K = 0,
      opt = FALSE,
      wiggle_penalty = exp(par[1]),
      flat_ridge_penalty = exp(par[2]),
      unique_penalty_per_predictor = FALSE,
      unique_penalty_per_partition = FALSE,
      just_linear_without_interactions = 1,
      standardize_response = FALSE,
      include_warnings = FALSE
    )
    as.numeric(predict(refit, cbind(case$t[i])))
  })

  loo_refit_value <- mean((case$y - loo_refit)^2)

  expect_equal(loo_obj$LOO_u, loo_refit_value, tolerance = 1e-4)
  expect_equal(loo_obj$criterion_value, loo_obj$LOO_u, tolerance = 1e-10)
})

test_that(".compute_loocv_gradient matches finite differences in a Gaussian case", {
  case <- make_gaussian_linear_fit(wiggle_penalty = 5e-3, flat_ridge_penalty = 0.3)
  env <- build_tuning_env_from_fit(case$fit, tuning_criterion = "loo")
  par <- log(c(5e-3, 0.3))

  analytic <- lgspline:::.compute_loocv_gradient(par, c(), env = env)$gradient[1:2]

  h <- 1e-5
  numeric <- vapply(seq_along(par), function(j) {
    step <- rep(0, length(par))
    step[j] <- h
    up <- lgspline:::.compute_loocv(par + step, c(), env)$criterion_value
    down <- lgspline:::.compute_loocv(par - step, c(), env)$criterion_value
    (up - down) / (2 * h)
  }, numeric(1))

  expect_equal(analytic, numeric, tolerance = 1e-4)
})

test_that("default tuning is LOO and gcv_gamma affects only the GCV objective", {
  t <- seq(-2, 2, length.out = 30)
  y <- sin(t) + 0.1 * t^2
  fit_gcv_case <- lgspline(
    cbind(t), y,
    K = 1,
    opt = FALSE,
    wiggle_penalty = 5e-3,
    flat_ridge_penalty = 0.3,
    unique_penalty_per_predictor = FALSE,
    unique_penalty_per_partition = FALSE,
    standardize_response = FALSE,
    include_warnings = FALSE
  )
  env_loo_1 <- build_tuning_env_from_fit(fit_gcv_case, tuning_criterion = "loo", gcv_gamma = 1)
  env_loo_9 <- build_tuning_env_from_fit(fit_gcv_case, tuning_criterion = "loo", gcv_gamma = 9)
  env_gcv_1 <- build_tuning_env_from_fit(fit_gcv_case, tuning_criterion = "gcv", gcv_gamma = 1)
  env_gcv_9 <- build_tuning_env_from_fit(fit_gcv_case, tuning_criterion = "gcv", gcv_gamma = 9)
  par <- log(c(5e-3, 0.3))

  loo_1 <- lgspline:::.compute_loocv(par, c(), env_loo_1)$criterion_value
  loo_9 <- lgspline:::.compute_loocv(par, c(), env_loo_9)$criterion_value
  gcv_1 <- lgspline:::.compute_gcvu(par, c(), env_gcv_1)$criterion_value
  gcv_9 <- lgspline:::.compute_gcvu(par, c(), env_gcv_9)$criterion_value

  expect_equal(loo_1, loo_9, tolerance = 1e-10)
  expect_false(isTRUE(all.equal(gcv_1, gcv_9, tolerance = 1e-10)))

  t <- seq(-1, 1, length.out = 20)
  y <- sin(t) + 0.1 * cos(3 * t)
  fit_default <- lgspline(
    cbind(t), y,
    K = 1,
    opt = TRUE,
    use_custom_bfgs = FALSE,
    gcv_gamma = 9,
    initial_wiggle = c(1e-4, 5e-4),
    initial_flat = c(0.2, 0.5),
    unique_penalty_per_predictor = FALSE,
    unique_penalty_per_partition = FALSE,
    just_linear_without_interactions = 1,
    standardize_response = FALSE,
    include_warnings = FALSE
  )
  fit_explicit_loo <- lgspline(
    cbind(t), y,
    K = 1,
    opt = TRUE,
    use_custom_bfgs = FALSE,
    tuning_criterion = "loo",
    gcv_gamma = 1,
    initial_wiggle = c(1e-4, 5e-4),
    initial_flat = c(0.2, 0.5),
    unique_penalty_per_predictor = FALSE,
    unique_penalty_per_partition = FALSE,
    just_linear_without_interactions = 1,
    standardize_response = FALSE,
    include_warnings = FALSE
  )

  expect_equal(fit_default$penalties$Lambda,
               fit_explicit_loo$penalties$Lambda,
               tolerance = 1e-8)
})

test_that("LOO and GCV tuning helpers stay finite for a non-Gaussian case", {
  set.seed(2)
  x <- seq(-1, 1, length.out = 40)
  eta <- -0.3 + 1.2 * x
  p <- plogis(eta)
  y <- rbinom(length(x), 1, p)

  fit <- lgspline(
    cbind(x),
    y,
    family = binomial(),
    K = 1,
    opt = FALSE,
    wiggle_penalty = 1e-2,
    flat_ridge_penalty = 0.5,
    unique_penalty_per_predictor = FALSE,
    unique_penalty_per_partition = FALSE,
    standardize_response = FALSE,
    include_warnings = FALSE
  )

  env_loo <- build_tuning_env_from_fit(fit, tuning_criterion = "loo")
  env_gcv <- build_tuning_env_from_fit(fit, tuning_criterion = "gcv", gcv_gamma = 1.4)
  par <- log(c(1e-2, 0.5))

  loo_obj <- lgspline:::.compute_loocv(par, c(), env_loo)
  loo_grad <- lgspline:::.compute_loocv_gradient(par, c(), env = env_loo)
  gcv_obj <- lgspline:::.compute_gcvu(par, c(), env_gcv)
  gcv_grad <- lgspline:::.compute_gcvu_gradient(par, c(), env = env_gcv)

  expect_true(is.finite(loo_obj$criterion_value))
  expect_true(all(is.finite(loo_grad$gradient)))
  expect_true(is.finite(gcv_obj$criterion_value))
  expect_true(all(is.finite(gcv_grad$gradient)))
})

test_that("use_custom_bfgs = FALSE works for both tuning criteria", {
  x <- seq(-1, 1, length.out = 12)
  y <- x^2 + 0.1 * x

  fit_loo <- lgspline(
    cbind(x), y,
    K = 0,
    opt = TRUE,
    use_custom_bfgs = FALSE,
    tuning_criterion = "loo",
    initial_wiggle = c(1e-4, 5e-4),
    initial_flat = c(0.2, 0.5),
    unique_penalty_per_predictor = FALSE,
    unique_penalty_per_partition = FALSE,
    just_linear_without_interactions = 1,
    standardize_response = FALSE,
    include_warnings = FALSE
  )
  fit_gcv <- lgspline(
    cbind(x), y,
    K = 0,
    opt = TRUE,
    use_custom_bfgs = FALSE,
    tuning_criterion = "gcv",
    gcv_gamma = 1.4,
    initial_wiggle = c(1e-4, 5e-4),
    initial_flat = c(0.2, 0.5),
    unique_penalty_per_predictor = FALSE,
    unique_penalty_per_partition = FALSE,
    just_linear_without_interactions = 1,
    standardize_response = FALSE,
    include_warnings = FALSE
  )

  expect_true(all(is.finite(fit_loo$penalties$Lambda)))
  expect_true(all(is.finite(fit_gcv$penalties$Lambda)))
})

test_that("post-optimization sample-size adjustment shrinks tuned penalties for both criteria", {
  x <- seq(-1, 1, length.out = 14)
  y <- sin(pi * x) + 0.05 * x^2
  initial_wiggle <- c(1e-4, 5e-4)
  initial_flat <- c(0.2, 0.5)
  sample_size_adjustment <- ((length(x) - 2) / (length(x) + 2))^2

  base_fit <- lgspline(
    cbind(x), y,
    K = 1,
    opt = FALSE,
    wiggle_penalty = 5e-3,
    flat_ridge_penalty = 0.3,
    unique_penalty_per_predictor = FALSE,
    unique_penalty_per_partition = FALSE,
    just_linear_without_interactions = 1,
    standardize_response = FALSE,
    include_warnings = FALSE
  )

  for (criterion in c("loo", "gcv")) {
    env <- build_tuning_env_from_fit(
      base_fit,
      tuning_criterion = criterion,
      gcv_gamma = 1.4
    )
    criterion_fxn <- if (criterion == "loo") {
      lgspline:::.compute_loocv
    } else {
      lgspline:::.compute_gcvu
    }
    best_start <- lgspline:::.tune_grid_search(
      log(initial_wiggle),
      log(initial_flat),
      c(),
      criterion_fxn,
      env,
      include_warnings = FALSE
    )
    opt_res <- optim(
      c(best_start, c()),
      fn = function(par) criterion_fxn(par, c(), env)$criterion_value,
      method = "BFGS"
    )
    raw_penalties <- exp(opt_res$par[1:2])

    fit_tuned <- lgspline(
      cbind(x), y,
      K = 1,
      opt = TRUE,
      use_custom_bfgs = FALSE,
      tuning_criterion = criterion,
      gcv_gamma = 1.4,
      initial_wiggle = initial_wiggle,
      initial_flat = initial_flat,
      unique_penalty_per_predictor = FALSE,
      unique_penalty_per_partition = FALSE,
      just_linear_without_interactions = 1,
      standardize_response = FALSE,
      include_warnings = FALSE
    )

    expect_equal(
      fit_tuned$penalties$wiggle_penalty,
      raw_penalties[1] * sample_size_adjustment,
      tolerance = 1e-8
    )
    expect_equal(
      fit_tuned$penalties$flat_ridge_penalty,
      raw_penalties[2] * sample_size_adjustment,
      tolerance = 1e-8
    )
    expect_lt(fit_tuned$penalties$wiggle_penalty, raw_penalties[1])
    expect_lt(fit_tuned$penalties$flat_ridge_penalty, raw_penalties[2])
  }
})

test_that(".damped_bfgs never returns a GCV solution worse than the grid start", {
  set.seed(123)
  x <- seq(-2, 2, length.out = 30)
  y <- sin(2 * x) + 0.15 * x^2

  fit <- lgspline(
    cbind(x), y,
    K = 1,
    opt = FALSE,
    wiggle_penalty = 5e-3,
    flat_ridge_penalty = 0.3,
    unique_penalty_per_predictor = FALSE,
    unique_penalty_per_partition = FALSE,
    just_linear_without_interactions = 1,
    standardize_response = FALSE,
    include_warnings = FALSE
  )

  env <- build_tuning_env_from_fit(fit, tuning_criterion = "gcv", gcv_gamma = 1.4)
  log_penalty_vec <- c()
  log_initial_wiggle <- log(c(1e-4, 5e-4, 5e-3))
  log_initial_flat <- log(c(0.2, 0.5, 1))

  best_start <- lgspline:::.tune_grid_search(
    log_initial_wiggle,
    log_initial_flat,
    log_penalty_vec,
    lgspline:::.compute_gcvu,
    env,
    include_warnings = FALSE
  )

  start_value <- lgspline:::.compute_gcvu(best_start, log_penalty_vec, env)$criterion_value
  res <- lgspline:::.damped_bfgs(
    c(best_start, log_penalty_vec),
    log_penalty_vec,
    lgspline:::.compute_gcvu,
    lgspline:::.compute_gcvu_gradient,
    env,
    tol = 1e-6
  )

  expect_lte(res$criterion_value, start_value + 1e-10)
})

test_that(".damped_bfgs preserves the start point when the first step is worse", {
  env <- list()
  start <- c(0, 0)

  criterion_fxn <- function(par, log_penalty_vec, env, ...) {
    list(criterion_value = sum(par^2))
  }
  gr_fxn <- function(par, log_penalty_vec, outlist, env, ...) {
    list(
      gradient = c(-10, -10),
      outlist = list(criterion_value = sum(par^2))
    )
  }

  res <- lgspline:::.damped_bfgs(
    par = start,
    log_penalty_vec = c(),
    criterion_fxn = criterion_fxn,
    gr_fxn = gr_fxn,
    env = env,
    tol = 1e-8,
    max_iter = 2
  )

  expect_equal(res$par, start, tolerance = 1e-12)
  expect_equal(res$criterion_value, 0, tolerance = 1e-12)
})

test_that("custom GCV tuning does not inflate partition penalties relative to finite differences", {
  set.seed(1234)
  x <- sort(runif(80, -3, 3))
  eta <- sin(x) - 0.2 * x^2
  y <- rbinom(length(x), 1, plogis(eta))
  dat <- data.frame(x = x, y = y)

  fit_fd <- lgspline(
    y ~ spl(x),
    K = 1,
    dat,
    tuning_criterion = "gcv",
    use_custom_bfgs = FALSE,
    unique_penalty_per_predictor = FALSE,
    unique_penalty_per_partition = TRUE,
    include_warnings = FALSE
  )
  fit_custom <- lgspline(
    y ~ spl(x),
    K = 1,
    dat,
    tuning_criterion = "gcv",
    use_custom_bfgs = TRUE,
    unique_penalty_per_predictor = FALSE,
    unique_penalty_per_partition = TRUE,
    include_warnings = FALSE
  )

  ratio <- max(fit_custom$penalties$other_penalties) /
    max(fit_fd$penalties$other_penalties)

  expect_lte(ratio, 10)
})

#' @useDynLib lgspline
#' @importFrom Rcpp sourceCpp
NULL

#' Create One-Hot Encoded Matrix
#'
#' @description
#' Converts a categorical vector into a one-hot encoded matrix where each unique value
#' becomes a binary column.
#'
#' @param x A vector containing categorical values (factors, character, etc.)
#' @param drop_first Logical; if \code{TRUE} and more than one dummy column is
#'   created, drop the first column.
#'
#' @return A data frame containing the one-hot encoded binary columns with cleaned column names
#'
#' @details
#' The function creates dummy variables for each unique value in the input vector using
#' \code{model.matrix()} with dummy-intercept coding. Column names are cleaned by removing the
#' 'x' prefix added by \code{model.matrix()}.
#'
#' @examples
#'
#'
#' ## lgspline will not accept this format of "catvar", because inputting data
#' # this way can cause difficult-to-diagnose issues in formula parsing
#' # all variables must be numeric
#' df <- data.frame(numvar = rnorm(100),
#'                  catvar = rep(LETTERS[1:4],
#'                               25))
#' print(head(df))
#'
#' ## Instead, replace with dummy-intercept coding by
#' # 1) applying one-hot encoding
#' # 2) dropping the first column
#' # 3) appending to our data
#'
#' dummy_intercept_coding <- create_onehot(df$catvar)[,-1]
#' df$catvar <- NULL
#' df <- cbind(df, dummy_intercept_coding)
#' print(head(df))
#'
#'
#'
#' @export
create_onehot <- function(x, drop_first = FALSE) {
  ## For convenience, coerce to factor to ensure stable level ordering
  if(!is.factor(x)) x <- as.factor(x)
  mat <- model.matrix(~ x - 1)
  colnames(mat) <- sub("^x", "", colnames(mat))
  result <- as.data.frame(mat)
  if(drop_first && ncol(result) > 1){
    result <- result[, -1, drop = FALSE]
  }
  return(result)
}

#' Standardize Vector to Z-Scores
#'
#' @description
#' Centers a vector by its sample mean, then scales it by its sample standard deviation
#' \eqn{(\text{x}-\text{mean}(\text{x}))/\text{sd}(\text{x})}.
#'
#'
#' @param x Numeric vector to standardize
#'
#' @return Standardized vector with sample mean 0 and standard deviation 1
#'
#' @examples
#'
#' x <- c(1, 2, 3, 4, 5)
#' std(x)
#' print(mean(x))
#' print(sd(x))
#'
#'
#' @export
#' @export
std <- function(x){
  (x-mean(x))/sd(x)
}

#' Compute softplus transform
#'
#' @description
#' Computes the softplus transform, equivalent to the cumulant generating function
#' of a logistic regression model: \eqn{\log(1+e^x)}.
#'
#'
#' @param x Numeric vector to apply softplus to
#'
#' @return Softplus transformed vector
#'
#' @examples
#'
#' x <- runif(5)
#' softplus(x)
#'
#'
#' @keywords internal
#' @export
softplus <- function(x){
  log(1+exp(x))
}

#' Efficient Matrix Multiplication Operator
#'
#' @description
#' Operator wrapper around C++ \code{efficient_matrix_mult()} for matrix multiplication syntax.
#'
#' This is an internal function meant to provide improvement over base R's operator for
#' certain large matrix operations, at a cost of potential slight slowdown for
#' smaller problems.
#'
#' @param x Left matrix
#' @param y Right matrix
#'
#' @return Matrix product of x and y
#'
#' @examples
#'
#' M1 <- matrix(1:4, 2, 2)
#' M2 <- matrix(5:8, 2, 2)
#' M1 %**% M2
#'
#'
#' @keywords internal
#' @export
`%**%` <- function(x, y) {
  efficient_matrix_mult(x, y)
}

#' Matrix Inversion with Fallback Methods
#'
#' @description
#' Attempts matrix inversion using multiple methods, falling back to more robust
#' approaches if standard inversion fails.
#'
#' @param mat Square matrix to invert
#' @param include_warnings Logical; default FALSE for current implementation.
#'
#' @return Inverted matrix, or the identity matrix if all inversion attempts fail
#'
#' @details
#' Tries methods in order:
#'
#' 1. Cholesky decomposition via \code{chol2inv(chol(...))} for symmetric
#'    positive-definite matrices (fastest for SPD)
#'
#' 2. Direct inversion using \code{armaInv()} as first fallback
#'
#' 3. Generalized inverse using eigendecomposition with small ridge
#'
#' 4. Returns identity matrix if all methods fail, with optional warning
#'
#' For eigendecomposition, uses a small ridge penalty (\code{1e-16}) for stability and
#' zeroes eigenvalues below machine precision.
#'
#' @examples
#'
#' ## Well-conditioned matrix
#' A <- matrix(c(4,2,2,4), 2, 2)
#' invert(A) %*% A
#'
#' ## Singular matrix falls back to M.P. generalized inverse
#' B <- matrix(c(1,1,1,1), 2, 2)
#' invert(B) %*% B
#'
#'
#' @keywords internal
#' @export
invert <- function(mat, include_warnings = FALSE){

  ## Try Cholesky first for SPD matrices
  #  chol2inv(chol(M)) is faster than solve(M) for SPD M
  t <- try({
    chol2inv(chol(mat))
  }, silent = TRUE)

  if(!any(inherits(t, 'try-error'))){
    return(t)
  }

  ## Fallback 1: direct inversion via Armadillo
  t <- try({
    armaInv(mat)
  }, silent = TRUE)

  if(!any(inherits(t, 'try-error'))){
    return(t)
  }

  ## generalized inverse with small ridge penalty
  t <- try({
    eig <- eigen(mat + 1e-16*diag(nrow(mat)), symmetric = TRUE)
    d_inv <- ifelse(eig$values <= sqrt(.Machine$double.eps),
                    0,
                    1/eig$values)
    tcrossprod(eig$vectors * rep(d_inv, each = nrow(mat)), eig$vectors)
  }, silent = TRUE)

  if(!any(inherits(t, 'try-error'))){
    return(t)
  }

  ## Fallback 3: return identity matrix with warning
  # Justification for identity is that for Newton-Raphson, this reduces
  # to approx. gradient-descent. But this is a terrible option for inference!
  if(include_warnings) warning('Matrix not inverted, returning identity: ',
                               print(t))
  return(diag(nrow(mat)))
}

#' Multiply Block Diagonal Matrices in Parallel
#'
#' @description
#' Multiplies two lists of matrices that form block diagonal structures, with optional
#' parallel processing.
#'
#' @param A List of matrices forming first block diagonal matrix
#' @param B List of matrices forming second block diagonal matrix
#' @param K Number of blocks minus 1 (\eqn{K})
#' @param parallel Logical; whether to use parallel processing
#' @param cl Cluster object for parallel processing
#' @param chunk_size Number of blocks per chunk for parallel processing
#' @param num_chunks Number of chunks for parallel processing
#' @param rem_chunks Remaining blocks after chunking
#'
#' @return List containing products of corresponding blocks
#'
#' @details
#' When \code{parallel=TRUE}, splits computation into chunks processed in parallel.
#' Handles remainder chunks separately. Uses \code{matmult_block_diagonal_cpp()} for
#' actual multiplication.
#'
#' The function expects A and B to contain corresponding blocks that can be
#' matrix multiplied.
#'
#' @examples
#'
#' A <- list(matrix(1:4,2,2), matrix(5:8,2,2))
#' B <- list(matrix(1:4,2,2), matrix(5:8,2,2))
#' matmult_block_diagonal(A, B, K=1, parallel=FALSE, cl=NULL,
#'                        chunk_size=1, num_chunks=1, rem_chunks=0)
#'
#' @keywords internal
#' @export
matmult_block_diagonal <- function(A,
                                   B,
                                   K,
                                   parallel,
                                   cl,
                                   chunk_size,
                                   num_chunks,
                                   rem_chunks){
  if(parallel & !is.null(cl)){
    ## Start with remainder chunks
    if(rem_chunks > 0){
      rem <- matmult_block_diagonal_cpp(A[num_chunks*chunk_size + 1:rem_chunks],
                                        B[num_chunks*chunk_size + 1:rem_chunks],
                                        rem_chunks-1)
    } else {
      rem <- list()
    }
    ## Handle the rest in parallel
    c(
      Reduce("c",parallel::parLapply(cl, 1:num_chunks, function(k){
        inds <- 1:chunk_size + (k-1)*chunk_size
        matmult_block_diagonal_cpp(A[inds],
                                   B[inds],
                                   chunk_size-1)
      })),
      rem
    )
  } else {
    matmult_block_diagonal_cpp(A, B, K)
  }
}

#' Generate Design Matrix with Polynomial and Interaction Terms
#'
#' @description
#' Internal function for creating a design matrix containing polynomial
#' expansions and interaction terms for predictor variables. Supports
#' customizable term generation including polynomial degrees up to quartic
#' terms, interaction types, and selective term exclusion.
#'
#' Column names take on the form "_v_" for linear terms, "_v_^d" for polynomial
#' powers up to d = 4, and "_v_x_w_" for interactions between variables v and w,
#' where v and w are column indices of the input predictor matrix.
#'
#' The \code{custom_basis_fxn} argument, if supplied, requires the same arguments
#' as this function, in the same order, excluding the argument
#' "custom_basis_fxn" itself.
#'
#' @param predictors Numeric matrix of predictor variables
#' @param numerics Integer vector; column indices for variables to expand as polynomials
#' @param just_linear_with_interactions Integer vector; column indices for variables to keep linear but allow interactions
#' @param just_linear_without_interactions Integer vector; column indices for variables to keep linear without interactions
#' @param exclude_interactions_for Integer vector; column indices to exclude from all interactions
#' @param include_quadratic_terms Logical; whether to include squared terms (default TRUE)
#' @param include_cubic_terms Logical; whether to include cubic terms (default TRUE)
#' @param include_quartic_terms Logical; whether to include 4th degree terms (default FALSE)
#' @param include_2way_interactions Logical; whether to include two-way interactions (default TRUE)
#' @param include_3way_interactions Logical; whether to include three-way interactions (default TRUE)
#' @param include_quadratic_interactions Logical; whether to include interactions with squared terms (default TRUE)
#' @param exclude_these_expansions Character vector; names of specific terms to exclude from final matrix
#' @param custom_basis_fxn Function; optional custom basis expansion function that accepts all arguments listed here except itself
#' @param ... Additional arguments passed to \code{custom_basis_fxn}
#'
#' @return Matrix with columns for intercept, polynomial terms, and specified interactions
#'
#' @keywords internal
#' @export
get_polynomial_expansions <- function(predictors,
                                      numerics,
                                      just_linear_with_interactions,
                                      just_linear_without_interactions,
                                      exclude_interactions_for = NULL,
                                      include_quadratic_terms = TRUE,
                                      include_cubic_terms = TRUE,
                                      include_quartic_terms = FALSE,
                                      include_2way_interactions = TRUE,
                                      include_3way_interactions = TRUE,
                                      include_quadratic_interactions = FALSE,
                                      exclude_these_expansions = NULL,
                                      custom_basis_fxn = NULL,
                                      ...) {
  if(any(!is.null(custom_basis_fxn))){
    ## Custom basis-expansion function, if desired
    return(custom_basis_fxn(predictors,
                            numerics,
                            just_linear_with_interactions,
                            just_linear_without_interactions,
                            exclude_interactions_for,
                            include_quadratic_terms,
                            include_cubic_terms,
                            include_quartic_terms,
                            include_2way_interactions,
                            include_3way_interactions,
                            include_quadratic_interactions,
                            exclude_these_expansions,
                            ...))
  } else {

    ## Dimension setup
    n_rows <- nrow(predictors)
    n_numerics <- length(numerics)
    n_linear_with <- length(just_linear_with_interactions)
    n_linear_without <- length(just_linear_without_interactions)
    n_vars <- ncol(predictors)

    ## Define which variables can participate in interactions
    # Start with potential interaction variables
    vars_that_interact <- unique(c(numerics, just_linear_with_interactions))

    ## Remove any variables specifically excluded from interactions
    if(!is.null(exclude_interactions_for)) {
      vars_that_interact <- setdiff(vars_that_interact,
                                    exclude_interactions_for)
    }

    n_interact_vars <- length(vars_that_interact)

    ## Multiplier for degree of polynomial expansions (3 or 4) we use
    if(include_quartic_terms){
      mult <- 4
    } else {
      mult <- 3
    }
    if(!include_quadratic_terms){
      mult <- mult - 1
    }
    if(!include_cubic_terms){
      mult <- mult - 1
    }

    ## Number of columns pre-allocated
    n_cols <- 1 +
      mult * n_numerics +  # polynomial terms for numerics
      n_linear_with +      # linear terms that can interact
      n_linear_without +   # linear terms that can't interact
      choose(n_interact_vars, 2)*
      include_2way_interactions +  # 2-way interactions
      choose(n_interact_vars, 3)*
      include_3way_interactions +  # 3-way interactions
      max(0, n_numerics * (length(vars_that_interact) - 1))*
      include_quadratic_interactions
    # Note: each numeric can interact quadratically with all other vars


    ## Pre-allocate once, then fill left-to-right in the same order used
    #  to build the column-name bookkeeping.
    result <- matrix(0, nrow = n_rows, ncol = n_cols)
    col_names <- character(n_cols)
    col_names[1] <- "intercept"
    col_index <- 2

    ## Intercept
    result[, 1] <- 1

    ## Numeric variables and their powers
    if (n_numerics > 0) {
      numeric_data <- predictors[, numerics, drop = FALSE]

      ## Linear terms
      result[, col_index:(col_index + n_numerics - 1)] <-
        numeric_data
      col_names[col_index:(col_index + n_numerics - 1)] <-
        paste0("_", numerics, "_")
      col_index <- col_index + n_numerics

      ## Quadratic terms
      if(include_quadratic_terms){
        result[, col_index:(col_index + n_numerics - 1)] <-
          numeric_data^2
        col_names[col_index:(col_index + n_numerics - 1)] <-
          paste0("_", numerics, "_^2")
        col_index <- col_index + n_numerics
      }

      ## Cubic terms
      if(include_cubic_terms){
        result[, col_index:(col_index + n_numerics - 1)] <-
          numeric_data^3
        col_names[col_index:(col_index + n_numerics - 1)] <-
          paste0("_", numerics, "_^3")
        col_index <- col_index + n_numerics
      }

      ## Quartic terms
      if(include_quartic_terms){
        result[, col_index:(col_index + n_numerics - 1)] <-
          numeric_data^4
        col_names[col_index:(col_index + n_numerics - 1)] <-
          paste0("_", numerics, "_^4")
        col_index <- col_index + n_numerics
      }
    }

    ## Linear terms that can interact
    if (length(just_linear_with_interactions) > 0) {
      result[, col_index:(col_index + n_linear_with - 1)] <-
        predictors[, just_linear_with_interactions, drop = FALSE]
      col_names[col_index:(col_index + n_linear_with - 1)] <-
        paste0("_", just_linear_with_interactions, "_")
      col_index <- col_index + n_linear_with
    }

    ## Linear terms that cannot interact
    if (n_linear_without > 0) {
      result[, col_index:(col_index + n_linear_without - 1)] <-
        predictors[, just_linear_without_interactions, drop = FALSE]
      col_names[col_index:(col_index + n_linear_without - 1)] <-
        paste0("_", just_linear_without_interactions, "_")
      col_index <- col_index + n_linear_without
    }

    ## 2-way Interactions (among variables that can interact)
    if(n_interact_vars > 1 & include_2way_interactions){
      for (i in 1:(length(vars_that_interact) - 1)) {
        for (j in (i+1):length(vars_that_interact)) {
          var_i <- vars_that_interact[i]
          var_j <- vars_that_interact[j]
          result[, col_index] <- predictors[, var_i] * predictors[, var_j]
          col_names[col_index] <- paste0("_", var_i, "_x_", var_j, "_")
          col_index <- col_index + 1
        }
      }
    }

    ## 3-way Interactions (among variables that can interact)
    if(n_interact_vars > 2 & include_3way_interactions){
      for (i in 1:(length(vars_that_interact) - 2)) {
        for (j in (i+1):(length(vars_that_interact) - 1)) {
          for (k in (j+1):length(vars_that_interact)) {
            var_i <- vars_that_interact[i]
            var_j <- vars_that_interact[j]
            var_k <- vars_that_interact[k]
            result[, col_index] <-
              predictors[, var_i] *
              predictors[, var_j] *
              predictors[, var_k]
            col_names[col_index] <- paste0("_",
                                           var_i,
                                           "_x_",
                                           var_j,
                                           "_x_",
                                           var_k,
                                           "_")
            col_index <- col_index + 1
          }
        }
      }
    }

    ## Quadratic interactions
    # Note: only numerics (not in exclude list) can have quadratic terms here
    if(include_quadratic_interactions &
       (length(numerics) + length(just_linear_with_interactions)) > 0 &
       length(intersect(numerics, vars_that_interact)) > 0){
      numerics_that_interact <- intersect(numerics, vars_that_interact)
      for (i in numerics_that_interact) {
        for (j in vars_that_interact) {
          if (i != j) {
            result[, col_index] <- predictors[, i]^2 * predictors[, j]
            col_names[col_index] <- paste0("_", j, "_x_", i, "_^2")
            col_index <- col_index + 1
          }
        }
      }
    }

    ## Set column names and filter empty
    colnames(result) <- col_names
    result <- result[,colnames(result) != ""]

    ## remove custom expansions if desired
    if(any(!is.null(exclude_these_expansions))){
      if(length(exclude_these_expansions) > 0){
        excl <- which(unlist(colnames(result)) %in%
                        unlist(exclude_these_expansions))
        if(length(excl) > 0){
          result <- result[,-excl, drop=FALSE]
        }
      }
    }

    return(result)
  }
}

#' Calculate Derivatives of Polynomial Terms
#'
#' @description
#' Computes first or second derivatives of polynomial terms in a design matrix with
#' respect to a specified variable. Handles polynomial terms up to fourth degree.
#'
#' @param dat Numeric matrix; design matrix containing polynomial basis expansions
#' @param var Character; column name of variable to differentiate with respect to
#' @param second Logical; if TRUE compute second derivative, if FALSE compute first derivative (default FALSE)
#' @param scale Numeric; scaling factor for normalization
#'
#' @return Numeric matrix containing derivatives of polynomial terms, with same dimensions as input matrix
#'
#' @keywords internal
#' @export
take_derivative <- function(dat, var, second = FALSE, scale) {

  ## Initialize matrix for returning
  n_cols <- ncol(dat)
  dat_deriv <- matrix(0, nrow = nrow(dat), ncol = n_cols)
  colnames(dat_deriv) <- colnames(dat)
  variable_values <- dat[, var]

  ## Walk the expansion columns and recover the derivative from the column
  #  naming convention rather than rebuilding the basis from scratch.
  for (i in 1:n_cols) {
    col_name <- colnames(dat)[i]
    match <- regexpr(var, col_name)[[1]]

    ## Check for quadratic terms of var
    qv <- paste0(var,'^2')
    prefix <- substr(col_name, 1, regexpr('_x', col_name)[[1]])
    col_name_sub <- substr(col_name, nchar(prefix)+2, nchar(col_name))
    match2 <- 1*(col_name_sub == qv) + 1*(col_name == qv)

    ## Derivative multiplier based on polynomial degree
    if(match2){
      mult <- 2
    } else if(substr(col_name, nchar(col_name)-1, nchar(col_name)) == '^3'){
      if(second){
        mult <- 6
      } else{
        mult <- 3
      }
    } else if(substr(col_name, nchar(col_name)-1, nchar(col_name)) == '^4'){
      if(second){
        mult <- 12
      } else{
        mult <- 4
      }
    } else {
      mult <- 1
    }

    ## Detect if basis expansion is a function of variable of interest
    # using column title
    if (match > 0) {

      ## Second derivative
      if(second){
        dat_deriv[, i] <- mult * dat[, i] /
          (variable_values^2 + 1*(variable_values == 0))

        ## First derivative
      } else {
        dat_deriv[, i] <- mult * dat[, i] /
          (variable_values + 1*(variable_values == 0))
      }
    }
  }

  ## For second derivatives, set linear terms to 0
  if(second){
    try(dat_deriv[,var] <- 0, silent = TRUE) # may be missing from manual excl.
  }

  return(dat_deriv)
}

#' Calculate Second Derivatives of Interaction Terms
#'
#' @description
#' Computes partial second derivatives for interaction terms including
#' two-way linear, quadratic, and three-way interactions. Handles special cases
#' for each type.
#'
#' This function is necessary to compute total second derivatives as the sum of
#' second partial "pure" derivatives (\eqn{d^2/dx^2}) plus second partial "mixed"
#' derivative (\eqn{d^2/dxdz}), for a predictor x and all other predictors z.
#'
#' @param dat Numeric matrix; design matrix containing basis expansions
#' @param var Character; variable name to differentiate with respect to
#' @param interaction_single_cols Integer vector; column indices for linear-linear interactions
#' @param interaction_quad_cols Integer vector; column indices for linear-quadratic interactions
#' @param triplet_cols Integer vector; column indices for three-way interactions
#' @param colnm_expansions Character vector; column names of expansions for each partition
#' @param power1_cols Integer vector; column indices of linear terms
#' @param power2_cols Integer vector; column indices of quadratic terms
#'
#' @return Numeric matrix of second derivatives, same dimensions as input
#'
#' @keywords internal
#' @export
take_interaction_2ndderivative <-
  function(dat,
           var,
           interaction_single_cols,
           interaction_quad_cols,
           triplet_cols,
           colnm_expansions,
           power1_cols,
           power2_cols) {

    ## Initialize output matrix
    n_cols <- ncol(dat)
    dat_deriv <- matrix(0, nrow = nrow(dat), ncol = n_cols)
    colnames(dat_deriv) <- colnames(dat)
    variable_values <- dat[,var]

    ## Index of linear term
    v <- which(colnm_expansions[power1_cols] == var)

    ## Detect interactions, if relevant
    if (length(interaction_single_cols) > 0) {
      interaction_singles <-
        interaction_single_cols[grep(paste0("_", var, "_"),
                                     colnm_expansions[interaction_single_cols])]
      if (length(interaction_singles) > 0) {
        ## 2nd derivative of two-way linear-linear interactions is always 1
        dat_deriv[, interaction_singles] <- 1
      }
    }

    ## For two-way linear-quadratic interactions, more work is needed
    if (length(interaction_quad_cols) > 0) {
      interaction_quads <-
        interaction_quad_cols[grep(paste0("_", var, "_"),
                                   colnm_expansions[interaction_quad_cols])]
      if (length(interaction_quads) > 0) {
        for (w in 1:length(power1_cols[-v])) {
          ## The other variable, with interactions affecting quadratic terms
          wvar <- c(power1_cols[-v])[w]

          ## Quadratic interaction indices
          interq <-
            interaction_quads[grep(colnm_expansions[wvar],
                                   colnm_expansions[interaction_quads])]
          if (length(interq) > 0) {
            ## This is the _var^2_x_w term
            if(length(power2_cols) > 0){
              nchv <- nchar(colnm_expansions[power2_cols[v]])
              interqv2 <-
                interq[substr(colnm_expansions[interq],
                              nchar(colnm_expansions[interq]) - nchv + 1,
                              nchar(colnm_expansions[interq])) ==
                         colnm_expansions[power2_cols[v]]]

              ## This is the _var_x_w_^2 term
              if(length(interqv2) > 0){
                interqv1 <- interq[-which(interq == interqv2)]
              }
            } else {
              ## This is the _var_x_w_^2 term
              interqv1 <- interq
            }

            ## 2nd derivative of each is 2*other, 2*other + 2*self respectively
            dat_deriv[, interqv1] <- 2 * dat[, colnm_expansions[wvar]]
            if(length(power2_cols) > 0){
              if(length(interqv2) > 0){
                dat_deriv[, interqv2] <-
                  dat_deriv[, interqv1] + 2 * dat[, var]
              }
            }
          }
        }
      }
    }

    ## 2nd deriv of triplets w.r.t.
    # one variable is always the sum of the other two variables
    if (length(triplet_cols) > 0) {
      triplets <-
        triplet_cols[grep(paste0("_", var, "_"),
                          colnm_expansions[triplet_cols])]
      if (length(triplets) > 0) {
        other2_vars <- lapply(triplets, function(tr) {
          vars <- unlist(strsplit(colnm_expansions[tr], 'x'))
          vars[vars != var]
        })
        for (tr in 1:length(other2_vars)) {
          dat_deriv[, triplets[tr]] <-
            dat[, other2_vars[[tr]][1]] +
            dat[, other2_vars[[tr]][2]]
        }
      }
    }

    return(dat_deriv)
  }


#' Expand Matrix into Partition Lists Based on Knot Boundaries
#'
#' @description
#' Takes an input \eqn{N \times p} matrix of polynomial expansions and outputs a list of
#' length \eqn{K+1}, isolating the rows of the input corresponding to assigned partition.
#'
#' @param partition_codes Numeric vector; values determining partition assignment for each row
#' @param partition_bounds Numeric vector; ordered knot locations defining partition boundaries
#' @param N Integer; number of rows in input matrix
#' @param mat Numeric matrix; data to be partitioned
#' @param K Integer; number of interior knots (resulting in \eqn{K+1} partitions)
#'
#' @return List of length \eqn{K+1}, each element containing the submatrix for that partition
#'
#' @keywords internal
#' @export
knot_expand_list <- function(partition_codes,
                             partition_bounds,
                             N,
                             mat,
                             K){
  ## No knots, return the input matrix
  if(K == 0){
    return(list(mat))
  } else {
    ## Lower and upper bounds are always infinite
    partition_bounds_app <- c(-Inf, partition_bounds, Inf)

    ## Isolate each partition
    expansions <- lapply(1:(K+1), function(partition){
      mat[partition_codes > partition_bounds_app[partition] &
            partition_codes <= partition_bounds_app[partition+1],,drop=FALSE]
    })
    return(expansions)
  }
}


#' Create Block Diagonal Matrix
#'
#' @param matrix_list List of matrices to arrange diagonally
#' @return Block diagonal matrix with input matrices on diagonal
#'
#' @details
#' Takes in a list of matrices, and returns a block-diagonal matrix with each
#' element of the list as one block. All off-diagonal elements are 0.
#' Matrices should be square and have compatible dimensions.
#'
#' @keywords internal
#' @export
create_block_diagonal <- function(matrix_list) {

  ## Calculate the total dimensions of the resulting matrix
  total_dim <- sum(sapply(matrix_list, nrow))

  ## Create an empty matrix filled with zeros
  result <- matrix(0, nrow = total_dim, ncol = total_dim)

  ## Fill the diagonal blocks
  current_row <- 1
  current_col <- 1

  for (mat in matrix_list) {
    dim_mat <- nrow(mat)
    result[current_row:(current_row + dim_mat - 1),
           current_col:(current_col + dim_mat - 1)] <- mat
    current_row <- current_row + dim_mat
    current_col <- current_col + dim_mat
  }

  return(result)
}


#' Compute Gram Matrix for Block Diagonal Structure
#'
#' @param list_in List of matrices
#' @param parallel Use parallel processing
#' @param cl Cluster object
#' @param chunk_size Chunk size for parallel
#' @param num_chunks Number of chunks
#' @param rem_chunks Remaining chunks
#'
#' @return List of Gram matrices (\eqn{\textbf{X}^{T}\textbf{X}}) for each block
#'
#' @details
#' For a list of matrices, will compute the gram matrix of each element of the
#' list.
#'
#'
#' @keywords internal
#' @export
compute_gram_block_diagonal <- function(list_in,
                                        parallel,
                                        cl,
                                        chunk_size,
                                        num_chunks,
                                        rem_chunks) {
  if(parallel & !is.null(cl)) {

    # Handle remainder chunks first
    if(rem_chunks > 0) {
      rem_indices <- num_chunks*chunk_size + 1:rem_chunks
      rem <- lapply(rem_indices, function(k, func) {
        crossprod(list_in[[k]])
      })
    } else {
      rem <- list()
    }

    # Process main chunks in parallel
    result <- c(
      Reduce("c", parallel::parLapply(cl, 1:num_chunks, function(chunk) {
        inds <- (chunk - 1)*chunk_size + 1:chunk_size
        lapply(inds, function(k) {
          crossprod(list_in[[k]])
        })
      })),
      rem
    )
  } else {
    result <- lapply(list_in, crossprod)
  }
  return(result)
}

#' Compute First and Second Derivative Matrices
#'
#' @param p_expansions Number of columns in the basis expansion per partition
#' @param Cpredictors Predictor matrix
#' @param power1_cols Indices of linear terms of spline effects
#' @param power2_cols Indices of quadratic terms of spline effects
#' @param nonspline_cols Indices of non-spline effects
#' @param interaction_single_cols Indices of first-order interactions
#' @param interaction_quad_cols Indices of quadratic interactions
#' @param triplet_cols Indices of three-way interactions
#' @param K Number of partitions minus 1 (\eqn{K})
#' @param include_2way_interactions Include 2-way interactions
#' @param include_3way_interactions Include 3-way interactions
#' @param include_quadratic_interactions Include quadratic interactions
#' @param colnm_expansions Column names
#' @param expansion_scales Scale factors
#' @param just_first_derivatives Only compute first derivatives
#' @param just_spline_effects Only compute derivatives for spline effects
#'
#' @return List containing first-derivative matrices and, unless
#'   \code{just_first_derivatives = TRUE}, second-derivative matrices
#'
#' @keywords internal
#' @export
make_derivative_matrix  <-  function(
    p_expansions,
    Cpredictors,
    power1_cols,
    power2_cols,
    nonspline_cols,
    interaction_single_cols,
    interaction_quad_cols,
    triplet_cols,
    K,
    include_2way_interactions,
    include_3way_interactions,
    include_quadratic_interactions,
    colnm_expansions,
    expansion_scales,
    just_first_derivatives = FALSE,
    just_spline_effects = TRUE){

  ## Include derivatives for non-spline effects, if desired
  if(!just_spline_effects){
    power1_cols <- c(power1_cols, nonspline_cols)
  }

  ## Build derivative matrices one variable at a time, keyed by the raw
  #  expansion names used throughout the fitting code.
  first_derivs <- lapply(colnm_expansions[power1_cols], function(v){
    take_derivative(dat = Cpredictors, var = v, scale = expansion_scales[v])
  })
  names(first_derivs) <- colnm_expansions[power1_cols]
  if(just_first_derivatives){
    return(list('first_derivatives' = first_derivs))
  }

  ## 2nd derivatives
  ## If interactions present
  if ((include_2way_interactions |
       include_3way_interactions |
       include_quadratic_interactions) &
      (length(power1_cols) > 1)){
    second_derivs <- lapply(
      colnm_expansions[power1_cols], function(v) {
        take_interaction_2ndderivative(
          dat = Cpredictors,
          var = v,
          interaction_single_cols,
          interaction_quad_cols,
          triplet_cols,
          colnm_expansions,
          power1_cols,
          power2_cols
        ) +
          take_derivative(dat = Cpredictors,
                          var = v,
                          second = TRUE,
                          scale = expansion_scales[v])
      }
    )
    names(second_derivs) <- colnm_expansions[power1_cols]
    ## No interactions present
  } else {
    second_derivs <- lapply(colnm_expansions[power1_cols], function(v){
      take_derivative(dat = Cpredictors,
                      var = v,
                      second = TRUE,
                      scale = expansion_scales[v])
    })
    names(second_derivs) <- colnm_expansions[power1_cols]
  }

  return(list('first_derivatives' = first_derivs,
              'second_derivatives' = second_derivs))
}

#' Create Smoothing Spline Constraint Matrix
#'
#' @description
#' Constructs constraint matrix \eqn{\textbf{A}} enforcing continuity and smoothness at knot boundaries
#' by constraining function values, derivatives, and interactions between partitions.
#'
#' @param p_expansions Integer; number of columns in basis expansion per partition
#' @param CKnots Matrix; basis expansions evaluated at knot points
#' @param power1_cols Integer vector; indices of linear terms
#' @param power2_cols Integer vector; indices of quadratic terms
#' @param nonspline_cols Integer vector; indices of non-spline terms
#' @param interaction_single_cols Integer vector; indices of linear interaction terms
#' @param interaction_quad_cols Integer vector; indices of quadratic interaction terms
#' @param triplet_cols Integer vector; indices of three-way interaction terms
#' @param K Integer; number of interior knots (\eqn{K+1} partitions)
#' @param include_constrain_fitted Logical; constrain function values at knots
#' @param include_constrain_first_deriv Logical; constrain first derivatives at knots
#' @param include_constrain_second_deriv Logical; constrain second derivatives at knots
#' @param include_constrain_interactions Logical; constrain interaction terms at knots
#' @param include_2way_interactions Logical; include two-way interactions
#' @param include_3way_interactions Logical; include three-way interactions
#' @param include_quadratic_interactions Logical; include quadratic interactions
#' @param colnm_expansions Character vector; column names for basis expansions
#' @param expansion_scales Numeric vector; scaling factors for standardization
#'
#' @return Matrix \eqn{\textbf{A}} of constraint coefficients. Columns correspond to
#' constraints, rows to coefficients across all \eqn{K+1} partitions.
#'
#' @keywords internal
#' @export
make_constraint_matrix <- function(p_expansions,
                                   CKnots,
                                   power1_cols,
                                   power2_cols,
                                   nonspline_cols,
                                   interaction_single_cols,
                                   interaction_quad_cols,
                                   triplet_cols,
                                   K,
                                   include_constrain_fitted,
                                   include_constrain_first_deriv,
                                   include_constrain_second_deriv,
                                   include_constrain_interactions,
                                   include_2way_interactions,
                                   include_3way_interactions,
                                   include_quadratic_interactions,
                                   colnm_expansions,
                                   expansion_scales){

  ## First, create a checkered matrix-each column alternates 1s/-1s,
  # 0s on off-diagonals
  checkered = matrix(0, nrow = K, ncol = K+1)

  ## Handle matrix construction based on dimension and CKnots structure
  if(!(any(is.null(rownames(CKnots))))){
    rwnms <- rownames(CKnots)
    if(length(rwnms) == 1){
      if(rwnms == 'CKnots'){
        ## For univariate case: Alternate 1/-1 in adjacent columns
        for(j in 1:(K+1)){
          checkered[,j] = -(2*(j %% 2) - 1)
        }
        ## Zero out non-adjacent entries
        for(i in 1:K){
          checkered[i,-c(i,i+1)] = 0
        }
      }
    } else {
      ## For multivariate case: Set pairwise constraints from rownames
      # Rownames code the neighbors i.e. 2_4 implies partitions 2 and 4
      # are neighbors
      for(i in 1:nrow(CKnots)){
        rown <- unlist(strsplit(rwnms[i], "_"))
        col1 <- as.numeric(rown[1])
        col2 <- as.numeric(rown[2])
        checkered[i,col1] <- 1
        checkered[i,col2] <- -1
      }
    }
  } else {
    ## Default univariate construction when no rownames present
    for(j in 1:(K+1)){
      checkered[,j] = -(2*(j %% 2) - 1)
    }
    for(i in 1:K){
      checkered[i,-c(i,i+1)] = 0
    }
  }

  ## Special case handling for single knot with multiple variables
  # (Fix for K = 1, q > 1)
  if(length(checkered) == 2 & !any(unique(checkered) != 0)){
    checkered[1] <- 1
    checkered[2] <- -1
  }

  ## Expand out the checkered matrix to match dimensions of
  # basis expansion dimensions
  checkered_fitted_expand = checkered[, rep(1:(K+1), each = p_expansions), drop = FALSE]

  ## Expand out the constraints, located at knots
  constrain_fitted <- CKnots[,rep(1:p_expansions, K+1), drop = FALSE] *
    checkered_fitted_expand

  ## When non-spline and spline present, repeat fitted constraint for
  # spline-only
  if(length(nonspline_cols) > 0 & length(power1_cols) > 0){
    CKnots0 <- CKnots
    CKnots0[,nonspline_cols] <- 0
    constrain_fitted0 <- CKnots0[,rep(1:p_expansions, K+1), drop = FALSE] *
      checkered_fitted_expand
    constrain_fitted <- rbind(constrain_fitted, constrain_fitted0)
  }

  ## Zero-out the fitted constraint if not desired
  if(!include_constrain_fitted){
    constrain_fitted <- 0 * constrain_fitted
  }

  ## First derivative, for all numeric variables
  if(include_constrain_first_deriv){
    first_derivs <- lapply(colnames(CKnots)[power1_cols], function(v){
      take_derivative(dat = CKnots, var = v, scale = expansion_scales[v])
    })
    first_deriv = Reduce('rbind',
                         first_derivs
    )
    constrain_first_deriv <- first_deriv[, rep(1:p_expansions, K+1),
                                         drop = FALSE] *
      checkered_fitted_expand[

        rep(c(1:nrow(checkered_fitted_expand)),
            length(power1_cols)),
      ]
  }

  ## First derivatives for non-spline effects
  # = identical coefficients across partitions
  if(length(nonspline_cols) > 0 &
     include_constrain_first_deriv){
    first_derivs <- lapply(colnames(CKnots)[nonspline_cols], function(v){
      take_derivative(dat = CKnots,
                      var = v,
                      scale = expansion_scales[v])
    })
    first_derivs <- Reduce('rbind',
                           first_derivs
    )

    ## New constraint
    new_constr <- first_derivs[, rep(1:p_expansions, K+1), drop = FALSE] *
      checkered_fitted_expand[
        rep(c(1:nrow(checkered_fitted_expand)),
            length(nonspline_cols)),
      ]
    new_constr <- unique(new_constr[rowSums(abs(new_constr)) > 0,,drop=FALSE])

    ## Constrain first derivative
    if(length(power1_cols) > 0){
      constrain_first_deriv <- rbind(
        constrain_first_deriv,
        new_constr
      )
    } else {
      constrain_first_deriv <- new_constr
    }
  }

  ## Second derivative, for all numeric variables
  if(include_constrain_second_deriv){
    second_deriv = Reduce('rbind',
                          lapply(colnames(CKnots)[c(power1_cols,
                                                    nonspline_cols)],
                                 function(v){
                                   take_derivative(dat = CKnots,
                                                   var = v,
                                                   second = TRUE,
                                                   scale = expansion_scales[v])
                                 })
    )
    constrain_second_deriv <- second_deriv[, rep(1:p_expansions, K+1),
                                           drop = FALSE] *
      checkered_fitted_expand[
        rep(c(1:nrow(checkered_fitted_expand)),
            length(c(power1_cols, nonspline_cols))),
      ]
  }

  ## If interactions present
  if ((include_2way_interactions |
       include_3way_interactions |
       include_quadratic_interactions) &
      (length(c(power1_cols, nonspline_cols)) > 1) &
      include_constrain_interactions){
    second_deriv_interaction <-
      Reduce('rbind',
             lapply(colnames(CKnots)[
               c(power1_cols, nonspline_cols)],
               function(v) {
                 if(v %in% nonspline_cols){
                   take_derivative(dat = CKnots,
                                   var = v,
                                   scale = expansion_scales[v],
                                   second = TRUE)
                 } else {
                   take_interaction_2ndderivative(
                     dat = CKnots,
                     var = v,
                     interaction_single_cols,
                     interaction_quad_cols,
                     triplet_cols,
                     colnm_expansions,
                     power1_cols,
                     power2_cols
                   )
                 }
               }))
    constrain_second_deriv_interactions <-
      second_deriv_interaction[, rep(1:p_expansions, K+1), drop = FALSE] *
      checkered_fitted_expand[
        rep(c(1:nrow(checkered_fitted_expand)),
            length(c(power1_cols,nonspline_cols))),
      ]
  }

  ## Combine the constraints into a single full-rank matrix A
  if ((include_2way_interactions |
       include_3way_interactions |
       include_quadratic_interactions) &
      (length(c(power1_cols, nonspline_cols)) > 1) &
      include_constrain_interactions){
    if (include_constrain_second_deriv) {
      if (include_constrain_first_deriv) {
        A = t(rbind(
          constrain_fitted,
          constrain_first_deriv,
          constrain_second_deriv +
            constrain_second_deriv_interactions
        ))
      } else {
        A = t(rbind(constrain_fitted,
                    constrain_second_deriv +
                      constrain_second_deriv_interactions))
      }
    } else {
      if (include_constrain_first_deriv) {
        A = t(rbind(constrain_fitted,
                    constrain_first_deriv,
                    constrain_second_deriv_interactions))
      } else {
        A = t(rbind(constrain_fitted,
                    constrain_second_deriv_interactions))
      }
    }
  } else {
    if (include_constrain_second_deriv) {
      if (include_constrain_first_deriv) {
        A = t(rbind(
          constrain_fitted,
          constrain_first_deriv,
          constrain_second_deriv
        ))
      } else {
        A = t(rbind(constrain_fitted,
                    constrain_second_deriv))
      }
    } else {
      if (include_constrain_first_deriv) {
        A = t(rbind(constrain_fitted,
                    constrain_first_deriv))
      } else {
        ## For compatibility
        A = t(rbind(constrain_fitted,
                    constrain_fitted))
      }
    }
  }

  ## If all 0s, return the 0s in conformable form
  if(!any(A != 0)){
    return(cbind(rep(0, (K+1)*p_expansions)))
  } else {
    ## Otherwise, return A
    return(A)
  }
}

#' Test if Vector is Binary
#'
#' @param x Vector to test
#' @return Logical indicating if x has at most 2 unique values
#'
#' @keywords internal
#' @export
is_binary <- function(x){
  if(length(unique(x)) > 2){
    return(FALSE)
  }
  TRUE
}


#' Compute Derivative of Inverse-Information Matrix G with Respect to Lambda
#'
#' @description
#' Calculates the derivative of the inverse-information matrix
#' \eqn{\textbf{G}} with respect to the smoothing parameter \eqn{\lambda},
#' supporting both shared and partition-specific penalties.
#'
#' @param G A list of inverse-information matrices \eqn{\textbf{G}} for each
#'   partition
#' @param dPenalty_dlambda Derivative of the shared penalty matrix with
#'   respect to \eqn{\lambda}.
#' @param K Number of partitions minus 1 (\eqn{K})
#' @param unique_penalty_per_partition Logical indicating partition-specific penalties
#' @param dPenalty_partition_list Optional list of derivatives of the
#'   partition-specific penalty matrices with respect to \eqn{\lambda}.
#' @param parallel Logical to enable parallel processing
#' @param cl Cluster object for parallel computation
#' @param chunk_size Size of chunks for parallel processing
#' @param num_chunks Number of chunks
#' @param rem_chunks Remainder chunks
#'
#' @return
#' A list of derivative matrices \eqn{d\textbf{G}/d\lambda} for each partition
#'
#' @keywords internal
#' @export
compute_dG_dlambda <- function(G,
                               dPenalty_dlambda,
                               K,
                               unique_penalty_per_partition,
                               dPenalty_partition_list,
                               parallel,
                               cl,
                               chunk_size,
                               num_chunks,
                               rem_chunks) {
  unique_penalty_per_partition <- isTRUE(unique_penalty_per_partition)

  if(!unique_penalty_per_partition){
    penalty_derivative <- dPenalty_dlambda
  } else {
    penalty_derivative <- lapply(dPenalty_partition_list, function(dP_k){
      dPenalty_dlambda + dP_k
    })
  }

  ## [2026-02-12] replaced %**% with crossprod/tcrossprod
  if(parallel & !is.null(cl)) {
    ## Handle remainder chunks first
    if(rem_chunks > 0) {
      rem_indices <- num_chunks*chunk_size + 1:rem_chunks
      rem <- lapply(rem_indices, function(k) {
        if(unique_penalty_per_partition){
          -crossprod(t(G[[k]]), crossprod(t(penalty_derivative[[k]]), G[[k]]))
        } else {
          -crossprod(t(G[[k]]), crossprod(t(penalty_derivative), G[[k]]))
        }
      })
    } else {
      rem <- list()
    }

    ## Process main chunks in parallel
    result <- c(
      Reduce("c",parallel::parLapply(cl, 1:num_chunks, function(chunk) {
        inds <- (chunk - 1)*chunk_size + 1:chunk_size
        lapply(inds, function(k) {
          if(unique_penalty_per_partition){
            -crossprod(t(G[[k]]), crossprod(t(penalty_derivative[[k]]), G[[k]]))
          } else {
            -crossprod(t(G[[k]]), crossprod(t(penalty_derivative), G[[k]]))
          }
        })
      })),
      rem
    )

  } else {
    ## Sequential computation
    result <- lapply(1:(K+1), function(k){
      if(unique_penalty_per_partition){
        -crossprod(t(G[[k]]), crossprod(t(penalty_derivative[[k]]), G[[k]]))
      } else {
        -crossprod(t(G[[k]]), crossprod(t(penalty_derivative), G[[k]]))
      }
    })
  }

  return(result)
}

## The tuning hat-diagonal helpers below are documented separately.
#' Compute the Diagonal of the Tuning Hat Matrix
#'
#' @description
#' Computes the per-observation diagonal of the constrained hat matrix used
#' during tuning, without forming the full \eqn{N \times N} matrix.
#'
#' @param X List of partition-specific design matrices.
#' @param G List of partition-specific \eqn{\mathbf{G}_k} matrices.
#' @param A Constraint matrix.
#' @param AGAInv Inverse of \eqn{\mathbf{A}^{\top}\mathbf{G}\mathbf{A}}.
#' @param K Number of interior knots.
#' @param p_expansions Number of columns per partition.
#' @param eps Numeric floor used to keep \eqn{1 - h_{ii}} away from zero.
#'
#' @return List with \code{h_diag} and \code{C_list}, both partition-wise.
#'
#' @keywords internal
.compute_hat_diag <- function(X,
                              G,
                              A,
                              AGAInv,
                              K,
                              p_expansions,
                              eps = 1e-7) {

  n_partitions <- K + 1
  h_diag <- vector("list", n_partitions)
  C_list <- vector("list", n_partitions)

  for (k in seq_len(n_partitions)) {
    rows <- (k - 1) * p_expansions + seq_len(p_expansions)
    A_k <- A[rows, , drop = FALSE]
    V_k <- G[[k]] %**% A_k
    C_k <- G[[k]] - V_k %**% AGAInv %**% t(V_k)
    h_k <- rowSums((X[[k]] %**% C_k) * X[[k]])
    h_diag[[k]] <- pmin(pmax(h_k, 0), 1 - eps)
    C_list[[k]] <- 0.5 * (C_k + t(C_k))
  }

  list(h_diag = h_diag, C_list = C_list)
}


#' Compute the Derivative of the Tuning Hat Matrix Diagonal
#'
#' @description
#' Computes \eqn{d h_{ii} / d\lambda_1} partition-wise for the exact
#' leave-one-out tuning criterion.
#'
#' @param X List of partition-specific design matrices.
#' @param G List of partition-specific \eqn{\mathbf{G}_k} matrices.
#' @param dG_dlambda List of derivatives \eqn{d\mathbf{G}_k / d\lambda_1}.
#' @param A Constraint matrix.
#' @param AGAInv Inverse of \eqn{\mathbf{A}^{\top}\mathbf{G}\mathbf{A}}.
#' @param K Number of interior knots.
#' @param p_expansions Number of columns per partition.
#'
#' @return List with \code{dh_diag} and \code{dC_list}, both partition-wise.
#'
#' @keywords internal
.compute_dhat_diag <- function(X,
                               G,
                               dG_dlambda,
                               A,
                               AGAInv,
                               K,
                               p_expansions) {

  n_partitions <- K + 1
  V_list <- vector("list", n_partitions)
  dV_list <- vector("list", n_partitions)
  D_theta <- matrix(0, ncol(A), ncol(A))

  for (k in seq_len(n_partitions)) {
    rows <- (k - 1) * p_expansions + seq_len(p_expansions)
    A_k <- A[rows, , drop = FALSE]
    V_k <- G[[k]] %**% A_k
    dV_k <- dG_dlambda[[k]] %**% A_k
    V_list[[k]] <- V_k
    dV_list[[k]] <- dV_k
    D_theta <- D_theta + crossprod(A_k, dV_k)
  }

  W <- AGAInv %**% D_theta %**% AGAInv
  dh_diag <- vector("list", n_partitions)
  dC_list <- vector("list", n_partitions)

  for (k in seq_len(n_partitions)) {
    X_k <- X[[k]]
    V_k <- V_list[[k]]
    dV_k <- dV_list[[k]]
    dC_k <- dG_dlambda[[k]] -
      dV_k %**% AGAInv %**% t(V_k) -
      V_k %**% AGAInv %**% t(dV_k) +
      V_k %**% W %**% t(V_k)
    dC_k <- 0.5 * (dC_k + t(dC_k))

    ## Compute the leverage derivative observation-wise rather than taking
    # diag(X dC X') from the full matrix product. This is algebraically
    # equivalent, but is more numerically stable for the local terms that are
    # heavily reweighted inside the exact LOO gradient.
    q_k <- X_k %**% V_k
    dq_k <- X_k %**% dV_k
    dh_diag[[k]] <- rowSums((X_k %**% dG_dlambda[[k]]) * X_k) -
      2 * rowSums((dq_k %**% AGAInv) * q_k) +
      rowSums((q_k %**% W) * q_k)

    dC_list[[k]] <- dC_k
  }

  list(dh_diag = dh_diag, dC_list = dC_list)
}


#' Evaluate Exact Leave-One-Out Criterion at a Given Penalty Configuration
#'
#' @description
#' Computes the exact leave-one-out tuning criterion for the same transformed
#' linearized problem used elsewhere in the tuning machinery.
#'
#' @inheritParams .compute_gcvu
#'
#' @return List containing the fitted objects plus \code{criterion_value},
#'   \code{LOO_u}, \code{h_diag}, and \code{C_list}.
#'
#' @keywords internal
.compute_loocv <- function(par,
                           log_penalty_vec,
                           env,
                           ...) {
  verbose <- env$verbose
  eps <- 1e-7

  if (verbose) cat("        loocv_fxn start\n")

  wiggle_penalty <- exp(par[1])
  flat_ridge_penalty <- exp(par[2])
  if (env$unique_penalty_per_predictor || env$unique_penalty_per_partition) {
    penalty_vec <- exp(c(par[-c(1:2)]))
  } else {
    penalty_vec <- c()
  }

  if (verbose) cat("        compute_Lambda\n")
  Lambda_list <- compute_Lambda(env$custom_penalty_mat,
                                env$smoothing_spline_penalty,
                                wiggle_penalty,
                                flat_ridge_penalty,
                                env$K,
                                env$p_expansions,
                                env$unique_penalty_per_predictor,
                                env$unique_penalty_per_partition,
                                penalty_vec,
                                env$colnm_expansions,
                                just_Lambda = FALSE)
  Lambda <- Lambda_list[[1]]
  L1 <- Lambda_list[[2]]
  L2 <- Lambda_list[[3]]

  if (verbose) cat("        compute_G_eigen\n")
  schur_corrections <- lapply(seq_len(env$K + 1), function(k) 0)
  G_list <- compute_G_eigen(env$X_gram,
                            Lambda,
                            env$K,
                            env$parallel & env$parallel_eigen,
                            env$cl,
                            env$chunk_size,
                            env$num_chunks,
                            env$rem_chunks,
                            env$family,
                            env$unique_penalty_per_partition,
                            Lambda_list$L_partition_list,
                            keep_G = TRUE,
                            schur_corrections)

  if (verbose) cat("        loocv_fxn fit coefficients\n")
  return_G_getB <- TRUE
  B_list <- .fit_coefficients(G_list, Lambda,
                              Lambda_list$L_partition_list,
                              env, return_G_getB, ...)
  G_list <- B_list$G_list
  B <- B_list$B

  if (verbose) cat("        loocv_fxn AGAmult_wrapper\n")
  AGAInv <- invert(AGAmult_wrapper(G_list$G,
                                   env$A,
                                   env$K,
                                   env$p_expansions,
                                   env$R_constraints,
                                   env$parallel & env$parallel_aga,
                                   env$cl,
                                   env$chunk_size,
                                   env$num_chunks,
                                   env$rem_chunks) +
                     1e-16 * diag(ncol(env$A)))

  if (verbose) cat("        loocv_fxn get predictions\n")
  preds <- .compute_tuning_predictions(env$X, B, env$K,
                                       env$parallel & env$parallel_matmult,
                                       env$cl,
                                       env$chunk_size,
                                       env$num_chunks,
                                       env$rem_chunks)

  if (verbose) cat("        loocv_fxn residuals\n")
  residuals <- .compute_tuning_residuals(env$y, preds, env$delta,
                                         env$family,
                                         env$observation_weights,
                                         env$K, env$order_list,
                                         ...)

  if (verbose) cat("        loocv_fxn hat diagonal\n")
  hat_result <- .compute_hat_diag(env$X,
                                  G_list$G,
                                  env$A,
                                  AGAInv,
                                  env$K,
                                  env$p_expansions,
                                  eps = eps)

  denom_list <- lapply(hat_result$h_diag, function(h_k) pmax(1 - h_k, eps))
  loo_resid_sq <- mapply(function(r_k, denom_k) {
    (r_k / denom_k)^2
  }, residuals, denom_list, SIMPLIFY = FALSE)
  LOO_u <- mean(unlist(loo_resid_sq))

  if (verbose) cat("        loocv_fxn penalization operations\n")
  mp <- .compute_meta_penalty(wiggle_penalty, penalty_vec,
                              env$meta_penalty,
                              env$unique_penalty_per_predictor,
                              env$unique_penalty_per_partition)

  criterion_value <- LOO_u + mp

  if (verbose) cat("        done LOO,", LOO_u, "\n")

  return(list(criterion_value = criterion_value,
              LOO_u          = LOO_u,
              B              = B,
              G_list         = G_list,
              Lambda         = Lambda,
              L1             = L1,
              L2             = L2,
              L_predictor_list = Lambda_list$L_predictor_list,
              L_partition_list = Lambda_list$L_partition_list,
              residuals      = residuals,
              AGAInv         = AGAInv,
              h_diag         = hat_result$h_diag,
              C_list         = hat_result$C_list))
}


#' Compute Closed-Form Gradient of Exact Leave-One-Out Criterion
#'
#' @description
#' Computes the analytical gradient of the exact leave-one-out tuning
#' criterion with respect to the log penalty parameters.
#'
#' @param par Numeric vector; log-scale penalty parameters.
#' @param log_penalty_vec Numeric vector; log-scale predictor/partition
#'   penalties.
#' @param outlist List or NULL; cached results from \code{.compute_loocv()}.
#' @param env List; tuning environment.
#' @param ... Additional arguments passed to fitting functions.
#'
#' @return List with \code{criterion_value}, \code{gradient}, and
#'   \code{outlist}.
#'
#' @details
#' The implementation below keeps the exact LOO gradient fully analytic,
#' including the leverage derivative. In empirical diagnostics, the
#' observation-wise leverage component can be numerically delicate even when
#' the aggregate gradient remains informative; see the developer diagnostics in
#' \code{inst/diagnostics/gradient_diagnostics.R} for inspection tools.
#'
#' @keywords internal
.compute_loocv_gradient <- function(par,
                                    log_penalty_vec,
                                    outlist = NULL,
                                    env,
                                    ...) {
  verbose <- env$verbose
  eps <- 1e-7

  if (verbose) cat("        loo gradient start\n")

  wiggle_penalty <- exp(par[1])
  flat_ridge_penalty <- exp(par[2])
  if (env$unique_penalty_per_predictor || env$unique_penalty_per_partition) {
    penalty_vec <- exp(c(par[-c(1:2)]))
  } else {
    penalty_vec <- c()
  }

  lambda_1 <- wiggle_penalty
  lambda_2 <- flat_ridge_penalty

  if (any(is.null(outlist))) {
    outlist <- .compute_loocv(par, log_penalty_vec, env, ...)
  }

  if (verbose) cat("        GhalfXy_temp_list \n")
  GhalfXy_temp_list <- compute_GhalfXy_temp_wrapper(
    outlist$G_list$G,
    outlist$G_list$Ghalf,
    env$A,
    outlist$AGAInv,
    env$Xy,
    env$p_expansions,
    env$K,
    env$parallel & env$parallel_aga,
    env$cl,
    env$chunk_size,
    env$num_chunks,
    env$rem_chunks)
  GhalfXy_temp <- GhalfXy_temp_list[[1]]
  AGAInvAGXy <- GhalfXy_temp_list[[2]]

  if (verbose) cat("        compute_dG_dlambda \n")
  dPenalty_dlambda1 <- outlist$Lambda / lambda_1
  dPenalty_partition_dlambda1 <- if (env$unique_penalty_per_partition) {
    lapply(outlist$L_partition_list, function(L_k) L_k / lambda_1)
  } else {
    list()
  }
  dG_dlambda <- compute_dG_dlambda(outlist$G_list$G,
                                   dPenalty_dlambda1,
                                   env$K,
                                   env$unique_penalty_per_partition,
                                   dPenalty_partition_dlambda1,
                                   env$parallel & env$parallel_matmult,
                                   env$cl,
                                   env$chunk_size,
                                   env$num_chunks,
                                   env$rem_chunks)

  if (verbose) cat("        compute_dGhalf \n")
  dGhalf <- compute_dGhalf(dG_dlambda,
                           env$p_expansions,
                           env$K,
                           env$parallel & env$parallel_eigen,
                           env$cl,
                           env$chunk_size,
                           env$num_chunks,
                           env$rem_chunks)

  if (verbose) cat("        compute_dG_u_dlambda_xy \n")
  dG_u_dlambda1_Xyr <- compute_dG_u_dlambda_xy(
    AGAInvAGXy,
    outlist$AGAInv,
    outlist$G_list$G,
    env$A,
    dG_dlambda,
    env$p_expansions,
    env$R_constraints,
    env$K,
    env$Xy,
    outlist$G_list$Ghalf,
    dGhalf,
    GhalfXy_temp,
    env$parallel & env$parallel_matmult,
    env$cl,
    env$chunk_size,
    env$num_chunks,
    env$rem_chunks)

  dPenalty_dlambda2 <- (lambda_1 / lambda_2) * outlist$L2
  dPenalty_partition_dlambda2 <- if (env$unique_penalty_per_partition) {
    lapply(outlist$L_partition_list, function(L_k) 0 * L_k)
  } else {
    list()
  }
  dG_dlambda2 <- compute_dG_dlambda(outlist$G_list$G,
                                    dPenalty_dlambda2,
                                    env$K,
                                    env$unique_penalty_per_partition,
                                    dPenalty_partition_dlambda2,
                                    env$parallel & env$parallel_matmult,
                                    env$cl,
                                    env$chunk_size,
                                    env$num_chunks,
                                    env$rem_chunks)
  dGhalf2 <- compute_dGhalf(dG_dlambda2,
                            env$p_expansions,
                            env$K,
                            env$parallel & env$parallel_eigen,
                            env$cl,
                            env$chunk_size,
                            env$num_chunks,
                            env$rem_chunks)
  dG_u_dlambda2_Xyr <- compute_dG_u_dlambda_xy(
    AGAInvAGXy,
    outlist$AGAInv,
    outlist$G_list$G,
    env$A,
    dG_dlambda2,
    env$p_expansions,
    env$R_constraints,
    env$K,
    env$Xy,
    outlist$G_list$Ghalf,
    dGhalf2,
    GhalfXy_temp,
    env$parallel & env$parallel_matmult,
    env$cl,
    env$chunk_size,
    env$num_chunks,
    env$rem_chunks)

  if (verbose) cat("        compute_dhat_diag \n")
  dhat_result <- .compute_dhat_diag(env$X,
                                    outlist$G_list$G,
                                    dG_dlambda,
                                    env$A,
                                    outlist$AGAInv,
                                    env$K,
                                    env$p_expansions)
  dhat_result2 <- .compute_dhat_diag(env$X,
                                     outlist$G_list$G,
                                     dG_dlambda2,
                                     env$A,
                                     outlist$AGAInv,
                                     env$K,
                                     env$p_expansions)

  ## .compute_loocv() clips h_diag into [0, 1 - eps]. The derivative used in
  ## the LOO gradient must respect that same clipping; once an observation is
  ## on a boundary, the derivative of the clipped value is zero.
  dh_diag_clipped1 <- mapply(
    function(h_k, dh_k) dh_k * ((h_k > 0) & (h_k < 1 - eps)),
    outlist$h_diag,
    dhat_result$dh_diag,
    SIMPLIFY = FALSE
  )
  dh_diag_clipped2 <- mapply(
    function(h_k, dh_k) dh_k * ((h_k > 0) & (h_k < 1 - eps)),
    outlist$h_diag,
    dhat_result2$dh_diag,
    SIMPLIFY = FALSE
  )

  dr_list <- vector("list", env$K + 1)
  for (k in seq_len(env$K + 1)) {
    rows <- (k - 1) * env$p_expansions + seq_len(env$p_expansions)
    q_k <- dG_u_dlambda1_Xyr[rows, , drop = FALSE]
    dr_list[[k]] <- -c(env$X[[k]] %**% q_k)
  }

  dLOO_dlambda1 <- (2 / env$N_obs) * sum(unlist(mapply(
    function(r_k, h_k, dr_k, dh_k) {
      denom_k <- pmax(1 - h_k, eps)
      w_k <- r_k / denom_k
      (w_k / denom_k) * (dr_k + w_k * dh_k)
    },
    outlist$residuals,
    outlist$h_diag,
    dr_list,
    dh_diag_clipped1,
    SIMPLIFY = FALSE
  )))

  dr_list2 <- vector("list", env$K + 1)
  for (k in seq_len(env$K + 1)) {
    rows <- (k - 1) * env$p_expansions + seq_len(env$p_expansions)
    q_k <- dG_u_dlambda2_Xyr[rows, , drop = FALSE]
    dr_list2[[k]] <- -c(env$X[[k]] %**% q_k)
  }

  dLOO_dlambda2 <- (2 / env$N_obs) * sum(unlist(mapply(
    function(r_k, h_k, dr_k, dh_k) {
      denom_k <- pmax(1 - h_k, eps)
      w_k <- r_k / denom_k
      (w_k / denom_k) * (dr_k + w_k * dh_k)
    },
    outlist$residuals,
    outlist$h_diag,
    dr_list2,
    dh_diag_clipped2,
    SIMPLIFY = FALSE
  )))

  ## The exact LOO gradient is kept fully analytic here. Empirically, the
  # observation-wise leverage derivative can be numerically delicate even when
  # the aggregated tuning gradient remains useful; see diagnostics for caveats.
  log_wiggle_gradient <- dLOO_dlambda1 * lambda_1
  log_flat_gradient <- dLOO_dlambda2 * lambda_2

  gradient <- cbind(c(log_wiggle_gradient,
                      log_flat_gradient))

  if (env$unique_penalty_per_predictor) {
    predictor_penalties <- penalty_vec[grep("predictor", names(penalty_vec))]
    predictor_penalty_gradient <- sapply(
      seq_along(predictor_penalties), function(j) {
        mean(diag(outlist$L_predictor_list[[j]])) /
          mean(diag(outlist$Lambda)) *
          log_wiggle_gradient
      })
    gradient <- cbind(c(c(gradient), predictor_penalty_gradient))
  }

  if (env$unique_penalty_per_partition) {
    partition_penalties <- penalty_vec[grep("partition", names(penalty_vec))]
    partition_penalty_gradient <- sapply(
      seq_along(partition_penalties), function(j) {
        mean(diag(outlist$L_partition_list[[j]])) /
          mean(diag(outlist$Lambda +
                      outlist$L_partition_list[[j]])) *
          log_wiggle_gradient
      })
    gradient <- cbind(c(gradient, partition_penalty_gradient))
  }

  regularizer <- .compute_meta_penalty_gradient(
    wiggle_penalty, penalty_vec, env$meta_penalty,
    env$unique_penalty_per_predictor,
    env$unique_penalty_per_partition)

  if (verbose) cat("        loo gradient end \n")

  return(list(criterion_value = outlist$criterion_value,
              LOO_u = outlist$LOO_u,
              gradient = c(gradient) + regularizer,
              outlist = outlist))
}


compute_dW_dlambda_wrapper <- function(G,
                                       A,
                                       GXX,
                                       Ghalf,
                                       dG_dlambda,
                                       dGhalf_dlambda,
                                       AGAInv,
                                       nc,
                                       K,
                                       parallel,
                                       cl,
                                       chunk_size,
                                       num_chunks,
                                       rem_chunks) {

  if(parallel & !is.null(cl)) {
    # Pre-calculate total size and indices
    total_chunks <- num_chunks + (rem_chunks > 0)
    chunk_indices <- vector("list", total_chunks)

    # Create all indices at once
    for(i in 1:num_chunks) {
      chunk_indices[[i]] <- (i-1)*chunk_size + 1:chunk_size
    }
    if(rem_chunks > 0) {
      chunk_indices[[total_chunks]] <- num_chunks*chunk_size + 1:rem_chunks
    }

    # Pre-compute dGXX for all partitions in one step
    dGXX <- matmult_block_diagonal_cpp(dG_dlambda, GXX, K)

    # Compute all traces and corrections in a single parallel operation
    results <- parallel::parLapply(cl, chunk_indices, function(inds) {
      start_idx <- min(inds) - 1

      # Compute initial trace
      trace_part <- sum(sapply(inds, function(k)
        sum(diag(crossprod(t(dG_dlambda[[k]]), GXX[[k]])))))

      # Get relevant chunk of A
      A_chunk <- A[(start_idx*nc + 1):min((max(inds))*nc, nrow(A)), ]
      len_inds <- length(inds) - 1

      # Compute both corrections at once
      correction1 <- compute_trace_correction(
        G[inds], A_chunk, GXX[inds],
        AGAInv, nc, len_inds)

      correction2 <- compute_trace_correction(
        G[inds], A_chunk, dGXX[inds],
        AGAInv, nc, len_inds)

      c(trace_part, correction1, correction2)
    })

    # Efficiently sum results
    all_results <- do.call(rbind, results)
    trace <- sum(all_results[,1])
    correction1 <- sum(all_results[,2])
    correction2 <- sum(all_results[,3])

    return(trace - correction1 - correction2)

  } else {
    return(compute_dW_dlambda(G,
                              A,
                              GXX,
                              Ghalf,
                              dG_dlambda,
                              dGhalf_dlambda,
                              AGAInv,
                              nc,
                              K))
  }
}


#' Calculate Trace of Matrix Product \eqn{\text{trace}(\textbf{X}\textbf{U}\textbf{G}\textbf{X}^{T})}
#'
#' @param G List of G matrices (\eqn{\textbf{G}})
#' @param A Constraint matrix (\eqn{\textbf{A}})
#' @param GXX List of \eqn{\textbf{G}\textbf{X}^{T}\textbf{X}} products
#' @param AGAInv Inverse of \eqn{\textbf{A}^{T}\textbf{G}\textbf{A}}
#' @param nc Number of columns
#' @param nca Number of constraint columns
#' @param K Number of partitions minus 1 (\eqn{K})
#' @param parallel Use parallel processing
#' @param cl Cluster object
#' @param chunk_size Size of parallel chunks
#' @param num_chunks Number of chunks
#' @param rem_chunks Remaining chunks
#'
#' @details
#' Computes \eqn{\text{trace}(\textbf{X}\textbf{U}\textbf{G}\textbf{X}^{T})} where \eqn{\textbf{U} = \textbf{I} - \textbf{G}\textbf{A}(\textbf{A}^{T}\textbf{G}\textbf{A})^{-1}\textbf{A}^{T}}.
#' Handles parallel computation by splitting into chunks.
#'
#' @return Trace value
#'
#' @keywords internal
#' @export
compute_trace_UGXX_wrapper <- function(G,
                                       A,
                                       GXX,
                                       AGAInv,
                                       nc,
                                       nca,
                                       K,
                                       parallel,
                                       cl,
                                       chunk_size,
                                       num_chunks,
                                       rem_chunks) {

  # P <- A %**% AGAInv %**% t(A)
  # U <- -t(matmult_U(P, G, nc, K)) + diag(nrow(P))
  # UGXX <- matmult_U(U, GXX, nc, K)
  # return(sum(diag(UGXX)))

  if(parallel & !is.null(cl)) {
    ## Pre-calculate total size and indices
    total_chunks <- num_chunks + (rem_chunks > 0)
    chunk_indices <- vector("list", total_chunks)

    ## Create all indices at once
    for(i in 1:num_chunks) {
      chunk_indices[[i]] <- (i-1)*chunk_size + 1:chunk_size
    }
    if(rem_chunks > 0) {
      chunk_indices[[total_chunks]] <- num_chunks*chunk_size + 1:rem_chunks
    }

    ## Compute traces in parallel with single parallel::parLapply call
    results <- parallel::parLapply(cl, chunk_indices, function(inds) {
      start_idx <- min(inds) - 1

      ## Compute first trace
      trace_part <- mean(sapply(GXX[inds], function(gxx)rowMeans(t(diag(gxx)))))

      ## Compute correction
      const <- length(inds) * nc
      A_chunk <- A[(start_idx*nc + 1):min((max(inds))*nc, nrow(A)), ]
      correction_part <- compute_trace_correction(G[inds],
                                                  A_chunk / sqrt(const),
                                                  GXX[inds],
                                                  AGAInv,
                                                  nc,
                                                  length(inds)-1)

      c(trace_part, correction_part) * const
    })

    ## Efficiently sum results
    all_results <- do.call(rbind, results)
    trace <- mean(all_results[,1])
    correction <- mean(all_results[,2])

    return(min(max((trace - correction)*nrow(all_results), 0), nc*(K+1)))

  } else {
    trace_part <- rowSums(
      rbind(sapply(GXX, function(gxx)rowSums(t(diag(gxx)))))
    )
    correction_part <-
      compute_trace_correction(G,
                               A,
                               GXX,
                               AGAInv,
                               nc,
                               K)
    return(min(max(trace_part - correction_part, 0), (K + 1)*nc))
  }
}

#' Effective degrees of freedom via trace of the hat matrix
#'
#' Computes \eqn{\mathrm{tr}(\mathbf{X}\mathbf{U}\mathbf{G}\mathbf{X}^\top
#' \mathbf{W}^{1/2}\mathbf{V}^{-1}\mathbf{W}^{1/2})} without forming
#' the \eqn{N \times N} hat matrix, by reducing the problem to
#' \eqn{P \times P} and \eqn{r \times r} operations.
#'
#' @details
#' \subsection{Derivation}{
#' Let \eqn{\mathbf{G} = (\mathbf{X}^\top\mathbf{W}^{1/2}\mathbf{V}^{-1}\mathbf{W}^{1/2}\mathbf{X} + \boldsymbol{\Lambda})^{-1}}
#' and \eqn{\mathbf{U} = \mathbf{I} - \mathbf{G}\mathbf{A}(\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^\top}.
#' By the cyclic property of the trace:
#' \deqn{\mathrm{tr}(\mathbf{X}\mathbf{U}\mathbf{G}\mathbf{X}^\top\mathbf{W}^{1/2}\mathbf{V}^{-1}\mathbf{W}^{1/2}) = \mathrm{tr}(\mathbf{U}\mathbf{G}\,\mathbf{X}^\top\mathbf{W}^{1/2}\mathbf{V}^{-1}\mathbf{W}^{1/2}\mathbf{X})}
#'
#' Substituting \eqn{\mathbf{X}^\top\mathbf{W}^{1/2}\mathbf{V}^{-1}\mathbf{W}^{1/2}\mathbf{X} = \mathbf{G}^{-1} - \boldsymbol{\Lambda}}:
#' \deqn{= \mathrm{tr}(\mathbf{U}\mathbf{G}(\mathbf{G}^{-1} - \boldsymbol{\Lambda})) = \mathrm{tr}(\mathbf{U}) - \mathrm{tr}(\mathbf{U}\mathbf{G}\boldsymbol{\Lambda})}
#'
#' Since \eqn{\mathbf{U}} is idempotent,
#' \eqn{\mathrm{tr}(\mathbf{U}) = \mathrm{rank}(\mathbf{U}) = P - r}
#' where \eqn{r = \mathrm{rank}(\mathbf{A})}. For the second term,
#' expand \eqn{\mathbf{U} = \mathbf{I} - \mathbf{G}\mathbf{A}(\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^\top}:
#' \deqn{\mathrm{tr}(\mathbf{U}\mathbf{G}\boldsymbol{\Lambda}) = \mathrm{tr}(\mathbf{G}\boldsymbol{\Lambda}) - \mathrm{tr}\!\left((\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^\top\mathbf{G}\boldsymbol{\Lambda}\mathbf{G}\mathbf{A}\right)}
#'
#' using the cyclic property on the second term. Both \eqn{\mathbf{G}}
#' and \eqn{\boldsymbol{\Lambda}} are block-diagonal with \eqn{K+1}
#' blocks of dimension \eqn{p \times p}, so:
#' \deqn{\mathrm{tr}(\mathbf{G}\boldsymbol{\Lambda}) = \sum_{k=1}^{K+1}\mathrm{tr}(\mathbf{G}_k\boldsymbol{\Lambda}_k)}
#' and \eqn{\mathbf{G}\boldsymbol{\Lambda}\mathbf{G}} is also block-diagonal
#' with blocks \eqn{\mathbf{G}_k\boldsymbol{\Lambda}_k\mathbf{G}_k}.
#'
#' Combining:
#' \deqn{\mathrm{tr}(\mathbf{X}\mathbf{U}\mathbf{G}\mathbf{X}^\top\mathbf{W}^{1/2}\mathbf{V}^{-1}\mathbf{W}^{1/2}) = (P - r) - \sum_{k=1}^{K+1}\mathrm{tr}(\mathbf{G}_k\boldsymbol{\Lambda}_k) + \mathrm{tr}\!\left((\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^\top\mathbf{G}\boldsymbol{\Lambda}\mathbf{G}\mathbf{A}\right)}
#'
#' The correction term
#' \eqn{\mathrm{tr}((\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^\top\mathbf{G}\boldsymbol{\Lambda}\mathbf{G}\mathbf{A})}
#' is an \eqn{r \times r} trace and captures the degrees of freedom
#' recovered because the constraints pin certain linear combinations of
#' penalized coefficients, removing them from estimation. When
#' \eqn{\boldsymbol{\Lambda} = \mathbf{0}} (no penalty), both
#' \eqn{\mathrm{tr}(\mathbf{G}\boldsymbol{\Lambda})} and the correction
#' vanish, giving \eqn{\mathrm{edf} = P - r}. When \eqn{\mathbf{U} = \mathbf{I}}
#' (no constraints), \eqn{r = 0} and the correction vanishes, giving
#' \eqn{\mathrm{edf} = P - \mathrm{tr}(\mathbf{G}\boldsymbol{\Lambda})}.
#'
#' The result is invariant to the correlation structure \eqn{\mathbf{V}}
#' and GLM weights \eqn{\mathbf{W}}, which enter only through
#' \eqn{\mathbf{G}}.
#' }
#'
#' \subsection{Partition-specific penalties}{
#' When \code{unique_penalty_per_partition = TRUE}, each partition \eqn{k}
#' has an additional penalty matrix \eqn{\mathbf{L}_k} from
#' \code{L_partition_list}, so the effective per-partition penalty becomes
#' \eqn{\boldsymbol{\Lambda}_k = \boldsymbol{\Lambda} + \mathbf{L}_k}.
#' All block-wise traces and products use \eqn{\boldsymbol{\Lambda}_k}
#' in place of the shared \eqn{\boldsymbol{\Lambda}}.
#' }
#'
#' \subsection{Computational cost}{
#' The partition-wise traces \eqn{\mathrm{tr}(\mathbf{G}_k\boldsymbol{\Lambda}_k)}
#' cost \eqn{O(p^2)} each and are parallelizable. The correction term
#' requires forming \eqn{\mathbf{G}\boldsymbol{\Lambda}\mathbf{G}\mathbf{A}}
#' (block-diagonal times sparse, \eqn{O((K+1)p^2 r)}) and a single
#' \eqn{r \times r} trace (\eqn{O(r^2)}). The total cost is
#' \eqn{O((K+1)p^2 + r^2)}, compared to \eqn{O(N^2)} or \eqn{O(NP)}
#' for forming and tracing the full hat matrix.
#' }
#'
#' @param G List of \eqn{K+1} matrices \eqn{\mathbf{G}_k}, each \eqn{p \times p}.
#' @param Lambda Base penalty matrix \eqn{\boldsymbol{\Lambda}}, \eqn{p \times p},
#'   shared across all partitions. When \code{unique_penalty_per_partition = FALSE},
#'   this is the full per-block penalty. When \code{TRUE}, the effective penalty
#'   for partition \eqn{k} is \eqn{\boldsymbol{\Lambda} + \mathbf{L}_k}.
#' @param A Constraint matrix, \eqn{P \times r}.
#' @param AGAInv Precomputed \eqn{(\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}},
#'   \eqn{r \times r}.
#' @param nc Number of columns per partition block (\eqn{p}).
#' @param nca Number of constraint columns (\eqn{r}).
#' @param K Number of interior knots (so there are \eqn{K+1} partitions).
#' @param parallel Logical; use parallel processing.
#' @param cl Cluster object for parallel computation.
#' @param chunk_size Size of parallel chunks.
#' @param num_chunks Number of full-size chunks.
#' @param rem_chunks Number of remaining partitions in the final chunk.
#' @param unique_penalty_per_partition Logical; if \code{TRUE}, each partition
#'   receives an additional penalty from \code{L_partition_list}.
#' @param L_partition_list List of partition-specific penalty matrices
#'   \eqn{\mathbf{L}_k}, each \eqn{p \times p}. Only used when
#'   \code{unique_penalty_per_partition = TRUE}; ignored otherwise.
#'
#' @return Scalar effective degrees of freedom, clamped to \eqn{[0, \, (K+1)p]}.
#'
#' @keywords internal
#' @export
compute_trace_H <- function(G,
                            Lambda,
                            A,
                            AGAInv,
                            nc,
                            nca,
                            K,
                            parallel,
                            cl,
                            chunk_size,
                            num_chunks,
                            rem_chunks,
                            unique_penalty_per_partition,
                            L_partition_list) {

  n_partitions <- K + 1
  P <- nc * n_partitions
  r <- nca

  ## Term 1: rank(U) = P - r
  rank_U <- P - r

  ## Helper: get the effective Lambda block for partition k
  #  When unique_penalty_per_partition, Lambda_k = Lambda + L_partition_list[[k]]
  #  Otherwise Lambda_k = Lambda for all k
  get_Lambda_k <- function(k) {
    if (unique_penalty_per_partition) {
      Lambda + L_partition_list[[k]]
    } else {
      Lambda
    }
  }

  ## Terms 2 and 3 computed jointly per partition
  # Term 2: sum_k tr(G_k Lambda_k)
  # Term 3: B = G Lambda G A (rows assembled block-wise)
  B <- matrix(0, P, r)

  if (parallel & !is.null(cl)) {
    ## Build chunk index lists
    total_chunks <- num_chunks + (rem_chunks > 0)
    chunk_indices <- vector("list", total_chunks)
    for (i in seq_len(num_chunks)) {
      chunk_indices[[i]] <- (i - 1) * chunk_size + seq_len(chunk_size)
    }
    if (rem_chunks > 0) {
      chunk_indices[[total_chunks]] <- num_chunks * chunk_size +
        seq_len(rem_chunks)
    }

    ## Parallel: each chunk computes its partition-wise traces and B rows
    results <- parallel::parLapply(cl, chunk_indices, function(inds) {
      tr_parts <- numeric(length(inds))
      B_chunk <- matrix(0, length(inds) * nc, r)
      for (j in seq_along(inds)) {
        k <- inds[j]
        Lk <- get_Lambda_k(k)
        tr_parts[j] <- sum(G[[k]] * Lk)
        rows_local <- (j - 1) * nc + seq_len(nc)
        rows_global <- (k - 1) * nc + seq_len(nc)
        A_k <- A[rows_global, , drop = FALSE]
        B_chunk[rows_local, ] <- G[[k]] %**% (Lk %**% (G[[k]] %**% A_k))
      }
      list(tr_parts = tr_parts, B_chunk = B_chunk, inds = inds)
    })

    ## Collect parallel results
    tr_GLambda <- 0
    for (res in results) {
      tr_GLambda <- tr_GLambda + sum(res$tr_parts)
      for (j in seq_along(res$inds)) {
        k <- res$inds[j]
        rows_global <- (k - 1) * nc + seq_len(nc)
        rows_local <- (j - 1) * nc + seq_len(nc)
        B[rows_global, ] <- res$B_chunk[rows_local, ]
      }
    }
  } else {
    ## Sequential
    tr_GLambda <- 0
    for (k in seq_len(n_partitions)) {
      Lk <- get_Lambda_k(k)
      tr_GLambda <- tr_GLambda + sum(G[[k]] * Lk)
      rows <- (k - 1) * nc + seq_len(nc)
      A_k <- A[rows, , drop = FALSE]
      B[rows, ] <- G[[k]] %**% (Lk %**% (G[[k]] %**% A_k))
    }
  }

  ## Term 3: tr((A'GA)^{-1} A' G Lambda G A)
  D <- crossprod(A, B)
  tr_correction <- sum(AGAInv * t(D))

  ## Combine
  edf <- rank_U - tr_GLambda + tr_correction

  return(min(max(edf, 0), P))
}

#' Derivative of Constrained Penalized Coefficients with Respect to Lambda
#'
#' Computes \eqn{d(\mathbf{U}\mathbf{G}\mathbf{X}^{T}\mathbf{y})/d\lambda},
#' the sensitivity of the constrained coefficient estimates
#' \eqn{\tilde{\boldsymbol{\beta}} = \mathbf{U}\mathbf{G}\mathbf{X}^T\mathbf{y}}
#' to changes in the smoothing parameter \eqn{\lambda}.
#'
#' @param AGAInv_AGXy Product of \eqn{(\mathbf{A}^{T}\mathbf{G}\mathbf{A})^{-1}} and \eqn{\mathbf{A}^{T}\mathbf{G}\mathbf{X}^{T}\mathbf{y}}
#' @param AGAInv Inverse of \eqn{\mathbf{A}^{T}\mathbf{G}\mathbf{A}}
#' @param G List of \eqn{\mathbf{G}} matrices
#' @param A Constraint matrix \eqn{\mathbf{A}}
#' @param dG_dlambda List of \eqn{d\mathbf{G}/d\lambda} matrices
#' @param nc Number of columns
#' @param nca Number of constraint columns
#' @param K Number of partitions minus 1 (\eqn{K})
#' @param Xy List of \eqn{\mathbf{X}^{T}\mathbf{y}} products
#' @param Ghalf List of \eqn{\mathbf{G}^{1/2}} matrices
#' @param dGhalf List of \eqn{d\mathbf{G}^{1/2}/d\lambda} matrices
#' @param GhalfXy_temp Temporary storage for \eqn{\mathbf{G}^{1/2}\mathbf{X}^{T}\mathbf{y}}
#' @param parallel Use parallel processing
#' @param cl Cluster object
#' @param chunk_size Size of parallel chunks
#' @param num_chunks Number of chunks
#' @param rem_chunks Remaining chunks
#'
#' @details
#' This function is called during GCV/penalty optimization and computes
#' how the constrained coefficient vector changes as the penalty weight
#' varies. Two implementations are provided depending on problem size.
#'
#' \subsection{Derivation}{
#' The constrained estimate is
#' \eqn{\tilde{\boldsymbol{\beta}} = \mathbf{U}\hat{\boldsymbol{\beta}}}
#' where
#' \eqn{\hat{\boldsymbol{\beta}} = \mathbf{G}\mathbf{X}^T\mathbf{y}}
#' and
#' \eqn{\mathbf{U} = \mathbf{I} - \mathbf{G}\mathbf{A}(\mathbf{A}^T\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^T}.
#' Both \eqn{\mathbf{G}} and \eqn{\mathbf{U}} depend on \eqn{\lambda}
#' through \eqn{\mathbf{G} = (\mathbf{X}^T\mathbf{X} + \lambda\boldsymbol{\Lambda})^{-1}}.
#'
#' Differentiating the product \eqn{\mathbf{U}\mathbf{G}\mathbf{X}^T\mathbf{y}}
#' requires the chain rule applied to three \eqn{\lambda}-dependent
#' components:
#' \deqn{\frac{d}{d\lambda}(\mathbf{U}\mathbf{G}\mathbf{X}^T\mathbf{y})}
#' \deqn{= \frac{d\mathbf{U}}{d\lambda}\hat{\boldsymbol{\beta}} + \mathbf{U}\frac{d\hat{\boldsymbol{\beta}}}{d\lambda}}
#'
#' The unconstrained part is straightforward:
#' \eqn{d\hat{\boldsymbol{\beta}}/d\lambda = (d\mathbf{G}/d\lambda)\mathbf{X}^T\mathbf{y}},
#' computed partition-wise since \eqn{\mathbf{G}} is block-diagonal.
#'
#' The constraint projection derivative is more involved. Writing
#' \eqn{\mathbf{C} = (\mathbf{A}^T\mathbf{G}\mathbf{A})^{-1}} and
#' expanding \eqn{d\mathbf{U}/d\lambda} gives three terms (after
#' applying the product rule and the identity
#' \eqn{d\mathbf{C}/d\lambda = -\mathbf{C}(d(\mathbf{A}^T\mathbf{G}\mathbf{A})/d\lambda)\mathbf{C}}):
#' \describe{
#'   \item{term2a}{\eqn{(d\mathbf{G}/d\lambda)\mathbf{A}\mathbf{C}\mathbf{A}^T\hat{\boldsymbol{\beta}}}
#'     direct effect of \eqn{d\mathbf{G}/d\lambda} on the projection}
#'   \item{term2b}{\eqn{\mathbf{G}\mathbf{A}\mathbf{C}(d(\mathbf{A}^T\mathbf{G}\mathbf{A})/d\lambda)\mathbf{C}\mathbf{A}^T\hat{\boldsymbol{\beta}}}
#'     effect through the change in \eqn{(\mathbf{A}^T\mathbf{G}\mathbf{A})^{-1}}}
#'   \item{term2c}{\eqn{\mathbf{G}\mathbf{A}\mathbf{C}\mathbf{A}^T(d\hat{\boldsymbol{\beta}}/d\lambda)}
#'     projection of the unconstrained derivative back through the constraint}
#' }
#'
#' The full derivative is
#' \eqn{d\hat{\boldsymbol{\beta}}/d\lambda} minus the sum of these
#' three correction terms.
#' }
#'
#' \subsection{Large problem path (K >= 10, nc > 4)}{
#' Computes the three correction terms explicitly using the block
#' structure of \eqn{\mathbf{G}} and \eqn{d\mathbf{G}/d\lambda}.
#' The intermediate quantity
#' \eqn{d(\mathbf{A}^T\mathbf{G}\mathbf{A})/d\lambda = \mathbf{A}^T(d\mathbf{G}/d\lambda)\mathbf{A}}
#' is accumulated partition-wise. Shared vectors
#' \eqn{\mathbf{A}\mathbf{C}\mathbf{A}^T\hat{\boldsymbol{\beta}}}
#' and related products are precomputed once and reused across
#' partitions. Parallelism is over chunks of partitions.
#' }
#'
#' \subsection{Small problem path}{
#' Reformulates the derivative using the matrix square root
#' \eqn{\mathbf{G}^{1/2}} and its derivative
#' \eqn{d\mathbf{G}^{1/2}/d\lambda}. The constraint
#' \eqn{\mathbf{A}^T\boldsymbol{\beta} = 0} is imposed via least
#' squares projection: the residuals from regressing
#' \eqn{\mathbf{G}^{1/2}\mathbf{X}^T\mathbf{y}} onto
#' \eqn{\mathbf{G}^{1/2}\mathbf{A}} give the constrained component,
#' and differentiating this projection with respect to \eqn{\lambda}
#' yields the derivative. Uses \code{.lm.fit} for speed with a
#' stabilizing rescaling factor to prevent numerical issues when the
#' constraint matrix is poorly scaled.
#' }
#'
#' @return \eqn{P \times 1} vector of derivatives
#'   \eqn{d\tilde{\boldsymbol{\beta}}/d\lambda}
#'
#' @keywords internal
#' @export
compute_dG_u_dlambda_xy <- function(AGAInv_AGXy,
                                    AGAInv,
                                    G,
                                    A,
                                    dG_dlambda,
                                    nc,
                                    nca,
                                    K,
                                    Xy,
                                    Ghalf,
                                    dGhalf,
                                    GhalfXy_temp,
                                    parallel,
                                    cl,
                                    chunk_size,
                                    num_chunks,
                                    rem_chunks) {

  if((K >= 10) & (nc > 4)){

    ##  ## Large problem path: explicit three-term derivative

    ## Helper function for chunk processing: computes partition-wise
    # GXy = G_k X_k'y_k, dGXy = (dG_k/dl) X_k'y_k, and accumulates
    # A_k'(dG_k/dl)A_k into the r x r matrix AdGA for the chain rule
    # on (A'GA)^{-1}.
    process_chunk <- function(inds, G_chunk, dG_chunk, A_chunk, Xy_chunk) {
      ## Process GXy and dGXy for chunk
      GXy_chunk <- lapply(seq_along(inds), function(i) {
        crossprod(t(G_chunk[[i]]), Xy_chunk[[i]])
      })

      dGXy_chunk <- lapply(seq_along(inds), function(i) {
        crossprod(t(dG_chunk[[i]]), Xy_chunk[[i]])
      })

      ## Process AdGA for chunk: accumulates A_k'(dG_k/dl)A_k
      AdGA_chunk <- Reduce("+", lapply(seq_along(inds), function(i) {
        k <- inds[i]
        start_idx <- (k-1)*nc + 1
        end_idx <- k*nc
        crossprod(A[start_idx:end_idx,],
                  crossprod(t(dG_chunk[[i]]), A[start_idx:end_idx,]))
      }))

      list(GXy = GXy_chunk, dGXy = dGXy_chunk, AdGA = AdGA_chunk)
    }

    if(parallel & !is.null(cl)) {
      ## Handle remainder chunks
      if(rem_chunks > 0) {
        rem_indices <- num_chunks*chunk_size + 1:rem_chunks
        rem_result <- process_chunk(
          rem_indices,
          G[rem_indices],
          dG_dlambda[rem_indices],
          A[(num_chunks*chunk_size)*nc + 1:((rem_chunks)*nc), ],
          Xy[rem_indices]
        )
      } else {
        rem_result <- list(GXy = list(), dGXy = list(), AdGA = 0)
      }

      ## Process main chunks in parallel
      chunk_results <- parallel::parLapply(cl, 1:num_chunks, function(chunk) {
        start_idx <- (chunk - 1) * chunk_size
        inds <- (start_idx + 1):min(start_idx + chunk_size, K+1)

        process_chunk(
          inds,
          G[inds],
          dG_dlambda[inds],
          A[(start_idx*nc + 1):min((start_idx + chunk_size)*nc, nrow(A)), ],
          Xy[inds]
        )
      })

      ## Combine results
      GXy <- Reduce("c", c(lapply(chunk_results, `[[`, "GXy"),
                           rem_result$GXy))
      dGXy <- Reduce("c", c(lapply(chunk_results, `[[`, "dGXy"),
                            rem_result$dGXy))
      AdGA <- Reduce("+", c(lapply(chunk_results, `[[`, "AdGA"),
                            list(rem_result$AdGA)))

    } else {
      ## Sequential computation
      GXy <- Reduce("c", lapply(1:(K+1), function(k)
        crossprod(t(G[[k]]), Xy[[k]])))
      dGXy <- Reduce("c", lapply(1:(K+1), function(k)
        crossprod(t(dG_dlambda[[k]]), Xy[[k]])))
      AdGA <- Reduce("+", lapply(1:(K+1), function(k) {
        start_idx <- (k-1)*nc + 1
        end_idx <- k*nc
        crossprod(A[start_idx:end_idx,],
                  crossprod(t(dG_dlambda[[k]]), A[start_idx:end_idx,]))
      }))
    }

    ## Precompute shared vectors used by all three correction terms:
    #    A_AGAInv_AGXy  = A * C * A' * beta_hat
    #      (the constraint correction applied to the unconstrained estimate)
    #    A_term2b_mat   = A * (-C * AdGA * C) * A' * beta_hat
    #      (effect of dC/dl on the projection, pre-expanded by A)
    #    A_term2c_mat   = A * C * A' * d(beta_hat)/dl
    #      (the constraint correction applied to the unconstrained derivative)
    A_AGAInv_AGXy <- cbind(crossprod(t(A), AGAInv_AGXy))
    A_term2b_mat <- cbind(crossprod(t(A),
                                    crossprod(t(-crossprod(t(AGAInv), AdGA)), AGAInv_AGXy)))
    A_term2c_mat <- crossprod(t(A),
                              cbind(crossprod(t(AGAInv),
                                              crossprod(A, cbind(unlist(dGXy))))))

    ## Compute the three correction terms partition-wise:
    #    term2a_k = (dG_k/dl) * [A C A' beta_hat]_k
    #    term2b_k = G_k * [A (-C (AdGA) C) A' beta_hat]_k
    #    term2c_k = G_k * [A C A' (d beta_hat/dl)]_k
    if(parallel & !is.null(cl)) {
      # Handle remainder chunks
      if(rem_chunks > 0) {
        rem_indices <- num_chunks*chunk_size + 1:rem_chunks

        term2a_rem <- lapply(rem_indices, function(k) {
          crossprod(t(dG_dlambda[[k]]),
                    A_AGAInv_AGXy[(k-1)*nc + 1:nc,,drop=FALSE])
        })

        term2b_rem <- lapply(rem_indices, function(k) {
          idx <- (k-1)*nc + 1:nc
          crossprod(t(G[[k]]), A_term2b_mat[idx,, drop=FALSE])
        })

        term2c_rem <- lapply(rem_indices, function(k) {
          idx <- (k-1)*nc + 1:nc
          crossprod(t(G[[k]]), A_term2c_mat[idx,, drop=FALSE])
        })
      } else {
        term2a_rem <- term2b_rem <- term2c_rem <- list()
      }

      ## Process main chunks in parallel
      chunk_results <- parallel::parLapply(cl, 1:num_chunks, function(chunk) {
        start_idx <- (chunk - 1) * chunk_size
        inds <- (start_idx + 1):min(start_idx + chunk_size, K+1)

        terms <- lapply(inds, function(k){
          t2a <- crossprod(t(dG_dlambda[[k]]),
                           A_AGAInv_AGXy[(k-1)*nc + 1:nc,, drop=FALSE])
          idx <- (k-1)*nc + 1:nc
          t2b <- crossprod(t(G[[k]]), A_term2b_mat[idx,, drop=FALSE])
          t2c <- crossprod(t(G[[k]]), A_term2c_mat[idx,, drop=FALSE])
          list(t2a,t2b,t2c)
        })
        list(term2a = lapply(terms, `[[`, 1),
             term2b = lapply(terms, `[[`, 2),
             term2c = lapply(terms, `[[`, 3))
      })

      term2a <- Reduce("c", c(lapply(chunk_results, `[[`, "term2a"),
                              term2a_rem))
      term2b <- Reduce("c", c(lapply(chunk_results, `[[`, "term2b"),
                              term2b_rem))
      term2c <- Reduce("c", c(lapply(chunk_results, `[[`, "term2c"),
                              term2c_rem))

    } else {
      terms <- lapply(1:(K+1), function(k){
        t2a <- crossprod(t(dG_dlambda[[k]]),
                         A_AGAInv_AGXy[(k-1)*nc + 1:nc,,drop=FALSE])
        idx <- (k-1)*nc + 1:nc
        t2b <- crossprod(t(G[[k]]), A_term2b_mat[idx,, drop=FALSE])
        t2c <- crossprod(t(G[[k]]), A_term2c_mat[idx,, drop=FALSE])
        list(t2a,t2b,t2c)
      })
      term2a <- lapply(terms, `[[`, 1)
      term2b <- lapply(terms, `[[`, 2)
      term2c <- lapply(terms, `[[`, 3)
    }

    term2a <- Reduce("c", term2a)
    term2b <- Reduce("c", term2b)
    term2c <- Reduce("c", term2c)

    ## Final result: unconstrained derivative minus three correction terms
    return(cbind(unlist(dGXy) - (unlist(term2a) +
                                   unlist(term2b) +
                                   unlist(term2c))))
  } else {

    ## ## Small problem path: least-squares formulation via G^{1/2}

    ## The constrained estimate can be written as
    #    beta_tilde = G^{1/2} * residuals(G^{1/2}X'y ~ G^{1/2}A)
    #  because projecting out the column space of G^{1/2}A from
    #  G^{1/2}X'y and then multiplying by G^{1/2} is equivalent to
    #  applying the constraint projection U. Differentiating this
    #  representation with respect to lambda using the product rule on
    #  G^{1/2} and dG^{1/2}/dl yields two residual vectors (resids1
    #  from the derivative of the basis, resids2 from the derivative
    #  of the right-hand side) that combine into the final derivative.

    ## G^{1/2} A and its derivative (block-diagonal times A)
    GhalfA <- Reduce("rbind", GAmult_wrapper(Ghalf, A, K, nc, nca,
                                             parallel, cl, chunk_size,
                                             num_chunks, rem_chunks))

    dGhalfA <- Reduce("rbind", GAmult_wrapper(dGhalf, A, K, nc, nca,
                                              parallel, cl, chunk_size,
                                              num_chunks, rem_chunks))

    ## (dG^{1/2}/dl) X'y (block-diagonal multiply)
    dGhalfXy <- cbind(unlist(
      matmult_block_diagonal(dGhalf, Xy, K, parallel, cl,
                             chunk_size, num_chunks, rem_chunks)))

    ## Stabilize the least squares solve: rescale by the average
    # magnitude of G^{1/2}A to prevent numerical issues when the
    # constraint matrix is poorly conditioned
    comp_stab_sc <- sqrt(mean(abs(GhalfA)))
    comp_stab_sc <- 1*(comp_stab_sc == 0) + comp_stab_sc

    ## resids1: residuals from regressing G^{1/2}X'y onto
    #           G^{1/2}A -- the constrained least-squares component
    resids1 <- do.call('.lm.fit', list(x = GhalfA / comp_stab_sc,
                                       y = GhalfXy_temp / comp_stab_sc)
    )$residuals * comp_stab_sc

    ## resids2: residuals from regressing (dG^{1/2}/dl)X'y onto
    #           (dG^{1/2}/dl)A -- derivative-side correction term
    resids2 <- do.call('.lm.fit', list(x = dGhalfA / comp_stab_sc,
                                       y = dGhalfXy / comp_stab_sc)
    )$residuals * comp_stab_sc

    ## Combine: d(beta_tilde)/dl = (dG^{1/2}/dl)*resids2 + G^{1/2}*resids1
    # applied partition-wise since G^{1/2} and dG^{1/2}/dl are block-diagonal
    if(parallel & !is.null(cl)) {
      if(rem_chunks > 0) {
        rem_indices <- num_chunks * chunk_size + 1:rem_chunks
        b2_rem <- lapply(rem_indices, function(k) {
          crossprod(t(dGhalf[[k]]), cbind(resids2[(k-1)*nc + 1:nc])) +
            crossprod(t(Ghalf[[k]]), cbind(resids1[(k-1)*nc + 1:nc]))
        })
      } else {
        b2_rem <- list()
      }

      b2_result <- c(
        Reduce("c",parallel::parLapply(cl, 1:num_chunks, function(chunk) {
          inds <- (chunk - 1) * chunk_size + 1:chunk_size
          lapply(inds, function(k) {
            crossprod(t(dGhalf[[k]]), cbind(resids2[(k-1)*nc + 1:nc])) +
              crossprod(t(Ghalf[[k]]), cbind(resids1[(k-1)*nc + 1:nc]))
          })
        })),
        b2_rem
      )
      b2 <- unlist(b2_result)
    } else {
      b2 <- Reduce("c", lapply(1:(K+1), function(k) {
        crossprod(t(dGhalf[[k]]), cbind(resids2[(k-1)*nc + 1:nc])) +
          crossprod(t(Ghalf[[k]]), cbind(resids1[(k-1)*nc + 1:nc]))
      }))
    }

    return(cbind(b2))
  }
}


#' Compute Eigenvalues and Related Matrices for G
#'
#' @description
#' Computes partition-wise inverse-information matrices and their matrix square
#' roots from the penalized information matrix via eigendecomposition.
#'
#' @param X_gram List of Gram matrices \eqn{X_k^T W_k X_k} by partition.
#' @param Lambda Penalty matrix \eqn{\Lambda} (shared across partitions).
#' @param K Integer; number of interior knots (partitions = \eqn{K+1}).
#' @param parallel Logical; use parallel processing across partitions.
#' @param cl Cluster object from \code{parallel::makeCluster}.
#' @param chunk_size,num_chunks,rem_chunks Partition distribution parameters.
#' @param family GLM family object.
#' @param unique_penalty_per_partition Logical; if \code{TRUE}, add
#'   partition-specific penalties from \code{L_partition_list}.
#' @param L_partition_list List of partition-specific penalty matrices.
#' @param keep_G Logical; if \code{TRUE}, return the full \eqn{G} matrix.
#' @param schur_corrections List of Schur complement correction matrices.
#'
#' @return A list with components \code{G} (or \code{NULL} blocks when
#'   \code{keep_G = FALSE}), \code{Ghalf}, and optionally \code{GhalfInv}.
#'
#' @keywords internal
#' @export
compute_G_eigen <- function(X_gram, Lambda, K, parallel, cl,
                            chunk_size, num_chunks, rem_chunks,
                            family, unique_penalty_per_partition,
                            L_partition_list, keep_G = TRUE,
                            schur_corrections) {

  ## Helper to compute eigen-based matrix powers efficiently
  compute_from_eigen <- function(eig, keep_G, need_GhalfInv) {
    eigen_values <- eig$values
    eigen_values[eig$values <= 0] <- 1
    inv_eigen_values <- 1/eigen_values
    inv_eigen_values[eig$values <= 0] <- 0
    sqrt_inv <- sqrt(inv_eigen_values)

    Ghalf <- tcrossprod(eig$vectors * rep(sqrt_inv, each = nrow(eig$vectors)),
                        eig$vectors)
    G_out <- NULL
    if(keep_G){
      G_out <- tcrossprod(eig$vectors * rep(inv_eigen_values, each = nrow(eig$vectors)),
                          eig$vectors)
    }
    if(need_GhalfInv){
      GhalfInv <- tcrossprod(eig$vectors / rep(sqrt_inv, each = nrow(eig$vectors)),
                             eig$vectors)
      return(list(G = G_out, Ghalf = Ghalf, GhalfInv = GhalfInv))
    } else {
      return(list(G = G_out, Ghalf = Ghalf))
    }
  }

  ## Gaussian identity only needs G and G^(1/2); the other families also
  #  need G^(-1/2) for transformed-score and variance calculations.
  need_GhalfInv <- (paste0(family)[2] != 'identity') |
    (paste0(family)[1] != 'gaussian')

  if(parallel & !is.null(cl)) {
    if(rem_chunks > 0) {
      rem_indices <- num_chunks * chunk_size + 1:rem_chunks
      rem <- lapply(rem_indices, function(k) {
        if(unique_penalty_per_partition){
          eig <- tryCatch(eigen(X_gram[[k]] + Lambda +
                                  L_partition_list[[k]] + schur_corrections[[k]],
                                symmetric = TRUE), error = function(e) NULL)
        } else {
          eig <- tryCatch(eigen(X_gram[[k]] + Lambda +
                                  schur_corrections[[k]], symmetric = TRUE),
                          error = function(e) NULL)
        }
        if(is.null(eig)) return(list(G = NULL, Ghalf = NULL))
        compute_from_eigen(eig, keep_G, need_GhalfInv)
      })
    } else {
      rem <- list()
    }

    result <- c(
      Reduce("c", parallel::parLapply(cl, 1:num_chunks, function(chunk) {
        inds <- (chunk - 1)*chunk_size + 1:chunk_size
        lapply(inds, function(k) {
          if(unique_penalty_per_partition){
            eig <- tryCatch(eigen(X_gram[[k]] + Lambda +
                                    L_partition_list[[k]] + schur_corrections[[k]],
                                  symmetric = TRUE), error = function(e) NULL)
          } else {
            eig <- tryCatch(eigen(X_gram[[k]] + Lambda +
                                    schur_corrections[[k]], symmetric = TRUE),
                            error = function(e) NULL)
          }
          if(is.null(eig)) return(list(G = NULL, Ghalf = NULL))
          compute_from_eigen(eig, keep_G, need_GhalfInv)
        })
      })),
      rem
    )
  } else {
    result <- lapply(1:(K+1), function(k) {
      if(unique_penalty_per_partition){
        eig <- tryCatch(eigen(X_gram[[k]] + Lambda +
                                L_partition_list[[k]] + schur_corrections[[k]],
                              symmetric = TRUE), error = function(e) NULL)
      } else {
        eig <- tryCatch(eigen(X_gram[[k]] + Lambda +
                                schur_corrections[[k]], symmetric = TRUE),
                        error = function(e) NULL)
      }
      if(is.null(eig)) return(list(G = NULL, Ghalf = NULL))
      compute_from_eigen(eig, keep_G, need_GhalfInv)
    })
  }

  ## Repackage the per-partition eigendecomposition output into the list
  #  structure expected downstream by get_B(), blockfit, and inference.
  if(need_GhalfInv){
    result_processed <- list(
      G = lapply(result, `[[`, "G"),
      Ghalf = lapply(result, `[[`, "Ghalf"),
      GhalfInv = lapply(result, `[[`, "GhalfInv")
    )
  } else {
    result_processed <- list(
      G = lapply(result, `[[`, "G"),
      Ghalf = lapply(result, `[[`, "Ghalf")
    )
  }
  return(result_processed)
}


#' Compute Component \eqn{\textbf{G}^{1/2}\textbf{A}(\textbf{A}^{T}\textbf{G}\textbf{A})^{-1}\textbf{A}^{T}\textbf{G}\textbf{X}^{T}\textbf{y}}
#'
#' @param G List of \eqn{\textbf{G}} matrices
#' @param Ghalf List of \eqn{\textbf{G}^{1/2}} matrices
#' @param A Constraint matrix \eqn{\textbf{A}}
#' @param AGAInv Inverse of \eqn{\textbf{A}^{T}\textbf{G}\textbf{A}}
#' @param Xy List of \eqn{\textbf{X}^{T}\textbf{y}} products
#' @param nc Number of columns
#' @param K Number of partitions minus 1 (\eqn{K})
#' @param parallel Use parallel processing
#' @param cl Cluster object
#' @param chunk_size Size of parallel chunks
#' @param num_chunks Number of chunks
#' @param rem_chunks Remaining chunks
#'
#' @details
#' Computes the least-squares projection component
#' \eqn{\mathbf{G}^{1/2}\mathbf{A}(\mathbf{A}^{T}\mathbf{G}\mathbf{A})^{-1}
#' \mathbf{A}^{T}\mathbf{G}\mathbf{X}^{T}\mathbf{y}} together with the
#' intermediate product
#' \eqn{(\mathbf{A}^{T}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{T}\mathbf{G}\mathbf{X}^{T}\mathbf{y}}
#' for reuse downstream.
#'
#' @return Unnamed two-element list containing the projected result vector and
#'   the intermediate
#'   \eqn{(\mathbf{A}^{T}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{T}\mathbf{G}\mathbf{X}^{T}\mathbf{y}}
#'   product.
#'
#' @keywords internal
#' @export
compute_GhalfXy_temp_wrapper <- function(G,
                                         Ghalf,
                                         A,
                                         AGAInv,
                                         Xy,
                                         nc,
                                         K,
                                         parallel,
                                         cl,
                                         chunk_size,
                                         num_chunks,
                                         rem_chunks) {
  if(parallel & !is.null(cl)) {
    ## First compute AGXy in parallel chunks
    if(rem_chunks > 0) {
      rem_start <- num_chunks * chunk_size + 1
      rem_end <- rem_start + rem_chunks - 1
      AGXy_rem <- compute_AGXy(G[rem_start:rem_end],
                               A,
                               Xy[rem_start:rem_end],
                               nc, K, rem_start-1, rem_end-1)
    } else {
      AGXy_rem <- matrix(0, ncol(A), 1)
    }

    ## Process main chunks for AGXy
    AGXy_chunks <- parallel::parLapply(cl, 1:num_chunks, function(chunk) {
      chunk_start <- (chunk - 1) * chunk_size + 1
      chunk_end <- min(chunk * chunk_size, K+1)
      compute_AGXy(G[chunk_start:chunk_end],
                   A,
                   Xy[chunk_start:chunk_end],
                   nc, K, chunk_start-1, chunk_end-1)
    })

    ## Sum up AGXy results
    AGXy <- Reduce('+', c(AGXy_chunks, list(AGXy_rem)))

    ## Compute AAGAInvAGXy once
    AGAInvAGXy <- AGAInv %**% AGXy
    AAGAInvAGXy <- A %**% AGAInvAGXy

    ## Now compute final blocks in parallel
    if(rem_chunks > 0) {
      rem_start <- num_chunks * chunk_size + 1
      rem_end <- rem_start + rem_chunks - 1
      result_rem <- compute_result_blocks(G[rem_start:rem_end],
                                          Ghalf[rem_start:rem_end],
                                          A,
                                          AAGAInvAGXy,
                                          nc,
                                          rem_start-1,
                                          rem_end-1)
    } else {
      result_rem <- numeric(0)
    }

    ## Process main chunks for final result
    result_chunks <- parallel::parLapply(cl, 1:num_chunks, function(chunk) {
      chunk_start <- (chunk - 1) * chunk_size + 1
      chunk_end <- min(chunk * chunk_size, K+1)
      compute_result_blocks(G[chunk_start:chunk_end],
                            Ghalf[chunk_start:chunk_end],
                            A,
                            AAGAInvAGXy,
                            nc,
                            chunk_start-1,
                            chunk_end-1)
    })

    ## Combine all results
    return(list(
      c(unlist(result_chunks), result_rem),
      AGAInvAGXy
    ))

  } else {

    ## Compute AAGAInvAGXy once
    AGAInvAGXy <- AGAInv %**% compute_AGXy(G,
                                           A,
                                           Xy,
                                           nc,
                                           K,
                                           0,
                                           K)

    ## Now compute final blocks in parallel
    result <- compute_result_blocks(G,
                                    Ghalf,
                                    A,
                                    A %**% AGAInvAGXy,
                                    nc,
                                    0,
                                    K)

    ## Combine all results
    return(list(
      c(unlist(result)),
      AGAInvAGXy
    ))
  }
}


#' Efficient Matrix Multiplication of G and A Matrices
#'
#' @param G List of G matrices
#' @param A Constraint matrix
#' @param K Number of partitions minus 1
#' @param p_expansions Number of columns per partition
#' @param R_constraints Number of constraint columns
#' @param parallel Use parallel processing
#' @param cl Cluster object
#' @param chunk_size Size of parallel chunks
#' @param num_chunks Number of chunks
#' @param rem_chunks Remaining chunks
#'
#' @details
#' Computes G %**% A when G has block diagonal structure and A is a matrix.
#' Processes in parallel chunks if enabled.
#'
#' @return List of matrix products, one per partition
#'
#' @keywords internal
#' @export
GAmult_wrapper <- function(G,
                           A,
                           K,
                           p_expansions,
                           R_constraints,
                           parallel,
                           cl,
                           chunk_size,
                           num_chunks,
                           rem_chunks) {

  if(parallel & !is.null(cl)) {

    ## Handle remainder chunks first
    if(rem_chunks > 0) {
      rem_indices <- num_chunks * chunk_size + 1:rem_chunks
      G_rem <- G[rem_indices]
      rem <- GAmult(G_rem, A[(num_chunks*chunk_size)*p_expansions + 1:((rem_chunks)*p_expansions),],
                    rem_chunks-1, p_expansions, R_constraints)
    } else {
      rem <- list()
    }

    ## Process main chunks in parallel
    result <- c(
      Reduce("c",parallel::parLapply(cl, 1:num_chunks, function(chunk) {
        start_idx <- (chunk - 1) * chunk_size
        G_chunk <- G[(start_idx + 1):min(start_idx + chunk_size, length(G))]
        A_chunk <- A[(start_idx*p_expansions + 1):min((start_idx + chunk_size)*p_expansions,
                                                      nrow(A)), ]
        GAmult(G_chunk,
               A_chunk,
               chunk_size-1,
               p_expansions,
               R_constraints)
      })),
      rem
    )

  } else {
    ## Sequential computation using original C++ function
    result <- GAmult(G, A, K, p_expansions, R_constraints)
  }

  return(result)
}



#' Efficient Matrix Multiplication for \eqn{\textbf{A}^{T}\textbf{G}\textbf{A}}
#'
#' @param G List of G matrices (\eqn{\textbf{G}})
#' @param A Constraint matrix (\eqn{\textbf{A}})
#' @param K Number of partitions minus 1 (\eqn{K})
#' @param p_expansions Number of columns per partition
#' @param R_constraints Number of constraint columns
#' @param parallel Use parallel processing
#' @param cl Cluster object
#' @param chunk_size Chunk size for parallel
#' @param num_chunks Number of chunks
#' @param rem_chunks Remaining chunks
#'
#' @details
#' Computes \eqn{\textbf{A}^{T}\textbf{G}\textbf{A}} efficiently in parallel chunks using \code{AGAmult_chunk()}.
#'
#' @return Matrix product \eqn{\textbf{A}^{T}\textbf{G}\textbf{A}}
#'
#' @keywords internal
#' @export
AGAmult_wrapper <- function(G,
                            A,
                            K,
                            p_expansions,
                            R_constraints,
                            parallel,
                            cl,
                            chunk_size,
                            num_chunks,
                            rem_chunks) {
  if(parallel & !is.null(cl)) {
    # Handle remainder chunks
    if(rem_chunks > 0) {
      rem_start <- num_chunks * chunk_size
      G_rem <- G[(rem_start + 1):(rem_start + rem_chunks)]
      rem_result <- AGAmult_chunk(G_rem,
                                  A,
                                  rem_start,
                                  rem_start + rem_chunks - 1,
                                  p_expansions)
    } else {
      rem_result <- matrix(0, R_constraints, R_constraints)
    }

    # Process main chunks in parallel
    chunk_results <- parallel::parLapply(cl, 1:num_chunks, function(chunk) {
      chunk_start <- (chunk - 1) * chunk_size
      G_chunk <- G[(chunk_start + 1):(chunk_start + chunk_size)]
      AGAmult_chunk(G_chunk, A, chunk_start, chunk_start + chunk_size - 1, p_expansions)
    })

    # Sum all results
    return(Reduce('+', c(chunk_results, list(rem_result))))

  } else {
    return(AGAmult(G, A, K, p_expansions, R_constraints))
  }
}


#' Efficiently Construct U Matrix
#'
#' @param G List of G matrices (\eqn{\textbf{G}})
#' @param A Constraint matrix (\eqn{\textbf{A}})
#' @param K Number of partitions minus 1 (\eqn{K})
#' @param p_expansions Number of columns per partition
#' @param R_constraints Number of constraint columns
#'
#' @return \eqn{\textbf{U}} matrix for constraints
#'
#' @details
#' Computes \eqn{\textbf{U} = \textbf{I} - \textbf{G}\textbf{A}(\textbf{A}^{T}\textbf{G}\textbf{A})^{-1}\textbf{A}^{T}} efficiently, avoiding unnecessary
#' multiplication of blocks of \eqn{\textbf{G}} with all-0 elements.
#'
#' @keywords internal
#' @export
get_U <- function(G, A, K, p_expansions, R_constraints){
  AGAInv <- invert(AGAmult(G, A, K, p_expansions, R_constraints))
  I_minus_U <- t(matmult_U(crossprod(t(A),
                          crossprod(t(AGAInv), (-t(A)))), G, p_expansions, K))
  return(I_minus_U + diag(p_expansions*(K+1)))
}

#' Generate Grid Indices Without expand.grid()
#'
#' @param vec_list List of vectors to combine
#' @param indices Indices of combinations to return
#'
#' @details
#' Returns selected combinations from the cartesian product of \code{vec_list}
#' without constructing full \code{expand.grid()} for memory efficiency.
#'
#' @return Data frame of selected combinations
#'
#' @keywords internal
#' @export
expgrid <- function(vec_list, indices) {
  # Calculate the total number of combinations
  total_combinations <- prod(sapply(vec_list, length))

  # Check if any indices are out of bounds
  if (any(indices > total_combinations) || any(indices < 1)) {
    stop("Invalid indices: some are out of bounds")
  }

  # Initialize the result list
  result <- vector("list", length(vec_list))
  names(result) <- names(vec_list)

  # Calculate the "stride" for each vector
  strides <- cumprod(c(1, sapply(vec_list[-length(vec_list)], length)))

  # For each requested index
  for (i in indices) {
    # Convert to 0-based index for easier calculation
    idx <- i - 1

    # For each vector in the predictors list
    for (j in seq_along(vec_list)) {
      # Calculate which element of this vector
      # corresponds to the current combination
      element_idx <- (idx %/% strides[j]) %% length(vec_list[[j]]) + 1
      result[[j]] <- c(result[[j]], vec_list[[j]][element_idx])
    }
  }

  # Convert the result to a data frame
  as.data.frame(result)
}

#' Construct Smoothing Spline Penalty Matrix
#'
#' @description
#' Builds penalty matrix combining smoothing spline and ridge penalties with optional
#' predictor/partition-specific components.
#'
#' @param custom_penalty_mat Matrix; optional custom ridge penalty structure
#' @param L1 Matrix; integrated squared second derivative penalty (\eqn{\textbf{L}_1})
#' @param wiggle_penalty,flat_ridge_penalty Numeric; smoothing and ridge penalty parameters
#' @param K Integer; number of interior knots (\eqn{K})
#' @param p_expansions Integer; number of basis columns per partition
#' @param unique_penalty_per_predictor,unique_penalty_per_partition Logical; enable predictor/partition-specific penalties
#' @param penalty_vec Named numeric; custom penalty values for predictors/partitions
#' @param colnm_expansions Character; column names for linking penalties to predictors
#' @param just_Lambda Logical; return only combined penalty matrix (\eqn{\boldsymbol{\Lambda}})
#'
#' @return List containing Lambda, L1, L2, L_predictor_list, L_partition_list;
#'   or just Lambda if \code{just_Lambda=TRUE} and no partition penalties.
#'
#' @keywords internal
#' @export
compute_Lambda <- function(custom_penalty_mat,
                           L1,
                           wiggle_penalty,
                           flat_ridge_penalty,
                           K,
                           p_expansions,
                           unique_penalty_per_predictor,
                           unique_penalty_per_partition,
                           penalty_vec,
                           colnm_expansions,
                           just_Lambda = TRUE){

  ## Custom or flat ridge penalty
  if(any(!is.null(custom_penalty_mat))){
    L2 <- custom_penalty_mat*flat_ridge_penalty
  } else {
    ## By default, = 1 if diagonal and smoothing spline penalty is 0 for
    # corresponding index (adds a ridge to all terms not penalized by the
    # smoothing spline penalty)
    L2 <- diag(p_expansions)*((diag(L1) == 0)*flat_ridge_penalty)
  }

  ## Combined shared penalty = wiggle * (L1 + L2)
  Lambda <- (L1 + L2)*wiggle_penalty
  L_predictor_list <- list()
  L_partition_list <- list()

  ## A unique penalty for each predictor
  if(unique_penalty_per_predictor){
    predictors <- names(penalty_vec)[grep('predictor', names(penalty_vec))]

    ## Scales the elements of the smoothing penalty uniquely by predictor
    for(j in 1:length(predictors)){
      inds <- grep(substr(predictors[j], 10, nchar(predictors[j])), colnm_expansions)
      L <- L1 * penalty_vec[predictors[j]] * wiggle_penalty

      ## All elements of smoothing penalty NOT touching this predictor are 0
      L[-inds,-inds] <- 0
      Lambda <- Lambda + L
      L_predictor_list <- c(L_predictor_list, list(L))
    }
  }

  ## A unique penalty for each partition
  if(unique_penalty_per_partition){
    partition_names <- names(penalty_vec)[grep('partition', names(penalty_vec))]
    L_partition_list <- lapply(partition_names, function(part){
      L1*penalty_vec[part]*wiggle_penalty
    })
    names(L_partition_list) <- partition_names
  }
  if(just_Lambda & !unique_penalty_per_partition){
    return(Lambda)
  } else{
    return(list(Lambda = Lambda,
                L1 = L1,
                L2 = L2,
                L_predictor_list = L_predictor_list,
                L_partition_list = L_partition_list))
  }
}


#' Compute Eigenvalues and Related Matrices for G
#'
#' @description
#' Computes partition-wise inverse-information matrices and their matrix square roots
#' from the penalized information matrix via eigendecomposition.
#'
#' @param X_gram List of Gram matrices \eqn{X_k^T W_k X_k} by partition.
#' @param Lambda Penalty matrix \eqn{\Lambda} (shared across partitions).
#' @param K Integer; number of interior knots (partitions = \eqn{K+1}).
#' @param parallel Logical; use parallel processing across partitions.
#' @param cl Cluster object from \code{parallel::makeCluster}.
#' @param chunk_size,num_chunks,rem_chunks Partition distribution parameters.
#' @param family GLM family object.
#' @param unique_penalty_per_partition Logical; if \code{TRUE}, add
#'   partition-specific penalties from \code{L_partition_list}.
#' @param L_partition_list List of partition-specific penalty matrices.
#' @param keep_G Logical; if \code{TRUE}, return the full \eqn{G} matrix.
#' @param schur_corrections List of Schur complement correction matrices.
#'
#' @return A list with components \code{G} (or \code{NULL} blocks when
#'   \code{keep_G = FALSE}), \code{Ghalf}, and optionally \code{GhalfInv}.
#'
#' @keywords internal
#' @export
compute_G_eigen <- function(X_gram, Lambda, K, parallel, cl,
                            chunk_size, num_chunks, rem_chunks,
                            family, unique_penalty_per_partition,
                            L_partition_list, keep_G = TRUE,
                            schur_corrections) {

  ## Determine once whether GhalfInv is needed
  need_GhalfInv <- (paste0(family)[2] != 'identity') |
    (paste0(family)[1] != 'gaussian')

  ## Core computation for a single partition k
  compute_one <- function(k) {
    ## Form G_k^{-1} = X^T W X + Lambda [+ L_k] + S_k
    if(unique_penalty_per_partition){
      eig <- tryCatch({
        eigen(X_gram[[k]] +
                Lambda +
                L_partition_list[[k]] +
                schur_corrections[[k]],
              symmetric = TRUE)
      }, error = function(e) NULL)
    } else {
      eig <- tryCatch({
        eigen(X_gram[[k]] +
                Lambda +
                schur_corrections[[k]],
              symmetric = TRUE)
      }, error = function(e) NULL)
    }
    if(is.null(eig)) {
      return(list(G = NULL, Ghalf = NULL, GhalfInv = NULL))
    }

    ## Clamp non-positive eigenvalues for pseudoinverse treatment
    eigen_values <- eig$values
    eigen_values[eig$values <= 0] <- 1
    inv_eigen_values <- 1/eigen_values
    inv_eigen_values[eig$values <= 0] <- 0
    sqrt_inv_eigen_values <- sqrt(inv_eigen_values)

    ## G^{1/2} = V diag(1/sqrt(d_i)) V^T
    Ghalf <- eig$vectors %**% (t(eig$vectors) * sqrt_inv_eigen_values)

    ## G = V diag(1/d_i) V^T (only if requested)
    if(keep_G){
      G <- eig$vectors %**% (t(eig$vectors) * inv_eigen_values)
    } else {
      G <- NULL
    }

    ## G^{-1/2} = V diag(sqrt(d_i)) V^T (only for non-Gaussian/non-identity)
    if(need_GhalfInv){
      sqrt_eigen_values <- sqrt(pmax(eig$values, 0))
      GhalfInv <- eig$vectors %**% (t(eig$vectors) * sqrt_eigen_values)
      return(list(G = G,
                  Ghalf = Ghalf,
                  GhalfInv = GhalfInv))
    } else {
      return(list(G = G,
                  Ghalf = Ghalf))
    }
  }

  if(parallel & !is.null(cl)) {
    if(rem_chunks > 0) {
      rem_indices <- num_chunks * chunk_size + 1:rem_chunks
      rem <- lapply(rem_indices, compute_one)
    } else {
      rem <- list()
    }

    result <- c(
      Reduce("c", parallel::parLapply(cl, 1:num_chunks, function(chunk) {
        inds <- (chunk - 1)*chunk_size + 1:chunk_size
        lapply(inds, compute_one)
      })),
      rem
    )
  } else {
    result <- lapply(1:(K+1), compute_one)
  }

  ## Reorganize from list-of-partitions to list-of-matrix-types
  if(need_GhalfInv){
    result_processed <- list(
      G = lapply(result, `[[`, "G"),
      Ghalf = lapply(result, `[[`, "Ghalf"),
      GhalfInv = lapply(result, `[[`, "GhalfInv")
    )
  } else {
    result_processed <- list(
      G = lapply(result, `[[`, "G"),
      Ghalf = lapply(result, `[[`, "Ghalf")
    )
  }
  return(result_processed)
}


#' Compute Eigen-Based Square-Root Factors for \eqn{d\textbf{G}/d\lambda}
#'
#' @description
#' Applies an eigendecomposition-based matrix square root to each partition-wise
#' \eqn{d\textbf{G}/d\lambda} matrix after replacing \code{NA} entries with 0
#' and treating all-0 inputs with an identity fallback.
#'
#' @param dG_dlambda List of \eqn{p \times p} \eqn{d\textbf{G}/d\lambda} matrices by partition
#' @param p_expansions Integer; number of columns per partition
#' @param K Integer; number of interior knots (\eqn{K})
#' @param parallel,cl,chunk_size,num_chunks,rem_chunks Parallel computation parameters
#'
#' @return List of \eqn{p \times p} eigen-based square-root factors derived from
#'   the partition-wise \eqn{d\textbf{G}/d\lambda} matrices
#'
#' @keywords internal
#' @export
compute_dGhalf <- function(dG_dlambda,
                           p_expansions,
                           K,
                           parallel,
                           cl,
                           chunk_size,
                           num_chunks,
                           rem_chunks) {
  if(parallel & !is.null(cl)) {
    ## Handle remainder chunks first
    if(rem_chunks > 0) {
      rem_indices <- num_chunks * chunk_size + 1:rem_chunks
      rem <- lapply(rem_indices, function(k) {
        mat_k <- dG_dlambda[[k]]
        mat_k[is.na(mat_k)] <- 0
        if(!any(!(mat_k == 0))){
          mat_k <- diag(nrow(mat_k))
        }
        eig <- eigen(mat_k, symmetric = TRUE)
        eigen_values <- eig$values
        eigen_values[eig$values <= 0] <- 1
        sqrt_eigen_values <- sqrt(eigen_values)
        sqrt_eigen_values[eig$values <= 0] <- 0
        eig$vectors %**% (t(eig$vectors) * sqrt_eigen_values)
      })
    } else {
      rem <- list()
    }

    ## Process main chunks in parallel
    result <- c(
      Reduce("c",parallel::parLapply(cl, 1:num_chunks, function(chunk) {
        inds <- (chunk - 1)*chunk_size + 1:chunk_size
        lapply(inds, function(k) {
          mat_k <- dG_dlambda[[k]]
          mat_k[is.na(mat_k)] <- 0
          if(!any(!(mat_k == 0))){
            mat_k <- diag(nrow(mat_k))
          }
          eig <- eigen(mat_k, symmetric = TRUE)
          eigen_values <- eig$values
          eigen_values[eig$values <= 0] <- 1
          sqrt_eigen_values <- sqrt(eigen_values)
          sqrt_eigen_values[eig$values <= 0] <- 0
          eig$vectors %**% (t(eig$vectors) * sqrt_eigen_values)
        })
      })),
      rem
    )

  } else {
    ## Sequential computation
    result <- lapply(dG_dlambda, function(mat_k) {
      mat_k[is.na(mat_k)] <- 0
      if(!any(!(mat_k == 0))){
        mat_k <- diag(nrow(mat_k))
      }
      eig <- eigen(mat_k, symmetric = TRUE)
      eigen_values <- eig$values
      eigen_values[eig$values <= 0] <- 1
      sqrt_eigen_values <- sqrt(eigen_values)
      sqrt_eigen_values[eig$values <= 0] <- 0
      eig$vectors %**% (t(eig$vectors) * sqrt_eigen_values)
    })
  }

  return(result)
}


#' Compute Integrated Squared Second Derivative Penalty Matrix
#'
#' @description
#' Computes the \eqn{p \times p} integrated squared second-derivative penalty matrix
#' \eqn{\boldsymbol{\Lambda}_s} for one partition of a monomial spline,
#' such that
#' \deqn{\boldsymbol{\beta}_k^\top \boldsymbol{\Lambda}_s
#'        \boldsymbol{\beta}_k
#'   = \int_{\mathbf{a}}^{\mathbf{b}}
#'     \|\tilde{f}_k''(\mathbf{t})\|^2 \, d\mathbf{t},}
#' where \eqn{\mathbf{a}} and \eqn{\mathbf{b}} are the observed
#' predictor minimums and maximums, and
#' \eqn{\tilde{f}_k(\mathbf{t}) = \mathbf{x}_k^\top
#' \boldsymbol{\beta}_k} is the fitted function for partition
#' \eqn{\mathcal{P}_k}.
#'
#' The implementation uses a general monomial derivative rule that
#' handles all term types (marginal powers, two-way interactions,
#' quadratic interactions, three-way interactions) in a single unified
#' loop.
#'
#' @param colnm_expansions Character vector of length \eqn{p} giving
#'   column names for the basis expansion matrix \eqn{\mathbf{X}_k}.
#'   Each name encodes the monomial structure: predictor names appear
#'   literally for linear terms, with \code{^d} suffixes for degree
#'   \eqn{d}, and an \code{x} separator between factors in interaction
#'   terms (e.g.\ \code{"_1_alphax_2_beta^2"} for
#'   \eqn{t_1 t_2^2}).  Predictor names must not be substrings of one
#'   another.
#' @param C Numeric \eqn{N \times p} matrix of basis expansions
#'   \eqn{\mathbf{X}_k}.  Used only to read the observed range
#'   \eqn{[a_j, b_j]} of each predictor via \code{min}/\code{max} of
#'   the linear columns.
#' @param power1_cols Integer vector.  Column indices of linear terms
#'   \eqn{t_j}.
#' @param power2_cols Integer vector.  Column indices of quadratic terms
#'   \eqn{t_j^2}.
#' @param power3_cols Integer vector.  Column indices of cubic terms
#'   \eqn{t_j^3}.
#' @param power4_cols Integer vector.  Column indices of quartic terms
#'   \eqn{t_j^4}.
#' @param interaction_single_cols Integer vector.  Column indices of
#'   linear-by-linear interaction terms \eqn{t_j t_l}.
#' @param interaction_quad_cols Integer vector.  Column indices of
#'   linear-by-quadratic interaction terms \eqn{t_j t_l^2} and
#'   \eqn{t_j^2 t_l}.
#' @param triplet_cols Integer vector.  Column indices of three-way
#'   interaction terms \eqn{t_j t_l t_m}.
#' @param select_cols Optional integer vector of predictor indices
#'   (positions within \code{power1_cols}) whose curvature operators
#'   \eqn{D_v} are summed.  Defaults to all \eqn{q} predictors.
#'
#' @return
#' A symmetric positive semi-definite \eqn{p \times p} matrix
#' \eqn{\boldsymbol{\Lambda}_s} with
#' \deqn{[\boldsymbol{\Lambda}_s]_{ij}
#'   = \sum_{v=1}^{q} \int_{\mathbf{a}}^{\mathbf{b}}
#'     D_v(\phi_i) \, D_v(\phi_j) \, d\mathbf{t}.}
#'
#' @details
#' \subsection{Mathematical framework}{
#'
#' Let the basis expansion for partition \eqn{\mathcal{P}_k} be
#' \eqn{\mathbf{x}_k = (\phi_1(\mathbf{t}), \ldots,
#' \phi_p(\mathbf{t}))^\top} where each \eqn{\phi_i} is a
#' multivariate monomial
#' \deqn{\phi_i(\mathbf{t})
#'   = \prod_{j=1}^{q} t_j^{\alpha_{ij}}.}
#'
#' The second derivative of \eqn{\tilde{f}_k} decomposes into
#' \eqn{q} total curvature operators, one per predictor.  For
#' predictor \eqn{v}:
#' \deqn{D_v = \frac{\partial^2}{\partial t_v^2}
#'   + \sum_{s \neq v} \frac{\partial^2}{\partial t_v \, \partial t_s}.}
#'
#' The monomial derivative rule gives each second partial derivative
#' in closed form.  For the pure second derivative (\eqn{r = s = v}):
#' \deqn{\frac{\partial^2}{\partial t_v^2}
#'       \prod_{j} t_j^{\alpha_j}
#'   = \alpha_v(\alpha_v - 1) \;
#'     t_v^{\alpha_v - 2} \prod_{j \neq v} t_j^{\alpha_j}.}
#'
#' For a mixed second derivative (\eqn{s \neq v}):
#' \deqn{\frac{\partial^2}{\partial t_v \, \partial t_s}
#'       \prod_{j} t_j^{\alpha_j}
#'   = \alpha_v \alpha_s \;
#'     t_v^{\alpha_v - 1} t_s^{\alpha_s - 1}
#'     \prod_{j \neq v,s} t_j^{\alpha_j}.}
#'
#' In both cases a term is zero when the required exponent would be
#' negative (e.g.\ \eqn{\alpha_v < 2} for the pure case).  Applying
#' \eqn{D_v} to \eqn{\phi_i} produces a sum of monomials with known
#' coefficients and exponent vectors.
#' }
#'
#' \subsection{Integration}{
#'
#' Because every \eqn{D_v(\phi_i)} is polynomial, the product
#' \eqn{D_v(\phi_i) \, D_v(\phi_j)} is also polynomial and the
#' multivariate integral factorises over predictors:
#' \deqn{\int_{\mathbf{a}}^{\mathbf{b}}
#'       \prod_{j=1}^{q} t_j^{e_j} \, d\mathbf{t}
#'   = \prod_{j=1}^{q}
#'     \frac{b_j^{e_j+1} - a_j^{e_j+1}}{e_j + 1}.}
#'
#' Notably, this integral runs over \emph{all} \eqn{q} predictor
#' ranges, including predictors that do not appear in the integrand
#' (for which \eqn{e_j = 0} and the factor reduces to
#' \eqn{b_j - a_j}).
#' }
#'
#' \subsection{Single-predictor verification}{
#'
#' For \eqn{q = 1} with expansion
#' \eqn{\mathbf{x} = (1, t, t^2, t^3)^\top} on \eqn{[a, b]}, the
#' penalty matrix reduces to
#' \deqn{\boldsymbol{\Lambda}_s
#'   = \int_a^b \mathbf{x}'' \mathbf{x}''^\top \, dt
#'   = \begin{pmatrix}
#'       0 & 0 & 0 & 0 \\
#'       0 & 0 & 0 & 0 \\
#'       0 & 0 & 4(b - a) & 6(b^2 - a^2) \\
#'       0 & 0 & 6(b^2 - a^2) & 12(b^3 - a^3)
#'     \end{pmatrix},}
#' matching the formula in Section 2.3.
#' }
#'
#' @section Naming convention:
#' Column names in \code{colnm_expansions} must encode each monomial's
#' predictor content.  The parser checks each predictor name (taken
#' from the linear columns) against the column name:
#' \itemize{
#'   \item \code{predname^d} (literal caret) signals exponent \eqn{d}.
#'   \item \code{predname} without \code{^} signals exponent 1.
#'   \item Absence of \code{predname} signals exponent 0.
#' }
#' Multiple factors in an interaction are separated by \code{x}.
#' Predictor names must be unique and must not be substrings of one
#' another.
#'
#' @references
#' Reinsch, C. H. (1967). Smoothing by spline functions.
#' \emph{Numerische Mathematik}, 10(3), 177--183.
#'
#' @seealso \code{\link{lgspline}} for the full model fitting
#'   interface.
#'
#' @examples
#'
#' ## Verification example: 3-predictor model with all term types:
#' #  This example constructs the penalty matrix analytically and then
#' #  verifies selected entries against closed-form hand calculations.
#' #  Users can extend or adapt this example to audit new basis
#' #  expansions.
#'
#' set.seed(1234)
#' n <- 2000
#' t1 <- runif(n, 1, 4)     # predictor 1, support [1, 4]
#' t2 <- runif(n, -2, 3)    # predictor 2, support [-2, 3]
#' t3 <- runif(n, 0.5, 2)   # predictor 3, support [0.5, 2]
#'
#' ## Predictor names (must not be substrings of one another)
#' pn <- c("_1_aa", "_2_bb", "_3_cc")
#'
#' ## Build column names encoding the monomial structure
#' col_names <- c(
#'   pn,                                                # linear
#'   paste0(pn, "^2"),                                  # quadratic
#'   paste0(pn, "^3"),                                  # cubic
#'   paste0(pn[1], "x", pn[2]),                         # t1*t2
#'   paste0(pn[1], "x", pn[3]),                         # t1*t3
#'   paste0(pn[2], "x", pn[3]),                         # t2*t3
#'   paste0(pn[2], "x", pn[1], "^2"),                   # t1^2*t2
#'   paste0(pn[1], "x", pn[2], "^2"),                   # t1*t2^2
#'   paste0(pn[1], "x", pn[2], "x", pn[3])              # t1*t2*t3
#' )
#' p_expansions <- length(col_names)   # 14 basis functions
#'
#' ## Build the expansion matrix C
#' C <- cbind(
#'   t1, t2, t3,
#'   t1^2, t2^2, t3^2,
#'   t1^3, t2^3, t3^3,
#'   t1*t2, t1*t3, t2*t3,
#'   t2*t1^2, t1*t2^2,
#'   t1*t2*t3
#' )
#' colnames(C) <- col_names
#'
#' ## Compute the penalty matrix
#' Ls <- get_2ndDerivPenalty(
#'   colnm_expansions       = col_names,
#'   C                      = C,
#'   power1_cols            = 1:3,
#'   power2_cols            = 4:6,
#'   power3_cols            = 7:9,
#'   power4_cols            = integer(0),
#'   interaction_single_cols = 10:12,
#'   interaction_quad_cols  = 13:14,
#'   triplet_cols           = 15,
#'   p_expansions           = p_expansions,
#'   select_cols            = 1:3
#' )
#'
#' ## Hand-computed reference values (exact, using true bounds) ---
#' #  Notation: dt1 = 4-1 = 3, dt2 = 3-(-2) = 5, dt3 = 2-0.5 = 1.5
#' #
#' #  Entry [4,4]: t1^2 diagonal.
#' #    D_1(t1^2) = 2.  No other D_v contributes.
#' #    integral (2)^2 dt1 dt2 dt3 = 4 * 3 * 5 * 1.5 = 90
#' #
#' #  Entry [10,10]: t1*t2 diagonal.
#' #    D_1(t1*t2) = 1 (mixed d^2/dt1 dt2).
#' #    D_2(t1*t2) = 1 (mixed d^2/dt2 dt1).
#' #    integral 1 dt1 dt2 dt3 + integral 1 dt1 dt2 dt3 = 2*3*5*1.5 = 45
#' #
#' #  Entry [15,15]: t1*t2*t3 diagonal.
#' #    D_1(t1*t2*t3) = t3 + t2  (mixed partials d^2/dt1 dt2 and d^2/dt1 dt3)
#' #    D_2(t1*t2*t3) = t1 + t3  (similarly)
#' #    D_3(t1*t2*t3) = t1 + t2  (similarly)
#' #    Full integral = sum of 3 terms = 120 + 337.5 + 266.25 = 723.75
#'
#' ## Compare (allowing ~0.5% tolerance for data-derived bounds)
#' cat("Entry [4,4]:   analytical =", round(Ls[4,4], 2),
#'     "  exact = 90.00\n")
#' cat("Entry [10,10]: analytical =", round(Ls[10,10], 2),
#'     "  exact = 45.00\n")
#' cat("Entry [15,15]: analytical =", round(Ls[15,15], 2),
#'     "  exact = 723.75\n")
#' cat("Symmetric:", isSymmetric(Ls), "\n")
#'
#'
#' @keywords internal
#' @export
get_2ndDerivPenalty <- function(colnm_expansions,
                                C,
                                power1_cols,
                                power2_cols,
                                power3_cols,
                                power4_cols,
                                interaction_single_cols,
                                interaction_quad_cols,
                                triplet_cols,
                                p_expansions,
                                select_cols = NULL) {

  ## Guard: return zero matrix when there are no linear columns
  mat <- matrix(0, nrow = p_expansions, ncol = p_expansions)
  if (length(power1_cols) == 0L) return(mat)

  if (!any(!is.null(select_cols))) {
    select_cols <- seq_along(power1_cols)
  }

  ## 1. Cache observed predictor bounds a_j, b_j for j = 1, ..., q
  q_predictors <- length(power1_cols)
  a_lo <- vapply(power1_cols, function(j) min(C[, j]), numeric(1))
  b_hi <- vapply(power1_cols, function(j) max(C[, j]), numeric(1))

  ## 2. Monomial integration helpers
  mono_int <- function(k, a, b) {
    (b^(k + 1L) - a^(k + 1L)) / (k + 1L)
  }

  int_product <- function(e_vec) {
    val <- 1
    for (j in seq_len(q_predictors)) {
      val <- val * mono_int(e_vec[j], a_lo[j], b_hi[j])
    }
    val
  }

  ## 3. Parse exponent vectors alpha_i from column names
  pred_names <- colnm_expansions[power1_cols]

  parse_exponents <- function(col_idx) {
    nm <- colnm_expansions[col_idx]
    e  <- integer(q_predictors)
    for (j in seq_len(q_predictors)) {
      pnm <- pred_names[j]
      pat <- paste0(pnm, "\\^([0-9]+)")
      m   <- regmatches(nm, regexpr(pat, nm, perl = TRUE))
      if (length(m) == 1L && nchar(m) > 0L) {
        e[j] <- as.integer(sub(paste0(pnm, "\\^"), "", m))
      } else if (grepl(pnm, nm, fixed = TRUE)) {
        e[j] <- 1L
      }
    }
    e
  }

  ## Collect all basis columns that can carry a non-zero penalty
  penalised <- sort(unique(c(power1_cols, power2_cols, power3_cols,
                             power4_cols, interaction_single_cols,
                             interaction_quad_cols, triplet_cols)))
  penalised <- penalised[penalised >= 1L & penalised <= p_expansions]

  ## Exponent matrix Alpha: Alpha[i, j] = alpha_{ij}
  Alpha <- matrix(0L, nrow = p_expansions, ncol = q_predictors)
  for (idx in penalised) {
    Alpha[idx, ] <- parse_exponents(idx)
  }

  ## 4. Total curvature operator D_v applied to monomial phi_i
  Dv <- function(alpha, v) {
    out <- list()

    ## Pure second derivative d^2/dt_v^2
    if (alpha[v] >= 2L) {
      e    <- alpha
      e[v] <- e[v] - 2L
      out[[length(out) + 1L]] <- list(c = alpha[v] * (alpha[v] - 1L),
                                      e = e)
    }

    ## Mixed second derivatives d^2/(dt_v dt_s) for each s != v
    for (s in seq_len(q_predictors)) {
      if (s == v) next
      if (alpha[v] >= 1L && alpha[s] >= 1L) {
        e    <- alpha
        e[v] <- e[v] - 1L
        e[s] <- e[s] - 1L
        out[[length(out) + 1L]] <- list(c = alpha[v] * alpha[s],
                                        e = e)
      }
    }
    out
  }

  ## 5. Penalty matrix entry
  entry <- function(i, j) {
    total <- 0
    for (v in select_cols) {
      ti <- Dv(Alpha[i, ], v)
      tj <- Dv(Alpha[j, ], v)
      if (length(ti) == 0L || length(tj) == 0L) next
      for (a in ti) {
        for (b in tj) {
          total <- total + a$c * b$c * int_product(a$e + b$e)
        }
      }
    }
    total
  }

  ## 6. Populate Lambda_s (symmetric; only upper triangle computed)
  Ls <- matrix(0, nrow = p_expansions, ncol = p_expansions)

  for (ii in seq_along(penalised)) {
    i <- penalised[ii]
    for (jj in ii:length(penalised)) {
      j <- penalised[jj]
      v <- entry(i, j)
      if (v != 0) {
        Ls[i, j] <- v
        Ls[j, i] <- v
      }
    }
  }

  Ls
}

#' Wrapper for Integrated Second-Derivative Penalty Computation
#'
#' @description
#' Computes the integrated squared second-derivative penalty matrix with
#' optional parallel processing.
#'
#' @param K Number of partitions (\eqn{K+1})
#' @param colnm_expansions Column names of basis expansions
#' @param C Basis expansion matrix used to determine observed predictor ranges
#' @param power1_cols Linear term columns
#' @param power2_cols Quadratic term columns
#' @param power3_cols Cubic term columns
#' @param power4_cols Quartic term columns
#' @param interaction_single_cols Single interaction columns
#' @param interaction_quad_cols Quadratic interaction columns
#' @param triplet_cols Triplet interaction columns
#' @param nonspline_cols Predictors not treated as spline effects
#' @param p_expansions Number of columns in the basis expansion per partition
#' @param parallel Logical to enable parallel processing
#' @param cl Cluster object for parallel computation
#'
#' @return A \eqn{p \times p} integrated squared second-derivative penalty
#'   matrix.
#'
#' @keywords internal
#' @export
get_2ndDerivPenalty_wrapper <- function(K,
                                        colnm_expansions,
                                        C,
                                        power1_cols,
                                        power2_cols,
                                        power3_cols,
                                        power4_cols,
                                        interaction_single_cols,
                                        interaction_quad_cols,
                                        triplet_cols,
                                        nonspline_cols,
                                        p_expansions,
                                        parallel,
                                        cl) {
  ## Temporarily append corresponding polynomial terms so nonspline columns
  #  are carried through the same analytic penalty bookkeeping.
  colnm_expansions_og <- colnm_expansions
  if(length(nonspline_cols) > 0){
    for(jj in 1:length(nonspline_cols)){
      power1_cols <- c(power1_cols, nonspline_cols[jj])
      if(length(power2_cols) > 0){
        colnm_expansions <- c(colnm_expansions,
                              paste0(colnm_expansions[nonspline_cols[jj]],'^2'))
        C <- cbind(C, 0)
        power2_cols <- c(power2_cols, ncol(C))
      }
      if(length(power3_cols) > 0){
        colnm_expansions <- c(colnm_expansions,
                              paste0(colnm_expansions[nonspline_cols[jj]],'^3'))
        C <- cbind(C, 0)
        power3_cols <- c(power3_cols, ncol(C))
      }
      if(length(power4_cols) > 0){
        colnm_expansions <- c(colnm_expansions,
                              paste0(colnm_expansions[nonspline_cols[jj]],'^4'))
        C <- cbind(C, 0)
        power4_cols <- c(power4_cols, ncol(C))
      }
    }
    ## Update colnames and number of columns with new nonspline power terms
    colnames(C) <- colnm_expansions
    p_expansions <- ncol(C)
  }

  ## If parallel processing is worthwhile, split the selected derivative
  if(parallel & (K > 1)){
    chunk_size <- max(1, floor(2 * length(cl)))
    total_cols <- length(power1_cols)
    result <- matrix(0, nrow = p_expansions, ncol = p_expansions)

    for(start in seq(1, total_cols, by = chunk_size)) {
      end <- min(start + chunk_size - 1, total_cols)
      chunk_result <- Reduce("+",
                             parallel::parLapply(cl,
                                                 start:end,
                                                 function(select_col) {
                                                   get_2ndDerivPenalty(
                                                     colnm_expansions,
                                                     C,
                                                     power1_cols,
                                                     power2_cols,
                                                     power3_cols,
                                                     power4_cols,
                                                     interaction_single_cols,
                                                     interaction_quad_cols,
                                                     triplet_cols,
                                                     p_expansions,
                                                     select_col)
                                                 }))
      result <- result + chunk_result
    }
  } else {
    result <- get_2ndDerivPenalty(colnm_expansions,
                                  C,
                                  power1_cols,
                                  power2_cols,
                                  power3_cols,
                                  power4_cols,
                                  interaction_single_cols,
                                  interaction_quad_cols,
                                  triplet_cols,
                                  p_expansions)
  }
  colnames(result) <- colnm_expansions
  rownames(result) <- colnm_expansions
  ## Isolate the entries excluding appended
  result <- result[colnm_expansions_og, colnm_expansions_og]
  return(result)
}


#' Unconstrained Generalized Linear Model Estimation
#'
#' @description
#' Fits generalized linear models without smoothing constraints
#' using penalized maximum likelihood estimation. This is applied to each
#' partition to obtain the unconstrained estimates, prior to imposing the
#' smoothing constraints.
#'
#' Hot-start estimates are initialized by treating the matrix square-root inverse
#' as pseudo-observations, appending to the rows of the design matrix, and
#' calling \code{\link{glm.fit}} replacing the response for pseudo-observations
#' with the inverse link function applied to the value of eta = XB = 0. For log
#' link, exp(0) = 1 is valid for most families. For cases like inverse link,
#' dividing by 0 is obviously not possible, so this is replaced with 1/tol
#' where tol is the convergence tolerance argument.
#'
#' @param X Design matrix of predictors
#' @param y Response variable vector
#' @param LambdaHalf Square root of penalty matrix
#'   (\eqn{\boldsymbol{\Lambda}^{1/2}})
#' @param Lambda Penalty matrix (\eqn{\boldsymbol{\Lambda}})
#' @param keep_weighted_Lambda Logical flag to control penalty matrix handling:
#'   - \code{TRUE}: Return coefficients directly from weighted penalty fitting
#'   - \code{FALSE}: Apply damped Newton-Raphson optimization to refine
#'     estimates
#' @param family Distribution family specification
#' @param tol Convergence tolerance
#' @param K Number of partitions minus one (\eqn{K})
#' @param parallel Flag for parallel processing
#' @param cl Cluster object for parallel computation
#' @param chunk_size Processing chunk size
#' @param num_chunks Number of computational chunks
#' @param rem_chunks Remaining chunks
#' @param order_indices Observation ordering indices
#' @param weights Optional observation weights
#' @param ... Additional arguments passed to \code{glm.fit}
#'
#' @return
#' Numeric column vector of unconstrained coefficient estimates.
#'
#' For fitting non-canonical GLMs, use \code{keep_weighted_Lambda = TRUE}
#' since the score and Hessian equations below are no longer valid.
#'
#' For Gamma(link='log') using \code{keep_weighted_Lambda = TRUE} is
#' misleading. The information is weighted by a constant (shape parameter)
#' rather than some mean-variance relationship. So
#' \code{keep_weighted_Lambda = TRUE} is highly recommended for log-link Gamma
#' models. This constant flushes into the penalty terms, and so the
#' formulation of the information matrix is valid.
#'
#' For other scenarios, like probit regression, there will be diagonal weights
#' incorporated into the penalty matrix for providing initial MLE estimates,
#' which technically imposes a prior distribution on beta coefficients that
#' isn't by intent.
#'
#' Heuristically, it shouldn't affect much, as these will be updated to their
#' proper form when providing estimates under constraint; lgspline otherwise
#' does use the correct form of score and information afterwards,
#' regardless of canonical/non-canonical status,
#' as long as 'glm_weight_function' and 'qp_score_function' are properly
#' specified.
#'
#'
#' @keywords internal
#' @export
unconstrained_fit_default <- function(X,
                                      y,
                                      LambdaHalf,
                                      Lambda,
                                      keep_weighted_Lambda,
                                      family,
                                      tol,
                                      K,
                                      parallel,
                                      cl,
                                      chunk_size,
                                      num_chunks,
                                      rem_chunks,
                                      order_indices,
                                      weights,
                                      ...){

  ## Return 0 vector for empty partitions
  if(nrow(X) == 0){
    return(cbind(rep(0, ncol(X))))
  }

  ## Weight if non-null
  # Yields first NR updated as
  # = (X^{T}V^{-1}X + Lambda)^{-1}X^{T}V^{-1}y for V^{-1} = diag(weights)
  if(any(!is.null(weights))){
    if(length(weights) == length(y)){
      weights <- c(weights)
    } else {
      weights <- rep(1, length(y))
    }
  } else {
    weights <- rep(1, length(y))
  }

  ## Determine if quasi family should be used for initialization
  family_init <- family

  ## Override with quasi when initialization approach,
  #  calling g(0) to get the vector of response corresponding to natural link
  #  eta = XB = 0,
  #  For example, 1/(1+exp(-0)) = 0.5 is not a valid value according to family=
  #  binomial(). Will only affect hot-start initialization.
  if(!is.null(family$family)){

    fam_name  <- family$family
    link_name <- family$link

    needs_quasi <- FALSE

    ## Binomial always needs quasi for augmented ridge trick
    if(fam_name == "binomial"){
      needs_quasi <- TRUE
    }

    ## Inverse-type links produce Inf at g(0)
    if(link_name %in% c("inverse", "1/mu^2")){
      needs_quasi <- TRUE
    }

    ## Other binomial-style links
    if(link_name %in% c("logit", "probit", "cloglog", "cauchit")){
      needs_quasi <- TRUE
    }

    if(needs_quasi){

      family_init <- switch(
        fam_name,
        "binomial" = quasibinomial(link = link_name),
        "Gamma" = quasi(link = link_name, variance = "mu^2"),
        "inverse.gaussian" = quasi(link = link_name, variance = "mu^3"),
        family
      )
    }
  }

  ## For cases like inverse link, we need finite values, so replace 0 with
  #  the "tol" value
  psuedoresponse <- family_init$linkinv(0)
  if(is.na(psuedoresponse)){
    psuedoresponse <- family_init$linkinv(tol)
  }
  if(!is.finite(psuedoresponse)){
    psuedoresponse <- family_init$linkinv(tol)
  }

  ## Ordinary fit using Tikhonov parameterization
  # This is not "correct" because the GLM weights are introduced into the
  # square-root penalty matrix
  # but makes good hot-start initialization
  mod <- try({
    glm.fit(
      x = rbind(X, LambdaHalf),
      y = cbind(c(y,
                  rep(psuedoresponse,
                      nrow(LambdaHalf)))),
      family = family_init,
      weights = c(weights,
                  rep(mean(weights),
                      nrow(LambdaHalf))),
      ...
    )
  }, silent = TRUE)

  if(keep_weighted_Lambda & any(!(inherits(mod, 'try-error')))){
    return(cbind(mod$coefficients))
  }

  if(any(inherits(mod, 'try-error'))){
    init <- c(family$linkfun(mean(y)), rep(0, ncol(X)-1))
  } else {
    init <- c(coef(mod))
  }

  if(any(is.na(init))){
    init <- rep(0, length(init))
  }
  if(any(!is.finite(init))){
    init <- rep(0, length(init))
  }

  ## Adjusts Tikhonov penalties using damped nr, with initial hot-start
  #  values from glm.fit using augmented data
  res <- cbind(damped_newton_r(
    ## initial guess
    init,
    ## proportional to log-likelihood
    function(par){
      -sum(weights*family$dev.resids(
        y,
        family$linkinv(c(X %**% cbind(par))),
        wt = 1))*0.5 -
        0.5*c(crossprod(par, Lambda) %**% cbind(par))
    },
    ## score
    function(par){
      c(crossprod(X, weights*cbind(y - family$linkinv(X %**% cbind(par)))) -
           Lambda %**% cbind(par))
    },
    ## information
    function(par){
      crossprod(X, weights*c(family$variance(X %**% cbind(par))) * X) +
        Lambda
    },
    tol))
  return(res)
}

#' Collapse Matrix List into a Single Dense Block-Layout Matrix
#'
#' @description
#' Transforms a list of matrices into a single dense block-layout matrix. This
#' is useful especially for quadratic programming problems, where operating on
#' lists of blocks is not convenient.
#'
#' @param matlist List of input matrices
#'
#' @return
#' Dense matrix with each input block placed in its own column range
#'
#' @keywords internal
#' @export
collapse_block_diagonal <- function(matlist){
  ## Lay each block into its own column range so dense routines such as
  #  solve.QP can work with the full system directly.
  nrows <- sapply(matlist, nrow)
  ncols <- sapply(matlist, ncol)
  Reduce('rbind', lapply(1:length(matlist), function(k){
    mat <- matrix(0,
                  nrow = nrows[k],
                  ncol = sum(ncols))
    mat[,sum(ncols[-c(k:length(nrows))]) +
          1:ncols[k]] <-
      matlist[[k]]
    mat
  }))
}

#' Compute Newton-Raphson Parameter Update with Numerical Stabilization
#'
#' @description
#' Performs parameter update in iterative optimization.
#'
#' Called by \code{\link{damped_newton_r}} in the update step
#'
#' @param gradient_val Numeric vector of gradient values (\eqn{\textbf{u}})
#' @param neghessian_val Negative Hessian matrix (\eqn{\textbf{G}^{-1}}
#'   approximately)
#'
#' @return
#' Numeric column vector of parameter updates (\eqn{\textbf{G}\textbf{u}})
#'
#' @details
#' This helper function is a core component of Newton-Raphson optimization.
#' It provides a computationally-stable approach to computing
#' \eqn{\textbf{G}\textbf{u}}, for information matrix \eqn{\textbf{G}} and
#' score vector \eqn{\textbf{u}}, where the Newton-Raphson update can be
#' expressed as
#' \eqn{\boldsymbol{\beta}^{(m+1)} = \boldsymbol{\beta}^{(m)} + \textbf{G}\textbf{u}}.
#'
#' @seealso \code{\link{damped_newton_r}} for the full optimization routine
#'
#' @keywords internal
#' @export
nr_iterate <- function(gradient_val, neghessian_val){
  sc <- sqrt(max(abs(diag(neghessian_val))))
  if(sc < .Machine$double.eps) sc <- 1
  H <- neghessian_val / sc
  g <- cbind(gradient_val / sc)
  out <- try(solve(H, g), silent = TRUE)
  if(inherits(out, 'try-error')) {
    out <- invert(H) %**% g
  }
  out
}


#' Damped Newton-Raphson Parameter Optimization
#'
#' @description
#' Performs iterative parameter estimation with adaptive step-size damping
#'
#' Internal function for fitting unconstrained model components using damped
#' Newton-Raphson updates.
#'
#' @param parameters Initial parameter vector to be optimized
#' @param loglikelihood Function computing log-likelihood for current parameters
#' @param gradient Function computing parameter gradients
#' @param neghessian Function computing negative Hessian matrix
#' @param tol Numeric convergence tolerance (default 1e-7)
#' @param max_cnt Maximum number of optimization iterations (default 64)
#' @param max_dmp_steps Maximum damping step attempts (default 16)
#'
#' @return
#' Final parameter vector returned at termination.
#'
#' @details
#' Implements a robust damped Newton-Raphson optimization algorithm.
#' The Newton direction is computed once per outer iteration and reused
#' across damping half-steps.
#'
#' @seealso
#' - \code{\link{nr_iterate}} for parameter update computation
#'
#' @keywords internal
#' @export
damped_newton_r <- function(parameters,
                            loglikelihood,
                            gradient,
                            neghessian,
                            tol = 1e-7,
                            max_cnt = 64,
                            max_dmp_steps = 16){

  ## Initialize
  new_param <- parameters
  old_param <- parameters
  eps <- Inf
  master_count <- 0

  ## Run the loop while change between coefficients > tol and
  #  max iterations have not been reached
  while(eps > tol & master_count < max_cnt){

    old_param <- c(new_param)
    prev_objective <- loglikelihood(old_param)

    if(!is.finite(prev_objective)){
      cat('\n \t Error Encountered, Number of N.R. steps so far: ',
          master_count, '\n')
      stop('\n \t NA/NaN/non-finite value detected when running',
           ' unconstrained damped Newton-Raphson.',
           ' \n \t Try re-fitting a simpler model, using',
           ' greater/smaller penalties, ',
           ' experimenting with different knot locations, or reducing the',
           ' number of knots.')
    }

    ## Compute Newton direction once, reuse across damping half-steps
    direction <- nr_iterate(gradient(old_param), neghessian(old_param))

    ## Damp: halve step size until objective improves
    accepted <- FALSE
    for(count in 0:(max_dmp_steps - 1)){
      candidate <- old_param + (2^(-count)) * c(direction)
      new_objective <- loglikelihood(candidate)
      if(!is.finite(new_objective)) new_objective <- -Inf
      if(new_objective > prev_objective){
        new_param <- candidate
        accepted <- TRUE
        break
      }
    }

    if(!accepted){
      ## No damping level improved; declare convergence at current point
      new_param <- old_param
      break
    }

    ## Otherwise re-compute difference between coefficient updates
    eps <- max(abs(old_param - new_param))
    if(!is.finite(eps)){
      new_param <- old_param
      break
    }

    master_count <- master_count + 1
  }

  return(new_param)
}


#' Generate Interaction Variable Patterns
#'
#' @description
#' Generates all possible interaction patterns for 2 or 3 variables. This is
#' used in part for identifying which interactions and expansions to exclude
#' (provided to "exclude_these_expansions" argument of lgspline) based on
#' formulas provided.
#'
#' @param vars Character vector of variable names
#'
#' @return
#' Character vector of interaction pattern strings for 2- or 3-variable
#' inputs, or \code{NULL} otherwise.
#'
#' @keywords internal
#' @export
get_interaction_patterns <- function(vars) {
  if(length(vars) == 2) {
    ## Linear: both orderings
    lin <- c(paste0(vars[1], "x", vars[2]),
             paste0(vars[2], "x", vars[1]))

    ## Quadratic: both orderings for each squared term
    quad <- c(paste0(vars[1], "x", vars[2], "^2"),
              paste0(vars[2], "x", vars[1], "^2"))

    return(c(lin, quad))
  } else if(length(vars) == 3) {
    ## All possible 3-way orderings
    return(c(paste0(vars[1], "x", vars[2], "x", vars[3]),
             paste0(vars[1], "x", vars[3], "x", vars[2]),
             paste0(vars[2], "x", vars[1], "x", vars[3]),
             paste0(vars[2], "x", vars[3], "x", vars[1]),
             paste0(vars[3], "x", vars[1], "x", vars[2]),
             paste0(vars[3], "x", vars[2], "x", vars[1])))
  }
}


#' BFGS Implementation for REML Parameter Estimation
#'
#' @description
#' BFGS optimizer designed for REML optimization of correlation parameters. Combines
#' function evaluation and gradient computation into single call to avoid redundant
#' model refitting.
#'
#' @param par Numeric vector of initial parameter values.
#' @param fn Function returning list(objective, gradient). Must return both objective
#' value and gradient vector matching length(par).
#' @param control List of control parameters:
#' \describe{
#'   \item{maxit}{Maximum iterations, default 50}
#'   \item{abstol}{Absolute convergence tolerance, default sqrt(.Machine$double.eps)}
#'   \item{reltol}{Relative convergence tolerance, default sqrt(.Machine$double.eps)}
#'   \item{initial_damp}{Initial damping factor, default 1}
#'   \item{min_damp}{Minimum damping before termination, default 2^-16}
#'   \item{trace}{Print iteration progress, default FALSE}
#' }
#'
#' @return List containing:
#' \describe{
#'   \item{par}{Parameter vector minimizing objective}
#'   \item{value}{Minimum objective value}
#'   \item{counts}{Number of iterations}
#'   \item{convergence}{TRUE if termination occurred before \code{maxit};
#'     this reflects the current stopping rule rather than a separate
#'     post-hoc convergence check}
#'   \item{message}{Description of termination status}
#'   \item{vcov}{Final approximation of inverse-Hessian, useful for inference}
#' }
#'
#' @details
#' Implements BFGS, used internally by \code{lgspline()} for optimizing
#' correlation parameters via REML when an analytic REML gradient is supplied.
#'
#' This is more efficient than native BFGS, since gradient and loss can be computed simultaneously,
#' avoiding re-computing components in "fn" and "gr" separately.
#'
#' @examples
#'
#'
#' ## Minimize Rosenbrock function
#' fn <- function(x) {
#'   # Objective
#'   f <- 100*(x[2] - x[1]^2)^2 + (1-x[1])^2
#'   # Gradient
#'   g <- c(-400*x[1]*(x[2] - x[1]^2) - 2*(1-x[1]),
#'          200*(x[2] - x[1]^2))
#'   list(f, g)
#' }
#' (res <- efficient_bfgs(c(0.5, 2.5), fn))
#'
#' ## Compare to
#' (res0 <- stats::optim(c(0.5, 2.5), function(x)fn(x)[[1]], hessian = TRUE))
#' solve(res0$hessian)
#'
#' @keywords internal
#' @export
efficient_bfgs <- function(par, fn, control = list()) {
  ctrl <- list(
    maxit = 50,
    abstol = sqrt(.Machine$double.eps),
    reltol = sqrt(.Machine$double.eps),
    initial_damp = 1,
    min_damp = 2^-16,
    trace = FALSE
  )
  ctrl[names(control)] <- control

  n_params <- length(par)
  x <- par
  Inv <- diag(n_params)
  damp <- ctrl$initial_damp
  best_x <- x
  best_f <- Inf
  best_Inv <- Inv

  result <- fn(c(x))
  if(length(result) != 2) stop("fn must return list of (objective, gradient)")
  f <- result[[1]]
  grad <- result[[2]]
  if(is.null(grad) | any(is.na(grad))) grad <- approx_grad(x, fn)
  if(length(grad) != n_params) stop("gradient must match parameter length")

  for(iter in 1:ctrl$maxit) {
    prev_x <- x
    prev_f <- f
    prev_grad <- grad

    p <- -crossprod(t(Inv), cbind(grad))
    x_new <- x + damp * p

    result <- fn(c(x_new))
    f_new <- result[[1]]
    grad_new <- result[[2]]
    if(is.null(grad_new) | any(is.na(grad_new))) grad_new <-
      approx_grad(x_new, fn)

    if(is.na(f_new) || is.nan(f_new) || !is.finite(f_new)) {
      damp <- damp/2
      if(damp < ctrl$min_damp) break
      next
    }

    if(f_new < f || iter <= 2) {
      if(f_new < best_f) {
        best_f <- f_new
        best_x <- x_new
        best_Inv <- Inv
      }

      s0 <- cbind(x_new - x)
      y0 <- cbind(grad_new - grad)
      denom <- sum(y0 * s0)

      if(abs(denom) > 1e-16) {
        rho0 <- 1/denom
        term1 <- diag(n_params) - rho0 * tcrossprod(s0, y0)
        term2 <- diag(n_params) - rho0 * tcrossprod(y0, s0)
        Inv <- crossprod(t(term1), crossprod(t(Inv), term2)) +
          rho0 * tcrossprod(s0)
      }

      x <- x_new
      f <- f_new
      grad <- grad_new
      damp <- 1

      if(iter > 2) {
        if(abs(f - prev_f) < ctrl$abstol * (abs(prev_f) + ctrl$reltol)) break
        if(max(abs(x - prev_x)) < ctrl$abstol) break
      }

    } else {
      x <- best_x*(1-damp) + x*damp
      f <- best_f*(1-damp) + f*damp
      Inv <- best_Inv*(1-damp) + Inv*damp
      damp <- damp/2
      if(damp < ctrl$min_damp) break
    }

    if(ctrl$trace) {
      cat(sprintf("Iter %d: f = %f, |grad| = %f, damp = %f\n",
                  iter, f, sqrt(sum(grad^2)), damp))
    }
  }

  list(
    par = c(best_x),
    value = best_f,
    counts = iter,
    convergence = (iter < ctrl$maxit),
    message = if(iter == ctrl$maxit)"Maximum iterations reached" else "Converged",
    vcov = best_Inv
  )
}

#' Finite-difference Gradient Computer
#'
#' @description
#' Computes a finite-difference approximation derived from the objective in
#' \code{fn} at \code{x}, returned in the sign convention currently used by
#' \code{efficient_bfgs}.
#'
#' @param x Numeric vector of function arguments
#' @param fn Function returning list(objective, gradient)
#' @param eps Numeric scalar, finite difference tolerance
#'
#' @return Numeric vector of finite-difference approximated gradient values in
#'   the current optimizer sign convention
#'
#' @details
#' Used within \code{efficient_bfgs} when \code{fn} does not supply a usable
#' gradient. In the main \code{lgspline()} correlation-optimization path,
#' \code{stats::optim()} is used instead when no analytic REML gradient is
#' available. Internally this helper returns the negated central-difference
#' approximation rather than the raw derivative.
#'
#' @keywords internal
#' @export
approx_grad <- function(x, fn, eps = sqrt(.Machine$double.eps)) {
  grad <- numeric(length(x))
  for(i in 1:length(x)) {
    ## Scale base epsilon by parameter magnitude
    h1 <- eps * max(1, abs(x[i]))
    h2 <- h1/2
    x_eps <- c(x)

    ## First gradient at h1
    x_eps[i] <- x[i] + 0.5*h1
    f1_plus <- fn(x_eps)[[1]]
    x_eps[i] <- x[i] - 0.5*h1
    f1_minus <- fn(x_eps)[[1]]
    g1 <- (f1_plus - f1_minus)/h1

    ## Second gradient at h2
    x_eps[i] <- x[i] + 0.5*h2
    f2_plus <- fn(x_eps)[[1]]
    x_eps[i] <- x[i] - 0.5*h2
    f2_minus <- fn(x_eps)[[1]]
    g2 <- (f2_plus - f2_minus)/h2

    ## Richardson extrapolation
    grad[i] <- (4*g2 - g1)/3
  }
  -grad
}

#' Calculate Matrix Square Root for Symmetric Matrices
#'
#' @param mat A symmetric matrix \eqn{\textbf{M}}
#'
#' @return A matrix \eqn{\textbf{B}} such that \eqn{\textbf{B}\textbf{B}}
#'   equals \eqn{\textbf{M}} on the positive-eigenvalue subspace, with
#'   non-positive components truncated to 0.
#'
#' @details
#'
#' Uses an eigenvalue-decomposition-based approach.
#'
#' Non-positive eigenvalues are set to 0 before taking fourth roots.
#'
#' This implementation is particularly useful for whitening procedures in GLMs
#' with correlation structures and for computing variance-covariance matrices
#' under constraints.
#'
#' You can use this to help construct a custom \code{Vhalf_fxn}, or more
#' directly to build the \eqn{\mathbf{V}^{1/2}} input supplied to
#' \code{\link{lgspline}} for correlation-aware fits.
#'
#' @examples
#'
#' ## Identity matrix
#' m1 <- diag(2)
#' matsqrt(m1)  # Returns identity matrix
#'
#' ## Compound symmetry correlation matrix
#' rho <- 0.5
#' m2 <- matrix(rho, 3, 3) + diag(1-rho, 3)
#' B <- matsqrt(m2)
#' # Verify: B %**% B approximately equals m2
#' all.equal(B %**% B, m2)
#'
#' ## Example for correlation structure
#' n_blocks <- 2  # Number of subjects
#' block_size <- 3  # Measurements per subject
#' rho <- 0.7  # Within-subject correlation
#' # Correlation matrix for one subject
#' R <- matrix(rho, block_size, block_size) +
#'      diag(1-rho, block_size)
#' # Full correlation matrix for all subjects
#' V <- kronecker(diag(n_blocks), R)
#' Vhalf <- matsqrt(V)
#'
#' @export
matsqrt <- function(mat) {
  eig <- eigen(mat, symmetric = TRUE)
  wts <- ifelse(eig$values > 0,
                eig$values^0.25,
                0)
  crossprod(t(eig$vectors) * wts)
}

#' Calculate Matrix Inverse Square Root for Symmetric Matrices
#'
#' @param mat A symmetric matrix \eqn{\textbf{M}}
#'
#' @return A matrix \eqn{\textbf{B}} such that \eqn{\textbf{B}\textbf{B}}
#'   equals the Moore-Penrose-style inverse on the positive-eigenvalue
#'   subspace, with non-positive components truncated to 0.
#'
#' @details
#'
#' Uses an eigenvalue-decomposition-based approach.
#'
#' Non-positive eigenvalues are set to 0 before taking inverse fourth roots.
#'
#' This implementation is particularly useful for whitening procedures in GLMs
#' with correlation structures and for computing variance-covariance matrices
#' under constraints.
#'
#' You can use this to help construct a custom \code{VhalfInv_fxn} for
#' \code{\link{lgspline}}.  When only \code{VhalfInv} is supplied there,
#' the corresponding \code{Vhalf} is reconstructed internally by inversion
#' for the GEE code paths.
#'
#' @examples
#'
#' ## Identity matrix
#' m1 <- diag(2)
#' matinvsqrt(m1)  # Returns identity matrix
#'
#' ## Compound symmetry correlation matrix
#' rho <- 0.5
#' m2 <- matrix(rho, 3, 3) + diag(1-rho, 3)
#' B <- matinvsqrt(m2)
#' # Verify: B %**% B approximately equals solve(m2)
#' all.equal(B %**% B, solve(m2))
#'
#' ## Example for GLM correlation structure
#' n_blocks <- 2  # Number of subjects
#' block_size <- 3  # Measurements per subject
#' rho <- 0.7  # Within-subject correlation
#' # Correlation matrix for one subject
#' R <- matrix(rho, block_size, block_size) +
#'      diag(1-rho, block_size)
#' ## Full correlation matrix for all subjects
#' V <- kronecker(diag(n_blocks), R)
#' ## Create whitening matrix
#' VhalfInv <- matinvsqrt(V)
#'
#' # Example construction of VhalfInv_fxn for lgspline
#' VhalfInv_fxn <- function(par) {
#'   rho <- tanh(par)  # Transform parameter to (-1, 1)
#'   R <- matrix(rho, block_size, block_size) +
#'        diag(1-rho, block_size)
#'   kronecker(diag(n_blocks), matinvsqrt(R))
#' }
#'
#' @export
matinvsqrt <- function(mat) {
  eig <- eigen(mat, symmetric = TRUE)
  wts <- ifelse(eig$values > 0,
                eig$values^-0.25,
                0)
  crossprod(t(eig$vectors) * wts)
}

#' Compute Euclidean distance matrix for a cluster block
#'
#' Returns pairwise Euclidean distances within the cluster indexed by
#' \code{inds}. When \code{spacetime} has multiple columns, squared distances
#' are averaged across dimensions before taking the square root.
#' @keywords internal
.compute_dist_block <- function(spacetime, inds) {
  spacetime <- cbind(spacetime)
  diffs <- outer(spacetime[inds, 1], spacetime[inds, 1],
                 function(x, y) (x - y)^2)
  if(ncol(spacetime) > 1){
    diffs <- diffs + Reduce("+", lapply(
      2:ncol(spacetime),
      function(v) outer(spacetime[inds, v], spacetime[inds, v],
                        function(x, y) (x - y)^2)
    ))
  }
  sqrt(diffs / ncol(spacetime))
}

#' Compute rank-based distance matrix for AR(1) structures
#'
#' Converts the unique within-block distances returned by
#' \code{.compute_dist_block()} to integer lags \code{0, 1, 2, \ldots} in
#' increasing order.
#' @keywords internal
.rank_dists <- function(spacetime, inds) {
  diffs <- .compute_dist_block(spacetime, inds)
  block_size <- length(inds)
  unique_diffs <- unique(as.vector(diffs))
  matrix(match(diffs, sort(unique_diffs)) - 1, block_size, block_size)
}

#' Evaluate the REML gradient with respect to a single correlation parameter
#'
#' Computes the derivative of the negative REML objective with respect to a
#' scalar correlation parameter, given the matrix derivative
#' \eqn{\partial \mathbf{V}/\partial \rho}.
#'
#' @details
#' \subsection{Notation}{
#' \eqn{\mathbf{D} = \mathrm{diag}(d_1,\ldots,d_N)} is the diagonal matrix of
#' observation weights (\code{observation_weights}), and
#' \eqn{\mathbf{W} = \mathrm{diag}(w_1,\ldots,w_N)} is the diagonal matrix of
#' GLM working weights evaluated at the current fitted values via
#' \code{glm_weight_function}, with the observation-weight contribution
#' carried separately by \eqn{\mathbf{D}}.  For canonical GLM families,
#' \eqn{w_i} is the usual IRWLS/Fisher-scoring weight on the mean--variance
#' scale; for example, logistic regression gives
#' \eqn{w_i = \mu_i(1-\mu_i)}.  The combined weighting entering the
#' information matrix is \eqn{\mathbf{W}\mathbf{D}}.  In Gaussian identity
#' models both reduce to scalar multiples of the identity.
#'
#' \eqn{\mathbf{V}} is the \eqn{N \times N} correlation matrix implied by
#' \code{VhalfInv}, with
#' \eqn{\mathbf{V}^{-1} = (\mathbf{V}^{-1/2})^\top \mathbf{V}^{-1/2}}.
#'
#' The penalized observed information at the current iterate is
#' \deqn{\mathbf{M} = (\mathbf{X}^*)^\top
#'   \mathbf{V}^{-1}\mathbf{W}\mathbf{D}\,\mathbf{X}^*
#'   + \mathbf{U}^\top\boldsymbol{\Lambda}\mathbf{U},}
#' where \eqn{\mathbf{X}^* = \mathbf{X}\mathbf{U}} is the constrained design
#' (\eqn{N \times P}) and the first term is the quadratic approximation to the
#' penalized log-likelihood Hessian at the current \eqn{\boldsymbol{\mu}}.
#' For non-Gaussian families this is a local approximation (the IRWLS/Fisher
#' scoring Hessian), not an exact GLS information matrix.
#'
#' The constraint projection is
#' \eqn{\mathbf{U} = \mathbf{I} -
#'   \mathbf{G}\mathbf{A}(\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}
#'   \mathbf{A}^\top}
#' with \eqn{\mathbf{G} = \mathbf{M}^{-1}}.
#' \eqn{\mathbf{U}} is idempotent (\eqn{\mathbf{U}^2 = \mathbf{U}}) but
#' \emph{not} symmetric, so
#' \eqn{\mathbf{U}^\top\boldsymbol{\Lambda}\mathbf{U} \neq
#'   \mathbf{U}\boldsymbol{\Lambda}\mathbf{U}^\top}.
#'
#' The \code{U} stored in \code{model_fit$U} is on the expansion- and
#' response-standardised scale.  The rescaled version used here is
#' \deqn{\tilde{\mathbf{U}} =
#'   \mathbf{U} \cdot \mathrm{diag}(1, s_1, \ldots, s_{p-1},
#'     1, s_1, \ldots) / \hat{\sigma}_y,}
#' where \eqn{s_j} are the expansion scales and \eqn{\hat{\sigma}_y}
#' standardises the response.  All quantities below use
#' \eqn{\tilde{\mathbf{U}}} in place of \eqn{\mathbf{U}}.
#' }
#'
#' \subsection{REML objective and its gradient}{
#' The REML objective is constructed by integrating out the fixed effects from
#' the penalized log-likelihood, using a Laplace approximation to the
#' marginal likelihood for non-Gaussian families.  This approximation is exact
#' for Gaussian identity models and is the standard extension used in
#' restricted maximum likelihood estimation for GLMMs.
#'
#' The REML correction term is
#' \eqn{-\tfrac{1}{2}\log|\mathbf{M}|}, where \eqn{\mathbf{M}} is the
#' penalized observed information defined above.  Differentiating with respect
#' to \eqn{\rho} (noting only \eqn{\mathbf{V}} depends on \eqn{\rho})
#' gives the REML correction gradient
#' \deqn{-\frac{1}{2}\mathrm{tr}\!\Bigl(
#'   \mathbf{M}^{-1}
#'   (\mathbf{X}^*)^\top\mathbf{V}^{-1}
#'   \frac{\partial\mathbf{V}}{\partial\rho}
#'   \mathbf{V}^{-1}\mathbf{W}\mathbf{D}\,\mathbf{X}^*
#'   \Bigr).}
#' }
#'
#' \subsection{Full gradient}{
#' \deqn{\frac{\partial(-\ell_R)}{\partial\rho}
#'   = \frac{1}{N}\Biggl[
#'     \underbrace{
#'       \frac{1}{2}\mathrm{tr}\!\Bigl(
#'         \mathbf{V}^{-1}\frac{\partial\mathbf{V}}{\partial\rho}
#'       \Bigr)
#'     }_{\text{(i) log-det of }\mathbf{V}}
#'     \underbrace{
#'       -\frac{1}{2\tilde{\sigma}^2}
#'       \mathbf{r}^\top\frac{\partial\mathbf{V}}{\partial\rho}\mathbf{r}
#'     }_{\text{(ii) residual quadratic form}}
#'     \underbrace{
#'       -\frac{1}{2}\mathrm{tr}\!\Bigl(
#'         \mathbf{M}^{-1}(\mathbf{X}^*)^\top\mathbf{V}^{-1}
#'         \frac{\partial\mathbf{V}}{\partial\rho}
#'         \mathbf{V}^{-1}\mathbf{W}\mathbf{D}\,\mathbf{X}^*
#'       \Bigr)
#'     }_{\text{(iii) REML correction}}
#'   \Biggr],}
#' where the whitened residual is
#' \deqn{\mathbf{r} =
#'   \mathrm{diag}\!\left(\sqrt{d_i}/\sqrt{w_i}\right)
#'   \mathbf{V}^{-1/2}(\mathbf{y} - \boldsymbol{\mu}),}
#' and \eqn{r_i =
#'   [\mathbf{V}^{-1/2}(\mathbf{y}-\boldsymbol{\mu})]_i
#'   \sqrt{d_i}/\sqrt{w_i}}.
#'
#' Term (i) does not involve \eqn{\mathbf{D}} or \eqn{\mathbf{W}}; the
#' log-determinant of \eqn{\mathbf{V}} depends only on the correlation
#' structure.  For non-Gaussian families terms (ii) and (iii) are evaluated
#' at the current IRWLS iterate and constitute a local approximation.
#' }
#'
#' @param dV \eqn{N \times N} numeric matrix giving
#'   \eqn{\partial\mathbf{V}/\partial\rho} for one scalar parameter
#'   \eqn{\rho}, with the chain rule through any reparameterization already
#'   applied.
#' @param model_fit A fitted \code{lgspline} object; see
#'   \code{\link{lgspline}}.
#' @param glm_weight_function The GLM weight function used during model
#'   fitting; see the \code{glm_weight_function} argument of
#'   \code{\link{lgspline}}.
#' @param ... Additional arguments forwarded to \code{glm_weight_function}.
#' @return A scalar: the gradient of the negative REML objective with respect
#'   to \eqn{\rho}, divided by \eqn{N}.
#' @seealso \code{\link{lgspline}}, \code{\link{lgspline.fit}}
#' @export
reml_grad_from_dV <- function(dV, model_fit, glm_weight_function, ...) {

  ## (ii) Residual quadratic form  -0.5 * r' (dV) r / sigma^2
  #
  #  r_i = [V^{-1/2} (y - mu)]_i * sqrt(d_i) / sqrt(w_i)
  #
  #  rep(1, N) is passed for observation_weights inside glm_weight_function
  #  to avoid double-counting D; D enters via sqrt(model_fit$weights) instead.
  #  For logistic: w_i = mu_i(1-mu_i), so r downweights residuals in regions
  #  of high curvature and upweights by observation weight d_i.
  glm_wt_inv <- sqrt(c(glm_weight_function(model_fit$ytilde,
                                           model_fit$y,
                                           1:model_fit$N,
                                           model_fit$family,
                                           model_fit$sigmasq_tilde,
                                           rep(1, model_fit$N),
                                           ...))) /
    sqrt(unlist(model_fit$weights)[model_fit$og_order])

  resid     <- model_fit$y - model_fit$ytilde
  VinvResid <- model_fit$VhalfInv %**% cbind(resid) / glm_wt_inv

  quad_term <- -0.5 * ((t(VinvResid) %**% dV) %**% VinvResid) /
    model_fit$sigmasq_tilde

  ## (i) Log-determinant trace  0.5 * tr(V^{-1} dV/dtheta)
  #
  #  tr(V^{-1} dV) = sum(VhalfInv * (VhalfInv %*% dV))  element-wise
  #  no D or W involvement -- log|V| depends only on the correlation structure
  trace_term <- 0.5 * sum((model_fit$VhalfInv %**% dV) * model_fit$VhalfInv)

  ## (iii) REML correction  -0.5 * tr( M^{-1} (X*)' V^{-1} dV V^{-1} W D X* )
  #
  #  Rescale U to match the unstandardised X in model_fit$X:
  #    U_tilde = U * diag(1, s_1,...,s_{p-1}, 1, s_1,...) / sd_y
  #
  #  VhalfInvX = V^{-1/2} X* = V^{-1/2} X U_tilde   (N x P)
  U <- t(t(model_fit$U) *
           rep(c(1, model_fit$expansion_scales), model_fit$K + 1)) /
    model_fit$sd_y

  VhalfInvX <- model_fit$VhalfInv %**%
    collapse_block_diagonal(model_fit$X)[unlist(model_fit$og_order), ] %**%
    U

  ## Raw block-diagonal penalty on the standardised scale
  if (length(model_fit$penalties$L_partition_list) != (model_fit$K + 1)) {
    model_fit$penalties$L_partition_list <-
      lapply(1:(model_fit$K + 1), function(k) 0)
  }

  Lambda_raw <- collapse_block_diagonal(
    lapply(1:(model_fit$K + 1), function(k)
      c(1, model_fit$expansion_scales) *
        (model_fit$penalties$L_partition_list[[k]] +
           model_fit$penalties$Lambda) %**%
        diag(c(1, model_fit$expansion_scales)) /
        model_fit$sd_y^2
    )
  )

  ## Penalty projected onto constrained subspace: U' Lambda_raw U
  #  U is NOT symmetric so  U' L U  !=  U L U'
  ULambdaU <- t(U) %**% Lambda_raw %**% U

  ## Gram block of M:  (X*)' V^{-1} W D X*
  #    = VhalfInvX' diag(w_i * d_i) VhalfInvX
  #
  #  glm_wt_i = sqrt(w_i * d_i); for logistic sqrt(mu_i(1-mu_i) * d_i).
  #  Scaling rows of VhalfInvX by glm_wt then taking crossprod gives
  #  VhalfInvX' diag(w_i d_i) VhalfInvX = (X*)' V^{-1} W D X* exactly.
  #  rep(1, N) passed again to avoid double-counting D.
  glm_wt <- sqrt(c(glm_weight_function(model_fit$ytilde,
                                       model_fit$y,
                                       1:model_fit$N,
                                       model_fit$family,
                                       model_fit$sigmasq_tilde,
                                       rep(1, model_fit$N),
                                       ...))) *
    sqrt(unlist(model_fit$weights)[model_fit$og_order])

  XVinvX_inv <- invert(crossprod(t(t(VhalfInvX) * c(glm_wt))) + ULambdaU)

  ## Trace computation, numerically stabilised by spectral norm of VInvX
  #
  #  VInvX    = V^{-1} X* = V^{-1/2} VhalfInvX   (N x P)
  #  VInvX_sc = VInvX / sc,  sc = ||VInvX||_2
  #
  #  tr( M^{-1} VInvX' dV VInvX )
  #    = sc^2 * sum( diag( (M^{-1} VInvX_sc') (dV VInvX_sc) ) )
  VInvX <- model_fit$VhalfInv %**% VhalfInvX
  sc    <- sqrt(norm(VInvX, "2"))
  VInvX <- VInvX / sc

  dXVinvX     <- (XVinvX_inv %**% t(VInvX)) %**% (dV %**% VInvX)
  XVinvX_term <- -0.5 * colSums(cbind(c(diag(dXVinvX) * sc))) * sc

  as.numeric(quad_term + trace_term + XVinvX_term) / model_fit$N
}


#' Safely Replace Variable Names in Printed Terms
#'
#' Replaces a variable name only when it appears as a standalone term,
#' avoiding accidental replacements inside other words.
#'
#' Useful for plot formula labels, for example replacing "t" with "that"
#' without changing "intercept" into "inthatercept".
#'
#' @param x Character vector to modify.
#' @param old Character scalar, original variable name.
#' @param new Character scalar, replacement variable name.
#'
#' @return Character vector with safe term-wise replacements.
#'
#' @keywords internal
safe_replace_var <- function(x, old, new) {
  old_esc <- gsub("([][{}()+*^$|\\\\.?])", "\\\\\\1", old)
  gsub(
    paste0("(?<![[:alnum:]_])", old_esc, "(?![[:alnum:]_])"),
    new,
    x,
    perl = TRUE
  )
}

## solver_utils.R
#  Shared helpers used by get_B() and blockfit_solve().
#
## Main pieces
#   .solver_build_lambda_block()
#   .solver_detect_qp_global()
#   .solver_recompute_G_at_estimate()
#   .solver_assemble_qp_info()




#' Build Block-Diagonal Penalty Matrix
#'
#' Assembles the full \eqn{P \times P} block-diagonal penalty matrix
#' \eqn{\boldsymbol{\Lambda}} from a shared per-partition penalty
#' \code{Lambda} and optional partition-specific additive terms.
#' \eqn{P = p \times (K+1)} where \eqn{p} is the number of basis terms
#' per partition.
#'
#' When \code{unique_penalty_per_partition = TRUE}, the \eqn{k}-th
#' diagonal block is \code{Lambda + L_partition_list[[k]]}; otherwise
#' every block is \code{Lambda}.
#'
#' @param Lambda Shared \eqn{p \times p} penalty matrix.
#' @param K Integer; number of interior knots (\eqn{K+1} partitions).
#' @param unique_penalty_per_partition Logical; if \code{TRUE}, add the
#'   corresponding element of \code{L_partition_list} to each block.
#' @param L_partition_list List of \eqn{K+1} partition-specific
#'   \eqn{p \times p} penalty matrices.
#'
#' @return A \eqn{P \times P} block-diagonal matrix,
#'   \eqn{P = p \times (K+1)}.
#'
#' @keywords internal
.solver_build_lambda_block <- function(Lambda,
                                       K,
                                       unique_penalty_per_partition,
                                       L_partition_list) {
  ## Lay down one penalty block per partition, with zeros off the diagonal.
  Reduce("rbind", lapply(seq_len(K + 1L), function(k) {
    Reduce("cbind", lapply(seq_len(K + 1L), function(j) {
      if (k != j) return(0 * Lambda)
      if (unique_penalty_per_partition) Lambda + L_partition_list[[k]]
      else Lambda
    }))
  }))
}




#' Detect Whether Inequality Constraints Require a Dense Global QP
#'
#' Inspects the columns of \code{qp_Amat} to determine whether every
#' inequality constraint is confined to a single partition block
#' (block-separable) or whether any constraint couples coefficients
#' across partitions.
#'
#' Returns \code{FALSE} (partition-wise active-set is valid) when all
#' columns have nonzeros in at most one block.  Returns \code{TRUE}
#' (dense SQP required) when any column spans multiple blocks.
#'
#' @param qp_Amat Inequality constraint matrix
#'   (\eqn{P \times n_{\mathrm{ineq}}}), or \code{NULL}.
#' @param p_expansions Integer; number of basis terms per partition.
#' @param K Integer; number of interior knots.
#'
#' @return Logical scalar.
#'
#' @keywords internal
.solver_detect_qp_global <- function(qp_Amat, p_expansions, K) {
  if (is.null(qp_Amat)) return(FALSE)
  if (!is.matrix(qp_Amat) || ncol(qp_Amat) == 0L) return(FALSE)

  tol <- sqrt(.Machine$double.eps)

  ## Scan each constraint column and see how many partition blocks it touches.
  for (j in seq_len(ncol(qp_Amat))) {
    nz <- which(abs(qp_Amat[, j]) > tol)
    if (length(nz) == 0L) next
    blocks_hit <- unique(ceiling(nz / p_expansions))
    if (length(blocks_hit) > 1L) return(TRUE)
  }
  FALSE
}




#' Recompute G, Ghalf, and GhalfInv at a Supplied Coefficient Estimate
#'
#' Given current constrained coefficient estimates \code{result},
#' recomputes the penalized information matrix
#' \eqn{\mathbf{G}_k = (\mathbf{X}_k^{\top}\mathbf{W}_k
#' \mathbf{D}_k\mathbf{X}_k + \boldsymbol{\Lambda}_k)^{-1}} and its
#' matrix square roots for each partition.
#'
#' This is used after Newton-Raphson convergence in Path 3 of
#' \code{get_B()} and at the final return in \code{blockfit_solve()}
#' when \code{return_G_getB = TRUE}.  The implementation matches
#' \code{.recompute_G_at_estimate()} exactly; it exists here so both
#' solvers can call the same numerical core.
#'
#' @param X List of partition-specific design matrices
#'   \eqn{\mathbf{X}_k}.
#' @param y List of response vectors \eqn{\mathbf{y}_k} by partition.
#' @param result List of current coefficient column vectors
#'   \eqn{\tilde{\boldsymbol{\beta}}_k} by partition.
#' @param K Integer; number of interior knots.
#' @param Lambda Shared \eqn{p \times p} penalty matrix.
#' @param family GLM family object.
#' @param order_list List of index vectors mapping partition rows to
#'   original data ordering.
#' @param glm_weight_function Function computing GLM working weights
#'   \eqn{\mathbf{W}_k}.
#' @param schur_correction_function Function computing Schur corrections
#'   to the information matrix.
#' @param need_dispersion_for_estimation Logical; if \code{TRUE},
#'   estimate dispersion before computing weights.
#' @param dispersion_function Dispersion estimation function.
#' @param observation_weights List of observation weights
#'   \eqn{\mathbf{D}_k} by partition.
#' @param VhalfInv Inverse square root correlation matrix, or
#'   \code{NULL}.
#' @param parallel_eigen,parallel_matmult Logical flags for parallel
#'   computation.
#' @param cl Parallel cluster object.
#' @param chunk_size,num_chunks,rem_chunks Parallel distribution
#'   parameters.
#' @param unique_penalty_per_partition Logical.
#' @param L_partition_list List of partition-specific penalty matrices.
#' @param ... Passed to weight, correction, and dispersion functions.
#'
#' @return A list with components \code{G}, \code{Ghalf}, and
#'   \code{GhalfInv}, each a list of \eqn{K+1} matrices.
#'   \code{G[[k]]} is computed as \code{tcrossprod(Ghalf[[k]])} to
#'   guarantee exact symmetry.
#'
#' @keywords internal
.solver_recompute_G_at_estimate <- function(X,
                                             y,
                                             result,
                                             K,
                                             Lambda,
                                             family,
                                             order_list,
                                             glm_weight_function,
                                             schur_correction_function,
                                             need_dispersion_for_estimation,
                                             dispersion_function,
                                             observation_weights,
                                             VhalfInv,
                                             parallel_eigen,
                                             parallel_matmult,
                                             cl,
                                             chunk_size,
                                             num_chunks,
                                             rem_chunks,
                                             unique_penalty_per_partition,
                                             L_partition_list,
                                             ...) {

  ## Some families need a current dispersion estimate before their
  #  working weights can be recomputed.
  if (need_dispersion_for_estimation) {
    mu <- family$linkinv(
      unlist(
        matmult_block_diagonal(X, result, K, parallel_matmult,
                               cl, chunk_size, num_chunks, rem_chunks)))
    dispersion_temp <- dispersion_function(
      mu                  = mu,
      y                   = unlist(y),
      order_indices       = unlist(order_list),
      family              = family,
      observation_weights = unlist(observation_weights),
      VhalfInv            = VhalfInv,
      ...)
  } else {
    dispersion_temp <- 1
  }

  ## Form the weighted design X_k * sqrt(W_k) for each partition.
  #  The Gram matrix X_k^T W_k D_k X_k follows from crossprod of this.
  Xw <- lapply(seq_len(K + 1L), function(k) {
    if (nrow(X[[k]]) == 0L) return(X[[k]])
    var <- glm_weight_function(
      family$linkinv(X[[k]] %**% cbind(c(result[[k]]))),
      y[[k]], order_list[[k]], family, dispersion_temp,
      observation_weights[[k]], ...)
    if (length(var) == 1L) {
      if (c(var) == 0) return(X[[k]] * 0)
      return(X[[k]] * c(sqrt(var)))
    }
    X[[k]] * c(sqrt(var))
  })

  X_gram <- compute_gram_block_diagonal(Xw, parallel_matmult, cl,
                                        chunk_size, num_chunks, rem_chunks)

  ## Rebuild any Schur corrections at the current coefficient estimate
  #  before handing the weighted Gram matrices back to compute_G_eigen.
  schur_corrections <- schur_correction_function(
    X, y, result, dispersion_temp, order_list, K, family,
    observation_weights, ...)

  G_list <- compute_G_eigen(X_gram, Lambda, K, parallel_eigen, cl,
                            chunk_size, num_chunks, rem_chunks, family,
                            unique_penalty_per_partition, L_partition_list,
                            keep_G = TRUE, schur_corrections)

  ## Recompute G from Ghalf to avoid rounding asymmetry that can arise
  ## when compute_G_eigen returns G and Ghalf through separate paths.
  G_list$G <- lapply(G_list$Ghalf, tcrossprod)

  G_list
}




#' Assemble qp_info from a solve.QP Solution
#'
#' Packages the output of \code{quadprog::solve.QP} into the
#' \code{qp_info} list expected by downstream code (inference,
#' \code{generate_posterior}, and \code{varcovmat} construction).
#'
#' Active constraint columns are identified as all equality columns
#' (\code{1:qp_meq_combined}) plus any inequality column whose
#' Lagrange multiplier exceeds \code{sqrt(.Machine$double.eps)}.
#' This matches the convention used in \code{.qp_refine()}.
#'
#' @param last_qp_sol Output of \code{quadprog::solve.QP}, or
#'   \code{NULL}.  When \code{NULL} the function returns \code{NULL}.
#' @param beta_block Final \eqn{P \times 1} coefficient vector.
#' @param qp_Amat_combined Combined equality + inequality constraint
#'   matrix.
#' @param qp_bvec_combined Combined constraint right-hand side.
#' @param qp_meq_combined Integer; number of leading equality
#'   constraints.
#' @param converged Logical; whether the outer loop converged.
#' @param final_deviance Scalar deviance at convergence.
#' @param info_matrix Optional information matrix; included in the
#'   returned list when non-\code{NULL}.
#'
#' @return A list with elements \code{solution}, \code{lagrangian},
#'   \code{active_constraints}, \code{iact}, \code{Amat_active},
#'   \code{bvec_active}, \code{meq_active}, \code{converged}, and
#'   \code{final_deviance}, plus \code{info_matrix},
#'   \code{Amat_combined}, \code{bvec_combined}, and
#'   \code{meq_combined} when \code{info_matrix} is supplied.
#'   Returns \code{NULL} if \code{last_qp_sol} is \code{NULL}.
#'
#' @keywords internal
.solver_assemble_qp_info <- function(last_qp_sol,
                                      beta_block,
                                      qp_Amat_combined,
                                      qp_bvec_combined,
                                      qp_meq_combined,
                                      converged,
                                      final_deviance,
                                      info_matrix = NULL) {
  if (is.null(last_qp_sol)) return(NULL)

  lag_mult <- last_qp_sol$Lagrangian
  n_eq     <- qp_meq_combined
  n_con    <- ncol(qp_Amat_combined)

  ## Active inequality columns: those with non-negligible multipliers.
  if (n_eq < n_con) {
    ineq_active <- n_eq + which(
      abs(lag_mult[-(seq_len(n_eq))]) > sqrt(.Machine$double.eps))
  } else {
    ineq_active <- integer(0L)
  }
  active_cols <- sort(unique(c(seq_len(n_eq), ineq_active)))

  ## Package the solve.QP output in the form used downstream by inference
  #  and posterior drawing code.
  out <- list(
    solution           = c(beta_block),
    lagrangian         = lag_mult,
    active_constraints = active_cols,
    iact               = last_qp_sol$iact,
    Amat_active        = qp_Amat_combined[, active_cols, drop = FALSE],
    bvec_active        = qp_bvec_combined[active_cols],
    meq_active         = qp_meq_combined,
    converged          = converged,
    final_deviance     = final_deviance
  )

  ## Keep the full combined constraint objects when the caller also wants
  #  the information matrix back.
  if (!is.null(info_matrix)) {
    out$info_matrix   <- info_matrix
    out$Amat_combined <- qp_Amat_combined
    out$bvec_combined <- qp_bvec_combined
    out$meq_combined  <- qp_meq_combined
  }

  out
}

#' Compute Log-Likelihood for Weibull Accelerated Failure Time Model
#'
#' @description
#' Calculates the log-likelihood for a Weibull accelerated failure time (AFT)
#' survival model, supporting right-censored survival data.
#'
#' @param log_y Numeric vector of logarithmic response/survival times
#' @param log_mu Numeric vector of logarithmic predicted survival times
#' @param status Numeric vector of censoring indicators
#'   (1 = event, 0 = censored)
#'   Indicates whether an event of interest occurred (1) or the observation was
#'   right-censored (0). In survival analysis, right-censoring occurs when the
#'   full survival time is unknown, typically because the study ended or the
#'   subject was lost to follow-up before the event of interest occurred.
#' @param scale Numeric scalar representing the Weibull scale parameter
#'   (sigma), equivalent to \code{survreg$scale}. This is the square root of
#'   the dispersion stored in \code{lgspline$sigmasq_tilde}.
#' @param weights Optional numeric vector of observation weights (default = 1)
#'
#' @return
#' A numeric scalar representing the weighted total log-likelihood of the model
#'
#' @details
#' The function computes log-likelihood contributions for a Weibull AFT model,
#' explicitly accounting for right-censored observations. It supports optional
#' observation weighting to accommodate complex sampling designs.
#'
#' This both provides a tool for actually fitting Weibull AFT models, and
#' boilerplate code for users who wish to incorporate Lagrangian multiplier
#' smoothing splines into their own custom models.
#'
#' @examples
#'
#' ## Minimal example of fitting a Weibull Accelerated Failure Time model
#' # Simulating survival data with right-censoring
#' set.seed(1234)
#' x1 <- rnorm(1000)
#' x2 <- rbinom(1000, 1, 0.5)
#' yraw <- rexp(exp(0.01*x1 + 0.01*x2))
#' # status: 1 = event occurred, 0 = right-censored
#' status <- rbinom(1000, 1, 0.25)
#' yobs <- ifelse(status, runif(length(yraw), 0, yraw), yraw)
#' df <- data.frame(
#'   y = yobs,
#'   x1 = x1,
#'   x2 = x2
#' )
#'
#' ## Fit model using lgspline with Weibull AFT specifics
#' model_fit <- lgspline(y ~ spl(x1) + x2,
#'                       df,
#'                       unconstrained_fit_fxn = unconstrained_fit_weibull,
#'                       family = weibull_family(),
#'                       need_dispersion_for_estimation = TRUE,
#'                       dispersion_function = weibull_dispersion_function,
#'                       glm_weight_function = weibull_glm_weight_function,
#'                       schur_correction_function = weibull_schur_correction,
#'                       status = status,
#'                       opt = FALSE,
#'                       K = 1)
#'
#' loglik_weibull(log(model_fit$y), log(model_fit$ytilde), status,
#'   sqrt(model_fit$sigmasq_tilde))
#'
#' @export
loglik_weibull <- function(log_y, log_mu, status, scale, weights = 1) {

  ## Log-likelihood contributions
  ## scale = sigma (Weibull scale), NOT sigma^2 (dispersion)
  z <- (log_y - log_mu) / scale
  logL <- status * (-log(scale) +
                      z -
                      log_y) -
    exp(z)

  ## Return sum of log likelihood
  return(sum(logL * weights))
}


#' Compute Gradient of Log-Likelihood of Weibull Accelerated Failure Model
#'
#' @description
#' Calculates the gradient of log-likelihood for a Weibull accelerated failure
#' time (AFT) survival model, supporting right-censored survival data.
#'
#' @param X Design matrix
#' @param y Response vector
#' @param mu Predicted mean vector
#' @param order_list List of observation indices per partition
#' @param dispersion Dispersion parameter (sigma^2 = scale^2). The lgspline
#'   framework stores and passes dispersion (sigma^2); the Weibull scale
#'   (sigma) matching \code{survreg$scale} is \code{sqrt(dispersion)}.
#' @param VhalfInv Inverse square root of correlation matrix; unused here and
#'   retained for interface compatibility.
#' @param observation_weights Observation weights
#' @param status Censoring indicator (1 = event, 0 = censored)
#'
#' @return
#' Numeric column vector representing the gradient with respect to
#' coefficients.
#'
#' @details
#' Needed if using "blockfit", correlation structures, or quadratic programming
#' with Weibull AFT models.
#'
#' The gradient is computed on a scale that omits the 1/sigma prefactor.
#' Specifically, the true score is (1/sigma) * X^T diag(w) (exp(z) - status),
#' but both this function and the corresponding information matrix used
#' internally omit 1/sigma and 1/sigma^2 respectively, so the Newton-Raphson
#' step G*u remains correct. This matches the convention in
#' \code{\link{unconstrained_fit_weibull}}.
#'
#' @examples
#'
#' set.seed(1234)
#' t1 <- rnorm(1000)
#' t2 <- rbinom(1000, 1, 0.5)
#' yraw <- rexp(exp(0.01*t1 + 0.01*t2))
#' status <- rbinom(1000, 1, 0.25)
#' yobs <- ifelse(status, runif(length(yraw), 0, yraw), yraw)
#' df <- data.frame(
#'   y = yobs,
#'   t1 = t1,
#'   t2 = t2
#' )
#'
#' ## Example using blockfit for t2 as a linear term - output does not look
#' # different, but internal methods used for fitting change
#' model_fit <- lgspline(y ~ spl(t1) + t2,
#'                       df,
#'                       unconstrained_fit_fxn = unconstrained_fit_weibull,
#'                       family = weibull_family(),
#'                       need_dispersion_for_estimation = TRUE,
#'                       qp_score_function = weibull_qp_score_function,
#'                       dispersion_function = weibull_dispersion_function,
#'                       glm_weight_function = weibull_glm_weight_function,
#'                       schur_correction_function = weibull_schur_correction,
#'                       K = 1,
#'                       blockfit = TRUE,
#'                       opt = FALSE,
#'                       status = status,
#'                       verbose = TRUE)
#'
#' print(summary(model_fit))
#'
#' @export
weibull_qp_score_function = function(X,
                                     y,
                                     mu,
                                     order_list,
                                     dispersion,
                                     VhalfInv,
                                     observation_weights,
                                     status){
  ## dispersion = sigma^2; scale = sigma = sqrt(dispersion)
  scale <- sqrt(dispersion)
  order_indices <- unlist(order_list)
  ## Score of beta (scaled by sigma^2 for numerical convenience).
  # True score: (1/sigma) * X^T diag(w) (exp(z) - status)
  #  Scaled score (x sigma^2): sigma * X^T diag(w) (exp(z) - status)
  #  This matches the convention in unconstrained_fit_weibull, where
  #  the information is also unscaled (missing 1/sigma^2), so the
  #  Newton-Raphson step G*u remains correct.
  crossprod(X , cbind(
    observation_weights *
      (exp((log(y) - log(mu)) / scale) - status[order_indices]) *
      scale)
  )
}

#' Correction for the Variance-Covariance Matrix for Uncertainty in Scale
#'
#' @description
#' Computes the Schur complement \eqn{\textbf{S}} such that
#' \eqn{\textbf{G}^* = (\textbf{G}^{-1} + \textbf{S})^{-1}} properly
#' accounts for uncertainty in estimating the Weibull scale parameter when
#' estimating the variance-covariance matrix. Otherwise, the
#' variance-covariance matrix is optimistic and assumes the scale is known,
#' when it was in fact estimated. Note that the parameterization adds the
#' output of this function elementwise (not subtract) so for most cases, the
#' output of this function will be negative or a negative
#' definite/semi-definite matrix.
#'
#' @param X Block-diagonal matrices of spline expansions
#' @param y Block-vector of response
#' @param B Block-vector of coefficient estimates
#' @param dispersion Scalar, estimate of dispersion (sigma^2 = scale^2). The
#'   lgspline framework stores and passes dispersion (sigma^2); the Weibull
#'   scale (sigma) matching \code{survreg$scale} is \code{sqrt(dispersion)}.
#' @param order_list List of partition orders
#' @param K Number of partitions minus 1 (\eqn{K})
#' @param family Distribution family
#' @param observation_weights Optional observation weights (default = 1)
#' @param status Censoring indicator (1 = event, 0 = censored)
#'   Indicates whether an event of interest occurred (1) or the observation was
#'   right-censored (0). In survival analysis, right-censoring occurs when the
#'   full survival time is unknown, typically because the study ended or the
#'   subject was lost to follow-up before the event of interest occurred.
#'
#' @return
#' List of blockwise Schur-complement corrections \eqn{\textbf{S}_k} to be
#' elementwise added to each block of the information matrix before inversion,
#' with \code{0} returned for empty partitions.
#'
#' @details
#' Adjusts the variance-covariance matrix unscaled for coefficients to account
#' for uncertainty in estimating the Weibull scale parameter, that otherwise
#' would be lost if simply using
#' \eqn{\textbf{G}=(\textbf{X}^{T}\textbf{W}\textbf{X} + \textbf{L})^{-1}}.
#' This is accomplished using a correction based on the Schur complement so we
#' avoid having to construct the entire variance-covariance matrix, or
#' modifying the procedure for \code{\link{lgspline}} substantially.
#' For any model with nuisance parameters that must have uncertainty accounted
#' for, this tool will be helpful.
#'
#' This both provides a tool for actually fitting Weibull accelerated failure
#' time (AFT) models, and boilerplate code for users who wish to incorporate
#' Lagrangian multiplier smoothing splines into their own custom models.
#'
#' @examples
#'
#' ## Minimal example of fitting a Weibull Accelerated Failure Time model
#' # Simulating survival data with right-censoring
#' set.seed(1234)
#' t1 <- rnorm(1000)
#' t2 <- rbinom(1000, 1, 0.5)
#' yraw <- rexp(exp(0.01*t1 + 0.01*t2))
#' # status: 1 = event occurred, 0 = right-censored
#' status <- rbinom(1000, 1, 0.25)
#' yobs <- ifelse(status, runif(length(yraw), 0, yraw), yraw)
#' df <- data.frame(
#'   y = yobs,
#'   t1 = t1,
#'   t2 = t2
#' )
#'
#' ## Fit model using lgspline with Weibull Schur correction
#' model_fit <- lgspline(y ~ spl(t1) + t2,
#'                       df,
#'                       unconstrained_fit_fxn = unconstrained_fit_weibull,
#'                       family = weibull_family(),
#'                       need_dispersion_for_estimation = TRUE,
#'                       dispersion_function = weibull_dispersion_function,
#'                       glm_weight_function = weibull_glm_weight_function,
#'                       schur_correction_function = weibull_schur_correction,
#'                       status = status,
#'                       opt = FALSE,
#'                       K = 1)
#'
#' print(summary(model_fit))
#'
#' ## Fit model using lgspline without Weibull Schur correction
#' naive_fit <- lgspline(y ~ spl(t1) + t2,
#'                       df,
#'                       unconstrained_fit_fxn = unconstrained_fit_weibull,
#'                       family = weibull_family(),
#'                       need_dispersion_for_estimation = TRUE,
#'                       dispersion_function = weibull_dispersion_function,
#'                       glm_weight_function = weibull_glm_weight_function,
#'                       status = status,
#'                       opt = FALSE,
#'                       K = 1)
#'
#' print(summary(naive_fit))
#'
#' @export
weibull_schur_correction <- function(X,
                                     y,
                                     B,
                                     dispersion,
                                     order_list,
                                     K,
                                     family,
                                     observation_weights,
                                     status){
  ## dispersion = sigma^2; scale = sigma = sqrt(dispersion)
  scale <- sqrt(dispersion)
  lapply(1:(K+1), function(k){
    if(nrow(X[[k]]) < 1){
      return(0)
    } else {
      mu <- family$linkinv(c(X[[k]] %**% B[[k]]))
      s <- status[order_list[[k]]]
      obs <- y[[k]]
      z <- (log(obs) - log(mu)) / scale
      exp_z <- exp(z)
      zexp_z <- z * exp_z
      weights <- c(observation_weights[[k]])

      ## Correction via Schur complement
      # Extract true conditional variance-covariance of beta coefficients
      # conditional upon estimate of scale (sigma).
      # I = ( I_bb I_bs^{T} )
      #     ( I_bs I_ss     )
      # for b = beta, s = scale (sigma)
      # Note that: I_bb = invert(G[[k]]) is incorrect, I_bb is part of
      # Fisher info
      I_bs <- t(X[[k]]) %**% cbind(weights * zexp_z * scale)
      I_ss <- -sum(
        weights * (
          (s + 2*s*z + zexp_z + exp_z * z^2)
        )
      )
      # compl gets elementwise added to G[[k]] for all k = 1...K+1
      compl <- I_bs %**% matrix(-1/I_ss) %**% t(I_bs)
      # Schur complement correction to pass on to compute_G_eigen()
      return(compl)
    }
  })
}

#' @rdname weibull_schur_correction
#' @export
weibull_shur_correction <- weibull_schur_correction

#' Estimate Scale for Weibull Accelerated Failure Time Model
#'
#' @description
#' Computes maximum log-likelihood scale estimate (sigma) for a Weibull
#' accelerated failure time (AFT) survival model.
#'
#' This both provides a tool for actually fitting Weibull AFT Models, and
#' boilerplate code for users who wish to incorporate Lagrangian multiplier
#' smoothing splines into their own custom models.
#'
#' @param log_y Logarithm of response/survival times
#' @param log_mu Logarithm of predicted survival times
#' @param status Censoring indicator (1 = event, 0 = censored)
#'   Indicates whether an event of interest occurred (1) or the observation was
#'   right-censored (0). In survival analysis, right-censoring occurs when the
#'   full survival time is unknown, typically because the study ended or the
#'   subject was lost to follow-up before the event of interest occurred.
#' @param weights Optional observation weights (default = 1)
#'
#' @return
#' Scalar representing the estimated Weibull scale (sigma), equivalent to
#' \code{survreg$scale}. The dispersion (as stored in
#' \code{lgspline$sigmasq_tilde}) is sigma^2.
#'
#' @details
#' Calculates maximum log-likelihood estimate of scale (sigma) for Weibull AFT
#' model accounting for right-censored observations using Brent's method for
#' optimization.
#'
#' @examples
#'
#' ## Simulate exponential data with censoring
#' set.seed(1234)
#' mu <- 2  # mean of exponential distribution
#' n <- 500
#' y <- rexp(n, rate = 1/mu)
#'
#' ## Introduce censoring (25% of observations)
#' status <- rbinom(n, 1, 0.75)
#' y_obs <- ifelse(status, y, NA)
#'
#' ## Compute scale estimate
#' scale_est <- weibull_scale(
#'   log_y = log(y_obs[!is.na(y_obs)]),
#'   log_mu = log(mu),
#'   status = status[!is.na(y_obs)]
#' )
#'
#' print(scale_est)
#'
#'
#' @export
weibull_scale <- function(log_y, log_mu, status, weights = 1){
  optim(
    1,
    fn = function(par){
      ## par is scale (sigma), passed directly to loglik_weibull
      -loglik_weibull(log_y, log_mu, status, par, weights)
    },
    method = 'Brent',
    lower = 1e-64,
    upper = 100
  )$par
}

## weibull_family, with $aic, $loglik, $dev.resids methods
#  for logLik.lgspline compatibility

#' Weibull Family for Survival Model Specification
#'
#' @description
#' Creates a family-like object for Weibull accelerated failure time (AFT)
#' models, including custom log-likelihood, AIC, and deviance helpers.
#'
#' This both provides a tool for actually fitting Weibull AFT Models, and
#' boilerplate code for users who wish to incorporate Lagrangian multiplier
#' smoothing splines into their own custom models.
#'
#' @return
#' A family-like list containing the link functions and Weibull-specific
#' methods used by \code{lgspline}.
#'
#' @details
#' Provides a comprehensive family specification for Weibull AFT models,
#' including family name, link function, inverse link function, custom loss
#' function for model tuning, and methods for AIC and log-likelihood
#' computation compatible with \code{logLik.lgspline}.
#'
#' Supports right-censored survival data with flexible parameter estimation.
#'
#' Note on scale vs. dispersion: throughout this package, the lgspline object
#' stores \code{sigmasq_tilde} which equals sigma^2 (dispersion), where
#' sigma is the Weibull scale parameter matching \code{survreg$scale}.
#' Functions that accept a \code{dispersion} argument receive sigma^2;
#' functions that accept a \code{scale} argument receive sigma.
#'
#' @examples
#'
#' ## Simulate survival data with covariates
#' set.seed(1234)
#' n <- 1000
#' t1 <- rnorm(n)
#' t2 <- rbinom(n, 1, 0.5)
#'
#' ## Generate survival times with Weibull-like structure
#' lambda <- exp(0.5 * t1 + 0.3 * t2)
#' yraw <- rexp(n, rate = 1/lambda)
#'
#' ## Introduce right-censoring
#' status <- rbinom(n, 1, 0.75)
#' y <- ifelse(status, yraw, runif(length(yraw), 0, yraw))
#'
#' ## Prepare data
#' df <- data.frame(y = y, t1 = t1, t2 = t2, status = status)
#'
#' ## Fit model using custom Weibull family
#' model_fit <- lgspline(y ~ spl(t1) + t2,
#'                       df,
#'                       unconstrained_fit_fxn = unconstrained_fit_weibull,
#'                       family = weibull_family(),
#'                       need_dispersion_for_estimation = TRUE,
#'                       dispersion_function = weibull_dispersion_function,
#'                       glm_weight_function = weibull_glm_weight_function,
#'                       schur_correction_function = weibull_schur_correction,
#'                       status = status,
#'                       opt = FALSE,
#'                       K = 1)
#'
#' summary(model_fit)
#'
#' ## Log-likelihood now works via logLik.lgspline:
#' # logLik(model_fit)
#'
#' @export
weibull_family <- function() list(
  family = "weibull",
  link = "log",
  linkfun = log,
  linkinv = exp,

  ## Added dev.resids for base R
  #  compatibility when logLik.lgspline checks for family$dev.resids
  dev.resids = function(y, mu, wt) {
    ## Deviance residuals for Weibull AFT: -2 * log-likelihood
    #  This is a simplified version; full version uses status via
    #  custom_dev.resids below
    wt * (y/mu - log(y/mu) - 1) * 2
  },

  ## Custom loss used in place of MSE for computing GCV
  custom_dev.resids =
    function(y, mu, order_indices, family, observation_weights, status){
      log_mu <- log(mu)
      log_y <- log(y)
      status <- status[order_indices]

      ## Initialize scale (sigma) using null model
      init_scale <-
        weibull_scale(log_y, mean(log_y), status,
                      observation_weights)
      ## Find scale (sigma) given current mu
      scale <- optim(
        init_scale,
        fn = function(par){
          ## par is scale (sigma)
          -loglik_weibull(log_y, log_mu, status, par, observation_weights)
        },
        lower = init_scale/10,
        upper = init_scale*10,
        method = 'Brent'
      )$par

      ## -2 * log-likelihood
      dev <- -2*(
        status * (-log(scale) +
                    (1/scale - 1)*log_y -
                    log_mu/scale) -
          (exp((log_y - log_mu)/scale))
      )
      return(dev * observation_weights)
    },

  ## Added $loglik method for
  #  logLik.lgspline compatibility
  #  Accepts a fitted lgspline model object and returns the scalar
  #  log-likelihood value
  loglik = function(model_fit) {
    ## Extract necessary components from the fitted model
    log_y <- log(model_fit$y)
    log_mu <- log(model_fit$ytilde)
    ## sigmasq_tilde is dispersion (sigma^2); take sqrt to get scale (sigma)
    scale <- sqrt(model_fit$sigmasq_tilde)

    ## Status must be stored in model_fit$extra or model_fit$call_args
    #  Try common locations
    status <- NULL
    if(!is.null(model_fit$status)) {
      status <- model_fit$status
    } else if(!is.null(model_fit$extra_args$status)) {
      status <- model_fit$extra_args$status
    } else {
      ## If status not found, assume all events observed
      warning("Censoring status not found in model object; ",
              "assuming all events observed (status = 1).")
      status <- rep(1, length(log_y))
    }

    weights <- if(!is.null(model_fit$observation_weights)){
      model_fit$observation_weights
    } else {
      rep(1, length(log_y))
    }

    loglik_weibull(log_y, log_mu, status, scale, weights)
  },

  ## Added $aic method for
  #  logLik.lgspline compatibility
  #  Returns AIC = -2*loglik + 2*edf
  #  edf = effective degrees of freedom (trace of hat matrix + 1 for scale)
  aic = function(model_fit) {
    ll <- weibull_family()$loglik(model_fit)
    ## edf: trace of smoother matrix + 1 for scale parameter
    edf <- model_fit$edf
    if(is.null(edf)){
      ## Fallback: use trace_XUGX if available
      edf <- model_fit$trace_XUGX
    }
    if(is.null(edf)){
      warning("Cannot compute AIC: effective degrees of freedom not found.")
      return(NA)
    }
    ## +1 for the scale parameter
    -2 * ll + 2 * (edf + 1)
  }
)


#' Estimate Weibull Dispersion for Accelerated Failure Time Model
#'
#' @description
#' Computes the dispersion parameter (sigma^2 = scale^2) for a Weibull
#' accelerated failure time (AFT) model, supporting right-censored survival
#' data. The returned value is sigma^2, where sigma is the Weibull scale
#' parameter matching \code{survreg$scale}.
#'
#' This both provides a tool for actually fitting Weibull AFT Models, and
#' boilerplate code for users who wish to incorporate Lagrangian multiplier
#' smoothing splines into their own custom models.
#'
#' @param mu Predicted survival times
#' @param y Observed response/survival times
#' @param order_indices Indices to align status with response
#' @param family Weibull AFT model family specification; unused here and
#'   retained for interface compatibility.
#' @param observation_weights Optional observation weights
#' @param VhalfInv Inverse square root of the correlation matrix; unused here
#'   and retained for interface compatibility.
#' @param status Censoring indicator (1 = event, 0 = censored)
#'   Indicates whether an event of interest occurred (1) or the observation was
#'   right-censored (0). In survival analysis, right-censoring occurs when the
#'   full survival time is unknown, typically because the study ended or the
#'   subject was lost to follow-up before the event of interest occurred.
#'
#' @return
#' Dispersion estimate (sigma^2) for the Weibull AFT model, i.e., the squared
#' scale parameter. The Weibull scale (sigma) matching \code{survreg$scale} is
#' \code{sqrt()} of this value.
#'
#' @seealso \code{\link{weibull_scale}} for the underlying scale estimation
#'   function
#'
#' @examples
#'
#' ## Simulate survival data with covariates
#' set.seed(1234)
#' n <- 1000
#' t1 <- rnorm(n)
#' t2 <- rbinom(n, 1, 0.5)
#'
#' ## Generate survival times with Weibull-like structure
#' lambda <- exp(0.5 * t1 + 0.3 * t2)
#' yraw <- rexp(n, rate = 1/lambda)
#'
#' ## Introduce right-censoring
#' status <- rbinom(n, 1, 0.75)
#' y <- ifelse(status, yraw, runif(length(yraw), 0, yraw))
#'
#' ## Example of using dispersion function
#' mu <- mean(y)
#' order_indices <- seq_along(y)
#' weights <- rep(1, n)
#'
#' ## Estimate dispersion (= scale^2 = sigma^2)
#' dispersion_est <- weibull_dispersion_function(
#'   mu = mu,
#'   y = y,
#'   order_indices = order_indices,
#'   family = weibull_family(),
#'   observation_weights = weights,
#'   VhalfInv = NULL,
#'   status = status
#' )
#'
#' print(dispersion_est)          # sigma^2
#' print(sqrt(dispersion_est))    # sigma (comparable to survreg$scale)
#'
#' @export
weibull_dispersion_function <- function(mu,
                                        y,
                                        order_indices,
                                        family,
                                        observation_weights,
                                        VhalfInv,
                                        status){

  ## Maximizes log-likelihood of right-censored data
  log_mu <- log(mu)
  log_y <- log(y)
  observation_weights <- c(observation_weights)
  status <- status[order_indices]

  ## Initialize scale (sigma) using null model
  init_scale <-
    weibull_scale(log_y,
                  mean(log_y),
                  status,
                  observation_weights)
  ## Find scale (sigma) given current mu
  scale <- optim(
    init_scale,
    fn = function(par){
      ## par is scale (sigma)
      -loglik_weibull(log_y,
                      log_mu,
                      status,
                      par,
                      observation_weights)
    },
    lower = init_scale/10, # Increased from 5 to 10
    upper = init_scale*10,
    method = 'Brent'
  )$par

  ## Return dispersion = sigma^2 = scale^2
  return(scale^2)
}

#' Weibull GLM Weight Function for Constructing Information Matrix
#'
#' @description
#' Computes the working weights used in the Weibull AFT information matrix,
#' including the observation-weight contribution returned on the vector scale.
#'
#' @param mu Predicted survival times
#' @param y Observed response/survival times
#' @param order_indices Order of observations when partitioned to match
#'   "status" to "response"
#' @param family Weibull AFT family; unused here and retained for interface
#'   compatibility.
#' @param dispersion Estimated dispersion parameter (sigma^2 = scale^2). The
#'   lgspline framework stores and passes dispersion (sigma^2); the Weibull
#'   scale (sigma) matching \code{survreg$scale} is \code{sqrt(dispersion)}.
#' @param observation_weights Weights of observations submitted to function
#' @param status Censoring indicator (1 = event, 0 = censored)
#'   Indicates whether an event of interest occurred (1) or the observation was
#'   right-censored (0). In survival analysis, right-censoring occurs when the
#'   full survival time is unknown, typically because the study ended or the
#'   subject was lost to follow-up before the event of interest occurred.
#'
#' @return
#' Numeric vector of working weights for the information matrix, including
#' observation weights when finite and a fallback of 1s when the natural
#' Weibull weights are non-finite.
#'
#' @details
#' This function generates weights used in constructing the information matrix
#' after unconstrained estimates have been found. Specifically, it is used in
#' the construction of the \eqn{\textbf{U}} and \eqn{\textbf{G}} matrices
#' following initial unconstrained parameter estimation.
#'
#' These weights are analogous to the variance terms in generalized linear
#' models (GLMs). Like logistic regression uses \eqn{\mu(1-\mu)}, Poisson
#' regression uses \eqn{e^{\mu}}, and Linear regression uses constant weights,
#' Weibull AFT models use \eqn{\exp((\log y - \log \mu)/\sigma)} where
#' \eqn{\sigma = \sqrt{\text{dispersion}}} is the scale parameter.
#'
#' @examples
#'
#' ## Demonstration of glm weight function in constrained model estimation
#' set.seed(1234)
#' n <- 1000
#' t1 <- rnorm(n)
#' t2 <- rbinom(n, 1, 0.5)
#'
#' ## Generate survival times
#' lambda <- exp(0.5 * t1 + 0.3 * t2)
#' yraw <- rexp(n, rate = 1/lambda)
#'
#' ## Introduce right-censoring
#' status <- rbinom(n, 1, 0.75)
#' y <- ifelse(status, yraw, runif(length(yraw), 0, yraw))
#'
#' ## Fit model demonstrating use of custom glm weight function
#' model_fit <- lgspline(y ~ spl(t1) + t2,
#'                       data.frame(y = y, t1 = t1, t2 = t2),
#'                       unconstrained_fit_fxn = unconstrained_fit_weibull,
#'                       family = weibull_family(),
#'                       need_dispersion_for_estimation = TRUE,
#'                       dispersion_function = weibull_dispersion_function,
#'                       glm_weight_function = weibull_glm_weight_function,
#'                       schur_correction_function = weibull_schur_correction,
#'                       status = status,
#'                       opt = FALSE,
#'                       K = 1)
#'
#' print(summary(model_fit))
#'
#'
#' @export
weibull_glm_weight_function <- function(mu,
                                        y,
                                        order_indices,
                                        family,
                                        dispersion,
                                        observation_weights,
                                        status){
  ## dispersion = sigma^2; scale = sigma = sqrt(dispersion)
  scale <- sqrt(dispersion)
  val <- exp((log(y) - log(mu)) / scale)
  if(any(!is.finite(val))){
    return(rep(1, length(val)))
  }
  newval <- val * c(observation_weights)
  return(newval)
}

#' Unconstrained Weibull Accelerated Failure Time Model Estimation
#'
#' @description
#' Estimates parameters for an unconstrained Weibull accelerated failure time
#' (AFT) model supporting right-censored survival data.
#'
#' This both provides a tool for actually fitting Weibull AFT Models, and
#' boilerplate code for users who wish to incorporate Lagrangian multiplier
#' smoothing splines into their own custom models.
#'
#' @param X Design matrix of predictors
#' @param y Survival/response times
#' @param LambdaHalf Square root of penalty matrix
#'   (\eqn{\boldsymbol{\Lambda}^{1/2}})
#' @param Lambda Penalty matrix (\eqn{\boldsymbol{\Lambda}})
#' @param keep_weighted_Lambda Flag to retain weighted penalties
#' @param family Distribution family specification
#' @param tol Convergence tolerance (default 1e-8)
#' @param K Number of partitions minus one (\eqn{K})
#' @param parallel Flag for parallel processing
#' @param cl Cluster object for parallel computation
#' @param chunk_size Processing chunk size
#' @param num_chunks Number of computational chunks
#' @param rem_chunks Remaining chunks
#' @param order_indices Observation ordering indices
#' @param weights Optional observation weights
#' @param status Censoring status indicator (1 = event, 0 = censored)
#'   Indicates whether an event of interest occurred (1) or the observation was
#'   right-censored (0). In survival analysis, right-censoring occurs when the
#'   full survival time is unknown, typically because the study ended or the
#'   subject was lost to follow-up before the event of interest occurred.
#'
#' @return
#' Numeric column vector of unconstrained coefficient estimates for the Weibull
#' AFT model.
#'
#' @details
#' Estimation Approach:
#' The function employs a two-stage optimization strategy for fitting
#' accelerated failure time models via maximum likelihood:
#'
#' 1. Outer Loop: Estimate Scale Parameter (sigma) using Brent's method
#'
#' 2. Inner Loop: Estimate Regression Coefficients (given sigma) using
#'    damped Newton-Raphson.
#'
#' Note: the score and information inside the Newton-Raphson are both scaled
#' by \eqn{\sigma^2} (i.e., both omit the \eqn{1/\sigma} and
#' \eqn{1/\sigma^2} prefactors respectively). Since the Newton-Raphson step is
#' \eqn{\mathbf{G}u = (\mathbf{X}^{\top}\mathbf{W}\mathbf{X})^{-1}
#' \mathbf{X}^{\top}\mathbf{v}}, the \eqn{\sigma^2} factors cancel and the
#' step remains correct.
#'
#' @examples
#'
#' ## Simulate survival data with covariates
#' set.seed(1234)
#' n <- 1000
#' t1 <- rnorm(n)
#' t2 <- rbinom(n, 1, 0.5)
#'
#' ## Generate survival times with Weibull-like structure
#' lambda <- exp(0.5 * t1 + 0.3 * t2)
#' yraw <- rexp(n, rate = 1/lambda)
#'
#' ## Introduce right-censoring
#' status <- rbinom(n, 1, 0.75)
#' y <- ifelse(status, yraw, runif(length(yraw), 0, yraw))
#' df <- data.frame(y = y, t1 = t1, t2 = t2)
#'
#' ## Fit model using lgspline with Weibull AFT unconstrained estimation
#' model_fit <- lgspline(y ~ spl(t1) + t2,
#'                       df,
#'                       unconstrained_fit_fxn = unconstrained_fit_weibull,
#'                       family = weibull_family(),
#'                       need_dispersion_for_estimation = TRUE,
#'                       dispersion_function = weibull_dispersion_function,
#'                       glm_weight_function = weibull_glm_weight_function,
#'                       schur_correction_function = weibull_schur_correction,
#'                       status = status,
#'                       opt = FALSE,
#'                       K = 1)
#'
#' ## Print model summary
#' summary(model_fit)
#'
#' @keywords internal
#' @export
unconstrained_fit_weibull <- function(X,
                                      y,
                                      LambdaHalf,
                                      Lambda,
                                      keep_weighted_Lambda,
                                      family,
                                      tol = 1e-8,
                                      K,
                                      parallel,
                                      cl,
                                      chunk_size,
                                      num_chunks,
                                      rem_chunks,
                                      order_indices,
                                      weights,
                                      status # passed through for censored Weibull fits
) {

  ## Weight if non-null
  if(any(!is.null(weights))){
    if(length(weights) > 0){
      weights <- c(weights)
    }
  } else {
    weights <- rep(1, length(y))
  }

  log_y <- log(y)

  ## Initialize scale (sigma)
  init_scale <- weibull_scale(log_y,
                              mean(log_y),
                              status[order_indices],
                              weights)

  ## First, use outer-loop to optimize scale (sigma)
  #  Then given scale, optimize beta.
  #  Note: the score and information inside damped_newton_r are both
  #  scaled by sigma^2 (omitting 1/sigma and 1/sigma^2 prefactors),
  #  so the Newton-Raphson step remains correct.
  scale <- optim(init_scale,
                 fn = function(par){
                   ## par is scale (sigma)
                   scale <- par
                   beta <- cbind(damped_newton_r(
                     c(mean(log_y), rep(0, ncol(X)-1)),
                     function(par){
                       beta <- cbind(par)
                       log_mu <- c(X %**% beta)
                       loglik_weibull(log_y,
                                      log_mu,
                                      status[order_indices],
                                      scale,
                                      weights) +
                         -0.5*c(crossprod(beta, Lambda) %**% beta)
                     },
                     function(par){
                       beta <- cbind(par)
                       eta <- c(X %**% beta)
                       z <- (log_y - eta) / scale
                       zeta <- exp(z)
                       ## Scaled score (x sigma^2): true score is
                       #  (1/sigma) * X^T diag(w)(exp(z) - status),
                       #  multiplied through by sigma^2 gives
                       #  sigma * X^T diag(w)(exp(z) - status)
                       grad_beta_sc <- crossprod(
                         X,
                         (weights*(zeta - status[order_indices])) * scale
                       ) +
                         Lambda %**% beta
                       cbind(grad_beta_sc)
                     },
                     function(par){
                       beta <- cbind(par)
                       eta <- X %**% beta
                       z <- (log_y - eta) / scale
                       zeta <- c(exp(z))
                       ## Scaled information (x sigma^2): true info is
                       #  (1/sigma^2) * X^T diag(w*exp(z)) X,
                       #  multiplied through by sigma^2 gives
                       #  X^T diag(w*exp(z)) X
                       info_sc <- crossprod(X, weights * zeta * X) + Lambda
                       info_sc
                     },
                     tol
                   ))
                   ## Return negated penalized log-likelihood
                   #  as the objective to minimize
                   log_mu <- X %**% beta
                   -loglik_weibull(log_y,
                                   log_mu,
                                   status[order_indices],
                                   par,
                                   weights) +
                     0.5*c(crossprod(beta, Lambda) %**% beta)
                 },
                 lower = init_scale/10, # Increased to 10
                 upper = init_scale*10, # (was 5)
                 method = 'Brent')$par

  ## Now optimize beta, given optimal scale (sigma)
  beta <- cbind(damped_newton_r(
    c(mean(log_y), rep(0, ncol(X)-1)),
    function(par){
      beta <- cbind(par)
      log_mu <- c(X %**% beta)
      loglik_weibull(log_y,
                     log_mu,
                     status[order_indices],
                     scale,
                     weights) +
        -0.5*c(crossprod(beta, Lambda) %**% beta)
    },
    function(par){
      beta <- cbind(par)
      eta <- c(X %**% beta)
      z <- (log_y - eta) / scale
      zeta <- exp(z)
      ## Scaled score (x sigma^2); see note above
      grad_beta_sc <- crossprod(
        X,
        (weights*(zeta - status[order_indices])) * scale
      ) +
        Lambda %**% beta
      cbind(grad_beta_sc)
    },
    function(par){
      beta <- cbind(par)
      eta <- X %**% beta
      z <- (log_y - eta) / scale
      zeta <- c(exp(z))
      ## Scaled information (x sigma^2); see note above
      info_sc <- crossprod(X, weights * zeta * X) + Lambda
      info_sc
    },
    tol
  ))

  return(beta)
}


#' Fit Weibull Accelerated Failure Time Model via lgspline
#'
#' @description
#' Convenience wrapper that calls \code{lgspline} with the correct
#' family, weight, dispersion, score, and unconstrained-fit functions for
#' Weibull accelerated failure time regression. All standard lgspline
#' arguments (knots, penalties, constraints, parallel, etc.) are passed
#' through.
#'
#' @param formula Formula specifying the model. The response should be
#'   survival time; \code{status} is passed separately.
#' @param data Data frame.
#' @param status Integer vector of event indicators (1 = event,
#'   0 = censored), same length as the number of rows in \code{data}.
#' @param ... Additional arguments passed to \code{lgspline}.
#'
#' @return An object of class \code{"lgspline"}.
#'
#' @details
#' Internally sets:
#' \itemize{
#'   \item \code{family = weibull_family()}
#'   \item \code{unconstrained_fit_fxn = unconstrained_fit_weibull}
#'   \item \code{glm_weight_function = weibull_glm_weight_function}
#'   \item \code{qp_score_function = weibull_qp_score_function}
#'   \item \code{dispersion_function = weibull_dispersion_function}
#'   \item \code{schur_correction_function = weibull_schur_correction}
#'   \item \code{need_dispersion_for_estimation = TRUE}
#'   \item \code{estimate_dispersion = TRUE}
#'   \item \code{standardize_response = FALSE}
#' }
#'
#' A formula interface is needed, e.g. \code{lgspline_weibull(t, y, ...)} won't work,
#' unlike for ordinary \code{\link{lgspline}}.
#'
#' @examples
#'
#' ## Weibull AFT with a nonlinear age effect on lung cancer survival
#' if(requireNamespace("survival", quietly = TRUE)) {
#'   library(survival)
#'   set.seed(1234)
#'   lung <- na.omit(lung[, c("time", "status", "age")])
#'   lung$age_std <- std(lung$age)
#'
#'   ## survival codes status as 1 = censored, 2 = dead
#'   event <- as.integer(lung$status == 2)
#'
#'   ## Spline on age
#'   fit <- lgspline_weibull(
#'     time ~ spl(age_std),
#'     data = lung,
#'     status = event,
#'     K = 1,
#'     opt = FALSE,
#'     wiggle_penalty = 1e-4,
#'     flat_ridge_penalty = 1
#'   )
#'   print(summary(fit))
#'   plot(fit,
#'        show_formulas = TRUE,
#'        custom_response_lab = 'Survival Time',
#'        custom_predictor_lab = 'Standardized Age')
#' }
#'
#' @seealso \code{\link{lgspline_cox}} for Cox proportional hazards,
#'   \code{\link{weibull_family}}, \code{\link{unconstrained_fit_weibull}}
#'
#' @export
lgspline_weibull <- function(formula, data, status, ...) {
  lgspline(
    formula = formula,
    data = data,
    family = weibull_family(),
    unconstrained_fit_fxn = unconstrained_fit_weibull,
    glm_weight_function = weibull_glm_weight_function,
    qp_score_function = weibull_qp_score_function,
    dispersion_function = weibull_dispersion_function,
    schur_correction_function = weibull_schur_correction,
    need_dispersion_for_estimation = TRUE,
    estimate_dispersion = TRUE,
    standardize_response = FALSE,
    status = status,
    ...
  )
}

test_that("Weibull accelerated failure time (AFT) models can be fit under custom constraints", {

  ## Setup
  set.seed(1234)
  x1 <- rnorm(1000)
  x2 <- rbinom(1000, 1, 0.5)
  yraw <- rexp(exp(0.01*x1 + 0.01*x2))
  status <- rbinom(1000, 1, 0.25)
  yobs <- ifelse(status, runif(1, 0, yraw), yraw)
  df <- data.frame(
    y = yobs,
    x1 = x1,
    x2 = x2
  )

  ## Weibull AFT
  model_fit_aft <- lgspline(
    y ~ spl(x1) + x2,
    df,
    unconstrained_fit_fxn = unconstrained_fit_weibull,
    family = weibull_family(),
    need_dispersion_for_estimation = TRUE,
    qp_score_function = weibull_qp_score_function,
    dispersion_function = weibull_dispersion_function,
    glm_weight_function = weibull_glm_weight_function,
    schur_correction_function = weibull_schur_correction,
    K = 1,
    include_constrain_fitted = FALSE,
    flat_ridge_penalty = 1e-2,
    wiggle_penalty = 1e-2,
    unique_penalty_per_predictor = FALSE,
    unique_penalty_per_partition = FALSE,
    opt = FALSE,
    return_varcovmat = TRUE,
    observation_weights = 0.5 + 0.5 * abs(rnorm(1000)),
    constraint_vectors = cbind(rep(1, 12)),
    null_constraint = matrix(12),
    status = status,
    verbose = TRUE,
    verbose_tune = TRUE
  )
  print(model_fit_aft)

  ## Check sum-to-P constraint only if model_fit_aft is a list
  if (is.list(model_fit_aft)) {
    expect_equal(round(sum(unlist(model_fit_aft$B)), 8), 12)
  } else {
  ## Skip the test on platforms when model returns a try-error instead
    expect_error(invisible(), NA)
  }
})

## get_B.R
## Constrained coefficient estimation for the main lgspline fitting paths.
##
## Main pieces
#   get_B()
#   the three fitting paths
#   Lagrangian projection helpers
#   QP refinement helpers
#   Woodbury helpers for the GEE paths


## Small wrappers around shared helpers

#' Build Block-Diagonal Penalty Matrix
#'
#' Thin wrapper around \code{.solver_build_lambda_block()}.
#' Keeping the local helper name preserves existing call sites in
#' \code{get_B()} while centralizing the shared implementation in
#' \code{solver_utils.R}.
#'
#' @param Lambda Shared \eqn{p \times p} penalty matrix.
#' @param K Integer; number of interior knots (\eqn{K+1} partitions).
#' @param unique_penalty_per_partition Logical; if \code{TRUE}, add
#'   \code{L_partition_list[[k]]} to the \eqn{k}-th diagonal block.
#' @param L_partition_list List of partition-specific \eqn{p \times p}
#'   penalty matrices.
#'
#' @return A \eqn{P \times P} block-diagonal matrix where
#'   \eqn{P = p \times (K+1)}.
#'
#' @keywords internal
.build_lambda_block <- function(Lambda, K, unique_penalty_per_partition,
                                L_partition_list) {
  .solver_build_lambda_block(
    Lambda = Lambda,
    K = K,
    unique_penalty_per_partition = unique_penalty_per_partition,
    L_partition_list = L_partition_list
  )
}



#' Extract Per-Partition Diagonal Blocks from a Full Matrix
#'
#' Given a \eqn{P \times P} matrix (e.g., the full-system
#' \eqn{\mathbf{G} = (\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X} +
#' \boldsymbol{\Lambda})^{-1}}), extracts the \eqn{p \times p} diagonal
#' block for each of the \eqn{K+1} partitions.
#'
#' @param M A \eqn{P \times P} matrix where \eqn{P = p\_expansions \times (K+1)}.
#' @param p_expansions Integer; number of basis terms (columns) per partition.
#' @param K Integer; number of interior knots.
#'
#' @return A list of length \eqn{K+1}, each element a
#'   \eqn{p\_expansions \times p\_expansions} matrix corresponding to the
#'   diagonal block for partition \eqn{k}.
#'
#' @keywords internal
.extract_G_diagonal <- function(M, p_expansions, K) {
  lapply(1:(K + 1), function(k) {
    M[(k - 1) * p_expansions + 1:p_expansions,
      (k - 1) * p_expansions + 1:p_expansions]
  })
}


## Lagrangian projection helpers

#' Lagrangian Projection via OLS Reformulation
#'
#' Given unconstrained (or GEE-initialized) coefficient estimates
#' \eqn{\hat{\boldsymbol{\beta}}}, computes the constrained estimate
#' \eqn{\tilde{\boldsymbol{\beta}} = \mathbf{U}\hat{\boldsymbol{\beta}}}
#' where \eqn{\mathbf{U} = \mathbf{I} - \mathbf{G}\mathbf{A}
#' (\mathbf{A}^{\top}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{\top}}.
#'
#' Rather than forming \eqn{\mathbf{U}} directly, the projection is
#' reformulated as a residual from an OLS problem (the
#' \eqn{\mathbf{G}^{1/2}\mathbf{r}^*} trick):
#' \deqn{\mathbf{y}^* = \mathbf{G}^{-1/2}\hat{\boldsymbol{\beta}}}
#' \deqn{\mathbf{X}^* = \mathbf{G}^{1/2}\mathbf{A}}
#' \deqn{\mathbf{r}^* = (\mathbf{I} - \mathbf{X}^*(\mathbf{X}^{*\top}
#'   \mathbf{X}^*)^{-1}\mathbf{X}^{*\top})\mathbf{y}^*}
#' \deqn{\tilde{\boldsymbol{\beta}} = \mathbf{G}^{1/2}\mathbf{r}^*}
#'
#' This avoids explicitly forming and inverting
#' \eqn{\mathbf{A}^{\top}\mathbf{G}\mathbf{A}}; the most expensive step is
#' the QR decomposition of the \eqn{R \times R} system inside
#' \code{.lm.fit}, which is far cheaper than the full \eqn{P \times P}
#' solve.
#'
#' @param GhalfXy Numeric column vector \eqn{\mathbf{y}^*} of length
#'   \eqn{P}; produced by multiplying \eqn{\mathbf{G}^{1/2}} (Path 2) or
#'   \eqn{\mathbf{G}^{-1/2}} (Path 3) into the unconstrained estimate.
#' @param Ghalf List of \eqn{\mathbf{G}^{1/2}_k} matrices (one per
#'   partition).
#' @param A Constraint matrix \eqn{\mathbf{A}} (\eqn{P \times R}).
#' @param K Integer; number of interior knots.
#' @param p_expansions Integer; number of basis terms per partition.
#' @param R_constraints Integer; number of columns of \eqn{\mathbf{A}}.
#' @param constraint_value_vectors List of constraint right-hand-side
#'   vectors encoding \eqn{\mathbf{A}^{\top}\boldsymbol{\beta} =
#'   \mathbf{c}}. When non-empty and nonzero, the particular solution is
#'   added.
#' @param family GLM family object (used for \code{linkinv(0)} fallback
#'   when \code{NA} values arise in \eqn{\mathbf{y}^*}).
#' @param parallel_aga,parallel_matmult Logical flags for parallel
#'   computation.
#' @param cl Parallel cluster object.
#' @param chunk_size,num_chunks,rem_chunks Parallel distribution parameters.
#'
#' @return A list of length \eqn{K+1}, each element a column vector of
#'   constrained coefficients \eqn{\tilde{\boldsymbol{\beta}}_k}.
#'
#' @keywords internal
.lagrangian_project <- function(GhalfXy, Ghalf, A, K, p_expansions,
                                R_constraints, constraint_value_vectors,
                                family, parallel_aga, parallel_matmult,
                                cl, chunk_size, num_chunks, rem_chunks) {

  ## Compute X* = G^{1/2} A by stacking partition blocks vertically.
  #  GAmult_wrapper handles the block-diagonal multiplication in parallel
  #  when parallel_aga = TRUE.
  GhalfA <- Reduce("rbind",
                   GAmult_wrapper(Ghalf,
                                  A,
                                  K,
                                  p_expansions,
                                  R_constraints,
                                  parallel_aga,
                                  cl,
                                  chunk_size,
                                  num_chunks,
                                  rem_chunks))

  ## Replace any NA values in y* with the link-transformed zero.
  #  NAs can arise from empty partitions (n_k = 0).
  GhalfXy <- ifelse(is.na(GhalfXy), family$linkinv(0), GhalfXy)

  ## OLS residuals: project y* onto the orthogonal complement of col(X*).
  #  Scaling by 1/sqrt(K+1) improves numerical stability of .lm.fit
  #  when the constraint matrix has many rows.
  comp_stab_sc <- 1 / sqrt(K + 1)
  resids_star <- do.call('.lm.fit', list(
    x = GhalfA * comp_stab_sc,
    y = GhalfXy * comp_stab_sc
  ))$residuals / comp_stab_sc

  ## If constraint values are nonzero (A^T beta = j with j != 0),
  #  add the particular solution (I - U) b_0 where A^T b_0 = j.
  #  The full constrained solution is: U beta_hat + (I - U) b_0.
  if (length(constraint_value_vectors) > 0) {
    if (any(unlist(constraint_value_vectors) != 0)) {
      comp_stab_sc <- 1 / sqrt(K + 1)
      preds_star <- GhalfA %**%
        (invert(crossprod(GhalfA) * comp_stab_sc) %**%
           crossprod(A,
                     Reduce("rbind", constraint_value_vectors) *
                       comp_stab_sc))
      resids_star <- resids_star + c(preds_star)
    }
  }

  ## Back-transform: beta_constrained_k = G^{1/2}_k r*_k (per partition).
  if (parallel_matmult & !is.null(cl)) {
    ## Handle remainder partitions that don't fill a full chunk.
    if (rem_chunks > 0) {
      rem_indices <- num_chunks * chunk_size + 1:rem_chunks
      rem <- lapply(rem_indices, function(k) {
        Ghalf[[k]] %**% cbind(resids_star[(k - 1) * p_expansions +
                                            1:p_expansions])
      })
    } else {
      rem <- list()
    }
    ## Parallel back-transformation across partition chunks.
    result <- c(
      Reduce("c", parallel::parLapply(cl, 1:num_chunks, function(chunk) {
        start_idx <- (chunk - 1) * chunk_size + 1
        end_idx <- chunk * chunk_size
        lapply(start_idx:end_idx, function(k) {
          Ghalf[[k]] %**% cbind(resids_star[(k - 1) * p_expansions +
                                              1:p_expansions])
        })
      })),
      rem
    )
  } else {
    result <- lapply(1:(K + 1), function(k) {
      Ghalf[[k]] %**% cbind(resids_star[(k - 1) * p_expansions +
                                          1:p_expansions])
    })
  }

  result
}


## Sub-function: .lagrangian_project_woodbury

#' Woodbury-Corrected Lagrangian Projection
#'
#' Applies the constrained projection for the structured-correlation
#' Woodbury path. In supplement notation,
#' \eqn{\mathbf{G}^{1/2} = \mathbf{G}_{\mathrm{on}}^{1/2}\mathbf{F}^{1/2}},
#' where \eqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{E}\mathbf{J}\mathbf{E}^{\top}}
#' and \eqn{\mathbf{F}^{1/2}} carries the low-rank correction.
#'
#' The routine follows the same four transformed-OLS steps as
#' \code{.lagrangian_project}: form \eqn{\mathbf{y}^{*}},
#' form \eqn{\mathbf{X}^{*}}, solve the OLS projection, then back-transform.
#' Only steps involving \eqn{\mathbf{G}^{1/2}} differ, and those are handled
#' as a block-diagonal multiplication through
#' \eqn{\mathbf{G}_{\mathrm{on}}^{1/2}} plus a rank-\eqn{r} update.
#'
#' Internally, \eqn{\mathbf{F}^{1/2}} is stored as
#' \eqn{\mathbf{I} - \mathbf{U}_Q\mathbf{C}\mathbf{U}_Q^{\top}}, where
#' \code{U_Q} spans the same subspace as the supplement matrix
#' \eqn{\mathbf{Q}} and \code{C} is diagonal.
#'
#' @param GhalfXy_V Numeric column vector of length \eqn{P};
#'   \eqn{\mathbf{G}_{\mathrm{on}}^{1/2}} applied to the per-partition GLS
#'   cross-products \eqn{\mathbf{X}_k^{\top}\mathbf{V}^{-1}\mathbf{y}_k}.
#' @param Ghalf_corrected List of \eqn{K+1} corrected
#'   \eqn{\mathbf{G}_{\mathrm{on},k}^{1/2}} matrices.
#' @param A Constraint matrix (\eqn{P \times R}).
#' @param K,p_expansions,R_constraints Integer dimensions.
#' @param constraint_value_vectors Constraint RHS list encoding
#'   \eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{c}}.
#' @param family GLM family object.
#' @param wb_sqrt Output of \code{.woodbury_halfsqrt_components},
#'   containing the internal rank-\eqn{r} representation of
#'   \eqn{\mathbf{F}^{1/2}}: \code{U_Q} / \code{Q_basis}
#'   (basis matching supplement \eqn{\mathbf{Q}}), \code{C} and \code{C_inv} (diagonal correction
#'   coefficients), and \code{GchalfUQ} (partition-wise copies of
#'   \eqn{\mathbf{G}_{\mathrm{on},k}^{1/2}\mathbf{U}_Q[rows_k,]}).
#' @param parallel_aga,parallel_matmult Logical flags.
#' @param cl,chunk_size,num_chunks,rem_chunks Parallel parameters.
#'
#' @return A list of \eqn{K+1} coefficient column vectors.
#'
#' @keywords internal
.lagrangian_project_woodbury <- function(GhalfXy_V, Ghalf_corrected,
                                         A, K, p_expansions,
                                         R_constraints,
                                         constraint_value_vectors,
                                         family,
                                         wb_sqrt,
                                         parallel_aga, parallel_matmult,
                                         cl, chunk_size, num_chunks,
                                         rem_chunks) {

  U_Q <- wb_sqrt$U_Q
  C <- wb_sqrt$C
  C_inv <- wb_sqrt$C_inv
  GchalfUQ <- wb_sqrt$GchalfUQ
  r <- ncol(U_Q)

  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  ## Step 1: Form y* = F^{1/2} GhalfXy_V.
  ## In supplement notation F^{1/2} = I - P + Q S^{-1/2} Q^T.
  ## Internally the same operator is stored as I - U_Q C U_Q^T.
  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  v_y <- crossprod(U_Q, GhalfXy_V)         # r x 1
  y_star <- GhalfXy_V - U_Q %**% (C %**% v_y)

  ## Replace NAs with link-transformed zero (same as .lagrangian_project).
  y_star <- ifelse(is.na(y_star), family$linkinv(0), y_star)

  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  ## Step 2: Form X* = G^{1/2} A = G_on^{1/2} F^{1/2} A.
  ## This is the Woodbury-corrected analogue of G^{1/2} A in the
  ## uncorrelated projection.
  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

  ## Block-diagonal part: G_on^{1/2} A.
  GhalfA <- Reduce("rbind",
                   GAmult_wrapper(Ghalf_corrected, A, K, p_expansions,
                                  R_constraints, parallel_aga, cl,
                                  chunk_size, num_chunks, rem_chunks))

  ## Rank-r correction restricted to the Woodbury subspace.
  UQtA <- crossprod(U_Q, A)                  # r x R
  C_UQtA <- C %**% UQtA                      # r x R
  GchalfUQ_full <- do.call(rbind, GchalfUQ)   # P x r
  correction_A <- GchalfUQ_full %**% C_UQtA   # P x R

  X_star <- GhalfA - correction_A

  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  ## Step 3: OLS residuals -- IDENTICAL to .lagrangian_project.
  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  comp_stab_sc <- 1 / sqrt(K + 1)
  resids_star <- do.call('.lm.fit', list(
    x = X_star * comp_stab_sc,
    y = y_star * comp_stab_sc
  ))$residuals / comp_stab_sc

  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  ## Nonzero constraint values (A^T beta = j with j != 0).
  ## Same logic as .lagrangian_project.
  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  if (length(constraint_value_vectors) > 0) {
    if (any(unlist(constraint_value_vectors) != 0)) {
      comp_stab_sc <- 1 / sqrt(K + 1)
      preds_star <- X_star %**%
        (invert(crossprod(X_star) * comp_stab_sc) %**%
           crossprod(A,
                     Reduce("rbind", constraint_value_vectors) *
                       comp_stab_sc))
      resids_star <- resids_star + c(preds_star)
    }
  }

  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  ## Step 4: Back-transform beta_tilde = G^{1/2} r* with the same
  ## block-plus-rank-r operator used in Steps 1 and 2.
  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  v_r <- crossprod(U_Q, cbind(resids_star))  # r x 1
  C_v_r <- C %**% v_r                        # r x 1

  result <- lapply(1:(K + 1), function(k) {
    rows_k <- ((k - 1) * p_expansions + 1):(k * p_expansions)
    ## Block-diagonal part.
    block_part <- Ghalf_corrected[[k]] %**%
      cbind(resids_star[rows_k])
    ## Rank-r correction.
    correction_part <- GchalfUQ[[k]] %**% C_v_r
    block_part - correction_part
  })

  result
}


## Shared G recomputation wrapper

#' Recompute G at Current Coefficient Estimates
#'
#' Thin wrapper around \code{.solver_recompute_G_at_estimate()}.
#' The numerical core is shared with \code{blockfit_solve()} so the
#' final \code{G}, \code{Ghalf}, and \code{GhalfInv} objects are
#' constructed through one code path.
#'
#' @inheritParams .solver_recompute_G_at_estimate
#'
#' @return A list with components \code{G}, \code{Ghalf}, and
#'   \code{GhalfInv}, each a list of \eqn{K+1} matrices.
#'
#' @keywords internal
.recompute_G_at_estimate <- function(X, y, result, K, Lambda, family,
                                     order_list, glm_weight_function,
                                     schur_correction_function,
                                     need_dispersion_for_estimation,
                                     dispersion_function,
                                     observation_weights, VhalfInv,
                                     parallel_eigen, parallel_matmult,
                                     cl, chunk_size, num_chunks, rem_chunks,
                                     unique_penalty_per_partition,
                                     L_partition_list, ...) {
  .solver_recompute_G_at_estimate(
    X = X,
    y = y,
    result = result,
    K = K,
    Lambda = Lambda,
    family = family,
    order_list = order_list,
    glm_weight_function = glm_weight_function,
    schur_correction_function = schur_correction_function,
    need_dispersion_for_estimation = need_dispersion_for_estimation,
    dispersion_function = dispersion_function,
    observation_weights = observation_weights,
    VhalfInv = VhalfInv,
    parallel_eigen = parallel_eigen,
    parallel_matmult = parallel_matmult,
    cl = cl,
    chunk_size = chunk_size,
    num_chunks = num_chunks,
    rem_chunks = rem_chunks,
    unique_penalty_per_partition = unique_penalty_per_partition,
    L_partition_list = L_partition_list,
    ...
  )
}


## QP helpers

#' Detect Whether Inequality Constraints Require Global (Dense) QP
#'
#' Thin wrapper around \code{.solver_detect_qp_global()}.
#' This preserves the original helper name inside \code{get_B()} while
#' using the shared detection logic defined in \code{solver_utils.R}.
#'
#' @param qp_Amat Inequality constraint matrix (\eqn{P \times n_{ineq}}),
#'   or \code{NULL}.
#' @param p_expansions Integer; number of basis terms per partition.
#' @param K Integer; number of interior knots.
#'
#' @return Logical; \code{TRUE} if dense QP is needed, \code{FALSE} if
#'   partition-wise active-set is valid.
#'
#' @keywords internal
.detect_qp_global <- function(qp_Amat, p_expansions, K) {
  .solver_detect_qp_global(
    qp_Amat = qp_Amat,
    p_expansions = p_expansions,
    K = K
  )
}


## Sub-function: .check_kkt_partitionwise

#' Check KKT Conditions for Partition-Wise Active-Set Method
#'
#' Given current constrained coefficient estimates and a set of active
#' inequality constraints (treated as equalities in the last Lagrangian
#' projection), checks primal feasibility of inactive constraints and
#' dual feasibility (non-negative multipliers) of active constraints.
#'
#' Multipliers for active inequality constraints are recovered from the
#' OLS fit used in the Lagrangian projection: the fitted coefficients
#' on \eqn{\mathbf{X}^* = \mathbf{G}^{1/2}\mathbf{A}_{\mathrm{aug}}}
#' give the Lagrangian multipliers (up to sign and scaling).
#'
#' @param result List of \eqn{K+1} coefficient column vectors.
#' @param Ghalf List of \eqn{\mathbf{G}^{1/2}_k} matrices.
#' @param GhalfInv List of \eqn{\mathbf{G}^{-1/2}_k} matrices.
#' @param Xy_or_uncon Either the list of cross-products
#'   \eqn{\mathbf{X}_k^{\top}\mathbf{y}_k} (Path 2) or unconstrained
#'   estimates (Path 3).
#' @param is_path3 Logical; if TRUE, \code{Xy_or_uncon} contains
#'   unconstrained estimates and GhalfInv is used for transformation.
#' @param A_aug Augmented constraint matrix (original A plus active
#'   inequality columns).
#' @param n_eq_orig Integer; number of original equality constraints
#'   (columns of A before augmentation).
#' @param qp_Amat Full inequality constraint matrix.
#' @param qp_bvec Full inequality constraint RHS.
#' @param active_ineq Integer vector; indices into columns of
#'   \code{qp_Amat} that are currently active.
#' @param K,p_expansions Integer dimensions.
#' @param family GLM family object.
#' @param parallel_matmult,parallel_aga Logical flags.
#' @param cl,chunk_size,num_chunks,rem_chunks Parallel parameters.
#' @param tol Numeric tolerance for feasibility and multiplier checks.
#'
#' @return A list with components:
#' \describe{
#'   \item{feasible}{Logical; TRUE if all inactive inequality constraints
#'     are satisfied within tolerance.}
#'   \item{dual_feasible}{Logical; TRUE if all active inequality
#'     multipliers are non-negative within tolerance.}
#'   \item{violated}{Integer vector; indices of violated inactive
#'     constraints (into \code{qp_Amat} columns).}
#'   \item{drop}{Integer vector; indices of active constraints with
#'     negative multipliers that should be dropped.}
#'   \item{multipliers}{Numeric vector of Lagrangian multipliers for
#'     the active inequality constraints.}
#' }
#'
#' @keywords internal
.check_kkt_partitionwise <- function(result, Ghalf, GhalfInv,
                                     Xy_or_uncon, is_path3,
                                     A_aug, n_eq_orig,
                                     qp_Amat, qp_bvec,
                                     active_ineq,
                                     K, p_expansions, family,
                                     parallel_matmult, parallel_aga,
                                     cl, chunk_size, num_chunks,
                                     rem_chunks, tol) {

  beta_full <- cbind(unlist(result))

  ## Check primal feasibility of ALL inequality constraints.
  #  Constraint form: qp_Amat^T beta >= qp_bvec.
  slack <- crossprod(qp_Amat, beta_full) - cbind(qp_bvec)
  violated_mask <- c(slack) < -tol
  all_ineq_idx <- 1:ncol(qp_Amat)
  inactive_idx <- setdiff(all_ineq_idx, active_ineq)
  violated <- inactive_idx[violated_mask[inactive_idx]]

  ## Compute Lagrangian multipliers for active inequality constraints.
  #  The multipliers come from the dual of the OLS problem in the
  #  Lagrangian projection.  Specifically, if y* = G^{-1/2} beta_hat
  #  and X* = G^{1/2} A_aug, then the OLS coefficients on X* give
  #  the multipliers (for the augmented system).
  drop <- integer(0)
  multipliers <- numeric(0)

  if (length(active_ineq) > 0) {
    ## Reconstruct y* from unconstrained or cross-product info.
    if (is_path3) {
      GhalfXy <- cbind(unlist(
        matmult_block_diagonal(GhalfInv, Xy_or_uncon, K,
                               parallel_matmult, cl,
                               chunk_size, num_chunks, rem_chunks)))
    } else {
      GhalfXy <- cbind(unlist(
        matmult_block_diagonal(Ghalf, Xy_or_uncon, K,
                               parallel_matmult, cl,
                               chunk_size, num_chunks, rem_chunks)))
    }
    GhalfXy <- ifelse(is.na(GhalfXy), family$linkinv(0), GhalfXy)

    ## X* = G^{1/2} A_aug (block-diagonal multiplication).
    GhalfA_aug <- Reduce("rbind",
                         GAmult_wrapper(Ghalf, A_aug, K, p_expansions,
                                        ncol(A_aug),
                                        parallel_aga, cl,
                                        chunk_size, num_chunks,
                                        rem_chunks))

    ## OLS fit to recover multipliers.
    comp_stab_sc <- 1 / sqrt(K + 1)
    ols_fit <- do.call('.lm.fit', list(
      x = GhalfA_aug * comp_stab_sc,
      y = GhalfXy * comp_stab_sc
    ))
    ## Coefficients correspond to multipliers for each column of A_aug.
    ## The first n_eq_orig are equality constraints; the rest are active
    ## inequality constraints.
    all_multipliers <- ols_fit$coefficients
    if (length(all_multipliers) > n_eq_orig) {
      ineq_multipliers <- all_multipliers[(n_eq_orig + 1):length(all_multipliers)]
      ## For constraints of the form A^T beta >= b, dual feasibility
      ## requires multipliers >= 0.  The sign convention from the OLS
      ## residual formulation means negative coefficients indicate
      ## the constraint is pulling the solution in the wrong direction.
      drop_mask <- ineq_multipliers < -tol
      drop <- which(drop_mask)
      multipliers <- ineq_multipliers
    }
  }

  list(feasible = length(violated) == 0,
       dual_feasible = length(drop) == 0,
       violated = violated,
       drop = drop,
       multipliers = multipliers)
}


## Sub-function: .active_set_refine

#' Partition-Wise Active-Set Refinement for Inequality Constraints
#'
#' Replaces the dense \code{.qp_refine} SQP loop with a partition-wise
#' active-set method that reuses the existing Lagrangian projection
#' machinery. The key insight is that an active inequality constraint
#' can be treated as an additional equality constraint and absorbed into
#' the augmented constraint matrix \eqn{\mathbf{A}_{\mathrm{aug}}},
#' after which the standard \eqn{\mathbf{G}^{1/2}\mathbf{r}^*} trick
#' applies without ever forming the full \eqn{P \times P} information
#' matrix.
#'
#' The active-set loop:
#' \enumerate{
#'   \item Solve the equality-constrained subproblem (Lagrangian projection)
#'     with current active set.
#'   \item Check primal feasibility: any violated inactive inequalities?
#'   \item Check dual feasibility: any negative multipliers on active
#'     inequalities?
#'   \item If both satisfied, return (KKT conditions met).
#'   \item Otherwise, add most-violated constraint or drop most-negative
#'     multiplier, and repeat.
#' }
#'
#' Falls back to \code{.qp_refine} if the active-set method does not
#' converge within \code{max_as_iter} iterations.
#'
#' @param result List of current coefficient column vectors by partition.
#' @param X,y Lists of partition-specific design matrices and responses.
#' @param K,p_expansions Integer dimensions.
#' @param A Original equality constraint matrix.
#' @param R_constraints Number of columns of \code{A}.
#' @param constraint_value_vectors Constraint RHS list.
#' @param Lambda Shared penalty matrix.
#' @param Ghalf,GhalfInv Lists of \eqn{\mathbf{G}^{1/2}_k} and
#'   \eqn{\mathbf{G}^{-1/2}_k} matrices.
#' @param family GLM family object.
#' @param qp_Amat Inequality constraint matrix.
#' @param qp_bvec Inequality constraint RHS.
#' @param qp_meq Number of equality constraints within \code{qp_Amat}
#'   (leading columns).
#' @param Xy_or_uncon Cross-products (Path 2) or unconstrained estimates
#'   (Path 3).
#' @param is_path3 Logical.
#' @param parallel_aga,parallel_matmult Logical flags.
#' @param cl,chunk_size,num_chunks,rem_chunks Parallel parameters.
#' @param tol Convergence tolerance.
#' @param max_as_iter Maximum active-set iterations (default 50).
#'
#' @return A list with components:
#' \describe{
#'   \item{result}{List of refined coefficient column vectors by partition.}
#'   \item{qp_info}{List with active constraint information, or NULL.}
#'   \item{converged}{Logical; TRUE if active-set method converged.}
#' }
#'
#' @keywords internal
.active_set_refine <- function(result, X, y, K, p_expansions,
                               A, R_constraints,
                               constraint_value_vectors,
                               Lambda, Ghalf, GhalfInv,
                               family,
                               qp_Amat, qp_bvec, qp_meq,
                               Xy_or_uncon, is_path3,
                               parallel_aga, parallel_matmult,
                               cl, chunk_size, num_chunks,
                               rem_chunks, tol,
                               max_as_iter = 50) {

  n_eq_orig <- ncol(A)
  n_ineq <- ncol(qp_Amat)

  ## Separate the leading qp_meq equality constraints from the true
  #  inequality constraints.  The qp_meq columns of qp_Amat are
  #  equalities that must always be active.
  if (!is.null(qp_meq) && qp_meq > 0) {
    permanent_eq_cols <- 1:qp_meq
    ineq_cols <- if (qp_meq < n_ineq) (qp_meq + 1):n_ineq else integer(0)
  } else {
    permanent_eq_cols <- integer(0)
    ineq_cols <- 1:n_ineq
  }

  ## Build the always-active augmented A: original A + permanent equalities.
  if (length(permanent_eq_cols) > 0) {
    A_base <- cbind(A, qp_Amat[, permanent_eq_cols, drop = FALSE])
  } else {
    A_base <- A
  }
  n_eq_base <- ncol(A_base)

  ## Initialize active set from primal feasibility check.
  beta_full <- cbind(unlist(result))
  if (length(ineq_cols) > 0) {
    slack <- crossprod(qp_Amat[, ineq_cols, drop = FALSE], beta_full) -
      cbind(qp_bvec[ineq_cols])
    active_ineq <- ineq_cols[which(c(slack) < tol)]
  } else {
    active_ineq <- integer(0)
  }

  converged <- FALSE

  for (as_iter in 1:max_as_iter) {

    ## Build augmented constraint matrix.
    if (length(active_ineq) > 0) {
      A_aug <- cbind(A_base, qp_Amat[, active_ineq, drop = FALSE])
    } else {
      A_aug <- A_base
    }

    ## Reduce to linearly independent columns via pivoted QR.
    qr_aug <- qr(A_aug)
    if (qr_aug$rank < ncol(A_aug)) {
      A_aug <- qr.Q(qr_aug)[, 1:qr_aug$rank, drop = FALSE]
    }
    n_aug <- ncol(A_aug)

    ## Build augmented constraint value vectors.
    #  For the original equality constraints, use constraint_value_vectors.
    #  For permanent equalities from qp_Amat, RHS is qp_bvec[permanent_eq_cols].
    #  For active inequalities, RHS is qp_bvec[active_ineq].
    combined_rhs <- rep(0, ncol(cbind(A_base,
                                      if (length(active_ineq) > 0)
                                        qp_Amat[, active_ineq, drop = FALSE]
                                      else NULL)))
    ## Original equality constraint RHS.
    if (length(constraint_value_vectors) > 0) {
      cv_vec <- Reduce("rbind", constraint_value_vectors)
      orig_rhs <- crossprod(A, cv_vec)
      combined_rhs[1:n_eq_orig] <- c(orig_rhs)
    }
    ## Permanent equality constraint RHS from qp_bvec.
    if (length(permanent_eq_cols) > 0) {
      combined_rhs[n_eq_orig + 1:length(permanent_eq_cols)] <-
        qp_bvec[permanent_eq_cols]
    }
    ## Active inequality RHS from qp_bvec.
    if (length(active_ineq) > 0) {
      combined_rhs[n_eq_base + 1:length(active_ineq)] <-
        qp_bvec[active_ineq]
    }

    ## Construct constraint_value_vectors for the augmented system.
    if (any(combined_rhs != 0)) {
      A_orig_full <- cbind(A_base,
                           if (length(active_ineq) > 0)
                             qp_Amat[, active_ineq, drop = FALSE]
                           else NULL)
      if (qr_aug$rank < ncol(A_orig_full)) {
        Q_aug <- qr.Q(qr_aug)[, 1:qr_aug$rank, drop = FALSE]
        rhs_proj <- crossprod(Q_aug, A_orig_full %**% cbind(combined_rhs))
        cv_for_proj <- list(rhs_proj)
      } else {
        cv_for_proj <- list(cbind(combined_rhs))
      }
    } else {
      cv_for_proj <- list()
    }

    ## Lagrangian projection with augmented constraints.
    if (is_path3) {
      GhalfXy <- cbind(unlist(
        matmult_block_diagonal(GhalfInv, Xy_or_uncon, K,
                               parallel_matmult, cl,
                               chunk_size, num_chunks, rem_chunks)))
    } else {
      GhalfXy <- cbind(unlist(
        matmult_block_diagonal(Ghalf, Xy_or_uncon, K,
                               parallel_matmult, cl,
                               chunk_size, num_chunks, rem_chunks)))
    }

    result_new <- .lagrangian_project(
      GhalfXy, Ghalf, A_aug, K, p_expansions,
      n_aug, cv_for_proj,
      family, parallel_aga, parallel_matmult,
      cl, chunk_size, num_chunks, rem_chunks)

    ## Check KKT conditions.
    kkt <- .check_kkt_partitionwise(
      result_new, Ghalf, GhalfInv,
      Xy_or_uncon, is_path3,
      A_aug, n_eq_base,
      qp_Amat[, ineq_cols, drop = FALSE],
      qp_bvec[ineq_cols],
      match(active_ineq, ineq_cols),
      K, p_expansions, family,
      parallel_matmult, parallel_aga,
      cl, chunk_size, num_chunks, rem_chunks,
      tol)

    if (kkt$feasible & kkt$dual_feasible) {
      result <- result_new
      converged <- TRUE
      break
    }

    ## Update active set.
    if (!kkt$feasible) {
      ## Add the most-violated constraint.
      beta_new <- cbind(unlist(result_new))
      slack_viol <- crossprod(
        qp_Amat[, ineq_cols[kkt$violated], drop = FALSE],
        beta_new) - cbind(qp_bvec[ineq_cols[kkt$violated]])
      worst <- kkt$violated[which.min(c(slack_viol))]
      active_ineq <- sort(unique(c(active_ineq, ineq_cols[worst])))
      result <- result_new
    } else if (!kkt$dual_feasible) {
      ## Drop the constraint with the most negative multiplier.
      worst_drop <- kkt$drop[which.min(kkt$multipliers[kkt$drop])]
      active_ineq <- active_ineq[-worst_drop]
      result <- result_new
    }
  }

  ## Build qp_info.
  qp_info <- NULL
  if (converged) {
    if (length(active_ineq) > 0) {
      Amat_active <- cbind(A, qp_Amat[, active_ineq, drop = FALSE])
    } else {
      Amat_active <- A
    }
    qp_info <- list(
      lagrangian = if (length(kkt$multipliers) > 0) kkt$multipliers
      else numeric(0),
      Amat_active = Amat_active,
      active_ineq = active_ineq,
      converged = TRUE
    )
  }

  list(result = result, qp_info = qp_info, converged = converged)
}


## Sub-function: .qp_refine

#' Quadratic Programming Refinement for Inequality Constraints
#'
#' After the Lagrangian projection handles smoothness equality constraints,
#' this function refines the estimate to satisfy additional inequality
#' constraints (monotonicity, derivative sign, range bounds, or user-supplied
#' constraints) via \code{quadprog::solve.QP}.
#'
#' The subproblem at each iteration is a second-order Taylor approximation
#' of the penalized log-likelihood around the current iterate
#' \eqn{\boldsymbol{\beta}^*}. Collecting terms, this yields:
#' \deqn{\tilde{\boldsymbol{\beta}} = \arg\min_{\boldsymbol{\beta}}
#'   \left\{-\mathbf{d}^{\top}\boldsymbol{\beta} + \frac{1}{2}
#'   \boldsymbol{\beta}^{\top}\mathbf{G}^{-1}\boldsymbol{\beta}\right\}
#'   \quad \text{s.t.} \quad
#'   \mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{0}, \quad
#'   \mathbf{C}^{\top}\boldsymbol{\beta} \succeq \mathbf{c}}
#' where \eqn{\mathbf{d}} is the score adjusted by the current iterate
#' and \eqn{\mathbf{c}} is the constraint value vector. Step acceptance
#' uses damped updates with deviance monitoring; see Nocedal and Wright
#' (2006) for the general SQP framework.
#'
#' @param result List of current coefficient column vectors by partition.
#' @param X List of partition-specific design matrices.
#' @param y List of response vectors by partition.
#' @param K Integer; number of interior knots.
#' @param p_expansions Integer; number of basis terms per partition.
#' @param A Equality constraint matrix \eqn{\mathbf{A}}.
#' @param Lambda Shared penalty matrix \eqn{\boldsymbol{\Lambda}}.
#' @param Lambda_block Full block-diagonal penalty matrix.
#' @param family GLM family object.
#' @param iterate Logical; if \code{TRUE}, iterate the SQP loop until
#'   convergence rather than taking a single step.
#' @param tol Convergence tolerance.
#' @param qp_Amat Inequality constraint matrix \eqn{\mathbf{C}} for
#'   \code{solve.QP}.
#' @param qp_bvec Inequality constraint vector \eqn{\mathbf{c}}.
#' @param qp_meq Number of equality constraints within \code{qp_Amat}.
#' @param qp_score_function Score function
#'   \eqn{\nabla_{\boldsymbol{\beta}}\ell(\boldsymbol{\beta}^*)} for the
#'   QP step.
#' @param order_list List of index vectors mapping partition rows to
#'   original data ordering.
#' @param glm_weight_function Function computing GLM working weights.
#' @param schur_correction_function Function computing Schur corrections.
#' @param need_dispersion_for_estimation Logical.
#' @param dispersion_function Dispersion estimation function.
#' @param observation_weights List of observation weights by partition.
#' @param VhalfInv Inverse square root of the working correlation matrix in
#'   the original observation ordering (or \code{NULL}); permuted internally
#'   via \code{order_list}.
#' @param ... Passed to weight, correction, dispersion, and score functions.
#'
#' @return A list with components:
#' \describe{
#'   \item{result}{List of refined coefficient column vectors by partition.}
#'   \item{qp_info}{List with QP solve metadata including Lagrangian
#'     multipliers and the active constraint matrix.}
#' }
#'
#' @keywords internal
.qp_refine <- function(result, X, y, K, p_expansions, A, Lambda,
                       Lambda_block, family, iterate, tol,
                       qp_Amat, qp_bvec, qp_meq,
                       qp_score_function,
                       order_list, glm_weight_function,
                       schur_correction_function,
                       need_dispersion_for_estimation,
                       dispersion_function, observation_weights,
                       VhalfInv, ...) {

  ## Assemble the full N x P block-diagonal design matrix by stacking
  #  partition blocks on the diagonal.
  X_block <- Reduce("rbind", lapply(1:(K + 1), function(k) {
    Reduce("cbind", lapply(1:(K + 1), function(j) {
      if (nrow(X[[k]]) == 0) {
        return(X[[k]])
      } else if (j == k) X[[k]] else 0 * X[[k]]
    }))
  }))
  beta_block <- cbind(unlist(result))
  y_block <- cbind(unlist(y))

  ## Damped SQP loop: iterate Newton-like QP steps with backtracking.
  damp_cnt <- 0
  master_cnt <- 0
  err <- Inf
  XB <- X_block %**% beta_block
  qp_info_out <- NULL

  while (err > tol & damp_cnt < 10 & master_cnt < 100) {
    master_cnt <- master_cnt + 1
    damp <- 2^(-(damp_cnt))

    if (need_dispersion_for_estimation) {
      dispersion_temp <- dispersion_function(
        mu = family$linkinv(XB),
        y = y_block,
        order_indices = unlist(order_list),
        family = family,
        observation_weights = unlist(observation_weights),
        VhalfInv = VhalfInv,
        ...
      )
    } else {
      dispersion_temp <- 1
    }
    W <- c(glm_weight_function(family$linkinv(XB),
                               y_block,
                               unlist(order_list),
                               family,
                               dispersion_temp,
                               unlist(observation_weights),
                               ...))

    ## Schur correction in per-partition form, then collapse to full matrix.
    result <- lapply(1:(K + 1), function(k) {
      cbind(beta_block[(k - 1) * p_expansions + 1:p_expansions])
    })
    schur_correction <-
      schur_correction_function(
        X, y, result, dispersion_temp, order_list, K, family,
        observation_weights, ...
      )
    if (!(any(unlist(schur_correction) != 0))) {
      schur_correction <- 0
    } else {
      schur_correction <- collapse_block_diagonal(schur_correction)
    }

    ## Information matrix: X^T W X + Lambda + Schur correction.
    info <- crossprod(X_block, W * X_block) + Lambda_block + schur_correction

    sc <- sqrt(mean(abs(info)))

    ## QP step: equality (smoothness) and inequality constraints combined.
    #  The dvec encodes the score adjusted for the current iterate so that
    #  the solution is a full position estimate, not a step direction.
    qp_score <- qp_score_function(
      X_block, y_block,
      cbind(family$linkinv(XB)),
      unlist(order_list), dispersion_temp, NULL,
      unlist(observation_weights), ...
    )
    qp_result <- try({
      quadprog::solve.QP(
        Dmat = info / sc,
        dvec = (qp_score -
                  Lambda_block %**% beta_block +
                  info %**% beta_block) / sc,
        Amat = cbind(A, qp_Amat),
        bvec = c(rep(0, ncol(A)), qp_bvec),
        meq = ncol(A) + qp_meq
      )
    }, silent = TRUE)

    if (any(inherits(qp_result, 'try-error'))) {
      beta_new <- 0 * beta_block
    } else {
      beta_new <- qp_result$solution
      ## Store most recent successful solve for multiplier extraction.
      qp_info_out <- list(
        lagrangian = qp_result$Lagrangian,
        Amat_active = cbind(A, qp_Amat)[,
                                        c(1:ncol(A),
                                          ncol(A) +
                                            which(abs(qp_result$Lagrangian[-(1:ncol(A))]) >
                                                    sqrt(.Machine$double.eps))),
                                        drop = FALSE]
      )
    }

    if (!iterate & master_cnt > 1) {
      ## Non-iterative mode: accept the single QP step and exit.
      beta_block <- beta_new
      damp_cnt <- 11
      master_cnt <- 101
      err <- tol - 1
    } else {
      ## Damped update: blend old and new iterates, check deviance.
      beta_new <- (1 - damp) * beta_block + damp * beta_new
      XB <- X_block %**% beta_new
      if (!is.null(family$custom_dev.resids) &
          is.null(family$dev.resids)) {
        err_new <- mean(
          family$custom_dev.resids(y_block,
                                   family$linkinv(c(XB)),
                                   unlist(order_list),
                                   family,
                                   unlist(observation_weights),
                                   ...))
      } else if (is.null(family$dev.resids)) {
        err_new <- mean(unlist(observation_weights) *
                          (y_block - family$linkinv(XB))^2)
      } else {
        err_new <-
          mean(unlist(observation_weights) *
                 family$dev.resids(y_block,
                                   family$linkinv(XB),
                                   wt = 1))
      }

      if (is.null(err_new) | is.na(err_new) | !is.finite(err_new)) {
        ## Non-finite deviance: increase damping and retry.
        damp_cnt <- damp_cnt + 1
      } else if (err_new <= err) {
        ## Deviance decreased: accept step and reset damping.
        prev_err <- err
        err <- err_new
        abs_diff <- max(abs(beta_new - beta_block))
        beta_block <- beta_new
        damp_cnt <- 0

        ## Early exit when both coefficient change and deviance decrease
        #  are below tolerance (after a burn-in of 10 iterations).
        if ((abs_diff < tol) &
            (prev_err - err < tol) &
            (master_cnt > 10)) {
          damp_cnt <- 11
          master_cnt <- 101
          err <- tol - 1
        }

      } else {
        ## Deviance increased: increase damping and retry.
        damp_cnt <- damp_cnt + 1
      }
    }

    ## Unpack full coefficient vector back into per-partition list.
    result <- lapply(1:(K + 1), function(k) {
      cbind(beta_block[1:p_expansions + (k - 1) * p_expansions])
    })
  }

  list(result = result, qp_info = qp_info_out)
}


## Sub-function: .woodbury_decompose_V

#' Woodbury Decomposition of Correlation Structure
#'
#' Builds the Woodbury factors for the correlated solver. The function
#' decomposes
#' \eqn{\mathbf{G}^{-1} = \mathbf{G}_{\mathrm{on}}^{-1} +
#' \mathbf{G}_{\mathrm{off}}^{-1}}
#' by absorbing the block-diagonal part of
#' \eqn{\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X} + \boldsymbol{\Lambda}}
#' into \eqn{\mathbf{G}_{\mathrm{on}}^{-1}}, then factoring the
#' cross-partition part as
#' \eqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{E}\mathbf{J}\mathbf{E}^{\top}}.
#'
#' The output is exactly the information needed by
#' \code{.woodbury_halfsqrt_components()} and
#' \code{.lagrangian_project_woodbury()}:
#' corrected block-diagonal \eqn{\mathbf{G}_{\mathrm{on}}},
#' the low-rank factor \eqn{\mathbf{E}}, the sign matrix diagonal
#' \eqn{\mathbf{J}}, and the small inverse
#' \eqn{(\mathbf{J}^{-1} +
#' \mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}\mathbf{E})^{-1}}.
#'
#' When the cross-partition rank is zero or too large to be worthwhile,
#' the helper returns \code{use_woodbury = FALSE} so the caller can use
#' the dense correlated path instead.
#'
#' @param VhalfInv_perm \eqn{\mathbf{V}^{-1/2}} permuted to partition
#'   ordering.
#' @param X List of partition-specific design matrices.
#' @param K,p_expansions Integer dimensions.
#' @param Lambda,Lambda_block Shared and full block-diagonal penalty
#'   matrices.
#' @param unique_penalty_per_partition Logical.
#' @param L_partition_list Partition-specific penalty matrices.
#' @param order_list Partition-to-data index mapping.
#' @param parallel_eigen Logical; parallel eigendecomposition.
#' @param cl,chunk_size,num_chunks,rem_chunks Parallel parameters.
#' @param rank_threshold_fraction Numeric; Woodbury is used only when
#'   \eqn{r < P \times} this fraction (default 1/3).
#' @param family GLM family object.
#'
#' @return A list with components:
#' \describe{
#'   \item{use_woodbury}{Logical; FALSE triggers dense Path 1 fallback.}
#'   \item{Ghalf_corrected}{List of corrected \eqn{\mathbf{G}_{\mathrm{on},k}^{1/2}}.}
#'   \item{GhalfInv_corrected}{List of corrected \eqn{\mathbf{G}_{\mathrm{on},k}^{-1/2}}.}
#'   \item{G_corrected}{List of corrected \eqn{\mathbf{G}_{\mathrm{on},k}}.}
#'   \item{E}{Off-diagonal low-rank factor (\eqn{P \times r}) from the
#'     supplement factorization
#'     \eqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{E}\mathbf{J}\mathbf{E}^{\top}}.}
#'   \item{J_signs}{Length-\eqn{r} vector giving the diagonal of
#'     \eqn{\mathbf{J}}.}
#'   \item{r}{Integer effective rank.}
#'   \item{inner_inv}{Precomputed \eqn{r \times r} Woodbury inner inverse. In
#'     supplement notation this is
#'     \eqn{(\mathbf{J}^{-1} + \mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}
#'     \mathbf{E})^{-1}}.}
#'   \item{GL}{List of \eqn{K+1} matrices storing
#'     \eqn{\mathbf{G}_{\mathrm{on},k}\mathbf{E}[rows_k,]} in supplement notation,
#'     each \eqn{p \times r}.}
#'   \item{L,S_signs,M}{Legacy internal aliases for
#'     \code{E}, \code{J_signs}, and \code{inner_inv}, retained for
#'     compatibility within the package code.}
#' }
#'
#' @keywords internal
.woodbury_decompose_V <- function(VhalfInv_perm, X, K, p_expansions,
                                  Lambda, Lambda_block,
                                  unique_penalty_per_partition,
                                  L_partition_list,
                                  order_list,
                                  parallel_eigen, cl,
                                  chunk_size, num_chunks, rem_chunks,
                                  rank_threshold_fraction = 1/3,
                                  family) {

  P <- p_expansions * (K + 1)

  ## Step 1a: form V^{-1} and the correlation-induced perturbation
  ## V^{-1} - I in partition ordering.
  Vinv <- tcrossprod(VhalfInv_perm)
  N <- nrow(Vinv)
  Delta_V <- Vinv - diag(N)

  ## Sparsity check: if V^{-1} - I is too dense, Woodbury is not
  #  worthwhile (the off-diagonal rank will be high).
  nnz <- sum(abs(Delta_V) > sqrt(.Machine$double.eps))
  if (nnz / (N * N) > 0.5) {
    return(list(use_woodbury = FALSE))
  }

  ## Step 1b: assemble M = X^T (V^{-1} - I) X.
  X_block <- collapse_block_diagonal(X)
  Delta_mat <- crossprod(X_block, Delta_V %**% X_block)

  ## Step 2: absorb the block-diagonal part into the corrected
  ## per-partition matrices defining G_on^{-1}.
  X_gram_corrected <- vector("list", K + 1)
  for (k in 1:(K + 1)) {
    rows_k <- ((k - 1) * p_expansions + 1):(k * p_expansions)
    Delta_diag_k <- Delta_mat[rows_k, rows_k]
    X_gram_corrected[[k]] <- crossprod(X[[k]]) + Delta_diag_k
  }

  ## Compute corrected G via eigendecomposition (reuses existing infra).
  #  Pass zero Schur corrections for the initial Gaussian decomposition.
  schur_zero <- lapply(1:(K + 1), function(k) {
    matrix(0, p_expansions, p_expansions)
  })
  G_list_c <- compute_G_eigen(X_gram_corrected, Lambda, K,
                              parallel_eigen, cl,
                              chunk_size, num_chunks, rem_chunks, family,
                              unique_penalty_per_partition,
                              L_partition_list,
                              keep_G = TRUE, schur_zero)

  ## Step 3: remove the absorbed diagonal blocks, leaving only the
  ## cross-partition remainder.
  Delta_offdiag <- Delta_mat
  for (k in 1:(K + 1)) {
    rows_k <- ((k - 1) * p_expansions + 1):(k * p_expansions)
    Delta_offdiag[rows_k, rows_k] <- 0
  }

  ## Step 4: eigentruncate the remainder and encode it as E J E^T.
  eig_off <- eigen(Delta_offdiag, symmetric = TRUE)
  r <- sum(abs(eig_off$values) >
             max(abs(eig_off$values)) * sqrt(.Machine$double.eps))

  if (r == 0) {
    ## No cross-partition coupling; corrected G is exact.
    return(list(
      use_woodbury = FALSE,
      Ghalf_corrected = G_list_c$Ghalf,
      GhalfInv_corrected = G_list_c$GhalfInv,
      G_corrected = G_list_c$G
    ))
  }

  if (r >= P * rank_threshold_fraction) {
    ## Low-rank approximation not worthwhile; fall back to dense.
    return(list(use_woodbury = FALSE))
  }

  ## Supplement notation: E J E^T for the retained off-diagonal remainder.
  E <- eig_off$vectors[, 1:r, drop = FALSE] %**%
    diag(sqrt(abs(eig_off$values[1:r])), r, r)
  J_signs <- sign(eig_off$values[1:r])

  ## Precompute G_{on,k} E_k blockwise for the later Woodbury inverse and
  ## square-root steps.
  GL <- lapply(1:(K + 1), function(k) {
    rows_k <- ((k - 1) * p_expansions + 1):(k * p_expansions)
    G_list_c$G[[k]] %**% E[rows_k, , drop = FALSE]
  })

  ## Assemble E^T G_on E on the retained subspace.
  GL_full <- do.call(rbind, GL)
  LtGL <- crossprod(E, GL_full)

  ## Precompute (J^{-1} + E^T G_on E)^{-1}; this is the only dense inverse
  ## in the Woodbury correction.
  inner_inv <- invert(diag(1 / J_signs, r, r) + LtGL)

  list(
    use_woodbury = TRUE,
    Ghalf_corrected = G_list_c$Ghalf,
    GhalfInv_corrected = G_list_c$GhalfInv,
    G_corrected = G_list_c$G,
    E = E,
    J_signs = J_signs,
    r = r,
    inner_inv = inner_inv,
    GL = GL,
    L = E,
    S_signs = J_signs,
    M = inner_inv
  )
}


## Sub-function: .woodbury_redecompose_weighted

#' Per-Iteration Weighted Woodbury Redecomposition
#'
#' Weighted analogue of \code{.woodbury_decompose_V()}. At each GLM
#' iteration, this helper rebuilds the current correlated information,
#' absorbs the block-diagonal part into
#' \eqn{\mathbf{G}_{\mathrm{on}}^{-1}}, and refactors the cross-partition
#' remainder as \eqn{\mathbf{E}\mathbf{J}\mathbf{E}^{\top}}.
#'
#' It is standalone in the sense that it returns the full Woodbury state
#' for the current working weights:
#' corrected block-diagonal factors, the low-rank basis, and the small
#' inverse \eqn{(\mathbf{J}^{-1} +
#' \mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}\mathbf{E})^{-1}}.
#'
#' @param Delta_V Fixed \eqn{N \times N} matrix
#'   \eqn{\mathbf{V}^{-1} - \mathbf{I}}.
#' @param X_block Full \eqn{N \times P} block-diagonal design matrix.
#' @param DV_X Precomputed \eqn{N \times P} product
#'   of \eqn{(\mathbf{V}^{-1} - \mathbf{I})\mathbf{X}}.
#' @param W Length-\eqn{N} vector of current GLM working weights.
#' @param X List of partition-specific design matrices.
#' @param K,p_expansions Integer dimensions.
#' @param X_gram_weighted List of \eqn{K+1} weighted Gram matrices
#'   \eqn{\mathbf{X}_k^{\top}\mathrm{diag}(\mathbf{W}_k)\mathbf{X}_k}.
#' @param Lambda Shared \eqn{p \times p} penalty matrix.
#' @param schur_corrections List of \eqn{K+1} Schur correction matrices.
#' @param unique_penalty_per_partition Logical.
#' @param L_partition_list Partition-specific penalty matrices.
#' @param parallel_eigen Logical; parallel eigendecomposition.
#' @param cl,chunk_size,num_chunks,rem_chunks Parallel parameters.
#' @param rank_threshold_fraction Numeric; Woodbury threshold (default 1/3).
#' @param family GLM family object.
#'
#' @return A list with the same structure as \code{.woodbury_decompose_V}:
#' \describe{
#'   \item{use_woodbury}{Logical; FALSE if rank exceeds threshold.}
#'   \item{Ghalf_corrected}{List of corrected \eqn{\mathbf{G}_{\mathrm{on},k}^{1/2}}.}
#'   \item{GhalfInv_corrected}{List of corrected \eqn{\mathbf{G}_{\mathrm{on},k}^{-1/2}}.}
#'   \item{G_corrected}{List of corrected \eqn{\mathbf{G}_{\mathrm{on},k}}.}
#'   \item{E}{Off-diagonal low-rank factor (\eqn{P \times r}), denoted by
#'     \eqn{\mathbf{E}} in the supplement.}
#'   \item{J_signs}{Length-\eqn{r} sign vector, i.e.\ the diagonal of
#'     \eqn{\mathbf{J}}.}
#'   \item{r}{Integer effective rank.}
#'   \item{inner_inv}{Precomputed \eqn{r \times r} Woodbury inner inverse; in
#'     supplement notation,
#'     \eqn{(\mathbf{J}^{-1} + \mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}
#'     \mathbf{E})^{-1}}.}
#'   \item{GL}{List of per-partition matrices corresponding to
#'     \eqn{\mathbf{G}_{\mathrm{on},k}\mathbf{E}[rows_k,]} in supplement notation.}
#'   \item{L,S_signs,M}{Legacy internal aliases for
#'     \code{E}, \code{J_signs}, and \code{inner_inv}.}
#' }
#'
#' @keywords internal
.woodbury_redecompose_weighted <- function(Delta_V, X_block, DV_X, W,
                                           X, K, p_expansions,
                                           X_gram_weighted,
                                           Lambda, schur_corrections,
                                           unique_penalty_per_partition,
                                           L_partition_list,
                                           parallel_eigen, cl,
                                           chunk_size, num_chunks,
                                           rem_chunks,
                                           rank_threshold_fraction = 1/3,
                                           family) {

  P <- p_expansions * (K + 1)

  ## Weighted analogue of manuscript M:
  ## X^T diag(W) (V^{-1} - I) X.
  ## Since DV_X = (V^{-1} - I) X is precomputed, this is:
  #    crossprod(X_block, W * DV_X)
  #  which costs O(N * P^2) -- cheaper than the P^3 eigen that the
  #  dense path requires.
  Delta_mat_W <- crossprod(X_block, c(W) * DV_X)

  ## Absorb the weighted block-diagonal portion into the corrected
  ## per-partition information defining G_on^{-1} at the current iterate.
  X_gram_corrected <- vector("list", K + 1)
  for (k in 1:(K + 1)) {
    rows_k <- ((k - 1) * p_expansions + 1):(k * p_expansions)
    Delta_diag_W_k <- Delta_mat_W[rows_k, rows_k]
    X_gram_corrected[[k]] <- X_gram_weighted[[k]] + Delta_diag_W_k
  }

  ## Eigendecompose per partition via compute_G_eigen.
  #  This adds Lambda + Schur corrections and produces Ghalf, GhalfInv, G.
  G_list_c <- compute_G_eigen(X_gram_corrected, Lambda, K,
                              parallel_eigen, cl,
                              chunk_size, num_chunks, rem_chunks, family,
                              unique_penalty_per_partition,
                              L_partition_list,
                              keep_G = TRUE, schur_corrections)

  ## Remove the absorbed diagonal blocks, leaving the weighted
  ## cross-partition remainder only.
  Delta_offdiag_W <- Delta_mat_W
  for (k in 1:(K + 1)) {
    rows_k <- ((k - 1) * p_expansions + 1):(k * p_expansions)
    Delta_offdiag_W[rows_k, rows_k] <- 0
  }

  ## Eigentruncate the weighted remainder and encode it as E J E^T.
  eig_off <- eigen(Delta_offdiag_W, symmetric = TRUE)
  r <- sum(abs(eig_off$values) >
             max(abs(eig_off$values)) * sqrt(.Machine$double.eps))

  if (r == 0) {
    ## No cross-partition coupling at this W; corrected G is exact.
    return(list(
      use_woodbury = TRUE,
      Ghalf_corrected = G_list_c$Ghalf,
      GhalfInv_corrected = G_list_c$GhalfInv,
      G_corrected = G_list_c$G,
      E = matrix(0, P, 0),
      J_signs = numeric(0),
      r = 0L,
      inner_inv = matrix(0, 0, 0),
      GL = lapply(1:(K + 1), function(k) matrix(0, p_expansions, 0)),
      L = matrix(0, P, 0),
      S_signs = numeric(0),
      M = matrix(0, 0, 0)
    ))
  }

  if (r >= P * rank_threshold_fraction) {
    ## Rank too high at this iteration; signal fallback to dense.
    return(list(use_woodbury = FALSE))
  }

  ## Supplement notation: E J E^T for the weighted off-diagonal remainder.
  E <- eig_off$vectors[, 1:r, drop = FALSE] %**%
    diag(sqrt(abs(eig_off$values[1:r])), r, r)
  J_signs <- sign(eig_off$values[1:r])

  ## Precompute G_{c,k} E_k blockwise for the current iterate.
  GL <- lapply(1:(K + 1), function(k) {
    rows_k <- ((k - 1) * p_expansions + 1):(k * p_expansions)
    G_list_c$G[[k]] %**% E[rows_k, , drop = FALSE]
  })

  ## Assemble E^T G_on E on the retained subspace.
  GL_full <- do.call(rbind, GL)
  LtGL <- crossprod(E, GL_full)

  ## Precompute (J^{-1} + E^T G_on E)^{-1} on the retained subspace.
  inner_inv <- invert(diag(1 / J_signs, r, r) + LtGL)

  list(
    use_woodbury = TRUE,
    Ghalf_corrected = G_list_c$Ghalf,
    GhalfInv_corrected = G_list_c$GhalfInv,
    G_corrected = G_list_c$G,
    E = E,
    J_signs = J_signs,
    r = r,
    inner_inv = inner_inv,
    GL = GL,
    L = E,
    S_signs = J_signs,
    M = inner_inv
  )
}


## Sub-function: .woodbury_halfsqrt_components

#' Compute Woodbury Half-Square-Root Components
#'
#' Converts the low-rank Woodbury decomposition into square-root objects
#' used directly by the solver. Starting from
#' \eqn{\mathbf{N} = \mathbf{G}_{\mathrm{on}}^{1/2}\mathbf{E}}, it builds
#' a basis for the supplement matrix \eqn{\mathbf{Q}} and stores
#' \eqn{\mathbf{F}^{1/2}} and \eqn{\mathbf{F}^{-1/2}} in the internal forms
#' \eqn{\mathbf{I} - \mathbf{U}_Q\mathbf{C}\mathbf{U}_Q^{\top}} and
#' \eqn{\mathbf{I} + \mathbf{U}_Q\mathbf{C}_{\mathrm{inv}}\mathbf{U}_Q^{\top}}.
#'
#' This helper is standalone: its output is the complete square-root state
#' consumed by \code{.lagrangian_project_woodbury()}.
#'
#' @param Ghalf_corrected List of corrected \eqn{\mathbf{G}_{\mathrm{on},k}^{1/2}}.
#' @param E Off-diagonal low-rank factor (\eqn{P \times r}), denoted
#'   \eqn{\mathbf{E}} in the supplement.
#' @param J_signs Length-\eqn{r} sign vector corresponding to the
#'   diagonal of \eqn{\mathbf{J}}.
#' @param r Integer effective rank.
#' @param K,p_expansions Integer dimensions.
#'
#' @return A list with components:
#' \describe{
#'   \item{valid}{Logical; FALSE if \eqn{\mathbf{F}} is not positive
#'     definite (signals fallback to dense).}
#'   \item{U_Q}{\eqn{P \times r} left singular vectors of
#'     \eqn{\mathbf{N} = \mathbf{G}_{\mathrm{on}}^{1/2}\mathbf{E}}. This spans the same
#'     subspace as the supplement QR factor \eqn{\mathbf{Q}}.}
#'   \item{Q_basis}{Alias for \code{U_Q}, included so the stored object can
#'     be read directly against the supplement notation \eqn{\mathbf{Q}}.}
#'   \item{C}{\eqn{r \times r} diagonal matrix
#'     encoding the rank-\eqn{r} update in the internal form
#'     \eqn{\mathbf{F}^{1/2} = \mathbf{I}_P - \mathbf{U}_Q\mathbf{C}
#'     \mathbf{U}_Q^{\top}}.}
#'   \item{C_inv}{\eqn{r \times r} diagonal matrix
#'     encoding the inverse update in the internal form
#'     \eqn{\mathbf{F}^{-1/2} = \mathbf{I}_P +
#'     \mathbf{U}_Q\mathbf{C}_{\mathrm{inv}}\mathbf{U}_Q^{\top}}.}
#'   \item{GchalfUQ}{List of \eqn{K+1} matrices, each \eqn{p \times r}:
#'     \eqn{\mathbf{G}_{\mathrm{on},k}^{1/2} \mathbf{U}_Q[rows_k,]}.}
#' }
#'
#' @keywords internal
.woodbury_halfsqrt_components <- function(Ghalf_corrected, E, J_signs,
                                          r, K, p_expansions) {
  P <- p_expansions * (K + 1)

  ## Handle the degenerate case r = 0 (no off-diagonal coupling).
  #  Return an identity-like structure with trivial correction matrices.
  if (r == 0) {
    GchalfUQ <- lapply(1:(K + 1), function(k) {
      matrix(0, p_expansions, 0)
    })
    return(list(
      valid = TRUE,
      U_Q = matrix(0, P, 0),
      Q_basis = matrix(0, P, 0),
      C = matrix(0, 0, 0),
      C_inv = matrix(0, 0, 0),
      GchalfUQ = GchalfUQ
    ))
  }

  ## Form the supplement matrix N = G_on^{1/2} E.
  N_mat <- matrix(0, P, r)
  for (k in 1:(K + 1)) {
    rows_k <- ((k - 1) * p_expansions + 1):(k * p_expansions)
    N_mat[rows_k, ] <- Ghalf_corrected[[k]] %**% E[rows_k, , drop = FALSE]
  }

  ## Thin SVD of N. The left singular vectors span the same subspace as
  ## the manuscript QR factor Q, which is all we need for the projector
  ## and square-root update.
  svd_Q <- svd(N_mat, nu = r, nv = r)
  U_Q <- svd_Q$u[, 1:r, drop = FALSE]
  sigma_Q <- svd_Q$d[1:r]

  ## The retained eigenvalues of the rank-r correction determine the
  ## diagonal coefficients in the internal representation
  ## F^{1/2} = I - U_Q C U_Q^T.
  d_F <- sigma_Q^2 / (1 / J_signs + sigma_Q^2)

  ## Check that 1 - d_F > 0 (F is positive definite).
  one_minus_dF <- 1 - d_F
  if (any(one_minus_dF <= sqrt(.Machine$double.eps))) {
    return(list(valid = FALSE))
  }

  C <- diag(1 - sqrt(one_minus_dF), r, r)
  C_inv <- diag(1 / sqrt(one_minus_dF) - 1, r, r)

  ## Precompute G_{c,k}^{1/2} U_Q blockwise so the Woodbury-corrected
  ## projection can apply the rank-r update partition by partition.
  GchalfUQ <- lapply(1:(K + 1), function(k) {
    rows_k <- ((k - 1) * p_expansions + 1):(k * p_expansions)
    Ghalf_corrected[[k]] %**% U_Q[rows_k, , drop = FALSE]
  })

  list(valid = TRUE,
       U_Q = U_Q,
       Q_basis = U_Q,
       C = C,
       C_inv = C_inv,
       GchalfUQ = GchalfUQ)
}


## Sub-function: .compute_Xy_V

#' Compute Partitioned GLS Cross-Products
#'
#' Computes \eqn{\mathbf{X}_k^{\top}\mathbf{V}^{-1}\mathbf{y}} for each
#' partition, accounting for cross-partition contributions from the
#' correlation structure.
#'
#' @param X List of partition-specific design matrices.
#' @param y List of response vectors by partition.
#' @param VhalfInv_perm \eqn{\mathbf{V}^{-1/2}} permuted to partition
#'   ordering.
#' @param K,p_expansions Integer dimensions.
#' @param order_list Partition-to-data index mapping (unused here but
#'   retained for interface consistency).
#'
#' @return A list of \eqn{K+1} column vectors, each
#'   \eqn{p\_expansions \times 1}.
#'
#' @keywords internal
.compute_Xy_V <- function(X, y, VhalfInv_perm, K, p_expansions,
                          order_list) {

  X_block <- collapse_block_diagonal(X)
  y_vec <- cbind(unlist(y))

  ## V^{-1} y in partition order (two-step multiply avoids forming V^{-1}).
  Vinv_y <- VhalfInv_perm %**% (VhalfInv_perm %**% y_vec)

  ## X^T V^{-1} y, split by partition.
  XtVinvy_full <- crossprod(X_block, Vinv_y)

  lapply(1:(K + 1), function(k) {
    rows_k <- ((k - 1) * p_expansions + 1):(k * p_expansions)
    XtVinvy_full[rows_k, , drop = FALSE]
  })
}


## Sub-function: .compute_score_V_partitioned

#' Compute Partitioned GEE Score Vector
#'
#' Computes the GEE score vector \eqn{\mathbf{X}^{\top}\mathrm{diag}
#' (\mathbf{W})\mathbf{V}^{-1}(\mathbf{y} - \boldsymbol{\mu})} split
#' into per-partition pieces of dimension \eqn{p \times 1} each. Uses
#' the identity \eqn{\mathbf{V}^{-1} = \mathbf{I} + \boldsymbol{\Delta}_V}
#' where \eqn{\boldsymbol{\Delta}_V = \mathbf{V}^{-1} - \mathbf{I}} is
#' precomputed and fixed across iterations, avoiding explicit formation
#' of \eqn{\mathbf{V}^{-1}}.
#'
#' @param X List of partition-specific design matrices.
#' @param X_block Full \eqn{N \times P} block-diagonal design.
#' @param y List of response vectors by partition.
#' @param result List of current coefficient column vectors by partition.
#' @param K,p_expansions Integer dimensions.
#' @param family GLM family object.
#' @param W Length-\eqn{N} vector of GLM working weights.
#' @param Delta_V Fixed \eqn{N \times N} matrix
#'   \eqn{\mathbf{V}^{-1} - \mathbf{I}}.
#' @param observation_weights List of observation weights by partition.
#'
#' @return A list of \eqn{K+1} column vectors, each
#'   \eqn{p\_expansions \times 1}, representing the partition-wise
#'   components of the full GEE score.
#'
#' @keywords internal
.compute_score_V_partitioned <- function(X, X_block, y, result,
                                         K, p_expansions,
                                         family, W, Delta_V,
                                         observation_weights) {

  ## Current linear predictor and mean.
  beta_block <- cbind(unlist(result))
  mu <- family$linkinv(X_block %**% beta_block)

  ## Raw residual: y - mu.
  y_block <- cbind(unlist(y))
  residual <- y_block - mu

  ## V^{-1} (y - mu) = (I + Delta_V)(y - mu) = residual + Delta_V residual.
  #  This avoids forming V^{-1} explicitly.
  Vinv_resid <- residual + Delta_V %**% residual

  ## Full score: X^T W V^{-1} (y - mu).
  score_full <- crossprod(X_block, c(W) * Vinv_resid)

  ## Split by partition.
  lapply(1:(K + 1), function(k) {
    rows_k <- ((k - 1) * p_expansions + 1):(k * p_expansions)
    score_full[rows_k, , drop = FALSE]
  })
}


## Sub-function: .get_B_gee_gaussian

#' Path 1a: Gaussian Identity + GEE (Closed-Form Full-System Solve)
#'
#' Computes the constrained penalized GLS estimate for Gaussian response
#' with identity link when a correlation structure is present. Because
#' \eqn{\mathbf{V}^{-1/2}} couples all partitions, fitting must operate on
#' the full \eqn{P}-dimensional whitened system rather than partition-wise.
#'
#' The unconstrained GLS estimate is:
#' \deqn{\hat{\boldsymbol{\beta}} = \mathbf{G}(\mathbf{X}^{*\top}
#'   \mathbf{y}^{*}), \quad
#'   \mathbf{G} = (\mathbf{X}^{*\top}\mathbf{X}^{*} +
#'   \boldsymbol{\Lambda})^{-1}}
#' where \eqn{\mathbf{X}^{*} = \mathbf{V}^{-1/2}\mathbf{X}} and
#' \eqn{\mathbf{y}^{*} = \mathbf{V}^{-1/2}\mathbf{y}}. The constrained
#' estimate is then obtained via the \eqn{\mathbf{G}^{1/2}\mathbf{r}^*}
#' trick in the full \eqn{P}-dimensional space.
#'
#' @param X_block Full \eqn{N \times P} unwhitened block-diagonal design.
#' @param X_tilde Whitened design \eqn{\mathbf{V}^{-1/2}\mathbf{X}}.
#' @param y_tilde Whitened response \eqn{\mathbf{V}^{-1/2}\mathbf{y}}.
#' @param VhalfInv_perm \eqn{\mathbf{V}^{-1/2}} permuted to partition
#'   ordering.
#' @param Lambda_block Full \eqn{P \times P} block-diagonal penalty.
#' @param A Constraint matrix (\eqn{P \times R}).
#' @param K,p_expansions Integer dimensions.
#' @param constraint_value_vectors Constraint RHS list encoding
#'   \eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{c}}.
#' @param family GLM family object.
#' @param return_G_getB Logical; return covariance components.
#' @param quadprog Logical; apply QP refinement.
#' @param qp_Amat,qp_bvec,qp_meq QP constraint specification.
#' @param qp_score_function Score function for QP step.
#' @param order_list,observation_weights Standard partition arguments.
#' @param ... Passed to score function.
#'
#' @return If \code{return_G_getB = TRUE}: list with \code{B},
#'   \code{G_list}, and \code{qp_info}. Otherwise: list of \code{B} and
#'   \code{qp_info}.
#'
#' @keywords internal
.get_B_gee_gaussian <- function(X_block, X_tilde, y_tilde, VhalfInv_perm,
                                Lambda_block, A, K, p_expansions,
                                constraint_value_vectors, family,
                                return_G_getB,
                                quadprog, qp_Amat, qp_bvec, qp_meq,
                                qp_score_function,
                                order_list, observation_weights, ...) {

  ## Observation weights enter the Gaussian GEE solve in the whitened system,
  #  exactly as they do in the iterative GLM GEE paths.
  obs_wt <- c(unlist(observation_weights))
  if (length(obs_wt) == 0L) {
    obs_wt <- rep(1, nrow(X_tilde))
  } else if (length(obs_wt) == 1L) {
    obs_wt <- rep(obs_wt, nrow(X_tilde))
  }

  ## Full P x P Gram matrix from the weighted whitened design:
  #  X_tilde^T D X_tilde = X^T V^{-1/2} D V^{-1/2} X.
  Gram_full <- crossprod(X_tilde, obs_wt * X_tilde)

  ## G = (X_tilde^T X_tilde + Lambda)^{-1}, a dense P x P matrix.
  G_full_inv <- Gram_full + Lambda_block
  G_full <- invert(G_full_inv)

  ## G^{1/2} via eigendecomposition of the full P x P G.
  eig_G <- eigen(G_full, symmetric = TRUE)
  vals_G <- eig_G$values
  vals_G[vals_G <= 0] <- 0
  G_full_half <- eig_G$vectors %**%
    (t(eig_G$vectors) * sqrt(vals_G))

  ## G^{-1/2} for per-partition extraction.
  inv_sqrt_vals_G <- ifelse(sqrt(pmax(eig_G$values, 0)) > 0,
                            1 / sqrt(pmax(eig_G$values, 0)), 0)
  G_full_half_inv <- eig_G$vectors %**%
    (t(eig_G$vectors) * inv_sqrt_vals_G)

  ## X_tilde^T D y_tilde: the P x 1 sufficient statistic for beta.
  Xy_tilde <- crossprod(X_tilde, obs_wt * y_tilde)

  ## Lagrangian projection in the full P-dimensional space.
  if (K == 0 | any(all.equal(unique(A), 0) == TRUE)) {
    A_proj <- cbind(rep(0, p_expansions * (K + 1)))
  } else {
    A_proj <- A
  }

  ## G^{1/2} r* trick.
  y_star <- G_full_half %**% Xy_tilde
  X_star <- G_full_half %**% A_proj

  comp_stab_sc <- 1 / sqrt(K + 1)
  resids_star <- .lm.fit(
    X_star * comp_stab_sc,
    y_star * comp_stab_sc
  )$residuals / comp_stab_sc

  ## Nonzero constraint values: add particular solution b_0.
  if (length(constraint_value_vectors) > 0) {
    if (any(unlist(constraint_value_vectors) != 0)) {
      preds_star <- X_star %**%
        (invert(crossprod(X_star) * comp_stab_sc) %**%
           crossprod(A_proj,
                     Reduce("rbind",
                            constraint_value_vectors) *
                       comp_stab_sc))
      resids_star <- resids_star + c(preds_star)
    }
  }

  ## beta_constrained = G^{1/2} r*.
  beta_block <- G_full_half %**% cbind(resids_star)

  ## Unpack into per-partition column vectors.
  result <- lapply(1:(K + 1), function(k) {
    cbind(beta_block[(k - 1) * p_expansions + 1:p_expansions])
  })

  qp_info <- NULL

  ## Optional QP refinement for inequality constraints.
  if (quadprog) {
    if (K == 0 | any(all.equal(unique(A), 0) == TRUE)) {
      A_qp <- cbind(rep(0, p_expansions * (K + 1)))
    } else {
      A_qp <- A
    }
    if (!is.null(qp_Amat)) {
      qp_Amat_c <- cbind(A_qp, qp_Amat)
    } else {
      qp_Amat_c <- A_qp
    }
    if (length(constraint_value_vectors) > 0) {
      constr_rhs <- Reduce('rbind', constraint_value_vectors)
      if (nrow(constr_rhs) < ncol(A_qp)) {
        constr_rhs <- c(rep(0, ncol(A_qp) - nrow(constr_rhs)),
                        constr_rhs)
      }
    } else {
      constr_rhs <- rep(0, ncol(A_qp))
    }
    if (!is.null(qp_bvec)) {
      qp_bvec_c <- c(constr_rhs, qp_bvec)
    } else {
      qp_bvec_c <- constr_rhs
    }
    if (!is.null(qp_meq)) {
      qp_meq_c <- ncol(A_qp) + qp_meq
    } else {
      qp_meq_c <- ncol(A_qp)
    }

    ## Information matrix for the QP: use full whitened Gram.
    info <- Gram_full + Lambda_block
    sc <- sqrt(mean(abs(info)))

    qp_score <- qp_score_function(
      X_tilde, y_tilde,
      VhalfInv_perm %**% cbind(family$linkinv(
        c(X_block %**% beta_block))),
      unlist(order_list), 1, VhalfInv_perm,
      unlist(observation_weights), ...
    )
    qp_result <- try({
      quadprog::solve.QP(
        Dmat = info / sc,
        dvec = (qp_score -
                  Lambda_block %**% beta_block +
                  info %**% beta_block) / sc,
        Amat = qp_Amat_c,
        bvec = qp_bvec_c,
        meq = qp_meq_c
      )
    }, silent = TRUE)

    if (!any(inherits(qp_result, 'try-error'))) {
      beta_block <- cbind(qp_result$solution)
      result <- lapply(1:(K + 1), function(k) {
        cbind(beta_block[(k - 1) * p_expansions + 1:p_expansions])
      })
      qp_info <- list(
        lagrangian = qp_result$Lagrangian,
        Amat_active = qp_Amat_c[,
                                c(1:ncol(A_qp),
                                  ncol(A_qp) + which(abs(qp_result$Lagrangian[-(1:ncol(A_qp))]) >
                                                       sqrt(.Machine$double.eps))),
                                drop = FALSE]
      )
    }
  }

  ## Return with per-partition G components.
  if (return_G_getB) {
    G_diag <- .extract_G_diagonal(G_full, p_expansions, K)
    Ghalf_diag <- .extract_G_diagonal(G_full_half, p_expansions, K)
    GhalfInv_diag <- .extract_G_diagonal(G_full_half_inv, p_expansions, K)
    G_diag_from_half <- lapply(1:(K + 1), function(k) {
      tcrossprod(Ghalf_diag[[k]])
    })
    return(list(
      B = result,
      G_list = list(G = G_diag_from_half,
                    Ghalf = Ghalf_diag,
                    GhalfInv = GhalfInv_diag),
      qp_info = qp_info
    ))
  } else {
    return(list(B = result, qp_info = qp_info))
  }
}


## Sub-function: .get_B_gee_woodbury

#' Path 1a-Woodbury: Gaussian GEE with Woodbury Acceleration
#'
#' Gaussian correlated solve using the Woodbury factorization from the
#' supplement. This replaces the dense GEE path when
#' \eqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{E}\mathbf{J}\mathbf{E}^{\top}}
#' has low rank and computes the constrained estimate through the same
#' transformed OLS projection as the independent case, with
#' \eqn{\mathbf{G}^{1/2} =
#' \mathbf{G}_{\mathrm{on}}^{1/2}\mathbf{F}^{1/2}}.
#'
#' This function is standalone for the Gaussian Woodbury path: it takes the
#' decomposition from \code{.woodbury_decompose_V()}, the square-root state
#' from \code{.woodbury_halfsqrt_components()}, performs the constrained
#' solve, and returns the same object structure as the dense Gaussian GEE
#' solver.
#'
#' Cost is \eqn{O(Kp^3 + Pr^2)} compared to \eqn{O(P^3)} for the dense
#' Path 1a.
#'
#' @param X,y Lists of partition-specific design matrices and responses.
#' @param K,p_expansions Integer dimensions.
#' @param VhalfInv_perm \eqn{\mathbf{V}^{-1/2}} permuted to partition
#'   ordering.
#' @param order_list Partition-to-data index mapping.
#' @param A Constraint matrix (\eqn{P \times R}).
#' @param R_constraints Number of columns of A.
#' @param constraint_value_vectors Constraint RHS list.
#' @param family GLM family object.
#' @param return_G_getB Logical; return covariance components.
#' @param quadprog Logical; apply QP refinement.
#' @param qp_Amat,qp_bvec,qp_meq QP constraint specification.
#' @param qp_score_function Score function for QP step.
#' @param observation_weights Observation weights.
#' @param wb_decomp Output of \code{.woodbury_decompose_V}.
#' @param wb_sqrt Output of \code{.woodbury_halfsqrt_components}.
#' @param parallel_aga,parallel_matmult Logical flags.
#' @param cl,chunk_size,num_chunks,rem_chunks Parallel parameters.
#' @param qp_global Logical; forced TRUE for GEE but passed for
#'   interface consistency.
#' @param ... Passed to sub-functions.
#'
#' @return Same structure as \code{.get_B_gee_gaussian}.
#'
#' @keywords internal
.get_B_gee_woodbury <- function(X, y, K, p_expansions,
                                VhalfInv_perm, order_list,
                                A, R_constraints,
                                constraint_value_vectors, family,
                                return_G_getB,
                                quadprog, qp_Amat, qp_bvec, qp_meq,
                                qp_score_function,
                                observation_weights,
                                wb_decomp, wb_sqrt,
                                parallel_aga, parallel_matmult,
                                cl, chunk_size, num_chunks, rem_chunks,
                                qp_global, ...) {

  ## GLS cross-products: X^T V^{-1} y per partition.
  Xy_V <- .compute_Xy_V(X, y, VhalfInv_perm, K, p_expansions,
                        order_list)

  ## Block-diagonal part of y*: G_on^{1/2} (X^T V^{-1} y).
  GhalfXy_V <- cbind(unlist(
    matmult_block_diagonal(wb_decomp$Ghalf_corrected, Xy_V, K,
                           parallel_matmult, cl,
                           chunk_size, num_chunks, rem_chunks)))

  ## Constrained projection via Woodbury-corrected Lagrangian.
  result <- .lagrangian_project_woodbury(
    GhalfXy_V, wb_decomp$Ghalf_corrected,
    A, K, p_expansions, R_constraints,
    constraint_value_vectors, family,
    wb_sqrt,
    parallel_aga, parallel_matmult,
    cl, chunk_size, num_chunks, rem_chunks)

  qp_info <- NULL

  ## QP refinement: GEE always uses dense SQP.
  if (quadprog) {
    X_block <- collapse_block_diagonal(X)
    beta_block <- cbind(unlist(result))
    y_block <- cbind(unlist(y))

    ## Whiten for the dense QP.
    X_tilde <- VhalfInv_perm %**% X_block
    y_tilde <- VhalfInv_perm %**% y_block

    Lambda_block <- .build_lambda_block(
      ## Reconstruct Lambda from corrected G -- use the shared Lambda
      ## that was passed to wb_decomp. Extract from Ghalf_corrected.
      ## Actually, we need Lambda_block here. Reconstruct from available.
      ## The simplest approach: the caller must supply Lambda_block.
      ## For now, compute from the Gram inverse structure.
      ## NOTE: In practice, Lambda_block is computed in the get_B()
      ## dispatcher and should be passed through. We reconstruct it
      ## from the wb_decomp's internal state.
      diag(0, p_expansions), K, FALSE, list())

    ## Full Gram from whitened design.
    Gram_full <- crossprod(X_tilde)
    info <- Gram_full + Lambda_block
    sc <- sqrt(mean(abs(info)))

    if (K == 0 | any(all.equal(unique(A), 0) == TRUE)) {
      A_qp <- cbind(rep(0, p_expansions * (K + 1)))
    } else {
      A_qp <- A
    }
    if (!is.null(qp_Amat)) {
      qp_Amat_c <- cbind(A_qp, qp_Amat)
    } else {
      qp_Amat_c <- A_qp
    }
    constr_rhs <- if (length(constraint_value_vectors) > 0) {
      cr <- Reduce('rbind', constraint_value_vectors)
      if (nrow(cr) < ncol(A_qp)) c(rep(0, ncol(A_qp) - nrow(cr)), cr)
      else cr
    } else {
      rep(0, ncol(A_qp))
    }
    qp_bvec_c <- if (!is.null(qp_bvec)) c(constr_rhs, qp_bvec) else constr_rhs
    qp_meq_c <- ncol(A_qp) + if (!is.null(qp_meq)) qp_meq else 0

    qp_score <- qp_score_function(
      X_tilde, y_tilde,
      VhalfInv_perm %**% cbind(family$linkinv(
        c(X_block %**% beta_block))),
      unlist(order_list), 1, VhalfInv_perm,
      unlist(observation_weights), ...
    )
    qp_result <- try({
      quadprog::solve.QP(
        Dmat = info / sc,
        dvec = (qp_score -
                  Lambda_block %**% beta_block +
                  info %**% beta_block) / sc,
        Amat = qp_Amat_c,
        bvec = qp_bvec_c,
        meq = qp_meq_c
      )
    }, silent = TRUE)

    if (!any(inherits(qp_result, 'try-error'))) {
      beta_block <- cbind(qp_result$solution)
      result <- lapply(1:(K + 1), function(k) {
        cbind(beta_block[(k - 1) * p_expansions + 1:p_expansions])
      })
      qp_info <- list(
        lagrangian = qp_result$Lagrangian,
        Amat_active = qp_Amat_c[,
                                c(1:ncol(A_qp),
                                  ncol(A_qp) + which(abs(qp_result$Lagrangian[-(1:ncol(A_qp))]) >
                                                       sqrt(.Machine$double.eps))),
                                drop = FALSE]
      )
    }
  }

  ## Return with per-partition G components.
  if (return_G_getB) {
    G_list <- list(
      G = lapply(wb_decomp$Ghalf_corrected, function(mat) tcrossprod(mat)),
      Ghalf = wb_decomp$Ghalf_corrected,
      GhalfInv = wb_decomp$GhalfInv_corrected
    )
    return(list(B = result, G_list = G_list, qp_info = qp_info))
  } else {
    return(list(B = result, qp_info = qp_info))
  }
}


## Sub-function: .get_B_gee_glm

#' Path 1b: Non-Gaussian GEE (Damped SQP with Full Whitened Design)
#'
#' Estimates constrained coefficients for non-Gaussian GLMs with correlation
#' structures using damped Sequential Quadratic Programming (SQP) in the
#' full \eqn{P}-dimensional whitened space. The whitened design
#' \eqn{\mathbf{X}^{*} = \mathbf{V}^{-1/2}\mathbf{X}} is required
#' because \eqn{\mathbf{V}^{-1/2}} couples all partitions.
#'
#' @inheritParams .get_B_gee_gaussian
#' @param y_block Unwhitened response vector \eqn{\mathbf{y}}.
#' @param iterate Logical; if \code{TRUE}, iterate until convergence.
#' @param tol Convergence tolerance.
#' @param glm_weight_function Function computing GLM working weights.
#' @param schur_correction_function Function computing Schur corrections.
#' @param need_dispersion_for_estimation Logical.
#' @param dispersion_function Dispersion estimation function.
#' @param VhalfInv Inverse square root correlation matrix.
#' @param ... Passed to weight, correction, dispersion, and score functions.
#'
#' @return Same structure as \code{.get_B_gee_gaussian}.
#'
#' @keywords internal
.get_B_gee_glm <- function(X_block, X_tilde, y_block, y_tilde,
                           VhalfInv_perm, Lambda_block, A, K,
                           p_expansions, constraint_value_vectors,
                           family, return_G_getB, iterate, tol,
                           qp_Amat, qp_bvec, qp_meq,
                           qp_score_function,
                           order_list, observation_weights,
                           glm_weight_function, schur_correction_function,
                           need_dispersion_for_estimation,
                           dispersion_function, VhalfInv, ...) {

  beta_block <- cbind(rep(0, ncol(X_block)))

  ## SQP iteration control.
  damp_cnt <- 0
  master_cnt <- 0
  err <- Inf

  ## Original-scale linear predictor (X_block is unwhitened).
  XB <- X_block %**% beta_block

  ## Combine smoothness equality and user inequality constraints.
  if (K == 0 | any(all.equal(unique(A), 0) == TRUE)) {
    A <- cbind(rep(0, p_expansions * (K + 1)))
  }
  if (!is.null(qp_Amat)) {
    qp_Amat <- cbind(A, qp_Amat)
  } else {
    qp_Amat <- A
  }
  if (length(constraint_value_vectors) > 0) {
    constr_A <- Reduce('rbind', constraint_value_vectors)
    if (nrow(constr_A) < ncol(A)) {
      constr_A <- c(rep(0, ncol(A) - nrow(constr_A)), constr_A)
    }
  } else {
    constr_A <- rep(0, ncol(A))
  }
  if (!is.null(qp_bvec)) {
    qp_bvec <- c(constr_A, qp_bvec)
  } else {
    qp_bvec <- constr_A
  }
  if (!is.null(qp_meq)) {
    qp_meq <- ncol(A) + qp_meq
  } else {
    qp_meq <- ncol(A)
  }

  qp_info <- NULL

  ## Damped SQP loop.
  while (err > tol & damp_cnt < 10 & master_cnt < 100) {
    master_cnt <- master_cnt + 1
    damp <- 2^(-(damp_cnt))

    if (need_dispersion_for_estimation) {
      dispersion_temp <- dispersion_function(
        mu = family$linkinv(XB),
        y = y_block,
        order_indices = unlist(order_list),
        family = family,
        observation_weights = unlist(observation_weights),
        VhalfInv = VhalfInv,
        ...
      )
    } else {
      dispersion_temp <- 1
    }

    W <- c(glm_weight_function(family$linkinv(XB),
                               y_block,
                               unlist(order_list),
                               family, dispersion_temp,
                               unlist(observation_weights), ...))

    ## Schur correction using the original-scale design.
    result <- lapply(1:(K + 1), function(k) {
      cbind(beta_block[(k - 1) * p_expansions + 1:p_expansions])
    })
    schur_correction <-
      schur_correction_function(
        list(X_block), list(y_block),
        list(cbind(unlist(result))),
        dispersion_temp, list(unlist(order_list)),
        0, family, unlist(observation_weights), ...
      )
    if (!(any(unlist(schur_correction) != 0))) {
      schur_correction_collapsed <- 0
    } else {
      schur_correction_collapsed <-
        collapse_block_diagonal(schur_correction)
    }

    info <- crossprod(X_tilde, W * X_tilde) +
      Lambda_block +
      schur_correction_collapsed

    sc <- sqrt(mean(abs(info)))

    if (master_cnt == 1) {
      ## First iteration: constrained Newton step.
      infoinv_block <- sc * invert(sc * info)
      U <- (diag(1, nrow(info)) -
              tcrossprod(infoinv_block %**%
                           A %**%
                           invert(crossprod(A,
                                            infoinv_block %**% A)),
                         A))
      mu_vec <- cbind(family$linkinv(c(XB)))
      beta_new <- c(
        U %**% infoinv_block %**%
          crossprod(X_tilde,
                    y_tilde - VhalfInv_perm %**% mu_vec)
      )
      infoinv_block <- NULL
    } else {
      ## Subsequent iterations: full QP solve.
      qp_score <- qp_score_function(
        X_tilde, y_tilde,
        VhalfInv_perm %**% cbind(family$linkinv(XB)),
        unlist(order_list), dispersion_temp,
        VhalfInv_perm,
        unlist(observation_weights), ...
      )
      qp_result <- try({
        quadprog::solve.QP(
          Dmat = info / sc,
          dvec = (qp_score -
                    Lambda_block %**% beta_block +
                    info %**% beta_block) / sc,
          Amat = qp_Amat,
          bvec = qp_bvec,
          meq = qp_meq
        )
      }, silent = TRUE)
      if (any(inherits(qp_result, 'try-error'))) {
        beta_new <- 0 * beta_block
      } else {
        beta_new <- qp_result$solution
        qp_info <- list(
          lagrangian = qp_result$Lagrangian,
          Amat_active = qp_Amat[,
                                c(1:ncol(A),
                                  ncol(A) + which(abs(qp_result$Lagrangian[-(1:ncol(A))]) >
                                                    sqrt(.Machine$double.eps))),
                                drop = FALSE]
        )
      }
    }

    ## Non-iterative mode: accept after the first Newton step.
    if (!iterate & master_cnt > 2) {
      beta_block <- beta_new
      damp_cnt <- 11
      master_cnt <- 101
      err <- tol - 1
    } else {
      beta_new <- (1 - damp) * beta_block + damp * beta_new
      XB <- X_block %**% cbind(beta_new)

      if (need_dispersion_for_estimation) {
        dispersion_temp <- dispersion_function(
          mu = family$linkinv(XB),
          y = y_block,
          order_indices = unlist(order_list),
          family = family,
          observation_weights = unlist(observation_weights),
          VhalfInv = VhalfInv,
          ...
        )
      } else {
        dispersion_temp <- 1
      }
      W <- c(glm_weight_function(family$linkinv(XB),
                                 y_block,
                                 unlist(order_list),
                                 family, dispersion_temp,
                                 unlist(observation_weights), ...))

      mu_new <- family$linkinv(c(XB))
      if (!is.null(family$custom_dev.resids)) {
        raw <- family$custom_dev.resids(
          y_block, mu_new, unlist(order_list),
          family, unlist(observation_weights), ...
        )
        W_safe <- pmax(W, sqrt(.Machine$double.eps))
        err_new <- mean((
          VhalfInv_perm %**%
            cbind(sign(raw) * sqrt(abs(raw)) / sqrt(c(W_safe)))
        )^2)
      } else if (is.null(family$dev.resids)) {
        err_new <- mean((unlist(observation_weights) *
                           (y_block - cbind(mu_new)))^2)
      } else {
        err_new <-
          mean(unlist(observation_weights) *
                 family$dev.resids(y_block, cbind(mu_new),
                                   wt = 1 / W))
      }

      if (is.null(err_new) | is.na(err_new) | !is.finite(err_new)) {
        damp_cnt <- damp_cnt + 1
      } else if (err_new <= err) {
        prev_err <- err
        err <- err_new
        abs_diff <- max(abs(beta_new - beta_block))
        beta_block <- beta_new
        damp_cnt <- 0
        if ((abs_diff < tol) & (prev_err - err < tol) &
            (master_cnt > 10)) {
          damp_cnt <- 11
          master_cnt <- 101
          err <- tol - 1
        }
      } else {
        damp_cnt <- damp_cnt + 1
      }
    }

    result <- lapply(1:(K + 1), function(k) {
      cbind(beta_block[(k - 1) * p_expansions + 1:p_expansions])
    })
  }

  ## Return with G components if requested.
  if (return_G_getB) {
    XB <- X_block %**% cbind(unlist(result))
    if (need_dispersion_for_estimation) {
      dispersion_temp <- dispersion_function(
        mu = family$linkinv(XB),
        y = y_block,
        order_indices = unlist(order_list),
        family = family,
        observation_weights = unlist(observation_weights),
        VhalfInv = VhalfInv,
        ...
      )
    } else {
      dispersion_temp <- 1
    }
    W <- c(glm_weight_function(family$linkinv(XB),
                               y_block,
                               unlist(order_list),
                               family, dispersion_temp,
                               unlist(observation_weights), ...))

    schur_correction <-
      schur_correction_function(
        list(X_block), list(y_block),
        list(cbind(unlist(result))),
        dispersion_temp, list(unlist(order_list)),
        0, family, unlist(observation_weights), ...
      )
    if (!(any(unlist(schur_correction) != 0))) {
      schur_correction_collapsed <- 0
    } else {
      schur_correction_collapsed <-
        collapse_block_diagonal(schur_correction)
    }

    info <- crossprod(X_tilde, W * X_tilde) +
      Lambda_block +
      schur_correction_collapsed

    G <- lapply(1:(K + 1), function(k) NA)
    Ghalf <- lapply(1:(K + 1), function(k) NA)
    GhalfInv <- lapply(1:(K + 1), function(k) NA)
    for (k in 1:(K + 1)) {
      info_kk <- info[(k - 1) * p_expansions + 1:p_expansions,
                      (k - 1) * p_expansions + 1:p_expansions]
      eig <- eigen(info_kk, symmetric = TRUE)
      vals <- eig$values
      vals_safe <- pmax(vals, 0)
      vals_safe[vals <= sqrt(.Machine$double.eps)] <- 0
      sqrt_vals <- sqrt(vals_safe)
      Ghalf[[k]] <- eig$vectors %**%
        (t(eig$vectors) * sqrt_vals)
      inv_sqvals <- ifelse(sqrt_vals > 0, 1 / sqrt_vals, 0)
      GhalfInv[[k]] <- eig$vectors %**%
        (t(eig$vectors) * inv_sqvals)
      G[[k]] <- tcrossprod(Ghalf[[k]])
    }

    return(list(
      B = result,
      G_list = list(Ghalf = Ghalf,
                    GhalfInv = GhalfInv,
                    G = G),
      qp_info = qp_info
    ))
  } else {
    return(list(B = result, qp_info = qp_info))
  }
}


## Sub-function: .get_B_gee_glm_woodbury

#' Path 1b-Woodbury: Non-Gaussian GEE with Woodbury Acceleration
#'
#' Non-Gaussian correlated solve using the Woodbury factorization. At each
#' damped Newton step, the current weighted information matrix is split into
#' a block-diagonal part defining \eqn{\mathbf{G}_{\mathrm{on}}} and a
#' low-rank cross-partition part
#' \eqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{E}\mathbf{J}\mathbf{E}^{\top}}.
#' The constrained step is then taken through
#' \code{.lagrangian_project_woodbury()}.
#'
#' This function is standalone for the weighted Woodbury path: it carries
#' the iteration, repeatedly rebuilds the Woodbury factors at the current
#' working weights, and falls back to the dense correlated GLM solver if the
#' rank is too large or \eqn{\mathbf{F}} is not positive definite.
#'
#' @inheritParams .get_B_gee_glm
#' @param X,y Lists of partition-specific design matrices and responses.
#' @param wb_decomp_init Initial Woodbury decomposition from
#'   \code{.woodbury_decompose_V} (used only to confirm that the
#'   unweighted decomposition was valid; the weighted decomposition is
#'   recomputed at each iteration).
#' @param wb_sqrt_init Initial half-sqrt components (unused; retained
#'   for interface consistency with the Gaussian Woodbury path).
#' @param Lambda,Lambda_block Shared and full penalty matrices.
#' @param unique_penalty_per_partition Logical.
#' @param L_partition_list Partition-specific penalty matrices.
#' @param parallel_eigen,parallel_aga,parallel_matmult Logical flags.
#' @param cl,chunk_size,num_chunks,rem_chunks Parallel parameters.
#' @param ... Passed to sub-functions.
#'
#' @return Same structure as \code{.get_B_gee_glm}.
#'
#' @keywords internal
.get_B_gee_glm_woodbury <- function(X, y, K, p_expansions,
                                    VhalfInv_perm, order_list,
                                    A, R_constraints,
                                    constraint_value_vectors, family,
                                    return_G_getB, iterate, tol,
                                    qp_Amat, qp_bvec, qp_meq,
                                    qp_score_function,
                                    observation_weights,
                                    Lambda, Lambda_block,
                                    unique_penalty_per_partition,
                                    L_partition_list,
                                    wb_decomp_init, wb_sqrt_init,
                                    glm_weight_function,
                                    schur_correction_function,
                                    need_dispersion_for_estimation,
                                    dispersion_function, VhalfInv,
                                    parallel_eigen, parallel_aga,
                                    parallel_matmult,
                                    cl, chunk_size, num_chunks,
                                    rem_chunks, ...) {

  P <- p_expansions * (K + 1)

  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  ## Precomputation (done once, fixed across Newton iterations).
  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

  X_block <- collapse_block_diagonal(X)
  y_block <- cbind(unlist(y))
  N <- nrow(X_block)

  ## Delta_V = V^{-1} - I: fixed across iterations.
  Delta_V <- tcrossprod(VhalfInv_perm) - diag(N)

  ## DV_X = (V^{-1} - I) X: fixed across iterations. O(N^2 P) or
  #  O(nnz * P) if Delta_V is sparse.
  DV_X <- Delta_V %**% X_block

  ## Partition sizes for splitting W.
  n_per_partition <- sapply(X, nrow)

  ## Initialize beta at zero.
  beta_block <- cbind(rep(0, P))

  ## Damped Newton control (mirrors .get_B_gee_glm).
  damp_cnt <- 0
  master_cnt <- 0
  err <- Inf
  XB <- X_block %**% beta_block

  ## Track fallback state.
  fell_back_to_dense <- FALSE

  qp_info <- NULL

  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  ## Newton loop.
  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  while (err > tol & damp_cnt < 10 & master_cnt < 100) {
    master_cnt <- master_cnt + 1
    damp <- 2^(-(damp_cnt))

    ## ##  1. Compute dispersion at current iterate ##
    if (need_dispersion_for_estimation) {
      dispersion_temp <- dispersion_function(
        mu = family$linkinv(XB),
        y = y_block,
        order_indices = unlist(order_list),
        family = family,
        observation_weights = unlist(observation_weights),
        VhalfInv = VhalfInv,
        ...
      )
    } else {
      dispersion_temp <- 1
    }

    ## ##  2. Compute GLM working weights W ##
    W <- c(glm_weight_function(family$linkinv(XB),
                               y_block,
                               unlist(order_list),
                               family, dispersion_temp,
                               unlist(observation_weights), ...))

    ## ##  3. Compute Schur corrections at current iterate ##
    #  Per-partition calling convention (what compute_G_eigen expects).
    result <- lapply(1:(K + 1), function(k) {
      cbind(beta_block[(k - 1) * p_expansions + 1:p_expansions])
    })
    schur_corrections <- schur_correction_function(
      X, y, result, dispersion_temp, order_list, K, family,
      observation_weights, ...
    )

    ## ##  4. Compute weighted Gram per partition ##
    #  Split W by partition according to actual partition sizes.
    W_split <- split(W, rep(1:(K + 1), n_per_partition))

    Xw <- lapply(1:(K + 1), function(k) {
      if (nrow(X[[k]]) == 0) return(X[[k]])
      X[[k]] * sqrt(W_split[[k]])
    })
    X_gram_weighted <- compute_gram_block_diagonal(
      Xw, parallel_matmult, cl, chunk_size, num_chunks, rem_chunks)

    ## ##  5. Woodbury redecomposition at current W ##
    wb <- .woodbury_redecompose_weighted(
      Delta_V, X_block, DV_X, W,
      X, K, p_expansions,
      X_gram_weighted,
      Lambda, schur_corrections,
      unique_penalty_per_partition, L_partition_list,
      parallel_eigen, cl,
      chunk_size, num_chunks, rem_chunks,
      rank_threshold_fraction = 1/3, family)

    if (!wb$use_woodbury) {
      ## Rank too high at this iteration; fall back to dense permanently.
      fell_back_to_dense <- TRUE
      break
    }

    ## ##  6. Compute halfsqrt components ##
    wb_sqrt <- .woodbury_halfsqrt_components(
      wb$Ghalf_corrected, wb$E, wb$J_signs,
      wb$r, K, p_expansions)

    if (!wb_sqrt$valid) {
      fell_back_to_dense <- TRUE
      break
    }

    ## ##  7. Compute GEE score per partition ##
    score_V <- .compute_score_V_partitioned(
      X, X_block, y, result, K, p_expansions,
      family, W, Delta_V, observation_weights)

    ## ##  8. Form GhalfXy_V = G_on^{1/2} score_V ##
    GhalfXy_V <- cbind(unlist(
      matmult_block_diagonal(wb$Ghalf_corrected, score_V, K,
                             parallel_matmult, cl,
                             chunk_size, num_chunks, rem_chunks)))

    ## ##  9. Constrained step via Lagrangian projection ##
    result_new <- .lagrangian_project_woodbury(
      GhalfXy_V, wb$Ghalf_corrected,
      A, K, p_expansions, R_constraints,
      constraint_value_vectors, family,
      wb_sqrt,
      parallel_aga, parallel_matmult,
      cl, chunk_size, num_chunks, rem_chunks)

    beta_new <- cbind(unlist(result_new))

    ## ##  10. Damped update and deviance check ##
    #  Logic mirrors .get_B_gee_glm exactly.
    if (!iterate & master_cnt > 1) {
      ## Non-iterative mode: accept step and exit.
      beta_block <- beta_new
      damp_cnt <- 11
      master_cnt <- 101
      err <- tol - 1
    } else {
      beta_new <- (1 - damp) * beta_block + damp * beta_new
      XB <- X_block %**% beta_new

      ## Deviance computation (original scale, not whitened).
      mu_new <- family$linkinv(c(XB))
      if (!is.null(family$custom_dev.resids) &
          is.null(family$dev.resids)) {
        err_new <- mean(
          family$custom_dev.resids(y_block,
                                   mu_new,
                                   unlist(order_list),
                                   family,
                                   unlist(observation_weights),
                                   ...))
      } else if (is.null(family$dev.resids)) {
        err_new <- mean(unlist(observation_weights) *
                          (y_block - mu_new)^2)
      } else {
        err_new <-
          mean(unlist(observation_weights) *
                 family$dev.resids(y_block,
                                   cbind(mu_new),
                                   wt = 1))
      }

      ## Step acceptance / damping logic.
      if (is.null(err_new) | is.na(err_new) | !is.finite(err_new)) {
        damp_cnt <- damp_cnt + 1
      } else if (err_new <= err) {
        prev_err <- err
        err <- err_new
        abs_diff <- max(abs(beta_new - beta_block))
        beta_block <- beta_new
        damp_cnt <- 0

        ## Early exit on convergence (after burn-in of 10 iterations).
        if ((abs_diff < tol) &
            (prev_err - err < tol) &
            (master_cnt > 10)) {
          damp_cnt <- 11
          master_cnt <- 101
          err <- tol - 1
        }

      } else {
        damp_cnt <- damp_cnt + 1
      }
    }

    ## Unpack into per-partition list.
    result <- lapply(1:(K + 1), function(k) {
      cbind(beta_block[(k - 1) * p_expansions + 1:p_expansions])
    })
  }

  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  ## Fallback to dense Path 1b.
  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  if (fell_back_to_dense) {
    X_tilde <- VhalfInv_perm %**% X_block
    y_tilde <- VhalfInv_perm %**% y_block
    return(.get_B_gee_glm(
      X_block, X_tilde, y_block, y_tilde,
      VhalfInv_perm, Lambda_block, A, K,
      p_expansions, constraint_value_vectors,
      family, return_G_getB, iterate, tol,
      qp_Amat, qp_bvec, qp_meq,
      qp_score_function,
      order_list, observation_weights,
      glm_weight_function, schur_correction_function,
      need_dispersion_for_estimation,
      dispersion_function, VhalfInv, ...))
  }

  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  ## Return.
  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  if (return_G_getB) {
    G_list <- list(
      G = lapply(wb$Ghalf_corrected, function(m) tcrossprod(m)),
      Ghalf = wb$Ghalf_corrected,
      GhalfInv = wb$GhalfInv_corrected
    )
    return(list(B = result, G_list = G_list, qp_info = qp_info))
  } else {
    return(list(B = result, qp_info = qp_info))
  }
}


## Sub-function: .get_B_gaussian_nocorr

#' Path 2: Gaussian Identity Link, No Correlation
#'
#' For the canonical Gaussian case without correlation structures, the
#' unconstrained estimate has the closed form
#' \eqn{\hat{\boldsymbol{\beta}}_k = \mathbf{G}_k \mathbf{X}_k^{\top}
#' \mathbf{y}_k} and the constrained estimate follows by a single
#' Lagrangian projection. No iteration is needed.
#'
#' When \eqn{K = 0} and there are no inequality constraints or nonzero
#' constraint values, the function returns
#' \eqn{\hat{\boldsymbol{\beta}} = \mathbf{G}\mathbf{X}^{\top}\mathbf{y}}
#' directly without forming the full \eqn{P}-dimensional OLS system.
#'
#' @param Xy List of cross-products \eqn{\mathbf{X}_k^{\top}\mathbf{y}_k}
#'   by partition.
#' @param Ghalf,GhalfInv Lists of \eqn{\mathbf{G}^{1/2}_k} and
#'   \eqn{\mathbf{G}^{-1/2}_k} matrices by partition.
#' @param A Constraint matrix \eqn{\mathbf{A}}.
#' @param K,p_expansions,R_constraints Integer dimensions.
#' @param constraint_value_vectors Constraint RHS list encoding
#'   \eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{c}}.
#' @param family GLM family object.
#' @param return_G_getB Logical; return covariance components.
#' @param quadprog Logical; apply QP refinement for inequality constraints.
#' @param qp_Amat,qp_bvec,qp_meq QP constraint specification.
#' @param qp_score_function Score function for QP step.
#' @param parallel_aga,parallel_matmult Logical flags.
#' @param cl,chunk_size,num_chunks,rem_chunks Parallel parameters.
#' @param X,y,Lambda,order_list,observation_weights Standard arguments
#'   passed through for optional QP refinement.
#' @param iterate,tol,glm_weight_function,schur_correction_function,
#'   need_dispersion_for_estimation,dispersion_function,
#'   unique_penalty_per_partition,L_partition_list,VhalfInv,
#'   homogenous_weights Standard arguments passed through.
#' @param ... Passed to sub-functions.
#'
#' @return Same structure as \code{get_B}.
#'
#' @keywords internal
.get_B_gaussian_nocorr <- function(Xy, Ghalf, GhalfInv, A, K,
                                   p_expansions, R_constraints,
                                   constraint_value_vectors, family,
                                   return_G_getB,
                                   quadprog, qp_Amat, qp_bvec, qp_meq,
                                   qp_score_function,
                                   parallel_aga, parallel_matmult,
                                   cl, chunk_size, num_chunks, rem_chunks,
                                   X, y, Lambda,
                                   order_list, observation_weights,
                                   iterate, tol,
                                   glm_weight_function,
                                   schur_correction_function,
                                   need_dispersion_for_estimation,
                                   dispersion_function,
                                   unique_penalty_per_partition,
                                   L_partition_list, VhalfInv,
                                   homogenous_weights, ...) {

  ## Fast path: K = 0, no inequality constraints, homogeneous RHS.
  #  Direct solution: beta = G X^T y (one partition, no projection needed).
  if (K == 0 & !quadprog & length(constraint_value_vectors) == 0) {
    G <- list(tcrossprod(Ghalf[[1]]))
    result <- list(G[[1]] %**% Xy[[1]])
    if (return_G_getB) {
      Ghalf_exact <- lapply(G, function(Gk) {
        matsqrt(Gk)
      })
      GhalfInv_exact <- lapply(G, function(Gk) {
        matinvsqrt(Gk)
      })
      return(list(
        B = result,
        G_list = list(G = G,
                      Ghalf = Ghalf_exact,
                      GhalfInv = GhalfInv_exact),
        qp_info = NULL
      ))
    } else {
      return(list(B = result, qp_info = NULL))
    }
  }

  ## General path: y* = G^{1/2} X^T y, then Lagrangian projection.
  GhalfXy <- cbind(
    unlist(
      matmult_block_diagonal(Ghalf, Xy, K, parallel_matmult, cl,
                             chunk_size, num_chunks, rem_chunks)))

  result <- .lagrangian_project(GhalfXy, Ghalf, A, K, p_expansions,
                                R_constraints, constraint_value_vectors,
                                family, parallel_aga, parallel_matmult,
                                cl, chunk_size, num_chunks, rem_chunks)

  qp_info <- NULL

  ## Optional QP refinement for inequality constraints.
  if (quadprog) {
    ## Auto-detect whether partition-wise active-set is valid.
    qp_needs_global <- .detect_qp_global(qp_Amat, p_expansions, K)

    if (!qp_needs_global) {
      ## Partition-wise active-set method: avoids forming dense P x P system.
      as_out <- try(.active_set_refine(
        result, X, y, K, p_expansions,
        A, R_constraints, constraint_value_vectors,
        Lambda, Ghalf, GhalfInv, family,
        qp_Amat, qp_bvec, qp_meq,
        Xy_or_uncon = Xy, is_path3 = FALSE,
        parallel_aga, parallel_matmult,
        cl, chunk_size, num_chunks, rem_chunks, tol), silent = TRUE)
      if (!inherits(as_out, "try-error") && as_out$converged) {
        result <- as_out$result
        qp_info <- as_out$qp_info
      } else {
        ## Fallback to dense SQP if active-set did not converge.
        Lambda_block <- .build_lambda_block(Lambda, K,
                                            unique_penalty_per_partition,
                                            L_partition_list)
        qp_out <- .qp_refine(result, X, y, K, p_expansions, A, Lambda,
                             Lambda_block, family, iterate, tol,
                             qp_Amat, qp_bvec, qp_meq,
                             qp_score_function,
                             order_list, glm_weight_function,
                             schur_correction_function,
                             need_dispersion_for_estimation,
                             dispersion_function, observation_weights,
                             VhalfInv, ...)
        result <- qp_out$result
        qp_info <- qp_out$qp_info
      }
    } else {
      ## Dense SQP (cross-partition constraints detected).
      Lambda_block <- .build_lambda_block(Lambda, K,
                                          unique_penalty_per_partition,
                                          L_partition_list)
      qp_out <- .qp_refine(result, X, y, K, p_expansions, A, Lambda,
                           Lambda_block, family, iterate, tol,
                           qp_Amat, qp_bvec, qp_meq,
                           qp_score_function,
                           order_list, glm_weight_function,
                           schur_correction_function,
                           need_dispersion_for_estimation,
                           dispersion_function, observation_weights,
                           VhalfInv, ...)
      result <- qp_out$result
      qp_info <- qp_out$qp_info
    }
  }

  if (return_G_getB) {
    G_exact <- lapply(Ghalf, function(mat) tcrossprod(mat))
    return(list(
      B = result,
      G_list = list(G = G_exact,
                    Ghalf = Ghalf,
                    GhalfInv = GhalfInv),
      qp_info = qp_info
    ))
  } else {
    return(list(B = result, qp_info = qp_info))
  }
}


## Sub-function: .get_B_glm_nocorr

#' Path 3: Non-Gaussian GLM, No Correlation
#'
#' For GLMs with non-identity links or non-Gaussian families (without
#' correlation structures), unconstrained partition-wise estimates are
#' first obtained via Newton-Raphson (or OLS for Gaussian identity), then
#' the constrained estimate is computed by a single Lagrangian projection.
#' For non-canonical links, \eqn{\mathbf{G}} depends on the current
#' fitted values through the GLM working weights \eqn{\mathbf{W}}; the
#' projection is therefore iterated, updating \eqn{\mathbf{G}} at the
#' current constrained estimate until convergence.
#'
#' @param X,y Lists of partition-specific design matrices and responses.
#' @param X_gram List of Gram matrices.
#' @param Xy List of cross-products.
#' @param Lambda Shared penalty matrix.
#' @param Ghalf,GhalfInv Lists of matrix square roots by partition.
#' @param A Constraint matrix.
#' @param K,p_expansions,R_constraints Integer dimensions.
#' @param constraint_value_vectors Constraint RHS list encoding
#'   \eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{c}}.
#' @param family GLM family object.
#' @param return_G_getB Logical; return covariance components.
#' @param iterate Logical; iterate for non-canonical links.
#' @param tol Convergence tolerance.
#' @param quadprog Logical; apply QP refinement.
#' @param qp_Amat,qp_bvec,qp_meq QP constraint specification.
#' @param qp_score_function Score function for QP step.
#' @param unconstrained_fit_fxn Function for partition-wise unconstrained
#'   estimation.
#' @param keep_weighted_Lambda,unique_penalty_per_partition Logical flags.
#' @param L_partition_list Partition-specific penalty matrices.
#' @param parallel_eigen,parallel_aga,parallel_matmult,
#'   parallel_unconstrained Logical flags.
#' @param cl,chunk_size,num_chunks,rem_chunks Parallel parameters.
#' @param order_list,observation_weights Standard partition arguments.
#' @param glm_weight_function,schur_correction_function,
#'   need_dispersion_for_estimation,dispersion_function GLM customization.
#' @param VhalfInv Inverse square root correlation (unused here).
#' @param homogenous_weights Logical.
#' @param ... Passed to fitting, weight, and score functions.
#'
#' @return Same structure as \code{get_B}.
#'
#' @keywords internal
.get_B_glm_nocorr <- function(X, y, X_gram, Xy, Lambda, Ghalf, GhalfInv,
                              A, K, p_expansions, R_constraints,
                              constraint_value_vectors, family,
                              return_G_getB, iterate, tol,
                              quadprog, qp_Amat, qp_bvec, qp_meq,
                              qp_score_function,
                              unconstrained_fit_fxn,
                              keep_weighted_Lambda,
                              unique_penalty_per_partition,
                              L_partition_list,
                              parallel_eigen, parallel_aga,
                              parallel_matmult, parallel_unconstrained,
                              cl, chunk_size, num_chunks, rem_chunks,
                              order_list, observation_weights,
                              glm_weight_function,
                              schur_correction_function,
                              need_dispersion_for_estimation,
                              dispersion_function, VhalfInv,
                              homogenous_weights, ...) {

  ## Lambda^{1/2} for augmented regression in unconstrained_fit_fxn.
  LambdaHalf <- matsqrt(Lambda)

  ## Obtain unconstrained estimates partition-wise.
  if (parallel_unconstrained) {
    unconstrained_estimate <-
      parallel::parLapply(cl, 1:(K + 1),
                          function(k,
                                   unique_penalty_per_partition,
                                   unconstrained_fit_fxn,
                                   keep_weighted_Lambda,
                                   family, tol, K,
                                   parallel_unconstrained, cl,
                                   chunk_size, num_chunks, rem_chunks,
                                   observation_weights) {
                            if (unique_penalty_per_partition) {
                              Lambda_temp <- Lambda + L_partition_list[[k]]
                              LambdaHalf_temp <- matsqrt(Lambda_temp)
                            } else {
                              LambdaHalf_temp <- LambdaHalf
                              Lambda_temp <- Lambda
                            }
                            cbind(c(unconstrained_fit_fxn(
                              X[[k]], y[[k]], LambdaHalf_temp, Lambda_temp,
                              keep_weighted_Lambda, family, tol, K,
                              parallel_unconstrained, cl,
                              chunk_size, num_chunks, rem_chunks,
                              order_list[[k]], observation_weights[[k]],
                              ...)))
                          },
                          unique_penalty_per_partition,
                          unconstrained_fit_fxn,
                          keep_weighted_Lambda,
                          family, tol, K,
                          parallel_unconstrained, cl,
                          chunk_size, num_chunks, rem_chunks,
                          observation_weights)
  } else {
    unconstrained_estimate <- lapply(1:(K + 1), function(k) {
      if (unique_penalty_per_partition) {
        Lambda_temp <- Lambda + L_partition_list[[k]]
        LambdaHalf_temp <- matsqrt(Lambda_temp)
      } else {
        LambdaHalf_temp <- LambdaHalf
        Lambda_temp <- Lambda
      }
      cbind(c(unconstrained_fit_fxn(
        X[[k]], y[[k]], LambdaHalf_temp, Lambda_temp,
        keep_weighted_Lambda, family, tol, K,
        parallel_unconstrained, cl,
        chunk_size, num_chunks, rem_chunks,
        order_list[[k]], observation_weights[[k]],
        ...)))
    })
  }

  ## Fast path: K = 0, no inequality constraints, homogeneous RHS.
  if (K == 0 & !quadprog & length(constraint_value_vectors) == 0) {
    if (return_G_getB) {
      G_list <- .recompute_G_at_estimate(
        X, y, unconstrained_estimate, K, Lambda, family,
        order_list, glm_weight_function, schur_correction_function,
        need_dispersion_for_estimation, dispersion_function,
        observation_weights, VhalfInv,
        parallel_eigen, parallel_matmult, cl,
        chunk_size, num_chunks, rem_chunks,
        unique_penalty_per_partition, L_partition_list, ...
      )
      return(list(B = unconstrained_estimate, G_list = G_list,
                  qp_info = NULL))
    } else {
      return(list(B = unconstrained_estimate, qp_info = NULL))
    }
  }

  ## Transform: y* = G^{-1/2} beta_hat.
  GhalfXy <- cbind(
    unlist(
      matmult_block_diagonal(
        GhalfInv, unconstrained_estimate, K, parallel_matmult,
        cl, chunk_size, num_chunks, rem_chunks)))

  ## Initial Lagrangian projection.
  result <- .lagrangian_project(GhalfXy, Ghalf, A, K, p_expansions,
                                R_constraints, constraint_value_vectors,
                                family, parallel_aga, parallel_matmult,
                                cl, chunk_size, num_chunks, rem_chunks)

  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  ## Iterative projection for non-canonical links.
  ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
  if (iterate) {
    prevB_IRWLS <- NULL
    prev_diff_IRWLS <- Inf

    for (IRWLS_iter in 1:100) {
      if (!is.null(prevB_IRWLS)) {
        diff_IRWLS <- mean(
          unlist(
            sapply(1:(K + 1),
                   function(k) mean(
                     abs(result[[k]] - prevB_IRWLS[[k]])))))
        if (diff_IRWLS < tol) break
        if (prev_diff_IRWLS <= diff_IRWLS) {
          result <- prevB_IRWLS
          break
        }
        prev_diff_IRWLS <- diff_IRWLS
      }

      prevB_IRWLS <- result

      if (need_dispersion_for_estimation) {
        dispersion_temp <- dispersion_function(
          mu = family$linkinv(unlist(matmult_block_diagonal(
            X, result, K, parallel_matmult, cl,
            chunk_size, num_chunks, rem_chunks))),
          y = unlist(y),
          order_indices = unlist(order_list),
          family = family,
          observation_weights = unlist(observation_weights),
          VhalfInv = VhalfInv,
          ...
        )
      } else {
        dispersion_temp <- 1
      }

      Xw <- lapply(1:(K + 1), function(k) {
        if (nrow(X[[k]]) == 0) {
          return(X[[k]])
        }
        var <- glm_weight_function(
          family$linkinv(X[[k]] %**% cbind(c(result[[k]]))),
          y[[k]], order_list[[k]], family, dispersion_temp,
          observation_weights[[k]], ...
        )
        if (length(var) == 1) {
          if (c(var) == 0) {
            return(X[[k]] * 0)
          } else {
            return(X[[k]] * c(sqrt(var)))
          }
        } else {
          var <- c(sqrt(var))
        }
        X[[k]] * var
      })
      X_gram_IRWLS <- compute_gram_block_diagonal(
        Xw, parallel_matmult, cl, chunk_size, num_chunks, rem_chunks)

      schur_corrections_IRWLS <- schur_correction_function(
        X, y, result, dispersion_temp, order_list, K, family,
        observation_weights, ...
      )
      G_list_IRWLS <- compute_G_eigen(
        X_gram_IRWLS, Lambda, K, parallel_eigen, cl,
        chunk_size, num_chunks, rem_chunks, family,
        unique_penalty_per_partition, L_partition_list,
        keep_G = FALSE, schur_corrections_IRWLS)

      Ghalf <- G_list_IRWLS$Ghalf
      GhalfInv <- G_list_IRWLS$GhalfInv

      GhalfXy <- cbind(
        unlist(
          matmult_block_diagonal(
            GhalfInv, unconstrained_estimate, K, parallel_matmult,
            cl, chunk_size, num_chunks, rem_chunks)))

      result <- .lagrangian_project(GhalfXy, Ghalf, A, K, p_expansions,
                                    R_constraints, constraint_value_vectors,
                                    family, parallel_aga, parallel_matmult,
                                    cl, chunk_size, num_chunks, rem_chunks)
    }
  }

  qp_info <- NULL

  ## Optional QP refinement for inequality constraints.
  if (quadprog) {
    ## Auto-detect whether partition-wise active-set is valid.
    qp_needs_global <- .detect_qp_global(qp_Amat, p_expansions, K)

    if (!qp_needs_global) {
      ## Partition-wise active-set method.
      #  For Path 3, pass unconstrained estimates; GhalfInv forms y*.
      as_out <- try(.active_set_refine(
        result, X, y, K, p_expansions,
        A, R_constraints, constraint_value_vectors,
        Lambda, Ghalf, GhalfInv, family,
        qp_Amat, qp_bvec, qp_meq,
        Xy_or_uncon = unconstrained_estimate, is_path3 = TRUE,
        parallel_aga, parallel_matmult,
        cl, chunk_size, num_chunks, rem_chunks, tol), silent = TRUE)
      if (!inherits(as_out, "try-error") && as_out$converged) {
        result <- as_out$result
        qp_info <- as_out$qp_info
      } else {
        ## Fallback to dense SQP.
        Lambda_block <- .build_lambda_block(Lambda, K,
                                            unique_penalty_per_partition,
                                            L_partition_list)
        qp_out <- .qp_refine(result, X, y, K, p_expansions, A, Lambda,
                             Lambda_block, family, iterate, tol,
                             qp_Amat, qp_bvec, qp_meq,
                             qp_score_function,
                             order_list, glm_weight_function,
                             schur_correction_function,
                             need_dispersion_for_estimation,
                             dispersion_function, observation_weights,
                             VhalfInv, ...)
        result <- qp_out$result
        qp_info <- qp_out$qp_info
      }
    } else {
      ## Dense SQP (cross-partition constraints detected).
      Lambda_block <- .build_lambda_block(Lambda, K,
                                          unique_penalty_per_partition,
                                          L_partition_list)
      qp_out <- .qp_refine(result, X, y, K, p_expansions, A, Lambda,
                           Lambda_block, family, iterate, tol,
                           qp_Amat, qp_bvec, qp_meq,
                           qp_score_function,
                           order_list, glm_weight_function,
                           schur_correction_function,
                           need_dispersion_for_estimation,
                           dispersion_function, observation_weights,
                           VhalfInv, ...)
      result <- qp_out$result
      qp_info <- qp_out$qp_info
    }
  }

  if (return_G_getB) {
    if (paste0(family)[1] == 'gaussian' &
        paste0(family)[2] == 'identity') {
      G_exact <- lapply(Ghalf, function(mat) tcrossprod(mat))
      return(list(
        B = result,
        G_list = list(G = G_exact,
                      Ghalf = Ghalf,
                      GhalfInv = GhalfInv),
        qp_info = qp_info
      ))
    }
    G_list <- .recompute_G_at_estimate(
      X, y, result, K, Lambda, family,
      order_list, glm_weight_function, schur_correction_function,
      need_dispersion_for_estimation, dispersion_function,
      observation_weights, VhalfInv,
      parallel_eigen, parallel_matmult, cl,
      chunk_size, num_chunks, rem_chunks,
      unique_penalty_per_partition, L_partition_list, ...
    )
    return(list(B = result, G_list = G_list, qp_info = qp_info))
  } else {
    return(list(B = result, qp_info = qp_info))
  }
}


## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
## Main function: get_B
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

#' Compute Constrained GLM Coefficient Estimates via Lagrangian Multipliers
#'
#' @description
#' Core estimation function for Lagrangian multiplier smoothing splines.
#' Computes penalized coefficient estimates subject to smoothness constraints
#' (continuity and derivative matching at knots) and optional user-supplied
#' linear equality or inequality constraints. Dispatches to one of three
#' computational paths depending on the model structure:
#'
#' \bold{Path 1. GEE (correlation structure present):}
#' When both \code{Vhalf} and \code{VhalfInv} are provided, the full
#' \eqn{N \times P} whitened design
#' \eqn{\mathbf{X}^{*} = \mathbf{V}_{\mathrm{perm}}^{-1/2}\mathbf{X}} is used
#' after permuting observations to partition ordering via \code{order_list}.
#' For structured correlation, the solver tries the supplement's
#' decomposition
#' \eqn{\mathbf{G}^{-1} =
#' \mathbf{G}_{\mathrm{on}}^{-1} + \mathbf{G}_{\mathrm{off}}^{-1}}
#' with
#' \eqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{E}\mathbf{J}\mathbf{E}^{\top}}.
#' This gives a Woodbury path with cost \eqn{O(Kp^3 + Pr^2)} when the
#' cross-partition rank \eqn{r} is small; otherwise the dense
#' \eqn{O(P^3)} path is used.
#' Two sub-paths:
#' \itemize{
#'   \item \bold{Path 1a.} Gaussian identity + GEE.
#'     See \code{\link{.get_B_gee_gaussian}} (dense) and
#'     \code{\link{.get_B_gee_woodbury}} (accelerated).
#'   \item \bold{Path 1b.} Non-Gaussian GEE.
#'     See \code{\link{.get_B_gee_glm}} (dense) and
#'     \code{\link{.get_B_gee_glm_woodbury}} (accelerated).
#' }
#'
#' \bold{Path 2. Gaussian identity link, no correlation:}
#' Unconstrained estimate \eqn{\hat{\boldsymbol{\beta}}_k =
#' \mathbf{G}_k\mathbf{X}_k^{\top}\mathbf{y}_k} per partition, then a
#' single Lagrangian projection. See \code{\link{.get_B_gaussian_nocorr}}.
#'
#' \bold{Path 3. Non-Gaussian GLM, no correlation:}
#' Partition-wise unconstrained estimates via Newton-Raphson, then
#' Lagrangian projection. See \code{\link{.get_B_glm_nocorr}}.
#'
#' \bold{Inequality constraint handling:}
#' When inequality constraints \eqn{\mathbf{C}^{\top}\boldsymbol{\beta}
#' \succeq \mathbf{c}} are present, the sparsity pattern of
#' \eqn{\mathbf{C}} is inspected automatically. If every constraint
#' column has nonzeros in only a single partition block (block-separable),
#' a partition-wise active-set method is used at cost
#' \eqn{O(Kp^3)} per iteration. If any constraint spans multiple
#' partition blocks (e.g.\ cross-knot monotonicity), the dense
#' \code{quadprog::solve.QP} SQP fallback is invoked.
#'
#' @param X List of length \eqn{K+1}; partition-specific design matrices.
#' @param X_gram List of Gram matrices by partition.
#' @param Lambda Combined penalty matrix.
#' @param keep_weighted_Lambda Logical.
#' @param unique_penalty_per_partition Logical.
#' @param L_partition_list List of partition-specific penalty matrices.
#' @param A Constraint matrix \eqn{\mathbf{A}} (\eqn{P \times R}).
#' @param Xy List of cross-products by partition.
#' @param y List of response vectors by partition.
#' @param K Integer; number of interior knots.
#' @param p_expansions Integer; basis terms per partition.
#' @param R_constraints Integer; columns of \eqn{\mathbf{A}}.
#' @param Ghalf List of \eqn{\mathbf{G}^{1/2}_k} matrices.
#' @param GhalfInv List of \eqn{\mathbf{G}^{-1/2}_k} matrices.
#' @param parallel_eigen,parallel_aga,parallel_matmult,parallel_unconstrained
#'   Logical flags.
#' @param cl Cluster object.
#' @param chunk_size,num_chunks,rem_chunks Parallel distribution parameters.
#' @param family GLM family object.
#' @param unconstrained_fit_fxn Partition-wise unconstrained estimator.
#' @param iterate Logical; iterate for non-canonical links.
#' @param qp_score_function Score function for QP steps.
#' @param quadprog Logical; use \code{quadprog::solve.QP} for inequality
#'   constraints.
#' @param qp_Amat,qp_bvec,qp_meq Inequality constraint specification
#'   \eqn{\mathbf{C}^{\top}\boldsymbol{\beta} \succeq \mathbf{c}}.
#' @param prevB,prevUnconB,iter_count,prev_diff Retired; ignored.
#' @param tol Convergence tolerance.
#' @param constraint_value_vectors List encoding nonzero RHS
#'   \eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{c}}.
#' @param order_list Partition-to-data index mapping.
#' @param glm_weight_function Function computing working weights.
#' @param schur_correction_function Function computing Schur corrections.
#' @param need_dispersion_for_estimation Logical.
#' @param dispersion_function Dispersion estimator.
#' @param observation_weights List of observation weights by partition.
#' @param homogenous_weights Logical.
#' @param return_G_getB Logical; return \eqn{\mathbf{G}},
#'   \eqn{\mathbf{G}^{1/2}}, \eqn{\mathbf{G}^{-1/2}} per partition.
#' @param blockfit,just_linear_without_interactions Retired; retained for
#'   call-site compatibility.
#' @param Vhalf,VhalfInv Square root and inverse square root of the
#'   working correlation matrix in the original observation ordering.
#'   When both are non-\code{NULL}, dispatches to Path 1; \code{VhalfInv}
#'   is permuted internally to partition ordering via \code{order_list}.
#' @param ... Passed to fitting, weight, correction, and dispersion
#'   functions.
#'
#' @return
#' If \code{return_G_getB = FALSE}: a list with \code{B} (coefficient
#' column vectors by partition) and \code{qp_info}.
#'
#' If \code{return_G_getB = TRUE}: a list with elements:
#' \describe{
#'   \item{B}{List of constrained coefficient column vectors
#'     \eqn{\tilde{\boldsymbol{\beta}}_k} by partition.}
#'   \item{G_list}{List with \code{G}, \code{Ghalf}, \code{GhalfInv},
#'     each a list of \eqn{K+1} matrices.}
#'   \item{qp_info}{QP or active-set metadata, including Lagrangian
#'     multipliers and active-constraint information when available, or
#'     \code{NULL}.}
#' }
#'
#' @keywords internal
#' @export
get_B <- function(X,
                  X_gram,
                  Lambda,
                  keep_weighted_Lambda,
                  unique_penalty_per_partition,
                  L_partition_list,
                  A,
                  Xy,
                  y,
                  K,
                  p_expansions,
                  R_constraints,
                  Ghalf,
                  GhalfInv,
                  parallel_eigen,
                  parallel_aga,
                  parallel_matmult,
                  parallel_unconstrained,
                  cl,
                  chunk_size,
                  num_chunks,
                  rem_chunks,
                  family,
                  unconstrained_fit_fxn,
                  iterate,
                  qp_score_function,
                  quadprog,
                  qp_Amat,
                  qp_bvec,
                  qp_meq,
                  prevB = NULL,
                  prevUnconB = NULL,
                  iter_count = 0,
                  prev_diff = Inf,
                  tol,
                  constraint_value_vectors,
                  order_list,
                  glm_weight_function,
                  schur_correction_function,
                  need_dispersion_for_estimation,
                  dispersion_function,
                  observation_weights,
                  homogenous_weights,
                  return_G_getB,
                  blockfit,
                  just_linear_without_interactions,
                  Vhalf,
                  VhalfInv,
                  ...) {

  ## Path 1: GEE (correlation structure present)
  if (!is.null(Vhalf) & !is.null(VhalfInv)) {

    ## Permute correlation matrices so rows/columns align with partition
    #  ordering rather than the original data ordering.
    perm <- unlist(order_list)
    Vhalf_perm    <- Vhalf[perm, perm]
    VhalfInv_perm <- VhalfInv[perm, perm]

    ## Default to unit observation weights when none were supplied.
    if (is.null(observation_weights) |
        length(unlist(observation_weights)) == 0) {
      observation_weights <- 1
    }

    ## Assemble the full N x P block-diagonal design.
    X_block <- collapse_block_diagonal(X)
    y_block <- cbind(unlist(y))

    ## Whiten: X_tilde = V^{-1/2} X, y_tilde = V^{-1/2} y.
    X_tilde <- VhalfInv_perm %**% X_block
    y_tilde <- VhalfInv_perm %**% y_block

    ## Full P x P block-diagonal penalty matrix.
    Lambda_block <- .build_lambda_block(Lambda, K,
                                        unique_penalty_per_partition,
                                        L_partition_list)

    ## Split the GEE path into the Gaussian one-step solve versus
    #  the iterative non-Gaussian path.
    is_gauss_id <- (paste0(family)[1] == 'gaussian' &
                      paste0(family)[2] == 'identity')

    ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##--
    ## Attempt Woodbury decomposition for structured V.
    ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##--
    wb_decomp <- .woodbury_decompose_V(
      VhalfInv_perm, X, K, p_expansions,
      Lambda, Lambda_block,
      unique_penalty_per_partition, L_partition_list,
      order_list,
      parallel_eigen, cl,
      chunk_size, num_chunks, rem_chunks,
      rank_threshold_fraction = 1/3,
      family)

    if (is_gauss_id) {
      obs_wt_full <- c(unlist(observation_weights))
      if (length(obs_wt_full) == 0L) {
        obs_wt_full <- rep(1, nrow(VhalfInv_perm))
      } else if (length(obs_wt_full) == 1L) {
        obs_wt_full <- rep(obs_wt_full, nrow(VhalfInv_perm))
      }
      can_use_gauss_woodbury <- all(abs(obs_wt_full - 1) <
                                      sqrt(.Machine$double.eps))

      if (wb_decomp$use_woodbury && can_use_gauss_woodbury) {
        ## Compute F^{1/2} components for the Woodbury path.
        wb_sqrt <- .woodbury_halfsqrt_components(
          wb_decomp$Ghalf_corrected, wb_decomp$E, wb_decomp$J_signs,
          wb_decomp$r, K, p_expansions)

        if (wb_sqrt$valid) {
          ## Path 1a-Woodbury: Gaussian identity + GEE, efficient.
          out <- .get_B_gee_woodbury(
            X, y, K, p_expansions,
            VhalfInv_perm, order_list,
            A, R_constraints,
            constraint_value_vectors, family,
            return_G_getB,
            quadprog, qp_Amat, qp_bvec, qp_meq,
            qp_score_function,
            observation_weights,
            wb_decomp, wb_sqrt,
            parallel_aga, parallel_matmult,
            cl, chunk_size, num_chunks, rem_chunks,
            qp_global = TRUE, ...)
        } else {
          ## F not positive definite; fall back to dense.
          out <- .get_B_gee_gaussian(
            X_block, X_tilde, y_tilde, VhalfInv_perm,
            Lambda_block, A, K, p_expansions,
            constraint_value_vectors, family,
            return_G_getB,
            quadprog, qp_Amat, qp_bvec, qp_meq,
            qp_score_function,
            order_list, observation_weights, ...)
        }
      } else {
        ## Otherwise use the dense Gaussian GEE path.
        out <- .get_B_gee_gaussian(
          X_block, X_tilde, y_tilde, VhalfInv_perm,
          Lambda_block, A, K, p_expansions,
          constraint_value_vectors, family,
          return_G_getB,
          quadprog, qp_Amat, qp_bvec, qp_meq,
          qp_score_function,
          order_list, observation_weights, ...)
      }
    } else {
      if (wb_decomp$use_woodbury) {
        wb_sqrt <- .woodbury_halfsqrt_components(
          wb_decomp$Ghalf_corrected, wb_decomp$E, wb_decomp$J_signs,
          wb_decomp$r, K, p_expansions)

        if (wb_sqrt$valid) {
          ## Path 1b-Woodbury: non-Gaussian GEE, efficient.
          out <- .get_B_gee_glm_woodbury(
            X, y, K, p_expansions,
            VhalfInv_perm, order_list,
            A, R_constraints,
            constraint_value_vectors, family,
            return_G_getB, iterate, tol,
            qp_Amat, qp_bvec, qp_meq,
            qp_score_function,
            observation_weights,
            Lambda, Lambda_block,
            unique_penalty_per_partition, L_partition_list,
            wb_decomp, wb_sqrt,
            glm_weight_function, schur_correction_function,
            need_dispersion_for_estimation, dispersion_function,
            VhalfInv,
            parallel_eigen, parallel_aga, parallel_matmult,
            cl, chunk_size, num_chunks, rem_chunks, ...)
        } else {
          ## F not positive definite; fall back to dense.
          out <- .get_B_gee_glm(
            X_block, X_tilde, y_block, y_tilde,
            VhalfInv_perm, Lambda_block, A, K, p_expansions,
            constraint_value_vectors, family,
            return_G_getB, iterate, tol,
            qp_Amat, qp_bvec, qp_meq,
            qp_score_function,
            order_list, observation_weights,
            glm_weight_function, schur_correction_function,
            need_dispersion_for_estimation, dispersion_function,
            VhalfInv, ...)
        }
      } else {
        ## Otherwise use the dense non-Gaussian GEE path.
        out <- .get_B_gee_glm(
          X_block, X_tilde, y_block, y_tilde,
          VhalfInv_perm, Lambda_block, A, K, p_expansions,
          constraint_value_vectors, family,
          return_G_getB, iterate, tol,
          qp_Amat, qp_bvec, qp_meq,
          qp_score_function,
          order_list, observation_weights,
          glm_weight_function, schur_correction_function,
          need_dispersion_for_estimation, dispersion_function,
          VhalfInv, ...)
      }
    }

    ## Normalize return format.
    if (return_G_getB) {
      return(list(B = out$B, G_list = out$G_list, qp_info = out$qp_info))
    } else {
      return(out$B)
    }
  }

  ## Paths 2 and 3: no correlation structure

  ## Without correlation, the family decides between the Gaussian
  #  closed-form path and the iterative GLM path.
  is_gauss_id <- (paste0(family)[2] == 'identity') &
    (paste0(family)[1] == 'gaussian')

  if (is_gauss_id) {
    ## Path 2: Gaussian identity, no correlation.
    out <- .get_B_gaussian_nocorr(
      Xy, Ghalf, GhalfInv, A, K, p_expansions, R_constraints,
      constraint_value_vectors, family,
      return_G_getB,
      quadprog, qp_Amat, qp_bvec, qp_meq,
      qp_score_function,
      parallel_aga, parallel_matmult,
      cl, chunk_size, num_chunks, rem_chunks,
      X, y, Lambda,
      order_list, observation_weights,
      iterate, tol,
      glm_weight_function, schur_correction_function,
      need_dispersion_for_estimation, dispersion_function,
      unique_penalty_per_partition, L_partition_list,
      VhalfInv, homogenous_weights, ...
    )
  } else {
    ## Path 3: Non-Gaussian GLM, no correlation.
    out <- .get_B_glm_nocorr(
      X, y, X_gram, Xy, Lambda, Ghalf, GhalfInv,
      A, K, p_expansions, R_constraints,
      constraint_value_vectors, family,
      return_G_getB, iterate, tol,
      quadprog, qp_Amat, qp_bvec, qp_meq,
      qp_score_function,
      unconstrained_fit_fxn,
      keep_weighted_Lambda,
      unique_penalty_per_partition, L_partition_list,
      parallel_eigen, parallel_aga, parallel_matmult,
      parallel_unconstrained,
      cl, chunk_size, num_chunks, rem_chunks,
      order_list, observation_weights,
      glm_weight_function, schur_correction_function,
      need_dispersion_for_estimation, dispersion_function,
      VhalfInv, homogenous_weights, ...
    )
  }

  ## Normalize return format.
  if (return_G_getB) {
    return(list(B = out$B, G_list = out$G_list, qp_info = out$qp_info))
  } else {
    return(out$B)
  }
}



#' @title Verification Examples for get_B
#'
#' @description
#' Simple, self-contained examples that reviewers can run to verify that
#' \code{get_B} produces correct output. These exercise Path 2 (Gaussian,
#' no correlation) and Path 3 (binomial GLM).
#'
#' @examples
#' \dontrun{
#' ## Example 1: Path 2 - Gaussian identity, with knots
#' set.seed(42)
#' t <- runif(200, -5, 5)
#' y <- sin(t) + rnorm(200, 0, 0.5)
#' fit1 <- lgspline(t, y, K = 3, opt = FALSE, wiggle_penalty = 1e-4)
#' stopifnot(inherits(fit1, "lgspline"))
#' stopifnot(length(fit1$B) == 4)  # K+1 = 4 partitions
#' cat("Example 1 passed: Gaussian identity, K=3\n")
#'
#' preds1 <- predict(fit1, new_predictors = rnorm(10))
#' stopifnot(all(is.finite(preds1)))
#' cat("  Predictions finite: OK\n")
#'
#' ## Example 2: Path 2 - Gaussian identity, K=0 (no constraints)
#' fit2 <- lgspline(t, y, K = 0, opt = FALSE, wiggle_penalty = 1e-4)
#' stopifnot(inherits(fit2, "lgspline"))
#' stopifnot(length(fit2$B) == 1)
#' preds2 <- predict(fit2, new_predictors = rnorm(10))
#' stopifnot(all(is.finite(preds2)))
#' cat("Example 2 passed: Gaussian identity, K=0\n")
#'
#' ## Example 3: Path 3 - Binomial GLM
#' y_bin <- rbinom(200, 1, plogis(sin(t)))
#' fit3 <- lgspline(t, y_bin, K = 2, family = binomial(),
#'                  opt = FALSE, wiggle_penalty = 1e-3)
#' stopifnot(inherits(fit3, "lgspline"))
#' preds3 <- predict(fit3, new_predictors = rnorm(10))
#' stopifnot(all(preds3 >= 0 & preds3 <= 1))
#' cat("Example 3 passed: Binomial GLM, K=2\n")
#'
#' ## Example 4: Path 2 with QP constraints (monotonic increase)
#' t_sorted <- sort(runif(100, -3, 3))
#' y_mono <- t_sorted + 0.5 * sin(t_sorted) + rnorm(100, 0, 0.3)
#' fit4 <- lgspline(t_sorted, y_mono, K = 2,
#'                  qp_monotonic_increase = TRUE,
#'                  opt = FALSE, wiggle_penalty = 1e-4)
#' preds4 <- predict(fit4, new_predictors = cbind(sort(rnorm(50))))
#' stopifnot(all(diff(preds4) >= -sqrt(.Machine$double.eps)))
#' cat("Example 4 passed: Monotonic increase QP constraint\n")
#'
#' ## Example 5: Path 2 with range constraints
#' fit5 <- lgspline(t, y, K = 3,
#'                  qp_range_lower = -2, qp_range_upper = 2,
#'                  opt = FALSE, wiggle_penalty = 1e-4)
#' stopifnot(all(fit5$ytilde >= -2 - 0.01))
#' stopifnot(all(fit5$ytilde <= 2 + 0.01))
#' cat("Example 5 passed: Range constraints\n")
#'
#' ## Example 6: Multi-predictor
#' set.seed(123)
#' x1 <- rnorm(300)
#' x2 <- rnorm(300)
#' y6 <- sin(x1) + cos(x2) + rnorm(300, 0, 0.5)
#' fit6 <- lgspline(cbind(x1, x2), y6, K = 4,
#'                  opt = FALSE, wiggle_penalty = 1e-5)
#' stopifnot(inherits(fit6, "lgspline"))
#' stopifnot(fit6$q_predictors == 2)
#' cat("Example 6 passed: 2D predictor, K=4\n")
#'
#' ## Example 7: Coefficient consistency (determinism)
#' set.seed(999)
#' t7 <- runif(150, -4, 4)
#' y7 <- 2 * cos(t7) + rnorm(150, 0, 0.4)
#' fit7a <- lgspline(t7, y7, K = 2, opt = FALSE, wiggle_penalty = 1e-5)
#' set.seed(999)
#' fit7b <- lgspline(t7, y7, K = 2, opt = FALSE, wiggle_penalty = 1e-5)
#' max_diff <- max(abs(unlist(fit7a$B) - unlist(fit7b$B)))
#' stopifnot(max_diff < 1e-12)
#' cat("Example 7 passed: Deterministic coefficient reproduction\n")
#'
#' cat("\nAll verification examples passed.\n")
#' }
#'
#' @name get_B_verification_examples
NULL

#' Get Centers for Partitioning
#'
#' @param data Matrix of predictor data (already processed: binary/excluded
#'   columns zeroed out by the caller when \code{data_already_processed = TRUE})
#' @param K Number of partitions minus 1 (\eqn{K})
#' @param cluster_args List with custom centers and kmeans args
#' @param cluster_on_indicators Include binary predictors in clustering
#' @param data_already_processed Logical; when TRUE the caller has already
#'   zeroed out binary / excluded columns, so \code{get_centers} skips its
#'   own binary-zeroing step.  Defaults to FALSE for backward compatibility.
#'
#' @details
#' Returns partition centers via:
#'
#' 1. Custom supplied centers if provided as a valid \eqn{(K+1) \times q} matrix
#'
#' 2. kmeans clustering on all non-spline variables if \code{cluster_on_indicators=TRUE}
#'    or if \code{data_already_processed=TRUE}
#'
#' 3. kmeans clustering excluding binary variables if \code{cluster_on_indicators=FALSE}
#'    and \code{data_already_processed=FALSE}
#'
#' @return Matrix of cluster centers
#'
#' @keywords internal
#' @export
get_centers <- function(data, K, cluster_args, cluster_on_indicators,
                        data_already_processed = FALSE) {

  ## If custom centers isn't null, return them
  #  These will be supplied to the first position
  #  Defaults to NA
  if(any(!is.na(cluster_args[[1]]))){
    return(cluster_args[[1]])

    ## Partition clusters including 0/1 predictors,
    #  or when the caller has pre-processed the data
    #  suppressWarnings prevents irrelevant "did not converge in 10 iterations"
  } else if(cluster_on_indicators || data_already_processed){
    km <- suppressWarnings(do.call(stats::kmeans,
                  c(list(x = data, centers = K + 1),
                    as.list(cluster_args[-1]))))

  } else {
    bin <- which(apply(data, 2, is_binary))
    data[,bin] <- 0
    km <- suppressWarnings(do.call(stats::kmeans,
                  c(list(x = data, centers = K + 1),
                    as.list(cluster_args[-1]))))
  }
  return(km$centers)
}

#' Find Neighboring Cluster Partitions Using Midpoint Distance Criterion
#'
#' @description
#' Identifies neighboring partitions by evaluating whether the midpoint between
#' cluster centers is closer to those centers than to any other center.
#'
#' @param centers Matrix; rows are cluster center coordinates
#' @param parallel Logical; use parallel processing
#' @param cl Cluster object for parallel execution
#' @param neighbor_tolerance Numeric; scaling factor for distance comparisons
#'
#' @return List where element i contains indices of centers neighboring center i
#'
#' @keywords internal
#' @export
find_neighbors <- function(centers, parallel, cl, neighbor_tolerance) {
  num_centers <- nrow(centers)
  neighbors <- vector("list", num_centers)

  if (!is.null(cl) & parallel) {
    ## Create index pairs for all comparisons
    pairs <- expand.grid(i = 1:(num_centers-1), j = 2:num_centers)
    pairs <- pairs[pairs$j > pairs$i, ]

    ## Split into chunks
    chunk_size <- max(1, ceiling(nrow(pairs) / length(cl)))
    chunks <- split(1:nrow(pairs), ceiling(seq_along(1:nrow(pairs))/chunk_size))

    ## Export data to cluster
    parallel::clusterExport(cl, "centers", envir = environment())

    ## Process chunks in parallel
    results <- parallel::parLapply(cl, chunks, function(chunk_indices) {
      chunk_results <- vector("list", length(chunk_indices))

      for (idx in seq_along(chunk_indices)) {
        pair_idx <- chunk_indices[idx]
        i <- pairs$i[pair_idx]
        j <- pairs$j[pair_idx]

        ## Check if midpoint is neighbor
        mid <- (centers[i,] + centers[j,])/2
        dist_ij <- sum((mid - centers[i,])^2) / neighbor_tolerance

        distances_to_others <- apply(centers[-c(i,j),,drop=FALSE], 1,
                                     function(k)sum((mid - k)^2))

        if(all(dist_ij < distances_to_others)) {
          chunk_results[[idx]] <- c(i, j)
        }
      }

      return(do.call(rbind, chunk_results[!sapply(chunk_results, is.null)]))
    })

    ## Combine results
    neighbor_pairs <- do.call(rbind, results)

    ## Convert to neighbor list
    if (!is.null(neighbor_pairs)) {
      for (k in 1:nrow(neighbor_pairs)) {
        i <- neighbor_pairs[k, 1]
        j <- neighbor_pairs[k, 2]
        neighbors[[i]] <- c(neighbors[[i]], j)
        neighbors[[j]] <- c(neighbors[[j]], i)
      }
    }

  } else {
    ## Sequential neighbor computation
    for (i in 1:(num_centers-1)) {
      for (j in (i+1):num_centers) {
        ## Get midpoint between centers i and j
        mid <- (centers[i,] + centers[j,])/2

        ## Compute scaled distance from midpoint to center i
        dist_ij <- sum((mid - centers[i,])^2) / neighbor_tolerance

        ## Get distances from midpoint to all other centers
        distances_to_others <- apply(centers[-c(i,j),,drop=FALSE], 1,
                                     function(k)sum((mid - k)^2))

        ## If midpoint closer to i,j than others, they are neighbors
        if(all(dist_ij < distances_to_others)) {
          neighbors[[i]] <- c(neighbors[[i]], j)
          neighbors[[j]] <- c(neighbors[[j]], i)
        }
      }
    }
  }

  return(neighbors)
}


#' Create Data Partitions Using Clustering
#'
#' @description
#' Partitions data support into clusters using Voronoi-like diagrams.
#' Standardization for clustering is handled internally; centers, knots, and
#' the returned \code{assign_partition} function are all on the **raw
#' (natural) scale** of the original predictors.
#'
#' @param data Numeric matrix of predictor variables (\strong{raw scale})
#' @param cluster_args Parameters for clustering
#' @param cluster_on_indicators Logical to include binary predictors
#' @param K Number of partitions minus 1 (\eqn{K})
#' @param parallel Logical to enable parallel processing
#' @param cl Cluster object for parallel computation
#' @param do_not_cluster_on_these Columns to exclude from clustering
#' @param neighbor_tolerance Scaling factor for neighbor detection
#' @param standardize Logical; whether to standardize data internally before
#'   clustering.  Should equal \code{standardize_predictors_for_knots} from
#'   \code{lgspline.fit}.  Default \code{TRUE}.
#' @param standardize_mode Character; \code{"auto"} (default) selects
#'   \code{"minmax"} for a single effective clustering dimension and
#'   \code{"normal"} for multiple dimensions.  Can be forced to either value.
#' @param dummy_adder Small constant added to numerator during standardization
#'   to avoid exact-zero values (matches \code{lgspline.fit}'s
#'   \code{dummy_adder}).  Default \code{0}.
#' @param dummy_dividor Small constant added to denominator during
#'   standardization to avoid division by zero (matches
#'   \code{lgspline.fit}'s \code{dummy_dividor}).  Default \code{0}.
#'
#' @return
#' A list containing:
#' \describe{
#'   \item{centers}{Cluster center coordinates on the \strong{raw} scale.}
#'   \item{knots}{Knot points between centers on the \strong{raw} scale.}
#'   \item{assign_partition}{Function that accepts \strong{raw}-scale new data
#'     and returns integer-like partition assignments (0.5, 1.5, \ldots).}
#'   \item{neighbors}{List of neighboring partition indices.}
#'   \item{standardize_transf}{The forward standardization function used
#'     internally (for diagnostic use only).}
#'   \item{standardize_inv_transf}{The inverse standardization function
#'     (for diagnostic use only).}
#'   \item{centers_std}{Cluster centers on the standardized scale (for
#'     diagnostic use only).}
#' }
#'
#' @keywords internal
#' @export
make_partitions <- function(data,
                            cluster_args,
                            cluster_on_indicators,
                            K,
                            parallel,
                            cl,
                            do_not_cluster_on_these,
                            neighbor_tolerance,
                            standardize      = TRUE,
                            standardize_mode = "auto",
                            dummy_adder      = 0,
                            dummy_dividor    = 0) {

  q <- ncol(data)

  ## If K is 0, return as is
  if(K == 0){
    return(list(
      centers                = matrix()[-1,,drop=FALSE],
      knots                  = matrix()[-1,,drop=FALSE],
      assign_partition       = function(x) 1,
      neighbors              = list(),
      standardize_transf     = function(X) X,
      standardize_inv_transf = function(X) X,
      centers_std            = matrix()[-1,,drop=FALSE]
    ))
  }

  ## Internal standardization for clustering
  #  Cluster on a standardized copy of the data so that variables measured on
  #  very different scales do not dominate the kmeans objective.  Centers and
  #  knots are back-transformed to raw scale before returning so that
  #  downstream consumers (constraint matrix, return_list, etc.) see raw-scale
  #  coordinates without any additional transformation.
  ##
  #  standardize_mode "auto": choose "minmax" when only one effective clustering
  #  dimension remains after zeroing excluded / binary columns, otherwise
  #  "normal".  Can be forced to "minmax" or "normal" by the caller.

  if(standardize_mode == "auto"){
    ## Count columns that will actually vary during clustering
    effective_dims <- q - length(do_not_cluster_on_these)
    if(!cluster_on_indicators){
      bin_check <- which(apply(data, 2, is_binary))
      active    <- setdiff(seq_len(q), do_not_cluster_on_these)
      effective_dims <- length(setdiff(active, bin_check))
    }
    standardize_mode <- if(effective_dims <= 1L) "minmax" else "normal"
  }

  if(standardize){
    if(standardize_mode == "minmax"){
      std_mins <- apply(data, 2, min)
      std_maxs <- apply(data, 2, max)
      std_transf <- function(X){
        for(j in seq_len(ncol(X))){
          X[,j] <- (X[,j] - std_mins[j] + dummy_adder) /
            (std_maxs[j] - std_mins[j] + dummy_dividor)
        }
        X
      }
      std_inv_transf <- function(X){
        for(j in seq_len(ncol(X))){
          X[,j] <- X[,j] * (std_maxs[j] - std_mins[j] + dummy_dividor) +
            std_mins[j] - dummy_adder
        }
        X
      }
    } else {
      ## Normal (z-score) standardization
      std_means <- apply(data, 2, mean)
      std_sds   <- apply(data, 2, function(x) tryCatch(sd(x), error = function(e) 1))
      std_transf <- function(X){
        for(j in seq_len(ncol(X))){
          X[,j] <- (X[,j] - std_means[j] + dummy_adder) /
            (std_sds[j] + dummy_dividor)
        }
        X
      }
      std_inv_transf <- function(X){
        for(j in seq_len(ncol(X))){
          X[,j] <- X[,j] * (std_sds[j] + dummy_dividor) +
            std_means[j] - dummy_adder
        }
        X
      }
    }
    data_std <- std_transf(data)
  } else {
    data_std       <- data
    std_transf     <- function(X) X
    std_inv_transf <- function(X) X
  }

  ## Zero out excluded / binary columns in the standardized copy   -
  #  We operate on data_std from here so that find_neighbors and kmeans work
  #  on the same scale.  The originals (data, data_std) are not modified.
  data_cluster <- data_std

  ## [Change 2026-03-14] Detect binary columns from the TRAINING data and
  #  store the indices for reuse in the assign_partition closure. Previously
  #  is_binary was re-evaluated on new data at prediction time, which caused
  #  small prediction sets (nrow <= 2) to misidentify continuous predictors
  #  as binary, zeroing them out and producing wrong partition assignments.
  if(!cluster_on_indicators){
    train_binary_cols <- which(apply(data_cluster, 2, is_binary))
    if(length(train_binary_cols) > 0){
      data_cluster[, train_binary_cols] <- 0
    }
  } else {
    train_binary_cols <- integer(0)
  }

  if(length(do_not_cluster_on_these) > 0){
    data_cluster[,do_not_cluster_on_these] <- 0
  }

  ## Compute CVT centers on the standardized scale       -
  #  Pass data_already_processed = TRUE so get_centers does not re-zero binary
  #  columns (we already did it above on data_cluster).
  #  Custom centers from cluster_args[[1]] are returned as-is; they are
  #  assumed to be on the STANDARDIZED scale when supplied via cluster_args
  #  (same assumption as before).  If the caller supplies raw-scale custom
  # centers they should pass standardize = FALSE.
  initial_points_std <- get_centers(data_cluster, K, cluster_args,
                                    cluster_on_indicators = TRUE,
                                    data_already_processed = TRUE)

  ## Re-order cluster indices by ascending L1-norm of their standardized
  #  center. This makes cluster 1 the "leftmost/smallest" and cluster K+1
  #  the "rightmost/largest", so that knot rownames like "1_2", "2_3" reflect
  #  actual spatial adjacency.  For 1-D this gives strict left-to-right order.
  #  For multi-D it gives a consistent, reproducible numbering that is
  #  insensitive to kmeans initialisation order.
  ## Re-index partitions so numbering follows the geometry more naturally
  #  and is reproducible across kmeans initialisations.
  l1_order <- order(apply(initial_points_std,
                          1,
                          function(r) sum(abs(r))))
  ## overwrite the initial_points_std above
  initial_points_std <- initial_points_std[l1_order, , drop = FALSE]

  ## Find neighboring partitions (on standardized scale)     -
  #  find_neighbors uses Euclidean distances; working on the standardized scale
  #  keeps comparisons numerically stable and scale-invariant.
  neighbors <- find_neighbors(initial_points_std,
                              parallel,
                              cl,
                              neighbor_tolerance)

  ## Compute knots as midpoints between neighboring centers
  find_knots <- function(pts, nbrs) {
    additional_points <- list()
    for (i in seq_along(nbrs)) {
      for (j in nbrs[[i]]) {
        if(j > i) next       ## avoid double-counting
        c1  <- pts[i,]
        c2  <- pts[j,]
        mid <- (c1 + c2) / 2
        additional_points[[length(additional_points) + 1]] <- mid
        names(additional_points)[length(additional_points)] <-
          paste0(i, '_', j)
      }
    }
    if(length(additional_points) > 0){
      return(do.call(rbind, additional_points))
    } else {
      return(matrix(nrow = 0, ncol = ncol(pts)))
    }
  }

  knots_std <- find_knots(initial_points_std, neighbors)

  ## Back-transform centers and knots to raw scale       -
  initial_points_raw <- std_inv_transf(initial_points_std)
  knots_raw <- if(nrow(knots_std) > 0) std_inv_transf(knots_std) else knots_std

  ## Partition-assignment closure
  #  The returned function accepts RAW-scale new data. It standardizes
  #  internally using the same transform fitted above, then uses nearest-
  #  neighbor lookup against the standardized centers.
  #
  #  Binary column mask (train_binary_cols) and excluded
  #  column mask (do_not_cluster_on_these) are captured from the training
  #  data at partition-creation time. Previously is_binary was re-evaluated
  #  on new_data, which misclassified continuous predictors as binary when
  #  the prediction set had <= 2 rows with distinct values.
  assign_partition <- function(new_data) {

    ## Standardize new data with the same transform used for training
    new_data_std <- std_transf(new_data)

    ## Apply the TRAINING-TIME column masks: binary cols and excluded cols
    #  Do NOT re-detect binary columns from new_data
    if(length(train_binary_cols) > 0){
      new_data_std[, train_binary_cols] <- 0
    }
    if(length(do_not_cluster_on_these) > 0){
      new_data_std[, do_not_cluster_on_these] <- 0
    }

    ## Batch the nearest-neighbor lookup so large prediction sets do not
    #  allocate one full distance matrix at once.
    bytes_per_row <- 8 * nrow(initial_points_std)
    batch_size    <- floor((0.5 * 1024^3) / bytes_per_row)
    total_rows    <- nrow(new_data_std)
    assignments   <- numeric(total_rows)

    for(i in seq(1, total_rows, batch_size)) {
      nd_idx <- min(i + batch_size - 1, total_rows)
      batch  <- new_data_std[i:nd_idx, , drop = FALSE]
      nn     <- FNN::get.knnx(initial_points_std, batch, k = 1)
      assignments[i:nd_idx] <- nn$nn.index
    }

    return(assignments - 0.5)
  }

  ## Return raw-scale centers/knots together with the assignment closure
  #  and the internal standardization objects used to build them.
  rownames(initial_points_raw) <- paste0("center_",
                                         seq_len(nrow(initial_points_raw)))
  return(list(
    centers                = initial_points_raw,   # raw scale
    knots                  = knots_raw,            # raw scale
    assign_partition       = assign_partition,     # Accepts raw scale input
    neighbors              = neighbors,
    standardize_transf     = std_transf,           # For diagnostics only
    standardize_inv_transf = std_inv_transf,       # For diagnostics only
    centers_std            = initial_points_std    # Standardized, for reference
  ))
}

#' Compute Leave-One-Out Cross-Validated Predictions for
#' Gaussian Response/Identity Link under Constraint
#'
#' @description
#' Computes the leave-one-out cross-validated predictions from a model
#' fit, assuming Gaussian-distributed response with identity link.
#'
#' The LOO closed-formula for observation \eqn{i} is \eqn{\hat{y}_{(-i)} = y_i -
#' \frac{1}{1 - H_{ii}}(y_i - \hat{y}_i)} where
#' \eqn{\mathbf{H}} is the effective hat matrix under
#' smoothing constraints, adjusted for weights and correlation structure if
#' present.
#'
#' Observations with leverage at or above \code{leverage_threshold} are flagged
#' in a warning, since extreme hat values can make the shortcut numerically
#' unreliable. The default \code{leverage_threshold = 100} is intentionally
#' permissive, so users who want diagnostic warnings for large \eqn{H_{ii}}
#' should set a smaller threshold explicitly.
#'
#' For related discussion of prediction-sum-of-squares calculations under
#' linear restrictions, see Tarpey (2000), who studies the PRESS statistic for
#' restricted least squares. That setting is closely related to the
#' constraint-adjusted hat-matrix shortcut used here.
#'
#' @param model_fit A fitted lgspline model object.
#' @param leverage_threshold Numeric scalar. Observations with
#'   \eqn{H_{ii} \geq} \code{leverage_threshold} are treated as
#'   high-leverage for the warning below. Default \code{100}.
#'
#' @return A vector of leave-one-out cross-validated predictions
#'
#' @examples
#'
#' ## Basic usage with Gaussian response, computing PRESS
#' set.seed(1234)
#' t <- rnorm(50)
#' y <- sin(t) + rnorm(50, 0, .25)
#' model_fit <- lgspline(t, y)
#' loo <- leave_one_out(model_fit)
#' press <- mean((y - loo)^2, na.rm = TRUE)
#'
#' plot(loo, y,
#'    main = "LOO Cross-Validation Prediction vs. Observed Response",
#'    xlab = 'Prediction', ylab = 'Response')
#' abline(0, 1)
#'
#' @references
#' Tarpey, T. (2000). A note on the prediction sum of squares statistic for
#' restricted least squares. \emph{The American Statistician}, 54(2), 116--118.
#' \doi{10.2307/2686028}
#'
#' @export
leave_one_out <- function(model_fit,
                          leverage_threshold = 100){

  ## Full N x P design matrix on the standardized scale
  X_block <- collapse_block_diagonal(
    lapply(model_fit$X, model_fit$std_X)
  )

  if(!is.null(model_fit$VhalfInv)){

    ## Under correlation, compute the hat diagonal from the same dense
    #  GLS system used by varcovmat and trace_XUGX.
    X_ordered <- X_block[unlist(model_fit$og_order), , drop = FALSE]
    VinvhalfX <- t(t(model_fit$VhalfInv %**% X_ordered) *
                     sqrt(model_fit$weights))

    has_part_pen <-
      length(model_fit$penalties$L_partition_list) == (model_fit$K + 1)
    Lambda_full <- collapse_block_diagonal(
      lapply(1:(model_fit$K + 1), function(k){
        if(has_part_pen){
          model_fit$penalties$Lambda +
            model_fit$penalties$L_partition_list[[k]]
        } else {
          model_fit$penalties$Lambda
        }
      })
    )

    Ghalf_correct <- matinvsqrt(crossprod(VinvhalfX) + Lambda_full)

    ## hat diagonal: diag(V^{-1/2} X U G X^T V^{-1/2})
    #  = rowSums((V^{-1/2} X U G^{1/2}) * (V^{-1/2} X G^{1/2}))
    const <- max(sqrt(norm(Ghalf_correct, "2")), sqrt(.Machine$double.eps))
    VinvhalfXUGhalf <-
      VinvhalfX %**% (model_fit$U %**% (Ghalf_correct / const))
    VinvhalfXGhalf <- VinvhalfX %**% (Ghalf_correct / const)
    diag_hat <- rowSums(VinvhalfXUGhalf * VinvhalfXGhalf) * const^2

  } else {

    ## Without correlation, use the block-diagonal shortcut for the
    #  hat diagonal.
    UGhalf <- matmult_U(model_fit$U,
                        model_fit$Ghalf,
                        model_fit$p,
                        model_fit$K)
    Ghalf_block <- collapse_block_diagonal(model_fit$Ghalf)
    const <- max(sqrt(norm(UGhalf, "2")), sqrt(.Machine$double.eps))
    X_block <- t(t(X_block) * sqrt(model_fit$weights))
    XUGhalf <- X_block %**% (UGhalf / const)
    XGhalf <- X_block %**% (Ghalf_block / const)
    diag_hat <- rowSums(XUGhalf * XGhalf) * const^2
    diag_hat <- diag_hat[unlist(model_fit$og_order)]
  }

  ## Flag high-leverage observations where 1/(1 - H_ii) is unreliable
  n_high <- sum(diag_hat >= leverage_threshold, na.rm = TRUE)
  if(n_high > 0){
    warning(
      n_high, " observation(s) have leverage >= ", leverage_threshold,
      "; LOO predictions for these are unreliable.",
      "Consider re-fitting with greater penalties or fewer knots.",
      call. = FALSE
    )
  }
  ## Apply the standard leave-one-out shortcut once the hat diagonal is in hand.
  denom <- 1 - diag_hat

  leave_one_out <- model_fit$y - (model_fit$y - model_fit$ytilde) / denom

  return(leave_one_out)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' Efficient Matrix Multiplication
#'
#' Performs matrix multiplication using RcppArmadillo
#'
#' @param A First input matrix
#' @param B Second input matrix
#' @return Matrix product of A and B
#' @export
#' @keywords internal
efficient_matrix_mult <- function(A, B) {
    .Call('_lgspline_efficient_matrix_mult', PACKAGE = 'lgspline', A, B)
}

#' Compute Gram Matrix
#'
#' Calculates X^T * X for the input matrix X
#'
#' @param X Input matrix
#' @return Gram matrix (X^T * X)
#' @export
#' @keywords internal
gramMatrix <- function(X) {
    .Call('_lgspline_gramMatrix', PACKAGE = 'lgspline', X)
}

#' Matrix Inversion using Armadillo
#'
#' Computes the inverse of a matrix using Armadillo's inversion method
#'
#' @param x Input matrix to be inverted
#' @return Inverted matrix
#' @export
#' @keywords internal
armaInv <- function(x) {
    .Call('_lgspline_armaInv', PACKAGE = 'lgspline', x)
}

#' Block Matrix Multiplication
#'
#' Performs multiplication of a list of matrices with a block matrix
#'
#' @param G List of matrices G
#' @param A Input matrix A
#' @param K Number of blocks
#' @param nc Number of columns
#' @param nca Number of columns in A
#' @return List of multiplied matrices
#' @noRd
#' @keywords internal
GAmult <- function(G, A, K, nc, nca) {
    .Call('_lgspline_GAmult', PACKAGE = 'lgspline', G, A, K, nc, nca)
}

#' Compute Trace Correction
#'
#' Calculates a trace correction for matrix computations
#'
#' @param G List of matrices G
#' @param A Input matrix A
#' @param GXX List of GX^{T}X matrices
#' @param AGAInv Matrix (A^{T}GA)^{-1}
#' @param nc Number of columns of each partition of G
#' @param K Number of blocks
#' @return Trace correction value
#' @noRd
#' @keywords internal
compute_trace_correction <- function(G, A, GXX, AGAInv, nc, K) {
    .Call('_lgspline_compute_trace_correction', PACKAGE = 'lgspline', G, A, GXX, AGAInv, nc, K)
}

#' AGAmult Computation Overall
#'
#' Performs A^{T}GA computation over all matrices in a list (G)
#'
#' @param G_chunk List of matrices for the chunk
#' @param A Input matrix A
#' @param K Number of partitions minus one
#' @param nc Number of columns
#' @return Resulting matrix
#' @noRd
#' @keywords internal
AGAmult <- function(G, A, K, nc, nca) {
    .Call('_lgspline_AGAmult', PACKAGE = 'lgspline', G, A, K, nc, nca)
}

#' Chunk-based AGAmult Computation
#'
#' Performs A^{T}GAmult computation on a specific chunk of matrices
#'
#' @param G_chunk List of matrices for the chunk
#' @param A Input matrix A
#' @param chunk_start Starting chunk index
#' @param chunk_end Ending chunk index
#' @param nc Number of columns
#' @return Resulting matrix
#' @noRd
#' @keywords internal
AGAmult_chunk <- function(G_chunk, A, chunk_start, chunk_end, nc) {
    .Call('_lgspline_AGAmult_chunk', PACKAGE = 'lgspline', G_chunk, A, chunk_start, chunk_end, nc)
}

#' Compute AGXy
#'
#' Computes A^T * G * Xy for specified range
#'
#' @param G List of G matrices
#' @param A Input matrix A
#' @param Xy List of Xy vectors
#' @param nc Number of columns
#' @param K Total number of blocks
#' @param start Start index
#' @param end End index
#' @return Resulting vector
#' @noRd
#' @keywords internal
compute_AGXy <- function(G, A, Xy, nc, K, start, end) {
    .Call('_lgspline_compute_AGXy', PACKAGE = 'lgspline', G, A, Xy, nc, K, start, end)
}

#' Compute Result Blocks
#'
#' Computes blocks of results for specific computations
#'
#' @param G List of G matrices
#' @param Ghalf List of Ghalf matrices
#' @param A Input matrix A
#' @param AAGAInvAGXy Input vector
#' @param nc Number of columns
#' @param start Start index
#' @param end End index
#' @return Resulting vector of blocks
#' @noRd
#' @keywords internal
compute_result_blocks <- function(G, Ghalf, A, AAGAInvAGXy, nc, start, end) {
    .Call('_lgspline_compute_result_blocks', PACKAGE = 'lgspline', G, Ghalf, A, AAGAInvAGXy, nc, start, end)
}

#' Block Diagonal Matrix Multiplication
#'
#' Performs multiplication of block diagonal matrices
#'
#' @param A List of matrices A
#' @param B List of matrices B
#' @param K Number of blocks
#' @return List of multiplied matrices
#' @noRd
#' @keywords internal
matmult_block_diagonal_cpp <- function(A, B, K) {
    .Call('_lgspline_matmult_block_diagonal_cpp', PACKAGE = 'lgspline', A, B, K)
}

#' Vector-Matrix Multiplication for Block Diagonal Matrices
#'
#' Performs vector-matrix multiplication for block diagonal matrices
#'
#' @param A List of matrices A
#' @param b List of vectors b
#' @param K Number of blocks
#' @return List of resulting vectors
#' @export
#' @keywords internal
vectorproduct_block_diagonal <- function(A, b, K) {
    .Call('_lgspline_vectorproduct_block_diagonal', PACKAGE = 'lgspline', A, b, K)
}

#' Matrix-Matrix Addition for Block Diagonal Matrices
#'
#' Performs matrix-matrix elementwise addition for block diagonal matrices
#'
#' @param A List of matrices A
#' @param B List of matrices B
#' @param K Number of blocks
#' @return List of resulting matrices
#' @noRd
#' @keywords internal
matadd_block_diagonal <- function(A, B, K) {
    .Call('_lgspline_matadd_block_diagonal', PACKAGE = 'lgspline', A, B, K)
}

#' Compute Derivative of W
#'
#' Computes derivative of diag(1/(1-XUGXt)) with respect to lambda, penalties of G
#'
#' @param A List of matrices A
#' @param GXX List of matrices of G times gram matrix of X transpose for each partition
#' @param Ghalf List of matrices of square-root of G for each partition
#' @param dG_dlambda List of derivatives of G matrix with respect to penalties lambda
#' @param dGhalf_dlambda List of derivatives of Ghalf matrix with respect to penalties lambda
#' @param AGAInv Matrix of A transpose times G times A within U = (I - GA(A^{T}GA)^{-1}A^{T})
#' @param nc Numeric numeric of basis expansions of predictors per partition
#' @param K Number of blocks
#' @return Matrix representing the derivative
#' @noRd
#' @keywords internal
compute_dW_dlambda <- function(G, A, GXX, Ghalf, dG_dlambda, dGhalf_dlambda, AGAInv, nc, K) {
    .Call('_lgspline_compute_dW_dlambda', PACKAGE = 'lgspline', G, A, GXX, Ghalf, dG_dlambda, dGhalf_dlambda, AGAInv, nc, K)
}

#' Compute G^{1/2}(GA(A^{T}GA)^{-1}A^{T})GX^{T}y 
#'
#' Computes the equivalent vector of G^{1/2}X^{T}y in the proposed model-fitting procedure but for G^{1/2}(GA(A^{T}GA)^{-1}A^{T})GX^{T}y instead.
#'
#' @param G List of matrices of G times gram matrix of X transpose for each partition
#' @param Ghalf List of matrices of square-root of G for each partition
#' @param A List of matrices A
#' @param AGAInv Matrix of A transpose times G times A within U = (I - GA(A^{T}GA)^{-1}A^{T})
#' @param Xy List of vectors for partition-wise dot product of transpose(X) times vector y
#' @param nc Numeric numeric of basis expansions of predictors per partition
#' @param K Number of blocks
#' @return Matrix representing the operation
#' @noRd
#' @keywords internal
compute_GhalfXy_temp <- function(G, Ghalf, A, AGAInv, Xy, nc, K) {
    .Call('_lgspline_compute_GhalfXy_temp', PACKAGE = 'lgspline', G, Ghalf, A, AGAInv, Xy, nc, K)
}

#' Left-Multiply a List of Block-Diagonal Matrices by U 
#'
#' Useful for computing UG where G is a list of block-diagonal matrices for each partition, and U is a square P by P dense matrix
#'
#' @param U Matrix of dimension P by P that projects onto the null space of A transpose
#' @param G List of G matrices for each partition
#' @param nc Numeric numeric of basis expansions of predictors per partition
#' @param K Number of blocks
#' @return Matrix of UG
#' @export
#' @keywords internal
matmult_U <- function(U, G, nc, K) {
    .Call('_lgspline_matmult_U', PACKAGE = 'lgspline', U, G, nc, K)
}

## process_qp.R
## Quadratic programming constraint preparation for lgspline
##
## [Change 2026-03-06] Refactored from lgspline.fit into a standalone
## helper so that (a) the main fitting function is easier to review,
## (b) the QP setup can be tested independently, and (c) per-variable
## derivative sign constraints are supported.
##
## Convention notes (for reviewers):
## - First comment line: ##   continuation: #
## - Long delimiter lines (## ## ## ##) used sparingly
## - LaTeX notation in Roxygen uses \eqn{}


#' Build Derivative QP Constraints in Full \eqn{P}-Dimensional Space
#'
#' @description
#' Internal helper that constructs the \code{Amat} / \code{bvec} pair
#' enforcing derivative sign constraints at every row of a block-diagonal
#' design matrix.
#'
#' Given an \eqn{N_{\mathrm{sub}} \times P} block-diagonal design matrix
#' \code{X_block}, where \eqn{P = p \times (K+1)}, this function:
#' \enumerate{
#'   \item Recovers the per-partition \eqn{p}-column expansion matrix
#'     \code{C_qp} and records each row's partition assignment.
#'   \item Calls \code{make_derivative_matrix} on \code{C_qp} to obtain
#'     first or second derivative matrices with respect to each predictor.
#'   \item Optionally selects only derivatives for a subset of predictor
#'     variables (\code{target_vars}).
#'   \item Maps each derivative row into the full \eqn{P}-dimensional
#'     coefficient space, yielding one constraint column per (observation,
#'     variable) pair.
#' }
#'
#' The result is a constraint pair \eqn{\mathbf{A}^{\top}\boldsymbol{\beta}
#' \ge \mathbf{b}} (with \eqn{\mathbf{b} = \mathbf{0}}) suitable for
#' \code{\link[quadprog]{solve.QP}}.
#'
#' @param X_block Numeric matrix, \eqn{N_{\mathrm{sub}} \times P}.
#'   Block-diagonal design matrix for the QP-selected observations.
#' @param sign_mult Numeric scalar, \code{+1} for positive constraints
#'   or \code{-1} for negative constraints.
#' @param just_first Logical. If \code{TRUE}, constrain first derivatives;
#'   if \code{FALSE}, constrain second derivatives.
#' @param p_expansions Integer. Number of basis expansions per partition.
#' @param K Integer. Number of interior knots (partitions minus 1).
#' @param colnm_expansions Character vector of length \code{p_expansions}.
#'   Column names of the expansion matrix.
#' @param power1_cols Integer vector of linear-term column indices.
#' @param power2_cols Integer vector of quadratic-term column indices.
#' @param nonspline_cols Integer vector of non-spline linear column indices.
#' @param interaction_single_cols Integer vector of linear-by-linear
#'   interaction column indices.
#' @param interaction_quad_cols Integer vector of linear-by-quadratic
#'   interaction column indices.
#' @param triplet_cols Integer vector of three-way interaction column indices.
#' @param include_2way_interactions Logical switch forwarded to
#'   \code{make_derivative_matrix}.
#' @param include_3way_interactions Logical switch forwarded to
#'   \code{make_derivative_matrix}.
#' @param include_quadratic_interactions Logical switch forwarded to
#'   \code{make_derivative_matrix}.
#' @param expansion_scales Numeric vector of length \code{p_expansions - 1}.
#'   Passed as unit scales (\code{1 + 0 * expansion_scales}) so that the
#'   constraint matrix stays on the expansion-standardized coefficient scale.
#' @param target_vars Optional. Integer vector of predictor column indices
#'   or character vector of predictor names identifying which predictors
#'   to constrain. When \code{NULL} (default), all derivative variables
#'   are constrained.
#' @param og_cols Optional character vector of original predictor column
#'   names, used to resolve character \code{target_vars} to integer indices.
#'
#' @return A list with components:
#' \describe{
#'   \item{Amat}{\eqn{P \times M} constraint matrix, where \eqn{M} is the
#'     number of unique constraint columns (after deduplication).}
#'   \item{bvec}{Numeric vector of length \eqn{M}, all zeros.}
#'   \item{meq}{Integer, always \code{0} (inequality constraints).}
#' }
#'
#' @keywords internal
#' @seealso \code{\link{process_qp}}, \code{\link[quadprog]{solve.QP}}
.build_deriv_qp <- function(X_block,
                            sign_mult,
                            just_first,
                            p_expansions,
                            K,
                            colnm_expansions,
                            power1_cols,
                            power2_cols,
                            nonspline_cols,
                            interaction_single_cols,
                            interaction_quad_cols,
                            triplet_cols,
                            include_2way_interactions,
                            include_3way_interactions,
                            include_quadratic_interactions,
                            expansion_scales,
                            target_vars = NULL,
                            og_cols = NULL) {

  N_sub <- nrow(X_block)

  ## Recover the p-column expansion matrix from the block-diagonal design.
  # Each row of X_block belongs to exactly one partition k, and its
  # p_expansions nonzero entries sit in columns [(k-1)*p + 1] : [k*p].
  C_qp <- matrix(0, N_sub, p_expansions)
  colnames(C_qp) <- colnm_expansions
  partition_assignment <- integer(N_sub)
  for (i in 1:N_sub) {
    for (k in 1:(K + 1)) {
      cols_k <- ((k - 1) * p_expansions + 1):(k * p_expansions)
      if (any(X_block[i, cols_k] != 0)) {
        C_qp[i, ] <- X_block[i, cols_k]
        partition_assignment[i] <- k
        break
      }
    }
  }

  ## Compute p-column derivative matrices on the standardized scale.
  # Unit scales ensure the constraint matrix matches the coefficient scale.
  derivs <- make_derivative_matrix(
    p_expansions, C_qp, power1_cols, power2_cols, nonspline_cols,
    interaction_single_cols, interaction_quad_cols, triplet_cols,
    K, include_2way_interactions, include_3way_interactions,
    include_quadratic_interactions, colnm_expansions,
    1 + 0 * expansion_scales,
    just_first_derivatives = just_first,
    just_spline_effects = FALSE
  )

  ## Select first or second derivative list
  if (just_first) {
    deriv_list <- derivs$first_derivatives
  } else {
    deriv_list <- derivs$second_derivatives
  }

  ## Resolve target_vars to integer indices into deriv_list.
  # deriv_list is named by predictor; names(deriv_list) holds the
  # expansion-name keys (e.g. "_1_", "_2_", or "Girth", "Height").
  if (!is.null(target_vars)) {
    deriv_names <- names(deriv_list)

    ## Convert character target_vars to the internal "_j_" form
    if (is.character(target_vars) && !is.null(og_cols)) {
      target_numeric <- unlist(lapply(target_vars, function(v) {
        idx <- which(og_cols == v)
        if (length(idx) == 0) idx <- grep(v, og_cols, fixed = TRUE)
        idx
      }))
      target_internal <- paste0("_", target_numeric, "_")
    } else if (is.numeric(target_vars)) {
      target_internal <- paste0("_", as.integer(target_vars), "_")
    } else {
      ## Fallback: treat as-is
      target_internal <- target_vars
    }

    ## Keep only derivative components whose name matches a target.
    # The derivative list names are the linear-term expansion names
    # (e.g. "_1_", "_2_"), so we match against target_internal.
    keep_idx <- which(sapply(deriv_names, function(nm) {
      any(sapply(target_internal, function(tv) grepl(tv, nm, fixed = TRUE)))
    }))

    if (length(keep_idx) == 0) {
      ## No matching variables: return empty constraint
      return(list(
        Amat = matrix(0, nrow = p_expansions * (K + 1), ncol = 0),
        bvec = numeric(0),
        meq = 0L
      ))
    }
    deriv_list <- deriv_list[keep_idx]
  }

  ## Map each row of each derivative matrix into P-dimensional space.
  # One constraint column per (observation, variable) pair.
  constraint_cols <- list()
  for (v in seq_along(deriv_list)) {
    deriv_mat <- deriv_list[[v]]
    for (i in 1:nrow(deriv_mat)) {
      k <- partition_assignment[i]
      col_P <- rep(0, p_expansions * (K + 1))
      col_P[((k - 1) * p_expansions + 1):(k * p_expansions)] <-
        sign_mult * deriv_mat[i, ]
      constraint_cols[[length(constraint_cols) + 1]] <- col_P
    }
  }

  qp_Amat <- do.call(cbind, constraint_cols)
  qp_Amat <- t(unique(t(qp_Amat)))
  qp_bvec <- rep(0, ncol(qp_Amat))

  list(Amat = qp_Amat, bvec = qp_bvec, meq = 0L)
}


#' Prepare Quadratic Programming Constraints for lgspline
#'
#' @description
#' Constructs \code{qp_Amat}, \code{qp_bvec}, and \code{qp_meq} from the
#' built-in quadratic programming constraints handled here: range bounds,
#' derivative sign constraints (first and second), monotonicity constraints,
#' and user-supplied custom constraint functions.
#'
#' This function was refactored from the inline QP setup block in
#' \code{\link{lgspline.fit}} to improve readability and testability.
#' The interface is fully backward-compatible: existing calls that pass
#' \code{TRUE}/\code{FALSE} for derivative flags continue to work.
#'
#' @section Per-Variable Derivative Constraints:
#' The arguments \code{qp_positive_derivative}, \code{qp_negative_derivative},
#' \code{qp_positive_2ndderivative}, and \code{qp_negative_2ndderivative}
#' now accept three forms:
#' \describe{
#'   \item{\code{FALSE}}{No constraint (default).}
#'   \item{\code{TRUE}}{Constrain \strong{all} predictor variables (backward
#'     compatible with previous behavior).}
#'   \item{Character or integer vector}{Constrain only the specified
#'     predictor variables. Character entries are resolved via
#'     \code{og_cols}; integer entries refer to column positions in the
#'     predictor matrix.}
#' }
#'
#' This allows, for example, enforcing a positive first derivative for
#' variable \code{"Dose"} and a negative first derivative for variable
#' \code{"Time"} simultaneously:
#' \preformatted{
#'   lgspline(...,
#'            qp_positive_derivative = "Dose",
#'            qp_negative_derivative = "Time")
#' }
#'
#' The arguments \code{qp_monotonic_increase} and
#' \code{qp_monotonic_decrease} remain \code{TRUE}/\code{FALSE} only,
#' because they constrain fitted values in observation order (not
#' per-variable).
#'
#' @param X List of per-partition design matrices.
#' @param K Integer. Number of interior knots.
#' @param p_expansions Integer. Number of basis expansions per partition.
#' @param order_list List of per-partition observation index vectors.
#' @param colnm_expansions Character vector of expansion column names.
#' @param expansion_scales Numeric vector of expansion standardization scales.
#' @param power1_cols Integer vector of linear-term column indices.
#' @param power2_cols Integer vector of quadratic-term column indices.
#' @param nonspline_cols Integer vector of non-spline linear column indices.
#' @param interaction_single_cols Integer vector of linear-by-linear
#'   interaction column indices.
#' @param interaction_quad_cols Integer vector of linear-by-quadratic
#'   interaction column indices.
#' @param triplet_cols Integer vector of three-way interaction column indices.
#' @param include_2way_interactions Logical switch controlling whether
#'   two-way interactions are included in derivative construction.
#' @param include_3way_interactions Logical switch controlling whether
#'   three-way interactions are included in derivative construction.
#' @param include_quadratic_interactions Logical switch controlling whether
#'   quadratic interactions are included in derivative construction.
#' @param family GLM family object.
#' @param mean_y,sd_y Numeric scalars for response standardization.
#' @param N_obs Integer. Total sample size.
#' @param qp_observations Optional integer vector of observation indices.
#' @param qp_positive_derivative,qp_negative_derivative Logical scalar,
#'   character vector, or integer vector. See section
#'   \emph{Per-Variable Derivative Constraints}.
#' @param qp_positive_2ndderivative,qp_negative_2ndderivative Same as above
#'   but for second derivatives.
#' @param qp_monotonic_increase,qp_monotonic_decrease Logical.
#'   Constrain fitted values to be monotonic in observation order.
#' @param qp_range_upper,qp_range_lower Optional numeric upper/lower bounds
#'   for fitted values.
#' @param qp_Amat_fxn,qp_bvec_fxn,qp_meq_fxn Optional user-supplied
#'   constraint-generating functions.
#' @param qp_Amat,qp_bvec,qp_meq Optional pre-built QP objects. Their presence
#'   marks QP handling as active, but this helper does not append them to the
#'   built-in constraints it constructs; they are expected to be handled
#'   outside this constructor.
#' @param all_derivatives_fxn Function to compute derivatives from expansion
#'   matrices (the \code{all_derivatives} closure from \code{lgspline.fit}).
#' @param og_cols Optional character vector of original predictor column names.
#' @param include_warnings Logical. Whether to issue warnings.
#' @param ... Additional arguments forwarded to custom constraint functions.
#'
#' @return A list with components:
#' \describe{
#'   \item{qp_Amat}{\eqn{P \times M} combined constraint matrix.}
#'   \item{qp_bvec}{Numeric vector of length \eqn{M}.}
#'   \item{qp_meq}{Integer. Number of leading equality constraints.}
#'   \item{quadprog}{Logical. \code{TRUE} if any QP constraints are active.}
#' }
#'
#' @examples
#' \dontrun{
#' ## Standalone verification: simple 1-D monotonic increase
#' set.seed(42)
#' t <- seq(-5, 5, length.out = 200)
#' y <- 3 * sin(t) + t + rnorm(200, 0, 0.5)
#'
#' ## Fit with positive first-derivative constraint on all variables
#' fit1 <- lgspline(t, y, K = 3,
#'                  qp_positive_derivative = TRUE)
#'
#' ## Verify: first derivative should be >= 0 everywhere
#' derivs1 <- predict(fit1, new_predictors = sort(t),
#'                    take_first_derivatives = TRUE)
#' stopifnot(all(derivs1$first_deriv >= -1e-8))
#'
#' ## Fit with monotonic increase (observation-order)
#' fit2 <- lgspline(t, y, K = 3,
#'                  qp_monotonic_increase = TRUE)
#' preds2 <- predict(fit2, new_predictors = sort(t))
#' stopifnot(all(diff(preds2) >= -1e-8))
#'
#' ## Per-variable constraints: 2-D example
#' x1 <- runif(500, -5, 5)
#' x2 <- runif(500, -5, 5)
#' y2 <- x1 + sin(x2) + rnorm(500, 0, 0.5)
#' dat2 <- data.frame(x1 = x1, x2 = x2, y = y2)
#'
#' ## Constrain x1 to have positive derivative, x2 to have negative
#' fit3 <- lgspline(y ~ spl(x1, x2), data = dat2, K = 2,
#'                  qp_positive_derivative = "x1",
#'                  qp_negative_derivative = "x2")
#'
#' ## Verify per-variable derivatives
#' newdat <- expand.grid(x1 = seq(-4, 4, length.out = 50),
#'                       x2 = seq(-4, 4, length.out = 50))
#' d3 <- predict(fit3, new_predictors = newdat,
#'               take_first_derivatives = TRUE)
#'
#' ## x1 derivative should be >= 0
#' stopifnot(all(unlist(d3$first_deriv[["_1_"]]) >= -1e-6))
#' ## x2 derivative should be <= 0
#' stopifnot(all(unlist(d3$first_deriv[["_2_"]]) <= 1e-6))
#'
#' ## Numeric column indices work identically
#' fit4 <- lgspline(y ~ spl(x1, x2), data = dat2, K = 2,
#'                  qp_positive_derivative = 1,
#'                  qp_negative_derivative = 2)
#'
#' ## Range + derivative constraints simultaneously
#' fit5 <- lgspline(t, y, K = 3,
#'                  qp_positive_derivative = TRUE,
#'                  qp_range_lower = -5,
#'                  qp_range_upper = 15)
#' preds5 <- predict(fit5)
#' stopifnot(all(preds5 >= -5 - 1e-6))
#' stopifnot(all(preds5 <= 15 + 1e-6))
#' }
#'
#' @export
process_qp <- function(X,
                       K,
                       p_expansions,
                       order_list,
                       colnm_expansions,
                       expansion_scales,
                       power1_cols,
                       power2_cols,
                       nonspline_cols,
                       interaction_single_cols,
                       interaction_quad_cols,
                       triplet_cols,
                       include_2way_interactions,
                       include_3way_interactions,
                       include_quadratic_interactions,
                       family,
                       mean_y,
                       sd_y,
                       N_obs,
                       qp_observations = NULL,
                       qp_positive_derivative = FALSE,
                       qp_negative_derivative = FALSE,
                       qp_positive_2ndderivative = FALSE,
                       qp_negative_2ndderivative = FALSE,
                       qp_monotonic_increase = FALSE,
                       qp_monotonic_decrease = FALSE,
                       qp_range_upper = NULL,
                       qp_range_lower = NULL,
                       qp_Amat_fxn = NULL,
                       qp_bvec_fxn = NULL,
                       qp_meq_fxn = NULL,
                       qp_Amat = NULL,
                       qp_bvec = NULL,
                       qp_meq = 0,
                       all_derivatives_fxn = NULL,
                       og_cols = NULL,
                       include_warnings = TRUE,
                       ...) {

  ## Determine whether any QP constraints are active.
  # The derivative flags can now be TRUE, FALSE, or a vector of
  # variable names/indices, so we check with .is_active().
  .is_active <- function(x) {
    if (is.logical(x)) return(any(x))
    if (is.numeric(x) || is.character(x)) return(length(x) > 0)
    FALSE
  }

  quadprog <- (
    .is_active(qp_negative_derivative) |
    .is_active(qp_monotonic_decrease) |
    .is_active(qp_negative_2ndderivative) |
    .is_active(qp_positive_derivative) |
    .is_active(qp_monotonic_increase) |
    .is_active(qp_positive_2ndderivative) |
    any(!is.null(qp_range_upper)) |
    any(!is.null(qp_range_lower)) |
    (any(!is.null(qp_Amat_fxn)) &
     any(!is.null(qp_bvec_fxn)) &
     any(!is.null(qp_meq_fxn))) |
    (any(!is.null(qp_Amat)) &
     any(!is.null(qp_bvec)) &
     any(!is.null(qp_meq)))
  )

  if (!quadprog) {
    return(list(
      qp_Amat = NULL,
      qp_bvec = NULL,
      qp_meq = 0L,
      quadprog = FALSE
    ))
  }

  ## Accumulate constraint triples from each source
  qp_Amat_list <- list()
  qp_bvec_list <- list()
  qp_meq_list <- list()

  ## Build the block-diagonal design matrix for QP observations.
  # This is N_sub x P where P = p_expansions * (K+1).
  X_block <- Reduce("rbind", lapply(1:(K + 1), function(k) {
    if (nrow(X[[k]]) == 0) return(X[[k]])
    Reduce("cbind", lapply(1:(K + 1), function(j) {
      if (j == k) X[[k]] else 0 * X[[k]]
    }))
  }))

  if (any(!is.null(qp_observations))) {
    qp_observations <- try(c(qp_observations), silent = TRUE)
    if (any(inherits(qp_observations, 'try-error'))) {
      stop('\n \t qp_observations must be coercible to a numeric vector \n')
    }
    X_block <- X_block[c(unlist(order_list)) %in% qp_observations, , drop = FALSE]
  }

  ## Resolve per-variable derivative flag to target_vars (NULL = all).
  # If TRUE, target_vars = NULL (constrain all); if FALSE, skip;
  # if character/numeric, pass through as target_vars.
  .resolve_target <- function(flag) {
    if (is.logical(flag)) {
      if (any(flag)) return(NULL)  # NULL = all variables
      else return(NA)              # NA = skip
    }
    ## Character or numeric: specific variables
    flag
  }

  ## Range constraints
  if (!any(is.null(qp_range_upper)) | !any(is.null(qp_range_lower))) {
    if (paste0(family)[2] != 'identity') {
      if (any(!is.null(qp_range_upper))) {
        qp_range_upper <- family$linkfun(qp_range_upper)
      }
      if (any(!is.null(qp_range_lower))) {
        qp_range_lower <- family$linkfun(qp_range_lower)
      }
      if (any(is.na(c(qp_range_upper, qp_range_lower))) |
          any(!is.finite(c(qp_range_upper, qp_range_lower)))) {
        stop('\n\tQuadratic programming upper/lower range constraints ',
             'are NA or not finite after link function transformation. ',
             'Supply bounds on raw-response scale, which are in-range of ',
             'link function transformations.\n')
      }
    }

    if (!any(is.null(qp_range_upper)) & !any(is.null(qp_range_lower))) {
      range_Amat <- cbind(t(X_block), -t(X_block))
      if (length(qp_range_lower) == 1) {
        if (length(qp_range_upper) == 1) {
          range_Amat <- t(unique(t(range_Amat)))
        }
        range_bvec_lower <- rep((qp_range_lower - mean_y) / sd_y,
                                ncol(range_Amat) / 2)
      } else {
        range_bvec_lower <- (qp_range_lower - mean_y) / sd_y
      }
      if (length(qp_range_upper) == 1) {
        range_bvec_upper <- rep(-(qp_range_upper - mean_y) / sd_y,
                                ncol(range_Amat) / 2)
      } else {
        range_bvec_upper <- -(qp_range_upper - mean_y) / sd_y
      }
      qp_Amat_list[[length(qp_Amat_list) + 1]] <- range_Amat
      qp_bvec_list[[length(qp_bvec_list) + 1]] <- c(range_bvec_lower,
                                                     range_bvec_upper)
      qp_meq_list[[length(qp_meq_list) + 1]] <- 0

    } else if (!any(is.null(qp_range_upper))) {
      range_Amat <- -t(X_block)
      if (length(qp_range_upper) == 1) {
        range_Amat <- t(unique(t(range_Amat)))
        range_bvec <- rep(qp_range_upper, ncol(range_Amat))
      } else {
        range_bvec <- qp_range_upper
      }
      range_bvec <- -(range_bvec - mean_y) / sd_y
      qp_Amat_list[[length(qp_Amat_list) + 1]] <- range_Amat
      qp_bvec_list[[length(qp_bvec_list) + 1]] <- range_bvec
      qp_meq_list[[length(qp_meq_list) + 1]] <- 0

    } else if (!any(is.null(qp_range_lower))) {
      range_Amat <- t(X_block)
      if (length(qp_range_lower) == 1) {
        range_Amat <- t(unique(t(range_Amat)))
        range_bvec <- rep(qp_range_lower, ncol(range_Amat))
      } else {
        range_bvec <- qp_range_lower
      }
      range_bvec <- (range_bvec - mean_y) / sd_y
      qp_Amat_list[[length(qp_Amat_list) + 1]] <- range_Amat
      qp_bvec_list[[length(qp_bvec_list) + 1]] <- range_bvec
      qp_meq_list[[length(qp_meq_list) + 1]] <- 0
    }
  }

  ## Bundle the common derivative-matrix inputs once so each sign / order
  #  case only needs to specify what is changing.
  deriv_shared <- list(
    X_block = X_block,
    p_expansions = p_expansions,
    K = K,
    colnm_expansions = colnm_expansions,
    power1_cols = power1_cols,
    power2_cols = power2_cols,
    nonspline_cols = nonspline_cols,
    interaction_single_cols = interaction_single_cols,
    interaction_quad_cols = interaction_quad_cols,
    triplet_cols = triplet_cols,
    include_2way_interactions = include_2way_interactions,
    include_3way_interactions = include_3way_interactions,
    include_quadratic_interactions = include_quadratic_interactions,
    expansion_scales = expansion_scales,
    og_cols = og_cols
  )

  ## Positive first derivative
  target_pos1 <- .resolve_target(qp_positive_derivative)
  if (!identical(target_pos1, NA)) {
    args <- c(deriv_shared, list(
      sign_mult = +1,
      just_first = TRUE,
      target_vars = target_pos1
    ))
    deriv_qp <- do.call(.build_deriv_qp, args)
    if (ncol(deriv_qp$Amat) > 0) {
      qp_Amat_list[[length(qp_Amat_list) + 1]] <- deriv_qp$Amat
      qp_bvec_list[[length(qp_bvec_list) + 1]] <- deriv_qp$bvec
      qp_meq_list[[length(qp_meq_list) + 1]] <- deriv_qp$meq
    }
  }

  ## Negative first derivative
  target_neg1 <- .resolve_target(qp_negative_derivative)
  if (!identical(target_neg1, NA)) {
    args <- c(deriv_shared, list(
      sign_mult = -1,
      just_first = TRUE,
      target_vars = target_neg1
    ))
    deriv_qp <- do.call(.build_deriv_qp, args)
    if (ncol(deriv_qp$Amat) > 0) {
      qp_Amat_list[[length(qp_Amat_list) + 1]] <- deriv_qp$Amat
      qp_bvec_list[[length(qp_bvec_list) + 1]] <- deriv_qp$bvec
      qp_meq_list[[length(qp_meq_list) + 1]] <- deriv_qp$meq
    }
  }

  ## Positive second derivative
  target_pos2 <- .resolve_target(qp_positive_2ndderivative)
  if (!identical(target_pos2, NA)) {
    args <- c(deriv_shared, list(
      sign_mult = +1,
      just_first = FALSE,
      target_vars = target_pos2
    ))
    deriv_qp <- do.call(.build_deriv_qp, args)
    if (ncol(deriv_qp$Amat) > 0) {
      qp_Amat_list[[length(qp_Amat_list) + 1]] <- deriv_qp$Amat
      qp_bvec_list[[length(qp_bvec_list) + 1]] <- deriv_qp$bvec
      qp_meq_list[[length(qp_meq_list) + 1]] <- deriv_qp$meq
    }
  }

  ## Negative second derivative
  target_neg2 <- .resolve_target(qp_negative_2ndderivative)
  if (!identical(target_neg2, NA)) {
    args <- c(deriv_shared, list(
      sign_mult = -1,
      just_first = FALSE,
      target_vars = target_neg2
    ))
    deriv_qp <- do.call(.build_deriv_qp, args)
    if (ncol(deriv_qp$Amat) > 0) {
      qp_Amat_list[[length(qp_Amat_list) + 1]] <- deriv_qp$Amat
      qp_bvec_list[[length(qp_bvec_list) + 1]] <- deriv_qp$bvec
      qp_meq_list[[length(qp_meq_list) + 1]] <- deriv_qp$meq
    }
  }

  ## Monotonic increase (observation order, not per-variable)
  if (qp_monotonic_increase || qp_monotonic_decrease) {
    sign_mono <- if(qp_monotonic_increase) +1 else -1

    ## Recover original observation order from partition-stacked X_block.
    #  order_list[[k]] gives the original row indices for partition k.
    #  Stacking them in partition order gives the permutation applied to
    #  build X_block; inverting it restores original observation order.
    obs_order    <- unlist(order_list)
    inv_order    <- order(obs_order)
    X_block_orig <- X_block[inv_order, , drop = FALSE]

    ## If qp_observations was applied, X_block may be a subset.
    #  In that case inv_order may be longer than nrow(X_block); guard.
    if(nrow(X_block_orig) != nrow(X_block)){
      ## Fallback: use X_block as-is (already filtered by qp_observations)
      X_block_orig <- X_block
    }

    value_constraints <- t(Reduce(
      'rbind',
      lapply(2:nrow(X_block_orig), function(i) {
        matrix(sign_mono * c(X_block_orig[i, ] - X_block_orig[i - 1, ]),
               nrow = 1)
      })
    ))
    mono_Amat <- cbind(value_constraints)
    mono_Amat <- t(unique(t(mono_Amat)))
    mono_bvec <- rep(0, ncol(mono_Amat))
    qp_Amat_list[[length(qp_Amat_list) + 1]] <- mono_Amat
    qp_bvec_list[[length(qp_bvec_list) + 1]] <- mono_bvec
    qp_meq_list[[length(qp_meq_list) + 1]] <- 0
  }

  ## User-supplied constraint builders are evaluated last and then appended
  #  to the built-in derivative / range / monotonicity constraints.
  if (!is.null(qp_Amat_fxn) &
      !is.null(qp_bvec_fxn) &
      !is.null(qp_meq_fxn)) {

    custom_Amat <- qp_Amat_fxn(
      N = N_obs, p = p_expansions, K = K,
      X = X_block, colnm = colnm_expansions,
      scales = expansion_scales,
      deriv_fxn = all_derivatives_fxn, ...
    )
    custom_bvec <- qp_bvec_fxn(
      qp_Amat = custom_Amat, N = N_obs,
      p = p_expansions, K = K,
      X = X_block, colnm = colnm_expansions,
      scales = expansion_scales,
      deriv_fxn = all_derivatives_fxn, ...
    )
    custom_meq <- qp_meq_fxn(
      qp_Amat = custom_Amat, N = N_obs,
      p = p_expansions, K = K,
      X = X_block, colnm = colnm_expansions,
      scales = expansion_scales,
      deriv_fxn = all_derivatives_fxn, ...
    )

    ## Validate dimensions
    P_coef <- p_expansions * (K + 1)
    if (!is.null(custom_Amat) && nrow(custom_Amat) != P_coef) {
      stop(
        '\n \t qp_Amat_fxn returned a matrix with ', nrow(custom_Amat),
        ' rows, but the coefficient vector has length p*(K+1) = ',
        P_coef, '. The constraint matrix must have nrow = p*(K+1).',
        '\n \t Your function receives named arguments: ',
        'p (expansions per partition), K (number of interior knots), ',
        'N (observations), X (block-diagonal design), ',
        'scales (expansion scales), colnm (column names), ',
        'deriv_fxn (derivative function).\n')
    }
    if (!is.null(custom_bvec) && length(custom_bvec) != ncol(custom_Amat)) {
      stop(
        '\n \t qp_bvec_fxn returned a vector of length ', length(custom_bvec),
        ' but qp_Amat has ', ncol(custom_Amat), ' columns (constraints).',
        ' These must match.\n')
    }

    qp_Amat_list[[length(qp_Amat_list) + 1]] <- custom_Amat
    qp_bvec_list[[length(qp_bvec_list) + 1]] <- custom_bvec
    qp_meq_list[[length(qp_meq_list) + 1]] <- custom_meq
  }

  ## Stack all active constraint sources into the single solve.QP interface.
  qp_Amat_out <- do.call(cbind, qp_Amat_list)
  qp_bvec_out <- do.call(c, qp_bvec_list)
  qp_meq_out <- sum(unlist(qp_meq_list))

  list(
    qp_Amat = qp_Amat_out,
    qp_bvec = qp_bvec_out,
    qp_meq = qp_meq_out,
    quadprog = TRUE
  )
}

#' Compute Deviance for Non-GEE Models
#'
#' Evaluates the mean deviance (or mean squared error as fallback) at
#' the current fitted values.  Used inside the damped Newton-Raphson outer loop of
#' \code{\link{blockfit_solve}} for convergence monitoring.
#'
#' @param y_vec Numeric vector of observed responses.
#' @param mu_vec Numeric vector of fitted means.
#' @param obs_wt Numeric vector of observation weights.
#' @param ord Integer vector of observation indices (passed to
#'   custom deviance residual functions).
#' @param fam GLM family object.
#' @param ... Additional arguments forwarded to
#'   \code{fam$custom_dev.resids}.
#'
#' @return Scalar; the mean deviance contribution.
#'
#' @keywords internal
.bf_deviance <- function(y_vec, mu_vec, obs_wt, ord, fam, ...){
  if(!is.null(fam$custom_dev.resids)){
    mean(fam$custom_dev.resids(y_vec, mu_vec, ord,
                               fam, obs_wt, ...))
  } else if(!is.null(fam$dev.resids)){
    mean(obs_wt * fam$dev.resids(y_vec, cbind(mu_vec), wt = 1))
  } else {
    mean(obs_wt * (y_vec - mu_vec)^2)
  }
}


#' Compute Whitened Deviance for GEE Convergence Monitoring
#'
#' Evaluates deviance in the whitened (decorrelated) space when the
#' family supplies \code{fam$custom_dev.resids}.  In that case the raw
#' deviance residuals are divided by \eqn{\sqrt{W}} and then
#' pre-multiplied by \eqn{\mathbf{V}^{-1/2}} before squaring and
#' averaging.  If only \code{fam$dev.resids} is available, the function
#' falls back to the usual mean deviance (and otherwise mean squared
#' error).  Used in the GEE refinement loop (Case c) of
#' \code{\link{blockfit_solve}}.
#'
#' @inheritParams .bf_deviance
#' @param W_vec Numeric vector of damped Newton-Raphson weights at current iterate.
#' @param VhInv \eqn{\mathbf{V}^{-1/2}} matrix (permuted to
#'   partition ordering).
#'
#' @return Scalar; the mean whitened deviance.
#'
#' @keywords internal
.bf_gee_deviance <- function(y_vec, mu_vec, W_vec, ord,
                             obs_wt, VhInv, fam, ...){
  if(!is.null(fam$custom_dev.resids)){
    raw <- fam$custom_dev.resids(y_vec, mu_vec, ord,
                                 fam, obs_wt, ...)
    W_safe <- pmax(W_vec, sqrt(.Machine$double.eps))
    mean((VhInv %**%
            cbind(sign(raw) * sqrt(abs(raw)) / sqrt(c(W_safe))))^2)
  } else if(!is.null(fam$dev.resids)){
    mean(obs_wt * fam$dev.resids(y_vec, cbind(mu_vec), wt = 1))
  } else {
    mean(obs_wt * (y_vec - mu_vec)^2)
  }
}


#' Numerical Derivative of the Inverse Link Function
#'
#' Returns \eqn{\partial\mu / \partial\eta} for the supplied family
#' object, using the family's own \code{mu.eta} method when available
#' and falling back to analytic forms for common links or central
#' differences otherwise.
#'
#' @param eta Numeric vector of linear predictor values.
#' @param fam GLM family object.
#'
#' @return Numeric vector of the same length as \code{eta}.
#'
#' @keywords internal
.bf_get_mu_eta <- function(eta, fam){
  if(!is.null(fam$mu.eta)) return(fam$mu.eta(eta))
  mu <- fam$linkinv(eta)
  switch(fam$link,
         'identity' = rep(1, length(eta)),
         'log'      = mu,
         'logit'    = mu * (1 - mu),
         'inverse'  = -mu^2,
         'probit'   = dnorm(eta),
         'sqrt'     = 0.5 / sqrt(pmax(mu, .Machine$double.eps)),
         'cloglog'  = exp(eta - exp(eta)),
         {
           eps <- sqrt(.Machine$double.eps)
           (fam$linkinv(eta + eps) -
               fam$linkinv(eta - eps)) / (2 * eps)
         })
}


#' Construct Full Block-Diagonal Penalty Matrix
#'
#' Thin wrapper around \code{.solver_build_lambda_block()}.
#' The block/backfitting solver keeps its original helper name so the
#' surrounding code stays unchanged, while the shared implementation now
#' lives in \code{solver_utils.R}.
#'
#' @param Lambda \eqn{p_expansions \times p_expansions} shared penalty matrix.
#' @param K Integer; number of interior knots.
#' @param unique_penalty_per_partition Logical.
#' @param L_partition_list List of \eqn{K+1} partition-specific
#'   penalty matrices (or zeros).
#'
#' @return A \eqn{(p_expansions \cdot (K+1)) \times (p_expansions \cdot (K+1))} matrix.
#'
#' @keywords internal
.bf_make_Lambda_block <- function(Lambda, K,
                                  unique_penalty_per_partition,
                                  L_partition_list){
  .solver_build_lambda_block(
    Lambda = Lambda,
    K = K,
    unique_penalty_per_partition = unique_penalty_per_partition,
    L_partition_list = L_partition_list
  )
}


#' Split Design and Penalty into Spline and Flat Components
#'
#' Partitions the per-partition design matrices, penalty matrices, and
#' constraint matrix into "spline" (columns receiving partition-specific
#' coefficients) and "flat" (columns receiving a single shared
#' coefficient across all partitions) subsets.
#'
#' @param X List of \eqn{K+1} design matrices (\eqn{N_k \times p_expansions}).
#' @param flat_cols Integer vector of flat column indices.
#' @param p_expansions Integer; total columns per partition.
#' @param K Integer; number of interior knots.
#' @param Lambda \eqn{p_expansions \times p_expansions} shared penalty matrix.
#' @param L_partition_list List of partition-specific penalty matrices.
#' @param A Full \eqn{P \times R_constraints} constraint matrix.
#' @param constraint_values List of constraint right-hand sides.
#'
#' @return A named list with elements:
#' \describe{
#'   \item{X_spline}{List of \eqn{K+1} matrices, spline columns only.}
#'   \item{X_flat}{List of \eqn{K+1} matrices, flat columns only.}
#'   \item{Lambda_spline}{\eqn{nc_s \times nc_s} penalty submatrix.}
#'   \item{Lambda_flat}{\eqn{nc_f \times nc_f} penalty submatrix.}
#'   \item{L_part_spline}{List of partition-specific penalty submatrices
#'     for the spline columns.}
#'   \item{A_spline}{Spline-only constraint matrix (columns pruned
#'     and rank-reduced via pivoted QR).}
#'   \item{nca_spline}{Integer; number of columns in
#'     \code{A_spline}.}
#'   \item{constraint_values_spline}{List of constraint RHS vectors
#'     restricted to spline rows.}
#'   \item{spline_cols}{Integer vector of spline column indices.}
#'   \item{nc_spline}{Integer; number of spline columns.}
#'   \item{nc_flat}{Integer; number of flat columns.}
#'   \item{A_flat}{Flat-only constraint matrix (rows for flat coefficients,
#'     columns pruned and rank-reduced). Used to enforce mixed constraints
#'     on the flat update step.}
#'   \item{nca_flat}{Integer; number of columns in \code{A_flat}.}
#'   \item{A_full}{The original full constraint matrix \code{A}, retained
#'     for mixed-constraint enforcement.}
#'   \item{flat_rows_all}{Integer vector of flat-coefficient row indices
#'     in the full P-space.}
#'   \item{spline_rows_all}{Integer vector of spline-coefficient row indices
#'     in the full P-space.}
#'   \item{has_mixed_constraints}{Logical; TRUE if any constraint column
#'     in A has nonzero entries on both spline and flat rows.}
#'   \item{mixed_constraint_cols}{Integer vector of column indices in the
#'     original A that are mixed (touch both spline and flat rows).}
#' }
#'
#' @keywords internal
.bf_split_components <- function(X, flat_cols, p_expansions, K, Lambda,
                                 L_partition_list, A,
                                 constraint_values){

  spline_cols <- setdiff(1:p_expansions, flat_cols)
  nc_spline <- length(spline_cols)
  nc_flat   <- length(flat_cols)

  X_spline <- lapply(X, function(Xk) Xk[, spline_cols, drop = FALSE])
  X_flat   <- lapply(X, function(Xk) Xk[, flat_cols,   drop = FALSE])

  Lambda_spline <- Lambda[spline_cols, spline_cols]
  Lambda_flat   <- Lambda[flat_cols,   flat_cols]

  L_part_spline <- lapply(L_partition_list, function(Lk){
    if(is.numeric(Lk) && length(Lk) == 1 && Lk == 0) return(0)
    Lk[spline_cols, spline_cols]
  })

  ## Extract spline-only and flat-only rows from full constraint matrix A
  flat_rows_all   <- unlist(lapply(0:K, function(k) k * p_expansions + flat_cols))
  spline_rows_all <- setdiff(1:(p_expansions * (K + 1)), flat_rows_all)
  A_spline <- A[spline_rows_all, , drop = FALSE]
  A_flat   <- A[flat_rows_all,   , drop = FALSE]

  ## Detect mixed constraints: columns of A that have nonzero entries
  #  on BOTH spline and flat rows
  spline_nz <- colSums(abs(A_spline)) > sqrt(.Machine$double.eps)
  flat_nz   <- colSums(abs(A_flat))   > sqrt(.Machine$double.eps)
  mixed_constraint_cols <- which(spline_nz & flat_nz)
  has_mixed_constraints <- length(mixed_constraint_cols) > 0

  ## Process A_spline: keep only columns with nonzero spline entries
  keep_cols <- which(colSums(abs(A_spline)) > sqrt(.Machine$double.eps))
  if(length(keep_cols) > 0){
    A_spline <- A_spline[, keep_cols, drop = FALSE]
    qr_As <- qr(A_spline)
    if(qr_As$rank < ncol(A_spline)){
      A_spline <- qr.Q(qr_As)[, 1:qr_As$rank, drop = FALSE]
    }
  } else {
    A_spline <- cbind(rep(0, nc_spline * (K + 1)))
  }
  nca_spline <- ncol(A_spline)

  ## Process A_flat: keep only columns with nonzero flat entries
  keep_cols_flat <- which(colSums(abs(A_flat)) > sqrt(.Machine$double.eps))
  if(length(keep_cols_flat) > 0){
    A_flat_kept <- A_flat[, keep_cols_flat, drop = FALSE]
    qr_Af <- qr(A_flat_kept)
    if(qr_Af$rank < ncol(A_flat_kept)){
      A_flat_kept <- qr.Q(qr_Af)[, 1:qr_Af$rank, drop = FALSE]
    }
    A_flat <- A_flat_kept
  } else {
    A_flat <- cbind(rep(0, length(flat_rows_all)))
  }
  nca_flat <- ncol(A_flat)

  ## Constraint values for spline-only subproblem
  if(length(constraint_values) > 0){
    constraint_values_spline <- lapply(constraint_values, function(cv){
      cv[spline_cols, , drop = FALSE]
    })
  } else {
    constraint_values_spline <- constraint_values
  }

  list(X_spline = X_spline,
       X_flat   = X_flat,
       Lambda_spline = Lambda_spline,
       Lambda_flat   = Lambda_flat,
       L_part_spline = L_part_spline,
       A_spline = A_spline,
       nca_spline = nca_spline,
       constraint_values_spline = constraint_values_spline,
       spline_cols = spline_cols,
       nc_spline = nc_spline,
       nc_flat   = nc_flat,
       A_flat = A_flat,
       nca_flat = nca_flat,
       A_full = A,
       flat_rows_all = flat_rows_all,
       spline_rows_all = spline_rows_all,
       has_mixed_constraints = has_mixed_constraints,
       mixed_constraint_cols = mixed_constraint_cols)
}


#' Constrained Flat Update Given Current Spline Solution
#'
#' Solves for flat coefficients subject to the residual equality
#' constraint imposed by the full constraint system, conditional on
#' the current spline block solution.
#'
#' When the full constraint system is \eqn{\mathbf{A}^\top \boldsymbol{\beta} = \mathbf{c}},
#' and we partition \eqn{\boldsymbol{\beta}} into spline and flat blocks, the flat
#' update must satisfy:
#' \deqn{
#' \mathbf{A}_{\mathrm{flat}}^\top \boldsymbol{\beta}_{\mathrm{flat}}
#' = \mathbf{c} - \mathbf{A}_{\mathrm{spline}}^\top \boldsymbol{\beta}_{\mathrm{spline}}
#' }
#'
#' This function solves the penalized least-squares problem for flat coefficients
#' subject to this residual equality constraint using a Lagrangian approach.
#'
#' @param XfWXf_pen Penalized flat Gram matrix
#'   (\eqn{nc_f \times nc_f}).
#' @param Xfr Right-hand side cross-product for flat update
#'   (\eqn{nc_f \times 1}).
#' @param A_full Full constraint matrix (\eqn{P \times R}).
#' @param constraint_values List of constraint RHS vectors.
#' @param beta_spline List of \eqn{K+1} current spline coefficient vectors.
#' @param flat_rows_all Integer vector of flat row indices in the full P-space.
#' @param spline_rows_all Integer vector of spline row indices in the full P-space.
#' @param spline_cols Integer vector of spline column indices within each partition.
#' @param flat_cols Integer vector of flat column indices within each partition.
#' @param p_expansions Integer; columns per partition.
#' @param K Integer; number of interior knots.
#' @param nc_flat Integer; number of flat columns.
#'
#' @return Numeric vector of flat coefficients (\eqn{nc_f \times 1}).
#'
#' @keywords internal
.bf_constrained_flat_update <- function(XfWXf_pen, Xfr,
                                        A_full, constraint_values,
                                        beta_spline,
                                        flat_rows_all, spline_rows_all,
                                        spline_cols, flat_cols,
                                        p_expansions, K, nc_flat){

  ## Build the full spline coefficient vector in full P-space ordering
  beta_spline_full <- rep(0, p_expansions * (K + 1))
  for(k in 1:(K + 1)){
    rows_k <- (k - 1) * p_expansions + spline_cols
    beta_spline_full[rows_k] <- beta_spline[[k]]
  }

  ## The full constraint is: A^T beta = c
  #  Partition: A_s^T beta_s + A_f^T beta_f = c
  #  So the flat constraint is: A_f^T beta_f = c - A_s^T beta_s
  #
  #  But flat coefficients are shared across partitions, so beta_f in
  #  the full P-space has the same nc_flat values replicated K+1 times.
  #  We need to express the constraint in terms of the unique nc_flat
  #  flat coefficients.

  ## Compute the spline contribution to each constraint
  spline_contribution <- crossprod(A_full, beta_spline_full)  # R x 1

  ## Get the constraint target
  if(length(constraint_values) > 0 && any(unlist(constraint_values) != 0)){
    cv_vec <- Reduce("rbind", constraint_values)
    constraint_target <- crossprod(A_full, cv_vec)  # R x 1
  } else {
    constraint_target <- rep(0, ncol(A_full))
  }

  ## Residual that the flat block must satisfy
  flat_target <- constraint_target - spline_contribution  # R x 1

  ## Build the flat constraint matrix in terms of unique flat coefficients.
  #  For each constraint column j of A, the flat contribution is:
  #    sum over k=0..K of A[flat_rows_k, j]^T * beta_flat_unique
  #  Since beta_flat is replicated, this equals:
  #    (sum over k of A[flat_rows_k, j])^T * beta_flat_unique
  A_flat_unique <- matrix(0, nrow = nc_flat, ncol = ncol(A_full))
  for(k in 1:(K + 1)){
    rows_k <- (k - 1) * p_expansions + flat_cols
    A_flat_unique <- A_flat_unique + A_full[rows_k, , drop = FALSE]
  }
  ## A_flat_unique is nc_flat x R

  ## Only keep constraints that actually involve flat coefficients
  active_cols <- which(colSums(abs(A_flat_unique)) > sqrt(.Machine$double.eps))

  if(length(active_cols) == 0){
    ## No constraints touch flat coefficients; unconstrained update
    return(c(invert(XfWXf_pen) %**% Xfr))
  }

  A_fc <- A_flat_unique[, active_cols, drop = FALSE]  # nc_flat x n_active
  b_fc <- flat_target[active_cols]                     # n_active x 1

  ## Solve the constrained problem via KKT / Lagrangian:
  #    minimize  0.5 * beta_f^T X'X beta_f - beta_f^T g
  #    subject to  A_fc^T beta_f = b_fc
  #
  #  KKT system:
  #    [ X'X    A_fc ] [ beta_f ] = [ g    ]
  #    [ A_fc^T  0 ] [ lambda ] = [ b_fc ]

  H <- XfWXf_pen
  g <- Xfr

  n_constr <- ncol(A_fc)

  ## Build and solve KKT system
  KKT <- rbind(
    cbind(H, A_fc),
    cbind(t(A_fc), matrix(0, n_constr, n_constr))
  )
  rhs <- rbind(g, cbind(b_fc))

  ## Use a robust solve
  kkt_sol <- try(solve(KKT, rhs), silent = TRUE)

  if(inherits(kkt_sol, "try-error")){
    ## If KKT is singular, try pseudoinverse approach
    #  This can happen if constraints are redundant
    kkt_sol <- try({
      ## Unconstrained solution
      beta_unc <- c(invert(H) %**% g)
      ## Project onto constraint
      residual <- b_fc - crossprod(A_fc, beta_unc)
      ## Correction
      HinvA <- invert(H) %**% A_fc
      correction <- HinvA %**% solve(crossprod(A_fc, HinvA), residual)
      cbind(beta_unc + c(correction))
    }, silent = TRUE)

    if(inherits(kkt_sol, "try-error")){
      ## Last resort: unconstrained
      return(c(invert(H) %**% g))
    }
    return(c(kkt_sol[1:nc_flat]))
  }

  c(kkt_sol[1:nc_flat])
}


#' Lagrangian Projection for the Spline-Only Subproblem
#'
#' Projects the adjusted cross-product \eqn{\mathbf{X}_s^{\top}
#' \tilde{\mathbf{y}}} through \eqn{\mathbf{G}_s^{1/2}} and
#' removes the component along \eqn{\mathbf{G}_s^{1/2}\mathbf{A}_s}
#' to enforce the smoothness constraints in the spline block.
#' When \code{cv_spline} contains nonzero values (e.g.\ from
#' \code{no_intercept}), the corresponding contribution is added
#' back after projection.
#'
#' @param Xy_adj List of \eqn{K+1} adjusted cross-product vectors
#'   (\eqn{nc_s \times 1}).
#' @param Ghalf_cur List of \eqn{K+1} \eqn{\mathbf{G}_s^{1/2}}
#'   matrices (each \eqn{nc_s \times nc_s}).
#' @param GhalfA_cur \eqn{(nc_s \cdot (K+1)) \times nca_s} matrix
#'   formed by row-stacking \eqn{\mathbf{G}_s^{1/2,(k)}
#'   \mathbf{A}_s^{(k)}}.
#' @param cv_spline List of constraint right-hand-side vectors for
#'   the spline subproblem (may be empty).
#' @param K Integer; number of interior knots.
#' @param nc_spline Integer; number of spline columns.
#' @param A_spline Spline-only constraint matrix.
#'
#' @return List of \eqn{K+1} coefficient vectors
#'   (\eqn{nc_s \times 1}).
#'
#' @keywords internal
.bf_lagrangian_project <- function(Xy_adj, Ghalf_cur, GhalfA_cur,
                                   cv_spline, K, nc_spline,
                                   A_spline){
  GhalfXy <- cbind(unlist(lapply(1:(K + 1), function(k){
    Ghalf_cur[[k]] %**% Xy_adj[[k]]
  })))

  sc <- 1 / sqrt(K + 1)
  resids_star <- .lm.fit(GhalfA_cur * sc,
                         GhalfXy * sc)$residuals / sc

  if(length(cv_spline) > 0){
    if(any(unlist(cv_spline) != 0)){
      cv_vec <- Reduce("rbind", cv_spline)
      preds_star <- GhalfA_cur %**%
        (invert(crossprod(GhalfA_cur) * sc) %**%
           (crossprod(A_spline, cv_vec) * sc))
      resids_star <- resids_star + c(preds_star)
    }
  }

  lapply(1:(K + 1), function(k){
    rows <- ((k - 1) * nc_spline + 1):(k * nc_spline)
    Ghalf_cur[[k]] %**% cbind(resids_star[rows])
  })
}


#' Gaussian Identity + GEE: Closed-Form Solve (Case a)
#'
#' When the response is Gaussian with identity link and a working
#' correlation structure is present, the whitened system
#' \eqn{\mathbf{V}^{-1/2}\mathbf{X}} is not block-diagonal, so
#' partition-wise backfitting does not apply.
#' This function performs a single closed-form solve that replicates
#' \code{get_B} Path 1a but in the backfitting context.
#'
#' In other words, this is the dense correlated solve on the pooled
#' coefficient space. The supplement's Woodbury factorization
#' \eqn{\mathbf{G}^{-1} =
#' \mathbf{G}_{\mathrm{on}}^{-1} + \mathbf{G}_{\mathrm{off}}^{-1}}
#' is handled in \code{get_B()}, not here, so this helper always uses the
#' dense correlated solve.
#'
#' @param X List of \eqn{K+1} unwhitened design matrices.
#' @param y List of \eqn{K+1} response vectors.
#' @param K Integer; number of interior knots.
#' @param p_expansions Integer; columns per partition.
#' @param order_list List of observation-index vectors by partition.
#' @param VhalfInv \eqn{\mathbf{V}^{-1/2}} matrix in the original
#'   observation ordering.
#' @param Lambda Shared penalty matrix.
#' @param L_partition_list Partition-specific penalty matrices.
#' @param unique_penalty_per_partition Logical.
#' @param A Full constraint matrix.
#' @param constraint_values List of constraint RHS vectors.
#' @param spline_cols,flat_cols Integer vectors of column indices.
#'
#' @return A named list:
#' \describe{
#'   \item{beta_spline}{List of \eqn{K+1} spline coefficient vectors.}
#'   \item{beta_flat}{Numeric vector of flat coefficients.}
#'   \item{result}{List of \eqn{K+1} full per-partition coefficient
#'     vectors.}
#' }
#'
#' @keywords internal
.bf_case_gauss_gee <- function(X, y, K, p_expansions, order_list, VhalfInv,
                               Lambda, L_partition_list,
                               unique_penalty_per_partition,
                               A, constraint_values,
                               spline_cols, flat_cols,
                               observation_weights){

  perm_bf <- unlist(order_list)
  VhalfInv_bf <- VhalfInv[perm_bf, perm_bf]
  obs_wt_bf <- c(unlist(observation_weights))
  if(length(obs_wt_bf) == 0L){
    obs_wt_bf <- rep(1, nrow(VhalfInv_bf))
  } else if(length(obs_wt_bf) == 1L){
    obs_wt_bf <- rep(obs_wt_bf, nrow(VhalfInv_bf))
  }

  X_block_bf <- collapse_block_diagonal(X)
  y_block_bf <- cbind(unlist(y))

  X_tilde_bf <- VhalfInv_bf %**% X_block_bf
  y_tilde_bf <- VhalfInv_bf %**% y_block_bf

  Gram_bf <- crossprod(X_tilde_bf, obs_wt_bf * X_tilde_bf)
  Lambda_block_bf <- .bf_make_Lambda_block(Lambda, K,
                                           unique_penalty_per_partition,
                                           L_partition_list)
  G_bf <- invert(Gram_bf + Lambda_block_bf)
  G_bf_half <- matsqrt(G_bf)
  Xy_bf <- crossprod(X_tilde_bf, obs_wt_bf * y_tilde_bf)

  ## Lagrangian projection in full P-space
  y_star <- G_bf_half %**% Xy_bf
  X_star <- G_bf_half %**% A

  sc <- 1 / sqrt(K + 1)
  resids_bf <- .lm.fit(X_star * sc, y_star * sc)$residuals / sc

  if(length(constraint_values) > 0){
    if(any(unlist(constraint_values) != 0)){
      cv_vec <- Reduce("rbind", constraint_values)
      preds_bf <- X_star %**%
        (invert(crossprod(X_star) * sc) %**%
           (crossprod(A, cv_vec) * sc))
      resids_bf <- resids_bf + c(preds_bf)
    }
  }

  beta_block <- G_bf_half %**% cbind(resids_bf)

  result <- lapply(1:(K + 1), function(k){
    cbind(beta_block[(k - 1) * p_expansions + 1:p_expansions])
  })

  beta_spline <- lapply(result, function(b) b[spline_cols, , drop = FALSE])
  beta_flat   <- result[[1]][flat_cols]

  list(beta_spline = beta_spline,
       beta_flat   = beta_flat,
       result      = result)
}


#' Gaussian Identity Backfitting Without Correlation (Case b)
#'
#' Standard block-coordinate descent on the convex quadratic
#' objective, alternating between the spline step (Lagrangian
#' projection) and the flat step (pooled penalized regression).
#'
#' @param X_spline,X_flat Lists of per-partition submatrices.
#' @param y List of response vectors.
#' @param K Integer.
#' @param Ghalf_sp List of \eqn{\mathbf{G}_s^{1/2}} matrices.
#' @param GhalfA_sp Pre-computed \eqn{\mathbf{G}_s^{1/2}
#'   \mathbf{A}_s}.
#' @param XfXf_pen_inv Penalised inverse of the pooled flat Gram
#'   matrix.
#' @param constraint_values_spline Spline-only constraint RHS.
#' @param nc_spline,nc_flat Integers.
#' @param A_spline Spline-only constraint matrix.
#' @param tol Convergence tolerance.
#' @param max_backfit_iter Maximum iterations.
#' @param verbose Logical.
#' @param split Output of \code{.bf_split_components}, needed for
#'   mixed-constraint enforcement on the flat step.
#' @param constraint_values Full constraint RHS list.
#' @param Lambda_flat Flat penalty submatrix.
#' @param p_expansions Integer; columns per partition.
#'
#' @return A named list with \code{beta_spline} and
#'   \code{beta_flat}.
#'
#' @keywords internal
.bf_case_gauss_no_corr <- function(X_spline, X_flat, y, K,
                                   Ghalf_sp, GhalfA_sp,
                                   XfXf_pen_inv,
                                   constraint_values_spline,
                                   nc_spline, nc_flat,
                                   A_spline,
                                   tol, max_backfit_iter,
                                   verbose,
                                   split = NULL,
                                   constraint_values = list(),
                                   Lambda_flat = NULL,
                                   p_expansions = NULL){

  beta_flat <- rep(0, nc_flat)
  beta_spline <- lapply(1:(K + 1), function(k) cbind(rep(0, nc_spline)))

  ## Determine whether we need constrained flat updates
  use_constrained_flat <- (!is.null(split) &&
                             split$has_mixed_constraints &&
                             !is.null(Lambda_flat) &&
                             !is.null(p_expansions))

  for(bf_iter in 1:max_backfit_iter){

    ## Spline step
    Xy_adj <- lapply(1:(K + 1), function(k){
      y_adj_k <- y[[k]] - X_flat[[k]] %**% cbind(beta_flat)
      crossprod(X_spline[[k]], y_adj_k)
    })
    beta_spline_new <- .bf_lagrangian_project(
      Xy_adj, Ghalf_sp, GhalfA_sp,
      constraint_values_spline, K, nc_spline, A_spline)

    ## Flat step
    Xfr <- Reduce("+", lapply(1:(K + 1), function(k){
      r_k <- y[[k]] - X_spline[[k]] %**% beta_spline_new[[k]]
      crossprod(X_flat[[k]], r_k)
    }))

    if(use_constrained_flat){
      ## Constrained flat update enforcing mixed equality constraints
      XfXf <- Reduce("+", lapply(1:(K + 1), function(k){
        crossprod(X_flat[[k]])
      }))
      XfXf_pen <- XfXf + Lambda_flat

      beta_flat_new <- .bf_constrained_flat_update(
        XfWXf_pen = XfXf_pen,
        Xfr = Xfr,
        A_full = split$A_full,
        constraint_values = constraint_values,
        beta_spline = beta_spline_new,
        flat_rows_all = split$flat_rows_all,
        spline_rows_all = split$spline_rows_all,
        spline_cols = split$spline_cols,
        flat_cols = setdiff(1:p_expansions, split$spline_cols),
        p_expansions = p_expansions,
        K = K,
        nc_flat = nc_flat)
    } else {
      beta_flat_new <- c(XfXf_pen_inv %**% Xfr)
    }

    ## Convergence check
    spline_diff <- max(abs(unlist(beta_spline_new) -
                             unlist(beta_spline)))
    flat_diff   <- max(abs(beta_flat_new - beta_flat))
    beta_spline <- beta_spline_new
    beta_flat   <- beta_flat_new

    if(verbose){
      cat('  Backfit iter', bf_iter,
          '| spline diff:', format(spline_diff, digits = 4),
          '| flat diff:',   format(flat_diff,   digits = 4), '\n')
    }

    if(max(spline_diff, flat_diff) < tol) break
  }

  list(beta_spline = beta_spline,
       beta_flat   = beta_flat)
}


#' GLM Backfitting Inner Loop
#'
#' Given damped Newton-Raphson weights and response partitioned into lists,
#' runs the inner backfitting loop that alternates between a weighted
#' spline step and a weighted flat step until convergence.  Shared by
#' both Case c (GEE warm start) and Case d (GLM without GEE).
#'
#' @param X_spline,X_flat Lists of per-partition submatrices.
#' @param z_list List of \eqn{K+1} working-response vectors.
#' @param W_list List of \eqn{K+1} weight vectors.
#' @param K Integer.
#' @param Lambda_spline,Lambda_flat Penalty submatrices.
#' @param L_part_spline Partition-specific spline penalty list.
#' @param unique_penalty_per_partition Logical.
#' @param A_spline Spline-only constraint matrix.
#' @param constraint_values_spline Constraint RHS list.
#' @param nc_spline,nc_flat Integers.
#' @param beta_spline_init,beta_flat_init Initial coefficient values.
#' @param tol Convergence tolerance.
#' @param max_backfit_iter Maximum iterations.
#' @param parallel_eigen,cl,chunk_size,num_chunks,rem_chunks
#'   Parallel arguments forwarded to \code{compute_G_eigen}.
#' @param split Output of \code{.bf_split_components}, needed for
#'   mixed-constraint enforcement on the flat step.
#' @param constraint_values Full constraint RHS list.
#' @param p_expansions Integer; columns per partition.
#'
#' @return A named list with \code{beta_spline} and
#'   \code{beta_flat}.
#'
#' @keywords internal
.bf_inner_weighted <- function(X_spline, X_flat,
                               z_list, W_list, K,
                               Lambda_spline, Lambda_flat,
                               L_part_spline,
                               unique_penalty_per_partition,
                               A_spline, constraint_values_spline,
                               nc_spline, nc_flat,
                               beta_spline_init, beta_flat_init,
                               tol, max_backfit_iter,
                               parallel_eigen, cl,
                               chunk_size, num_chunks,
                               rem_chunks,
                               split = NULL,
                               constraint_values = list(),
                               p_expansions = NULL){

  beta_spline <- beta_spline_init
  beta_flat   <- beta_flat_init
  schur_zero <- lapply(1:(K + 1), function(k) 0)

  ## Determine whether we need constrained flat updates
  use_constrained_flat <- (!is.null(split) &&
                             split$has_mixed_constraints &&
                             !is.null(p_expansions))

  ## Weighted spline Gram and G
  X_gram_sp_w <- lapply(1:(K + 1), function(k){
    crossprod(X_spline[[k]] * sqrt(W_list[[k]]))
  })
  G_sp_w <- compute_G_eigen(X_gram_sp_w,
                            Lambda_spline, K,
                            parallel_eigen, cl,
                            chunk_size, num_chunks, rem_chunks,
                            gaussian(),
                            unique_penalty_per_partition,
                            L_part_spline,
                            keep_G = TRUE,
                            schur_zero)
  Ghalf_sp_w <- G_sp_w$Ghalf

  GhalfA_sp_w <- Reduce("rbind", lapply(1:(K + 1), function(k){
    rows <- ((k - 1) * nc_spline + 1):(k * nc_spline)
    Ghalf_sp_w[[k]] %**% A_spline[rows, , drop = FALSE]
  }))

  for(bf_iter in 1:max_backfit_iter){

    ## Spline step (weighted)
    Xy_adj <- lapply(1:(K + 1), function(k){
      z_adj_k <- z_list[[k]] - X_flat[[k]] %**% cbind(beta_flat)
      crossprod(X_spline[[k]] * W_list[[k]], cbind(z_adj_k))
    })
    beta_spline_new <- .bf_lagrangian_project(
      Xy_adj, Ghalf_sp_w, GhalfA_sp_w,
      constraint_values_spline, K, nc_spline, A_spline)

    ## Flat step (weighted, pooled)
    XfWXf <- Reduce("+", lapply(1:(K + 1), function(k){
      crossprod(X_flat[[k]] * sqrt(W_list[[k]]))
    }))
    XfWXf_pen <- XfWXf + Lambda_flat
    Xfr <- Reduce("+", lapply(1:(K + 1), function(k){
      r_k <- z_list[[k]] - X_spline[[k]] %**% beta_spline_new[[k]]
      crossprod(X_flat[[k]] * W_list[[k]], cbind(r_k))
    }))

    if(use_constrained_flat){
      ## Constrained flat update enforcing mixed equality constraints
      beta_flat_new <- .bf_constrained_flat_update(
        XfWXf_pen = XfWXf_pen,
        Xfr = Xfr,
        A_full = split$A_full,
        constraint_values = constraint_values,
        beta_spline = beta_spline_new,
        flat_rows_all = split$flat_rows_all,
        spline_rows_all = split$spline_rows_all,
        spline_cols = split$spline_cols,
        flat_cols = setdiff(1:p_expansions, split$spline_cols),
        p_expansions = p_expansions,
        K = K,
        nc_flat = nc_flat)
    } else {
      XfWXf_pen_inv <- invert(XfWXf_pen)
      beta_flat_new <- c(XfWXf_pen_inv %**% Xfr)
    }

    flat_diff <- max(abs(beta_flat_new - beta_flat))
    beta_spline <- beta_spline_new
    beta_flat   <- beta_flat_new

    if(flat_diff < tol) break
  }

  list(beta_spline = beta_spline,
       beta_flat   = beta_flat)
}


#' Damped SQP Refinement on the Full (Optionally Whitened) System
#'
#' Runs a damped sequential quadratic programming loop using
#' \code{quadprog::solve.QP} at each iteration, enforcing all
#' smoothness equalities, flat-equality constraints, and any
#' user-supplied inequality constraints.
#'
#' This function serves two roles within \code{\link{blockfit_solve}}:
#' \enumerate{
#'   \item \strong{GEE Stage 2} (Case c): operates on the whitened
#'     system \eqn{\tilde{\mathbf{X}} = \mathbf{V}^{-1/2}
#'     \mathbf{X}_{\mathrm{block}}}, initialized from the damped Newton-Raphson
#'     backfitting warm start.
#'   \item \strong{Non-GEE SQP refinement}: operates on the
#'     unwhitened block-diagonal design, applying inequality
#'     constraints after backfitting convergence.
#' }
#'
#' @param X_design \eqn{N \times P} design matrix (whitened for GEE,
#'   unwhitened otherwise).
#' @param y_design \eqn{N \times 1} response vector (whitened for
#'   GEE, unwhitened otherwise).
#' @param X_block_raw Unwhitened block-diagonal design matrix (used
#'   for computing the linear predictor on the original scale).
#'   For the non-GEE path this equals \code{X_design}.
#' @param beta_init \eqn{P \times 1} initial coefficient vector.
#' @param Lambda_block \eqn{P \times P} block-diagonal penalty.
#' @param qp_Amat_combined Combined constraint matrix (equalities
#'   then inequalities).
#' @param qp_bvec_combined Combined constraint RHS.
#' @param qp_meq_combined Number of leading equality constraints.
#' @param K,p_expansions Integers.
#' @param family GLM family object.
#' @param order_list Observation-index lists.
#' @param glm_weight_function,schur_correction_function,
#'   qp_score_function Functions.
#' @param need_dispersion_for_estimation Logical.
#' @param dispersion_function Function.
#' @param observation_weights List or vector of weights.
#' @param iterate Logical; if FALSE, takes at most two steps.
#' @param tol Convergence tolerance.
#' @param VhalfInv,VhalfInv_perm \eqn{\mathbf{V}^{-1/2}} matrices
#'   (original and permuted ordering).  NULL for non-GEE path.
#' @param is_gee Logical; TRUE when called from the GEE refinement
#'   path (affects deviance computation).
#' @param deviance_fun Function computing deviance; either
#'   \code{.bf_deviance} or \code{.bf_gee_deviance}.
#' @param X_partitions List of per-partition design matrices
#'   (unwhitened); needed for Schur correction in non-GEE path.
#' @param y_partitions List of per-partition response vectors
#'   (unwhitened); needed for Schur correction in non-GEE path.
#' @param verbose Logical.
#' @param ... Forwarded to weight, dispersion, and score functions.
#'
#' @return A named list with elements \code{beta_block} (final
#'   coefficient vector), \code{result} (per-partition coefficient
#'   list), \code{last_qp_sol} (last successful \code{solve.QP}
#'   output or NULL), \code{converged} (logical), and
#'   \code{final_deviance} (scalar).
#'
#' @keywords internal
.bf_sqp_loop <- function(X_design, y_design, X_block_raw,
                         beta_init, Lambda_block,
                         qp_Amat_combined, qp_bvec_combined,
                         qp_meq_combined,
                         K, p_expansions, family, order_list,
                         glm_weight_function,
                         schur_correction_function,
                         qp_score_function,
                         need_dispersion_for_estimation,
                         dispersion_function,
                         observation_weights,
                         iterate, tol,
                         VhalfInv, VhalfInv_perm,
                         is_gee, deviance_fun,
                         X_partitions, y_partitions,
                         verbose, ...){

  beta_block <- beta_init
  damp_cnt   <- 0
  master_cnt <- 0
  err <- Inf
  converged <- FALSE
  last_qp_sol <- NULL

  XB <- X_block_raw %**% beta_block
  y_block <- if(is_gee) cbind(unlist(y_partitions)) else y_design

  while(err > tol & damp_cnt < 10 & master_cnt < 100){
    master_cnt <- master_cnt + 1
    damp <- 2^(-(damp_cnt))

    ## Dispersion
    if(need_dispersion_for_estimation){
      dispersion_temp <- dispersion_function(
        mu = family$linkinv(XB),
        y = y_block,
        order_indices = unlist(order_list),
        family = family,
        observation_weights = unlist(observation_weights),
        VhalfInv = VhalfInv,
        ...
      )
    } else {
      dispersion_temp <- 1
    }

    ## Working weights
    W <- c(glm_weight_function(
      family$linkinv(XB), y_block,
      unlist(order_list), family, dispersion_temp,
      unlist(observation_weights), ...
    ))
    W <- pmax(W, sqrt(.Machine$double.eps))

    ## Schur correction
    result_temp <- lapply(1:(K + 1), function(k){
      cbind(beta_block[(k - 1) * p_expansions + 1:p_expansions])
    })

    if(is_gee){
      schur_correction <- schur_correction_function(
        list(X_block_raw), list(y_block),
        list(cbind(unlist(result_temp))),
        dispersion_temp, list(unlist(order_list)),
        0, family, unlist(observation_weights), ...
      )
    } else {
      schur_correction <- schur_correction_function(
        X_partitions, y_partitions,
        result_temp, dispersion_temp,
        order_list, K, family,
        observation_weights, ...
      )
    }

    if(!(any(unlist(schur_correction) != 0))){
      schur_correction_coll <- 0
    } else {
      schur_correction_coll <- collapse_block_diagonal(schur_correction)
    }

    ## Information matrix
    info <- crossprod(X_design, W * X_design) +
      Lambda_block + schur_correction_coll

    sc <- sqrt(mean(abs(info)))

    ## Score
    if(is_gee){
      qp_score <- qp_score_function(
        X_design, y_design,
        VhalfInv_perm %**% cbind(family$linkinv(c(XB))),
        unlist(order_list), dispersion_temp,
        VhalfInv_perm, unlist(observation_weights), ...
      )
    } else {
      qp_score <- qp_score_function(
        X_design, y_design,
        cbind(family$linkinv(XB)),
        unlist(order_list), dispersion_temp,
        NULL, unlist(observation_weights), ...
      )
    }

    ## Solve QP
    qp_sol <- try({quadprog::solve.QP(
      Dmat = info / sc,
      dvec = (qp_score -
                Lambda_block %**% beta_block +
                info %**% beta_block) / sc,
      Amat = qp_Amat_combined,
      bvec = qp_bvec_combined,
      meq  = qp_meq_combined
    )}, silent = TRUE)

    if(any(inherits(qp_sol, 'try-error'))){
      if(verbose){
        cat("    SQP iter", master_cnt,
            "- solve.QP failed, retaining current\n")
      }
      beta_new <- beta_block
    } else {
      last_qp_sol <- qp_sol
      beta_new    <- cbind(qp_sol$solution)
    }

    ## Step acceptance
    if(!iterate & master_cnt > 1){
      beta_block <- beta_new
      converged  <- TRUE
      damp_cnt   <- 11
      master_cnt <- 101
      err <- tol - 1
    } else {
      beta_new_damped <- (1 - damp) * beta_block + damp * beta_new
      XB <- X_block_raw %**% beta_new_damped

      if(need_dispersion_for_estimation){
        dispersion_temp <- dispersion_function(
          mu = family$linkinv(XB),
          y = y_block,
          order_indices = unlist(order_list),
          family = family,
          observation_weights = unlist(observation_weights),
          VhalfInv = VhalfInv,
          ...
        )
      } else {
        dispersion_temp <- 1
      }

      W <- c(glm_weight_function(
        family$linkinv(XB), y_block,
        unlist(order_list), family, dispersion_temp,
        unlist(observation_weights), ...
      ))
      W <- pmax(W, sqrt(.Machine$double.eps))

      if(is_gee){
        err_new <- deviance_fun(
          y_block, family$linkinv(c(XB)), W,
          unlist(order_list), unlist(observation_weights),
          VhalfInv_perm, family, ...
        )
      } else {
        err_new <- deviance_fun(
          y_block, family$linkinv(c(XB)),
          unlist(observation_weights),
          unlist(order_list), family, ...
        )
      }

      if(is.null(err_new) | is.na(err_new) | !is.finite(err_new)){
        damp_cnt <- damp_cnt + 1
      } else if(err_new <= err){
        prev_err <- err
        err <- err_new
        abs_diff <- max(abs(beta_new_damped - beta_block))
        beta_block <- beta_new_damped
        damp_cnt <- max(0, damp_cnt - 1)

        if((abs_diff < tol) &
           (prev_err - err < tol) &
           (master_cnt > 5)){
          converged  <- TRUE
          damp_cnt   <- 11
          master_cnt <- 101
          err <- tol - 1
        }
      } else {
        damp_cnt <- damp_cnt + 1
      }
    }

    if(verbose & (master_cnt <= 100)){
      cat("    SQP iter", master_cnt,
          "| deviance:", format(err, digits = 6),
          "| damp:", format(damp, digits = 4), "\n")
    }

    XB <- X_block_raw %**% beta_block
  }

  result <- lapply(1:(K + 1), function(k){
    cbind(beta_block[1:p_expansions + (k - 1) * p_expansions])
  })

  list(beta_block = beta_block,
       result     = result,
       last_qp_sol = last_qp_sol,
       converged  = converged,
       final_deviance = err)
}


#' Assemble qp_info from a solve.QP Solution
#'
#' Thin wrapper around \code{.solver_assemble_qp_info()}.
#' Keeping the local helper name avoids changing the public return shape
#' or downstream references inside \code{blockfit_solve()}.
#'
#' @param last_qp_sol Output of \code{quadprog::solve.QP}, or NULL.
#' @param beta_block Final coefficient vector.
#' @param qp_Amat_combined Combined constraint matrix.
#' @param qp_bvec_combined Combined constraint RHS.
#' @param qp_meq_combined Number of equalities.
#' @param converged Logical.
#' @param final_deviance Scalar.
#' @param info_matrix Information matrix (non-GEE path only; NULL
#'   otherwise).
#'
#' @return A list suitable for the \code{qp_info} slot, or NULL if
#'   \code{last_qp_sol} is NULL.
#'
#' @keywords internal
.bf_assemble_qp_info <- function(last_qp_sol, beta_block,
                                 qp_Amat_combined,
                                 qp_bvec_combined,
                                 qp_meq_combined,
                                 converged, final_deviance,
                                 info_matrix = NULL){
  .solver_assemble_qp_info(
    last_qp_sol = last_qp_sol,
    beta_block = beta_block,
    qp_Amat_combined = qp_Amat_combined,
    qp_bvec_combined = qp_bvec_combined,
    qp_meq_combined = qp_meq_combined,
    converged = converged,
    final_deviance = final_deviance,
    info_matrix = info_matrix
  )
}


#' GLM + GEE Two-Stage Estimation (Case c)
#'
#' Stage 1 runs damped Newton-Raphson backfitting on the unwhitened link-scale working
#' response to produce a warm start.  Stage 2 refines via damped SQP
#' on the full whitened system, replicating \code{get_B} Path 1b.
#'
#' @inheritParams blockfit_solve
#' @param split Output of \code{.bf_split_components}.
#' @param schur_zero List of zeros (one per partition).
#'
#' @return A named list with \code{result}, \code{beta_spline},
#'   \code{beta_flat}, and \code{qp_info}.
#'
#' @keywords internal
.bf_case_glm_gee <- function(X, y, K, p_expansions, flat_cols,
                             split, family, order_list,
                             glm_weight_function,
                             schur_correction_function,
                             qp_score_function,
                             need_dispersion_for_estimation,
                             dispersion_function,
                             observation_weights,
                             iterate, tol, max_backfit_iter,
                             parallel_eigen, cl,
                             chunk_size, num_chunks, rem_chunks,
                             schur_zero,
                             quadprog, qp_Amat, qp_bvec, qp_meq,
                             Lambda, L_partition_list,
                             unique_penalty_per_partition,
                             A, constraint_values,
                             Vhalf, VhalfInv,
                             verbose, ...){

  spline_cols <- split$spline_cols
  nc_spline <- split$nc_spline
  nc_flat   <- split$nc_flat

  beta_flat   <- rep(0, nc_flat)
  beta_spline <- lapply(1:(K + 1), function(k) cbind(rep(0, nc_spline)))

  dnr_err_ws <- Inf
  damp_cnt_ws  <- 0
  disp_ws      <- 1

  ## Stage 1: damped Newton-Raphson backfitting warm start
  for(dnr_iter_ws in 1:50){

    eta_ws <- unlist(lapply(1:(K + 1), function(k){
      c(split$X_spline[[k]] %**% beta_spline[[k]] +
          split$X_flat[[k]] %**% cbind(beta_flat))
    }))
    mu_ws  <- family$linkinv(eta_ws)
    mu_eta_ws <- .bf_get_mu_eta(eta_ws, family)
    mu_eta_ws <- ifelse(abs(mu_eta_ws) < sqrt(.Machine$double.eps),
                        sqrt(.Machine$double.eps), mu_eta_ws)

    y_vec_ws <- unlist(y)
    z_ws <- eta_ws + (y_vec_ws - mu_ws) / mu_eta_ws

    if(need_dispersion_for_estimation){
      disp_ws <- dispersion_function(
        mu = mu_ws, y = y_vec_ws,
        order_indices = unlist(order_list),
        family = family,
        observation_weights = unlist(observation_weights),
        VhalfInv = NULL, ...)
    }

    W_ws <- c(glm_weight_function(mu_ws, y_vec_ws,
                                  unlist(order_list),
                                  family, disp_ws,
                                  unlist(observation_weights), ...))
    W_ws <- pmax(W_ws, sqrt(.Machine$double.eps))

    ## Partition z and W
    idx_ws <- 0L
    z_list_ws <- W_list_ws <- vector("list", K + 1)
    for(k in 1:(K + 1)){
      nk <- length(y[[k]])
      rows_ws <- (idx_ws + 1):(idx_ws + nk)
      z_list_ws[[k]] <- z_ws[rows_ws]
      W_list_ws[[k]] <- W_ws[rows_ws]
      idx_ws <- idx_ws + nk
    }

    ## Inner weighted backfitting
    inner <- .bf_inner_weighted(
      split$X_spline, split$X_flat,
      z_list_ws, W_list_ws, K,
      split$Lambda_spline, split$Lambda_flat,
      split$L_part_spline,
      unique_penalty_per_partition,
      split$A_spline, split$constraint_values_spline,
      nc_spline, nc_flat,
      beta_spline, beta_flat,
      tol, max_backfit_iter,
      parallel_eigen, cl,
      chunk_size, num_chunks, rem_chunks,
      split = split,
      constraint_values = constraint_values,
      p_expansions = p_expansions)

    beta_spline <- inner$beta_spline
    beta_flat   <- inner$beta_flat

    ## Damped Newton-Raphson convergence
    eta_new_ws <- unlist(lapply(1:(K + 1), function(k){
      c(split$X_spline[[k]] %**% beta_spline[[k]] +
          split$X_flat[[k]] %**% cbind(beta_flat))
    }))
    mu_new_ws <- family$linkinv(eta_new_ws)
    err_new_ws <- .bf_deviance(y_vec_ws, mu_new_ws,
                               unlist(observation_weights),
                               unlist(order_list), family, ...)

    if(verbose){
      cat('  GEE warm-start DNR iter', dnr_iter_ws,
          '| deviance:', format(err_new_ws, digits = 6), '\n')
    }

    if(is.na(err_new_ws) | !is.finite(err_new_ws)){
      damp_cnt_ws <- damp_cnt_ws + 1
      if(damp_cnt_ws >= 10) break
    } else if(err_new_ws <= dnr_err_ws){
      prev_dnr_ws_err <- dnr_err_ws
      dnr_err_ws <- err_new_ws
      damp_cnt_ws <- 0
      if(abs(prev_dnr_ws_err - dnr_err_ws) < tol &
         dnr_iter_ws > 5) break
    } else {
      damp_cnt_ws <- damp_cnt_ws + 1
      if(damp_cnt_ws >= 10) break
    }

    if(!iterate & dnr_iter_ws >= 2) break
  }

  ## Assemble warm start into full block vector
  result_ws <- lapply(1:(K + 1), function(k){
    b <- rep(0, p_expansions)
    b[spline_cols] <- beta_spline[[k]]
    b[flat_cols]   <- beta_flat
    cbind(b)
  })
  beta_block <- cbind(unlist(result_ws))

  ## Stage 2: Damped SQP on full whitened system
  if(verbose) cat("  GEE full-system refinement (Path 1b)\n")

  perm <- unlist(order_list)
  VhalfInv_perm <- VhalfInv[perm, perm]

  X_block   <- collapse_block_diagonal(X)
  y_block   <- cbind(unlist(y))
  X_tilde   <- VhalfInv_perm %**% X_block
  y_tilde   <- VhalfInv_perm %**% y_block
  Lambda_block <- .bf_make_Lambda_block(Lambda, K,
                                        unique_penalty_per_partition,
                                        L_partition_list)

  ## Combined constraint matrix
  if(quadprog && !is.null(qp_Amat) && ncol(cbind(qp_Amat)) > 0){
    qp_Amat_combined <- cbind(A, qp_Amat)
    qp_bvec_combined <- c(rep(0, ncol(A)), qp_bvec)
    qp_meq_combined  <- ncol(A) + qp_meq
  } else {
    qp_Amat_combined <- A
    qp_bvec_combined <- rep(0, ncol(A))
    qp_meq_combined  <- ncol(A)
  }

  if(length(constraint_values) > 0){
    cv_vec <- Reduce("rbind", constraint_values)
    constr_rhs <- crossprod(A, cv_vec)
    if(length(constr_rhs) == ncol(A)){
      qp_bvec_combined[1:ncol(A)] <- c(constr_rhs)
    }
  }

  ## [Diagnostic 2026-03-05] Verify qp_Amat_combined dimensions and
  #  equality count for GEE + blockfit path.
  #
  #  Expected:
  #    qp_Amat_combined has nrow = P = p_expansions * (K+1)
  #    qp_meq_combined  = ncol(A) + qp_meq
  #    ncol(qp_Amat_combined) = ncol(A) + ncol(qp_Amat)  [if qp present]
  #                           = ncol(A)                    [if no ineq]
  #
  #  If qp_meq_combined > ncol(qp_Amat_combined), solve.QP will fail
  #  silently or produce garbage Lagrangians.
  #
  #  If ncol(A) doesn't match the number of smoothness constraints
  #  (should be ~K * p_expansions - K for typical models), something
  #  upstream changed A unexpectedly.
  if(verbose){
    cat("  GEE SQP setup:",
        "nrow(qp_Amat_combined) =", nrow(qp_Amat_combined),
        "| ncol(qp_Amat_combined) =", ncol(qp_Amat_combined),
        "| qp_meq_combined =", qp_meq_combined,
        "| ncol(A) =", ncol(A),
        "| qp_meq =", ifelse(is.null(qp_meq), "NULL", qp_meq),
        "| P =", p_expansions * (K + 1), "\n")
    if(qp_meq_combined > ncol(qp_Amat_combined)){
      warning("\n\t [BUG] qp_meq_combined (", qp_meq_combined,
              ") > ncol(qp_Amat_combined) (", ncol(qp_Amat_combined),
              "). solve.QP will treat inequality columns as equalities.\n")
    }
  }

  sqp <- .bf_sqp_loop(
    X_design = X_tilde,
    y_design = y_tilde,
    X_block_raw = X_block,
    beta_init = beta_block,
    Lambda_block = Lambda_block,
    qp_Amat_combined = qp_Amat_combined,
    qp_bvec_combined = qp_bvec_combined,
    qp_meq_combined  = qp_meq_combined,
    K = K, p_expansions = p_expansions, family = family,
    order_list = order_list,
    glm_weight_function = glm_weight_function,
    schur_correction_function = schur_correction_function,
    qp_score_function = qp_score_function,
    need_dispersion_for_estimation = need_dispersion_for_estimation,
    dispersion_function = dispersion_function,
    observation_weights = observation_weights,
    iterate = iterate, tol = tol,
    VhalfInv = VhalfInv, VhalfInv_perm = VhalfInv_perm,
    is_gee = TRUE, deviance_fun = .bf_gee_deviance,
    X_partitions = X, y_partitions = y,
    verbose = verbose, ...)

  qp_info <- .bf_assemble_qp_info(
    sqp$last_qp_sol, sqp$beta_block,
    qp_Amat_combined, qp_bvec_combined, qp_meq_combined,
    sqp$converged, sqp$final_deviance)

  ## [Diagnostic 2026-03-05] Verify Amat_active from GEE path.
  #  Should have ncol = qp_meq_combined + (number of binding inequalities).
  #  The number of binding inequalities should be small relative to total.
  #  If ncol(Amat_active) == ncol(qp_Amat_combined), all constraints
  #  are "active" which likely indicates the old iact bug resurfacing.
  if(verbose && !is.null(qp_info)){
    n_ineq_active <- ncol(qp_info$Amat_active) - qp_meq_combined
    cat("  GEE qp_info:",
        "ncol(Amat_active) =", ncol(qp_info$Amat_active),
        "| n_equalities =", qp_meq_combined,
        "| n_active_ineq =", n_ineq_active,
        "| lagrangian_nonzero =",
        sum(abs(qp_info$lagrangian) > sqrt(.Machine$double.eps)),
        "\n")
    if(n_ineq_active > 0.5 * (ncol(qp_Amat_combined) - qp_meq_combined)){
      warning("\n\t [SUSPECT] More than half of inequality constraints ",
              "are active (", n_ineq_active, " of ",
              ncol(qp_Amat_combined) - qp_meq_combined,
              "). Check .bf_assemble_qp_info logic.\n")
    }
  }

  list(result = sqp$result,
       beta_spline = lapply(sqp$result, function(b) b[spline_cols, , drop = FALSE]),
       beta_flat   = sqp$result[[1]][flat_cols],
       qp_info     = qp_info)
}


#' GLM Without GEE: Damped Newton-Raphson + Backfitting (Case d)
#'
#' Damped Newton-Raphson outer loop wrapping the weighted backfitting inner loop.
#' Each damped Newton-Raphson step computes weights from the
#' current linear predictor, then calls \code{.bf_inner_weighted}.
#'
#' @inheritParams blockfit_solve
#' @param split Output of \code{.bf_split_components}.
#' @param schur_zero List of zeros.
#'
#' @return A named list with \code{beta_spline} and
#'   \code{beta_flat}.
#'
#' @keywords internal
.bf_case_glm_no_corr <- function(X, y, K, p_expansions,
                                 split, family, order_list,
                                 glm_weight_function,
                                 need_dispersion_for_estimation,
                                 dispersion_function,
                                 observation_weights,
                                 iterate, tol, max_backfit_iter,
                                 parallel_eigen, cl,
                                 chunk_size, num_chunks,
                                 rem_chunks, schur_zero,
                                 unique_penalty_per_partition,
                                 VhalfInv,
                                 verbose,
                                 constraint_values = list(),
                                 ...){

  nc_spline <- split$nc_spline
  nc_flat   <- split$nc_flat

  beta_flat   <- rep(0, nc_flat)
  beta_spline <- lapply(1:(K + 1), function(k) cbind(rep(0, nc_spline)))

  dnr_err  <- Inf
  damp_cnt   <- 0
  disp_temp  <- 1

  for(dnr_iter in 1:100){

    eta <- unlist(lapply(1:(K + 1), function(k){
      c(split$X_spline[[k]] %**% beta_spline[[k]] +
          split$X_flat[[k]] %**% cbind(beta_flat))
    }))
    mu  <- family$linkinv(eta)

    mu_eta <- .bf_get_mu_eta(eta, family)
    mu_eta <- ifelse(abs(mu_eta) < sqrt(.Machine$double.eps),
                     sqrt(.Machine$double.eps), mu_eta)

    y_vec <- unlist(y)
    z <- eta + (y_vec - mu) / mu_eta

    if(need_dispersion_for_estimation){
      disp_temp <- dispersion_function(
        mu = mu, y = y_vec,
        order_indices = unlist(order_list),
        family = family,
        observation_weights = unlist(observation_weights),
        VhalfInv = VhalfInv, ...)
    }

    W <- c(glm_weight_function(mu, y_vec, unlist(order_list),
                               family, disp_temp,
                               unlist(observation_weights), ...))
    W <- pmax(W, sqrt(.Machine$double.eps))

    ## Partition z and W
    idx <- 0L
    z_list <- W_list <- vector("list", K + 1)
    for(k in 1:(K + 1)){
      nk <- length(y[[k]])
      rows <- (idx + 1):(idx + nk)
      z_list[[k]] <- z[rows]
      W_list[[k]] <- W[rows]
      idx <- idx + nk
    }

    ## Inner weighted backfitting
    inner <- .bf_inner_weighted(
      split$X_spline, split$X_flat,
      z_list, W_list, K,
      split$Lambda_spline, split$Lambda_flat,
      split$L_part_spline,
      unique_penalty_per_partition,
      split$A_spline, split$constraint_values_spline,
      nc_spline, nc_flat,
      beta_spline, beta_flat,
      tol, max_backfit_iter,
      parallel_eigen, cl,
      chunk_size, num_chunks, rem_chunks,
      split = split,
      constraint_values = constraint_values,
      p_expansions = p_expansions)

    beta_spline <- inner$beta_spline
    beta_flat   <- inner$beta_flat

    ## Damped Newton-Raphson convergence
    eta_new <- unlist(lapply(1:(K + 1), function(k){
      c(split$X_spline[[k]] %**% beta_spline[[k]] +
          split$X_flat[[k]] %**% cbind(beta_flat))
    }))
    mu_new <- family$linkinv(eta_new)

    err_new <- .bf_deviance(y_vec, mu_new, unlist(observation_weights),
                            unlist(order_list), family, ...)

    if(verbose){
      cat('  Backfitting DNR iter', dnr_iter,
          '| deviance:', format(err_new, digits = 6), '\n')
    }

    if(is.na(err_new) | !is.finite(err_new)){
      damp_cnt <- damp_cnt + 1
      if(damp_cnt >= 10) break
    } else if(err_new <= dnr_err){
      prev_dnr_err <- dnr_err
      dnr_err <- err_new
      damp_cnt <- 0
      if(abs(prev_dnr_err - dnr_err) < tol & dnr_iter > 5) break
    } else {
      damp_cnt <- damp_cnt + 1
      if(damp_cnt >= 10) break
    }

    if(!iterate & dnr_iter >= 2) break
  }

  list(beta_spline = beta_spline,
       beta_flat   = beta_flat)
}


#' Backfitting Solver for Blockfit Models
#'
#' @description
#' Fits models with mixed spline and non-interactive linear ("flat") terms
#' using an iterative backfitting approach. Flat terms receive a single
#' shared coefficient across partitions (rather than \eqn{K+1}
#' partition-specific coefficients constrained to equality), reducing the
#' effective parameter count and improving efficiency when the number of
#' flat terms is large relative to spline terms.
#'
#' The backfitting loop alternates between:
#' \enumerate{
#'
#' \item \strong{Spline step}
#'
#' Fit spline terms on the response adjusted for the current flat
#' contribution using a constrained penalized least squares solve.
#'
#' If \eqn{\mathbf{y}_k} is the response vector for partition \eqn{k},
#' \eqn{\mathbf{Z}_k} the spline design matrix,
#' \eqn{\mathbf{X}_{\mathrm{flat}}^{(k)}} the flat design matrix,
#' and \eqn{\mathbf{v}} the flat coefficient vector, then
#'
#' \deqn{
#' \boldsymbol{\beta}_{\mathrm{spline}}^{(k)} =
#' \arg\min_{\boldsymbol{\beta}}
#' \left\|
#' \mathbf{y}_k
#' - \mathbf{Z}_k \boldsymbol{\beta}
#' - \mathbf{X}_{\mathrm{flat}}^{(k)} \mathbf{v}
#' \right\|^2
#' +
#' \boldsymbol{\beta}^{\top}
#' \mathbf{\Lambda}_{\mathrm{spline}}
#' \boldsymbol{\beta}
#' }
#'
#' subject to
#'
#' \deqn{
#' \mathbf{A}_{\mathrm{spline}}^{\top}
#' \boldsymbol{\beta}
#' =
#' \mathbf{c}_{\mathrm{spline}}.
#' }
#'
#' \item \strong{Flat step}
#'
#' Update flat coefficients via pooled penalized regression on residuals,
#' subject to the residual equality constraint from any mixed constraints:
#'
#' \deqn{
#' \mathbf{v}
#' =
#' \arg\min_{\mathbf{v}}
#' \sum_{k=0}^{K}
#' \left\|
#' \mathbf{y}_k
#' -
#' \mathbf{Z}_k
#' \boldsymbol{\beta}_{\mathrm{spline}}^{(k)}
#' -
#' \mathbf{X}_{\mathrm{flat}}^{(k)}
#' \mathbf{v}
#' \right\|^2
#' +
#' \mathbf{v}^{\top}
#' \mathbf{\Lambda}_{\mathrm{flat}}
#' \mathbf{v}
#' }
#'
#' subject to
#'
#' \deqn{
#' \tilde{\mathbf{A}}_{\mathrm{flat}}^{\top}
#' \mathbf{v}
#' =
#' \mathbf{c}
#' -
#' \mathbf{A}_{\mathrm{spline}}^{\top}
#' \boldsymbol{\beta}_{\mathrm{spline}}.
#' }
#' }
#'
#' Four estimation paths are selected automatically based on the model
#' configuration:
#' \describe{
#'   \item{Case (a)}{Gaussian identity + GEE: closed-form solve via
#'     \code{.bf_case_gauss_gee}.}
#'   \item{Case (b)}{Gaussian identity, no correlation: standard
#'     backfitting via \code{.bf_case_gauss_no_corr}.}
#'   \item{Case (c)}{GLM + GEE: two-stage (damped Newton-Raphson warm start then
#'     damped SQP) via \code{.bf_case_glm_gee}.}
#'   \item{Case (d)}{GLM without GEE: damped Newton-Raphson + backfitting via
#'     \code{.bf_case_glm_no_corr}.}
#' }
#'
#' When \code{quadprog = TRUE} without GEE, inequality constraints are
#' enforced after backfitting convergence.  The constraint handling
#' method is selected automatically by inspecting the sparsity pattern
#' of \code{qp_Amat}: if every column has nonzeros in only one
#' partition block (e.g.\ derivative sign or range constraints),
#' a partition-wise active-set method is used via
#' \code{.active_set_refine}, avoiding the dense \eqn{P \times P}
#' system.  If any column spans multiple partition blocks (e.g.\
#' cross-knot monotonicity), or if the active-set method does not
#' converge, the dense SQP fallback via \code{.bf_sqp_loop} is used.
#' GEE paths always use the dense system.
#'
#' After convergence, coefficients are reassembled into the standard
#' per-partition format (flat coefficients replicated across partitions)
#' for compatibility with downstream inference.
#'
#' @param X List of \eqn{K+1} design matrices
#'   (\eqn{N_k \times p_expansions}). Unwhitened even when GEE.
#' @param y List of \eqn{K+1} response vectors. Unwhitened even when GEE.
#' @param flat_cols Integer vector indicating flat columns of
#'   \eqn{\mathbf{X}^{(k)}}.
#' @param K Integer; number of interior knots.
#' @param p_expansions Integer; number of coefficients per partition.
#' @param Lambda \eqn{p_expansions \times p_expansions} penalty matrix.
#' @param L_partition_list List of partition-specific penalty matrices.
#' @param unique_penalty_per_partition Logical.
#' @param A Full \eqn{P \times R_constraints} constraint matrix.
#' @param R_constraints Integer; number of columns of \eqn{\mathbf{A}}.
#' @param constraint_values List of constraint right-hand sides.
#' @param X_gram List of Gram matrices
#'   \eqn{\mathbf{X}_k^{\top} \mathbf{X}_k}.
#' @param Ghalf_full,GhalfInv_full Lists of
#'   \eqn{\mathbf{G}^{1/2}} and \eqn{\mathbf{G}^{-1/2}} matrices.
#' @param family GLM family object.
#' @param order_list List of observation index vectors by partition.
#' @param glm_weight_function GLM weight function.
#' @param schur_correction_function Schur complement correction function.
#' @param need_dispersion_for_estimation Logical.
#' @param dispersion_function Dispersion estimation function.
#' @param observation_weights List of observation weights.
#' @param homogenous_weights Logical.
#' @param iterate Logical; if FALSE, single pass (no iteration).
#' @param tol Convergence tolerance.
#' @param parallel_eigen,cl,chunk_size,num_chunks,rem_chunks Parallel arguments.
#' @param return_G_getB Logical.
#' @param quadprog Logical; apply inequality constraint refinement if TRUE.
#' @param qp_Amat Inequality constraint matrix for
#'   \code{quadprog::solve.QP}.
#' @param qp_bvec Inequality constraint right-hand side.
#' @param qp_meq Number of leading equality constraints.
#' @param qp_score_function Score function for QP subproblem.
#' @param keep_weighted_Lambda Logical.
#' @param max_backfit_iter Integer.
#' @param Vhalf Square root of the working correlation matrix in the
#'   original observation ordering.
#' @param VhalfInv Inverse square root of the working correlation matrix
#'   in the original observation ordering.  When both are non-\code{NULL},
#'   the GEE paths are used.
#' @param include_warnings Logical.
#' @param verbose Logical.
#' @param ... Additional arguments passed to weight and dispersion functions.
#'
#' @return A list with elements:
#' \describe{
#'   \item{B}{
#'     List of \eqn{K+1} coefficient vectors
#'     (\eqn{p_expansions \times 1}), flat coefficients replicated across partitions.
#'   }
#'   \item{G_list}{
#'     List containing \eqn{\mathbf{G}},
#'     \eqn{\mathbf{G}^{1/2}}, and
#'     \eqn{\mathbf{G}^{-1/2}},
#'     each a list of \eqn{K+1}
#'     \eqn{p_expansions \times p_expansions} matrices,
#'     or NULL if \code{return_G_getB} is FALSE.
#'
#'     \eqn{\mathbf{G}} satisfies
#'     \eqn{\mathbf{G}
#'     =
#'     \mathbf{G}^{1/2}
#'     (\mathbf{G}^{1/2})^{\top}} exactly.
#'
#'     When a correlation structure is present, these matrices are obtained
#'     from the dense whitened information matrix and then split back into
#'     per-partition blocks.  When flat columns are present, the returned
#'     matrices also reflect the pooled flat-column information across
#'     partitions.
#'   }
#'   \item{qp_info}{
#'     NULL when no inequality-refinement metadata are produced.
#'     Otherwise a list of available QP or active-set diagnostics.
#'     Dense SQP solves include \code{solution}, \code{lagrangian},
#'     \code{active_constraints}, \code{iact}, \code{Amat_active},
#'     \code{bvec_active}, \code{meq_active}, \code{converged}, and
#'     \code{final_deviance}; non-GEE dense solves also carry
#'     \code{info_matrix}, \code{Amat_combined}, \code{bvec_combined},
#'     and \code{meq_combined}.  Partition-wise active-set refinement
#'     returns the corresponding active-constraint summary only.
#'   }
#' }
#'
#' @examples
#' \dontrun{
#' ## Minimal verification example: Gaussian identity, no correlation,
#' ## 2 partitions, 3 spline columns + 1 flat column per partition.
#' ##
#' ## This confirms that blockfit_solve returns compatible output and
#' ## that flat coefficients are replicated across partitions.
#'
#' set.seed(1234)
#' n1 <- 50; n2 <- 50
#' p_expansions <- 4  # intercept + x + x^2 + z (flat)
#' K  <- 1  # 2 partitions
#'
#' X1 <- cbind(1, rnorm(n1), rnorm(n1)^2, rnorm(n1))
#' X2 <- cbind(1, rnorm(n2), rnorm(n2)^2, rnorm(n2))
#' beta_true <- c(1, 0.5, -0.3, 0.8)
#' y1 <- X1 %*% beta_true + rnorm(n1, 0, 0.5)
#' y2 <- X2 %*% beta_true + rnorm(n2, 0, 0.5)
#'
#' ## Constraint: spline coefficients equal across partitions
#' ## (columns 1:3 constrained, column 4 is flat)
#' A <- matrix(0, nrow = p_expansions * (K + 1), ncol = 3)
#' for(j in 1:3){
#'   A[j, j]      <-  1
#'   A[p_expansions + j, j]  <- -1
#' }
#' qr_A <- qr(A)
#' A <- qr.Q(qr_A)[, 1:qr_A$rank, drop = FALSE]
#'
#' Lambda <- diag(p_expansions) * 1e-4
#' L_partition_list <- list(0, 0)
#' X_gram <- list(crossprod(X1), crossprod(X2))
#'
#' G_list <- compute_G_eigen(
#'   X_gram, Lambda, K,
#'   parallel = FALSE, cl = NULL,
#'   chunk_size = NULL, num_chunks = NULL, rem_chunks = NULL,
#'   family = gaussian(),
#'   unique_penalty_per_partition = FALSE,
#'   L_partition_list = L_partition_list,
#'   keep_G = TRUE,
#'   schur_corrections = list(0, 0))
#'
#' result <- blockfit_solve(
#'   X = list(X1, X2),
#'   y = list(y1, y2),
#'   flat_cols = 4L,
#'   K = K, p_expansions = p_expansions,
#'   Lambda = Lambda,
#'   L_partition_list = L_partition_list,
#'   unique_penalty_per_partition = FALSE,
#'   A = A, R_constraints = ncol(A),
#'   constraint_values = list(),
#'   X_gram = X_gram,
#'   Ghalf_full = G_list$Ghalf,
#'   GhalfInv_full = G_list$GhalfInv,
#'   family = gaussian(),
#'   order_list = list(1:n1, (n1+1):(n1+n2)),
#'   glm_weight_function = function(mu, y, oi, fam, d, ow, ...) rep(1, length(y)),
#'   schur_correction_function = function(X, y, B, d, ol, K, fam, ow, ...) {
#'     lapply(1:(K + 1), function(k) 0)
#'   },
#'   need_dispersion_for_estimation = FALSE,
#'   dispersion_function = function(...) 1,
#'   observation_weights = list(rep(1, n1), rep(1, n2)),
#'   homogenous_weights = TRUE,
#'   iterate = TRUE, tol = 1e-8,
#'   parallel = FALSE, cl = NULL,
#'   chunk_size = NULL, num_chunks = NULL, rem_chunks = NULL,
#'   return_G_getB = TRUE,
#'   verbose = FALSE)
#'
#' ## Verify: flat coefficient (col 4) is identical across partitions
#' stopifnot(abs(result$B[[1]][4] - result$B[[2]][4]) < 1e-10)
#'
#' ## Verify: output structure
#' stopifnot(length(result$B) == K + 1)
#' stopifnot(length(result$B[[1]]) == p_expansions)
#' stopifnot(!is.null(result$G_list))
#' }
#'
#' @seealso \code{\link{lgspline}}, \code{\link{lgspline.fit}},
#'   \code{\link{get_B}}
#'
#' @keywords internal
#' @export
blockfit_solve <- function(
    X,
    y,
    flat_cols,
    K,
    p_expansions,
    Lambda,
    L_partition_list,
    unique_penalty_per_partition,
    A,
    R_constraints,
    constraint_values,
    X_gram,
    Ghalf_full,
    GhalfInv_full,
    family,
    order_list,
    glm_weight_function,
    schur_correction_function,
    need_dispersion_for_estimation,
    dispersion_function,
    observation_weights,
    homogenous_weights = TRUE,
    iterate,
    tol,
    parallel_eigen,
    cl,
    chunk_size,
    num_chunks,
    rem_chunks,
    return_G_getB,
    quadprog = FALSE,
    qp_Amat = NULL,
    qp_bvec = NULL,
    qp_meq = NULL,
    qp_score_function = NULL,
    keep_weighted_Lambda = FALSE,
    max_backfit_iter = 100,
    Vhalf = NULL,
    VhalfInv = NULL,
    include_warnings = TRUE,
    verbose = FALSE,
    ...
){

  ## Dimensions and case detection
  nc_flat <- length(flat_cols)
  spline_cols <- setdiff(1:p_expansions, flat_cols)
  nc_spline <- length(spline_cols)
  nr <- sum(sapply(y, length))
  is_gauss_id <- (paste0(family)[1] == 'gaussian' &
                    paste0(family)[2] == 'identity')

  has_corr <- !is.null(Vhalf) & !is.null(VhalfInv)
  needs_gee_glm <- has_corr & !is_gauss_id

  ## Auto-detect whether inequality constraints require global (dense)
  ## QP or can be handled partition-wise. GEE always forces global.
  if(quadprog){
    qp_global <- has_corr | .solver_detect_qp_global(qp_Amat, p_expansions, K)
  } else {
    qp_global <- FALSE
  }
  if(has_corr) qp_global <- TRUE

  ## Split design matrices, penalty, and constraints
  #  The spline-only objects are then reused across the case-specific solvers.
  split <- .bf_split_components(X, flat_cols, p_expansions, K, Lambda,
                                L_partition_list, A,
                                constraint_values)

  ## Compute spline-only G^{1/2}
  X_gram_spline <- lapply(1:(K + 1), function(k){
    crossprod(split$X_spline[[k]])
  })
  schur_zero <- lapply(1:(K + 1), function(k) 0)

  G_sp <- compute_G_eigen(X_gram_spline,
                          split$Lambda_spline,
                          K,
                          parallel_eigen,
                          cl,
                          chunk_size,
                          num_chunks,
                          rem_chunks,
                          gaussian(),
                          unique_penalty_per_partition,
                          split$L_part_spline,
                          keep_G = TRUE,
                          schur_zero)
  Ghalf_sp <- G_sp$Ghalf

  ## G^{1/2} A for Lagrangian projection
  GhalfA_sp <- Reduce("rbind", lapply(1:(K + 1), function(k){
    rows <- ((k - 1) * nc_spline + 1):(k * nc_spline)
    Ghalf_sp[[k]] %**% split$A_spline[rows, , drop = FALSE]
  }))

  ## Pooled flat Gram and penalized inverse
  XfXf <- Reduce("+", lapply(1:(K + 1), function(k){
    crossprod(split$X_flat[[k]])
  }))
  XfXf_pen_inv <- invert(XfXf + split$Lambda_flat)

  ## Initialize
  beta_flat <- rep(0, nc_flat)
  beta_spline <- lapply(1:(K + 1), function(k) cbind(rep(0, nc_spline)))
  qp_info <- NULL

  ## Dispatch to the appropriate blockfit solver path.
  #  Cases are ordered from most specialized to most general.

  ## Case (a): Gaussian identity + GEE
  if(is_gauss_id & has_corr){

    out_a <- .bf_case_gauss_gee(
      X, y, K, p_expansions, order_list, VhalfInv,
      Lambda, L_partition_list,
      unique_penalty_per_partition,
      A, constraint_values,
      spline_cols, flat_cols,
      observation_weights)
    beta_spline <- out_a$beta_spline
    beta_flat   <- out_a$beta_flat
    result      <- out_a$result

    ## Case (b): Gaussian identity, no correlation
  } else if(is_gauss_id & !has_corr){

    out_b <- .bf_case_gauss_no_corr(
      split$X_spline, split$X_flat, y, K,
      Ghalf_sp, GhalfA_sp, XfXf_pen_inv,
      split$constraint_values_spline,
      nc_spline, nc_flat, split$A_spline,
      tol, max_backfit_iter, verbose,
      split = split,
      constraint_values = constraint_values,
      Lambda_flat = split$Lambda_flat,
      p_expansions = p_expansions)

    beta_spline <- out_b$beta_spline
    beta_flat   <- out_b$beta_flat

    ## Case (c): GLM + GEE
  } else if(has_corr){

    out_c <- .bf_case_glm_gee(
      X, y, K, p_expansions, flat_cols, split,
      family, order_list,
      glm_weight_function, schur_correction_function,
      qp_score_function,
      need_dispersion_for_estimation, dispersion_function,
      observation_weights,
      iterate, tol, max_backfit_iter,
      parallel_eigen, cl,
      chunk_size, num_chunks, rem_chunks, schur_zero,
      quadprog, qp_Amat, qp_bvec, qp_meq,
      Lambda, L_partition_list,
      unique_penalty_per_partition,
      A, constraint_values,
      Vhalf, VhalfInv, verbose, ...)

    result      <- out_c$result
    beta_spline <- out_c$beta_spline
    beta_flat   <- out_c$beta_flat
    qp_info     <- out_c$qp_info

    ## Case (d): GLM without GEE
  } else {

    out_d <- .bf_case_glm_no_corr(
      X, y, K, p_expansions, split, family, order_list,
      glm_weight_function,
      need_dispersion_for_estimation, dispersion_function,
      observation_weights,
      iterate, tol, max_backfit_iter,
      parallel_eigen, cl,
      chunk_size, num_chunks, rem_chunks, schur_zero,
      unique_penalty_per_partition, VhalfInv,
      verbose,
      constraint_values = constraint_values,
      ...)

    beta_spline <- out_d$beta_spline
    beta_flat   <- out_d$beta_flat
  }

  ## Assemble full per-partition coefficients (Cases b and d)
  if(!(is_gauss_id & has_corr) & !needs_gee_glm){
    result <- lapply(1:(K + 1), function(k){
      b <- rep(0, p_expansions)
      b[spline_cols] <- beta_spline[[k]]
      b[flat_cols]   <- beta_flat
      cbind(b)
    })
  }

  ## Inequality constraint refinement (non-GEE cases only)
  if(quadprog & !needs_gee_glm & !(is_gauss_id & has_corr)){

    if(!qp_global){
      ## Partition-wise active-set refinement.
      # Reuse the full-dimensional Ghalf/GhalfInv from above.
      # For Gaussian identity, Xy_or_uncon = Xy (cross-products);
      # for GLM, use result as unconstrained proxy.
      if(verbose) cat("  QP refinement (partition-wise active-set)\n")

      ## Compute Xy = X_k^T y_adj_k per partition for Gaussian,
      # or use result as unconstrained proxy for GLM.
      if(is_gauss_id){
        Xy_for_as <- lapply(1:(K + 1), function(k){
          crossprod(X[[k]], y[[k]])
        })
        is_p3 <- FALSE
      } else {
        Xy_for_as <- result
        is_p3 <- TRUE
      }

      as_out <- .active_set_refine(
        result, X, y, K, p_expansions,
        A, R_constraints, constraint_values,
        Lambda, Ghalf_full, GhalfInv_full, family,
        qp_Amat, qp_bvec, qp_meq,
        Xy_or_uncon = Xy_for_as, is_path3 = is_p3,
        parallel_aga = FALSE, parallel_matmult = FALSE,
        cl = cl, chunk_size = chunk_size,
        num_chunks = num_chunks, rem_chunks = rem_chunks,
        tol = tol)

      if(as_out$converged){
        result <- as_out$result
        qp_info <- as_out$qp_info
      } else {
        ## Fallback to dense SQP.
        if(verbose) cat("  Active-set did not converge, falling back to dense SQP\n")

        X_block <- collapse_block_diagonal(X)
        beta_block <- cbind(unlist(result))
        Lambda_block <- .bf_make_Lambda_block(
          Lambda, K, unique_penalty_per_partition, L_partition_list)
        y_block <- cbind(unlist(y))

        qp_Amat_combined_std <- cbind(A, qp_Amat)
        qp_bvec_combined_std <- c(rep(0, ncol(A)), qp_bvec)
        qp_meq_combined_std  <- ncol(A) + qp_meq

        sqp <- .bf_sqp_loop(
          X_design = X_block,
          y_design = y_block,
          X_block_raw = X_block,
          beta_init = beta_block,
          Lambda_block = Lambda_block,
          qp_Amat_combined = qp_Amat_combined_std,
          qp_bvec_combined = qp_bvec_combined_std,
          qp_meq_combined  = qp_meq_combined_std,
          K = K, p_expansions = p_expansions, family = family,
          order_list = order_list,
          glm_weight_function = glm_weight_function,
          schur_correction_function = schur_correction_function,
          qp_score_function = qp_score_function,
          need_dispersion_for_estimation = need_dispersion_for_estimation,
          dispersion_function = dispersion_function,
          observation_weights = observation_weights,
          iterate = iterate, tol = tol,
          VhalfInv = VhalfInv, VhalfInv_perm = NULL,
          is_gee = FALSE, deviance_fun = .bf_deviance,
          X_partitions = X, y_partitions = y,
          verbose = verbose, ...)

        result <- sqp$result

        qp_info <- .bf_assemble_qp_info(
          sqp$last_qp_sol, cbind(unlist(sqp$result)),
          qp_Amat_combined_std, qp_bvec_combined_std,
          qp_meq_combined_std,
          sqp$converged, sqp$final_deviance,
          info_matrix = NULL)
      }

    } else {
      ## Dense SQP (original behavior, cross-partition constraints detected).
      if(verbose) cat("  QP refinement (SQP loop)\n")

      X_block <- collapse_block_diagonal(X)
      beta_block <- cbind(unlist(result))
      Lambda_block <- .bf_make_Lambda_block(
        Lambda, K, unique_penalty_per_partition, L_partition_list)
      y_block <- cbind(unlist(y))

      qp_Amat_combined_std <- cbind(A, qp_Amat)
      qp_bvec_combined_std <- c(rep(0, ncol(A)), qp_bvec)
      qp_meq_combined_std  <- ncol(A) + qp_meq

      sqp <- .bf_sqp_loop(
        X_design = X_block,
        y_design = y_block,
        X_block_raw = X_block,
        beta_init = beta_block,
        Lambda_block = Lambda_block,
        qp_Amat_combined = qp_Amat_combined_std,
        qp_bvec_combined = qp_bvec_combined_std,
        qp_meq_combined  = qp_meq_combined_std,
        K = K, p_expansions = p_expansions, family = family,
        order_list = order_list,
        glm_weight_function = glm_weight_function,
        schur_correction_function = schur_correction_function,
        qp_score_function = qp_score_function,
        need_dispersion_for_estimation = need_dispersion_for_estimation,
        dispersion_function = dispersion_function,
        observation_weights = observation_weights,
        iterate = iterate, tol = tol,
        VhalfInv = VhalfInv, VhalfInv_perm = NULL,
        is_gee = FALSE, deviance_fun = .bf_deviance,
        X_partitions = X, y_partitions = y,
        verbose = verbose, ...)

      result <- sqp$result

      qp_info <- .bf_assemble_qp_info(
        sqp$last_qp_sol, cbind(unlist(sqp$result)),
        qp_Amat_combined_std, qp_bvec_combined_std,
        qp_meq_combined_std,
        sqp$converged, sqp$final_deviance,
        info_matrix = NULL)
    }
  }

  ## If downstream inference was not requested, stop at the
  #  coefficient estimates and any qp metadata gathered above.
  if(!return_G_getB){
    return(list(B = result, G_list = NULL, qp_info = qp_info))
  }

  ## Recompute G at final estimates using the same path as get_B Path 3.
  #  Since point estimates coincide with get_B, this produces identical
  #  G/Ghalf/GhalfInv and therefore identical U, varcovmat, and
  #  generate_posterior draws.
  if(is_gauss_id){
    G_list_out <- compute_G_eigen(
      X_gram, Lambda, K,
      parallel_eigen, cl,
      chunk_size, num_chunks, rem_chunks,
      family,
      unique_penalty_per_partition,
      L_partition_list,
      keep_G = TRUE,
      lapply(1:(K + 1), function(k) 0)
    )
    G_list_out$G <- lapply(G_list_out$Ghalf, tcrossprod)

    ## [Debug 2026-03-05] Flat-column pooling for blockfit.
    #  The flat columns were estimated with pooled information across
    #  all K+1 partitions (single shared coefficient), but compute_G_eigen
    #  treats each partition independently. The per-partition G blocks
    #  therefore reflect only partition-k's X_flat_k^T X_flat_k, not the
    #  pooled sum. Rather than patching G/Ghalf here (which breaks the
    #  identity G = tcrossprod(Ghalf) when cross-terms are zeroed), the
    #  fix is applied in lgspline.fit by augmenting the constraint matrix
    #  A with flat-equality constraints before U is constructed. This
    #  ensures U projects out the flat-equality directions, correctly
    #  pooling flat-column information in varcovmat and generate_posterior.
    #
    #  If posterior draws or confint still show inflated variance for
    #  blockfit vs non-blockfit, check:
    #    1. Is A augmented with flat-equality constraints in lgspline.fit
    #       BEFORE U is constructed? (Search for "flat_eq_cols")
    #    2. Is return_list$A updated after augmentation?
    #    3. Are G_list$Ghalf[[k]] identical between blockfit and non-blockfit?
    #       They should be, since both call compute_G_eigen on the same
    #       X_gram and Lambda.
    #    4. Is model_fit$Ghalf populated? (requires return_Ghalf = TRUE)
    #       generate_posterior uses model_fit$Ghalf to build L_post.
    #    5. Compare diag(model_fit$varcovmat) between blockfit and non-blockfit.
    #       Spline-column variances should be similar; flat-column variances
    #       will differ slightly because blockfit pools via U projection
    #       while non-blockfit pools via the original A constraint.
    #
    #  Diagnostic snippet (run after fitting):
    #    cat("G[[1]] diag:", diag(G_list_out$G[[1]]), "\n")
    #    cat("Ghalf[[1]] diag:", diag(G_list_out$Ghalf[[1]]), "\n")
    #    cat("flat_cols:", flat_cols, "\n")
    #    cat("spline_cols:", spline_cols, "\n")

  } else {
    G_list_out <- .solver_recompute_G_at_estimate(
      X = X, y = y, result = result, K = K, Lambda = Lambda,
      family = family, order_list = order_list,
      glm_weight_function = glm_weight_function,
      schur_correction_function = schur_correction_function,
      need_dispersion_for_estimation = need_dispersion_for_estimation,
      dispersion_function = dispersion_function,
      observation_weights = observation_weights,
      VhalfInv = VhalfInv,
      parallel_eigen = parallel_eigen, parallel_matmult = FALSE,
      cl = cl, chunk_size = chunk_size,
      num_chunks = num_chunks, rem_chunks = rem_chunks,
      unique_penalty_per_partition = unique_penalty_per_partition,
      L_partition_list = L_partition_list, ...
    )

    ## [Debug 2026-03-05] Non-Gaussian blockfit G recomputation.
    #  .solver_recompute_G_at_estimate computes per-partition weighted Gram
    #  matrices using GLM working weights at the final estimates. For
    #  flat columns, this again reflects only per-partition information.
    #  The A augmentation in lgspline.fit handles pooling via U, same
    #  as the Gaussian case.
    #
    #  If non-Gaussian blockfit shows inflated variance:
    #    1. Verify .solver_recompute_G_at_estimate returns proper matrices
    #       (not NULL for GhalfInv -- it should exist for non-Gaussian).
    #    2. Check that the A augmentation in lgspline.fit fires for
    #       non-Gaussian families (use_blockfit && length(flat_cols) > 0).
    #    3. Compare G_list_out$G[[k]] diagonal between blockfit and
    #       non-blockfit for the spline columns -- these should match
    #       if working weights are identical (same point estimates).
  }

  if(verbose){
    cat("  blockfit_solve returning: K =", K,
        "| p =", p_expansions,
        "| flat_cols =", paste(flat_cols, collapse = ","),
        "| n_result =", length(result),
        "| has_G =", !is.null(G_list_out$G[[1]]),
        "| has_Ghalf =", !is.null(G_list_out$Ghalf[[1]]),
        "| has_GhalfInv =", !is.null(G_list_out$GhalfInv),
        "| has_qp_info =", !is.null(qp_info), "\n")
  }

  return(list(B = result, G_list = G_list_out, qp_info = qp_info))
}

test_that("basic parallel processing works", {
  skip_on_cran()  # Skip on CRAN out of common courtesy

  # Setup test data
  set.seed(1234)
  x <- seq(-9, 9, length.out = 1000)
  y <- sin(x) + rnorm(1000, 0, 0.1)
  dat <- cbind(x, y)

  # Test parallel vs non-parallel results match
  cl <- parallel::makeCluster(2)
  on.exit(parallel::stopCluster(cl))
  ## Ensure cluster is stopped even if test fails

  set.seed(1234)
  fit_parallel <- lgspline(cbind(dat[,'x']),
                          dat[,'y'],
                          cl = cl,
                          K = 2)

  set.seed(1234)
  fit_serial <- lgspline(cbind(dat[,'x']),
                        dat[,'y'],
                        K = 2)

  # Compare results
  expect_equal(fit_parallel$ytilde,
              fit_serial$ytilde,
              tolerance = 1e-5)

  # Test predictions match
  newx <- matrix(seq(-9, 9, length.out = 100))
  pred_parallel <- predict(fit_parallel, newx)
  pred_serial <- predict(fit_serial, newx)

  expect_equal(pred_parallel, pred_serial, tolerance = 1e-5)
})

test_that("predictions are reasonable, derivative can be obtained", {
  set.seed(1234)
  # Test predictions follow true function pattern
  x <- seq(-9, 9, length.out = 100)
  y <- sin(x) + rnorm(100, 0, 0.1)
  fit <- lgspline(cbind(x), y, K = 5)

  newx <- matrix(seq(-9, 9, length.out = 10))
  pred <- predict(fit, newx)

  # Check predictions are within reasonable bounds
  expect_true(all(abs(pred) < max(abs(y)) * 1.5))

  ## Test first derivatives
  if ("take_first_derivatives" %in% names(formals(fit$predict))) {
    deriv <- predict(fit, newx, take_first_derivatives = TRUE)
    expect_type(deriv, "list")
    expect_true("first_deriv" %in% names(deriv))
  }
})

test_that("basic plotting functions work", {
  set.seed(1234)
  x <- seq(-9, 9, length.out = 100)
  y <- sin(x) + rnorm(100, 0, 0.1)
  fit <- lgspline(cbind(x), y, K = 5)

  # Test 1D plot
  expect_error(plot(fit), NA)
  expect_error(plot(fit, show_formulas = TRUE), NA)

  # Test 2D plot, include quartic terms
  data(volcano)
  volcano_long <- Reduce('rbind', lapply(1:nrow(volcano), function(i){
    t(sapply(1:ncol(volcano), function(j){
      c(i, j, volcano[i,j])
    }))
  }))
  fit2d <- lgspline(volcano_long[,1:2],
                    include_quartic_terms = TRUE,
                    volcano_long[,3], K = 5)
  plot(fit2d,
       show_formulas = TRUE,
       text_size_formula  = 7,
       digits = 2,
       custom_predictor_lab1 = 'My Predictor 1',
       custom_predictor_lab2 = 'My Predictor 2',
       custom_response_lab = 'My Response')
})

test_that("basic lgspline fits 1D data correctly using base R BFGS for tuning", {
  x <- seq(-9, 9, length.out = 100)
  y <- sin(x) + rnorm(100, 0, 0.1)

  fit <- lgspline(cbind(x), y, K = 5, use_custom_bfgs = FALSE)

  expect_s3_class(fit, "lgspline")
  expect_length(fit$B, fit$K + 1)
  expect_true(all(!is.na(fit$ytilde)))
})

test_that("Basic lgspline fits 2D data correctly", {
  data(volcano)
  volcano_long <- Reduce('rbind', lapply(1:nrow(volcano), function(i){
    t(sapply(1:ncol(volcano), function(j){
      c(i, j, volcano[i,j])
    }))
  }))

  fit <- lgspline(volcano_long[,1:2], volcano_long[,3], K = 1, opt = FALSE,
                  unique_penalty_per_partition = FALSE)

  expect_s3_class(fit, "lgspline")
  expect_true(fit$q == 2)
  expect_true(all(!is.na(fit$ytilde)))
})

test_that("Basic lgspline handles no knots", {
  # No knots
  x <- 1:10
  y <- x + rnorm(10)
  expect_error(lgspline(cbind(x), y, K = 0), NA)
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{confint.wald_lgspline}
\alias{confint.wald_lgspline}
\title{Extract Confidence Intervals from a wald_lgspline Object}
\usage{
\method{confint}{wald_lgspline}(object, parm = NULL, level = NULL, ...)
}
\arguments{
\item{object}{A \code{"wald_lgspline"} object.}

\item{parm}{Parameter specification (ignored; all returned).}

\item{level}{Confidence level (ignored; uses the object's critical value).}

\item{...}{Not used.}
}
\value{
Matrix with columns lower and upper, or NULL if confidence
  limits are not available.
}
\description{
Extract Confidence Intervals from a wald_lgspline Object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{wald_univariate}
\alias{wald_univariate}
\title{Univariate Wald Tests and Confidence Intervals for lgspline Coefficients}
\usage{
wald_univariate(object, scale_vcovmat_by = 1, cv, ...)
}
\arguments{
\item{object}{A fitted lgspline object. Must have been fit with
\code{return_varcovmat = TRUE}.}

\item{scale_vcovmat_by}{Numeric; scaling factor for the variance-covariance
matrix. Default 1.}

\item{cv}{Numeric; critical value for confidence intervals. If missing,
defaults to \code{object$critical_value} or \code{qnorm(0.975)}.}

\item{...}{Additional arguments passed to the internal \code{wald_univariate}
method.}
}
\value{
An object of class \code{"wald_lgspline"}, a list with:
\describe{
  \item{coefficients}{Matrix with columns: Estimate, Std. Error,
        t value or z value, Pr(>|t|) or Pr(>|z|), CI LB, CI UB.}
  \item{critical_value}{Critical value used.}
  \item{family}{GLM family from the fitted model.}
  \item{N}{Number of observations.}
  \item{trace_XUGX}{Effective df trace term.}
  \item{statistic_name}{"t value" or "z value".}
  \item{p_value_name}{"Pr(>|t|)" or "Pr(>|z|)".}
  \item{df.residual}{Residual degrees of freedom when supplied by the
        internal Wald method.}
}
Print, summary, and plot methods are available; see
\code{\link{print.wald_lgspline}}, \code{\link{summary.wald_lgspline}},
\code{\link{plot.wald_lgspline}}.
}
\description{
Computes per-coefficient Wald tests and confidence intervals from a fitted
lgspline. For Gaussian identity-link models, t-statistics and t-intervals
are used; otherwise z-statistics.
}
\examples{

set.seed(1234)
t <- runif(1000, -10, 10)
y <- 2*sin(t) + -0.06*t^2 + rnorm(length(t))
model_fit <- lgspline(t, y, return_varcovmat = TRUE)

wald_default <- wald_univariate(model_fit)
print(wald_default)

## t-distribution critical value
eff_df <- model_fit$N - model_fit$trace_XUGX
wald_t <- wald_univariate(model_fit, cv = qt(0.975, eff_df))
print(wald_t)

coef_table <- wald_default$coefficients
plot(wald_default)

}
\seealso{
\code{\link{lgspline}}, \code{\link{confint.lgspline}},
\code{\link{print.wald_lgspline}}, \code{\link{summary.wald_lgspline}},
\code{\link{plot.wald_lgspline}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cox_helpers.R
\name{cox_glm_weight_function}
\alias{cox_glm_weight_function}
\title{Cox PH GLM Weight Function}
\usage{
cox_glm_weight_function(
  mu,
  y,
  order_indices,
  family,
  dispersion,
  observation_weights,
  status
)
}
\arguments{
\item{mu}{Predicted values (exp(eta), i.e., relative hazard).}

\item{y}{Observed survival times.}

\item{order_indices}{Observation indices in partition order.}

\item{family}{Cox family object (unused, for interface compatibility).}

\item{dispersion}{Dispersion parameter (fixed at 1 for Cox PH).}

\item{observation_weights}{Observation weights.}

\item{status}{Event indicator (1 = event, 0 = censored).}
}
\value{
Numeric vector of working weights, length N.
}
\description{
Computes working weights for the Cox PH information matrix, used by
lgspline when updating \eqn{\mathbf{G}} after obtaining constrained
estimates. The weights are a diagonal approximation built from the Breslow
tied-event information contributions.
}
\details{
For a tied event-time block \eqn{g}, the diagonal approximation uses
\deqn{W_{jj}^{(g)} = d_g^{(w)} \frac{h_j}{S_g}
  \Bigl(1 - \frac{h_j}{S_g}\Bigr), \qquad j \in R_g}
where \eqn{h_j = w_j \exp(\eta_j)},
\eqn{S_g = \sum_{k \in R_g} h_k}, and
\eqn{d_g^{(w)} = \sum_{i \in D_g} w_i}.

When the natural weights are degenerate (all zero or non-finite), falls
back to a vector of ones.
}
\examples{
## Used internally by lgspline; see cox_helpers examples below.

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{equation}
\alias{equation}
\alias{equation.lgspline}
\alias{print.equation}
\title{Print Closed-Form Fitted Equation from lgspline Model}
\usage{
equation(object, ...)

\method{equation}{lgspline}(
  object,
  digits = 4,
  scientific = FALSE,
  show_bounds = TRUE,
  predictor_names = NULL,
  response_name = NULL,
  collapse_zero = TRUE,
  first_derivative = NULL,
  second_derivative = NULL,
  antiderivative = NULL,
  ...
)

\method{print}{equation}(x, ...)

\method{equation}{lgspline}(
  object,
  digits = 4,
  scientific = FALSE,
  show_bounds = TRUE,
  predictor_names = NULL,
  response_name = NULL,
  collapse_zero = TRUE,
  first_derivative = NULL,
  second_derivative = NULL,
  antiderivative = NULL,
  ...
)

\method{print}{equation}(x, ...)
}
\arguments{
\item{object}{A fitted lgspline model object.}

\item{...}{Not used.}

\item{digits}{Integer; decimal places for coefficient display. Default 4.}

\item{scientific}{Logical; use scientific notation for coefficients with
absolute value < 1e-3 or > 1e4. Default FALSE.}

\item{show_bounds}{Logical; display partition bounds (1D) or knot midpoint
boundaries (multi-D). Default TRUE.}

\item{predictor_names}{Character vector; custom names for predictor variables.
If NULL (default), uses original column names or "_j_" labels.}

\item{response_name}{Character; label for response. If NULL (default), uses
"y" for identity link Gaussian, or "link(E[y])" otherwise.}

\item{collapse_zero}{Logical; omit terms with coefficient exactly 0.
Default TRUE.}

\item{first_derivative}{Default: NULL. Character name or integer index of
the predictor variable with respect to which the first derivative
is printed. Only one variable at a time is supported. When non-NULL,
the printed equations show \eqn{df/dx_j} for each partition.}

\item{second_derivative}{Default: NULL. Character name or integer index of
the predictor variable with respect to which the second derivative
is printed. Only one variable at a time is supported. When non-NULL,
the printed equations show \eqn{d^2f/dx_j^2} for each partition.
Ignored if \code{first_derivative} is also non-NULL.}

\item{antiderivative}{Default: NULL. Character name or integer index of
the predictor variable with respect to which the antiderivative
(indefinite integral) is printed. Only one variable at a time is
supported. When non-NULL, the printed equations show
\eqn{\int f\, dx_j} for each partition, with an unspecified
constant of integration \eqn{C}. Ignored if \code{first_derivative}
or \code{second_derivative} is also non-NULL.}

\item{x}{An object returned by \code{equation()} for printing.}
}
\value{
Invisibly returns a list with components:
\describe{
  \item{formulas}{Character vector of equation strings per partition.}
  \item{bounds}{Matrix or list of partition boundary information.}
  \item{link}{Character; link function name.}
  \item{mode}{Character; one of "equation", "first_derivative",
        "second_derivative", or "antiderivative".}
  \item{variable}{Character; the variable name for the calculus operation,
        or NULL if mode is "equation".}
}
}
\description{
Displays the closed-form polynomial equation for each partition of a fitted
lgspline model, along with partition boundary or cluster center information.
Optionally prints the first derivative, second derivative, or antiderivative
of the fitted equation with respect to a single specified variable.
}
\details{
For 1D models with K knots, partition boundaries are displayed as intervals
on the predictor scale. For multi-predictor models, partition boundaries are
computed as the midpoints between adjacent cluster centers along each
predictor dimension. When the model's \code{make_partition_list} contains
\code{knots} (midpoint boundaries between clusters), those are used directly.
Otherwise, cluster centers are displayed.

Coefficients are displayed on the original (unstandardized) predictor scale.
For GLMs with non-identity link, the left-hand side shows the link function
applied to the expected response.

\strong{Derivative and antiderivative modes.}
Only one of \code{first_derivative}, \code{second_derivative}, or
\code{antiderivative} may be non-NULL. If more than one is supplied, the
priority order is: first derivative, second derivative, antiderivative.

Derivatives and antiderivatives are computed symbolically from the
polynomial coefficients. For a term \eqn{a x^n}, the first derivative is
\eqn{n a x^{n-1}}, the second derivative is \eqn{n(n-1) a x^{n-2}}, and
the antiderivative is \eqn{a x^{n+1}/(n+1)}. Cross-terms (interactions)
involving the target variable are differentiated or integrated with respect
to that variable only, treating all other variables as constants.

A warning is emitted if the user attempts to differentiate or integrate with
respect to more than one variable simultaneously. Multi-variable calculus
operations should be performed one variable at a time by calling
\code{equation()} repeatedly.
}
\examples{

## 1D example
set.seed(1234)
t <- runif(500, -5, 5)
y <- 2*sin(t) + 0.1*t^2 + rnorm(length(t), 0, 0.5)
fit <- lgspline(t, y, K = 2)
equation(fit)
equation(fit, digits = 2, predictor_names = "time")

## First derivative with respect to predictor
equation(fit, first_derivative = 1)

## Second derivative
equation(fit, second_derivative = 1)

## Antiderivative
equation(fit, antiderivative = 1)

## 2D example with named predictors
x1 <- runif(300, 0, 10)
x2 <- runif(300, 0, 10)
y <- x1 + 0.5*x2 + 0.1*x1*x2 + rnorm(300)
fit2d <- lgspline(cbind(x1, x2), y, K = 3)
equation(fit2d, predictor_names = c("Length", "Width"))

## Derivative w.r.t. first variable only
equation(fit2d, first_derivative = "Length",
         predictor_names = c("Length", "Width"))

## GLM example
y_bin <- rbinom(500, 1, plogis(0.5*t))
fit_glm <- lgspline(t, y_bin, family = binomial(), K = 1)
equation(fit_glm)

}
\seealso{
\code{\link{lgspline}}, \code{\link{plot.lgspline}},
  \code{\link{coef.lgspline}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_Lambda.R
\name{.fit_get_B}
\alias{.fit_get_B}
\title{Call get_B During Penalty Tuning}
\usage{
.fit_get_B(G_list, Lambda, L_partition_list, env, return_G_getB, ...)
}
\arguments{
\item{G_list}{List; eigendecomposition results from \code{compute_G_eigen},
containing \code{Ghalf} and \code{GhalfInv}.}

\item{Lambda}{Matrix; current combined penalty matrix.}

\item{L_partition_list}{List; partition-specific penalty matrices.}

\item{env}{List; tuning environment from \code{.build_tuning_env}.}

\item{return_G_getB}{Logical; whether to return G inside the fit. Set to
TRUE within \code{.compute_gcvu} and \code{.compute_gcvu_gradient} so
that \code{B_list$G_list} carries the (possibly GLM-iterated) G matrices
required for the subsequent \code{AGAmult_wrapper} and trace computations.}

\item{...}{Additional arguments forwarded to the fitting routine.}
}
\description{
Internal wrapper that calls \code{get_B} with all arguments drawn from
the tuning environment \code{env}. Separated from \code{.fit_coefficients}
so the fallback path in \code{.fit_coefficients} is clean and does not
repeat the full argument list.

This helper is the point at which penalty tuning inherits the full
constrained solver described in \code{lgspline-details}, including GLM
reweighting, dense correlation whitening, and the structured-correlation
Woodbury acceleration when available. No additional tuning-specific
approximation is introduced here beyond whatever \code{get_B} itself uses.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{create_onehot}
\alias{create_onehot}
\title{Create One-Hot Encoded Matrix}
\usage{
create_onehot(x, drop_first = FALSE)
}
\arguments{
\item{x}{A vector containing categorical values (factors, character, etc.)}

\item{drop_first}{Logical; if \code{TRUE} and more than one dummy column is
created, drop the first column.}
}
\value{
A data frame containing the one-hot encoded binary columns with cleaned column names
}
\description{
Converts a categorical vector into a one-hot encoded matrix where each unique value
becomes a binary column.
}
\details{
The function creates dummy variables for each unique value in the input vector using
\code{model.matrix()} with dummy-intercept coding. Column names are cleaned by removing the
'x' prefix added by \code{model.matrix()}.
}
\examples{


## lgspline will not accept this format of "catvar", because inputting data
# this way can cause difficult-to-diagnose issues in formula parsing
# all variables must be numeric
df <- data.frame(numvar = rnorm(100),
                 catvar = rep(LETTERS[1:4],
                              25))
print(head(df))

## Instead, replace with dummy-intercept coding by
# 1) applying one-hot encoding
# 2) dropping the first column
# 3) appending to our data

dummy_intercept_coding <- create_onehot(df$catvar)[,-1]
df$catvar <- NULL
df <- cbind(df, dummy_intercept_coding)
print(head(df))



}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blockfit_solve.R
\name{.bf_case_gauss_no_corr}
\alias{.bf_case_gauss_no_corr}
\title{Gaussian Identity Backfitting Without Correlation (Case b)}
\usage{
.bf_case_gauss_no_corr(
  X_spline,
  X_flat,
  y,
  K,
  Ghalf_sp,
  GhalfA_sp,
  XfXf_pen_inv,
  constraint_values_spline,
  nc_spline,
  nc_flat,
  A_spline,
  tol,
  max_backfit_iter,
  verbose,
  split = NULL,
  constraint_values = list(),
  Lambda_flat = NULL,
  p_expansions = NULL
)
}
\arguments{
\item{X_spline, X_flat}{Lists of per-partition submatrices.}

\item{y}{List of response vectors.}

\item{K}{Integer.}

\item{Ghalf_sp}{List of \eqn{\mathbf{G}_s^{1/2}} matrices.}

\item{GhalfA_sp}{Pre-computed \eqn{\mathbf{G}_s^{1/2}
\mathbf{A}_s}.}

\item{XfXf_pen_inv}{Penalised inverse of the pooled flat Gram
matrix.}

\item{constraint_values_spline}{Spline-only constraint RHS.}

\item{nc_spline, nc_flat}{Integers.}

\item{A_spline}{Spline-only constraint matrix.}

\item{tol}{Convergence tolerance.}

\item{max_backfit_iter}{Maximum iterations.}

\item{verbose}{Logical.}

\item{split}{Output of \code{.bf_split_components}, needed for
mixed-constraint enforcement on the flat step.}

\item{constraint_values}{Full constraint RHS list.}

\item{Lambda_flat}{Flat penalty submatrix.}

\item{p_expansions}{Integer; columns per partition.}
}
\value{
A named list with \code{beta_spline} and
  \code{beta_flat}.
}
\description{
Standard block-coordinate descent on the convex quadratic
objective, alternating between the spline step (Lagrangian
projection) and the flat step (pooled penalized regression).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.lgspline}
\alias{print.lgspline}
\title{Print Method for lgspline Objects}
\usage{
\method{print}{lgspline}(x, ...)
}
\arguments{
\item{x}{An lgspline model object.}

\item{...}{Not used.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Prints a concise summary of the fitted model to the console.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weibull_helpers.R
\name{weibull_scale}
\alias{weibull_scale}
\title{Estimate Scale for Weibull Accelerated Failure Time Model}
\usage{
weibull_scale(log_y, log_mu, status, weights = 1)
}
\arguments{
\item{log_y}{Logarithm of response/survival times}

\item{log_mu}{Logarithm of predicted survival times}

\item{status}{Censoring indicator (1 = event, 0 = censored)
Indicates whether an event of interest occurred (1) or the observation was
right-censored (0). In survival analysis, right-censoring occurs when the
full survival time is unknown, typically because the study ended or the
subject was lost to follow-up before the event of interest occurred.}

\item{weights}{Optional observation weights (default = 1)}
}
\value{
Scalar representing the estimated Weibull scale (sigma), equivalent to
\code{survreg$scale}. The dispersion (as stored in
\code{lgspline$sigmasq_tilde}) is sigma^2.
}
\description{
Computes maximum log-likelihood scale estimate (sigma) for a Weibull
accelerated failure time (AFT) survival model.

This both provides a tool for actually fitting Weibull AFT Models, and
boilerplate code for users who wish to incorporate Lagrangian multiplier
smoothing splines into their own custom models.
}
\details{
Calculates maximum log-likelihood estimate of scale (sigma) for Weibull AFT
model accounting for right-censored observations using Brent's method for
optimization.
}
\examples{

## Simulate exponential data with censoring
set.seed(1234)
mu <- 2  # mean of exponential distribution
n <- 500
y <- rexp(n, rate = 1/mu)

## Introduce censoring (25\% of observations)
status <- rbinom(n, 1, 0.75)
y_obs <- ifelse(status, y, NA)

## Compute scale estimate
scale_est <- weibull_scale(
  log_y = log(y_obs[!is.na(y_obs)]),
  log_mu = log(mu),
  status = status[!is.na(y_obs)]
)

print(scale_est)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.check_kkt_partitionwise}
\alias{.check_kkt_partitionwise}
\title{Check KKT Conditions for Partition-Wise Active-Set Method}
\usage{
.check_kkt_partitionwise(
  result,
  Ghalf,
  GhalfInv,
  Xy_or_uncon,
  is_path3,
  A_aug,
  n_eq_orig,
  qp_Amat,
  qp_bvec,
  active_ineq,
  K,
  p_expansions,
  family,
  parallel_matmult,
  parallel_aga,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  tol
)
}
\arguments{
\item{result}{List of \eqn{K+1} coefficient column vectors.}

\item{Ghalf}{List of \eqn{\mathbf{G}^{1/2}_k} matrices.}

\item{GhalfInv}{List of \eqn{\mathbf{G}^{-1/2}_k} matrices.}

\item{Xy_or_uncon}{Either the list of cross-products
\eqn{\mathbf{X}_k^{\top}\mathbf{y}_k} (Path 2) or unconstrained
estimates (Path 3).}

\item{is_path3}{Logical; if TRUE, \code{Xy_or_uncon} contains
unconstrained estimates and GhalfInv is used for transformation.}

\item{A_aug}{Augmented constraint matrix (original A plus active
inequality columns).}

\item{n_eq_orig}{Integer; number of original equality constraints
(columns of A before augmentation).}

\item{qp_Amat}{Full inequality constraint matrix.}

\item{qp_bvec}{Full inequality constraint RHS.}

\item{active_ineq}{Integer vector; indices into columns of
\code{qp_Amat} that are currently active.}

\item{K, p_expansions}{Integer dimensions.}

\item{family}{GLM family object.}

\item{parallel_matmult, parallel_aga}{Logical flags.}

\item{cl, chunk_size, num_chunks, rem_chunks}{Parallel parameters.}

\item{tol}{Numeric tolerance for feasibility and multiplier checks.}
}
\value{
A list with components:
\describe{
  \item{feasible}{Logical; TRUE if all inactive inequality constraints
    are satisfied within tolerance.}
  \item{dual_feasible}{Logical; TRUE if all active inequality
    multipliers are non-negative within tolerance.}
  \item{violated}{Integer vector; indices of violated inactive
    constraints (into \code{qp_Amat} columns).}
  \item{drop}{Integer vector; indices of active constraints with
    negative multipliers that should be dropped.}
  \item{multipliers}{Numeric vector of Lagrangian multipliers for
    the active inequality constraints.}
}
}
\description{
Given current constrained coefficient estimates and a set of active
inequality constraints (treated as equalities in the last Lagrangian
projection), checks primal feasibility of inactive constraints and
dual feasibility (non-negative multipliers) of active constraints.
}
\details{
Multipliers for active inequality constraints are recovered from the
OLS fit used in the Lagrangian projection: the fitted coefficients
on \eqn{\mathbf{X}^* = \mathbf{G}^{1/2}\mathbf{A}_{\mathrm{aug}}}
give the Lagrangian multipliers (up to sign and scaling).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_input.R
\name{process_input}
\alias{process_input}
\title{Process Input Arguments for lgspline}
\usage{
process_input(
  predictors = NULL,
  y = NULL,
  formula = NULL,
  response = NULL,
  data = NULL,
  weights = NULL,
  observation_weights = NULL,
  family = gaussian(),
  K = NULL,
  custom_knots = NULL,
  auto_encode_factors = TRUE,
  include_2way_interactions = TRUE,
  include_3way_interactions = TRUE,
  just_linear_with_interactions = NULL,
  just_linear_without_interactions = NULL,
  exclude_interactions_for = NULL,
  exclude_these_expansions = NULL,
  offset = c(),
  no_intercept = FALSE,
  do_not_cluster_on_these = c(),
  include_quartic_terms = NULL,
  cluster_args = c(custom_centers = NA, nstart = 10),
  include_warnings = TRUE,
  dummy_fit = FALSE,
  include_constrain_second_deriv = TRUE,
  standardize_response = TRUE,
  ...
)
}
\arguments{
\item{predictors}{Default: NULL. Formula or numeric matrix/data frame
of predictor variables.}

\item{y}{Default: NULL. Numeric response vector.}

\item{formula}{Default: NULL. Optional formula; alias for predictors
when a formula object.}

\item{response}{Default: NULL. Alternative name for \code{y}.}

\item{data}{Default: NULL. Data frame for formula interface.}

\item{weights}{Default: NULL. Alias for \code{observation_weights}.}

\item{observation_weights}{Default: NULL. Numeric observation weight
vector.}

\item{family}{Default: \code{gaussian()}. GLM family object.}

\item{K}{Default: NULL. Number of interior knots.}

\item{custom_knots}{Default: NULL. Custom knot matrix.}

\item{auto_encode_factors}{Default: TRUE. Logical; auto one-hot encode
factor and character columns when using the formula interface.}

\item{include_2way_interactions}{Default: TRUE. Logical.}

\item{include_3way_interactions}{Default: TRUE. Logical.}

\item{just_linear_with_interactions}{Default: NULL. Integer vector or
character vector of column names.}

\item{just_linear_without_interactions}{Default: NULL. Integer vector or
character vector of column names.}

\item{exclude_interactions_for}{Default: NULL. Integer vector or character
vector of column names.}

\item{exclude_these_expansions}{Default: NULL. Character vector of
expansion names to exclude.}

\item{offset}{Default: \code{c()}. Vector of column indices or names
to include as offsets.}

\item{no_intercept}{Default: FALSE. Logical; remove intercept.}

\item{do_not_cluster_on_these}{Default: \code{c()}. Vector of column
indices or names to exclude from clustering.}

\item{include_quartic_terms}{Default: NULL. Logical or NULL.}

\item{cluster_args}{Default: \code{c(custom_centers = NA, nstart = 10)}.
Named vector of clustering arguments.}

\item{include_warnings}{Default: TRUE. Logical.}

\item{dummy_fit}{Default: FALSE. Logical; run the full preprocessing path but stop short of fitting nonzero coefficients.}

\item{include_constrain_second_deriv}{Default: TRUE. Logical.}

\item{standardize_response}{Default: TRUE. Logical.}

\item{...}{Additional arguments (checked for depreciated names).}
}
\value{
A named list containing:
\describe{
  \item{predictors}{Numeric matrix of predictor variables with column
    names stripped for positional indexing.}
  \item{y}{Numeric response vector.}
  \item{og_cols}{Character vector of original predictor column names,
    or NULL if none were available.}
  \item{replace_colnames}{Logical; TRUE if og_cols is available and
    column renaming should be applied post-fit.}
  \item{just_linear_with_interactions}{Integer vector of column indices,
    or NULL.}
  \item{just_linear_without_interactions}{Integer vector of column indices,
    or NULL.}
  \item{exclude_interactions_for}{Integer vector of column indices,
    or NULL.}
  \item{exclude_these_expansions}{Character vector of positional-notation
    expansion names, or NULL.}
  \item{offset}{Integer vector of column indices, or \code{c()}.}
  \item{no_intercept}{Logical.}
  \item{do_not_cluster_on_these}{Numeric vector of column indices,
    or \code{c()}.}
  \item{observation_weights}{Numeric vector or NULL.}
  \item{K}{Integer or NULL, possibly updated by cluster_args or
    all-linear detection.}
  \item{include_3way_interactions}{Logical, possibly updated by
    formula parsing.}
  \item{include_quartic_terms}{Logical or NULL, possibly updated
    by number of predictors.}
  \item{data}{Data frame, possibly with factor columns one-hot encoded.}
  \item{include_constrain_second_deriv}{Logical, possibly set FALSE
    when no numeric predictors remain.}
  \item{factor_groups}{Named list mapping original factor column names to
    integer vectors of their one-hot indicator column positions within
    the predictor matrix. Used by lgspline.fit to impose sum-to-zero
    constraints on encoded factor levels. NULL when no factors were
    encoded, or when the formula interface was not used.}
}
}
\description{
Parses formula and data arguments, performs factor encoding,
resolves variable roles (spline, linear-with-interactions,
linear-without-interactions), constructs exclusion patterns,
and validates inputs. Called internally by \code{\link{lgspline}}.

Users may call this function directly to inspect how their formula
and data are interpreted before fitting.
}
\examples{
\dontrun{
data("Theoph")
df <- Theoph[, c("Time", "Dose", "conc", "Subject")]
processed <- process_input(
  predictors = conc ~ spl(Time) + Time*Dose,
  data = df,
  auto_encode_factors = TRUE,
  include_warnings = TRUE
)
str(processed$predictors)
processed$og_cols
processed$just_linear_without_interactions
processed$factor_groups
}

}
\seealso{
\code{\link{lgspline}} for the main fitting interface.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_Lambda.R
\name{.compute_meta_penalty}
\alias{.compute_meta_penalty}
\title{Compute Regularization (Meta) Penalty on Penalty Parameters}
\usage{
.compute_meta_penalty(
  wiggle_penalty,
  penalty_vec,
  meta_penalty_coef,
  unique_penalty_per_predictor,
  unique_penalty_per_partition
)
}
\arguments{
\item{wiggle_penalty}{Numeric; current wiggle penalty on raw scale.}

\item{penalty_vec}{Numeric vector; current predictor/partition penalties
on raw scale. May be empty (\code{c()}).}

\item{meta_penalty_coef}{Numeric; coefficient for the meta-penalty.}

\item{unique_penalty_per_predictor}{Logical; whether predictor-specific
penalties are active.}

\item{unique_penalty_per_partition}{Logical; whether partition-specific
penalties are active.}
}
\value{
Numeric scalar; the regularization penalty value.
}
\description{
Computes the regularization term that pulls predictor- and partition-specific
penalty parameters toward 1 on the raw (positive) scale. This acts as a
"meta-penalty" on the penalty magnitudes themselves.
}
\details{
The penalty takes the form:
\deqn{0.5 \times c_{\mathrm{meta}} \times \sum_{j} (\lambda_{j} - 1)^{2}
      + 0.5 \times 10^{-32} \times (\lambda_{w} - 1)^{2}}
where \eqn{\lambda_{j}} are predictor/partition penalties and
\eqn{\lambda_{w}} is the wiggle penalty.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{get_U}
\alias{get_U}
\title{Efficiently Construct U Matrix}
\usage{
get_U(G, A, K, p_expansions, R_constraints)
}
\arguments{
\item{G}{List of G matrices (\eqn{\textbf{G}})}

\item{A}{Constraint matrix (\eqn{\textbf{A}})}

\item{K}{Number of partitions minus 1 (\eqn{K})}

\item{p_expansions}{Number of columns per partition}

\item{R_constraints}{Number of constraint columns}
}
\value{
\eqn{\textbf{U}} matrix for constraints
}
\description{
Efficiently Construct U Matrix
}
\details{
Computes \eqn{\textbf{U} = \textbf{I} - \textbf{G}\textbf{A}(\textbf{A}^{T}\textbf{G}\textbf{A})^{-1}\textbf{A}^{T}} efficiently, avoiding unnecessary
multiplication of blocks of \eqn{\textbf{G}} with all-0 elements.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{plot.lgspline}
\alias{plot.lgspline}
\title{Plot Method for lgspline Objects}
\usage{
\method{plot}{lgspline}(
  x,
  show_formulas = FALSE,
  include_all_terms_in_formulas = FALSE,
  digits = 4,
  legend_pos = "topright",
  custom_response_lab = "y",
  custom_predictor_lab = NULL,
  custom_predictor_lab1 = NULL,
  custom_predictor_lab2 = NULL,
  custom_formula_lab = NULL,
  custom_title = "Fitted Function",
  text_size_formula = NULL,
  legend_args = list(),
  new_predictors = NULL,
  xlim = NULL,
  ylim = NULL,
  color_function = NULL,
  add = FALSE,
  vars = c(),
  legend_order = NULL,
  se.fit = FALSE,
  cv = 1,
  band_col = "grey80",
  band_border = NA,
  fixed_values = NULL,
  n_grid = 200,
  ...
)
}
\arguments{
\item{x}{A fitted lgspline model object.}

\item{show_formulas}{Logical; display partition-level polynomial formulas.
Default FALSE.}

\item{include_all_terms_in_formulas}{Logical; when \code{show_formulas = TRUE}
and plotting only a subset of predictors via \code{vars}, include all
fitted terms in the displayed formulas rather than only the terms
involving the plotted predictor(s). Default FALSE retains the current
marginal-only formula display.}

\item{digits}{Integer; decimal places for formula coefficients. Default 4.}

\item{legend_pos}{Character; legend position for 1D plots. Default
\code{"topright"}.}

\item{custom_response_lab}{Character; response axis label. Default
\code{"y"}.}

\item{custom_predictor_lab}{Character; predictor axis label (1D). Default
NULL uses the column name.}

\item{custom_predictor_lab1}{Character; first predictor axis label (2D).
Default NULL.}

\item{custom_predictor_lab2}{Character; second predictor axis label (2D).
Default NULL.}

\item{custom_formula_lab}{Character; fitted response label on the link
scale. Default NULL.}

\item{custom_title}{Character; plot title. Default \code{"Fitted Function"}.}

\item{text_size_formula}{Numeric; formula text size. Passed to \code{cex}
(1D) or hover font size (2D). Default NULL (0.8 for 1D, 8 for 2D).}

\item{legend_args}{List; additional arguments passed to \code{legend()}
(1D only).}

\item{new_predictors}{Matrix; optional predictor values for prediction.
Default NULL. When \code{vars} is specified and
\code{new_predictors} is NULL, a grid is automatically generated
with non-plotted variables set to zero (or values from
\code{fixed_values}).}

\item{xlim}{Numeric vector; x-axis limits (1D only). Default NULL.}

\item{ylim}{Numeric vector; y-axis limits (1D only). Default NULL.}

\item{color_function}{Function; returns K+1 colors, one per partition.
Default NULL uses \code{grDevices::rainbow(K+1)} for 1D and a
Spectral palette for 2D.}

\item{add}{Logical; add to an existing plot (1D only). Default FALSE.}

\item{vars}{Numeric or character vector; predictor indices or names to
plot. Default \code{c()} plots all.}

\item{legend_order}{Numeric; re-ordered partition indices for the legend.}

\item{se.fit}{Logical; if TRUE, plot pointwise confidence bands. Default
FALSE.}

\item{cv}{Numeric; critical value for confidence bands. Default 1.}

\item{band_col}{Character; color for confidence band fill. Default
\code{"grey80"}.}

\item{band_border}{Character or NA; border color for confidence band
polygon. Default NA (no border).}

\item{fixed_values}{Named list; fixed values for non-plotted predictors
when \code{vars} is specified. Names should match predictor names.
Default NULL sets non-plotted predictors to zero.}

\item{n_grid}{Integer; number of grid points for automatic grid
generation when \code{vars} is specified and
\code{new_predictors} is NULL. Default 200.}

\item{...}{Additional arguments passed to \code{\link[graphics]{plot}}
(1D) or \code{\link[plotly]{plot_ly}} (2D).}
}
\value{
For 1D models: invisibly returns NULL (base R plot drawn to
  device). For 2D models: returns a plotly object.
}
\description{
Wrapper for the internal lgspline plot function. Produces a 1D line plot
(base R) or interactive 3D surface plot (plotly) depending on the number
of predictors, with optional formula overlays per partition. When plotting
a subset of variables via \code{vars}, non-plotted predictors are
automatically set to zero (or a user-specified value via
\code{fixed_values}).
}
\details{
Partition boundaries are indicated by color changes. For 1D models,
observation points can be overlaid. For 2D models, plotly is used.

When using \code{vars} to plot a subset of predictors, the non-plotted
predictors are automatically set to zero. This can be overridden by
passing a named list to \code{fixed_values} (e.g.,
\code{fixed_values = list(Height = 75)}). The automatic zeroing replaces
the previous behavior where the user had to manually construct
\code{new_predictors} with non-plotted variables set to fixed values.

When \code{se.fit = TRUE}, pointwise confidence bands are drawn around
the fitted function. These are Wald-type intervals constructed on the
link scale and back-transformed to the response scale, using
\code{cv} as the critical value to multiply se.fit by (default 1 for actual se).

The function extracts predictor positions from linear expansion terms.
If linear terms are excluded (e.g., via \code{exclude_these_expansions}),
plotting will fail. As a workaround, constrain those terms to zero via
\code{constraint_vectors} / \code{constraint_values} so they remain in
the expansion but are zeroed out.
}
\examples{

set.seed(1234)
t_data <- runif(1000, -10, 10)
y_data <- 2*sin(t_data) + -0.06*t_data^2 + rnorm(length(t_data))
model_fit <- lgspline(t_data, y_data, K = 9)

## Basic plot
plot(model_fit)

## Plot with confidence bands
plot(model_fit,
     se.fit = TRUE,
     cv = 1.96,
     custom_title = 'Fitted Function with 95\% CI')

## Multi-predictor: automatically zeros non-plotted variables
# plot(model_fit_2d, vars = 'x1', se.fit = TRUE)

}
\seealso{
\code{\link{lgspline}},
\code{\link[graphics]{plot}},
\code{\link[plotly]{plot_ly}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_qp.R
\name{process_qp}
\alias{process_qp}
\title{Prepare Quadratic Programming Constraints for lgspline}
\usage{
process_qp(
  X,
  K,
  p_expansions,
  order_list,
  colnm_expansions,
  expansion_scales,
  power1_cols,
  power2_cols,
  nonspline_cols,
  interaction_single_cols,
  interaction_quad_cols,
  triplet_cols,
  include_2way_interactions,
  include_3way_interactions,
  include_quadratic_interactions,
  family,
  mean_y,
  sd_y,
  N_obs,
  qp_observations = NULL,
  qp_positive_derivative = FALSE,
  qp_negative_derivative = FALSE,
  qp_positive_2ndderivative = FALSE,
  qp_negative_2ndderivative = FALSE,
  qp_monotonic_increase = FALSE,
  qp_monotonic_decrease = FALSE,
  qp_range_upper = NULL,
  qp_range_lower = NULL,
  qp_Amat_fxn = NULL,
  qp_bvec_fxn = NULL,
  qp_meq_fxn = NULL,
  qp_Amat = NULL,
  qp_bvec = NULL,
  qp_meq = 0,
  all_derivatives_fxn = NULL,
  og_cols = NULL,
  include_warnings = TRUE,
  ...
)
}
\arguments{
\item{X}{List of per-partition design matrices.}

\item{K}{Integer. Number of interior knots.}

\item{p_expansions}{Integer. Number of basis expansions per partition.}

\item{order_list}{List of per-partition observation index vectors.}

\item{colnm_expansions}{Character vector of expansion column names.}

\item{expansion_scales}{Numeric vector of expansion standardization scales.}

\item{power1_cols}{Integer vector of linear-term column indices.}

\item{power2_cols}{Integer vector of quadratic-term column indices.}

\item{nonspline_cols}{Integer vector of non-spline linear column indices.}

\item{interaction_single_cols}{Integer vector of linear-by-linear
interaction column indices.}

\item{interaction_quad_cols}{Integer vector of linear-by-quadratic
interaction column indices.}

\item{triplet_cols}{Integer vector of three-way interaction column indices.}

\item{include_2way_interactions}{Logical switch controlling whether
two-way interactions are included in derivative construction.}

\item{include_3way_interactions}{Logical switch controlling whether
three-way interactions are included in derivative construction.}

\item{include_quadratic_interactions}{Logical switch controlling whether
quadratic interactions are included in derivative construction.}

\item{family}{GLM family object.}

\item{mean_y, sd_y}{Numeric scalars for response standardization.}

\item{N_obs}{Integer. Total sample size.}

\item{qp_observations}{Optional integer vector of observation indices.}

\item{qp_positive_derivative, qp_negative_derivative}{Logical scalar,
character vector, or integer vector. See section
\emph{Per-Variable Derivative Constraints}.}

\item{qp_positive_2ndderivative, qp_negative_2ndderivative}{Same as above
but for second derivatives.}

\item{qp_monotonic_increase, qp_monotonic_decrease}{Logical.
Constrain fitted values to be monotonic in observation order.}

\item{qp_range_upper, qp_range_lower}{Optional numeric upper/lower bounds
for fitted values.}

\item{qp_Amat_fxn, qp_bvec_fxn, qp_meq_fxn}{Optional user-supplied
constraint-generating functions.}

\item{qp_Amat, qp_bvec, qp_meq}{Optional pre-built QP objects. Their presence
marks QP handling as active, but this helper does not append them to the
built-in constraints it constructs; they are expected to be handled
outside this constructor.}

\item{all_derivatives_fxn}{Function to compute derivatives from expansion
matrices (the \code{all_derivatives} closure from \code{lgspline.fit}).}

\item{og_cols}{Optional character vector of original predictor column names.}

\item{include_warnings}{Logical. Whether to issue warnings.}

\item{...}{Additional arguments forwarded to custom constraint functions.}
}
\value{
A list with components:
\describe{
  \item{qp_Amat}{\eqn{P \times M} combined constraint matrix.}
  \item{qp_bvec}{Numeric vector of length \eqn{M}.}
  \item{qp_meq}{Integer. Number of leading equality constraints.}
  \item{quadprog}{Logical. \code{TRUE} if any QP constraints are active.}
}
}
\description{
Constructs \code{qp_Amat}, \code{qp_bvec}, and \code{qp_meq} from the
built-in quadratic programming constraints handled here: range bounds,
derivative sign constraints (first and second), monotonicity constraints,
and user-supplied custom constraint functions.

This function was refactored from the inline QP setup block in
\code{\link{lgspline.fit}} to improve readability and testability.
The interface is fully backward-compatible: existing calls that pass
\code{TRUE}/\code{FALSE} for derivative flags continue to work.
}
\section{Per-Variable Derivative Constraints}{

The arguments \code{qp_positive_derivative}, \code{qp_negative_derivative},
\code{qp_positive_2ndderivative}, and \code{qp_negative_2ndderivative}
now accept three forms:
\describe{
  \item{\code{FALSE}}{No constraint (default).}
  \item{\code{TRUE}}{Constrain \strong{all} predictor variables (backward
    compatible with previous behavior).}
  \item{Character or integer vector}{Constrain only the specified
    predictor variables. Character entries are resolved via
    \code{og_cols}; integer entries refer to column positions in the
    predictor matrix.}
}

This allows, for example, enforcing a positive first derivative for
variable \code{"Dose"} and a negative first derivative for variable
\code{"Time"} simultaneously:
\preformatted{
  lgspline(...,
           qp_positive_derivative = "Dose",
           qp_negative_derivative = "Time")
}

The arguments \code{qp_monotonic_increase} and
\code{qp_monotonic_decrease} remain \code{TRUE}/\code{FALSE} only,
because they constrain fitted values in observation order (not
per-variable).
}

\examples{
\dontrun{
## Standalone verification: simple 1-D monotonic increase
set.seed(42)
t <- seq(-5, 5, length.out = 200)
y <- 3 * sin(t) + t + rnorm(200, 0, 0.5)

## Fit with positive first-derivative constraint on all variables
fit1 <- lgspline(t, y, K = 3,
                 qp_positive_derivative = TRUE)

## Verify: first derivative should be >= 0 everywhere
derivs1 <- predict(fit1, new_predictors = sort(t),
                   take_first_derivatives = TRUE)
stopifnot(all(derivs1$first_deriv >= -1e-8))

## Fit with monotonic increase (observation-order)
fit2 <- lgspline(t, y, K = 3,
                 qp_monotonic_increase = TRUE)
preds2 <- predict(fit2, new_predictors = sort(t))
stopifnot(all(diff(preds2) >= -1e-8))

## Per-variable constraints: 2-D example
x1 <- runif(500, -5, 5)
x2 <- runif(500, -5, 5)
y2 <- x1 + sin(x2) + rnorm(500, 0, 0.5)
dat2 <- data.frame(x1 = x1, x2 = x2, y = y2)

## Constrain x1 to have positive derivative, x2 to have negative
fit3 <- lgspline(y ~ spl(x1, x2), data = dat2, K = 2,
                 qp_positive_derivative = "x1",
                 qp_negative_derivative = "x2")

## Verify per-variable derivatives
newdat <- expand.grid(x1 = seq(-4, 4, length.out = 50),
                      x2 = seq(-4, 4, length.out = 50))
d3 <- predict(fit3, new_predictors = newdat,
              take_first_derivatives = TRUE)

## x1 derivative should be >= 0
stopifnot(all(unlist(d3$first_deriv[["_1_"]]) >= -1e-6))
## x2 derivative should be <= 0
stopifnot(all(unlist(d3$first_deriv[["_2_"]]) <= 1e-6))

## Numeric column indices work identically
fit4 <- lgspline(y ~ spl(x1, x2), data = dat2, K = 2,
                 qp_positive_derivative = 1,
                 qp_negative_derivative = 2)

## Range + derivative constraints simultaneously
fit5 <- lgspline(t, y, K = 3,
                 qp_positive_derivative = TRUE,
                 qp_range_lower = -5,
                 qp_range_upper = 15)
preds5 <- predict(fit5)
stopifnot(all(preds5 >= -5 - 1e-6))
stopifnot(all(preds5 <= 15 + 1e-6))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blockfit_solve.R
\name{.bf_deviance}
\alias{.bf_deviance}
\title{Compute Deviance for Non-GEE Models}
\usage{
.bf_deviance(y_vec, mu_vec, obs_wt, ord, fam, ...)
}
\arguments{
\item{y_vec}{Numeric vector of observed responses.}

\item{mu_vec}{Numeric vector of fitted means.}

\item{obs_wt}{Numeric vector of observation weights.}

\item{ord}{Integer vector of observation indices (passed to
custom deviance residual functions).}

\item{fam}{GLM family object.}

\item{...}{Additional arguments forwarded to
\code{fam$custom_dev.resids}.}
}
\value{
Scalar; the mean deviance contribution.
}
\description{
Evaluates the mean deviance (or mean squared error as fallback) at
the current fitted values.  Used inside the damped Newton-Raphson outer loop of
\code{\link{blockfit_solve}} for convergence monitoring.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cox_helpers.R
\name{cox_family}
\alias{cox_family}
\title{Cox Proportional Hazards Family for lgspline}
\usage{
cox_family()
}
\value{
A list with family components used by \code{lgspline}.
}
\description{
Creates a family-like object for Cox PH models. The link function is
\code{log} (the linear predictor is log-relative-hazard), but unlike
standard GLM families there is no dispersion parameter and no closed-form
mean-variance relationship.

The family provides \code{$loglik} and \code{$aic} methods compatible
with \code{\link{logLik.lgspline}}.
}
\details{
Cox PH is semiparametric: the baseline hazard is unspecified. The
partial log-likelihood depends only on the order of event times and the
linear predictor \eqn{\eta = \mathbf{X}\boldsymbol{\beta}}. Consequently:
\itemize{
  \item No dispersion parameter is estimated (\code{sigmasq_tilde} is
        fixed at 1).
  \item \code{dev.resids} returns martingale-style residuals for GCV
        tuning compatibility.
  \item The response variable is survival time (positive), and the link
        is \code{log}.
}
}
\examples{
fam <- cox_family()
fam$family
fam$link

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{create_block_diagonal}
\alias{create_block_diagonal}
\title{Create Block Diagonal Matrix}
\usage{
create_block_diagonal(matrix_list)
}
\arguments{
\item{matrix_list}{List of matrices to arrange diagonally}
}
\value{
Block diagonal matrix with input matrices on diagonal
}
\description{
Create Block Diagonal Matrix
}
\details{
Takes in a list of matrices, and returns a block-diagonal matrix with each
element of the list as one block. All off-diagonal elements are 0.
Matrices should be square and have compatible dimensions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/leave_one_out.R
\name{leave_one_out}
\alias{leave_one_out}
\title{Compute Leave-One-Out Cross-Validated Predictions for
Gaussian Response/Identity Link under Constraint}
\usage{
leave_one_out(model_fit, leverage_threshold = 100)
}
\arguments{
\item{model_fit}{A fitted lgspline model object.}

\item{leverage_threshold}{Numeric scalar. Observations with
\eqn{H_{ii} \geq} \code{leverage_threshold} are treated as
high-leverage for the warning below. Default \code{100}.}
}
\value{
A vector of leave-one-out cross-validated predictions
}
\description{
Computes the leave-one-out cross-validated predictions from a model
fit, assuming Gaussian-distributed response with identity link.

The LOO closed-formula for observation \eqn{i} is \eqn{\hat{y}_{(-i)} = y_i -
\frac{1}{1 - H_{ii}}(y_i - \hat{y}_i)} where
\eqn{\mathbf{H}} is the effective hat matrix under
smoothing constraints, adjusted for weights and correlation structure if
present.

Observations with leverage at or above \code{leverage_threshold} are flagged
in a warning, since extreme hat values can make the shortcut numerically
unreliable. The default \code{leverage_threshold = 100} is intentionally
permissive, so users who want diagnostic warnings for large \eqn{H_{ii}}
should set a smaller threshold explicitly.

For related discussion of prediction-sum-of-squares calculations under
linear restrictions, see Tarpey (2000), who studies the PRESS statistic for
restricted least squares. That setting is closely related to the
constraint-adjusted hat-matrix shortcut used here.
}
\examples{

## Basic usage with Gaussian response, computing PRESS
set.seed(1234)
t <- rnorm(50)
y <- sin(t) + rnorm(50, 0, .25)
model_fit <- lgspline(t, y)
loo <- leave_one_out(model_fit)
press <- mean((y - loo)^2, na.rm = TRUE)

plot(loo, y,
   main = "LOO Cross-Validation Prediction vs. Observed Response",
   xlab = 'Prediction', ylab = 'Response')
abline(0, 1)

}
\references{
Tarpey, T. (2000). A note on the prediction sum of squares statistic for
restricted least squares. \emph{The American Statistician}, 54(2), 116--118.
\doi{10.2307/2686028}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.build_lambda_block}
\alias{.build_lambda_block}
\title{Build Block-Diagonal Penalty Matrix}
\usage{
.build_lambda_block(Lambda, K, unique_penalty_per_partition, L_partition_list)
}
\arguments{
\item{Lambda}{Shared \eqn{p \times p} penalty matrix.}

\item{K}{Integer; number of interior knots (\eqn{K+1} partitions).}

\item{unique_penalty_per_partition}{Logical; if \code{TRUE}, add
\code{L_partition_list[[k]]} to the \eqn{k}-th diagonal block.}

\item{L_partition_list}{List of partition-specific \eqn{p \times p}
penalty matrices.}
}
\value{
A \eqn{P \times P} block-diagonal matrix where
  \eqn{P = p \times (K+1)}.
}
\description{
Thin wrapper around \code{.solver_build_lambda_block()}.
Keeping the local helper name preserves existing call sites in
\code{get_B()} while centralizing the shared implementation in
\code{solver_utils.R}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.wald_lgspline}
\alias{print.wald_lgspline}
\title{Print Method for wald_lgspline Objects}
\usage{
\method{print}{wald_lgspline}(
  x,
  digits = max(3, getOption("digits") - 3),
  signif.stars = getOption("show.signif.stars"),
  ...
)
}
\arguments{
\item{x}{A \code{"wald_lgspline"} object from \code{\link{wald_univariate}}.}

\item{digits}{Number of significant digits.}

\item{signif.stars}{Logical; show significance stars.}

\item{...}{Additional arguments passed to \code{\link[stats]{printCoefmat}}.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Prints the coefficient table using \code{\link[stats]{printCoefmat}} with
significance stars.
}
\seealso{
\code{\link{wald_univariate}}, \code{\link[stats]{printCoefmat}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\docType{methods}
\name{lgspline-methods}
\alias{lgspline-methods}
\title{lgspline: S3 Methods}
\description{
S3 methods for lgspline objects: print, summary, coef, plot, predict,
confint, logLik, and inference helpers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_Lambda.R
\name{.compute_tuning_predictions}
\alias{.compute_tuning_predictions}
\title{Compute Predictions During Penalty Tuning}
\usage{
.compute_tuning_predictions(
  X,
  B,
  K,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks
)
}
\arguments{
\item{X}{List; design matrices by partition.}

\item{B}{List; coefficient vectors by partition.}

\item{K}{Integer; number of interior knots.}

\item{parallel}{Logical; use parallel computation.}

\item{cl}{Parallel cluster object.}

\item{chunk_size, num_chunks, rem_chunks}{Integer; parallel chunking parameters.}
}
\value{
List of prediction vectors, one per partition.
}
\description{
Wrapper around \code{matmult_block_diagonal} for computing partition-wise
predictions \eqn{\mathbf{X}_{k} \boldsymbol{\beta}_{k}} during penalty
tuning.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.woodbury_redecompose_weighted}
\alias{.woodbury_redecompose_weighted}
\title{Per-Iteration Weighted Woodbury Redecomposition}
\usage{
.woodbury_redecompose_weighted(
  Delta_V,
  X_block,
  DV_X,
  W,
  X,
  K,
  p_expansions,
  X_gram_weighted,
  Lambda,
  schur_corrections,
  unique_penalty_per_partition,
  L_partition_list,
  parallel_eigen,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  rank_threshold_fraction = 1/3,
  family
)
}
\arguments{
\item{Delta_V}{Fixed \eqn{N \times N} matrix
\eqn{\mathbf{V}^{-1} - \mathbf{I}}.}

\item{X_block}{Full \eqn{N \times P} block-diagonal design matrix.}

\item{DV_X}{Precomputed \eqn{N \times P} product
of \eqn{(\mathbf{V}^{-1} - \mathbf{I})\mathbf{X}}.}

\item{W}{Length-\eqn{N} vector of current GLM working weights.}

\item{X}{List of partition-specific design matrices.}

\item{K, p_expansions}{Integer dimensions.}

\item{X_gram_weighted}{List of \eqn{K+1} weighted Gram matrices
\eqn{\mathbf{X}_k^{\top}\mathrm{diag}(\mathbf{W}_k)\mathbf{X}_k}.}

\item{Lambda}{Shared \eqn{p \times p} penalty matrix.}

\item{schur_corrections}{List of \eqn{K+1} Schur correction matrices.}

\item{unique_penalty_per_partition}{Logical.}

\item{L_partition_list}{Partition-specific penalty matrices.}

\item{parallel_eigen}{Logical; parallel eigendecomposition.}

\item{cl, chunk_size, num_chunks, rem_chunks}{Parallel parameters.}

\item{rank_threshold_fraction}{Numeric; Woodbury threshold (default 1/3).}

\item{family}{GLM family object.}
}
\value{
A list with the same structure as \code{.woodbury_decompose_V}:
\describe{
  \item{use_woodbury}{Logical; FALSE if rank exceeds threshold.}
  \item{Ghalf_corrected}{List of corrected \eqn{\mathbf{G}_{\mathrm{on},k}^{1/2}}.}
  \item{GhalfInv_corrected}{List of corrected \eqn{\mathbf{G}_{\mathrm{on},k}^{-1/2}}.}
  \item{G_corrected}{List of corrected \eqn{\mathbf{G}_{\mathrm{on},k}}.}
  \item{E}{Off-diagonal low-rank factor (\eqn{P \times r}), denoted by
    \eqn{\mathbf{E}} in the supplement.}
  \item{J_signs}{Length-\eqn{r} sign vector, i.e.\ the diagonal of
    \eqn{\mathbf{J}}.}
  \item{r}{Integer effective rank.}
  \item{inner_inv}{Precomputed \eqn{r \times r} Woodbury inner inverse; in
    supplement notation,
    \eqn{(\mathbf{J}^{-1} + \mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}
    \mathbf{E})^{-1}}.}
  \item{GL}{List of per-partition matrices corresponding to
    \eqn{\mathbf{G}_{\mathrm{on},k}\mathbf{E}[rows_k,]} in supplement notation.}
  \item{L,S_signs,M}{Legacy internal aliases for
    \code{E}, \code{J_signs}, and \code{inner_inv}.}
}
}
\description{
Weighted analogue of \code{.woodbury_decompose_V()}. At each GLM
iteration, this helper rebuilds the current correlated information,
absorbs the block-diagonal part into
\eqn{\mathbf{G}_{\mathrm{on}}^{-1}}, and refactors the cross-partition
remainder as \eqn{\mathbf{E}\mathbf{J}\mathbf{E}^{\top}}.
}
\details{
It is standalone in the sense that it returns the full Woodbury state
for the current working weights:
corrected block-diagonal factors, the low-rank basis, and the small
inverse \eqn{(\mathbf{J}^{-1} +
\mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}\mathbf{E})^{-1}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cox_helpers.R
\name{cox_helpers}
\alias{cox_helpers}
\title{Cox Proportional Hazards Helpers for lgspline}
\description{
Functions for fitting Cox proportional hazards regression models within the
lgspline framework. Analogous to the Weibull AFT helpers, these provide
the partial log-likelihood, score, information, and all interface functions
needed by lgspline's unconstrained fitting, penalty tuning, and inference
machinery.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{matmult_block_diagonal}
\alias{matmult_block_diagonal}
\title{Multiply Block Diagonal Matrices in Parallel}
\usage{
matmult_block_diagonal(
  A,
  B,
  K,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks
)
}
\arguments{
\item{A}{List of matrices forming first block diagonal matrix}

\item{B}{List of matrices forming second block diagonal matrix}

\item{K}{Number of blocks minus 1 (\eqn{K})}

\item{parallel}{Logical; whether to use parallel processing}

\item{cl}{Cluster object for parallel processing}

\item{chunk_size}{Number of blocks per chunk for parallel processing}

\item{num_chunks}{Number of chunks for parallel processing}

\item{rem_chunks}{Remaining blocks after chunking}
}
\value{
List containing products of corresponding blocks
}
\description{
Multiplies two lists of matrices that form block diagonal structures, with optional
parallel processing.
}
\details{
When \code{parallel=TRUE}, splits computation into chunks processed in parallel.
Handles remainder chunks separately. Uses \code{matmult_block_diagonal_cpp()} for
actual multiplication.

The function expects A and B to contain corresponding blocks that can be
matrix multiplied.
}
\examples{

A <- list(matrix(1:4,2,2), matrix(5:8,2,2))
B <- list(matrix(1:4,2,2), matrix(5:8,2,2))
matmult_block_diagonal(A, B, K=1, parallel=FALSE, cl=NULL,
                       chunk_size=1, num_chunks=1, rem_chunks=0)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blockfit_solve.R
\name{.bf_make_Lambda_block}
\alias{.bf_make_Lambda_block}
\title{Construct Full Block-Diagonal Penalty Matrix}
\usage{
.bf_make_Lambda_block(
  Lambda,
  K,
  unique_penalty_per_partition,
  L_partition_list
)
}
\arguments{
\item{Lambda}{\eqn{p_expansions \times p_expansions} shared penalty matrix.}

\item{K}{Integer; number of interior knots.}

\item{unique_penalty_per_partition}{Logical.}

\item{L_partition_list}{List of \eqn{K+1} partition-specific
penalty matrices (or zeros).}
}
\value{
A \eqn{(p_expansions \cdot (K+1)) \times (p_expansions \cdot (K+1))} matrix.
}
\description{
Thin wrapper around \code{.solver_build_lambda_block()}.
The block/backfitting solver keeps its original helper name so the
surrounding code stays unchanged, while the shared implementation now
lives in \code{solver_utils.R}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/negbin_helpers.R
\name{negbin_family}
\alias{negbin_family}
\title{Negative Binomial Family for lgspline}
\usage{
negbin_family()
}
\value{
A list with family components used by \code{lgspline}.
}
\description{
Creates a family-like object for NB2 regression. The link is log, the
variance function is \eqn{V(\mu) = \mu + \mu^2/\theta}, and the
dispersion stored by lgspline (\code{sigmasq_tilde}) is the shape
parameter \eqn{\theta}.
}
\details{
The NB2 model has a nuisance shape parameter \eqn{\theta} analogous
to the Weibull scale parameter. It is estimated jointly with
\eqn{\boldsymbol{\beta}} and its uncertainty is propagated via the
Schur complement correction.
}
\examples{
fam <- negbin_family()
fam$family
fam$link

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{nr_iterate}
\alias{nr_iterate}
\title{Compute Newton-Raphson Parameter Update with Numerical Stabilization}
\usage{
nr_iterate(gradient_val, neghessian_val)
}
\arguments{
\item{gradient_val}{Numeric vector of gradient values (\eqn{\textbf{u}})}

\item{neghessian_val}{Negative Hessian matrix (\eqn{\textbf{G}^{-1}}
approximately)}
}
\value{
Numeric column vector of parameter updates (\eqn{\textbf{G}\textbf{u}})
}
\description{
Performs parameter update in iterative optimization.

Called by \code{\link{damped_newton_r}} in the update step
}
\details{
This helper function is a core component of Newton-Raphson optimization.
It provides a computationally-stable approach to computing
\eqn{\textbf{G}\textbf{u}}, for information matrix \eqn{\textbf{G}} and
score vector \eqn{\textbf{u}}, where the Newton-Raphson update can be
expressed as
\eqn{\boldsymbol{\beta}^{(m+1)} = \boldsymbol{\beta}^{(m)} + \textbf{G}\textbf{u}}.
}
\seealso{
\code{\link{damped_newton_r}} for the full optimization routine
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/negbin_helpers.R
\name{loglik_negbin}
\alias{loglik_negbin}
\title{Compute Negative Binomial Log-Likelihood}
\usage{
loglik_negbin(y, mu, theta, weights = 1)
}
\arguments{
\item{y}{Non-negative integer response vector.}

\item{mu}{Positive mean vector, same length as \code{y}.}

\item{theta}{Positive scalar shape parameter.}

\item{weights}{Optional observation weights (default 1).}
}
\value{
Scalar log-likelihood value.
}
\description{
Evaluates the NB2 log-likelihood for given mean vector and shape parameter.
}
\details{
The log-likelihood is
\deqn{\ell(\mu, \theta) = \sum_i w_i \bigl[
  \log\Gamma(y_i + \theta) - \log\Gamma(\theta) - \log\Gamma(y_i + 1)
  + \theta\log\theta - \theta\log(\mu_i + \theta)
  + y_i\log\mu_i - y_i\log(\mu_i + \theta)\bigr]}
}
\examples{
set.seed(1234)
mu <- exp(rnorm(50))
y <- rpois(50, mu)
loglik_negbin(y, mu, theta = 5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/solver_utils.R
\name{.solver_build_lambda_block}
\alias{.solver_build_lambda_block}
\title{Build Block-Diagonal Penalty Matrix}
\usage{
.solver_build_lambda_block(
  Lambda,
  K,
  unique_penalty_per_partition,
  L_partition_list
)
}
\arguments{
\item{Lambda}{Shared \eqn{p \times p} penalty matrix.}

\item{K}{Integer; number of interior knots (\eqn{K+1} partitions).}

\item{unique_penalty_per_partition}{Logical; if \code{TRUE}, add the
corresponding element of \code{L_partition_list} to each block.}

\item{L_partition_list}{List of \eqn{K+1} partition-specific
\eqn{p \times p} penalty matrices.}
}
\value{
A \eqn{P \times P} block-diagonal matrix,
  \eqn{P = p \times (K+1)}.
}
\description{
Assembles the full \eqn{P \times P} block-diagonal penalty matrix
\eqn{\boldsymbol{\Lambda}} from a shared per-partition penalty
\code{Lambda} and optional partition-specific additive terms.
\eqn{P = p \times (K+1)} where \eqn{p} is the number of basis terms
per partition.
}
\details{
When \code{unique_penalty_per_partition = TRUE}, the \eqn{k}-th
diagonal block is \code{Lambda + L_partition_list[[k]]}; otherwise
every block is \code{Lambda}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{efficient_bfgs}
\alias{efficient_bfgs}
\title{BFGS Implementation for REML Parameter Estimation}
\usage{
efficient_bfgs(par, fn, control = list())
}
\arguments{
\item{par}{Numeric vector of initial parameter values.}

\item{fn}{Function returning list(objective, gradient). Must return both objective
value and gradient vector matching length(par).}

\item{control}{List of control parameters:
\describe{
  \item{maxit}{Maximum iterations, default 50}
  \item{abstol}{Absolute convergence tolerance, default sqrt(.Machine$double.eps)}
  \item{reltol}{Relative convergence tolerance, default sqrt(.Machine$double.eps)}
  \item{initial_damp}{Initial damping factor, default 1}
  \item{min_damp}{Minimum damping before termination, default 2^-16}
  \item{trace}{Print iteration progress, default FALSE}
}}
}
\value{
List containing:
\describe{
  \item{par}{Parameter vector minimizing objective}
  \item{value}{Minimum objective value}
  \item{counts}{Number of iterations}
  \item{convergence}{TRUE if termination occurred before \code{maxit};
    this reflects the current stopping rule rather than a separate
    post-hoc convergence check}
  \item{message}{Description of termination status}
  \item{vcov}{Final approximation of inverse-Hessian, useful for inference}
}
}
\description{
BFGS optimizer designed for REML optimization of correlation parameters. Combines
function evaluation and gradient computation into single call to avoid redundant
model refitting.
}
\details{
Implements BFGS, used internally by \code{lgspline()} for optimizing
correlation parameters via REML when an analytic REML gradient is supplied.

This is more efficient than native BFGS, since gradient and loss can be computed simultaneously,
avoiding re-computing components in "fn" and "gr" separately.
}
\examples{


## Minimize Rosenbrock function
fn <- function(x) {
  # Objective
  f <- 100*(x[2] - x[1]^2)^2 + (1-x[1])^2
  # Gradient
  g <- c(-400*x[1]*(x[2] - x[1]^2) - 2*(1-x[1]),
         200*(x[2] - x[1]^2))
  list(f, g)
}
(res <- efficient_bfgs(c(0.5, 2.5), fn))

## Compare to
(res0 <- stats::optim(c(0.5, 2.5), function(x)fn(x)[[1]], hessian = TRUE))
solve(res0$hessian)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_Lambda.R
\name{.fit_coefficients}
\alias{.fit_coefficients}
\title{Fit Coefficients During Penalty Tuning: blockfit_solve or get_B}
\usage{
.fit_coefficients(G_list, Lambda, L_partition_list, env, return_G_getB, ...)
}
\arguments{
\item{G_list}{List; eigendecomposition results from \code{compute_G_eigen},
containing \code{Ghalf} and \code{GhalfInv}.}

\item{Lambda}{Matrix; current combined penalty matrix.}

\item{L_partition_list}{List; partition-specific penalty matrices.}

\item{env}{List; tuning environment from \code{.build_tuning_env}.}

\item{return_G_getB}{Logical; whether to return G inside the fit. Set to
TRUE within \code{.compute_gcvu} and \code{.compute_gcvu_gradient} so
that \code{B_list$G_list} carries the (possibly GLM-iterated) G matrices
required for the subsequent \code{AGAmult_wrapper} and trace computations.}

\item{...}{Additional arguments forwarded to the fitting routine.}
}
\value{
List; output of \code{blockfit_solve} or \code{get_B}, containing
  at minimum \code{$B} (coefficient list) and \code{$G_list}.
}
\description{
Dispatches to \code{blockfit_solve} when the blockfit conditions are met
(i.e. \code{env$use_blockfit} is TRUE), otherwise calls \code{get_B}.
On \code{blockfit_solve} failure, falls back to \code{get_B} automatically.
}
\details{
The blockfit condition mirrors \code{lgspline.fit}:
\code{blockfit && length(flat_cols) > 0 && K > 0}, pre-computed in
\code{tune_Lambda} and stored in \code{env$use_blockfit}.

\code{return_G_getB} is set to TRUE by the callers so that
\code{B_list$G_list} contains the updated G matrices (after any GLM
weight iteration inside \code{get_B} or \code{blockfit_solve}).
These are needed immediately after this call for \code{AGAmult_wrapper},
\code{GXX}, and the trace computation.

When correlation structure inputs are supplied, this wrapper does not
introduce a separate tuning-specific notation or solver path. Instead, the
same correlated coefficient estimator used in the final model fit is called
here inside each tuning-objective evaluation. In particular, any
structured-correlation Woodbury factorization
\eqn{\mathbf{G}^{-1} =
\mathbf{G}_{\mathrm{on}}^{-1} + \mathbf{G}_{\mathrm{off}}^{-1}}
is handled inside \code{get_B} and documented in \code{lgspline-details}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{.compute_loocv_gradient}
\alias{.compute_loocv_gradient}
\title{Compute Closed-Form Gradient of Exact Leave-One-Out Criterion}
\usage{
.compute_loocv_gradient(par, log_penalty_vec, outlist = NULL, env, ...)
}
\arguments{
\item{par}{Numeric vector; log-scale penalty parameters.}

\item{log_penalty_vec}{Numeric vector; log-scale predictor/partition
penalties.}

\item{outlist}{List or NULL; cached results from \code{.compute_loocv()}.}

\item{env}{List; tuning environment.}

\item{...}{Additional arguments passed to fitting functions.}
}
\value{
List with \code{criterion_value}, \code{gradient}, and
  \code{outlist}.
}
\description{
Computes the analytical gradient of the exact leave-one-out tuning
criterion with respect to the log penalty parameters.
}
\details{
The implementation below keeps the exact LOO gradient fully analytic,
including the leverage derivative. In empirical diagnostics, the
observation-wise leverage component can be numerically delicate even when
the aggregate gradient remains informative; see the developer diagnostics in
\code{inst/diagnostics/gradient_diagnostics.R} for inspection tools.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{get_polynomial_expansions}
\alias{get_polynomial_expansions}
\title{Generate Design Matrix with Polynomial and Interaction Terms}
\usage{
get_polynomial_expansions(
  predictors,
  numerics,
  just_linear_with_interactions,
  just_linear_without_interactions,
  exclude_interactions_for = NULL,
  include_quadratic_terms = TRUE,
  include_cubic_terms = TRUE,
  include_quartic_terms = FALSE,
  include_2way_interactions = TRUE,
  include_3way_interactions = TRUE,
  include_quadratic_interactions = FALSE,
  exclude_these_expansions = NULL,
  custom_basis_fxn = NULL,
  ...
)
}
\arguments{
\item{predictors}{Numeric matrix of predictor variables}

\item{numerics}{Integer vector; column indices for variables to expand as polynomials}

\item{just_linear_with_interactions}{Integer vector; column indices for variables to keep linear but allow interactions}

\item{just_linear_without_interactions}{Integer vector; column indices for variables to keep linear without interactions}

\item{exclude_interactions_for}{Integer vector; column indices to exclude from all interactions}

\item{include_quadratic_terms}{Logical; whether to include squared terms (default TRUE)}

\item{include_cubic_terms}{Logical; whether to include cubic terms (default TRUE)}

\item{include_quartic_terms}{Logical; whether to include 4th degree terms (default FALSE)}

\item{include_2way_interactions}{Logical; whether to include two-way interactions (default TRUE)}

\item{include_3way_interactions}{Logical; whether to include three-way interactions (default TRUE)}

\item{include_quadratic_interactions}{Logical; whether to include interactions with squared terms (default TRUE)}

\item{exclude_these_expansions}{Character vector; names of specific terms to exclude from final matrix}

\item{custom_basis_fxn}{Function; optional custom basis expansion function that accepts all arguments listed here except itself}

\item{...}{Additional arguments passed to \code{custom_basis_fxn}}
}
\value{
Matrix with columns for intercept, polynomial terms, and specified interactions
}
\description{
Internal function for creating a design matrix containing polynomial
expansions and interaction terms for predictor variables. Supports
customizable term generation including polynomial degrees up to quartic
terms, interaction types, and selective term exclusion.

Column names take on the form "_v_" for linear terms, "_v_^d" for polynomial
powers up to d = 4, and "_v_x_w_" for interactions between variables v and w,
where v and w are column indices of the input predictor matrix.

The \code{custom_basis_fxn} argument, if supplied, requires the same arguments
as this function, in the same order, excluding the argument
"custom_basis_fxn" itself.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{compute_trace_UGXX_wrapper}
\alias{compute_trace_UGXX_wrapper}
\title{Calculate Trace of Matrix Product \eqn{\text{trace}(\textbf{X}\textbf{U}\textbf{G}\textbf{X}^{T})}}
\usage{
compute_trace_UGXX_wrapper(
  G,
  A,
  GXX,
  AGAInv,
  nc,
  nca,
  K,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks
)
}
\arguments{
\item{G}{List of G matrices (\eqn{\textbf{G}})}

\item{A}{Constraint matrix (\eqn{\textbf{A}})}

\item{GXX}{List of \eqn{\textbf{G}\textbf{X}^{T}\textbf{X}} products}

\item{AGAInv}{Inverse of \eqn{\textbf{A}^{T}\textbf{G}\textbf{A}}}

\item{nc}{Number of columns}

\item{nca}{Number of constraint columns}

\item{K}{Number of partitions minus 1 (\eqn{K})}

\item{parallel}{Use parallel processing}

\item{cl}{Cluster object}

\item{chunk_size}{Size of parallel chunks}

\item{num_chunks}{Number of chunks}

\item{rem_chunks}{Remaining chunks}
}
\value{
Trace value
}
\description{
Calculate Trace of Matrix Product \eqn{\text{trace}(\textbf{X}\textbf{U}\textbf{G}\textbf{X}^{T})}
}
\details{
Computes \eqn{\text{trace}(\textbf{X}\textbf{U}\textbf{G}\textbf{X}^{T})} where \eqn{\textbf{U} = \textbf{I} - \textbf{G}\textbf{A}(\textbf{A}^{T}\textbf{G}\textbf{A})^{-1}\textbf{A}^{T}}.
Handles parallel computation by splitting into chunks.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.woodbury_halfsqrt_components}
\alias{.woodbury_halfsqrt_components}
\title{Compute Woodbury Half-Square-Root Components}
\usage{
.woodbury_halfsqrt_components(Ghalf_corrected, E, J_signs, r, K, p_expansions)
}
\arguments{
\item{Ghalf_corrected}{List of corrected \eqn{\mathbf{G}_{\mathrm{on},k}^{1/2}}.}

\item{E}{Off-diagonal low-rank factor (\eqn{P \times r}), denoted
\eqn{\mathbf{E}} in the supplement.}

\item{J_signs}{Length-\eqn{r} sign vector corresponding to the
diagonal of \eqn{\mathbf{J}}.}

\item{r}{Integer effective rank.}

\item{K, p_expansions}{Integer dimensions.}
}
\value{
A list with components:
\describe{
  \item{valid}{Logical; FALSE if \eqn{\mathbf{F}} is not positive
    definite (signals fallback to dense).}
  \item{U_Q}{\eqn{P \times r} left singular vectors of
    \eqn{\mathbf{N} = \mathbf{G}_{\mathrm{on}}^{1/2}\mathbf{E}}. This spans the same
    subspace as the supplement QR factor \eqn{\mathbf{Q}}.}
  \item{Q_basis}{Alias for \code{U_Q}, included so the stored object can
    be read directly against the supplement notation \eqn{\mathbf{Q}}.}
  \item{C}{\eqn{r \times r} diagonal matrix
    encoding the rank-\eqn{r} update in the internal form
    \eqn{\mathbf{F}^{1/2} = \mathbf{I}_P - \mathbf{U}_Q\mathbf{C}
    \mathbf{U}_Q^{\top}}.}
  \item{C_inv}{\eqn{r \times r} diagonal matrix
    encoding the inverse update in the internal form
    \eqn{\mathbf{F}^{-1/2} = \mathbf{I}_P +
    \mathbf{U}_Q\mathbf{C}_{\mathrm{inv}}\mathbf{U}_Q^{\top}}.}
  \item{GchalfUQ}{List of \eqn{K+1} matrices, each \eqn{p \times r}:
    \eqn{\mathbf{G}_{\mathrm{on},k}^{1/2} \mathbf{U}_Q[rows_k,]}.}
}
}
\description{
Converts the low-rank Woodbury decomposition into square-root objects
used directly by the solver. Starting from
\eqn{\mathbf{N} = \mathbf{G}_{\mathrm{on}}^{1/2}\mathbf{E}}, it builds
a basis for the supplement matrix \eqn{\mathbf{Q}} and stores
\eqn{\mathbf{F}^{1/2}} and \eqn{\mathbf{F}^{-1/2}} in the internal forms
\eqn{\mathbf{I} - \mathbf{U}_Q\mathbf{C}\mathbf{U}_Q^{\top}} and
\eqn{\mathbf{I} + \mathbf{U}_Q\mathbf{C}_{\mathrm{inv}}\mathbf{U}_Q^{\top}}.
}
\details{
This helper is standalone: its output is the complete square-root state
consumed by \code{.lagrangian_project_woodbury()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{compute_gram_block_diagonal}
\alias{compute_gram_block_diagonal}
\title{Compute Gram Matrix for Block Diagonal Structure}
\usage{
compute_gram_block_diagonal(
  list_in,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks
)
}
\arguments{
\item{list_in}{List of matrices}

\item{parallel}{Use parallel processing}

\item{cl}{Cluster object}

\item{chunk_size}{Chunk size for parallel}

\item{num_chunks}{Number of chunks}

\item{rem_chunks}{Remaining chunks}
}
\value{
List of Gram matrices (\eqn{\textbf{X}^{T}\textbf{X}}) for each block
}
\description{
Compute Gram Matrix for Block Diagonal Structure
}
\details{
For a list of matrices, will compute the gram matrix of each element of the
list.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/negbin_helpers.R
\name{lgspline_negbin}
\alias{lgspline_negbin}
\title{Fit Negative Binomial Model via lgspline}
\usage{
lgspline_negbin(formula, data, ...)
}
\arguments{
\item{formula}{Formula specifying the model. The response should be
non-negative integer counts.}

\item{data}{Data frame.}

\item{...}{Additional arguments passed to \code{lgspline}, including
\code{Vhalf} and \code{VhalfInv} for correlation structures.}
}
\value{
An object of class \code{"lgspline"}.
}
\description{
Convenience wrapper that calls \code{lgspline} with the correct
family, weight, dispersion, score, and unconstrained-fit functions for
NB2 regression. All standard lgspline arguments (knots, penalties,
constraints, parallel, correlation structures, etc.) are passed through.
}
\details{
Internally sets:
\itemize{
  \item \code{family = negbin_family()}
  \item \code{unconstrained_fit_fxn = unconstrained_fit_negbin}
  \item \code{glm_weight_function = negbin_glm_weight_function}
  \item \code{qp_score_function = negbin_qp_score_function}
  \item \code{dispersion_function = negbin_dispersion_function}
  \item \code{schur_correction_function = negbin_schur_correction}
  \item \code{need_dispersion_for_estimation = TRUE}
  \item \code{estimate_dispersion = TRUE}
  \item \code{standardize_response = FALSE}
}

A formula interface is needed, e.g. \code{lgspline_negbin(t, y)} won't work,
unlike for ordinary \code{\link{lgspline}}.

When a correlation structure is supplied via \code{Vhalf}/\code{VhalfInv},
the model is fitted through the GEE Path 1b machinery in \code{get_B}.
The dispersion function uses \code{VhalfInv} to whiten Pearson
residuals for a better moment-based initialization of \eqn{\theta},
which stabilizes the profile MLE under moderate to strong correlation.
The score function handles the whitened design consistently with the
Weibull AFT GEE convention.
}
\examples{

set.seed(1234)
N <- 300
t <- rnorm(N)
mu <- exp(1 + 0.5 * sin(2 * t))
y <- rnbinom(N, size = 3, mu = mu)
df <- data.frame(response = y, predictor = t)

fit <- lgspline_negbin(
  response ~ spl(predictor),
  data = df,
  K = 2,
  opt = FALSE,
  wiggle_penalty = 1e-2
)
print(summary(fit))
plot(fit, show_formulas = TRUE,
     custom_response_lab = 'Count')
points(t, mu, col = 'grey', cex=0.67)

}
\seealso{
\code{\link{lgspline_cox}} for Cox PH,
  \code{\link{lgspline_weibull}} for Weibull AFT,
  \code{\link{negbin_family}}, \code{\link{unconstrained_fit_negbin}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blockfit_solve.R
\name{.bf_sqp_loop}
\alias{.bf_sqp_loop}
\title{Damped SQP Refinement on the Full (Optionally Whitened) System}
\usage{
.bf_sqp_loop(
  X_design,
  y_design,
  X_block_raw,
  beta_init,
  Lambda_block,
  qp_Amat_combined,
  qp_bvec_combined,
  qp_meq_combined,
  K,
  p_expansions,
  family,
  order_list,
  glm_weight_function,
  schur_correction_function,
  qp_score_function,
  need_dispersion_for_estimation,
  dispersion_function,
  observation_weights,
  iterate,
  tol,
  VhalfInv,
  VhalfInv_perm,
  is_gee,
  deviance_fun,
  X_partitions,
  y_partitions,
  verbose,
  ...
)
}
\arguments{
\item{X_design}{\eqn{N \times P} design matrix (whitened for GEE,
unwhitened otherwise).}

\item{y_design}{\eqn{N \times 1} response vector (whitened for
GEE, unwhitened otherwise).}

\item{X_block_raw}{Unwhitened block-diagonal design matrix (used
for computing the linear predictor on the original scale).
For the non-GEE path this equals \code{X_design}.}

\item{beta_init}{\eqn{P \times 1} initial coefficient vector.}

\item{Lambda_block}{\eqn{P \times P} block-diagonal penalty.}

\item{qp_Amat_combined}{Combined constraint matrix (equalities
then inequalities).}

\item{qp_bvec_combined}{Combined constraint RHS.}

\item{qp_meq_combined}{Number of leading equality constraints.}

\item{K, p_expansions}{Integers.}

\item{family}{GLM family object.}

\item{order_list}{Observation-index lists.}

\item{glm_weight_function, schur_correction_function, }{qp_score_function Functions.}

\item{need_dispersion_for_estimation}{Logical.}

\item{dispersion_function}{Function.}

\item{observation_weights}{List or vector of weights.}

\item{iterate}{Logical; if FALSE, takes at most two steps.}

\item{tol}{Convergence tolerance.}

\item{VhalfInv, VhalfInv_perm}{\eqn{\mathbf{V}^{-1/2}} matrices
(original and permuted ordering).  NULL for non-GEE path.}

\item{is_gee}{Logical; TRUE when called from the GEE refinement
path (affects deviance computation).}

\item{deviance_fun}{Function computing deviance; either
\code{.bf_deviance} or \code{.bf_gee_deviance}.}

\item{X_partitions}{List of per-partition design matrices
(unwhitened); needed for Schur correction in non-GEE path.}

\item{y_partitions}{List of per-partition response vectors
(unwhitened); needed for Schur correction in non-GEE path.}

\item{verbose}{Logical.}

\item{...}{Forwarded to weight, dispersion, and score functions.}
}
\value{
A named list with elements \code{beta_block} (final
  coefficient vector), \code{result} (per-partition coefficient
  list), \code{last_qp_sol} (last successful \code{solve.QP}
  output or NULL), \code{converged} (logical), and
  \code{final_deviance} (scalar).
}
\description{
Runs a damped sequential quadratic programming loop using
\code{quadprog::solve.QP} at each iteration, enforcing all
smoothness equalities, flat-equality constraints, and any
user-supplied inequality constraints.
}
\details{
This function serves two roles within \code{\link{blockfit_solve}}:
\enumerate{
  \item \strong{GEE Stage 2} (Case c): operates on the whitened
    system \eqn{\tilde{\mathbf{X}} = \mathbf{V}^{-1/2}
    \mathbf{X}_{\mathrm{block}}}, initialized from the damped Newton-Raphson
    backfitting warm start.
  \item \strong{Non-GEE SQP refinement}: operates on the
    unwhitened block-diagonal design, applying inequality
    constraints after backfitting convergence.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{plot.wald_lgspline}
\alias{plot.wald_lgspline}
\title{Plot Method for wald_lgspline Objects}
\usage{
\method{plot}{wald_lgspline}(
  x,
  parm = NULL,
  which = NULL,
  main = "Coefficient Estimates and CIs",
  xlab = "Estimate",
  ...
)
}
\arguments{
\item{x}{A \code{"wald_lgspline"} object.}

\item{parm}{Integer vector of coefficient indices or character vector of
names to plot. Default NULL plots all.}

\item{which}{Integer vector of coefficient indices to plot (alternative to
\code{parm}). Default NULL.}

\item{main}{Plot title. Default \code{"Coefficient Estimates and CIs"}.}

\item{xlab}{x-axis label. Default \code{"Estimate"}.}

\item{...}{Additional arguments passed to \code{\link[graphics]{plot}}.}
}
\value{
Invisibly returns NULL.
}
\description{
Forest-style plot of coefficient estimates with confidence intervals.
}
\seealso{
\code{\link{wald_univariate}}, \code{\link{confint.lgspline}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cox_helpers.R
\name{cox_qp_score_function}
\alias{cox_qp_score_function}
\title{Cox PH Score Function for Quadratic Programming and Blockfit}
\usage{
cox_qp_score_function(
  X,
  y,
  mu,
  order_list,
  dispersion,
  VhalfInv,
  observation_weights,
  status
)
}
\arguments{
\item{X}{Block-diagonal design matrix (N x P).}

\item{y}{Response vector (survival times, N x 1).}

\item{mu}{Predicted values (N x 1), same order as X and y.}

\item{order_list}{List of observation indices per partition.}

\item{dispersion}{Dispersion (fixed at 1).}

\item{VhalfInv}{Inverse square root correlation matrix (NULL for
independent observations).}

\item{observation_weights}{Observation weights.}

\item{status}{Event indicator (1 = event, 0 = censored).}
}
\value{
Numeric column vector of length P (gradient w.r.t. coefficients).
}
\description{
Computes the score (gradient of partial log-likelihood) in the format
expected by lgspline's \code{qp_score_function} interface. The block-
diagonal design matrix \eqn{\mathbf{X}} and response \eqn{\mathbf{y}} are
in partition order; this function internally sorts by event time, computes
the Cox score using the Breslow approximation for tied event times, and
returns the result in the original partition order.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{compute_Lambda}
\alias{compute_Lambda}
\title{Construct Smoothing Spline Penalty Matrix}
\usage{
compute_Lambda(
  custom_penalty_mat,
  L1,
  wiggle_penalty,
  flat_ridge_penalty,
  K,
  p_expansions,
  unique_penalty_per_predictor,
  unique_penalty_per_partition,
  penalty_vec,
  colnm_expansions,
  just_Lambda = TRUE
)
}
\arguments{
\item{custom_penalty_mat}{Matrix; optional custom ridge penalty structure}

\item{L1}{Matrix; integrated squared second derivative penalty (\eqn{\textbf{L}_1})}

\item{wiggle_penalty, flat_ridge_penalty}{Numeric; smoothing and ridge penalty parameters}

\item{K}{Integer; number of interior knots (\eqn{K})}

\item{p_expansions}{Integer; number of basis columns per partition}

\item{unique_penalty_per_predictor, unique_penalty_per_partition}{Logical; enable predictor/partition-specific penalties}

\item{penalty_vec}{Named numeric; custom penalty values for predictors/partitions}

\item{colnm_expansions}{Character; column names for linking penalties to predictors}

\item{just_Lambda}{Logical; return only combined penalty matrix (\eqn{\boldsymbol{\Lambda}})}
}
\value{
List containing Lambda, L1, L2, L_predictor_list, L_partition_list;
  or just Lambda if \code{just_Lambda=TRUE} and no partition penalties.
}
\description{
Builds penalty matrix combining smoothing spline and ridge penalties with optional
predictor/partition-specific components.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{compute_dGhalf}
\alias{compute_dGhalf}
\title{Compute Eigen-Based Square-Root Factors for \eqn{d\textbf{G}/d\lambda}}
\usage{
compute_dGhalf(
  dG_dlambda,
  p_expansions,
  K,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks
)
}
\arguments{
\item{dG_dlambda}{List of \eqn{p \times p} \eqn{d\textbf{G}/d\lambda} matrices by partition}

\item{p_expansions}{Integer; number of columns per partition}

\item{K}{Integer; number of interior knots (\eqn{K})}

\item{parallel, cl, chunk_size, num_chunks, rem_chunks}{Parallel computation parameters}
}
\value{
List of \eqn{p \times p} eigen-based square-root factors derived from
  the partition-wise \eqn{d\textbf{G}/d\lambda} matrices
}
\description{
Applies an eigendecomposition-based matrix square root to each partition-wise
\eqn{d\textbf{G}/d\lambda} matrix after replacing \code{NA} entries with 0
and treating all-0 inputs with an identity fallback.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cox_helpers.R
\name{info_cox}
\alias{info_cox}
\title{Compute Cox Observed Information Matrix}
\usage{
info_cox(X, eta, status, y = NULL, weights = 1)
}
\arguments{
\item{X}{Design matrix (N x p), sorted by ascending event time.}

\item{eta}{Linear predictor vector.}

\item{status}{Event indicator (1 = event, 0 = censored).}

\item{y}{Optional numeric vector of observed event/censor times, same length
and order as \code{eta}. When supplied, tied event times are handled with
the Breslow approximation. When omitted, the function assumes there are no
ties.}

\item{weights}{Observation weights (default 1).}
}
\value{
Symmetric p x p observed information matrix.
}
\description{
Negative Hessian of the Cox partial log-likelihood under the Breslow
approximation for tied event times. Data must be sorted by ascending event
time.
}
\details{
Under the Breslow approximation, the observed information is
\deqn{\sum_g d_g^{(w)}
  \Bigl[\frac{S_{2g}}{S_{0g}} -
  \Bigl(\frac{S_{1g}}{S_{0g}}\Bigr)
  \Bigl(\frac{S_{1g}}{S_{0g}}\Bigr)^{\top}\Bigr]}
where
\eqn{S_{0g} = \sum_{j \in R_g} w_j e^{\eta_j}},
\eqn{S_{1g} = \sum_{j \in R_g} w_j e^{\eta_j}\mathbf{x}_j},
and \eqn{S_{2g} = \sum_{j \in R_g} w_j e^{\eta_j}\mathbf{x}_j\mathbf{x}_j^{\top}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{take_derivative}
\alias{take_derivative}
\title{Calculate Derivatives of Polynomial Terms}
\usage{
take_derivative(dat, var, second = FALSE, scale)
}
\arguments{
\item{dat}{Numeric matrix; design matrix containing polynomial basis expansions}

\item{var}{Character; column name of variable to differentiate with respect to}

\item{second}{Logical; if TRUE compute second derivative, if FALSE compute first derivative (default FALSE)}

\item{scale}{Numeric; scaling factor for normalization}
}
\value{
Numeric matrix containing derivatives of polynomial terms, with same dimensions as input matrix
}
\description{
Computes first or second derivatives of polynomial terms in a design matrix with
respect to a specified variable. Handles polynomial terms up to fourth degree.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{make_derivative_matrix}
\alias{make_derivative_matrix}
\title{Compute First and Second Derivative Matrices}
\usage{
make_derivative_matrix(
  p_expansions,
  Cpredictors,
  power1_cols,
  power2_cols,
  nonspline_cols,
  interaction_single_cols,
  interaction_quad_cols,
  triplet_cols,
  K,
  include_2way_interactions,
  include_3way_interactions,
  include_quadratic_interactions,
  colnm_expansions,
  expansion_scales,
  just_first_derivatives = FALSE,
  just_spline_effects = TRUE
)
}
\arguments{
\item{p_expansions}{Number of columns in the basis expansion per partition}

\item{Cpredictors}{Predictor matrix}

\item{power1_cols}{Indices of linear terms of spline effects}

\item{power2_cols}{Indices of quadratic terms of spline effects}

\item{nonspline_cols}{Indices of non-spline effects}

\item{interaction_single_cols}{Indices of first-order interactions}

\item{interaction_quad_cols}{Indices of quadratic interactions}

\item{triplet_cols}{Indices of three-way interactions}

\item{K}{Number of partitions minus 1 (\eqn{K})}

\item{include_2way_interactions}{Include 2-way interactions}

\item{include_3way_interactions}{Include 3-way interactions}

\item{include_quadratic_interactions}{Include quadratic interactions}

\item{colnm_expansions}{Column names}

\item{expansion_scales}{Scale factors}

\item{just_first_derivatives}{Only compute first derivatives}

\item{just_spline_effects}{Only compute derivatives for spline effects}
}
\value{
List containing first-derivative matrices and, unless
  \code{just_first_derivatives = TRUE}, second-derivative matrices
}
\description{
Compute First and Second Derivative Matrices
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.woodbury_decompose_V}
\alias{.woodbury_decompose_V}
\title{Woodbury Decomposition of Correlation Structure}
\usage{
.woodbury_decompose_V(
  VhalfInv_perm,
  X,
  K,
  p_expansions,
  Lambda,
  Lambda_block,
  unique_penalty_per_partition,
  L_partition_list,
  order_list,
  parallel_eigen,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  rank_threshold_fraction = 1/3,
  family
)
}
\arguments{
\item{VhalfInv_perm}{\eqn{\mathbf{V}^{-1/2}} permuted to partition
ordering.}

\item{X}{List of partition-specific design matrices.}

\item{K, p_expansions}{Integer dimensions.}

\item{Lambda, Lambda_block}{Shared and full block-diagonal penalty
matrices.}

\item{unique_penalty_per_partition}{Logical.}

\item{L_partition_list}{Partition-specific penalty matrices.}

\item{order_list}{Partition-to-data index mapping.}

\item{parallel_eigen}{Logical; parallel eigendecomposition.}

\item{cl, chunk_size, num_chunks, rem_chunks}{Parallel parameters.}

\item{rank_threshold_fraction}{Numeric; Woodbury is used only when
\eqn{r < P \times} this fraction (default 1/3).}

\item{family}{GLM family object.}
}
\value{
A list with components:
\describe{
  \item{use_woodbury}{Logical; FALSE triggers dense Path 1 fallback.}
  \item{Ghalf_corrected}{List of corrected \eqn{\mathbf{G}_{\mathrm{on},k}^{1/2}}.}
  \item{GhalfInv_corrected}{List of corrected \eqn{\mathbf{G}_{\mathrm{on},k}^{-1/2}}.}
  \item{G_corrected}{List of corrected \eqn{\mathbf{G}_{\mathrm{on},k}}.}
  \item{E}{Off-diagonal low-rank factor (\eqn{P \times r}) from the
    supplement factorization
    \eqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{E}\mathbf{J}\mathbf{E}^{\top}}.}
  \item{J_signs}{Length-\eqn{r} vector giving the diagonal of
    \eqn{\mathbf{J}}.}
  \item{r}{Integer effective rank.}
  \item{inner_inv}{Precomputed \eqn{r \times r} Woodbury inner inverse. In
    supplement notation this is
    \eqn{(\mathbf{J}^{-1} + \mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}
    \mathbf{E})^{-1}}.}
  \item{GL}{List of \eqn{K+1} matrices storing
    \eqn{\mathbf{G}_{\mathrm{on},k}\mathbf{E}[rows_k,]} in supplement notation,
    each \eqn{p \times r}.}
  \item{L,S_signs,M}{Legacy internal aliases for
    \code{E}, \code{J_signs}, and \code{inner_inv}, retained for
    compatibility within the package code.}
}
}
\description{
Builds the Woodbury factors for the correlated solver. The function
decomposes
\eqn{\mathbf{G}^{-1} = \mathbf{G}_{\mathrm{on}}^{-1} +
\mathbf{G}_{\mathrm{off}}^{-1}}
by absorbing the block-diagonal part of
\eqn{\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X} + \boldsymbol{\Lambda}}
into \eqn{\mathbf{G}_{\mathrm{on}}^{-1}}, then factoring the
cross-partition part as
\eqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{E}\mathbf{J}\mathbf{E}^{\top}}.
}
\details{
The output is exactly the information needed by
\code{.woodbury_halfsqrt_components()} and
\code{.lagrangian_project_woodbury()}:
corrected block-diagonal \eqn{\mathbf{G}_{\mathrm{on}}},
the low-rank factor \eqn{\mathbf{E}}, the sign matrix diagonal
\eqn{\mathbf{J}}, and the small inverse
\eqn{(\mathbf{J}^{-1} +
\mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}\mathbf{E})^{-1}}.

When the cross-partition rank is zero or too large to be worthwhile,
the helper returns \code{use_woodbury = FALSE} so the caller can use
the dense correlated path instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{find_extremum}
\alias{find_extremum}
\title{Find the Extremum of a Fitted lgspline}
\usage{
find_extremum(
  object,
  vars = NULL,
  quick_heuristic = TRUE,
  initial = NULL,
  B_predict = NULL,
  minimize = FALSE,
  stochastic = FALSE,
  stochastic_draw = function(mu, sigma, ...) {
     N <- length(mu)
     rnorm(N, mu,
    sigma)
 },
  sigmasq_predict = object$sigmasq_tilde,
  custom_objective_function = NULL,
  custom_objective_derivative = NULL,
  ...
)
}
\arguments{
\item{object}{A fitted lgspline model object.}

\item{vars}{Integer or character vector; indices or names of predictors to
optimize over. Default NULL optimizes all predictors.}

\item{quick_heuristic}{Logical; if TRUE (default) searches only the
best-performing partition. If FALSE, initiates searches from all
partition local maxima.}

\item{initial}{Numeric vector; optional starting values. Useful for
fixing binary predictors. Default NULL.}

\item{B_predict}{List; optional coefficient list for prediction, e.g.
from \code{\link{generate_posterior}}. Default NULL uses
\code{object$B}.}

\item{minimize}{Logical; find minimum instead of maximum. Default FALSE.}

\item{stochastic}{Logical; add noise during optimization for exploration.
Default FALSE.}

\item{stochastic_draw}{Function; generates noise for stochastic
optimization. Takes \code{mu}, \code{sigma}, and \code{...}.
Default \code{rnorm(length(mu), mu, sigma)}.}

\item{sigmasq_predict}{Numeric; variance for stochastic draws.
Default \code{object$sigmasq_tilde}.}

\item{custom_objective_function}{Function; optional custom objective.
Takes \code{mu}, \code{sigma}, \code{y_best}, \code{...}.
Default NULL.}

\item{custom_objective_derivative}{Function; optional gradient of
\code{custom_objective_function}. Takes \code{mu}, \code{sigma},
\code{y_best}, \code{d_mu}, \code{...}. Default NULL.}

\item{...}{Additional arguments passed to internal optimization routines.}
}
\value{
A list with elements:
\describe{
  \item{t}{Numeric vector; predictor values at the extremum.}
  \item{y}{Numeric; objective value at the extremum.}
}
}
\description{
Finds the global maximum or minimum of a fitted lgspline using L-BFGS-B,
with options for partition-based heuristics, stochastic exploration, and
custom objective functions (e.g., acquisition functions for Bayesian
optimization).
}
\examples{

set.seed(1234)
t <- runif(1000, -10, 10)
y <- 2*sin(t) + -0.06*t^2 + rnorm(length(t))
model_fit <- lgspline(t, y)
plot(model_fit)

max_point <- find_extremum(model_fit)
min_point <- find_extremum(model_fit, minimize = TRUE)
abline(v = max_point$t, col = 'blue')
abline(v = min_point$t, col = 'red')

## Expected improvement acquisition function
ei_obj <- function(mu, sigma, y_best, ...) {
  d <- y_best - mu
  d * pnorm(d/sigma) + sigma * dnorm(d/sigma)
}
ei_deriv <- function(mu, sigma, y_best, d_mu, ...) {
  d <- y_best - mu
  z <- d/sigma
  d_z <- -d_mu/sigma
  pnorm(z)*d_mu - d*dnorm(z)*d_z + sigma*z*dnorm(z)*d_z
}

post_draw <- generate_posterior(model_fit)
acq <- find_extremum(model_fit,
                     stochastic = TRUE,
                     B_predict = post_draw$post_draw_coefficients,
                     sigmasq_predict = post_draw$post_draw_sigmasq,
                     custom_objective_function = ei_obj,
                     custom_objective_derivative = ei_deriv)
abline(v = acq$t, col = 'green')

}
\seealso{
\code{\link{lgspline}}, \code{\link{generate_posterior}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{get_2ndDerivPenalty_wrapper}
\alias{get_2ndDerivPenalty_wrapper}
\title{Wrapper for Integrated Second-Derivative Penalty Computation}
\usage{
get_2ndDerivPenalty_wrapper(
  K,
  colnm_expansions,
  C,
  power1_cols,
  power2_cols,
  power3_cols,
  power4_cols,
  interaction_single_cols,
  interaction_quad_cols,
  triplet_cols,
  nonspline_cols,
  p_expansions,
  parallel,
  cl
)
}
\arguments{
\item{K}{Number of partitions (\eqn{K+1})}

\item{colnm_expansions}{Column names of basis expansions}

\item{C}{Basis expansion matrix used to determine observed predictor ranges}

\item{power1_cols}{Linear term columns}

\item{power2_cols}{Quadratic term columns}

\item{power3_cols}{Cubic term columns}

\item{power4_cols}{Quartic term columns}

\item{interaction_single_cols}{Single interaction columns}

\item{interaction_quad_cols}{Quadratic interaction columns}

\item{triplet_cols}{Triplet interaction columns}

\item{nonspline_cols}{Predictors not treated as spline effects}

\item{p_expansions}{Number of columns in the basis expansion per partition}

\item{parallel}{Logical to enable parallel processing}

\item{cl}{Cluster object for parallel computation}
}
\value{
A \eqn{p \times p} integrated squared second-derivative penalty
  matrix.
}
\description{
Computes the integrated squared second-derivative penalty matrix with
optional parallel processing.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{.compute_dhat_diag}
\alias{.compute_dhat_diag}
\title{Compute the Derivative of the Tuning Hat Matrix Diagonal}
\usage{
.compute_dhat_diag(X, G, dG_dlambda, A, AGAInv, K, p_expansions)
}
\arguments{
\item{X}{List of partition-specific design matrices.}

\item{G}{List of partition-specific \eqn{\mathbf{G}_k} matrices.}

\item{dG_dlambda}{List of derivatives \eqn{d\mathbf{G}_k / d\lambda_1}.}

\item{A}{Constraint matrix.}

\item{AGAInv}{Inverse of \eqn{\mathbf{A}^{\top}\mathbf{G}\mathbf{A}}.}

\item{K}{Number of interior knots.}

\item{p_expansions}{Number of columns per partition.}
}
\value{
List with \code{dh_diag} and \code{dC_list}, both partition-wise.
}
\description{
Computes \eqn{d h_{ii} / d\lambda_1} partition-wise for the exact
leave-one-out tuning criterion.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{get_interaction_patterns}
\alias{get_interaction_patterns}
\title{Generate Interaction Variable Patterns}
\usage{
get_interaction_patterns(vars)
}
\arguments{
\item{vars}{Character vector of variable names}
}
\value{
Character vector of interaction pattern strings for 2- or 3-variable
inputs, or \code{NULL} otherwise.
}
\description{
Generates all possible interaction patterns for 2 or 3 variables. This is
used in part for identifying which interactions and expansions to exclude
(provided to "exclude_these_expansions" argument of lgspline) based on
formulas provided.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{is_binary}
\alias{is_binary}
\title{Test if Vector is Binary}
\usage{
is_binary(x)
}
\arguments{
\item{x}{Vector to test}
}
\value{
Logical indicating if x has at most 2 unique values
}
\description{
Test if Vector is Binary
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_Lambda.R
\name{.tune_grid_search}
\alias{.tune_grid_search}
\title{Grid Search Initialization for Penalty Tuning}
\usage{
.tune_grid_search(
  log_initial_wiggle,
  log_initial_flat,
  log_penalty_vec,
  criterion_fxn,
  env,
  include_warnings,
  ...
)
}
\arguments{
\item{log_initial_wiggle}{Numeric vector; log-scale candidate values for
the wiggle penalty.}

\item{log_initial_flat}{Numeric vector; log-scale candidate values for
the flat ridge penalty.}

\item{log_penalty_vec}{Numeric vector; log-scale predictor/partition
penalties.}

\item{criterion_fxn}{Function; tuning-objective evaluation function.}

\item{env}{List; tuning environment.}

\item{include_warnings}{Logical; whether to print warnings on failure.}

\item{...}{Additional arguments passed to \code{criterion_fxn}.}
}
\value{
Numeric vector of length 2; the best (log_wiggle, log_flat) found.
}
\description{
Evaluates the selected tuning criterion over a grid of initial wiggle and ridge
penalty values to find a good starting point for BFGS optimization.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_qp.R
\name{.build_deriv_qp}
\alias{.build_deriv_qp}
\title{Build Derivative QP Constraints in Full \eqn{P}-Dimensional Space}
\usage{
.build_deriv_qp(
  X_block,
  sign_mult,
  just_first,
  p_expansions,
  K,
  colnm_expansions,
  power1_cols,
  power2_cols,
  nonspline_cols,
  interaction_single_cols,
  interaction_quad_cols,
  triplet_cols,
  include_2way_interactions,
  include_3way_interactions,
  include_quadratic_interactions,
  expansion_scales,
  target_vars = NULL,
  og_cols = NULL
)
}
\arguments{
\item{X_block}{Numeric matrix, \eqn{N_{\mathrm{sub}} \times P}.
Block-diagonal design matrix for the QP-selected observations.}

\item{sign_mult}{Numeric scalar, \code{+1} for positive constraints
or \code{-1} for negative constraints.}

\item{just_first}{Logical. If \code{TRUE}, constrain first derivatives;
if \code{FALSE}, constrain second derivatives.}

\item{p_expansions}{Integer. Number of basis expansions per partition.}

\item{K}{Integer. Number of interior knots (partitions minus 1).}

\item{colnm_expansions}{Character vector of length \code{p_expansions}.
Column names of the expansion matrix.}

\item{power1_cols}{Integer vector of linear-term column indices.}

\item{power2_cols}{Integer vector of quadratic-term column indices.}

\item{nonspline_cols}{Integer vector of non-spline linear column indices.}

\item{interaction_single_cols}{Integer vector of linear-by-linear
interaction column indices.}

\item{interaction_quad_cols}{Integer vector of linear-by-quadratic
interaction column indices.}

\item{triplet_cols}{Integer vector of three-way interaction column indices.}

\item{include_2way_interactions}{Logical switch forwarded to
\code{make_derivative_matrix}.}

\item{include_3way_interactions}{Logical switch forwarded to
\code{make_derivative_matrix}.}

\item{include_quadratic_interactions}{Logical switch forwarded to
\code{make_derivative_matrix}.}

\item{expansion_scales}{Numeric vector of length \code{p_expansions - 1}.
Passed as unit scales (\code{1 + 0 * expansion_scales}) so that the
constraint matrix stays on the expansion-standardized coefficient scale.}

\item{target_vars}{Optional. Integer vector of predictor column indices
or character vector of predictor names identifying which predictors
to constrain. When \code{NULL} (default), all derivative variables
are constrained.}

\item{og_cols}{Optional character vector of original predictor column
names, used to resolve character \code{target_vars} to integer indices.}
}
\value{
A list with components:
\describe{
  \item{Amat}{\eqn{P \times M} constraint matrix, where \eqn{M} is the
    number of unique constraint columns (after deduplication).}
  \item{bvec}{Numeric vector of length \eqn{M}, all zeros.}
  \item{meq}{Integer, always \code{0} (inequality constraints).}
}
}
\description{
Internal helper that constructs the \code{Amat} / \code{bvec} pair
enforcing derivative sign constraints at every row of a block-diagonal
design matrix.

Given an \eqn{N_{\mathrm{sub}} \times P} block-diagonal design matrix
\code{X_block}, where \eqn{P = p \times (K+1)}, this function:
\enumerate{
  \item Recovers the per-partition \eqn{p}-column expansion matrix
    \code{C_qp} and records each row's partition assignment.
  \item Calls \code{make_derivative_matrix} on \code{C_qp} to obtain
    first or second derivative matrices with respect to each predictor.
  \item Optionally selects only derivatives for a subset of predictor
    variables (\code{target_vars}).
  \item Maps each derivative row into the full \eqn{P}-dimensional
    coefficient space, yielding one constraint column per (observation,
    variable) pair.
}

The result is a constraint pair \eqn{\mathbf{A}^{\top}\boldsymbol{\beta}
\ge \mathbf{b}} (with \eqn{\mathbf{b} = \mathbf{0}}) suitable for
\code{\link[quadprog]{solve.QP}}.
}
\seealso{
\code{\link{process_qp}}, \code{\link[quadprog]{solve.QP}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_partitions.R
\name{make_partitions}
\alias{make_partitions}
\title{Create Data Partitions Using Clustering}
\usage{
make_partitions(
  data,
  cluster_args,
  cluster_on_indicators,
  K,
  parallel,
  cl,
  do_not_cluster_on_these,
  neighbor_tolerance,
  standardize = TRUE,
  standardize_mode = "auto",
  dummy_adder = 0,
  dummy_dividor = 0
)
}
\arguments{
\item{data}{Numeric matrix of predictor variables (\strong{raw scale})}

\item{cluster_args}{Parameters for clustering}

\item{cluster_on_indicators}{Logical to include binary predictors}

\item{K}{Number of partitions minus 1 (\eqn{K})}

\item{parallel}{Logical to enable parallel processing}

\item{cl}{Cluster object for parallel computation}

\item{do_not_cluster_on_these}{Columns to exclude from clustering}

\item{neighbor_tolerance}{Scaling factor for neighbor detection}

\item{standardize}{Logical; whether to standardize data internally before
clustering.  Should equal \code{standardize_predictors_for_knots} from
\code{lgspline.fit}.  Default \code{TRUE}.}

\item{standardize_mode}{Character; \code{"auto"} (default) selects
\code{"minmax"} for a single effective clustering dimension and
\code{"normal"} for multiple dimensions.  Can be forced to either value.}

\item{dummy_adder}{Small constant added to numerator during standardization
to avoid exact-zero values (matches \code{lgspline.fit}'s
\code{dummy_adder}).  Default \code{0}.}

\item{dummy_dividor}{Small constant added to denominator during
standardization to avoid division by zero (matches
\code{lgspline.fit}'s \code{dummy_dividor}).  Default \code{0}.}
}
\value{
A list containing:
\describe{
  \item{centers}{Cluster center coordinates on the \strong{raw} scale.}
  \item{knots}{Knot points between centers on the \strong{raw} scale.}
  \item{assign_partition}{Function that accepts \strong{raw}-scale new data
    and returns integer-like partition assignments (0.5, 1.5, \ldots).}
  \item{neighbors}{List of neighboring partition indices.}
  \item{standardize_transf}{The forward standardization function used
    internally (for diagnostic use only).}
  \item{standardize_inv_transf}{The inverse standardization function
    (for diagnostic use only).}
  \item{centers_std}{Cluster centers on the standardized scale (for
    diagnostic use only).}
}
}
\description{
Partitions data support into clusters using Voronoi-like diagrams.
Standardization for clustering is handled internally; centers, knots, and
the returned \code{assign_partition} function are all on the **raw
(natural) scale** of the original predictors.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_partitions.R
\name{find_neighbors}
\alias{find_neighbors}
\title{Find Neighboring Cluster Partitions Using Midpoint Distance Criterion}
\usage{
find_neighbors(centers, parallel, cl, neighbor_tolerance)
}
\arguments{
\item{centers}{Matrix; rows are cluster center coordinates}

\item{parallel}{Logical; use parallel processing}

\item{cl}{Cluster object for parallel execution}

\item{neighbor_tolerance}{Numeric; scaling factor for distance comparisons}
}
\value{
List where element i contains indices of centers neighboring center i
}
\description{
Identifies neighboring partitions by evaluating whether the midpoint between
cluster centers is closer to those centers than to any other center.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cox_helpers.R
\name{score_cox}
\alias{score_cox}
\title{Compute Cox Partial Log-Likelihood Score Vector}
\usage{
score_cox(X, eta, status, y = NULL, weights = 1)
}
\arguments{
\item{X}{Design matrix (N x p), sorted by ascending event time.}

\item{eta}{Linear predictor vector \eqn{\mathbf{X}\boldsymbol{\beta}}.}

\item{status}{Event indicator (1 = event, 0 = censored).}

\item{y}{Optional numeric vector of observed event/censor times, same length
and order as \code{eta}. When supplied, tied event times are handled with
the Breslow approximation. When omitted, the function assumes there are no
ties.}

\item{weights}{Observation weights (default 1).}
}
\value{
Numeric column vector of length p (gradient).
}
\description{
Gradient of the Cox partial log-likelihood with respect to
\eqn{\boldsymbol{\beta}}. Data must be sorted by ascending event time.
}
\details{
Under the Breslow approximation, the score is
\deqn{\mathbf{u} =
  \sum_g \Bigl[\sum_{i \in D_g} w_i \mathbf{x}_i -
  d_g^{(w)} \frac{\sum_{j \in R_g} w_j e^{\eta_j}\mathbf{x}_j}
  {\sum_{j \in R_g} w_j e^{\eta_j}}\Bigr]}
where \eqn{D_g} is the tied event set at time \eqn{t_g},
\eqn{R_g = \{j : t_j \ge t_g\}} is the risk set, and
\eqn{d_g^{(w)} = \sum_{i \in D_g} w_i}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weibull_helpers.R
\name{weibull_family}
\alias{weibull_family}
\title{Weibull Family for Survival Model Specification}
\usage{
weibull_family()
}
\value{
A family-like list containing the link functions and Weibull-specific
methods used by \code{lgspline}.
}
\description{
Creates a family-like object for Weibull accelerated failure time (AFT)
models, including custom log-likelihood, AIC, and deviance helpers.

This both provides a tool for actually fitting Weibull AFT Models, and
boilerplate code for users who wish to incorporate Lagrangian multiplier
smoothing splines into their own custom models.
}
\details{
Provides a comprehensive family specification for Weibull AFT models,
including family name, link function, inverse link function, custom loss
function for model tuning, and methods for AIC and log-likelihood
computation compatible with \code{logLik.lgspline}.

Supports right-censored survival data with flexible parameter estimation.

Note on scale vs. dispersion: throughout this package, the lgspline object
stores \code{sigmasq_tilde} which equals sigma^2 (dispersion), where
sigma is the Weibull scale parameter matching \code{survreg$scale}.
Functions that accept a \code{dispersion} argument receive sigma^2;
functions that accept a \code{scale} argument receive sigma.
}
\examples{

## Simulate survival data with covariates
set.seed(1234)
n <- 1000
t1 <- rnorm(n)
t2 <- rbinom(n, 1, 0.5)

## Generate survival times with Weibull-like structure
lambda <- exp(0.5 * t1 + 0.3 * t2)
yraw <- rexp(n, rate = 1/lambda)

## Introduce right-censoring
status <- rbinom(n, 1, 0.75)
y <- ifelse(status, yraw, runif(length(yraw), 0, yraw))

## Prepare data
df <- data.frame(y = y, t1 = t1, t2 = t2, status = status)

## Fit model using custom Weibull family
model_fit <- lgspline(y ~ spl(t1) + t2,
                      df,
                      unconstrained_fit_fxn = unconstrained_fit_weibull,
                      family = weibull_family(),
                      need_dispersion_for_estimation = TRUE,
                      dispersion_function = weibull_dispersion_function,
                      glm_weight_function = weibull_glm_weight_function,
                      schur_correction_function = weibull_schur_correction,
                      status = status,
                      opt = FALSE,
                      K = 1)

summary(model_fit)

## Log-likelihood now works via logLik.lgspline:
# logLik(model_fit)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.active_set_refine}
\alias{.active_set_refine}
\title{Partition-Wise Active-Set Refinement for Inequality Constraints}
\usage{
.active_set_refine(
  result,
  X,
  y,
  K,
  p_expansions,
  A,
  R_constraints,
  constraint_value_vectors,
  Lambda,
  Ghalf,
  GhalfInv,
  family,
  qp_Amat,
  qp_bvec,
  qp_meq,
  Xy_or_uncon,
  is_path3,
  parallel_aga,
  parallel_matmult,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  tol,
  max_as_iter = 50
)
}
\arguments{
\item{result}{List of current coefficient column vectors by partition.}

\item{X, y}{Lists of partition-specific design matrices and responses.}

\item{K, p_expansions}{Integer dimensions.}

\item{A}{Original equality constraint matrix.}

\item{R_constraints}{Number of columns of \code{A}.}

\item{constraint_value_vectors}{Constraint RHS list.}

\item{Lambda}{Shared penalty matrix.}

\item{Ghalf, GhalfInv}{Lists of \eqn{\mathbf{G}^{1/2}_k} and
\eqn{\mathbf{G}^{-1/2}_k} matrices.}

\item{family}{GLM family object.}

\item{qp_Amat}{Inequality constraint matrix.}

\item{qp_bvec}{Inequality constraint RHS.}

\item{qp_meq}{Number of equality constraints within \code{qp_Amat}
(leading columns).}

\item{Xy_or_uncon}{Cross-products (Path 2) or unconstrained estimates
(Path 3).}

\item{is_path3}{Logical.}

\item{parallel_aga, parallel_matmult}{Logical flags.}

\item{cl, chunk_size, num_chunks, rem_chunks}{Parallel parameters.}

\item{tol}{Convergence tolerance.}

\item{max_as_iter}{Maximum active-set iterations (default 50).}
}
\value{
A list with components:
\describe{
  \item{result}{List of refined coefficient column vectors by partition.}
  \item{qp_info}{List with active constraint information, or NULL.}
  \item{converged}{Logical; TRUE if active-set method converged.}
}
}
\description{
Replaces the dense \code{.qp_refine} SQP loop with a partition-wise
active-set method that reuses the existing Lagrangian projection
machinery. The key insight is that an active inequality constraint
can be treated as an additional equality constraint and absorbed into
the augmented constraint matrix \eqn{\mathbf{A}_{\mathrm{aug}}},
after which the standard \eqn{\mathbf{G}^{1/2}\mathbf{r}^*} trick
applies without ever forming the full \eqn{P \times P} information
matrix.
}
\details{
The active-set loop:
\enumerate{
  \item Solve the equality-constrained subproblem (Lagrangian projection)
    with current active set.
  \item Check primal feasibility: any violated inactive inequalities?
  \item Check dual feasibility: any negative multipliers on active
    inequalities?
  \item If both satisfied, return (KKT conditions met).
  \item Otherwise, add most-violated constraint or drop most-negative
    multiplier, and repeat.
}

Falls back to \code{.qp_refine} if the active-set method does not
converge within \code{max_as_iter} iterations.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{collapse_block_diagonal}
\alias{collapse_block_diagonal}
\title{Collapse Matrix List into a Single Dense Block-Layout Matrix}
\usage{
collapse_block_diagonal(matlist)
}
\arguments{
\item{matlist}{List of input matrices}
}
\value{
Dense matrix with each input block placed in its own column range
}
\description{
Transforms a list of matrices into a single dense block-layout matrix. This
is useful especially for quadratic programming problems, where operating on
lists of blocks is not convenient.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/negbin_helpers.R
\name{score_negbin}
\alias{score_negbin}
\title{Compute Negative Binomial Score Vector}
\usage{
score_negbin(X, y, mu, theta, weights = 1)
}
\arguments{
\item{X}{Design matrix (N x p).}

\item{y}{Response vector.}

\item{mu}{Mean vector \eqn{\exp(\mathbf{X}\boldsymbol{\beta})}.}

\item{theta}{Shape parameter.}

\item{weights}{Observation weights (default 1).}
}
\value{
Numeric column vector of length p (gradient).
}
\description{
Gradient of the NB2 log-likelihood with respect to
\eqn{\boldsymbol{\beta}} under the log link.
}
\details{
Under log link, \eqn{d\mu_i/d\eta_i = \mu_i}, so the score is
\deqn{\mathbf{u} = \mathbf{X}^{\top}\mathbf{w}_{\mathrm{obs}}
  \odot \frac{(y_i - \mu_i)\theta}{\theta + \mu_i}}
where the per-observation contribution
\eqn{(y_i - \mu_i)\theta/(\theta + \mu_i)} is the derivative of the
log-likelihood with respect to \eqn{\eta_i}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_Lambda.R
\name{.compute_tuning_delta}
\alias{.compute_tuning_delta}
\title{Compute Pseudocount Delta for Link Function Stabilization}
\usage{
.compute_tuning_delta(family, unl_y, N_obs, observation_weights, opt)
}
\arguments{
\item{family}{GLM family object.}

\item{unl_y}{Numeric vector; unlisted response values (concatenated across
partitions).}

\item{N_obs}{Integer; total sample size.}

\item{observation_weights}{List or NULL; observation weights by partition.}

\item{opt}{Logical; whether optimization is being performed.}
}
\value{
Numeric scalar; the pseudocount \eqn{\delta \geq 0}.
}
\description{
Determines the pseudocount \eqn{\delta} used to stabilize link function
transformations during penalty tuning. For identity link or when
the response is naturally in the domain of the link function, returns 0.
Otherwise, finds the \eqn{\delta} that makes the transformed response
distribution most closely approximate a t-distribution.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.summary.lgspline}
\alias{print.summary.lgspline}
\title{Print Method for lgspline Summaries}
\usage{
\method{print}{summary.lgspline}(x, ...)
}
\arguments{
\item{x}{A \code{summary.lgspline} object.}

\item{...}{Not used.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Displays a formatted model summary using \code{\link[stats]{printCoefmat}}
for the coefficient table.
}
\seealso{
\code{\link[stats]{printCoefmat}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weibull_helpers.R
\name{lgspline_weibull}
\alias{lgspline_weibull}
\title{Fit Weibull Accelerated Failure Time Model via lgspline}
\usage{
lgspline_weibull(formula, data, status, ...)
}
\arguments{
\item{formula}{Formula specifying the model. The response should be
survival time; \code{status} is passed separately.}

\item{data}{Data frame.}

\item{status}{Integer vector of event indicators (1 = event,
0 = censored), same length as the number of rows in \code{data}.}

\item{...}{Additional arguments passed to \code{lgspline}.}
}
\value{
An object of class \code{"lgspline"}.
}
\description{
Convenience wrapper that calls \code{lgspline} with the correct
family, weight, dispersion, score, and unconstrained-fit functions for
Weibull accelerated failure time regression. All standard lgspline
arguments (knots, penalties, constraints, parallel, etc.) are passed
through.
}
\details{
Internally sets:
\itemize{
  \item \code{family = weibull_family()}
  \item \code{unconstrained_fit_fxn = unconstrained_fit_weibull}
  \item \code{glm_weight_function = weibull_glm_weight_function}
  \item \code{qp_score_function = weibull_qp_score_function}
  \item \code{dispersion_function = weibull_dispersion_function}
  \item \code{schur_correction_function = weibull_schur_correction}
  \item \code{need_dispersion_for_estimation = TRUE}
  \item \code{estimate_dispersion = TRUE}
  \item \code{standardize_response = FALSE}
}

A formula interface is needed, e.g. \code{lgspline_weibull(t, y, ...)} won't work,
unlike for ordinary \code{\link{lgspline}}.
}
\examples{

## Weibull AFT with a nonlinear age effect on lung cancer survival
if(requireNamespace("survival", quietly = TRUE)) {
  library(survival)
  set.seed(1234)
  lung <- na.omit(lung[, c("time", "status", "age")])
  lung$age_std <- std(lung$age)

  ## survival codes status as 1 = censored, 2 = dead
  event <- as.integer(lung$status == 2)

  ## Spline on age
  fit <- lgspline_weibull(
    time ~ spl(age_std),
    data = lung,
    status = event,
    K = 1,
    opt = FALSE,
    wiggle_penalty = 1e-4,
    flat_ridge_penalty = 1
  )
  print(summary(fit))
  plot(fit,
       show_formulas = TRUE,
       custom_response_lab = 'Survival Time',
       custom_predictor_lab = 'Standardized Age')
}

}
\seealso{
\code{\link{lgspline_cox}} for Cox proportional hazards,
  \code{\link{weibull_family}}, \code{\link{unconstrained_fit_weibull}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.get_B_glm_nocorr}
\alias{.get_B_glm_nocorr}
\title{Path 3: Non-Gaussian GLM, No Correlation}
\usage{
.get_B_glm_nocorr(
  X,
  y,
  X_gram,
  Xy,
  Lambda,
  Ghalf,
  GhalfInv,
  A,
  K,
  p_expansions,
  R_constraints,
  constraint_value_vectors,
  family,
  return_G_getB,
  iterate,
  tol,
  quadprog,
  qp_Amat,
  qp_bvec,
  qp_meq,
  qp_score_function,
  unconstrained_fit_fxn,
  keep_weighted_Lambda,
  unique_penalty_per_partition,
  L_partition_list,
  parallel_eigen,
  parallel_aga,
  parallel_matmult,
  parallel_unconstrained,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  order_list,
  observation_weights,
  glm_weight_function,
  schur_correction_function,
  need_dispersion_for_estimation,
  dispersion_function,
  VhalfInv,
  homogenous_weights,
  ...
)
}
\arguments{
\item{X, y}{Lists of partition-specific design matrices and responses.}

\item{X_gram}{List of Gram matrices.}

\item{Xy}{List of cross-products.}

\item{Lambda}{Shared penalty matrix.}

\item{Ghalf, GhalfInv}{Lists of matrix square roots by partition.}

\item{A}{Constraint matrix.}

\item{K, p_expansions, R_constraints}{Integer dimensions.}

\item{constraint_value_vectors}{Constraint RHS list encoding
\eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{c}}.}

\item{family}{GLM family object.}

\item{return_G_getB}{Logical; return covariance components.}

\item{iterate}{Logical; iterate for non-canonical links.}

\item{tol}{Convergence tolerance.}

\item{quadprog}{Logical; apply QP refinement.}

\item{qp_Amat, qp_bvec, qp_meq}{QP constraint specification.}

\item{qp_score_function}{Score function for QP step.}

\item{unconstrained_fit_fxn}{Function for partition-wise unconstrained
estimation.}

\item{keep_weighted_Lambda, unique_penalty_per_partition}{Logical flags.}

\item{L_partition_list}{Partition-specific penalty matrices.}

\item{parallel_eigen, parallel_aga, parallel_matmult, }{parallel_unconstrained Logical flags.}

\item{cl, chunk_size, num_chunks, rem_chunks}{Parallel parameters.}

\item{order_list, observation_weights}{Standard partition arguments.}

\item{glm_weight_function, schur_correction_function, }{need_dispersion_for_estimation,dispersion_function GLM customization.}

\item{VhalfInv}{Inverse square root correlation (unused here).}

\item{homogenous_weights}{Logical.}

\item{...}{Passed to fitting, weight, and score functions.}
}
\value{
Same structure as \code{get_B}.
}
\description{
For GLMs with non-identity links or non-Gaussian families (without
correlation structures), unconstrained partition-wise estimates are
first obtained via Newton-Raphson (or OLS for Gaussian identity), then
the constrained estimate is computed by a single Lagrangian projection.
For non-canonical links, \eqn{\mathbf{G}} depends on the current
fitted values through the GLM working weights \eqn{\mathbf{W}}; the
projection is therefore iterated, updating \eqn{\mathbf{G}} at the
current constrained estimate until convergence.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/negbin_helpers.R
\name{negbin_glm_weight_function}
\alias{negbin_glm_weight_function}
\title{NB GLM Weight Function}
\usage{
negbin_glm_weight_function(
  mu,
  y,
  order_indices,
  family,
  dispersion,
  observation_weights
)
}
\arguments{
\item{mu}{Predicted values \eqn{\exp(\eta)}.}

\item{y}{Observed counts.}

\item{order_indices}{Observation indices in partition order.}

\item{family}{NB family object (unused, for interface compatibility).}

\item{dispersion}{Shape parameter \eqn{\theta}.}

\item{observation_weights}{Observation weights.}
}
\value{
Numeric vector of working weights, length N.
}
\description{
Computes working weights for the NB2 information matrix used by
lgspline when updating \eqn{\mathbf{G}} after obtaining constrained
estimates.
}
\details{
The IRLS weight for NB2 with log link is
\deqn{W_i = w_i \mu_i \theta / (\theta + \mu_i)}

Falls back to observation weights when natural weights are degenerate.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.lagrangian_project}
\alias{.lagrangian_project}
\title{Lagrangian Projection via OLS Reformulation}
\usage{
.lagrangian_project(
  GhalfXy,
  Ghalf,
  A,
  K,
  p_expansions,
  R_constraints,
  constraint_value_vectors,
  family,
  parallel_aga,
  parallel_matmult,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks
)
}
\arguments{
\item{GhalfXy}{Numeric column vector \eqn{\mathbf{y}^*} of length
\eqn{P}; produced by multiplying \eqn{\mathbf{G}^{1/2}} (Path 2) or
\eqn{\mathbf{G}^{-1/2}} (Path 3) into the unconstrained estimate.}

\item{Ghalf}{List of \eqn{\mathbf{G}^{1/2}_k} matrices (one per
partition).}

\item{A}{Constraint matrix \eqn{\mathbf{A}} (\eqn{P \times R}).}

\item{K}{Integer; number of interior knots.}

\item{p_expansions}{Integer; number of basis terms per partition.}

\item{R_constraints}{Integer; number of columns of \eqn{\mathbf{A}}.}

\item{constraint_value_vectors}{List of constraint right-hand-side
vectors encoding \eqn{\mathbf{A}^{\top}\boldsymbol{\beta} =
\mathbf{c}}. When non-empty and nonzero, the particular solution is
added.}

\item{family}{GLM family object (used for \code{linkinv(0)} fallback
when \code{NA} values arise in \eqn{\mathbf{y}^*}).}

\item{parallel_aga, parallel_matmult}{Logical flags for parallel
computation.}

\item{cl}{Parallel cluster object.}

\item{chunk_size, num_chunks, rem_chunks}{Parallel distribution parameters.}
}
\value{
A list of length \eqn{K+1}, each element a column vector of
  constrained coefficients \eqn{\tilde{\boldsymbol{\beta}}_k}.
}
\description{
Given unconstrained (or GEE-initialized) coefficient estimates
\eqn{\hat{\boldsymbol{\beta}}}, computes the constrained estimate
\eqn{\tilde{\boldsymbol{\beta}} = \mathbf{U}\hat{\boldsymbol{\beta}}}
where \eqn{\mathbf{U} = \mathbf{I} - \mathbf{G}\mathbf{A}
(\mathbf{A}^{\top}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{\top}}.
}
\details{
Rather than forming \eqn{\mathbf{U}} directly, the projection is
reformulated as a residual from an OLS problem (the
\eqn{\mathbf{G}^{1/2}\mathbf{r}^*} trick):
\deqn{\mathbf{y}^* = \mathbf{G}^{-1/2}\hat{\boldsymbol{\beta}}}
\deqn{\mathbf{X}^* = \mathbf{G}^{1/2}\mathbf{A}}
\deqn{\mathbf{r}^* = (\mathbf{I} - \mathbf{X}^*(\mathbf{X}^{*\top}
  \mathbf{X}^*)^{-1}\mathbf{X}^{*\top})\mathbf{y}^*}
\deqn{\tilde{\boldsymbol{\beta}} = \mathbf{G}^{1/2}\mathbf{r}^*}

This avoids explicitly forming and inverting
\eqn{\mathbf{A}^{\top}\mathbf{G}\mathbf{A}}; the most expensive step is
the QR decomposition of the \eqn{R \times R} system inside
\code{.lm.fit}, which is far cheaper than the full \eqn{P \times P}
solve.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blockfit_solve.R
\name{.bf_split_components}
\alias{.bf_split_components}
\title{Split Design and Penalty into Spline and Flat Components}
\usage{
.bf_split_components(
  X,
  flat_cols,
  p_expansions,
  K,
  Lambda,
  L_partition_list,
  A,
  constraint_values
)
}
\arguments{
\item{X}{List of \eqn{K+1} design matrices (\eqn{N_k \times p_expansions}).}

\item{flat_cols}{Integer vector of flat column indices.}

\item{p_expansions}{Integer; total columns per partition.}

\item{K}{Integer; number of interior knots.}

\item{Lambda}{\eqn{p_expansions \times p_expansions} shared penalty matrix.}

\item{L_partition_list}{List of partition-specific penalty matrices.}

\item{A}{Full \eqn{P \times R_constraints} constraint matrix.}

\item{constraint_values}{List of constraint right-hand sides.}
}
\value{
A named list with elements:
\describe{
  \item{X_spline}{List of \eqn{K+1} matrices, spline columns only.}
  \item{X_flat}{List of \eqn{K+1} matrices, flat columns only.}
  \item{Lambda_spline}{\eqn{nc_s \times nc_s} penalty submatrix.}
  \item{Lambda_flat}{\eqn{nc_f \times nc_f} penalty submatrix.}
  \item{L_part_spline}{List of partition-specific penalty submatrices
    for the spline columns.}
  \item{A_spline}{Spline-only constraint matrix (columns pruned
    and rank-reduced via pivoted QR).}
  \item{nca_spline}{Integer; number of columns in
    \code{A_spline}.}
  \item{constraint_values_spline}{List of constraint RHS vectors
    restricted to spline rows.}
  \item{spline_cols}{Integer vector of spline column indices.}
  \item{nc_spline}{Integer; number of spline columns.}
  \item{nc_flat}{Integer; number of flat columns.}
  \item{A_flat}{Flat-only constraint matrix (rows for flat coefficients,
    columns pruned and rank-reduced). Used to enforce mixed constraints
    on the flat update step.}
  \item{nca_flat}{Integer; number of columns in \code{A_flat}.}
  \item{A_full}{The original full constraint matrix \code{A}, retained
    for mixed-constraint enforcement.}
  \item{flat_rows_all}{Integer vector of flat-coefficient row indices
    in the full P-space.}
  \item{spline_rows_all}{Integer vector of spline-coefficient row indices
    in the full P-space.}
  \item{has_mixed_constraints}{Logical; TRUE if any constraint column
    in A has nonzero entries on both spline and flat rows.}
  \item{mixed_constraint_cols}{Integer vector of column indices in the
    original A that are mixed (touch both spline and flat rows).}
}
}
\description{
Partitions the per-partition design matrices, penalty matrices, and
constraint matrix into "spline" (columns receiving partition-specific
coefficients) and "flat" (columns receiving a single shared
coefficient across all partitions) subsets.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{get_B_verification_examples}
\alias{get_B_verification_examples}
\title{Verification Examples for get_B}
\description{
Simple, self-contained examples that reviewers can run to verify that
\code{get_B} produces correct output. These exercise Path 2 (Gaussian,
no correlation) and Path 3 (binomial GLM).
}
\examples{
\dontrun{
## Example 1: Path 2 - Gaussian identity, with knots
set.seed(42)
t <- runif(200, -5, 5)
y <- sin(t) + rnorm(200, 0, 0.5)
fit1 <- lgspline(t, y, K = 3, opt = FALSE, wiggle_penalty = 1e-4)
stopifnot(inherits(fit1, "lgspline"))
stopifnot(length(fit1$B) == 4)  # K+1 = 4 partitions
cat("Example 1 passed: Gaussian identity, K=3\n")

preds1 <- predict(fit1, new_predictors = rnorm(10))
stopifnot(all(is.finite(preds1)))
cat("  Predictions finite: OK\n")

## Example 2: Path 2 - Gaussian identity, K=0 (no constraints)
fit2 <- lgspline(t, y, K = 0, opt = FALSE, wiggle_penalty = 1e-4)
stopifnot(inherits(fit2, "lgspline"))
stopifnot(length(fit2$B) == 1)
preds2 <- predict(fit2, new_predictors = rnorm(10))
stopifnot(all(is.finite(preds2)))
cat("Example 2 passed: Gaussian identity, K=0\n")

## Example 3: Path 3 - Binomial GLM
y_bin <- rbinom(200, 1, plogis(sin(t)))
fit3 <- lgspline(t, y_bin, K = 2, family = binomial(),
                 opt = FALSE, wiggle_penalty = 1e-3)
stopifnot(inherits(fit3, "lgspline"))
preds3 <- predict(fit3, new_predictors = rnorm(10))
stopifnot(all(preds3 >= 0 & preds3 <= 1))
cat("Example 3 passed: Binomial GLM, K=2\n")

## Example 4: Path 2 with QP constraints (monotonic increase)
t_sorted <- sort(runif(100, -3, 3))
y_mono <- t_sorted + 0.5 * sin(t_sorted) + rnorm(100, 0, 0.3)
fit4 <- lgspline(t_sorted, y_mono, K = 2,
                 qp_monotonic_increase = TRUE,
                 opt = FALSE, wiggle_penalty = 1e-4)
preds4 <- predict(fit4, new_predictors = cbind(sort(rnorm(50))))
stopifnot(all(diff(preds4) >= -sqrt(.Machine$double.eps)))
cat("Example 4 passed: Monotonic increase QP constraint\n")

## Example 5: Path 2 with range constraints
fit5 <- lgspline(t, y, K = 3,
                 qp_range_lower = -2, qp_range_upper = 2,
                 opt = FALSE, wiggle_penalty = 1e-4)
stopifnot(all(fit5$ytilde >= -2 - 0.01))
stopifnot(all(fit5$ytilde <= 2 + 0.01))
cat("Example 5 passed: Range constraints\n")

## Example 6: Multi-predictor
set.seed(123)
x1 <- rnorm(300)
x2 <- rnorm(300)
y6 <- sin(x1) + cos(x2) + rnorm(300, 0, 0.5)
fit6 <- lgspline(cbind(x1, x2), y6, K = 4,
                 opt = FALSE, wiggle_penalty = 1e-5)
stopifnot(inherits(fit6, "lgspline"))
stopifnot(fit6$q_predictors == 2)
cat("Example 6 passed: 2D predictor, K=4\n")

## Example 7: Coefficient consistency (determinism)
set.seed(999)
t7 <- runif(150, -4, 4)
y7 <- 2 * cos(t7) + rnorm(150, 0, 0.4)
fit7a <- lgspline(t7, y7, K = 2, opt = FALSE, wiggle_penalty = 1e-5)
set.seed(999)
fit7b <- lgspline(t7, y7, K = 2, opt = FALSE, wiggle_penalty = 1e-5)
max_diff <- max(abs(unlist(fit7a$B) - unlist(fit7b$B)))
stopifnot(max_diff < 1e-12)
cat("Example 7 passed: Deterministic coefficient reproduction\n")

cat("\nAll verification examples passed.\n")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weibull_helpers.R
\name{unconstrained_fit_weibull}
\alias{unconstrained_fit_weibull}
\title{Unconstrained Weibull Accelerated Failure Time Model Estimation}
\usage{
unconstrained_fit_weibull(
  X,
  y,
  LambdaHalf,
  Lambda,
  keep_weighted_Lambda,
  family,
  tol = 1e-08,
  K,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  order_indices,
  weights,
  status
)
}
\arguments{
\item{X}{Design matrix of predictors}

\item{y}{Survival/response times}

\item{LambdaHalf}{Square root of penalty matrix
(\eqn{\boldsymbol{\Lambda}^{1/2}})}

\item{Lambda}{Penalty matrix (\eqn{\boldsymbol{\Lambda}})}

\item{keep_weighted_Lambda}{Flag to retain weighted penalties}

\item{family}{Distribution family specification}

\item{tol}{Convergence tolerance (default 1e-8)}

\item{K}{Number of partitions minus one (\eqn{K})}

\item{parallel}{Flag for parallel processing}

\item{cl}{Cluster object for parallel computation}

\item{chunk_size}{Processing chunk size}

\item{num_chunks}{Number of computational chunks}

\item{rem_chunks}{Remaining chunks}

\item{order_indices}{Observation ordering indices}

\item{weights}{Optional observation weights}

\item{status}{Censoring status indicator (1 = event, 0 = censored)
Indicates whether an event of interest occurred (1) or the observation was
right-censored (0). In survival analysis, right-censoring occurs when the
full survival time is unknown, typically because the study ended or the
subject was lost to follow-up before the event of interest occurred.}
}
\value{
Numeric column vector of unconstrained coefficient estimates for the Weibull
AFT model.
}
\description{
Estimates parameters for an unconstrained Weibull accelerated failure time
(AFT) model supporting right-censored survival data.

This both provides a tool for actually fitting Weibull AFT Models, and
boilerplate code for users who wish to incorporate Lagrangian multiplier
smoothing splines into their own custom models.
}
\details{
Estimation Approach:
The function employs a two-stage optimization strategy for fitting
accelerated failure time models via maximum likelihood:

1. Outer Loop: Estimate Scale Parameter (sigma) using Brent's method

2. Inner Loop: Estimate Regression Coefficients (given sigma) using
   damped Newton-Raphson.

Note: the score and information inside the Newton-Raphson are both scaled
by \eqn{\sigma^2} (i.e., both omit the \eqn{1/\sigma} and
\eqn{1/\sigma^2} prefactors respectively). Since the Newton-Raphson step is
\eqn{\mathbf{G}u = (\mathbf{X}^{\top}\mathbf{W}\mathbf{X})^{-1}
\mathbf{X}^{\top}\mathbf{v}}, the \eqn{\sigma^2} factors cancel and the
step remains correct.
}
\examples{

## Simulate survival data with covariates
set.seed(1234)
n <- 1000
t1 <- rnorm(n)
t2 <- rbinom(n, 1, 0.5)

## Generate survival times with Weibull-like structure
lambda <- exp(0.5 * t1 + 0.3 * t2)
yraw <- rexp(n, rate = 1/lambda)

## Introduce right-censoring
status <- rbinom(n, 1, 0.75)
y <- ifelse(status, yraw, runif(length(yraw), 0, yraw))
df <- data.frame(y = y, t1 = t1, t2 = t2)

## Fit model using lgspline with Weibull AFT unconstrained estimation
model_fit <- lgspline(y ~ spl(t1) + t2,
                      df,
                      unconstrained_fit_fxn = unconstrained_fit_weibull,
                      family = weibull_family(),
                      need_dispersion_for_estimation = TRUE,
                      dispersion_function = weibull_dispersion_function,
                      glm_weight_function = weibull_glm_weight_function,
                      schur_correction_function = weibull_schur_correction,
                      status = status,
                      opt = FALSE,
                      K = 1)

## Print model summary
summary(model_fit)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{safe_replace_var}
\alias{safe_replace_var}
\title{Safely Replace Variable Names in Printed Terms}
\usage{
safe_replace_var(x, old, new)
}
\arguments{
\item{x}{Character vector to modify.}

\item{old}{Character scalar, original variable name.}

\item{new}{Character scalar, replacement variable name.}
}
\value{
Character vector with safe term-wise replacements.
}
\description{
Replaces a variable name only when it appears as a standalone term,
avoiding accidental replacements inside other words.
}
\details{
Useful for plot formula labels, for example replacing "t" with "that"
without changing "intercept" into "inthatercept".
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/solver_utils.R
\name{.solver_recompute_G_at_estimate}
\alias{.solver_recompute_G_at_estimate}
\title{Recompute G, Ghalf, and GhalfInv at a Supplied Coefficient Estimate}
\usage{
.solver_recompute_G_at_estimate(
  X,
  y,
  result,
  K,
  Lambda,
  family,
  order_list,
  glm_weight_function,
  schur_correction_function,
  need_dispersion_for_estimation,
  dispersion_function,
  observation_weights,
  VhalfInv,
  parallel_eigen,
  parallel_matmult,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  unique_penalty_per_partition,
  L_partition_list,
  ...
)
}
\arguments{
\item{X}{List of partition-specific design matrices
\eqn{\mathbf{X}_k}.}

\item{y}{List of response vectors \eqn{\mathbf{y}_k} by partition.}

\item{result}{List of current coefficient column vectors
\eqn{\tilde{\boldsymbol{\beta}}_k} by partition.}

\item{K}{Integer; number of interior knots.}

\item{Lambda}{Shared \eqn{p \times p} penalty matrix.}

\item{family}{GLM family object.}

\item{order_list}{List of index vectors mapping partition rows to
original data ordering.}

\item{glm_weight_function}{Function computing GLM working weights
\eqn{\mathbf{W}_k}.}

\item{schur_correction_function}{Function computing Schur corrections
to the information matrix.}

\item{need_dispersion_for_estimation}{Logical; if \code{TRUE},
estimate dispersion before computing weights.}

\item{dispersion_function}{Dispersion estimation function.}

\item{observation_weights}{List of observation weights
\eqn{\mathbf{D}_k} by partition.}

\item{VhalfInv}{Inverse square root correlation matrix, or
\code{NULL}.}

\item{parallel_eigen, parallel_matmult}{Logical flags for parallel
computation.}

\item{cl}{Parallel cluster object.}

\item{chunk_size, num_chunks, rem_chunks}{Parallel distribution
parameters.}

\item{unique_penalty_per_partition}{Logical.}

\item{L_partition_list}{List of partition-specific penalty matrices.}

\item{...}{Passed to weight, correction, and dispersion functions.}
}
\value{
A list with components \code{G}, \code{Ghalf}, and
  \code{GhalfInv}, each a list of \eqn{K+1} matrices.
  \code{G[[k]]} is computed as \code{tcrossprod(Ghalf[[k]])} to
  guarantee exact symmetry.
}
\description{
Given current constrained coefficient estimates \code{result},
recomputes the penalized information matrix
\eqn{\mathbf{G}_k = (\mathbf{X}_k^{\top}\mathbf{W}_k
\mathbf{D}_k\mathbf{X}_k + \boldsymbol{\Lambda}_k)^{-1}} and its
matrix square roots for each partition.
}
\details{
This is used after Newton-Raphson convergence in Path 3 of
\code{get_B()} and at the final return in \code{blockfit_solve()}
when \code{return_G_getB = TRUE}.  The implementation matches
\code{.recompute_G_at_estimate()} exactly; it exists here so both
solvers can call the same numerical core.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cox_helpers.R
\name{loglik_cox}
\alias{loglik_cox}
\title{Compute Cox Partial Log-Likelihood}
\usage{
loglik_cox(eta, status, y = NULL, weights = 1)
}
\arguments{
\item{eta}{Numeric vector of linear predictors \eqn{\mathbf{X}\boldsymbol{\beta}},
length N, sorted by ascending event time.}

\item{status}{Integer/logical vector of event indicators (1 = event,
0 = censored), same length and order as \code{eta}.}

\item{y}{Optional numeric vector of observed event/censor times, same length
and order as \code{eta}. When supplied, tied event times are handled by the
Breslow approximation using a common risk-set denominator within each tied
event-time block. When omitted, the function assumes there are no ties (or
that ties have already been expanded appropriately).}

\item{weights}{Optional numeric vector of observation weights (default 1).}
}
\value{
Scalar partial log-likelihood value.
}
\description{
Evaluates the Cox partial log-likelihood for a given coefficient vector,
using the Breslow approximation for tied event times.

Observations must be sorted in ascending order of survival time before
calling this function. The internal helpers handle sorting automatically;
this function is exposed for diagnostics and testing.
}
\details{
The partial log-likelihood (Breslow) is
\deqn{\ell(\boldsymbol{\beta}) =
  \sum_{g} \Bigl[\sum_{i \in D_g} w_i \eta_i -
  d_g^{(w)} \log\Bigl(\sum_{j \in R_g} w_j \exp(\eta_j)\Bigr)\Bigr]}
where \eqn{D_g} is the event set at tied event time \eqn{t_g},
\eqn{R_g = \{j : t_j \ge t_g\}} is the corresponding risk set, and
\eqn{d_g^{(w)} = \sum_{i \in D_g} w_i}.
}
\examples{
set.seed(1234)
eta <- rnorm(50)
status <- rbinom(50, 1, 0.6)
y <- rexp(50)
loglik_cox(eta, status, y)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_Lambda.R
\name{.compute_gcvu_gradient}
\alias{.compute_gcvu_gradient}
\title{Compute Closed-Form Gradient of GCV_u Criterion}
\usage{
.compute_gcvu_gradient(par, log_penalty_vec, outlist = NULL, env, ...)
}
\arguments{
\item{par}{Numeric vector; log-scale penalty parameters.}

\item{log_penalty_vec}{Numeric vector; log-scale predictor/partition
penalties.}

\item{outlist}{List or NULL; pre-computed GCV_u components from
\code{.compute_gcvu}. If NULL, they are computed internally.}

\item{env}{List; pre-computed objects and tuning configuration (same
structure as in \code{.compute_gcvu}).}

\item{...}{Additional arguments passed to fitting functions.}
}
\value{
List containing:
\describe{
  \item{GCV_u}{Numeric; GCV_u criterion value including meta-penalty.}
  \item{gradient}{Numeric vector; gradient on the log penalty scale.}
  \item{outlist}{List; GCV_u components (for reuse to avoid recomputation).}
}
}
\description{
Computes the gradient of the modified GCV_u criterion with respect to the
log-scale penalty parameters using analytical derivatives of the hat matrix
trace and residual sum of squares.
}
\details{
The gradient is computed via:
\deqn{\frac{\partial \mathrm{GCV}_u}{\partial \theta}
  = \frac{1}{D^{2}} \left(
    \frac{\partial N}{\partial \theta} D
    - N \frac{\partial D}{\partial \theta}
  \right)}
where \eqn{N = \sum r_{i}^{2}} (numerator),
\eqn{D = n(1 - \gamma\bar{W})^{2}} (denominator), \eqn{\theta} is the
log-scale penalty parameter, and the chain rule
\eqn{d\lambda / d\theta = \lambda} (exp parameterization) is applied.

The shared wiggle and flat directions are differentiated directly from the
fitted criterion. For predictor- and partition-specific penalties, a
lower-cost ratio heuristic is used on the log-penalty scale:
\deqn{\frac{\partial \mathrm{GCV}_u}{\partial \theta_{j}}
  \approx \frac{\mathrm{tr}(\mathbf{L}_{j})}{\mathrm{tr}(\boldsymbol{\Lambda})}
  \cdot \frac{\partial \mathrm{GCV}_u}{\partial \theta_{w}},}
where \eqn{\mathbf{L}_{j}} denotes the current contribution of that extra
penalty to \eqn{\boldsymbol{\Lambda}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weibull_helpers.R
\name{loglik_weibull}
\alias{loglik_weibull}
\title{Compute Log-Likelihood for Weibull Accelerated Failure Time Model}
\usage{
loglik_weibull(log_y, log_mu, status, scale, weights = 1)
}
\arguments{
\item{log_y}{Numeric vector of logarithmic response/survival times}

\item{log_mu}{Numeric vector of logarithmic predicted survival times}

\item{status}{Numeric vector of censoring indicators
(1 = event, 0 = censored)
Indicates whether an event of interest occurred (1) or the observation was
right-censored (0). In survival analysis, right-censoring occurs when the
full survival time is unknown, typically because the study ended or the
subject was lost to follow-up before the event of interest occurred.}

\item{scale}{Numeric scalar representing the Weibull scale parameter
(sigma), equivalent to \code{survreg$scale}. This is the square root of
the dispersion stored in \code{lgspline$sigmasq_tilde}.}

\item{weights}{Optional numeric vector of observation weights (default = 1)}
}
\value{
A numeric scalar representing the weighted total log-likelihood of the model
}
\description{
Calculates the log-likelihood for a Weibull accelerated failure time (AFT)
survival model, supporting right-censored survival data.
}
\details{
The function computes log-likelihood contributions for a Weibull AFT model,
explicitly accounting for right-censored observations. It supports optional
observation weighting to accommodate complex sampling designs.

This both provides a tool for actually fitting Weibull AFT models, and
boilerplate code for users who wish to incorporate Lagrangian multiplier
smoothing splines into their own custom models.
}
\examples{

## Minimal example of fitting a Weibull Accelerated Failure Time model
# Simulating survival data with right-censoring
set.seed(1234)
x1 <- rnorm(1000)
x2 <- rbinom(1000, 1, 0.5)
yraw <- rexp(exp(0.01*x1 + 0.01*x2))
# status: 1 = event occurred, 0 = right-censored
status <- rbinom(1000, 1, 0.25)
yobs <- ifelse(status, runif(length(yraw), 0, yraw), yraw)
df <- data.frame(
  y = yobs,
  x1 = x1,
  x2 = x2
)

## Fit model using lgspline with Weibull AFT specifics
model_fit <- lgspline(y ~ spl(x1) + x2,
                      df,
                      unconstrained_fit_fxn = unconstrained_fit_weibull,
                      family = weibull_family(),
                      need_dispersion_for_estimation = TRUE,
                      dispersion_function = weibull_dispersion_function,
                      glm_weight_function = weibull_glm_weight_function,
                      schur_correction_function = weibull_schur_correction,
                      status = status,
                      opt = FALSE,
                      K = 1)

loglik_weibull(log(model_fit$y), log(model_fit$ytilde), status,
  sqrt(model_fit$sigmasq_tilde))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blockfit_solve.R
\name{.bf_get_mu_eta}
\alias{.bf_get_mu_eta}
\title{Numerical Derivative of the Inverse Link Function}
\usage{
.bf_get_mu_eta(eta, fam)
}
\arguments{
\item{eta}{Numeric vector of linear predictor values.}

\item{fam}{GLM family object.}
}
\value{
Numeric vector of the same length as \code{eta}.
}
\description{
Returns \eqn{\partial\mu / \partial\eta} for the supplied family
object, using the family's own \code{mu.eta} method when available
and falling back to analytic forms for common links or central
differences otherwise.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blockfit_solve.R
\name{.bf_case_glm_gee}
\alias{.bf_case_glm_gee}
\title{GLM + GEE Two-Stage Estimation (Case c)}
\usage{
.bf_case_glm_gee(
  X,
  y,
  K,
  p_expansions,
  flat_cols,
  split,
  family,
  order_list,
  glm_weight_function,
  schur_correction_function,
  qp_score_function,
  need_dispersion_for_estimation,
  dispersion_function,
  observation_weights,
  iterate,
  tol,
  max_backfit_iter,
  parallel_eigen,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  schur_zero,
  quadprog,
  qp_Amat,
  qp_bvec,
  qp_meq,
  Lambda,
  L_partition_list,
  unique_penalty_per_partition,
  A,
  constraint_values,
  Vhalf,
  VhalfInv,
  verbose,
  ...
)
}
\arguments{
\item{X}{List of \eqn{K+1} design matrices
(\eqn{N_k \times p_expansions}). Unwhitened even when GEE.}

\item{y}{List of \eqn{K+1} response vectors. Unwhitened even when GEE.}

\item{K}{Integer; number of interior knots.}

\item{p_expansions}{Integer; number of coefficients per partition.}

\item{flat_cols}{Integer vector indicating flat columns of
\eqn{\mathbf{X}^{(k)}}.}

\item{split}{Output of \code{.bf_split_components}.}

\item{family}{GLM family object.}

\item{order_list}{List of observation index vectors by partition.}

\item{glm_weight_function}{GLM weight function.}

\item{schur_correction_function}{Schur complement correction function.}

\item{qp_score_function}{Score function for QP subproblem.}

\item{need_dispersion_for_estimation}{Logical.}

\item{dispersion_function}{Dispersion estimation function.}

\item{observation_weights}{List of observation weights.}

\item{iterate}{Logical; if FALSE, single pass (no iteration).}

\item{tol}{Convergence tolerance.}

\item{max_backfit_iter}{Integer.}

\item{parallel_eigen, cl, chunk_size, num_chunks, rem_chunks}{Parallel arguments.}

\item{schur_zero}{List of zeros (one per partition).}

\item{quadprog}{Logical; apply inequality constraint refinement if TRUE.}

\item{qp_Amat}{Inequality constraint matrix for
\code{quadprog::solve.QP}.}

\item{qp_bvec}{Inequality constraint right-hand side.}

\item{qp_meq}{Number of leading equality constraints.}

\item{Lambda}{\eqn{p_expansions \times p_expansions} penalty matrix.}

\item{L_partition_list}{List of partition-specific penalty matrices.}

\item{unique_penalty_per_partition}{Logical.}

\item{A}{Full \eqn{P \times R_constraints} constraint matrix.}

\item{constraint_values}{List of constraint right-hand sides.}

\item{Vhalf}{Square root of the working correlation matrix in the
original observation ordering.}

\item{VhalfInv}{Inverse square root of the working correlation matrix
in the original observation ordering.  When both are non-\code{NULL},
the GEE paths are used.}

\item{verbose}{Logical.}

\item{...}{Additional arguments passed to weight and dispersion functions.}
}
\value{
A named list with \code{result}, \code{beta_spline},
  \code{beta_flat}, and \code{qp_info}.
}
\description{
Stage 1 runs damped Newton-Raphson backfitting on the unwhitened link-scale working
response to produce a warm start.  Stage 2 refines via damped SQP
on the full whitened system, replicating \code{get_B} Path 1b.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{vectorproduct_block_diagonal}
\alias{vectorproduct_block_diagonal}
\title{Vector-Matrix Multiplication for Block Diagonal Matrices}
\usage{
vectorproduct_block_diagonal(A, b, K)
}
\arguments{
\item{A}{List of matrices A}

\item{b}{List of vectors b}

\item{K}{Number of blocks}
}
\value{
List of resulting vectors
}
\description{
Performs vector-matrix multiplication for block diagonal matrices
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integrate.R
\name{integrate.lgspline}
\alias{integrate.lgspline}
\title{Definite Integral of a Fitted lgspline}
\usage{
\method{integrate}{lgspline}(
  f,
  lower,
  upper,
  vars = NULL,
  initial_values = NULL,
  B_predict = NULL,
  link_scale = FALSE,
  n_quad = 50L,
  ...
)
}
\arguments{
\item{f}{A fitted \code{lgspline} object.}

\item{lower}{Numeric vector of lower bounds, one per integration variable.
Scalar values are recycled.}

\item{upper}{Numeric vector of upper bounds, one per integration variable.
Scalar values are recycled.}

\item{vars}{Default: NULL.  Character or integer vector identifying which
predictor(s) to integrate over.  When NULL all numeric predictors are
integrated simultaneously.}

\item{initial_values}{Default: NULL.  Numeric vector of length \eqn{q}
supplying fixed values for predictors not among the integration
variables.  When NULL, non-integration predictors are held at the
midpoint of their training range.}

\item{B_predict}{Default: NULL.  Optional list of coefficient vectors,
one per partition.  When NULL the fitted coefficients are used.}

\item{link_scale}{Default: FALSE.  Logical; when TRUE the integral is
computed on the link (linear predictor) scale \eqn{\eta} rather than
the response scale \eqn{\mu}.}

\item{n_quad}{Default: 50.  Number of Gauss--Legendre nodes per
integration dimension.}

\item{\ldots}{Additional arguments (currently unused; present for S3
method compatibility).}
}
\value{
A numeric scalar: the estimated definite integral.
}
\description{
Given a fitted \code{lgspline} object, computes the definite integral of
the fitted surface over a rectangular domain using Gauss--Legendre
quadrature on \code{predict()}.
}
\section{Method}{

The integration domain is discretised into a tensor-product grid of
Gauss--Legendre quadrature nodes.  Predicted values at each node come
from the model's \code{predict()} method, which correctly handles
partition assignment and piecewise polynomial evaluation.  The integral
is the weighted sum
\deqn{\int_{a_1}^{b_1} \cdots \int_{a_d}^{b_d}
      \hat{f}(\mathbf{t})\,\mathrm{d}t_1 \cdots \mathrm{d}t_d
      \;\approx\;
      \sum_{i=1}^{M} w_i\,\hat{f}(\mathbf{t}_i)}
where \eqn{M = n_{\mathrm{quad}}^d} and each weight incorporates the
Jacobian \eqn{(b_j - a_j)/2} for the affine map from \eqn{[-1, 1]} to
\eqn{[a_j, b_j]}.  Nodes and weights on \eqn{[-1, 1]} are computed via
the Golub--Welsch algorithm (eigenvalues of the symmetric tridiagonal
Jacobi matrix).

For smooth polynomials, 30--50 nodes per dimension is typically
sufficient; highly partitioned models (large \eqn{K}) may benefit from
more.  Total evaluation points scale as \eqn{n_{\mathrm{quad}}^d}, so
problems with \eqn{d \ge 4} may require reducing \code{n_quad}.
}

\section{Integration scale}{

By default (\code{link_scale = FALSE}), integration is on the response
scale \eqn{\mu = g^{-1}(\eta)}.  Setting \code{link_scale = TRUE}
integrates the linear predictor \eqn{\eta = \mathbf{x}^{\top}\boldsymbol{\beta}}
directly, which is useful when the quantity of interest is the area
under the link-transformed surface rather than the response.  For the
identity link the two coincide.
}

\examples{

## 1-D: integral of fitted sin(t) over [-pi, pi] should be near 0
set.seed(1234)
t <- seq(-pi, pi, length.out = 1000)
y <- sin(t) + rnorm(length(t), 0, 0.01)
fit <- lgspline(t, y, K = 4, opt = FALSE)
integrate(fit, lower = -pi, upper = pi)

## Base R integrate still works as expected
integrate(sin, lower = -pi, upper = pi)

## 2-D: volume under fitted volcano surface
data(volcano)
vlong <- cbind(
  rep(seq_len(nrow(volcano)), ncol(volcano)),
  rep(seq_len(ncol(volcano)), each = nrow(volcano)),
  as.vector(volcano)
)
colnames(vlong) <- c("Length", "Width", "Height")
fit_v <- lgspline(vlong[, 1:2], vlong[, 3], K = 18,
                  include_quadratic_interactions = TRUE, opt = FALSE)
integrate(fit_v, lower = c(1, 1), upper = c(87, 61))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{compute_GhalfXy_temp_wrapper}
\alias{compute_GhalfXy_temp_wrapper}
\title{Compute Component \eqn{\textbf{G}^{1/2}\textbf{A}(\textbf{A}^{T}\textbf{G}\textbf{A})^{-1}\textbf{A}^{T}\textbf{G}\textbf{X}^{T}\textbf{y}}}
\usage{
compute_GhalfXy_temp_wrapper(
  G,
  Ghalf,
  A,
  AGAInv,
  Xy,
  nc,
  K,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks
)
}
\arguments{
\item{G}{List of \eqn{\textbf{G}} matrices}

\item{Ghalf}{List of \eqn{\textbf{G}^{1/2}} matrices}

\item{A}{Constraint matrix \eqn{\textbf{A}}}

\item{AGAInv}{Inverse of \eqn{\textbf{A}^{T}\textbf{G}\textbf{A}}}

\item{Xy}{List of \eqn{\textbf{X}^{T}\textbf{y}} products}

\item{nc}{Number of columns}

\item{K}{Number of partitions minus 1 (\eqn{K})}

\item{parallel}{Use parallel processing}

\item{cl}{Cluster object}

\item{chunk_size}{Size of parallel chunks}

\item{num_chunks}{Number of chunks}

\item{rem_chunks}{Remaining chunks}
}
\value{
Unnamed two-element list containing the projected result vector and
  the intermediate
  \eqn{(\mathbf{A}^{T}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{T}\mathbf{G}\mathbf{X}^{T}\mathbf{y}}
  product.
}
\description{
Compute Component \eqn{\textbf{G}^{1/2}\textbf{A}(\textbf{A}^{T}\textbf{G}\textbf{A})^{-1}\textbf{A}^{T}\textbf{G}\textbf{X}^{T}\textbf{y}}
}
\details{
Computes the least-squares projection component
\eqn{\mathbf{G}^{1/2}\mathbf{A}(\mathbf{A}^{T}\mathbf{G}\mathbf{A})^{-1}
\mathbf{A}^{T}\mathbf{G}\mathbf{X}^{T}\mathbf{y}} together with the
intermediate product
\eqn{(\mathbf{A}^{T}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{T}\mathbf{G}\mathbf{X}^{T}\mathbf{y}}
for reuse downstream.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_partitions.R
\name{get_centers}
\alias{get_centers}
\title{Get Centers for Partitioning}
\usage{
get_centers(
  data,
  K,
  cluster_args,
  cluster_on_indicators,
  data_already_processed = FALSE
)
}
\arguments{
\item{data}{Matrix of predictor data (already processed: binary/excluded
columns zeroed out by the caller when \code{data_already_processed = TRUE})}

\item{K}{Number of partitions minus 1 (\eqn{K})}

\item{cluster_args}{List with custom centers and kmeans args}

\item{cluster_on_indicators}{Include binary predictors in clustering}

\item{data_already_processed}{Logical; when TRUE the caller has already
zeroed out binary / excluded columns, so \code{get_centers} skips its
own binary-zeroing step.  Defaults to FALSE for backward compatibility.}
}
\value{
Matrix of cluster centers
}
\description{
Get Centers for Partitioning
}
\details{
Returns partition centers via:

1. Custom supplied centers if provided as a valid \eqn{(K+1) \times q} matrix

2. kmeans clustering on all non-spline variables if \code{cluster_on_indicators=TRUE}
   or if \code{data_already_processed=TRUE}

3. kmeans clustering excluding binary variables if \code{cluster_on_indicators=FALSE}
   and \code{data_already_processed=FALSE}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.compute_Xy_V}
\alias{.compute_Xy_V}
\title{Compute Partitioned GLS Cross-Products}
\usage{
.compute_Xy_V(X, y, VhalfInv_perm, K, p_expansions, order_list)
}
\arguments{
\item{X}{List of partition-specific design matrices.}

\item{y}{List of response vectors by partition.}

\item{VhalfInv_perm}{\eqn{\mathbf{V}^{-1/2}} permuted to partition
ordering.}

\item{K, p_expansions}{Integer dimensions.}

\item{order_list}{Partition-to-data index mapping (unused here but
retained for interface consistency).}
}
\value{
A list of \eqn{K+1} column vectors, each
  \eqn{p\_expansions \times 1}.
}
\description{
Computes \eqn{\mathbf{X}_k^{\top}\mathbf{V}^{-1}\mathbf{y}} for each
partition, accounting for cross-partition contributions from the
correlation structure.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{predict.lgspline}
\alias{predict.lgspline}
\title{Predict Method for lgspline Objects}
\usage{
\method{predict}{lgspline}(
  object,
  newdata = NULL,
  parallel = FALSE,
  cl = NULL,
  chunk_size = NULL,
  num_chunks = NULL,
  rem_chunks = NULL,
  B_predict = NULL,
  take_first_derivatives = FALSE,
  take_second_derivatives = FALSE,
  expansions_only = FALSE,
  new_predictors = NULL,
  ...
)
}
\arguments{
\item{object}{A fitted lgspline model object.}

\item{newdata}{Matrix or data.frame; new predictor values. Default NULL.}

\item{parallel}{Logical; use parallel processing (experimental).
Default FALSE.}

\item{cl}{Cluster object for parallel processing. Default NULL.}

\item{chunk_size}{Integer; chunk size for parallel processing. Default NULL.}

\item{num_chunks}{Integer; number of chunks. Default NULL.}

\item{rem_chunks}{Integer; remainder chunks. Default NULL.}

\item{B_predict}{List; per-partition coefficient list for prediction, e.g.
from \code{\link{generate_posterior}}. Default NULL uses
\code{object$B}.}

\item{take_first_derivatives}{Logical; compute first derivatives.
Default FALSE.}

\item{take_second_derivatives}{Logical; compute second derivatives.
Default FALSE.}

\item{expansions_only}{Logical; return basis expansion matrix only.
Default FALSE.}

\item{new_predictors}{Matrix or data.frame; overrides \code{newdata}.}

\item{...}{Additional arguments passed to the internal predict method.}
}
\value{
A numeric vector of predictions, or a list when derivatives or
  interval summaries are requested:
\describe{
  \item{preds}{Numeric vector of predictions when derivatives are requested.}
  \item{fit}{Numeric vector of predictions when \code{se.fit = TRUE} and no
        derivatives are requested.}
  \item{first_deriv}{Numeric vector or named list of first derivatives (if
        requested).}
  \item{second_deriv}{Numeric vector or named list of second derivatives (if
        requested).}
  \item{se.fit}{Pointwise standard errors on the link scale (if requested).}
  \item{lower}{Pointwise lower interval bound (if requested).}
  \item{upper}{Pointwise upper interval bound (if requested).}
  \item{cv}{Critical value returned when \code{se.fit = TRUE} without
        derivative requests.}
}
If \code{expansions_only = TRUE}, returns a list of basis expansions.
}
\description{
Generates predictions, derivatives, and basis expansions from a fitted
lgspline model. Wrapper for the internal predict closure stored in the
object.
}
\details{
\code{new_predictors} takes priority over \code{newdata} when both are
supplied. When both are NULL, the training data is used.

Fitted values are also accessible directly as \code{model_fit$ytilde} or
via \code{model_fit$predict()}.

The parallel processing feature is experimental.

Additional arguments passed through \code{...} include \code{se.fit}
and \code{cv} for pointwise interval summaries.

Predictor input should use the original predictor columns. Named extra
columns are dropped when they can be identified as irrelevant to the
fitted expansions.
}
\examples{

set.seed(1234)
t <- runif(1000, -10, 10)
y <- 2*sin(t) + -0.06*t^2 + rnorm(length(t))
model_fit <- lgspline(t, y)

newdata <- matrix(sort(rnorm(10000)), ncol = 1)
preds <- predict(model_fit, newdata)

deriv1_res <- predict(model_fit, newdata, take_first_derivatives = TRUE)
deriv2_res <- predict(model_fit, newdata, take_second_derivatives = TRUE)

oldpar <- par(no.readonly = TRUE)
layout(matrix(c(1,1,2,2,3,3), byrow = TRUE, ncol = 2))

plot(newdata[,1], preds, main = 'Fitted Function',
     xlab = 't', ylab = "f(t)", type = 'l')
plot(newdata[,1], deriv1_res$first_deriv, main = 'First Derivative',
     xlab = 't', ylab = "f'(t)", type = 'l')
plot(newdata[,1], deriv2_res$second_deriv, main = 'Second Derivative',
     xlab = 't', ylab = "f''(t)", type = 'l')

par(oldpar)

}
\seealso{
\code{\link{lgspline}}, \code{\link{plot.lgspline}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{GAmult_wrapper}
\alias{GAmult_wrapper}
\title{Efficient Matrix Multiplication of G and A Matrices}
\usage{
GAmult_wrapper(
  G,
  A,
  K,
  p_expansions,
  R_constraints,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks
)
}
\arguments{
\item{G}{List of G matrices}

\item{A}{Constraint matrix}

\item{K}{Number of partitions minus 1}

\item{p_expansions}{Number of columns per partition}

\item{R_constraints}{Number of constraint columns}

\item{parallel}{Use parallel processing}

\item{cl}{Cluster object}

\item{chunk_size}{Size of parallel chunks}

\item{num_chunks}{Number of chunks}

\item{rem_chunks}{Remaining chunks}
}
\value{
List of matrix products, one per partition
}
\description{
Efficient Matrix Multiplication of G and A Matrices
}
\details{
Computes G %**% A when G has block diagonal structure and A is a matrix.
Processes in parallel chunks if enabled.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weibull_helpers.R
\name{weibull_qp_score_function}
\alias{weibull_qp_score_function}
\title{Compute Gradient of Log-Likelihood of Weibull Accelerated Failure Model}
\usage{
weibull_qp_score_function(
  X,
  y,
  mu,
  order_list,
  dispersion,
  VhalfInv,
  observation_weights,
  status
)
}
\arguments{
\item{X}{Design matrix}

\item{y}{Response vector}

\item{mu}{Predicted mean vector}

\item{order_list}{List of observation indices per partition}

\item{dispersion}{Dispersion parameter (sigma^2 = scale^2). The lgspline
framework stores and passes dispersion (sigma^2); the Weibull scale
(sigma) matching \code{survreg$scale} is \code{sqrt(dispersion)}.}

\item{VhalfInv}{Inverse square root of correlation matrix; unused here and
retained for interface compatibility.}

\item{observation_weights}{Observation weights}

\item{status}{Censoring indicator (1 = event, 0 = censored)}
}
\value{
Numeric column vector representing the gradient with respect to
coefficients.
}
\description{
Calculates the gradient of log-likelihood for a Weibull accelerated failure
time (AFT) survival model, supporting right-censored survival data.
}
\details{
Needed if using "blockfit", correlation structures, or quadratic programming
with Weibull AFT models.

The gradient is computed on a scale that omits the 1/sigma prefactor.
Specifically, the true score is (1/sigma) * X^T diag(w) (exp(z) - status),
but both this function and the corresponding information matrix used
internally omit 1/sigma and 1/sigma^2 respectively, so the Newton-Raphson
step G*u remains correct. This matches the convention in
\code{\link{unconstrained_fit_weibull}}.
}
\examples{

set.seed(1234)
t1 <- rnorm(1000)
t2 <- rbinom(1000, 1, 0.5)
yraw <- rexp(exp(0.01*t1 + 0.01*t2))
status <- rbinom(1000, 1, 0.25)
yobs <- ifelse(status, runif(length(yraw), 0, yraw), yraw)
df <- data.frame(
  y = yobs,
  t1 = t1,
  t2 = t2
)

## Example using blockfit for t2 as a linear term - output does not look
# different, but internal methods used for fitting change
model_fit <- lgspline(y ~ spl(t1) + t2,
                      df,
                      unconstrained_fit_fxn = unconstrained_fit_weibull,
                      family = weibull_family(),
                      need_dispersion_for_estimation = TRUE,
                      qp_score_function = weibull_qp_score_function,
                      dispersion_function = weibull_dispersion_function,
                      glm_weight_function = weibull_glm_weight_function,
                      schur_correction_function = weibull_schur_correction,
                      K = 1,
                      blockfit = TRUE,
                      opt = FALSE,
                      status = status,
                      verbose = TRUE)

print(summary(model_fit))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cox_helpers.R
\name{cox_schur_correction}
\alias{cox_schur_correction}
\title{Cox PH Schur Correction}
\usage{
cox_schur_correction(
  X,
  y,
  B,
  dispersion,
  order_list,
  K,
  family,
  observation_weights,
  ...
)
}
\arguments{
\item{X}{List of partition design matrices.}

\item{y}{List of partition response vectors.}

\item{B}{List of partition coefficient vectors.}

\item{dispersion}{Dispersion (fixed at 1).}

\item{order_list}{List of observation indices per partition.}

\item{K}{Number of knots.}

\item{family}{Family object.}

\item{observation_weights}{Observation weights.}

\item{...}{Additional arguments.}
}
\value{
List of K+1 zeros.
}
\description{
Returns zero corrections for all partitions. Cox PH has no nuisance
dispersion parameter, so no Schur complement correction to the
information matrix is needed. This function exists for interface
compatibility with lgspline's \code{schur_correction_function}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{reml_grad_from_dV}
\alias{reml_grad_from_dV}
\title{Evaluate the REML gradient with respect to a single correlation parameter}
\usage{
reml_grad_from_dV(dV, model_fit, glm_weight_function, ...)
}
\arguments{
\item{dV}{\eqn{N \times N} numeric matrix giving
\eqn{\partial\mathbf{V}/\partial\rho} for one scalar parameter
\eqn{\rho}, with the chain rule through any reparameterization already
applied.}

\item{model_fit}{A fitted \code{lgspline} object; see
\code{\link{lgspline}}.}

\item{glm_weight_function}{The GLM weight function used during model
fitting; see the \code{glm_weight_function} argument of
\code{\link{lgspline}}.}

\item{...}{Additional arguments forwarded to \code{glm_weight_function}.}
}
\value{
A scalar: the gradient of the negative REML objective with respect
  to \eqn{\rho}, divided by \eqn{N}.
}
\description{
Computes the derivative of the negative REML objective with respect to a
scalar correlation parameter, given the matrix derivative
\eqn{\partial \mathbf{V}/\partial \rho}.
}
\details{
\subsection{Notation}{
\eqn{\mathbf{D} = \mathrm{diag}(d_1,\ldots,d_N)} is the diagonal matrix of
observation weights (\code{observation_weights}), and
\eqn{\mathbf{W} = \mathrm{diag}(w_1,\ldots,w_N)} is the diagonal matrix of
GLM working weights evaluated at the current fitted values via
\code{glm_weight_function}, with the observation-weight contribution
carried separately by \eqn{\mathbf{D}}.  For canonical GLM families,
\eqn{w_i} is the usual IRWLS/Fisher-scoring weight on the mean--variance
scale; for example, logistic regression gives
\eqn{w_i = \mu_i(1-\mu_i)}.  The combined weighting entering the
information matrix is \eqn{\mathbf{W}\mathbf{D}}.  In Gaussian identity
models both reduce to scalar multiples of the identity.

\eqn{\mathbf{V}} is the \eqn{N \times N} correlation matrix implied by
\code{VhalfInv}, with
\eqn{\mathbf{V}^{-1} = (\mathbf{V}^{-1/2})^\top \mathbf{V}^{-1/2}}.

The penalized observed information at the current iterate is
\deqn{\mathbf{M} = (\mathbf{X}^*)^\top
  \mathbf{V}^{-1}\mathbf{W}\mathbf{D}\,\mathbf{X}^*
  + \mathbf{U}^\top\boldsymbol{\Lambda}\mathbf{U},}
where \eqn{\mathbf{X}^* = \mathbf{X}\mathbf{U}} is the constrained design
(\eqn{N \times P}) and the first term is the quadratic approximation to the
penalized log-likelihood Hessian at the current \eqn{\boldsymbol{\mu}}.
For non-Gaussian families this is a local approximation (the IRWLS/Fisher
scoring Hessian), not an exact GLS information matrix.

The constraint projection is
\eqn{\mathbf{U} = \mathbf{I} -
  \mathbf{G}\mathbf{A}(\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}
  \mathbf{A}^\top}
with \eqn{\mathbf{G} = \mathbf{M}^{-1}}.
\eqn{\mathbf{U}} is idempotent (\eqn{\mathbf{U}^2 = \mathbf{U}}) but
\emph{not} symmetric, so
\eqn{\mathbf{U}^\top\boldsymbol{\Lambda}\mathbf{U} \neq
  \mathbf{U}\boldsymbol{\Lambda}\mathbf{U}^\top}.

The \code{U} stored in \code{model_fit$U} is on the expansion- and
response-standardised scale.  The rescaled version used here is
\deqn{\tilde{\mathbf{U}} =
  \mathbf{U} \cdot \mathrm{diag}(1, s_1, \ldots, s_{p-1},
    1, s_1, \ldots) / \hat{\sigma}_y,}
where \eqn{s_j} are the expansion scales and \eqn{\hat{\sigma}_y}
standardises the response.  All quantities below use
\eqn{\tilde{\mathbf{U}}} in place of \eqn{\mathbf{U}}.
}

\subsection{REML objective and its gradient}{
The REML objective is constructed by integrating out the fixed effects from
the penalized log-likelihood, using a Laplace approximation to the
marginal likelihood for non-Gaussian families.  This approximation is exact
for Gaussian identity models and is the standard extension used in
restricted maximum likelihood estimation for GLMMs.

The REML correction term is
\eqn{-\tfrac{1}{2}\log|\mathbf{M}|}, where \eqn{\mathbf{M}} is the
penalized observed information defined above.  Differentiating with respect
to \eqn{\rho} (noting only \eqn{\mathbf{V}} depends on \eqn{\rho})
gives the REML correction gradient
\deqn{-\frac{1}{2}\mathrm{tr}\!\Bigl(
  \mathbf{M}^{-1}
  (\mathbf{X}^*)^\top\mathbf{V}^{-1}
  \frac{\partial\mathbf{V}}{\partial\rho}
  \mathbf{V}^{-1}\mathbf{W}\mathbf{D}\,\mathbf{X}^*
  \Bigr).}
}

\subsection{Full gradient}{
\deqn{\frac{\partial(-\ell_R)}{\partial\rho}
  = \frac{1}{N}\Biggl[
    \underbrace{
      \frac{1}{2}\mathrm{tr}\!\Bigl(
        \mathbf{V}^{-1}\frac{\partial\mathbf{V}}{\partial\rho}
      \Bigr)
    }_{\text{(i) log-det of }\mathbf{V}}
    \underbrace{
      -\frac{1}{2\tilde{\sigma}^2}
      \mathbf{r}^\top\frac{\partial\mathbf{V}}{\partial\rho}\mathbf{r}
    }_{\text{(ii) residual quadratic form}}
    \underbrace{
      -\frac{1}{2}\mathrm{tr}\!\Bigl(
        \mathbf{M}^{-1}(\mathbf{X}^*)^\top\mathbf{V}^{-1}
        \frac{\partial\mathbf{V}}{\partial\rho}
        \mathbf{V}^{-1}\mathbf{W}\mathbf{D}\,\mathbf{X}^*
      \Bigr)
    }_{\text{(iii) REML correction}}
  \Biggr],}
where the whitened residual is
\deqn{\mathbf{r} =
  \mathrm{diag}\!\left(\sqrt{d_i}/\sqrt{w_i}\right)
  \mathbf{V}^{-1/2}(\mathbf{y} - \boldsymbol{\mu}),}
and \eqn{r_i =
  [\mathbf{V}^{-1/2}(\mathbf{y}-\boldsymbol{\mu})]_i
  \sqrt{d_i}/\sqrt{w_i}}.

Term (i) does not involve \eqn{\mathbf{D}} or \eqn{\mathbf{W}}; the
log-determinant of \eqn{\mathbf{V}} depends only on the correlation
structure.  For non-Gaussian families terms (ii) and (iii) are evaluated
at the current IRWLS iterate and constitute a local approximation.
}
}
\seealso{
\code{\link{lgspline}}, \code{\link{lgspline.fit}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.lagrangian_project_woodbury}
\alias{.lagrangian_project_woodbury}
\title{Woodbury-Corrected Lagrangian Projection}
\usage{
.lagrangian_project_woodbury(
  GhalfXy_V,
  Ghalf_corrected,
  A,
  K,
  p_expansions,
  R_constraints,
  constraint_value_vectors,
  family,
  wb_sqrt,
  parallel_aga,
  parallel_matmult,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks
)
}
\arguments{
\item{GhalfXy_V}{Numeric column vector of length \eqn{P};
\eqn{\mathbf{G}_{\mathrm{on}}^{1/2}} applied to the per-partition GLS
cross-products \eqn{\mathbf{X}_k^{\top}\mathbf{V}^{-1}\mathbf{y}_k}.}

\item{Ghalf_corrected}{List of \eqn{K+1} corrected
\eqn{\mathbf{G}_{\mathrm{on},k}^{1/2}} matrices.}

\item{A}{Constraint matrix (\eqn{P \times R}).}

\item{K, p_expansions, R_constraints}{Integer dimensions.}

\item{constraint_value_vectors}{Constraint RHS list encoding
\eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{c}}.}

\item{family}{GLM family object.}

\item{wb_sqrt}{Output of \code{.woodbury_halfsqrt_components},
containing the internal rank-\eqn{r} representation of
\eqn{\mathbf{F}^{1/2}}: \code{U_Q} / \code{Q_basis}
(basis matching supplement \eqn{\mathbf{Q}}), \code{C} and \code{C_inv} (diagonal correction
coefficients), and \code{GchalfUQ} (partition-wise copies of
\eqn{\mathbf{G}_{\mathrm{on},k}^{1/2}\mathbf{U}_Q[rows_k,]}).}

\item{parallel_aga, parallel_matmult}{Logical flags.}

\item{cl, chunk_size, num_chunks, rem_chunks}{Parallel parameters.}
}
\value{
A list of \eqn{K+1} coefficient column vectors.
}
\description{
Applies the constrained projection for the structured-correlation
Woodbury path. In supplement notation,
\eqn{\mathbf{G}^{1/2} = \mathbf{G}_{\mathrm{on}}^{1/2}\mathbf{F}^{1/2}},
where \eqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{E}\mathbf{J}\mathbf{E}^{\top}}
and \eqn{\mathbf{F}^{1/2}} carries the low-rank correction.
}
\details{
The routine follows the same four transformed-OLS steps as
\code{.lagrangian_project}: form \eqn{\mathbf{y}^{*}},
form \eqn{\mathbf{X}^{*}}, solve the OLS projection, then back-transform.
Only steps involving \eqn{\mathbf{G}^{1/2}} differ, and those are handled
as a block-diagonal multiplication through
\eqn{\mathbf{G}_{\mathrm{on}}^{1/2}} plus a rank-\eqn{r} update.

Internally, \eqn{\mathbf{F}^{1/2}} is stored as
\eqn{\mathbf{I} - \mathbf{U}_Q\mathbf{C}\mathbf{U}_Q^{\top}}, where
\code{U_Q} spans the same subspace as the supplement matrix
\eqn{\mathbf{Q}} and \code{C} is diagonal.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weibull_helpers.R
\name{weibull_schur_correction}
\alias{weibull_schur_correction}
\alias{weibull_shur_correction}
\title{Correction for the Variance-Covariance Matrix for Uncertainty in Scale}
\usage{
weibull_schur_correction(
  X,
  y,
  B,
  dispersion,
  order_list,
  K,
  family,
  observation_weights,
  status
)

weibull_shur_correction(
  X,
  y,
  B,
  dispersion,
  order_list,
  K,
  family,
  observation_weights,
  status
)
}
\arguments{
\item{X}{Block-diagonal matrices of spline expansions}

\item{y}{Block-vector of response}

\item{B}{Block-vector of coefficient estimates}

\item{dispersion}{Scalar, estimate of dispersion (sigma^2 = scale^2). The
lgspline framework stores and passes dispersion (sigma^2); the Weibull
scale (sigma) matching \code{survreg$scale} is \code{sqrt(dispersion)}.}

\item{order_list}{List of partition orders}

\item{K}{Number of partitions minus 1 (\eqn{K})}

\item{family}{Distribution family}

\item{observation_weights}{Optional observation weights (default = 1)}

\item{status}{Censoring indicator (1 = event, 0 = censored)
Indicates whether an event of interest occurred (1) or the observation was
right-censored (0). In survival analysis, right-censoring occurs when the
full survival time is unknown, typically because the study ended or the
subject was lost to follow-up before the event of interest occurred.}
}
\value{
List of blockwise Schur-complement corrections \eqn{\textbf{S}_k} to be
elementwise added to each block of the information matrix before inversion,
with \code{0} returned for empty partitions.
}
\description{
Computes the Schur complement \eqn{\textbf{S}} such that
\eqn{\textbf{G}^* = (\textbf{G}^{-1} + \textbf{S})^{-1}} properly
accounts for uncertainty in estimating the Weibull scale parameter when
estimating the variance-covariance matrix. Otherwise, the
variance-covariance matrix is optimistic and assumes the scale is known,
when it was in fact estimated. Note that the parameterization adds the
output of this function elementwise (not subtract) so for most cases, the
output of this function will be negative or a negative
definite/semi-definite matrix.
}
\details{
Adjusts the variance-covariance matrix unscaled for coefficients to account
for uncertainty in estimating the Weibull scale parameter, that otherwise
would be lost if simply using
\eqn{\textbf{G}=(\textbf{X}^{T}\textbf{W}\textbf{X} + \textbf{L})^{-1}}.
This is accomplished using a correction based on the Schur complement so we
avoid having to construct the entire variance-covariance matrix, or
modifying the procedure for \code{\link{lgspline}} substantially.
For any model with nuisance parameters that must have uncertainty accounted
for, this tool will be helpful.

This both provides a tool for actually fitting Weibull accelerated failure
time (AFT) models, and boilerplate code for users who wish to incorporate
Lagrangian multiplier smoothing splines into their own custom models.
}
\examples{

## Minimal example of fitting a Weibull Accelerated Failure Time model
# Simulating survival data with right-censoring
set.seed(1234)
t1 <- rnorm(1000)
t2 <- rbinom(1000, 1, 0.5)
yraw <- rexp(exp(0.01*t1 + 0.01*t2))
# status: 1 = event occurred, 0 = right-censored
status <- rbinom(1000, 1, 0.25)
yobs <- ifelse(status, runif(length(yraw), 0, yraw), yraw)
df <- data.frame(
  y = yobs,
  t1 = t1,
  t2 = t2
)

## Fit model using lgspline with Weibull Schur correction
model_fit <- lgspline(y ~ spl(t1) + t2,
                      df,
                      unconstrained_fit_fxn = unconstrained_fit_weibull,
                      family = weibull_family(),
                      need_dispersion_for_estimation = TRUE,
                      dispersion_function = weibull_dispersion_function,
                      glm_weight_function = weibull_glm_weight_function,
                      schur_correction_function = weibull_schur_correction,
                      status = status,
                      opt = FALSE,
                      K = 1)

print(summary(model_fit))

## Fit model using lgspline without Weibull Schur correction
naive_fit <- lgspline(y ~ spl(t1) + t2,
                      df,
                      unconstrained_fit_fxn = unconstrained_fit_weibull,
                      family = weibull_family(),
                      need_dispersion_for_estimation = TRUE,
                      dispersion_function = weibull_dispersion_function,
                      glm_weight_function = weibull_glm_weight_function,
                      status = status,
                      opt = FALSE,
                      K = 1)

print(summary(naive_fit))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blockfit_solve.R
\name{blockfit_solve}
\alias{blockfit_solve}
\title{Backfitting Solver for Blockfit Models}
\usage{
blockfit_solve(
  X,
  y,
  flat_cols,
  K,
  p_expansions,
  Lambda,
  L_partition_list,
  unique_penalty_per_partition,
  A,
  R_constraints,
  constraint_values,
  X_gram,
  Ghalf_full,
  GhalfInv_full,
  family,
  order_list,
  glm_weight_function,
  schur_correction_function,
  need_dispersion_for_estimation,
  dispersion_function,
  observation_weights,
  homogenous_weights = TRUE,
  iterate,
  tol,
  parallel_eigen,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  return_G_getB,
  quadprog = FALSE,
  qp_Amat = NULL,
  qp_bvec = NULL,
  qp_meq = NULL,
  qp_score_function = NULL,
  keep_weighted_Lambda = FALSE,
  max_backfit_iter = 100,
  Vhalf = NULL,
  VhalfInv = NULL,
  include_warnings = TRUE,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{X}{List of \eqn{K+1} design matrices
(\eqn{N_k \times p_expansions}). Unwhitened even when GEE.}

\item{y}{List of \eqn{K+1} response vectors. Unwhitened even when GEE.}

\item{flat_cols}{Integer vector indicating flat columns of
\eqn{\mathbf{X}^{(k)}}.}

\item{K}{Integer; number of interior knots.}

\item{p_expansions}{Integer; number of coefficients per partition.}

\item{Lambda}{\eqn{p_expansions \times p_expansions} penalty matrix.}

\item{L_partition_list}{List of partition-specific penalty matrices.}

\item{unique_penalty_per_partition}{Logical.}

\item{A}{Full \eqn{P \times R_constraints} constraint matrix.}

\item{R_constraints}{Integer; number of columns of \eqn{\mathbf{A}}.}

\item{constraint_values}{List of constraint right-hand sides.}

\item{X_gram}{List of Gram matrices
\eqn{\mathbf{X}_k^{\top} \mathbf{X}_k}.}

\item{Ghalf_full, GhalfInv_full}{Lists of
\eqn{\mathbf{G}^{1/2}} and \eqn{\mathbf{G}^{-1/2}} matrices.}

\item{family}{GLM family object.}

\item{order_list}{List of observation index vectors by partition.}

\item{glm_weight_function}{GLM weight function.}

\item{schur_correction_function}{Schur complement correction function.}

\item{need_dispersion_for_estimation}{Logical.}

\item{dispersion_function}{Dispersion estimation function.}

\item{observation_weights}{List of observation weights.}

\item{homogenous_weights}{Logical.}

\item{iterate}{Logical; if FALSE, single pass (no iteration).}

\item{tol}{Convergence tolerance.}

\item{parallel_eigen, cl, chunk_size, num_chunks, rem_chunks}{Parallel arguments.}

\item{return_G_getB}{Logical.}

\item{quadprog}{Logical; apply inequality constraint refinement if TRUE.}

\item{qp_Amat}{Inequality constraint matrix for
\code{quadprog::solve.QP}.}

\item{qp_bvec}{Inequality constraint right-hand side.}

\item{qp_meq}{Number of leading equality constraints.}

\item{qp_score_function}{Score function for QP subproblem.}

\item{keep_weighted_Lambda}{Logical.}

\item{max_backfit_iter}{Integer.}

\item{Vhalf}{Square root of the working correlation matrix in the
original observation ordering.}

\item{VhalfInv}{Inverse square root of the working correlation matrix
in the original observation ordering.  When both are non-\code{NULL},
the GEE paths are used.}

\item{include_warnings}{Logical.}

\item{verbose}{Logical.}

\item{...}{Additional arguments passed to weight and dispersion functions.}
}
\value{
A list with elements:
\describe{
  \item{B}{
    List of \eqn{K+1} coefficient vectors
    (\eqn{p_expansions \times 1}), flat coefficients replicated across partitions.
  }
  \item{G_list}{
    List containing \eqn{\mathbf{G}},
    \eqn{\mathbf{G}^{1/2}}, and
    \eqn{\mathbf{G}^{-1/2}},
    each a list of \eqn{K+1}
    \eqn{p_expansions \times p_expansions} matrices,
    or NULL if \code{return_G_getB} is FALSE.

    \eqn{\mathbf{G}} satisfies
    \eqn{\mathbf{G}
    =
    \mathbf{G}^{1/2}
    (\mathbf{G}^{1/2})^{\top}} exactly.

    When a correlation structure is present, these matrices are obtained
    from the dense whitened information matrix and then split back into
    per-partition blocks.  When flat columns are present, the returned
    matrices also reflect the pooled flat-column information across
    partitions.
  }
  \item{qp_info}{
    NULL when no inequality-refinement metadata are produced.
    Otherwise a list of available QP or active-set diagnostics.
    Dense SQP solves include \code{solution}, \code{lagrangian},
    \code{active_constraints}, \code{iact}, \code{Amat_active},
    \code{bvec_active}, \code{meq_active}, \code{converged}, and
    \code{final_deviance}; non-GEE dense solves also carry
    \code{info_matrix}, \code{Amat_combined}, \code{bvec_combined},
    and \code{meq_combined}.  Partition-wise active-set refinement
    returns the corresponding active-constraint summary only.
  }
}
}
\description{
Fits models with mixed spline and non-interactive linear ("flat") terms
using an iterative backfitting approach. Flat terms receive a single
shared coefficient across partitions (rather than \eqn{K+1}
partition-specific coefficients constrained to equality), reducing the
effective parameter count and improving efficiency when the number of
flat terms is large relative to spline terms.

The backfitting loop alternates between:
\enumerate{

\item \strong{Spline step}

Fit spline terms on the response adjusted for the current flat
contribution using a constrained penalized least squares solve.

If \eqn{\mathbf{y}_k} is the response vector for partition \eqn{k},
\eqn{\mathbf{Z}_k} the spline design matrix,
\eqn{\mathbf{X}_{\mathrm{flat}}^{(k)}} the flat design matrix,
and \eqn{\mathbf{v}} the flat coefficient vector, then

\deqn{
\boldsymbol{\beta}_{\mathrm{spline}}^{(k)} =
\arg\min_{\boldsymbol{\beta}}
\left\|
\mathbf{y}_k
- \mathbf{Z}_k \boldsymbol{\beta}
- \mathbf{X}_{\mathrm{flat}}^{(k)} \mathbf{v}
\right\|^2
+
\boldsymbol{\beta}^{\top}
\mathbf{\Lambda}_{\mathrm{spline}}
\boldsymbol{\beta}
}

subject to

\deqn{
\mathbf{A}_{\mathrm{spline}}^{\top}
\boldsymbol{\beta}
=
\mathbf{c}_{\mathrm{spline}}.
}

\item \strong{Flat step}

Update flat coefficients via pooled penalized regression on residuals,
subject to the residual equality constraint from any mixed constraints:

\deqn{
\mathbf{v}
=
\arg\min_{\mathbf{v}}
\sum_{k=0}^{K}
\left\|
\mathbf{y}_k
-
\mathbf{Z}_k
\boldsymbol{\beta}_{\mathrm{spline}}^{(k)}
-
\mathbf{X}_{\mathrm{flat}}^{(k)}
\mathbf{v}
\right\|^2
+
\mathbf{v}^{\top}
\mathbf{\Lambda}_{\mathrm{flat}}
\mathbf{v}
}

subject to

\deqn{
\tilde{\mathbf{A}}_{\mathrm{flat}}^{\top}
\mathbf{v}
=
\mathbf{c}
-
\mathbf{A}_{\mathrm{spline}}^{\top}
\boldsymbol{\beta}_{\mathrm{spline}}.
}
}

Four estimation paths are selected automatically based on the model
configuration:
\describe{
  \item{Case (a)}{Gaussian identity + GEE: closed-form solve via
    \code{.bf_case_gauss_gee}.}
  \item{Case (b)}{Gaussian identity, no correlation: standard
    backfitting via \code{.bf_case_gauss_no_corr}.}
  \item{Case (c)}{GLM + GEE: two-stage (damped Newton-Raphson warm start then
    damped SQP) via \code{.bf_case_glm_gee}.}
  \item{Case (d)}{GLM without GEE: damped Newton-Raphson + backfitting via
    \code{.bf_case_glm_no_corr}.}
}

When \code{quadprog = TRUE} without GEE, inequality constraints are
enforced after backfitting convergence.  The constraint handling
method is selected automatically by inspecting the sparsity pattern
of \code{qp_Amat}: if every column has nonzeros in only one
partition block (e.g.\ derivative sign or range constraints),
a partition-wise active-set method is used via
\code{.active_set_refine}, avoiding the dense \eqn{P \times P}
system.  If any column spans multiple partition blocks (e.g.\
cross-knot monotonicity), or if the active-set method does not
converge, the dense SQP fallback via \code{.bf_sqp_loop} is used.
GEE paths always use the dense system.

After convergence, coefficients are reassembled into the standard
per-partition format (flat coefficients replicated across partitions)
for compatibility with downstream inference.
}
\examples{
\dontrun{
## Minimal verification example: Gaussian identity, no correlation,
## 2 partitions, 3 spline columns + 1 flat column per partition.
##
## This confirms that blockfit_solve returns compatible output and
## that flat coefficients are replicated across partitions.

set.seed(1234)
n1 <- 50; n2 <- 50
p_expansions <- 4  # intercept + x + x^2 + z (flat)
K  <- 1  # 2 partitions

X1 <- cbind(1, rnorm(n1), rnorm(n1)^2, rnorm(n1))
X2 <- cbind(1, rnorm(n2), rnorm(n2)^2, rnorm(n2))
beta_true <- c(1, 0.5, -0.3, 0.8)
y1 <- X1 \%*\% beta_true + rnorm(n1, 0, 0.5)
y2 <- X2 \%*\% beta_true + rnorm(n2, 0, 0.5)

## Constraint: spline coefficients equal across partitions
## (columns 1:3 constrained, column 4 is flat)
A <- matrix(0, nrow = p_expansions * (K + 1), ncol = 3)
for(j in 1:3){
  A[j, j]      <-  1
  A[p_expansions + j, j]  <- -1
}
qr_A <- qr(A)
A <- qr.Q(qr_A)[, 1:qr_A$rank, drop = FALSE]

Lambda <- diag(p_expansions) * 1e-4
L_partition_list <- list(0, 0)
X_gram <- list(crossprod(X1), crossprod(X2))

G_list <- compute_G_eigen(
  X_gram, Lambda, K,
  parallel = FALSE, cl = NULL,
  chunk_size = NULL, num_chunks = NULL, rem_chunks = NULL,
  family = gaussian(),
  unique_penalty_per_partition = FALSE,
  L_partition_list = L_partition_list,
  keep_G = TRUE,
  schur_corrections = list(0, 0))

result <- blockfit_solve(
  X = list(X1, X2),
  y = list(y1, y2),
  flat_cols = 4L,
  K = K, p_expansions = p_expansions,
  Lambda = Lambda,
  L_partition_list = L_partition_list,
  unique_penalty_per_partition = FALSE,
  A = A, R_constraints = ncol(A),
  constraint_values = list(),
  X_gram = X_gram,
  Ghalf_full = G_list$Ghalf,
  GhalfInv_full = G_list$GhalfInv,
  family = gaussian(),
  order_list = list(1:n1, (n1+1):(n1+n2)),
  glm_weight_function = function(mu, y, oi, fam, d, ow, ...) rep(1, length(y)),
  schur_correction_function = function(X, y, B, d, ol, K, fam, ow, ...) {
    lapply(1:(K + 1), function(k) 0)
  },
  need_dispersion_for_estimation = FALSE,
  dispersion_function = function(...) 1,
  observation_weights = list(rep(1, n1), rep(1, n2)),
  homogenous_weights = TRUE,
  iterate = TRUE, tol = 1e-8,
  parallel = FALSE, cl = NULL,
  chunk_size = NULL, num_chunks = NULL, rem_chunks = NULL,
  return_G_getB = TRUE,
  verbose = FALSE)

## Verify: flat coefficient (col 4) is identical across partitions
stopifnot(abs(result$B[[1]][4] - result$B[[2]][4]) < 1e-10)

## Verify: output structure
stopifnot(length(result$B) == K + 1)
stopifnot(length(result$B[[1]]) == p_expansions)
stopifnot(!is.null(result$G_list))
}

}
\seealso{
\code{\link{lgspline}}, \code{\link{lgspline.fit}},
  \code{\link{get_B}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{std}
\alias{std}
\title{Standardize Vector to Z-Scores}
\usage{
std(x)
}
\arguments{
\item{x}{Numeric vector to standardize}
}
\value{
Standardized vector with sample mean 0 and standard deviation 1
}
\description{
Centers a vector by its sample mean, then scales it by its sample standard deviation
\eqn{(\text{x}-\text{mean}(\text{x}))/\text{sd}(\text{x})}.
}
\examples{

x <- c(1, 2, 3, 4, 5)
std(x)
print(mean(x))
print(sd(x))


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blockfit_solve.R
\name{.bf_inner_weighted}
\alias{.bf_inner_weighted}
\title{GLM Backfitting Inner Loop}
\usage{
.bf_inner_weighted(
  X_spline,
  X_flat,
  z_list,
  W_list,
  K,
  Lambda_spline,
  Lambda_flat,
  L_part_spline,
  unique_penalty_per_partition,
  A_spline,
  constraint_values_spline,
  nc_spline,
  nc_flat,
  beta_spline_init,
  beta_flat_init,
  tol,
  max_backfit_iter,
  parallel_eigen,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  split = NULL,
  constraint_values = list(),
  p_expansions = NULL
)
}
\arguments{
\item{X_spline, X_flat}{Lists of per-partition submatrices.}

\item{z_list}{List of \eqn{K+1} working-response vectors.}

\item{W_list}{List of \eqn{K+1} weight vectors.}

\item{K}{Integer.}

\item{Lambda_spline, Lambda_flat}{Penalty submatrices.}

\item{L_part_spline}{Partition-specific spline penalty list.}

\item{unique_penalty_per_partition}{Logical.}

\item{A_spline}{Spline-only constraint matrix.}

\item{constraint_values_spline}{Constraint RHS list.}

\item{nc_spline, nc_flat}{Integers.}

\item{beta_spline_init, beta_flat_init}{Initial coefficient values.}

\item{tol}{Convergence tolerance.}

\item{max_backfit_iter}{Maximum iterations.}

\item{parallel_eigen, cl, chunk_size, num_chunks, rem_chunks}{Parallel arguments forwarded to \code{compute_G_eigen}.}

\item{split}{Output of \code{.bf_split_components}, needed for
mixed-constraint enforcement on the flat step.}

\item{constraint_values}{Full constraint RHS list.}

\item{p_expansions}{Integer; columns per partition.}
}
\value{
A named list with \code{beta_spline} and
  \code{beta_flat}.
}
\description{
Given damped Newton-Raphson weights and response partitioned into lists,
runs the inner backfitting loop that alternates between a weighted
spline step and a weighted flat step until convergence.  Shared by
both Case c (GEE warm start) and Case d (GLM without GEE).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{armaInv}
\alias{armaInv}
\title{Matrix Inversion using Armadillo}
\usage{
armaInv(x)
}
\arguments{
\item{x}{Input matrix to be inverted}
}
\value{
Inverted matrix
}
\description{
Computes the inverse of a matrix using Armadillo's inversion method
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/negbin_helpers.R
\name{negbin_schur_correction}
\alias{negbin_schur_correction}
\title{NB Schur Correction}
\usage{
negbin_schur_correction(
  X,
  y,
  B,
  dispersion,
  order_list,
  K,
  family,
  observation_weights
)
}
\arguments{
\item{X}{List of partition design matrices.}

\item{y}{List of partition response vectors.}

\item{B}{List of partition coefficient vectors.}

\item{dispersion}{Scalar shape parameter \eqn{\theta}.}

\item{order_list}{List of observation indices per partition.}

\item{K}{Number of knots.}

\item{family}{Family object.}

\item{observation_weights}{Observation weights.}
}
\value{
List of K+1 correction matrices, with \code{0} for empty
  partitions.
}
\description{
Computes the Schur complement correction to account for uncertainty in
estimating \eqn{\theta}. Structure is identical to
\code{\link{weibull_schur_correction}}: the joint information is
partitioned into \eqn{(\boldsymbol{\beta}, \theta)} blocks and the
correction is \eqn{-\mathbf{I}_{\beta\theta}
I_{\theta\theta}^{-1}\mathbf{I}_{\beta\theta}^{\top}}.
}
\details{
The cross-derivative (score of \eqn{\eta_i} w.r.t. \eqn{\theta}) is
\deqn{\frac{\partial^2 \ell}{\partial \eta_i \partial \theta}
  = \frac{(y_i - \mu_i)\mu_i}{(\theta + \mu_i)^2}}
The second derivative of the log-likelihood w.r.t. \eqn{\theta} is
\deqn{I_{\theta\theta} = -\sum_i w_i \bigl[
  \psi'(y_i + \theta) - \psi'(\theta) +
  \frac{1}{\theta} - \frac{2}{\mu_i + \theta} +
  \frac{y_i + \theta}{(\mu_i + \theta)^2}\bigr]}
where \eqn{\psi'} is the trigamma function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{matsqrt}
\alias{matsqrt}
\title{Calculate Matrix Square Root for Symmetric Matrices}
\usage{
matsqrt(mat)
}
\arguments{
\item{mat}{A symmetric matrix \eqn{\textbf{M}}}
}
\value{
A matrix \eqn{\textbf{B}} such that \eqn{\textbf{B}\textbf{B}}
  equals \eqn{\textbf{M}} on the positive-eigenvalue subspace, with
  non-positive components truncated to 0.
}
\description{
Calculate Matrix Square Root for Symmetric Matrices
}
\details{
Uses an eigenvalue-decomposition-based approach.

Non-positive eigenvalues are set to 0 before taking fourth roots.

This implementation is particularly useful for whitening procedures in GLMs
with correlation structures and for computing variance-covariance matrices
under constraints.

You can use this to help construct a custom \code{Vhalf_fxn}, or more
directly to build the \eqn{\mathbf{V}^{1/2}} input supplied to
\code{\link{lgspline}} for correlation-aware fits.
}
\examples{

## Identity matrix
m1 <- diag(2)
matsqrt(m1)  # Returns identity matrix

## Compound symmetry correlation matrix
rho <- 0.5
m2 <- matrix(rho, 3, 3) + diag(1-rho, 3)
B <- matsqrt(m2)
# Verify: B \%**\% B approximately equals m2
all.equal(B \%**\% B, m2)

## Example for correlation structure
n_blocks <- 2  # Number of subjects
block_size <- 3  # Measurements per subject
rho <- 0.7  # Within-subject correlation
# Correlation matrix for one subject
R <- matrix(rho, block_size, block_size) +
     diag(1-rho, block_size)
# Full correlation matrix for all subjects
V <- kronecker(diag(n_blocks), R)
Vhalf <- matsqrt(V)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blockfit_solve.R
\name{.bf_lagrangian_project}
\alias{.bf_lagrangian_project}
\title{Lagrangian Projection for the Spline-Only Subproblem}
\usage{
.bf_lagrangian_project(
  Xy_adj,
  Ghalf_cur,
  GhalfA_cur,
  cv_spline,
  K,
  nc_spline,
  A_spline
)
}
\arguments{
\item{Xy_adj}{List of \eqn{K+1} adjusted cross-product vectors
(\eqn{nc_s \times 1}).}

\item{Ghalf_cur}{List of \eqn{K+1} \eqn{\mathbf{G}_s^{1/2}}
matrices (each \eqn{nc_s \times nc_s}).}

\item{GhalfA_cur}{\eqn{(nc_s \cdot (K+1)) \times nca_s} matrix
formed by row-stacking \eqn{\mathbf{G}_s^{1/2,(k)}
\mathbf{A}_s^{(k)}}.}

\item{cv_spline}{List of constraint right-hand-side vectors for
the spline subproblem (may be empty).}

\item{K}{Integer; number of interior knots.}

\item{nc_spline}{Integer; number of spline columns.}

\item{A_spline}{Spline-only constraint matrix.}
}
\value{
List of \eqn{K+1} coefficient vectors
  (\eqn{nc_s \times 1}).
}
\description{
Projects the adjusted cross-product \eqn{\mathbf{X}_s^{\top}
\tilde{\mathbf{y}}} through \eqn{\mathbf{G}_s^{1/2}} and
removes the component along \eqn{\mathbf{G}_s^{1/2}\mathbf{A}_s}
to enforce the smoothness constraints in the spline block.
When \code{cv_spline} contains nonzero values (e.g.\ from
\code{no_intercept}), the corresponding contribution is added
back after projection.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_Lambda.R
\name{.compute_tuning_residuals}
\alias{.compute_tuning_residuals}
\title{Compute Residuals During Penalty Tuning}
\usage{
.compute_tuning_residuals(
  y,
  preds,
  delta,
  family,
  observation_weights,
  K,
  order_list,
  ...
)
}
\arguments{
\item{y}{List; response vectors by partition.}

\item{preds}{List; prediction vectors by partition.}

\item{delta}{Numeric; pseudocount for link function stabilization.}

\item{family}{GLM family object.}

\item{observation_weights}{List; observation weights by partition.}

\item{K}{Integer; number of interior knots (partitions - 1).}

\item{order_list}{List; observation indices per partition.}

\item{...}{Additional arguments passed to \code{family$custom_dev.resids}.}
}
\value{
List of residual vectors, one per partition.
}
\description{
Computes the residuals used by the tuning criteria. Handles identity link,
general GLM link functions with pseudocount delta, custom deviance residual
functions, and observation weights.
}
\details{
Three computation paths:
\enumerate{
  \item \strong{Identity link or no custom deviance}: Standard
    \eqn{g(y) - \hat{\eta}} residuals on the link scale, optionally
    weighted by observation weights for non-Gaussian families.
    For Gaussian identity link with heterogeneous weights, the weights
    have already been absorbed into X and y prior to this call.
  \item \strong{Custom deviance residuals}: Delegates to
    \code{family$custom_dev.resids(y, mu, order_indices, family,
    observation_weights, ...)}.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{generate_posterior}
\alias{generate_posterior}
\title{Generate Posterior Samples from a Fitted lgspline}
\usage{
generate_posterior(
  object,
  new_sigmasq_tilde = object$sigmasq_tilde,
  new_predictors = NULL,
  theta_1 = 0,
  theta_2 = 0,
  posterior_predictive_draw = function(N, mean, sqrt_dispersion, ...) {
     rnorm(N,
    mean, sqrt_dispersion)
 },
  draw_dispersion = TRUE,
  include_posterior_predictive = FALSE,
  num_draws = 1,
  enforce_qp_constraints = TRUE,
  draw_correlation = FALSE,
  correlation_param_mean = NULL,
  correlation_param_vcov = NULL,
  correlation_VhalfInv_fxn = NULL,
  correlation_Vhalf_fxn = NULL,
  correlation_param_vcov_scale = NULL,
  include_warnings = TRUE,
  ...
)
}
\arguments{
\item{object}{A fitted \code{lgspline} model object.}

\item{new_sigmasq_tilde}{Numeric; dispersion \eqn{\tilde{\sigma}^2} used
as the point estimate when \code{draw_dispersion = FALSE}.
Default \code{object$sigmasq_tilde}.}

\item{new_predictors}{Matrix; predictor matrix for posterior predictive
sampling. Default uses in-sample predictors.}

\item{theta_1}{Numeric; shape increment for the inverse-gamma prior on
\eqn{\sigma^2}. Default 0.}

\item{theta_2}{Numeric; rate increment for the inverse-gamma prior.
Default 0.}

\item{posterior_predictive_draw}{Function; sampler for posterior predictive
realisations. Must accept \code{N}, \code{mean},
\code{sqrt_dispersion}, \code{...}. Default \code{rnorm}.}

\item{draw_dispersion}{Logical; sample \eqn{\sigma^2} from its posterior.
Default TRUE.}

\item{include_posterior_predictive}{Logical; generate posterior predictive
draws at \code{new_predictors}. Default FALSE.}

\item{num_draws}{Positive integer; number of draws. Default 1.}

\item{enforce_qp_constraints}{Logical; if TRUE, enforce active QP
inequality constraints during posterior sampling via the stored
elliptical-slice constrained sampler. Default TRUE.}

\item{draw_correlation}{Logical; propagate correlation parameter
uncertainty. Requires \code{VhalfInv_fxn} and
\code{VhalfInv_params_estimates} in the fitted object.
Default FALSE.}

\item{correlation_param_mean}{Numeric vector; mean of the approximate
normal posterior for correlation parameters on the unbounded
(working) scale. Default: \code{object$VhalfInv_params_estimates}.}

\item{correlation_param_vcov}{Matrix; variance-covariance for correlation
parameter draws. Default: inverse Hessian from BFGS
(\code{object$VhalfInv_params_vcov}).}

\item{correlation_VhalfInv_fxn}{Function; maps correlation parameter
vector to \eqn{\mathbf{V}^{-1/2}}. Default \code{object$VhalfInv_fxn}.}

\item{correlation_Vhalf_fxn}{Function or NULL; maps to
\eqn{\mathbf{V}^{1/2}}. Passed through to
\code{\link{generate_posterior_correlation}}; the current correlation-aware posterior path only requires \code{correlation_VhalfInv_fxn}.}

\item{correlation_param_vcov_scale}{NULL or numeric; if supplied,
divides a user-supplied \code{correlation_param_vcov} by this
value before passing it to
\code{\link{generate_posterior_correlation}}. When NULL,
no additional scaling is applied.}

\item{include_warnings}{Logical; emit warnings for degenerate draws,
constraint violations, etc. Default TRUE.}

\item{...}{Additional arguments forwarded to the GLM weight function,
dispersion function, and \code{posterior_predictive_draw}.}
}
\value{
When \code{num_draws = 1}, a named list:
\describe{
  \item{post_draw_coefficients}{List of length K+1; per-partition
        coefficient vectors on the original scale.}
  \item{post_draw_sigmasq}{Drawn (or fixed) dispersion.}
  \item{post_pred_draw}{Posterior predictive vector (only when
        \code{include_posterior_predictive = TRUE}).}
  \item{post_draw_correlation_params}{Drawn correlation parameters
        on the working scale (only when \code{draw_correlation = TRUE}).}
}
When \code{num_draws > 1}, each element becomes a list of length
\code{num_draws}, and \code{post_pred_draw} (if requested) is an
\eqn{N_{\mathrm{new}} \times M} matrix, where \eqn{M = \mathrm{num\_draws}}.
}
\description{
Draws from the posterior distribution of model coefficients, with optional
dispersion sampling, posterior predictive draws, and propagation of
uncertainty in estimated correlation parameters.
}
\details{
Uses a Laplace approximation centred at the MAP estimate for non-Gaussian
responses.


\strong{Dispersion posterior.}
When \code{draw_dispersion = TRUE}, \eqn{\sigma^2} is drawn from
\deqn{
  \sigma^2 \mid \mathbf{y} \sim
  \mathrm{InvGamma}(\alpha_1, \alpha_2),
}
where
\deqn{
  \alpha_1 = \theta_1 + \tfrac{1}{2}(N - s \cdot \mathrm{tr}(\mathbf{H})),
  \quad
  \alpha_2 = \theta_2 + \tfrac{1}{2}(N - s \cdot \mathrm{tr}(\mathbf{H}))
             \tilde{\sigma}^2,
}
\eqn{\mathbf{H} = \mathbf{XUGX}^\top} is the hat matrix, \eqn{s = 1}
when \code{unbias_dispersion = TRUE} (else \eqn{s = 0}), and
\eqn{\theta_1 = \theta_2 = 0} recovers an improper uniform prior.

\strong{Correlation parameter posterior.}
When \code{draw_correlation = TRUE} and the fitted model contains an
estimated correlation structure, each draw first samples
\eqn{\boldsymbol{\rho}} from
\deqn{
  \boldsymbol{\rho}^{(m)} \sim
  \mathcal{N}(\hat{\boldsymbol{\rho}}_{\mathrm{REML}},
              \mathbf{H}^{-1}_{\mathrm{BFGS}}),
}
rebuilds the posterior covariance under the drawn correlation
structure (reusing all pre-computed design matrices, constraints, and
penalty matrices) and then draws coefficients from the updated
posterior. Knot placement, partitioning, coefficient re-estimation,
and GCV tuning are skipped entirely.
Draws producing non-positive-definite correlation matrices are
rejected and redrawn (up to 50 attempts).

When \code{draw_correlation = FALSE} (default), correlation parameters
are fixed at their estimated values.

\strong{Inequality constraints.}
Active QP inequalities can be enforced during posterior sampling via
elliptical slice sampling, producing draws from the corresponding
truncated multivariate normal posterior on the coefficient scale.
The public \code{enforce_qp_constraints} argument is forwarded to the
stored sampler for both the standard and correlation-aware posterior
paths.
}
\examples{

\donttest{
set.seed(1234)
n_blocks <- 100; block_size <- 5; N <- n_blocks * block_size
rho_true <- 0.3
t <- seq(-5, 5, length.out = N)
true_mean <- sin(t)
errors <- Reduce("rbind",
  lapply(1:n_blocks, function(i) {
    sigma <- diag(block_size) + rho_true *
      (matrix(1, block_size, block_size) - diag(block_size))
    matsqrt(sigma) \%*\% rnorm(block_size)
  })
)
y <- true_mean + errors * 0.5

model_fit <- lgspline(t, y,
  K = 3,
  correlation_id = rep(1:n_blocks, each = block_size),
  correlation_structure = "exchangeable",
  include_warnings = FALSE
)

## Propagate correlation uncertainty across 50 draws
post <- generate_posterior(model_fit,
  draw_correlation = TRUE, num_draws = 50,
  include_warnings = FALSE
)

## Fixed correlation parameters for comparison
post_fixed <- generate_posterior(model_fit, num_draws = 50)

corr_draws <- unlist(post$post_draw_correlation_params)
rho_draws <- exp(-exp(corr_draws))
print(summary(rho_draws))
}

}
\seealso{
\code{\link{lgspline}},
\code{\link{generate_posterior_correlation}},
\code{\link{wald_univariate}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_Lambda.R
\name{tune_Lambda}
\alias{tune_Lambda}
\title{Tune Smoothing and Ridge Penalties via Leave-One-Out or Generalized Cross-Validation}
\usage{
tune_Lambda(
  y,
  X,
  X_gram,
  smoothing_spline_penalty,
  A,
  K,
  p_expansions,
  N_obs,
  opt,
  use_custom_bfgs,
  C,
  colnm_expansions,
  wiggle_penalty,
  flat_ridge_penalty,
  initial_wiggle,
  initial_flat,
  unique_penalty_per_predictor,
  unique_penalty_per_partition,
  penalty_vec,
  meta_penalty,
  family,
  unconstrained_fit_fxn,
  keep_weighted_Lambda,
  iterate,
  qp_score_function,
  quadprog,
  qp_Amat,
  qp_bvec,
  qp_meq,
  tol,
  sd_y,
  delta,
  tuning_criterion,
  gcv_gamma,
  constraint_value_vectors,
  parallel,
  parallel_eigen,
  parallel_trace,
  parallel_aga,
  parallel_matmult,
  parallel_unconstrained,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  shared_env,
  custom_penalty_mat,
  order_list,
  glm_weight_function,
  schur_correction_function,
  need_dispersion_for_estimation,
  dispersion_function,
  observation_weights,
  homogenous_weights,
  blockfit,
  just_linear_without_interactions,
  Vhalf,
  VhalfInv,
  verbose,
  include_warnings,
  ...
)
}
\arguments{
\item{y}{List; response vectors by partition.}

\item{X}{List; design matrices by partition.}

\item{X_gram}{List; Gram matrices by partition.}

\item{smoothing_spline_penalty}{Matrix; integrated squared second derivative
penalty.}

\item{A}{Matrix; smoothness constraints at knots.}

\item{K}{Integer; number of interior knots in 1-D, number of partitions - 1
in higher dimensions.}

\item{p_expansions}{Integer; columns per partition.}

\item{N_obs}{Integer; total sample size.}

\item{opt}{Logical; TRUE to optimize penalties, FALSE to use initial values.}

\item{use_custom_bfgs}{Logical; TRUE for analytic gradient BFGS as natively
implemented, FALSE for finite differences as implemented by
\code{stats::optim()}. The analytic path is typically faster; however, for
exact LOO tuning the leverage derivative can be numerically delicate in
some problems, in which case \code{use_custom_bfgs = FALSE} remains a
conservative fallback.}

\item{C}{Matrix; polynomial expansion matrix (used for initialization).
This is the monomial expansion design used to derive starting values and
is not the inequality-constraint matrix sometimes denoted by
\eqn{\mathbf{C}} elsewhere in the package documentation.}

\item{colnm_expansions}{Character vector; column names of the expansion
matrix.}

\item{wiggle_penalty, flat_ridge_penalty}{Fixed penalty values if provided.}

\item{initial_wiggle, initial_flat}{Numeric vectors; candidate values for grid
search initialization on the raw (non-negative) scale. Converted to log
scale internally for optimization.}

\item{unique_penalty_per_predictor, unique_penalty_per_partition}{Logical;
allow predictor/partition-specific penalties.}

\item{penalty_vec}{Numeric vector; initial values for predictor/partition
penalties on the raw (non-negative) scale. Converted to log scale
internally for optimization. Use \code{c()} when no per-predictor or
per-partition penalties are needed.}

\item{meta_penalty}{The "meta" ridge penalty, a regularization for
predictor/partition penalties to pull them towards 1 on the raw scale.}

\item{family}{GLM family with optional custom tuning loss.}

\item{unconstrained_fit_fxn}{Function for unconstrained fitting.}

\item{keep_weighted_Lambda, iterate}{Logical controlling GLM fitting.}

\item{qp_score_function, quadprog, qp_Amat, qp_bvec, qp_meq}{Quadratic
programming parameters (see arguments of \code{\link[lgspline]{lgspline}}).}

\item{tol}{Numeric; convergence tolerance.}

\item{sd_y, delta}{Response standardization parameters.}

\item{tuning_criterion}{Character scalar selecting the tuning objective:
\code{"loo"} for exact leave-one-out on the transformed tuning problem or
\code{"gcv"} for generalized cross-validation. The LOO criterion itself is
still evaluated exactly on that transformed problem; the caveat is that
its fully analytic leverage derivative can be numerically sensitive in some
datasets. For very large samples, \code{"gcv"} is often the more practical
choice; as a rough guideline, it is recommended once the sample size is
above about 250,000.}

\item{gcv_gamma}{Numeric scalar, at least 1, used only when
\code{tuning_criterion = "gcv"}. Multiplies the effective degrees of
freedom in the GCV denominator.}

\item{constraint_value_vectors}{List; constraint values.}

\item{parallel}{Logical; enable parallel computation.}

\item{parallel_eigen, parallel_trace, parallel_aga}{Logical; specific parallel
flags.}

\item{parallel_matmult, parallel_unconstrained}{Logical; specific parallel
flags.}

\item{cl}{Parallel cluster object.}

\item{chunk_size, num_chunks, rem_chunks}{Integer; parallel computation
parameters.}

\item{shared_env}{Environment; shared variables exported to cluster workers.}

\item{custom_penalty_mat}{Optional custom penalty matrix.}

\item{order_list}{List; observation ordering by partition.}

\item{glm_weight_function, schur_correction_function}{Functions for GLM
weights and corrections.}

\item{need_dispersion_for_estimation, dispersion_function}{Control dispersion
estimation.}

\item{observation_weights}{Optional observation weights.}

\item{homogenous_weights}{Logical; TRUE if all weights equal.}

\item{blockfit}{Logical; when TRUE, the backfitting block decomposition
(\code{blockfit_solve}) is used in place of \code{get_B} for the
coefficient estimation step at each tuning-objective evaluation, provided
\code{just_linear_without_interactions} is non-empty and \code{K > 0}.
The dispatch condition is identical to that in \code{lgspline.fit} so
that penalties are tuned under the same estimator that will be used for
the final fit. Falls back to \code{get_B} automatically on failure.}

\item{just_linear_without_interactions}{Numeric; vector of columns for
non-spline effects without interactions.}

\item{Vhalf, VhalfInv}{Square root and inverse square root correlation
structure matrices. These are passed through to the coefficient-estimation
step used inside each tuning-objective evaluation, so any dense or
Woodbury-accelerated
correlated solve is the same one used for the final fitted model.}

\item{verbose}{Logical; print progress.}

\item{include_warnings}{Logical; print warnings/try-errors.}

\item{...}{Additional arguments passed to fitting functions.}
}
\value{
List containing:
\describe{
  \item{Lambda}{Final combined penalty matrix.}
  \item{flat_ridge_penalty}{Optimized ridge penalty.}
  \item{wiggle_penalty}{Optimized smoothing penalty.}
  \item{other_penalties}{Optimized predictor/partition penalties.}
  \item{L_predictor_list}{Predictor-specific penalty matrices.}
  \item{L_partition_list}{Partition-specific penalty matrices.}
}
}
\description{
Optimizes smoothing spline and ridge regression penalties by minimizing the
selected tuning criterion, either exact leave-one-out on the transformed
tuning problem or modified GCV. Uses BFGS optimization with analytical
gradients or finite differences. This is the top-level entry point that
orchestrates grid search initialization and quasi-Newton optimization via
internal subfunctions.
}
\details{
The tuning procedure consists of the following steps:
\enumerate{
  \item \strong{Preprocessing}: Convert raw-scale penalties to log scale,
    compute cross-products, determine pseudocount delta, ensure constraint
    matrix compatibility.
  \item \strong{Blockfit dispatch}: Pre-compute \code{flat_cols} and the
    \code{use_blockfit} flag so that every tuning-objective evaluation uses the same
    coefficient estimator as the final fit. \code{flat_cols} are identified
    by matching column names against
    \code{paste0("_", just_linear_without_interactions, "_")} in
    \code{colnm_expansions}.
  \item \strong{Grid search}: Evaluate the selected tuning criterion over a grid of
    (wiggle, ridge) penalty candidates to find a good starting point
    (see \code{.tune_grid_search}).
  \item \strong{BFGS optimization}: Minimize the selected tuning criterion via either the custom
    damped BFGS with closed-form gradients (see \code{.damped_bfgs},
    \code{.compute_gcvu_gradient}, \code{.compute_loocv_gradient}) or base R's \code{stats::optim} with
    finite-difference gradients. The GCV gradient is differentiated directly
    in the shared wiggle and flat directions; for LOO, the same direct
    differentiation is used, but the leverage derivative may be numerically
    delicate in some problems and can motivate the finite-difference
    fallback.
  \item \strong{Criterion choice}: \code{"loo"} uses exact per-observation
    leverages on the transformed tuning problem, while \code{"gcv"} uses the
    existing trace-based denominator with optional \code{gcv_gamma}
    inflation.
  \item \strong{Sample-size adjustment}: After optimization, divide the tuned
    penalties by \eqn{((N+2)/(N-2))^{2}} (equivalently multiply by
    \eqn{((N-2)/(N+2))^{2}}) for both GCV- and LOO-based tuning so that the
    final penalties are decreased at small sample sizes.
  \item \strong{Final Lambda}: Compute the final penalty matrix from
    optimized parameters via \code{compute_Lambda}.
}

Parameterization: initial penalty values are accepted on the raw
(non-negative) scale and converted to natural log-scale internally,
i.e. raw_penalty = exp(theta), so that raw penalties are always positive.
The chain rule factor d(exp(theta))/d(theta) = exp(theta) = raw_penalty.

For predictor- and partition-specific penalties, the current implementation
still uses the existing low-cost ratio approximation rather than a full
direct derivative in each extra penalty direction. Thus the most accurate
closed-form tuning gradients are presently the shared wiggle and flat
directions.

The resulting penalty matrix follows the same decomposition used throughout
the paper and package documentation:
\deqn{
  \boldsymbol{\Lambda}
  =
  \lambda_{w}\mathbf{L}_{1}
  +
  \lambda_{r}\mathbf{L}_{2}
  +
  \sum_{j}\nu_{j}\mathbf{L}^{(\mathrm{pred})}_{j}
  +
  \sum_{k}\tau_{k}\mathbf{L}^{(\mathrm{part})}_{k},
}
where the predictor- and partition-specific sums are included only when the
corresponding options are active. Internally these components are returned as
\code{L1}, \code{L2}, \code{L_predictor_list}, and \code{L_partition_list}.

If correlation structure inputs are supplied, each tuning-objective evaluation calls the
same constrained correlated solver used by the final model fit. In the
structured-correlation case, the low-rank Woodbury correction described in
\code{lgspline-details} is therefore inherited automatically through
\code{get_B}; no separate tuning-specific notation is introduced here.
}
\examples{
\dontrun{
## ## Example 1: Basic usage within lgspline ## ## ## ## ## ## ## ## ## ## ##
## tune_Lambda is called internally by lgspline; direct calls are
#  for advanced users. Here we verify that the refactored version
#  produces identical output to the original.

set.seed(42)
t <- runif(200, -5, 5)
y <- sin(t) + rnorm(200, 0, 0.5)

## Fit with automatic penalty tuning (calls tune_Lambda internally)
fit1 <- lgspline(t, y, K = 3)
cat("Wiggle penalty:", fit1$penalties$wiggle_penalty, "\n")
cat("Ridge penalty:", fit1$penalties$flat_ridge_penalty, "\n")
cat("Trace (edf):", fit1$trace_XUGX, "\n")

## ## Example 2: Fixed penalties (no tuning) ## ## ## ## ## ## ## ## ## ## ##
fit2 <- lgspline(t, y, K = 3, opt = FALSE,
                 wiggle_penalty = 1e-4,
                 flat_ridge_penalty = 0.1)
cat("Fixed wiggle:", fit2$penalties$wiggle_penalty, "\n")

## ## Example 3: blockfit path #### ## ## ## ## ## ## ## ## ## ## ## ## ## ##
## When blockfit = TRUE and just_linear_without_interactions is non-empty,
#  tune_Lambda dispatches to blockfit_solve at each GCV evaluation,
#  ensuring penalties are tuned under the same estimator used in the final
#  fit. Verify that tuned penalties are consistent across both paths.

set.seed(7)
n  <- 300
x1 <- runif(n, 0, 5)
x2 <- rnorm(n)
y2 <- sin(x1) + 0.5 * x2 + rnorm(n, 0, 0.3)
df <- data.frame(x1 = x1, x2 = x2)

## blockfit = TRUE uses blockfit_solve during tuning
fit_bf  <- lgspline(df, y2, K = 2, blockfit = TRUE,
                    just_linear_without_interactions = 2)
## blockfit = FALSE uses get_B during tuning (original path)
fit_std <- lgspline(df, y2, K = 2, blockfit = FALSE,
                    just_linear_without_interactions = 2)

cat("blockfit wiggle  :", fit_bf$penalties$wiggle_penalty, "\n")
cat("standard wiggle  :", fit_std$penalties$wiggle_penalty, "\n")
## Penalties may differ slightly; predictions should be close.
cat("Max pred diff:", max(abs(fit_bf$ytilde - fit_std$ytilde)), "\n")

## ## Example 4: Verify refactored subfunctions ## ## ## ## ## ## ## ## ## ##
## The internal .compute_meta_penalty should match hand calculation
mp <- lgspline:::.compute_meta_penalty(
  wiggle_penalty = 0.5,
  penalty_vec = c(predictor1 = 1.2, partition1 = 0.8),
  meta_penalty_coef = 1e-8,
  unique_penalty_per_predictor = TRUE,
  unique_penalty_per_partition = TRUE
)
expected <- 0.5 * 1e-8 * ((1.2 - 1)^2 + (0.8 - 1)^2) +
            0.5 * 1e-32 * (0.5 - 1)^2
stopifnot(abs(mp - expected) < 1e-20)
cat("Meta-penalty check passed.\n")

## ## Example 5: Verify gradient of meta-penalty ## ## ## ## ## ## ## ## ## #
gr <- lgspline:::.compute_meta_penalty_gradient(
  wiggle_penalty = 2.0,
  penalty_vec = c(predictor1 = 1.5),
  meta_penalty_coef = 1e-8,
  unique_penalty_per_predictor = TRUE,
  unique_penalty_per_partition = FALSE
)
# gr[1] should be 1e-32 * (2 - 1) * 2 = 2e-32
# gr[2] should be 0
# gr[3] should be 1e-8 * (1.5 - 1) * 1.5 = 7.5e-9
stopifnot(abs(gr[1] - 2e-32) < 1e-40)
stopifnot(gr[2] == 0)
stopifnot(abs(gr[3] - 7.5e-9) < 1e-17)
cat("Meta-penalty gradient check passed.\n")

## ## Example 6: Residual computation paths ## ##
## Identity link ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
r1 <- lgspline:::.compute_tuning_residuals(
  y = list(c(1, 2, 3)),
  preds = list(c(1.1, 1.9, 3.2)),
  delta = 0,
  family = gaussian(),
  observation_weights = list(NULL),
  K = 0,
  order_list = list(1:3)
)
stopifnot(max(abs(r1[[1]] - c(-0.1, 0.1, -0.2))) < 1e-10)
cat("Residual check passed.\n")
}

}
\seealso{
\itemize{
  \item \code{\link[stats]{optim}} for Hessian-free optimization
  \item \code{\link{compute_Lambda}} for penalty matrix construction
  \item \code{\link{compute_G_eigen}} for eigendecomposition of penalized
    Gram matrices
  \item \code{\link{get_B}} for constrained coefficient estimation
  \item \code{\link{blockfit_solve}} for the backfitting block-decomposition
    estimator used when \code{blockfit = TRUE}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/negbin_helpers.R
\name{negbin_helpers}
\alias{negbin_helpers}
\title{Negative Binomial Regression Helpers for lgspline}
\description{
Functions for fitting negative binomial (NB2) regression models within the
lgspline framework. Analogous to the Weibull AFT and Cox PH helpers, these
provide the log-likelihood, score, information, and all interface functions
needed by lgspline's unconstrained fitting, penalty tuning, and inference
machinery.
}
\details{
The parameterization follows NB2: \eqn{Y \sim \mathrm{NB}(\mu, \theta)}
with \eqn{\mathrm{Var}(Y) = \mu + \mu^2/\theta}, where \eqn{\theta > 0}
is the shape (size) parameter. The canonical link is log:
\eqn{\eta = \log\mu}. The dispersion stored in
\code{lgspline$sigmasq_tilde} is \eqn{\theta} itself, not
\eqn{1/\theta}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lgspline-details.R
\name{Details}
\alias{Details}
\title{Lagrangian Multiplier Smoothing Splines: Mathematical Details}
\description{
This document provides the mathematical and implementation details for
Lagrangian Multiplier Smoothing Splines as implemented in \pkg{lgspline}.

This package provides exhaustive resources for fitting multivariate smoothing
smoothing splines with a monomial basis and analytical form cubic smoothing
spline penalty.

The material is presented such that a programmer or statistician of
reasonable experience and background can understand and implement the
procedure from scratch, and also potentially critique some of the modelling
choices that went into designing this package.

Informally, \pkg{lgspline} answers the following question:
How can we best adapt a useful functionality of basis splines, under the
alternative interpretation of smoothing as explicit external constraints instead?

The obvious benefit is a much more flexible and interpretable final model that for
non-experienced users is simply easier to understand without post-hoc processing, and
for experienced users can be used to customize models more easily.

The drawback is that the interpretation of constraints as external adds a new layer of complexity
to each step of the model fitting process, whereas for implicit design matrix
construction these complications are bypassed.

While it is true a B-spline can always be converted back into monomial form,
tensor-product splines that generalize this to multiple dimensions often
explodes the number and degree of interaction terms, the conversion may not
be computationally stable, and it is not available in standard software.
}
\section{Statistical Problem Formulation}{


Consider an \eqn{N \times q} matrix of predictors
\eqn{\mathbf{T} = (\mathbf{t}_1, \dots, \mathbf{t}_N)^{\top}} and an
\eqn{N \times 1} response vector \eqn{\mathbf{y} = (y_1, \dots, y_N)^{\top}}.
We assume the relationship follows a generalized linear model with unknown
smooth function \eqn{f}:
\deqn{y_i \sim \mathcal{D}(g^{-1}(f(\mathbf{t}_i)),\, \sigma^2)}
where \eqn{\mathcal{D}} is a distribution (e.g. exponential family or related) with mean
\eqn{\mu_i = g^{-1}(f(\mathbf{t}_i))}, link function \eqn{g(\cdot)}, and
dispersion parameter \eqn{\sigma^2}. For Gaussian response with identity
link, observations are independently distributed as
\eqn{y_i \mid \mathbf{t}_i, \sigma^2 \sim \mathcal{N}(f(\mathbf{t}_i), \sigma^2)}.

The objective is to estimate \eqn{f} by:
\enumerate{
  \item Partitioning the predictor space into \eqn{K+1} mutually exclusive regions.
  \item Fitting local polynomial models within each partition.
  \item Enforcing smoothness at partition boundaries via Lagrangian multipliers.
  \item Penalizing the integrated squared second derivative to discourage roughness.
}

Unlike other smoothing spline formulations, no post-fitting algebraic
rearrangement or disentanglement of a spline basis is needed to obtain
interpretable models. The polynomial expansions are homogeneous across
partitions, and the relationship between predictor and response is explicit
at the coefficient level.

To anchor the notation, in the single-predictor cubic case one would write
\deqn{\hat{f}(t_i) = \hat{\beta}_{(0)} + \hat{\beta}_{(1)}t_i +
  \hat{\beta}_{(2)}t_i^2 + \hat{\beta}_{(3)}t_i^3 =
  \mathbf{x}_i^{\top}\hat{\boldsymbol{\beta}},}
where \eqn{\mathbf{x}_i = (1, t_i, t_i^2, t_i^3)^{\top}}. The LMSS
formulation preserves exactly this kind of polynomial representation, but now
does so within each partition and then forces neighboring pieces to agree in
the smoothness conditions described below.

Core notation used throughout:
\itemize{
  \item \eqn{\mathbf{y}_{(N \times 1)}}: Response vector.
  \item \eqn{\mathbf{T}_{(N \times q)}}: Matrix of predictors.
  \item \eqn{\mathbf{X}_{(N \times P)}}: Block-diagonal matrix of polynomial
    expansions, with diagonal blocks \eqn{\mathbf{X}_k} of dimension
    \eqn{n_k \times p}.
  \item \eqn{\boldsymbol{\Lambda}_{(P \times P)}}: Block-diagonal penalty matrix,
    with blocks \eqn{\boldsymbol{\Lambda}_k} of dimension \eqn{p \times p}.
  \item \eqn{\hat{\boldsymbol{\beta}}_{(P \times 1)}}: Unconstrained penalized
    estimate.
  \item \eqn{\tilde{\boldsymbol{\beta}}_{(P \times 1)}}: Constrained coefficient
    estimates.
  \item \eqn{\mathbf{G}_{(P \times P)}}: Block-diagonal matrix with blocks
    \eqn{\mathbf{G}_k = (\mathbf{X}_k^{\top}\mathbf{W}_k\mathbf{D}_k\mathbf{X}_k + \boldsymbol{\Lambda}_k)^{-1}},
    where \eqn{\mathbf{W}_k} and \eqn{\mathbf{D}_k} are defined below.
  \item \eqn{\mathbf{A}_{(P \times r)}}: Constraint matrix encoding smoothness
    conditions. Reduced to linearly independent columns via pivoted QR
    decomposition.
  \item \eqn{\mathbf{U}_{(P \times P)}}:
    \eqn{\mathbf{I} - \mathbf{G}\mathbf{A}(\mathbf{A}^{\top}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{\top}}.
  \item \eqn{\mathbf{D}_{(N \times N)}}: Diagonal matrix of user-supplied
    observation weights (\code{observation_weights} or \code{weights}).
    Defaults to the identity. These play the role of prior precision on
    individual observations: a weight of 2 is equivalent to seeing that
    observation twice.
  \item \eqn{\mathbf{W}_{(N \times N)}}: Diagonal matrix of GLM working
    weights. In the implementation these diagonal entries are whatever is
    returned by \code{glm_weight_function}; by default this is
    \code{family$variance(mu)}, optionally multiplied by user-supplied
    observation weights. For Gaussian response with identity link,
    \eqn{\mathbf{W} = \mathbf{I}}. For other families,
    \eqn{\mathbf{W}} depends on the current fitted values and is updated at
    each Newton-Raphson iteration. For the common canonical families used
    by default, this matches the familiar Fisher-scoring weighting role.
  \item \eqn{\mathbf{V}_{(N \times N)}}: Correlation matrix of errors.
    When no correlation structure is specified, \eqn{\mathbf{V} = \mathbf{I}}.
    Otherwise supplied via \code{VhalfInv} or estimated through
    \code{VhalfInv_fxn}.
}

In the Gaussian identity case with unit weights and no correlation,
\eqn{\mathbf{G}_k = (\mathbf{X}_k^{\top}\mathbf{X}_k + \boldsymbol{\Lambda}_k)^{-1}}
and most formulas simplify accordingly. When \eqn{\mathbf{D}} or
\eqn{\mathbf{W}} appear in a formula, the product \eqn{\mathbf{W}\mathbf{D}}
means ``GLM working weights times observation weights''; whenever one of
them is the identity it drops out.

Before these quantities reach the main fitting stage, the user-facing inputs
are parsed, standardized, and organized by \code{\link{process_input}}. When
the formula interface is used and \code{auto_encode_factors = TRUE}, that
preprocessing step also relies on helpers such as \code{\link{create_onehot}}
to encode factor levels before the design reaches \code{\link{lgspline.fit}()}.
The notation in the remainder of this document therefore refers to the internal
objects that actually enter \code{\link{lgspline.fit}()}, not necessarily the raw
objects originally supplied by the user.

From the user side, many of the arguments that control these internal objects
can be supplied either individually or through the grouped lists
\code{penalty_args}, \code{tuning_args}, \code{expansion_args},
\code{constraint_args}, \code{qp_args}, \code{parallel_args},
\code{covariance_args}, \code{return_args}, and \code{glm_args}, as
documented in \code{\link{lgspline}}. These grouped lists are unpacked before
dispatch into the same fitting pipeline, so they are a convenience layer
rather than a separate modeling abstraction. A closely related exploratory
mode is \code{dummy_fit = TRUE} in \code{\link{lgspline}} or
\code{\link{lgspline.fit}}, which runs the preprocessing, partition
construction, expansion building, and penalty setup without solving for
nonzero coefficients, making it a practical way to inspect objects such as
\code{X}, \code{A}, the returned \code{make_partition_list} from
\code{\link{make_partitions}}, and the assembled \code{penalties} from
\code{\link{compute_Lambda}} before a full fit.
}

\section{Model Formulation and Estimation}{


\subsection{Piecewise Polynomial Structure}{
For \eqn{K} knots (one predictor) or \eqn{K+1} partitions (multiple
predictors) there are \eqn{K+1} mutually exclusive partitions
\eqn{\mathcal{P}_0, \dots, \mathcal{P}_{K}}. Each observation \eqn{i}
belongs to exactly one partition. Within partition \eqn{k}, the function is
represented as a polynomial of degree \eqn{p-1} in each predictor:
\deqn{\hat{f}_k(\mathbf{t}) = \mathbf{x}^{\top}\tilde{\boldsymbol{\beta}}_k}
where \eqn{\mathbf{x}} collects the polynomial basis terms (intercept,
linear, quadratic, cubic, and optionally quartic and interaction terms) and
\eqn{\tilde{\boldsymbol{\beta}}_k} are the corresponding coefficients. In
one predictor, the same idea can be written more explicitly as
\deqn{\hat{f}(t_i) = \sum_{k=0}^{K}
  \mathbf{x}_{ik}^{\top}\hat{\boldsymbol{\beta}}_k
  \mathbf{1}(t_i \in \mathcal{P}_k),}
which highlights that the unconstrained problem is just a collection of
local polynomial regressions. The expansions are homogeneous across
partitions, so coefficients are directly comparable. This is implemented via
\code{\link{get_polynomial_expansions}}.

The exact contents of \eqn{\mathbf{x}} are controlled by the basis-expansion
arguments documented in \code{\link{lgspline}}: \code{include_quadratic_terms},
\code{include_cubic_terms}, \code{include_quartic_terms},
\code{include_2way_interactions}, \code{include_3way_interactions},
\code{include_quadratic_interactions}, \code{exclude_interactions_for},
\code{exclude_these_expansions}, and \code{custom_basis_fxn}. Likewise,
\code{just_linear_with_interactions} and
\code{just_linear_without_interactions} determine which predictors remain
structurally linear even though they still participate in the same
partition-wise polynomial bookkeeping described here.

Letting \eqn{p} denote the number of basis terms per partition,
\eqn{P = p(K+1)} is the total number of coefficients. The full design matrix
\eqn{\mathbf{X}} and penalty matrix \eqn{\boldsymbol{\Lambda}} are
block-diagonal with \eqn{K+1} blocks, so unconstrained estimation reduces to
\eqn{K+1} independent penalized regressions, which appears as follows for the identity link case:
\deqn{\hat{\boldsymbol{\beta}}_k = \mathbf{G}_k \mathbf{X}_k^{\top} \mathbf{W}_k\mathbf{D}_k\mathbf{y}_k, \quad
  \mathbf{G}_k = (\mathbf{X}_k^{\top}\mathbf{W}_k\mathbf{D}_k\mathbf{X}_k + \boldsymbol{\Lambda}_k)^{-1}.}
For Gaussian identity with unit weights this reduces to the familiar
\eqn{\mathbf{G}_k = (\mathbf{X}_k^{\top}\mathbf{X}_k + \boldsymbol{\Lambda}_k)^{-1}}.
The block-diagonal structure means these can be computed in parallel across
partitions. In the user-facing interface this is realized by supplying a
cluster through \code{cl}, optionally controlling work splitting with
\code{chunk_size}, and enabling stages such as \code{parallel_eigen}
for the eigendecompositions and, in non-Gaussian Path 3, \code{parallel_unconstrained}
for the partition-wise unconstrained fits; nearby stages can likewise use
\code{parallel_penalty} and \code{parallel_make_constraint}. The eigenvalue
decomposition and matrix square roots of each \eqn{\mathbf{G}_k} are
computed by \code{\link{compute_G_eigen}}, and can be returned in the fitted
object as \code{G} and \code{Ghalf} when \code{return_G = TRUE} and
\code{return_Ghalf = TRUE}.

Fitted values for the canonical Gaussian case appear as
\eqn{\tilde{\mathbf{y}} = \mathbf{X}\tilde{\boldsymbol{\beta}} = \mathbf{H}\mathbf{y}}
for \eqn{\mathbf{H} = \mathbf{X}\mathbf{U}\mathbf{G}\mathbf{X}^{\top}}.
}

\subsection{Smoothing Constraints and the Constraint Matrix}{
Without further intervention the piecewise polynomial will be discontinuous.
The central idea of LMSS is that smoothness is not hidden inside a special
basis, but instead imposed directly where neighboring partitions meet.
At each knot \eqn{t_{k,k+1}} between neighboring partitions \eqn{k} and
\eqn{k+1}, up to three smoothing constraints are imposed:
\enumerate{
  \item Continuity: \eqn{\mathbf{x}_{k,k+1}^{\top}\boldsymbol{\beta}_k = \mathbf{x}_{k,k+1}^{\top}\boldsymbol{\beta}_{k+1}}.
  \item First-derivative continuity: \eqn{\mathbf{x}_{k,k+1}^{\prime\top}\boldsymbol{\beta}_k = \mathbf{x}_{k,k+1}^{\prime\top}\boldsymbol{\beta}_{k+1}}.
  \item Second-derivative continuity: \eqn{\mathbf{x}_{k,k+1}^{\prime\prime\top}\boldsymbol{\beta}_k = \mathbf{x}_{k,k+1}^{\prime\prime\top}\boldsymbol{\beta}_{k+1}}.
}
where \eqn{\mathbf{x}^{\prime}} and \eqn{\mathbf{x}^{\prime\prime}} are
elementwise first and second derivatives of the basis with respect to
\eqn{\mathbf{t}}. For the familiar cubic single-predictor basis
\eqn{\mathbf{x} = (1, t, t^2, t^3)^{\top}}, these derivative vectors are
\deqn{\mathbf{x}' = (0, 1, 2t, 3t^2)^{\top}, \qquad
  \mathbf{x}'' = (0, 0, 2, 6t)^{\top}.}
With \eqn{K} knots this yields up to \eqn{3K} scalar
constraints (for a single predictor; more for multiple predictors with
interactions), collected as linear equations
\eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{0}}
in a \eqn{P \times r} matrix \eqn{\mathbf{A}}. The constraint matrix is
built by \code{\link{make_constraint_matrix}} and returned in the fitted
object as \code{A}.

In higher dimensions or with many partitions, the constraints can become
over-specified and force the model toward a single global polynomial. In these
cases it is recommended to drop second-derivative constraints or include quartic
terms, allowing the model to fit a richer surface while maintaining
perceived smoothness at knots. The appropriate constraint level can be
controlled via \code{include_constrain_fitted},
\code{include_constrain_first_deriv}, and
\code{include_constrain_second_deriv}.
The companion flag \code{include_constrain_interactions} determines whether
the analogous mixed-partial constraints are imposed for interaction terms,
and \code{no_intercept} adds the special homogeneous equality constraint that
fixes the intercept at zero (the same behavior triggered by using
\code{0 +} in the formula interface).

Before computing the projection \eqn{\mathbf{U}}, the constraint matrix is
reduced to a linearly independent subset of columns via pivoted QR
decomposition. This avoids numerical instability from redundant constraints
and ensures \eqn{\mathbf{A}^{\top}\mathbf{G}\mathbf{A}} is invertible.
}

\subsection{Lagrangian Projection}{
The constrained estimate is derived via Lagrangian multipliers. Define the
\eqn{P \times P} projection matrix:
\deqn{\mathbf{U} = \mathbf{I} - \mathbf{G}\mathbf{A}(\mathbf{A}^{\top}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{\top}.}
Then the constrained estimate is:
\deqn{\tilde{\boldsymbol{\beta}} = \mathbf{U}\hat{\boldsymbol{\beta}}.}
The matrix \eqn{\mathbf{U}} has the property that
\eqn{\mathbf{U}\mathbf{G}\mathbf{U}^{\top} = \mathbf{U}\mathbf{G}}, which is
used extensively in variance estimation and posterior draws. In words, the
unconstrained penalized estimate is projected back into the coefficient space
that satisfies the smoothness restrictions, and all subsequent uncertainty
calculations inherit that same projected geometry.
The projection is computed via \code{\link{get_U}} and, when requested,
returned in the fitted object as \code{U} through \code{return_U = TRUE}.

When the constraints are inhomogeneous
(\eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{c}} with
\eqn{\mathbf{c} \neq \mathbf{0}}), a particular solution
\eqn{\boldsymbol{\beta}_0} satisfying
\eqn{\mathbf{A}^{\top}\boldsymbol{\beta}_0 = \mathbf{c}} is added back
after projection, yielding the full Lagrangian solution
\eqn{\mathbf{U}\hat{\boldsymbol{\beta}} + (\mathbf{I} - \mathbf{U})\boldsymbol{\beta}_0}.
In \code{\link{lgspline}} and \code{\link{lgspline.fit}}, users realize
this by supplying extra equality columns in \code{constraint_vectors}
together with matching right-hand sides in \code{constraint_values};
\code{null_constraint} provides the alternate shorthand documented in
\code{\link{lgspline}} when \code{constraint_vectors} is supplied and
\code{constraint_values} is left empty.

In practice \eqn{\mathbf{U}} is never explicitly formed during fitting.
The constrained estimate is obtained from a transformed OLS residual problem (the
\eqn{\mathbf{G}^{1/2}\mathbf{r}^{*}} trick) in four steps:
\enumerate{
  \item Obtain the unconstrained partition-wise
    unconstrained estimate \eqn{\hat{\boldsymbol{\beta}}}.
  \item Set \eqn{\mathbf{y}^{*} = \mathbf{G}^{-1/2}\hat{\boldsymbol{\beta}}} and
    \eqn{\mathbf{X}^{*} = \mathbf{G}^{1/2}\mathbf{A}}.
  \item Fit the linear model \eqn{\mathbb{E}[\mathbf{y}^{*}] = \mathbf{X}^{*}\boldsymbol{\gamma}}
    by OLS using QR decomposition.
  \item Compute the residuals \eqn{\mathbf{r}^{*} = \mathbf{y}^{*} - \mathbf{X}^{*}(\mathbf{X}^{*\top}\mathbf{X}^{*})^{-1}\mathbf{X}^{*\top}\mathbf{y}^{*}} from that transformed OLS
    fit and recover the constrained estimate by
    \eqn{\tilde{\boldsymbol{\beta}} = \mathbf{G}^{1/2}\mathbf{r}^{*}}.
}

A scaling factor \eqn{1/\sqrt{K+1}} is applied to both
\eqn{\mathbf{X}^{*}} and \eqn{\mathbf{y}^{*}} prior to the OLS call
and divided out afterward, improving numerical conditioning when the
constraint matrix has many rows.

The most expensive operation in this approach is the QR decomposition of the
\eqn{P \times r} matrix \eqn{\mathbf{X}^{*} = \mathbf{G}^{1/2}\mathbf{A}},
which is far cheaper than working with the full \eqn{P \times P} system directly.
Without correlation or SQP constraints, \eqn{\mathbf{G}} is stored and operated upon as a
list of \eqn{K+1} small \eqn{p \times p} matrices rather than the full
\eqn{P \times P} block-diagonal, saving substantial memory when \eqn{K}
is large and allowing for parallelism.

When correlation is present, \eqn{\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X}}
is no longer block-diagonal, so the full-dimensional system must be handled
directly unless the Woodbury acceleration
(see \code{\link{.woodbury_decompose_V}}) applies.
When additional inequality constraints are present, the code either augments the equality system with a
partition-wise active-set refinement (block-separable case) or falls back to
dense SQP via the Goldfarb-Idnani dual active-set method implemented in
\code{\link[quadprog]{solve.QP}}.
}
}

\section{GLM Extension and Iterative Updates}{


\subsection{Working Quantities}{
For GLM responses with mean \eqn{\mu_i = g^{-1}(\eta_i)} and working
weights \eqn{w_i = [V(\mu_i)\{g'(\mu_i)\}^2]^{-1}}, the penalized
information becomes
\eqn{\mathbf{G}_k^{-1} = \mathbf{X}_k^{\top}\mathbf{W}_k\mathbf{X}_k
+ \boldsymbol{\Lambda}_k} with
\eqn{\mathbf{W}_k = \mathrm{diag}(w_i : i \in \mathcal{P}_k)}.
Because \eqn{\mathbf{W}} is diagonal,
\eqn{\mathbf{X}^{\top}\mathbf{W}\mathbf{X}} remains block-diagonal and
the four-step procedure carries through with
\eqn{\mathbf{G}_k = (\mathbf{X}_k^{\top}\mathbf{W}_k\mathbf{X}_k +
\boldsymbol{\Lambda}_k)^{-1}} in place of
\eqn{(\mathbf{X}_k^{\top}\mathbf{X}_k + \boldsymbol{\Lambda}_k)^{-1}}.
Under the default \code{glm_weight_function}, the diagonal entries reduce
to \code{family$variance(mu)} (optionally scaled by observation weights),
which for canonical families is the usual Fisher scoring weight.

The partition-wise unconstrained estimates are obtained by
\code{unconstrained_fit_fxn}, by default
\code{\link{unconstrained_fit_default}}, which initializes via an augmented
ridge trick (appending \eqn{\boldsymbol{\Lambda}^{1/2}} as
pseudo-observations to \code{glm.fit}) then refines using
\code{\link{damped_newton_r}} with \code{\link{nr_iterate}}. For
non-canonical log-link gamma regression, the \code{keep_weighted_Lambda = TRUE}
option correctly returns the augmented ridge estimate directly.
}

\subsection{Three Fitting Paths}{

\strong{Path 1: Correlation structure present.}
When a working correlation \eqn{\mathbf{V}} is present, such as for
marginal means models or generalized estimating equation (GEE)-like models,
\eqn{\mathbf{V}^{-1}} couples observations across partitions, so
partition-wise fitting is not directly available. Two sub-paths handle
this.

\emph{Path 1a} (Gaussian identity + GEE) solves the whitened system
directly via
\eqn{\tilde{\mathbf{G}} =
(\tilde{\mathbf{X}}^{\top}\tilde{\mathbf{X}} +
\boldsymbol{\Lambda}_{\mathrm{block}})^{-1}} where
\eqn{\tilde{\mathbf{X}} = \mathbf{V}^{-1/2}\mathbf{X}}, then applies
the Lagrangian projection in the full \eqn{P}-space.
See \code{\link{.get_B_gee_gaussian}}.

\emph{Path 1b} (non-Gaussian GEE) uses a damped SQP iteration on the
whitened system. The first iterate is a constrained Newton step from the
projection matrix \eqn{\mathbf{U}}; subsequent iterates solve the
quadratic subproblem via \code{\link[quadprog]{solve.QP}}.
See \code{\link{.get_B_gee_glm}}.

Both sub-paths have Woodbury-accelerated variants described below.

\strong{Path 2: Gaussian identity, no correlation.}
The constrained estimate is obtained by a single Lagrangian projection
from the per-partition penalized least-squares cross-products (the four
steps in closed form). No outer iteration is needed. When \eqn{K = 0} and
there are no additional constraints, this reduces to the ordinary
penalized closed form
\eqn{\hat{\boldsymbol{\beta}} = \mathbf{G}\mathbf{X}^{\top}\mathbf{y}}.
Implemented in \code{\link{.get_B_gaussian_nocorr}}.

\strong{Path 3: Non-Gaussian GLM, no correlation.}
Unconstrained estimates are obtained separately within each partition,
then projected onto the constraint space. When the link is non-identity,
the information \eqn{\mathbf{G}} depends on the current fitted values
through the working weights, so the projection must be iterated. The
unconstrained anchor \eqn{\hat{\boldsymbol{\beta}}} is held fixed while
\eqn{\mathbf{G}} is recomputed at each iterate's constrained estimate:
\deqn{\tilde{\boldsymbol{\beta}}^{(s+1)} =
  \mathbf{U}^{(s)}\hat{\boldsymbol{\beta}}, \quad
  \mathbf{U}^{(s)} = \mathbf{I} - \mathbf{G}^{(s)}\mathbf{A}
  (\mathbf{A}^{\top}\mathbf{G}^{(s)}\mathbf{A})^{-1}\mathbf{A}^{\top}.}
Iteration stops when the mean absolute coefficient change falls below
\code{tol}, or when the update begins increasing (in which case the
previous iterate is restored). The recomputation of weighted Gram
matrices, Schur corrections, and square-root information factors at each
step is handled by \code{.solver_recompute_G_at_estimate}. Implemented
in \code{\link{.get_B_glm_nocorr}}.
}

\subsection{Woodbury Acceleration for Structured Correlation}{
For structured correlation, write the penalized information as
\deqn{\mathbf{G}^{-1}
  = \mathbf{G}_{\mathrm{on}}^{-1} + \mathbf{G}_{\mathrm{off}}^{-1},}
where \eqn{\mathbf{G}_{\mathrm{on}}^{-1}} contains the \eqn{K+1}
block-diagonal \eqn{p \times p} pieces of
\eqn{\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X} + \boldsymbol{\Lambda}},
and \eqn{\mathbf{G}_{\mathrm{off}}^{-1}} contains the cross-partition part.
Without correlation, \eqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{0}}.

When the cross-partition part is low rank, the code factors it as
\deqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{E}\mathbf{J}\mathbf{E}^{\top},}
with \eqn{\mathbf{E}} of size \eqn{P \times r} and \eqn{\mathbf{J}}
diagonal with entries \eqn{\pm 1}. Then
\deqn{\mathbf{G}
  = \mathbf{G}_{\mathrm{on}}
  - \mathbf{G}_{\mathrm{on}}\mathbf{E}
  (\mathbf{J}^{-1} + \mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}\mathbf{E})^{-1}
  \mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}.}
The only dense inverse is the small \eqn{r \times r} matrix
\eqn{(\mathbf{J}^{-1} + \mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}\mathbf{E})^{-1}}.

Define \eqn{\mathbf{N} = \mathbf{G}_{\mathrm{on}}^{1/2}\mathbf{E}} and a
thin QR decomposition \eqn{\mathbf{N} = \mathbf{Q}\mathbf{R}}. With
\eqn{\mathbf{P} = \mathbf{Q}\mathbf{Q}^{\top}} and
\eqn{\mathbf{S} = \mathbf{I}_r + \mathbf{R}\mathbf{J}\mathbf{R}^{\top}},
the supplement writes
\deqn{\mathbf{F}
  = (\mathbf{I}_P + \mathbf{N}\mathbf{J}\mathbf{N}^{\top})^{-1}
  = \mathbf{I}_P - \mathbf{P} + \mathbf{Q}\mathbf{S}^{-1}\mathbf{Q}^{\top},}
so that
\deqn{\mathbf{G}^{1/2}
  = \mathbf{G}_{\mathrm{on}}^{1/2}\mathbf{F}^{1/2}, \qquad
  \mathbf{G}^{-1/2}
  = \mathbf{F}^{-1/2}\mathbf{G}_{\mathrm{on}}^{-1/2},}
with
\deqn{\mathbf{F}^{1/2}
  = (\mathbf{I}_P - \mathbf{P}) +
  \mathbf{Q}\mathbf{S}^{-1/2}\mathbf{Q}^{\top}, \qquad
  \mathbf{F}^{-1/2}
  = (\mathbf{I}_P - \mathbf{P}) +
  \mathbf{Q}\mathbf{S}^{1/2}\mathbf{Q}^{\top}.}
This preserves the same transformed OLS projection used in the
independent case:
\deqn{\mathbf{y}^{*} = \mathbf{G}^{1/2}\mathbf{w}, \qquad
  \mathbf{X}^{*} = \mathbf{G}^{1/2}\mathbf{A}, \qquad
  \tilde{\boldsymbol{\beta}} = \mathbf{G}^{1/2}\mathbf{r}^{*},}
where \eqn{\mathbf{w} = \mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{y}}.
The full-space operations remain the QR on \eqn{\mathbf{X}^{*}} and a
rank-\eqn{r} correction through \eqn{\mathbf{Q}}.

\strong{Implementation map.} The code uses the same algebra but stores a
numerically convenient representation. The correspondence is:
\itemize{
  \item manuscript \eqn{\mathbf{E}} \eqn{\leftrightarrow} helper output
    \code{E} (legacy alias \code{L});
  \item manuscript diagonal sign matrix \eqn{\mathbf{J}}
    \eqn{\leftrightarrow} helper output \code{J_signs}
    (legacy alias \code{S_signs});
  \item manuscript inverse
    \eqn{(\mathbf{J}^{-1} + \mathbf{E}^{\top}\mathbf{G}_{\mathrm{on}}\mathbf{E})^{-1}}
    \eqn{\leftrightarrow} helper output \code{inner_inv}
    (legacy alias \code{M});
  \item manuscript projector basis \eqn{\mathbf{Q}}
    \eqn{\leftrightarrow} internal basis \code{U_Q}
    (alias \code{Q_basis});
  \item manuscript projector \eqn{\mathbf{P} = \mathbf{Q}\mathbf{Q}^{\top}}
    is not stored explicitly, but is represented through rank-\eqn{r}
    multiplies involving \code{U_Q};
  \item manuscript factors \eqn{\mathbf{F}^{1/2}} and
    \eqn{\mathbf{F}^{-1/2}} are stored internally as
    \eqn{\mathbf{I} - \mathbf{U}_Q\mathbf{C}\mathbf{U}_Q^{\top}} and
    \eqn{\mathbf{I} + \mathbf{U}_Q\mathbf{C}_{\mathrm{inv}}\mathbf{U}_Q^{\top}},
    where \code{C} and \code{C_inv} are diagonal.
}
These are just storage choices; the notation matches the supplement.

For the non-Gaussian Woodbury path
(\code{\link{.get_B_gee_glm_woodbury}}), the fixed correlation
perturbation \eqn{\mathbf{V}^{-1} - \mathbf{I}} and its product with
\eqn{\mathbf{X}} are precomputed once and held fixed
across Newton iterations. At each step, the weighted perturbation
\eqn{\mathbf{X}^{\top}\mathbf{W}(\mathbf{V}^{-1} -
\mathbf{I})\mathbf{X}} is formed using the
precomputed product, then split and re-factored via
\code{\link{.woodbury_redecompose_weighted}}.

The Woodbury path is used when \eqn{r < P/3}; otherwise the code falls
back to the dense whitened solve. If \eqn{\mathbf{F}} is not positive
definite, the dense path is also used.
See \code{\link{.get_B_gee_woodbury}} (Gaussian) and
\code{\link{.get_B_gee_glm_woodbury}} (non-Gaussian).
}

\subsection{Step Control}{
Different paths use different step-control strategies.

In the GEE paths (Paths 1a and 1b), the coefficient update is damped:
\deqn{\boldsymbol{\beta}^{(s+1)} = (1 - \alpha_s)\boldsymbol{\beta}^{(s)}
  + \alpha_s\boldsymbol{\beta}_{\mathrm{cand}}^{(s)},}
with \eqn{\alpha_s = 2^{-d_s}} and \eqn{d_s} increased whenever the
candidate gives non-finite or larger deviance. The loop terminates after
10 consecutive rejections or 100 total iterations, and exits early when
both coefficient change and deviance improvement fall below \code{tol}
after a burn-in period.

In Path 3 (non-Gaussian, no correlation), the outer projection loop has
no line search. The algorithm recomputes \eqn{\mathbf{G}} at the current
constrained estimate and applies a fresh projection. If the mean absolute
coefficient change begins increasing, the previous iterate is restored
and the loop stops. The partition-wise unconstrained estimates themselves
are obtained by damped Newton-Raphson inside
\code{\link{unconstrained_fit_default}}, where
\code{\link{damped_newton_r}} computes a Newton direction once per
iteration and halves the step size until the penalized log-likelihood
improves.
}
}

\section{Accommodating Correlation Structures}{


\subsection{Parametric Correlation Structures}{
Suppose \eqn{\mathrm{Cov}(\mathbf{y}) = \sigma^2\,\mathbf{V}(\boldsymbol{\theta})}
for a known parametric family indexed by \eqn{\boldsymbol{\theta}} (e.g.,
AR(1) with \eqn{\theta = \rho}, Matern with
\eqn{\boldsymbol{\theta} = (\ell, \nu)}, exchangeable with
\eqn{\theta = \rho}). The penalized generalized least-squares problem
becomes
\deqn{\min_{\boldsymbol{\beta}}\;
  (\mathbf{y} - \mathbf{X}\boldsymbol{\beta})^{\top}\mathbf{V}^{-1}
  (\mathbf{y} - \mathbf{X}\boldsymbol{\beta})
  + \boldsymbol{\beta}^{\top}\boldsymbol{\Lambda}\boldsymbol{\beta}
  \quad \text{s.t.}\; \mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{0}.}
The correlation matrix \eqn{\mathbf{V}} is supplied through the
fitted-object components \code{Vhalf} and \code{VhalfInv}, either
directly or via user functions \code{Vhalf_fxn} and \code{VhalfInv_fxn}.
When both are non-\code{NULL}, \code{\link{get_B}} dispatches to the GEE
paths (Path 1a or 1b). For built-in correlation structures, the required
square-root matrices are assembled numerically using
\code{\link{matsqrt}}, \code{\link{matinvsqrt}}, and \code{\link{invert}}.
}

\subsection{Whitening and Permutation}{
Because the data are stored in partition ordering (all observations from
partition 0, then partition 1, etc.) while \eqn{\mathbf{V}} is in the
original observation ordering, a permutation is applied internally:
\eqn{\mathbf{V}_{\mathrm{perm}}^{-1/2} =
\mathbf{V}^{-1/2}[\boldsymbol{\pi}, \boldsymbol{\pi}]}, where
\eqn{\boldsymbol{\pi} = \texttt{unlist(order\_list)}} maps original
indices to partition-ordered indices. The whitened design and response are
\eqn{\tilde{\mathbf{X}} =
\mathbf{V}_{\mathrm{perm}}^{-1/2}\mathbf{X}_{\mathrm{block}}} and
\eqn{\tilde{\mathbf{y}} =
\mathbf{V}_{\mathrm{perm}}^{-1/2}\mathbf{y}}.

The \eqn{\mathbf{X}} and \eqn{\mathbf{y}} inputs to
\code{\link{lgspline.fit}} are preserved in their unwhitened form.
Whitening is applied inside \code{\link{get_B}} and
\code{\link{blockfit_solve}} where the full \eqn{N \times P}
block-diagonal design is available, since applying
\eqn{\mathbf{V}^{-1/2}} to only the diagonal blocks of the partitioned
design would silently discard cross-partition contributions and corrupt
the Gram matrix.
}

\subsection{Loss of Block-Diagonal Structure}{
Unlike the independent-errors case,
\eqn{\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X}} is not block-diagonal
because \eqn{\mathbf{V}^{-1}} introduces cross-partition coupling. The
unconstrained estimator
\deqn{\hat{\boldsymbol{\beta}} =
  (\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X} +
  \boldsymbol{\Lambda})^{-1}
  \mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{y}}
requires a full \eqn{P \times P} solve, and the constraint projection
proceeds with
\eqn{\mathbf{G} = (\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X} +
\boldsymbol{\Lambda})^{-1}}.

For structured correlation matrices (AR(1), exchangeable, banded), the
perturbation \eqn{\mathbf{V}^{-1} - \mathbf{I}} is sparse and the
cross-partition coupling has low effective rank. In these cases the
Woodbury-accelerated paths
(\code{\link{.get_B_gee_woodbury}},
\code{\link{.get_B_gee_glm_woodbury}}) recover the partition-wise
computational structure by decomposing the coupling into a
block-diagonal correction plus a low-rank remainder, as described in the
GLM Extension section. For dense or high-rank
\eqn{\mathbf{V}^{-1}}, the code falls back to the full whitened system
(\code{\link{.get_B_gee_gaussian}}, \code{\link{.get_B_gee_glm}}).
}

\subsection{GEE Deviance Monitoring}{
For non-Gaussian models with correlation, the deviance used for
convergence monitoring is computed in the whitened space by
\code{.bf_gee_deviance}. When the family supplies
\code{custom_dev.resids}, the raw deviance residuals
\eqn{r_i = \mathrm{sign}(d_i)\sqrt{|d_i|}} are divided by
\eqn{\sqrt{w_i}} and pre=dmultiplied by
\eqn{\mathbf{V}_{\mathrm{perm}}^{-1/2}} before squaring and averaging:
\deqn{D_{\mathrm{GEE}} = \frac{1}{N}\left\|
  \mathbf{V}_{\mathrm{perm}}^{-1/2}\,
  \mathbf{w}^{-1/2}\mathbf{r}\right\|^{2},}
where \eqn{\mathbf{w}} is the vector of working weights at the current
iterate, clamped below at \eqn{\sqrt{\varepsilon_{\mathrm{mach}}}}. When
only \code{dev.resids} is available, the function falls back to the
standard mean deviance; otherwise it uses mean squared error.
}

\subsection{REML Estimation of Correlation Parameters}{
Correlation parameters \eqn{\boldsymbol{\theta}} are estimated by minimizing
a negative restricted log-likelihood (REML) objective. The criterion
implemented in \pkg{lgspline} is a central-limit-theorem-based working
approximation to a Laplace-style marginal likelihood criterion,
applied here solely to correlation structure estimation rather than
penalty parameter selection.
Let \eqn{\mathbf{D} = \mathrm{diag}(d_i)} be the observation weight matrix,
\eqn{\tilde{\mathbf{W}} = \mathrm{diag}(\tilde{w}_i)} the GLM working weight
matrix at the current fitted values, \eqn{\mathbf{V}} the correlation matrix
parameterized by \eqn{\boldsymbol{\rho}} (a vector on the unconstrained
real line), and \eqn{\tilde{\sigma}^{2}} the dispersion profiled at its
restricted maximum likelihood estimate. The negative REML objective
implemented in \pkg{lgspline}, scaled by \eqn{1/N}, is
\deqn{-\ell_{R}(\boldsymbol{\rho}) = \frac{1}{N}\!\left[
  -\log|\mathbf{V}^{-1/2}|
  + \frac{N}{2}\log\tilde{\sigma}^{2}
  + \frac{1}{2\tilde{\sigma}^{2}}
    (\mathbf{y} - \boldsymbol{\mu})^{\top}
    \mathbf{D}\tilde{\mathbf{W}}^{-1}\mathbf{V}^{-1}
    (\mathbf{y} - \boldsymbol{\mu})
  + \frac{1}{2}\log|(\tilde{\sigma}^{2}\mathbf{U}\mathbf{G})^{-1}|^{+}
\right],}
where \eqn{|\cdot|^{+}} denotes the generalized determinant (product of
nonzero eigenvalues), and
\eqn{\boldsymbol{\mu} = g^{-1}(\mathbf{X}\tilde{\boldsymbol{\beta}})} are
the fitted values on the response scale.

Gradients with respect to correlation parameters are available in
closed form for all built-in structures except Matern, which uses
finite-difference approximation due to the complexity of differentiating
the modified Bessel function \eqn{K_{\nu}} with respect to \eqn{\nu}. See
\code{\link{reml_grad_from_dV}} for the full gradient derivation and
notation. Custom analytic gradients can be supplied through \code{REML_grad},
and a fully custom criterion can replace REML through
\code{custom_VhalfInv_loss}. The Toeplitz example in \code{\link{lgspline}}
demonstrates how to supply custom correlation structures with user-defined
gradient functions. Optimization over these working correlation parameters is
then carried out by the same quasi-Newton engine used elsewhere in the
package, with a finite-difference fallback when needed. In the user-facing interface, this
machinery is activated through \code{correlation_structure},
\code{correlation_id}, \code{spacetime}, \code{VhalfInv}, \code{Vhalf},
\code{VhalfInv_fxn}, \code{Vhalf_fxn}, \code{VhalfInv_par_init},
\code{REML_grad}, \code{custom_VhalfInv_loss}, and
\code{VhalfInv_logdet}.

The gradient of the negative REML has three terms per parameter:
\enumerate{
  \item \eqn{\frac{1}{2}\mathrm{tr}(\mathbf{V}^{-1}\partial\mathbf{V}/\partial\theta_j)}:
    the log-determinant contribution.
  \item \eqn{-\frac{1}{2\tilde{\sigma}^{2}}\mathbf{r}^{\top}
    (\partial\mathbf{V}/\partial\theta_j)\mathbf{r}}:
    the residual quadratic form contribution, where
    \eqn{\mathbf{r} = \mathrm{diag}(\sqrt{d_i}/\sqrt{\tilde{w}_i})
    \mathbf{V}^{-1/2}(\mathbf{y} - \boldsymbol{\mu})}.
  \item \eqn{-\frac{1}{2}\mathrm{tr}\!\left(\mathbf{M}^{+}\mathbf{X}_{*}^{\top}
    \mathbf{V}^{-1}(\partial\mathbf{V}/\partial\theta_j)\mathbf{V}^{-1}
    \tilde{\mathbf{W}}\mathbf{D}\mathbf{X}_{*}\right)}:
    the REML correction, where
    \eqn{\mathbf{X}_{*} = \mathbf{X}\mathbf{U}} is the constrained design and
    \eqn{\mathbf{M} = \mathbf{X}_{*}^{\top}\mathbf{V}^{-1}
    \tilde{\mathbf{W}}\mathbf{D}\mathbf{X}_{*}
    + \mathbf{U}^{\top}\boldsymbol{\Lambda}\mathbf{U}} is the projected
    penalized information.
}

For each supported correlation family, the derivatives
\eqn{\partial\mathbf{V}/\partial\theta_j} are available in closed form,
enabling analytic gradient computation for use with the quasi-Newton
optimizer.
}

\subsection{Connection to Standard Mixed Model REML}{
The quadratic penalty \eqn{\boldsymbol{\Lambda}} acts as the inverse prior
covariance of a Gaussian random effect on the spline coefficients, with
the smoothing parameter satisfying \eqn{\lambda = \tau^{2}/\sigma^{2}};
this is the same mixed model representation used by \pkg{mgcv}, and
Monte Carlo draws under the resulting Laplace-style approximation are
available via \code{\link{generate_posterior}}. For Gaussian responses
with identity link and no penalization, the implemented criterion
coincides exactly with classical REML. For non-Gaussian responses, the
criterion substitutes the Fisher information for the full penalized
log-likelihood Hessian, exploiting the CLT approximation that
\eqn{\tilde{\mathbf{W}}^{-1/2}(\mathbf{y} - \boldsymbol{\mu})} is
approximately Gaussian; this yields a method-of-moments style estimator
for the correlation and variance parameters that depends only on
mean-variance relationships through \eqn{\mathbf{W}}, and therefore
generalizes naturally to quasi-likelihood and other settings where a
fully specified log-likelihood is unavailable. The REML correction term
\eqn{\log|(\tilde{\sigma}^{2}\mathbf{U}\mathbf{G})^{-1}|^{+}} uses a
generalized log-determinant so that only nonzero eigenvalues contribute
when rank deficiency arises from smoothness constraints or
identifiability conditions in \eqn{\mathbf{A}}. When
\eqn{\boldsymbol{\Lambda}} is full rank, the criterion coincides with
the exact marginal likelihood from integrating out
\eqn{\boldsymbol{\beta}} under its Gaussian prior; when
\eqn{\boldsymbol{\Lambda}} is rank-deficient, unpenalized coefficients
are projected out in the REML sense while penalized coefficients are
integrated through their prior. During penalty tuning, the block-diagonal
approximation is retained for GCV criteria and gradients; since GCV is
rotation-invariant the practical effect on automatic selection is expected
to be negligible, though this is not confirmed and the tuned penalties can
always be overridden.
}

\subsection{Built-In Correlation Structures}{
The package provides several built-in correlation structures for modeling
spatial and temporal dependence. These are specified via
\code{correlation_structure} with group membership in \code{correlation_id}
and spatial or temporal coordinates in \code{spacetime} (an \eqn{N}-row
matrix). Exchangeable correlation does not require \code{spacetime}.

All positive scale parameters are estimated on the log scale,
with back-transform \eqn{\exp(\cdot)}. Parameters constrained to
\eqn{(0, 1)} use a double-exponential back-transform of the form
\eqn{\exp(-\exp(\eta))}, so optimization still occurs on the
unconstrained real line while the correlation remains bounded.

\describe{
  \item{Exchangeable}{
    Aliases: \code{'exchangeable'}, \code{'cs'}, \code{'CS'},
    \code{'compoundsymmetric'}, \code{'compound-symmetric'}.

    A constant correlation \eqn{\nu} between any two observations within
    the same cluster. Parameterization: \eqn{\nu = \exp(-\exp(\rho))},
    so \eqn{\nu \in (0, 1)}. Only positive within-cluster correlation is
    supported under this parameterization.
  }
  \item{Spatial Exponential}{
    Aliases: \code{'spatial-exponential'}, \code{'spatialexponential'},
    \code{'exp'}, \code{'exponential'}.

    Correlation decays exponentially with distance:
    \eqn{\exp(-\omega d)} where \eqn{d} is Euclidean distance and
    \eqn{\omega > 0}. Parameterization: \eqn{\omega = \exp(\rho)}.
    Mathematically equivalent to the power correlation \eqn{\theta^{d}}
    with \eqn{\theta = e^{-\omega}}, but with better numerical properties
    during optimization.
  }
  \item{AR(1)}{
    Aliases: \code{'ar1'}, \code{'ar(1)'}, \code{'AR(1)'}, \code{'AR1'}.

    Correlation depends on rank difference between observations:
    \eqn{\nu^{r}} where \eqn{r} is the rank difference within cluster.
    Parameterization: \eqn{\nu = \exp(-\exp(\rho))},
    so \eqn{\nu \in (0, 1)}. Only positive autocorrelation is supported.
  }
  \item{Gaussian / Squared Exponential}{
    Aliases: \code{'gaussian'}, \code{'rbf'}, \code{'squared-exponential'}.

    Smooth decay with squared distance:
    \eqn{\exp(-d^{2}/(2\ell^{2}))} where \eqn{\ell} is the length scale.
    Parameterization: \eqn{\ell = \exp(\rho)}.
  }
  \item{Spherical}{
    Aliases: \code{'spherical'}, \code{'Spherical'}, \code{'cubic'},
    \code{'sphere'}.

    Polynomial decay with a hard cutoff at range \eqn{r}:
    \eqn{1 - 1.5(d/r) + 0.5(d/r)^{3}} for \eqn{d \le r}, and \eqn{0}
    otherwise. Parameterization: \eqn{r = \exp(\rho)}.
  }
  \item{Matern}{
    Aliases: \code{'matern'}, \code{'Matern'}.

    Flexible correlation with adjustable smoothness:
    \eqn{(2^{1-\nu}/\Gamma(\nu))(\sqrt{2\nu}\,d/\ell)^{\nu}K_{\nu}(\sqrt{2\nu}\,d/\ell)}.
    Two parameters: length scale \eqn{\ell = \exp(\rho_1)} and smoothness
    \eqn{\nu = \exp(\rho_2)}. No analytical gradient is available for
    \eqn{\nu} due to the difficulty of differentiating the modified Bessel
    function \eqn{K_{\nu}} with respect to its order, so finite differences
    are used; this makes Matern slower and potentially less stable than
    other structures.
  }
  \item{Gamma-Cosine}{
    Aliases: \code{'gamma-cosine'}, \code{'gammacosine'},
    \code{'GammaCosine'}.

    Oscillatory dependence:
    \eqn{(d^{\alpha-1}e^{-\gamma d})/(\Gamma(\alpha)/\gamma^{\alpha})\cdot\cos(\omega d)}.
    Three parameters: shape \eqn{\alpha = \exp(\rho_1)}, rate
    \eqn{\gamma = \exp(\rho_2)}, frequency \eqn{\omega = \exp(\rho_3)}.
    Reduces to exponential when \eqn{\alpha = 1} and \eqn{\omega \approx 0}.
  }
  \item{Gaussian-Cosine}{
    Aliases: \code{'gaussian-cosine'}, \code{'gaussiancosine'},
    \code{'GaussianCosine'}.

    Smooth oscillatory correlation:
    \eqn{\exp(-d^{2}/(2\ell^{2}))\cdot\cos(\omega d)}.
    Two parameters: length scale \eqn{\ell = \exp(\rho_1)} and frequency
    \eqn{\omega = \exp(\rho_2)}. Reduces to Gaussian when
    \eqn{\omega \approx 0}.
  }
}
}

\subsection{Interpreting Estimated Correlation Parameters}{
Correlation parameters are estimated on transformed scales; they must be
back-transformed for interpretation. When \code{\link{confint.lgspline}} is called and the
inverse Hessian from BFGS is available, confidence intervals are returned
on the untransformed (working) scale and should be back-transformed as
described in the examples for \code{\link{lgspline}}.
}

\subsection{Custom Correlation Structures}{
Custom correlation structures can be specified through:
\itemize{
  \item \code{VhalfInv_fxn}: Creates \eqn{\mathbf{V}^{-1/2}}.
  \item \code{Vhalf_fxn}: Creates \eqn{\mathbf{V}^{1/2}}. When omitted,
    the code computes it by explicit inversion of \code{VhalfInv}.
  \item \code{REML_grad}: Provides the analytical gradient of the REML
    objective.
  \item \code{VhalfInv_logdet}: Efficient log-determinant computation.
  \item \code{custom_VhalfInv_loss}: Replaces the REML objective entirely.
}
These functions enter \code{\link{lgspline}} through
\code{correlation_structure}, \code{VhalfInv_fxn}, \code{Vhalf_fxn},
\code{REML_grad}, and \code{custom_VhalfInv_loss}, and the fitted object
retains the resulting correlation machinery in components such as
\code{VhalfInv_fxn}, \code{Vhalf_fxn}, and
\code{VhalfInv_params_estimates}. When \code{VhalfInv} is supplied but
\code{Vhalf} is not, \code{Vhalf} is computed unconditionally as the inverse
of \code{VhalfInv} for all family/link combinations, since both
\code{\link{get_B}} and \code{\link{blockfit_solve}} require it for GEE
estimation.
}
}

\section{Variance and Dispersion Estimation}{


Once the constrained estimate has been obtained, the next questions are how
much flexibility the fitted model effectively used and how uncertainty should
be propagated through the same constrained geometry. The quantities in this
section are therefore all built on the projected information matrices from
the previous sections.

\subsection{Effective Degrees of Freedom and Dispersion}{
The effective degrees of freedom is the trace of the hat matrix. In the
Gaussian identity case with observation weights and correlation, the fitted
linear operator is built from the dense GLS analogue
\eqn{\mathbf{G}_{\mathrm{correct}}} and can be written schematically as
\deqn{\mathbf{H} =
  \mathbf{V}^{-1/2}(\mathbf{W}\mathbf{D})^{1/2}
  \mathbf{X}\mathbf{U}\mathbf{G}_{\mathrm{correct}}\mathbf{X}^{\top}
  (\mathbf{W}\mathbf{D})^{1/2}\mathbf{V}^{-1/2},}
where for Gaussian identity \eqn{\mathbf{W} = \mathbf{I}}. In the
no-correlation Gaussian case this reduces to the familiar
\eqn{\mathbf{H} = \mathbf{X}\mathbf{U}\mathbf{G}\mathbf{X}^{\top}\mathbf{D}}.

For Gaussian identity fits, the dispersion estimate is computed as a
weighted mean squared residual, optionally scaled by
\eqn{N/(N - \mathrm{tr}(\mathbf{H}))} when
\code{unbias_dispersion = TRUE}:
\deqn{\tilde{\sigma}^{2} =
  \frac{1}{N - \mathrm{tr}(\mathbf{H})}\|\mathbf{y} - \mathbf{\tilde{y}}_i \|^{2}.}

More generally with weights, a correlation structure and non-linear link function:
\deqn{\tilde{\sigma}^{2} =
  \frac{1}{N - \mathrm{tr}(\mathbf{H})}\| \mathbf{V}^{-1/2}\mathbf{W}^{-1/2}\mathbf{D}^{1/2}(\mathbf{y} - \tilde{\mathbf{y}})\|^{2}.}

This estimated dispersion is returned as \code{sigmasq_tilde}, and the
corresponding effective degrees of freedom trace is returned as
\code{trace_XUGX}. For non-Gaussian families, the fitting code delegates
dispersion estimation to \code{dispersion_function}; thus the package does
not assume a single closed-form Pearson-style formula outside the Gaussian
identity setting. The hat-matrix trace itself is assembled by
\code{\link{compute_trace_H}} in the dense correlation-aware case and by the
same blockwise products summarized by \code{trace_XUGX} in the simpler
no-correlation paths.
A concrete built-in non-Gaussian example is the Weibull AFT path, which pairs
\code{\link{weibull_family}} with \code{\link{weibull_dispersion_function}},
\code{\link{weibull_glm_weight_function}}, and
\code{\link{weibull_schur_correction}}.
Users who want these quantities available for downstream inference should
keep \code{estimate_dispersion = TRUE} and \code{return_varcovmat = TRUE}
(the defaults), since \code{\link{wald_univariate}},
\code{\link{confint.lgspline}}, and the prediction-standard-error path in
\code{\link{predict.lgspline}} all rely on the post-fit dispersion and
covariance components documented here.
}

\subsection{Variance-Covariance Matrix}{
The variance-covariance matrix of \eqn{\tilde{\boldsymbol{\beta}}} is
estimated as:
\deqn{\mathrm{Var}(\tilde{\boldsymbol{\beta}}) = \tilde{\sigma}^{2}(\mathbf{U}\mathbf{G}^{1/2})(\mathbf{U}\mathbf{G}^{1/2})^{\top}}
using the outer-product form for numerical stability. The result is returned
as \code{varcovmat} when \code{return_varcovmat = TRUE}. The algebraically
equivalent expression \eqn{\tilde{\sigma}^{2}\mathbf{U}\mathbf{G}} is not
used because \eqn{\mathbf{G}} is only positive semi-definite when the
penalty matrix \eqn{\boldsymbol{\Lambda}} has zero eigenvalues (e.g., the
intercept and linear terms under the smoothing spline penalty when
\code{flat_ridge_penalty} = 0), which can introduce negative diagonal
entries in finite precision arithmetic. The outer-product form also
guarantees symmetry.

This is the Bayesian posterior covariance, treating the penalty as a
Gaussian prior on the coefficients. When \code{exact_varcovmat = TRUE},
a frequentist correction is additionally computed:
\deqn{\mathrm{Var}_{\mathrm{exact}}(\tilde{\boldsymbol{\beta}}) =
  \tilde{\sigma^2}\mathbf{U}\mathbf{G}^{1/2}(\mathbf{X}^{\top}\mathbf{W}\mathbf{D}\mathbf{V}^{-1}\mathbf{X})\mathbf{G}^{1/2}\mathbf{U}^{\top} =
  \tilde{\sigma}^{2}\mathbf{U}\mathbf{G}\mathbf{U}^{\top}
  - \tilde{\sigma}^{2}\mathbf{U}\mathbf{G}\boldsymbol{\Lambda}\mathbf{G}\mathbf{U}^{\top}.}
The first term is the Bayesian posterior covariance; the second is a frequentist
correction such that for Gaussian identity link (with or without correlation),
this is the exact variance-covariance matrix of the constrained estimator.

When a correlation structure is present (\code{VhalfInv} non-\code{NULL}),
the block-diagonal \eqn{\mathbf{G}} is replaced by the full weighted GLS
analogue
\deqn{\mathbf{G}_{\mathrm{correct}} =
  \left(\mathbf{X}^{\top}\mathbf{W}\mathbf{D}\mathbf{V}^{-1}\mathbf{X}
  + \boldsymbol{\Lambda}\right)^{-1},}
where \eqn{\mathbf{W} = \mathbf{I}} in the Gaussian identity case.
This dense matrix is what enters the correlation-aware \eqn{\mathbf{U}},
\eqn{\mathrm{Var}(\tilde{\boldsymbol{\beta}})}, and
\eqn{\mathrm{Var}_{\mathrm{exact}}(\tilde{\boldsymbol{\beta}})}
computations.

In user-facing terms, \code{return_varcovmat} controls whether this matrix is
stored at all, while \code{exact_varcovmat} switches between the default
posterior/Laplace approximation and the exact frequentist correction in the
Gaussian-identity setting. The stored covariance is what powers
\code{\link{wald_univariate}}, \code{\link{confint.lgspline}}, and
\code{se.fit = TRUE} in \code{\link{predict.lgspline}}; the
\code{critical_value} argument supplied at fit time is carried forward as the
default cutoff for those interval-producing helpers.
}

\subsection{Recomputation of G at Convergence}{
At the final iterate, \eqn{\mathbf{G}} is recomputed to reflect the
converged working weights and Schur corrections. The implementation
computes the weighted design
\eqn{\mathbf{X}_{w}^{(k)} = \mathbf{X}_k \cdot \mathrm{diag}(\sqrt{\mathbf{w}_k})},
forms the weighted Gram matrix
\eqn{\mathbf{X}_{w}^{(k)\top}\mathbf{X}_{w}^{(k)}}, adds the Schur
correction, and performs eigendecomposition via
\code{\link{compute_G_eigen}} to obtain \eqn{\mathbf{G}_k},
\eqn{\mathbf{G}_k^{1/2}}, and \eqn{\mathbf{G}_k^{-1/2}}. The relationship
\eqn{\mathbf{G}_k = \mathbf{G}_k^{1/2}(\mathbf{G}_k^{1/2})^{\top}} is
enforced exactly by construction, and the fitted object can retain these as
\code{G} and \code{Ghalf} when \code{return_G = TRUE} and
\code{return_Ghalf = TRUE}. The square-root factors are numerically
stabilized through the helper routines \code{\link{matsqrt}} and
\code{\link{matinvsqrt}}, which are also used elsewhere in the package when
dense analogues of \eqn{\mathbf{G}^{1/2}} or \eqn{\mathbf{G}^{-1/2}} are
required.
}
}

\section{Bayesian Interpretation}{


The penalty has a natural Gaussian-prior interpretation, so once the
constrained estimator and its covariance are available, Bayesian-style
posterior simulation follows almost immediately. This section records the
interpretation that is already implicit in the fitted object and in the
package's posterior simulation helpers.

A Bayesian interpretation follows from viewing the penalty as a Gaussian
prior on the coefficients. Conditional on the fitted smoothing parameters,
the code samples on the coefficient scale from
\deqn{\boldsymbol{\beta}^{(m)} =
  \tilde{\boldsymbol{\beta}} +
  \sqrt{\tilde{\sigma}^{2}},\mathbf{U}\mathbf{G}^{1/2}\mathbf{z}^{(m)},
  \qquad \mathbf{z}^{(m)} \sim \mathcal{N}(\mathbf{0}, \mathbf{I}).}
The coefficients are then back-transformed to the original response and
predictor scales.

When inequality constraints are absent, these draws are i.i.d. Gaussian
posterior draws around the fitted mode.
The underlying coefficient-draw closure also contains an elliptical slice
sampling route for active inequality constraints, using the same covariance
factor to keep retained draws in the feasible region.

At the implementation level, standard Gaussian posterior draws may place
positive mass on the infeasible region, so the constrained-draw closure
instead targets the truncated posterior
\deqn{\pi(\boldsymbol{\beta} \mid \mathbf{y}) \propto
  \exp\!\left(-\frac{1}{2}(\boldsymbol{\beta} - \tilde{\boldsymbol{\beta}})^{\top}
  \mathbf{G}^{-1}(\boldsymbol{\beta} - \tilde{\boldsymbol{\beta}})\right)
  \mathbf{1}(\mathbf{C}^{\top}\boldsymbol{\beta} \succeq \mathbf{c}),}
yielding credible intervals that respect the constraint boundaries.
The public \code{\link{generate_posterior}} wrapper forwards
\code{enforce_qp_constraints} to the stored constrained-draw closure, so
constrained draws can be requested directly from the user-facing interface.
When a working correlation structure is present, the companion helper
\code{\link{generate_posterior_correlation}} extends this idea by propagating
uncertainty in the fitted correlation parameters through the same
\code{VhalfInv_fxn}/\code{Vhalf_fxn} machinery described in the correlation
section, rather than conditioning only on fixed covariance parameters. Correlation
parameters are drawn from a multivariate normal distribution centered about their estimates
with the inverse approxiamte BFGS Hessian of the REML optimization problem
\code{VhalfInv_params_vcov} used by default (or a custom alternative as
supplied to the argument \code{correlation_param_vcov_sc}).
}

\section{Inequality Constraints via Sequential Quadratic Programming}{


\subsection{Overview}{
Inequality constraints of the form
\eqn{\mathbf{C}^{\top}\boldsymbol{\beta} \succeq \mathbf{c}} handle
shape restrictions such as monotonicity, convexity, or boundedness. In
the monomial basis, these are linear in \eqn{\boldsymbol{\beta}}.
Monotonicity at a grid of points requires the first derivative polynomial
to be non-negative there; convexity requires the second derivative to be
non-negative; range constraints bound the fitted values directly. All of
these translate to linear inequality constraints on the coefficient
vector.

The inequality pieces are assembled by \code{\link{process_qp}}, which
returns the \code{qp_Amat}, \code{qp_bvec}, and \code{qp_meq} objects
passed to \code{\link[quadprog]{solve.QP}}, together with a
\code{quadprog} flag indicating whether any inequality constraints are
active. For derivative-sign constraints, \code{\link{process_qp}} calls
\code{.build_deriv_qp}, which uses
\code{\link{make_derivative_matrix}} on the expansion-standardized
design and maps the derivative rows into the full \eqn{P}-dimensional
coefficient space partition by partition.
}

\subsection{Partition-Wise Active-Set Method}{
The sparsity pattern of \eqn{\mathbf{C}} is inspected automatically by
\code{\link{.detect_qp_global}} (equivalently
\code{.solver_detect_qp_global}). When every column of
\eqn{\mathbf{C}} has nonzero entries in only a single partition block,
the constraint system is block-separable and an active-set method can
replace the dense QP.

At each iteration, the active set \eqn{\mathcal{A}} (constraints
satisfied at equality) is appended to \eqn{\mathbf{A}} as additional
equality constraints:
\deqn{\mathbf{A}_{\mathrm{aug}} = [\mathbf{A} \mid
  \mathbf{C}_{\mathcal{A}}].}
The constrained estimate is obtained by the same OLS projection with
\eqn{\mathbf{A}_{\mathrm{aug}}} in place of \eqn{\mathbf{A}}, and
since \eqn{\mathbf{A}_{\mathrm{aug}}} retains block-diagonal
compatibility, all operations remain partition-wise.

The active set is updated by checking primal feasibility (adding the
most-violated inactive constraint) and dual feasibility (dropping the
active constraint with the most negative Lagrange multiplier) until the
KKT conditions are satisfied. Lagrange multipliers for active
inequalities are recovered from the OLS fit used in the Lagrangian
projection: the fitted coefficients on
\eqn{\mathbf{X}^* = \mathbf{G}^{1/2}\mathbf{A}_{\mathrm{aug}}} give
the multipliers up to sign. Implemented in
\code{\link{.active_set_refine}} and
\code{\link{.check_kkt_partitionwise}}.

When any column of \eqn{\mathbf{C}} spans multiple partition blocks
(for example, cross-knot monotonicity constraints), the block-diagonal
structure is broken and the dense SQP approach is required. The
selection is made automatically. If the active-set method does not
converge within its iteration limit (default 50), the code falls back
to dense SQP as well.
}

\subsection{Dense SQP Iteration}{
The dense SQP approach, implemented in \code{\link{.qp_refine}},
solves a sequence of quadratic subproblems approximating the penalized
log-likelihood. At each iteration \eqn{s}:
\enumerate{
  \item Compute the information matrix
    \eqn{\mathbf{M}^{(s)} =
    \mathbf{X}^{\top}\mathbf{W}^{(s)}\mathbf{X} +
    \boldsymbol{\Lambda}_{\mathrm{block}} + \mathbf{S}^{(s)}},
    where \eqn{\mathbf{S}^{(s)}} is the Schur complement correction.
  \item Compute the score vector via \code{qp_score_function}.
  \item Solve the QP with \code{\link[quadprog]{solve.QP}}, where the
    combined constraint matrix is
    \eqn{[\mathbf{A} \mid \mathbf{C}]} with the first \eqn{R} columns
    treated as equalities.
  \item Apply a damped update
    \eqn{\boldsymbol{\beta}^{(s+1)} =
    (1 - \alpha)\boldsymbol{\beta}^{(s)} +
    \alpha\boldsymbol{\beta}_{\mathrm{QP}}}, with
    \eqn{\alpha = 2^{-d}} and \eqn{d} incremented upon deviance
    increase.
}
A rescaling factor
\eqn{\mathrm{sc} = \sqrt{\mathrm{mean}(|\mathbf{M}^{(s)}|)}} is applied
to the Hessian and linear term before calling the QP solver. The
equality-constrained estimate from the Lagrangian projection serves as
a warm start. The \code{qp_score_function} defaults to the canonical GLM
score
\eqn{\mathbf{X}^{\top}(\mathbf{y} - \boldsymbol{\mu})}; for custom
models a different score can be supplied.
}

\subsection{Active Set and Lagrange Multipliers}{
The active set at the solution identifies binding inequality constraints.
The Lagrange multipliers quantify the cost of each: a multiplier of zero
means the constraint is not binding. The implementation stores active
constraint indices, the corresponding submatrix of the constraint matrix,
and the multiplier vector in the \code{qp_info} list returned alongside
coefficient estimates, with components \code{lagrangian}, \code{iact},
and \code{Amat_active}. When
\code{return_lagrange_multipliers = TRUE}, the fitted object also stores
the final multiplier vector directly as \code{lagrange_multipliers}.

The original assembled inequality data are retained in the fitted
object's \code{quadprog_list} component (containing \code{qp_Amat},
\code{qp_bvec}, and \code{qp_meq} from \code{\link{process_qp}}) so
the final active set can be interpreted relative to the full
specification. The final \eqn{\mathbf{U}} used in constructing the
posterior variance-covariance matrix is built from both the equality
constraints and the active inequality constraints at the solution.
}

\subsection{Built-In Constraints}{
Built-in inequality constraints include:
\itemize{
  \item Monotonicity: \code{qp_monotonic_increase},
    \code{qp_monotonic_decrease}. Enforced by requiring consecutive
    fitted values to be non-decreasing (or non-increasing):
    \eqn{(\mathbf{x}_i - \mathbf{x}_{i-1})^{\top}\boldsymbol{\beta}
    \geq 0}. These are constructed by \code{\link{process_qp}} from
    the partition-stacked block design reordered to observation order.
  \item Derivative sign: \code{qp_positive_derivative},
    \code{qp_negative_derivative}. Enforced through the
    first-derivative design matrix from
    \code{\link{make_derivative_matrix}}. May be \code{TRUE}/\code{FALSE}
    or a character/integer vector selecting specific predictors.
  \item Second-derivative sign: \code{qp_positive_2ndderivative},
    \code{qp_negative_2ndderivative}. Same construction using the
    second-derivative design matrix.
  \item Response range: \code{qp_range_lower}, \code{qp_range_upper}.
    Bounds on the linear predictor; for non-identity links, the bounds
    are transformed to the link scale inside \code{\link{process_qp}}.
  \item Custom constraints via \code{qp_Amat_fxn}, \code{qp_bvec_fxn},
    and \code{qp_meq_fxn}, which receive the design matrix structure
    and return the constraint matrix, bound vector, and number of
    equalities. These are commonly paired with a custom
    \code{qp_score_function} when the quadratic approximation uses a
    non-default likelihood. The low-level objects \code{qp_Amat},
    \code{qp_bvec}, and \code{qp_meq} remain documented in
    \code{\link{lgspline}} but in the current implementation serve as
    activation markers rather than being merged into the constraint set
    assembled by \code{\link{process_qp}}.
}
All constraints can be thinned to a user-specified subset of rows via
\code{qp_observations}, which \code{\link{process_qp}} applies before
assembly.
}
}

\section{Blockfit Backfitting for Linear Non-Interactive Effects}{


\subsection{Motivation}{
When a model contains both spline terms (receiving \eqn{K+1}
partition-specific coefficient vectors constrained to smoothness) and
non-interactive linear terms (``flat'' terms, specified via
\code{just_linear_without_interactions}, receiving a single shared
coefficient vector \eqn{\mathbf{v}} across all partitions), the standard
solver carries \eqn{K+1} copies of \eqn{\mathbf{v}} linked by equality
constraints. Backfitting avoids this inflation by solving a
lower-dimensional problem at each step. Write the partition-\eqn{k}
design as
\eqn{\mathbf{X}_k = [\mathbf{Z}_k \mid
\mathbf{X}_{\mathrm{flat}}^{(k)}]}, where \eqn{\mathbf{Z}_k} contains
the spline columns and \eqn{\mathbf{X}_{\mathrm{flat}}^{(k)}} the flat
columns. This is invoked when \code{blockfit = TRUE}, flat columns are
non-empty, and \eqn{K > 0}.

The design, penalty, and constraint matrices are split into spline and
flat components by \code{\link{.bf_split_components}}, which extracts
spline rows from \eqn{\mathbf{A}}, drops null columns, rank-reduces via
QR, and detects mixed constraints (columns of \eqn{\mathbf{A}} with
nonzero entries on both spline and flat rows).
}

\subsection{Block-Coordinate Descent}{
\strong{Spline step.} Holding \eqn{\mathbf{v}} fixed, the code forms
the adjusted response
\eqn{\mathbf{y}_k - \mathbf{X}_{\mathrm{flat}}^{(k)}\mathbf{v}} and
applies the spline-only Lagrangian projection via
\code{\link{.bf_lagrangian_project}}:
\deqn{\tilde{\boldsymbol{\beta}}_{\mathrm{spline}}^{(k)} =
  \mathbf{U}_{\mathrm{spline}}\mathbf{G}_{\mathrm{spline}}
  \mathbf{Z}_k^{\top}(\mathbf{y}_k -
  \mathbf{X}_{\mathrm{flat}}^{(k)}\mathbf{v}).}

\strong{Flat step.} Holding spline coefficients fixed, the shared flat
vector is updated by pooled penalized regression on residuals:
\deqn{\mathbf{v} = \left(\sum_{k=0}^{K}
  \mathbf{X}_{\mathrm{flat}}^{(k)\top}
  \mathbf{X}_{\mathrm{flat}}^{(k)} +
  \boldsymbol{\Lambda}_{\mathrm{flat}}\right)^{-1}
  \sum_{k=0}^{K}
  \mathbf{X}_{\mathrm{flat}}^{(k)\top}
  (\mathbf{y}_k - \mathbf{Z}_k
  \tilde{\boldsymbol{\beta}}_{\mathrm{spline}}^{(k)}).}
When the constraint matrix \eqn{\mathbf{A}} has mixed columns (nonzero
entries on both spline and flat rows), the flat update instead solves a
KKT system enforcing the residual equality constraint
\eqn{\mathbf{A}_{\mathrm{flat}}^{\top}\mathbf{v} = \mathbf{c} -
\mathbf{A}_{\mathrm{spline}}^{\top}
\tilde{\boldsymbol{\beta}}_{\mathrm{spline}}}
via \code{\link{.bf_constrained_flat_update}}.

Convergence is checked using the maximum absolute change across both
spline and flat coefficients. In the weighted inner loop used by the GLM
solvers, the flat-block change alone determines stopping.
}

\subsection{Four Estimation Cases}{
\code{\link{blockfit_solve}} dispatches to one of four paths.

\strong{Case (a): Gaussian identity + GEE}
(\code{\link{.bf_case_gauss_gee}}).
Whitening destroys block-diagonal structure, so the code skips
backfitting and performs the same full-system Gaussian GEE projection
used by \code{\link{get_B}} Path 1a. The result is split back into
spline and flat components for downstream assembly.

\strong{Case (b): Gaussian identity, no correlation}
(\code{\link{.bf_case_gauss_no_corr}}).
Standard block-coordinate descent as described above. The spline-only
\eqn{\mathbf{G}_{\mathrm{spline}}} factors are precomputed once, and the
pooled flat penalized inverse is reused across iterations.

\strong{Case (c): GLM + GEE}
(\code{\link{.bf_case_glm_gee}}).
Two stages. Stage 1 forms a warm start by running damped Newton-Raphson
on the unwhitened working response: each outer iteration computes
working responses and weights at the current linear predictor, then
calls \code{\link{.bf_inner_weighted}} for the inner backfitting loop.
Stage 2 refines the warm start on the full whitened system using the
damped SQP loop (\code{\link{.bf_sqp_loop}}), replicating the approach
used by \code{\link{.get_B_gee_glm}}.

\strong{Case (d): GLM without GEE}
(\code{\link{.bf_case_glm_no_corr}}).
A damped Newton-Raphson outer loop updates working responses and weights,
while each inner iteration alternates between weighted spline and
weighted flat updates via \code{\link{.bf_inner_weighted}}. Deviance is
monitored across outer iterations for convergence and damping.
}

\subsection{Inequality Constraints and Reassembly}{
Because flat coefficients are shared by construction, the corresponding
equality constraints are satisfied exactly. Smoothness constraints on the
spline block are enforced by the spline-only Lagrangian projection.
After backfitting convergence, inequality constraints are enforced via
the same partition-wise active-set or dense SQP refinement used by
\code{\link{get_B}}. The method is selected automatically by
\code{.solver_detect_qp_global}: block-separable constraints use the
active-set method through \code{\link{.active_set_refine}}, while
cross-partition constraints trigger the dense SQP loop through
\code{\link{.bf_sqp_loop}}. For GEE (Case c), inequality handling
occurs inside Stage 2 on the whitened system.

After convergence, the shared flat vector \eqn{\mathbf{v}} is copied
into each partition's coefficient vector, yielding
\eqn{\boldsymbol{\beta}_k =
[\tilde{\boldsymbol{\beta}}_{\mathrm{spline}}^{(k)\top},
\mathbf{v}^{\top}]^{\top}} for compatibility with downstream inference.
If \code{\link{blockfit_solve}} throws an error, a warning is issued
and the code falls back to \code{\link{get_B}}.
}
}

\section{Knot Selection and Partitioning}{


The topic of knot selection is not the main focus of the package, but the
partition structure is central because every later design matrix, penalty,
and smoothness constraint depends on it. The defaults in \pkg{lgspline} are
therefore meant to be practical and transparent rather than theoretically
final.

\subsection{Univariate Case}{
For a single predictor, the default partitioning is now handled by
\code{\link{make_partitions}} in the same \eqn{k}-means framework used more
generally: \eqn{K+1} centers are fit on an internally standardized copy
of the predictor, controlled by \code{standardize_predictors_for_knots}, and
then returned on the raw scale. Custom knots can still be supplied via
\code{custom_knots}, in which case partition assignment is built directly
from those raw-scale breakpoints. The default number of knots \eqn{K} is
chosen adaptively based on \eqn{N}, \eqn{p}, \eqn{q}, and the GLM family.
For multivariate fits, the resulting partition metadata are returned as
\code{make_partition_list} and can be re-used in later calls to
\code{\link{lgspline}}. This is particularly useful when one wants to hold
the partition geometry fixed across repeated fits, for example while varying
penalties, families, or correlation structures.
}
\subsection{Multivariate Case}{
For multiple predictors, \eqn{K+1} cluster centers are identified by
\eqn{k}-means on an internally standardized predictor matrix via
\code{\link{make_partitions}}. This is the partitioning mechanism used to
determine the multivariate spline regions; see MacQueen (1967) for the
classical clustering formulation and Kisi et al. (2025) for a recent applied
example of \eqn{k}-means-driven partitioning in a nonlinear prediction
setting. Midpoints between neighboring centers
(those whose midpoint  does not fall into a third cluster) serve as knot
locations. Observations are assigned to the nearest cluster center using
\code{\link[FNN]{get.knnx}}, and the returned centers and knots are on
the original predictor scale. The resulting partition structure is a type
of Voronoi diagram and is stored in the fitted object as
\code{make_partition_list}. The \code{do_not_cluster_on_these} argument can
exclude certain predictors from clustering (e.g., a treatment indicator that
should not drive partitioning). The lower-level clustering behavior can be
further controlled by \code{cluster_args} and \code{neighbor_tolerance},
while \code{cluster_on_indicators} determines whether binary predictors are
allowed to influence the partition geometry at all.
}

\subsection{Standardizing Predictors}{
Higher-order polynomial terms can dramatically inflate or deflate the
magnitude of basis expansions, introducing numerical instability. All
polynomial basis expansions are scaled by
\eqn{q_{0.69} - q_{0.31}}, where \eqn{q_{\zeta}} is the \eqn{\zeta}-th
quantile of the expansion. For a standard normal distribution this quantity
is approximately 1, so the scaling is close to one standard deviation for
symmetric distributions. This fitting-stage rescaling is controlled by
\code{standardize_expansions_for_fitting}, while knot construction is
controlled separately by \code{standardize_predictors_for_knots}. The same
scaling is applied to the constraint matrix to maintain smoothness, and
coefficients are back-transformed to the original scale after fitting.
}
}

\section{Smoothing Spline Penalty}{


\subsection{Penalty Construction}{
The penalty matrix \eqn{\boldsymbol{\Lambda}_s} penalizes the integrated
squared total curvature of the fitted function over the observed predictor
ranges. This is the step that makes the piecewise polynomial fit genuinely
behave like a smoothing spline rather than merely a constrained regression
spline. The package computes this penalty directly from the monomial
structure of the basis rather than by appealing to a pre-tabulated spline
basis. For a single partition \eqn{k} with basis expansion
\eqn{\mathbf{x}_k = (\phi_1(\mathbf{t}), \ldots, \phi_p(\mathbf{t}))^{\top}}
where each \eqn{\phi_i(\mathbf{t}) = \prod_{j=1}^{q} t_j^{\alpha_{ij}}}
is a multivariate monomial:
\deqn{\boldsymbol{\beta}_k^{\top}\boldsymbol{\Lambda}_s\boldsymbol{\beta}_k
  = \int_{\mathbf{a}}^{\mathbf{b}}
    \|\tilde{f}_k''(\mathbf{t})\|^{2}\,d\mathbf{t},}
where \eqn{\mathbf{a}} and \eqn{\mathbf{b}} are the observed predictor
minimums and maximums (computed globally from the data, not
partition-specific), and
\eqn{\tilde{f}_k(\mathbf{t}) = \mathbf{x}_k^{\top}\boldsymbol{\beta}_k}
is the fitted function for partition \eqn{\mathcal{P}_k}.

\strong{Total curvature operator.}
The integrated squared second derivative decomposes into \eqn{q}
curvature operators, one per predictor. For predictor \eqn{v}, the
curvature operator \eqn{D_v} is defined as
\deqn{D_v = \frac{\partial^{2}}{\partial t_v^{2}}
  + \sum_{s \neq v}\frac{\partial^{2}}{\partial t_v\,\partial t_s}.}
That is, \eqn{D_v} captures both the pure second derivative with respect
to \eqn{t_v} and all mixed second partial derivatives involving \eqn{t_v}.
The penalty matrix entries are then
\deqn{[\boldsymbol{\Lambda}_s]_{ij}
  = \sum_{v=1}^{q}\int_{\mathbf{a}}^{\mathbf{b}}
    D_v(\phi_i)\,D_v(\phi_j)\,d\mathbf{t}.}

\strong{Monomial derivative rule.}
For a monomial \eqn{\phi(\mathbf{t}) = \prod_j t_j^{\alpha_j}}, the
derivatives entering \eqn{D_v} have closed forms. The pure second
derivative is
\deqn{\frac{\partial^{2}}{\partial t_v^{2}}\prod_j t_j^{\alpha_j}
  = \alpha_v(\alpha_v - 1)\,t_v^{\alpha_v - 2}\prod_{j \neq v} t_j^{\alpha_j},}
which is zero when \eqn{\alpha_v < 2}. The mixed second derivative is
\deqn{\frac{\partial^{2}}{\partial t_v\,\partial t_s}\prod_j t_j^{\alpha_j}
  = \alpha_v\alpha_s\,t_v^{\alpha_v - 1}t_s^{\alpha_s - 1}
  \prod_{j \neq v,s} t_j^{\alpha_j},}
which is zero when \eqn{\alpha_v < 1} or \eqn{\alpha_s < 1}. Applying
\eqn{D_v} to a monomial \eqn{\phi_i} produces a sum of monomials with
known coefficients and exponent vectors.

\strong{Factorized integration.}
Because every \eqn{D_v(\phi_i)} is polynomial, the product
\eqn{D_v(\phi_i)\,D_v(\phi_j)} is also polynomial and the multivariate
integral factorizes over predictors:
\deqn{\int_{\mathbf{a}}^{\mathbf{b}}\prod_{j=1}^{q} t_j^{e_j}\,d\mathbf{t}
  = \prod_{j=1}^{q}\frac{b_j^{e_j+1} - a_j^{e_j+1}}{e_j + 1}.}
Crucially, this integral runs over \emph{all} \eqn{q} predictor ranges,
including predictors that do not appear in the integrand (for which
\eqn{e_j = 0} and the factor reduces to \eqn{b_j - a_j}). This ensures
that the penalty is properly scaled relative to the volume of the
predictor space.

\strong{Single-predictor verification.}
For \eqn{q = 1} with expansion
\eqn{\mathbf{x} = (1, t, t^{2}, t^{3})^{\top}} on \eqn{[a, b]}, the
curvature operator reduces to \eqn{D_1 = \partial^{2}/\partial t^{2}} (no
mixed partials exist), and the penalty matrix reduces to
\deqn{\boldsymbol{\Lambda}_s
  = \int_a^b \mathbf{x}''\mathbf{x}''^{\top}\,dt
  = \begin{pmatrix}
      0 & 0 & 0 & 0 \\
      0 & 0 & 0 & 0 \\
      0 & 0 & 4(b - a) & 6(b^{2} - a^{2}) \\
      0 & 0 & 6(b^{2} - a^{2}) & 12(b^{3} - a^{3})
    \end{pmatrix},}
equivalent to the classical cubic smoothing spline penalty originally proposed by Reinsch.

\strong{Handling of non-spline predictors.}
Predictors specified via \code{just_linear_without_interactions} or
\code{just_linear_with_interactions} do not receive higher-order
polynomial expansions in the design matrix. To ensure their curvature
contributions are still correctly computed (particularly through
interaction terms), the implementation temporarily appends phantom
higher-order columns (with zero data) for these predictors, computes
the full curvature penalty on the augmented basis, and then subsets the
result back to the original \eqn{p \times p} dimensions. This ensures
that interaction terms involving non-spline predictors receive appropriate
penalty contributions without affecting the rest of the estimation
pipeline.

\strong{Parallel computation.}
Because the total penalty is an additive sum over predictors
(\eqn{\boldsymbol{\Lambda}_s = \sum_{v=1}^{q}\boldsymbol{\Lambda}_{s,v}}),
the computation can be parallelized by distributing the per-predictor
curvature matrices across workers via \code{parallel::parLapply} and
summing the results. This is controlled by the \code{parallel_penalty}
argument and is beneficial when \eqn{q} is large.

The penalty is computed by \code{\link{get_2ndDerivPenalty}} (single
predictor or subset) and \code{\link{get_2ndDerivPenalty_wrapper}}
(full assembly with optional parallelism and non-spline handling).

Because the smoothing penalty has zero eigenvalues for the intercept and
linear terms (whose second derivatives vanish), an optional ridge penalty on
lower-order terms is added for computational stability. The full penalty
block for partition \eqn{k} is:
\deqn{\boldsymbol{\Lambda}_k = \lambda_w\bigl(\boldsymbol{\Lambda}_s + \lambda_r\boldsymbol{\Lambda}_r + \sum_{m=1}^{M}\xi_{mk}\mathbf{L}_{mk}\bigr)}
where \eqn{\lambda_w} is the global wiggle penalty (\code{wiggle_penalty}),
\eqn{\lambda_r} is ridge penalty on linear and intercept terms
(\code{flat_ridge_penalty}) multiplied by the wiggle penalty, and
\eqn{\xi_{mk}} and \eqn{\mathbf{L}_{mk}} denote optional additional
penalty multipliers and matrices, including the predictor- and
partition-specific components activated through
\code{unique_penalty_per_predictor}, \code{unique_penalty_per_partition},
\code{predictor_penalties}, and \code{partition_penalties}. This assembly is handled by
\code{\link{compute_Lambda}}.

The penalty matrix \eqn{\boldsymbol{\Lambda}} is stored as a list
of \eqn{K+1} \eqn{p \times p} square, symmetric, positive semi-definite
matrices.
}

\subsection{Penalty Optimization via Leave-One-Out or Generalized Cross-Validation}{
After the structural pieces of the model are fixed, the main remaining
question is how much smoothing to apply. In \pkg{lgspline}, that tuning is
performed with exact leave-one-out by default, or with generalized
cross-validation when \code{tuning_criterion = "gcv"}. In either case, the
criterion is evaluated using the same constrained estimator that will be
used in the final model fit.
Penalty parameters are estimated on the log scale via exponential
parameterization (\eqn{\lambda = \exp(\theta)},
\eqn{\theta \in \mathbb{R}}), ensuring positivity. The chain rule factor
\eqn{\partial\exp(\theta)/\partial\theta = \exp(\theta) = \lambda} is
applied throughout. User-facing arguments (\code{initial_wiggle},
\code{initial_flat}, \code{predictor_penalties},
\code{partition_penalties}) accept values on the raw, natural scale;
conversion to log scale is handled internally. The final tuned values and
assembled penalty pieces are returned in the fitted object's
\code{penalties} component.

The total penalty matrix \eqn{\boldsymbol{\Lambda}} is constructed as:
\deqn{\boldsymbol{\Lambda} = \lambda_w\mathbf{L}_{w} + \lambda_r\mathbf{L}_{r}
  + \sum_{j}\nu_j\mathbf{L}_j^{(\mathrm{pred})}
  + \sum_{k}\tau_k\mathbf{L}_k^{(\mathrm{part})},}
where \eqn{\mathbf{L}_{w}} is the integrated squared second-derivative
penalty (i.e., \eqn{\boldsymbol{\Lambda}_s} above), \eqn{\mathbf{L}_{r}} is
a ridge penalty on intercept and linear coefficients,
\eqn{\mathbf{L}_j^{(\mathrm{pred})}} are predictor-specific penalties, and
\eqn{\mathbf{L}_k^{(\mathrm{part})}} are partition-specific penalties. The
scalars \eqn{\lambda_w} (\code{wiggle_penalty}), \eqn{\lambda_r}
(\code{flat_ridge_penalty}), \eqn{\{\nu_j\}}
(\code{predictor_penalties}), and \eqn{\{\tau_k\}}
(\code{partition_penalties}) are tuned.
With \code{tuning_criterion = "loo"}, the criterion is
\deqn{\mathrm{LOO}
  = \frac{1}{N}\sum_{i=1}^{N}\left(\frac{r_i}{1-h_{ii}}\right)^2,}
where \eqn{h_{ii}} is the diagonal of the tuning hat matrix. For Gaussian
identity-link tuning with fixed transformed design, this is exact. With
correlation structure and/or non-identity links, \pkg{lgspline} applies the
same transformed or linearized tuning problem already used by GCV, so the
leave-one-out calculation is carried out on that working scale. In empirical
diagnostics, the criterion itself behaves stably, while the fully analytic
derivative of the observation-wise leverage term can be numerically delicate
in some problems; when that sensitivity matters in practice, the finite-
difference optimizer path (\code{use_custom_bfgs = FALSE}) remains
available. This exact observation-wise calculation is also the main reason
LOO tuning becomes more expensive than GCV as \eqn{N} grows; in routine use,
\code{tuning_criterion = "gcv"} is often the more practical choice once the
sample size is above about 250,000.

With \code{tuning_criterion = "gcv"}, the package instead uses
\deqn{\mathrm{GCV}_{u,\gamma}
  = \frac{\sum_{i=1}^{N}D_{ii}\,r_i^{2}}
          {N(1 - \gamma\bar{W})^{2}}}
where \eqn{\bar{W} = \mathrm{tr}(\mathbf{H})/N} is the mean of the
hat-matrix diagonal. The symbol \eqn{\gamma} is set by
\code{gcv_gamma} and controls optional inflation of the
effective-degrees-of-freedom term.

For identity link, \eqn{r_i = y_i - \hat{\eta}_i}. For non-identity links,
the residuals are
\eqn{r_i = g((y_i + \delta)/(1+2\delta)) - (\hat{\eta}_i + \delta)/(1+2\delta)},
where \eqn{\delta \geq 0} is a pseudocount that stabilizes the link
transformation, automatically tuned within \code{\link{tune_Lambda}} if not supplied
to \code{delta}.
Non-Gaussian families with observation weights
\eqn{\omega_i} have their residuals scaled by \eqn{\omega_i}. When the
family provides a custom deviance residual function, that function is used
in place of the link-scale residuals.

\strong{Pseudocount selection.} The pseudocount \eqn{\delta} is chosen to
make the transformed response distribution most closely approximate a
\eqn{t}-distribution with \eqn{N-1} degrees of freedom, in the sense of
minimizing the (optionally weighted) mean absolute deviation between the
sorted standardized transformed responses and the corresponding
\eqn{t}-quantiles. This is solved via Brent's method over
\eqn{[10^{-64}, 1]}. When the link is identity, or when the response
naturally lies in the domain of the link function, \eqn{\delta = 0}.
This behavior is exposed through the \code{delta} argument in
\code{\link{lgspline}}: supplying a fixed numeric value bypasses the internal
search, while leaving it \code{NULL} allows the tuning code to choose the
stabilizing pseudocount automatically when needed.

\strong{Meta-penalty regularization.} A regularization term pulls the
predictor- and partition-specific penalty parameters toward 1 on the raw
scale:
\deqn{P_{\mathrm{meta}}(\lambda_w, \nu_j, \tau_k)
  = \frac{1}{2}c_{\mathrm{meta}}\sum_j(\nu_j - 1)^{2}
  + \frac{1}{2}\cdot 10^{-32}(\lambda_w - 1)^{2}}
where \eqn{c_{\mathrm{meta}}} is a user-specified coefficient
(\code{meta_penalty}). The gradient of \eqn{P_{\mathrm{meta}}} on the log
scale, incorporating the exp chain rule, is
\eqn{\partial P_{\mathrm{meta}}/\partial\theta_j = c_{\mathrm{meta}}(\nu_j - 1)\nu_j}
and
\eqn{\partial P_{\mathrm{meta}}/\partial\theta_1 = 10^{-32}(\lambda_w - 1)\lambda_w}.
The total objective is the selected tuning criterion plus
\eqn{P_{\mathrm{meta}}}.
}

\subsection{Closed-Form Gradients for Tuning Criteria}{
The gradient of the GCV objective with respect to
\eqn{\theta_1 = \log\lambda_w} is computed analytically via the quotient
rule:
\deqn{\frac{\partial\mathrm{GCV}_{u,\gamma}}{\partial\theta_1}
  = \frac{1}{D^{2}}\left(\frac{\partial\mathcal{N}}{\partial\theta_1}D
  - \mathcal{N}\frac{\partial D}{\partial\theta_1}\right),}
where \eqn{\mathcal{N} = \sum r_i^{2}} (numerator) and
\eqn{D = N(1 - \gamma\bar{W})^{2}} (denominator). The key intermediates are:
\itemize{
  \item \eqn{\partial\mathbf{G}/\partial\lambda_w}, computed from the
    matrix identity
    \eqn{\partial(\mathbf{X}^{\top}\mathbf{X} + \boldsymbol{\Lambda})^{-1}/\partial\lambda = -\mathbf{G}(\partial\boldsymbol{\Lambda}/\partial\lambda)\mathbf{G}}.
  \item \eqn{\partial\mathbf{G}^{1/2}/\partial\lambda_w}, derived from
    \eqn{\partial\mathbf{G}/\partial\lambda_w} via the eigendecomposition
    chain rule.
  \item \eqn{\partial\bar{W}/\partial\lambda_w}, the derivative of the
    trace of the hat matrix
    \eqn{\mathbf{H} = \mathbf{X}\mathbf{U}\mathbf{G}\mathbf{X}^{\top}},
    which depends on both
    \eqn{\partial\mathbf{G}/\partial\lambda_w} and
    \eqn{\partial\mathbf{G}^{1/2}/\partial\lambda_w}.
  \item \eqn{\partial\mathcal{N}/\partial\theta_1 = -2\mathbf{r}^{\top}\mathbf{X}(\partial(\mathbf{U}\mathbf{G})/\partial\lambda_w)\mathbf{X}^{\top}\mathbf{y}\cdot\lambda_w},
    via the chain rule applied to the residual vector.
  \item \eqn{\partial D/\partial\theta_1 =
    2(1 - \gamma\bar{W})(-\gamma\,\partial\bar{W}/\partial\lambda_w)\cdot\lambda_w}.
}
In the implementation, these quantities are assembled by a small set of
helper routines: \code{\link{compute_dG_dlambda}} for
\eqn{\partial\mathbf{G}/\partial\lambda}, \code{\link{compute_dGhalf}}
for \eqn{\partial\mathbf{G}^{1/2}/\partial\lambda},
\code{\link{compute_trace_H}} together with the tuning hat-diagonal
derivative helpers for the effective-degrees-of-freedom term, and
\code{\link{compute_dG_u_dlambda_xy}} for the derivative of the fitted-value
quadratic form.

For exact leave-one-out, the implementation instead differentiates the
per-observation PRESS form directly, requiring both
\eqn{\partial r_i / \partial\lambda_w} and
\eqn{\partial h_{ii} / \partial\lambda_w}. The latter is assembled
partition-wise from the same constrained-\eqn{\mathbf{G}} ingredients used
elsewhere in tuning. The exact LOO path does not explicitly form the full
projection matrix \eqn{\mathbf{U}} or the full hat matrix \eqn{\mathbf{H}};
instead it works with blockwise matrices
\eqn{\mathbf{C}_k = \mathbf{G}_k - \mathbf{G}_k\mathbf{A}_k(\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}\mathbf{A}_k^\top\mathbf{G}_k}
and their derivatives. Empirically, this observation-wise leverage
derivative is the most numerically delicate part of the analytic LOO
gradient: most observations are typically well behaved, but a diffuse set of
small local errors can remain even when the overall LOO objective and the
corresponding GCV gradient are stable.

The shared wiggle and flat directions are differentiated directly. For
predictor- and partition-specific penalties, a lower-cost ratio heuristic is
used:
\deqn{\frac{\partial\mathrm{GCV}_{u,\gamma}}{\partial\theta_l} \approx
  \frac{\mathrm{mean}(\mathrm{diag}(\mathbf{L}_l))}{\mathrm{mean}(\mathrm{diag}(\boldsymbol{\Lambda}))}
  \frac{\partial\mathrm{GCV}_{u,\gamma}}{\partial\theta_w},}
where \eqn{\theta_l = \log \lambda_l} and
\eqn{\mathbf{L}_l} denotes the current contribution of the corresponding
predictor- or partition-specific penalty to \eqn{\boldsymbol{\Lambda}}. This
keeps the extra penalty directions inexpensive while preserving the direct
derivatives for the shared wiggle and flat penalties.
}

\subsection{Optimization Procedure}{
\strong{Grid search initialization.} The selected tuning criterion is
evaluated over a grid of candidate values for
\eqn{(\lambda_w, \lambda_r)} on the log scale. All combinations of
user-supplied candidate vectors (\code{initial_wiggle} and
\code{initial_flat}) are formed, and the combination yielding
the smallest finite criterion value is selected as the starting
point for BFGS optimization. Grid points producing non-finite
objective values are discarded. If all grid points fail, an error
is raised advising the user to check the data or adjust the grid.

\strong{Damped BFGS optimizer.} A custom damped BFGS quasi-Newton optimizer,
implemented in \code{\link{tune_Lambda}} through
\code{.damped_bfgs()}, minimizes the selected tuning criterion plus
\eqn{P_{\mathrm{meta}}}. When analytic gradients are not preferred, or when
the exact LOO leverage derivative appears too noisy for a given problem, a
base-R \code{\link[stats]{optim}} call with method \code{"BFGS"} provides
the finite-difference fallback.

\emph{Iterations 1-2: steepest descent.} The first two iterations use
steepest descent with a damping factor \eqn{\alpha}:
\eqn{\boldsymbol{\phi}^{(t+1)} = \boldsymbol{\phi}^{(t)} - \alpha\nabla_{\boldsymbol{\phi}}}.

\emph{Iterations 3+: BFGS.} From iteration 3, an inverse Hessian
approximation \eqn{\mathbf{J}^{(t)}} is maintained via the standard secant
update. Let
\eqn{\mathbf{s}^{(t)} = \boldsymbol{\phi}^{(t)} - \boldsymbol{\phi}^{(t-1)}}
and
\eqn{\mathbf{v}^{(t)} = \nabla^{(t)} - \nabla^{(t-1)}}. The BFGS update
is:
\deqn{\mathbf{J}^{(t+1)}
  = (\mathbf{I} - \mathbf{u}\mathbf{s}\mathbf{v}^{\top})
  \mathbf{J}^{(t)}
  (\mathbf{I} - \mathbf{u}\mathbf{v}\mathbf{s}^{\top})
  + \mathbf{u}\mathbf{s}\mathbf{s}^{\top},
  \qquad \mathbf{u} = (\mathbf{v}^{\top}\mathbf{s})^{-1}.}
When \eqn{|\mathbf{v}^{\top}\mathbf{s}| < 10^{-64}}, the approximation is
reset to \eqn{\mathbf{I}} and the iteration is flagged for restart. The
search direction is
\eqn{\mathbf{d}^{(t)} = -\mathbf{J}^{(t)}\nabla^{(t)}}.

\emph{Step acceptance.} A step is accepted if the new tuning criterion value
is no larger than the old one.
On rejection, \eqn{\alpha} is halved. If \eqn{\alpha < 2^{-10}} (early
iterations) or \eqn{\alpha < 2^{-12}} (later iterations), the optimizer
terminates with the best solution found.

\emph{Convergence.} The optimizer terminates when
the absolute change in the tuning criterion is less than \eqn{\epsilon}, or
\eqn{\|\boldsymbol{\phi}^{(t)} - \boldsymbol{\phi}^{(t-1)}\|_{\infty} < \epsilon},
provided at least 10 iterations have elapsed, for penalties
\eqn{\boldsymbol{\phi} = \lambda_w, \lambda_r, \nu_j, \tau_j, ...}.

\emph{Alternative.} A base-R
\code{\link[stats]{optim}} call with method \code{"BFGS"} and
finite-difference gradients can be used instead via
\code{use_custom_bfgs = FALSE}.

\strong{Post-optimization sample-size adjustment.} After optimization, the
tuned penalty parameters are divided by \eqn{((N+2)/(N-2))^{2}}
(equivalently multiplied by \eqn{((N-2)/(N+2))^{2}}) for both GCV- and
LOO-based tuning. This decreases the final penalties at small sample sizes.

The tuning loop is implemented in \code{\link{tune_Lambda}}.
}
}

\section{Incorporating Non-Spline Effects}{


Multiple fixed effects are accommodated naturally in the LMSS framework
because spline effects, linear effects, and many interaction terms all live
in the same partition-wise polynomial expansion. The distinction is therefore
not whether a term is "allowed" by the solver, but whether it receives full
spline treatment or remains structurally linear across partitions.

The constrained framework naturally accommodates non-spline terms. If only
linear terms are included for a predictor (via
\code{just_linear_without_interactions} or
\code{just_linear_with_interactions}), the first-derivative smoothing
constraint forces the linear coefficient to be identical across all
partitions, since the derivative of a linear function is its slope. This
is not an algorithmic modification but a natural consequence of the
constraint structure.

For example, a model with one spline effect and a linear treatment indicator
interaction will naturally keep the treatment-time interaction coefficient
constant across partitions while allowing the time effect to vary
nonlinearly. This conveniently extends to arbitrary combinations of spline
and linear terms without requiring special handling.

When \code{blockfit = TRUE} is specified alongside
\code{just_linear_without_interactions}, the flat-block path provides an
alternative enforcement mechanism. Rather than relying on constraint
projection, flat coefficients are pooled structurally across partitions
during backfitting. The two approaches agree at the point estimate but
differ in their uncertainty quantification; see the Blockfit section above.
}

\section{Integration}{


Because the fitted object retains an explicit polynomial representation in
each partition, numerical integration can be carried out in a fairly direct
way. The package wraps that calculation in a user-facing S3 method so the
user does not need to manage knot boundaries or partition membership by hand.

In the user-facing interface, numerical integration is applied through
\code{\link{integrate.lgspline}}, which applies Gauss-Legendre quadrature to
predictions from the fitted model produced by \code{\link{predict.lgspline}}.

\subsection{Implementation}{
For a user-supplied rectangular domain, \code{\link{integrate.lgspline}} constructs a
tensor-product grid of Gauss-Legendre nodes, evaluates the fitted model at
those points, and forms the weighted sum. This works for both univariate and
multivariate models, respects the fitted partition structure automatically,
and avoids requiring the user to keep track of knot boundaries by hand.

The \code{vars} argument selects which predictors are integrated over.
Predictors not listed in \code{vars} are held fixed at
\code{initial_values} when supplied, or otherwise at the midpoint of their
observed training range. The optional \code{B_predict} argument makes it
possible to integrate posterior draws or other alternate coefficient sets,
and \code{n_quad} controls the number of Gauss-Legendre nodes used per
integrated dimension.

Integration is performed on the response scale by default. Setting
\code{link_scale = TRUE} instead integrates the linear predictor
\eqn{\eta = f(\mathbf{t})}. For identity-link Gaussian models the two scales
coincide.
}
}

\section{Lagrange Multipliers}{


When \code{return_lagrange_multipliers = TRUE}, the multiplier vector
\deqn{\boldsymbol{\lambda} = (\mathbf{A}^{\top}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{\top}\hat{\boldsymbol{\beta}}}
is returned. These quantify the sensitivity of the penalized objective to
relaxing each smoothness or user-supplied equality constraint. When
constraint target values are nonzero
(\eqn{\mathbf{A}^{\top}\boldsymbol{\beta}_0 \neq \mathbf{0}}), the modified
formulation is used:
\deqn{\boldsymbol{\lambda} = (\mathbf{A}^{\top}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{\top}(\hat{\boldsymbol{\beta}} - \boldsymbol{\beta}_0)}
where \eqn{\mathbf{A}^{\top}\boldsymbol{\beta}_0} is the vector of
constraint target values. Multipliers are \code{NULL} when no constraints
are active (\eqn{\mathbf{A}} is \code{NULL} or \eqn{K = 0}).

For inequality constraints, multipliers are returned as computed by
\code{\link[quadprog]{solve.QP}}. The Lagrange multipliers for active
inequality constraints can be used diagnostically to identify which shape
constraints are most costly in terms of goodness of fit.
}

\section{S3 Methods}{


Standard S3 methods are provided for objects of class \code{lgspline}:
\itemize{
  \item \code{\link{print.lgspline}} and \code{\link{summary.lgspline}}:
    Provide concise model summaries, with
    \code{\link{print.summary.lgspline}} formatting coefficient tables in a
    familiar regression-style layout.

  \item \code{\link{logLik.lgspline}}: Returns a standard \code{logLik}
    object. For Gaussian responses with identity link, the exact
    log-likelihood is computed. When a correlation structure is present via
    \code{VhalfInv}, the log-likelihood includes the
    \eqn{\log|\mathbf{V}^{-1/2}|} adjustment and the corresponding
    whitened quadratic form. For other families, the method falls back to
    \code{family$aic} or a deviance-based approximation. An
    \code{include_prior} argument (default \code{TRUE}) optionally adds
    the Gaussian prior penalty interpretation of the smoothing spline
    penalty
    \eqn{-\frac{1}{2\sigma^{2}}\tilde{\boldsymbol{\beta}}^{\top}\boldsymbol{\Lambda}\tilde{\boldsymbol{\beta}}}
    to obtain a penalized MAP log-likelihood.

  \item \code{\link{predict.lgspline}}: Produces fitted values and related
    quantities (e.g., derivatives and standard errors through
    \code{se.fit = TRUE}), lets \code{new_predictors} override
    \code{newdata}, accepts alternate coefficient lists through
    \code{B_predict}, and supports prediction on new predictor matrices
    consistent with the original spline expansions.

  \item \code{\link{coef.lgspline}}: Extracts partition-specific
    coefficient vectors.

  \item \code{\link{confint.lgspline}}: Extracts confidence intervals.
    When the inverse Hessian from BFGS optimization is available for
    correlation parameters, intervals for those correlation parameters are
    returned on the working (transformed) scale and should be
    back-transformed as described in the correlation section.
  \item \code{\link{plot.lgspline}}: For one-dimensional fits, produces
    base R graphics showing the fitted function (with optional partition-wise
    formulas) and supports overlay via \code{add = TRUE}. For two or more
    predictors, an interactive \pkg{plotly}-based visualization is
    returned. Specific predictors may be selected via \code{vars}.

  \item \code{\link{integrate.lgspline}}: Computes definite integrals of
    the fitted surface over rectangular domains by Gauss-Legendre
    quadrature.
}

Additional user-facing helpers include \code{\link{wald_univariate}} for
coefficient-wise Wald inference, \code{\link{generate_posterior}} for
posterior and posterior-predictive sampling,
\code{\link{generate_posterior_correlation}} for correlation-aware posterior
simulation, \code{\link{equation}} for closed-form display of the fitted
partition formulas, and
\code{\link{find_extremum}} for optimizing the fitted surface or a custom
acquisition function built from it.
}

\references{
Buse, A. and Lim, L. (1977). Cubic Splines as a Special Case of
Restricted Least Squares. \emph{Journal of the American Statistical
Association}, 72, 64-68.

Eilers, P. H. and Marx, B. D. (1996). Flexible Smoothing with B-splines
and Penalties. \emph{Statistical Science}, 11(2), 89-121.

Ezhov, N., Neitzel, F. and Petrovic, S. (2018). Spline Approximation,
Part 1: Basic Methodology. \emph{Journal of Applied Geodesy}, 12(2),
139-155.

Goldfarb, D. and Idnani, A. (1983). A Numerically Stable Dual Method for
Solving Strictly Convex Quadratic Programs. \emph{Mathematical
Programming}, 27(1), 1-33.

Harville, D. A. (1977). Maximum Likelihood Approaches to Variance
Component Estimation and to Related Problems. \emph{Journal of the
American Statistical Association}, 72(358), 320-338.

Hastie, T. J. and Tibshirani, R. J. (1990). \emph{Generalized Additive
Models}. Chapman & Hall/CRC.

Kisi, O., Heddam, S., Parmar, K. S., Petroselli, A., Kulls, C. and
Zounemat-Kermani, M. (2025). Integration of Gaussian Process Regression
and K Means Clustering for Enhanced Short Term Rainfall Runoff Modeling.
\emph{Scientific Reports}, 15, 7444.

MacQueen, J. B. (1967). Some Methods for Classification and Analysis of
Multivariate Observations. In \emph{Proceedings of the Fifth Berkeley
Symposium on Mathematical Statistics and Probability}, Volume 1,
281-297. University of California Press.

McCullagh, P. and Nelder, J. A. (1989). \emph{Generalized Linear Models}.
Chapman & Hall, 2nd edition.

Murray, I., Adams, R. P. and MacKay, D. J. C. (2010). Elliptical Slice
Sampling. \emph{Proceedings of the 13th International Conference on
Artificial Intelligence and Statistics (AISTATS)}, 9, 541-548.

Nocedal, J. and Wright, S. J. (2006). \emph{Numerical Optimization}
(2nd ed.). Springer.

Kim, Y.-J. and Gu, C. (2004). Smoothing Spline Gaussian Regression:
More Scalable Computation via Efficient Approximation. \emph{Journal of
the Royal Statistical Society: Series B}, 66(2), 337-356.

Patterson, H. D. and Thompson, R. (1971). Recovery of Inter-Block
Information When Block Sizes Are Unequal. \emph{Biometrika}, 58,
545-554.

Pya, N. and Wood, S. N. (2015). Shape Constrained Additive Models.
\emph{Statistics and Computing}, 25(3), 543-559.

Reinsch, C. H. (1967). Smoothing by Spline Functions. \emph{Numerische
Mathematik}, 10, 177-183.

Ruppert, D., Wand, M. P. and Carroll, R. J. (2003).
\emph{Semiparametric Regression}. Cambridge University Press.

Searle, S. R., Casella, G. and McCulloch, C. E. (2006). \emph{Variance
Components}. Wiley.

Wahba, G. (1990). \emph{Spline Models for Observational Data}. SIAM.

Wood, S. N. (2006). On Confidence Intervals for Generalized Additive
Models Based on Penalized Regression Splines. \emph{Australian & New
Zealand Journal of Statistics}, 48(4), 445-464.

Wood, S. N. (2011). Fast Stable Restricted Maximum Likelihood and Marginal
Likelihood Estimation of Semiparametric Generalized Linear Models.
\emph{Journal of the Royal Statistical Society: Series B}, 73(1), 3-36.

Wood, S. N. (2017). \emph{Generalized Additive Models: An Introduction
with R}. CRC Press, 2nd edition.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{get_2ndDerivPenalty}
\alias{get_2ndDerivPenalty}
\title{Compute Integrated Squared Second Derivative Penalty Matrix}
\usage{
get_2ndDerivPenalty(
  colnm_expansions,
  C,
  power1_cols,
  power2_cols,
  power3_cols,
  power4_cols,
  interaction_single_cols,
  interaction_quad_cols,
  triplet_cols,
  p_expansions,
  select_cols = NULL
)
}
\arguments{
\item{colnm_expansions}{Character vector of length \eqn{p} giving
column names for the basis expansion matrix \eqn{\mathbf{X}_k}.
Each name encodes the monomial structure: predictor names appear
literally for linear terms, with \code{^d} suffixes for degree
\eqn{d}, and an \code{x} separator between factors in interaction
terms (e.g.\ \code{"_1_alphax_2_beta^2"} for
\eqn{t_1 t_2^2}).  Predictor names must not be substrings of one
another.}

\item{C}{Numeric \eqn{N \times p} matrix of basis expansions
\eqn{\mathbf{X}_k}.  Used only to read the observed range
\eqn{[a_j, b_j]} of each predictor via \code{min}/\code{max} of
the linear columns.}

\item{power1_cols}{Integer vector.  Column indices of linear terms
\eqn{t_j}.}

\item{power2_cols}{Integer vector.  Column indices of quadratic terms
\eqn{t_j^2}.}

\item{power3_cols}{Integer vector.  Column indices of cubic terms
\eqn{t_j^3}.}

\item{power4_cols}{Integer vector.  Column indices of quartic terms
\eqn{t_j^4}.}

\item{interaction_single_cols}{Integer vector.  Column indices of
linear-by-linear interaction terms \eqn{t_j t_l}.}

\item{interaction_quad_cols}{Integer vector.  Column indices of
linear-by-quadratic interaction terms \eqn{t_j t_l^2} and
\eqn{t_j^2 t_l}.}

\item{triplet_cols}{Integer vector.  Column indices of three-way
interaction terms \eqn{t_j t_l t_m}.}

\item{select_cols}{Optional integer vector of predictor indices
(positions within \code{power1_cols}) whose curvature operators
\eqn{D_v} are summed.  Defaults to all \eqn{q} predictors.}
}
\value{
A symmetric positive semi-definite \eqn{p \times p} matrix
\eqn{\boldsymbol{\Lambda}_s} with
\deqn{[\boldsymbol{\Lambda}_s]_{ij}
  = \sum_{v=1}^{q} \int_{\mathbf{a}}^{\mathbf{b}}
    D_v(\phi_i) \, D_v(\phi_j) \, d\mathbf{t}.}
}
\description{
Computes the \eqn{p \times p} integrated squared second-derivative penalty matrix
\eqn{\boldsymbol{\Lambda}_s} for one partition of a monomial spline,
such that
\deqn{\boldsymbol{\beta}_k^\top \boldsymbol{\Lambda}_s
       \boldsymbol{\beta}_k
  = \int_{\mathbf{a}}^{\mathbf{b}}
    \|\tilde{f}_k''(\mathbf{t})\|^2 \, d\mathbf{t},}
where \eqn{\mathbf{a}} and \eqn{\mathbf{b}} are the observed
predictor minimums and maximums, and
\eqn{\tilde{f}_k(\mathbf{t}) = \mathbf{x}_k^\top
\boldsymbol{\beta}_k} is the fitted function for partition
\eqn{\mathcal{P}_k}.

The implementation uses a general monomial derivative rule that
handles all term types (marginal powers, two-way interactions,
quadratic interactions, three-way interactions) in a single unified
loop.
}
\details{
\subsection{Mathematical framework}{

Let the basis expansion for partition \eqn{\mathcal{P}_k} be
\eqn{\mathbf{x}_k = (\phi_1(\mathbf{t}), \ldots,
\phi_p(\mathbf{t}))^\top} where each \eqn{\phi_i} is a
multivariate monomial
\deqn{\phi_i(\mathbf{t})
  = \prod_{j=1}^{q} t_j^{\alpha_{ij}}.}

The second derivative of \eqn{\tilde{f}_k} decomposes into
\eqn{q} total curvature operators, one per predictor.  For
predictor \eqn{v}:
\deqn{D_v = \frac{\partial^2}{\partial t_v^2}
  + \sum_{s \neq v} \frac{\partial^2}{\partial t_v \, \partial t_s}.}

The monomial derivative rule gives each second partial derivative
in closed form.  For the pure second derivative (\eqn{r = s = v}):
\deqn{\frac{\partial^2}{\partial t_v^2}
      \prod_{j} t_j^{\alpha_j}
  = \alpha_v(\alpha_v - 1) \;
    t_v^{\alpha_v - 2} \prod_{j \neq v} t_j^{\alpha_j}.}

For a mixed second derivative (\eqn{s \neq v}):
\deqn{\frac{\partial^2}{\partial t_v \, \partial t_s}
      \prod_{j} t_j^{\alpha_j}
  = \alpha_v \alpha_s \;
    t_v^{\alpha_v - 1} t_s^{\alpha_s - 1}
    \prod_{j \neq v,s} t_j^{\alpha_j}.}

In both cases a term is zero when the required exponent would be
negative (e.g.\ \eqn{\alpha_v < 2} for the pure case).  Applying
\eqn{D_v} to \eqn{\phi_i} produces a sum of monomials with known
coefficients and exponent vectors.
}

\subsection{Integration}{

Because every \eqn{D_v(\phi_i)} is polynomial, the product
\eqn{D_v(\phi_i) \, D_v(\phi_j)} is also polynomial and the
multivariate integral factorises over predictors:
\deqn{\int_{\mathbf{a}}^{\mathbf{b}}
      \prod_{j=1}^{q} t_j^{e_j} \, d\mathbf{t}
  = \prod_{j=1}^{q}
    \frac{b_j^{e_j+1} - a_j^{e_j+1}}{e_j + 1}.}

Notably, this integral runs over \emph{all} \eqn{q} predictor
ranges, including predictors that do not appear in the integrand
(for which \eqn{e_j = 0} and the factor reduces to
\eqn{b_j - a_j}).
}

\subsection{Single-predictor verification}{

For \eqn{q = 1} with expansion
\eqn{\mathbf{x} = (1, t, t^2, t^3)^\top} on \eqn{[a, b]}, the
penalty matrix reduces to
\deqn{\boldsymbol{\Lambda}_s
  = \int_a^b \mathbf{x}'' \mathbf{x}''^\top \, dt
  = \begin{pmatrix}
      0 & 0 & 0 & 0 \\
      0 & 0 & 0 & 0 \\
      0 & 0 & 4(b - a) & 6(b^2 - a^2) \\
      0 & 0 & 6(b^2 - a^2) & 12(b^3 - a^3)
    \end{pmatrix},}
matching the formula in Section 2.3.
}
}
\section{Naming convention}{

Column names in \code{colnm_expansions} must encode each monomial's
predictor content.  The parser checks each predictor name (taken
from the linear columns) against the column name:
\itemize{
  \item \code{predname^d} (literal caret) signals exponent \eqn{d}.
  \item \code{predname} without \code{^} signals exponent 1.
  \item Absence of \code{predname} signals exponent 0.
}
Multiple factors in an interaction are separated by \code{x}.
Predictor names must be unique and must not be substrings of one
another.
}

\examples{

## Verification example: 3-predictor model with all term types:
#  This example constructs the penalty matrix analytically and then
#  verifies selected entries against closed-form hand calculations.
#  Users can extend or adapt this example to audit new basis
#  expansions.

set.seed(1234)
n <- 2000
t1 <- runif(n, 1, 4)     # predictor 1, support [1, 4]
t2 <- runif(n, -2, 3)    # predictor 2, support [-2, 3]
t3 <- runif(n, 0.5, 2)   # predictor 3, support [0.5, 2]

## Predictor names (must not be substrings of one another)
pn <- c("_1_aa", "_2_bb", "_3_cc")

## Build column names encoding the monomial structure
col_names <- c(
  pn,                                                # linear
  paste0(pn, "^2"),                                  # quadratic
  paste0(pn, "^3"),                                  # cubic
  paste0(pn[1], "x", pn[2]),                         # t1*t2
  paste0(pn[1], "x", pn[3]),                         # t1*t3
  paste0(pn[2], "x", pn[3]),                         # t2*t3
  paste0(pn[2], "x", pn[1], "^2"),                   # t1^2*t2
  paste0(pn[1], "x", pn[2], "^2"),                   # t1*t2^2
  paste0(pn[1], "x", pn[2], "x", pn[3])              # t1*t2*t3
)
p_expansions <- length(col_names)   # 14 basis functions

## Build the expansion matrix C
C <- cbind(
  t1, t2, t3,
  t1^2, t2^2, t3^2,
  t1^3, t2^3, t3^3,
  t1*t2, t1*t3, t2*t3,
  t2*t1^2, t1*t2^2,
  t1*t2*t3
)
colnames(C) <- col_names

## Compute the penalty matrix
Ls <- get_2ndDerivPenalty(
  colnm_expansions       = col_names,
  C                      = C,
  power1_cols            = 1:3,
  power2_cols            = 4:6,
  power3_cols            = 7:9,
  power4_cols            = integer(0),
  interaction_single_cols = 10:12,
  interaction_quad_cols  = 13:14,
  triplet_cols           = 15,
  p_expansions           = p_expansions,
  select_cols            = 1:3
)

## Hand-computed reference values (exact, using true bounds) ---
#  Notation: dt1 = 4-1 = 3, dt2 = 3-(-2) = 5, dt3 = 2-0.5 = 1.5
#
#  Entry [4,4]: t1^2 diagonal.
#    D_1(t1^2) = 2.  No other D_v contributes.
#    integral (2)^2 dt1 dt2 dt3 = 4 * 3 * 5 * 1.5 = 90
#
#  Entry [10,10]: t1*t2 diagonal.
#    D_1(t1*t2) = 1 (mixed d^2/dt1 dt2).
#    D_2(t1*t2) = 1 (mixed d^2/dt2 dt1).
#    integral 1 dt1 dt2 dt3 + integral 1 dt1 dt2 dt3 = 2*3*5*1.5 = 45
#
#  Entry [15,15]: t1*t2*t3 diagonal.
#    D_1(t1*t2*t3) = t3 + t2  (mixed partials d^2/dt1 dt2 and d^2/dt1 dt3)
#    D_2(t1*t2*t3) = t1 + t3  (similarly)
#    D_3(t1*t2*t3) = t1 + t2  (similarly)
#    Full integral = sum of 3 terms = 120 + 337.5 + 266.25 = 723.75

## Compare (allowing ~0.5\% tolerance for data-derived bounds)
cat("Entry [4,4]:   analytical =", round(Ls[4,4], 2),
    "  exact = 90.00\n")
cat("Entry [10,10]: analytical =", round(Ls[10,10], 2),
    "  exact = 45.00\n")
cat("Entry [15,15]: analytical =", round(Ls[15,15], 2),
    "  exact = 723.75\n")
cat("Symmetric:", isSymmetric(Ls), "\n")


}
\references{
Reinsch, C. H. (1967). Smoothing by spline functions.
\emph{Numerische Mathematik}, 10(3), 177--183.
}
\seealso{
\code{\link{lgspline}} for the full model fitting
  interface.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{\%**\%}
\alias{\%**\%}
\title{Efficient Matrix Multiplication Operator}
\usage{
x \%**\% y
}
\arguments{
\item{x}{Left matrix}

\item{y}{Right matrix}
}
\value{
Matrix product of x and y
}
\description{
Operator wrapper around C++ \code{efficient_matrix_mult()} for matrix multiplication syntax.

This is an internal function meant to provide improvement over base R's operator for
certain large matrix operations, at a cost of potential slight slowdown for
smaller problems.
}
\examples{

M1 <- matrix(1:4, 2, 2)
M2 <- matrix(5:8, 2, 2)
M1 \%**\% M2


}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{integrate}
\alias{integrate}
\alias{integrate.default}
\title{Generic for Numerical Integration}
\usage{
integrate(f, ...)

\method{integrate}{default}(f, ...)
}
\arguments{
\item{f}{A fitted model object or a function.}

\item{...}{Arguments passed to methods.}
}
\description{
S3 generic that dispatches to \code{integrate.lgspline} for fitted
\code{lgspline} objects and falls back to \code{\link[stats]{integrate}}
for ordinary functions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{damped_newton_r}
\alias{damped_newton_r}
\title{Damped Newton-Raphson Parameter Optimization}
\usage{
damped_newton_r(
  parameters,
  loglikelihood,
  gradient,
  neghessian,
  tol = 1e-07,
  max_cnt = 64,
  max_dmp_steps = 16
)
}
\arguments{
\item{parameters}{Initial parameter vector to be optimized}

\item{loglikelihood}{Function computing log-likelihood for current parameters}

\item{gradient}{Function computing parameter gradients}

\item{neghessian}{Function computing negative Hessian matrix}

\item{tol}{Numeric convergence tolerance (default 1e-7)}

\item{max_cnt}{Maximum number of optimization iterations (default 64)}

\item{max_dmp_steps}{Maximum damping step attempts (default 16)}
}
\value{
Final parameter vector returned at termination.
}
\description{
Performs iterative parameter estimation with adaptive step-size damping

Internal function for fitting unconstrained model components using damped
Newton-Raphson updates.
}
\details{
Implements a robust damped Newton-Raphson optimization algorithm.
The Newton direction is computed once per outer iteration and reused
across damping half-steps.
}
\seealso{
- \code{\link{nr_iterate}} for parameter update computation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.extract_G_diagonal}
\alias{.extract_G_diagonal}
\title{Extract Per-Partition Diagonal Blocks from a Full Matrix}
\usage{
.extract_G_diagonal(M, p_expansions, K)
}
\arguments{
\item{M}{A \eqn{P \times P} matrix where \eqn{P = p\_expansions \times (K+1)}.}

\item{p_expansions}{Integer; number of basis terms (columns) per partition.}

\item{K}{Integer; number of interior knots.}
}
\value{
A list of length \eqn{K+1}, each element a
  \eqn{p\_expansions \times p\_expansions} matrix corresponding to the
  diagonal block for partition \eqn{k}.
}
\description{
Given a \eqn{P \times P} matrix (e.g., the full-system
\eqn{\mathbf{G} = (\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X} +
\boldsymbol{\Lambda})^{-1}}), extracts the \eqn{p \times p} diagonal
block for each of the \eqn{K+1} partitions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blockfit_solve.R
\name{.bf_case_gauss_gee}
\alias{.bf_case_gauss_gee}
\title{Gaussian Identity + GEE: Closed-Form Solve (Case a)}
\usage{
.bf_case_gauss_gee(
  X,
  y,
  K,
  p_expansions,
  order_list,
  VhalfInv,
  Lambda,
  L_partition_list,
  unique_penalty_per_partition,
  A,
  constraint_values,
  spline_cols,
  flat_cols,
  observation_weights
)
}
\arguments{
\item{X}{List of \eqn{K+1} unwhitened design matrices.}

\item{y}{List of \eqn{K+1} response vectors.}

\item{K}{Integer; number of interior knots.}

\item{p_expansions}{Integer; columns per partition.}

\item{order_list}{List of observation-index vectors by partition.}

\item{VhalfInv}{\eqn{\mathbf{V}^{-1/2}} matrix in the original
observation ordering.}

\item{Lambda}{Shared penalty matrix.}

\item{L_partition_list}{Partition-specific penalty matrices.}

\item{unique_penalty_per_partition}{Logical.}

\item{A}{Full constraint matrix.}

\item{constraint_values}{List of constraint RHS vectors.}

\item{spline_cols, flat_cols}{Integer vectors of column indices.}
}
\value{
A named list:
\describe{
  \item{beta_spline}{List of \eqn{K+1} spline coefficient vectors.}
  \item{beta_flat}{Numeric vector of flat coefficients.}
  \item{result}{List of \eqn{K+1} full per-partition coefficient
    vectors.}
}
}
\description{
When the response is Gaussian with identity link and a working
correlation structure is present, the whitened system
\eqn{\mathbf{V}^{-1/2}\mathbf{X}} is not block-diagonal, so
partition-wise backfitting does not apply.
This function performs a single closed-form solve that replicates
\code{get_B} Path 1a but in the backfitting context.
}
\details{
In other words, this is the dense correlated solve on the pooled
coefficient space. The supplement's Woodbury factorization
\eqn{\mathbf{G}^{-1} =
\mathbf{G}_{\mathrm{on}}^{-1} + \mathbf{G}_{\mathrm{off}}^{-1}}
is handled in \code{get_B()}, not here, so this helper always uses the
dense correlated solve.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{invert}
\alias{invert}
\title{Matrix Inversion with Fallback Methods}
\usage{
invert(mat, include_warnings = FALSE)
}
\arguments{
\item{mat}{Square matrix to invert}

\item{include_warnings}{Logical; default FALSE for current implementation.}
}
\value{
Inverted matrix, or the identity matrix if all inversion attempts fail
}
\description{
Attempts matrix inversion using multiple methods, falling back to more robust
approaches if standard inversion fails.
}
\details{
Tries methods in order:

1. Cholesky decomposition via \code{chol2inv(chol(...))} for symmetric
   positive-definite matrices (fastest for SPD)

2. Direct inversion using \code{armaInv()} as first fallback

3. Generalized inverse using eigendecomposition with small ridge

4. Returns identity matrix if all methods fail, with optional warning

For eigendecomposition, uses a small ridge penalty (\code{1e-16}) for stability and
zeroes eigenvalues below machine precision.
}
\examples{

## Well-conditioned matrix
A <- matrix(c(4,2,2,4), 2, 2)
invert(A) \%*\% A

## Singular matrix falls back to M.P. generalized inverse
B <- matrix(c(1,1,1,1), 2, 2)
invert(B) \%*\% B


}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{confint.lgspline}
\alias{confint.lgspline}
\title{Confidence Intervals for lgspline Coefficients}
\usage{
\method{confint}{lgspline}(object, parm, level = 0.95, ...)
}
\arguments{
\item{object}{A fitted lgspline object with \code{return_varcovmat = TRUE}.}

\item{parm}{Optional vector of parameter indices or names. Default returns
all regression parameters; working-scale correlation parameters are
appended when available.}

\item{level}{Confidence level. Default 0.95.}

\item{...}{Additional arguments passed to \code{\link{wald_univariate}}.}
}
\value{
A matrix with columns giving lower and upper confidence limits,
  named e.g. \code{2.5 \%} and \code{97.5 \%} for 95\% intervals.
  When available, rows for working-scale correlation parameters are
  appended after the regression coefficients.
}
\description{
Wald-based confidence intervals for regression coefficients and, when
available, correlation parameters (on the working scale).
}
\details{
For Gaussian identity-link models, t-distribution quantiles are used with
effective degrees of freedom
\eqn{N - \mathrm{trace}(\mathbf{XUGX}^\top)}.
All other families use normal quantiles.

Correlation parameter intervals (if \code{VhalfInv_params_estimates} and
\code{VhalfInv_params_vcov} are present) are computed on the unbounded
working scale via a Wald interval.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lgspline.R
\docType{package}
\name{lgspline-package}
\alias{lgspline-package}
\title{Lagrangian Multiplier Smoothing Splines}
\description{
Implements Lagrangian multiplier smoothing splines for flexible nonparametric regression and function estimation. Provides tools for fitting, prediction, and inference using a constrained optimization approach to enforce smoothness. Supports generalized linear models, Weibull accelerated failure time (AFT) models, quadratic programming constraints, and customizable working-correlation structures, with options for parallel fitting. The core spline construction builds on Ezhov et al. (2018) \doi{10.1515/jag-2017-0029}. Quadratic-programming and SQP details follow Goldfarb & Idnani (1983) \doi{10.1007/BF02591962} and Nocedal & Wright (2006) \doi{10.1007/978-0-387-40065-5}. For smoothing spline and penalized spline background, see Wahba (1990) \doi{10.1137/1.9781611970128} and Wood (2017) \doi{10.1201/9781315370279}. For variance-component and correlation-parameter estimation, see Searle et al. (2006) <ISBN:978-0470009598>. The default multivariate partitioning step uses k-means clustering as in MacQueen (1967).
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/matthewlouisdavisBioStat/lgspline}
  \item Report bugs at \url{https://github.com/matthewlouisdavisBioStat/lgspline/issues}
}

}
\author{
\strong{Maintainer}: Matthew Davis \email{matthewlouisdavis@gmail.com} (\href{https://orcid.org/0000-0001-9714-1018}{ORCID})

}
\keyword{constrained}
\keyword{internal}
\keyword{lagrangian}
\keyword{monomial}
\keyword{multiplier}
\keyword{parametric}
\keyword{regression}
\keyword{smoothing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{coef.lgspline}
\alias{coef.lgspline}
\title{Extract Coefficients from a Fitted lgspline}
\usage{
\method{coef}{lgspline}(object, ...)
}
\arguments{
\item{object}{A fitted lgspline model object.}

\item{...}{Not used.}
}
\value{
A list of per-partition coefficient vectors. Returns NULL with a
  warning if \code{object$B} is not found.
}
\description{
Returns the per-partition polynomial coefficient lists from a fitted
lgspline model.
}
\details{
Coefficient names reflect the polynomial expansion terms, e.g.:
\itemize{
  \item intercept
  \item v: linear term for predictor v
  \item v_^2: quadratic term
  \item v^3: cubic term
  \item _v_x_w_: two-way interaction
}
Column/variable names replace numeric indices when available.

To get all coefficients as a single matrix:
\code{Reduce('cbind', coef(model_fit))}.
}
\examples{

set.seed(1234)
t <- runif(1000, -10, 10)
y <- 2*sin(t) + -0.06*t^2 + rnorm(length(t))
model_fit <- lgspline(t, y)

coefficients <- coef(model_fit)
print(coefficients[[1]])
print(Reduce('cbind', coefficients))

}
\seealso{
\code{\link{lgspline}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.compute_score_V_partitioned}
\alias{.compute_score_V_partitioned}
\title{Compute Partitioned GEE Score Vector}
\usage{
.compute_score_V_partitioned(
  X,
  X_block,
  y,
  result,
  K,
  p_expansions,
  family,
  W,
  Delta_V,
  observation_weights
)
}
\arguments{
\item{X}{List of partition-specific design matrices.}

\item{X_block}{Full \eqn{N \times P} block-diagonal design.}

\item{y}{List of response vectors by partition.}

\item{result}{List of current coefficient column vectors by partition.}

\item{K, p_expansions}{Integer dimensions.}

\item{family}{GLM family object.}

\item{W}{Length-\eqn{N} vector of GLM working weights.}

\item{Delta_V}{Fixed \eqn{N \times N} matrix
\eqn{\mathbf{V}^{-1} - \mathbf{I}}.}

\item{observation_weights}{List of observation weights by partition.}
}
\value{
A list of \eqn{K+1} column vectors, each
  \eqn{p\_expansions \times 1}, representing the partition-wise
  components of the full GEE score.
}
\description{
Computes the GEE score vector \eqn{\mathbf{X}^{\top}\mathrm{diag}
(\mathbf{W})\mathbf{V}^{-1}(\mathbf{y} - \boldsymbol{\mu})} split
into per-partition pieces of dimension \eqn{p \times 1} each. Uses
the identity \eqn{\mathbf{V}^{-1} = \mathbf{I} + \boldsymbol{\Delta}_V}
where \eqn{\boldsymbol{\Delta}_V = \mathbf{V}^{-1} - \mathbf{I}} is
precomputed and fixed across iterations, avoiding explicit formation
of \eqn{\mathbf{V}^{-1}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/negbin_helpers.R
\name{negbin_dispersion_function}
\alias{negbin_dispersion_function}
\title{NB Dispersion Function}
\usage{
negbin_dispersion_function(
  mu,
  y,
  order_indices,
  family,
  observation_weights,
  VhalfInv
)
}
\arguments{
\item{mu}{Predicted values.}

\item{y}{Observed counts.}

\item{order_indices}{Observation indices.}

\item{family}{NB family object.}

\item{observation_weights}{Observation weights.}

\item{VhalfInv}{Inverse square root of the correlation matrix, or NULL
for independent observations. When non-NULL, used to whiten residuals
for the moment-based initialization of \eqn{\theta}.}
}
\value{
Scalar \eqn{\theta} estimate (stored as \code{sigmasq_tilde}).
}
\description{
Estimates the shape parameter \eqn{\theta} from current fitted values.
When a correlation structure is present (\code{VhalfInv} is non-NULL),
the Pearson residuals are whitened before computing the moment-based
initial value, giving a better starting point for the profile MLE
under correlated data. The final estimate is always the profile MLE
over \eqn{\theta}.
}
\details{
The profile MLE maximizes \eqn{\ell(\theta \mid \mu)} via Brent's
method. When \code{VhalfInv} is provided, the Pearson residuals
\eqn{r_i = (y_i - \mu_i) / \sqrt{V(\mu_i)}} are pre-whitened as
\eqn{\tilde{r} = V^{-1/2} r} before computing the moment estimator
used for initialization. This accounts for the correlation structure
in the variance decomposition and produces a more stable starting
point for the optimizer, particularly when the correlation inflates
the marginal variance beyond what the NB model alone would predict.

The profile MLE itself does not use \code{VhalfInv} because the NB
log-likelihood is a marginal quantity; the correlation structure
affects estimation only through the mean model (handled by the GEE
paths in \code{get_B}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{compute_trace_H}
\alias{compute_trace_H}
\title{Effective degrees of freedom via trace of the hat matrix}
\usage{
compute_trace_H(
  G,
  Lambda,
  A,
  AGAInv,
  nc,
  nca,
  K,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  unique_penalty_per_partition,
  L_partition_list
)
}
\arguments{
\item{G}{List of \eqn{K+1} matrices \eqn{\mathbf{G}_k}, each \eqn{p \times p}.}

\item{Lambda}{Base penalty matrix \eqn{\boldsymbol{\Lambda}}, \eqn{p \times p},
shared across all partitions. When \code{unique_penalty_per_partition = FALSE},
this is the full per-block penalty. When \code{TRUE}, the effective penalty
for partition \eqn{k} is \eqn{\boldsymbol{\Lambda} + \mathbf{L}_k}.}

\item{A}{Constraint matrix, \eqn{P \times r}.}

\item{AGAInv}{Precomputed \eqn{(\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}},
\eqn{r \times r}.}

\item{nc}{Number of columns per partition block (\eqn{p}).}

\item{nca}{Number of constraint columns (\eqn{r}).}

\item{K}{Number of interior knots (so there are \eqn{K+1} partitions).}

\item{parallel}{Logical; use parallel processing.}

\item{cl}{Cluster object for parallel computation.}

\item{chunk_size}{Size of parallel chunks.}

\item{num_chunks}{Number of full-size chunks.}

\item{rem_chunks}{Number of remaining partitions in the final chunk.}

\item{unique_penalty_per_partition}{Logical; if \code{TRUE}, each partition
receives an additional penalty from \code{L_partition_list}.}

\item{L_partition_list}{List of partition-specific penalty matrices
\eqn{\mathbf{L}_k}, each \eqn{p \times p}. Only used when
\code{unique_penalty_per_partition = TRUE}; ignored otherwise.}
}
\value{
Scalar effective degrees of freedom, clamped to \eqn{[0, \, (K+1)p]}.
}
\description{
Computes \eqn{\mathrm{tr}(\mathbf{X}\mathbf{U}\mathbf{G}\mathbf{X}^\top
\mathbf{W}^{1/2}\mathbf{V}^{-1}\mathbf{W}^{1/2})} without forming
the \eqn{N \times N} hat matrix, by reducing the problem to
\eqn{P \times P} and \eqn{r \times r} operations.
}
\details{
\subsection{Derivation}{
Let \eqn{\mathbf{G} = (\mathbf{X}^\top\mathbf{W}^{1/2}\mathbf{V}^{-1}\mathbf{W}^{1/2}\mathbf{X} + \boldsymbol{\Lambda})^{-1}}
and \eqn{\mathbf{U} = \mathbf{I} - \mathbf{G}\mathbf{A}(\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^\top}.
By the cyclic property of the trace:
\deqn{\mathrm{tr}(\mathbf{X}\mathbf{U}\mathbf{G}\mathbf{X}^\top\mathbf{W}^{1/2}\mathbf{V}^{-1}\mathbf{W}^{1/2}) = \mathrm{tr}(\mathbf{U}\mathbf{G}\,\mathbf{X}^\top\mathbf{W}^{1/2}\mathbf{V}^{-1}\mathbf{W}^{1/2}\mathbf{X})}

Substituting \eqn{\mathbf{X}^\top\mathbf{W}^{1/2}\mathbf{V}^{-1}\mathbf{W}^{1/2}\mathbf{X} = \mathbf{G}^{-1} - \boldsymbol{\Lambda}}:
\deqn{= \mathrm{tr}(\mathbf{U}\mathbf{G}(\mathbf{G}^{-1} - \boldsymbol{\Lambda})) = \mathrm{tr}(\mathbf{U}) - \mathrm{tr}(\mathbf{U}\mathbf{G}\boldsymbol{\Lambda})}

Since \eqn{\mathbf{U}} is idempotent,
\eqn{\mathrm{tr}(\mathbf{U}) = \mathrm{rank}(\mathbf{U}) = P - r}
where \eqn{r = \mathrm{rank}(\mathbf{A})}. For the second term,
expand \eqn{\mathbf{U} = \mathbf{I} - \mathbf{G}\mathbf{A}(\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^\top}:
\deqn{\mathrm{tr}(\mathbf{U}\mathbf{G}\boldsymbol{\Lambda}) = \mathrm{tr}(\mathbf{G}\boldsymbol{\Lambda}) - \mathrm{tr}\!\left((\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^\top\mathbf{G}\boldsymbol{\Lambda}\mathbf{G}\mathbf{A}\right)}

using the cyclic property on the second term. Both \eqn{\mathbf{G}}
and \eqn{\boldsymbol{\Lambda}} are block-diagonal with \eqn{K+1}
blocks of dimension \eqn{p \times p}, so:
\deqn{\mathrm{tr}(\mathbf{G}\boldsymbol{\Lambda}) = \sum_{k=1}^{K+1}\mathrm{tr}(\mathbf{G}_k\boldsymbol{\Lambda}_k)}
and \eqn{\mathbf{G}\boldsymbol{\Lambda}\mathbf{G}} is also block-diagonal
with blocks \eqn{\mathbf{G}_k\boldsymbol{\Lambda}_k\mathbf{G}_k}.

Combining:
\deqn{\mathrm{tr}(\mathbf{X}\mathbf{U}\mathbf{G}\mathbf{X}^\top\mathbf{W}^{1/2}\mathbf{V}^{-1}\mathbf{W}^{1/2}) = (P - r) - \sum_{k=1}^{K+1}\mathrm{tr}(\mathbf{G}_k\boldsymbol{\Lambda}_k) + \mathrm{tr}\!\left((\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^\top\mathbf{G}\boldsymbol{\Lambda}\mathbf{G}\mathbf{A}\right)}

The correction term
\eqn{\mathrm{tr}((\mathbf{A}^\top\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^\top\mathbf{G}\boldsymbol{\Lambda}\mathbf{G}\mathbf{A})}
is an \eqn{r \times r} trace and captures the degrees of freedom
recovered because the constraints pin certain linear combinations of
penalized coefficients, removing them from estimation. When
\eqn{\boldsymbol{\Lambda} = \mathbf{0}} (no penalty), both
\eqn{\mathrm{tr}(\mathbf{G}\boldsymbol{\Lambda})} and the correction
vanish, giving \eqn{\mathrm{edf} = P - r}. When \eqn{\mathbf{U} = \mathbf{I}}
(no constraints), \eqn{r = 0} and the correction vanishes, giving
\eqn{\mathrm{edf} = P - \mathrm{tr}(\mathbf{G}\boldsymbol{\Lambda})}.

The result is invariant to the correlation structure \eqn{\mathbf{V}}
and GLM weights \eqn{\mathbf{W}}, which enter only through
\eqn{\mathbf{G}}.
}

\subsection{Partition-specific penalties}{
When \code{unique_penalty_per_partition = TRUE}, each partition \eqn{k}
has an additional penalty matrix \eqn{\mathbf{L}_k} from
\code{L_partition_list}, so the effective per-partition penalty becomes
\eqn{\boldsymbol{\Lambda}_k = \boldsymbol{\Lambda} + \mathbf{L}_k}.
All block-wise traces and products use \eqn{\boldsymbol{\Lambda}_k}
in place of the shared \eqn{\boldsymbol{\Lambda}}.
}

\subsection{Computational cost}{
The partition-wise traces \eqn{\mathrm{tr}(\mathbf{G}_k\boldsymbol{\Lambda}_k)}
cost \eqn{O(p^2)} each and are parallelizable. The correction term
requires forming \eqn{\mathbf{G}\boldsymbol{\Lambda}\mathbf{G}\mathbf{A}}
(block-diagonal times sparse, \eqn{O((K+1)p^2 r)}) and a single
\eqn{r \times r} trace (\eqn{O(r^2)}). The total cost is
\eqn{O((K+1)p^2 + r^2)}, compared to \eqn{O(N^2)} or \eqn{O(NP)}
for forming and tracing the full hat matrix.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{matinvsqrt}
\alias{matinvsqrt}
\title{Calculate Matrix Inverse Square Root for Symmetric Matrices}
\usage{
matinvsqrt(mat)
}
\arguments{
\item{mat}{A symmetric matrix \eqn{\textbf{M}}}
}
\value{
A matrix \eqn{\textbf{B}} such that \eqn{\textbf{B}\textbf{B}}
  equals the Moore-Penrose-style inverse on the positive-eigenvalue
  subspace, with non-positive components truncated to 0.
}
\description{
Calculate Matrix Inverse Square Root for Symmetric Matrices
}
\details{
Uses an eigenvalue-decomposition-based approach.

Non-positive eigenvalues are set to 0 before taking inverse fourth roots.

This implementation is particularly useful for whitening procedures in GLMs
with correlation structures and for computing variance-covariance matrices
under constraints.

You can use this to help construct a custom \code{VhalfInv_fxn} for
\code{\link{lgspline}}.  When only \code{VhalfInv} is supplied there,
the corresponding \code{Vhalf} is reconstructed internally by inversion
for the GEE code paths.
}
\examples{

## Identity matrix
m1 <- diag(2)
matinvsqrt(m1)  # Returns identity matrix

## Compound symmetry correlation matrix
rho <- 0.5
m2 <- matrix(rho, 3, 3) + diag(1-rho, 3)
B <- matinvsqrt(m2)
# Verify: B \%**\% B approximately equals solve(m2)
all.equal(B \%**\% B, solve(m2))

## Example for GLM correlation structure
n_blocks <- 2  # Number of subjects
block_size <- 3  # Measurements per subject
rho <- 0.7  # Within-subject correlation
# Correlation matrix for one subject
R <- matrix(rho, block_size, block_size) +
     diag(1-rho, block_size)
## Full correlation matrix for all subjects
V <- kronecker(diag(n_blocks), R)
## Create whitening matrix
VhalfInv <- matinvsqrt(V)

# Example construction of VhalfInv_fxn for lgspline
VhalfInv_fxn <- function(par) {
  rho <- tanh(par)  # Transform parameter to (-1, 1)
  R <- matrix(rho, block_size, block_size) +
       diag(1-rho, block_size)
  kronecker(diag(n_blocks), matinvsqrt(R))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blockfit_solve.R
\name{.bf_gee_deviance}
\alias{.bf_gee_deviance}
\title{Compute Whitened Deviance for GEE Convergence Monitoring}
\usage{
.bf_gee_deviance(y_vec, mu_vec, W_vec, ord, obs_wt, VhInv, fam, ...)
}
\arguments{
\item{y_vec}{Numeric vector of observed responses.}

\item{mu_vec}{Numeric vector of fitted means.}

\item{W_vec}{Numeric vector of damped Newton-Raphson weights at current iterate.}

\item{ord}{Integer vector of observation indices (passed to
custom deviance residual functions).}

\item{obs_wt}{Numeric vector of observation weights.}

\item{VhInv}{\eqn{\mathbf{V}^{-1/2}} matrix (permuted to
partition ordering).}

\item{fam}{GLM family object.}

\item{...}{Additional arguments forwarded to
\code{fam$custom_dev.resids}.}
}
\value{
Scalar; the mean whitened deviance.
}
\description{
Evaluates deviance in the whitened (decorrelated) space when the
family supplies \code{fam$custom_dev.resids}.  In that case the raw
deviance residuals are divided by \eqn{\sqrt{W}} and then
pre-multiplied by \eqn{\mathbf{V}^{-1/2}} before squaring and
averaging.  If only \code{fam$dev.resids} is available, the function
falls back to the usual mean deviance (and otherwise mean squared
error).  Used in the GEE refinement loop (Case c) of
\code{\link{blockfit_solve}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lgspline.R
\name{lgspline.fit}
\alias{lgspline.fit}
\title{Low-Level Fitting for Lagrangian Smoothing Splines}
\usage{
lgspline.fit(predictors, y = NULL, standardize_response = TRUE,
             standardize_predictors_for_knots = TRUE,
             standardize_expansions_for_fitting = TRUE, family = gaussian(),
             glm_weight_function, schur_correction_function,
             need_dispersion_for_estimation = FALSE,
             dispersion_function,
             K = NULL, custom_knots = NULL, cluster_on_indicators = FALSE,
             make_partition_list = NULL, previously_tuned_penalties = NULL,
             smoothing_spline_penalty = NULL, opt = TRUE, use_custom_bfgs = TRUE,
             delta = NULL, tol = 10*sqrt(.Machine$double.eps),
             tuning_criterion = "loo", gcv_gamma = 1.4,
             initial_wiggle = c(2e-12, 2e-7, 2e-4, 0.2),
             initial_flat = c(0.5, 5), wiggle_penalty = 2e-07,
             flat_ridge_penalty = 0.5, unique_penalty_per_partition = TRUE,
             unique_penalty_per_predictor = TRUE, meta_penalty = 1e-08,
             predictor_penalties = NULL, partition_penalties = NULL,
             include_quadratic_terms = TRUE, include_cubic_terms = TRUE,
             include_quartic_terms = FALSE, include_2way_interactions = TRUE,
             include_3way_interactions = TRUE,
             include_quadratic_interactions = FALSE,
             offset = c(), just_linear_with_interactions = NULL,
             just_linear_without_interactions = NULL,
             exclude_interactions_for = NULL,
             exclude_these_expansions = NULL, custom_basis_fxn = NULL,
             include_constrain_fitted = TRUE,
             include_constrain_first_deriv = TRUE,
             include_constrain_second_deriv = TRUE,
             include_constrain_interactions = TRUE, cl = NULL, chunk_size = NULL,
             parallel_eigen = TRUE, parallel_trace = FALSE, parallel_aga = FALSE,
             parallel_matmult = FALSE, parallel_unconstrained = FALSE,
             parallel_find_neighbors = FALSE, parallel_penalty = FALSE,
             parallel_make_constraint = FALSE,
             unconstrained_fit_fxn = unconstrained_fit_default,
             keep_weighted_Lambda = FALSE, iterate_tune = TRUE,
             iterate_final_fit = TRUE, blockfit = TRUE,
             qp_score_function,
             qp_observations = NULL, qp_Amat = NULL, qp_bvec = NULL, qp_meq = 0,
             qp_positive_derivative = FALSE, qp_negative_derivative = FALSE,
             qp_positive_2ndderivative = FALSE, qp_negative_2ndderivative = FALSE,
             qp_monotonic_increase = FALSE, qp_monotonic_decrease = FALSE,
             qp_range_upper = NULL, qp_range_lower = NULL, qp_Amat_fxn = NULL,
             qp_bvec_fxn = NULL, qp_meq_fxn = NULL, constraint_values = cbind(),
             constraint_vectors = cbind(), return_G = TRUE, return_Ghalf = TRUE,
             return_U = TRUE, estimate_dispersion = TRUE,
             unbias_dispersion = TRUE,
             return_varcovmat = TRUE, exact_varcovmat = FALSE,
             return_lagrange_multipliers = FALSE,
             custom_penalty_mat = NULL,
             cluster_args = c(custom_centers = NA, nstart = 10),
             dummy_dividor = 1.2345672152894e-22,
             dummy_adder = 2.234567210529e-18,
             verbose = FALSE, verbose_tune = FALSE,
             dummy_fit = FALSE, auto_encode_factors = TRUE,
             observation_weights = NULL, do_not_cluster_on_these = c(),
             neighbor_tolerance = 1 + 1e-16, no_intercept = FALSE,
             VhalfInv = NULL, Vhalf = NULL, include_warnings = TRUE,
             og_cols = NULL,
             factor_groups = NULL, ...)
}
\arguments{
\item{predictors}{Numeric matrix or data frame of predictor variables on the
low-level input scale expected by \code{lgspline.fit}. Unlike
\code{\link{lgspline}}, this interface does not parse formulas or a
separate \code{data} argument.}

\item{y}{Default: NULL. Numeric response variable vector.}

\item{standardize_response}{Default: TRUE. Logical indicator controlling whether
the response variable should be centered and scaled before model fitting. Only
offered for identity link functions.}

\item{standardize_predictors_for_knots}{Default: TRUE. Logical flag controlling
whether predictors are internally standardized for partitioning / knot
placement. The exact transformation is handled inside
\code{\link{make_partitions}} and depends on the effective clustering
dimension.}

\item{standardize_expansions_for_fitting}{Default: TRUE. Logical switch to
standardize polynomial basis expansions during model fitting. Design matrices,
variance-covariance matrices, and coefficients are backtransformed after fitting.
\eqn{\mathbf{U}} and \eqn{\mathbf{G}} remain on the transformed scale;
\code{B_raw} corresponds to coefficients on the expansion-standardized scale.}

\item{family}{Default: \code{gaussian()}. GLM family specifying the error distribution
and link function. Minimally requires: family name, link name, linkfun, linkinv,
variance.}

\item{glm_weight_function}{Default: function returning \code{family$variance(mu)},
optionally multiplied by \code{observation_weights}. Codes the mean-variance
relationship (the diagonal \eqn{\mathbf{W}} matrix) used for updating
\eqn{\mathbf{G}} after obtaining constrained estimates.}

\item{schur_correction_function}{Default: function returning list of zeros.
Computes Schur complements \eqn{\mathbf{S}} added to \eqn{\mathbf{G}}:
\eqn{\mathbf{G}^{*} = (\mathbf{G}^{-1} + \mathbf{S})^{-1}}.}

\item{need_dispersion_for_estimation}{Default: FALSE. Logical indicator specifying
whether a dispersion parameter is required for coefficient estimation (e.g.
Weibull AFT).}

\item{dispersion_function}{Default: function returning mean squared residuals.
Custom function for estimating the exponential dispersion parameter.}

\item{K}{Default: NULL. Integer specifying the number of knot locations. Intuitively,
total partitions minus 1.}

\item{custom_knots}{Default: NULL. Optional matrix providing user-specified knot
locations in 1-D.}

\item{cluster_on_indicators}{Default: FALSE. Logical flag for whether indicator
variables should be used for clustering knot locations.}

\item{make_partition_list}{Default: NULL. Optional list allowing direct specification
of custom partition assignments. The \code{make_partition_list} returned by one
model can be supplied here to reuse knot locations.}

\item{previously_tuned_penalties}{Default: NULL. Optional list of pre-computed penalty
components from a previous model fit.}

\item{smoothing_spline_penalty}{Default: NULL. Optional custom smoothing spline penalty
matrix.}

\item{opt}{Default: TRUE. Logical switch controlling automatic penalty optimization.}

\item{use_custom_bfgs}{Default: TRUE. Selects between a native damped-BFGS
implementation with closed-form gradients or base R's BFGS with finite-difference
gradients. The native path is usually faster, while the finite-difference
fallback can be preferable when tuning under exact LOO and the leverage
derivative is numerically noisy.}

\item{delta}{Default: NULL. Numeric pseudocount for stabilizing optimization in
non-identity link function scenarios.}

\item{tol}{Default: \code{10*sqrt(.Machine$double.eps)}. Numeric convergence tolerance.}

\item{tuning_criterion}{Default: \code{"loo"}. Character scalar selecting the
tuning criterion. Use \code{"loo"} for exact leave-one-out on the transformed
tuning problem, or \code{"gcv"} for the generalized cross-validation
criterion. The LOO path computes the needed hat-matrix diagonal exactly
from blockwise constrained-\eqn{\mathbf{G}} quantities, without explicitly
forming the full projection matrix or full hat matrix. In empirical
diagnostics, the observation-wise derivative of the LOO leverage term can
be numerically delicate even when the overall tuning criterion and fitted
penalties remain well behaved; users who prefer a more conservative
optimization path can set \code{use_custom_bfgs = FALSE}. Shared wiggle
and flat tuning directions are differentiated directly, while optional
predictor- and partition-specific penalties continue to use a lower-cost
ratio approximation. For very large samples, generalized cross-validation
is often the more practical choice; as a rough guideline, \code{"gcv"}
is recommended once the sample size is above about 250,000.}

\item{gcv_gamma}{Default: 1.4. Numeric scalar, at least 1, used only when
\code{tuning_criterion = "gcv"}. Multiplies the effective degrees of freedom
in the GCV denominator during automatic penalty tuning. It is accepted but
ignored when \code{tuning_criterion = "loo"}.}

\item{initial_wiggle}{Default: \code{c(2e-12, 2e-7, 2e-4, 0.2)}. Numeric vector of
initial grid points for wiggle penalty optimization, on the raw (non-negative) scale.}

\item{initial_flat}{Default: \code{c(0.5, 5)}. Numeric vector of initial grid points
for ridge penalty optimization, on the raw scale (ratio of ridge to wiggle).}

\item{wiggle_penalty}{Default: 2e-7. Numeric penalty on the integrated squared second
derivative, governing function smoothness.}

\item{flat_ridge_penalty}{Default: 0.5. Numeric flat ridge penalty for intercepts and
linear terms only. Multiplied by \code{wiggle_penalty} to obtain total ridge penalty.}

\item{unique_penalty_per_partition}{Default: TRUE. Logical flag allowing penalty
magnitude to differ across partitions.}

\item{unique_penalty_per_predictor}{Default: TRUE. Logical flag allowing penalty
magnitude to differ between predictors.}

\item{meta_penalty}{Default: 1e-8. Numeric regularization coefficient for
predictor- and partition-specific penalties during tuning. On the raw scale,
the implemented meta-penalty shrinks these penalty multipliers toward 1; the
wiggle penalty receives only a tiny stabilizing penalty by default.}

\item{predictor_penalties}{Default: NULL. Optional vector of custom penalties per
predictor, on the raw (positive) scale.}

\item{partition_penalties}{Default: NULL. Optional vector of custom penalties per
partition, on the raw (positive) scale.}

\item{include_quadratic_terms}{Default: TRUE. Logical switch to include squared
predictor terms.}

\item{include_cubic_terms}{Default: TRUE. Logical switch to include cubic predictor
terms.}

\item{include_quartic_terms}{Default: FALSE. Logical switch to include quartic
predictor terms at this low-level interface.}

\item{include_2way_interactions}{Default: TRUE. Logical switch for linear two-way
interactions.}

\item{include_3way_interactions}{Default: TRUE. Logical switch for three-way
interactions.}

\item{include_quadratic_interactions}{Default: FALSE. Logical switch for
linear-quadratic interaction terms.}

\item{offset}{Default: Empty vector. Column indices/names to include as offsets.
Coefficients for offset terms are automatically constrained to 1.}

\item{just_linear_with_interactions}{Default: NULL. Integer or character vector
specifying predictors to retain as linear terms while still allowing
interactions.}

\item{just_linear_without_interactions}{Default: NULL. Integer or character
vector specifying predictors to retain only as linear terms without
interactions. Eligible for blockfitting.}

\item{exclude_interactions_for}{Default: NULL. Integer or character vector of
predictors to exclude from all interaction terms.}

\item{exclude_these_expansions}{Default: NULL. Character vector of basis expansions to
exclude. Named columns of data, or in the form \code{"_1_"}, \code{"_2_"},
\code{"_1_x_2_"}, \code{"_2_^2"} etc.}

\item{custom_basis_fxn}{Default: NULL. Optional user-defined function for custom basis
expansions. See \code{\link{get_polynomial_expansions}}.}

\item{include_constrain_fitted}{Default: TRUE. Logical switch to constrain fitted
values at knot points.}

\item{include_constrain_first_deriv}{Default: TRUE. Logical switch to constrain first
derivatives at knot points.}

\item{include_constrain_second_deriv}{Default: TRUE. Logical switch to constrain second
derivatives at knot points.}

\item{include_constrain_interactions}{Default: TRUE. Logical switch to constrain
interaction terms at knot points.}

\item{cl}{Default: NULL. Parallel processing cluster object
(use \code{parallel::makeCluster()}).}

\item{chunk_size}{Default: NULL. Integer specifying custom chunk size for parallel
processing.}

\item{parallel_eigen}{Default: TRUE. Logical flag for parallel eigenvalue decomposition.}

\item{parallel_trace}{Default: FALSE. Logical flag for parallel trace computation.}

\item{parallel_aga}{Default: FALSE. Logical flag for parallel \eqn{\mathbf{G}} and
\eqn{\mathbf{A}} matrix operations.}

\item{parallel_matmult}{Default: FALSE. Logical flag for parallel block-diagonal matrix
multiplication.}

\item{parallel_unconstrained}{Default: FALSE. Logical flag for parallel
unconstrained MLE for non-identity-link-Gaussian models.}

\item{parallel_find_neighbors}{Default: FALSE. Logical flag for parallel neighbor
identification.}

\item{parallel_penalty}{Default: FALSE. Logical flag for parallel penalty matrix
construction.}

\item{parallel_make_constraint}{Default: FALSE. Logical flag for parallel constraint
matrix generation.}

\item{unconstrained_fit_fxn}{Default: \code{\link{unconstrained_fit_default}}.
Custom function for fitting unconstrained models per partition.}

\item{keep_weighted_Lambda}{Default: FALSE. Logical flag to retain GLM weights in
penalty constraints using Tikhonov parameterization. Advised for non-canonical GLMs.}

\item{iterate_tune}{Default: TRUE. Logical switch for iterative optimization during
penalty tuning.}

\item{iterate_final_fit}{Default: TRUE. Logical switch for iterative optimization in
final model fitting.}

\item{blockfit}{Default: TRUE. Logical switch for backfitting with mixed spline and
non-interactive linear terms. Requires flat columns, \code{K > 0}, and no active
correlation structure. Falls back to \code{get_B} on failure.}

\item{qp_score_function}{Default: \eqn{\mathbf{X}^{\top}(\mathbf{y} - \boldsymbol{\mu})}.
Score function for quadratic programming, blockfit, and GEE formulations.
Accepts arguments \code{"X, y, mu, order_list, dispersion, VhalfInv,
observation_weights, ..."}.}

\item{qp_observations}{Default: NULL. Numeric vector of observation indices
at which built-in QP constraints are evaluated. Useful for reducing the
size of the constrained system.}

\item{qp_Amat}{Default: NULL. Optional pre-built QP constraint matrix.
In the current pipeline its presence marks QP handling as active, but the
built-in constructor does not merge it into the assembled constraint set;
use \code{qp_Amat_fxn} for custom assembled constraints.}

\item{qp_bvec}{Default: NULL. Optional pre-built QP right-hand side paired
with \code{qp_Amat}. Like \code{qp_Amat}, it is currently treated as an
advanced placeholder rather than merged into the built-in constructor.}

\item{qp_meq}{Default: 0. Optional number of equality constraints paired
with \code{qp_Amat}. Like \code{qp_Amat}, it is currently treated as an
advanced placeholder rather than merged into the built-in constructor.}

\item{qp_positive_derivative}{Default: FALSE. Constrain function to have positive first derivatives. Accepts: \code{FALSE} (no constraint), \code{TRUE} (all predictors), or a character/integer vector naming specific predictor variables to constrain. For example, \code{qp_positive_derivative = "Dose"} constrains only the Dose variable, while \code{qp_positive_derivative = c(1, 3)} constrains columns 1 and 3 of the predictor matrix.}

\item{qp_negative_derivative}{Default: FALSE. Constrain function to have negative first derivatives. Same input types as \code{qp_positive_derivative}. Can be used simultaneously with \code{qp_positive_derivative} on different variables.}

\item{qp_positive_2ndderivative}{Default: FALSE. Constrain function to have positive (convex) second derivatives. Same input types as \code{qp_positive_derivative}.}

\item{qp_negative_2ndderivative}{Default: FALSE. Constrain function to have negative (concave) second derivatives. Same input types as \code{qp_positive_derivative}.}

\item{qp_monotonic_increase}{Default: FALSE. Logical only. Constrain fitted values to be monotonically increasing in observation order.}

\item{qp_monotonic_decrease}{Default: FALSE. Logical only. Constrain fitted values to be monotonically decreasing in observation order.}

\item{qp_range_upper}{Default: NULL. Numeric upper bound for constrained fitted values.}

\item{qp_range_lower}{Default: NULL. Numeric lower bound for constrained fitted values.}

\item{qp_Amat_fxn}{Default: NULL. Custom function generating Amat.}

\item{qp_bvec_fxn}{Default: NULL. Custom function generating bvec.}

\item{qp_meq_fxn}{Default: NULL. Custom function generating meq.}

\item{constraint_values}{Default: \code{cbind()}. Optional matrix encoding
nonzero equality targets paired with \code{constraint_vectors}. When left
empty, added equality constraints are treated as homogeneous.}

\item{constraint_vectors}{Default: \code{cbind()}. Optional matrix of
user-supplied equality-constraint vectors, appended to the internally
generated smoothness constraints.}

\item{return_G}{Default: TRUE. Logical switch to return the unscaled unconstrained
variance-covariance matrix \eqn{\mathbf{G}}.}

\item{return_Ghalf}{Default: TRUE. Logical switch to return
\eqn{\mathbf{G}^{1/2}}.}

\item{return_U}{Default: TRUE. Logical switch to return the constraint projection
matrix \eqn{\mathbf{U}}.}

\item{estimate_dispersion}{Default: TRUE. Logical flag to estimate dispersion after
fitting.}

\item{unbias_dispersion}{Default: TRUE. Logical switch to multiply
dispersion by \eqn{N/(N - \mathrm{trace}(\mathbf{H}))}. Unlike
\code{\link{lgspline}}, no wrapper-level auto-resolution is performed here.}

\item{return_varcovmat}{Default: TRUE. Logical switch to return the variance-covariance
matrix of estimated coefficients. Needed for Wald inference.}

\item{exact_varcovmat}{Default: FALSE. Logical switch to replace the default
asymptotic (Bayesian posterior) variance-covariance matrix with the exact
frequentist variance-covariance matrix of the constrained estimator. The
asymptotic version uses the Hessian of
the penalized log-likelihood:
\eqn{\tilde{\sigma}^{2}\mathbf{U}\mathbf{G}\mathbf{U}^{\top}}.
The exact version additionally corrects for the penalty's contribution as a shrinkage
prior, giving:
\deqn{\tilde{\sigma}^{2}\mathbf{U}\mathbf{G}\mathbf{U}^{\top}
      - \tilde{\sigma}^{2}\mathbf{U}\mathbf{G}\boldsymbol{\Lambda}\mathbf{G}\mathbf{U}^{\top}}
When a correlation structure is present (\code{VhalfInv} non-NULL),
\eqn{\mathbf{G}_{\mathrm{correct}}} replaces the block-diagonal \eqn{\mathbf{G}}.
For Gaussian identity link (with or without correlation structure), the result is
the exact variance-covariance matrix of the constrained estimate.
The returned object still stores the result in \code{varcovmat}. Requires
\code{return_varcovmat = TRUE}.}

\item{return_lagrange_multipliers}{Default: FALSE. Logical switch to return the
Lagrangian multiplier vector.}

\item{custom_penalty_mat}{Default: NULL. Optional \eqn{p \times p} custom penalty
matrix for individual partitions, replacing the default ridge on linear/intercept
terms. Run with \code{dummy_fit = TRUE} first to inspect expansion structure.}

\item{cluster_args}{Default: \code{c(custom_centers = NA, nstart = 10)}. Named
vector of arguments controlling clustering. If the first argument is not
\code{NA}, it is treated as custom cluster centers (typically an
\eqn{(K+1) \times q} matrix). Otherwise, default k-means is used.}

\item{dummy_dividor}{Default: 0.00000000000000000000012345672152894. Small numeric
constant to prevent division by zero.}

\item{dummy_adder}{Default: 0.000000000000000002234567210529. Small numeric constant
to prevent division by zero.}

\item{verbose}{Default: FALSE. Logical flag to print general progress messages.}

\item{verbose_tune}{Default: FALSE. Logical flag to print detailed progress during
penalty tuning.}

\item{dummy_fit}{Default: FALSE. Runs the full pipeline but sets coefficients to zero,
allowing inspection of design matrix structure, penalty matrices, and partitioning.
Replaces the deprecated \code{expansions_only} argument.}

\item{auto_encode_factors}{Default: TRUE. Compatibility flag carried through
from higher-level preprocessing. Direct calls to \code{lgspline.fit}
should usually supply already encoded predictors and use
\code{factor_groups} when sum-to-zero constraints are needed.}

\item{observation_weights}{Default: NULL. Numeric vector of observation-specific
weights for generalized least squares estimation.}

\item{do_not_cluster_on_these}{Default: \code{c()}. Predictor columns to exclude from
clustering. Accepts numeric column indices or character column names.}

\item{neighbor_tolerance}{Default: \code{1 + 1e-16}. Numeric tolerance for
determining neighboring partitions using k-means clustering. Intended for
internal use.}

\item{no_intercept}{Default: FALSE. Logical flag to constrain intercept to 0.
Formulas with \code{"0+"} set this to TRUE automatically.}

\item{VhalfInv}{Default: NULL. Fixed custom \eqn{N \times N} square-root-inverse
covariance matrix \eqn{\mathbf{V}^{-1/2}}. Triggers GLS with known covariance.
Post-fit inference recomputed from whitened Gram matrices.}

\item{Vhalf}{Default: NULL. Fixed custom \eqn{N \times N} square-root covariance
\eqn{\mathbf{V}^{1/2}}. Computed as inverse of \code{VhalfInv} if not supplied.}

\item{include_warnings}{Default: TRUE. Logical switch to control display of warnings.}

\item{og_cols}{Default: NULL. Original predictor names}

\item{factor_groups}{Named list mapping original factor variable names to
integer vectors of their corresponding one-hot indicator column positions
within the predictor matrix. Each element enforces a sum-to-zero equality
constraint on the linear-term coefficients of its indicator columns within
every partition, ensuring identifiability when all factor levels are
included without a reference/dropped level. For a group with indicator
columns at positions \code{j1, j2, ..., jm}, the constraint is
\eqn{\sum_{i=1}^{m} \beta_{ji,k} = 0} for each partition \eqn{k}.
Groups with fewer than two resolved positions are silently ignored.
Populated automatically by \code{\link{process_input}} when
\code{auto_encode_factors = TRUE}; users calling \code{lgspline.fit}
directly should construct this list manually when passing one-hot encoded
predictors without a reference level. Default \code{NULL} (no
sum-to-zero constraints imposed).}

\item{...}{Additional arguments passed to the unconstrained model fitting function.}
}
\value{
A list containing the fitted model components, forming the core
structure used internally by \code{\link{lgspline}} and its associated methods.
This function is primarily intended for internal use or advanced users needing
direct access to fitting components. The returned list contains numerous elements,
typically including:
\describe{
  \item{y}{The original response vector provided.}
  \item{ytilde}{The fitted values on the original response scale. Set to
    \code{rep(0, N)} when \code{dummy_fit = TRUE}.}
  \item{X}{A list, with each element the design matrix (\eqn{\mathbf{X}_{k}})
    for partition k, on the unstandardized expansion scale.}
  \item{A}{The constraint matrix (\eqn{\mathbf{A}}) encoding smoothness and
    any other linear equality constraints. Reduced to linearly independent
    columns via pivoted QR decomposition.}
  \item{B}{A list of the final fitted coefficient vectors
    (\eqn{\boldsymbol{\beta}_{k}}) for each partition k, on the original
    predictor/response scale.}
  \item{B_raw}{A list of fitted coefficient vectors on the internally
    standardized scale used during fitting.}
  \item{K, p, q, P, N}{Key dimensions: number of internal knots (K), basis
    functions per partition (p), original predictors (q), total coefficients
    (P), and sample size (N).}
  \item{penalties}{A list containing the final penalty components used
    (e.g., \code{Lambda}, \code{L1}, \code{L2}, \code{L_predictor_list},
    \code{L_partition_list}). See \code{\link{compute_Lambda}}.}
  \item{knot_scale_transf, knot_scale_inv_transf}{Functions to transform
    predictors to/from the scale used for knot placement.}
  \item{knots}{Matrix or vector of knot locations on the original predictor
    scale (NULL if K=0 or q > 1).}
  \item{partition_codes}{Vector assigning each original observation to a partition.}
  \item{partition_bounds}{Internal representation of partition boundaries.}
  \item{make_partition_list}{List containing centers, knot midpoints, neighbor
    info, and assignment function from partitioning (NULL if K=0 or 1D).
    See \code{\link{make_partitions}}.}
  \item{knot_expand_function, assign_partition}{Internal functions for
    partitioning data. See \code{\link{knot_expand_list}}.}
  \item{predict}{The primary function embedded in the object for generating
    predictions on new data. For multi-predictor models,
    \code{take_first_derivatives = TRUE} returns derivatives as a named list
    of per-variable derivative vectors rather than a concatenated vector.
    See \code{\link{predict.lgspline}}.}
  \item{family}{The \code{\link[stats]{family}} object or custom list used.}
  \item{estimate_dispersion, unbias_dispersion}{Logical flags related to
    dispersion estimation settings.}
  \item{sigmasq_tilde}{The estimated (or fixed) dispersion parameter
    \eqn{\tilde{\sigma}^{2}}. For Gaussian identity fits with
    \code{VhalfInv} non-\code{NULL}, this is computed from whitened
    residuals \eqn{\mathbf{V}^{-1/2}(\mathbf{y} - \hat{\mathbf{y}})},
    multiplied by the observation weights and the optional
    bias-correction factor. When \code{estimate_dispersion = FALSE},
    set to 1. Omitted when \code{dummy_fit = TRUE}.}
  \item{backtransform_coefficients, forwtransform_coefficients}{Functions to
    convert coefficients between standardized and original scales.}
  \item{mean_y, sd_y}{Mean and standard deviation used for standardizing
    the response.}
  \item{og_order, order_list}{Information mapping original data order to
    partitioned order.}
  \item{constraint_values, constraint_vectors}{User-supplied additional linear
    equality constraints.}
  \item{expansion_scales}{Scaling factors applied to basis expansions during
    fitting (if \code{standardize_expansions_for_fitting = TRUE}).}
  \item{take_derivative, take_interaction_2ndderivative,
    get_all_derivatives_insample}{Functions related to computing derivatives
    of the fitted spline.}
  \item{numerics, power1_cols, ..., nonspline_cols}{Integer vectors storing
    column indices identifying different types of terms in the basis expansion.}
  \item{return_varcovmat}{Logical indicating if variance matrix calculation
    was requested.}
  \item{exact_varcovmat}{Not returned as a standalone component; this
    argument only controls whether \code{varcovmat}, when requested, is
    left as the default asymptotic/Laplace version or replaced by the
    exact frequentist correction available for Gaussian identity fits.}
  \item{raw_expansion_names}{Original generated names for basis expansion
    columns (before potential renaming if input predictors had names).}
  \item{std_X, unstd_X}{Functions to standardize/unstandardize design matrices
    according to \code{expansion_scales}.}
  \item{parallel_cluster_supplied}{Logical indicating if a parallel cluster
    was used.}
  \item{weights}{The original observation weights provided (potentially
    reformatted).}
  \item{VhalfInv}{The fixed \eqn{\mathbf{V}^{-1/2}} matrix if supplied.}
  \item{quadprog_list}{List containing components related to quadratic
    programming constraints, if used.}
  \item{G}{List of unscaled variance-covariance matrices
    \eqn{\mathbf{G}_{k}} per partition, returned if \code{return_G = TRUE}.
    When \code{VhalfInv} is non-\code{NULL}, recomputed from whitened Gram
    matrices. Omitted when \code{dummy_fit = TRUE}.}
  \item{Ghalf}{List of \eqn{\mathbf{G}_{k}^{1/2}} matrices, returned if
    \code{return_Ghalf = TRUE}. When \code{VhalfInv} is non-\code{NULL},
    the full \eqn{\mathbf{G}_{\mathrm{correct}}^{1/2}} is used for
    posterior draws and variance-covariance computation.
    Omitted when \code{dummy_fit = TRUE}.}
  \item{U}{Constraint projection matrix \eqn{\mathbf{U}}, returned if
    \code{return_U = TRUE}. Omitted when \code{dummy_fit = TRUE}.}
  \item{trace_XUGX}{The effective degrees-of-freedom trace term.
    When \code{VhalfInv} is non-\code{NULL}, it is computed from the
    full penalized GLS information rather than the block-diagonal
    approximation. Omitted when \code{dummy_fit = TRUE}.}
  \item{varcovmat}{The final variance-covariance matrix of the estimated
    coefficients. Computed via the outer-product form
    \eqn{\sigma^{2}(\mathbf{U}\mathbf{G}^{1/2})(\mathbf{U}\mathbf{G}^{1/2})^{\top}}
    for numerical stability. When \code{VhalfInv} is non-\code{NULL}, uses
    the full \eqn{\mathbf{G}_{\mathrm{correct}}^{1/2}} in place of
    block-diagonal \eqn{\mathbf{G}^{1/2}}. Returned if
    \code{return_varcovmat = TRUE}. By default this is the asymptotic
    (Laplace/posterior) variance-covariance matrix; when
    \code{exact_varcovmat = TRUE}, it is replaced in-place by the exact
    frequentist correction available for Gaussian identity fits.
    Omitted when \code{dummy_fit = TRUE}.}
  \item{lagrange_multipliers}{Vector of Lagrangian multipliers if
    \code{return_lagrange_multipliers = TRUE}. For equality-only fits these
    follow the formulation
    \eqn{(\mathbf{A}^{\top}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{\top}(\hat{\boldsymbol{\beta}} - \boldsymbol{\beta_0})}.
    When quadratic-programming constraints are active they are taken
    directly from \code{solve.QP} and therefore refer to the combined
    equality/inequality constraint system. \code{NULL} if no constraints
    are active}.
}
Note that the exact components returned depend heavily on the function
arguments (e.g., values of \code{return_G}, \code{return_varcovmat}, etc.)
and whether \code{dummy_fit = TRUE}.
}
\description{
The core function for fitting Lagrangian smoothing splines with
less user-friendliness. Called internally by \code{\link{lgspline}} after
formula parsing, factor encoding, and correlation-structure setup.
}
\details{
\code{lgspline.fit} performs the following steps:
\enumerate{
  \item Polynomial expansion and predictor standardization.
  \item Knot placement and partitioning (k-means or custom).
  \item Constraint matrix \eqn{\mathbf{A}} construction. Only a linearly
        independent subset of columns is retained via pivoted QR decomposition.
  \item Penalty tuning via exact leave-one-out by default, or generalized
        cross-validation when \code{tuning_criterion = "gcv"}, or use of
        previously tuned penalties.
  \item Final coefficient estimation via one of three paths:
        \itemize{
          \item \strong{Blockfit option} (when \code{blockfit = TRUE},
                flat columns are non-empty, \code{K > 0}, and no correlation
                structure): Routes through \code{blockfit_solve} for
                backfitting with mixed spline and non-interactive linear terms.
                Falls back to \code{get_B} on failure.
          \item \strong{Standard \code{get_B}} path: Three internal
                computational paths: GEE (damped SQP with correlation
                structures), Gaussian identity (closed-form OLS projection),
                and general GLM (unconstrained fit + Lagrangian projection
                with optional IRLS loop).
        }
  \item Post-fit inference: \eqn{\mathbf{U}}, trace, dispersion,
        variance-covariance matrix, and optionally Lagrange multipliers.
        When \code{VhalfInv} is non-\code{NULL}, these are computed from
        the whitened Gram matrices
        \eqn{\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X}} via the full
        penalized GLS information
        \eqn{\mathbf{G}_{\mathrm{correct}} =
        (\mathbf{X}^{\top}\mathbf{V}^{-1}\mathbf{X} +
        \boldsymbol{\Lambda})^{-1}}.
}

\strong{Dummy fit.} When \code{dummy_fit = TRUE}, an early-return path
skips the expensive fitting steps (\code{compute_G_eigen}, \code{get_B},
trace computation, variance-covariance matrix) while retaining all
penalty, partitioning, and design matrix information. Coefficients are
set to zero. This replaces the deprecated \code{expansions_only} argument.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.qp_refine}
\alias{.qp_refine}
\title{Quadratic Programming Refinement for Inequality Constraints}
\usage{
.qp_refine(
  result,
  X,
  y,
  K,
  p_expansions,
  A,
  Lambda,
  Lambda_block,
  family,
  iterate,
  tol,
  qp_Amat,
  qp_bvec,
  qp_meq,
  qp_score_function,
  order_list,
  glm_weight_function,
  schur_correction_function,
  need_dispersion_for_estimation,
  dispersion_function,
  observation_weights,
  VhalfInv,
  ...
)
}
\arguments{
\item{result}{List of current coefficient column vectors by partition.}

\item{X}{List of partition-specific design matrices.}

\item{y}{List of response vectors by partition.}

\item{K}{Integer; number of interior knots.}

\item{p_expansions}{Integer; number of basis terms per partition.}

\item{A}{Equality constraint matrix \eqn{\mathbf{A}}.}

\item{Lambda}{Shared penalty matrix \eqn{\boldsymbol{\Lambda}}.}

\item{Lambda_block}{Full block-diagonal penalty matrix.}

\item{family}{GLM family object.}

\item{iterate}{Logical; if \code{TRUE}, iterate the SQP loop until
convergence rather than taking a single step.}

\item{tol}{Convergence tolerance.}

\item{qp_Amat}{Inequality constraint matrix \eqn{\mathbf{C}} for
\code{solve.QP}.}

\item{qp_bvec}{Inequality constraint vector \eqn{\mathbf{c}}.}

\item{qp_meq}{Number of equality constraints within \code{qp_Amat}.}

\item{qp_score_function}{Score function
\eqn{\nabla_{\boldsymbol{\beta}}\ell(\boldsymbol{\beta}^*)} for the
QP step.}

\item{order_list}{List of index vectors mapping partition rows to
original data ordering.}

\item{glm_weight_function}{Function computing GLM working weights.}

\item{schur_correction_function}{Function computing Schur corrections.}

\item{need_dispersion_for_estimation}{Logical.}

\item{dispersion_function}{Dispersion estimation function.}

\item{observation_weights}{List of observation weights by partition.}

\item{VhalfInv}{Inverse square root of the working correlation matrix in
the original observation ordering (or \code{NULL}); permuted internally
via \code{order_list}.}

\item{...}{Passed to weight, correction, dispersion, and score functions.}
}
\value{
A list with components:
\describe{
  \item{result}{List of refined coefficient column vectors by partition.}
  \item{qp_info}{List with QP solve metadata including Lagrangian
    multipliers and the active constraint matrix.}
}
}
\description{
After the Lagrangian projection handles smoothness equality constraints,
this function refines the estimate to satisfy additional inequality
constraints (monotonicity, derivative sign, range bounds, or user-supplied
constraints) via \code{quadprog::solve.QP}.
}
\details{
The subproblem at each iteration is a second-order Taylor approximation
of the penalized log-likelihood around the current iterate
\eqn{\boldsymbol{\beta}^*}. Collecting terms, this yields:
\deqn{\tilde{\boldsymbol{\beta}} = \arg\min_{\boldsymbol{\beta}}
  \left\{-\mathbf{d}^{\top}\boldsymbol{\beta} + \frac{1}{2}
  \boldsymbol{\beta}^{\top}\mathbf{G}^{-1}\boldsymbol{\beta}\right\}
  \quad \text{s.t.} \quad
  \mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{0}, \quad
  \mathbf{C}^{\top}\boldsymbol{\beta} \succeq \mathbf{c}}
where \eqn{\mathbf{d}} is the score adjusted by the current iterate
and \eqn{\mathbf{c}} is the constraint value vector. Step acceptance
uses damped updates with deviance monitoring; see Nocedal and Wright
(2006) for the general SQP framework.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{summary.wald_lgspline}
\alias{summary.wald_lgspline}
\title{Summary Method for wald_lgspline Objects}
\usage{
\method{summary}{wald_lgspline}(object, ...)
}
\arguments{
\item{object}{A \code{"wald_lgspline"} object.}

\item{...}{Passed to \code{\link{print.wald_lgspline}}.}
}
\value{
Invisibly returns \code{object}.
}
\description{
Prints a header with model info then delegates to
\code{\link{print.wald_lgspline}}.
}
\seealso{
\code{\link{wald_univariate}}, \code{\link{print.wald_lgspline}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{.compute_hat_diag}
\alias{.compute_hat_diag}
\title{Compute the Diagonal of the Tuning Hat Matrix}
\usage{
.compute_hat_diag(X, G, A, AGAInv, K, p_expansions, eps = 1e-07)
}
\arguments{
\item{X}{List of partition-specific design matrices.}

\item{G}{List of partition-specific \eqn{\mathbf{G}_k} matrices.}

\item{A}{Constraint matrix.}

\item{AGAInv}{Inverse of \eqn{\mathbf{A}^{\top}\mathbf{G}\mathbf{A}}.}

\item{K}{Number of interior knots.}

\item{p_expansions}{Number of columns per partition.}

\item{eps}{Numeric floor used to keep \eqn{1 - h_{ii}} away from zero.}
}
\value{
List with \code{h_diag} and \code{C_list}, both partition-wise.
}
\description{
Computes the per-observation diagonal of the constrained hat matrix used
during tuning, without forming the full \eqn{N \times N} matrix.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prior_loglik.R
\name{prior_loglik}
\alias{prior_loglik}
\title{Log-Prior Distribution Evaluation for lgspline Models}
\usage{
prior_loglik(model_fit, sigmasq = NULL, include_constant = TRUE)
}
\arguments{
\item{model_fit}{An lgspline model object.}

\item{sigmasq}{Numeric scalar dispersion parameter. If NULL,
\code{model_fit$sigmasq_tilde} is used.}

\item{include_constant}{Logical; if TRUE (default), include the
multivariate normal normalizing constant.}
}
\value{
A numeric scalar representing the prior log-likelihood.
}
\description{
Evaluates the log-prior on the spline coefficients conditional on the
dispersion and penalty matrices.
}
\details{
Returns the quadratic form of \eqn{\beta^{T}\Lambda\beta} evaluated at the
tuned or fixed penalties, scaled by negative one-half inverse dispersion.

Assuming fixed penalties, the prior on \eqn{\beta} is taken to be
\deqn{\beta | \sigma^2 \sim \mathcal{N}(\textbf{0}, \sigma^2\Lambda^{-1})}
so that, up to a normalizing constant \eqn{C} with respect to \eqn{\beta},
\deqn{\implies \log P(\beta|\sigma^2) = C-\frac{1}{2\sigma^2}\beta^{T}\Lambda\beta}

The value of \eqn{C} is included when \code{include_constant = TRUE}, and
omitted when \code{FALSE}.

This is useful for computing joint penalized log-likelihoods and related
MAP-style diagnostics for a fitted \code{lgspline} object.
}
\examples{

\donttest{
## Data
t <- sort(runif(100, -5, 5))
y <- sin(t) - 0.1*t^2 + rnorm(100)

## Model keeping penalties fixed
model_fit <- lgspline(t, y, opt = FALSE)

## Full joint log-likelihood, conditional upon known sigma^2 = 1
jntloglik <- sum(dnorm(model_fit$y,
                    model_fit$ytilde,
                    1,
                    log = TRUE)) +
          prior_loglik(model_fit, sigmasq = 1)
print(jntloglik)
}

}
\seealso{
\code{\link{lgspline}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{gramMatrix}
\alias{gramMatrix}
\title{Compute Gram Matrix}
\usage{
gramMatrix(X)
}
\arguments{
\item{X}{Input matrix}
}
\value{
Gram matrix (X^T * X)
}
\description{
Calculates X^T * X for the input matrix X
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{unconstrained_fit_default}
\alias{unconstrained_fit_default}
\title{Unconstrained Generalized Linear Model Estimation}
\usage{
unconstrained_fit_default(
  X,
  y,
  LambdaHalf,
  Lambda,
  keep_weighted_Lambda,
  family,
  tol,
  K,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  order_indices,
  weights,
  ...
)
}
\arguments{
\item{X}{Design matrix of predictors}

\item{y}{Response variable vector}

\item{LambdaHalf}{Square root of penalty matrix
(\eqn{\boldsymbol{\Lambda}^{1/2}})}

\item{Lambda}{Penalty matrix (\eqn{\boldsymbol{\Lambda}})}

\item{keep_weighted_Lambda}{Logical flag to control penalty matrix handling:
- \code{TRUE}: Return coefficients directly from weighted penalty fitting
- \code{FALSE}: Apply damped Newton-Raphson optimization to refine
  estimates}

\item{family}{Distribution family specification}

\item{tol}{Convergence tolerance}

\item{K}{Number of partitions minus one (\eqn{K})}

\item{parallel}{Flag for parallel processing}

\item{cl}{Cluster object for parallel computation}

\item{chunk_size}{Processing chunk size}

\item{num_chunks}{Number of computational chunks}

\item{rem_chunks}{Remaining chunks}

\item{order_indices}{Observation ordering indices}

\item{weights}{Optional observation weights}

\item{...}{Additional arguments passed to \code{glm.fit}}
}
\value{
Numeric column vector of unconstrained coefficient estimates.

For fitting non-canonical GLMs, use \code{keep_weighted_Lambda = TRUE}
since the score and Hessian equations below are no longer valid.

For Gamma(link='log') using \code{keep_weighted_Lambda = TRUE} is
misleading. The information is weighted by a constant (shape parameter)
rather than some mean-variance relationship. So
\code{keep_weighted_Lambda = TRUE} is highly recommended for log-link Gamma
models. This constant flushes into the penalty terms, and so the
formulation of the information matrix is valid.

For other scenarios, like probit regression, there will be diagonal weights
incorporated into the penalty matrix for providing initial MLE estimates,
which technically imposes a prior distribution on beta coefficients that
isn't by intent.

Heuristically, it shouldn't affect much, as these will be updated to their
proper form when providing estimates under constraint; lgspline otherwise
does use the correct form of score and information afterwards,
regardless of canonical/non-canonical status,
as long as 'glm_weight_function' and 'qp_score_function' are properly
specified.
}
\description{
Fits generalized linear models without smoothing constraints
using penalized maximum likelihood estimation. This is applied to each
partition to obtain the unconstrained estimates, prior to imposing the
smoothing constraints.

Hot-start estimates are initialized by treating the matrix square-root inverse
as pseudo-observations, appending to the rows of the design matrix, and
calling \code{\link{glm.fit}} replacing the response for pseudo-observations
with the inverse link function applied to the value of eta = XB = 0. For log
link, exp(0) = 1 is valid for most families. For cases like inverse link,
dividing by 0 is obviously not possible, so this is replaced with 1/tol
where tol is the convergence tolerance argument.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{compute_G_eigen}
\alias{compute_G_eigen}
\title{Compute Eigenvalues and Related Matrices for G}
\usage{
compute_G_eigen(
  X_gram,
  Lambda,
  K,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  family,
  unique_penalty_per_partition,
  L_partition_list,
  keep_G = TRUE,
  schur_corrections
)

compute_G_eigen(
  X_gram,
  Lambda,
  K,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  family,
  unique_penalty_per_partition,
  L_partition_list,
  keep_G = TRUE,
  schur_corrections
)
}
\arguments{
\item{X_gram}{List of Gram matrices \eqn{X_k^T W_k X_k} by partition.}

\item{Lambda}{Penalty matrix \eqn{\Lambda} (shared across partitions).}

\item{K}{Integer; number of interior knots (partitions = \eqn{K+1}).}

\item{parallel}{Logical; use parallel processing across partitions.}

\item{cl}{Cluster object from \code{parallel::makeCluster}.}

\item{chunk_size, num_chunks, rem_chunks}{Partition distribution parameters.}

\item{family}{GLM family object.}

\item{unique_penalty_per_partition}{Logical; if \code{TRUE}, add
partition-specific penalties from \code{L_partition_list}.}

\item{L_partition_list}{List of partition-specific penalty matrices.}

\item{keep_G}{Logical; if \code{TRUE}, return the full \eqn{G} matrix.}

\item{schur_corrections}{List of Schur complement correction matrices.}
}
\value{
A list with components \code{G} (or \code{NULL} blocks when
  \code{keep_G = FALSE}), \code{Ghalf}, and optionally \code{GhalfInv}.

A list with components \code{G} (or \code{NULL} blocks when
  \code{keep_G = FALSE}), \code{Ghalf}, and optionally \code{GhalfInv}.
}
\description{
Computes partition-wise inverse-information matrices and their matrix square
roots from the penalized information matrix via eigendecomposition.

Computes partition-wise inverse-information matrices and their matrix square roots
from the penalized information matrix via eigendecomposition.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/negbin_helpers.R
\name{unconstrained_fit_negbin}
\alias{unconstrained_fit_negbin}
\title{Unconstrained NB Estimation for lgspline}
\usage{
unconstrained_fit_negbin(
  X,
  y,
  LambdaHalf,
  Lambda,
  keep_weighted_Lambda,
  family,
  tol = 1e-08,
  K,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  order_indices,
  weights
)
}
\arguments{
\item{X}{Design matrix (N_k x p) for partition k.}

\item{y}{Response counts for partition k.}

\item{LambdaHalf}{Square root of penalty matrix.}

\item{Lambda}{Penalty matrix.}

\item{keep_weighted_Lambda}{Logical; if TRUE, return hot-start estimates
from augmented Poisson regression without refinement.}

\item{family}{NB family object.}

\item{tol}{Convergence tolerance.}

\item{K}{Number of knots.}

\item{parallel}{Logical for parallel processing.}

\item{cl}{Cluster object.}

\item{chunk_size, }{num_chunks, rem_chunks Parallelism parameters.}

\item{order_indices}{Observation indices mapping partition to full data.}

\item{weights}{Observation weights.}
}
\value{
Numeric column vector of penalized coefficient estimates.
}
\description{
Estimates penalized NB2 coefficients for a single partition (or the
full model when K = 0) using a two-stage optimization: outer loop
profiles over \eqn{\theta} via Brent's method, inner loop estimates
\eqn{\boldsymbol{\beta}} via damped Newton-Raphson on the penalized
log-likelihood.
}
\details{
The penalized log-likelihood is
\deqn{\ell_p(\boldsymbol{\beta}, \theta) = \ell(\boldsymbol{\beta},
  \theta) - \tfrac{1}{2}\boldsymbol{\beta}^{\top}
  \boldsymbol{\Lambda}\boldsymbol{\beta}}

The outer loop optimizes \eqn{\theta} given the inner-loop-optimal
\eqn{\boldsymbol{\beta}(\theta)}. This mirrors the Weibull AFT
approach where scale (\eqn{\sigma}) is profiled out.

Initialization uses Poisson regression coefficients as a hot start
(equivalent to \eqn{\theta \to \infty}).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{efficient_matrix_mult}
\alias{efficient_matrix_mult}
\title{Efficient Matrix Multiplication}
\usage{
efficient_matrix_mult(A, B)
}
\arguments{
\item{A}{First input matrix}

\item{B}{Second input matrix}
}
\value{
Matrix product of A and B
}
\description{
Performs matrix multiplication using RcppArmadillo
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{generate_posterior_correlation}
\alias{generate_posterior_correlation}
\title{Generate Posterior Samples Propagating Correlation Parameter Uncertainty}
\usage{
generate_posterior_correlation(
  object,
  new_sigmasq_tilde = object$sigmasq_tilde,
  new_predictors = NULL,
  theta_1 = 0,
  theta_2 = 0,
  posterior_predictive_draw = function(N, mean, sqrt_dispersion, ...) {
     rnorm(N,
    mean, sqrt_dispersion)
 },
  draw_dispersion = TRUE,
  include_posterior_predictive = FALSE,
  num_draws = 1,
  enforce_qp_constraints = TRUE,
  correlation_param_mean = NULL,
  correlation_param_vcov_sc = NULL,
  correlation_VhalfInv_fxn = NULL,
  correlation_Vhalf_fxn = NULL,
  include_warnings = TRUE,
  ...
)
}
\arguments{
\item{object}{A fitted \code{lgspline} object with a correlation structure
(i.e., \code{VhalfInv_fxn} and \code{VhalfInv_params_estimates}
present, or supplied via override arguments).}

\item{new_sigmasq_tilde}{Numeric; dispersion starting value when
\code{draw_dispersion = FALSE}. Default \code{object$sigmasq_tilde}.}

\item{new_predictors}{Matrix or NULL; predictor matrix for posterior
predictive sampling. Default uses in-sample predictors.}

\item{theta_1}{Numeric; shape increment for the inverse-gamma prior.
Default 0.}

\item{theta_2}{Numeric; rate increment for the inverse-gamma prior.
Default 0.}

\item{posterior_predictive_draw}{Function; sampler for posterior predictive
realisations. Default \code{rnorm}.}

\item{draw_dispersion}{Logical; sample \eqn{\sigma^2} within each
draw. Default TRUE.}

\item{include_posterior_predictive}{Logical; generate posterior predictive
draws. Default FALSE.}

\item{num_draws}{Positive integer; number of draws (each requires one
correlation parameter sample and one covariance rebuild).
Default 1.}

\item{enforce_qp_constraints}{Logical; if TRUE, enforce active QP
inequality constraints during each coefficient draw via the
stored elliptical-slice constrained sampler. Default TRUE.}

\item{correlation_param_mean}{Numeric vector or NULL; mean of the
approximate normal posterior on the working scale. Default:
\code{object$VhalfInv_params_estimates}. Supplying this allows
correlation draws for models fit with a fixed (non-optimised)
correlation structure.}

\item{correlation_param_vcov_sc}{Matrix or NULL; variance-covariance
on the working scale. Default:
\code{object$VhalfInv_params_vcov}. No further scaling is
applied within this function.}

\item{correlation_VhalfInv_fxn}{Function or NULL; maps parameter vector
to \eqn{\mathbf{V}^{-1/2}}. Default \code{object$VhalfInv_fxn}.}

\item{correlation_Vhalf_fxn}{Function or NULL; maps to
\eqn{\mathbf{V}^{1/2}}. Not consumed in the current method body
but resolved for interface consistency and potential future use.}

\item{include_warnings}{Logical; emit warnings. Default TRUE.}

\item{...}{Additional arguments forwarded to the GLM weight function,
dispersion function, and \code{posterior_predictive_draw}.}
}
\value{
When \code{num_draws = 1}, a named list:
\describe{
  \item{post_draw_coefficients}{List of length K+1; per-partition
        coefficient vectors on the original scale.}
  \item{post_draw_sigmasq}{Drawn dispersion.}
  \item{post_pred_draw}{Posterior predictive vector (only when
        \code{include_posterior_predictive = TRUE}).}
  \item{post_draw_correlation_params}{Drawn correlation parameters
        on the working scale.}
}
When \code{num_draws > 1}:
\describe{
  \item{post_draw_coefficients}{List of \code{num_draws} lists of K+1
        coefficient vectors.}
  \item{post_draw_sigmasq}{List of \code{num_draws} scalars.}
  \item{post_pred_draw}{\eqn{N_{\mathrm{new}} \times M} matrix, where
        \eqn{M = \mathrm{num\_draws}} (only when
        \code{include_posterior_predictive = TRUE}).}
  \item{post_draw_correlation_params}{List of \code{num_draws} vectors.}
}
}
\description{
Called internally by \code{\link{generate_posterior}} when
\code{draw_correlation = TRUE}, but can be used directly for finer control.
For each draw, samples the correlation parameter vector from its approximate
normal posterior, rebuilds the posterior covariance under that drawn
correlation structure without re-solving for a new coefficient mode, then
draws coefficients from the updated posterior.
}
\details{
Each draw proceeds in three steps:
\enumerate{
  \item \strong{Draw correlation parameters.}
    \eqn{\boldsymbol{\rho}^{(m)} \sim
    \mathcal{N}(\hat{\boldsymbol{\rho}}_{\mathrm{REML}},
              \mathbf{H}^{-1}_{\mathrm{BFGS}})}
    on the unbounded working scale. Draws producing a non-PD correlation
    matrix are rejected and redrawn (up to 50 attempts); if all fail, the
    point estimate is used with a warning.

  \item \strong{Rebuild posterior covariance.}
    Using the already-expanded \eqn{\mathbf{X}_k}, \eqn{\mathbf{A}}, and
    \eqn{\boldsymbol{\Lambda}} from the original fit, recompute only the
    covariance-side quantities implied by the drawn correlation structure:
    \deqn{
      \mathbf{G}_{\mathrm{correct}}^{(m)} =
      \left(\mathbf{X}^{\top} \mathbf{W}\mathbf{D}
      \mathbf{V}^{-1}(\boldsymbol{\rho}^{(m)}) \mathbf{X}
      + \boldsymbol{\Lambda}\right)^{-1},
    }
    and from this the updated constraint projection
    \eqn{\mathbf{U}^{(m)}} and effective degrees of freedom
    \eqn{\mathrm{trace}(\mathbf{H}^{(m)})}.

    The coefficient mode \eqn{\hat{\boldsymbol{\beta}}_{\mathrm{raw}}}
    and fitted mean \eqn{\tilde{\mathbf{y}}} are held fixed at the
    original fit values. Knot placement, partitioning, polynomial
    expansion, penalty tuning, and coefficient re-estimation are all
    skipped entirely.

  \item \strong{Draw coefficients.}
    Updated quantities (\code{U}, \code{Ghalf_correct}, \code{VhalfInv},
    \code{sigmasq_tilde}, \code{trace_XUGX}) are passed to the stored
    closure via \code{override_*} arguments so that the draw is centred
    at the original mode but uses the covariance implied by the drawn
    correlation structure. The stored mode \code{object$B_raw} is passed
    as \code{override_B_raw} so it is not recomputed.
}

\strong{Why the mode is held fixed.}
Re-solving for a new MAP estimate under each drawn correlation structure
is expensive, requires iterative solvers, and risks convergence failures
on draws far from the REML estimate. The posterior draw is centred at the
original mode, which remains a reasonable approximation when the REML
surface is not sharply peaked. The covariance update captures the primary
effect of correlation uncertainty on posterior width and shape.

\strong{BFGS inverse Hessian caveat.}
The BFGS inverse Hessian approximation for the correlation parameter
covariance is asymptotically valid but may be poor for small samples,
near-boundary estimates, or multimodal REML surfaces. It is not guaranteed
to converge to the observed information matrix. Users should inspect
\code{object$VhalfInv_params_vcov} before relying on these draws.
}
\examples{
## See ?generate_posterior for a complete worked example.

}
\seealso{
\code{\link{generate_posterior}},
\code{\link{lgspline}},
\code{\link{lgspline.fit}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{matmult_U}
\alias{matmult_U}
\title{Left-Multiply a List of Block-Diagonal Matrices by U}
\usage{
matmult_U(U, G, nc, K)
}
\arguments{
\item{U}{Matrix of dimension P by P that projects onto the null space of A transpose}

\item{G}{List of G matrices for each partition}

\item{nc}{Numeric numeric of basis expansions of predictors per partition}

\item{K}{Number of blocks}
}
\value{
Matrix of UG
}
\description{
Useful for computing UG where G is a list of block-diagonal matrices for each partition, and U is a square P by P dense matrix
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_Lambda.R
\name{.compute_gcvu}
\alias{.compute_gcvu}
\title{Evaluate Modified GCV_u Criterion at a Given Penalty Configuration}
\usage{
.compute_gcvu(par, log_penalty_vec, env, ...)
}
\arguments{
\item{par}{Numeric vector; log-scale penalty parameters. First two elements
are log(wiggle_penalty) and log(flat_ridge_penalty). Remaining elements
(if any) are log-scale predictor/partition penalties.}

\item{log_penalty_vec}{Numeric vector; log-scale predictor/partition
penalties (passed separately for compatibility with the grid search).}

\item{env}{List; pre-computed objects and tuning configuration. Contains:
\describe{
  \item{X, y, X_gram, Xy}{Design matrices, response, Gram matrices,
    cross-products.}
  \item{A, R_constraints, K, p_expansions, N_obs}{Constraint matrix and
    dimensions.}
  \item{smoothing_spline_penalty}{Integrated second-derivative penalty
    matrix.}
  \item{custom_penalty_mat}{Optional custom penalty matrix.}
  \item{colnm_expansions}{Character vector of expansion column names.}
  \item{unique_penalty_per_predictor, unique_penalty_per_partition}{Logicals.}
  \item{family}{GLM family object.}
  \item{delta}{Pseudocount for link function stabilization.}
  \item{gcv_gamma}{Modified-GCV inflation factor applied to the
    effective-degrees-of-freedom term in the denominator.}
  \item{meta_penalty}{Meta-penalty coefficient.}
  \item{order_list}{List of observation indices per partition.}
  \item{observation_weights, homogenous_weights}{Observation weighting.}
  \item{parallel flags, cl, chunk_size, num_chunks, rem_chunks}{Parallel
    config.}
  \item{unconstrained_fit_fxn, keep_weighted_Lambda, iterate}{Fitting
    config.}
  \item{qp_score_function, quadprog, qp_Amat, qp_bvec, qp_meq}{QP config.}
  \item{tol, sd_y}{Convergence tolerance and response scale.}
  \item{constraint_value_vectors}{Constraint values.}
  \item{glm_weight_function, schur_correction_function}{GLM functions.}
  \item{need_dispersion_for_estimation, dispersion_function}{Dispersion.}
  \item{blockfit, just_linear_without_interactions}{Blockfit config.}
  \item{use_blockfit}{Logical; pre-computed dispatch flag for blockfit.}
  \item{flat_cols}{Integer vector; pre-computed flat column indices.}
  \item{Vhalf, VhalfInv}{Correlation structure matrices.}
  \item{verbose, include_warnings}{Output control.}
}}

\item{...}{Additional arguments passed to fitting functions.}
}
\value{
List containing:
\describe{
  \item{GCV_u}{Numeric; GCV_u criterion value including meta-penalty.}
  \item{B}{List; fitted coefficient vectors by partition.}
  \item{GXX}{List; \eqn{\mathbf{G}_{k} \mathbf{X}_{k}^{\top}\mathbf{X}_{k}}
    matrices.}
  \item{G_list}{List; eigendecomposition results from
    \code{compute_G_eigen}.}
  \item{mean_W}{Numeric; \eqn{\mathrm{tr}(\mathbf{H})/N}, the average
    leverage entering the denominator of \eqn{\mathrm{GCV}_u}.}
  \item{sum_W}{Numeric; \eqn{\mathrm{tr}(\mathbf{H})}, the effective
    degrees of freedom.}
  \item{Lambda}{Matrix; combined penalty matrix
    \eqn{\boldsymbol{\Lambda}}.}
  \item{L1}{Matrix; baseline smoothness penalty component.}
  \item{L2}{Matrix; baseline ridge penalty component.}
  \item{L_predictor_list}{List; predictor-specific penalty matrices.}
  \item{L_partition_list}{List; partition-specific penalty matrices.}
  \item{numerator}{Numeric; sum of squared residuals.}
  \item{denominator}{Numeric; modified-GCV denominator
    \eqn{N(1 - \gamma\bar{W})^{2}}.}
  \item{residuals}{List; residual vectors by partition.}
  \item{denom_sq}{Numeric; squared denominator.}
  \item{AGAInv}{Matrix; \eqn{(\mathbf{A}^{\top}\mathbf{G}\mathbf{A})^{-1}}.}
}
}
\description{
Computes the modified GCV_u criterion for a given set of penalty
parameters. This is the objective function minimized during penalty tuning;
\code{gcv_gamma = 1} recovers ordinary GCV.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_Lambda.R
\name{.build_tuning_env}
\alias{.build_tuning_env}
\title{Build Tuning Environment}
\usage{
.build_tuning_env(
  y,
  X,
  X_gram,
  Xy,
  smoothing_spline_penalty,
  A,
  R_constraints,
  K,
  p_expansions,
  N_obs,
  custom_penalty_mat,
  colnm_expansions,
  unique_penalty_per_predictor,
  unique_penalty_per_partition,
  meta_penalty,
  family,
  delta,
  order_list,
  observation_weights,
  homogenous_weights,
  parallel,
  parallel_eigen,
  parallel_trace,
  parallel_aga,
  parallel_matmult,
  parallel_unconstrained,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  unconstrained_fit_fxn,
  keep_weighted_Lambda,
  iterate,
  qp_score_function,
  quadprog,
  qp_Amat,
  qp_bvec,
  qp_meq,
  tol,
  sd_y,
  tuning_criterion,
  gcv_gamma,
  constraint_value_vectors,
  glm_weight_function,
  schur_correction_function,
  need_dispersion_for_estimation,
  dispersion_function,
  blockfit,
  just_linear_without_interactions,
  Vhalf,
  VhalfInv,
  verbose,
  include_warnings,
  flat_cols,
  use_blockfit
)
}
\arguments{
\item{gcv_gamma}{Numeric scalar, at least 1; modified-GCV multiplier for the
effective-degrees-of-freedom term.}

\item{flat_cols}{Integer vector; pre-computed flat column indices (passed
in from \code{tune_Lambda} to avoid recomputation).}

\item{use_blockfit}{Logical; pre-computed dispatch flag (passed in from
\code{tune_Lambda}).}
}
\value{
Named list (the "tuning environment").
}
\description{
Assembles a named list containing all pre-computed objects and configuration
needed by the tuning-objective evaluation and gradient functions during
penalty tuning.
This avoids deep nesting of closures and makes the dependencies explicit.
}
\details{
In addition to the standard tuning arguments, the environment stores two
pre-computed blockfit dispatch items:
\describe{
  \item{use_blockfit}{Logical; TRUE when \code{blockfit} is enabled,
    \code{flat_cols} is non-empty, and \code{K > 0}. Mirrors the dispatch
    logic in \code{lgspline.fit} so that the same fitting path is used
    during tuning as during the final fit.}
  \item{flat_cols}{Integer vector; column indices of non-interactive linear
    terms derived from \code{just_linear_without_interactions} and
    \code{colnm_expansions}. Pre-computed once here rather than re-derived
    at every tuning-objective evaluation.}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{compute_dG_dlambda}
\alias{compute_dG_dlambda}
\title{Compute Derivative of Inverse-Information Matrix G with Respect to Lambda}
\usage{
compute_dG_dlambda(
  G,
  dPenalty_dlambda,
  K,
  unique_penalty_per_partition,
  dPenalty_partition_list,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks
)
}
\arguments{
\item{G}{A list of inverse-information matrices \eqn{\textbf{G}} for each
partition}

\item{dPenalty_dlambda}{Derivative of the shared penalty matrix with
respect to \eqn{\lambda}.}

\item{K}{Number of partitions minus 1 (\eqn{K})}

\item{unique_penalty_per_partition}{Logical indicating partition-specific penalties}

\item{dPenalty_partition_list}{Optional list of derivatives of the
partition-specific penalty matrices with respect to \eqn{\lambda}.}

\item{parallel}{Logical to enable parallel processing}

\item{cl}{Cluster object for parallel computation}

\item{chunk_size}{Size of chunks for parallel processing}

\item{num_chunks}{Number of chunks}

\item{rem_chunks}{Remainder chunks}
}
\value{
A list of derivative matrices \eqn{d\textbf{G}/d\lambda} for each partition
}
\description{
Calculates the derivative of the inverse-information matrix
\eqn{\textbf{G}} with respect to the smoothing parameter \eqn{\lambda},
supporting both shared and partition-specific penalties.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{coef.wald_lgspline}
\alias{coef.wald_lgspline}
\title{Extract Coefficients from a wald_lgspline Object}
\usage{
\method{coef}{wald_lgspline}(object, ...)
}
\arguments{
\item{object}{A \code{"wald_lgspline"} object.}

\item{...}{Not used.}
}
\value{
Named numeric vector of coefficient estimates, or NULL if no
  estimate column is available.
}
\description{
Extract Coefficients from a wald_lgspline Object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_Lambda.R
\name{.damped_bfgs}
\alias{.damped_bfgs}
\title{Damped BFGS Optimizer for Penalty Tuning}
\usage{
.damped_bfgs(
  par,
  log_penalty_vec,
  criterion_fxn,
  gr_fxn,
  env,
  tol,
  max_iter = 100,
  ...
)
}
\arguments{
\item{par}{Numeric vector; initial log-scale penalty parameters (first two
elements are log(wiggle) and log(flat_ridge)).}

\item{log_penalty_vec}{Numeric vector; log-scale predictor/partition
penalties appended to the optimization vector.}

\item{criterion_fxn}{Function; tuning-objective evaluation function with signature
\code{function(par, log_penalty_vec, env, ...)}.}

\item{gr_fxn}{Function; gradient function with signature
\code{function(par, log_penalty_vec, outlist, env, ...)}.}

\item{env}{List; tuning environment (passed through to
\code{criterion_fxn} and \code{gr_fxn}).}

\item{tol}{Numeric; convergence tolerance for both criterion change and
parameter change.}

\item{max_iter}{Integer; maximum number of BFGS iterations (default 100).}

\item{...}{Additional arguments passed to fitting functions.}
}
\value{
List containing the best parameter vector found, the corresponding
  criterion value, and the number of iterations performed.
}
\description{
Custom implementation of damped BFGS quasi-Newton optimization for
minimizing the selected tuning criterion. Uses step-size damping with
backtracking and Sherman-Morrison-Woodbury inverse Hessian updates.
}
\details{
The optimizer uses the following strategy:
\enumerate{
  \item Iterations 1-2: steepest descent with damping.
  \item Iteration 3+: BFGS quasi-Newton with inverse Hessian approximation
    updated via the standard secant condition. Falls back to identity
    matrix when the update is numerically unstable.
  \item Step acceptance: Armijo-like criterion (accept if
    \eqn{\mathrm{GCV}_{u}^{(\mathrm{new})} \leq
    \mathrm{GCV}_{u}^{(\mathrm{old})}}).
  \item Backtracking: damping factor halved on rejection; terminates
    when damp < \eqn{2^{-10}} (early iterations) or \eqn{2^{-12}}
    (later iterations).
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{expgrid}
\alias{expgrid}
\title{Generate Grid Indices Without expand.grid()}
\usage{
expgrid(vec_list, indices)
}
\arguments{
\item{vec_list}{List of vectors to combine}

\item{indices}{Indices of combinations to return}
}
\value{
Data frame of selected combinations
}
\description{
Generate Grid Indices Without expand.grid()
}
\details{
Returns selected combinations from the cartesian product of \code{vec_list}
without constructing full \code{expand.grid()} for memory efficiency.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{approx_grad}
\alias{approx_grad}
\title{Finite-difference Gradient Computer}
\usage{
approx_grad(x, fn, eps = sqrt(.Machine$double.eps))
}
\arguments{
\item{x}{Numeric vector of function arguments}

\item{fn}{Function returning list(objective, gradient)}

\item{eps}{Numeric scalar, finite difference tolerance}
}
\value{
Numeric vector of finite-difference approximated gradient values in
  the current optimizer sign convention
}
\description{
Computes a finite-difference approximation derived from the objective in
\code{fn} at \code{x}, returned in the sign convention currently used by
\code{efficient_bfgs}.
}
\details{
Used within \code{efficient_bfgs} when \code{fn} does not supply a usable
gradient. In the main \code{lgspline()} correlation-optimization path,
\code{stats::optim()} is used instead when no analytic REML gradient is
available. Internally this helper returns the negated central-difference
approximation rather than the raw derivative.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/negbin_helpers.R
\name{info_negbin}
\alias{info_negbin}
\title{Compute Negative Binomial Observed Information Matrix}
\usage{
info_negbin(X, y, mu, theta, weights = 1)
}
\arguments{
\item{X}{Design matrix (N x p).}

\item{mu}{Mean vector.}

\item{theta}{Shape parameter.}

\item{weights}{Observation weights (default 1).}
}
\value{
Symmetric p x p observed information matrix.
}
\description{
Negative Hessian of the NB2 log-likelihood with respect to
\eqn{\boldsymbol{\beta}} under the log link.
}
\details{
The expected information under log link is
\deqn{\mathbf{I} = \mathbf{X}^{\top}\mathbf{W}\mathbf{X}}
where \eqn{W_{ii} = w_i \mu_i \theta / (\theta + \mu_i)}.
This is the IRLS weight for NB2 with log link.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{compute_dG_u_dlambda_xy}
\alias{compute_dG_u_dlambda_xy}
\title{Derivative of Constrained Penalized Coefficients with Respect to Lambda}
\usage{
compute_dG_u_dlambda_xy(
  AGAInv_AGXy,
  AGAInv,
  G,
  A,
  dG_dlambda,
  nc,
  nca,
  K,
  Xy,
  Ghalf,
  dGhalf,
  GhalfXy_temp,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks
)
}
\arguments{
\item{AGAInv_AGXy}{Product of \eqn{(\mathbf{A}^{T}\mathbf{G}\mathbf{A})^{-1}} and \eqn{\mathbf{A}^{T}\mathbf{G}\mathbf{X}^{T}\mathbf{y}}}

\item{AGAInv}{Inverse of \eqn{\mathbf{A}^{T}\mathbf{G}\mathbf{A}}}

\item{G}{List of \eqn{\mathbf{G}} matrices}

\item{A}{Constraint matrix \eqn{\mathbf{A}}}

\item{dG_dlambda}{List of \eqn{d\mathbf{G}/d\lambda} matrices}

\item{nc}{Number of columns}

\item{nca}{Number of constraint columns}

\item{K}{Number of partitions minus 1 (\eqn{K})}

\item{Xy}{List of \eqn{\mathbf{X}^{T}\mathbf{y}} products}

\item{Ghalf}{List of \eqn{\mathbf{G}^{1/2}} matrices}

\item{dGhalf}{List of \eqn{d\mathbf{G}^{1/2}/d\lambda} matrices}

\item{GhalfXy_temp}{Temporary storage for \eqn{\mathbf{G}^{1/2}\mathbf{X}^{T}\mathbf{y}}}

\item{parallel}{Use parallel processing}

\item{cl}{Cluster object}

\item{chunk_size}{Size of parallel chunks}

\item{num_chunks}{Number of chunks}

\item{rem_chunks}{Remaining chunks}
}
\value{
\eqn{P \times 1} vector of derivatives
  \eqn{d\tilde{\boldsymbol{\beta}}/d\lambda}
}
\description{
Computes \eqn{d(\mathbf{U}\mathbf{G}\mathbf{X}^{T}\mathbf{y})/d\lambda},
the sensitivity of the constrained coefficient estimates
\eqn{\tilde{\boldsymbol{\beta}} = \mathbf{U}\mathbf{G}\mathbf{X}^T\mathbf{y}}
to changes in the smoothing parameter \eqn{\lambda}.
}
\details{
This function is called during GCV/penalty optimization and computes
how the constrained coefficient vector changes as the penalty weight
varies. Two implementations are provided depending on problem size.

\subsection{Derivation}{
The constrained estimate is
\eqn{\tilde{\boldsymbol{\beta}} = \mathbf{U}\hat{\boldsymbol{\beta}}}
where
\eqn{\hat{\boldsymbol{\beta}} = \mathbf{G}\mathbf{X}^T\mathbf{y}}
and
\eqn{\mathbf{U} = \mathbf{I} - \mathbf{G}\mathbf{A}(\mathbf{A}^T\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^T}.
Both \eqn{\mathbf{G}} and \eqn{\mathbf{U}} depend on \eqn{\lambda}
through \eqn{\mathbf{G} = (\mathbf{X}^T\mathbf{X} + \lambda\boldsymbol{\Lambda})^{-1}}.

Differentiating the product \eqn{\mathbf{U}\mathbf{G}\mathbf{X}^T\mathbf{y}}
requires the chain rule applied to three \eqn{\lambda}-dependent
components:
\deqn{\frac{d}{d\lambda}(\mathbf{U}\mathbf{G}\mathbf{X}^T\mathbf{y})}
\deqn{= \frac{d\mathbf{U}}{d\lambda}\hat{\boldsymbol{\beta}} + \mathbf{U}\frac{d\hat{\boldsymbol{\beta}}}{d\lambda}}

The unconstrained part is straightforward:
\eqn{d\hat{\boldsymbol{\beta}}/d\lambda = (d\mathbf{G}/d\lambda)\mathbf{X}^T\mathbf{y}},
computed partition-wise since \eqn{\mathbf{G}} is block-diagonal.

The constraint projection derivative is more involved. Writing
\eqn{\mathbf{C} = (\mathbf{A}^T\mathbf{G}\mathbf{A})^{-1}} and
expanding \eqn{d\mathbf{U}/d\lambda} gives three terms (after
applying the product rule and the identity
\eqn{d\mathbf{C}/d\lambda = -\mathbf{C}(d(\mathbf{A}^T\mathbf{G}\mathbf{A})/d\lambda)\mathbf{C}}):
\describe{
  \item{term2a}{\eqn{(d\mathbf{G}/d\lambda)\mathbf{A}\mathbf{C}\mathbf{A}^T\hat{\boldsymbol{\beta}}}
    direct effect of \eqn{d\mathbf{G}/d\lambda} on the projection}
  \item{term2b}{\eqn{\mathbf{G}\mathbf{A}\mathbf{C}(d(\mathbf{A}^T\mathbf{G}\mathbf{A})/d\lambda)\mathbf{C}\mathbf{A}^T\hat{\boldsymbol{\beta}}}
    effect through the change in \eqn{(\mathbf{A}^T\mathbf{G}\mathbf{A})^{-1}}}
  \item{term2c}{\eqn{\mathbf{G}\mathbf{A}\mathbf{C}\mathbf{A}^T(d\hat{\boldsymbol{\beta}}/d\lambda)}
    projection of the unconstrained derivative back through the constraint}
}

The full derivative is
\eqn{d\hat{\boldsymbol{\beta}}/d\lambda} minus the sum of these
three correction terms.
}

\subsection{Large problem path (K >= 10, nc > 4)}{
Computes the three correction terms explicitly using the block
structure of \eqn{\mathbf{G}} and \eqn{d\mathbf{G}/d\lambda}.
The intermediate quantity
\eqn{d(\mathbf{A}^T\mathbf{G}\mathbf{A})/d\lambda = \mathbf{A}^T(d\mathbf{G}/d\lambda)\mathbf{A}}
is accumulated partition-wise. Shared vectors
\eqn{\mathbf{A}\mathbf{C}\mathbf{A}^T\hat{\boldsymbol{\beta}}}
and related products are precomputed once and reused across
partitions. Parallelism is over chunks of partitions.
}

\subsection{Small problem path}{
Reformulates the derivative using the matrix square root
\eqn{\mathbf{G}^{1/2}} and its derivative
\eqn{d\mathbf{G}^{1/2}/d\lambda}. The constraint
\eqn{\mathbf{A}^T\boldsymbol{\beta} = 0} is imposed via least
squares projection: the residuals from regressing
\eqn{\mathbf{G}^{1/2}\mathbf{X}^T\mathbf{y}} onto
\eqn{\mathbf{G}^{1/2}\mathbf{A}} give the constrained component,
and differentiating this projection with respect to \eqn{\lambda}
yields the derivative. Uses \code{.lm.fit} for speed with a
stabilizing rescaling factor to prevent numerical issues when the
constraint matrix is poorly scaled.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.get_B_gee_glm_woodbury}
\alias{.get_B_gee_glm_woodbury}
\title{Path 1b-Woodbury: Non-Gaussian GEE with Woodbury Acceleration}
\usage{
.get_B_gee_glm_woodbury(
  X,
  y,
  K,
  p_expansions,
  VhalfInv_perm,
  order_list,
  A,
  R_constraints,
  constraint_value_vectors,
  family,
  return_G_getB,
  iterate,
  tol,
  qp_Amat,
  qp_bvec,
  qp_meq,
  qp_score_function,
  observation_weights,
  Lambda,
  Lambda_block,
  unique_penalty_per_partition,
  L_partition_list,
  wb_decomp_init,
  wb_sqrt_init,
  glm_weight_function,
  schur_correction_function,
  need_dispersion_for_estimation,
  dispersion_function,
  VhalfInv,
  parallel_eigen,
  parallel_aga,
  parallel_matmult,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  ...
)
}
\arguments{
\item{X, y}{Lists of partition-specific design matrices and responses.}

\item{K, p_expansions}{Integer dimensions.}

\item{VhalfInv_perm}{\eqn{\mathbf{V}^{-1/2}} permuted to partition
ordering.}

\item{order_list, observation_weights}{Standard partition arguments.}

\item{A}{Constraint matrix (\eqn{P \times R}).}

\item{constraint_value_vectors}{Constraint RHS list encoding
\eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{c}}.}

\item{family}{GLM family object.}

\item{return_G_getB}{Logical; return covariance components.}

\item{iterate}{Logical; if \code{TRUE}, iterate until convergence.}

\item{tol}{Convergence tolerance.}

\item{qp_Amat, qp_bvec, qp_meq}{QP constraint specification.}

\item{qp_score_function}{Score function for QP step.}

\item{Lambda, Lambda_block}{Shared and full penalty matrices.}

\item{unique_penalty_per_partition}{Logical.}

\item{L_partition_list}{Partition-specific penalty matrices.}

\item{wb_decomp_init}{Initial Woodbury decomposition from
\code{.woodbury_decompose_V} (used only to confirm that the
unweighted decomposition was valid; the weighted decomposition is
recomputed at each iteration).}

\item{wb_sqrt_init}{Initial half-sqrt components (unused; retained
for interface consistency with the Gaussian Woodbury path).}

\item{glm_weight_function}{Function computing GLM working weights.}

\item{schur_correction_function}{Function computing Schur corrections.}

\item{need_dispersion_for_estimation}{Logical.}

\item{dispersion_function}{Dispersion estimation function.}

\item{VhalfInv}{Inverse square root correlation matrix.}

\item{parallel_eigen, parallel_aga, parallel_matmult}{Logical flags.}

\item{cl, chunk_size, num_chunks, rem_chunks}{Parallel parameters.}

\item{...}{Passed to sub-functions.}
}
\value{
Same structure as \code{.get_B_gee_glm}.
}
\description{
Non-Gaussian correlated solve using the Woodbury factorization. At each
damped Newton step, the current weighted information matrix is split into
a block-diagonal part defining \eqn{\mathbf{G}_{\mathrm{on}}} and a
low-rank cross-partition part
\eqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{E}\mathbf{J}\mathbf{E}^{\top}}.
The constrained step is then taken through
\code{.lagrangian_project_woodbury()}.
}
\details{
This function is standalone for the weighted Woodbury path: it carries
the iteration, repeatedly rebuilds the Woodbury factors at the current
working weights, and falls back to the dense correlated GLM solver if the
rank is too large or \eqn{\mathbf{F}} is not positive definite.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{softplus}
\alias{softplus}
\title{Compute softplus transform}
\usage{
softplus(x)
}
\arguments{
\item{x}{Numeric vector to apply softplus to}
}
\value{
Softplus transformed vector
}
\description{
Computes the softplus transform, equivalent to the cumulant generating function
of a logistic regression model: \eqn{\log(1+e^x)}.
}
\examples{

x <- runif(5)
softplus(x)


}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{knot_expand_list}
\alias{knot_expand_list}
\title{Expand Matrix into Partition Lists Based on Knot Boundaries}
\usage{
knot_expand_list(partition_codes, partition_bounds, N, mat, K)
}
\arguments{
\item{partition_codes}{Numeric vector; values determining partition assignment for each row}

\item{partition_bounds}{Numeric vector; ordered knot locations defining partition boundaries}

\item{N}{Integer; number of rows in input matrix}

\item{mat}{Numeric matrix; data to be partitioned}

\item{K}{Integer; number of interior knots (resulting in \eqn{K+1} partitions)}
}
\value{
List of length \eqn{K+1}, each element containing the submatrix for that partition
}
\description{
Takes an input \eqn{N \times p} matrix of polynomial expansions and outputs a list of
length \eqn{K+1}, isolating the rows of the input corresponding to assigned partition.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{.compute_dist_block}
\alias{.compute_dist_block}
\title{Compute Euclidean distance matrix for a cluster block}
\usage{
.compute_dist_block(spacetime, inds)
}
\description{
Returns pairwise Euclidean distances within the cluster indexed by
\code{inds}. When \code{spacetime} has multiple columns, squared distances
are averaged across dimensions before taking the square root.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weibull_helpers.R
\name{weibull_dispersion_function}
\alias{weibull_dispersion_function}
\title{Estimate Weibull Dispersion for Accelerated Failure Time Model}
\usage{
weibull_dispersion_function(
  mu,
  y,
  order_indices,
  family,
  observation_weights,
  VhalfInv,
  status
)
}
\arguments{
\item{mu}{Predicted survival times}

\item{y}{Observed response/survival times}

\item{order_indices}{Indices to align status with response}

\item{family}{Weibull AFT model family specification; unused here and
retained for interface compatibility.}

\item{observation_weights}{Optional observation weights}

\item{VhalfInv}{Inverse square root of the correlation matrix; unused here
and retained for interface compatibility.}

\item{status}{Censoring indicator (1 = event, 0 = censored)
Indicates whether an event of interest occurred (1) or the observation was
right-censored (0). In survival analysis, right-censoring occurs when the
full survival time is unknown, typically because the study ended or the
subject was lost to follow-up before the event of interest occurred.}
}
\value{
Dispersion estimate (sigma^2) for the Weibull AFT model, i.e., the squared
scale parameter. The Weibull scale (sigma) matching \code{survreg$scale} is
\code{sqrt()} of this value.
}
\description{
Computes the dispersion parameter (sigma^2 = scale^2) for a Weibull
accelerated failure time (AFT) model, supporting right-censored survival
data. The returned value is sigma^2, where sigma is the Weibull scale
parameter matching \code{survreg$scale}.

This both provides a tool for actually fitting Weibull AFT Models, and
boilerplate code for users who wish to incorporate Lagrangian multiplier
smoothing splines into their own custom models.
}
\examples{

## Simulate survival data with covariates
set.seed(1234)
n <- 1000
t1 <- rnorm(n)
t2 <- rbinom(n, 1, 0.5)

## Generate survival times with Weibull-like structure
lambda <- exp(0.5 * t1 + 0.3 * t2)
yraw <- rexp(n, rate = 1/lambda)

## Introduce right-censoring
status <- rbinom(n, 1, 0.75)
y <- ifelse(status, yraw, runif(length(yraw), 0, yraw))

## Example of using dispersion function
mu <- mean(y)
order_indices <- seq_along(y)
weights <- rep(1, n)

## Estimate dispersion (= scale^2 = sigma^2)
dispersion_est <- weibull_dispersion_function(
  mu = mu,
  y = y,
  order_indices = order_indices,
  family = weibull_family(),
  observation_weights = weights,
  VhalfInv = NULL,
  status = status
)

print(dispersion_est)          # sigma^2
print(sqrt(dispersion_est))    # sigma (comparable to survreg$scale)

}
\seealso{
\code{\link{weibull_scale}} for the underlying scale estimation
  function
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{make_constraint_matrix}
\alias{make_constraint_matrix}
\title{Create Smoothing Spline Constraint Matrix}
\usage{
make_constraint_matrix(
  p_expansions,
  CKnots,
  power1_cols,
  power2_cols,
  nonspline_cols,
  interaction_single_cols,
  interaction_quad_cols,
  triplet_cols,
  K,
  include_constrain_fitted,
  include_constrain_first_deriv,
  include_constrain_second_deriv,
  include_constrain_interactions,
  include_2way_interactions,
  include_3way_interactions,
  include_quadratic_interactions,
  colnm_expansions,
  expansion_scales
)
}
\arguments{
\item{p_expansions}{Integer; number of columns in basis expansion per partition}

\item{CKnots}{Matrix; basis expansions evaluated at knot points}

\item{power1_cols}{Integer vector; indices of linear terms}

\item{power2_cols}{Integer vector; indices of quadratic terms}

\item{nonspline_cols}{Integer vector; indices of non-spline terms}

\item{interaction_single_cols}{Integer vector; indices of linear interaction terms}

\item{interaction_quad_cols}{Integer vector; indices of quadratic interaction terms}

\item{triplet_cols}{Integer vector; indices of three-way interaction terms}

\item{K}{Integer; number of interior knots (\eqn{K+1} partitions)}

\item{include_constrain_fitted}{Logical; constrain function values at knots}

\item{include_constrain_first_deriv}{Logical; constrain first derivatives at knots}

\item{include_constrain_second_deriv}{Logical; constrain second derivatives at knots}

\item{include_constrain_interactions}{Logical; constrain interaction terms at knots}

\item{include_2way_interactions}{Logical; include two-way interactions}

\item{include_3way_interactions}{Logical; include three-way interactions}

\item{include_quadratic_interactions}{Logical; include quadratic interactions}

\item{colnm_expansions}{Character vector; column names for basis expansions}

\item{expansion_scales}{Numeric vector; scaling factors for standardization}
}
\value{
Matrix \eqn{\textbf{A}} of constraint coefficients. Columns correspond to
constraints, rows to coefficients across all \eqn{K+1} partitions.
}
\description{
Constructs constraint matrix \eqn{\textbf{A}} enforcing continuity and smoothness at knot boundaries
by constraining function values, derivatives, and interactions between partitions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cox_helpers.R
\name{cox_dispersion_function}
\alias{cox_dispersion_function}
\title{Cox PH Dispersion Function}
\usage{
cox_dispersion_function(
  mu,
  y,
  order_indices,
  family,
  observation_weights,
  VhalfInv,
  ...
)
}
\arguments{
\item{mu}{Predicted values.}

\item{y}{Observed survival times.}

\item{order_indices}{Observation indices.}

\item{family}{Family object.}

\item{observation_weights}{Observation weights.}

\item{VhalfInv}{Inverse square root of correlation matrix.}

\item{...}{Additional arguments (including \code{status}).}
}
\value{
Scalar 1.
}
\description{
Returns 1 unconditionally. Cox PH has no dispersion parameter; this
function exists solely for interface compatibility with lgspline's
\code{dispersion_function} argument.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.get_B_gee_glm}
\alias{.get_B_gee_glm}
\title{Path 1b: Non-Gaussian GEE (Damped SQP with Full Whitened Design)}
\usage{
.get_B_gee_glm(
  X_block,
  X_tilde,
  y_block,
  y_tilde,
  VhalfInv_perm,
  Lambda_block,
  A,
  K,
  p_expansions,
  constraint_value_vectors,
  family,
  return_G_getB,
  iterate,
  tol,
  qp_Amat,
  qp_bvec,
  qp_meq,
  qp_score_function,
  order_list,
  observation_weights,
  glm_weight_function,
  schur_correction_function,
  need_dispersion_for_estimation,
  dispersion_function,
  VhalfInv,
  ...
)
}
\arguments{
\item{X_block}{Full \eqn{N \times P} unwhitened block-diagonal design.}

\item{X_tilde}{Whitened design \eqn{\mathbf{V}^{-1/2}\mathbf{X}}.}

\item{y_block}{Unwhitened response vector \eqn{\mathbf{y}}.}

\item{y_tilde}{Whitened response \eqn{\mathbf{V}^{-1/2}\mathbf{y}}.}

\item{VhalfInv_perm}{\eqn{\mathbf{V}^{-1/2}} permuted to partition
ordering.}

\item{Lambda_block}{Full \eqn{P \times P} block-diagonal penalty.}

\item{A}{Constraint matrix (\eqn{P \times R}).}

\item{K, p_expansions}{Integer dimensions.}

\item{constraint_value_vectors}{Constraint RHS list encoding
\eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{c}}.}

\item{family}{GLM family object.}

\item{return_G_getB}{Logical; return covariance components.}

\item{iterate}{Logical; if \code{TRUE}, iterate until convergence.}

\item{tol}{Convergence tolerance.}

\item{qp_Amat, qp_bvec, qp_meq}{QP constraint specification.}

\item{qp_score_function}{Score function for QP step.}

\item{order_list, observation_weights}{Standard partition arguments.}

\item{glm_weight_function}{Function computing GLM working weights.}

\item{schur_correction_function}{Function computing Schur corrections.}

\item{need_dispersion_for_estimation}{Logical.}

\item{dispersion_function}{Dispersion estimation function.}

\item{VhalfInv}{Inverse square root correlation matrix.}

\item{...}{Passed to weight, correction, dispersion, and score functions.}
}
\value{
Same structure as \code{.get_B_gee_gaussian}.
}
\description{
Estimates constrained coefficients for non-Gaussian GLMs with correlation
structures using damped Sequential Quadratic Programming (SQP) in the
full \eqn{P}-dimensional whitened space. The whitened design
\eqn{\mathbf{X}^{*} = \mathbf{V}^{-1/2}\mathbf{X}} is required
because \eqn{\mathbf{V}^{-1/2}} couples all partitions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.get_B_gaussian_nocorr}
\alias{.get_B_gaussian_nocorr}
\title{Path 2: Gaussian Identity Link, No Correlation}
\usage{
.get_B_gaussian_nocorr(
  Xy,
  Ghalf,
  GhalfInv,
  A,
  K,
  p_expansions,
  R_constraints,
  constraint_value_vectors,
  family,
  return_G_getB,
  quadprog,
  qp_Amat,
  qp_bvec,
  qp_meq,
  qp_score_function,
  parallel_aga,
  parallel_matmult,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  X,
  y,
  Lambda,
  order_list,
  observation_weights,
  iterate,
  tol,
  glm_weight_function,
  schur_correction_function,
  need_dispersion_for_estimation,
  dispersion_function,
  unique_penalty_per_partition,
  L_partition_list,
  VhalfInv,
  homogenous_weights,
  ...
)
}
\arguments{
\item{Xy}{List of cross-products \eqn{\mathbf{X}_k^{\top}\mathbf{y}_k}
by partition.}

\item{Ghalf, GhalfInv}{Lists of \eqn{\mathbf{G}^{1/2}_k} and
\eqn{\mathbf{G}^{-1/2}_k} matrices by partition.}

\item{A}{Constraint matrix \eqn{\mathbf{A}}.}

\item{K, p_expansions, R_constraints}{Integer dimensions.}

\item{constraint_value_vectors}{Constraint RHS list encoding
\eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{c}}.}

\item{family}{GLM family object.}

\item{return_G_getB}{Logical; return covariance components.}

\item{quadprog}{Logical; apply QP refinement for inequality constraints.}

\item{qp_Amat, qp_bvec, qp_meq}{QP constraint specification.}

\item{qp_score_function}{Score function for QP step.}

\item{parallel_aga, parallel_matmult}{Logical flags.}

\item{cl, chunk_size, num_chunks, rem_chunks}{Parallel parameters.}

\item{X, y, Lambda, order_list, observation_weights}{Standard arguments
passed through for optional QP refinement.}

\item{iterate, tol, glm_weight_function, schur_correction_function, }{need_dispersion_for_estimation,dispersion_function,
unique_penalty_per_partition,L_partition_list,VhalfInv,
homogenous_weights Standard arguments passed through.}

\item{...}{Passed to sub-functions.}
}
\value{
Same structure as \code{get_B}.
}
\description{
For the canonical Gaussian case without correlation structures, the
unconstrained estimate has the closed form
\eqn{\hat{\boldsymbol{\beta}}_k = \mathbf{G}_k \mathbf{X}_k^{\top}
\mathbf{y}_k} and the constrained estimate follows by a single
Lagrangian projection. No iteration is needed.
}
\details{
When \eqn{K = 0} and there are no inequality constraints or nonzero
constraint values, the function returns
\eqn{\hat{\boldsymbol{\beta}} = \mathbf{G}\mathbf{X}^{\top}\mathbf{y}}
directly without forming the full \eqn{P}-dimensional OLS system.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.get_B_gee_woodbury}
\alias{.get_B_gee_woodbury}
\title{Path 1a-Woodbury: Gaussian GEE with Woodbury Acceleration}
\usage{
.get_B_gee_woodbury(
  X,
  y,
  K,
  p_expansions,
  VhalfInv_perm,
  order_list,
  A,
  R_constraints,
  constraint_value_vectors,
  family,
  return_G_getB,
  quadprog,
  qp_Amat,
  qp_bvec,
  qp_meq,
  qp_score_function,
  observation_weights,
  wb_decomp,
  wb_sqrt,
  parallel_aga,
  parallel_matmult,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  qp_global,
  ...
)
}
\arguments{
\item{X, y}{Lists of partition-specific design matrices and responses.}

\item{K, p_expansions}{Integer dimensions.}

\item{VhalfInv_perm}{\eqn{\mathbf{V}^{-1/2}} permuted to partition
ordering.}

\item{order_list}{Partition-to-data index mapping.}

\item{A}{Constraint matrix (\eqn{P \times R}).}

\item{R_constraints}{Number of columns of A.}

\item{constraint_value_vectors}{Constraint RHS list.}

\item{family}{GLM family object.}

\item{return_G_getB}{Logical; return covariance components.}

\item{quadprog}{Logical; apply QP refinement.}

\item{qp_Amat, qp_bvec, qp_meq}{QP constraint specification.}

\item{qp_score_function}{Score function for QP step.}

\item{observation_weights}{Observation weights.}

\item{wb_decomp}{Output of \code{.woodbury_decompose_V}.}

\item{wb_sqrt}{Output of \code{.woodbury_halfsqrt_components}.}

\item{parallel_aga, parallel_matmult}{Logical flags.}

\item{cl, chunk_size, num_chunks, rem_chunks}{Parallel parameters.}

\item{qp_global}{Logical; forced TRUE for GEE but passed for
interface consistency.}

\item{...}{Passed to sub-functions.}
}
\value{
Same structure as \code{.get_B_gee_gaussian}.
}
\description{
Gaussian correlated solve using the Woodbury factorization from the
supplement. This replaces the dense GEE path when
\eqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{E}\mathbf{J}\mathbf{E}^{\top}}
has low rank and computes the constrained estimate through the same
transformed OLS projection as the independent case, with
\eqn{\mathbf{G}^{1/2} =
\mathbf{G}_{\mathrm{on}}^{1/2}\mathbf{F}^{1/2}}.
}
\details{
This function is standalone for the Gaussian Woodbury path: it takes the
decomposition from \code{.woodbury_decompose_V()}, the square-root state
from \code{.woodbury_halfsqrt_components()}, performs the constrained
solve, and returns the same object structure as the dense Gaussian GEE
solver.

Cost is \eqn{O(Kp^3 + Pr^2)} compared to \eqn{O(P^3)} for the dense
Path 1a.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{take_interaction_2ndderivative}
\alias{take_interaction_2ndderivative}
\title{Calculate Second Derivatives of Interaction Terms}
\usage{
take_interaction_2ndderivative(
  dat,
  var,
  interaction_single_cols,
  interaction_quad_cols,
  triplet_cols,
  colnm_expansions,
  power1_cols,
  power2_cols
)
}
\arguments{
\item{dat}{Numeric matrix; design matrix containing basis expansions}

\item{var}{Character; variable name to differentiate with respect to}

\item{interaction_single_cols}{Integer vector; column indices for linear-linear interactions}

\item{interaction_quad_cols}{Integer vector; column indices for linear-quadratic interactions}

\item{triplet_cols}{Integer vector; column indices for three-way interactions}

\item{colnm_expansions}{Character vector; column names of expansions for each partition}

\item{power1_cols}{Integer vector; column indices of linear terms}

\item{power2_cols}{Integer vector; column indices of quadratic terms}
}
\value{
Numeric matrix of second derivatives, same dimensions as input
}
\description{
Computes partial second derivatives for interaction terms including
two-way linear, quadratic, and three-way interactions. Handles special cases
for each type.

This function is necessary to compute total second derivatives as the sum of
second partial "pure" derivatives (\eqn{d^2/dx^2}) plus second partial "mixed"
derivative (\eqn{d^2/dxdz}), for a predictor x and all other predictors z.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_Lambda.R
\name{.compute_meta_penalty_gradient}
\alias{.compute_meta_penalty_gradient}
\title{Compute Gradient of Regularization (Meta) Penalty}
\usage{
.compute_meta_penalty_gradient(
  wiggle_penalty,
  penalty_vec,
  meta_penalty_coef,
  unique_penalty_per_predictor,
  unique_penalty_per_partition
)
}
\arguments{
\item{wiggle_penalty}{Numeric; current wiggle penalty on raw scale.}

\item{penalty_vec}{Numeric vector; current predictor/partition penalties
on raw scale. May be empty (\code{c()}).}

\item{meta_penalty_coef}{Numeric; coefficient for the meta-penalty.}

\item{unique_penalty_per_predictor}{Logical; whether predictor-specific
penalties are active.}

\item{unique_penalty_per_partition}{Logical; whether partition-specific
penalties are active.}
}
\value{
Numeric vector; gradient of the meta-penalty on the log scale.
  Length equals 2 + length(penalty_vec).
}
\description{
Computes the gradient of the meta-penalty with respect to the log-scale
penalty parameters, incorporating the exp parameterization chain rule.
}
\details{
Under exp parameterization \eqn{\lambda = \exp(\theta)}:
\deqn{\frac{\partial}{\partial \theta}
  \left[ 0.5 c (\exp(\theta) - 1)^{2} \right]
  = c (\lambda - 1) \lambda}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/solver_utils.R
\name{.solver_assemble_qp_info}
\alias{.solver_assemble_qp_info}
\title{Assemble qp_info from a solve.QP Solution}
\usage{
.solver_assemble_qp_info(
  last_qp_sol,
  beta_block,
  qp_Amat_combined,
  qp_bvec_combined,
  qp_meq_combined,
  converged,
  final_deviance,
  info_matrix = NULL
)
}
\arguments{
\item{last_qp_sol}{Output of \code{quadprog::solve.QP}, or
\code{NULL}.  When \code{NULL} the function returns \code{NULL}.}

\item{beta_block}{Final \eqn{P \times 1} coefficient vector.}

\item{qp_Amat_combined}{Combined equality + inequality constraint
matrix.}

\item{qp_bvec_combined}{Combined constraint right-hand side.}

\item{qp_meq_combined}{Integer; number of leading equality
constraints.}

\item{converged}{Logical; whether the outer loop converged.}

\item{final_deviance}{Scalar deviance at convergence.}

\item{info_matrix}{Optional information matrix; included in the
returned list when non-\code{NULL}.}
}
\value{
A list with elements \code{solution}, \code{lagrangian},
  \code{active_constraints}, \code{iact}, \code{Amat_active},
  \code{bvec_active}, \code{meq_active}, \code{converged}, and
  \code{final_deviance}, plus \code{info_matrix},
  \code{Amat_combined}, \code{bvec_combined}, and
  \code{meq_combined} when \code{info_matrix} is supplied.
  Returns \code{NULL} if \code{last_qp_sol} is \code{NULL}.
}
\description{
Packages the output of \code{quadprog::solve.QP} into the
\code{qp_info} list expected by downstream code (inference,
\code{generate_posterior}, and \code{varcovmat} construction).
}
\details{
Active constraint columns are identified as all equality columns
(\code{1:qp_meq_combined}) plus any inequality column whose
Lagrange multiplier exceeds \code{sqrt(.Machine$double.eps)}.
This matches the convention used in \code{.qp_refine()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.detect_qp_global}
\alias{.detect_qp_global}
\title{Detect Whether Inequality Constraints Require Global (Dense) QP}
\usage{
.detect_qp_global(qp_Amat, p_expansions, K)
}
\arguments{
\item{qp_Amat}{Inequality constraint matrix (\eqn{P \times n_{ineq}}),
or \code{NULL}.}

\item{p_expansions}{Integer; number of basis terms per partition.}

\item{K}{Integer; number of interior knots.}
}
\value{
Logical; \code{TRUE} if dense QP is needed, \code{FALSE} if
  partition-wise active-set is valid.
}
\description{
Thin wrapper around \code{.solver_detect_qp_global()}.
This preserves the original helper name inside \code{get_B()} while
using the shared detection logic defined in \code{solver_utils.R}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/negbin_helpers.R
\name{negbin_qp_score_function}
\alias{negbin_qp_score_function}
\title{NB Score Function for Quadratic Programming and Blockfit}
\usage{
negbin_qp_score_function(
  X,
  y,
  mu,
  order_list,
  dispersion,
  VhalfInv,
  observation_weights
)
}
\arguments{
\item{X}{Block-diagonal design matrix (N x P).}

\item{y}{Response vector (N x 1).}

\item{mu}{Predicted values (N x 1), same order as X and y.}

\item{order_list}{List of observation indices per partition.}

\item{dispersion}{Shape parameter \eqn{\theta}.}

\item{VhalfInv}{Inverse square root of correlation matrix; when non-NULL
the score is computed on the whitened scale as
\eqn{\tilde{X}^{\top}\tilde{r}} where
\eqn{\tilde{X} = V^{-1/2}X} and the residual vector accounts for
the correlation.}

\item{observation_weights}{Observation weights.}
}
\value{
Numeric column vector of length P.
}
\description{
Computes the score (gradient of NB log-likelihood) in the format
expected by lgspline's \code{qp_score_function} interface.
}
\details{
Without correlation (\code{VhalfInv = NULL}), the score is
\eqn{\mathbf{X}^{\top}\mathbf{w} \odot (y - \mu)\theta/(\theta + \mu)}.

With correlation, the GEE score is
\eqn{\tilde{\mathbf{X}}^{\top}\mathrm{diag}(\mathbf{W})
\mathbf{V}^{-1}(\mathbf{y} - \boldsymbol{\mu})}
where \eqn{\mathbf{W}} contains the NB working weights. The
whitening is absorbed by pre-multiplying both \eqn{\mathbf{X}} and
the residual by \eqn{\mathbf{V}^{-1/2}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blockfit_solve.R
\name{.bf_constrained_flat_update}
\alias{.bf_constrained_flat_update}
\title{Constrained Flat Update Given Current Spline Solution}
\usage{
.bf_constrained_flat_update(
  XfWXf_pen,
  Xfr,
  A_full,
  constraint_values,
  beta_spline,
  flat_rows_all,
  spline_rows_all,
  spline_cols,
  flat_cols,
  p_expansions,
  K,
  nc_flat
)
}
\arguments{
\item{XfWXf_pen}{Penalized flat Gram matrix
(\eqn{nc_f \times nc_f}).}

\item{Xfr}{Right-hand side cross-product for flat update
(\eqn{nc_f \times 1}).}

\item{A_full}{Full constraint matrix (\eqn{P \times R}).}

\item{constraint_values}{List of constraint RHS vectors.}

\item{beta_spline}{List of \eqn{K+1} current spline coefficient vectors.}

\item{flat_rows_all}{Integer vector of flat row indices in the full P-space.}

\item{spline_rows_all}{Integer vector of spline row indices in the full P-space.}

\item{spline_cols}{Integer vector of spline column indices within each partition.}

\item{flat_cols}{Integer vector of flat column indices within each partition.}

\item{p_expansions}{Integer; columns per partition.}

\item{K}{Integer; number of interior knots.}

\item{nc_flat}{Integer; number of flat columns.}
}
\value{
Numeric vector of flat coefficients (\eqn{nc_f \times 1}).
}
\description{
Solves for flat coefficients subject to the residual equality
constraint imposed by the full constraint system, conditional on
the current spline block solution.
}
\details{
When the full constraint system is \eqn{\mathbf{A}^\top \boldsymbol{\beta} = \mathbf{c}},
and we partition \eqn{\boldsymbol{\beta}} into spline and flat blocks, the flat
update must satisfy:
\deqn{
\mathbf{A}_{\mathrm{flat}}^\top \boldsymbol{\beta}_{\mathrm{flat}}
= \mathbf{c} - \mathbf{A}_{\mathrm{spline}}^\top \boldsymbol{\beta}_{\mathrm{spline}}
}

This function solves the penalized least-squares problem for flat coefficients
subject to this residual equality constraint using a Lagrangian approach.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{.get_B_gee_gaussian}
\alias{.get_B_gee_gaussian}
\title{Path 1a: Gaussian Identity + GEE (Closed-Form Full-System Solve)}
\usage{
.get_B_gee_gaussian(
  X_block,
  X_tilde,
  y_tilde,
  VhalfInv_perm,
  Lambda_block,
  A,
  K,
  p_expansions,
  constraint_value_vectors,
  family,
  return_G_getB,
  quadprog,
  qp_Amat,
  qp_bvec,
  qp_meq,
  qp_score_function,
  order_list,
  observation_weights,
  ...
)
}
\arguments{
\item{X_block}{Full \eqn{N \times P} unwhitened block-diagonal design.}

\item{X_tilde}{Whitened design \eqn{\mathbf{V}^{-1/2}\mathbf{X}}.}

\item{y_tilde}{Whitened response \eqn{\mathbf{V}^{-1/2}\mathbf{y}}.}

\item{VhalfInv_perm}{\eqn{\mathbf{V}^{-1/2}} permuted to partition
ordering.}

\item{Lambda_block}{Full \eqn{P \times P} block-diagonal penalty.}

\item{A}{Constraint matrix (\eqn{P \times R}).}

\item{K, p_expansions}{Integer dimensions.}

\item{constraint_value_vectors}{Constraint RHS list encoding
\eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{c}}.}

\item{family}{GLM family object.}

\item{return_G_getB}{Logical; return covariance components.}

\item{quadprog}{Logical; apply QP refinement.}

\item{qp_Amat, qp_bvec, qp_meq}{QP constraint specification.}

\item{qp_score_function}{Score function for QP step.}

\item{order_list, observation_weights}{Standard partition arguments.}

\item{...}{Passed to score function.}
}
\value{
If \code{return_G_getB = TRUE}: list with \code{B},
  \code{G_list}, and \code{qp_info}. Otherwise: list of \code{B} and
  \code{qp_info}.
}
\description{
Computes the constrained penalized GLS estimate for Gaussian response
with identity link when a correlation structure is present. Because
\eqn{\mathbf{V}^{-1/2}} couples all partitions, fitting must operate on
the full \eqn{P}-dimensional whitened system rather than partition-wise.
}
\details{
The unconstrained GLS estimate is:
\deqn{\hat{\boldsymbol{\beta}} = \mathbf{G}(\mathbf{X}^{*\top}
  \mathbf{y}^{*}), \quad
  \mathbf{G} = (\mathbf{X}^{*\top}\mathbf{X}^{*} +
  \boldsymbol{\Lambda})^{-1}}
where \eqn{\mathbf{X}^{*} = \mathbf{V}^{-1/2}\mathbf{X}} and
\eqn{\mathbf{y}^{*} = \mathbf{V}^{-1/2}\mathbf{y}}. The constrained
estimate is then obtained via the \eqn{\mathbf{G}^{1/2}\mathbf{r}^*}
trick in the full \eqn{P}-dimensional space.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/negbin_helpers.R
\name{negbin_theta}
\alias{negbin_theta}
\title{Estimate Negative Binomial Shape Parameter}
\usage{
negbin_theta(y, mu, weights = 1, init = NULL)
}
\arguments{
\item{y}{Response vector.}

\item{mu}{Mean vector.}

\item{weights}{Observation weights (default 1).}

\item{init}{Optional initial value for \eqn{\theta}. If NULL, uses a
moment-based estimate.}
}
\value{
Scalar MLE of \eqn{\theta}.
}
\description{
Computes the profile MLE of the shape parameter \eqn{\theta} given
current mean estimates \eqn{\mu}.
}
\details{
Maximizes the profile log-likelihood over \eqn{\theta} via Brent's
method on \eqn{[10^{-4}, 10^4]}.
}
\examples{
set.seed(1234)
mu <- rep(5, 200)
y <- rnbinom(200, size = 3, mu = 5)
negbin_theta(y, mu)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cox_helpers.R
\name{unconstrained_fit_cox}
\alias{unconstrained_fit_cox}
\title{Unconstrained Cox PH Estimation for lgspline}
\usage{
unconstrained_fit_cox(
  X,
  y,
  LambdaHalf,
  Lambda,
  keep_weighted_Lambda,
  family,
  tol = 1e-08,
  K,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  order_indices,
  weights,
  status
)
}
\arguments{
\item{X}{Design matrix (N_k x p) for partition k.}

\item{y}{Survival times for partition k.}

\item{LambdaHalf}{Square root of penalty matrix (unused here, retained
for interface compatibility).}

\item{Lambda}{Penalty matrix.}

\item{keep_weighted_Lambda}{Logical; if TRUE, return hot-start estimates
without refinement (not recommended for Cox).}

\item{family}{Cox family object.}

\item{tol}{Convergence tolerance.}

\item{K}{Number of knots.}

\item{parallel}{Logical for parallel processing.}

\item{cl}{Cluster object.}

\item{chunk_size, }{num_chunks, rem_chunks Parallelism parameters.}

\item{order_indices}{Observation indices mapping partition to full data.}

\item{weights}{Observation weights.}

\item{status}{Event indicator (1 = event, 0 = censored), full-data length.}
}
\value{
Numeric column vector of penalized partial-likelihood coefficient
  estimates.
}
\description{
Estimates penalized Cox PH coefficients for a single partition (or the
full model when K = 0) using damped Newton-Raphson on the penalized
partial log-likelihood with the Breslow approximation for tied event times.

This function is passed to lgspline's \code{unconstrained_fit_fxn}
argument. It receives the partition-level design matrix and response,
sorts internally by event time, and returns the coefficient vector.
}
\details{
The penalized partial log-likelihood is
\deqn{\ell_p(\boldsymbol{\beta}) = \ell(\boldsymbol{\beta})
  - \tfrac{1}{2}\boldsymbol{\beta}^{\top}
  \boldsymbol{\Lambda}\boldsymbol{\beta}}
Newton-Raphson updates use the Cox score and observed information
(computed via \code{score_cox} and \code{info_cox}), plus the penalty
term. The step is damped: the step size is halved until the penalized
log-likelihood improves.
}
\examples{
## Used internally by lgspline; see the full-model example below.

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_B.R
\name{get_B}
\alias{get_B}
\title{Compute Constrained GLM Coefficient Estimates via Lagrangian Multipliers}
\usage{
get_B(
  X,
  X_gram,
  Lambda,
  keep_weighted_Lambda,
  unique_penalty_per_partition,
  L_partition_list,
  A,
  Xy,
  y,
  K,
  p_expansions,
  R_constraints,
  Ghalf,
  GhalfInv,
  parallel_eigen,
  parallel_aga,
  parallel_matmult,
  parallel_unconstrained,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks,
  family,
  unconstrained_fit_fxn,
  iterate,
  qp_score_function,
  quadprog,
  qp_Amat,
  qp_bvec,
  qp_meq,
  prevB = NULL,
  prevUnconB = NULL,
  iter_count = 0,
  prev_diff = Inf,
  tol,
  constraint_value_vectors,
  order_list,
  glm_weight_function,
  schur_correction_function,
  need_dispersion_for_estimation,
  dispersion_function,
  observation_weights,
  homogenous_weights,
  return_G_getB,
  blockfit,
  just_linear_without_interactions,
  Vhalf,
  VhalfInv,
  ...
)
}
\arguments{
\item{X}{List of length \eqn{K+1}; partition-specific design matrices.}

\item{X_gram}{List of Gram matrices by partition.}

\item{Lambda}{Combined penalty matrix.}

\item{keep_weighted_Lambda}{Logical.}

\item{unique_penalty_per_partition}{Logical.}

\item{L_partition_list}{List of partition-specific penalty matrices.}

\item{A}{Constraint matrix \eqn{\mathbf{A}} (\eqn{P \times R}).}

\item{Xy}{List of cross-products by partition.}

\item{y}{List of response vectors by partition.}

\item{K}{Integer; number of interior knots.}

\item{p_expansions}{Integer; basis terms per partition.}

\item{R_constraints}{Integer; columns of \eqn{\mathbf{A}}.}

\item{Ghalf}{List of \eqn{\mathbf{G}^{1/2}_k} matrices.}

\item{GhalfInv}{List of \eqn{\mathbf{G}^{-1/2}_k} matrices.}

\item{parallel_eigen, parallel_aga, parallel_matmult, parallel_unconstrained}{Logical flags.}

\item{cl}{Cluster object.}

\item{chunk_size, num_chunks, rem_chunks}{Parallel distribution parameters.}

\item{family}{GLM family object.}

\item{unconstrained_fit_fxn}{Partition-wise unconstrained estimator.}

\item{iterate}{Logical; iterate for non-canonical links.}

\item{qp_score_function}{Score function for QP steps.}

\item{quadprog}{Logical; use \code{quadprog::solve.QP} for inequality
constraints.}

\item{qp_Amat, qp_bvec, qp_meq}{Inequality constraint specification
\eqn{\mathbf{C}^{\top}\boldsymbol{\beta} \succeq \mathbf{c}}.}

\item{prevB, prevUnconB, iter_count, prev_diff}{Retired; ignored.}

\item{tol}{Convergence tolerance.}

\item{constraint_value_vectors}{List encoding nonzero RHS
\eqn{\mathbf{A}^{\top}\boldsymbol{\beta} = \mathbf{c}}.}

\item{order_list}{Partition-to-data index mapping.}

\item{glm_weight_function}{Function computing working weights.}

\item{schur_correction_function}{Function computing Schur corrections.}

\item{need_dispersion_for_estimation}{Logical.}

\item{dispersion_function}{Dispersion estimator.}

\item{observation_weights}{List of observation weights by partition.}

\item{homogenous_weights}{Logical.}

\item{return_G_getB}{Logical; return \eqn{\mathbf{G}},
\eqn{\mathbf{G}^{1/2}}, \eqn{\mathbf{G}^{-1/2}} per partition.}

\item{blockfit, just_linear_without_interactions}{Retired; retained for
call-site compatibility.}

\item{Vhalf, VhalfInv}{Square root and inverse square root of the
working correlation matrix in the original observation ordering.
When both are non-\code{NULL}, dispatches to Path 1; \code{VhalfInv}
is permuted internally to partition ordering via \code{order_list}.}

\item{...}{Passed to fitting, weight, correction, and dispersion
functions.}
}
\value{
If \code{return_G_getB = FALSE}: a list with \code{B} (coefficient
column vectors by partition) and \code{qp_info}.

If \code{return_G_getB = TRUE}: a list with elements:
\describe{
  \item{B}{List of constrained coefficient column vectors
    \eqn{\tilde{\boldsymbol{\beta}}_k} by partition.}
  \item{G_list}{List with \code{G}, \code{Ghalf}, \code{GhalfInv},
    each a list of \eqn{K+1} matrices.}
  \item{qp_info}{QP or active-set metadata, including Lagrangian
    multipliers and active-constraint information when available, or
    \code{NULL}.}
}
}
\description{
Core estimation function for Lagrangian multiplier smoothing splines.
Computes penalized coefficient estimates subject to smoothness constraints
(continuity and derivative matching at knots) and optional user-supplied
linear equality or inequality constraints. Dispatches to one of three
computational paths depending on the model structure:

\bold{Path 1. GEE (correlation structure present):}
When both \code{Vhalf} and \code{VhalfInv} are provided, the full
\eqn{N \times P} whitened design
\eqn{\mathbf{X}^{*} = \mathbf{V}_{\mathrm{perm}}^{-1/2}\mathbf{X}} is used
after permuting observations to partition ordering via \code{order_list}.
For structured correlation, the solver tries the supplement's
decomposition
\eqn{\mathbf{G}^{-1} =
\mathbf{G}_{\mathrm{on}}^{-1} + \mathbf{G}_{\mathrm{off}}^{-1}}
with
\eqn{\mathbf{G}_{\mathrm{off}}^{-1} = \mathbf{E}\mathbf{J}\mathbf{E}^{\top}}.
This gives a Woodbury path with cost \eqn{O(Kp^3 + Pr^2)} when the
cross-partition rank \eqn{r} is small; otherwise the dense
\eqn{O(P^3)} path is used.
Two sub-paths:
\itemize{
  \item \bold{Path 1a.} Gaussian identity + GEE.
    See \code{\link{.get_B_gee_gaussian}} (dense) and
    \code{\link{.get_B_gee_woodbury}} (accelerated).
  \item \bold{Path 1b.} Non-Gaussian GEE.
    See \code{\link{.get_B_gee_glm}} (dense) and
    \code{\link{.get_B_gee_glm_woodbury}} (accelerated).
}

\bold{Path 2. Gaussian identity link, no correlation:}
Unconstrained estimate \eqn{\hat{\boldsymbol{\beta}}_k =
\mathbf{G}_k\mathbf{X}_k^{\top}\mathbf{y}_k} per partition, then a
single Lagrangian projection. See \code{\link{.get_B_gaussian_nocorr}}.

\bold{Path 3. Non-Gaussian GLM, no correlation:}
Partition-wise unconstrained estimates via Newton-Raphson, then
Lagrangian projection. See \code{\link{.get_B_glm_nocorr}}.

\bold{Inequality constraint handling:}
When inequality constraints \eqn{\mathbf{C}^{\top}\boldsymbol{\beta}
\succeq \mathbf{c}} are present, the sparsity pattern of
\eqn{\mathbf{C}} is inspected automatically. If every constraint
column has nonzeros in only a single partition block (block-separable),
a partition-wise active-set method is used at cost
\eqn{O(Kp^3)} per iteration. If any constraint spans multiple
partition blocks (e.g.\ cross-knot monotonicity), the dense
\code{quadprog::solve.QP} SQP fallback is invoked.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blockfit_solve.R
\name{.bf_assemble_qp_info}
\alias{.bf_assemble_qp_info}
\title{Assemble qp_info from a solve.QP Solution}
\usage{
.bf_assemble_qp_info(
  last_qp_sol,
  beta_block,
  qp_Amat_combined,
  qp_bvec_combined,
  qp_meq_combined,
  converged,
  final_deviance,
  info_matrix = NULL
)
}
\arguments{
\item{last_qp_sol}{Output of \code{quadprog::solve.QP}, or NULL.}

\item{beta_block}{Final coefficient vector.}

\item{qp_Amat_combined}{Combined constraint matrix.}

\item{qp_bvec_combined}{Combined constraint RHS.}

\item{qp_meq_combined}{Number of equalities.}

\item{converged}{Logical.}

\item{final_deviance}{Scalar.}

\item{info_matrix}{Information matrix (non-GEE path only; NULL
otherwise).}
}
\value{
A list suitable for the \code{qp_info} slot, or NULL if
  \code{last_qp_sol} is NULL.
}
\description{
Thin wrapper around \code{.solver_assemble_qp_info()}.
Keeping the local helper name avoids changing the public return shape
or downstream references inside \code{blockfit_solve()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{logLik.lgspline}
\alias{logLik.lgspline}
\title{Extract Log-Likelihood from a Fitted lgspline}
\usage{
\method{logLik}{lgspline}(object, include_prior = TRUE, new_weights = NULL, ...)
}
\arguments{
\item{object}{A fitted lgspline model object.}

\item{include_prior}{Logical; add the log-prior penalty term. Default TRUE.}

\item{new_weights}{Numeric scalar or N-vector; optional observation weights
overriding \code{object$weights}.}

\item{...}{Not used.}
}
\value{
A \code{"logLik"} object with attributes \code{df} (effective
  degrees of freedom) and \code{nobs} (number of observations).
}
\description{
Returns the log-likelihood as a \code{"logLik"} object for use with
\code{\link[stats]{AIC}}, \code{\link[stats]{BIC}}, and other model
comparison tools.
}
\details{
\strong{Gaussian identity, no correlation.}
\deqn{
  \ell = -\frac{N}{2}\log(2\pi\tilde{\sigma}^2) -
  \frac{1}{2\tilde{\sigma}^2}\sum_{i}(y_i - \hat{y}_i)^2
}

\strong{Gaussian identity, with correlation.}
GLS log-likelihood:
\deqn{
  \ell = -\frac{N}{2}\log(2\pi\tilde{\sigma}^2)
  + \log|\mathbf{V}^{-1/2}|
  - \frac{1}{2\tilde{\sigma}^2}
    \|\mathbf{V}^{-1/2}(\mathbf{y} - \hat{\mathbf{y}})\|^2
}
\eqn{\log|\mathbf{V}^{-1/2}|} is obtained from \code{VhalfInv_logdet}
when available, or computed directly from \code{VhalfInv}.

\strong{Prior contribution.}
When \code{include_prior = TRUE} (default), the log-prior
\deqn{
  -\frac{1}{2\tilde{\sigma}^2}
  \sum_{k}\boldsymbol{\beta}_k^\top\boldsymbol{\Lambda}_k
  \boldsymbol{\beta}_k
}
is added, giving the penalised MAP log-likelihood coherent with the
smoothing spline objective. Set \code{include_prior = FALSE} for the
unpenalised marginal likelihood, which is more appropriate when comparing
models with different penalty structures or numbers of knots.

\strong{Other GLM families.}
Uses \code{family$aic()} when available. For correlated models the
whitened residuals and fitted values are passed. When \code{family$aic()}
is unavailable, a deviance-based approximation is used (valid for
relative comparisons; a warning is emitted).

This function returns the marginal (full) GLS log-likelihood, not the
REML log-likelihood. This is consistent with \code{REML = FALSE} in
\code{lme} and \code{gls}, and is the conventional choice for AIC/BIC
comparisons of fixed-effects structure.

The \code{df} attribute is set to \eqn{N - \mathrm{trace}(\mathbf{XUGX}^\top)}.
}
\examples{

set.seed(1234)
t <- runif(1000, -10, 10)
y <- 2*sin(t) + -0.06*t^2 + rnorm(length(t))
model_fit <- lgspline(t, y)

logLik(model_fit)
logLik(model_fit, include_prior = FALSE)

AIC(model_fit)
BIC(model_fit)

## Compare models with different K using unpenalized likelihood
fit_k3 <- lgspline(t, y, K = 3)
fit_k7 <- lgspline(t, y, K = 7)
AIC(fit_k3, fit_k7)

}
\seealso{
\code{\link{lgspline}}, \code{\link{prior_loglik}},
\code{\link[stats]{logLik}}, \code{\link[stats]{AIC}},
\code{\link[stats]{BIC}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/solver_utils.R
\name{.solver_detect_qp_global}
\alias{.solver_detect_qp_global}
\title{Detect Whether Inequality Constraints Require a Dense Global QP}
\usage{
.solver_detect_qp_global(qp_Amat, p_expansions, K)
}
\arguments{
\item{qp_Amat}{Inequality constraint matrix
(\eqn{P \times n_{\mathrm{ineq}}}), or \code{NULL}.}

\item{p_expansions}{Integer; number of basis terms per partition.}

\item{K}{Integer; number of interior knots.}
}
\value{
Logical scalar.
}
\description{
Inspects the columns of \code{qp_Amat} to determine whether every
inequality constraint is confined to a single partition block
(block-separable) or whether any constraint couples coefficients
across partitions.
}
\details{
Returns \code{FALSE} (partition-wise active-set is valid) when all
columns have nonzeros in at most one block.  Returns \code{TRUE}
(dense SQP required) when any column spans multiple blocks.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{.rank_dists}
\alias{.rank_dists}
\title{Compute rank-based distance matrix for AR(1) structures}
\usage{
.rank_dists(spacetime, inds)
}
\description{
Converts the unique within-block distances returned by
\code{.compute_dist_block()} to integer lags \code{0, 1, 2, \ldots} in
increasing order.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{AGAmult_wrapper}
\alias{AGAmult_wrapper}
\title{Efficient Matrix Multiplication for \eqn{\textbf{A}^{T}\textbf{G}\textbf{A}}}
\usage{
AGAmult_wrapper(
  G,
  A,
  K,
  p_expansions,
  R_constraints,
  parallel,
  cl,
  chunk_size,
  num_chunks,
  rem_chunks
)
}
\arguments{
\item{G}{List of G matrices (\eqn{\textbf{G}})}

\item{A}{Constraint matrix (\eqn{\textbf{A}})}

\item{K}{Number of partitions minus 1 (\eqn{K})}

\item{p_expansions}{Number of columns per partition}

\item{R_constraints}{Number of constraint columns}

\item{parallel}{Use parallel processing}

\item{cl}{Cluster object}

\item{chunk_size}{Chunk size for parallel}

\item{num_chunks}{Number of chunks}

\item{rem_chunks}{Remaining chunks}
}
\value{
Matrix product \eqn{\textbf{A}^{T}\textbf{G}\textbf{A}}
}
\description{
Efficient Matrix Multiplication for \eqn{\textbf{A}^{T}\textbf{G}\textbf{A}}
}
\details{
Computes \eqn{\textbf{A}^{T}\textbf{G}\textbf{A}} efficiently in parallel chunks using \code{AGAmult_chunk()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{.compute_loocv}
\alias{.compute_loocv}
\title{Evaluate Exact Leave-One-Out Criterion at a Given Penalty Configuration}
\usage{
.compute_loocv(par, log_penalty_vec, env, ...)
}
\arguments{
\item{par}{Numeric vector; log-scale penalty parameters. First two elements
are log(wiggle_penalty) and log(flat_ridge_penalty). Remaining elements
(if any) are log-scale predictor/partition penalties.}

\item{log_penalty_vec}{Numeric vector; log-scale predictor/partition
penalties (passed separately for compatibility with the grid search).}

\item{env}{List; pre-computed objects and tuning configuration. Contains:
\describe{
  \item{X, y, X_gram, Xy}{Design matrices, response, Gram matrices,
    cross-products.}
  \item{A, R_constraints, K, p_expansions, N_obs}{Constraint matrix and
    dimensions.}
  \item{smoothing_spline_penalty}{Integrated second-derivative penalty
    matrix.}
  \item{custom_penalty_mat}{Optional custom penalty matrix.}
  \item{colnm_expansions}{Character vector of expansion column names.}
  \item{unique_penalty_per_predictor, unique_penalty_per_partition}{Logicals.}
  \item{family}{GLM family object.}
  \item{delta}{Pseudocount for link function stabilization.}
  \item{gcv_gamma}{Modified-GCV inflation factor applied to the
    effective-degrees-of-freedom term in the denominator.}
  \item{meta_penalty}{Meta-penalty coefficient.}
  \item{order_list}{List of observation indices per partition.}
  \item{observation_weights, homogenous_weights}{Observation weighting.}
  \item{parallel flags, cl, chunk_size, num_chunks, rem_chunks}{Parallel
    config.}
  \item{unconstrained_fit_fxn, keep_weighted_Lambda, iterate}{Fitting
    config.}
  \item{qp_score_function, quadprog, qp_Amat, qp_bvec, qp_meq}{QP config.}
  \item{tol, sd_y}{Convergence tolerance and response scale.}
  \item{constraint_value_vectors}{Constraint values.}
  \item{glm_weight_function, schur_correction_function}{GLM functions.}
  \item{need_dispersion_for_estimation, dispersion_function}{Dispersion.}
  \item{blockfit, just_linear_without_interactions}{Blockfit config.}
  \item{use_blockfit}{Logical; pre-computed dispatch flag for blockfit.}
  \item{flat_cols}{Integer vector; pre-computed flat column indices.}
  \item{Vhalf, VhalfInv}{Correlation structure matrices.}
  \item{verbose, include_warnings}{Output control.}
}}

\item{...}{Additional arguments passed to fitting functions.}
}
\value{
List containing the fitted objects plus \code{criterion_value},
  \code{LOO_u}, \code{h_diag}, and \code{C_list}.
}
\description{
Computes the exact leave-one-out tuning criterion for the same transformed
linearized problem used elsewhere in the tuning machinery.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weibull_helpers.R
\name{weibull_glm_weight_function}
\alias{weibull_glm_weight_function}
\title{Weibull GLM Weight Function for Constructing Information Matrix}
\usage{
weibull_glm_weight_function(
  mu,
  y,
  order_indices,
  family,
  dispersion,
  observation_weights,
  status
)
}
\arguments{
\item{mu}{Predicted survival times}

\item{y}{Observed response/survival times}

\item{order_indices}{Order of observations when partitioned to match
"status" to "response"}

\item{family}{Weibull AFT family; unused here and retained for interface
compatibility.}

\item{dispersion}{Estimated dispersion parameter (sigma^2 = scale^2). The
lgspline framework stores and passes dispersion (sigma^2); the Weibull
scale (sigma) matching \code{survreg$scale} is \code{sqrt(dispersion)}.}

\item{observation_weights}{Weights of observations submitted to function}

\item{status}{Censoring indicator (1 = event, 0 = censored)
Indicates whether an event of interest occurred (1) or the observation was
right-censored (0). In survival analysis, right-censoring occurs when the
full survival time is unknown, typically because the study ended or the
subject was lost to follow-up before the event of interest occurred.}
}
\value{
Numeric vector of working weights for the information matrix, including
observation weights when finite and a fallback of 1s when the natural
Weibull weights are non-finite.
}
\description{
Computes the working weights used in the Weibull AFT information matrix,
including the observation-weight contribution returned on the vector scale.
}
\details{
This function generates weights used in constructing the information matrix
after unconstrained estimates have been found. Specifically, it is used in
the construction of the \eqn{\textbf{U}} and \eqn{\textbf{G}} matrices
following initial unconstrained parameter estimation.

These weights are analogous to the variance terms in generalized linear
models (GLMs). Like logistic regression uses \eqn{\mu(1-\mu)}, Poisson
regression uses \eqn{e^{\mu}}, and Linear regression uses constant weights,
Weibull AFT models use \eqn{\exp((\log y - \log \mu)/\sigma)} where
\eqn{\sigma = \sqrt{\text{dispersion}}} is the scale parameter.
}
\examples{

## Demonstration of glm weight function in constrained model estimation
set.seed(1234)
n <- 1000
t1 <- rnorm(n)
t2 <- rbinom(n, 1, 0.5)

## Generate survival times
lambda <- exp(0.5 * t1 + 0.3 * t2)
yraw <- rexp(n, rate = 1/lambda)

## Introduce right-censoring
status <- rbinom(n, 1, 0.75)
y <- ifelse(status, yraw, runif(length(yraw), 0, yraw))

## Fit model demonstrating use of custom glm weight function
model_fit <- lgspline(y ~ spl(t1) + t2,
                      data.frame(y = y, t1 = t1, t2 = t2),
                      unconstrained_fit_fxn = unconstrained_fit_weibull,
                      family = weibull_family(),
                      need_dispersion_for_estimation = TRUE,
                      dispersion_function = weibull_dispersion_function,
                      glm_weight_function = weibull_glm_weight_function,
                      schur_correction_function = weibull_schur_correction,
                      status = status,
                      opt = FALSE,
                      K = 1)

print(summary(model_fit))


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{summary.lgspline}
\alias{summary.lgspline}
\title{Summary Method for lgspline Objects}
\usage{
\method{summary}{lgspline}(object, ...)
}
\arguments{
\item{object}{An lgspline model object.}

\item{...}{Not used.}
}
\value{
An object of class \code{summary.lgspline}, a list containing:
\describe{
  \item{model_family}{The \code{\link[stats]{family}} object.}
  \item{observations}{Number of observations N.}
  \item{predictors}{Number of predictor variables q.}
  \item{knots}{Number of knots K.}
  \item{basis_functions}{Basis functions per partition p.}
  \item{estimate_dispersion}{"Yes" or "No".}
  \item{cv}{Critical value used for confidence intervals.}
  \item{coefficients}{Coefficient matrix from \code{\link{wald_univariate}},
        or a single-column estimate matrix if \code{return_varcovmat = FALSE}.}
  \item{sigmasq_tilde}{Estimated dispersion \eqn{\tilde{\sigma}^2}.}
  \item{trace_XUGX}{Trace of the hat matrix \eqn{\mathrm{trace}(\mathbf{XUGX}^\top)}.}
  \item{N}{Number of observations.}
}
}
\description{
Summary Method for lgspline Objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lgspline.R
\name{lgspline}
\alias{lgspline}
\title{Fit Lagrangian Multiplier Smoothing Splines}
\usage{
lgspline(
  predictors = NULL,
  y = NULL,
  formula = NULL,
  response = NULL,
  standardize_response = TRUE,
  standardize_predictors_for_knots = TRUE,
  standardize_expansions_for_fitting = TRUE,
  family = gaussian(),
  glm_weight_function = function(mu, y, order_indices, family, dispersion,
    observation_weights, ...) {
     if (any(!is.null(observation_weights))) {
        
    family$variance(mu) * observation_weights
     }
     else {
        
    family$variance(mu)
     }
 },
  schur_correction_function = function(X, y, B, dispersion, order_list, K, family,
    observation_weights, ...) {
     lapply(1:(K + 1), function(k) 0)
 },
  need_dispersion_for_estimation = FALSE,
  dispersion_function = function(mu, y, order_indices, family, observation_weights,
    VhalfInv, ...) {
     if (!is.null(VhalfInv)) {
         VhalfInv <-
    VhalfInv[order_indices, order_indices]
         c(mean((tcrossprod(VhalfInv, t(y -
    mu)))^2/family$variance(mu)))
     }
     else {
         c(mean((y -
    mu)^2/family$variance(mu)))
     }
 },
  K = NULL,
  custom_knots = NULL,
  cluster_on_indicators = FALSE,
  make_partition_list = NULL,
  previously_tuned_penalties = NULL,
  smoothing_spline_penalty = NULL,
  opt = TRUE,
  use_custom_bfgs = TRUE,
  delta = NULL,
  tol = 10 * sqrt(.Machine$double.eps),
  tuning_criterion = "loo",
  gcv_gamma = 1.4,
  initial_wiggle = c(2e-12, 2e-07, 2e-04, 0.2),
  initial_flat = c(0.5, 5),
  wiggle_penalty = 2e-07,
  flat_ridge_penalty = 0.5,
  unique_penalty_per_partition = TRUE,
  unique_penalty_per_predictor = TRUE,
  meta_penalty = 1e-08,
  predictor_penalties = NULL,
  partition_penalties = NULL,
  include_quadratic_terms = TRUE,
  include_cubic_terms = TRUE,
  include_quartic_terms = NULL,
  include_2way_interactions = TRUE,
  include_3way_interactions = TRUE,
  include_quadratic_interactions = FALSE,
  offset = c(),
  just_linear_with_interactions = NULL,
  just_linear_without_interactions = NULL,
  exclude_interactions_for = NULL,
  exclude_these_expansions = NULL,
  custom_basis_fxn = NULL,
  include_constrain_fitted = TRUE,
  include_constrain_first_deriv = TRUE,
  include_constrain_second_deriv = TRUE,
  include_constrain_interactions = TRUE,
  cl = NULL,
  chunk_size = NULL,
  parallel_eigen = TRUE,
  parallel_trace = FALSE,
  parallel_aga = FALSE,
  parallel_matmult = FALSE,
  parallel_unconstrained = TRUE,
  parallel_find_neighbors = FALSE,
  parallel_penalty = FALSE,
  parallel_make_constraint = FALSE,
  unconstrained_fit_fxn = unconstrained_fit_default,
  keep_weighted_Lambda = FALSE,
  iterate_tune = TRUE,
  iterate_final_fit = TRUE,
  blockfit = TRUE,
  qp_score_function = function(X, y, mu, order_list, dispersion, VhalfInv,
    observation_weights, ...) {
     if (!is.null(observation_weights)) {
        
    crossprod(X, cbind((y - mu) * observation_weights))
     }
     else {
        
    crossprod(X, cbind(y - mu))
     }
 },
  qp_observations = NULL,
  qp_Amat = NULL,
  qp_bvec = NULL,
  qp_meq = 0,
  qp_positive_derivative = FALSE,
  qp_negative_derivative = FALSE,
  qp_positive_2ndderivative = FALSE,
  qp_negative_2ndderivative = FALSE,
  qp_monotonic_increase = FALSE,
  qp_monotonic_decrease = FALSE,
  qp_range_upper = NULL,
  qp_range_lower = NULL,
  qp_Amat_fxn = NULL,
  qp_bvec_fxn = NULL,
  qp_meq_fxn = NULL,
  constraint_values = cbind(),
  constraint_vectors = cbind(),
  return_G = TRUE,
  return_Ghalf = TRUE,
  return_U = TRUE,
  estimate_dispersion = TRUE,
  unbias_dispersion = NULL,
  return_varcovmat = TRUE,
  exact_varcovmat = FALSE,
  return_lagrange_multipliers = FALSE,
  custom_penalty_mat = NULL,
  cluster_args = c(custom_centers = NA, nstart = 10),
  dummy_dividor = 1.2345672152894e-22,
  dummy_adder = 2.234567210529e-18,
  verbose = FALSE,
  verbose_tune = FALSE,
  dummy_fit = FALSE,
  auto_encode_factors = TRUE,
  observation_weights = NULL,
  do_not_cluster_on_these = c(),
  neighbor_tolerance = 1 + 1e-08,
  null_constraint = NULL,
  critical_value = qnorm(1 - 0.05/2),
  data = NULL,
  weights = NULL,
  no_intercept = FALSE,
  correlation_id = NULL,
  spacetime = NULL,
  correlation_structure = NULL,
  VhalfInv = NULL,
  Vhalf = NULL,
  VhalfInv_fxn = NULL,
  Vhalf_fxn = NULL,
  VhalfInv_par_init = c(),
  REML_grad = NULL,
  custom_VhalfInv_loss = NULL,
  VhalfInv_logdet = NULL,
  include_warnings = TRUE,
  penalty_args = NULL,
  tuning_args = NULL,
  expansion_args = NULL,
  constraint_args = NULL,
  qp_args = NULL,
  parallel_args = NULL,
  covariance_args = NULL,
  return_args = NULL,
  glm_args = NULL,
  ...
)
}
\arguments{
\item{predictors}{Default: NULL. Numeric matrix or data frame of predictor
variables, or a formula when using the formula interface.}

\item{y}{Default: NULL. Numeric response variable vector.}

\item{formula}{Default: NULL. Optional statistical formula for model specification,
supporting \code{spl()} (and the alias \code{s()}) for spline terms.}

\item{response}{Default: NULL. Alternative name for response variable.}

\item{standardize_response}{Default: TRUE. Logical indicator controlling whether
the response variable should be centered and scaled before model fitting. Only
offered for identity link functions.}

\item{standardize_predictors_for_knots}{Default: TRUE. Logical flag controlling
whether predictors are internally standardized for partitioning / knot
placement. The exact transformation is handled inside
\code{\link{make_partitions}} and depends on the effective clustering
dimension.}

\item{standardize_expansions_for_fitting}{Default: TRUE. Logical switch to
standardize polynomial basis expansions during model fitting. Design matrices,
variance-covariance matrices, and coefficients are backtransformed after fitting.
\eqn{\mathbf{U}} and \eqn{\mathbf{G}} remain on the transformed scale;
\code{B_raw} corresponds to coefficients on the expansion-standardized scale.}

\item{family}{Default: \code{gaussian()}. GLM family specifying the error distribution
and link function. Minimally requires: family name, link name, linkfun, linkinv,
variance.}

\item{glm_weight_function}{Default: function returning \code{family$variance(mu)},
optionally multiplied by \code{observation_weights}. Codes the mean-variance
relationship (the diagonal \eqn{\mathbf{W}} matrix) used for updating
\eqn{\mathbf{G}} after obtaining constrained estimates.}

\item{schur_correction_function}{Default: function returning list of zeros.
Computes Schur complements \eqn{\mathbf{S}} added to \eqn{\mathbf{G}}:
\eqn{\mathbf{G}^{*} = (\mathbf{G}^{-1} + \mathbf{S})^{-1}}.}

\item{need_dispersion_for_estimation}{Default: FALSE. Logical indicator specifying
whether a dispersion parameter is required for coefficient estimation (e.g.
Weibull AFT).}

\item{dispersion_function}{Default: function returning mean squared residuals.
Custom function for estimating the exponential dispersion parameter.}

\item{K}{Default: NULL. Integer specifying the number of knot locations. Intuitively,
total partitions minus 1.}

\item{custom_knots}{Default: NULL. Optional matrix providing user-specified knot
locations in 1-D.}

\item{cluster_on_indicators}{Default: FALSE. Logical flag for whether indicator
variables should be used for clustering knot locations.}

\item{make_partition_list}{Default: NULL. Optional list allowing direct specification
of custom partition assignments. The \code{make_partition_list} returned by one
model can be supplied here to reuse knot locations.}

\item{previously_tuned_penalties}{Default: NULL. Optional list of pre-computed penalty
components from a previous model fit.}

\item{smoothing_spline_penalty}{Default: NULL. Optional custom smoothing spline penalty
matrix.}

\item{opt}{Default: TRUE. Logical switch controlling automatic penalty optimization.}

\item{use_custom_bfgs}{Default: TRUE. Selects between a native damped-BFGS
implementation with closed-form gradients or base R's BFGS with finite-difference
gradients. The native path is usually faster, while the finite-difference
fallback can be preferable when tuning under exact LOO and the leverage
derivative is numerically noisy.}

\item{delta}{Default: NULL. Numeric pseudocount for stabilizing optimization in
non-identity link function scenarios.}

\item{tol}{Default: \code{10*sqrt(.Machine$double.eps)}. Numeric convergence tolerance.}

\item{tuning_criterion}{Default: \code{"loo"}. Character scalar selecting the
tuning criterion. Use \code{"loo"} for exact leave-one-out on the transformed
tuning problem, or \code{"gcv"} for the generalized cross-validation
criterion. The LOO path computes the needed hat-matrix diagonal exactly
from blockwise constrained-\eqn{\mathbf{G}} quantities, without explicitly
forming the full projection matrix or full hat matrix. In empirical
diagnostics, the observation-wise derivative of the LOO leverage term can
be numerically delicate even when the overall tuning criterion and fitted
penalties remain well behaved; users who prefer a more conservative
optimization path can set \code{use_custom_bfgs = FALSE}. Shared wiggle
and flat tuning directions are differentiated directly, while optional
predictor- and partition-specific penalties continue to use a lower-cost
ratio approximation. For very large samples, generalized cross-validation
is often the more practical choice; as a rough guideline, \code{"gcv"}
is recommended once the sample size is above about 250,000.}

\item{gcv_gamma}{Default: 1.4. Numeric scalar, at least 1, used only when
\code{tuning_criterion = "gcv"}. Multiplies the effective degrees of freedom
in the GCV denominator during automatic penalty tuning. It is accepted but
ignored when \code{tuning_criterion = "loo"}.}

\item{initial_wiggle}{Default: \code{c(2e-12, 2e-7, 2e-4, 0.2)}. Numeric vector of
initial grid points for wiggle penalty optimization, on the raw (non-negative) scale.}

\item{initial_flat}{Default: \code{c(0.5, 5)}. Numeric vector of initial grid points
for ridge penalty optimization, on the raw scale (ratio of ridge to wiggle).}

\item{wiggle_penalty}{Default: 2e-7. Numeric penalty on the integrated squared second
derivative, governing function smoothness.}

\item{flat_ridge_penalty}{Default: 0.5. Numeric flat ridge penalty for intercepts and
linear terms only. Multiplied by \code{wiggle_penalty} to obtain total ridge penalty.}

\item{unique_penalty_per_partition}{Default: TRUE. Logical flag allowing penalty
magnitude to differ across partitions.}

\item{unique_penalty_per_predictor}{Default: TRUE. Logical flag allowing penalty
magnitude to differ between predictors.}

\item{meta_penalty}{Default: 1e-8. Numeric regularization coefficient for
predictor- and partition-specific penalties during tuning. On the raw scale,
the implemented meta-penalty shrinks these penalty multipliers toward 1; the
wiggle penalty receives only a tiny stabilizing penalty by default.}

\item{predictor_penalties}{Default: NULL. Optional vector of custom penalties per
predictor, on the raw (positive) scale.}

\item{partition_penalties}{Default: NULL. Optional vector of custom penalties per
partition, on the raw (positive) scale.}

\item{include_quadratic_terms}{Default: TRUE. Logical switch to include squared
predictor terms.}

\item{include_cubic_terms}{Default: TRUE. Logical switch to include cubic predictor
terms.}

\item{include_quartic_terms}{Default: NULL. Includes quartic terms; when NULL, set to
FALSE for single predictor and TRUE otherwise. Highly recommended for multi-predictor
models to avoid over-specified constraints.}

\item{include_2way_interactions}{Default: TRUE. Logical switch for linear two-way
interactions.}

\item{include_3way_interactions}{Default: TRUE. Logical switch for three-way
interactions.}

\item{include_quadratic_interactions}{Default: FALSE. Logical switch for
linear-quadratic interaction terms.}

\item{offset}{Default: Empty vector. Column indices/names to include as offsets.
Coefficients for offset terms are automatically constrained to 1.}

\item{just_linear_with_interactions}{Default: NULL. Integer or character vector
specifying predictors to retain as linear terms while still allowing
interactions.}

\item{just_linear_without_interactions}{Default: NULL. Integer or character
vector specifying predictors to retain only as linear terms without
interactions. Eligible for blockfitting.}

\item{exclude_interactions_for}{Default: NULL. Integer or character vector of
predictors to exclude from all interaction terms.}

\item{exclude_these_expansions}{Default: NULL. Character vector of basis expansions to
exclude. Named columns of data, or in the form \code{"_1_"}, \code{"_2_"},
\code{"_1_x_2_"}, \code{"_2_^2"} etc.}

\item{custom_basis_fxn}{Default: NULL. Optional user-defined function for custom basis
expansions. See \code{\link{get_polynomial_expansions}}.}

\item{include_constrain_fitted}{Default: TRUE. Logical switch to constrain fitted
values at knot points.}

\item{include_constrain_first_deriv}{Default: TRUE. Logical switch to constrain first
derivatives at knot points.}

\item{include_constrain_second_deriv}{Default: TRUE. Logical switch to constrain second
derivatives at knot points.}

\item{include_constrain_interactions}{Default: TRUE. Logical switch to constrain
interaction terms at knot points.}

\item{cl}{Default: NULL. Parallel processing cluster object
(use \code{parallel::makeCluster()}).}

\item{chunk_size}{Default: NULL. Integer specifying custom chunk size for parallel
processing.}

\item{parallel_eigen}{Default: TRUE. Logical flag for parallel eigenvalue decomposition.}

\item{parallel_trace}{Default: FALSE. Logical flag for parallel trace computation.}

\item{parallel_aga}{Default: FALSE. Logical flag for parallel \eqn{\mathbf{G}} and
\eqn{\mathbf{A}} matrix operations.}

\item{parallel_matmult}{Default: FALSE. Logical flag for parallel block-diagonal matrix
multiplication.}

\item{parallel_unconstrained}{Default: TRUE. Logical flag for parallel unconstrained
MLE for non-identity-link-Gaussian models.}

\item{parallel_find_neighbors}{Default: FALSE. Logical flag for parallel neighbor
identification.}

\item{parallel_penalty}{Default: FALSE. Logical flag for parallel penalty matrix
construction.}

\item{parallel_make_constraint}{Default: FALSE. Logical flag for parallel constraint
matrix generation.}

\item{unconstrained_fit_fxn}{Default: \code{\link{unconstrained_fit_default}}.
Custom function for fitting unconstrained models per partition.}

\item{keep_weighted_Lambda}{Default: FALSE. Logical flag to retain GLM weights in
penalty constraints using Tikhonov parameterization. Advised for non-canonical GLMs.}

\item{iterate_tune}{Default: TRUE. Logical switch for iterative optimization during
penalty tuning.}

\item{iterate_final_fit}{Default: TRUE. Logical switch for iterative optimization in
final model fitting.}

\item{blockfit}{Default: TRUE. Logical switch for backfitting with mixed spline and
non-interactive linear terms. Requires flat columns, \code{K > 0}, and no active
correlation structure. Falls back to \code{get_B} on failure.}

\item{qp_score_function}{Default: \eqn{\mathbf{X}^{\top}(\mathbf{y} - \boldsymbol{\mu})}.
Score function for quadratic programming, blockfit, and GEE formulations.
Accepts arguments \code{"X, y, mu, order_list, dispersion, VhalfInv,
observation_weights, ..."}.}

\item{qp_observations}{Default: NULL. Numeric vector of observation indices
at which built-in QP constraints are evaluated. Useful for reducing the
size of the constrained system.}

\item{qp_Amat}{Default: NULL. Optional pre-built QP constraint matrix.
In the current pipeline its presence marks QP handling as active, but the
built-in constructor does not merge it into the assembled constraint set;
use \code{qp_Amat_fxn} for custom assembled constraints.}

\item{qp_bvec}{Default: NULL. Optional pre-built QP right-hand side paired
with \code{qp_Amat}. Like \code{qp_Amat}, it is currently treated as an
advanced placeholder rather than merged into the built-in constructor.}

\item{qp_meq}{Default: 0. Optional number of equality constraints paired
with \code{qp_Amat}. Like \code{qp_Amat}, it is currently treated as an
advanced placeholder rather than merged into the built-in constructor.}

\item{qp_positive_derivative}{Default: FALSE. Constrain function to have positive first derivatives. Accepts: \code{FALSE} (no constraint), \code{TRUE} (all predictors), or a character/integer vector naming specific predictor variables to constrain. For example, \code{qp_positive_derivative = "Dose"} constrains only the Dose variable, while \code{qp_positive_derivative = c(1, 3)} constrains columns 1 and 3 of the predictor matrix.}

\item{qp_negative_derivative}{Default: FALSE. Constrain function to have negative first derivatives. Same input types as \code{qp_positive_derivative}. Can be used simultaneously with \code{qp_positive_derivative} on different variables.}

\item{qp_positive_2ndderivative}{Default: FALSE. Constrain function to have positive (convex) second derivatives. Same input types as \code{qp_positive_derivative}.}

\item{qp_negative_2ndderivative}{Default: FALSE. Constrain function to have negative (concave) second derivatives. Same input types as \code{qp_positive_derivative}.}

\item{qp_monotonic_increase}{Default: FALSE. Logical only. Constrain fitted values to be monotonically increasing in observation order.}

\item{qp_monotonic_decrease}{Default: FALSE. Logical only. Constrain fitted values to be monotonically decreasing in observation order.}

\item{qp_range_upper}{Default: NULL. Numeric upper bound for constrained fitted values.}

\item{qp_range_lower}{Default: NULL. Numeric lower bound for constrained fitted values.}

\item{qp_Amat_fxn}{Default: NULL. Custom function generating Amat.}

\item{qp_bvec_fxn}{Default: NULL. Custom function generating bvec.}

\item{qp_meq_fxn}{Default: NULL. Custom function generating meq.}

\item{constraint_values}{Default: \code{cbind()}. Optional matrix encoding
nonzero equality targets paired with \code{constraint_vectors}. When left
empty, added equality constraints are treated as homogeneous.}

\item{constraint_vectors}{Default: \code{cbind()}. Optional matrix of
user-supplied equality-constraint vectors, appended to the internally
generated smoothness constraints.}

\item{return_G}{Default: TRUE. Logical switch to return the unscaled unconstrained
variance-covariance matrix \eqn{\mathbf{G}}.}

\item{return_Ghalf}{Default: TRUE. Logical switch to return
\eqn{\mathbf{G}^{1/2}}.}

\item{return_U}{Default: TRUE. Logical switch to return the constraint projection
matrix \eqn{\mathbf{U}}.}

\item{estimate_dispersion}{Default: TRUE. Logical flag to estimate dispersion after
fitting.}

\item{unbias_dispersion}{Default: NULL. Logical switch to multiply dispersion by
\eqn{N/(N - \mathrm{trace}(\mathbf{H}))}. When NULL, set to TRUE for Gaussian
identity link and FALSE otherwise.}

\item{return_varcovmat}{Default: TRUE. Logical switch to return the variance-covariance
matrix of estimated coefficients. Needed for Wald inference.}

\item{exact_varcovmat}{Default: FALSE. Logical switch to replace the default
asymptotic (Bayesian posterior) variance-covariance matrix with the exact
frequentist variance-covariance matrix of the constrained estimator. The
asymptotic version uses the Hessian of
the penalized log-likelihood:
\eqn{\tilde{\sigma}^{2}\mathbf{U}\mathbf{G}\mathbf{U}^{\top}}.
The exact version additionally corrects for the penalty's contribution as a shrinkage
prior, giving:
\deqn{\tilde{\sigma}^{2}\mathbf{U}\mathbf{G}\mathbf{U}^{\top}
      - \tilde{\sigma}^{2}\mathbf{U}\mathbf{G}\boldsymbol{\Lambda}\mathbf{G}\mathbf{U}^{\top}}
When a correlation structure is present (\code{VhalfInv} non-NULL),
\eqn{\mathbf{G}_{\mathrm{correct}}} replaces the block-diagonal \eqn{\mathbf{G}}.
For Gaussian identity link (with or without correlation structure), the result is
the exact variance-covariance matrix of the constrained estimate.
The returned object still stores the result in \code{varcovmat}. Requires
\code{return_varcovmat = TRUE}.}

\item{return_lagrange_multipliers}{Default: FALSE. Logical switch to return the
Lagrangian multiplier vector.}

\item{custom_penalty_mat}{Default: NULL. Optional \eqn{p \times p} custom penalty
matrix for individual partitions, replacing the default ridge on linear/intercept
terms. Run with \code{dummy_fit = TRUE} first to inspect expansion structure.}

\item{cluster_args}{Default: \code{c(custom_centers = NA, nstart = 10)}. Named
vector of arguments controlling clustering. If the first argument is not
\code{NA}, it is treated as custom cluster centers (typically an
\eqn{(K+1) \times q} matrix). Otherwise, default k-means is used.}

\item{dummy_dividor}{Default: 0.00000000000000000000012345672152894. Small numeric
constant to prevent division by zero.}

\item{dummy_adder}{Default: 0.000000000000000002234567210529. Small numeric constant
to prevent division by zero.}

\item{verbose}{Default: FALSE. Logical flag to print general progress messages.}

\item{verbose_tune}{Default: FALSE. Logical flag to print detailed progress during
penalty tuning.}

\item{dummy_fit}{Default: FALSE. Runs the full pipeline but sets coefficients to zero,
allowing inspection of design matrix structure, penalty matrices, and partitioning.
Replaces the deprecated \code{expansions_only} argument.}

\item{auto_encode_factors}{Default: TRUE. Logical switch to automatically one-hot encode
factor or character variables when using the formula interface.}

\item{observation_weights}{Default: NULL. Numeric vector of observation-specific
weights for generalized least squares estimation.}

\item{do_not_cluster_on_these}{Default: \code{c()}. Predictor columns to exclude from
clustering. Accepts numeric column indices or character column names.}

\item{neighbor_tolerance}{Default: \code{1 + 1e-8}. Numeric tolerance for determining
neighboring partitions using k-means clustering. Intended for internal use.}

\item{null_constraint}{Default: NULL. Alternative parameterization for a
nonzero equality target when \code{constraint_vectors} is supplied and
\code{constraint_values} is left empty.}

\item{critical_value}{Default: \code{qnorm(1-0.05/2)}. Numeric critical value for
Wald confidence intervals.}

\item{data}{Default: NULL. Optional data frame for formula-based model specification.}

\item{weights}{Default: NULL. Alias for \code{observation_weights}.}

\item{no_intercept}{Default: FALSE. Logical flag to constrain intercept to 0.
Formulas with \code{"0+"} set this to TRUE automatically.}

\item{correlation_id, spacetime}{Default: NULL. N-length vector and N-row matrix of
cluster ids and longitudinal/spatial variables, respectively.}

\item{correlation_structure}{Default: NULL. Native implementations: \code{"exchangeable"},
\code{"spatial-exponential"}, \code{"squared-exponential"}, \code{"ar(1)"},
\code{"spherical"}, \code{"gaussian-cosine"}, \code{"gamma-cosine"},
\code{"matern"}, and aliases. Estimated via REML.}

\item{VhalfInv}{Default: NULL. Fixed custom \eqn{N \times N} square-root-inverse
covariance matrix \eqn{\mathbf{V}^{-1/2}}. Triggers GLS with known covariance.
Post-fit inference recomputed from whitened Gram matrices.}

\item{Vhalf}{Default: NULL. Fixed custom \eqn{N \times N} square-root covariance
\eqn{\mathbf{V}^{1/2}}. Computed as inverse of \code{VhalfInv} if not supplied.}

\item{VhalfInv_fxn}{Default: NULL. Parametric function for \eqn{\mathbf{V}^{-1/2}};
takes single numeric vector \code{"par"}, returns \eqn{N \times N} matrix.
Optimized via BFGS when \code{VhalfInv_par_init} is provided.}

\item{Vhalf_fxn}{Default: NULL. Optional function for efficient computation of
\eqn{\mathbf{V}^{1/2}} from the same parameter vector used by
\code{VhalfInv_fxn}. When omitted, \code{Vhalf} is obtained by explicit
matrix inversion of \code{VhalfInv}.}

\item{VhalfInv_par_init}{Default: \code{c()}. Initial parameter values for
\code{VhalfInv_fxn} optimization, on unbounded transformed scale.}

\item{REML_grad}{Default: NULL. Function for the gradient of the negative REML (or
custom loss) with respect to the parameters of \code{VhalfInv_fxn}. Takes
\code{"par"}, \code{"model_fit"}, and \code{"..."}.}

\item{custom_VhalfInv_loss}{Default: NULL. Alternative to negative REML for the
correlation parameter objective function. Takes \code{"par"}, \code{"model_fit"},
and \code{"..."}.}

\item{VhalfInv_logdet}{Default: NULL. Function for efficient \eqn{\log|\mathbf{V}^{-1/2}|}
computation. Takes same \code{"par"} as \code{VhalfInv_fxn}.}

\item{include_warnings}{Default: TRUE. Logical switch to control display of warnings.}

\item{penalty_args}{Default: NULL. Optional named list grouping penalty-related
arguments. See section "Grouped Argument Lists".}

\item{tuning_args}{Default: NULL. Optional named list grouping tuning-related
arguments.}

\item{expansion_args}{Default: NULL. Optional named list grouping basis expansion
arguments.}

\item{constraint_args}{Default: NULL. Optional named list grouping constraint arguments.}

\item{qp_args}{Default: NULL. Optional named list grouping quadratic programming
arguments.}

\item{parallel_args}{Default: NULL. Optional named list grouping parallel processing
arguments.}

\item{covariance_args}{Default: NULL. Optional named list grouping correlation
structure arguments.}

\item{return_args}{Default: NULL. Optional named list grouping return-control
arguments.}

\item{glm_args}{Default: NULL. Optional named list grouping GLM customization
arguments.}

\item{...}{Additional arguments passed to the unconstrained model fitting function.}
}
\value{
A list of class \code{"lgspline"} containing model components:
\describe{
  \item{y}{Original response vector.}
  \item{ytilde}{Fitted/predicted values on the scale of the response.}
  \item{X}{List of design matrices \eqn{\mathbf{X}_{k}} for each partition k, containing basis expansions including intercept, linear, quadratic, cubic, and interaction terms as specified. Returned on the unstandardized scale.}
  \item{A}{Constraint matrix \eqn{\mathbf{A}} encoding smoothness constraints at knot points and any user-specified linear constraints. Only a linearly independent subset of columns is retained (via pivoted QR decomposition).}
  \item{B}{List of fitted coefficients \eqn{\boldsymbol{\beta}_{k}} for each partition k on the original, unstandardized scale of the predictors and response.}
  \item{B_raw}{List of fitted coefficients for each partition on the predictor-and-response standardized scale.}
  \item{K}{Number of interior knots with one predictor (number of partitions minus 1 with > 1 predictor).}
  \item{p}{Number of basis expansions of predictors per partition.}
  \item{q}{Number of predictor variables.}
  \item{P}{Total number of coefficients (\eqn{p \times (K+1)}).}
  \item{N}{Number of observations.}
  \item{penalties}{List containing optimized penalty matrices and components:
    \itemize{
      \item Lambda: Combined penalty matrix (\eqn{\boldsymbol{\Lambda}}), includes \eqn{\mathbf{L}_{\mathrm{predictor\_list}}} contributions but not \eqn{\mathbf{L}_{\mathrm{partition\_list}}}.
      \item L1: Smoothing spline penalty matrix (\eqn{\mathbf{L}_{1}}).
      \item L2: Ridge penalty matrix (\eqn{\mathbf{L}_{2}}).
      \item L_predictor_list: Predictor-specific penalty matrices (\eqn{\mathbf{L}_{\mathrm{predictor\_list}}}).
      \item L_partition_list: Partition-specific penalty matrices (\eqn{\mathbf{L}_{\mathrm{partition\_list}}}).
    }
  }
  \item{knot_scale_transf}{Function for transforming predictors to standardized scale used for knot placement.}
  \item{knot_scale_inv_transf}{Function for transforming standardized predictors back to original scale.}
  \item{knots}{Matrix of knot locations on original unstandardized predictor scale for one predictor.}
  \item{partition_codes}{Vector assigning observations to partitions.}
  \item{partition_bounds}{Vector or matrix specifying the boundaries between partitions.}
  \item{knot_expand_function}{Internal function for expanding data according to partition structure.}
  \item{predict}{Function for generating predictions on new data. For multi-predictor models, \code{take_first_derivatives = TRUE, take_second_derivatives} returns derivatives as a named list of components per predictor variable, rather than a concatenated vector. When \code{new_predictors} contains columns not present in the data, extraneous columns are silently dropped before prediction.}
  \item{assign_partition}{Function for assigning new observations to partitions.}
  \item{family}{GLM family object specifying the error distribution and link function.}
  \item{estimate_dispersion}{Logical indicating whether dispersion parameter was estimated.}
  \item{unbias_dispersion}{Logical indicating whether dispersion estimates should be unbiased.}
  \item{backtransform_coefficients}{Function for converting standardized coefficients to original scale.}
  \item{forwtransform_coefficients}{Function for converting coefficients to standardized scale.}
  \item{mean_y, sd_y}{Mean and standard deviation of response if standardized.}
  \item{og_order}{Original ordering of observations before partitioning.}
  \item{order_list}{List containing observation indices for each partition.}
  \item{constraint_values, constraint_vectors}{Matrices specifying linear equality constraints if provided.}
  \item{make_partition_list}{List containing partition information for > 1-D cases.}
  \item{expansion_scales}{Vector of scaling factors used for standardizing basis expansions.}
  \item{take_derivative, take_interaction_2ndderivative}{Functions for computing derivatives of basis expansions.}
  \item{get_all_derivatives_insample}{Function for computing all derivatives on training data.}
  \item{numerics}{Indices of numeric predictors used in basis expansions.}
  \item{power1_cols, power2_cols, power3_cols, power4_cols}{Column indices for linear through quartic terms.}
  \item{quad_cols}{Column indices for all quadratic terms (including interactions).}
  \item{interaction_single_cols, interaction_quad_cols}{Column indices for linear-linear and linear-quadratic interactions.}
  \item{triplet_cols}{Column indices for three-way interactions.}
  \item{nonspline_cols}{Column indices for terms excluded from spline expansion.}
  \item{return_varcovmat}{Logical indicating whether variance-covariance matrix was computed.}
  \item{raw_expansion_names}{Names of basis expansion terms.}
  \item{std_X, unstd_X}{Functions for standardizing/unstandardizing design matrices.}
  \item{parallel_cluster_supplied}{Logical indicating whether a parallel cluster was supplied.}
  \item{weights}{Original observation weights on the data scale. When no
    weights were supplied, this is a vector of ones.}
  \item{G}{List of unscaled partition-wise information inverses
    \eqn{\mathbf{G}_{k}} if \code{return_G = TRUE}. These are the blockwise
    quantities stored on the fitting scale; correlation-aware trace,
    posterior, and variance calculations additionally use dense GLS analogues
    internally when needed.}
  \item{Ghalf}{List of \eqn{\mathbf{G}_{k}^{1/2}} matrices if
    \code{return_Ghalf = TRUE}. As with \code{G}, dense GLS square-root
    factors may also be constructed internally for correlation-aware
    post-fit calculations.}
  \item{U}{Constraint projection matrix \eqn{\mathbf{U}} if \code{return_U = TRUE}. For K=0 and no constraints, returns identity. Otherwise, returns \eqn{\mathbf{U} = \mathbf{I} - \mathbf{G}\mathbf{A}(\mathbf{A}^{\top}\mathbf{G}\mathbf{A})^{-1}\mathbf{A}^{\top}}. Used for computing the variance-covariance matrix \eqn{\sigma^{2}\mathbf{U}\mathbf{G}}.}
  \item{sigmasq_tilde}{Estimated (or fixed) dispersion parameter
    \eqn{\tilde{\sigma}^{2}}. For Gaussian identity fits without correlation,
    this is the weighted mean squared residual with optional bias correction.
    When \code{VhalfInv} is non-\code{NULL}, Gaussian-identity residuals are
    whitened before this calculation.}
  \item{trace_XUGX}{Effective degrees of freedom (\eqn{\mathrm{trace}(\mathbf{X}\mathbf{U}\mathbf{G}\mathbf{X}^{\top})}), where \eqn{\mathbf{X}\mathbf{U}\mathbf{G}\mathbf{X}^{\top}} serves as the "hat" matrix. When \code{VhalfInv} is non-\code{NULL}, computed as \eqn{\|\mathbf{V}^{-1/2}\mathbf{X}\mathbf{U}\mathbf{G}_{\mathrm{correct}}^{1/2}\|_{F}^{2}} using the full penalized GLS information.}
  \item{varcovmat}{Variance-covariance matrix of coefficient estimates if \code{return_varcovmat = TRUE}. Computed as \eqn{\sigma^{2}(\mathbf{U}\mathbf{G}^{1/2})(\mathbf{U}\mathbf{G}^{1/2})^{\top}} for numerical stability. When \code{VhalfInv} is non-\code{NULL}, uses the full \eqn{\mathbf{G}_{\mathrm{correct}}^{1/2}} in place of the block-diagonal \eqn{\mathbf{G}^{1/2}}.}
  \item{lagrange_multipliers}{Vector of Lagrangian multipliers if \code{return_lagrange_multipliers = TRUE}. For equality-only fits these correspond to the active columns of \eqn{\mathbf{A}}; when quadratic-programming constraints are active they are taken directly from \code{solve.QP} and therefore refer to the combined equality/inequality constraint system. \code{NULL} if no constraints are active (\eqn{\mathbf{A}} is \code{NULL} or \code{K == 0}).}
  \item{VhalfInv}{The \eqn{\mathbf{V}^{-1/2}} matrix used for implementing correlation structures, if specified.}
  \item{VhalfInv_fxn, Vhalf_fxn, VhalfInv_logdet, REML_grad}{Functions for generating \eqn{\mathbf{V}^{-1/2}}, \eqn{\mathbf{V}^{1/2}}, \eqn{\log|\mathbf{V}^{-1/2}|}, and gradient of REML if provided.}
  \item{VhalfInv_params_estimates}{Vector of estimated correlation parameters when using \code{VhalfInv_fxn}.}
  \item{VhalfInv_params_vcov}{Approximate variance-covariance matrix of estimated correlation parameters from BFGS optimization.}
  \item{wald_univariate}{Function for computing univariate Wald statistics and confidence intervals. Returns an S3 object of class \code{"wald_lgspline"} with dedicated \code{print}, \code{summary}, \code{plot}, \code{coef}, and \code{confint} methods. The \code{print} method uses \code{printCoefmat()} for standard R coefficient table formatting with significance stars.}
  \item{critical_value}{Critical value used for confidence interval construction.}
  \item{generate_posterior}{Function for drawing from the posterior distribution of coefficients. When \code{VhalfInv} is non-\code{NULL}, draws are from the correct joint posterior \eqn{\mathbf{U}\mathbf{G}_{\mathrm{correct}}^{1/2}\mathbf{z}} using the full penalized GLS information, reflecting cross-partition posterior covariance induced by off-diagonal blocks of \eqn{\mathbf{V}^{-1/2}}.}
  \item{find_extremum}{Function for optimizing the fitted function. Accepts both numeric column indices and character column names for \code{vars}. When \code{select_vars_fl = TRUE}, L-BFGS-B bounds are correctly subsetted to the optimized variables.}
  \item{plot}{Function for visualizing fitted curves.}
  \item{quadprog_list}{List containing quadratic programming components if applicable.}
  \item{.fit_call_args}{List containing the arguments passed to
    \code{\link{lgspline}}.}
}

The returned object has class \code{"lgspline"} and provides comprehensive tools for
model interpretation, inference, prediction, and visualization. All
coefficients and predictions can be transformed between standardized and
original scales using the provided transformation functions. The object includes
both frequentist and Bayesian inference capabilities through Wald statistics
and posterior sampling. S3 methods \code{\link{logLik.lgspline}} and
\code{\link{confint.lgspline}} are available for standard log-likelihood
extraction and confidence interval computation, respectively.
Advanced customization options are available for
analyzing arbitrarily complex study designs.
}
\description{
A comprehensive software package for fitting a variant of smoothing splines
as a constrained optimization problem, avoiding the need to algebraically
disentangle a spline basis after fitting, and allowing for interpretable
interactions and non-spline effects to be included.

\code{lgspline} fits piecewise polynomial regression splines constrained to be smooth where
they meet, penalized by the squared, integrated, second-derivative of the
estimated function with respect to predictors, using a monomial basis.

The method of Lagrangian multipliers is used to derive a polynomial regression spline
that enforces the following smoothing constraints:
\itemize{
  \item Equivalent fitted values at knots
  \item Equivalent first derivatives at knots, with respect to predictors
  \item Equivalent second derivatives at knots, with respect to predictors
}

The coefficients are penalized by a closed-form of the traditional
cubic smoothing spline penalty, as well as tunable modifications that allow
for unique penalization of multiple predictors and partitions.

This package supports model fitting for multiple spline and non-spline effects,
GLM families, Weibull accelerated failure time (AFT) models,
Cox proportional-Hazards models, negative-binomial regression,
arbitrary correlation structures, shape constraints, and extensive
customization for user-defined models and constraints.

In addition, parallel processing capabilities and comprehensive
tools for visualization, frequentist, and Bayesian inference are provided.
}
\details{
A flexible and interpretable implementation of smoothing splines including:
\itemize{
  \item Multiple predictors and interaction terms
  \item Various GLM families and link functions
  \item Correlation structures for longitudinal/clustered data
  \item Shape constraints via quadratic programming
  \item Parallel computation for large datasets
  \item Comprehensive inference tools
}
}
\section{Response and Predictor Setup}{

These arguments control the primary data inputs and the initial
standardization steps applied before knot placement and fitting.
}

\section{GLM Customization}{

These options let you override the default GLM working-weight,
dispersion, and partition-wise unconstrained fitting behavior.
}

\section{Knots and Partitioning}{

These arguments determine how the predictor space is partitioned and how
knot locations are chosen or reused.
}

\section{Penalty}{

These arguments configure the smoothing penalty itself and the optional
tuning procedure, using exact leave-one-out by default and GCV optionally.
}

\section{Basis Expansions}{

These arguments control which polynomial and interaction terms are included
in the partition-specific design matrices.
}

\section{Constraints}{

These arguments govern the smoothness equalities and any additional user
supplied linear equality constraints.
}

\section{Quadratic Programming}{

These arguments activate built-in or custom inequality constraints handled
through quadratic programming.
}

\section{Parallel Processing}{

These arguments control which computational subroutines may run in parallel
and how work is chunked across cluster workers.
}

\section{Tuning Control}{

These options control iterative updates during penalty tuning and the final
constrained fit.
}

\section{Return Control}{

These arguments determine which intermediate matrices and inferential
quantities are retained in the returned fit object.
}

\section{Correlation Structures}{

These arguments enable built-in or custom working-correlation structures for
longitudinal, clustered, or spatially indexed responses. In the notation
used throughout \code{\link{Details}}, the correlated penalized
information is written as
\eqn{\mathbf{G}^{-1} =
\mathbf{G}_{\mathrm{on}}^{-1} + \mathbf{G}_{\mathrm{off}}^{-1}}.
When the cross-partition part is low rank, the internal Woodbury helpers
factor
\eqn{\mathbf{G}_{\mathrm{off}}^{-1} =
\mathbf{E}\mathbf{J}\mathbf{E}^{\top}}
and accelerate the correlated solve without changing the final estimator.
}

\section{Grouped Argument Lists}{

For convenience, related arguments can be bundled into named lists.
When a grouped argument is non-NULL, its entries overwrite the corresponding
individual arguments. Individual arguments remain available for backward compatibility.

\describe{
  \item{\code{penalty_args}}{Groups: \code{wiggle_penalty},
    \code{flat_ridge_penalty}, \code{unique_penalty_per_partition},
    \code{unique_penalty_per_predictor}, \code{meta_penalty},
    \code{predictor_penalties}, \code{partition_penalties},
    \code{custom_penalty_mat}, \code{previously_tuned_penalties},
    \code{smoothing_spline_penalty}.}
  \item{\code{tuning_args}}{Groups: \code{opt},
    \code{use_custom_bfgs}, \code{delta}, \code{tol},
    \code{tuning_criterion},
    \code{gcv_gamma},
    \code{initial_wiggle}, \code{initial_flat},
    \code{iterate_tune}, \code{iterate_final_fit}.}
  \item{\code{expansion_args}}{Groups: \code{include_quadratic_terms},
    \code{include_cubic_terms}, \code{include_quartic_terms},
    \code{include_2way_interactions}, \code{include_3way_interactions},
    \code{include_quadratic_interactions},
    \code{just_linear_with_interactions},
    \code{just_linear_without_interactions},
    \code{exclude_interactions_for}, \code{exclude_these_expansions},
    \code{custom_basis_fxn}, \code{offset}.}
  \item{\code{constraint_args}}{Groups:
    \code{include_constrain_fitted},
    \code{include_constrain_first_deriv},
    \code{include_constrain_second_deriv},
    \code{include_constrain_interactions},
    \code{constraint_values}, \code{constraint_vectors},
    \code{no_intercept}.}
  \item{\code{qp_args}}{Groups all \code{qp_*} arguments.}
  \item{\code{parallel_args}}{Groups: \code{cl},
    \code{chunk_size}, and all \code{parallel_*} flags.}
  \item{\code{covariance_args}}{Groups: \code{correlation_id},
    \code{spacetime}, \code{correlation_structure},
    \code{VhalfInv}, \code{Vhalf}, \code{VhalfInv_fxn},
    \code{Vhalf_fxn}, \code{VhalfInv_par_init},
    \code{REML_grad}, \code{custom_VhalfInv_loss},
    \code{VhalfInv_logdet}.}
  \item{\code{return_args}}{Groups: \code{return_G},
    \code{return_Ghalf}, \code{return_U},
    \code{estimate_dispersion}, \code{unbias_dispersion},
    \code{return_varcovmat}, \code{exact_varcovmat},
    \code{return_lagrange_multipliers}.}
  \item{\code{glm_args}}{Groups: \code{glm_weight_function},
    \code{schur_correction_function},
    \code{need_dispersion_for_estimation},
    \code{dispersion_function}, \code{unconstrained_fit_fxn},
    \code{keep_weighted_Lambda}.}
}
}

\section{Miscellaneous}{

These remaining arguments affect inference defaults, numerical safeguards,
verbosity, and developer-oriented diagnostics.
}

\examples{

## ## ## ## Simple Examples ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
## Simulate some data, fit using default settings without tuning, and plot
set.seed(1234)
t <- runif(2500, -10, 10)
y <- 2*sin(t) + -0.06*t^2 + rnorm(length(t))
model_fit <- lgspline(t, y, opt = FALSE)
plot(t, y, main = 'Observed Data vs. Fitted Function, Colored by Partition',
     ylim = c(-10, 10))
plot(model_fit, add = TRUE)

\donttest{
## Repeat using logistic regression, with univariate inference shown
# and alternative function call
y <- rbinom(length(y), 1, 1/(1+exp(-std(y))))
df <- data.frame(t = t, y = y)
model_fit <- lgspline(y ~ spl(t),
                      df,
                      family = binomial())
plot(t, y, main = 'Observed Data vs Fitted Function with Formulas and Derivatives',
  ylim = c(-0.5, 1.05), cex.main = 0.8)
plot(model_fit,
     show_formulas = TRUE,
     text_size_formula = 0.65,
     legend_pos = 'bottomleft',
     legend_args = list(y.intersp = 1.1),
     add = TRUE)
## Notice how the coefficients match the formula, and expansions are
# homogenous across partitions without reparameterization
print(summary(model_fit))

## Overlay first and second derivatives of fitted function respectively
derivs <- predict(model_fit,
                  new_predictors = sort(t),
                  take_first_derivatives = TRUE,
                  take_second_derivatives = TRUE)
points(sort(t), derivs$first_deriv, col = 'gold', type = 'l')
points(sort(t), derivs$second_deriv, col = 'goldenrod', type = 'l')
legend('bottomright',
       col = c('gold','goldenrod'),
       lty = 1,
       legend = c('First Derivative', 'Second Derivative'))

## Simple 2D example - including a non-spline effect
z <- seq(-2, 2, length.out = length(y))
df <- data.frame(Predictor1 = t,
                 Predictor2 = z,
                 Response = sin(y)+0.1*z)
model_fit <- lgspline(Response ~ spl(Predictor1) + Predictor1*Predictor2,
                      df)

## Notice, while spline effects change over partitions,
# interactions and non-spline effects are constrained to remain the same
coefficients <- Reduce('cbind', coef(model_fit))
colnames(coefficients) <- paste0('Partition ', 1:(model_fit$K+1))
print(coefficients)

## One or two variables can be selected for plotting at a time
# even when >= 3 predictors are present
plot(model_fit,
      custom_title = 'Marginal Relationship of Predictor 1 and Response',
      vars = 'Predictor1',
      custom_response_lab = 'Response',
      show_formulas = TRUE,
      legend_pos = 'bottomright',
      digits = 4,
      text_size_formula = 0.5)

## 3D plots are implemented as well, retaining closed-formulas
my_plot <- plot(model_fit,
                show_formulas = TRUE,
                custom_response_lab = 'Response')
my_plot


## ## ## ## More Detailed 1D Example ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
## 1D data generating functions
t <- seq(-9, 9, length.out = 1000)
slinky <- function(x) {
  (50 * cos(x * 2) -2 * x^2 + (0.25 * x)^4 + 80)
}
coil <- function(x) {
  (100 * cos(x * 2) +-1.5 * x^2 + (0.1 * x)^4 +
  (0.05 * x^3) + (-0.01 * x^5) +
     (0.00002 * x^6) -(0.000001 * x^7) + 100)
}
exponential_log <- function(x) {
  unlist(c(sapply(x, function(xx) {
    if (xx <= 1) {
      100 * (exp(xx) - exp(1))
    } else {
      100 * (log(xx))
    }
  })))
}
scaled_abs_gamma <- function(x) {
  2*sqrt(gamma(abs(x)))
}

## Composite function
fxn <- function(x)(slinky(t) +
                   coil(t) +
                   exponential_log(t) +
                   scaled_abs_gamma(t))

## Bind together with random noise
dat <- cbind(t, fxn(t) + rnorm(length(t), 0, 50))
colnames(dat) <- c('t', 'y')
x <- dat[,'t']
y <- dat[,'y']

## Fit Model, 4 equivalent ways are shown below
model_fit <- lgspline(t, y, opt = FALSE)
model_fit <- lgspline(y ~ spl(t), as.data.frame(dat), opt = FALSE)
model_fit <- lgspline(response = y, predictors = t, opt = FALSE)
model_fit <- lgspline(data = as.data.frame(dat), formula = y ~ ., opt = FALSE)

# This is not valid: lgspline(y ~ ., t)
# This is not valid: lgspline(y, data = as.data.frame(dat))
# Do not put operations in formulas, not valid: lgspline(y ~ log(t) + spl(t))

## Basic Functionality
predict(model_fit, new_predictors = rnorm(1)) # make prediction on new data
loo_vals <- suppressWarnings(head(leave_one_out(model_fit)))
loo_vals # may contain NA when leverage is too high
coef(model_fit) # extract coefficients
summary(model_fit) # model information and Wald inference
generate_posterior(model_fit) # generate draws of parameters from posterior distribution
find_extremum(model_fit, minimize = TRUE) # find the minimum of the fitted function

## Incorporate range constraints, custom knots, keep penalization identical
# across partitions
model_fit <- lgspline(y ~ spl(t),
                      unique_penalty_per_partition = FALSE,
                      custom_knots = cbind(c(-2, -1, 0, 1, 2)),
                      data = data.frame(t = t, y = y),
                      qp_range_lower = -150,
                      qp_range_upper = 150)

## Plotting the constraints and knots
plot(model_fit,
     custom_title = 'Fitted Function Constrained to Lie Between (-150, 150)',
     cex.main = 0.75)
# knot locations
abline(v = model_fit$knots)
# lower bound from quadratic program
abline(h = -150, lty = 2)
# upper bound from quadratic program
abline(h = 150, lty = 2)
# observed data
points(t, y, cex = 0.24)

## Enforce monotonic increasing constraints on fitted values
# K = 4 => 5 partitions
t <- seq(-10, 10, length.out = 100)
y <- 5*sin(t) + t + 2*rnorm(length(t))
model_fit <- lgspline(t,
                      y,
                      K = 4,
                      qp_monotonic_increase = TRUE)
plot(t, y, main = 'Monotonic Increasing Function with Respect to Fitted Values')
plot(model_fit,
     add = TRUE,
     show_formulas = TRUE,
     legend_pos = 'bottomright',
     custom_predictor_lab = 't',
     custom_response_lab = 'y')

## Posterior draws under constraint
draw <- generate_posterior(model_fit, enforce_qp_constraints = TRUE)
pr <- predict(model_fit, B_predict = draw$post_draw_coefficients)
points(t, pr, col = 'grey')

## ## ## ## 2D Example using Volcano Dataset ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
## Prep
data('volcano')
volcano_long <-
  Reduce('rbind', lapply(1:nrow(volcano), function(i){
    t(sapply(1:ncol(volcano), function(j){
      c(i, j, volcano[i,j])
    }))
  }))
colnames(volcano_long) <- c('Length', 'Width', 'Height')

## Fit, with 50 partitions
# When fitting with > 1 predictor and large K, including quartic terms
# is highly recommended, and/or dropping the second-derivative constraint.
# Otherwise, the constraints can impose all partitions to be equal, with one
# cubic function fit for all (there is not enough degrees of freedom to fit
# unique cubic functions due to the massive amount of constraints).
# Below, quartic terms are included and the constraint of second-derivative
# smoothness at knots is ignored.
model_fit <- lgspline(volcano_long[,c(1, 2)],
                      volcano_long[,3],
                      include_quadratic_interactions = TRUE,
                      K = 49,
                      opt = FALSE,
                      return_U = FALSE,
                      return_varcovmat = FALSE,
                      estimate_dispersion = TRUE,
                      return_Ghalf = FALSE,
                      return_G = FALSE,
                      include_constrain_second_deriv = FALSE,
                      unique_penalty_per_predictor = FALSE,
                      unique_penalty_per_partition = FALSE,
                      wiggle_penalty = 1e-10, # the fixed wiggle penalty
                      flat_ridge_penalty = 1e-2) # the ridge penalty / wiggle penalty

## Plotting on new data with interactive visual + formulas
new_input <- expand.grid(seq(min(volcano_long[,1]),
                             max(volcano_long[,1]),
                             length.out = 250),
                         seq(min(volcano_long[,2]),
                             max(volcano_long[,2]),
                             length.out = 250))
plot(model_fit,
     new_predictors = new_input,
     show_formulas = TRUE,
     custom_response_lab = "Height",
     custom_title = 'Volcano 3-D Map',
     digits = 2)

## Get AUC
area_under_volcano <- integrate(model_fit,
                               lower = apply(volcano_long, 2, min)[1:2],
                               upper = apply(volcano_long, 2, max)[1:2])

## ## ## ## Advanced Techniques using Trees Dataset ## ## ## ## ## ## ## ## ## ## ## ## ##
## Goal here is to introduce how lgspline works with non-canonical GLMs and
# demonstrate some custom features
data('trees')

## L1-regularization constraint function on standardized coefficients
# Bound all coefficients to be less than a certain value (l1_bound) in absolute
# magnitude such that | B^{(j)}_k | < lambda for all j = 1....p coefficients,
# and k = 1...K+1 partitions.
l1_constraint_matrix <- function(p, K) {
  ## Total number of coefficients
  P <- p * (K + 1)

  ## Create diagonal matrices for L1 constraint
  # First matrix: lamdba > -bound
  # Second matrix: -lambda > -bound
  first_diag <- diag(P)
  second_diag <- -diag(P)

  ## Combine matrices
  l1_Amat <- cbind(first_diag, second_diag)

  return(l1_Amat)
}

## Bounds absolute value of coefficients to be < l1_bound
l1_bound_vector <- function(qp_Amat,
                            scales,
                            l1_bound) {

  ## Combine matrices
  l1_bvec <- rep(-l1_bound, ncol(qp_Amat)) * c(1, scales)

  return(l1_bvec)
}

## Fit model, using predictor-response formulation, assuming
# Gamma-distributed response, and custom quadratic-programming constraints,
# with qp_score_function/glm_weight_function updated for non-canonical GLMs
# as well as quartic terms, keeping the effect of height constant across
# partitions, and 3 partitions in total. Hence, this is an advanced-usage
# case.
# You can modify this code for performing l1-regularization in general.
# For canonical GLMs, the default qp_score_function/glm_weight_function are
# correct and do not need to be changed
# (custom functionality is not needed for canonical GLMs).
model_fit <- lgspline(
  Volume ~ spl(Girth) + Height*Girth,
  data = with(trees, cbind(Girth, Height, Volume)),
  family = Gamma(link = 'log'),
  keep_weighted_Lambda = TRUE,
  glm_weight_function = function(
    mu,
    y,
    order_indices,
    family,
    dispersion,
    observation_weights,
   ...){
     rep(1/dispersion, length(y))
   },
   dispersion_function = function(
     mu,
     y,
     order_indices,
     family,
     observation_weights,
     VhalfInv,
     ...){
    mean(
      mu^2/((y-mu)^2)
    )
  }, # = biased estimate of 1/shape parameter
  need_dispersion_for_estimation = TRUE,
  unbias_dispersion = TRUE, # multiply dispersion by N/(N-trace(XUGX^{T}))
  K = 2, # 3 partitions
  opt = FALSE, # keep penalties fixed
  unique_penalty_per_partition = FALSE,
  unique_penalty_per_predictor = FALSE,
  flat_ridge_penalty = 1e-64,
  wiggle_penalty = 1e-64,
  qp_score_function = function(X, y, mu, order_list, dispersion, VhalfInv,
    observation_weights, ...){
   t(X) \%*\% diag(c(1/mu * 1/dispersion)) \%*\% cbind(y - mu)
  }, # updated score for gamma regression with log link
  qp_Amat_fxn = function(N, p, K, X, colnm, scales, deriv_fxn, ...) {
    l1_constraint_matrix(p, K)
  },
  qp_bvec_fxn = function(qp_Amat, N, p, K, X, colnm, scales, deriv_fxn, ...) {
    l1_bound_vector(qp_Amat, scales, 25)
  },
  qp_meq_fxn = function(qp_Amat, N, p, K, X, colnm, scales, deriv_fxn, ...) 0
)

## Notice, interaction effect is constant across partitions as is the effect
# of Height alone
Reduce('cbind', coef(model_fit))

## Many constraints, many coefficients, and small sample size makes inference
#  using asymptotic variance-covariance matrix untrustworthy.
#  Confidence intervals are often too wide or narrow, even for "good" fit.
#  Consider bootstrapping or alternative.
print(summary(model_fit))

## Plot results
plot(model_fit, custom_predictor_lab1 = 'Girth',
     custom_predictor_lab2 = 'Height',
     custom_response_lab = 'Volume',
     custom_title = 'Girth and Height Predicting Volume of Trees',
     show_formulas = TRUE)

## Verify magnitude of unstandardized coefficients does not exceed bound (25)
print(max(abs(unlist(model_fit$B))))

## Find height and girth where tree volume is closest to 42
# Uses custom objective that minimizes MSE discrepancy between predicted
# value and 42.
# The vanilla find_extremum function can be thought of as
# using "function(mu)mu" aka the identity function as the
# objective, where mu = "f(t)", our estimated function. The derivative is then
# d_mu = "df/dt" with respect to predictors t.
# But with more creative objectives, and since we have machinery for
# df/dt already available, we can compute gradients for (and optimize)
# arbitrary differentiable functions of our predictors too.
# For any objective, differentiate w.r.t. to mu, then multiply by d_mu to
# satisfy chain rule.
# Here, we have objective function: 0.5*(mu-42)^2
# and gradient                    : (mu-42)*d_mu
# and L-BFGS-B will be used to find the height and girth that most closely
# yields a prediction of 42 within the bounds of the observed data.
# The d_mu also takes into account link function transforms automatically
# for most common link functions, and will return warning + instructions
# on how to program the link-function derivatives otherwise.

## Custom acquisition functions for Bayesian optimization could be coded here.
find_extremum(
  model_fit,
  minimize = TRUE,
  custom_objective_function = function(mu, sigma, ybest, ...){
    0.5*(mu - 42)^2
  },
  custom_objective_derivative = function(mu, sigma, ybest, d_mu, ...){
    (mu - 42) * d_mu
  }
)

## ## ## ## How to Use Formulas in lgspline ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
## Demonstrates splines with multiple mixed predictors and interactions

## Generate data
n <- 2500
x <- rnorm(n)
y <- rnorm(n)
z <- sin(x)*mean(abs(y))/2

## Categorical predictors
cat1 <- rbinom(n, 1, 0.5)
cat2 <- rbinom(n, 1, 0.5)
cat3 <- rbinom(n, 1, 0.5)

## Response with mix of effects
response <- y + z + 0.1*(2*cat1 - 1)

## Continuous predictors re-named
continuous1 <- x
continuous2 <- z

## Combine data
dat <- data.frame(
  response = response,
  continuous1 = continuous1,
  continuous2 = continuous2,
  cat1 = cat1,
  cat2 = cat2,
  cat3 = cat3
)

## Example 1: Basic Model with Default Terms, No Intercept
# standardize_response = FALSE often needed when constraining intercepts to 0
fit1 <- lgspline(
  formula = response ~ 0 + spl(continuous1, continuous2) +
    cat1*cat2*continuous1 + cat3,
  K = 2,
  standardize_response = FALSE,
  data = dat
)
## Examine coefficients included
rownames(fit1$B$partition1)
## Verify intercept term is near 0 up to some numeric tolerance
abs(fit1$B[[1]][1]) < 1e-8

## Example 2: Similar Model with Intercept, Other Terms Excluded
fit2 <- lgspline(
  formula = response ~ spl(continuous1, continuous2) +
    cat1*cat2*continuous1 + cat3,
  K = 1,
  standardize_response = FALSE,
  include_cubic_terms = FALSE,
  exclude_these_expansions = c( # Not all need to actually be present
    '_batman_x_robin_',
    '_3_x_4_', # no cat1 x cat2 interaction, coded using column indices
    'continuous1xcontinuous2', # no continuous1 x continuous2 interaction
    'thejoker'
  ),
  data = dat
)
## Examine coefficients included
rownames(Reduce('cbind',coef(fit2)))
# Intercept will probably be present and non-0 now
abs(fit2$B[[1]][1]) < 1e-8

## ## ## ## Compare Inference to survreg for Weibull AFT Model Validation ##
# Only linear predictors, no knots, no penalties, using Weibull AFT Model
# The goal here is to ensure that for the special case of no spline effects
# and no knots, this implementation will be consistent with other model
# implementations.
# Also note, that when using models (like Weibull AFT) where dispersion is
# being estimated and is required for estimating beta coefficients,
# we use a schur complement correction function to adjust (or "correct") our
# variance-covariance matrix for both estimation and inference to account for
# uncertainty in estimating the dispersion.
# Typically the schur_correction_function would return a negative-definite
# matrix, as its output is elementwise added to the information matrix prior
# to inversion.
if (requireNamespace("survival", quietly = TRUE)) {
  data("pbc", package = "survival")
  df <- data.frame(na.omit(
    pbc[, c("time", "trt", "stage", "hepato", "bili", "age", "status")]
  ))

  ## Weibull AFT using lgspline, showing how some custom options can be used to
  # fit more complicated models
  model_fit <- lgspline(time ~ trt + stage + hepato + bili + age,
                        df,
                        family = weibull_family(),
                        need_dispersion_for_estimation = TRUE,
                        dispersion_function = weibull_dispersion_function,
                        glm_weight_function = weibull_glm_weight_function,
                        schur_correction_function = weibull_schur_correction,
                        unconstrained_fit_fxn = unconstrained_fit_weibull,
                        opt = FALSE,
                        wiggle_penalty = 0,
                        flat_ridge_penalty = 0,
                        K = 0,
                        status = df$status != 0)
  print(summary(model_fit))

  ## Survreg results match closely on estimates and inference for coefficients
  survreg_fit <- survival::survreg(
    survival::Surv(time, status != 0) ~ trt + stage + hepato + bili + age,
    df
  )
  print(summary(survreg_fit))

  ## sigmasq_tilde = scale^2 of survreg
  print(c(sqrt(model_fit$sigmasq_tilde), survreg_fit$scale))
}

## ## ## ## Modelling Correlation Structures ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
## Setup
n_blocks <- 200 # Number of correlation_ids (subjects)
block_size <- 5 # Size of each correlation_ids (number of repeated measures per subj.)
N <- n_blocks * block_size # total sample size (balanced here)
rho_true <- 0.25  # True correlation

## Generate predictors and mean structure
t <- seq(-9, 9, length.out = N)
true_mean <- sin(t)

## Create block compound symmetric errors = I(1-p) + Jp
errors <- Reduce('rbind',
                 lapply(1:n_blocks,
                        function(i){
                          sigma <- diag(block_size) + rho_true *
                            (matrix(1, block_size, block_size) -
                               diag(block_size))
                          matsqrt(sigma) \%*\% rnorm(block_size)
                        }))

## Generate response with correlated errors
y <- true_mean + errors * 0.5

## Fit model with correlation structure
# include_warnings = FALSE is a good idea here, since many proposed
# correlations will not work
model_fit <- lgspline(t,
                      y,
                      K = 4,
                      correlation_id = rep(1:n_blocks, each = block_size),
                      correlation_structure = 'exchangeable',
                      include_warnings = FALSE
)

## Assess overall fit
plot(t, y, main = 'Sinosudial Fit Under Correlation Structure')
plot(model_fit, add = TRUE, show_formulas = TRUE, custom_predictor_lab = 't')

## Compare estimated vs true correlation
# Built-in exchangeable uses rho = exp(-exp(par)), so par in (-Inf, Inf)
# maps to rho in (0, 1). Only positive correlation is supported.
rho_est <- exp(-exp(model_fit$VhalfInv_params_estimates))
print(c("True correlation:" = rho_true,
        "Estimated correlation:" = rho_est))

## Quantify uncertainty in correlation estimate with 95\% confidence interval
#  CI is constructed on the working scale and back-transformed
ci_transformed <- confint(model_fit)['Correlation parameter 1',]
ci_natural <- sort(exp(-exp(ci_transformed)))
print("95\% CI for correlation:")
print(ci_natural)

## Also check SD (should be close to 0.5)
print(sqrt(model_fit$sigmasq_tilde))

## Toeplitz Simulation Setup, with demonstration of custom functions
# and boilerplate. Toep is not implemented by default, because it makes
# strong assumptions on the study design and missingness that are rarely met,
# with non-obvious workarounds.
# If a GLM was to-be-fit, you would also submit a function "Vhalf_fxn" analogous
# to VhalfInv_fxn with same argument (par) and an output of an N x N matrix
# that yields the inverse of VhalfInv_fxn output.
n_blocks <- 250   # Number of correlation_ids
block_size <- 8   # Observations per correlation_id
N <- n_blocks * block_size # total sample size
sigma_true <- 2   # Marginal standard deviation

## True Toeplitz components
# This example uses a convex combination of two geometric lag kernels:
# corr(h) = mix * rho_fast^h + (1 - mix) * rho_slow^h
# which is Toeplitz and positive definite for mix in (0, 1) and
# rho_fast, rho_slow in (0, 1).
rho_fast_true <- 0.25
rho_slow_true <- 0.75
mix_true <- 0.40

## Create time and correlation_id variables
time_var <- rep(1:block_size, n_blocks)
correlation_id_var <- rep(1:n_blocks, each = block_size)

## Create nonlinear predictor-response relationship
# Not sinusoidal and not polynomial.
t_base <- seq(-2, 2, length.out = block_size)
t <- rep(t_base, n_blocks) + rnorm(N, sd = 0.10)
f_true <- function(t) {
  1.4 + 0.9 * atan(1.8 * t) + 0.8 * exp(-1.2 * (t - 0.4)^2)
}

## Generate mean structure
mu_true <- f_true(t)

## Toeplitz correlation helper
corr_from_components <- function(rho_fast, rho_slow, mix) {
  corr <- matrix(0, block_size, block_size)
  for(i in 1:block_size) {
    for(j in 1:block_size) {
      lag <- abs(i - j)
      if(lag == 0) {
        corr[i, j] <- 1
      } else {
        corr[i, j] <- mix * rho_fast^lag + (1 - mix) * rho_slow^lag
      }
    }
  }
  corr
}

## Toeplitz correlation function
# Custom functions can use any parameterization. Here we map:
#   par[1] -> rho_fast = exp(-exp(par[1]))
#   par[2] -> rho_slow = exp(-exp(par[2]))
#   par[3] -> mix      = plogis(par[3])
# so the parameter space is unconstrained, while the resulting Toeplitz
# correlation matrix remains valid.
corr_from_par <- function(par) {
  rho_fast <- exp(-exp(par[1]))
  rho_slow <- exp(-exp(par[2]))
  mix <- plogis(par[3])
  corr_from_components(rho_fast, rho_slow, mix)
}

## Create block Toeplitz errors from the same family we will fit
corr_true <- corr_from_components(rho_fast_true, rho_slow_true, mix_true)
errors <- Reduce('c',
                 lapply(1:n_blocks, function(i) {
                   c(matsqrt(corr_true) \%*\% rnorm(block_size))
                 }))

## Generate response with correlated errors and nonlinear covariate effect
y <- mu_true + sigma_true * errors

VhalfInv_fxn <- function(par) {
  corr <- corr_from_par(par)
  kronecker(diag(n_blocks), matinvsqrt(corr))
}

Vhalf_fxn <- function(par) {
  corr <- corr_from_par(par)
  kronecker(diag(n_blocks), matsqrt(corr))
}

## Determinant function (for efficiency)
# This avoids taking determinant of N by N matrix
VhalfInv_logdet <- function(par) {
  corr <- corr_from_par(par)
  log_det_invsqrt_corr <- -0.5 * determinant(corr, logarithm = TRUE)$modulus[1]
  n_blocks * log_det_invsqrt_corr
}

## GLM weights for REML gradient helper
# For Gaussian identity, these are all 1.
glm_weight_function <- function(mu, y, order_indices, family,
                                dispersion, observation_weights, ...) {
  rep(1, length(mu))
}

## REML gradient function
# The helper reml_grad_from_dV computes the three REML terms once dV / dpar
# is supplied. For this parameterization, dV / dpar has closed form.
REML_grad <- function(par, model_fit, ...) {
  rho_fast <- exp(-exp(par[1]))
  rho_slow <- exp(-exp(par[2]))
  mix <- plogis(par[3])

  dV1_block <- matrix(0, block_size, block_size)
  dV2_block <- matrix(0, block_size, block_size)
  dV3_block <- matrix(0, block_size, block_size)

  for(i in 1:block_size) {
    for(j in 1:block_size) {
      lag <- abs(i - j)
      if(lag > 0) {
        ## d/dpar[1] through rho_fast = exp(-exp(par[1]))
        dV1_block[i, j] <- -mix * lag * exp(par[1]) * rho_fast^lag
        ## d/dpar[2] through rho_slow = exp(-exp(par[2]))
        dV2_block[i, j] <- -(1 - mix) * lag * exp(par[2]) * rho_slow^lag
        ## d/dpar[3] through mix = plogis(par[3])
        dV3_block[i, j] <- mix * (1 - mix) * (rho_fast^lag - rho_slow^lag)
      }
    }
  }

  dV1 <- kronecker(diag(n_blocks), dV1_block)
  dV2 <- kronecker(diag(n_blocks), dV2_block)
  dV3 <- kronecker(diag(n_blocks), dV3_block)

  gradient <- numeric(3)
  gradient[1] <- lgspline::reml_grad_from_dV(dV1, model_fit,
                                   glm_weight_function, ...)
  gradient[2] <- reml_grad_from_dV(dV2, model_fit,
                                   glm_weight_function, ...)
  gradient[3] <- reml_grad_from_dV(dV3, model_fit,
                                   glm_weight_function, ...)
  gradient
}

## Visualization
plot(t, y, col = correlation_id_var,
     main = 'Simulated Data with Toeplitz Correlation')

## Fit model with custom Toeplitz
model_fit <- lgspline(
  response = y,
  predictors = t,
  K = 4,
  standardize_response = FALSE,
  VhalfInv_fxn = VhalfInv_fxn,
  Vhalf_fxn = Vhalf_fxn,
  VhalfInv_logdet = VhalfInv_logdet,
  REML_grad = REML_grad,
  VhalfInv_par_init = c(0, -1, 0),
  include_warnings = FALSE
)

## Print comparison of true and estimated correlations
lag_values <- 1:(block_size - 1)
corr_true_by_lag <- sapply(lag_values, function(h) {
  mix_true * rho_fast_true^h + (1 - mix_true) * rho_slow_true^h
})
rho_fast_est <- exp(-exp(model_fit$VhalfInv_params_estimates[1]))
rho_slow_est <- exp(-exp(model_fit$VhalfInv_params_estimates[2]))
mix_est <- plogis(model_fit$VhalfInv_params_estimates[3])
corr_est_by_lag <- sapply(lag_values, function(h) {
  mix_est * rho_fast_est^h + (1 - mix_est) * rho_slow_est^h
})
cat('Toeplitz Correlation Estimates by Lag:\n')
print(data.frame(
  Lag = lag_values,
  True.Correlation = round(corr_true_by_lag, 4),
  Estimated.Correlation = round(corr_est_by_lag, 4)
))

## Should be ~ 2
print(sqrt(model_fit$sigmasq_tilde))

## ## ## ## Parallelism ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
if (requireNamespace("parallel", quietly = TRUE)) {
  ## Data generating function
  a <- runif(500000, -9, 9)
  b <- runif(500000, -9, 9)
  c <- rnorm(500000)
  d <- rpois(500000, 1)
  y <- sin(a) + cos(b) - 0.2*sqrt(a^2 + b^2) +
    abs(a) + b +
    0.5*(a^2 + b^2) +
    (1/6)*(a^3 + b^3) +
    a*b*c -
    c +
    d +
    rnorm(500000, 0, 5)

  ## Set up cores
  cl <- parallel::makeCluster(1)
  on.exit(try(parallel::stopCluster(cl), silent = TRUE), add = TRUE)

  ## This example shows some options for what operations can be parallelized
  # By default, only parallel_eigen and parallel_unconstrained are TRUE
  # parallel_unconstrained is only for GLMs, for identity link Gaussian
  # response, use parallel_matmult=TRUE to ensure parallel fitting across
  # partitions.
  # G, G^{-1/2}, and G^{1/2} are computed in parallel across each of the
  # K+1 partitions.
  # However, parallel_unconstrained only affects GLMs without corr. components
  # - it does not affect fitting here
  system.time({
    parfit <- lgspline(y ~ spl(a, b) + a*b*c + d,
                       data = data.frame(y = y,
                                         a = a,
                                         b = b,
                                         c = c,
                                         d = d),
                       cl = cl,
                       K = 1,
                       parallel_eigen = TRUE,
                       parallel_unconstrained = TRUE,
                       parallel_aga = FALSE,
                       parallel_find_neighbors = FALSE,
                       parallel_trace = FALSE,
                       parallel_matmult = TRUE,
                       parallel_make_constraint = FALSE,
                       parallel_penalty = FALSE)
  })
  print(summary(parfit))
}
}


}
\seealso{
\itemize{
  \item \code{\link{lgspline.fit}} for the low-level fitting interface
  \item \code{\link{logLik.lgspline}} for log-likelihood extraction
  \item \code{\link{confint.lgspline}} for confidence interval extraction
  \item \code{\link{leave_one_out}} for leave-one-out cross-validated predictions
  \item \code{\link{blockfit_solve}} for the standalone backfitting solver
  \item \code{\link[quadprog]{solve.QP}} for quadratic programming optimization
  \item \code{\link[plotly]{plot_ly}} for interactive plotting
  \item \code{\link[stats]{kmeans}} for k-means clustering
  \item \code{\link[stats]{optim}} for general purpose optimization routines
}
}
